(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 11.2' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[   8203308,     230506]
NotebookOptionsPosition[   7654365,     219918]
NotebookOutlinePosition[   7673339,     220310]
CellTagsIndexPosition[   7671505,     220268]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["xAct`xTensor`", "Section",
 CellChangeTimes->{{3.72897608609194*^9, 3.728976087819756*^9}},
 FontSize->24,
 CellTags->{"S5.18.1", "1.1"},
 CellID->733679349,ExpressionUUID->"8f751253-cb93-40fd-ba41-fbaa77bd1a49"],

Cell[CellGroupData[{

Cell["Author", "Subsection",ExpressionUUID->"9bdcbc5a-2e86-4a77-9fb5-564dc8632094"],

Cell[TextData[{
 "\[Copyright] 2002-2018, under the GNU General Public License (GPL)\n     ",
 StyleBox["Jos\[EAcute] M. Mart\[IAcute]n-Garc\[IAcute]a",
  FontWeight->"Bold"],
 "\n\t",
 StyleBox["jose@xact.es",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox[ButtonBox["http://www.xAct.es",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://www.xact.es/"], None},
  ButtonNote->"http://www.xact.es/"],
  FontFamily->"Courier"],
 StyleBox["\n\t",
  FontFamily->"Courier"],
 "\t",
 StyleBox[ButtonBox["http://groups.google.com/group/xAct/",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://groups.google.com/group/xAct/"], None},
  ButtonNote->"http://groups.google.com/group/xAct/"],
  FontFamily->"Courier"]
}], "Text",
 FontFamily->"Arial",ExpressionUUID->"e5ed7587-e24d-4f73-b8f3-63f0c64152fc"],

Cell[TextData[{
 "Helped by:\n\t",
 StyleBox["Alfonso Garc\[IAcute]a-Parrado",
  FontWeight->"Bold"],
 " : design of vector bundles, complex structure, products and mappings.\n\t\
",
 StyleBox["Guillaume Faye",
  FontWeight->"Bold"],
 " : design of n+m decompositions."
}], "Text",ExpressionUUID->"55526d07-7028-4064-b2ac-5696ca58a58a"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Intro", "Subsection",ExpressionUUID->"58295eb0-9c58-4bc2-922e-93dc4f966c53"],

Cell[TextData[{
 StyleBox["xTensor`", "MR"],
 " extends ",
 StyleBox["Mathematica", "TI"],
 " capabilities in abstract tensor calculus, focusing on General Relativity. \
It works with tensor fields with arbitrary symmetries under permutations of \
indices, defined on one or more vector bundles and manifolds. It computes \
covariant derivatives with respect to multiple connections, including gauge \
connections in complex vector bundles. ",
 StyleBox["xTensor`", "MR"],
 " can work with Lie derivatives, parametric derivatives and variational \
derivatives. It allows the presence of a metric in each manifold and defines \
all the associated torsion and curvature tensors (Riemann, Ricci, Einstein, \
Weyl, etc.), handling automatically their definitions and relations in terms \
of Christoffel objects. It can work with mappings between any two manifolds \
and push/pull tensor fields along those maps.\n",
 StyleBox["xTensor`", "MR"],
 " does not perform component calculations. See the twin package ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 ".\n",
 StyleBox["xTensor`", "MR"],
 " needs the twin package ",
 StyleBox["xPerm`", "MR"],
 " in order to compute the canonical form of a list of indices under certain \
symmetry group.\n",
 StyleBox["xTensor`", "MR"],
 " has been designed with the following priorities in mind:\n\t1.- \
Mathematical structure\n\t2.- Efficiency\n\t3.- Compliance with ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " style\n\t4.- Simplicity of input/output\nIn particular, concerning the \
mathematical structure, ",
 StyleBox["xTensor`", "MR"],
 " imitates the usual cycle in mathematics ``let ",
 StyleBox["X", "Input"],
 " be a (",
 StyleBox["Type", "Input"],
 ") with properties ",
 StyleBox["props", "Input"],
 "; then use it to do some ",
 StyleBox["computations", "Input"],
 "\.b4\.b4 with the general form of declaration\n\t",
 StyleBox["DefType[X, props]", "Input"],
 "\n\t",
 StyleBox["computations", "Input"]
}], "Text",ExpressionUUID->"dacae0c5-11a2-4fcd-a41c-5ae57fcad21a"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Load the package", "Subsection",ExpressionUUID->"8914d7d8-31e0-4a20-a60f-843051cb7ed6"],

Cell[CellGroupData[{

Cell["\<\
This loads the package from the default directory, for example \
$HOME/.Mathematica/Applications/xAct/ for a single-user installation under \
Linux (see the installation notes for other possibilities),\
\>", "MathCaption",
 CellChangeTimes->{{3.728977519955257*^9, 
  3.728977521376769*^9}},ExpressionUUID->"63a736e0-39af-4e8d-9b57-\
37c116af6eb8"],

Cell[CellGroupData[{

Cell["<<xAct`xTensor`", "Input",
 CellLabel->"In[1]:=",ExpressionUUID->"1fb58714-653b-4365-aa43-7b82564c3a6d"],

Cell[CellGroupData[{

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.7928040618302517`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->551875384,ExpressionUUID->"37e67236-d784-488e-b78b-539616ac34ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xPerm`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.2.3\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2015", ",", "8", ",", "23"}], "}"}]}],
  SequenceForm["Package xAct`xPerm`  version ", "1.2.3", ", ", {2015, 8, 23}],
  
  Editable->False]], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.792804061838923*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->387707495,ExpressionUUID->"96aae411-9519-4432-ad32-11c55c5de4dc"],

Cell[BoxData["\<\"CopyRight (C) 2003-2020, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.792804061847281*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->734483467,ExpressionUUID->"a7e6bf32-dba2-4298-bbcd-4fd1c90f3784"],

Cell[BoxData["\<\"Connecting to external mac executable...\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.7928040618534327`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->727626913,ExpressionUUID->"b5e01409-19db-4b30-985d-669d58f2ae42"],

Cell[BoxData["\<\"Connection established.\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.792804061903558*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1453640773,ExpressionUUID->"2cd76ea9-cff8-44c3-87a5-3eb8839e6499"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.7928040619097023`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1566503539,ExpressionUUID->"71b7ba64-d48e-4e69-90ab-9547595f46c7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Package xAct`xTensor`  version \"\>", 
   "\[InvisibleSpace]", "\<\"1.1.4\"\>", "\[InvisibleSpace]", "\<\", \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2020", ",", "2", ",", "16"}], "}"}]}],
  SequenceForm[
  "Package xAct`xTensor`  version ", "1.1.4", ", ", {2020, 2, 16}],
  Editable->False]], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.792804061915328*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->75467753,ExpressionUUID->"6cf8339f-c224-46ca-9cbc-91d7126fc1ca"],

Cell[BoxData["\<\"CopyRight (C) 2002-2020, Jose M. Martin-Garcia, under the \
General Public License.\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.7928040619210787`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1371172941,ExpressionUUID->"1fddaec9-f1f6-420e-912d-b2569353640f"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.792804061927359*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->656603430,ExpressionUUID->"5f1102af-0d9a-4abf-ab95-163e9b78483a"],

Cell[BoxData["\<\"These packages come with ABSOLUTELY NO WARRANTY; for \
details type Disclaimer[]. This is free software, and you are welcome to \
redistribute it under certain conditions. See the General Public License for \
details.\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.7928040619336157`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1762869724,ExpressionUUID->"cb79e0ae-c180-4725-bb3a-218a63bce74d"],

Cell[BoxData["\<\"------------------------------------------------------------\
\"\>"], "Print",
 CellChangeTimes->{{3.7289775339237213`*^9, 3.728977542747715*^9}, 
   3.7928040619392548`*^9},
 CellLabel->"During evaluation of In[1]:=",
 CellID->1867057256,ExpressionUUID->"32c53a62-a534-4654-8d67-c2495637767f"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["0. Basics", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->853073756,ExpressionUUID->"ca3135f0-5abf-4775-a4fc-678f1dffad3e"],

Cell[CellGroupData[{

Cell["0.1. Example session", "Subsection",ExpressionUUID->"49ccbd85-b1ee-4ab9-957e-62b0ff70d533"],

Cell[TextData[{
 "This is a very simple example of the use of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ". We start by defining a manifold and some tensor fields on its tangent \
bundle."
}], "Text",ExpressionUUID->"23d37c49-4c87-4fd8-8d70-984ab004bf6e"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a 3d manifold ",
 StyleBox["M3",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"733d4119-e5b1-4b6b-8ed7-5c97b457d9d7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[2]:=",ExpressionUUID->"67a3701a-8ac8-4629-aa60-772ddf0caed8"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977572682373*^9},
 CellLabel->"During evaluation of In[2]:=",
 CellID->280435421,ExpressionUUID->"a9b233b4-d6a5-4736-b9bd-6d4662bf9526"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977572701936*^9},
 CellLabel->"During evaluation of In[2]:=",
 CellID->1698999229,ExpressionUUID->"c96430fd-4926-488a-b3f2-2d098ec42435"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a contravariant vector ",
 StyleBox["v",
  FontFamily->"Courier"],
 " and a covariant antisymmetric tensor ",
 StyleBox["F",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"008634dc-0c9e-4eb9-9a34-e678dec0fab0"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "}"}], "]"}]}], "]"}]}], "Input",
 CellLabel->"In[3]:=",ExpressionUUID->"eaa10a04-da2d-4175-bb11-da65b21403e2"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977574012474*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->1672621724,ExpressionUUID->"9cc31f94-3ae3-4381-ae36-b22473dbeef1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`F[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289775740317717`*^9},
 CellLabel->"During evaluation of In[3]:=",
 CellID->1192376038,ExpressionUUID->"77953aca-f5eb-4a0d-bb4b-0ad8ddd8843f"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Tensor product of both tensors, based on the abstract index notation:\
\>", "MathCaption",ExpressionUUID->"979b6ed1-be88-4046-84bd-753c6f388b2c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"F", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->"In[5]:=",ExpressionUUID->"6b294b34-4699-4a9f-b31f-f5523ab89b3b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"F", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`F[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897757504585*^9},
 CellLabel->"Out[5]=",
 CellID->41016295,ExpressionUUID->"4bc2242d-cd73-45a7-a86b-249b53db34f0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", " ", 
  RowBox[{"v", "[", "a", "]"}]}]], "Input",
 CellLabel->"In[6]:=",ExpressionUUID->"74ab546e-7541-401f-b7a8-89446501a6b4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"F", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`F[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977575711482*^9},
 CellLabel->"Out[6]=",
 CellID->378027064,ExpressionUUID->"46e5f676-c74b-4032-8930-c05720f5e47b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
There is no automatic simplification. It must be explicitly asked for:\
\>", "MathCaption",ExpressionUUID->"ea0f31b7-36a9-4045-b483-43613ab687c9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellLabel->"In[7]:=",ExpressionUUID->"18eb2e93-cb8e-4c33-b436-517c3a4c17d7"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977576723857*^9},
 CellLabel->"Out[7]=",
 CellID->1898734097,ExpressionUUID->"53fe00b1-86b1-4a8d-9b45-f3b363f21829"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example of dummy handling:", "MathCaption",ExpressionUUID->"3fd18b89-5491-4bc0-8950-24f3dccb5d5a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"v", "[", "d", "]"}], 
   RowBox[{"v", "[", "b", "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}]}], "+", 
  RowBox[{"3", 
   RowBox[{"v", "[", "b", "]"}], 
   RowBox[{"v", "[", "c", "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "c"}], "]"}], 
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", "d"}], "]"}], 
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "a"}]}], "]"}]}]}]], "Input",
 CellLabel->"In[8]:=",ExpressionUUID->"76307278-faac-44e1-907f-0f2948d60a9d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"F", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`F[-$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"F", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`F[-$CellContext`d, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"F", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`F[-$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289775778040648`*^9},
 CellLabel->"Out[8]=",
 CellID->1570938722,ExpressionUUID->"ccc3b734-4654-40e2-b0b6-4574e306b524"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->"In[9]:=",ExpressionUUID->"2c856d83-a0bf-4f3c-ac46-75e5062541f5"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"F", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`F[-$CellContext`a, -$CellContext`c],
     Editable->False], "-", 
    RowBox[{"3", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"F", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`F[-$CellContext`a, -$CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"F", 
          StyleBox[GridBox[{
             {" ", "d"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`F[-$CellContext`c, $CellContext`d],
      Editable->False]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977579827976*^9},
 CellLabel->"Out[9]=",
 CellID->136888559,ExpressionUUID->"9e732302-e714-475a-88c4-36f34c51e7d1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There is not a metric and hence ",
 StyleBox["v",
  FontFamily->"Courier"],
 " cannot be covariant. Note the position of the head ",
 Cell[BoxData[
  StyleBox["ERROR",
   FontColor->RGBColor[1, 0, 0]]], "Output",ExpressionUUID->
  "cf231807-8085-4d29-9f69-f6ad6cc466aa"],
 ":"
}], "MathCaption",ExpressionUUID->"86bdeab5-7032-45af-ae1e-7e5083ca90a9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "%", "]"}]], "Input",
 CellLabel->"In[10]:=",ExpressionUUID->"e1714e59-b838-40c8-95a5-b1120d3c6fd6"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor F\\\\\
\\\"\\\"}]\\)\"", 2, 10, 1, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977580963031*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->276709775,ExpressionUUID->"a22bf1ed-0245-4f15-9d91-91dd92545138"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor v\\\\\
\\\"\\\"}]\\)\"", 2, 10, 2, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977581048689*^9},
 CellLabel->"During evaluation of In[10]:=",
 CellID->1359092670,ExpressionUUID->"f982f1fb-0bbf-486f-8511-146b6ba801cc"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["ERROR",
      FontColor->Hue[0.]], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`b],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ERROR[
    $CellContext`v[-$CellContext`b]],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"F", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`F[-$CellContext`a, -$CellContext`c],
     Editable->False], "-", 
    RowBox[{"3", " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        StyleBox["ERROR",
         FontColor->Hue[0.]], "[", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"F", 
             StyleBox[GridBox[{
                {" ", "d"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`F[-$CellContext`c, $CellContext`d],
         Editable->False], "]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`ERROR[
       $CellContext`F[-$CellContext`c, $CellContext`d]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"F", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`F[-$CellContext`a, -$CellContext`d],
      Editable->False]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977581114188*^9},
 CellLabel->"Out[10]=",
 CellID->1634435198,ExpressionUUID->"c8dc07e1-6548-42dc-bda7-bc1f06ba7eb9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Internal structure:", "MathCaption",ExpressionUUID->"cfe5bdde-9bf4-4c42-962e-fa585d6d69fa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%%", "]"}]], "Input",
 CellLabel->"In[11]:=",ExpressionUUID->"0a91f51a-3d40-43b9-acec-8beb86650730"],

Cell["(F[-a, -c] - 3*F[-a, -d]*F[-c, d])*v[-b]*v[b]*v[c]", "Output",
 CellChangeTimes->{3.728977582567274*^9},
 CellLabel->"Out[11]//InputForm=",
 CellID->9239946,ExpressionUUID->"e272e9ab-b216-43ab-9d1d-924ae292e416"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Now we define a covariant derivative operator and check the first Bianchi \
identity (valid for any symmetric connection). Note that ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " uses unique (dollar-) variables as internal dummy indices to avoid index \
collisions. The function ",
 StyleBox["ScreenDollarIndices",
  FontFamily->"Courier"],
 " hides those indices replacing them by \"normal\" indices. "
}], "Text",ExpressionUUID->"e18353cd-049b-480f-baa8-56723d506293"],

Cell[CellGroupData[{

Cell["\<\
Covariant derivatives are defined by default with curvature but without \
torsion. Note also the default associated symbols: semicolon in Postfix \
notation and nabla in Prefix notation:\
\>", "MathCaption",ExpressionUUID->"af4360a0-05fb-459c-933d-7b0d5575970d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Options", "[", "DefCovD", "]"}]], "Input",
 CellLabel->"In[12]:=",ExpressionUUID->"7746ad4a-dbf0-481b-b66c-676aab67bc23"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"SymbolOfCovD", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\<\";\"\>", ",", "\<\"\[EmptyDownTriangle]\"\>"}], "}"}]}], ",", 
   RowBox[{"Torsion", "\[Rule]", "False"}], ",", 
   RowBox[{"Curvature", "\[Rule]", "True"}], ",", 
   RowBox[{"FromMetric", "\[Rule]", "Null"}], ",", 
   RowBox[{"CurvatureRelations", "\[Rule]", "True"}], ",", 
   RowBox[{"ExtendedFrom", "\[Rule]", "Null"}], ",", 
   RowBox[{"OtherDependencies", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"OrthogonalTo", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"ProjectedWith", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"WeightedWithBasis", "\[Rule]", "Null"}], ",", 
   RowBox[{"ProtectNewSymbol", "\[RuleDelayed]", "$ProtectNewSymbols"}], ",", 
   
   RowBox[{"Master", "\[Rule]", "Null"}], ",", 
   RowBox[{"DefInfo", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\<\"covariant derivative\"\>", ",", "\<\"\"\>"}], "}"}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728977586485179*^9},
 CellLabel->"Out[12]=",
 CellID->1951053507,ExpressionUUID->"8a759938-1bbc-4c24-9eb4-169bff48d812"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "]"}]], "Input",
 CellLabel->"In[13]:=",ExpressionUUID->"bec91a2d-7597-45cf-ba53-e02d48a72bf7"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"Cd", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`Cd[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977587231658*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->2078150768,ExpressionUUID->"80e097f9-1af2-4673-9cad-fe343f30725b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCd[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289775872500896`*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->805922968,ExpressionUUID->"3c6ed23c-f5e7-4c4d-86b3-5c6ac9216fac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977587267672*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->967961547,ExpressionUUID->"86c37c43-9cd6-4ff3-8c98-034d338fe7d8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977587284672*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->1341094404,ExpressionUUID->"825ddd1b-2bbc-4ad2-8d31-52e1d933b591"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289775873021*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->1932772763,ExpressionUUID->"18cd6e80-9b20-4e8b-997f-d8de4fab9462"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728977587319776*^9},
 CellLabel->"During evaluation of In[13]:=",
 CellID->1274183219,ExpressionUUID->"4b2e3e33-9cf9-4be1-acea-008a936c0deb"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Leibnitz rule is automatic:", "MathCaption",ExpressionUUID->"3a2a50f3-9824-47ba-b62e-832dea04f25a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", " ", 
  RowBox[{
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], 
   RowBox[{"v", "[", "c", "]"}]}], " ", "]"}]], "Input",
 CellLabel->"In[14]:=",ExpressionUUID->"cebc62b2-b1ef-48b4-8815-3d59f557291c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"F", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`F[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"F", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`F[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`v[$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728977588480983*^9},
 CellLabel->"Out[14]=",
 CellID->94542730,ExpressionUUID->"1c81698b-9f28-4553-aefb-c433f14b9e3e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", " ", 
   RowBox[{"v", "[", "b", "]"}]}], "//", "Simplification"}]], "Input",
 CellLabel->"In[15]:=",ExpressionUUID->"80c61803-d97e-40f7-84bb-9afdfe0934e9"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"F", 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`F[-$CellContext`b, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Cd[-$CellContext`a][
     $CellContext`v[$CellContext`c]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.7289775891127853`*^9},
 CellLabel->"Out[15]=",
 CellID->170734857,ExpressionUUID->"288343d5-ce25-4c41-bcd1-211a6280a42e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The tensors have been defined with their expected symmetries:\
\>", "MathCaption",ExpressionUUID->"93dd870e-0f51-4948-a667-ca0fc7cbfaf7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "+", 
   RowBox[{"RiemannCd", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}]}], "//", 
  "Simplification"}]], "Input",
 CellLabel->"In[16]:=",ExpressionUUID->"1a9c7bde-e83e-4455-a414-9c8430fec8ec"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289775911757*^9},
 CellLabel->"Out[16]=",
 CellID->1527135633,ExpressionUUID->"b02ea8d3-2324-463b-ab74-25874d16cf63"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
but it is possible to deduce them from more basic principles:\
\>", "MathCaption",ExpressionUUID->"8dd8161c-5d40-40c2-aeac-7216f5523a2e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"RiemannCd", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", "d"}], "]"}], "+", 
    RowBox[{"RiemannCd", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "c"}], ",", "d"}], "]"}]}], "//", "RiemannToChristoffel"}],
   "//", "Simplification"}]], "Input",
 CellLabel->"In[17]:=",ExpressionUUID->"11a3a435-4ae6-42db-bdea-c0e37b4d6181"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977592291641*^9},
 CellLabel->"Out[17]=",
 CellID->1550310597,ExpressionUUID->"40d712c8-5e76-454e-abf2-344a3f6e3f03"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"RiemannCd", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", "d"}], "]"}], "+", 
    RowBox[{"RiemannCd", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", "d"}], "]"}], "+", 
    RowBox[{"RiemannCd", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "d"}], "]"}]}], "//", "RiemannToChristoffel"}],
   "//", "Simplification"}]], "Input",
 CellLabel->"In[18]:=",ExpressionUUID->"04ae657f-ef8f-481b-90f2-04e18cd0faa3"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.72897759312184*^9},
 CellLabel->"Out[18]=",
 CellID->1920925839,ExpressionUUID->"c65979cb-e594-4aa7-bff5-d9b4ebba6cfa"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Check of the second Bianchi identity:", "MathCaption",ExpressionUUID->"f2445a32-362f-49d2-9e09-ba5fd0af47d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Antisymmetrize", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Cd", "[", 
     RowBox[{"-", "e"}], "]"}], "[", 
    RowBox[{"RiemannCd", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "b"}], ",", "a"}], "]"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[19]:=",ExpressionUUID->"6dddded9-a24e-4f60-bf55-ca1d47508409"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "6"], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"d", "e", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`c][
      $CellContext`RiemannCd[-$CellContext`d, -$CellContext`e, \
-$CellContext`b, $CellContext`a]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"e", "d", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`c][
      $CellContext`RiemannCd[-$CellContext`e, -$CellContext`d, \
-$CellContext`b, $CellContext`a]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"c", "e", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`d][
      $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e, \
-$CellContext`b, $CellContext`a]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"e", "c", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`d][
      $CellContext`RiemannCd[-$CellContext`e, -$CellContext`c, \
-$CellContext`b, $CellContext`a]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"c", "d", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`e][
      $CellContext`RiemannCd[-$CellContext`c, -$CellContext`d, \
-$CellContext`b, $CellContext`a]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"d", "c", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`e][
      $CellContext`RiemannCd[-$CellContext`d, -$CellContext`c, \
-$CellContext`b, $CellContext`a]],
     Editable->False]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728977594674554*^9},
 CellLabel->"Out[19]=",
 CellID->668680751,ExpressionUUID->"726abc39-2aa2-4162-a69a-986a72514773"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"3", "%"}], "//", "Simplification"}]], "Input",
 CellLabel->"In[20]:=",ExpressionUUID->"6863af84-7e1b-4a1b-9378-08de11211671"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"d", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`c][
    $CellContext`RiemannCd[-$CellContext`d, -$CellContext`e, -$CellContext`b, \
$CellContext`a]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`d][
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e, -$CellContext`b, \
$CellContext`a]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`e][
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`d, -$CellContext`b, \
$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289775953432837`*^9},
 CellLabel->"Out[20]=",
 CellID->904014618,ExpressionUUID->"bc3317e6-ea2a-45cf-98e4-2816c9180f38"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "CovDToChristoffel"}]], "Input",
 CellLabel->"In[21]:=",ExpressionUUID->"ae6f8bce-4062-45bc-9de5-24742e3cb5d3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "e", "e$2842"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e$2842],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e$2842"},
           {"c", "d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`d, -$CellContext`b, \
$CellContext`e$2842],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2843", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2843, -$CellContext`e, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "d", "e$2843", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`d, \
-$CellContext`e$2843, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "e$2833"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e$2833],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e$2833"},
           {"c", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e, -$CellContext`b, \
$CellContext`e$2833],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2834", " ", " "},
           {" ", "d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2834, -$CellContext`d, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "e", "e$2834", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e, \
-$CellContext`e$2834, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2836", " ", " "},
           {" ", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2836, -$CellContext`d, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "e$2836", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e$2836, \
-$CellContext`b, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2845", " ", " "},
           {" ", "e", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2845, -$CellContext`e, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "e$2845", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e$2845, \
-$CellContext`b, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "e$2824"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e$2824],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e$2824"},
           {"d", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`d, -$CellContext`e, -$CellContext`b, \
$CellContext`e$2824],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2825", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2825, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"d", "e", "e$2825", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`d, -$CellContext`e, \
-$CellContext`e$2825, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2827", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2827, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"d", "e$2827", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`d, -$CellContext`e$2827, \
-$CellContext`b, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2826", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2826, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"e$2826", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`e$2826, -$CellContext`e, \
-$CellContext`b, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2835", " ", " "},
           {" ", "d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2835, -$CellContext`d, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"e$2835", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`e$2835, -$CellContext`e, \
-$CellContext`b, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e$2844", " ", " "},
           {" ", "e", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e$2844, -$CellContext`e, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"e$2844", "d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`e$2844, -$CellContext`d, \
-$CellContext`b, $CellContext`a],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"d", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`RiemannCd[-$CellContext`d, -$CellContext`e, -$CellContext`b, \
$CellContext`a]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "e", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`e, -$CellContext`b, \
$CellContext`a]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "e"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`e][
    $CellContext`RiemannCd[-$CellContext`c, -$CellContext`d, -$CellContext`b, \
$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977597281951*^9},
 CellLabel->"Out[21]=",
 CellID->1314426350,ExpressionUUID->"2cd089c2-762c-4ee3-a704-c928c26b9289"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "RiemannToChristoffel"}], "//", 
  "ScreenDollarIndices"}]], "Input",
 CellLabel->"In[22]:=",ExpressionUUID->"2d413902-e71d-4cdd-adbf-cb5808b0d9b1"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "e", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", " ", " "},
            {" ", "d", "e1"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", " ", " "},
            {" ", "e", "e1"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "e", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "d", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "e", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], 
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`b]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], 
     SubscriptBox["\[PartialD]", "e"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", " ", " "},
            {" ", "c", "e1"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "c", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`c, \
-$CellContext`e1],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "d", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`d, \
-$CellContext`e1],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", " ", " "},
            {" ", "e", "e1"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "e", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "e", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "d", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`c, \
-$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "c", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`d, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "e", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], 
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`b]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], 
     SubscriptBox["\[PartialD]", "e"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", " ", " "},
            {" ", "c", "e1"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "c", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`c, \
-$CellContext`e1],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e, \
-$CellContext`e1],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", " ", " "},
            {" ", "d", "e1"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "d", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`d, \
-$CellContext`e1],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e, \
-$CellContext`e1],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e1]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "e1"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e1],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "e", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`c, \
-$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "c", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e][
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e1", " ", " "},
            {" ", "d", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e1],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "e", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`d, \
-$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "d", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`b]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e1", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e][
       $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "e"], 
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`e][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`b]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "e"], 
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`e][
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`c, \
-$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e1, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`b]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e1"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e1][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "e", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`c, \
-$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e1, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`b]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e1"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "c", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e1][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`c, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "e", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`e, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"a", " ", " "},
               {" ", "e1", "e2"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`e2],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`d, \
-$CellContext`b],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e1, \
-$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`b]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e1"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e1][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`d, \
-$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e1, \
-$CellContext`b],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`b]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e1"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "d", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e1][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`d, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"a", " ", " "},
               {" ", "e1", "e2"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`e2],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e, \
-$CellContext`b],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e1, \
-$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`b]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e1"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e1][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`b]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`e1, -$CellContext`d, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"a", " ", " "},
               {" ", "e1", "e2"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`e2],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e, \
-$CellContext`b],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "e2"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`e2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e2", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`e2, -$CellContext`e1, \
-$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e1", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e1, \
-$CellContext`b]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "e1"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"a", " ", " "},
              {" ", "e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`e1][
       $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`e, \
-$CellContext`b]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728977598133354*^9},
 CellLabel->"Out[22]=",
 CellID->582528635,ExpressionUUID->"25be150a-b1e1-4eb5-9854-99f870cd86e7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Simplification", "[", "%", "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->"In[23]:=",ExpressionUUID->"d823ac90-1c4b-4d64-bf23-46879bb96cad"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.174235`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977599928913*^9},
 CellLabel->"Out[23]=",
 CellID->112544729,ExpressionUUID->"2ca7c995-2ce7-4fe5-89cf-c8aa071b9346"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Commutation of covariant derivatives. It is possible to choose the indices \
(use ",
 StyleBox["CommuteCovDs",
  FontFamily->"Courier"],
 ") or just let ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " to bring the derivative indices into canonical order (use \[LineSeparator]",
 StyleBox["SortCovDs",
  FontFamily->"Courier"],
 "):"
}], "MathCaption",ExpressionUUID->"4d7774b3-33c3-457c-be81-ab513a1cd1a1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"Cd", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{"v", "[", "c", "]"}]}]}]], "Input",
 CellLabel->"In[24]:=",ExpressionUUID->"c94edf35-faa8-4fd7-8444-acc6ad3f86f9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], 
    SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`Cd[-$CellContext`b][
    $CellContext`v[$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289776028339863`*^9},
 CellLabel->"Out[24]=",
 CellID->1474478125,ExpressionUUID->"a1983510-b84c-48f2-801d-6560ef830e6d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CommuteCovDs", "[", 
   RowBox[{"%", ",", "Cd", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "a"}]}], "}"}]}], "]"}], "//", 
  "ScreenDollarIndices"}]], "Input",
 CellLabel->"In[25]:=",ExpressionUUID->"cb171f09-f5a5-440e-9e92-c9952c3f6c7b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "c"},
           {"b", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`b, -$CellContext`a, -$CellContext`d, \
$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], 
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`b][
    $CellContext`Cd[-$CellContext`a][
     $CellContext`v[$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977603548317*^9},
 CellLabel->"Out[25]=",
 CellID->603282111,ExpressionUUID->"5c8fbeca-a455-4972-9450-e02b2b6de003"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SortCovDs", "[", "%%", "]"}], "//", 
  "ScreenDollarIndices"}]], "Input",
 CellLabel->"In[26]:=",ExpressionUUID->"1b96d04a-0ec9-47a2-bdae-5d5758cf4214"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "c"},
           {"b", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`b, -$CellContext`a, -$CellContext`d, \
$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], 
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`b][
    $CellContext`Cd[-$CellContext`a][
     $CellContext`v[$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977604234456*^9},
 CellLabel->"Out[26]=",
 CellID->1412296109,ExpressionUUID->"37cab0ed-b2d1-4d83-bc57-a414052bcace"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Now we introduce a metric on the manifold. Its associated curvature tensors \
are defined: ",
 StyleBox["RiemannCD",
  FontFamily->"Courier"],
 ", ",
 StyleBox["RiciCD",
  FontFamily->"Courier"],
 ", ",
 StyleBox["EinsteinCD",
  FontFamily->"Courier"],
 ", etc:"
}], "Text",ExpressionUUID->"51e18220-a1c8-4ed2-b295-357653e5fbaf"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define the metric ",
 StyleBox["g",
  FontFamily->"Courier"],
 ", with negative determinant. Its Levi-Civita connection will be called ",
 StyleBox["CD",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"3271f7c4-a29c-4892-8b8b-ec36185f6654"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"g", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<|\>\"", ",", "\"\<D\>\""}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[27]:=",ExpressionUUID->"ac761454-d7f9-4d33-9eed-0dad0b039e15"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"g", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`g[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605760187*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->60842753,ExpressionUUID->"cd4a8451-9375-48df-a0eb-8a364f4841e6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilong", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilong[-$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605777892*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->381933699,ExpressionUUID->"6bff225e-8ec5-43b9-ab2f-57985493db0f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605796771*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->331058210,ExpressionUUID->"5fbc85a6-50d0-40e5-968d-e37598c06121"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605817031*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->2112689740,ExpressionUUID->"514c17b8-75ef-4170-b500-49b022339039"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605836664*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->1481840682,ExpressionUUID->"876ec043-77b4-46ab-a983-87051b44317f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605856462*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->928488115,ExpressionUUID->"c25eacb1-13f8-491e-9f34-482431b3227e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605875146*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->1483569703,ExpressionUUID->"e24d0c23-7a7d-40b9-9ab7-15e0b71d2165"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.7289776058945932`*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->2070856651,ExpressionUUID->"9b1348f6-aaf0-49f8-8d89-8b310913d49c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605915947*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->620266428,ExpressionUUID->"8aff0875-f24a-4b75-ad64-a11d063fb7d8"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.7289776059368057`*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->932232858,ExpressionUUID->"5f799f0e-b882-4d37-92d1-0033bbfbed60"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977605956944*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->2144439541,ExpressionUUID->"4883ee58-57ef-4a85-b54b-03e1edf78e1f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776059788027`*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->377317986,ExpressionUUID->"10d712eb-17e0-41d0-b476-672ea12170b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977606000885*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->64181111,ExpressionUUID->"bf248ee2-2699-41b2-9513-57190ffb4409"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977606021306*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->655141783,ExpressionUUID->"e7295a78-6d23-491e-9f21-6f252da0b3e7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977606041692*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->819288712,ExpressionUUID->"cc479cf0-18e6-4b31-a4e2-30bf036da699"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977606061721*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->120238020,ExpressionUUID->"7fd59c0a-7e0f-4627-ba0d-efd4564d1800"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977606082439*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->58290313,ExpressionUUID->"8d634f01-72dc-4572-866f-3c47ad9863f5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detg", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detg[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897760610283*^9},
 CellLabel->"During evaluation of In[27]:=",
 CellID->432747559,ExpressionUUID->"d3012688-4195-4d92-a883-598285d5af02"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Standard properties of the metric and its associated objects:\
\>", "MathCaption",ExpressionUUID->"a0034d48-2255-40ac-abff-8e8cefcfa0cf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", " ", 
  RowBox[{"g", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], " ", "]"}]], "Input",
 CellLabel->"In[28]:=",ExpressionUUID->"e931d299-b6fe-4d3c-93bd-cf0902fb6299"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977608043634*^9},
 CellLabel->"Out[28]=",
 CellID->994858890,ExpressionUUID->"f541b0d6-0f48-4277-8807-48b4859e9d62"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", " ", 
  RowBox[{"EinsteinCD", "[", 
   RowBox[{"a", ",", "b"}], "]"}], " ", "]"}]], "Input",
 CellLabel->"In[29]:=",ExpressionUUID->"c17da91a-18b8-44db-a347-2cdcb8ab1800"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977608869609*^9},
 CellLabel->"Out[29]=",
 CellID->945402066,ExpressionUUID->"3f28828a-aa7e-45e8-9197-8f9f556c3ea9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"g", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "a"}]}], "]"}]], "Input",
 CellLabel->"In[30]:=",ExpressionUUID->"12cb5df4-aab9-4a7f-a5b8-4e3f3fd5e041"],

Cell[BoxData["3"], "Output",
 CellChangeTimes->{3.728977609716557*^9},
 CellLabel->"Out[30]=",
 CellID->2032759218,ExpressionUUID->"430ca87d-03b6-4250-adbd-dbc43d3bb04c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"g", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"g", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->"In[31]:=",ExpressionUUID->"e6d23bb3-801f-4340-925c-9fa6adb26f72"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "a"},
         {"c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`c, $CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897761048624*^9},
 CellLabel->"Out[31]=",
 CellID->1928400476,ExpressionUUID->"6af945d8-0592-445c-b8f3-000e55be5d3d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Let us check the antisymmetry of the second pair of indices in the Riemann \
tensor:\
\>", "MathCaption",ExpressionUUID->"17a21e64-ee23-46dc-b27f-a0e8c5598a4c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}]}], "]"}], "+", 
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->"In[32]:=",ExpressionUUID->"3f44d07e-b351-4f9f-876b-b747f7d0ebd0"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "d", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977615082753*^9},
 CellLabel->"Out[32]=",
 CellID->773979084,ExpressionUUID->"9adf5748-ee7d-4a27-b08e-44fb57a1214f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Because canonicalization is not automatic, the previous expression is not \
automatically zero. We expand it in derivatives of Christoffels and \
canonicalize. The partial derivative does not commute with the metric, so \
that internally the code replaces temporarily those derivatives with \
Levi-Civita connections:\
\>", "MathCaption",ExpressionUUID->"a9007a20-2ac3-4bae-ab9f-f32f221dc61a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "//", "RiemannToChristoffel"}], "//", "Simplification"}], "//",
   "ScreenDollarIndices"}]], "Input",
 CellLabel->"In[33]:=",ExpressionUUID->"4d13ad75-10b3-453f-bf3b-dcfbbc50d6f2"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 33, 3, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977616773161*^9},
 CellLabel->"During evaluation of In[33]:=",
 CellID->901495669,ExpressionUUID->"3ebbfba0-68a6-4c4c-ae0e-5e8770c0a313"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 33, 4, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977616828382*^9},
 CellLabel->"During evaluation of In[33]:=",
 CellID->940579298,ExpressionUUID->"3109e3f3-3662-49e9-9ca7-d57187803cea"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 33, 5, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289776169045067`*^9},
 CellLabel->"During evaluation of In[33]:=",
 CellID->689135919,ExpressionUUID->"e1c40249-2ec5-496f-9f12-dc516588d066"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"ToCanonical\\\", \\\"::\
\\\", \\\"cmods\\\"}], \\\"MessageName\\\"]\\) will be suppressed during this \
calculation.\"", 2, 33, 6, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977616931631*^9},
 CellLabel->"During evaluation of In[33]:=",
 CellID->928082362,ExpressionUUID->"b59aac10-b288-4ec5-a390-2d93b0189ed0"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "e"},
           {"d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`b, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"e", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "e"},
           {"c", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`b, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"e", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`e, -$CellContext`a, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "e"},
           {"d", "a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`a, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"e", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "e"},
           {"c", "a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`a, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"e", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`e, -$CellContext`b, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e1, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[-$CellContext`c, -$CellContext`e1],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e1, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[-$CellContext`c, -$CellContext`e1],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e1, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[-$CellContext`d, -$CellContext`e1],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e1", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e1, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e1"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`g[-$CellContext`d, -$CellContext`e1],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`b, \
-$CellContext`d]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`a, \
-$CellContext`d]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289776170000973`*^9},
 CellLabel->"Out[33]=",
 CellID->1106529973,ExpressionUUID->"17eb4c04-ed79-44dd-a6f5-b598c1fb6b10"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Metric contraction is not automatic either. The resulting expression is not \
obviously zero:\
\>", "MathCaption",ExpressionUUID->"9acd67ce-9b8f-4773-bd43-7d014da9f836"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ContractMetric"}], "//", "Simplification"}]], "Input",
 CellLabel->"In[34]:=",ExpressionUUID->"d29c4f3d-bc25-43f4-9a45-f6305c49c24c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"c", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`b, \
-$CellContext`d]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`a, \
-$CellContext`d]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289776188033953`*^9},
 CellLabel->"Out[34]=",
 CellID->1308641128,ExpressionUUID->"c13f4b23-11c7-43c5-8278-4ec635c0f151"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Actually it is only zero for connections deriving from a metric:\
\>", "MathCaption",ExpressionUUID->"71998cc3-cd1d-4173-88eb-8ca18b5a5173"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ChristoffelToGradMetric"}], "//", "Expand"}]], "Input",
 CellLabel->"In[35]:=",ExpressionUUID->"ea659810-066e-48e3-8459-279068cdef0b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "2"]}], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`g[-$CellContext`c, -$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`g[-$CellContext`d, -$CellContext`c]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`g[-$CellContext`c, -$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`g[-$CellContext`d, -$CellContext`c]]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289776201840467`*^9},
 CellLabel->"Out[35]=",
 CellID->1675831686,ExpressionUUID->"9851259a-4014-4d4b-b3a6-fb13c32eaf04"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->"In[36]:=",ExpressionUUID->"765c0c44-a175-4bd6-aade-764ecee082b1"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289776209335938`*^9},
 CellLabel->"Out[36]=",
 CellID->1541888581,ExpressionUUID->"6f9dafc6-7ca6-4916-9e22-43a8e646edf0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In this case a simple reordering of the partial derivatives (this is not \
automatic) would have been enough:\
\>", "MathCaption",ExpressionUUID->"4d479e15-4ba2-4b16-8537-f5dd10af7403"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "SortCovDs"}]], "Input",
 CellLabel->"In[37]:=",ExpressionUUID->"084f69a4-8874-48c6-9086-571b72382de7"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977622139701*^9},
 CellLabel->"Out[37]=",
 CellID->588011226,ExpressionUUID->"0fc2aabc-9475-44b8-b32d-78834a554db7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Examples of Lie derivative and bracket:", "MathCaption",ExpressionUUID->"db3e68e8-3010-4329-979f-789cb337196a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LieD", "[", 
   RowBox[{"3", 
    RowBox[{"v", "[", "a", "]"}]}], "]"}], "[", 
  RowBox[{
   RowBox[{"F", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], 
   RowBox[{"v", "[", "b", "]"}]}], "]"}]], "Input",
 CellLabel->"In[38]:=",ExpressionUUID->"abdd0cc2-c5c4-4a11-bf96-26120436579b"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[ScriptCapitalL]", "v"], GridBox[{
        {"F", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`LieD[
     $CellContext`v[$CellContext`a]][
     $CellContext`F[-$CellContext`a, -$CellContext`b]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728977623367653*^9},
 CellLabel->"Out[38]=",
 CellID->887726534,ExpressionUUID->"7683e2d6-d063-4c43-bd0b-37b4b87caf19"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Bracket", "[", 
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], ",", 
    RowBox[{
     RowBox[{"F", "[", 
      RowBox[{"a", ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"v", "[", "b", "]"}]}]}], "]"}], "[", "a", "]"}]], "Input",
 CellLabel->"In[39]:=",ExpressionUUID->"1bbfeb63-fe02-4dd5-90f1-24f64e4f141c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"[", GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}], ",", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"F", 
             StyleBox[GridBox[{
                {"a", " "},
                {" ", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`F[$CellContext`a, -$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}], "]"}], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Bracket[
   $CellContext`v[$CellContext`a], $CellContext`F[$CellContext`a, \
-$CellContext`b] $CellContext`v[$CellContext`b]][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289776241010838`*^9},
 CellLabel->"Out[39]=",
 CellID->207576009,ExpressionUUID->"2ac1cfa7-707c-4ca0-ab41-9e276c415056"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Undefine all objects on the manifold and the manifold itself:\
\>", "MathCaption",ExpressionUUID->"3f77be99-090a-49e6-b0c5-bb118d0b1692"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "g", "]"}]], "Input",
 CellLabel->"In[40]:=",ExpressionUUID->"abeaa576-2e54-4724-9711-781417301268"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776253131*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1828814815,ExpressionUUID->"3a29ab3d-9582-41e9-b8ea-a295e535af89"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625331307*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->2104639841,ExpressionUUID->"48cff370-5f01-49b5-a6a2-f99863def3c0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625350045*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1478657950,ExpressionUUID->"09cfae9b-700a-4eef-9153-05e38e820a3f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776253719397`*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1123402566,ExpressionUUID->"dba1c412-6958-4141-af3d-bf8af67bf99c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776253898783`*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->350702338,ExpressionUUID->"d59ed287-1150-420e-9b0e-e534ffe442ef"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625411035*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1559067398,ExpressionUUID->"7ac86513-5211-4652-9b5c-ac92ee705cf9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625432012*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1797781139,ExpressionUUID->"0161eb23-2136-4ae0-9553-4238645f5e8e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625451996*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->212689883,ExpressionUUID->"2c3c0f16-270b-4dc9-a3ed-9586f457b607"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625473542*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1266074973,ExpressionUUID->"e7809ed8-dbb8-4cea-9717-950f98cbf247"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625494875*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1654096735,ExpressionUUID->"84270404-7932-4fcc-b76d-c401e8c5fa07"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776255162086`*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1228014730,ExpressionUUID->"d0e91fcb-64c8-4210-b4fc-1f3d772ff66e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilong"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilong],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897762553723*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->1925719440,ExpressionUUID->"dbfea2e9-e9bb-45e6-8527-48bad9dee745"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "g"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`g],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977625560272*^9},
 CellLabel->"During evaluation of In[40]:=",
 CellID->840584143,ExpressionUUID->"4245844c-4c89-43fd-a30d-817c279304d4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"v", ",", "F"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"UndefCovD", "[", "Cd", "]"}]}], "Input",
 CellLabel->"In[41]:=",ExpressionUUID->"91e11ed1-820b-4c86-89df-2be8b24baa86"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897762647059*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->1457251913,ExpressionUUID->"e9715e34-33d3-4607-8ff6-7438d65a1aa6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "F"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`F],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977626487185*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->2129391403,ExpressionUUID->"1418575a-76a2-44c3-b12b-e04cb38f80a9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977626505556*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->866332966,ExpressionUUID->"ca333a7e-fc91-43a0-a37c-fedf286fdeb1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977626523974*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->519032902,ExpressionUUID->"9bdb7152-61f3-4137-b7a5-6f8a1be95005"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776265424023`*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->307409800,ExpressionUUID->"50a7d844-5da4-41ee-8c0f-1b760a96adf2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977626563184*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->1938337565,ExpressionUUID->"5e36962d-7b18-4d18-a363-9076df425baf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Cd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`Cd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977626584396*^9},
 CellLabel->"During evaluation of In[41]:=",
 CellID->312529965,ExpressionUUID->"ac560e78-b987-458f-a447-57c1f6c03c6b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellLabel->"In[43]:=",ExpressionUUID->"5a9af476-fba0-4d6d-a72e-c1eabd4c9ce8"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776278574057`*^9},
 CellLabel->"During evaluation of In[43]:=",
 CellID->132224728,ExpressionUUID->"543a0863-baf2-4e44-87b4-8a91dc82adb6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776278788967`*^9},
 CellLabel->"During evaluation of In[43]:=",
 CellID->1802912498,ExpressionUUID->"a8f07585-618c-4e5a-befa-d81792a06a29"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There are no symbols left in the ",
 StyleBox["Global`",
  FontFamily->"Courier"],
 " context"
}], "MathCaption",ExpressionUUID->"701669b9-41a3-46d5-abd1-294561a8e615"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Global`*"}]], "Input",
 CellLabel->"In[44]:=",ExpressionUUID->"e8679f61-e704-4ff9-a662-fa5da7ef1d14"],

Cell["Global`$UseTextSearchQ", "Print", "PrintUsage",
 CellChangeTimes->{3.7289776289837217`*^9},
 CellTags->"Info443728981228-9181623",
 CellID->1995674241,ExpressionUUID->"94eae8c2-9b3a-4e88-b43e-adc509ad506c"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["0.2. Summary", "Subsection",ExpressionUUID->"5889dd12-3523-4e67-99cd-4929c03562c1"],

Cell[TextData[{
 "These are the objects that currently can be defined in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ".\n\t- Manifolds\n\t- Mappings\n\t- Vector bundles\n\t- Indices (of \
different types)\n\t- Tensors (in particular, Metrics)\n\t- Covariant \
derivative operators (including partial derivatives)\n\t- Constant symbols\n\t\
- Parameters\n\t- Inert heads (wrappers for tensors)\n\t- Scalar functions\n\t\
- Charts (coordinate systems): use the twin package ",
 StyleBox["xCoba` ",
  FontFamily->"Courier"],
 "(in collaboration with D. Yllanes)\n\t- Bases (frames): use the twin \
package ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 "\nWe also have predefined mathematical operators:\n\t- Lie derivative\n\t- \
Lie bracket\n\t- Parametric derivative\n\t- Variational derivative\nand \
useful mathematical operations:\n\t- Relations and rules among indexed \
objects\n\t- Indexed equations solving\n\t- Canonicalization and \
simplification of expressions\n\t- Automatic definition of curvature tensors, \
and relations among them\n\t- Complex conjugation\nThere are also specialized \
algorithms, frequently used in General Relativity:\n\t- Warped metric \
decomposition\n\t- ADM-like decomposition\n\t- Perturbation theory in General \
Relativity: use the twin package ",
 StyleBox["xPert` ",
  FontFamily->"Courier"],
 "(in collaboration with D. Brizuela and G. Mena Marug\[AAcute]n)\n\t- \
Spherical tensor harmonics: use the twin package ",
 StyleBox["Harmonics` ",
  FontFamily->"Courier"],
 "(in collaboration with D. Brizuela and G. Mena Marug\[AAcute]n)\n\t- \
Manipulation of invariants of the Riemann tensor: use the twin package ",
 StyleBox["Invar` ",
  FontFamily->"Courier"],
 "(in collaboration with R. Portugal, L. Manssur and D. Yllanes)\n\t- Spinor \
calculus (in collaboration with A. Garc\[IAcute]a-Parrado)"
}], "Text",ExpressionUUID->"6d0e05f7-f5f5-43a1-9fe1-f0b6022a9853"]
}, Closed]],

Cell[CellGroupData[{

Cell["0.3. Getting info", "Subsection",ExpressionUUID->"49d482a0-f3b5-4c3e-8b1f-dc27e05614d1"],

Cell[CellGroupData[{

Cell[TextData[{
 "Every piece of information in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " is associated to a symbol and can be obtained using the ",
 StyleBox["?",
  FontFamily->"Courier"],
 " and ",
 StyleBox["??",
  FontFamily->"Courier"],
 " symbols (see the command ",
 StyleBox["Information",
  FontFamily->"Courier"],
 " in the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " help). ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " strictly follows that rule. (In particular, we have explicitly avoided \
notations based on string input/output which would require a parser, much \
restricting the capabilities of the user.)"
}], "Text",ExpressionUUID->"cd41f060-d80b-4635-8e01-65bab609a452"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "DefManifold"}]], "Input",
 CellLabel->"In[45]:=",ExpressionUUID->"89d6fdb0-314a-4340-85fe-3cecc475cf03"],

Cell[BoxData["\<\"DefManifold[M, dim, {a, b, c,...}] defines M to be an \
n-dimensional differentiable manifold with dimension dim (a positive integer \
or a constant symbol) and tensor abstract indices a, b, c, ... .  \
DefManifold[M, {M1, ..., Mm}, {a, b, c,...}] defines M to be the product \
manifold of previously defined manifolds M1 ... Mm. For backward \
compatibility dim can be a list of positive integers, whose length is \
interpreted as the dimension of the defined manifold.\"\>"], "Print", \
"PrintUsage",
 CellChangeTimes->{3.728977653717033*^9},
 CellTags->"Info453728981253-9181623",
 CellID->1862154774,ExpressionUUID->"e4d5af28-bacd-48e1-a58c-0f67bfc59604"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"??", "ERROR"}]], "Input",
 CellLabel->"In[46]:=",ExpressionUUID->"fc20be92-4173-4af0-95d3-9b1a35edd183"],

Cell[CellGroupData[{

Cell[BoxData["\<\"ERROR[expr] is an expression where an error has been \
detected. ERROR is an inert head with a single argument.\"\>"], "Print", \
"PrintUsage",
 CellChangeTimes->{3.728977653918168*^9},
 CellTags->"Info463728981253-9181623",
 CellID->1956098887,ExpressionUUID->"61ca6cff-8a7c-4bca-b966-a93a55cad7c9"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {
     RowBox[{
      RowBox[{"Attributes", "[", "ERROR", "]"}], "=", 
      RowBox[{"{", 
       RowBox[{"HoldFirst", ",", "Protected"}], "}"}]}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "ERROR", "]"}], "^=", 
         RowBox[{"{", 
          
          RowBox[{"\<\"inert head\"\>", 
           ",", "\<\"Generic head to wrap expressions with errors.\"\>"}], 
          "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"InertHeadQ", "[", "ERROR", "]"}], "^=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"LinearQ", "[", "ERROR", "]"}], "^=", "False"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "ERROR", "]"}], 
         "^=", "\<\"\\!\\(\\*StyleBox[\\\"ERROR\\\",FontColor->Hue[0.]]\\)\"\>\
"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[xAct`xTensor`ERROR],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776539419117`*^9},
 CellTags->"Info463728981253-9181623",
 CellID->326841886,ExpressionUUID->"560a3844-0f62-4961-bffb-79612314e255"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["0.4. Too many brackets!", "Subsection",ExpressionUUID->"421f06b6-b5e8-46e6-9b8c-07e14a614d08"],

Cell[TextData[{
 "This is a common complaint from most users after their first contact with \
",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ". My answer to that is the following:"
}], "Text",ExpressionUUID->"5cffb5aa-7bf8-46f7-9de6-c8a1c83f4f25"],

Cell[CellGroupData[{

Cell[TextData[{
 "1. That is how ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " internally works. For example, did you expect this simple derivative \
having internally six pairs of brackets, when it has just one in the \
\"user-notation\"?"
}], "MathCaption",ExpressionUUID->"8f8513b0-163c-471e-a39e-3e01852ab6c4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "'"}], "[", 
   RowBox[{
    RowBox[{"2", "x"}], "-", "y"}], "]"}], "//", "FullForm"}]], "Input",
 CellLabel->"In[47]:=",ExpressionUUID->"3483e09f-16f4-4b30-b162-c683f4ed5e3a"],

Cell[BoxData[
 TagBox[
  StyleBox[
   RowBox[{
    RowBox[{
     RowBox[{"Derivative", "[", "1", "]"}], "[", "f", "]"}], "[", 
    RowBox[{"Plus", "[", 
     RowBox[{
      RowBox[{"Times", "[", 
       RowBox[{"2", ",", "x"}], "]"}], ",", 
      RowBox[{"Times", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "y"}], "]"}]}], "]"}], "]"}],
   ShowSpecialCharacters->False,
   ShowStringCharacters->True,
   NumberMarks->True],
  FullForm]], "Output",
 CellChangeTimes->{3.728977658205041*^9},
 CellLabel->"Out[47]//FullForm=",
 CellID->1240949882,ExpressionUUID->"6919d85c-3f2b-4b91-a9d2-0b8ac99f42b1"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "2. ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " does not have different internal and external notations (at least not \
now). This helps understanding what is happening internally and allows the \
user to build pattern rules which always work. However, you are continually \
exposed to the internal notation."
}], "MathCaption",ExpressionUUID->"4d9bf58f-f25b-4a42-b26d-97acdd4601dd"],

Cell[TextData[{
 "3. The highest accummulation of brackets happens in derivatives. This \
simply follows the general idea in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " (made clear in the example above) of treating a derivative as an operator \
on an expression, and not as a simple additional argument of the expression, \
as the traditional \"comma-index\" notation might suggest."
}], "MathCaption",ExpressionUUID->"403cc0c0-39e7-440d-ac09-bfcd7e3d1699"],

Cell[TextData[{
 "4. The ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " language is very similar to LISP, in which you have a lot of parenthesis."
}], "MathCaption",
 CellChangeTimes->{
  3.494606778168723*^9},ExpressionUUID->"36e8529a-7053-4ef3-a291-\
9628ad03698f"],

Cell[CellGroupData[{

Cell["\<\
5. If you are still not convinced by the previous three arguments, then I \
must say that it is possible to construct whatever other notation you like \
for a derivative, or for any other operator, and several examples of this are \
given in section 6.1 below.\
\>", "MathCaption",ExpressionUUID->"47104d1e-2ff2-4c2c-9286-76126b9bf7ee"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"f", ",", "x", ",", "y"}], "]"}]], "Input",
 CellLabel->"In[48]:=",ExpressionUUID->"e615b8ea-04d7-41a0-8745-10fae41ccf8a"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["0.5. Warnings", "Subsection",ExpressionUUID->"326cef8c-f330-45b8-9cd1-30477fc02a2d"],

Cell[TextData[{
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " is a very powerful tool in the hands of an average ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " user, but could be difficult for an unexperienced ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " user. This is because I have not tried to hide ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " subtleties (like delayed assignments, unique symbols or upvalues), but \
rather I have tried to make as much use as possible of such powerful \
instruments. If you don't understand those three listed concepts, please \
spend some (not really much) time with a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " tutorial. It will pay off, even if you finally don't use ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"795b24da-db86-46d7-acda-cf427ab5e34a"],

Cell[TextData[{
 "If you think you can do your tensor computation by hand in a few lines, \
then ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " is probably too much. This system is designed to be efficient when \
developing big projects, not small computations: a general-purpose tensor \
computer algebra system always has hundreds of commands, and only big \
projects give you enough motivation to learn them."
}], "Text",ExpressionUUID->"23fc648e-4045-4311-b83d-7549addc71c7"],

Cell[TextData[{
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " is a manipulator of abstract expressions. It does not compute (at least \
easily) the components of the Riemann tensor associated with a given matrix \
of metric components, and things like that. ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " is specialized in the manipulation of identities obeyed by families of \
objects with some properties (like being antisymmetric). Many users feel \
dissapointed at the beginning because they were expecting component-type \
computations. Use the twin package ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 " for that."
}], "Text",ExpressionUUID->"9e78cb1c-454f-4695-b5d1-507e97f3f100"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["1. Manifolds, vector bundles and parameters", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->390546124,ExpressionUUID->"f7d60e03-9b2e-438d-919f-8d580da78398"],

Cell[CellGroupData[{

Cell["1.1. Define a manifold and its tangent bundle", "Subsection",ExpressionUUID->"e8030db3-1fe6-4a08-824b-ae44525c0ac0"],

Cell[TextData[{
 "Just after loading,  ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " does not have any manifold defined. Here we see how to declare our own \
abstract manifolds, on which tensor fields will be defined."
}], "Text",ExpressionUUID->"0e6cf3af-6b03-4690-9088-015deab2bb53"],

Cell[TextData[{
 StyleBox[ButtonBox["DefManifold",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefManifold.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Define a manifold or a product of manifolds\n", "Text"],
 StyleBox[ButtonBox["UndefManifold",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefManifold.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Undefine a manifold\n", "Text"],
 StyleBox[ButtonBox["ManifoldQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ManifoldQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check manifold", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["$Manifolds",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$Manifolds.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined manifolds\n", "Text"],
 StyleBox[ButtonBox["$ProductManifolds",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$ProductManifolds.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["List of defined product manifolds", "Text"]
}], "DefinitionBox",ExpressionUUID->"1ef92702-7204-4829-825f-4d0221cb4c06"],

Cell["Definition of a manifold.", "Text",ExpressionUUID->"98f97fb0-2352-438c-9eee-7798834bb3fd"],

Cell[CellGroupData[{

Cell[TextData[{
 "We define a 3-dimensional manifold ",
 StyleBox["M3",
  FontFamily->"Courier"],
 " with abstract indices {a, b, ..., h} associated to its tangent vector \
bundle:"
}], "MathCaption",ExpressionUUID->"a102bb63-2a7b-4eae-8d1b-25fabc2ddb5a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", 
     "h"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[49]:=",ExpressionUUID->"d39feba1-d953-4465-be4f-f38923b0a0ad"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776736609592`*^9},
 CellLabel->"During evaluation of In[49]:=",
 CellID->894423641,ExpressionUUID->"c0dd4935-216e-43a3-81ba-4651922e51f3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977673681849*^9},
 CellLabel->"During evaluation of In[49]:=",
 CellID->192605287,ExpressionUUID->"af7cc6fa-cafc-4978-8540-2cced64763b0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DimOfManifold", "[", "M3", "]"}]], "Input",
 CellLabel->"In[50]:=",ExpressionUUID->"39af99de-0f20-4702-96c6-86ce614679e2"],

Cell[BoxData["3"], "Output",
 CellChangeTimes->{3.728977673836411*^9},
 CellLabel->"Out[50]=",
 CellID->859202580,ExpressionUUID->"7079644f-84ec-4707-bbb3-c2328f7b02a9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We can ask ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " about this manifold using the question mark ",
 StyleBox["?",
  FontFamily->"Courier"],
 " (command ",
 StyleBox["Information",
  FontFamily->"Courier"],
 "). We see that the data is stored in several property-functions using the \
concept of \"upvalue\" (this is why we find the  ^= sign below)"
}], "MathCaption",ExpressionUUID->"8ad557fd-7d55-44df-b518-57f5ee4392ca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "M3"}]], "Input",
 CellLabel->"In[51]:=",ExpressionUUID->"97c5c513-db25-4ea5-9b9f-083bc3d3e4cd"],

Cell[CellGroupData[{

Cell["Global`M3", "Print", "PrintUsage",
 CellChangeTimes->{3.728977675687475*^9},
 CellTags->"Info513728981275-9181623",
 CellID->580545217,ExpressionUUID->"0c64b26f-1b6a-426c-8169-13f835a49e44"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "M3", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"manifold\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DimOfManifold", "[", "M3", "]"}], "^=", "3"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ManifoldQ", "[", "M3", "]"}], "^=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "M3", "]"}], "^=", "\<\"M3\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ServantsOf", "[", "M3", "]"}], "^=", 
         RowBox[{"{", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SplittingsOfManifold", "[", "M3", "]"}], "^=", 
         RowBox[{"{", "}"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776757100477`*^9},
 CellTags->"Info513728981275-9181623",
 CellID->1205253798,ExpressionUUID->"794dc6e3-9199-42ce-b23c-afc9b0990f95"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Together with the manifold we have also defined its tangent vector bundle, \
which stores all the information related to indices and indexed objects. By \
default, the tangent bundle is named by joining the symbol ",
 StyleBox["Tangent",
  FontFamily->"Courier"],
 " with the name of the manifold, but that name can be freely choosen. To \
conform with the standard notation, it is typeset by default as the character \
",
 StyleBox["\[DoubleStruckCapitalT]",
  FontFamily->"Courier"],
 " followed by the typesetting of the manifold, but the typsetting can also \
be freely chosen. Note that we shall often abbreviate \"vector bundle\" to \
\"vbundle\"."
}], "Text",ExpressionUUID->"9a5ff845-8fc9-4b04-966b-e13a0861acde"],

Cell[TextData[{
 StyleBox[ButtonBox["Tangent",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TangentReals.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["[M]\t\t", "DisplayMath"],
 StyleBox["Tangent bundle of the manifold M", "Text"]
}], "DefinitionBox",ExpressionUUID->"93c7beb0-5818-4811-b987-edb23c890510"],

Cell["Tangent bundle of a manifold.", "Text",ExpressionUUID->"ec987b65-22b9-4127-82c4-1ff98ca47192"],

Cell[CellGroupData[{

Cell["\<\
By default, tangent bundles are real and have no metric. The dimension of the \
tangent bundle is defined as that of the vector space at each point, and \
coincides with the dimension of the base manifold.\
\>", "MathCaption",ExpressionUUID->"4387b72c-5344-483c-9fc7-70a077da29a1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Tangent", "[", "M3", "]"}]], "Input",
 CellLabel->"In[52]:=",ExpressionUUID->"6a3d6e27-68d1-4ed0-b12f-b7b5f0d81b17"],

Cell[BoxData[
 InterpretationBox[
  StyleBox["\[DoubleStruckCapitalT]M3",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TangentM3,
  Editable->False]], "Output",
 CellChangeTimes->{3.728977684825652*^9},
 CellLabel->"Out[52]=",
 CellID->702233062,ExpressionUUID->"49510770-32ed-48b3-a843-9df916670925"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->"In[53]:=",ExpressionUUID->"77e79572-07a7-407f-8d75-25de623e6c28"],

Cell["TangentM3", "Output",
 CellChangeTimes->{3.728977686296956*^9},
 CellLabel->"Out[53]//InputForm=",
 CellID->163721661,ExpressionUUID->"a922cc88-5ad6-4a59-9153-80931a54f97c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "TangentM3"}]], "Input",
 CellLabel->"In[54]:=",ExpressionUUID->"d8756b4b-befd-4be5-98f3-fca67301574e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  StyleBox["\[DoubleStruckCapitalT]M3",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TangentM3,
  Editable->False]], "Print", "PrintUsage",
 CellChangeTimes->{3.728977686955414*^9},
 CellTags->"Info543728981286-9181623",
 CellID->76239291,ExpressionUUID->"a14a8109-2565-4fe7-84ec-88b5522828b6"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"BaseOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", "M3"}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False], "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`TangentM3],
          Editable->False], "^=", 
         InterpretationBox[
          StyleBox["\[DoubleStruckCapitalT]M3",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`TangentM3,
          Editable->False]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"vbundle\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DimOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", "3"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"IndicesOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
            "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", 
             ",", "h"}], "}"}], ",", 
           RowBox[{"{", "}"}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"LastIndex", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"h", ",", "1"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", "M3"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MetricsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], 
         "^=", "\<\"\[DoubleStruckCapitalT]M3\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SplittingsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"VBundleQ", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977686980287*^9},
 CellTags->"Info543728981286-9181623",
 CellID->1453510225,ExpressionUUID->"2ec7dae1-8a02-4a21-a5a7-9fd5c71f3517"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We define a second manifold ",
 StyleBox["S2",
  FontFamily->"Courier"],
 ". Note that we can use ",
 StyleBox["C",
  FontFamily->"Courier"],
 ", ",
 StyleBox["D",
  FontFamily->"Courier"],
 ", ",
 StyleBox["K",
  FontFamily->"Courier"],
 ", ",
 StyleBox["N",
  FontFamily->"Courier"],
 ", ",
 StyleBox["O",
  FontFamily->"Courier"],
 " as indices, even though ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " has reserved meanings for those symbols. The capitals ",
 StyleBox["E",
  FontFamily->"Courier"],
 " and ",
 StyleBox["I",
  FontFamily->"Courier"],
 " cannot be used as indices because they are always understood as the base \
of natural logarithms and the square root of -1, respectively. ",
 StyleBox["xTensor`", "DisplayMath"],
 " issues warnings to remind you of this point:"
}], "MathCaption",ExpressionUUID->"045d3424-06d0-4dc2-b568-5739921d2361"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"S2", ",", " ", "2", ",", " ", 
   RowBox[{"{", 
    RowBox[{"A", ",", "B", ",", "C", ",", "D", ",", "F", ",", "G", ",", "H"}],
     "}"}]}], "]"}]], "Input",
 CellLabel->"In[55]:=",ExpressionUUID->"c829f48e-d04d-4c93-bf2a-d1bb043cdf03"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "S2", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`S2, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776889445963`*^9},
 CellLabel->"During evaluation of In[55]:=",
 CellID->391777724,ExpressionUUID->"96aa42ed-b72f-4b51-8ebc-18fbe500705d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentS2", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentS2, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977688962801*^9},
 CellLabel->"During evaluation of In[55]:=",
 CellID->423345566,ExpressionUUID->"70605633-1a52-4f41-95cf-52a2d3d355c9"]
}, Open  ]],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"C\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 55, 7, 19253573948027831207, "Local", 
   "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977688991045*^9},
 CellLabel->"During evaluation of In[55]:=",
 CellID->555605251,ExpressionUUID->"0e4c297c-4466-49fc-98f1-255a4fe9dcda"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"D\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 55, 8, 19253573948027831207, "Local", 
   "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977689016057*^9},
 CellLabel->"During evaluation of In[55]:=",
 CellID->798556401,ExpressionUUID->"3ac2529b-dc46-4b1a-bdab-d9a8bcb005a6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The lists of manifolds and vbundles defined in the current session are \
given by the global variables ",
 StyleBox["$Manifolds",
  FontFamily->"Courier"],
 " and ",
 StyleBox["$VBundles",
  FontFamily->"Courier"],
 ", respectively:"
}], "MathCaption",ExpressionUUID->"b035769e-73b9-4ef0-8395-d8fa1b839b26"],

Cell[CellGroupData[{

Cell[BoxData["$Manifolds"], "Input",
 CellLabel->"In[56]:=",ExpressionUUID->"5c36480d-b101-471e-90cb-85cdc6a583e1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"M3", ",", "S2"}], "}"}]], "Output",
 CellChangeTimes->{3.728977690706039*^9},
 CellLabel->"Out[56]=",
 CellID->1667145854,ExpressionUUID->"aa376e0c-a3ea-4522-b6ab-259052088c62"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$VBundles"], "Input",
 CellLabel->"In[57]:=",ExpressionUUID->"374d1cc9-99d6-48b1-87e5-0c7b36bc0c82"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]S2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentS2,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728977691314704*^9},
 CellLabel->"Out[57]=",
 CellID->2144759790,ExpressionUUID->"c310b3d2-229c-46b7-9e58-826d511bbc1a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A manifold can be undefined (its properties are lost and the symbol is \
removed). In the process, the associated tangent bundle is also undefined:\
\>", "MathCaption",ExpressionUUID->"7ec5d8f7-916b-4409-a627-cbed7a34d8c3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellLabel->"In[58]:=",ExpressionUUID->"d628234a-4a57-4cf7-8f87-bd66ea255b24"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776926395283`*^9},
 CellLabel->"During evaluation of In[58]:=",
 CellID->390229355,ExpressionUUID->"accd7dfe-6599-404b-ac43-3a3fe8ac4b64"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776926583223`*^9},
 CellLabel->"During evaluation of In[58]:=",
 CellID->169711547,ExpressionUUID->"a7a278e1-17e9-438d-8a92-6e3862171a1a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "M3"}]], "Input",
 CellLabel->"In[59]:=",ExpressionUUID->"d14ddd2c-fbda-4970-bdf8-38eafa357efd"],

Cell[BoxData[
 TemplateBox[{
  "Information", "notfound", 
   "\"Symbol \\!\\(\\*RowBox[{\\\"\\\\\\\"M3\\\\\\\"\\\"}]\\) not found.\"", 
   2, 59, 9, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977693143483*^9},
 CellLabel->"During evaluation of In[59]:=",
 CellID->111960166,ExpressionUUID->"16d090a3-6e27-402c-9f0c-7d3148fa067e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$Manifolds"], "Input",
 CellLabel->"In[60]:=",ExpressionUUID->"1f4fe908-f907-469c-bebb-69cb37f5e101"],

Cell[BoxData[
 RowBox[{"{", "S2", "}"}]], "Output",
 CellChangeTimes->{3.728977693744076*^9},
 CellLabel->"Out[60]=",
 CellID->730204304,ExpressionUUID->"3c5798aa-4f7b-46c8-b82b-58bd42b5b6fa"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "And redefined (note the use of the command ",
 StyleBox["IndexRange",
  FontFamily->"Courier"],
 ", quite handy for long lists of indices):"
}], "MathCaption",
 CellChangeTimes->{{3.494606924743266*^9, 3.4946069429822206`*^9}, {
  3.494606979734871*^9, 
  3.494606985640581*^9}},ExpressionUUID->"2865d1d0-4b87-4877-a573-\
37bb04904229"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", " ", "3", ",", " ", 
   RowBox[{"IndexRange", "[", 
    RowBox[{"a", ",", "h"}], "]"}]}], "]"}]], "Input",
 CellLabel->"In[61]:=",ExpressionUUID->"66c7fb81-9d85-4f47-b94b-23f14098af8e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289776952238607`*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->442594184,ExpressionUUID->"bd1f5700-9263-42b6-909a-c1f719d85a2b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977695243499*^9},
 CellLabel->"During evaluation of In[61]:=",
 CellID->1944580927,ExpressionUUID->"31490e21-67ca-4617-adac-be7e1b06701b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$Manifolds"], "Input",
 CellLabel->"In[62]:=",ExpressionUUID->"d208da62-5962-4155-93ee-f6e9ea12979b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"S2", ",", "M3"}], "}"}]], "Output",
 CellChangeTimes->{3.72897769581172*^9},
 CellLabel->"Out[62]=",
 CellID->863685934,ExpressionUUID->"cbd49fc0-aad4-45da-bc9e-481109a90198"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "It is possible to define 1-dimensional manifolds, with a corresponding \
1-dimensional tangent vbundle. They are treated in exactly the same way as \
other manifolds and vbundles. Other systems, like ",
 StyleBox["Ricci`", "MR"],
 ", only allow a single abstract index (which can be repeated), but that is a \
large internal exception, and ",
 StyleBox["xTensor`", "MR"],
 "does not follow that idea. Instead, we shall declare all indices of the \
same 1-dimensional space to be symmetric. See subsection 4.4 below."
}], "Text",ExpressionUUID->"87931585-5c99-4960-96b8-0cbc8e05ae00"],

Cell["\<\
The dimension of a manifold does not have to be an integer necessarily. It \
can be any expression. It is advisable to use constant-symbols (see 6.5 \
below) as manifold dimensions.\
\>", "Text",ExpressionUUID->"899896a8-1b63-4664-9c40-f6a921808b78"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "1.2. Product manifolds. ",
 StyleBox["V1.1.0 UPDATE",
  FontColor->RGBColor[0, 0, 1]]
}], "Subsection",ExpressionUUID->"cf275d8f-dc19-4950-a603-a03bd3072d50"],

Cell[TextData[{
 "Given two manifolds, it is always possible to define another manifold as \
the Cartesian product of those. The tangent bundle of the product will be \
then constructed as the (external, or disjoint) direct sum of the tangent \
bundles of the submanifolds. There are two ways to approach this problem, and \
we will find this duality repeatedly in all of ",
 StyleBox["xAct`",
  FontFamily->"Courier"],
 ". On the one hand we can declare a new symbolic manifold using a ",
 StyleBox["DefType",
  FontFamily->"Courier"],
 " command, which will prepare everything for us in advance. On the other \
hand we can describe the new manifold using a ",
 StyleBox["CartesianProduct",
  FontFamily->"Courier"],
 " command, whose properties will be computed in real time whenever they are \
needed."
}], "Text",ExpressionUUID->"33450aa3-d5e7-4cc4-ac55-fcb1bd7e05bd"],

Cell[CellGroupData[{

Cell["We define a product manifold as", "MathCaption",ExpressionUUID->"b118a9ae-0d52-4943-91b2-0cc2f95370a6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M5", ",", 
   RowBox[{"{", 
    RowBox[{"M3", ",", "S2"}], "}"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{
    "\[Mu]", ",", " ", "\[Nu]", ",", " ", "\[Lambda]", ",", " ", "\[Sigma]", 
     ",", " ", "\[Eta]", ",", " ", "\[Rho]"}], "}"}]}], "]"}]], "Input",
 CellLabel->"In[63]:=",ExpressionUUID->"e62bf63c-d578-475a-88d2-b90e2f37f7ef"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M5", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M5, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897770371828*^9},
 CellLabel->"During evaluation of In[63]:=",
 CellID->422190744,ExpressionUUID->"50c1bd67-8f9b-4465-96ed-1712efdb71cd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM5", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM5, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977703736231*^9},
 CellLabel->"During evaluation of In[63]:=",
 CellID->473864720,ExpressionUUID->"bafbbe97-0fae-440d-9a24-d7ec43e0b226"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "M5"}]], "Input",
 CellLabel->"In[64]:=",ExpressionUUID->"2b6c1e02-5bab-41b5-ac57-419456dcbcec"],

Cell[CellGroupData[{

Cell["Global`M5", "Print", "PrintUsage",
 CellChangeTimes->{3.728977704745297*^9},
 CellTags->"Info643728981304-9181623",
 CellID->1678239341,ExpressionUUID->"495fb146-4919-4a2e-a6a9-01c1ea7ad2db"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "M5", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"manifold\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DimOfManifold", "[", "M5", "]"}], "^=", "5"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ManifoldQ", "[", "M5", "]"}], "^=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "M5", "]"}], "^=", "\<\"M5\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ServantsOf", "[", "M5", "]"}], "^=", 
         RowBox[{"{", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SplittingsOfManifold", "[", "M5", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", "\[Times]", "S2"}], "}"}]}]},
       {" "},
       {
        RowBox[{"M5", "/:", 
         RowBox[{"SubmanifoldQ", "[", 
          RowBox[{"M5", ",", "M3"}], "]"}], "=", "True"}]},
       {" "},
       {
        RowBox[{"M5", "/:", 
         RowBox[{"SubmanifoldQ", "[", 
          RowBox[{"M5", ",", "S2"}], "]"}], "=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`M5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777047699747`*^9},
 CellTags->"Info643728981304-9181623",
 CellID->1118384489,ExpressionUUID->"696209d9-520d-4605-8a2b-cc1e264a6a62"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
whose tangent bundle is a direct sum of the corresponding subvbundles:\
\>", "MathCaption",ExpressionUUID->"7b872029-8c5d-42c0-b8c4-bbe8ae53acfa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "TangentM5"}]], "Input",
 CellLabel->"In[65]:=",ExpressionUUID->"819acafa-417f-4430-b4a2-8cde5aee1b6e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  StyleBox["\[DoubleStruckCapitalT]M5",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TangentM5,
  Editable->False]], "Print", "PrintUsage",
 CellChangeTimes->{3.7289777059244003`*^9},
 CellTags->"Info653728981305-9181623",
 CellID->1548585086,ExpressionUUID->"c67e0a3a-aab2-4e76-b372-ae1b984fb9b1"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"BaseOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", "M5"}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M5",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM5,
             Editable->False], "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`TangentM5],
          Editable->False], "^=", 
         InterpretationBox[
          StyleBox["\[DoubleStruckCapitalT]M5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`TangentM5,
          Editable->False]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"vbundle\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DimOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", "5"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"IndicesOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
            "\[Mu]", ",", "\[Nu]", ",", "\[Lambda]", ",", "\[Sigma]", ",", 
             "\[Eta]", ",", "\[Rho]"}], "}"}], ",", 
           RowBox[{"{", "}"}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"LastIndex", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\[Rho]", ",", "1"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", "M5"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MetricsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], 
         "^=", "\<\"\[DoubleStruckCapitalT]M5\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SplittingsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], "\[CirclePlus]", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]S2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentS2,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox["\[DoubleStruckCapitalT]M5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`TangentM5,
          Editable->False], "/:", 
         RowBox[{"SubvbundleQ", "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM5,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "]"}], "=", "True"}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox["\[DoubleStruckCapitalT]M5",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`TangentM5,
          Editable->False], "/:", 
         RowBox[{"SubvbundleQ", "[", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M5",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM5,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]S2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentS2,
            Editable->False]}], "]"}], "=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"VBundleQ", "[", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M5",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM5,
           Editable->False], "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`TangentM5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777059798737`*^9},
 CellTags->"Info653728981305-9181623",
 CellID->1715222409,ExpressionUUID->"4f8c5a66-9828-4a8c-a4c2-2c05d8fba923"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The list of product manifolds is given by another global variable:\
\>", "MathCaption",ExpressionUUID->"19b5b905-c580-4aaa-8d99-1fa6e4b1cd0f"],

Cell[CellGroupData[{

Cell[BoxData["$ProductManifolds"], "Input",
 CellLabel->"In[66]:=",ExpressionUUID->"8a20050f-5b08-40ba-ac20-3a2851925ba5"],

Cell[BoxData[
 RowBox[{"{", "M5", "}"}]], "Output",
 CellChangeTimes->{3.728977707212233*^9},
 CellLabel->"Out[66]=",
 CellID->1383227708,ExpressionUUID->"572978e1-b92a-47e1-b5e1-87f76555d6c8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The command ",
 StyleBox["SplitManifold",
  FontFamily->"Courier"],
 " also allows constructing a product manifold, or splitting a previously \
existing manifold."
}], "Text",ExpressionUUID->"3f032dea-7b04-4075-bf98-58359ab38213"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "SplitManifold"}]], "Input",
 CellLabel->"In[67]:=",ExpressionUUID->"a82a7eff-216c-4076-ada8-b0c9515a073e"],

Cell[BoxData["\<\"SplitManifold[M, {m1, ..., mn}] identifies the manifold M \
as the (external) Cartesian product of the manifolds mi. Both the M manifold \
and the mi manifolds are assumed to exist already. The dimension of M must \
coincide with the sum of the dimensions of the mi manifolds.\"\>"], "Print", \
"PrintUsage",
 CellChangeTimes->{3.72897770837747*^9},
 CellTags->"Info673728981308-9181623",
 CellID->1069478814,ExpressionUUID->"f89e4fe2-2986-45ad-b58e-63513099d114"]
}, Open  ]]
}, Open  ]],

Cell["\<\
We can also work directly with an expression representing the product \
manifold.\
\>", "Text",ExpressionUUID->"c5abdda0-498c-44d1-9cee-e7f2ca7d7d83"],

Cell[TextData[{
 StyleBox[ButtonBox["CartesianProduct",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TangentReals.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["[M1, M2, ...]\t\t", "DisplayMath"],
 StyleBox["Cartesian product of the manifolds M1, M2, ...", "Text"]
}], "DefinitionBox",ExpressionUUID->"2b1d3866-bcef-4cf6-9cfc-536e2910f6bb"],

Cell["Cartesian product of manifolds.", "Text",ExpressionUUID->"6bbcb532-2ca1-4ca1-9dc2-dec86556317b"],

Cell[CellGroupData[{

Cell["Cartesian product of our two manifolds :", "MathCaption",ExpressionUUID->"80005056-4171-41b4-9f96-b546be1999e0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"prod", "=", 
  RowBox[{"CartesianProduct", "[", 
   RowBox[{"M3", ",", "S2"}], "]"}]}]], "Input",
 CellLabel->"In[68]:=",ExpressionUUID->"933e0218-cae1-4abb-b4b7-17711468fe81"],

Cell[BoxData[
 RowBox[{"M3", "\[Times]", "S2"}]], "Output",
 CellChangeTimes->{3.7289777153598022`*^9},
 CellLabel->"Out[68]=",
 CellID->839651,ExpressionUUID->"837146aa-d13f-4041-b227-6069f282aa7f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ManifoldQ", "[", "prod", "]"}]], "Input",
 CellLabel->"In[69]:=",ExpressionUUID->"ceee3743-da5b-4f41-a2b8-f82e37c94a6b"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977717134571*^9},
 CellLabel->"Out[69]=",
 CellID->1946601836,ExpressionUUID->"c2853399-a2a9-47a4-be99-edf76576abdb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DimOfManifold", "[", "prod", "]"}]], "Input",
 CellLabel->"In[70]:=",ExpressionUUID->"38297d9f-4f24-4c5a-b946-0dd0e2432b18"],

Cell[BoxData["5"], "Output",
 CellChangeTimes->{3.728977719151785*^9},
 CellLabel->"Out[70]=",
 CellID->1089153226,ExpressionUUID->"b1828f0b-4666-4361-ac7b-9dc736f33796"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Tangent", "[", "prod", "]"}]], "Input",
 CellLabel->"In[71]:=",ExpressionUUID->"bdede592-7995-498d-aa4d-f58c7be3bf5d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox["\[DoubleStruckCapitalT]M3",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TangentM3,
   Editable->False], "\[CirclePlus]", 
  InterpretationBox[
   StyleBox["\[DoubleStruckCapitalT]S2",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TangentS2,
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289777197591763`*^9},
 CellLabel->"Out[71]=",
 CellID->718487895,ExpressionUUID->"ec06b306-d0f0-4217-ad59-0120d74adccb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "FullForm"}]], "Input",
 CellLabel->"In[72]:=",ExpressionUUID->"ad47a5b6-d7e8-40c3-8c03-ca9a42bee6d3"],

Cell[BoxData[
 TagBox[
  StyleBox[
   RowBox[{"CirclePlus", "[", 
    RowBox[{"TangentM3", ",", "TangentS2"}], "]"}],
   ShowSpecialCharacters->False,
   ShowStringCharacters->True,
   NumberMarks->True],
  FullForm]], "Output",
 CellChangeTimes->{3.7289777203193283`*^9},
 CellLabel->"Out[72]//FullForm=",
 CellID->2004106280,ExpressionUUID->"a5049b68-0212-412c-8656-d998e2a2dd32"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"4405d533-bd0c-4e8e-9732-2b0026dda536"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M5", "]"}]], "Input",
 CellLabel->"In[73]:=",ExpressionUUID->"b46e64fa-0fdb-4f84-85e1-2343055f28f4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M5",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM5,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777214320993`*^9},
 CellLabel->"During evaluation of In[73]:=",
 CellID->1972642216,ExpressionUUID->"9461cfbc-d200-4bfd-ba4f-c6d0962844f5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977721452014*^9},
 CellLabel->"During evaluation of In[73]:=",
 CellID->1345181926,ExpressionUUID->"fcd65b7e-ed9b-4d27-b95d-8e2edaebfae4"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "prod", "]"}]], "Input",
 CellLabel->"In[74]:=",ExpressionUUID->"9372b455-0c25-43f5-bda9-2d171282f8c9"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.3. Inner vector bundles", "Subsection",ExpressionUUID->"41df9b57-d4a7-402c-a52f-449095ba56f2"],

Cell[TextData[{
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " allows the definition of vector bundles other than tangent bundles. These \
``inner vbundles\.b4\.b4 can be real or complex, which forces the definition \
of the complex conjugated vbundle. Complex conjugation is controlled by the \
function ",
 StyleBox["Dagger",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"4dcae6d3-f9c5-4c84-a1c5-aa7d6ce12852"],

Cell[TextData[{
 StyleBox[ButtonBox["DefVBundle",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefVBundle.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Define a vbundle or a sum of vbundles\n", "Text"],
 StyleBox[ButtonBox["UndefVBundle",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefVBundle.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Undefine a vbundle\n", "Text"],
 StyleBox[ButtonBox["VBundleQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "VBundleQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check a vbundle", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["$VBundles",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$VBundles.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined vbundles\n", "Text"],
 StyleBox[ButtonBox["$SumVBundles",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$SumVBundles.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["List of defined sum vbundles", "Text"]
}], "DefinitionBox",ExpressionUUID->"e41db040-0eeb-4478-8fa3-25c3c0a85e46"],

Cell["Definition of a vbundle.", "Text",ExpressionUUID->"41d0679c-1a13-485d-b03c-a867b3250886"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a complex inner vbundle with dimension 4 (this is the fiber \
dimension) and base manifold ",
 StyleBox["M3",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"a610520d-1f6c-4913-96fc-32f9ceab89c9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefVBundle", "[", 
  RowBox[{"InnerC", ",", "M3", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", "\[GothicCapitalC]", 
     ",", "\[GothicCapitalD]", ",", "\[GothicCapitalE]", ",", 
     "\[GothicCapitalF]", ",", "\[GothicCapitalG]", ",", 
     "\[GothicCapitalH]"}], "}"}], ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->"In[75]:=",ExpressionUUID->"e1f8cde2-3386-4628-9c85-b2d541380ea4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "InnerC", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`InnerC, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977734125264*^9},
 CellLabel->"During evaluation of In[75]:=",
 CellID->960640895,ExpressionUUID->"24e59c23-5cf5-4355-9fa4-771438c3c4e0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle \"\>", "\[InvisibleSpace]", 
   "InnerC\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Assuming fixed anti-isomorphism between InnerC \
and InnerC\[Dagger]\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "conjugated vbundle ", $CellContext`InnerC\[Dagger], ". ", 
   "Assuming fixed anti-isomorphism between InnerC and InnerC\[Dagger]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977734146121*^9},
 CellLabel->"During evaluation of In[75]:=",
 CellID->709464119,ExpressionUUID->"a8cd3616-0b7a-4b96-a654-72359909efb5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "InnerC"}]], "Input",
 CellLabel->"In[76]:=",ExpressionUUID->"49a3e91a-e823-44b0-bcc3-c0adf147d061"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  StyleBox["InnerC",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`InnerC,
  Editable->False]], "Print", "PrintUsage",
 CellChangeTimes->{3.7289777343278437`*^9},
 CellTags->"Info763728981334-9181623",
 CellID->1134239889,ExpressionUUID->"78f21f23-a24d-4d19-9550-1ea66aeba6ec"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"BaseOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", "M3"}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", 
            InterpretationBox[
             StyleBox["InnerC",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`InnerC,
             Editable->False], "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`InnerC],
          Editable->False], "^=", 
         InterpretationBox[
          StyleBox["InnerC\[Dagger]",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`InnerC\[Dagger],
          Editable->False]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"vbundle\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DimOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", "4"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"IndicesOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
            "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
             "\[GothicCapitalC]", ",", "\[GothicCapitalD]", ",", 
             "\[GothicCapitalE]", ",", "\[GothicCapitalF]", ",", 
             "\[GothicCapitalG]", ",", "\[GothicCapitalH]"}], "}"}], ",", 
           RowBox[{"{", "}"}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"LastIndex", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\[GothicCapitalH]", ",", "1"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MetricsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", "\<\"InnerC\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ServantsOf", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SplittingsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"VBundleQ", "[", 
          InterpretationBox[
           StyleBox["InnerC",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC,
           Editable->False], "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`InnerC],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777344030867`*^9},
 CellTags->"Info763728981334-9181623",
 CellID->2104963651,ExpressionUUID->"86d299a7-dd5d-4f3d-ba67-a9c99a8f3c23"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In the process of definition, the conjugated vbundle ",
 StyleBox["InnerC\[Dagger]",
  FontFamily->"Courier"],
 ", with corresponding conjugated indices, has been created. It is a servant \
of ",
 StyleBox["InnerC",
  FontFamily->"Courier"],
 ", and hence can only be undefined through undefinition of the latter. By \
default, all conjugated symbols are constructed by appending the dagger \
character ",
 StyleBox["\[Dagger]",
  FontFamily->"Courier"],
 ". (This character is stored in the goblal variable ",
 StyleBox["$DaggerCharacter",
  FontFamily->"Courier"],
 ", which can be changed.)"
}], "MathCaption",ExpressionUUID->"51b25888-11e0-4f34-9197-3d72c4553576"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "InnerC\[Dagger]"}]], "Input",
 CellLabel->"In[77]:=",ExpressionUUID->"27f253d2-1c2f-4182-a2a5-a4d6b96ff92c"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  StyleBox["InnerC\[Dagger]",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`InnerC\[Dagger],
  Editable->False]], "Print", "PrintUsage",
 CellChangeTimes->{3.728977734573596*^9},
 CellTags->"Info773728981334-9181623",
 CellID->478481679,ExpressionUUID->"e45eb883-9712-4a14-bb63-71266d89737a"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"BaseOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", "M3"}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", 
            InterpretationBox[
             StyleBox["InnerC\[Dagger]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`InnerC\[Dagger],
             Editable->False], "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`InnerC\[Dagger]],
          Editable->False], "^=", 
         InterpretationBox[
          StyleBox["InnerC",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`InnerC,
          Editable->False]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          
          RowBox[{"\<\"conjugated vbundle\"\>", 
           ",", "\<\"Assuming fixed anti-isomorphism between InnerC and \
InnerC\[Dagger]\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DimOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", "4"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"IndicesOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{
            "\[GothicCapitalA]\[Dagger]", ",", "\[GothicCapitalB]\[Dagger]", 
             ",", "\[GothicCapitalC]\[Dagger]", ",", 
             "\[GothicCapitalD]\[Dagger]", ",", "\[GothicCapitalE]\[Dagger]", 
             ",", "\[GothicCapitalF]\[Dagger]", ",", 
             "\[GothicCapitalG]\[Dagger]", ",", 
             "\[GothicCapitalH]\[Dagger]"}], "}"}], ",", 
           RowBox[{"{", "}"}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"LastIndex", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\[GothicCapitalH]\[Dagger]", ",", "1"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         InterpretationBox[
          StyleBox["InnerC",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`InnerC,
          Editable->False]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MetricsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", "\<\"InnerC\[Dagger]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SplittingsOfVBundle", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"VBundleQ", "[", 
          InterpretationBox[
           StyleBox["InnerC\[Dagger]",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`InnerC\[Dagger],
           Editable->False], "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`InnerC\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977734631353*^9},
 CellTags->"Info773728981334-9181623",
 CellID->1662804585,ExpressionUUID->"1b43b27e-5ef8-4e34-ad3f-b46f987e1e90"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We have already defined four different vector bundles:", "MathCaption",ExpressionUUID->"281293b9-d69c-4c8b-8a41-b10a3c4440fb"],

Cell[CellGroupData[{

Cell[BoxData["$VBundles"], "Input",
 CellLabel->"In[78]:=",ExpressionUUID->"a4e76137-ed7d-4a74-b496-6490cc2cc946"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]S2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentS2,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["InnerC",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["InnerC\[Dagger]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC\[Dagger],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728977734778001*^9},
 CellLabel->"Out[78]=",
 CellID->1551961684,ExpressionUUID->"c3a2e3f5-8e19-4730-a516-9a28e96094d1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There is also a function ",
 StyleBox["SplitVBundle",
  FontFamily->"Courier"],
 ", parallel to ",
 StyleBox["SplitManifold",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"fa959902-aa79-407c-940f-2285556b445f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "SplitVBundle"}]], "Input",
 CellLabel->"In[79]:=",ExpressionUUID->"3c96b45f-e102-4392-bf32-84e64578de9c"],

Cell[BoxData["\<\"SplitVBundle[V, {v1, ..., vn}] identifies the vbundle V as \
the direct sum of the vbundles vi. Both the V vbundle and the vi vbundles are \
assumed to exist already. The dimension of V must coincide with the sum of \
the dimensions of the vi vbundles.\"\>"], "Print", "PrintUsage",
 CellChangeTimes->{3.728977734943226*^9},
 CellTags->"Info793728981334-9181623",
 CellID->483646526,ExpressionUUID->"b624df61-586c-41cd-84c6-be2034fb0f05"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["1.4. Parameters", "Subsection",ExpressionUUID->"e88af057-5080-4dba-a53c-e91d18080d3c"],

Cell[TextData[{
 "In ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " the indexed objects can be fields on manifolds and/or functions of real \
parameters (for example proper time)."
}], "Text",ExpressionUUID->"ecadba5a-a0cd-4981-ae56-1e080a567e56"],

Cell[TextData[{
 StyleBox[ButtonBox["DefParameter",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefParameter.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Define a real parameter\n", "Text"],
 StyleBox[ButtonBox["UndefParameter",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefParameter.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Undefine a parameter\n", "Text"],
 StyleBox[ButtonBox["ParameterQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ParameterQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check a real parameter", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["$Parameters",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$Parameters.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined parameters", "Text"]
}], "DefinitionBox",ExpressionUUID->"7b133ce1-4008-486c-bf5b-9db872b0a2db"],

Cell["Definition of a parameter", "Text",ExpressionUUID->"af2f997a-c293-4dc2-abb5-f953b1d04d9b"],

Cell[CellGroupData[{

Cell[TextData[{
 "We define a real parameter that will be used below to compute parametric \
derivatives (see ",
 StyleBox["ParamD",
  FontFamily->"Courier"],
 " in subsection 6.6)."
}], "MathCaption",ExpressionUUID->"703ba00d-9850-441c-a4d1-fa1478287f89"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefParameter", "[", "time", "]"}]], "Input",
 CellLabel->"In[80]:=",ExpressionUUID->"17213dac-b2f6-46c4-bafe-dc0ad267bac7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefParameter", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parameter \"\>", "\[InvisibleSpace]", "time", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefParameter, ": Defining ", 
   "parameter ", $CellContext`time, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777427296352`*^9},
 CellLabel->"During evaluation of In[80]:=",
 CellID->229219215,ExpressionUUID->"af4bb84c-e5cf-4949-8f6d-962879c7ad1d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "time"}]], "Input",
 CellLabel->"In[81]:=",ExpressionUUID->"ad324b11-367e-4fda-bdea-1eaf4530fe4d"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  StyleBox["time",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`time,
  Editable->False]], "Print", "PrintUsage",
 CellChangeTimes->{3.728977742871666*^9},
 CellTags->"Info813728981342-9181623",
 CellID->2049167434,ExpressionUUID->"c8622f06-4d6e-4201-bb88-bb26ce55c8ab"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"DefInfo", "[", 
          InterpretationBox[
           StyleBox["time",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`time,
           Editable->False], "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"parameter\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ParameterQ", "[", 
          InterpretationBox[
           StyleBox["time",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`time,
           Editable->False], "]"}], "^=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", 
          InterpretationBox[
           StyleBox["time",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`time,
           Editable->False], "]"}], "^=", "\<\"time\"\>"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`time],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977742895159*^9},
 CellTags->"Info813728981342-9181623",
 CellID->1161816299,ExpressionUUID->"3f1bc420-c985-4a39-be0e-84a2da604eb8"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Manifolds and parameters are the only possible \"dependencies\" of the \
objects in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ". We shall later explain this in more detail."
}], "Text",ExpressionUUID->"b54369c3-fed7-4946-bce8-8e08a33b3a19"],

Cell["\<\
An important question is the relation between fields on 1d manifolds and \
functions of a parameter. In decompositions of ADM-type it would be nice to \
have a way of going from a 4d structure to a 3+1 structure and then from here \
to a 3d structure plus a time parameter. I still don't know how to program \
this idea. There are many things to change in a nontrivial way: tensor \
fields, covariant derivatives, metrics, etc. Please, let me know if you have \
suggestions.\
\>", "Text",
 CellChangeTimes->{{3.4946070715633287`*^9, 
  3.4946070970365133`*^9}},ExpressionUUID->"869cb9c3-a5d7-4433-a5c8-\
0af58d5e7af5"]
}, Closed]],

Cell[CellGroupData[{

Cell["1.5. Vector spaces: bundles over a point", "Subsection",ExpressionUUID->"f4a065ec-8746-4f74-990d-0ab375474513"],

Cell[TextData[{
 "It is possible to work  ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " with a single vector space, considered as a bundle over a manifold that \
happens to be a point."
}], "Text",ExpressionUUID->"5bf2dfdc-8a1a-4613-9ba4-36ee49938748"],

Cell[CellGroupData[{

Cell["\<\
Define a zero-dimensional manifold, a point. We do not need to specify \
indices because there is no tangent space:\
\>", "MathCaption",ExpressionUUID->"c7cf903b-a7b6-4bfd-a118-2acc357d11a5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"P", ",", "0", ",", 
   RowBox[{"{", "}"}]}], "]"}]], "Input",
 CellLabel->"In[82]:=",ExpressionUUID->"24f3f38a-1ee0-4ea8-8948-eb7857b82456"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "P", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", "manifold ", $CellContext`P,
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777467353354`*^9},
 CellLabel->"During evaluation of In[82]:=",
 CellID->2012913861,ExpressionUUID->"877caff4-d563-41eb-8204-14522d3e65f3"],

Cell[BoxData["\<\"** DefManifold:  Tangent bundle does not exist for a 0-dim \
manifold\"\>"], "Print",
 CellChangeTimes->{3.7289777467546864`*^9},
 CellLabel->"During evaluation of In[82]:=",
 CellID->1733370636,ExpressionUUID->"45226275-4177-442a-97af-57c5f2c2008a"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we can define a vector bundle over this point. It has a single fiber, \
which is a vector space of the required dimension:\
\>", "MathCaption",ExpressionUUID->"8071bcb4-9e39-4e02-9c88-eaff93fed879"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefVBundle", "[", 
  RowBox[{"V", ",", "P", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"\[GothicA]", ",", "\[GothicB]", ",", "\[GothicC]"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->"In[83]:=",ExpressionUUID->"36d36bdc-c191-40c9-a208-1e01bfb4120e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "V", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", "vbundle ", $CellContext`V, 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977746918551*^9},
 CellLabel->"During evaluation of In[83]:=",
 CellID->2144607697,ExpressionUUID->"d3397308-8b78-4841-bb77-be581f7cfcae"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
And now we can define vectors and tensors on this space, with no dependencies:\
\>", "MathCaption",ExpressionUUID->"b85a38d6-fa67-46e7-b45d-96783d8b665a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"X", "[", "\[GothicA]", "]"}], ",", 
   RowBox[{"{", "}"}]}], "]"}]], "Input",
 CellLabel->"In[84]:=",ExpressionUUID->"b069a155-d510-4b25-bc89-3e5500b1c931"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"X", "[", "\[GothicA]", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`X[$CellContext`\[GothicA]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977747050667*^9},
 CellLabel->"During evaluation of In[84]:=",
 CellID->1056032231,ExpressionUUID->"20f1dcde-4f43-4ee5-bcdf-6e76b39e110f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine all objects:", "MathCaption",ExpressionUUID->"e293c462-6e57-4668-b017-4a2de63c24dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "X", "]"}]], "Input",
 CellLabel->"In[85]:=",ExpressionUUID->"4747e910-5168-427c-9784-59dd97f21fc5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "X"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`X],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977747163076*^9},
 CellLabel->"During evaluation of In[85]:=",
 CellID->1928847609,ExpressionUUID->"e9753a4c-e1bb-463f-bcdf-0d23b383f27a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefVBundle", "[", "V", "]"}]], "Input",
 CellLabel->"In[86]:=",ExpressionUUID->"b2001d89-3fa4-49a6-bc39-dfb1098e92c8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["V",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`V,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977747311757*^9},
 CellLabel->"During evaluation of In[86]:=",
 CellID->1876112998,ExpressionUUID->"da90fa35-b075-4eab-ae0b-ec9c68303200"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "P", "]"}]], "Input",
 CellLabel->"In[87]:=",ExpressionUUID->"b02d2530-6567-4f11-8dd4-445ef9080e93"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "P"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`P],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977747428424*^9},
 CellLabel->"During evaluation of In[87]:=",
 CellID->328768485,ExpressionUUID->"a03abb13-cb34-493a-9369-4f47f122a6a7"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["2. Indices", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1215012216,ExpressionUUID->"e50a4bbd-8e74-4fce-82b6-11d15ef76e1d"],

Cell[CellGroupData[{

Cell["2.0. Important comments", "Subsection",ExpressionUUID->"f3e6e9da-0f4e-40b2-b1ed-c50ec9fe6a74"],

Cell[TextData[{
 "1) Indices are the key objects in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", and the source of the generality of the system. They are carriers of \
contextual information, much like the decimal digits in numerical \
computations, or the letters in a text. An index by itself has little \
information (only an association to a vector space), but a group of indices \
together can encode a lot of information (rank, symmetries, metrics, bases, \
etc). In general, an index is anything which can be placed at an index-slot: \
that is, we shall later define four places in which we can put indices and \
anything which we can meaningfully put there will be called an index."
}], "Text",ExpressionUUID->"e6ece487-8eb2-43a6-ace3-ff0e4f43fa3c"],

Cell[TextData[{
 "2) Starting version 0.9 of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", indices are associated to vbundles, and not to manifolds, as was the case \
in previous versions. This allows us to work with vbundles other than tangent \
bundles to manifolds."
}], "Text",ExpressionUUID->"d4d91dd1-913d-4b76-a80e-e65db9625f77"],

Cell["\<\
3) We need a \"language\" to talk about indices, and though we usually \
understand what a dummy index or a covariant index is, such a language is not \
rigorously defined anywhere. I have invented my own language for this. \
Please, suggest changes if you find inconsistencies, gaps or bad choices.\
\>", "Text",ExpressionUUID->"0499bb65-c6b2-43a8-bc89-8dece0141f98"],

Cell["\<\
4) Taken from the Internet:
\tThe plural of \"index\" for books, journals and newspapers is \"indexes\"; \
the plural of \"index\" for graphs and benchmarks for economics for example, \
is \"indices\"; but the plural of \"appendix\" is always \"appendices\".
I think the \"index\" of a tensor must follow the rule for the second group, \
and hence its plural will be \"indices\".\
\>", "Text",ExpressionUUID->"ba883f62-4a3d-4909-9d34-4850e872ed9d"]
}, Closed]],

Cell[CellGroupData[{

Cell["2.1. Properties of indices: type, character, state", "Subsection",ExpressionUUID->"1d6d10f7-5284-4d78-8010-4890039bf855"],

Cell[TextData[{
 "1) There are five known types of index in ",
 StyleBox["xTensor` ",
  FontFamily->"Courier"],
 "(other types could be added, but there is no user-defined way to do this):\n\
\t- Abstract index (",
 StyleBox[ButtonBox["AIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "AIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "). See subsection 2.2.\n\t- Basis index (",
 StyleBox[ButtonBox["BIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "BIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "), associated to some basis (coordinated or not) of vector fields. \
Described in ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 ".\n\t- Component index (",
 StyleBox[ButtonBox["CIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "CIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "), also associated to some basis. Described in ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 ".\n\t- Directional index (",
 StyleBox[ButtonBox["DIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "). See subsection 3.7.\n\t- Label index (",
 StyleBox[ButtonBox["LIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "). See subsection 3.8.\nCollectively we refer to all five classes as \
\"generalized indices\" (",
 StyleBox["GIndex",
  FontFamily->"Courier"],
 "). Additionally, we can have patterns (",
 StyleBox[ButtonBox["PIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 ") for g-indices at those positions where g-indices are expected, but \
patterns are not considered as g-indices. In the following \"indices\" will \
mean \"g-indices\" unless we specify the type."
}], "Text",ExpressionUUID->"78c412bc-f1ee-408f-a244-2126222cca9b"],

Cell[TextData[{
 "2) All indices have a \"character\", which can be either covariant (",
 StyleBox[ButtonBox["Down",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Down.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "-index) or contravariant (",
 StyleBox[ButtonBox["Up",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Up.nb", CharacterEncoding -> "iso8859-1"],
     None}],
  FontFamily->"Courier"],
 "-index)."
}], "Text",ExpressionUUID->"7bcc6a39-f7e2-45a5-9117-6844b149e1cb"],

Cell[TextData[{
 "3) Some indices can be used to represent contractions, following the \
Einstein convention: two repeated indices in the same tensor or tensor \
product, but each having a different character.  We call them \
\"contractible\" or \"Einstein\" indices (",
 StyleBox["EIndex",
  FontFamily->"Courier"],
 ") and currently only abstract and basis indices are allowed to be \
contractible. (Note that having contractible indices with the same character \
is a syntactic error in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ".) E-indices have a \"state\": they can be ",
 StyleBox[ButtonBox["Free",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Free.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 " or contracted (aka ",
 StyleBox[ButtonBox["Dummy",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Dummy.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "). For completeness, the state of components, directions and labels is said \
to be always \"",
 StyleBox[ButtonBox["Blocked",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Blocked.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 "\", what in particular means that they can be repeated (i.e. truly \
repeated, not simply staggered)."
}], "Text",ExpressionUUID->"c9dc8329-8018-4ea8-a951-b43f659228d4"],

Cell[TextData[{
 "4) Basis and component indices are known by part of the routines of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " (for example the formatting routines, the index selectors, the \
canonicalizer or the constructors of rules), but used in the twin package ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 ". From now on we shall not consider those two types of indices. See ",
 StyleBox["xCobaDoc.nb",
  FontFamily->"Courier"],
 " for further information.\n5) There is a further property of an index, only \
used internally: its metric-state, saying whether an index can be raised or \
lowered using a given metric (not whether it has been or not actually raised \
already). This will be important in section 7.3."
}], "Text",ExpressionUUID->"5979c260-3a9b-46c4-9832-5a7d8bb9dd1e"]
}, Closed]],

Cell[CellGroupData[{

Cell["2.2. Covariant and contravariant abstract indices", "Subsection",ExpressionUUID->"05082b76-880a-4928-83c2-67d4901eca40"],

Cell[TextData[{
 "Ab",
 "stract ind",
 "ices in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " are valid atomic symbols without numeric value (excluding therefore ",
 StyleBox["E",
  FontFamily->"Courier"],
 " and ",
 StyleBox["I",
  FontFamily->"Courier"],
 ") and with no special output (excluding symbols like ",
 StyleBox["Space",
  FontFamily->"Courier"],
 ", ",
 StyleBox["Tab",
  FontFamily->"Courier"],
 ", etc.) One of the most important decisions when working with tensors is \
the notation for covariant and contravariant indices. Given the binary \
character of the problem there is only one natural simple choice: the use of \
signs ",
 StyleBox["+",
  FontFamily->"Courier"],
 " and ",
 StyleBox["-",
  FontFamily->"Courier"],
 " . We shall denote contravariant indices as ",
 StyleBox["+a",
  FontFamily->"Courier"],
 " or ",
 StyleBox["a",
  FontFamily->"Courier"],
 ", and covariant indices as ",
 StyleBox["-a",
  FontFamily->"Courier"],
 ". This choice is very simple, but not particularly convenient when doing \
pattern matching, due to the asymmetry of ",
 StyleBox["a",
  FontFamily->"Courier"],
 " being a symbol but ",
 StyleBox["-a",
  FontFamily->"Courier"],
 " being a composite expression."
}], "Text",ExpressionUUID->"0384c176-8639-4d9d-bcff-4069a4f44e53"],

Cell[TextData[{
 StyleBox[ButtonBox["AIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "AIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["Validate an ab", "Text"],
 "stract ind",
 StyleBox["ex", "Text"],
 StyleBox["\nVB`Q\t\t\t\t\t\t", "DisplayMath"],
 StyleBox["Check that an abstract up-index belongs to vbundle", "Text"],
 StyleBox[" VB\nVB`pmQ\t\t\t\t\t", "DisplayMath"],
 StyleBox["Check that an abstract (up- or down-) index belongs to vbundle", 
  "Text"],
 StyleBox[" VB", "DisplayMath"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.7289777607842493`*^9},ExpressionUUID->"4360fc4d-6fca-4955-90c0-\
3a73c56ba2de"],

Cell["Validation of indices.", "Text",ExpressionUUID->"965569c5-aac7-4c15-9732-9ccba76719fc"],

Cell[CellGroupData[{

Cell[TextData[{
 "We can check whether a symbol has been registered as an abstract index \
using the function ",
 StyleBox["AIndexQ",
  FontFamily->"Courier"]
}], "MathCaption",ExpressionUUID->"e7ce011f-252a-484a-b040-bb1673e24c97"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", "A", "]"}]], "Input",
 CellLabel->"In[88]:=",ExpressionUUID->"80c5e0ea-d526-4a38-8257-b213c41e432d"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977766549274*^9},
 CellLabel->"Out[88]=",
 CellID->1077251231,ExpressionUUID->"397657ef-c0bb-4fb7-98d5-84c48bf79008"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", 
  RowBox[{"-", "b"}], "]"}]], "Input",
 CellLabel->"In[89]:=",ExpressionUUID->"71a605ae-4f3e-4c73-8f9e-3d25178f3cc9"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.72897776663459*^9},
 CellLabel->"Out[89]=",
 CellID->2074295207,ExpressionUUID->"0e7c5339-1ef9-43bd-a44e-ef433d673c19"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", "q", "]"}]], "Input",
 CellLabel->"In[90]:=",ExpressionUUID->"721b5a43-7732-4161-817e-96026db6d291"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.728977766723879*^9},
 CellLabel->"Out[90]=",
 CellID->430018651,ExpressionUUID->"0d5eb9fd-25d8-47b8-bda6-1928e31800df"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "q", "]"}]], "Input",
 CellLabel->"In[91]:=",ExpressionUUID->"9546b9f7-b9d9-4966-9d1f-79a7adf1f4f4"]
}, Open  ]],

Cell[CellGroupData[{

Cell["In particular, we can specify the vbundle it must belong to:", \
"MathCaption",ExpressionUUID->"81bd29c2-5f25-41da-8ecd-d8a5ab274c11"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", 
  RowBox[{"A", ",", "TangentS2"}], "]"}]], "Input",
 CellLabel->"In[92]:=",ExpressionUUID->"3fdf3655-d045-4c16-9085-809d392e337b"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977766881269*^9},
 CellLabel->"Out[92]=",
 CellID->1040995202,ExpressionUUID->"44b8a245-1440-459f-9abe-3ae9e96fa183"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", 
  RowBox[{"\[GothicCapitalA]", ",", "TangentM3"}], "]"}]], "Input",
 CellLabel->"In[93]:=",ExpressionUUID->"b52ffada-af64-4caf-bcee-7d5d4b44afa4"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.7289777670221148`*^9},
 CellLabel->"Out[93]=",
 CellID->457087905,ExpressionUUID->"3f4c39d4-08f4-4a80-b461-b85d542f94df"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", 
  RowBox[{"\[GothicCapitalA]", ",", "InnerC"}], "]"}]], "Input",
 CellLabel->"In[94]:=",ExpressionUUID->"4523e501-bcf2-44ca-84c5-8d130fd60953"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977767390678*^9},
 CellLabel->"Out[94]=",
 CellID->727414016,ExpressionUUID->"dfdb62b5-4db2-4525-8d5e-bf8d2599b854"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", 
  RowBox[{"\[GothicCapitalA]\[Dagger]", ",", "InnerC"}], "]"}]], "Input",
 CellLabel->"In[95]:=",ExpressionUUID->"e6831a30-14c5-4617-98bf-db129f2a99bb"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.72897776769016*^9},
 CellLabel->"Out[95]=",
 CellID->22964541,ExpressionUUID->"7803c63a-3420-4271-b6e9-971bb2fcce3b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AIndexQ", "[", 
  RowBox[{"\[GothicCapitalA]\[Dagger]", ",", "InnerC\[Dagger]"}], 
  "]"}]], "Input",
 CellLabel->"In[96]:=",ExpressionUUID->"5a1dba44-ace2-434d-9f99-d5f162116f4a"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977767919726*^9},
 CellLabel->"Out[96]=",
 CellID->1565467774,ExpressionUUID->"396e0ea3-8712-4e7b-bce8-654dec092a3e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Each vbundle has two special functions that select abstract indices on that \
manifold. For example for the vbundle ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 " they are ",
 StyleBox["TangentM3`Q",
  FontFamily->"Courier"],
 " and ",
 StyleBox["TangentM3`pmQ",
  FontFamily->"Courier"],
 ". The former only accepts contravariant indices; the latter accepts both \
characters. The use of context notation (the ` ) is purely historical : \
before using upvalues to store information I tried to use contexts, but it is \
less convenient."
}], "MathCaption",ExpressionUUID->"b8055217-df79-4c70-bd9b-bb7fe710db97"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentM3`Q", "[", "a", "]"}]], "Input",
 CellLabel->"In[97]:=",ExpressionUUID->"dad30ed6-0200-4030-8551-9bb366ebb246"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7289777680481977`*^9},
 CellLabel->"Out[97]=",
 CellID->1994718221,ExpressionUUID->"fd6f2a5d-fd49-48ef-aba8-e713f1cdee68"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentM3`Q", "[", 
  RowBox[{"-", "a"}], "]"}]], "Input",
 CellLabel->"In[98]:=",ExpressionUUID->"454c5e63-6461-496f-b8d3-f13466121428"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.7289777682948627`*^9},
 CellLabel->"Out[98]=",
 CellID->861927030,ExpressionUUID->"0529e84c-04d4-4088-9a8e-03f6e590dbe5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentM3`Q", "[", "A", "]"}]], "Input",
 CellLabel->"In[99]:=",ExpressionUUID->"a2fc3080-0eb3-4b50-a88e-afe72570ae57"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.7289777684142857`*^9},
 CellLabel->"Out[99]=",
 CellID->1630394195,ExpressionUUID->"2623b412-737b-4874-9c53-ccb18932f5d6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentM3`pmQ", "[", "a", "]"}]], "Input",
 CellLabel->
  "In[100]:=",ExpressionUUID->"a0dcff7c-244e-4028-9ca0-29524d13b40c"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977768521274*^9},
 CellLabel->"Out[100]=",
 CellID->143899645,ExpressionUUID->"e0fb5b01-4b2e-4d13-8f42-45af8107588b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentM3`pmQ", "[", 
  RowBox[{"-", "a"}], "]"}]], "Input",
 CellLabel->
  "In[101]:=",ExpressionUUID->"812c0777-64e1-4c21-ac43-2d4636fce7e3"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7289777687744207`*^9},
 CellLabel->"Out[101]=",
 CellID->454746993,ExpressionUUID->"55567e55-84f0-4f2a-aaf2-a057c5fdedca"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["2.3. Get (new) indices", "Subsection",ExpressionUUID->"8c80a1fc-ae59-4563-9ed8-74a9291dbd67"],

Cell[TextData[{
 "Sometimes we need to use many indices. There are two ways to do it. We can \
register more symbols using ",
 StyleBox["AddIndices",
  FontFamily->"Courier"],
 " or we can ask ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " to generate new indices for a vbundle using ",
 StyleBox["NewIndexIn",
  FontFamily->"Courier"],
 " (it uses a combination of the last registered symbol and integer numbers). \
User-defined indices and computer-defined indices are kept in two different \
lists."
}], "Text",ExpressionUUID->"abf05f7d-06d5-41e5-84d5-fb9be5879686"],

Cell[TextData[{
 StyleBox[ButtonBox["AddIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "AddIndices.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Add user-defined indices to a vbundle", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["RemoveIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "RemoveIndices.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Remove user-defined indices from a vbundle\n", "Text"],
 StyleBox[ButtonBox["NewIndexIn",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "NewIndexIn.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Generate a computer-defined index", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["GetIndicesOfVBundle",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "GetIndicesOfVBundle.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Give ", "Text"],
 StyleBox["n", "Text",
  FontFamily->"Courier"],
 StyleBox[" indices", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.72897777804665*^9},ExpressionUUID->"c3266cf3-6f91-4b2f-bc1c-b8156c655034"],

Cell["Functions that return or change the available indices.", "Text",ExpressionUUID->"afe27f77-65c3-4bd6-981b-82149b8b6a85"],

Cell[CellGroupData[{

Cell["\<\
Currently there are only user-defined indices. We add some more:\
\>", "MathCaption",ExpressionUUID->"d5be2c62-63d1-4d6d-a3c0-9998bb7664e7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[102]:=",ExpressionUUID->"55710039-c479-47a7-b899-8d71ae9e59e9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", 
     "h"}], "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977780854219*^9},
 CellLabel->"Out[102]=",
 CellID->391059037,ExpressionUUID->"077638e7-1096-4d3c-b211-46856bc1d771"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"AddIndices", "[", 
  RowBox[{"TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"i", ",", "j", ",", "k", ",", "l"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[103]:=",ExpressionUUID->"604a62e1-604b-408e-b03c-cf0fbe723691"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[104]:=",ExpressionUUID->"c452fb46-83d3-45b7-a523-573fb51ae540"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", "h",
      ",", "i", ",", "j", ",", "k", ",", "l"}], "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977781043365*^9},
 CellLabel->"Out[104]=",
 CellID->1229275039,ExpressionUUID->"865fcbc3-138c-43e9-b678-60093bcab966"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can also remove indices, but this is a very dangerous operation because \
some previous expressions could get corrupted if they contain removed \
indices. There is no built-in command to check whether this is happening or \
not.\
\>", "MathCaption",ExpressionUUID->"fe691718-bca7-490d-810f-0e942c00e537"],

Cell[BoxData[
 RowBox[{"RemoveIndices", "[", 
  RowBox[{"TangentM3", ",", 
   RowBox[{"{", 
    RowBox[{"j", ",", "i", ",", "l", ",", "k"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[105]:=",ExpressionUUID->"82c84a3a-3c9b-4bc5-ac1d-227eb375a76d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[106]:=",ExpressionUUID->"f875e084-0440-429b-870c-7f0247c6e44f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", 
     "h"}], "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977781234248*^9},
 CellLabel->"Out[106]=",
 CellID->1632896333,ExpressionUUID->"8b2ccf57-90e0-4b2c-983a-e04cbc5807bc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " generates new indices using ",
 StyleBox["NewIndexIn",
  FontFamily->"Courier"]
}], "MathCaption",ExpressionUUID->"4baf3557-ec02-446c-bbdf-962e9c3128bf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NewIndexIn", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[107]:=",ExpressionUUID->"39cc2e40-646d-4538-8615-073d65df5cf3"],

Cell[BoxData["h1"], "Output",
 CellChangeTimes->{3.728977781475079*^9},
 CellLabel->"Out[107]=",
 CellID->1395143718,ExpressionUUID->"4dc095b7-56c6-4548-8b0e-57086340a0c6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[108]:=",ExpressionUUID->"82886089-dd34-47f4-815f-ea7b75deb5f5"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", 
     "h"}], "}"}], ",", 
   RowBox[{"{", "h1", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977781742084*^9},
 CellLabel->"Out[108]=",
 CellID->1954010480,ExpressionUUID->"9aba1e9d-6933-42b7-aa8a-ff59927dcc6c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The user-function to get any number of indices is ",
 StyleBox["GetIndicesOfVBundle",
  FontFamily->"Courier"],
 ". Note that the last argument is a list of symbols that we do not want to \
get:"
}], "MathCaption",ExpressionUUID->"a0ea6373-2bad-4d25-bee0-9053c79469c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GetIndicesOfVBundle", "[", 
  RowBox[{"TangentM3", ",", "14", ",", 
   RowBox[{"{", 
    RowBox[{"d", ",", "e", ",", "h3"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[109]:=",ExpressionUUID->"438e6e57-0cde-47b7-a514-ca941c825f25"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "a", ",", "b", ",", "c", ",", "f", ",", "g", ",", "h", ",", "h1", ",", "h2",
    ",", "h4", ",", "h5", ",", "h6", ",", "h7", ",", "h8", ",", "h9"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.72897778201334*^9},
 CellLabel->"Out[109]=",
 CellID->1734207436,ExpressionUUID->"8211a7e3-a7a6-4cc6-9ef0-1124f422f01a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[110]:=",ExpressionUUID->"0a2288d1-079a-4af0-847c-7c30de238ef1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", 
     "h"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "h1", ",", "h2", ",", "h3", ",", "h4", ",", "h5", ",", "h6", ",", "h7", 
     ",", "h8", ",", "h9"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289777821868963`*^9},
 CellLabel->"Out[110]=",
 CellID->1176319768,ExpressionUUID->"1bdc9522-c8a1-464b-8788-d3f155c0f8b7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The indices of complex conjugated vbundles are paired, and therefore if we \
generate an index in one of them, automatically the complex conjugated index \
is generated in the other one:\
\>", "MathCaption",ExpressionUUID->"45b89367-4d86-4a1d-a75f-ea3654f77f5c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NewIndexIn", "[", "InnerC", "]"}]], "Input",
 CellLabel->
  "In[111]:=",ExpressionUUID->"ec29ef21-2bac-478f-b4a9-e1751f168b64"],

Cell[BoxData["\[GothicCapitalH]1"], "Output",
 CellChangeTimes->{3.728977782609326*^9},
 CellLabel->"Out[111]=",
 CellID->671886494,ExpressionUUID->"ac8f83d9-0c91-49ba-b023-d82be5b0f99e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "InnerC", "]"}]], "Input",
 CellLabel->
  "In[112]:=",ExpressionUUID->"30d30216-18dc-4d89-90ef-f0f90582887c"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", "\[GothicCapitalC]", 
     ",", "\[GothicCapitalD]", ",", "\[GothicCapitalE]", ",", 
     "\[GothicCapitalF]", ",", "\[GothicCapitalG]", ",", 
     "\[GothicCapitalH]"}], "}"}], ",", 
   RowBox[{"{", "\[GothicCapitalH]1", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977782769821*^9},
 CellLabel->"Out[112]=",
 CellID->927288732,ExpressionUUID->"1d7446f7-5815-4f36-8b31-4bc84990cb50"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "InnerC\[Dagger]", "]"}]], "Input",
 CellLabel->
  "In[113]:=",ExpressionUUID->"325aebc8-1c3b-4e25-a3ae-62faf917301a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "\[GothicCapitalA]\[Dagger]", ",", "\[GothicCapitalB]\[Dagger]", ",", 
     "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]", ",", 
     "\[GothicCapitalE]\[Dagger]", ",", "\[GothicCapitalF]\[Dagger]", ",", 
     "\[GothicCapitalG]\[Dagger]", ",", "\[GothicCapitalH]\[Dagger]"}], "}"}],
    ",", 
   RowBox[{"{", "\[GothicCapitalH]1\[Dagger]", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977782932928*^9},
 CellLabel->"Out[113]=",
 CellID->652892445,ExpressionUUID->"c9fa1af9-cab6-4c31-b82a-fbc0f0914464"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GetIndicesOfVBundle", "[", 
  RowBox[{"InnerC\[Dagger]", ",", "10"}], "]"}]], "Input",
 CellLabel->
  "In[114]:=",ExpressionUUID->"8de04d3a-5ca4-47ae-8ef3-87ed5685e24e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "\[GothicCapitalA]\[Dagger]", ",", "\[GothicCapitalB]\[Dagger]", ",", 
   "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]", ",", 
   "\[GothicCapitalE]\[Dagger]", ",", "\[GothicCapitalF]\[Dagger]", ",", 
   "\[GothicCapitalG]\[Dagger]", ",", "\[GothicCapitalH]1\[Dagger]", ",", 
   "\[GothicCapitalH]\[Dagger]", ",", "\[GothicCapitalH]2"}], "}"}]], "Output",\

 CellChangeTimes->{3.728977783054936*^9},
 CellLabel->"Out[114]=",
 CellID->1669370592,ExpressionUUID->"916ea33a-4ee0-48db-b622-b0b015a67b2d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "InnerC", "]"}]], "Input",
 CellLabel->
  "In[115]:=",ExpressionUUID->"909ff47a-feb3-49a9-9fcb-f832d17797de"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", "\[GothicCapitalC]", 
     ",", "\[GothicCapitalD]", ",", "\[GothicCapitalE]", ",", 
     "\[GothicCapitalF]", ",", "\[GothicCapitalG]", ",", 
     "\[GothicCapitalH]"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"\[GothicCapitalH]1", ",", "\[GothicCapitalH]2"}], "}"}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.72897778315714*^9},
 CellLabel->"Out[115]=",
 CellID->237691689,ExpressionUUID->"ca4b4e19-05c6-407d-8e66-253b4e4167b5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Note the information associated to an abstract index:", "MathCaption",ExpressionUUID->"684bdef6-822b-477a-8bf3-54d8fbcc0ec1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "\[GothicCapitalA]"}]], "Input",
 CellLabel->
  "In[116]:=",ExpressionUUID->"ca31c5f4-8ea5-457c-96ef-72abed459a54"],

Cell[CellGroupData[{

Cell["Global`\[GothicCapitalA]", "Print", "PrintUsage",
 CellChangeTimes->{3.728977783291326*^9},
 CellTags->"Info1163728981383-9181623",
 CellID->564633603,ExpressionUUID->"58195580-63ce-46a6-9406-193b5711864f"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"AbstractIndexQ", "[", "\[GothicCapitalA]", "]"}], "^=", 
         "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DaggerIndex", "[", "\[GothicCapitalA]", "]"}], "^=", 
         "\[GothicCapitalA]\[Dagger]"}]},
       {" "},
       {
        RowBox[{
         RowBox[{
         "xAct`xTensor`Private`NoDollar", "[", "\[GothicCapitalA]", "]"}], "^=",
          "\[GothicCapitalA]"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "\[GothicCapitalA]", "]"}], 
         "^=", "\<\"\[GothicCapitalA]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"InnerC`Q", "[", "\[GothicCapitalA]", "]"}], "^=", 
         "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"VBundleOfIndex", "[", "\[GothicCapitalA]", "]"}], "^=", 
         InterpretationBox[
          StyleBox["InnerC",
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`InnerC,
          Editable->False]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`\[GothicCapitalA]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977783312869*^9},
 CellTags->"Info1163728981383-9181623",
 CellID->66729528,ExpressionUUID->"0c3eff53-b245-4530-b9f2-53af8ab16efe"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["3. Tensors and tensor slots", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1339531280,ExpressionUUID->"f4f44c18-1ef4-4eff-98cd-53767a44bc77"],

Cell[CellGroupData[{

Cell["3.1. Define a tensor", "Subsection",ExpressionUUID->"6e665f91-497f-409e-85c8-983c34cca83a"],

Cell[TextData[{
 "Now we can define tensors. In ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " we always work with tensor fields on zero, one or several manifolds, \
having indices on zero, one or several of their vbundles."
}], "Text",ExpressionUUID->"b6150745-e0ba-469a-bd7c-7fa6677052d8"],

Cell[TextData[{
 StyleBox[ButtonBox["DefTensor",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefTensor.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Define a tensor\n", "Text"],
 StyleBox[ButtonBox["UndefTensor",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefTensor.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Undefine a tensor\n", "Text"],
 StyleBox[ButtonBox["xTensorQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "xTensorQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check a tensor\n", "Text"],
 StyleBox[ButtonBox["$Tensors",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$Tensors.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined tensors", "Text"]
}], "DefinitionBox",ExpressionUUID->"77b34b2e-6302-4f68-8cb8-b9dec729d00a"],

Cell["Definition of tensors.", "Text",ExpressionUUID->"52e4347c-1f61-48cc-96cf-1d5af49db5ed"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a tensor field ",
 StyleBox["T",
  FontFamily->"Courier"],
 " on ",
 StyleBox["M3",
  FontFamily->"Courier"],
 " with two contravariant indices and one covariant index, and two other \
tensors. The actual indices used in the definition are irrelevant, and are \
only meant to specify the vbundle associated to each slot"
}], "MathCaption",ExpressionUUID->"cb266cfd-184e-44cb-b1ab-3d2cd4c92074"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[117]:=",ExpressionUUID->"e909f5d5-4eb4-4e66-aba4-6c821b279003"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c], ". ", ""],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977796474069*^9},
 CellLabel->"During evaluation of In[117]:=",
 CellID->1083110202,ExpressionUUID->"1e41444b-094e-4b97-964c-4d443f6a9be6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[118]:=",ExpressionUUID->"23ddea17-1aa7-4ccd-8c2c-534254357ebd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`S[$CellContext`a, $CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977796624576*^9},
 CellLabel->"During evaluation of In[118]:=",
 CellID->1761773042,ExpressionUUID->"7622fc00-a7ba-483e-9758-0f9fdc824475"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[119]:=",ExpressionUUID->"bc8594f3-abaa-44d1-b2d7-49f49663aa87"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777967579813`*^9},
 CellLabel->"During evaluation of In[119]:=",
 CellID->614992058,ExpressionUUID->"e8c8137e-392b-4e08-b7cf-be74ce60110d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a scalar field ",
 StyleBox["r",
  FontFamily->"Courier"],
 " (do not forget the empty pair of brackets!)"
}], "MathCaption",ExpressionUUID->"6986e7f3-fc49-4f57-b7fd-35cc6fd8a5c9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"r", "[", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[120]:=",ExpressionUUID->"3b7640fa-e6ef-4593-8ddb-ab5073518a94"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"r", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`r[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777968733473`*^9},
 CellLabel->"During evaluation of In[120]:=",
 CellID->332725795,ExpressionUUID->"1d47e5f9-33a7-4449-9f5e-7ff8aff31037"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 " on ",
 StyleBox["M3",
  FontFamily->"Courier"],
 " and ",
 StyleBox["S2",
  FontFamily->"Courier"],
 " with two contravariant indices and one covariant index, The tensor is also \
a function of the parameter time:"
}], "MathCaption",ExpressionUUID->"f22202b1-e4a5-4b3d-a847-f93f27bfc2ab"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "C"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"M3", ",", "S2", ",", "time"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[121]:=",ExpressionUUID->"20af02d0-e91b-4822-904e-334ddfcd005d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"U", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`U[$CellContext`a, $CellContext`b, -C], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977796991355*^9},
 CellLabel->"During evaluation of In[121]:=",
 CellID->75019777,ExpressionUUID->"4b4d64c5-3907-4145-9d39-5a9fecca207e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The tensor is identified by a symbol and we shall associate to that symbol \
all definitions related to the tensor. This is internally done using ",
 StyleBox["UpSet",
  FontFamily->"Courier"],
 " (^=) rather than ",
 StyleBox["Set",
  FontFamily->"Courier"],
 " (=). The definitions can then be collected using ",
 StyleBox["?",
  FontFamily->"Courier"]
}], "Text",ExpressionUUID->"a932613c-c10f-409d-bce5-8b63a0006442"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "U"}]], "Input",
 CellLabel->
  "In[122]:=",ExpressionUUID->"3325b290-4960-4c9d-bcc0-7bee99df921f"],

Cell[CellGroupData[{

Cell["Global`U", "Print", "PrintUsage",
 CellChangeTimes->{3.7289777971240253`*^9},
 CellTags->"Info1223728981397-9181623",
 CellID->284491250,ExpressionUUID->"51354a4a-940d-40e8-8e3e-9b1c29871105"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "U", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`U],
          Editable->False], "^=", "U"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["time",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`time,
            Editable->False], ",", "M3", ",", "S2"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", "S2", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]S2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentS2,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["time",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`time,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "U", "]"}], "^=", "\<\"U\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]S2",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentS2,
             Editable->False]}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "U", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "U", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977797147203*^9},
 CellTags->"Info1223728981397-9181623",
 CellID->1145840528,ExpressionUUID->"14242c38-c27d-43a7-955c-6a18caddad3f"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
A tensor can be undefined. Its properties are lost, and the symbol is removed.\
\>", "MathCaption",ExpressionUUID->"5bec7d37-3441-42aa-ac77-fe2c5ff31cb9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "U", "]"}]], "Input",
 CellLabel->
  "In[123]:=",ExpressionUUID->"44a1d28c-a478-43de-b726-175b2cbe502b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777973574677`*^9},
 CellLabel->"During evaluation of In[123]:=",
 CellID->1655891447,ExpressionUUID->"04722cc1-71bb-48de-bba7-38ed39adda76"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "U"}]], "Input",
 CellLabel->
  "In[124]:=",ExpressionUUID->"372ddd7d-2650-4c09-9477-f0fcfeae250d"],

Cell[BoxData[
 TemplateBox[{
  "Information", "notfound", 
   "\"Symbol \\!\\(\\*RowBox[{\\\"\\\\\\\"U\\\\\\\"\\\"}]\\) not found.\"", 2,
    124, 10, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977797693821*^9},
 CellLabel->"During evaluation of In[124]:=",
 CellID->1326968256,ExpressionUUID->"45eed1e1-1aaa-479b-8f10-034e7d87737f"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "An object with indices on a given vbundle is considered to be necessarily a \
field on the base manifold of that vbundle. ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " checks that point. See the discussion below on Constant objects."
}], "Text",ExpressionUUID->"eac2a4fa-aaf4-46e1-8eb8-fac4551dd604"],

Cell[CellGroupData[{

Cell[TextData[{
 "The tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 " with indices on vbundles ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 " and ",
 StyleBox["TangentS2",
  FontFamily->"Courier"],
 " is defined to live only on ",
 StyleBox["M3",
  FontFamily->"Courier"],
 ". ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " adds the manifold ",
 StyleBox["S2",
  FontFamily->"Courier"],
 " to the list ",
 StyleBox["DependenciesOfTensor[U].",
  FontFamily->"Courier"]
}], "MathCaption",ExpressionUUID->"43e5cec0-c28c-457a-b300-2ba1c20f74e9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{"a", ",", "A"}], "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[125]:=",ExpressionUUID->"d9bacb74-5924-467a-9802-2a4ff6b729a6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"U", "[", 
    RowBox[{"a", ",", "A"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`U[$CellContext`a, $CellContext`A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977797853896*^9},
 CellLabel->"During evaluation of In[125]:=",
 CellID->737804497,ExpressionUUID->"1203a021-fd2d-4d39-872b-2692e1646fc9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "U"}]], "Input",
 CellLabel->
  "In[126]:=",ExpressionUUID->"2a6b19c0-33cf-4265-968a-23e42c6a4d2c"],

Cell[CellGroupData[{

Cell["Global`U", "Print", "PrintUsage",
 CellChangeTimes->{3.7289777979419537`*^9},
 CellTags->"Info1263728981397-9181623",
 CellID->709603301,ExpressionUUID->"125cac5c-4025-4a7f-a7b8-a2538f16f51f"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "U", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`U],
          Editable->False], "^=", "U"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]S2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentS2,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "U", "]"}], "^=", "\<\"U\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]S2",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentS2,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "U", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "U", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977797967342*^9},
 CellTags->"Info1263728981397-9181623",
 CellID->1325988116,ExpressionUUID->"493ff12b-20ae-40f5-99b4-5f0b62a6615a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["UndefTensor[U]", "Input",
 CellLabel->
  "In[127]:=",ExpressionUUID->"b7df524a-d99c-4041-8abb-9cbaca17e715"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289777981022*^9},
 CellLabel->"During evaluation of In[127]:=",
 CellID->1444460693,ExpressionUUID->"25395c65-cc6f-46bd-8ab8-6a451c4d2ffc"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Note that, from version 1.0 of",
 StyleBox[" xTensor`",
  FontFamily->"Courier"],
 ", it is now possible to work with tensors whose head is not a symbol, but a \
more complicated expression. See the update in section 3.7."
}], "Text",ExpressionUUID->"c8c17be7-a18e-43f2-8cd3-4b7a909f5c69"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.2. Standard output", "Subsection",ExpressionUUID->"46b8adf0-3b18-4471-8061-37060ea25fd1"],

Cell[TextData[{
 "It is important to have nice-looking output expressions, mainly when we \
deal with very large expressions. ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has built-in rules to convert tensors and derivatives into 2-dimensional \
boxes for ",
 StyleBox["StandardForm",
  FontFamily->"Courier"],
 " output. Those boxes can be cut and pasted, using ",
 StyleBox["InterpretationBox",
  FontFamily->"Courier"],
 ". Additionally, translation to ",
 Cell[BoxData[
  StyleBox[
   RowBox[{"L", 
    StyleBox[
     AdjustmentBox["A",
      BoxBaselineShift->-0.2,
      BoxMargins->{{-0.36, -0.1}, {0, 0}}],
     FontSize->Smaller], "T", 
    AdjustmentBox["E",
     BoxBaselineShift->0.5,
     BoxMargins->{{-0.075, -0.085}, {0, 0}}], "X"}]]],ExpressionUUID->
  "9834a37d-1481-4be5-a931-06d4e16d22da"],
 " will be shown below."
}], "Text",ExpressionUUID->"a8f8d757-5dbf-417d-8a23-8240bd97d0c1"],

Cell[CellGroupData[{

Cell[TextData[{
 "The output of tensors in ",
 StyleBox["StandardForm",
  FontFamily->"Courier"],
 " has been defined with 2-dimensional boxes, but the internal structure is \
kept in 1-dimensional form (as usual in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "):"
}], "MathCaption",ExpressionUUID->"9e6f37b7-0fbd-4a21-8e30-f2652596eaaa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"r", "[", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[128]:=",ExpressionUUID->"cebd0320-5f16-4328-9816-e722b7b75061"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", " "},
           {" ", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289778087250147`*^9},
 CellLabel->"Out[128]=",
 CellID->713664600,ExpressionUUID->"71ec79e4-14b4-4f43-a8b2-63acb13d242d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[129]:=",ExpressionUUID->"697f1067-67f7-4b49-82b5-bad6dacbd53e"],

Cell["{T[a, b, -c], S[a, b], v[-a], r[]}", "Output",
 CellChangeTimes->{3.7289778088743753`*^9},
 CellLabel->"Out[129]//InputForm=",
 CellID->864763705,ExpressionUUID->"e2de2219-b328-4620-b812-354bc0a1597c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "No rules have been defined for ",
 StyleBox["OutputForm",
  FontFamily->"Courier"],
 " because the output is simple enough:"
}], "MathCaption",ExpressionUUID->"14262219-47b3-426f-a774-55ed1891688f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"OutputForm", "[", "%%", "]"}]], "Input",
 CellLabel->
  "In[130]:=",ExpressionUUID->"b9f3d553-d9ac-4b3b-b24e-8396dd824f22"],

Cell[OutputFormData["\<\
{T[a, b, -c], S[a, b], v[-a], r[]}\
\>", "\<\
{T[a, b, -c], S[a, b], v[-a], r[]}\
\>"], "Output",
 CellChangeTimes->{3.728977809003673*^9},
 CellLabel->"Out[130]//OutputForm=",
 CellID->2061750686,ExpressionUUID->"bc22edc7-46e3-45e7-b207-5cc70217bf74"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The ouput representation of the heads can be given at definition time, or \
changed later. Again, this is only possible in ",
 StyleBox["StandardForm",
  FontFamily->"Courier"],
 ". When using this option, it is very important to remember that the input \
and output names of the same tensor are different!"
}], "Text",ExpressionUUID->"2ddfb04f-79b2-4628-8a54-be7dd6c5d127"],

Cell[TextData[{
 StyleBox[ButtonBox["PrintAs",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PrintAs.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["Option of ", "Text"],
 StyleBox["DefType", "Text",
  FontFamily->"Courier"],
 StyleBox[" functions to define the output of a head", "Text"]
}], "DefinitionBox",ExpressionUUID->"1fa6c783-fc39-42cb-a9e6-0fadd5dabb13"],

Cell["Output of defined heads.", "Text",ExpressionUUID->"dfc0d1ae-5874-420d-81a8-7f54fde2895f"],

Cell[CellGroupData[{

Cell[TextData[{
 "The tensor ",
 StyleBox["TT",
  FontFamily->"Courier"],
 " will be output as ",
 StyleBox["\[Tau]",
  FontFamily->"Courier"],
 ". ",
 StyleBox["PrintAs",
  FontFamily->"Courier"],
 " must give a string or a function that applied on the defined symbol gives \
a string"
}], "MathCaption",ExpressionUUID->"a35238a9-4a66-4702-9e53-5dc85bd52dc6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"TT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", " ", "M3", ",", " ", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<\[Tau]\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[131]:=",ExpressionUUID->"34e02ad1-a042-44c9-9e4c-81a5b7cb5bcc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`TT[$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778091252203`*^9},
 CellLabel->"During evaluation of In[131]:=",
 CellID->47651610,ExpressionUUID->"3240a1b8-1453-41c1-9b58-db31ae5c1d0e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TT", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[132]:=",ExpressionUUID->"d9ac9c61-3d48-4166-8427-2d96affd6e96"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Tau]", 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TT[$CellContext`a, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977809395767*^9},
 CellLabel->"Out[132]=",
 CellID->2016692586,ExpressionUUID->"3d862ef4-be35-4b67-80cc-e593b9070726"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[133]:=",ExpressionUUID->"069da075-c092-4052-a1f2-7d90e01fe546"],

Cell["TT[a, -c]", "Output",
 CellChangeTimes->{3.728977809696003*^9},
 CellLabel->"Out[133]//InputForm=",
 CellID->817892301,ExpressionUUID->"fb4f208e-488c-48ac-8ab1-7a871566ffea"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["As an example of a printas function we can use:", "MathCaption",ExpressionUUID->"dc0a0313-a72b-46d0-9678-ef1987cf6b04"],

Cell[BoxData[
 RowBox[{
  RowBox[{"FirstCharacter", "[", "symbol_", "]"}], ":=", 
  RowBox[{"First", "[", 
   RowBox[{"Characters", "[", 
    RowBox[{"ToString", "[", "symbol", "]"}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[134]:=",ExpressionUUID->"f4f3cf21-8ae1-400e-8d8c-071acd584d55"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Force", "[", "a", "]"}], ",", " ", "M3", ",", " ", 
   RowBox[{"PrintAs", "\[Rule]", "FirstCharacter"}]}], "]"}]], "Input",
 CellLabel->
  "In[135]:=",ExpressionUUID->"9db0fada-cdf3-40eb-b024-f57f8c6e6539"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Force", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Force[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977810108201*^9},
 CellLabel->"During evaluation of In[135]:=",
 CellID->273078460,ExpressionUUID->"335e854b-29c4-4062-b84b-0e61fbaab3fc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Force", "[", "a", "]"}]], "Input",
 CellLabel->
  "In[136]:=",ExpressionUUID->"f957743c-546a-43ae-a38c-a06061a88412"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"F", 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Force[$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289778102434187`*^9},
 CellLabel->"Out[136]=",
 CellID->457123581,ExpressionUUID->"08818c66-c99a-48ca-b9ac-3e5c24b6a929"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The formatting of a tensor can be manually changed at any time:\
\>", "MathCaption",ExpressionUUID->"a7509d5c-c4bf-4d58-88f5-9155a82b7505"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PrintAs", "[", "Force", "]"}], "^=", 
  "\"\<\[ScriptCapitalF]\>\""}]], "Input",
 CellLabel->
  "In[137]:=",ExpressionUUID->"e4a72725-8685-4ace-a4f0-ff911667b4e2"],

Cell[BoxData["\<\"\[ScriptCapitalF]\"\>"], "Output",
 CellChangeTimes->{3.7289778104118433`*^9},
 CellLabel->"Out[137]=",
 CellID->1327020548,ExpressionUUID->"f2898a4f-b588-42d6-95e9-4d5a9d8fb805"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Force", "[", "a", "]"}]], "Input",
 CellLabel->
  "In[138]:=",ExpressionUUID->"addd8d6b-1be4-43e2-be3c-d488149b96e7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[ScriptCapitalF]", 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Force[$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977810845522*^9},
 CellLabel->"Out[138]=",
 CellID->1616676141,ExpressionUUID->"031a0caf-12e7-4ec4-b794-3660389bac1c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The 2-dimensional boxes are based on a combination of three built-in's: \n\t\
",
 StyleBox["SubsuperscriptBox/GridBox",
  FontFamily->"Courier"],
 ", which construct the sub/super-indices structures,\n\t",
 StyleBox["StyleBox",
  FontFamily->"Courier"],
 ", which arranges the indices with the proper interspaces, and\n\t",
 StyleBox["InterpretationBox",
  FontFamily->"Courier"],
 ", which stores both the box-form and the input-form of the expression. Note \
the option ",
 StyleBox["Editable->False",
  FontFamily->"Courier"],
 ", which prevents desynchronization. \nBefore version 1.0.0 we always used a \
formatting based on ",
 StyleBox["SubsuperscriptBox",
  FontFamily->"Courier"],
 ". Now we use a formatting based on ",
 StyleBox["GridBox",
  FontFamily->"Courier"],
 ", which allows vertical alignment of indices. This can be controlled with a \
global variable:"
}], "MathCaption",
 CellChangeTimes->{{3.494607480806465*^9, 3.494607484000292*^9}, {
  3.494607534420059*^9, 3.49460753608278*^9}, {3.494607574200631*^9, 
  3.494607584272273*^9}},ExpressionUUID->"87db2a4b-67fc-4759-a55e-\
3009a584e9df"],

Cell[CellGroupData[{

Cell[BoxData["$TensorBoxes"], "Input",
 CellChangeTimes->{{3.494607345412532*^9, 3.494607386932415*^9}},
 CellLabel->
  "In[139]:=",ExpressionUUID->"9410297d-59b2-4490-a6df-4f3c20857999"],

Cell[BoxData["\<\"GridBox\"\>"], "Output",
 CellChangeTimes->{3.7289778111741877`*^9},
 CellLabel->"Out[139]=",
 CellID->1512799943,ExpressionUUID->"2439ebe4-7ad2-441a-93f8-178a2e3caa29"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Force", "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.494607407743402*^9, 3.4946074100472403`*^9}},
 CellLabel->
  "In[140]:=",ExpressionUUID->"393998c1-3a41-44f3-a7ad-22def2da8c61"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[ScriptCapitalF]", 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Force[$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977811356537*^9},
 CellLabel->"Out[140]=",
 CellID->1569745168,ExpressionUUID->"b2fed207-b25e-4660-9772-2e601c5da022"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBoxes", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[141]:=",ExpressionUUID->"e67ac65b-ec58-4095-80dd-11bf04d39008"],

Cell[BoxData[
 RowBox[{"InterpretationBox", "[", 
  RowBox[{
   RowBox[{"StyleBox", "[", 
    RowBox[{
     RowBox[{"GridBox", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"{", 
         RowBox[{"\<\"\[ScriptCapitalF]\"\>", ",", 
          RowBox[{"StyleBox", "[", 
           RowBox[{
            RowBox[{"GridBox", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{
                RowBox[{"{", "\<\"a\"\>", "}"}], ",", 
                RowBox[{"{", "\<\" \"\>", "}"}]}], "}"}], ",", 
              RowBox[{"RowSpacings", "\[Rule]", "0"}], ",", 
              RowBox[{"ColumnSpacings", "\[Rule]", "0.05`"}], ",", 
              RowBox[{"GridFrameMargins", "\[Rule]", 
               RowBox[{"{", 
                RowBox[{
                 RowBox[{"{", 
                  RowBox[{"0", ",", "0"}], "}"}], ",", 
                 RowBox[{"{", 
                  RowBox[{"0", ",", "0"}], "}"}]}], "}"}]}]}], "]"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "9"}]}], "]"}]}], "}"}], "}"}], 
       ",", 
       RowBox[{"ColumnSpacings", "\[Rule]", "0.05`"}], ",", 
       RowBox[{"RowAlignments", "\[Rule]", "Center"}]}], "]"}], ",", 
     RowBox[{"AutoSpacing", "\[Rule]", "False"}], ",", 
     RowBox[{"ShowAutoStyles", "\[Rule]", "False"}]}], "]"}], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[ScriptCapitalF]", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Force[$CellContext`a],
    Editable->False], ",", 
   RowBox[{"Editable", "\[Rule]", "False"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7289778115604753`*^9},
 CellLabel->"Out[141]=",
 CellID->870435181,ExpressionUUID->"0f371ab1-ed8f-49ce-b193-3141a4b7f69d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"$TensorBoxes", "=", "\"\<ScriptBox\>\""}]], "Input",
 CellChangeTimes->{{3.4946074131505013`*^9, 3.494607440428686*^9}},
 CellLabel->
  "In[142]:=",ExpressionUUID->"506b43dd-d44b-42ab-a324-bb96123030a9"],

Cell[BoxData["\<\"ScriptBox\"\>"], "Output",
 CellChangeTimes->{3.728977811692133*^9},
 CellLabel->"Out[142]=",
 CellID->745483703,ExpressionUUID->"43b9f0f4-6917-4366-887c-97c2d0fa29fb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Force", "[", "a", "]"}]], "Input",
 CellChangeTimes->{{3.494607427005692*^9, 3.4946074286021557`*^9}},
 CellLabel->
  "In[143]:=",ExpressionUUID->"ba03388a-e27c-4842-8b54-5886190efa55"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   SubsuperscriptBox["\[ScriptCapitalF]", " ", "a"],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Force[$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977811819394*^9},
 CellLabel->"Out[143]=",
 CellID->924999723,ExpressionUUID->"2b055c64-9192-4499-a02b-d53c578b909a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToBoxes", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.494607447216296*^9, 3.4946074492725477`*^9}},
 CellLabel->
  "In[144]:=",ExpressionUUID->"f5b63544-f31a-4d9e-9dd2-9df447de8ecf"],

Cell[BoxData[
 RowBox[{"InterpretationBox", "[", 
  RowBox[{
   RowBox[{"StyleBox", "[", 
    RowBox[{
     RowBox[{"SubsuperscriptBox", "[", 
      RowBox[{"\<\"\[ScriptCapitalF]\"\>", ",", "\<\" \"\>", 
       ",", "\<\"a\"\>"}], "]"}], ",", 
     RowBox[{"AutoSpacing", "\[Rule]", "False"}], ",", 
     RowBox[{"ShowAutoStyles", "\[Rule]", "False"}]}], "]"}], ",", 
   InterpretationBox[
    StyleBox[
     SubsuperscriptBox["\[ScriptCapitalF]", " ", "a"],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Force[$CellContext`a],
    Editable->False], ",", 
   RowBox[{"Editable", "\[Rule]", "False"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7289778119795628`*^9},
 CellLabel->"Out[144]=",
 CellID->836223828,ExpressionUUID->"6bcf6bcd-c4db-4dcb-8fee-bb29a0a7dd4a"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$TensorBoxes", "=", "\"\<GridBox\>\""}], ";"}]], "Input",
 CellChangeTimes->{{3.494607453298777*^9, 3.4946074658144817`*^9}},
 CellLabel->
  "In[145]:=",ExpressionUUID->"0d58476b-aa43-4d92-a113-7f9bffbce3ae"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can cut-and-paste the output, but not edit it, nor construct a tensor \
directly in box-form (do this manually; the \"help\" format of this notebook \
does not allow cut and paste through different sessions):\
\>", "MathCaption",ExpressionUUID->"8fa345cd-27c2-48f9-82e7-c7044c3487d4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   SubsuperscriptBox["\[ScriptCapitalF]", " ", "a"],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Force[$CellContext`a],
  Editable->False]], "Input",
 CellLabel->
  "In[146]:=",ExpressionUUID->"49841c43-6c3c-4ed6-a911-1f859d13a4cb"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[ScriptCapitalF]", 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Force[$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289778123386993`*^9},
 CellLabel->"Out[146]=",
 CellID->1280891228,ExpressionUUID->"ff67b1e0-c66c-4bb9-bfe8-0e6ad18750a1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[147]:=",ExpressionUUID->"cfccd1c6-a210-40b1-a1f2-e191d5504a3c"],

Cell["Force[a]", "Output",
 CellChangeTimes->{3.728977812655814*^9},
 CellLabel->"Out[147]//InputForm=",
 CellID->1562791468,ExpressionUUID->"1d42a209-0066-4317-8506-ac987e7738c2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"4820e426-b458-4fec-a2cb-fefb100c0a0d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "Force", "]"}]], "Input",
 CellLabel->
  "In[148]:=",ExpressionUUID->"268521b9-9c50-43e9-a5a4-953628137e2d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Force"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Force],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977812786703*^9},
 CellLabel->"During evaluation of In[148]:=",
 CellID->2111199866,ExpressionUUID->"e0b88a5b-6ddf-4927-a354-4a04b4b52a3a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "TT", "]"}]], "Input",
 CellLabel->
  "In[149]:=",ExpressionUUID->"6ab9ba5c-51fb-45df-8ac5-ad6366316025"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "TT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`TT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977812894014*^9},
 CellLabel->"During evaluation of In[149]:=",
 CellID->63848481,ExpressionUUID->"1f52acbc-edfe-4289-aa1c-14500633e66d"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"FirstCharacter", ",", "symbol"}], "]"}]], "Input",
 CellLabel->
  "In[150]:=",ExpressionUUID->"f371e37b-52dc-43a6-9e13-b1aed7548114"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.3. Expressions and validation", "Subsection",ExpressionUUID->"017626f1-d144-499d-99e2-9e9a8cf6e2f6"],

Cell[TextData[{
 "Once we use abstract indices, there is no need to have a special \
TensorTimes. The usual ",
 StyleBox["Times",
  FontFamily->"Courier"],
 " product is enough and has one important advantage: we can use many \
built-in rules in ",
 StyleBox["Mathematica ",
  FontSlant->"Italic"],
 "to simplify ",
 StyleBox["Times",
  FontFamily->"Courier"],
 " expressions. As usual, contractions are given using Einstein rule of \
staggered up/down repeated indices."
}], "Text",ExpressionUUID->"e1a6ff6a-6972-41d5-b2c0-8f6244714a7e"],

Cell[CellGroupData[{

Cell[TextData[{
 "The order of factors is irrelevant in ",
 StyleBox["Times",
  FontFamily->"Courier"],
 " expressions:"
}], "MathCaption",ExpressionUUID->"f070f89e-4c73-4430-a12f-a750214df2b3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}], 
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b"}], "]"}]}]], "Input",
 CellLabel->
  "In[151]:=",ExpressionUUID->"fa8c85c3-cc64-4834-b23f-9cd7f4af216e"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977820557081*^9},
 CellLabel->"Out[151]=",
 CellID->123842069,ExpressionUUID->"f4c9549e-60ca-4461-95c0-1e2d16be9f11"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}], "+", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[152]:=",ExpressionUUID->"a02e5133-37f8-4ce1-b18a-7f9d4b09d9c9"],

Cell[BoxData[
 RowBox[{"2", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289778207075033`*^9},
 CellLabel->"Out[152]=",
 CellID->501999740,ExpressionUUID->"fd0ce06f-9dd5-4d6d-bba0-773ca51b9900"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The syntax of any expression can be checked using the function ",
 StyleBox["Validate",
  FontFamily->"Courier"],
 ". This function works recursively from the most external structures to the \
most internal ones, calling private functions ",
 StyleBox["ValidateTensor",
  FontFamily->"Courier"],
 ", ",
 StyleBox["ValidateCovD",
  FontFamily->"Courier"],
 ", etc. That gives you an idea of which problems will be detected first."
}], "Text",ExpressionUUID->"cc580805-3a9b-48f6-9f1c-49130c8d2f70"],

Cell[TextData[{
 StyleBox[ButtonBox["Validate",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Validate.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Validate an input expression", "Text"]
}], "DefinitionBox",ExpressionUUID->"40129e61-34b7-4eed-b175-dd7b2996c8ae"],

Cell["Functions that deal with validation.", "Text",ExpressionUUID->"b47508f7-78da-4e01-84bd-9fee7e825728"],

Cell[CellGroupData[{

Cell["If we get no errors, the input is syntactically correct:", "MathCaption",ExpressionUUID->"befda43b-6f4b-4822-9008-0849e2738b6d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b", ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"S", "[", 
     RowBox[{"c", ",", "d"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "d"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "3"}], 
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[153]:=",ExpressionUUID->"4956aadf-f9da-4fcb-b9c7-5b0b99bf9fa2"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor v\\\\\
\\\"\\\"}]\\)\"", 2, 153, 11, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977820848768*^9},
 CellLabel->"During evaluation of In[153]:=",
 CellID->1793864764,ExpressionUUID->"2be0e124-2f12-49a2-8156-7d6047845521"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["ERROR",
       FontColor->Hue[0.]], "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`d],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`ERROR[
     $CellContext`v[-$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", " "},
           {" ", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289778209255743`*^9},
 CellLabel->"Out[153]=",
 CellID->962168315,ExpressionUUID->"d5bcfe7d-3aab-4be6-b30c-a1a6fbba3e61"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", "d"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "d"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{
     RowBox[{"r", "[", "]"}], "^", "3"}], 
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[154]:=",ExpressionUUID->"01322a7f-0a83-41ff-ac7f-5d6c21a6ddb8"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor T\\\\\
\\\"\\\"}]\\)\"", 2, 154, 12, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977821147854*^9},
 CellLabel->"During evaluation of In[154]:=",
 CellID->611008295,ExpressionUUID->"d0656a01-a745-4c40-8c04-cbfad3c5af75"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor S\\\\\
\\\"\\\"}]\\)\"", 2, 154, 13, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.72897782126476*^9},
 CellLabel->"During evaluation of In[154]:=",
 CellID->2097097758,ExpressionUUID->"ba717ef6-1131-4b02-acfa-9386fe6852b7"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor v\\\\\
\\\"\\\"}]\\)\"", 2, 154, 14, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977821317252*^9},
 CellLabel->"During evaluation of In[154]:=",
 CellID->1420404142,ExpressionUUID->"bbf1aa9a-b438-4ec3-9410-c70092f3ba68"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"Validate\\\", \
\\\"::\\\", \\\"error\\\"}], \\\"MessageName\\\"]\\) will be suppressed \
during this calculation.\"", 2, 154, 15, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289778213421164`*^9},
 CellLabel->"During evaluation of In[154]:=",
 CellID->1304194542,ExpressionUUID->"c411dd95-1b9e-4b8b-8991-0897856f2c60"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["ERROR",
       FontColor->Hue[0.]], "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"S", 
           StyleBox[GridBox[{
              {" ", "d"},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`S[-$CellContext`c, $CellContext`d],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`ERROR[
     $CellContext`S[-$CellContext`c, $CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["ERROR",
       FontColor->Hue[0.]], "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b", "c"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`ERROR[
     $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["ERROR",
       FontColor->Hue[0.]], "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`d],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`ERROR[
     $CellContext`v[-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`b],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728977821547344*^9},
 CellLabel->"Out[154]=",
 CellID->232082465,ExpressionUUID->"9118eedc-3111-4f41-b860-c22a1798fa23"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Some errors cannot be always localized in a single object. We do not return \
the expression:\
\>", "MathCaption",ExpressionUUID->"536827e2-c1d8-4869-8491-d688c27cabf4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "c"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[155]:=",ExpressionUUID->"3af3fe6f-bd80-450b-b313-a5e88fe9b355"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "repeated", 
   "\"Found indices with the same name \
\\!\\(\\*RowBox[{\\\"\\\\\\\"-c\\\\\\\"\\\"}]\\).\"", 2, 155, 16, 
   19253573948027831207, "Local", "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289778217031927`*^9},
 CellLabel->"During evaluation of In[155]:=",
 CellID->779686303,ExpressionUUID->"077c2b58-223e-4f52-8d28-ec7954fee910"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], "+", 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[156]:=",ExpressionUUID->"3449fe64-203f-44fa-a558-cb77be406c9c"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "inhom", 
   "\"Found inhomogeneous \
\\!\\(\\*RowBox[{\\\"\\\\\\\"indices\\\\\\\"\\\"}]\\): \
\\!\\(\\*RowBox[{\\\"\\\\\\\"{IndexList[-a], \
IndexList[a]}\\\\\\\"\\\"}]\\).\"", 2, 156, 17, 19253573948027831207, "Local",
    "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977822039587*^9},
 CellLabel->"During evaluation of In[156]:=",
 CellID->242480527,ExpressionUUID->"1f63d481-27fd-4ed1-a644-1ba8f8164e21"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Because validation takes up some time, it is not automatic. ",
 StyleBox["xTensor`", "DisplayMath"],
 " has the command ",
 StyleBox["Validate",
  FontFamily->"Courier"],
 " to be used by the user as required. However, if validation is needed in \
all computations it is possible to switch on automatic validation, by acting \
on the ",
 StyleBox["$Pre",
  FontFamily->"Courier"],
 " variable:"
}], "Text",ExpressionUUID->"5e6710a7-dc09-44b1-8fa2-06705999d232"],

Cell[CellGroupData[{

Cell["Here, there is not automatic validation:", "MathCaption",ExpressionUUID->"305da4d5-e080-41b0-8e15-3d2100d9a205"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"S", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->
  "In[157]:=",ExpressionUUID->"42e5d121-21ae-4256-b11c-fe044e3262b7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897782233939*^9},
 CellLabel->"Out[157]=",
 CellID->1593738930,ExpressionUUID->"1f832a38-7fd8-4c23-992f-44282a76e042"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[158]:=",ExpressionUUID->"acb84b21-d6d6-45ec-96ef-f30ebc1c8529"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "repeated", 
   "\"Found indices with the same name \\!\\(\\*RowBox[{\\\"\\\\\\\"b\\\\\\\"\
\\\"}]\\).\"", 2, 158, 18, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.72897782248877*^9},
 CellLabel->"During evaluation of In[158]:=",
 CellID->406667479,ExpressionUUID->"22466f4f-fd2c-425d-aaec-702be3ba7b60"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This switches on automatic validation:", "MathCaption",ExpressionUUID->"ca259f97-b295-45a4-996b-160232aef5aa"],

Cell[CellGroupData[{

Cell[BoxData["$Pre"], "Input",
 CellLabel->
  "In[159]:=",ExpressionUUID->"34a653ac-1b3f-449a-8729-badfd84b360b"],

Cell[BoxData["$Pre"], "Output",
 CellChangeTimes->{3.728977822621489*^9},
 CellLabel->"Out[159]=",
 CellID->1567354799,ExpressionUUID->"158d8a28-b8ee-4cad-aa8d-86c0729509a0"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$Pre", "=", "Validate"}], ";"}]], "Input",
 CellLabel->
  "In[160]:=",ExpressionUUID->"844c637f-5953-432c-b8bf-3f04eeb8154b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"S", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->
  "In[161]:=",ExpressionUUID->"6636582f-d60b-4b13-a9e6-0fc4c5341dae"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "repeated", 
   "\"Found indices with the same name \\!\\(\\*RowBox[{\\\"\\\\\\\"b\\\\\\\"\
\\\"}]\\).\"", 2, 161, 19, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977822928875*^9},
 CellLabel->"During evaluation of In[161]:=",
 CellID->1497692691,ExpressionUUID->"2388bc2d-ab73-442a-a79d-862836bed025"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This switches off automatic validation:", "MathCaption",ExpressionUUID->"3aebedc9-4ca5-4337-b640-2bab07ef3774"],

Cell[BoxData[
 RowBox[{"$Pre", "=."}]], "Input",
 CellLabel->
  "In[162]:=",ExpressionUUID->"26ec62f1-c428-4513-8ce6-5b8470b1ff54"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"S", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"v", "[", "b", "]"}]}]], "Input",
 CellLabel->
  "In[163]:=",ExpressionUUID->"1156007c-1e8b-4b2b-ac7d-b53b586c1164"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977823492058*^9},
 CellLabel->"Out[163]=",
 CellID->1065231150,ExpressionUUID->"c14e54cf-6bd0-4ec8-b99b-b29eb18daa5a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Note the ",
 StyleBox["HoldFirst",
  FontFamily->"Courier"],
 "  attribute of ",
 StyleBox["Validate",
  FontFamily->"Courier"],
 ", to make sure that we validate the input expression, and not whatever it \
evaluates to. That means that ",
 StyleBox["Evaluate",
  FontFamily->"Courier"],
 " must be used when the expression is not explicit:"
}], "MathCaption",
 CellChangeTimes->{{3.494607640047133*^9, 3.494607707850284*^9}, {
  3.4946079169214354`*^9, 
  3.4946080112297497`*^9}},ExpressionUUID->"b6c5c2b8-a08e-4a7f-bae0-\
f2098ca48b36"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Attributes", "[", "Validate", "]"}]], "Input",
 CellChangeTimes->{{3.494607625005177*^9, 3.494607629182807*^9}},
 CellLabel->
  "In[164]:=",ExpressionUUID->"3b7f4ae2-cd4a-4d7b-a804-b6f8929f371e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"HoldFirst", ",", "Protected"}], "}"}]], "Output",
 CellChangeTimes->{3.7289778236211357`*^9},
 CellLabel->"Out[164]=",
 CellID->1916256658,ExpressionUUID->"f191540b-7955-46b8-a972-70276d7e7d9e"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", "b", "]"}]}]}], ";"}]], "Input",
 CellChangeTimes->{{3.494607881583022*^9, 3.4946078879440603`*^9}},
 CellLabel->
  "In[165]:=",ExpressionUUID->"fe93f975-ccb6-4219-88ca-fdfe1dcd00e8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "expr", "]"}]], "Input",
 CellChangeTimes->{{3.4946078898602533`*^9, 3.4946078918098183`*^9}},
 CellLabel->
  "In[166]:=",ExpressionUUID->"4174f44f-7bfd-426a-b259-e9f1853fee8f"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289778239259357`*^9},
 CellLabel->"Out[166]=",
 CellID->713396947,ExpressionUUID->"3afbc5ff-cbb4-4716-a67d-58a02e723225"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{"Evaluate", "[", "expr", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.4946078938599787`*^9, 3.494607898075199*^9}},
 CellLabel->
  "In[167]:=",ExpressionUUID->"c83ef039-57cb-4f2f-a087-d860f334c746"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "repeated", 
   "\"Found indices with the same name \\!\\(\\*RowBox[{\\\"\\\\\\\"b\\\\\\\"\
\\\"}]\\).\"", 2, 167, 20, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977824040086*^9},
 CellLabel->"During evaluation of In[167]:=",
 CellID->661287028,ExpressionUUID->"fc08460c-4b2d-4862-97ec-117ad058d10b"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[168]:=",ExpressionUUID->"10db3ef5-88a4-4dc2-b2ac-3b432b42e8a5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "S", "]"}]], "Input",
 CellLabel->
  "In[169]:=",ExpressionUUID->"0c53c218-9311-4fad-83b2-f6041a1654da"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977824254533*^9},
 CellLabel->"During evaluation of In[169]:=",
 CellID->1023206371,ExpressionUUID->"1b056f06-41f5-46d8-b32b-4a3434ab21be"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.4. Symmetries", "Subsection",ExpressionUUID->"39425ff7-2e00-4f6f-9dc3-2e26809f04cc"],

Cell[TextData[{
 "There are two kinds of symmetries for a tensor. We can have monoterm (or \
permutation) symmetries, such that a tensor remains equal or changes sign \
under a permutation of its indices. Or we can have multiterm (for example \
cyclic) symmetries, such that a linear combination of several of those \
permuted tensors is equal to zero. In the former case the symmetry of a \
tensor of n indices can be described as a subgroup of the group {1,-1}x S_n. \
In the latter, the symmetry can be described as an algebra. Currently ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " can deal with an arbitrary case (n not too large: say 200 indices at most) \
of monoterm symmetries. However, it does not implement any general algorithm \
for multiterm symmetries, because no efficient algorithm is known for that \
case.\nThe twin package ",
 StyleBox["xPerm`",
  FontFamily->"Courier"],
 " is used to work with permutations. See ",
 StyleBox["xPermDoc.nb",
  FontFamily->"Courier"],
 " for further information."
}], "Text",ExpressionUUID->"16786579-b120-4641-8f02-b8857c0e200d"],

Cell[TextData[{
 StyleBox["GenSet\t\t\t\t\t", "DisplayMath"],
 StyleBox["Generating set of a permutations group", "Text"],
 StyleBox["\nStrongGenSet\t\t\t", "DisplayMath"],
 StyleBox["Strong generating set of a permutations group", "Text"],
 StyleBox["\nSymmetric\t\t\t\t", "DisplayMath"],
 StyleBox["Completely symmetric permutation group", "Text"],
 StyleBox["\nAntisymmetric\t\t\t", "DisplayMath"],
 StyleBox["Completely antisymmetric permutation group\n", "Text"],
 StyleBox["RiemannSymmetric\t\t", "DisplayMath"],
 StyleBox["Group of symmetry of the Riemann tensor\t", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["SymmetryOf",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SymmetryOf.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Symmetry properties of an expression.\n", "Text"],
 StyleBox[ButtonBox["ForceSymmetries",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ForceSymmetries.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Option of ", "Text"],
 StyleBox["DefTensor", "Text",
  FontFamily->"Courier"],
 StyleBox[" to allow for symmetries among up/down indices", "Text"]
}], "DefinitionBox",ExpressionUUID->"453cb721-f0a1-4ea4-8d34-26a216b7be59"],

Cell["Functions related to symmetries", "Text",ExpressionUUID->"819c8d03-ab78-40e2-b22b-81a1753ffb63"],

Cell[CellGroupData[{

Cell[TextData[{
 "Any permutation group on a set of n points can be given using a set of \
generators (head ",
 StyleBox["GenSet",
  FontFamily->"Courier"],
 ") or, better, a strong set of generators (head ",
 StyleBox["StrongGenSet",
  FontFamily->"Courier"],
 "). The functions ",
 StyleBox["Symmetric",
  FontFamily->"Courier"],
 " and ",
 StyleBox["Antisymmetric",
  FontFamily->"Courier"],
 " are example: ",
 StyleBox["From version 1.0.5, following ",
  FontColor->GrayLevel[0]],
 StyleBox["Mathematica",
  FontSlant->"Italic",
  FontColor->GrayLevel[0]],
 StyleBox[" 9, the expressions ",
  FontColor->GrayLevel[0]],
 StyleBox["Symmetric[inds]",
  FontFamily->"Courier"],
 StyleBox[" and ",
  FontColor->GrayLevel[0]],
 StyleBox["Antisymmetric[inds]",
  FontFamily->"Courier"],
 StyleBox[" are inert. The use of a second argument is needed to produce \
actually a strong generating set. The second argument will be in red, but \
that is a harmless issue.",
  FontColor->GrayLevel[0]]
}], "MathCaption",ExpressionUUID->"d2ff414e-6268-45ff-b8e5-d7a6a227c370"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Symmetric", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"a", ",", "b"}], "}"}], ",", "Cycles"}], "]"}]], "Input",
 CellLabel->
  "In[170]:=",ExpressionUUID->"3b28144c-f079-463e-bef9-d1d71dd7d723"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"a", ",", "b"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "a", ",", "b", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{$CellContext`a, $CellContext`b}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728977831810829*^9},
 CellLabel->"Out[170]=",
 CellID->1356062270,ExpressionUUID->"86b51675-63c7-4700-af43-c76ddbe9662d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Antisymmetric", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"a", ",", "b"}], "}"}], ",", "Cycles"}], "]"}]], "Input",
 CellLabel->
  "In[171]:=",ExpressionUUID->"b110fb19-8a02-41c4-a9df-484a1974f4ac"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"a", ",", "b"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "a", ",", "b", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{$CellContext`a, $CellContext`b}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7289778319301023`*^9},
 CellLabel->"Out[171]=",
 CellID->94626279,ExpressionUUID->"0b170118-49a9-4dfe-b528-c58eb4102ee2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a symmetric tensor ",
 StyleBox["S",
  FontFamily->"Courier"],
 " on ",
 StyleBox["M3",
  FontFamily->"Courier"]
}], "MathCaption",ExpressionUUID->"283dc12e-c20d-4c3a-9071-0e8b7bd7488e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", " ", "M3", ",", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "b"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[172]:=",ExpressionUUID->"2f768cee-963c-41c4-a51e-395c139dda24"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`S[$CellContext`a, $CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977832062229*^9},
 CellLabel->"During evaluation of In[172]:=",
 CellID->1245972373,ExpressionUUID->"169e4780-3f91-4f30-ba4d-3ad8d73c477e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "S"}]], "Input",
 CellLabel->
  "In[173]:=",ExpressionUUID->"8dd2f949-19f4-451c-96f5-37969fe1aa56"],

Cell[CellGroupData[{

Cell["Global`S", "Print", "PrintUsage",
 CellChangeTimes->{3.7289778321776447`*^9},
 CellTags->"Info1733728981432-9181623",
 CellID->944631114,ExpressionUUID->"ab3ee4a8-7bf1-41bf-a3b2-652eaf0137c9"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "S", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`S],
          Editable->False], "^=", "S"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "S", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "S", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "S", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "S", "]"}], "^=", "\<\"S\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "S", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "S", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"1", ",", "2"}], "}"}], ",", 
           RowBox[{"GenSet", "[", 
            InterpretationBox[
             StyleBox[
              RowBox[{"(", "1", ",", "2", ")"}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             xAct`xPerm`Cycles[{1, 2}],
             Editable->False], "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "S", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "S", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977832202317*^9},
 CellTags->"Info1733728981432-9181623",
 CellID->68530727,ExpressionUUID->"579ee2dc-80f0-49f8-9e65-24f5efae300c"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a totally antisymmetric tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 " on ",
 StyleBox["M3",
  FontFamily->"Courier"],
 ". We can use the names of the indices or their positions in the tensor"
}], "MathCaption",ExpressionUUID->"1617e0b4-1388-459a-be01-55a7eed330b4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", " ", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[174]:=",ExpressionUUID->"1ef15a81-375b-4a70-be6c-b3b4374cc5da"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c], ". ", 
   ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778324726*^9},
 CellLabel->"During evaluation of In[174]:=",
 CellID->744693897,ExpressionUUID->"27d19452-2983-48a0-9792-a6c3c013858f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "U"}]], "Input",
 CellLabel->
  "In[175]:=",ExpressionUUID->"1a077c8d-5660-4c49-ad01-4030aa0bf0b8"],

Cell[CellGroupData[{

Cell["Global`U", "Print", "PrintUsage",
 CellChangeTimes->{3.728977832911667*^9},
 CellTags->"Info1753728981432-9181623",
 CellID->535632163,ExpressionUUID->"a67b04c3-c3c7-4ac1-8a2a-674cc955be23"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "U", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`U],
          Editable->False], "^=", "U"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "U", "]"}], "^=", "\<\"U\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "U", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
           RowBox[{"GenSet", "[", 
            RowBox[{
             InterpretationBox[
              StyleBox[
               RowBox[{"-", 
                RowBox[{"(", "1", ",", "2", ")"}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              -xAct`xPerm`Cycles[{1, 2}],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox[
               RowBox[{"-", 
                RowBox[{"(", "2", ",", "3", ")"}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              -xAct`xPerm`Cycles[{2, 3}],
              Editable->False]}], "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "U", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "U", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778329350767`*^9},
 CellTags->"Info1753728981432-9181623",
 CellID->118315317,ExpressionUUID->"78dd0d02-24ec-4d49-9b35-f8d81c30bc4b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "For example, for the tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 " the group is generated by the two signed permutations -",
 StyleBox["Cycles",
  FontFamily->"Courier"],
 "[{1,2}] (representing a change of sign under the exchange of slots 1 and 2) \
and -",
 StyleBox["Cycles",
  FontFamily->"Courier"],
 "[{2,3}] (representing a change of sign under the exchange of slots 2 and \
3). Again, ",
 StyleBox["Cycles",
  FontFamily->"Courier"],
 " is colored in red because there is a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " symbol with the same name. "
}], "Text",ExpressionUUID->"a6abed5b-67e9-4968-b506-10e9f89ff031"],

Cell["Note however that not every generating set is valid:", "Text",ExpressionUUID->"e415ec05-ead8-47aa-9cf4-ce605f0a1841"],

Cell[CellGroupData[{

Cell["\<\
For example, in a group of three indices, antisymmetry in the first two \
indices is not consistent with symmetry in the second and third indices, \
unless the whole tensor is zero:\
\>", "MathCaption",ExpressionUUID->"b2a6ac3f-ae13-4448-adea-85f6ca073e36"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "M3", ",", 
   RowBox[{"GenSet", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"Cycles", "[", 
       RowBox[{"{", 
        RowBox[{"1", ",", "2"}], "}"}], "]"}]}], ",", 
     RowBox[{"Cycles", "[", 
      RowBox[{"{", 
       RowBox[{"2", ",", "3"}], "}"}], "]"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[176]:=",ExpressionUUID->"0c245624-a1f5-404a-889e-0815cb126fd1"],

Cell[BoxData[
 TemplateBox[{
  "DefTensor", "zero", 
   "\"Symmetry makes tensor zero. Use VanishingQ instead.\"", 2, 176, 21, 
   19253573948027831207, "Local", "xAct`xTensor`DefTensor"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977833166712*^9},
 CellLabel->"During evaluation of In[176]:=",
 CellID->2103159344,ExpressionUUID->"93b808d8-1d22-4f0b-a682-11b237aa5ea7"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The symmetries must be consistent with the character of the indices \
(irrespectively of whether there is a metric or not). This is checked.\
\>", "Text",ExpressionUUID->"f6d502ca-da3b-422b-9dd4-34a538cb82a2"],

Cell[CellGroupData[{

Cell["\<\
We cannot define a symmetric tensor with two indices of different character:\
\>", "MathCaption",ExpressionUUID->"69bcd9ca-2ef7-44df-b030-faa030a529ff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Q", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[177]:=",ExpressionUUID->"193fa296-2a2f-4655-87ec-91cfd5a8c8a8"],

Cell[BoxData[
 TemplateBox[{
  "DefTensor", "wrongsym", 
   "\"Symmetry properties are inconsistent with indices of tensor.\"", 2, 177,
    22, 19253573948027831207, "Local", "xAct`xTensor`DefTensor"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289778334498796`*^9},
 CellLabel->"During evaluation of In[177]:=",
 CellID->218264988,ExpressionUUID->"2a3d9b51-3e5e-4126-9965-ef065f2714d1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["unless we force it (so that the check is not performed):", "MathCaption",ExpressionUUID->"e5709dcc-1cf8-4aa9-8d43-01c6fc204c2c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Q", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}], ",", 
   RowBox[{"ForceSymmetries", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[178]:=",ExpressionUUID->"f805382a-8507-41e8-9736-3487e5adf1f8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Q", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Q[$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977833694573*^9},
 CellLabel->"During evaluation of In[178]:=",
 CellID->477584240,ExpressionUUID->"8bfeaef2-2de7-4fd6-a814-87005107fcc7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
For symmetry groups of order larger than 10000 the check would take too long \
and is not performed:\
\>", "MathCaption",ExpressionUUID->"ef20c19f-bdf0-4443-84dc-bd477feedc20"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"supereta", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", 
     RowBox[{"-", "f"}], ",", 
     RowBox[{"-", "g"}], ",", 
     RowBox[{"-", "h"}], ",", 
     RowBox[{"-", "h1"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"Range", "[", "8", "]"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[179]:=",ExpressionUUID->"578b54ef-1fa1-436f-af19-ac42c58fb1b8"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"  Order of group of symmetry: \"\>", "\[InvisibleSpace]", 
   "40320", "\[InvisibleSpace]", "\<\" > 10000. ForceSymmetry check not \
performed.\"\>"}],
  SequenceForm[
  "  Order of group of symmetry: ", 40320, 
   " > 10000. ForceSymmetry check not performed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778339522038`*^9},
 CellLabel->"During evaluation of In[179]:=",
 CellID->1611932841,ExpressionUUID->"779ed934-9078-4257-abea-a144b075fd31"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"supereta", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", "e", ",", 
     RowBox[{"-", "f"}], ",", 
     RowBox[{"-", "g"}], ",", 
     RowBox[{"-", "h"}], ",", 
     RowBox[{"-", "h1"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`supereta[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d, $CellContext`e, -$CellContext`f, -$CellContext`g, \
-$CellContext`h, -$CellContext`h1], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977833970067*^9},
 CellLabel->"During evaluation of In[179]:=",
 CellID->1249051782,ExpressionUUID->"acae08af-1839-4f77-8dc7-978450fb1098"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"Q", ",", "supereta", ",", "U"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[180]:=",ExpressionUUID->"545b076d-d3be-4b68-8acb-4c0166b5fd8a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Q"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Q],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897783411943*^9},
 CellLabel->"During evaluation of In[180]:=",
 CellID->1160650649,ExpressionUUID->"5e14588e-2f8d-49ef-bd33-b9b801e998c8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "supereta"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`supereta],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778341372747`*^9},
 CellLabel->"During evaluation of In[180]:=",
 CellID->1461492042,ExpressionUUID->"54908e96-ff70-48c2-b1d9-ee06bf3ac8bd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977834156289*^9},
 CellLabel->"During evaluation of In[180]:=",
 CellID->378795508,ExpressionUUID->"8cd86757-6047-4c48-bbb8-fa507b9c5aa0"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "W", "]"}]], "Input",
 CellLabel->
  "In[181]:=",ExpressionUUID->"426b0a8c-5cea-49d7-b87d-6103d81733c3"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "3.5. delta, Gdelta and Sdelta. ",
 StyleBox["V1.1.0 UPDATE",
  FontColor->RGBColor[0, 0, 1]]
}], "Subsection",ExpressionUUID->"8149cdd7-01de-4e5a-a962-368625084a4c"],

Cell[TextData[{
 "The identity tensor on vbundles is called ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " and is the same for all vbundles. This has changed from old versions of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", in which each vbundle had its own identity tensor.\nThe ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " is a very special object. For example Penrose & Rindler use a delta object \
with one index on top of the other, rather than staggered, not following the \
\"sorted slots\" convention. That means that indices of such an object cannot \
be moved up and down with a metric. Such a construction is not natural in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " and hence we must look for an alternative. Up to version 0.9.6 ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " was considered to be always a symmetric tensor, though this didn't make \
much sense in the absence of a metric, because an object with mixed indices \
cannot have symmetry. The problem becomes even worse in the presence of an \
antisymmetric metric (spinor and symplectic calculus), and hence this has now \
changed. The main requirement is that ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " is a tensor and has a preferred index configuration, which will be taken \
as ",
 StyleBox["delta[-a, b]",
  FontFamily->"Courier"],
 ", which is more natural in the usual conventions of spinor calculus. In the \
presence of a metric other configurations are possible, and actually ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " becomes equal to the metric. That means that in the presence of a \
symmetric metric ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " is symmetric, and in the presence of an antisymmetric metric ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " is antisymmetric."
}], "Text",ExpressionUUID->"e31ffbb3-2a58-491f-8c79-34c63a181061"],

Cell[TextData[{
 StyleBox[ButtonBox["delta",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "delta.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["Identity tensor on every vbundle", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["Gdelta",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Gdelta.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["Generalized ", "Text"],
 StyleBox["delta", "Text",
  FontFamily->"Courier"],
 StyleBox[" tensor on every vbundle\n", "Text"],
 StyleBox[ButtonBox["ExpandGdelta",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ExpandGdelta.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Expansion of ", "Text"],
 StyleBox["Gdelta", "Text",
  FontFamily->"Courier"],
 StyleBox[" in products of ", "Text"],
 StyleBox["delta", "Text",
  FontFamily->"Courier"],
 StyleBox["'s", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728977840948724*^9},ExpressionUUID->"1f261d1d-2917-4ae1-b6ee-\
54bc6f595eb0"],

Cell["Identity tensors.", "Text",ExpressionUUID->"72ef4872-06ea-48da-b3b6-afb1ab2c255c"],

Cell[CellGroupData[{

Cell[TextData[{
 "The ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " tensor on the the vector bundle ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"c45cc6ad-6baa-44ca-9d48-154461fc9297"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"delta", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", "b"}], "]"}]], "Input",
 CellLabel->
  "In[182]:=",ExpressionUUID->"366abbff-f545-4141-a509-5e0b85de9b47"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "b"},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977843895678*^9},
 CellLabel->"Out[182]=",
 CellID->732934548,ExpressionUUID->"cd4a9c43-f473-4fef-8b4a-4a4da31f78a6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The opposite configuration is forbidden without a metric:", \
"MathCaption",ExpressionUUID->"0d09ccd5-64ff-4b3f-9b4b-f3315184fe41"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"delta", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}]}], "]"}], "//", "Validate"}]], "Input",
 CellLabel->
  "In[183]:=",ExpressionUUID->"7dc92267-4c32-4279-a993-0e54963e03a1"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor delta\
\\\\\\\"\\\"}]\\)\"", 2, 183, 23, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289778440457287`*^9},
 CellLabel->"During evaluation of In[183]:=",
 CellID->464749044,ExpressionUUID->"6a3b8160-081e-41d8-a69e-dfd42390d676"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["ERROR",
     FontColor->Hue[0.]], "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"a", " "},
            {" ", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[$CellContext`a, -$CellContext`b],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`ERROR[
   xAct`xTensor`delta[$CellContext`a, -$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897784407661*^9},
 CellLabel->"Out[183]=",
 CellID->639158369,ExpressionUUID->"73175e5d-beaf-4346-8221-18586e9f4390"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There is automatic simplification with most of the expected rules for the \
",
 StyleBox["delta",
  FontFamily->"Courier"],
 " tensors:"
}], "MathCaption",ExpressionUUID->"157e4412-5a39-43b4-a985-82ca3495540e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"delta", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", "a"}], "]"}]], "Input",
 CellLabel->
  "In[184]:=",ExpressionUUID->"755d651c-c673-41b5-8f5c-80a7583a391b"],

Cell[BoxData["3"], "Output",
 CellChangeTimes->{3.728977844196336*^9},
 CellLabel->"Out[184]=",
 CellID->621505489,ExpressionUUID->"bf619c5f-bb3f-4f3e-8adf-3aadc847ecf3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"delta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "b"}], "]"}], 
  RowBox[{"delta", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", "c"}], "]"}]}]], "Input",
 CellLabel->
  "In[185]:=",ExpressionUUID->"6784a59f-6539-4a7b-bf2b-2d5c4c65213d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "c"},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`a, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289778443316708`*^9},
 CellLabel->"Out[185]=",
 CellID->282294615,ExpressionUUID->"3d056001-eccf-4097-b2d5-2bfd63fce57f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"delta", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", "a"}], "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellLabel->
  "In[186]:=",ExpressionUUID->"343a025f-5253-43dc-9409-35f5abf9ddb0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {" "},
         {"b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[-$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977844442029*^9},
 CellLabel->"Out[186]=",
 CellID->734218082,ExpressionUUID->"67fedc8b-b7b8-438b-b291-02765d56a23f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "When both indices can be contracted, the second index of ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " is contracted first:"
}], "MathCaption",ExpressionUUID->"726a0338-d3d6-4ba2-810a-f4507b70c42a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"delta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "b"}], "]"}], 
  RowBox[{"S", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[187]:=",ExpressionUUID->"a847ef3d-b428-44b6-8e14-9ffb14882973"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977844717101*^9},
 CellLabel->"Out[187]=",
 CellID->562268512,ExpressionUUID->"b307c2f2-4cd9-4554-aed8-73e61afd5207"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The indices of ",
 StyleBox["delta",
  FontFamily->"Courier"],
 " can never have the same character, unless automatic conversion to a metric \
is possible:"
}], "MathCaption",ExpressionUUID->"5dc6d362-d4d8-4cc4-818a-4fe61da0798d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"delta", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "//", "Catch"}]], "Input",
 CellLabel->
  "In[188]:=",ExpressionUUID->"71ce46af-1159-4ac5-85eb-571ad624f945"],

Cell[BoxData[
 TemplateBox[{
  "MetricsOfVBundle", "missing", 
   "\"There is no \\!\\(\\*RowBox[{\\\"\\\\\\\"metric\\\\\\\"\\\"}]\\) in \\!\
\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]M3\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentM3, \
Rule[Editable, False]]\\).\"", 2, 188, 24, 19253573948027831207, "Local", 
   "xAct`xTensor`MetricsOfVBundle"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977845114039*^9},
 CellLabel->"During evaluation of In[188]:=",
 CellID->1118834384,ExpressionUUID->"8fe1e60f-c66f-4ea8-8c49-61ae816d53fa"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In version 0.9.0 ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " added a generalized delta tensor, with two sets of antisymmetric slots. In \
version 1.1.0 its contraction properties have been improved. In particular we \
now try to avoid expanding generalized deltas into single deltas whenever \
possible."
}], "Text",ExpressionUUID->"a4ed2797-8ad1-4985-b7ea-46784be6a640"],

Cell[CellGroupData[{

Cell["\<\
This is the generalized tensor with 2 lower indices and 2 upper indices:\
\>", "MathCaption",ExpressionUUID->"98862426-6ea5-4f99-81ff-4a54255b1f23"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Gdelta", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", "c", ",", "d"}], "]"}]], "Input",
 CellLabel->
  "In[189]:=",ExpressionUUID->"1c3da9f7-cd50-46bd-b11f-9d4d8584573f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", " ", "c", "d"},
         {"a", "b", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977845378893*^9},
 CellLabel->"Out[189]=",
 CellID->882678756,ExpressionUUID->"d868bf2b-aff7-40e2-bbf1-9e3db75aadc8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is antisymmetric in the first group of indices and separately \
antisymmetric in the second group of indices. Note that there is no \
group-exchange symmetry because there is no metric on this vbundle yet.\
\>", "MathCaption",ExpressionUUID->"e15c1a03-1540-4b6b-b08d-72c240c353f8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"Gdelta", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "a"}], ",", "c", ",", "d"}], "]"}], ",", 
    RowBox[{"Gdelta", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "d", ",", "c"}], "]"}], ",", 
    RowBox[{"Gdelta", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "c"}], ",", "b", ",", "a"}], "]"}]}], "}"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[190]:=",ExpressionUUID->"e12c9aee-b83d-4156-a0ff-03a9cbff5b0d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", " ", "c", "d"},
            {"a", "b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`d],
     Editable->False]}], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", " ", "c", "d"},
            {"a", "b", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, $CellContext`c, \
$CellContext`d],
     Editable->False]}], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", " ", "a", "b"},
           {"c", "d", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-$CellContext`c, -$CellContext`d, $CellContext`a, \
$CellContext`b],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728977845523748*^9},
 CellLabel->"Out[190]=",
 CellID->364818616,ExpressionUUID->"f53f8d93-7578-4196-bcc5-0cf42b58e1f7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Staggered indices are automatically contracted (there is no need of a metric \
for this):\
\>", "MathCaption",ExpressionUUID->"72cbd1b8-6a92-4701-b5bc-d0c4cc2f4075"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Gdelta", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "a", ",", "b", ",", "c"}], "]"}]], "Input",
 CellLabel->
  "In[191]:=",ExpressionUUID->"98ef51f8-6244-49af-a06a-656ae2311db1"],

Cell[BoxData["6"], "Output",
 CellChangeTimes->{3.7289778456619368`*^9},
 CellLabel->"Out[191]=",
 CellID->440666927,ExpressionUUID->"210705cc-2b56-43d1-a859-0ac1785174df"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Gdelta", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "d", ",", "e", ",", "c"}], "]"}]], "Input",
 CellLabel->
  "In[192]:=",ExpressionUUID->"9dd30a31-f8ea-4d1f-850b-861ccf17b282"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", " ", "d", "e"},
         {"a", "b", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, $CellContext`d, \
$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977845907393*^9},
 CellLabel->"Out[192]=",
 CellID->1765427233,ExpressionUUID->"c554b6b6-5dfa-44e8-b4e7-10c6016bae22"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Contractions in the same group give zero. A metric is actually needed here, \
and Gdelta does not check its existence.\
\>", "MathCaption",ExpressionUUID->"3b3bc965-2ee0-495d-be14-a74b91891f73"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Gdelta", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", "a", ",", 
   RowBox[{"-", "c"}], ",", "d", ",", "e"}], "]"}]], "Input",
 CellLabel->
  "In[193]:=",ExpressionUUID->"0021ce72-5cab-4323-85fb-230a79a5d574"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977846045349*^9},
 CellLabel->"Out[193]=",
 CellID->1901438079,ExpressionUUID->"839bb3e3-92a8-4c0b-9bd2-5d144a5e3219"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is always possible to convert it into a determinant of normal delta \
tensors:\
\>", "MathCaption",ExpressionUUID->"0b0ba427-fb05-4c18-9116-1fdc8f0ef43e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ExpandGdelta", "[", 
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", "d", ",", "e"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[194]:=",ExpressionUUID->"7e41b45b-70d7-4105-94bd-5f73ee803904"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "e"},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`a, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728977846499954*^9},
 CellLabel->"Out[194]=",
 CellID->765642746,ExpressionUUID->"ee0cc355-b587-47ed-be74-c069785507dc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Conversion into delta is automatic:", "MathCaption",ExpressionUUID->"25d26999-d31f-4671-be2c-11c244552727"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "b"}], "]"}], "//", "InputForm"}]], "Input",
 CellLabel->
  "In[195]:=",ExpressionUUID->"3d9ab215-f7d9-4e34-9a2a-3c2af8cdc9bf"],

Cell["delta[-a, b]", "Output",
 CellChangeTimes->{3.728977846727206*^9},
 CellLabel->"Out[195]//InputForm=",
 CellID->435619198,ExpressionUUID->"e116bce3-18ed-4cb6-9378-2281b5458da8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can contract two generalized deltas into another generalized delta:\
\>", "MathCaption",ExpressionUUID->"28251a08-e568-4e50-881a-657ad1f6af19"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", "c", ",", "d"}], "]"}], 
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}], ",", "e", ",", "f"}], "]"}]}]], "Input",
 CellLabel->
  "In[196]:=",ExpressionUUID->"09a691f4-571b-48ec-9bad-578ffb55b9b5"],

Cell[BoxData[
 RowBox[{"2", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[Delta]", 
       StyleBox[GridBox[{
          {" ", " ", "e", "f"},
          {"a", "b", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, $CellContext`e, \
$CellContext`f],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289778468670607`*^9},
 CellLabel->"Out[196]=",
 CellID->680108163,ExpressionUUID->"14fc77f9-3179-4615-b67c-5c8c28d7c5af"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "d", ",", "e", ",", "f"}], "]"}], 
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}], ",", 
    RowBox[{"-", "f"}], ",", "g", ",", "h", ",", "h1"}], "]"}]}]], "Input",
 CellLabel->
  "In[197]:=",ExpressionUUID->"16151d9c-97cf-49e7-bad8-bdc4a3b1176d"],

Cell[BoxData[
 RowBox[{"6", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[Delta]", 
       StyleBox[GridBox[{
          {" ", " ", " ", "g", "h", "h1"},
          {"a", "b", "c", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`g, $CellContext`h, $CellContext`h1],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977846981103*^9},
 CellLabel->"Out[197]=",
 CellID->1669926207,ExpressionUUID->"91962f4e-9b68-4ac4-831d-f1487261b83c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "d", ",", "e", ",", "f"}], "]"}], 
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}], ",", 
    RowBox[{"-", "g"}], ",", "h", ",", "h1", ",", "h2"}], "]"}]}]], "Input",
 CellLabel->
  "In[198]:=",ExpressionUUID->"10c5c03d-e424-4675-823e-44b6aaca6fe9"],

Cell[BoxData[
 RowBox[{"2", " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", " ", " ", "h", "h1", "h2"},
             {"a", "b", "g", " ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`b, -$CellContext`g, \
$CellContext`h, $CellContext`h1, $CellContext`h2],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", " ", " ", "h", "h1", "h2"},
             {"a", "c", "g", " ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Gdelta[-$CellContext`a, -$CellContext`c, -$CellContext`g, \
$CellContext`h, $CellContext`h1, $CellContext`h2],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", " ", " ", "h", "h1", "h2"},
             {"b", "c", "g", " ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Gdelta[-$CellContext`b, -$CellContext`c, -$CellContext`g, \
$CellContext`h, $CellContext`h1, $CellContext`h2],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.7289778474838247`*^9},
 CellLabel->"Out[198]=",
 CellID->985978406,ExpressionUUID->"a7cda249-47e2-48fb-a5aa-7a0add85a3e9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us check this last result.", "MathCaption",ExpressionUUID->"5a9f661f-51fa-4b51-af0a-47294192773f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ExpandGdelta", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[199]:=",ExpressionUUID->"e8e57c9a-eddb-49a3-90e6-c5b2161937b4"],

Cell[BoxData[
 RowBox[{"2", " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
         Editable->False]}]}], ")"}]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
         Editable->False]}]}], ")"}]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h"},
                {"b", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h1"},
                {"c", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "h2"},
                {"g", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
         Editable->False]}]}], ")"}]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728977848052606*^9},
 CellLabel->"Out[199]=",
 CellID->1355249390,ExpressionUUID->"a362d989-54ec-4920-aa0b-6ebcfadb8aef"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "-", 
  RowBox[{
   RowBox[{"ExpandGdelta", "[", 
    RowBox[{"Gdelta", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", "d", ",", "e", ",", "f"}], "]"}], "]"}], 
   RowBox[{"ExpandGdelta", "[", 
    RowBox[{"Gdelta", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}], ",", 
      RowBox[{"-", "g"}], ",", "h", ",", "h1", ",", "h2"}], "]"}], 
    "]"}]}]}]], "Input",
 CellLabel->
  "In[200]:=",ExpressionUUID->"d463af76-bfd5-436a-8d5a-7ebf8b92edc6"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {" ", "f"},
                {"a", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
         Editable->False]}], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "e"},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`c, $CellContext`d],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "e"},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "f"},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`c, $CellContext`d],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "f"},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`b, $CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "e"},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`c, $CellContext`e],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`a, $CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "f"},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "e"},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`c, $CellContext`e],
        Editable->False]}], "-", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "e"},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`b, $CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "f"},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
        Editable->False]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`a, $CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "e"},
               {"b", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "f"},
               {"c", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
        Editable->False]}]}], ")"}]}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "h2"},
               {"d", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-$CellContext`d, $CellContext`h2],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h1"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`h1],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h"},
              {"g", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h1"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`d, $CellContext`h1],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h2"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`h2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h"},
              {"g", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h2"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`d, $CellContext`h2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h1"},
              {"g", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`d, $CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h2"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`h2],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h1"},
              {"g", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h1"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`d, $CellContext`h1],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h2"},
              {"g", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`d, $CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h1"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`h1],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "h2"},
              {"g", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "f"},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"\[Delta]", 
               StyleBox[GridBox[{
                  {" ", "h2"},
                  {"a", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
          Editable->False]}]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "f"},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"\[Delta]", 
               StyleBox[GridBox[{
                  {" ", "h2"},
                  {"a", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"a", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
          Editable->False]}]}], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "f"},
              {"a", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"\[Delta]", 
               StyleBox[GridBox[{
                  {" ", "h2"},
                  {"b", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h2],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h1],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
          Editable->False]}], "+", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h"},
                 {"b", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h1"},
                 {"c", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`c, $CellContext`h1],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"\[Delta]", 
              StyleBox[GridBox[{
                 {" ", "h2"},
                 {"g", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`delta[-$CellContext`g, $CellContext`h2],
          Editable->False]}]}], ")"}]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728977848560082*^9},
 CellLabel->"Out[200]=",
 CellID->1464937648,ExpressionUUID->"cf1b929f-8cc6-4925-8133-f37dbd7acf27"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[201]:=",ExpressionUUID->"fa5ae39d-4fb6-4938-a829-4150b5c4ade0"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977848900332*^9},
 CellLabel->"Out[201]=",
 CellID->307037081,ExpressionUUID->"29b782fa-86be-469d-bbef-006de51552bd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can also contract generalized deltas with antisymmetric tensors, without \
expanding into standard deltas:\
\>", "MathCaption",ExpressionUUID->"e2e35eb6-332c-4f4e-8d88-9eb3057813e4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[202]:=",ExpressionUUID->"76ecc4d4-f804-4230-9c93-e82e2a1c80c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"U", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897784906539*^9},
 CellLabel->"During evaluation of In[202]:=",
 CellID->2116500273,ExpressionUUID->"43303e10-cfc4-41c9-bb86-f92a0c2797fa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "e", ",", "f", ",", "g"}], "]"}]}]], "Input",
 CellLabel->
  "In[203]:=",ExpressionUUID->"d1e0b314-4b3b-4963-b343-05780c425cd9"],

Cell[BoxData[
 RowBox[{"6", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"e", "f", "g"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`e, $CellContext`f, $CellContext`g],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289778492329893`*^9},
 CellLabel->"Out[203]=",
 CellID->1100959403,ExpressionUUID->"6ff29cb9-2753-410d-9f64-3fdf16552c45"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Again, let us check the result:", "MathCaption",ExpressionUUID->"86d20593-4145-416e-91d0-37390eb6c98c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{"ExpandGdelta", "[", 
   RowBox[{"Gdelta", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "e", ",", "f", ",", "g"}], "]"}], 
   "]"}]}]], "Input",
 CellLabel->
  "In[204]:=",ExpressionUUID->"a3a4479d-daa2-4923-8612-a8a2a7906341"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "g"},
              {"a", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`a, $CellContext`g],
       Editable->False]}], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`e],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`e],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`g],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`g],
      Editable->False]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977849948433*^9},
 CellLabel->"Out[204]=",
 CellID->1105583669,ExpressionUUID->"2c26fab6-a96e-462f-8340-145e0ee3c7f2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[205]:=",ExpressionUUID->"66652138-f1d8-4903-b1f6-e3d5da7825d4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"e", "f", "g"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`e, $CellContext`f, $CellContext`g],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"e", "g", "f"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`e, $CellContext`g, $CellContext`f],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"f", "e", "g"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`f, $CellContext`e, $CellContext`g],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"f", "g", "e"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`f, $CellContext`g, $CellContext`e],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"g", "e", "f"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`g, $CellContext`e, $CellContext`f],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"g", "f", "e"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`g, $CellContext`f, $CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977850113667*^9},
 CellLabel->"Out[205]=",
 CellID->1516211039,ExpressionUUID->"be4c3da9-440e-4b89-a4a9-598f0b07f3d6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[206]:=",ExpressionUUID->"8791f0ad-41ea-4796-9cd5-530d3a6e8512"],

Cell[BoxData[
 RowBox[{"6", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"e", "f", "g"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`e, $CellContext`f, $CellContext`g],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977850278228*^9},
 CellLabel->"Out[206]=",
 CellID->917977171,ExpressionUUID->"41c15348-4ae4-4a05-b22a-62dc81984341"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "U", "]"}]], "Input",
 CellLabel->
  "In[207]:=",ExpressionUUID->"47b8b85c-2421-4c20-b56d-adc94d7c0e33"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977850580768*^9},
 CellLabel->"During evaluation of In[207]:=",
 CellID->919792949,ExpressionUUID->"c36347f0-a8d6-43d2-bc01-bc2234b6976a"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In version 1.0.0, ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has added yet another delta tensor, the \"superdelta\", which accepts any \
symmetry group. This object is still under development."
}], "Text",ExpressionUUID->"bc3c20b0-8a9d-47cf-863d-7c09c054085c"],

Cell[CellGroupData[{

Cell["\<\
This tensor projects any other tensor onto an expression with the symmetries \
of the Riemann tensor :\
\>", "MathCaption",
 CellChangeTimes->{{3.4946082225858593`*^9, 
  3.494608249178817*^9}},ExpressionUUID->"699aa6cc-c06d-4108-99d4-\
60028dfac6c1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Sdelta", "[", 
   RowBox[{"RiemannSymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}], "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}], ",", "e", ",", "f", ",", "g", ",", "h"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.494608169706142*^9, 3.494608189645891*^9}},
 CellLabel->
  "In[208]:=",ExpressionUUID->"f76963c5-a5ff-4f92-bfd8-1ddbb8ba850c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", " ", " ", " ", "e", "f", "g", "h"},
         {"a", "b", "c", "d", " ", " ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Sdelta[
   xAct`xPerm`StrongGenSet[{1, 2, 3, 4}, 
    xAct`xPerm`GenSet[
     xAct`xPerm`Cycles[{1, 3}, {2, 4}], -xAct`xPerm`Cycles[{1, 2}], -
     xAct`xPerm`Cycles[{3, 
       4}]]]][-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d, $CellContext`e, $CellContext`f, $CellContext`g, \
$CellContext`h],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977850934519*^9},
 CellLabel->"Out[208]=",
 CellID->1372802737,ExpressionUUID->"9a3eba78-774d-450c-84e9-4c10988d96c9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ExpandSdelta"}]], "Input",
 CellChangeTimes->{{3.494608209390026*^9, 3.4946082133171883`*^9}},
 CellLabel->
  "In[209]:=",ExpressionUUID->"06139917-d37f-46b5-8d9b-afe2b7303981"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "8"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`e],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`e],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`h],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`f],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`h],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`f],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`g],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`h],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`g],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`h],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "f"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "g"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`c, $CellContext`g],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "h"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`d, $CellContext`h],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728977851197719*^9},
 CellLabel->"Out[209]=",
 CellID->131297391,ExpressionUUID->"47795137-627c-42fb-8617-858465a90876"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.6. Directional indices", "Subsection",ExpressionUUID->"cee63d08-2f11-43e6-a4bc-9b8db80f5a44"],

Cell[TextData[{
 "From the abstract point of view, covariant (contravariant) indices of \
tensors represent slots where vectors (covectors) can be \"contracted\". It \
is interesting sometimes to write that contraction explicitly: ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " uses the head ",
 StyleBox["Dir",
  FontFamily->"Courier"],
 " to denote that kind of \"directional index\". The introduction of a \
special head is a technical decision: we expect this feature to be seldom \
used and therefore it makes no sense to force ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " to check continuously whether a given slot contains an ab",
 "stract ind",
 "ex or a vector. All definitions concerning directional-indices are \
associated to ",
 StyleBox["Dir",
  FontFamily->"Courier"],
 ", and hence they are not even considered unless the head ",
 StyleBox["Dir",
  FontFamily->"Courier"],
 " is found."
}], "Text",ExpressionUUID->"dae88e9f-94c7-466a-b8a0-987970ef7265"],

Cell[TextData[{
 StyleBox[ButtonBox["Dir",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Dir.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t\t", "DisplayMath"],
 StyleBox["Head denoting directional slots in tensors", "Text"]
}], "DefinitionBox",ExpressionUUID->"1ff6fab4-5050-41f8-b2f5-d1b26ae8b70f"],

Cell["Directional indices.", "Text",ExpressionUUID->"4882a813-adcd-480b-9aaf-e30e26795a3b"],

Cell[CellGroupData[{

Cell["This is the usual representation for a contraction:", "MathCaption",ExpressionUUID->"092e5e45-7da0-4d07-a504-9698540dd06e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"-", "b"}], "]"}], 
  RowBox[{"S", "[", 
   RowBox[{"a", ",", "b"}], "]"}]}]], "Input",
 CellLabel->
  "In[210]:=",ExpressionUUID->"7120c12d-c856-431e-a321-b43f633d8a94"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977861231162*^9},
 CellLabel->"Out[210]=",
 CellID->923251014,ExpressionUUID->"423213de-2dc4-40e8-ab9a-46d3a6588a72"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
But it can also be given as (note the different results of the validation \
routine below)\
\>", "MathCaption",ExpressionUUID->"05a19e42-7531-406c-a3df-d2c31fa5bbe0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"S", "[", 
  RowBox[{"a", ",", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}], "]"}]}], " ", "]"}]], "Input",
 CellLabel->
  "In[211]:=",ExpressionUUID->"6ccf66fd-e2b0-48b9-ba72-1da7e66739ab"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "v"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[-$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977861366091*^9},
 CellLabel->"Out[211]=",
 CellID->1510442723,ExpressionUUID->"0bd88b6e-44c7-4e12-800c-d09d143242fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[212]:=",ExpressionUUID->"84a69629-63e3-467d-94d2-cffbfeedea99"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor v\\\\\
\\\"\\\"}]\\)\"", 2, 212, 25, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977861514696*^9},
 CellLabel->"During evaluation of In[212]:=",
 CellID->606654274,ExpressionUUID->"29eda692-6e50-417b-8e3d-5f0fdce60f07"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "v"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[-$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977861544901*^9},
 CellLabel->"Out[212]=",
 CellID->1933510809,ExpressionUUID->"8824e6ab-bda5-48d9-83ef-c50c9dcea625"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Directional slots containing vectors are represented downstairs, and slots \
containing covectors are represented upstairs:\
\>", "MathCaption",ExpressionUUID->"a4ebcff7-4407-4381-9825-209611ee90b4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"S", "[", 
  RowBox[{"a", ",", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", "c", "]"}], "]"}]}], " ", "]"}]], "Input",
 CellLabel->
  "In[213]:=",ExpressionUUID->"d3df5355-2307-473c-a818-af8f0aea3651"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "v"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977861665*^9},
 CellLabel->"Out[213]=",
 CellID->793620459,ExpressionUUID->"3a8634b7-4bb6-4d11-a99f-57a34efeded1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[214]:=",ExpressionUUID->"2ad9d6db-cc99-4e6b-8ffb-62caf9d74a9a"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor S\\\\\
\\\"\\\"}]\\)\"", 2, 214, 26, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289778619528*^9},
 CellLabel->"During evaluation of In[214]:=",
 CellID->358122896,ExpressionUUID->"8de8e366-bc8a-45ba-afd4-5208958efdb1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["ERROR",
     FontColor->Hue[0.]], "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", " "},
            {" ", "v"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, 
      xAct`xTensor`Dir[
       $CellContext`v[$CellContext`c]]],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`ERROR[
   $CellContext`S[$CellContext`a, 
    xAct`xTensor`Dir[
     $CellContext`v[$CellContext`c]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977861983584*^9},
 CellLabel->"Out[214]=",
 CellID->1722334178,ExpressionUUID->"e0a7ee8c-6c48-4550-b190-f04da1e89f06"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"S", "[", 
  RowBox[{"a", ",", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}], "]"}]}], " ", "]"}]], "Input",
 CellLabel->
  "In[215]:=",ExpressionUUID->"b8f584b9-a518-4ae7-9e9f-96139f65c0a9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "v"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[-$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977862353179*^9},
 CellLabel->"Out[215]=",
 CellID->1338451733,ExpressionUUID->"4189be2a-852b-41da-8ad0-f4180255af00"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[216]:=",ExpressionUUID->"366ee27f-f619-4c08-9667-7cc58346bbd9"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor v\\\\\
\\\"\\\"}]\\)\"", 2, 216, 27, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977862669281*^9},
 CellLabel->"During evaluation of In[216]:=",
 CellID->467074206,ExpressionUUID->"65dcb444-608b-4a8b-9a2d-070df1c03e24"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "v"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[-$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977862700863*^9},
 CellLabel->"Out[216]=",
 CellID->1063450457,ExpressionUUID->"1386a5b6-b773-40cc-bbd2-74c5435b44e9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It has the expected properties for a tensorial slot:", "MathCaption",ExpressionUUID->"c37cbf54-17b1-4660-ac3c-48e8fce6feca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[217]:=",ExpressionUUID->"39c61074-f698-423c-b63e-fbdf19c13967"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"w", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`w[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977862969079*^9},
 CellLabel->"During evaluation of In[217]:=",
 CellID->2052920465,ExpressionUUID->"81c123f3-d8b8-4bf8-b2bb-b85a5ee6e89a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"S", "[", 
  RowBox[{"a", ",", 
   RowBox[{"Dir", "[", 
    RowBox[{
     RowBox[{"3", 
      RowBox[{"v", "[", 
       RowBox[{"-", "c"}], "]"}]}], "+", 
     RowBox[{"w", "[", 
      RowBox[{"-", "c"}], "]"}]}], "]"}]}], " ", "]"}]], "Input",
 CellLabel->
  "In[218]:=",ExpressionUUID->"b8e3ceec-38d5-4fb2-aa3b-889cd257c8ee"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "v"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, 
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`c]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "w"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, 
    xAct`xTensor`Dir[
     $CellContext`w[-$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977863100268*^9},
 CellLabel->"Out[218]=",
 CellID->1961982553,ExpressionUUID->"21651e12-eb84-4a6a-bb97-4d85a7b5065e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "w", "]"}]], "Input",
 CellLabel->
  "In[219]:=",ExpressionUUID->"3716cbf8-85e4-4d70-8284-d73d1f695200"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "w"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`w],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977863227611*^9},
 CellLabel->"During evaluation of In[219]:=",
 CellID->429368586,ExpressionUUID->"db977d41-2503-4ec2-86d6-5c720a6f0297"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In particular, ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " understands the whole expression as a vector, and not as a 2-tensor:"
}], "MathCaption",ExpressionUUID->"52eb0882-e869-4c24-8ba9-47a038d4b5b3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UpVectorQ", "[", 
  RowBox[{"S", "[", 
   RowBox[{"a", ",", 
    RowBox[{"Dir", "[", 
     RowBox[{"v", "[", 
      RowBox[{"-", "c"}], "]"}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[220]:=",ExpressionUUID->"67cd41f0-8aeb-4eff-a6f6-699f8c96edb4"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977863348957*^9},
 CellLabel->"Out[220]=",
 CellID->852770982,ExpressionUUID->"908c2215-0161-4142-a94a-02eba1119bcc"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "It is important to realize that the free index ",
 StyleBox["c",
  FontFamily->"Courier"],
 " in ",
 StyleBox["Dir[v[c]]",
  FontFamily->"Courier"],
 " is completely irrelevant except for its sign and the tangent bundle it \
belongs to. It is there for consistency reasons: it identifies the argument \
of ",
 StyleBox["Dir",
  FontFamily->"Courier"],
 " as an upvector on ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 ". I call those indices \"ultraindices\"."
}], "Text",ExpressionUUID->"6608ce0a-e5ee-491a-8663-9165f031c619"],

Cell[CellGroupData[{

Cell[TextData[{
 "If you find annoying to write ",
 StyleBox["Dir[v[c]]",
  FontFamily->"Courier"],
 " many times, you can name this expression:"
}], "MathCaption",ExpressionUUID->"742c9853-0aee-4193-aeeb-203e626202cf"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"dirv", "=", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", "c", "]"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"S", "[", 
  RowBox[{"a", ",", "dirv"}], "]"}]}], "Input",
 CellLabel->
  "In[221]:=",ExpressionUUID->"d278930a-0155-492b-9fcb-723efc59c857"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "v"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977863580779*^9},
 CellLabel->"Out[222]=",
 CellID->216034396,ExpressionUUID->"dc01e41e-9864-4c36-bcc1-39a215b4ed7c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Finally, there is a pair of functions to convert the Dir indices into normal \
contractions, and viceversa :\
\>", "MathCaption",
 CellChangeTimes->{{3.461934006077577*^9, 
  3.4619340174393377`*^9}},ExpressionUUID->"41c015fd-f582-4c55-90f0-\
e6f4af8adef3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SeparateDir", "[", 
   RowBox[{"%", ",", "v"}], "]"}], "//", "ScreenDollarIndices"}]], "Input",
 CellChangeTimes->{{3.4619340245346127`*^9, 3.461934024837811*^9}},
 CellLabel->
  "In[223]:=",ExpressionUUID->"abb6c5fc-d64a-458f-acc4-d056b6fe3fce"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", " "},
          {" ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977863765894*^9},
 CellLabel->"Out[223]=",
 CellID->1642730142,ExpressionUUID->"f2107841-ef55-43ad-9162-bc3066ef2b42"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractDir", "[", 
  RowBox[{"%", ",", "v"}], "]"}]], "Input",
 CellChangeTimes->{{3.461934029883997*^9, 3.461934034911304*^9}},
 CellLabel->
  "In[224]:=",ExpressionUUID->"6047e091-bcef-4164-85c5-fa5638b79224"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "v"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`b]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977863883071*^9},
 CellLabel->"Out[224]=",
 CellID->1315730762,ExpressionUUID->"ec13ccc4-8d12-4a50-b4e1-b805742c0685"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "dirv", "]"}]], "Input",
 CellLabel->
  "In[225]:=",ExpressionUUID->"1dea45a0-20ba-4a4e-8326-86027e224e7d"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.7. Label indices", "Subsection",ExpressionUUID->"5bcbb47f-5dd2-4b4c-a475-129578912a65"],

Cell[TextData[{
 "Sometimes we want to have indices with no tensorial meaning. For instance \
the labels l, m of the spherical harmonics can be defined as label indices. \
This is done using a new head ",
 StyleBox["LI",
  FontFamily->"Courier"],
 ", following similar ideas to those bringing us to the head ",
 StyleBox["Dir",
  FontFamily->"Courier"],
 ". There are no definitions associated to ",
 StyleBox["LI",
  FontFamily->"Courier"],
 "; the user can use ",
 StyleBox["LI",
  FontFamily->"Courier"],
 " for any purpose he likes."
}], "Text",ExpressionUUID->"cee59409-3b4a-4d59-822f-63cc0832e3e6"],

Cell[TextData[{
 StyleBox[ButtonBox["LI",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LI.nb", CharacterEncoding -> "iso8859-1"],
     None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t\t", "DisplayMath"],
 StyleBox["Head denoting label slots in tensors\n", "Text"],
 StyleBox[ButtonBox["Labels",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Labels.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "Text",
  FontFamily->"Courier"],
 StyleBox["A ficticious vbundle to which all label indices belong", "Text"]
}], "DefinitionBox",ExpressionUUID->"f77ec4b8-27b1-4d56-9f5b-4d842fa65f7d"],

Cell["Label indices.", "Text",ExpressionUUID->"2cc6deed-6db7-4c12-b135-12b8060806f1"],

Cell[CellGroupData[{

Cell["We can define the spherical harmonics as", "MathCaption",ExpressionUUID->"d0c96438-ab0b-4de1-847b-cf9aaea23a4a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Y", "[", 
    RowBox[{
     RowBox[{"LI", "@", "l"}], ",", 
     RowBox[{"LI", "@", "m"}]}], "]"}], ",", "S2"}], "]"}]], "Input",
 CellLabel->
  "In[226]:=",ExpressionUUID->"3917762c-0856-45e2-8a62-473dccb37ee9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Y", "[", 
    RowBox[{
     RowBox[{"LI", "[", "l", "]"}], ",", 
     RowBox[{"LI", "[", "m", "]"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Y[
    xAct`xTensor`LI[$CellContext`l], 
    xAct`xTensor`LI[$CellContext`m]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977869603508*^9},
 CellLabel->"During evaluation of In[226]:=",
 CellID->736429224,ExpressionUUID->"9bfd987d-3644-454e-bd9f-7362b6c6ed41"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Y"}]], "Input",
 CellLabel->
  "In[227]:=",ExpressionUUID->"9201c517-efba-4712-b236-84adf1bbfe92"],

Cell[CellGroupData[{

Cell["Global`Y", "Print", "PrintUsage",
 CellChangeTimes->{3.7289778697823763`*^9},
 CellTags->"Info2273728981469-9181623",
 CellID->1217710107,ExpressionUUID->"d09061f2-d245-4708-9251-86b611186377"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "Y", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`Y],
          Editable->False], "^=", "Y"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "Y", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "Y", "]"}], "^=", 
         RowBox[{"{", "S2", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "Y", "]"}], "^=", 
         RowBox[{"{", "S2", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "Y", "]"}], "^=", "\<\"Y\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "Y", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["xAct`xTensor`Labels",
             ShowAutoStyles->False,
             AutoSpacing->False],
            xAct`xTensor`Labels,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["xAct`xTensor`Labels",
             ShowAutoStyles->False,
             AutoSpacing->False],
            xAct`xTensor`Labels,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "Y", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "Y", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "Y", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977869806344*^9},
 CellTags->"Info2273728981469-9181623",
 CellID->534612439,ExpressionUUID->"17855b62-38ea-437b-aaa7-7df201779033"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The up/down character of the labels only indicates the output position of \
the indices. Note that the signs inside ",
 StyleBox["LI",
  FontFamily->"Courier"],
 " are part of the label, and do not indicate the character."
}], "MathCaption",ExpressionUUID->"2b50a1c0-7f05-4f3c-a83b-e3913ff7ab34"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Y", "[", 
  RowBox[{
   RowBox[{"LI", "@", "L"}], ",", 
   RowBox[{"LI", "@", "M"}]}], "]"}]], "Input",
 CellLabel->
  "In[228]:=",ExpressionUUID->"323759bd-c4a0-4395-8b17-1d69a22c20e1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"Y", 
      StyleBox[GridBox[{
         {"L", "M"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Y[
   xAct`xTensor`LI[$CellContext`L], 
   xAct`xTensor`LI[$CellContext`M]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977869925173*^9},
 CellLabel->"Out[228]=",
 CellID->159683225,ExpressionUUID->"342c4704-c948-4176-ae45-511858e7b0e4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Y", "[", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"LI", "[", "L", "]"}]}], ",", 
     RowBox[{"-", 
      RowBox[{"LI", "[", "2", "]"}]}]}], "]"}], ",", 
   RowBox[{"Y", "[", 
    RowBox[{
     RowBox[{"LI", "[", 
      RowBox[{"-", "L"}], "]"}], ",", 
     RowBox[{"LI", "[", 
      RowBox[{"-", "2"}], "]"}]}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[229]:=",ExpressionUUID->"4380e29d-41c7-48a5-af56-629b92f3aab4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Y", 
        StyleBox[GridBox[{
           {" ", " "},
           {"L", "2"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Y[-xAct`xTensor`LI[$CellContext`L], -xAct`xTensor`LI[2]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Y", 
        StyleBox[GridBox[{
           {
            RowBox[{"-", "L"}], 
            RowBox[{"-", "2"}]},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Y[
     xAct`xTensor`LI[-$CellContext`L], 
     xAct`xTensor`LI[-2]],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.72897787004317*^9},
 CellLabel->"Out[229]=",
 CellID->1588887472,ExpressionUUID->"2c7a8f35-cb62-4308-9956-320fe89a669b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "Y", "]"}]], "Input",
 CellLabel->
  "In[230]:=",ExpressionUUID->"399acf47-e35f-47a7-b7de-9b5cc6a2ab37"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Y"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Y],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977870184417*^9},
 CellLabel->"During evaluation of In[230]:=",
 CellID->472591459,ExpressionUUID->"396107e9-6a05-45c0-b1ed-f6579ab9a925"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In version 1.0.0 ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has added the possibility of having tensors with nonatomic heads. However, \
this is not accepted in ",
 StyleBox["DefTensor",
  FontFamily->"Courier"],
 ", which means that we need to declare by hand all properties of the tensors."
}], "Text",ExpressionUUID->"8b831627-c064-4d91-b9df-20c4447657ba"],

Cell[CellGroupData[{

Cell["\<\
For example, we could have (symmetric) tensor spherical harmonics, with a \
variable number of indices, expressed as (note the use of upvalues)\
\>", "MathCaption",
 CellChangeTimes->{{3.494608378815723*^9, 3.494608393260377*^9}, {
  3.494608430088417*^9, 3.494608432445478*^9}, {3.494608499378916*^9, 
  3.494608503771892*^9}, {3.494608595644052*^9, 3.494608621812394*^9}, {
  3.4946086559276447`*^9, 
  3.494608660664278*^9}},ExpressionUUID->"05837fd2-6990-400e-85af-\
320fda22566d"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"xTensorQ", "[", 
    RowBox[{"Y", "[", 
     RowBox[{"l_", ",", "m_", ",", "s_"}], "]"}], "]"}], "^=", "True"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SlotsOfTensor", "[", 
    RowBox[{"Y", "[", 
     RowBox[{"l_", ",", "m_", ",", "s_"}], "]"}], "]"}], "^:=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"-", "TangentS2"}], ",", 
     RowBox[{"{", "s", "}"}]}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"DependenciesOfTensor", "[", 
    RowBox[{"Y", "[", 
     RowBox[{"l_", ",", "m_", ",", "s_"}], "]"}], "]"}], "^=", 
   RowBox[{"{", "S2", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SymmetryGroupOfTensor", "[", 
    RowBox[{
     RowBox[{"Y", "[", 
      RowBox[{"l_", ",", "m_", ",", "s_"}], "]"}], "[", "inds___", "]"}], 
    "]"}], "^:=", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"Range", "[", 
     RowBox[{"Length", "[", 
      RowBox[{"{", "inds", "}"}], "]"}], "]"}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"PrintAs", "[", 
    RowBox[{"Y", "[", 
     RowBox[{"l_", ",", "m_", ",", "s_"}], "]"}], "]"}], "^:=", 
   RowBox[{"SubscriptBox", "[", 
    RowBox[{"\"\<Y\>\"", ",", 
     RowBox[{"ToBoxes", "[", 
      RowBox[{"{", 
       RowBox[{"l", ",", "m", ",", "s"}], "}"}], "]"}]}], "]"}]}], 
  ";"}]}], "Input",
 CellLabel->
  "In[231]:=",ExpressionUUID->"fd427259-6fb3-4f69-a6e1-5b856b707520"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Y", "[", 
   RowBox[{"3", ",", 
    RowBox[{"-", "2"}], ",", "1"}], "]"}], "[", 
  RowBox[{"a", ",", "b", ",", "c"}], "]"}]], "Input",
 CellChangeTimes->{{3.4946084074152927`*^9, 3.494608427147748*^9}, {
  3.4946086661771183`*^9, 3.494608680361247*^9}},
 CellLabel->
  "In[236]:=",ExpressionUUID->"ee6e225e-5363-4f55-ab8e-7e15915a4868"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      SubscriptBox["Y", 
       RowBox[{"{", 
        RowBox[{"3", ",", 
         RowBox[{"-", "2"}], ",", "1"}], "}"}]], 
      StyleBox[GridBox[{
         {"a", "b", "c"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Y[3, -2, 1][$CellContext`a, $CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977870617581*^9},
 CellLabel->"Out[236]=",
 CellID->1450124807,ExpressionUUID->"c2f754ea-eadb-440f-acd9-a9d8b5653437"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Sometimes we do not need to define values for all functions. The only \
essential functions are ",
 StyleBox["xTensorQ",
  FontFamily->"Courier"],
 ", ",
 StyleBox["SlotsOfTensor",
  FontFamily->"Courier"],
 " and",
 StyleBox[" DependenciesOfTensor",
  FontFamily->"Courier"],
 ", that identify the tensor field character of the object. Then ",
 StyleBox["SymmetryGroupOfTensor",
  FontFamily->"Courier"],
 " is needed if the tensor has a symmetry, and ",
 StyleBox["PrintAs",
  FontFamily->"Courier"],
 " if special formatting is needed. The only other functions that might be \
needed are ",
 StyleBox["Dagger",
  FontFamily->"Courier"],
 " and ",
 StyleBox["WeightOfTensor",
  FontFamily->"Courier"],
 ", the latter having a default of zero. The remaining function are used by \
",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " for bookkeping purposes, namely ",
 StyleBox["DefInfo",
  FontFamily->"Courier"],
 ", ",
 StyleBox["HostsOf",
  FontFamily->"Courier"],
 " and ",
 StyleBox["TensorID",
  FontFamily->"Courier"],
 ". These non-atomic-head tensors cannot be defined via ",
 StyleBox["DefTensor",
  FontFamily->"Courier"],
 ". They must necessarily be defined by hand."
}], "Text",ExpressionUUID->"8ac046d8-9cb4-4b92-ba43-d7ee9f63bc36"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "Y", ",", "l", ",", "m", ",", "s", ",", "inds", ",", "L", ",", "M"}], 
  "]"}]], "Input",
 CellLabel->
  "In[237]:=",ExpressionUUID->"d840339c-6241-45e3-962e-97933d64631d"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.8. Complex conjugation", "Subsection",ExpressionUUID->"ea33d657-7f0c-4e11-b67d-e460d1a2abaf"],

Cell[TextData[{
 "Complex conjugation is performed by the function ",
 StyleBox["Dagger",
  FontFamily->"Courier"],
 ", except on indices, where we shall use the variant ",
 StyleBox["DaggerIndex",
  FontFamily->"Courier"],
 ", to avoid overloading ",
 StyleBox["Dagger",
  FontFamily->"Courier"],
 " too much."
}], "Text",ExpressionUUID->"eb2c7916-e5c5-4dde-a668-0bc55813358f"],

Cell[TextData[{
 StyleBox[ButtonBox["Dagger",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Dagger.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Complex conjugation of tensor expressions\n", "Text"],
 StyleBox[ButtonBox["DaggerIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DaggerIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Complex conjugation of indices\n", "Text"],
 StyleBox[ButtonBox["DaggerQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DaggerQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check for complex objects\n", "Text"],
 StyleBox[ButtonBox["TransposeDagger",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TransposeDagger.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Transposition of indices of complex conjugate vbundles\n", 
  "Text"],
 StyleBox[ButtonBox["$DaggerCharacter",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$DaggerCharacter.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t", "DisplayMath"],
 StyleBox["Character which denotes complex conjugates of other objects", 
  "Text"]
}], "DefinitionBox",ExpressionUUID->"159dc468-ea03-412f-b8d2-10faa70ec4a4"],

Cell["Complex conjugation", "Text",ExpressionUUID->"687e2a48-aee8-46e3-88a2-b01d2d89ee12"],

Cell[CellGroupData[{

Cell["\<\
A tensor with indices on a complex vbundle always has a dagger pair:\
\>", "MathCaption",ExpressionUUID->"e31ec0b5-a939-48cb-843b-1ad3f66cb1df"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}], ",", "TangentM3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->
  "In[238]:=",ExpressionUUID->"234e76b2-f59f-4f60-915a-a90ed8a4347c"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[$CellContext`a, $CellContext`\[GothicCapitalB]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977879434466*^9},
 CellLabel->"During evaluation of In[238]:=",
 CellID->34886677,ExpressionUUID->"5f3cd4d9-0c91-4c79-a027-27aba6f3165d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V\[Dagger]", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalB]\
\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977879452157*^9},
 CellLabel->"During evaluation of In[238]:=",
 CellID->77344708,ExpressionUUID->"bf2013e1-2abd-4a10-8895-95a7e3f8d863"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "V"}]], "Input",
 CellLabel->
  "In[239]:=",ExpressionUUID->"46acec2c-8acb-4ee1-864f-be8dc3d9316e"],

Cell[CellGroupData[{

Cell["Global`V", "Print", "PrintUsage",
 CellChangeTimes->{3.728977879613552*^9},
 CellTags->"Info2393728981479-9181623",
 CellID->704900921,ExpressionUUID->"729ec76a-285d-4e29-a392-3a0b59f569e5"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "V", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`V],
          Editable->False], "^=", "V\[Dagger]"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "V", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "V", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "V", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC,
            Editable->False], ",", "M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "V", "]"}], "^=", "\<\"V\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ServantsOf", "[", "V", "]"}], "^=", 
         RowBox[{"{", "V\[Dagger]", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "V", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["InnerC",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "V", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "V", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "V", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897787963589*^9},
 CellTags->"Info2393728981479-9181623",
 CellID->1749982395,ExpressionUUID->"55bb2ae3-f92f-49b6-a6b4-e800ab4bf2c8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "V\[Dagger]"}]], "Input",
 CellLabel->
  "In[240]:=",ExpressionUUID->"3ff8171b-7faa-4e1a-b515-08f53ee376dd"],

Cell[CellGroupData[{

Cell["Global`V\[Dagger]", "Print", "PrintUsage",
 CellChangeTimes->{3.728977879795273*^9},
 CellTags->"Info2403728981479-9181623",
 CellID->18083453,ExpressionUUID->"f4ae4598-eca7-4561-b34c-69b4a8d91214"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "V\[Dagger]", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`V\[Dagger]],
          Editable->False], "^=", "V"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "V\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "V\[Dagger]", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "V\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC\[Dagger]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC\[Dagger],
            Editable->False], ",", "M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", "V\[Dagger]", "]"}], "^=", "V"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "V\[Dagger]", "]"}], 
         "^=", "\<\"V\[Dagger]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "V\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["InnerC\[Dagger]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC\[Dagger],
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "V\[Dagger]", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "V\[Dagger]", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "V\[Dagger]", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`V\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778798193073`*^9},
 CellTags->"Info2403728981479-9181623",
 CellID->1133400344,ExpressionUUID->"50884fe2-237c-4232-b241-7b33a9fb2528"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Complex conjugation acts on the name of the tensor and on its indices as well:\
\>", "MathCaption",ExpressionUUID->"4868cfeb-8e11-41dc-82b0-fa75de150ccd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", " ", 
  RowBox[{"V", "[", 
   RowBox[{"a", ",", "\[GothicCapitalC]"}], "]"}], " ", "]"}]], "Input",
 CellLabel->
  "In[241]:=",ExpressionUUID->"99cccb93-2181-412a-b0f6-257cc8584006"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"V\[Dagger]", 
      StyleBox[GridBox[{
         {"a", "\[GothicCapitalC]\[Dagger]"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`V\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalC]\
\[Dagger]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977880170309*^9},
 CellLabel->"Out[241]=",
 CellID->879632718,ExpressionUUID->"eae13a93-1ae9-482b-a738-13895d0fad76"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "V", "]"}]], "Input",
 CellLabel->
  "In[242]:=",ExpressionUUID->"69d3165b-e826-4dbe-89cb-cbe7731681f8"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897788053129*^9},
 CellLabel->"During evaluation of In[242]:=",
 CellID->709665478,ExpressionUUID->"7c23daf0-2746-4780-b2b1-8d7343ea062b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977880551731*^9},
 CellLabel->"During evaluation of In[242]:=",
 CellID->223513308,ExpressionUUID->"12b06724-e87f-4d37-b0a7-6e1be6257e48"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Conjugation of indices (this should be used only for programming):\
\>", "MathCaption",ExpressionUUID->"57c9d282-d824-46e2-92d7-5f5219f6f197"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DaggerIndex", "/@", 
  RowBox[{"{", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}], ",", "\[GothicCapitalA]", ",", 
    RowBox[{"-", "\[GothicCapitalB]\[Dagger]"}]}], "}"}]}]], "Input",
 CellLabel->
  "In[243]:=",ExpressionUUID->"f4ae1704-1355-4c19-b0ac-ebc7d50ba99e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", "\[GothicCapitalA]\[Dagger]", ",", 
   RowBox[{"-", "\[GothicCapitalB]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289778807171717`*^9},
 CellLabel->"Out[243]=",
 CellID->639268953,ExpressionUUID->"bd61ac98-b6ce-4fa7-86c5-37011cb7da88"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Tensors can be defined with different properties against complex conjugation.\
\
\>", "Text",ExpressionUUID->"5f6c124e-74f3-4d15-bcc5-1bc82c7c10f9"],

Cell[CellGroupData[{

Cell["On real vbundles tensors are real by default:", "MathCaption",ExpressionUUID->"8ee12e9d-f5dc-4751-a553-eba159640d0a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Options", "[", "DefTensor", "]"}]], "Input",
 CellLabel->
  "In[244]:=",ExpressionUUID->"ba71e427-4a2a-4770-9a44-9e4da06317b9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Dagger", "\[Rule]", "Real"}], ",", 
   RowBox[{"Master", "\[Rule]", "Null"}], ",", 
   RowBox[{"PrintAs", "\[Rule]", "Identity"}], ",", 
   RowBox[{"VanishingQ", "\[Rule]", "False"}], ",", 
   RowBox[{"ForceSymmetries", "\[Rule]", "False"}], ",", 
   RowBox[{"WeightOfTensor", "\[Rule]", "0"}], ",", 
   RowBox[{"GradeOfTensor", "\[Rule]", "0"}], ",", 
   RowBox[{"FrobeniusQ", "\[Rule]", "False"}], ",", 
   RowBox[{"OrthogonalTo", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"ProjectedWith", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"ProtectNewSymbol", "\[RuleDelayed]", "$ProtectNewSymbols"}], ",", 
   
   RowBox[{"DefInfo", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}], ",", 
   RowBox[{"TensorID", "\[Rule]", 
    RowBox[{"{", "}"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977880993339*^9},
 CellLabel->"Out[244]=",
 CellID->332607130,ExpressionUUID->"d467c0fc-2330-4794-a200-34f5434b5c0e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[245]:=",ExpressionUUID->"0b10d262-d244-44b9-932b-72be1dba2bc6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778813214293`*^9},
 CellLabel->"During evaluation of In[245]:=",
 CellID->228100769,ExpressionUUID->"676b09d7-1fdc-4fb7-a628-0f7f99a861b2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"3", "I", " ", 
   RowBox[{"V", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[246]:=",ExpressionUUID->"521a9185-17fa-4e06-8e11-9081d4d46fa2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "3"}], " ", "\[ImaginaryI]", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"V", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`V[$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289778815615053`*^9},
 CellLabel->"Out[246]=",
 CellID->431314688,ExpressionUUID->"3bb42abd-eab7-4305-a2cf-1305b38ec1ea"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "V", "]"}]], "Input",
 CellLabel->
  "In[247]:=",ExpressionUUID->"ced94c9d-6356-4fdb-8eee-75b5e7c18ab2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977881750795*^9},
 CellLabel->"During evaluation of In[247]:=",
 CellID->1348652591,ExpressionUUID->"a291bb32-be87-4f18-b9aa-77c5980dd45b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However, it is possible to \"complexify\" a real vbundle, in such a way that \
we duplicate its real dimension. Then vectors can have nontrivial complex \
conjugates:\
\>", "MathCaption",ExpressionUUID->"68df0777-34c1-4a2e-b230-8c62e528c97f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", "a", "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->
  "In[248]:=",ExpressionUUID->"e90deda9-b906-4960-a78f-8434db06178e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778818866673`*^9},
 CellLabel->"During evaluation of In[248]:=",
 CellID->114202852,ExpressionUUID->"ac33456f-e11c-4ae1-9c50-6a4516561df6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V\[Dagger]", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V\[Dagger][$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778819059477`*^9},
 CellLabel->"During evaluation of In[248]:=",
 CellID->1704436793,ExpressionUUID->"9baa71f6-8bde-42bf-9826-efd354a4580b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"3", "I", " ", 
   RowBox[{"V", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[249]:=",ExpressionUUID->"544a2b37-a264-429b-8c03-2e884a1e6c95"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "3"}], " ", "\[ImaginaryI]", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"V\[Dagger]", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`V\[Dagger][$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977882052754*^9},
 CellLabel->"Out[249]=",
 CellID->1184812618,ExpressionUUID->"024d95dd-ce4f-4e67-ab15-0e7d50851c2a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "V", "]"}]], "Input",
 CellLabel->
  "In[250]:=",ExpressionUUID->"d5c3ee27-e635-4347-ab22-78bf4f8bf4d3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778821951847`*^9},
 CellLabel->"During evaluation of In[250]:=",
 CellID->1677295742,ExpressionUUID->"320d4b5b-3a58-40a1-b36e-9043996b2719"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778822137337`*^9},
 CellLabel->"During evaluation of In[250]:=",
 CellID->291701625,ExpressionUUID->"f2899bff-c11b-4843-bb66-53e086f9b551"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If the vbundle is complex then all tensors on it must be complex:\
\>", "MathCaption",ExpressionUUID->"8fb825ea-90cb-4f52-b12f-bd6f233b4ee5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", "\[GothicCapitalA]", "]"}], ",", "M3"}], "]"}]], "Input",\

 CellLabel->
  "In[251]:=",ExpressionUUID->"51615578-25c7-4472-819c-fc996d122322"],

Cell[BoxData[
 TemplateBox[{
  "DefTensor", "invalid", 
   "\"\\!\\(\\*RowBox[{\\\"Real\\\"}]\\) is not a valid \\!\\(\\*RowBox[{\\\"\
\\\\\\\"value for Dagger: complex indices\\\\\\\"\\\"}]\\).\"", 2, 251, 28, 
   19253573948027831207, "Local", "xAct`xTensor`DefTensor"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728977882469781*^9},
 CellLabel->"During evaluation of In[251]:=",
 CellID->2071528921,ExpressionUUID->"8398fd3c-ec66-4827-8e80-54be58173303"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", "\[GothicCapitalA]", "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->
  "In[252]:=",ExpressionUUID->"572a8af7-b017-43af-9a35-bcd91804c95d"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", "\[GothicCapitalA]", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[$CellContext`\[GothicCapitalA]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977882921338*^9},
 CellLabel->"During evaluation of In[252]:=",
 CellID->1181690083,ExpressionUUID->"dc81dfc3-bbbc-4d3f-8390-40848bfde7be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V\[Dagger]", "[", "\[GothicCapitalA]\[Dagger]", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V\[Dagger][$CellContext`\[GothicCapitalA]\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778829402323`*^9},
 CellLabel->"During evaluation of In[252]:=",
 CellID->736457525,ExpressionUUID->"a4c429ee-fcf1-494e-bb12-0b9c89159343"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"3", "I", " ", 
   RowBox[{"V", "[", "\[GothicCapitalA]", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[253]:=",ExpressionUUID->"e38a45d1-4b2c-4184-8703-2cea9adc08c8"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "3"}], " ", "\[ImaginaryI]", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"V\[Dagger]", 
       StyleBox[GridBox[{
          {"\[GothicCapitalA]\[Dagger]"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`V\[Dagger][$CellContext`\[GothicCapitalA]\[Dagger]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977883115734*^9},
 CellLabel->"Out[253]=",
 CellID->1693365014,ExpressionUUID->"dec04b48-cfde-4a66-8a02-849b265f337a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "V", "]"}]], "Input",
 CellLabel->
  "In[254]:=",ExpressionUUID->"eece5558-b330-469d-ae57-2da5e8bf9946"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897788326652*^9},
 CellLabel->"During evaluation of In[254]:=",
 CellID->1932500254,ExpressionUUID->"2ba7993c-bfb3-45d9-bbda-466b265efa57"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977883284461*^9},
 CellLabel->"During evaluation of In[254]:=",
 CellID->989091750,ExpressionUUID->"f0363b49-b156-4437-b21b-920cf0851a5f"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
On complex vbundles it is possible to define Hermitian tensors if they have \
equal number of indices on both conjugated vbundles (though there is no \
built-in operation of Hermitian conjugation):\
\>", "MathCaption",ExpressionUUID->"94d5af63-1576-430f-96ee-ba46a6d17bb2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
     "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]"}], "]"}],
    ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Hermitian"}]}], "]"}]], "Input",
 CellLabel->
  "In[255]:=",ExpressionUUID->"49de26f0-e4b5-465d-bb04-9fcda065a04c"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
     "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]"}], "]"}],
    "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[$CellContext`\[GothicCapitalA], $CellContext`\
\[GothicCapitalB], $CellContext`\[GothicCapitalC]\[Dagger], $CellContext`\
\[GothicCapitalD]\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977883422408*^9},
 CellLabel->"During evaluation of In[255]:=",
 CellID->2141614470,ExpressionUUID->"fa48d62d-7b46-4675-acc0-65b2103e9c80"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V\[Dagger]", "[", 
    RowBox[{
    "\[GothicCapitalA]\[Dagger]", ",", "\[GothicCapitalB]\[Dagger]", ",", 
     "\[GothicCapitalC]", ",", "\[GothicCapitalD]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V\[Dagger][$CellContext`\[GothicCapitalA]\[Dagger], \
$CellContext`\[GothicCapitalB]\[Dagger], $CellContext`\[GothicCapitalC], \
$CellContext`\[GothicCapitalD]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778834406853`*^9},
 CellLabel->"During evaluation of In[255]:=",
 CellID->140989923,ExpressionUUID->"1b982c69-2c28-4e35-8298-13f45a5f6191"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"V", "[", 
   RowBox[{
   "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
    "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]"}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[256]:=",ExpressionUUID->"f2be409f-f389-4683-945f-166b3198ef4f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"V", 
      StyleBox[GridBox[{
         {"\[GothicCapitalC]", "\[GothicCapitalD]", 
          "\[GothicCapitalA]\[Dagger]", "\[GothicCapitalB]\[Dagger]"},
         {" ", " ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`V[$CellContext`\[GothicCapitalC], $CellContext`\
\[GothicCapitalD], $CellContext`\[GothicCapitalA]\[Dagger], $CellContext`\
\[GothicCapitalB]\[Dagger]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977883586383*^9},
 CellLabel->"Out[256]=",
 CellID->943173220,ExpressionUUID->"91ecfb7a-e9a1-4728-b042-288a0f2f39fd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"V\[Dagger]", "[", 
   RowBox[{
   "\[GothicCapitalA]\[Dagger]", ",", "\[GothicCapitalB]\[Dagger]", ",", 
    "\[GothicCapitalC]", ",", "\[GothicCapitalD]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[257]:=",ExpressionUUID->"0af59942-68cd-4700-b253-5b265ecddbac"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"V", 
      StyleBox[GridBox[{
         {"\[GothicCapitalA]", "\[GothicCapitalB]", 
          "\[GothicCapitalC]\[Dagger]", "\[GothicCapitalD]\[Dagger]"},
         {" ", " ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`V[$CellContext`\[GothicCapitalA], $CellContext`\
\[GothicCapitalB], $CellContext`\[GothicCapitalC]\[Dagger], $CellContext`\
\[GothicCapitalD]\[Dagger]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289778837092113`*^9},
 CellLabel->"Out[257]=",
 CellID->1335510033,ExpressionUUID->"3353dae0-284c-4fdf-b6ae-98619fd6c94d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
     "\[GothicCapitalC]\[Dagger]"}], "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Hermitian"}]}], "]"}]], "Input",
 CellLabel->
  "In[258]:=",ExpressionUUID->"a174b86d-f4e2-4ed3-abd6-365721f1981d"],

Cell[BoxData[
 TemplateBox[{
  "TransposeDagger", "error3", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Different number of indices \
of\\\\\\\"\\\"}]\\) \\!\\(\\*InterpretationBox[StyleBox[\\\"InnerC\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], InnerC, \
Rule[Editable, False]]\\) \\!\\(\\*RowBox[{\\\"\\\\\\\"and its \
conjugate\\\\\\\"\\\"}]\\) \\!\\(\\*InterpretationBox[StyleBox[\\\"InnerC\
\[Dagger]\\\", Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], InnerC\
\\\\[Dagger], Rule[Editable, False]]\\)\"", 2, 258, 29, 19253573948027831207, 
   "Local", "xAct`xTensor`TransposeDagger"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289778838262053`*^9},
 CellLabel->"During evaluation of In[258]:=",
 CellID->418046187,ExpressionUUID->"54c70a22-939f-404f-8183-91e1b9716b23"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "[", "V", "]"}], ";"}]], "Input",
 CellLabel->
  "In[259]:=",ExpressionUUID->"a8cac00d-fec6-48a4-9511-791a28520dcf"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778842217007`*^9},
 CellLabel->"During evaluation of In[259]:=",
 CellID->1683255572,ExpressionUUID->"80d5bd83-3433-4cc4-8ae3-4eed20bfc38d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977884240902*^9},
 CellLabel->"During evaluation of In[259]:=",
 CellID->1674316673,ExpressionUUID->"469395c4-fc63-4ef3-b2ef-b5022b75d228"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Note that the transposition involved for Hermitian objects does not require \
any particular ordering of indices of the tensors. However it follows the \
convention of exchanging the first slot of a given complex vbundle with the \
first slot of its complex conjugate:\
\>", "MathCaption",ExpressionUUID->"1468d4f7-eedf-43ee-854c-0616d4346519"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
     "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]", ",", 
     "\[GothicCapitalE]", ",", "\[GothicCapitalF]\[Dagger]"}], "]"}], ",", 
   "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Hermitian"}]}], "]"}]], "Input",
 CellLabel->
  "In[260]:=",ExpressionUUID->"91584caa-7b9b-4295-90a2-dbc7dd057173"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
     "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]", ",", 
     "\[GothicCapitalE]", ",", "\[GothicCapitalF]\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[$CellContext`\[GothicCapitalA], $CellContext`\
\[GothicCapitalB], $CellContext`\[GothicCapitalC]\[Dagger], $CellContext`\
\[GothicCapitalD]\[Dagger], $CellContext`\[GothicCapitalE], $CellContext`\
\[GothicCapitalF]\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778845830593`*^9},
 CellLabel->"During evaluation of In[260]:=",
 CellID->1762096470,ExpressionUUID->"fb5b66ba-32e6-4d91-8d1f-ba55870a25e0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V\[Dagger]", "[", 
    RowBox[{
    "\[GothicCapitalA]\[Dagger]", ",", "\[GothicCapitalB]\[Dagger]", ",", 
     "\[GothicCapitalC]", ",", "\[GothicCapitalD]", ",", 
     "\[GothicCapitalE]\[Dagger]", ",", "\[GothicCapitalF]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V\[Dagger][$CellContext`\[GothicCapitalA]\[Dagger], \
$CellContext`\[GothicCapitalB]\[Dagger], $CellContext`\[GothicCapitalC], \
$CellContext`\[GothicCapitalD], $CellContext`\[GothicCapitalE]\[Dagger], \
$CellContext`\[GothicCapitalF]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897788460075*^9},
 CellLabel->"During evaluation of In[260]:=",
 CellID->1755879160,ExpressionUUID->"be4c0973-0c7e-4c18-a0b5-d63de7433988"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dagger", "[", 
  RowBox[{"V", "[", 
   RowBox[{
   "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", 
    "\[GothicCapitalC]\[Dagger]", ",", "\[GothicCapitalD]\[Dagger]", ",", 
    "\[GothicCapitalE]", ",", "\[GothicCapitalF]\[Dagger]"}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[261]:=",ExpressionUUID->"43aa06d0-6176-4176-ae70-09442e01734e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"V", 
      StyleBox[GridBox[{
         {"\[GothicCapitalC]", "\[GothicCapitalD]", 
          "\[GothicCapitalA]\[Dagger]", "\[GothicCapitalB]\[Dagger]", 
          "\[GothicCapitalF]", "\[GothicCapitalE]\[Dagger]"},
         {" ", " ", " ", " ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`V[$CellContext`\[GothicCapitalC], $CellContext`\
\[GothicCapitalD], $CellContext`\[GothicCapitalA]\[Dagger], $CellContext`\
\[GothicCapitalB]\[Dagger], $CellContext`\[GothicCapitalF], $CellContext`\
\[GothicCapitalE]\[Dagger]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977884736393*^9},
 CellLabel->"Out[261]=",
 CellID->340765046,ExpressionUUID->"aa333319-0399-4a6c-aa64-32743ea920ff"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "V", "]"}]], "Input",
 CellLabel->
  "In[262]:=",ExpressionUUID->"561e6c71-f23b-404f-8fed-141a9cb9aaf5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977884863521*^9},
 CellLabel->"During evaluation of In[262]:=",
 CellID->2046725220,ExpressionUUID->"577390e9-8641-4691-8755-fec59989f09e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977884882667*^9},
 CellLabel->"During evaluation of In[262]:=",
 CellID->917714894,ExpressionUUID->"79c1785d-d182-47de-8b28-6101ba8ba080"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The symbol for conjugation can be changed using the global variable ",
 StyleBox["$DaggerCharacter",
  FontFamily->"Courier"],
 ", but of course the command for complex conjugation will still be called ",
 StyleBox["Dagger",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"3aa3f2c4-eaf9-424f-b420-7f34e6bae3fd"],

Cell[CellGroupData[{

Cell[BoxData["$DaggerCharacter"], "Input",
 CellLabel->
  "In[263]:=",ExpressionUUID->"24820903-b510-4139-a564-1360d91496f7"],

Cell[BoxData["\<\"\[Dagger]\"\>"], "Output",
 CellChangeTimes->{3.728977885048777*^9},
 CellLabel->"Out[263]=",
 CellID->806670704,ExpressionUUID->"0612ca62-1934-467e-bbd6-038136d4c3c0"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "W", "]"}]], "Input",
 CellLabel->
  "In[264]:=",ExpressionUUID->"08a20120-5f20-4b41-b7e4-95cc4ebb2320"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["3.9. Tensors with a variable number of slots", "Subsection",ExpressionUUID->"85b1e558-0468-4c77-849b-9e1fc7db8f9e"],

Cell[TextData[{
 "Sometimes it is desirable to work with tensors with a variable number of \
slots, and ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " supports it. For example, this is important when dealing with tensor \
harmonics, and actually the special package ",
 StyleBox["xAct`Harmonics`",
  FontFamily->"Courier"],
 " makes use of it.\nIn general, once a tensor has been defined with a number \
of slots and the symmetry properties of those slots, it is possible to use \
the tensor with more slots, and the system will assume that there are no \
additional symmetry properties on the added slots. The problem, however, \
comes when we want all slots to participate in the symmetry of the tensor. \
The solution is very simple:\nThe symmetry of a tensor is given by the \
function ",
 StyleBox["SymmetryGroupOfTensor",
  FontFamily->"Courier"],
 ", which stores the symmetry as an upvalue for the name of the tensor. \
However, this function tries first to get the symmetry from the whole \
expression of the tensor (indices included), and only if this is not defined \
it will use the stored definition. Therefore we can give additional symmetry \
definitions for the tensor which will overwrite the initial symmetry \
assignment. From version 0.9.4 we include a special notation to indicate from \
the beginning the presence of a variable number of slots in vbundle."
}], "Text",ExpressionUUID->"0d0845de-9f05-4c78-8761-3b37fb35c005"],

Cell[CellGroupData[{

Cell["\<\
Define a tensor with a variable number of indices on a given vbundle:\
\>", "MathCaption",ExpressionUUID->"a673768f-7822-4a7c-bfcc-e01965389728"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Z", "[", 
    RowBox[{"AnyIndices", "@", "TangentM3"}], "]"}], ",", "M3"}], 
  "]"}]], "Input",
 CellLabel->
  "In[265]:=",ExpressionUUID->"66ddc65a-3c82-4c76-9e58-c165170672ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Z", "[", 
    RowBox[{"AnyIndices", "[", 
     InterpretationBox[
      StyleBox["\[DoubleStruckCapitalT]M3",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TangentM3,
      Editable->False], "]"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Z[
    xAct`xTensor`AnyIndices[$CellContext`TangentM3]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977892327051*^9},
 CellLabel->"During evaluation of In[265]:=",
 CellID->432104373,ExpressionUUID->"8a06c4a2-ebca-4f41-82b0-bfe02d4a4f82"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Z"}]], "Input",
 CellLabel->
  "In[266]:=",ExpressionUUID->"5c0076a7-85aa-4dcd-94d3-ef9a7df51381"],

Cell[CellGroupData[{

Cell["Global`Z", "Print", "PrintUsage",
 CellChangeTimes->{3.728977892498336*^9},
 CellTags->"Info2663728981492-9181623",
 CellID->1735986504,ExpressionUUID->"8b37f71b-9bc8-4c6b-a966-48e42cfdf8b6"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "Z", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`Z],
          Editable->False], "^=", "Z"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "Z", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "Z", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "Z", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "Z", "]"}], "^=", "\<\"Z\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "Z", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"AnyIndices", "[", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], "]"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "Z", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "Z", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`Z],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977892524729*^9},
 CellTags->"Info2663728981492-9181623",
 CellID->1012344586,ExpressionUUID->"bf829cfa-6dbb-4861-8773-7a3aefbd4dba"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The tensor is given no symmetry. By default we return the identity group:\
\>", "MathCaption",ExpressionUUID->"64da12c3-3e4e-4553-8e4d-699d0b9dd4ba"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "Z", "]"}]], "Input",
 CellLabel->
  "In[267]:=",ExpressionUUID->"11ed7705-3f7a-47e5-a176-de2207f51c74"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"GenSet", "[", "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728977892680461*^9},
 CellLabel->"Out[267]=",
 CellID->2110005368,ExpressionUUID->"d6204b88-6155-4d78-a651-9589f277ada0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"Z", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[268]:=",ExpressionUUID->"6ed271e6-2d44-4b7f-9aab-7280d936bcc7"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"GenSet", "[", "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728977892807494*^9},
 CellLabel->"Out[268]=",
 CellID->1462757065,ExpressionUUID->"dd655c45-e0bf-4bec-84be-2427894cf6ac"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now make it symmetric for any number of indices:", "MathCaption",ExpressionUUID->"23d0c9d9-168f-4b4a-b6f5-e6472fa42c9f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"SymmetryGroupOfTensor", "[", 
   RowBox[{"Z", "[", "inds___", "]"}], "]"}], "^:=", 
  RowBox[{"Symmetric", "[", 
   RowBox[{"Range", "[", 
    RowBox[{"Length", "[", 
     RowBox[{"{", "inds", "}"}], "]"}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[269]:=",ExpressionUUID->"611f91bc-969d-4a77-bf4b-c0da37358482"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"Z", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[270]:=",ExpressionUUID->"7a70475a-99e6-40bc-bc9b-6ad0244aaed7"],

Cell[BoxData[
 RowBox[{"Symmetric", "[", 
  RowBox[{"{", "}"}], "]"}]], "Output",
 CellChangeTimes->{3.7289778929727707`*^9},
 CellLabel->"Out[270]=",
 CellID->165200599,ExpressionUUID->"56378924-4e4f-4098-a161-52e53ad92f97"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"Z", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[271]:=",ExpressionUUID->"c4949cd5-aee2-45f2-afb9-e3010500ed40"],

Cell[BoxData[
 RowBox[{"Symmetric", "[", 
  RowBox[{"{", 
   RowBox[{"1", ",", "2"}], "}"}], "]"}]], "Output",
 CellChangeTimes->{3.728977893103332*^9},
 CellLabel->"Out[271]=",
 CellID->1977595469,ExpressionUUID->"8491ec76-bc40-40c1-b887-5bccc693370b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"Z", "[", 
   RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[272]:=",ExpressionUUID->"b8ef38d1-a6ac-418c-bf18-3d7aaddbffb6"],

Cell[BoxData[
 RowBox[{"Symmetric", "[", 
  RowBox[{"{", 
   RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]], "Output",
 CellChangeTimes->{3.728977893252224*^9},
 CellLabel->"Out[272]=",
 CellID->842642864,ExpressionUUID->"cd5ebad8-825d-4576-85ed-dd14ce98fbcf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"2e9d9408-7cc3-455e-b0c2-f382085b96da"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "Z", "]"}]], "Input",
 CellLabel->
  "In[273]:=",ExpressionUUID->"d701fdee-130c-48f5-bfa9-7a7f34e3dcd2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Z"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Z],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897789336397*^9},
 CellLabel->"During evaluation of In[273]:=",
 CellID->1791446930,ExpressionUUID->"f03f6b90-fbe8-43bd-a0d9-33862eede570"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "inds", "]"}]], "Input",
 CellLabel->
  "In[274]:=",ExpressionUUID->"6ca2900a-4c7f-4594-98ef-910509f5c712"]
}, Open  ]],

Cell[TextData[{
 "In case you need symmetries other  than total symmetry or total \
antisymmetry, you will need to use the general notation for permutation \
groups (see ",
 StyleBox["xPermDoc.nb",
  FontFamily->"Courier"],
 " for a description)."
}], "Text",ExpressionUUID->"e71a8723-3263-4bf9-ab2d-372e7953824e"]
}, Closed]],

Cell[CellGroupData[{

Cell["3.10. Clean up", "Subsection",ExpressionUUID->"f35a0194-1aa1-46e0-a2f9-7b2af9e581fa"],

Cell[TextData[{
 "It is good practice in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ", and in all of ",
 StyleBox["xAct",
  FontSlant->"Italic"],
 ", to remove those symbols that are no longer needed. This makes the symbols \
available for other uses, and also frees memory that may be used to store \
properties of those symbols.\nThis documentation notebook is quite large, and \
hence at the end of some sections we will have \[OpenCurlyDoubleQuote]Clean \
up\[CloseCurlyDoubleQuote] sections removing symbols used in examples in the \
previous sections."
}], "Text",ExpressionUUID->"f3ad4d17-e94d-4713-a7e2-d0f8e9f17935"],

Cell[CellGroupData[{

Cell["Undefine tensors:", "MathCaption",ExpressionUUID->"399f8699-c3cc-44b3-a2e0-64736fb7b37f"],

Cell[CellGroupData[{

Cell[BoxData["$Tensors"], "Input",
 CellLabel->
  "In[275]:=",ExpressionUUID->"f6fe5140-dea6-415a-a0c0-a1156ed9cb59"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"T", ",", "v", ",", "r", ",", "S"}], "}"}]], "Output",
 CellChangeTimes->{3.728977896905465*^9},
 CellLabel->"Out[275]=",
 CellID->1584742292,ExpressionUUID->"37d12554-7e46-4621-9f53-f7d8b1e0793e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", "%"}], ";"}]], "Input",
 CellLabel->
  "In[276]:=",ExpressionUUID->"470f14bf-c925-403c-9055-0f1752574e97"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778970201883`*^9},
 CellLabel->"During evaluation of In[276]:=",
 CellID->1671511550,ExpressionUUID->"bbcf6719-8d9a-4f1d-b4bc-b363589df69f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977897040285*^9},
 CellLabel->"During evaluation of In[276]:=",
 CellID->1861034443,ExpressionUUID->"104d60d0-4c44-495e-ae8d-424a96a8547e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "r"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`r],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977897057859*^9},
 CellLabel->"During evaluation of In[276]:=",
 CellID->306564946,ExpressionUUID->"0603fbf9-f633-4904-b585-769e13837363"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977897074111*^9},
 CellLabel->"During evaluation of In[276]:=",
 CellID->623778269,ExpressionUUID->"5523e2f4-1511-42f4-ae27-ebcc4277cf2d"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We cannot yet undefine the manifold M3 because other geometric structures (in \
this case an inner vector bundle) are defined on it. We call them \
\[OpenCurlyDoubleQuote]visitors\[CloseCurlyDoubleQuote] of the \
\[OpenCurlyDoubleQuote]host\[CloseCurlyDoubleQuote] structure, the manifold \
M3 in this case.\
\>", "MathCaption",ExpressionUUID->"d25fb823-4acb-427a-92a2-b9b7348d71c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Catch", "[", 
  RowBox[{"UndefManifold", "[", "M3", "]"}], "]"}]], "Input",
 CellLabel->
  "In[277]:=",ExpressionUUID->"12570b0e-504c-4e69-8a89-5d8f2024ca5e"],

Cell[BoxData[
 TemplateBox[{
  "General", "noundef", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Symbol\\\\\\\"\\\"}]\\) \
\\!\\(\\*RowBox[{\\\"M3\\\"}]\\) cannot be undefined because \
\\!\\(\\*RowBox[{\\\"\\\\\\\"it has visitors\\\\\\\"\\\"}]\\).\"", 2, 277, 30,
    19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977897217198*^9},
 CellLabel->"During evaluation of In[277]:=",
 CellID->1973686076,ExpressionUUID->"0cea6aed-94ef-44f7-85d0-83ba02f74565"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine the inner vbundle explicitly:", "MathCaption",ExpressionUUID->"ce746ea1-b3f1-444b-9406-bed61163aebe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefVBundle", "[", "InnerC", "]"}]], "Input",
 CellLabel->
  "In[278]:=",ExpressionUUID->"ad50ddba-0e40-4de4-a129-4c668b7df482"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC\[Dagger]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC\[Dagger],
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "conjugated vbundle", 
   " ", $CellContext`InnerC\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778975046177`*^9},
 CellLabel->"During evaluation of In[278]:=",
 CellID->1709467770,ExpressionUUID->"cdf7af0e-9284-423d-ae18-328219308bd4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`InnerC],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778976230783`*^9},
 CellLabel->"During evaluation of In[278]:=",
 CellID->1988580824,ExpressionUUID->"2c9db178-cfb9-447e-aef8-8e22dab60246"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now we can undefine the manifolds:", "MathCaption",ExpressionUUID->"e03a86d6-80c1-4fb6-a2d0-5dd23d7c5510"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "S2", "]"}]], "Input",
 CellLabel->
  "In[279]:=",ExpressionUUID->"7bf774fe-644d-4207-93c7-e7de4221511b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]S2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentS2,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentS2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977898043635*^9},
 CellLabel->"During evaluation of In[279]:=",
 CellID->789910707,ExpressionUUID->"ab857645-62ec-40cc-9e08-a5bb4f418850"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`S2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977898066807*^9},
 CellLabel->"During evaluation of In[279]:=",
 CellID->285346795,ExpressionUUID->"da4651e4-a813-4773-b8dd-6eefcd6914ad"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellLabel->
  "In[280]:=",ExpressionUUID->"d06634ee-92e5-474e-a3f1-a4c25d21fa70"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289778984385033`*^9},
 CellLabel->"During evaluation of In[280]:=",
 CellID->492599814,ExpressionUUID->"a48e0cb1-6e38-467e-8c27-fabfae5bc59c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977898461198*^9},
 CellLabel->"During evaluation of In[280]:=",
 CellID->575614713,ExpressionUUID->"83f604d3-0a00-436d-9637-383d92a0f904"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And the parameter:", "MathCaption",ExpressionUUID->"ba377a37-bed2-41da-b013-76cb48ccc3a2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefParameter", "[", "time", "]"}]], "Input",
 CellLabel->
  "In[281]:=",ExpressionUUID->"3fea85b3-4646-4287-afc8-ebe1f75c26bc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefParameter", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parameter\"\>", "\[InvisibleSpace]", "\<\" \"\>",
    "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["time",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`time,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefParameter, ": Undefined ", "parameter", 
   " ", $CellContext`time],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977898581707*^9},
 CellLabel->"During evaluation of In[281]:=",
 CellID->2090135303,ExpressionUUID->"311ad552-186b-43a2-88e4-c83072ef668b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Our session is empty again:", "MathCaption",ExpressionUUID->"d925ef7f-fc5c-49c6-86bf-1742ab62a8fc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Global`*"}]], "Input",
 CellLabel->
  "In[282]:=",ExpressionUUID->"560455fe-a648-4382-b925-9c27aa88e4f2"],

Cell["Global`$UseTextSearchQ", "Print", "PrintUsage",
 CellChangeTimes->{3.7289778987059317`*^9},
 CellTags->"Info2823728981498-9181623",
 CellID->366539895,ExpressionUUID->"944a01b4-5f58-4e66-a631-2beb8f017d35"]
}, Open  ]]
}, Open  ]],

Cell["\<\
TODO: Where are those two symbols coming from? They should not be there...\
\>", "Text",ExpressionUUID->"37ab47cf-a37b-466f-872f-1400001b4863"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["4. Canonicalization", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1330943070,ExpressionUUID->"e07ebc50-8899-4702-bbbe-f4d58d3ca5cc"],

Cell[CellGroupData[{

Cell["4.1. Basics", "Subsection",ExpressionUUID->"27be6501-297f-4c01-a07b-643e49a7ce19"],

Cell[TextData[{
 "In any calculation it is essential to bring all tensors and products of \
tensors down to a canonical form, in order to be compared. This is done by \
the function ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 ", probably the most important ingredient of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", and certainly the most complicated function.\nDo not confuse the \
canonicalization proces, by which an expression is brought to a canonical \
form (not necessarily simple) from the simplification process, by which an \
expression is rewritten as an equivalent, but simpler form. The former is \
uniquely defined once the canonical form of every sintactycally correct \
expression is decided. The latter is largely subjective, and generally more \
difficult to work with. Fortunately ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " has a very good algorithm for simplification (",
 StyleBox["Simplify",
  FontFamily->"Courier"],
 "). This is one of the main reasons to avoid introducing TensorTimes or \
TensorPlus, which would force us to build a new TensorSimplify, and another \
strong reason to use abstract index notation. In ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " we have defined a command ",
 StyleBox["Simplification",
  FontFamily->"Courier"],
 ", which simply applies ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 " first, and then ",
 StyleBox["Simplify",
  FontFamily->"Courier"],
 ".\nThe hardest part of the process (the canonicalization of indices) is \
performed by the companion package ",
 StyleBox["xPerm`",
  FontFamily->"Courier"],
 ". The canonicalization code is duplicated in that package: there is first a \
pure ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " version of the code, and then there is an external C-executable, which is \
much faster, but requires a ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " connection, which does not work for some operating systems. Which code is \
used is controlled with several switches, as we will see later on."
}], "Text",ExpressionUUID->"0fc444e8-cd2c-4c05-a2ac-fa11fa4967e1"],

Cell[TextData[{
 StyleBox[ButtonBox["ToCanonical",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ToCanonical.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Canonicalization of indices of an expression\n", "Text"],
 StyleBox[ButtonBox["Simplification",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Simplification.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Apply ", "Text"],
 StyleBox["ToCanonical", "Text",
  FontFamily->"Courier"],
 StyleBox[" and then ", "Text"],
 StyleBox["Simplify", "Text",
  FontFamily->"Courier"]
}], "DefinitionBox",ExpressionUUID->"a4853172-1344-4f00-a08e-1d0f978a6c6b"],

Cell["Simplification functions", "Text",ExpressionUUID->"ae327cb7-0805-4d8f-a39a-36c11b296e46"],

Cell[CellGroupData[{

Cell["Define a manifold and some tensors fields:", "MathCaption",ExpressionUUID->"86ecd186-81dc-4d3f-b9be-378a3ba87e08"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", "h",
      ",", "i", ",", "j"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[283]:=",ExpressionUUID->"9528b0dc-60db-4da0-a1f8-7bad3c6faab4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779077202787`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->985284070,ExpressionUUID->"ba18926e-63ab-46ea-bd00-bd15d6d59c87"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM3", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM3, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779077382593`*^9},
 CellLabel->"During evaluation of In[283]:=",
 CellID->422962473,ExpressionUUID->"d0c3d622-7ff1-4eef-8750-2f106d0d466b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[284]:=",ExpressionUUID->"41ba7f86-db7d-4177-bd90-7743bf5e9519"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c], ". ", 
   ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977907888639*^9},
 CellLabel->"During evaluation of In[284]:=",
 CellID->559412141,ExpressionUUID->"f96744ac-c004-4ac5-96b7-00ca5d603efd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[285]:=",ExpressionUUID->"362271f3-1082-489b-a812-99f1a5ed85d0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`S[$CellContext`a, $CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779080213137`*^9},
 CellLabel->"During evaluation of In[285]:=",
 CellID->356447221,ExpressionUUID->"aa088c12-f943-4acb-8d22-b9661ba00c6e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[286]:=",ExpressionUUID->"f8413efe-4e9d-4480-9a74-52c66952f1c0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977908137731*^9},
 CellLabel->"During evaluation of In[286]:=",
 CellID->1476551864,ExpressionUUID->"4573a8e6-df04-4851-82a3-cf83084f370b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Canonical form of the totally antisymmetric tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 ": sort indices in alphabetical order:"
}], "MathCaption",ExpressionUUID->"4123c80d-f87b-4825-ba5d-1a693af48dbf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", " ", 
  RowBox[{"U", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], " ", "]"}]], "Input",
 CellLabel->
  "In[287]:=",ExpressionUUID->"f5c1bc7a-1e28-48db-a0ad-afefe3521473"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"a", "b", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977908408895*^9},
 CellLabel->"Out[287]=",
 CellID->86153516,ExpressionUUID->"ced005e8-4ea8-4813-ba0d-e3039a0f927d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "+", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "a"}]}], "]"}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[288]:=",ExpressionUUID->"97fc91d0-dbe0-4991-a802-ba00273cb21f"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.72897790889884*^9},
 CellLabel->"Out[288]=",
 CellID->578165334,ExpressionUUID->"3178cd84-80d9-4780-840e-6bf9e9c0b296"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Dummy indices are replaced so that the total number of different dummies is \
minimal:\
\>", "MathCaption",ExpressionUUID->"8b0e0cc6-8067-48ad-9c1d-7adeb12e7a26"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}]}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[289]:=",ExpressionUUID->"1cf5e7ce-0395-4517-83f9-621717c4b8ed"],

Cell[BoxData[
 RowBox[{"2", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977909121375*^9},
 CellLabel->"Out[289]=",
 CellID->2008749510,ExpressionUUID->"e1be20c2-0433-4c46-98a2-64c9f1bfe34f"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The process of canonicalization is far from trivial (see next section). We \
can monitor it using ",
 StyleBox["Verbose",
  FontFamily->"Courier"],
 " options."
}], "Text",ExpressionUUID->"51d3dc2d-1f73-404e-991c-653ffde168ef"],

Cell[TextData[{
 StyleBox[ButtonBox["Verbose",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Verbose.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Gives a real-time report of the tensorial canonicalization process\
\n", "Text"],
 StyleBox["TimeVerbose\t\t\t\t", "DisplayMath"],
 StyleBox["Gives timings of several canonicalization steps\n", "Text"],
 StyleBox["xPermVerbose\t\t\t\t", "DisplayMath"],
 StyleBox["Gives a real-time report of the permutation canonicalization \
process", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.7289779174249477`*^9, 
  3.728977919327373*^9}},ExpressionUUID->"4ff7dfea-d01e-40d7-85e9-\
92d66ceb3358"],

Cell[TextData[{
 "Verbosing options of ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"03367d2e-3fd7-494d-8711-ac60fa5e1ecd"],

Cell[CellGroupData[{

Cell[TextData[{
 "These three cells show the output of the three ",
 StyleBox["Verbose",
  FontFamily->"Courier"],
 " options of ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 " in a simple case. The output of ",
 StyleBox["xPermVerbose",
  FontFamily->"Courier"],
 " can be very large even for small expressions. We do not use the external \
executable because there is not verbose information from it:"
}], "MathCaption",ExpressionUUID->"e92c0676-9ea4-404d-bd5a-65b63e06582e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"CanonicalPerm", ",", 
   RowBox[{"MathLink", "\[Rule]", "False"}]}], "]"}]], "Input",
 CellLabel->
  "In[290]:=",ExpressionUUID->"a467c9e8-2413-4cb1-b28b-884dd2a2a75a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MathLink", "\[Rule]", "False"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"xPermVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"OrderedBase", "\[Rule]", "True"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977909255715*^9},
 CellLabel->"Out[290]=",
 CellID->462600083,ExpressionUUID->"25b1023d-bb6c-438f-9caa-2de7678b0675"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"b", ",", "c", ",", "a"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[291]:=",ExpressionUUID->"df9663b7-68f8-44d2-a111-46dabfd5ed75"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 291, 31, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977909388124*^9},
 CellLabel->"During evaluation of In[291]:=",
 CellID->1028792841,ExpressionUUID->"22d685c6-29dd-496a-a846-88629da5daf3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.009223000000000425`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.009223000000000425, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977909467868*^9},
 CellLabel->"During evaluation of In[291]:=",
 CellID->1530436690,ExpressionUUID->"21d29b50-cad9-47ce-9775-2d17f3c4e66c"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 291, 32, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977909507205*^9},
 CellLabel->"During evaluation of In[291]:=",
 CellID->1637808711,ExpressionUUID->"5a44fe39-3612-4fd7-8ad9-058c1dd8927c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.007001000000000701`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.007001000000000701, " secs."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779095209217`*^9},
 CellLabel->"During evaluation of In[291]:=",
 CellID->205618531,ExpressionUUID->"bde5710b-d9f1-4b13-abb0-0cd668c66e91"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "c", "b"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`c, $CellContext`b],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977909590954*^9},
 CellLabel->"Out[291]=",
 CellID->529462494,ExpressionUUID->"abf6fbcc-a539-4231-a624-bcdffc4488af"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"b", ",", "c", ",", "a"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], ",", 
   RowBox[{"Verbose", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[292]:=",ExpressionUUID->"14565ad5-3787-4fce-9650-543266d8ea7c"],

Cell[CellGroupData[{

Cell[BoxData["\<\"***********************************************************\
\"\>"], "Print",
 CellChangeTimes->{3.7289779097215233`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1651387137,ExpressionUUID->"b4c221b4-4f7b-49b5-825f-3eb0399bdc57"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonical:: object: \"\>", "\[InvisibleSpace]", 
   TagBox[
    TooltipBox[
     UnderscriptBox[
      RowBox[{
       RowBox[{"(", 
        TagBox[
         TooltipBox[
          UnderscriptBox["Times", "_"],
          RowBox[{"\"VerbatimProduct\"", "[", "Times", "]"}]],
         #& ], ")"}], "[", 
       RowBox[{
        TagBox[
         TooltipBox[
          UnderscriptBox[
           InterpretationBox[
            StyleBox[GridBox[{
               {"v", 
                StyleBox[GridBox[{
                   {" "},
                   {"b"}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`v[-$CellContext`b],
            Editable->False], "_"],
          RowBox[{"\"Object\"", "[", 
            RowBox[{
              InterpretationBox[
               StyleBox[
                GridBox[{{"v", 
                   StyleBox[
                    
                    GridBox[{{" "}, {"b"}}, RowSpacings -> 0, ColumnSpacings -> 
                    0.05, GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 
                    9]}}, ColumnSpacings -> 0.05, RowAlignments -> Center], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`v[-$CellContext`b], Editable -> False], ",", 
              RowBox[{"{", 
                RowBox[{"Tensor", ",", "1", ",", "v"}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"-", "b"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[-$CellContext`b], Editable -> 
                   False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"-", "b"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[-$CellContext`b], Editable -> 
                   False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False]}], ",", "Null"}], "}"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[{-$CellContext`TangentM3, Null}], 
                   Editable -> False]}], "}"}]}], "]"}]],
         #& ], ",", 
        TagBox[
         TooltipBox[
          UnderscriptBox[
           InterpretationBox[
            StyleBox[GridBox[{
               {"U", 
                StyleBox[GridBox[{
                   {"b", "c", "a"},
                   {" ", " ", " "}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`U[$CellContext`b, $CellContext`c, $CellContext`a],
            Editable->False], "_"],
          RowBox[{"\"Object\"", "[", 
            RowBox[{
              InterpretationBox[
               StyleBox[
                GridBox[{{"U", 
                   StyleBox[
                    
                    GridBox[{{"b", "c", "a"}, {" ", " ", " "}}, RowSpacings -> 
                    0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                 ColumnSpacings -> 0.05, RowAlignments -> Center], 
                AutoSpacing -> False, ShowAutoStyles -> False], 
               $CellContext`U[$CellContext`b, $CellContext`c, $CellContext`a],
                Editable -> False], ",", 
              RowBox[{"{", 
                RowBox[{"Tensor", ",", "6", ",", "U"}], "}"}], ",", 
              RowBox[{"{", 
                RowBox[{
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{"b", ",\[ThinSpace]", "c", ",\[ThinSpace]", "a"}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`IndexList[$CellContext`b, $CellContext`c, \
$CellContext`a], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{"a", ",\[ThinSpace]", "b", ",\[ThinSpace]", "c"}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   
                   xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, \
$CellContext`c], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[], Editable -> False], ",", 
                  InterpretationBox[
                   StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}], ",\[ThinSpace]", 
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}], ",\[ThinSpace]", 
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                   xAct`xTensor`IndexList[{$CellContext`TangentM3, 
                    Null}, {$CellContext`TangentM3, 
                    Null}, {$CellContext`TangentM3, Null}], Editable -> 
                   False]}], "}"}]}], "]"}]],
         #& ]}], "]"}], "_"],
     RowBox[{"\"Object\"", "[", 
       RowBox[{
         RowBox[{
           RowBox[{"(", 
             TagBox[
              TooltipBox[
               UnderscriptBox["Times", "_"], 
               RowBox[{"\"VerbatimProduct\"", "[", "Times", "]"}]], #& ], 
             ")"}], "[", 
           RowBox[{
             TagBox[
              TooltipBox[
               UnderscriptBox[
                InterpretationBox[
                 StyleBox[
                  GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{" "}, {"b"}}, RowSpacings -> 0, ColumnSpacings -> 
                    0.05, GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 
                    9]}}, ColumnSpacings -> 0.05, RowAlignments -> Center], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`v[-$CellContext`b], Editable -> False], "_"], 
               RowBox[{"\"Object\"", "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    GridBox[{{"v", 
                    StyleBox[
                    
                    GridBox[{{" "}, {"b"}}, RowSpacings -> 0, ColumnSpacings -> 
                    0.05, GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 
                    9]}}, ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`v[-$CellContext`b], Editable -> False], ",", 
                   
                   RowBox[{"{", 
                    RowBox[{"Tensor", ",", "1", ",", "v"}], "}"}], ",", 
                   RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"-", "b"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[-$CellContext`b], Editable -> 
                    False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"-", "b"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[-$CellContext`b], Editable -> 
                    False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False]}], ",", "Null"}], "}"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[{-$CellContext`TangentM3, Null}], 
                    Editable -> False]}], "}"}]}], "]"}]], #& ], ",", 
             TagBox[
              TooltipBox[
               UnderscriptBox[
                InterpretationBox[
                 StyleBox[
                  GridBox[{{"U", 
                    StyleBox[
                    
                    GridBox[{{"b", "c", "a"}, {" ", " ", " "}}, RowSpacings -> 
                    0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                   ColumnSpacings -> 0.05, RowAlignments -> Center], 
                  AutoSpacing -> False, ShowAutoStyles -> False], 
                 $CellContext`U[$CellContext`b, $CellContext`c, \
$CellContext`a], Editable -> False], "_"], 
               RowBox[{"\"Object\"", "[", 
                 RowBox[{
                   InterpretationBox[
                    StyleBox[
                    GridBox[{{"U", 
                    StyleBox[
                    
                    GridBox[{{"b", "c", "a"}, {" ", " ", " "}}, RowSpacings -> 
                    0, ColumnSpacings -> 0.05, 
                    GridFrameMargins -> {{0, 0}, {0, 0}}], FontSize -> 9]}}, 
                    ColumnSpacings -> 0.05, RowAlignments -> Center], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    $CellContext`U[$CellContext`b, $CellContext`c, \
$CellContext`a], Editable -> False], ",", 
                   RowBox[{"{", 
                    RowBox[{"Tensor", ",", "6", ",", "U"}], "}"}], ",", 
                   RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{"b", ",\[ThinSpace]", "c", ",\[ThinSpace]", "a"}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    xAct`xTensor`IndexList[$CellContext`b, $CellContext`c, \
$CellContext`a], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{"a", ",\[ThinSpace]", "b", ",\[ThinSpace]", "c"}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, \
$CellContext`c], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    xAct`xTensor`IndexList[], Editable -> False], ",", 
                    InterpretationBox[
                    StyleBox[
                    RowBox[{
                    StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                    RowBox[{
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}], ",\[ThinSpace]", 
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}], ",\[ThinSpace]", 
                    RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}]}], 
                    StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], 
                    AutoSpacing -> False, ShowAutoStyles -> False], 
                    
                    xAct`xTensor`IndexList[{$CellContext`TangentM3, 
                    Null}, {$CellContext`TangentM3, 
                    Null}, {$CellContext`TangentM3, Null}], Editable -> 
                    False]}], "}"}]}], "]"}]], #& ]}], "]"}], ",", 
         RowBox[{"{", 
           RowBox[{
             RowBox[{"{", 
               RowBox[{"Times", ",", 
                 RowBox[{"{", 
                   RowBox[{"Tensor", ",", "Tensor"}], "}"}]}], "}"}], ",", 
             RowBox[{"{", "6", "}"}], ",", 
             RowBox[{"{", 
               RowBox[{"Times", ",", 
                 RowBox[{"{", 
                   RowBox[{"v", ",", "U"}], "}"}]}], "}"}]}], "}"}], ",", 
         RowBox[{"{", 
           RowBox[{
             InterpretationBox[
              StyleBox[
               RowBox[{
                 StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                 RowBox[{
                   RowBox[{"-", "b"}], ",\[ThinSpace]", "b", ",\[ThinSpace]", 
                   "c", ",\[ThinSpace]", "a"}], 
                 StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], AutoSpacing -> 
               False, ShowAutoStyles -> False], 
              
              xAct`xTensor`IndexList[-$CellContext`b, $CellContext`b, \
$CellContext`c, $CellContext`a], Editable -> False], ",", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                 StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                 RowBox[{"a", ",\[ThinSpace]", "c"}], 
                 StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], AutoSpacing -> 
               False, ShowAutoStyles -> False], 
              xAct`xTensor`IndexList[$CellContext`a, $CellContext`c], 
              Editable -> False], ",", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                 StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                 RowBox[{
                   RowBox[{"-", "b"}], ",\[ThinSpace]", "b"}], 
                 StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], AutoSpacing -> 
               False, ShowAutoStyles -> False], 
              xAct`xTensor`IndexList[-$CellContext`b, $CellContext`b], 
              Editable -> False], ",", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                 StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                 RowBox[{}], 
                 StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], AutoSpacing -> 
               False, ShowAutoStyles -> False], 
              xAct`xTensor`IndexList[], Editable -> False], ",", 
             InterpretationBox[
              StyleBox[
               RowBox[{
                 StyleBox["{", FontColor -> RGBColor[0, 0, 1]], 
                 RowBox[{
                   RowBox[{"{", 
                    RowBox[{
                    RowBox[{"-", 
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False]}], ",", "Null"}], "}"}], 
                   ",\[ThinSpace]", 
                   RowBox[{"{", 
                    RowBox[{
                    InterpretationBox[
                    StyleBox[
                    "\[DoubleStruckCapitalT]M3", AutoSpacing -> False, 
                    ShowAutoStyles -> False], $CellContext`TangentM3, 
                    Editable -> False], ",", "Null"}], "}"}]}], 
                 StyleBox["}", FontColor -> RGBColor[0, 0, 1]]}], AutoSpacing -> 
               False, ShowAutoStyles -> False], 
              
              xAct`xTensor`IndexList[{-$CellContext`TangentM3, 
                Null}, {$CellContext`TangentM3, Null}], Editable -> False]}], 
           "}"}]}], "]"}]],
    #& ]}],
  SequenceForm["ToCanonical:: object: ", 
   xAct`xTensor`Private`Object[
    xAct`xTensor`Private`VerbatimProduct[Times][
     xAct`xTensor`Private`Object[
      $CellContext`v[-$CellContext`b], {
      xAct`xTensor`Tensor, 1, $CellContext`v}, {
       xAct`xTensor`IndexList[-$CellContext`b], 
       xAct`xTensor`IndexList[-$CellContext`b], 
       xAct`xTensor`IndexList[], 
       xAct`xTensor`IndexList[], 
       xAct`xTensor`IndexList[{-$CellContext`TangentM3, Null}]}], 
     xAct`xTensor`Private`Object[
      $CellContext`U[$CellContext`b, $CellContext`c, $CellContext`a], {
      xAct`xTensor`Tensor, 6, $CellContext`U}, {
       xAct`xTensor`IndexList[$CellContext`b, $CellContext`c, $CellContext`a], 
       xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, $CellContext`c], 
       xAct`xTensor`IndexList[], 
       xAct`xTensor`IndexList[], 
       xAct`xTensor`IndexList[{$CellContext`TangentM3, 
         Null}, {$CellContext`TangentM3, Null}, {$CellContext`TangentM3, 
         Null}]}]], {{Times, {xAct`xTensor`Tensor, xAct`xTensor`Tensor}}, {
     6}, {Times, {$CellContext`v, $CellContext`U}}}, {
     xAct`xTensor`IndexList[-$CellContext`b, $CellContext`b, $CellContext`c, \
$CellContext`a], 
     xAct`xTensor`IndexList[$CellContext`a, $CellContext`c], 
     xAct`xTensor`IndexList[-$CellContext`b, $CellContext`b], 
     xAct`xTensor`IndexList[], 
     xAct`xTensor`IndexList[{-$CellContext`TangentM3, 
       Null}, {$CellContext`TangentM3, Null}]}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897790979312*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->784957731,ExpressionUUID->"a5d885ca-9c7e-497a-8ce6-362cb8054b45"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonical:: sym: \"\>", "\[InvisibleSpace]", 
   RowBox[{"Symmetry", "[", 
    RowBox[{"4", ",", 
     RowBox[{
      RowBox[{"(", 
       TagBox[
        TooltipBox[
         UnderscriptBox["Times", "_"],
         RowBox[{"\"VerbatimProduct\"", "[", "Times", "]"}]],
        #& ], ")"}], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"\[FilledCircle]1"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[
         xAct`xTensor`Private`slot[1]],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"U", 
            StyleBox[GridBox[{
               {"\[FilledCircle]2", "\[FilledCircle]3", "\[FilledCircle]4"},
               {" ", " ", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`U[
         xAct`xTensor`Private`slot[2], 
         xAct`xTensor`Private`slot[3], 
         xAct`xTensor`Private`slot[4]],
        Editable->False]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\<\"\[FilledCircle]1\"\>", "\[Rule]", 
        RowBox[{"-", "b"}]}], ",", 
       RowBox[{"\<\"\[FilledCircle]2\"\>", "\[Rule]", "b"}], ",", 
       RowBox[{"\<\"\[FilledCircle]3\"\>", "\[Rule]", "c"}], ",", 
       RowBox[{"\<\"\[FilledCircle]4\"\>", "\[Rule]", "a"}]}], "}"}], ",", 
     RowBox[{"StrongGenSet", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"2", ",", "3", ",", "4"}], "}"}], ",", 
       RowBox[{"GenSet", "[", 
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"-", 
            RowBox[{"(", "2", ",", "3", ")"}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          -xAct`xPerm`Cycles[{2, 3}],
          Editable->False], ",", 
         InterpretationBox[
          StyleBox[
           RowBox[{"-", 
            RowBox[{"(", "3", ",", "4", ")"}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          -xAct`xPerm`Cycles[{3, 4}],
          Editable->False]}], "]"}]}], "]"}]}], "]"}]}],
  SequenceForm["ToCanonical:: sym: ", 
   xAct`xTensor`Symmetry[4, 
    xAct`xTensor`Private`VerbatimProduct[Times][
     $CellContext`v[
      xAct`xTensor`Private`slot[1]], 
     $CellContext`U[
      xAct`xTensor`Private`slot[2], 
      xAct`xTensor`Private`slot[3], 
      xAct`xTensor`Private`slot[4]]], {
    xAct`xTensor`Private`slot[1] -> -$CellContext`b, 
     xAct`xTensor`Private`slot[2] -> $CellContext`b, 
     xAct`xTensor`Private`slot[3] -> $CellContext`c, 
     xAct`xTensor`Private`slot[4] -> $CellContext`a}, 
    xAct`xPerm`StrongGenSet[{2, 3, 4}, 
     xAct`xPerm`GenSet[-xAct`xPerm`Cycles[{2, 3}], -
      xAct`xPerm`Cycles[{3, 4}]]]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779098752947`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->792541396,ExpressionUUID->"72cfd32d-c6da-44a4-bdde-c77589d2745f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Actual configuration: \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["{",
       FontColor->RGBColor[0, 0, 1]], 
      RowBox[{
       RowBox[{"-", "b"}], ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", "c",
        ",", "\[ThinSpace]", "a"}], 
      StyleBox["}",
       FontColor->RGBColor[0, 0, 1]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`IndexList[-$CellContext`b, $CellContext`b, $CellContext`c, \
$CellContext`a],
    Editable->False]}],
  SequenceForm["ToCanonicalOne:: Actual configuration: ", 
   xAct`xTensor`IndexList[-$CellContext`b, $CellContext`b, $CellContext`c, \
$CellContext`a]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779098933477`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1549663878,ExpressionUUID->"b5d01558-7dbf-4b52-9c2f-69d5a39abea7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Standard configuration: \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["{",
       FontColor->RGBColor[0, 0, 1]], 
      RowBox[{
      "a", ",", "\[ThinSpace]", "c", ",", "\[ThinSpace]", "b", ",", 
       "\[ThinSpace]", 
       RowBox[{"-", "b"}]}], 
      StyleBox["}",
       FontColor->RGBColor[0, 0, 1]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`IndexList[$CellContext`a, $CellContext`c, $CellContext`b, \
-$CellContext`b],
    Editable->False]}],
  SequenceForm["ToCanonicalOne:: Standard configuration: ", 
   xAct`xTensor`IndexList[$CellContext`a, $CellContext`c, $CellContext`b, \
-$CellContext`b]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977909912537*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->519126730,ExpressionUUID->"9cdd7a1f-ca81-400b-8ec5-35dd8938e42d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Repeated indices: \"\>", "\[InvisibleSpace]", 
   
   InterpretationBox[
    StyleBox[
     RowBox[{"{", "}"}],
     ShowAutoStyles->False,
     AutoSpacing->False,
     FontColor->RGBColor[0, 0, 1]],
    xAct`xTensor`IndexList[],
    Editable->False]}],
  SequenceForm["ToCanonicalOne:: Repeated indices: ", 
   xAct`xTensor`IndexList[]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779099328003`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1884376996,ExpressionUUID->"73872195-1695-4c2d-944b-9dd319f008cc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Repeated indices: \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"{", "}"}]}],
  SequenceForm["ToCanonicalOne:: Repeated indices: ", {}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779099540977`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1782829463,ExpressionUUID->"8abfab5c-f343-4fc8-bc05-99323e6feeaa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Permutation to be canonicalized: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], "]"}]}],
  SequenceForm["ToCanonicalOne:: Permutation to be canonicalized: ", 
   xAct`xPerm`Images[{4, 3, 2, 1}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977909972831*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->175458050,ExpressionUUID->"b355a96f-982f-41b3-bf61-2dd347fb5592"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"dummies: \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["{",
       FontColor->RGBColor[0, 0, 1]], "b", 
      StyleBox["}",
       FontColor->RGBColor[0, 0, 1]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`IndexList[$CellContext`b],
    Editable->False]}],
  SequenceForm["dummies: ", 
   xAct`xTensor`IndexList[$CellContext`b]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977909991448*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->428399019,ExpressionUUID->"df60d226-111b-45b1-b068-9ea8b89941c3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: dummysets_tmp: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"DummySet", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False], ",", 
      RowBox[{"{", "b", "}"}], ",", "0"}], "]"}], "}"}]}],
  SequenceForm["ToCanonicalOne:: dummysets_tmp: ", {
    xAct`xPerm`DummySet[$CellContext`TangentM3, {$CellContext`b}, 0]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779100113153`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1737127589,ExpressionUUID->"0420c512-3394-48dc-8702-8e61367d8569"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: dummysets: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"DummySet", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False], ",", 
      RowBox[{"{", 
       RowBox[{"{", 
        RowBox[{"3", ",", "4"}], "}"}], "}"}], ",", "0"}], "]"}], "}"}]}],
  SequenceForm["ToCanonicalOne:: dummysets: ", {
    xAct`xPerm`DummySet[$CellContext`TangentM3, {{3, 4}}, 0]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779100299473`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->863207358,ExpressionUUID->"e07cf8cb-c6be-4ad1-8c40-7d7603eb69b9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Free indices: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}]}],
  SequenceForm["ToCanonicalOne:: Free indices: ", {1, 2}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910050907*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1674030421,ExpressionUUID->"00b32bf5-5862-481e-a1e8-737e83b8d2bd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: calling: \"\>", 
   "\[InvisibleSpace]", "\<\"CanonicalPerm[\"\>", "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], "]"}], 
   "\[InvisibleSpace]", "\<\",\"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\",\"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", "3", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[
         RowBox[{"-", 
          RowBox[{"(", "2", ",", "3", ")"}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        -xAct`xPerm`Cycles[{2, 3}],
        Editable->False], ",", 
       InterpretationBox[
        StyleBox[
         RowBox[{"-", 
          RowBox[{"(", "3", ",", "4", ")"}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        -xAct`xPerm`Cycles[{3, 4}],
        Editable->False]}], "]"}]}], "]"}], "\[InvisibleSpace]", "\<\",\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], "\[InvisibleSpace]", "\<\",\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"DummySet", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False], ",", 
      RowBox[{"{", 
       RowBox[{"{", 
        RowBox[{"3", ",", "4"}], "}"}], "}"}], ",", "0"}], "]"}], "}"}], 
   "\[InvisibleSpace]", "\<\",\"\>", "\[InvisibleSpace]", 
   RowBox[{"Verbose", "\[Rule]", "True"}], "\[InvisibleSpace]", "\<\"]\"\>"}],
  
  SequenceForm["ToCanonicalOne:: calling: ", "CanonicalPerm[", 
   xAct`xPerm`Images[{4, 3, 2, 1}], ",", 4, ",", 
   xAct`xPerm`StrongGenSet[{2, 3, 4}, 
    xAct`xPerm`GenSet[-xAct`xPerm`Cycles[{2, 3}], -
     xAct`xPerm`Cycles[{3, 4}]]], ",", {1, 2}, ",", {
    xAct`xPerm`DummySet[$CellContext`TangentM3, {{3, 4}}, 0]}, ",", Verbose -> 
   True, "]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910072837*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1246026400,ExpressionUUID->"811cbc84-89b8-47c2-834c-513d5ed30d01"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonicalOne:: Canonical permutation: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Images", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "ToCanonicalOne:: Canonical permutation: ", -
   xAct`xPerm`Images[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779100927677`*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1852662903,ExpressionUUID->"111d435f-fc9d-45a5-a1dc-ac15dc27932d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonical:: newindices: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1"}], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        StyleBox["{",
         FontColor->RGBColor[0, 0, 1]], 
        RowBox[{
         RowBox[{"-", "b"}], ",", "\[ThinSpace]", "a", ",", "\[ThinSpace]", 
         "c", ",", "\[ThinSpace]", "b"}], 
        StyleBox["}",
         FontColor->RGBColor[0, 0, 1]]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`IndexList[-$CellContext`b, $CellContext`a, $CellContext`c, \
$CellContext`b],
      Editable->False]}], "}"}]}],
  SequenceForm["ToCanonical:: newindices: ", {-1, 
    xAct`xTensor`IndexList[-$CellContext`b, $CellContext`a, $CellContext`c, \
$CellContext`b]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910111841*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1340012868,ExpressionUUID->"4dab8de6-07e4-4a95-97c0-93a22154ba95"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonical:: riskyders: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}]}],
  SequenceForm["ToCanonical:: riskyders: ", {}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910130909*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->935266387,ExpressionUUID->"bd738d51-aad3-4b89-9423-ccd2c495057b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"ToCanonical:: result: \"\>", "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{
     RowBox[{"(", 
      TagBox[
       TooltipBox[
        UnderscriptBox["Times", "_"],
        RowBox[{"\"VerbatimProduct\"", "[", "Times", "]"}]],
       #& ], ")"}], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`b],
       Editable->False], ",", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"a", "c", "b"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`a, $CellContext`c, $CellContext`b],
       Editable->False]}], "]"}]}]}],
  SequenceForm[
  "ToCanonical:: result: ", -xAct`xTensor`Private`VerbatimProduct[Times][
     $CellContext`v[-$CellContext`b], 
     $CellContext`U[$CellContext`a, $CellContext`c, $CellContext`b]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791022159*^9},
 CellLabel->"During evaluation of In[292]:=",
 CellID->1589353916,ExpressionUUID->"253fc659-50a2-4a7a-9b8c-553024b42fd1"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "c", "b"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`c, $CellContext`b],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289779103025503`*^9},
 CellLabel->"Out[292]=",
 CellID->1869664826,ExpressionUUID->"4050f88c-e41b-4557-b8fc-8d8745644cff"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"b", ",", "c", ",", "a"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], ",", 
   RowBox[{"xPermVerbose", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[293]:=",ExpressionUUID->"c2889709-49ef-47a9-b5ec-0758cfa9dcc7"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"xPermVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 293, 33, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289779104614964`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1476007678,ExpressionUUID->"5231426f-eb63-4329-9d37-b841d7c96db4"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"xPermVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 293, 34, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289779105251513`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1677437649,ExpressionUUID->"398913a0-557a-4c49-a6a9-5555bf9edf36"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free indices at slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"4", ",", "3"}], "}"}]}],
  SequenceForm["Free indices at slots: ", {4, 3}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779105831013`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->529248927,ExpressionUUID->"243e4bdd-4ba1-4f47-aba7-57169bcbe72d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"RIGHT-COSET-REPRESENTATIVE ALGORITHM for \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], "]"}]}],
  SequenceForm["RIGHT-COSET-REPRESENTATIVE ALGORITHM for ", 
   xAct`xPerm`Images[{4, 3, 2, 1}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779105970087`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->209450215,ExpressionUUID->"5ba65501-e9ee-4ed1-9859-6830a916f654"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"which corresponds to the index list: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Perm", "[", 
    RowBox[{"{", 
     RowBox[{"4", ",", "3", ",", "2", ",", "1"}], "}"}], "]"}]}],
  SequenceForm["which corresponds to the index list: ", 
   xAct`xPerm`Perm[{4, 3, 2, 1}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910617559*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->504066072,ExpressionUUID->"7e63b828-dfea-4162-bf2f-d96e5462c44c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"base: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}]}],
  SequenceForm["base: ", {1, 2, 3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910637019*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->15484689,ExpressionUUID->"79bb82e2-cd27-4d2f-b5f9-acab31769755"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"****** Analysing element i=\"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of base: slot \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" ******\"\>"}],
  SequenceForm[
  "****** Analysing element i=", 1, " of base: slot ", 1, " ******"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910655272*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->405646736,ExpressionUUID->"b0317c6e-0c90-4e7a-bdea-e38cc1b05486"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Symmetry orbit Delta of slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}]}],
  SequenceForm["Symmetry orbit Delta of slots: ", {1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910675168*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1591596374,ExpressionUUID->"48bca0d5-9802-4c64-ba31-e76d2c4200c0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"4", ",", "3"}], "}"}]}],
  SequenceForm["Free slots: ", {4, 3}],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791069357*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->903950786,ExpressionUUID->"ef60a04c-4c31-4dbb-8f15-1a1cb250ff99"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free slots that can go to that slot: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "}"}]}],
  SequenceForm["Free slots that can go to that slot: ", {}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910712183*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1401541424,ExpressionUUID->"2543264f-f02a-4ffe-a651-7f74d0666bc8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"****** Analysing element i=\"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\" of base: slot \"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\" ******\"\>"}],
  SequenceForm[
  "****** Analysing element i=", 2, " of base: slot ", 2, " ******"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779107308598`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2106072872,ExpressionUUID->"099001d9-f82f-40fa-85f0-d9195f40a54c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Symmetry orbit Delta of slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2", ",", "3", ",", "4"}], "}"}]}],
  SequenceForm["Symmetry orbit Delta of slots: ", {2, 3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779107495327`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1941289850,ExpressionUUID->"1e4ea39c-d562-4ef2-80f7-4f2716e92ce7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"4", ",", "3"}], "}"}]}],
  SequenceForm["Free slots: ", {4, 3}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910767071*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->539460079,ExpressionUUID->"34b134f6-bf7e-4afa-81c1-e492ec2b84cb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free slots that can go to that slot: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"3", ",", "4"}], "}"}]}],
  SequenceForm["Free slots that can go to that slot: ", {3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910784815*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->29590119,ExpressionUUID->"211bb0f0-f78d-435e-9159-aac2286bb8f9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"At those slots we respectively find indices Deltap: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2", ",", "1"}], "}"}]}],
  SequenceForm["At those slots we respectively find indices Deltap: ", {2, 1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779108042*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1252407576,ExpressionUUID->"53b396f8-0c09-49e9-a603-bf9b74b25fec"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The least index is \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\", found at position pk: \"\>", 
   "\[InvisibleSpace]", "2", "\[InvisibleSpace]", "\<\" of Deltap\"\>"}],
  SequenceForm[
  "The least index is ", 1, ", found at position pk: ", 2, " of Deltap"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779108221827`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1726905585,ExpressionUUID->"0762422c-c3b0-46ac-a106-92ba42cd9c3d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"That index is found in tensor at slot pp: \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["That index is found in tensor at slot pp: ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910905601*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->860677890,ExpressionUUID->"7580a49c-1430-491e-b3eb-9b90235b551d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"We can move slot \"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" to slot \"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\" using permutation om: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "4", ",", "2", ",", "3"}], "}"}], "]"}], 
   "\[InvisibleSpace]", "\<\" in S\"\>"}],
  SequenceForm[
  "We can move slot ", 4, " to slot ", 2, " using permutation om: ", 
   xAct`xPerm`Images[{1, 4, 2, 3}], " in S"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977910924819*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2110934685,ExpressionUUID->"f112c2a3-6ac7-4570-a84d-2999ecb9dfd9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New indices list: \"\>", "\[InvisibleSpace]", 
   RowBox[{"Perm", "[", 
    RowBox[{"{", 
     RowBox[{"4", ",", "1", ",", "3", ",", "2"}], "}"}], "]"}]}],
  SequenceForm["New indices list: ", 
   xAct`xPerm`Perm[{4, 1, 3, 2}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779109445887`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2049402355,ExpressionUUID->"ac3becae-d0e3-4690-8314-4a6f8354addd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computing stabilizer in S of slot \"\>", "\[InvisibleSpace]", 
   "2"}],
  SequenceForm["Computing stabilizer in S of slot ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779109637127`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->815166252,ExpressionUUID->"c486b8f9-27ab-4811-8235-6ad32de39d0d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"SGS before change: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{"Images", "[", 
         RowBox[{"{", 
          RowBox[{"1", ",", "3", ",", "2", ",", "4"}], "}"}], "]"}]}], ",", 
       RowBox[{"-", 
        RowBox[{"Images", "[", 
         RowBox[{"{", 
          RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], "]"}]}]}], 
      "]"}]}], "]"}]}],
  SequenceForm["SGS before change: ", 
   xAct`xPerm`StrongGenSet[{1, 2, 3, 4}, 
    xAct`xPerm`GenSet[-xAct`xPerm`Images[{1, 3, 2, 4}], -
     xAct`xPerm`Images[{1, 2, 4, 3}]]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779109820547`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1701013151,ExpressionUUID->"7a46fb19-db54-4da5-8808-4d927d052c16"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"SGS after change: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", "1", ",", "3", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{"Images", "[", 
         RowBox[{"{", 
          RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], "]"}]}], ",", 
       RowBox[{"-", 
        RowBox[{"Images", "[", 
         RowBox[{"{", 
          RowBox[{"1", ",", "3", ",", "2", ",", "4"}], "}"}], "]"}]}]}], 
      "]"}]}], "]"}]}],
  SequenceForm["SGS after change: ", 
   xAct`xPerm`StrongGenSet[{2, 1, 3, 4}, 
    xAct`xPerm`GenSet[-xAct`xPerm`Images[{1, 2, 4, 3}], -
     xAct`xPerm`Images[{1, 3, 2, 4}]]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911000969*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1625719993,ExpressionUUID->"76466d00-b884-4bc5-87a7-362c43a15fba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"SGS after stabilization: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "3", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      RowBox[{"-", 
       RowBox[{"Images", "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], "]"}]}], "]"}]}],
     "]"}]}],
  SequenceForm["SGS after stabilization: ", 
   xAct`xPerm`StrongGenSet[{1, 3, 4}, 
    xAct`xPerm`GenSet[-xAct`xPerm`Images[{1, 2, 4, 3}]]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911021999*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1041481570,ExpressionUUID->"8a8d9258-eec0-4e1c-b3a7-410245c8bed1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"newbase after change: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "3", ",", "4"}], "}"}]}],
  SequenceForm["newbase after change: ", {1, 3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779110403833`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->41540293,ExpressionUUID->"fffe2cc6-9187-41c1-b4af-f983868935b9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"****** Analysing element i=\"\>", "\[InvisibleSpace]", "3", 
   "\[InvisibleSpace]", "\<\" of base: slot \"\>", "\[InvisibleSpace]", "3", 
   "\[InvisibleSpace]", "\<\" ******\"\>"}],
  SequenceForm[
  "****** Analysing element i=", 3, " of base: slot ", 3, " ******"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911059689*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->235519176,ExpressionUUID->"d3a48a63-f401-4915-aaf0-9cc34acc2ae2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Symmetry orbit Delta of slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"3", ",", "4"}], "}"}]}],
  SequenceForm["Symmetry orbit Delta of slots: ", {3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911078595*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->558429611,ExpressionUUID->"b59bf584-0952-41c4-995e-c2f427b4b464"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free slots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"2", ",", "4"}], "}"}]}],
  SequenceForm["Free slots: ", {2, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779110969763`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1022934687,ExpressionUUID->"69a067ff-37e7-4652-8197-80a4f86ea1c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free slots that can go to that slot: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["Free slots that can go to that slot: ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911115944*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1856690174,ExpressionUUID->"ee1da9f2-76a8-4137-af25-b7368925e3a2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"At those slots we respectively find indices Deltap: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "2", "}"}]}],
  SequenceForm["At those slots we respectively find indices Deltap: ", {2}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911134411*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->306879861,ExpressionUUID->"6ee556bb-f0e1-40c9-b2b8-8cfe61ec716e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The least index is \"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\", found at position pk: \"\>", 
   "\[InvisibleSpace]", "1", "\[InvisibleSpace]", "\<\" of Deltap\"\>"}],
  SequenceForm[
  "The least index is ", 2, ", found at position pk: ", 1, " of Deltap"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911152946*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1621587380,ExpressionUUID->"9f36bd8d-d7c6-49d0-882f-9841a5e5e556"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"That index is found in tensor at slot pp: \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["That index is found in tensor at slot pp: ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911171125*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->304131583,ExpressionUUID->"cbc9c307-277c-4be2-8a60-802c4435fd56"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"We can move slot \"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" to slot \"\>", "\[InvisibleSpace]", "3", 
   "\[InvisibleSpace]", "\<\" using permutation om: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Images", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], "]"}]}], 
   "\[InvisibleSpace]", "\<\" in S\"\>"}],
  SequenceForm[
  "We can move slot ", 4, " to slot ", 3, " using permutation om: ", -
   xAct`xPerm`Images[{1, 2, 4, 3}], " in S"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911189516*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->528435621,ExpressionUUID->"7502fc87-ab52-4a88-8b43-bc17a28358bb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New indices list: \"\>", "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Perm", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm["New indices list: ", -xAct`xPerm`Perm[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911208036*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1572601802,ExpressionUUID->"f80b17f0-b380-49ef-92f0-dbb7b47f75aa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Computing stabilizer in S of slot \"\>", "\[InvisibleSpace]", 
   "3"}],
  SequenceForm["Computing stabilizer in S of slot ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911226294*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1637242585,ExpressionUUID->"952832eb-83c8-4f92-be1b-23ecaaa6f3ad"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"SGS before change: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "3", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      RowBox[{"-", 
       RowBox[{"Images", "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], "]"}]}], "]"}]}],
     "]"}]}],
  SequenceForm["SGS before change: ", 
   xAct`xPerm`StrongGenSet[{1, 3, 4}, 
    xAct`xPerm`GenSet[-xAct`xPerm`Images[{1, 2, 4, 3}]]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911245967*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1428157084,ExpressionUUID->"c67cd88e-7046-437e-89f0-00726e4bfd9b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"SGS after change: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"3", ",", "1", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      RowBox[{"-", 
       RowBox[{"Images", "[", 
        RowBox[{"{", 
         RowBox[{"1", ",", "2", ",", "4", ",", "3"}], "}"}], "]"}]}], "]"}]}],
     "]"}]}],
  SequenceForm["SGS after change: ", 
   xAct`xPerm`StrongGenSet[{3, 1, 4}, 
    xAct`xPerm`GenSet[-xAct`xPerm`Images[{1, 2, 4, 3}]]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911264657*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->611388269,ExpressionUUID->"94abc7bb-3c81-4087-9620-3a17c282f029"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"SGS after stabilization: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["SGS after stabilization: ", 
   xAct`xPerm`StrongGenSet[{1, 4}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779112868032`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2040869835,ExpressionUUID->"4df27b2a-8e93-4a04-807e-ff01b4a9c115"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"newbase after change: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "4"}], "}"}]}],
  SequenceForm["newbase after change: ", {1, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779113716297`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1290506961,ExpressionUUID->"3e9aebfd-fe05-4fb6-91a7-5582772a8b13"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Canonical Permutation after RightCosetRepresentative: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Images", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "Canonical Permutation after RightCosetRepresentative: ", -
   xAct`xPerm`Images[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911391652*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2113142474,ExpressionUUID->"ead7db8e-237b-4454-9847-3964fb40bec2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"DOUBLE-COSET-REPRESENTATIVE ALGORITHM for \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Images", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "DOUBLE-COSET-REPRESENTATIVE ALGORITHM for ", -
   xAct`xPerm`Images[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779114107533`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->501006432,ExpressionUUID->"69b8e4de-12a4-4cae-b11b-bdfb1d9ec0c2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"index-dummysets: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"DummySet", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False], ",", 
      RowBox[{"{", 
       RowBox[{"{", 
        RowBox[{"3", ",", "4"}], "}"}], "}"}], ",", "0"}], "]"}], "}"}]}],
  SequenceForm["index-dummysets: ", {
    xAct`xPerm`DummySet[$CellContext`TangentM3, {{3, 4}}, 0]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911429545*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->131722737,ExpressionUUID->"1825dc49-cf5f-4434-bcdf-11174e901eae"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"dummyindices: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"3", ",", "4"}], "}"}]}],
  SequenceForm["dummyindices: ", {3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911447947*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1955884349,ExpressionUUID->"3a294659-444b-4ad1-93a7-4f6b566f055b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"dummyslots: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"4", ",", "1"}], "}"}]}],
  SequenceForm["dummyslots: ", {4, 1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779114664297`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1016069014,ExpressionUUID->"0d882871-b7bf-4ea4-9754-c435d5f566f8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Extended \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "4"}], "}"}], "\[InvisibleSpace]", "\<\" to \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "4"}], "}"}]}],
  SequenceForm["Extended ", {1, 4}, " to ", {1, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911484767*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->281957354,ExpressionUUID->"b2ce695d-e9fc-494f-aca8-4fa7821a388d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Initial SGSS: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "4"}], "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["Initial SGSS: ", 
   xAct`xPerm`StrongGenSet[{1, 4}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779115032187`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1508323396,ExpressionUUID->"df2bec7b-a199-47d8-9aca-c78c89744293"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"base for sorting: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"3", ",", "4"}], "}"}]}],
  SequenceForm["base for sorting: ", {3, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779115210543`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1494347384,ExpressionUUID->"03d3b3e1-cafe-4f90-9b11-5dcff5108a03"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Initial SGSD: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["Initial SGSD: ", 
   xAct`xPerm`StrongGenSet[{3}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911539119*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->675231223,ExpressionUUID->"ffc00945-9edb-4198-a6a3-71753593521c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"******************* Loop i= \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" *********************\"\>"}],
  SequenceForm["******************* Loop i= ", 1, " *********************"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911558138*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1307574213,ExpressionUUID->"71e759a6-8198-4788-bb41-bf0b9ac0dda7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Analyzing slot \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" of tensor\"\>"}],
  SequenceForm["Analyzing slot ", 1, " of tensor"],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779115773067`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1199638140,ExpressionUUID->"aaf95585-25ed-46a3-a17e-b42b524fe27d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"nuS: \"\>", "\[InvisibleSpace]", 
   RowBox[{"Schreier", "[", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\" with first element \"\>", "\[InvisibleSpace]", 
   "1"}],
  SequenceForm["nuS: ", 
   xAct`xPerm`Schreier[{1}, {2}, {3}, {4}, {0, 0, 0, 0}, {0, 0, 0, 0}], 
   " with first element ", 1],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911596715*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1741740944,ExpressionUUID->"9cbcaa9c-a651-456f-b2fc-634451b0bd1c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Under S, slot \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" can go to slots Deltab: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}]}],
  SequenceForm["Under S, slot ", 1, " can go to slots Deltab: ", {1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911617423*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1090453551,ExpressionUUID->"bc3643dc-d875-46b4-906d-d1a97ad90ec8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Orbits of indices under D: DeltaD: \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}]}], "}"}]}],
  SequenceForm["Orbits of indices under D: DeltaD: ", {{1}, {2}, {3}, {4}}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911639064*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->627388159,ExpressionUUID->"7c706081-31f7-4b43-bb31-cca6de4c6750"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"With L=\"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}], "\[InvisibleSpace]", "\<\" we get sgd: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Images", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "With L=", {}, " we get sgd: ", -xAct`xPerm`Images[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911657456*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->368081543,ExpressionUUID->"7bfec5e3-9c3f-401e-8d4d-c7368053d47a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"which maps slots in Deltab to indices list: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["which maps slots in Deltab to indices list: ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779116754923`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1593589841,ExpressionUUID->"45bcdb9b-7b7a-474c-9cec-a4223302a81a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"whose points belong to orbits \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["whose points belong to orbits ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791169462*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->48412303,ExpressionUUID->"4a900eb7-3558-44ee-a994-71855baf7b95"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Therefore at slot \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" we can have indices IMAGES: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["Therefore at slot ", 1, " we can have indices IMAGES: ", {4}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911717071*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1916757955,ExpressionUUID->"81e1563a-a93e-4c94-9e62-a12da6e6b395"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The least of them is p[[\"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\"]]: \"\>", "\[InvisibleSpace]", "4"}],
  SequenceForm["The least of them is p[[", 1, "]]: ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911736676*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->548761038,ExpressionUUID->"5c94b873-4f4d-4818-9bbf-f07c33934a74"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Moved pairs \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"DummySet", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False], ",", 
      RowBox[{"{", 
       RowBox[{"{", 
        RowBox[{"4", ",", "3"}], "}"}], "}"}], ",", "0"}], "]"}], "}"}]}],
  SequenceForm["Moved pairs ", {
    xAct`xPerm`DummySet[$CellContext`TangentM3, {{4, 3}}, 0]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791188356*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1731985448,ExpressionUUID->"5f539234-b2da-4e92-b955-27f1e0bcbd3c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New SGS of D: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "4", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["New SGS of D: ", 
   xAct`xPerm`StrongGenSet[{4}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911901607*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->338589451,ExpressionUUID->"62a1b77f-24e9-41d8-b866-ba601938fc58"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"with Schreier vector nuD: \"\>", "\[InvisibleSpace]", 
   RowBox[{"Schreier", "[", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "]"}]}],
  SequenceForm["with Schreier vector nuD: ", 
   xAct`xPerm`Schreier[{1}, {2}, {3}, {4}, {0, 0, 0, 0}, {0, 0, 0, 0}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911920732*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1988722054,ExpressionUUID->"d547ba90-005c-4cd6-99df-f8833b8c100f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"In particular, the orbit of index \"\>", "\[InvisibleSpace]", 
   "4", "\[InvisibleSpace]", "\<\" is Deltap: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["In particular, the orbit of index ", 4, " is Deltap: ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779119405518`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1016079670,ExpressionUUID->"6720eb18-d867-4266-a7d7-0c9587507852"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Now looking for all permutations sgd that move index \"\>", 
   "\[InvisibleSpace]", "4", "\[InvisibleSpace]", "\<\" to slot \"\>", 
   "\[InvisibleSpace]", "1"}],
  SequenceForm[
  "Now looking for all permutations sgd that move index ", 4, " to slot ", 1],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911960988*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->461052746,ExpressionUUID->"3ba0ddcc-e929-46d8-913b-a6a800586033"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Loop with l=\"\>", "\[InvisibleSpace]", "1"}],
  SequenceForm["Loop with l=", 1],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977911981635*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->792503646,ExpressionUUID->"18d4fdfa-9312-4ca3-8437-b16bb79100b9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"L= \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "}"}]}],
  SequenceForm["L= ", {}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779119996*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->915199628,ExpressionUUID->"ed9eae32-bae6-4184-bbeb-3f35c116ed1f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"TAB[L] = {s, d} = \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"Images", "[", 
      RowBox[{"{", 
       RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}], ",", 
     RowBox[{"Images", "[", 
      RowBox[{"{", 
       RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}], "}"}]}],
  SequenceForm["TAB[L] = {s, d} = ", {
    xAct`xPerm`Images[{1, 2, 3, 4}], 
    xAct`xPerm`Images[{1, 2, 3, 4}]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779120181828`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1482267982,ExpressionUUID->"3b2a13c1-7676-45da-9bac-cfcae48e9f3f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Calculating NEXT. We need the intersection of sets of slots \"\
\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}], "\[InvisibleSpace]", "\<\" and \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}]}],
  SequenceForm[
  "Calculating NEXT. We need the intersection of sets of slots ", {1}, 
   " and ", {1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912036674*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1667668274,ExpressionUUID->"edbb2532-4ccc-4238-ae0e-4361b8af4d33"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Intermediate slots NEXT= \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}]}],
  SequenceForm["Intermediate slots NEXT= ", {1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779120551243`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1691604121,ExpressionUUID->"e07c81b0-6d0c-4680-9a5b-a056f4052fee"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"From slot \"\>", "\[InvisibleSpace]", "1", 
   "\[InvisibleSpace]", "\<\" to intermediate slot \"\>", "\[InvisibleSpace]",
    "1", "\[InvisibleSpace]", "\<\" use s1=\"\>", "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}],
  SequenceForm["From slot ", 1, " to intermediate slot ", 1, " use s1=", 
   xAct`xPerm`Images[{1, 2, 3, 4}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791207405*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->254459014,ExpressionUUID->"e2f78199-e2d9-4eb7-a658-09ce598fd232"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"d1= \"\>", "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}],
  SequenceForm["d1= ", 
   xAct`xPerm`Images[{1, 2, 3, 4}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779120936193`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1052707121,ExpressionUUID->"f7256c8d-784d-4408-9d9f-40e91682fb4e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"L1= \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}]}],
  SequenceForm["L1= ", {1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912112686*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1878298661,ExpressionUUID->"7f841b7c-bc6c-4822-96fa-15cec466db02"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"This gives us the new index configuration: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Perm", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "This gives us the new index configuration: ", -
   xAct`xPerm`Perm[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779121322193`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->555984527,ExpressionUUID->"ac9d1d54-075f-41a2-8bb1-98956e51a64d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New ALPHA: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", "1", "}"}], "}"}]}],
  SequenceForm["New ALPHA: ", {{1}}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912151641*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->730597192,ExpressionUUID->"4561b92d-cdf2-434e-a065-ada9cfd93c08"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Checking consistency in set \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"-", 
     RowBox[{"Images", "[", 
      RowBox[{"{", 
       RowBox[{"2", ",", "3", ",", "4", ",", "1"}], "}"}], "]"}]}], "}"}]}],
  SequenceForm[
  "Checking consistency in set ", {-xAct`xPerm`Images[{2, 3, 4, 1}]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912170149*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1982996045,ExpressionUUID->"f843d385-e3d7-406b-97b5-efb29926892b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Removing permutations from SGS of S that move slot \"\>", 
   "\[InvisibleSpace]", "1"}],
  SequenceForm["Removing permutations from SGS of S that move slot ", 1],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912189602*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1415810522,ExpressionUUID->"82e5b943-24d9-4334-b5c9-42976217a92e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New SGS of S: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "4", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["New SGS of S: ", 
   xAct`xPerm`StrongGenSet[{4}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912210618*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1408634494,ExpressionUUID->"00579408-3e7d-45d8-8edf-20294f115387"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Removing permutations from SGS of D that move index \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["Removing permutations from SGS of D that move index ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779122290087`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1060216130,ExpressionUUID->"5db8ec80-7cd5-4e4b-8d4e-e18d396fff55"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New SGS of D: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["New SGS of D: ", 
   xAct`xPerm`StrongGenSet[{}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791224673*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1380677751,ExpressionUUID->"72df186b-5195-4727-bce0-87d0171e7711"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"******************* Loop i= \"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\" *********************\"\>"}],
  SequenceForm["******************* Loop i= ", 2, " *********************"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912265108*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1490155222,ExpressionUUID->"e52729ff-0f25-4255-aa29-4a0a041c58e4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Analyzing slot \"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" of tensor\"\>"}],
  SequenceForm["Analyzing slot ", 4, " of tensor"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912405506*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1084524840,ExpressionUUID->"f644a698-9277-43cb-8995-43c287ce0990"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"nuS: \"\>", "\[InvisibleSpace]", 
   RowBox[{"Schreier", "[", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\" with first element \"\>", "\[InvisibleSpace]", 
   "4"}],
  SequenceForm["nuS: ", 
   xAct`xPerm`Schreier[{1}, {2}, {3}, {4}, {0, 0, 0, 0}, {0, 0, 0, 0}], 
   " with first element ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912426899*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1915273791,ExpressionUUID->"70760835-627a-4317-a158-59f4a52d0748"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Under S, slot \"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" can go to slots Deltab: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["Under S, slot ", 4, " can go to slots Deltab: ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779124460497`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1374220454,ExpressionUUID->"65a16a5f-bf60-444b-99d3-f710fda3ce95"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Orbits of indices under D: DeltaD: \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}]}], "}"}]}],
  SequenceForm["Orbits of indices under D: DeltaD: ", {{1}, {2}, {3}, {4}}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779124643183`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->336546893,ExpressionUUID->"abf82003-886d-4547-b309-d60354f08913"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"With L=\"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}], "\[InvisibleSpace]", "\<\" we get sgd: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Images", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "With L=", {1}, " we get sgd: ", -xAct`xPerm`Images[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912482259*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2019188100,ExpressionUUID->"2796f95e-2205-426f-b6e4-82eb89cfc206"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"which maps slots in Deltab to indices list: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "3", "}"}]}],
  SequenceForm["which maps slots in Deltab to indices list: ", {3}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912500437*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1957806806,ExpressionUUID->"bdd910f6-dc85-4526-a663-de5612d3f05d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"whose points belong to orbits \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "3", "}"}]}],
  SequenceForm["whose points belong to orbits ", {3}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779125188713`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1114549108,ExpressionUUID->"1450ba03-7bfa-4e2b-b89d-843d8bc46c4f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Therefore at slot \"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" we can have indices IMAGES: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "3", "}"}]}],
  SequenceForm["Therefore at slot ", 4, " we can have indices IMAGES: ", {3}],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912537353*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1874489022,ExpressionUUID->"6ae1841c-1764-4b88-8775-d299767047b6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The least of them is p[[\"\>", "\[InvisibleSpace]", "2", 
   "\[InvisibleSpace]", "\<\"]]: \"\>", "\[InvisibleSpace]", "3"}],
  SequenceForm["The least of them is p[[", 2, "]]: ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912558123*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->951891264,ExpressionUUID->"b15c3bcb-ee4b-4658-8120-c59dd5d025b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Moved pairs \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"DummySet", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False], ",", 
      RowBox[{"{", "}"}], ",", "0"}], "]"}], "}"}]}],
  SequenceForm["Moved pairs ", {
    xAct`xPerm`DummySet[$CellContext`TangentM3, {}, 0]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912578216*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1830787545,ExpressionUUID->"3becc5dd-c42a-4a1d-a03f-f8c2983c291d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New SGS of D: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["New SGS of D: ", 
   xAct`xPerm`StrongGenSet[{}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912598184*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->900574599,ExpressionUUID->"bee12b09-57e9-40fb-be21-6e311df1f64a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"with Schreier vector nuD: \"\>", "\[InvisibleSpace]", 
   RowBox[{"Schreier", "[", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}], ",", 
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]}], "]"}]}],
  SequenceForm["with Schreier vector nuD: ", 
   xAct`xPerm`Schreier[{1}, {2}, {3}, {4}, {0, 0, 0, 0}, {0, 0, 0, 0}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912616877*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->492541337,ExpressionUUID->"202131ae-cca0-4387-ae1b-f466f66e0d25"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"In particular, the orbit of index \"\>", "\[InvisibleSpace]", 
   "3", "\[InvisibleSpace]", "\<\" is Deltap: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "3", "}"}]}],
  SequenceForm["In particular, the orbit of index ", 3, " is Deltap: ", {3}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912634995*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1740275421,ExpressionUUID->"21ab58db-d32b-4faf-824b-91a3f5c7639a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Now looking for all permutations sgd that move index \"\>", 
   "\[InvisibleSpace]", "3", "\[InvisibleSpace]", "\<\" to slot \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm[
  "Now looking for all permutations sgd that move index ", 3, " to slot ", 4],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912653327*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->81728607,ExpressionUUID->"c1a1fa7a-dfea-45ec-b1ed-3e126aa016eb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Loop with l=\"\>", "\[InvisibleSpace]", "1"}],
  SequenceForm["Loop with l=", 1],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779126719093`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1370647416,ExpressionUUID->"12ddd58c-28f0-4450-bbd0-1acc80c79a02"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"L= \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}]}],
  SequenceForm["L= ", {1}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912690185*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->414039343,ExpressionUUID->"35071e22-8725-4cab-8153-5261168752b7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"TAB[L] = {s, d} = \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"Images", "[", 
      RowBox[{"{", 
       RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}], ",", 
     RowBox[{"Images", "[", 
      RowBox[{"{", 
       RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}], "}"}]}],
  SequenceForm["TAB[L] = {s, d} = ", {
    xAct`xPerm`Images[{1, 2, 3, 4}], 
    xAct`xPerm`Images[{1, 2, 3, 4}]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779127089043`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->912221521,ExpressionUUID->"0a06c632-3210-4bd9-8a0f-3cbba4e1166f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Calculating NEXT. We need the intersection of sets of slots \"\
\>", "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}], "\[InvisibleSpace]", "\<\" and \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm[
  "Calculating NEXT. We need the intersection of sets of slots ", {4}, 
   " and ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912729225*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->821347514,ExpressionUUID->"a8b6a7c2-7702-4275-8611-28dc5da42b0b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Intermediate slots NEXT= \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "4", "}"}]}],
  SequenceForm["Intermediate slots NEXT= ", {4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912748232*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->362784882,ExpressionUUID->"ac544d9a-fd18-4860-b039-e523b9f5e397"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"From slot \"\>", "\[InvisibleSpace]", "4", 
   "\[InvisibleSpace]", "\<\" to intermediate slot \"\>", "\[InvisibleSpace]",
    "4", "\[InvisibleSpace]", "\<\" use s1=\"\>", "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}],
  SequenceForm["From slot ", 4, " to intermediate slot ", 4, " use s1=", 
   xAct`xPerm`Images[{1, 2, 3, 4}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912767539*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->619437941,ExpressionUUID->"7ace7537-c0f8-42b8-93e9-2e2c9b77113e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"d1= \"\>", "\[InvisibleSpace]", 
   RowBox[{"Images", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}],
  SequenceForm["d1= ", 
   xAct`xPerm`Images[{1, 2, 3, 4}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912855109*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2040237378,ExpressionUUID->"3debb102-04d1-413d-bdab-e6f02f705fe9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"L1= \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "4"}], "}"}]}],
  SequenceForm["L1= ", {1, 4}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912874899*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1398220397,ExpressionUUID->"90f3c059-dff5-4bab-aead-0cee2b35c538"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"This gives us the new index configuration: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", 
    RowBox[{"Perm", "[", 
     RowBox[{"{", 
      RowBox[{"4", ",", "1", ",", "2", ",", "3"}], "}"}], "]"}]}]}],
  SequenceForm[
  "This gives us the new index configuration: ", -
   xAct`xPerm`Perm[{4, 1, 2, 3}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912894504*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2124146905,ExpressionUUID->"ffe9ca4b-b9b8-458b-a24d-cc2beef48c07"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New ALPHA: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"{", 
     RowBox[{"1", ",", "4"}], "}"}], "}"}]}],
  SequenceForm["New ALPHA: ", {{1, 4}}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912914178*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->768884607,ExpressionUUID->"f880cd5e-0a6c-4dc6-b929-f418fd3dd237"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Checking consistency in set \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"-", 
     RowBox[{"Images", "[", 
      RowBox[{"{", 
       RowBox[{"2", ",", "3", ",", "4", ",", "1"}], "}"}], "]"}]}], "}"}]}],
  SequenceForm[
  "Checking consistency in set ", {-xAct`xPerm`Images[{2, 3, 4, 1}]}],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779129342337`*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1234485850,ExpressionUUID->"2bc2ea7e-4d80-456c-b160-e55926f5f70f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Removing permutations from SGS of S that move slot \"\>", 
   "\[InvisibleSpace]", "4"}],
  SequenceForm["Removing permutations from SGS of S that move slot ", 4],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912952841*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->1164167283,ExpressionUUID->"03bfd093-09bf-48a1-9a3c-9d8016b8327f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New SGS of S: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["New SGS of S: ", 
   xAct`xPerm`StrongGenSet[{}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912971648*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->436017564,ExpressionUUID->"0f588641-4eec-4a13-bc66-6379995b3b41"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Removing permutations from SGS of D that move index \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["Removing permutations from SGS of D that move index ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977912991315*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->932994584,ExpressionUUID->"dd044807-a668-441e-bda1-8733dd9f315a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"New SGS of D: \"\>", "\[InvisibleSpace]", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{"GenSet", "[", "]"}]}], "]"}]}],
  SequenceForm["New SGS of D: ", 
   xAct`xPerm`StrongGenSet[{}, 
    xAct`xPerm`GenSet[]]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977913011422*^9},
 CellLabel->"During evaluation of In[293]:=",
 CellID->2061866541,ExpressionUUID->"6996463d-8d33-4d8c-a32e-2a2ec3d9a26b"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "c", "b"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`c, $CellContext`b],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289779131182747`*^9},
 CellLabel->"Out[293]=",
 CellID->1994826595,ExpressionUUID->"0866f0c0-1356-45ee-8ff6-26aa7f3b2859"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The canonicalization of expressions involving a large number of indices is \
a slow process, mainly when most of those indices are dummies. In those cases \
the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code for ",
 StyleBox["xPerm`",
  FontFamily->"Courier"],
 ", which is interpreted, becomes too slow. A C executable called xperm is \
used to compute the hardest part of the calculations; ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " communicates with it using a ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " connection. Currently this executable exists for Linux and Mac for any \
version of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", and for MS Windows starting ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " 6.0."
}], "Text",ExpressionUUID->"a4507fde-768f-477b-9982-eb13d4221dcf"],

Cell[CellGroupData[{

Cell[TextData[{
 "In order to be able to connect to the external executable, the following \
global variable must return ",
 StyleBox["True",
  FontFamily->"Courier"],
 ". If you get ",
 StyleBox["False",
  FontFamily->"Courier"],
 ", contact JMM to see what is the problem, and whether it can be solved."
}], "MathCaption",ExpressionUUID->"0372fef3-e7c8-4c7e-90fa-c9885d121816"],

Cell[CellGroupData[{

Cell[BoxData["$xpermQ"], "Input",
 CellLabel->
  "In[294]:=",ExpressionUUID->"148749d1-a620-47ee-9076-03e9bf70f5ab"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.72897791327989*^9},
 CellLabel->"Out[294]=",
 CellID->535885201,ExpressionUUID->"aeda6b3e-e1c3-4c8b-84c3-6b13612b6536"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["MathLink",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MathLink.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Option of ", "Text"],
 StyleBox["CanonicalPerm", "Text",
  FontFamily->"Courier"],
 StyleBox[" to use the external C executable", "Text"],
 StyleBox[" xperm", "DisplayMath"]
}], "DefinitionBox",ExpressionUUID->"27cb64b0-2b65-4816-aab4-cd60a38eaa6b"],

Cell["MathLink connection.", "Text",ExpressionUUID->"8e34f80c-c37d-4984-bf4d-7e8f11abeef1"],

Cell["\<\
Let us show an example of the difference in timings between the interpreted \
Mathematica code and the external executable.\
\>", "Text",
 CellChangeTimes->{{3.473764934580714*^9, 
  3.473764964296207*^9}},ExpressionUUID->"71fe7dc1-1a37-43e7-b943-\
64f39767fca6"],

Cell[CellGroupData[{

Cell[TextData[{
 "We define an antisymmetric tensor ",
 StyleBox["Anti",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"ed36a11c-e4a3-4cec-8330-d25081aa7142"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Anti", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"a", ",", "b"}], "}"}], "]"}], ",", 
   RowBox[{"PrintAs", "->", "\"\<A\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[295]:=",ExpressionUUID->"fd3869cf-952b-4a2b-a3df-9716ebaffe6b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Anti", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Anti[$CellContext`a, $CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897791340944*^9},
 CellLabel->"During evaluation of In[295]:=",
 CellID->76157689,ExpressionUUID->"71b9e845-0598-4eae-a23e-61516b68ba6b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now we fake that there is a metric on ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 ". We shall later explain how to define a metric properly."
}], "MathCaption",ExpressionUUID->"ea8a247c-ee40-4523-868a-e0c9ffa61f66"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"MetricsOfVBundle", "[", "TangentM3", "]"}], "^=", 
   RowBox[{"{", "metricg", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"SymmetryGroupOfTensor", "[", "metricg", "]"}], "^=", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], ";"}]}], "Input",
 CellLabel->
  "In[296]:=",ExpressionUUID->"3e783a9e-1f33-4b90-a49f-3fd68b7e0e05"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
By default the external executable is used, if the link was possible,\
\>", "MathCaption",
 CellChangeTimes->{{3.4737647946772738`*^9, 
  3.473764820520563*^9}},ExpressionUUID->"55f21979-4e3f-4aa9-bff7-\
757a6061ffc0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Options", "[", "CanonicalPerm", "]"}]], "Input",
 CellChangeTimes->{{3.4737648311459846`*^9, 3.47376483743442*^9}},
 CellLabel->
  "In[298]:=",ExpressionUUID->"688a4ec1-f134-4f88-9d92-d06a0f0e01b4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MathLink", "\[Rule]", "False"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"xPermVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"OrderedBase", "\[Rule]", "True"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977913629044*^9},
 CellLabel->"Out[298]=",
 CellID->82025411,ExpressionUUID->"83d38c86-5951-43e8-b7ed-7a93da3171e0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Let us temporarily use the internal interpreted ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code to see the difference in timings :"
}], "MathCaption",
 CellChangeTimes->{{3.4737648465425797`*^9, 3.47376487263594*^9}, {
  3.473764915406534*^9, 
  3.4737649175880013`*^9}},ExpressionUUID->"e4869e3c-055e-492e-a642-\
4b1246131c7e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"CanonicalPerm", ",", 
   RowBox[{"MathLink", "->", "False"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.473764878869479*^9, 3.4737648869508247`*^9}},
 CellLabel->
  "In[299]:=",ExpressionUUID->"55d129b9-cdb5-4785-abc2-6a627d133ccd"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MathLink", "\[Rule]", "False"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"xPermVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"OrderedBase", "\[Rule]", "True"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977913795958*^9},
 CellLabel->"Out[299]=",
 CellID->2064064950,ExpressionUUID->"87cc3c79-d14a-4073-ba68-17ecdea9c79f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Powers of an odd number of antisymmetric tensors are 0 :", "MathCaption",
 CellChangeTimes->{{3.4737649026543837`*^9, 
  3.47376490443298*^9}},ExpressionUUID->"78e69ffc-f2d4-40df-82b9-\
36f3f9d5e196"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Options", "[", "ToCanonical", "]"}]], "Input",
 CellLabel->
  "In[300]:=",ExpressionUUID->"20567bd8-92e5-4972-9584-5a2e5080ee48"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Verbose", "\[Rule]", "False"}], ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "All"}], ",", 
   RowBox[{"Method", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"ChangeCovD", ",", 
      RowBox[{"\<\"ExpandChristoffel\"\>", "\[Rule]", "False"}]}], "}"}]}], 
   ",", 
   RowBox[{"MathLink", "\[RuleDelayed]", "$xpermQ"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977913936892*^9},
 CellLabel->"Out[300]=",
 CellID->288781012,ExpressionUUID->"11e34d29-76e2-4e24-823f-7d1a9987ee9e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"Anti", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "a"}]}], "]"}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[301]:=",ExpressionUUID->"3d4c6d62-82a5-47c7-a497-e7643dede8bd"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 301, 35, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977914063034*^9},
 CellLabel->"During evaluation of In[301]:=",
 CellID->229426464,ExpressionUUID->"7da13220-0e9d-4ad5-a563-87a9a3163466"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.003973000000000226`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.003973000000000226, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779140917683`*^9},
 CellLabel->"During evaluation of In[301]:=",
 CellID->1237434083,ExpressionUUID->"a9970acb-21da-4a3e-8e79-e671d975a424"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 301, 36, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977914102929*^9},
 CellLabel->"During evaluation of In[301]:=",
 CellID->1322995433,ExpressionUUID->"f3bece03-70d4-4485-8d74-7771c2ad94c6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.0053520000000002454`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm[
  "Dummy algorithm applied in ", 0.0053520000000002454`, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779141260777`*^9},
 CellLabel->"During evaluation of In[301]:=",
 CellID->538561393,ExpressionUUID->"0321ceef-c804-4e3d-9675-49ab579b0f62"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.010406`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977914138227*^9},
 CellLabel->"Out[301]=",
 CellID->2105243693,ExpressionUUID->"d197824e-22d1-4a9d-a7be-250e1d82ee59"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Anti", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "a"}]}], "]"}]}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[302]:=",ExpressionUUID->"72fe68e8-2b7e-41fa-a107-cfc5ffa0cce8"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 302, 37, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977914276167*^9},
 CellLabel->"During evaluation of In[302]:=",
 CellID->890180279,ExpressionUUID->"7085419d-961f-4152-94be-d976e7c96c83"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.005964999999999776`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.005964999999999776, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977914335657*^9},
 CellLabel->"During evaluation of In[302]:=",
 CellID->1888864151,ExpressionUUID->"a20e79d2-b94f-4b22-9af4-a9892eb86305"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 302, 38, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977914346916*^9},
 CellLabel->"During evaluation of In[302]:=",
 CellID->2036515903,ExpressionUUID->"3098ebff-1ec6-4a5a-97e7-cd95107e26eb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.01528199999999913`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.01528199999999913, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977914395899*^9},
 CellLabel->"During evaluation of In[302]:=",
 CellID->1128415835,ExpressionUUID->"36f3dee7-9205-45ea-8b5b-6eda1f150274"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.024889`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Anti[-$CellContext`a, -$CellContext`b],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289779144730463`*^9},
 CellLabel->"Out[302]=",
 CellID->856852188,ExpressionUUID->"d97fd5b1-c399-493d-8bbc-b5312f46c9f4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Anti", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"c", ",", 
       RowBox[{"-", "a"}]}], "]"}]}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[303]:=",ExpressionUUID->"27a0dff3-517b-47cb-bdde-53cae0f29f6f"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 303, 39, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977914610838*^9},
 CellLabel->"During evaluation of In[303]:=",
 CellID->89386934,ExpressionUUID->"40717f26-7dc5-4f93-8321-fad7efb1c7d6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.0073659999999993175`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.0073659999999993175`, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779146415854`*^9},
 CellLabel->"During evaluation of In[303]:=",
 CellID->1235130485,ExpressionUUID->"f911e7de-0418-4ad4-8d1d-5ca6da6b356d"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 303, 40, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977914653624*^9},
 CellLabel->"During evaluation of In[303]:=",
 CellID->1261129493,ExpressionUUID->"bf492051-a3a3-4035-b17d-60047159d926"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.031121999999999872`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.031121999999999872`, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977914676134*^9},
 CellLabel->"During evaluation of In[303]:=",
 CellID->139792961,ExpressionUUID->"5f2434c5-bc90-49f7-a950-357958d744f7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.044653`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977914687769*^9},
 CellLabel->"Out[303]=",
 CellID->153185709,ExpressionUUID->"bf7b4454-96da-453a-875e-29bafda01b0b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Anti", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"c", ",", "d"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "a"}]}], "]"}]}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[304]:=",ExpressionUUID->"bc2e1e6d-c305-4307-84cf-e27cebd21a76"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 304, 41, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289779148272257`*^9},
 CellLabel->"During evaluation of In[304]:=",
 CellID->1906802057,ExpressionUUID->"8a910d4e-0ed3-4bfb-9d9b-f6efc8aa1e91"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.007920999999999623`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.007920999999999623, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779148557796`*^9},
 CellLabel->"During evaluation of In[304]:=",
 CellID->723707818,ExpressionUUID->"38ebff8b-96c0-42cf-9b63-c86e24051a54"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 304, 42, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.72897791486759*^9},
 CellLabel->"During evaluation of In[304]:=",
 CellID->9920518,ExpressionUUID->"39248624-227e-437b-b75c-e51be1d7390e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.0660469999999993`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.0660469999999993, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779148932533`*^9},
 CellLabel->"During evaluation of In[304]:=",
 CellID->1377905207,ExpressionUUID->"56bbc755-8eac-4e1c-bff4-b4625081566a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.083074`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "c"},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`a, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`c, -$CellContext`d],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977914973757*^9},
 CellLabel->"Out[304]=",
 CellID->2136648448,ExpressionUUID->"6fb881b7-7708-4341-9a01-d45606b5cdd0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Anti", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"c", ",", "d"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"e", ",", 
       RowBox[{"-", "a"}]}], "]"}]}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[305]:=",ExpressionUUID->"62944ed6-4977-42c2-927e-5ecaf2ca13aa"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 305, 43, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977915127383*^9},
 CellLabel->"During evaluation of In[305]:=",
 CellID->428920402,ExpressionUUID->"88ee9675-3407-41ac-bf10-6082be9544a2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.008774999999999977`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.008774999999999977, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977915156088*^9},
 CellLabel->"During evaluation of In[305]:=",
 CellID->1910077082,ExpressionUUID->"38b9bd6f-b9ef-435c-a17c-d9e7a70d3203"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 305, 44, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977915167234*^9},
 CellLabel->"During evaluation of In[305]:=",
 CellID->2079961818,ExpressionUUID->"f7239f83-2376-427e-b83c-58ee9d650581"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.10780200000000129`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.10780200000000129`, " secs."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779152555647`*^9},
 CellLabel->"During evaluation of In[305]:=",
 CellID->1179469260,ExpressionUUID->"3ba135a4-5c02-4273-bf50-51386a64c3a9"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.128523`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977915269899*^9},
 CellLabel->"Out[305]=",
 CellID->2065791010,ExpressionUUID->"f1e324b4-7e44-42c3-a4b9-260c9e7d155c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Anti", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"c", ",", "d"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"e", ",", "f"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "f"}], ",", 
       RowBox[{"-", "a"}]}], "]"}]}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[306]:=",ExpressionUUID->"c90256d8-4702-47d3-9679-451d7e5f3452"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 306, 45, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977915476822*^9},
 CellLabel->"During evaluation of In[306]:=",
 CellID->1381384021,ExpressionUUID->"23bb41c2-651f-4849-9388-f494abfe84d6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.011490999999999474`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.011490999999999474`, " secs."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977915504113*^9},
 CellLabel->"During evaluation of In[306]:=",
 CellID->475533230,ExpressionUUID->"6dcb4702-2af2-4b63-b666-845486a36db1"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 306, 46, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977915515664*^9},
 CellLabel->"During evaluation of In[306]:=",
 CellID->1500346123,ExpressionUUID->"d86318cb-1101-4c9a-bc49-c538d48ac6e0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.2090720000000008`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.2090720000000008, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977915690283*^9},
 CellLabel->"During evaluation of In[306]:=",
 CellID->564306187,ExpressionUUID->"c9b57d7d-6167-4214-b208-4a526244f897"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.235312`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Anti[-$CellContext`a, $CellContext`c],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "e"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`c, $CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "f"},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`d, $CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`e, -$CellContext`f],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977915772709*^9},
 CellLabel->"Out[306]=",
 CellID->1006299757,ExpressionUUID->"c38f0545-96d8-483a-ba7d-719d01d89626"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Anti", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"c", ",", "d"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"e", ",", "f"}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{
       RowBox[{"-", "f"}], ",", 
       RowBox[{"-", "g"}]}], "]"}], 
     RowBox[{"Anti", "[", 
      RowBox[{"g", ",", 
       RowBox[{"-", "a"}]}], "]"}]}], ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[307]:=",ExpressionUUID->"e3293a73-d6ff-4aef-86fa-7244bd2ba255"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 307, 47, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977915977116*^9},
 CellLabel->"During evaluation of In[307]:=",
 CellID->1446295774,ExpressionUUID->"8a712f1f-8243-47bd-b3ed-b8bc58cce75d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.011877999999999389`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.011877999999999389`, " secs."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977916005131*^9},
 CellLabel->"During evaluation of In[307]:=",
 CellID->525673225,ExpressionUUID->"9a37b2f3-af3d-4143-ba06-487471f6a300"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 307, 48, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977916017119*^9},
 CellLabel->"During evaluation of In[307]:=",
 CellID->242168817,ExpressionUUID->"57480c3a-bf20-48e5-8f56-93943f10ab9f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.314331000000001`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.314331000000001, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977916290555*^9},
 CellLabel->"During evaluation of In[307]:=",
 CellID->649796254,ExpressionUUID->"3bea37a0-ae26-4279-a172-5af0097721ef"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.344869`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977916301962*^9},
 CellLabel->"Out[307]=",
 CellID->1171295983,ExpressionUUID->"adebffc6-9e2b-449a-b1fc-9222426d370e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are the corresponding timings with the C executable:", \
"MathCaption",ExpressionUUID->"bb635e68-c167-4451-ae8f-ce5009812c1d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"CanonicalPerm", ",", 
   RowBox[{"MathLink", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[308]:=",ExpressionUUID->"649c518f-45c4-459d-86ba-899131779675"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"MathLink", "\[Rule]", "True"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"xPermVerbose", "\[Rule]", "False"}], ",", 
   RowBox[{"OrderedBase", "\[Rule]", "True"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977916463723*^9},
 CellLabel->"Out[308]=",
 CellID->2136477879,ExpressionUUID->"2177cfae-2db9-48c0-8eca-7e6ebfa1f6ee"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"Anti", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"c", ",", "d"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"e", ",", 
      RowBox[{"-", "a"}]}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[309]:=",ExpressionUUID->"7ec7d309-e127-4be5-a47a-8236224af3f2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.006754`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977916610849*^9},
 CellLabel->"Out[309]=",
 CellID->1857134513,ExpressionUUID->"34e0ea7a-90ef-4ef0-8728-1c0674f189f1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"Anti", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"c", ",", "d"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"e", ",", "f"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "f"}], ",", 
      RowBox[{"-", "a"}]}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[310]:=",ExpressionUUID->"d909efeb-86ba-48c1-881d-7033a9cd54d8"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.007632`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Anti[-$CellContext`a, $CellContext`c],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "e"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`c, $CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "f"},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`d, $CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`e, -$CellContext`f],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977916788908*^9},
 CellLabel->"Out[310]=",
 CellID->1332434493,ExpressionUUID->"f30b2d64-a861-4898-8c3d-45032e7fe687"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"Anti", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"c", ",", "d"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"e", ",", "f"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "f"}], ",", 
      RowBox[{"-", "g"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"g", ",", 
      RowBox[{"-", "a"}]}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[311]:=",ExpressionUUID->"77099857-70ce-4350-8820-30c9a9b83aca"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.008334`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7289779169632463`*^9},
 CellLabel->"Out[311]=",
 CellID->267052935,ExpressionUUID->"aa210262-173d-4ed7-9fb7-feb8cdbc586c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"Anti", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"c", ",", "d"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"e", ",", "f"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "f"}], ",", 
      RowBox[{"-", "g"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"g", ",", "h"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "h"}], ",", 
      RowBox[{"-", "a"}]}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[312]:=",ExpressionUUID->"f0230162-5e6a-46fa-9503-bcdc742d135e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.009496`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "c"},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`a, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "e"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`c, $CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "f"},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`d, $CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "g"},
            {"e", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`e, $CellContext`g],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "h"},
            {"f", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`f, $CellContext`h],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"g", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`g, -$CellContext`h],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289779171122713`*^9},
 CellLabel->"Out[312]=",
 CellID->117856315,ExpressionUUID->"5ccc7ea2-b619-4e47-a540-5d51f1e784b9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"Anti", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"c", ",", "d"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"e", ",", "f"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "f"}], ",", 
      RowBox[{"-", "g"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"g", ",", "h"}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{
      RowBox[{"-", "h"}], ",", 
      RowBox[{"-", "i"}]}], "]"}], 
    RowBox[{"Anti", "[", 
     RowBox[{"i", ",", 
      RowBox[{"-", "a"}]}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[313]:=",ExpressionUUID->"3395f34e-bedc-453c-a4b7-a6a08b9794e4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.010742`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977917278748*^9},
 CellLabel->"Out[313]=",
 CellID->831163671,ExpressionUUID->"fadb01c1-e854-47be-bf58-cdcc1f028e8e"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.2. Technical details", "Subsection",ExpressionUUID->"3a8486f2-1233-4b73-8f71-ca4c9d665f6e"],

Cell[TextData[{
 StyleBox["Canonicalization:",
  FontWeight->"Bold"],
 "\nThe canonicalization process has essentially six steps, of which only \
steps 4 and 5 are really hard. Assume we start from a generic syntactically \
correct input in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " on which we act with ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 ":\n\t1) Expand the expression to convert it into a sum (",
 StyleBox["Plus",
  FontFamily->"Courier"],
 ") of terms, each being a product (",
 StyleBox["Times",
  FontFamily->"Courier"],
 ") of indexed objects, sharing dummy indices. This can be considered a \
polynomial in tensor variables, and as with any normal polynomial, it cannot \
be canonicalized without first expanding it. We use the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " command ",
 StyleBox["Expand",
  FontFamily->"Courier"],
 ". It might happen that during the expansion some parts of the expression \
are evaluated and give new products to expand. That means that ",
 StyleBox["Expand",
  FontFamily->"Courier"],
 " must be used repeatedly until the expression does no longer change (we use \
the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " command ",
 StyleBox["FixedPoint",
  FontFamily->"Courier"],
 " for that).\n\t2) Minimize the number of different dummies in the whole \
expression. We use the ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " command ",
 StyleBox["SameDummies",
  FontFamily->"Courier"],
 ".\n\t3) ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 " is threaded on each term. Steps 4 and 5 cannot compare different terms in \
the expression. From now on we only deal with products of objects.\n\t4) The \
objects in each term must be sorted according to certain priorities depending \
on the types and properties of the objects, but not on the actual indices \
they have. The term is then considered to be a single tensor, whose \
properties can be deduced from the properties of the elementary objects \
composing it. This idea is taken from R. Portugal (put journal ref. here).\n\t\
5) We extract the list of indices of that composite tensor and canonicalize \
that list according to the symmetries of the composite object and comparing \
with the \"ideal\" ordering of those indices (given in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " by the function ",
 StyleBox["IndexSort",
  FontFamily->"Courier"],
 "). This \"ideal\" ordering is also decided by a number of priorities \
depending on the properties of the indices (type, character and state), as \
explained below. In general, the indices can change slot, but there is no \
creation of new indices, the exception being the treatment of derivatives \
which generate new terms. A very important issue at this step is the \
interaction between metrics and derivatives, discussed below. The problem \
will be even harder if frames are involved. The canonicalization algorithms \
for tensors are basically those of R. Portugal et al (put journal refs here).\
\n\t6) By now each term is already canonicalized. The function ",
 StyleBox["Plus",
  FontFamily->"Courier"],
 " will add up equal terms and the result is returned, in canonical form. No \
simplification (e.g. factorization) is attempted by ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 ".\nThe priorities for sorting objects in step 4 are fixed, even though they \
could be easily changed if a user needs it, but the priorities for sorting \
indices in step 5 can be configured by the user. In this section we shall \
describe what those priorities are and how some of them affect the efficiency \
of the whole process.\n"
}], "Text",ExpressionUUID->"b45e0955-72d2-4720-8b66-69f514156352"],

Cell[TextData[{
 StyleBox["Sorting objects:",
  FontWeight->"Bold"],
 "\nObjects are sorted according to seven considerations, most based in \
lexicographic ordering (this is also what ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " does). Objects are sorted with respect to the first of these criteria \
applying:\n\t1) Type: use alphabetical order in the types:\n\t\tConstant\n\t\t\
ConstantSymbol\n\t\tParameter\n\t\tScalar (see subsection 8.7 for a \
description of this head)\n\t\tTensor\n\t     In subexpressions sometimes we \
need the type of a composite object: the type of a product is the product of \
types, and the type of a sum is the sum of types. Both Times and Plus sort \
their arguments lexicographically.\n\t2) Order (in the sense of number of \
elements) of the group of symmetry of the objects. More symmetric (higher \
order) objects come last (to improve efficiency for tensors with lots of \
indices), though I have found that the opposite choice is aesthetically \
better in general.\n\t3) Names of objects. Every object is given a name which \
depends on the names and the structure of its components, but not on its \
indices.\n\t4) Number of indices of objects. Sort first the objects with less \
indices. This is also more efficient.\n\t5) Number of free indices (of the \
whole term) of objects. Sort first the objects with more free indices.\n\t6) \
If the global variable ",
 StyleBox["$CommuteFreeIndices",
  FontFamily->"Courier"],
 " is ",
 StyleBox["False",
  FontFamily->"Courier"],
 " then sort first the object with the least free index according to ",
 StyleBox["IndexSort",
  FontFamily->"Courier"],
 ". If it is ",
 StyleBox["True",
  FontFamily->"Courier"],
 " (the default) then do not take into account the actual free indices and \
jump to step 6.\n\t7) If several objects cannot be sorted according to the \
previous considerations, then they are considered equivalent (I call them \
\"commuting\") and left then marked as CommutingObjects[o1, o2, ...], what \
will be used by the permutations algorithms to add new symmetries among the \
indices of those equivalent tensors.\n"
}], "Text",ExpressionUUID->"656ed42e-59f5-4270-9e76-248059fcbae8"],

Cell[TextData[{
 StyleBox["Sorting indices:",
  FontWeight->"Bold"],
 "\nThese are all possible things we can consider when sorting indices, in no \
special order.\n\t- Relative position of free and dummy indices.\n\t- \
Relative position of the members of the same dummy pair.\n\t- Relative \
position of up and down indices.\n\t- Alphabetic order of indices.\n\t- The \
order that the indices have in ",
 StyleBox["IndicesOfManifold",
  FontFamily->"Courier"],
 " (the order at definition time).\nThe default is the following, chosen \
\"experimentally\" as the most efficient, in average (I don't really have an \
explanation of why this seems to be more efficient than other choices):\n\t1) \
free indices come before dummy indices.\n\t2) sort the indices according to \
lexicographic order.\n\t3) up-indices come before down-indices.\nThose \
priorities can be changed. We need three priorities, that must be chosen from \
four pairs of strings, not choosing two of the same pair: ",
 StyleBox["\"free\"",
  FontFamily->"Courier"],
 " / ",
 StyleBox["\"dummy\"",
  FontFamily->"Courier"],
 ", ",
 StyleBox["\"up\"",
  FontFamily->"Courier"],
 " / ",
 StyleBox["\"down\"",
  FontFamily->"Courier"],
 ", ",
 StyleBox["\"lexicographic\"",
  FontFamily->"Courier"],
 " / ",
 StyleBox["\"antilexicographic\"",
  FontFamily->"Courier"],
 " and ",
 StyleBox["\"positional\"",
  FontFamily->"Courier"],
 " / ",
 StyleBox["\"antipositional\"",
  FontFamily->"Courier"],
 ". "
}], "Text",ExpressionUUID->"e355b849-e774-490e-bf88-f6b718337e2c"],

Cell[TextData[{
 StyleBox[ButtonBox["IndexSort",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexSort.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Sort a list of indices\n", "Text"],
 StyleBox[ButtonBox["SetIndexSortPriorities",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SetIndexSortPriorities.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t", "DisplayMath"],
 StyleBox["Define priorities for index sorting", "Text"]
}], "DefinitionBox",ExpressionUUID->"e59d4de5-3442-44e3-bbbe-abcf86b5212a"],

Cell[CellGroupData[{

Cell["Functions that sort indices.", "Text",ExpressionUUID->"0a1890cf-def1-478c-bf42-a12395efe46e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "IndexSort"}]], "Input",
 CellLabel->
  "In[314]:=",ExpressionUUID->"8e3f8703-33bc-44fe-8e15-7e30b7adefd0"],

Cell[BoxData["\<\"IndexSort[list] sorts the elements of list (assumed to be \
g-indices) according to three priorities set up by SetIndexSortPriorities. \
Currently: first: free; second: lexicographic; third: up.\"\>"], "Print", \
"PrintUsage",
 CellChangeTimes->{3.728977946132901*^9},
 CellTags->"Info3143728981546-9181623",
 CellID->799431715,ExpressionUUID->"b286d54a-110c-4101-a947-c1930e20065e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Indices of manifold ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"5188daf7-454e-438e-beb3-a072c3832375"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[315]:=",ExpressionUUID->"985532c9-3af5-4014-9c1e-6e3ac0238427"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", "h",
      ",", "i", ",", "j"}], "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977946234346*^9},
 CellLabel->"Out[315]=",
 CellID->1391138373,ExpressionUUID->"afb82529-fe52-4833-acc1-225f7b962461"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Study this case, taking into account the default priorities given above:\
\>", "MathCaption",ExpressionUUID->"0e9c8122-11ee-4368-aca5-314b0f9764d4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSort", "[", 
  RowBox[{"{", 
   RowBox[{"a", ",", "b", ",", "i", ",", "j", ",", 
    RowBox[{"-", "e"}], ",", "d", ",", "c", ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "j"}], ",", "e", ",", "f"}], "}"}], "]"}]], "Input",
 CellLabel->
  "In[316]:=",ExpressionUUID->"28af3cd7-98ec-418d-9c0b-5e1d2d2dbb16"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
    "a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", "d", ",", 
     "\[ThinSpace]", "f", ",", "\[ThinSpace]", "i", ",", "\[ThinSpace]", "c", 
     ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", "e", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}], ",", "\[ThinSpace]", "j", ",", "\[ThinSpace]", 
     RowBox[{"-", "j"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, $CellContext`d, \
$CellContext`f, $CellContext`i, $CellContext`c, -$CellContext`c, \
$CellContext`e, -$CellContext`e, $CellContext`j, -$CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977946329524*^9},
 CellLabel->"Out[316]=",
 CellID->1208170122,ExpressionUUID->"4fe2e442-3337-4744-b763-5d7f31fff981"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now we change priorities. The list is sorted differently", "MathCaption",ExpressionUUID->"80277130-8651-4d89-9c61-0169b67139c9"],

Cell[BoxData[
 RowBox[{"SetIndexSortPriorities", "[", 
  RowBox[{
  "\"\<down\>\"", ",", "\"\<free\>\"", ",", "\"\<antilexicographic\>\""}], 
  "]"}]], "Input",
 CellLabel->
  "In[317]:=",ExpressionUUID->"f64c2818-f3b5-4001-8795-1e453885c655"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "IndexSort"}]], "Input",
 CellLabel->
  "In[318]:=",ExpressionUUID->"9983e5d6-1ee7-490a-b9d9-ff18cb7b2c56"],

Cell[BoxData["\<\"IndexSort[list] sorts the elements of list (assumed to be \
g-indices) according to three priorities set up by SetIndexSortPriorities. \
Currently: first: down; second: free; third: antilexicographic.\"\>"], "Print",\
 "PrintUsage",
 CellChangeTimes->{3.728977946524436*^9},
 CellTags->"Info3183728981546-9181623",
 CellID->1257358069,ExpressionUUID->"bfc1f777-98e2-405b-8b4a-375be9c9d56c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSort", "[", 
  RowBox[{"{", 
   RowBox[{"a", ",", "b", ",", "i", ",", "j", ",", 
    RowBox[{"-", "e"}], ",", "d", ",", "c", ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "j"}], ",", "e", ",", "f"}], "}"}], "]"}]], "Input",
 CellLabel->
  "In[319]:=",ExpressionUUID->"987e8edb-93c1-4239-86af-d40316e247f6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "j"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", "i", ",", "\[ThinSpace]", "f", 
     ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     "a", ",", "\[ThinSpace]", "j", ",", "\[ThinSpace]", "e", ",", 
     "\[ThinSpace]", "c"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`j, -$CellContext`e, -$CellContext`c, \
$CellContext`i, $CellContext`f, $CellContext`d, $CellContext`b, \
$CellContext`a, $CellContext`j, $CellContext`e, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289779466361017`*^9},
 CellLabel->"Out[319]=",
 CellID->1177979203,ExpressionUUID->"28fd081e-9b2c-49b7-8005-d85ee731cd73"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Note that once ",
 StyleBox["\"positional\"",
  FontFamily->"Courier"],
 " or ",
 StyleBox["\"lexicographic\"",
  FontFamily->"Courier"],
 " is chosen (or their opposite), then the priority ",
 StyleBox["\"free\"",
  FontFamily->"Courier"],
 "/",
 StyleBox["\"dummy\"",
  FontFamily->"Courier"],
 " becomes irrelevant:"
}], "MathCaption",ExpressionUUID->"cee1fcb7-94c3-43ef-8bd9-85d4691ea749"],

Cell[BoxData[
 RowBox[{"SetIndexSortPriorities", "[", 
  RowBox[{"\"\<positional\>\"", ",", "\"\<down\>\"", ",", "\"\<dummy\>\""}], 
  "]"}]], "Input",
 CellLabel->
  "In[320]:=",ExpressionUUID->"76a2a6da-9d40-4b88-8662-e74e1cd952fd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSort", "[", 
  RowBox[{"{", 
   RowBox[{"a", ",", "b", ",", "i", ",", "j", ",", 
    RowBox[{"-", "e"}], ",", "d", ",", "c", ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "j"}], ",", "e", ",", "f"}], "}"}], "]"}]], "Input",
 CellLabel->
  "In[321]:=",ExpressionUUID->"70008b35-5ee4-4f74-a3f7-21bb5273f958"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", "c", ",", "\[ThinSpace]", "d", 
     ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}], ",", "\[ThinSpace]", "e", ",", "\[ThinSpace]", "f", 
     ",", "\[ThinSpace]", "i", ",", "\[ThinSpace]", 
     RowBox[{"-", "j"}], ",", "\[ThinSpace]", "j"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, -$CellContext`c, \
$CellContext`c, $CellContext`d, -$CellContext`e, $CellContext`e, \
$CellContext`f, $CellContext`i, -$CellContext`j, $CellContext`j],
  Editable->False]], "Output",
 CellChangeTimes->{3.728977946774921*^9},
 CellLabel->"Out[321]=",
 CellID->644723128,ExpressionUUID->"15af1cda-c5a7-4964-a931-270805949b8a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The choice of priorities can make a huge difference in the efficiency of the \
dummy-canonicalization process. In principle, it is not possible to predict \
which set of priorities will be better in a given situation, but \
experimentally we have found that trying to match pairs of dummies with \
symmetry pairs of S gives better results:\
\>", "Text",ExpressionUUID->"84e02928-6c04-4c4d-8e97-2fed344ab395"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{"Anti", "[", 
    RowBox[{"a", ",", "b"}], "]"}], 
   RowBox[{"Anti", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], 
   RowBox[{"Anti", "[", 
    RowBox[{"c", ",", "d"}], "]"}], 
   RowBox[{"Anti", "[", 
    RowBox[{
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], 
   RowBox[{"Anti", "[", 
    RowBox[{"e", ",", "f"}], "]"}], 
   RowBox[{"Anti", "[", 
    RowBox[{
     RowBox[{"-", "f"}], ",", 
     RowBox[{"-", "a"}]}], "]"}]}]}]], "Input",
 CellLabel->
  "In[322]:=",ExpressionUUID->"418e6062-1975-413d-8064-6ec9003d3f33"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[-$CellContext`b, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"c", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {" ", " "},
          {"d", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[-$CellContext`d, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"e", "f"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`e, $CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {" ", " "},
          {"f", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[-$CellContext`f, -$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977946895467*^9},
 CellLabel->"Out[322]=",
 CellID->1806613181,ExpressionUUID->"0614c7cb-56dd-4451-8bc2-8ca2bef911a0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Having one of ",
 StyleBox["\"up\"",
  FontFamily->"Courier"],
 "/",
 StyleBox["\"down\"",
  FontFamily->"Courier"],
 " before ",
 StyleBox["\"free\"",
  FontFamily->"Courier"],
 "/",
 StyleBox["\"up\"",
  FontFamily->"Courier"],
 " seems terrible in the Dummy algorithm:"
}], "MathCaption",ExpressionUUID->"8a4bbb3d-20c8-48c2-b074-8e7dc059a85c"],

Cell[BoxData[
 RowBox[{"SetIndexSortPriorities", "[", 
  RowBox[{"\"\<up\>\"", ",", "\"\<free\>\"", ",", "\"\<lexicographic\>\""}], 
  "]"}]], "Input",
 CellLabel->
  "In[323]:=",ExpressionUUID->"dc15dac0-4473-4a73-8bdc-93d65ffe3e3c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{"expr", ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}], ",", 
    RowBox[{"MathLink", "\[Rule]", "False"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[324]:=",ExpressionUUID->"3bc09fe8-ac08-4acd-9363-cfdf672f9ec0"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 324, 49, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977947044683*^9},
 CellLabel->"During evaluation of In[324]:=",
 CellID->476448134,ExpressionUUID->"bbfac71d-e079-4011-9533-6ced83690e66"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.009828000000000614`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.009828000000000614, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977947078545*^9},
 CellLabel->"During evaluation of In[324]:=",
 CellID->1844752122,ExpressionUUID->"130fc8fa-eb87-43aa-80f8-b5a8f241e38b"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 324, 50, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977947093156*^9},
 CellLabel->"During evaluation of In[324]:=",
 CellID->1527502715,ExpressionUUID->"32507cba-cedf-4833-a991-fb6ae135efb8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.7761649999999989`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.7761649999999989, " secs."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289779479139547`*^9},
 CellLabel->"During evaluation of In[324]:=",
 CellID->1718614767,ExpressionUUID->"defa3453-2797-4719-88dd-4622dc278a90"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.799659`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Anti[-$CellContext`a, -$CellContext`c],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"c", "d"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`c, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`d, -$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"e", "f"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`e, $CellContext`f],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977948172535*^9},
 CellLabel->"Out[324]=",
 CellID->1103336819,ExpressionUUID->"883d24fd-0908-480a-9271-af30ba786f47"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the default choice, instead. Note that the final result is different. \
This is not a problem: what we want is a uniquely defined process of \
canonicalization, but which one we choose is \"only\" a matter of efficiency \
and aesthetics.\
\>", "MathCaption",ExpressionUUID->"f89b7f57-0db5-415a-a00a-cd93065f6d01"],

Cell[BoxData[
 RowBox[{"SetIndexSortPriorities", "[", 
  RowBox[{"\"\<free\>\"", ",", "\"\<lexicographic\>\"", ",", "\"\<up\>\""}], 
  "]"}]], "Input",
 CellLabel->
  "In[325]:=",ExpressionUUID->"e6ec04c6-ffb5-4a41-b9fb-4b84a56bd6b7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"ToCanonical", "[", 
   RowBox[{"expr", ",", 
    RowBox[{"TimeVerbose", "\[Rule]", "True"}], ",", 
    RowBox[{"MathLink", "\[Rule]", "False"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[326]:=",ExpressionUUID->"88025c4c-5c88-42e9-b93f-d560278aa43f"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"RightCosetRepresentative\\\"}]\\).\"", 2, 326, 51, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977948505438*^9},
 CellLabel->"During evaluation of In[326]:=",
 CellID->260625088,ExpressionUUID->"c159de9a-91d6-45e6-9098-01593e54d90b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Free algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.010289999999999466`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Free algorithm applied in ", 0.010289999999999466`, " secs."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977948532962*^9},
 CellLabel->"During evaluation of In[326]:=",
 CellID->1606697783,ExpressionUUID->"6e784f90-9cd1-4739-9480-b251230ff2d5"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \
\\!\\(\\*RowBox[{\\\"\\\\\\\"TimeVerbose\\\\\\\"\\\"}]\\) for \
\\!\\(\\*RowBox[{\\\"DoubleCosetRepresentative\\\"}]\\).\"", 2, 326, 52, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728977948546899*^9},
 CellLabel->"During evaluation of In[326]:=",
 CellID->1310261503,ExpressionUUID->"0eac8c16-d240-4142-8fb1-3be4f40fb96b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Dummy algorithm applied in \"\>", "\[InvisibleSpace]", 
   "0.21134700000000173`", "\[InvisibleSpace]", "\<\" secs.\"\>"}],
  SequenceForm["Dummy algorithm applied in ", 0.21134700000000173`, " secs."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728977948731965*^9},
 CellLabel->"During evaluation of In[326]:=",
 CellID->1267764583,ExpressionUUID->"91adbdd2-ba82-4fb3-a883-e6fc8c0324af"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.237522`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Anti[-$CellContext`a, $CellContext`c],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "e"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`c, $CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "f"},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`d, $CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`e, -$CellContext`f],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977948813918*^9},
 CellLabel->"Out[326]=",
 CellID->2134084594,ExpressionUUID->"83460eac-6841-4d0a-a08c-80b9083e0187"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[327]:=",ExpressionUUID->"41afebc0-4860-4472-b64a-9f47dbab7a79"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["4.3. Benchmarking", "Subsection",ExpressionUUID->"218a8fe1-d10f-406f-b7b2-bf7d5c336cf5"],

Cell[TextData[{
 "This section contains a series of examples performed in a Linux box with a \
1.7 GHz Dual core processor. At least 256 Mb RAM are required to repeat some \
of these examples.\nWe use the external executable ",
 StyleBox["xperm",
  FontSlant->"Italic"],
 ", compiled from the code xperm.c and linked from ",
 StyleBox["xPerm`",
  FontFamily->"Courier"],
 " through the ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " protocol (xperm.tm template). The code xperm.c contains some C99 \
extensions, supported by GNU gcc, and therefore Linux and Mac can always use \
this executable. However, all other Windows C-compilers that ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " can communicate to (Borland, MS Visual 2003, etc) are still C90 complier \
only. The gcc compiler under Windows (under the cygwin system) is compatible \
with ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " only from version 6.0 of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", but not before and therefore pre-6.0 versions of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " cannot use ",
 StyleBox["xperm",
  FontSlant->"Italic"],
 "."
}], "Text",ExpressionUUID->"ce5ef912-807c-49fd-8faf-c335f3310335"],

Cell[CellGroupData[{

Cell["\<\
This variable says whether the connection to the external executable has been \
performed correctly. If you get False, don't try to repeat these examples; \
they will take too long.\
\>", "MathCaption",ExpressionUUID->"4d850949-7562-413c-af18-bfd3c0f00b6d"],

Cell[CellGroupData[{

Cell[BoxData["$xpermQ"], "Input",
 CellLabel->
  "In[328]:=",ExpressionUUID->"d5c06c29-8c1b-4f4b-9b81-9489407ee455"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728977958941482*^9},
 CellLabel->"Out[328]=",
 CellID->517171343,ExpressionUUID->"9d67c1c7-8b4f-48eb-9a5a-86d46bfaaa3a"]
}, Open  ]]
}, Open  ]],

Cell[TextData[StyleBox["Example 1:",
 FontWeight->"Bold"]], "Text",ExpressionUUID->"50af9ae9-f931-41f8-b8a2-4315729dacf8"],

Cell["\<\
Let us come back to the problem of products of antisymmetric tensors.\
\>", "Text",ExpressionUUID->"3c3e1fd3-dada-4a5f-939c-63c71cfe19e0"],

Cell[CellGroupData[{

Cell["\<\
This simple code constructs the products we want to canonicalize:\
\>", "MathCaption",ExpressionUUID->"440cb950-21db-4f09-a20d-881ff9003f3c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"productAnti", "[", "n_", "]"}], ":=", 
  RowBox[{"Times", "@@", 
   RowBox[{"Apply", "[", 
    RowBox[{"Anti", ",", 
     RowBox[{"Partition", "[", 
      RowBox[{
       RowBox[{"RotateLeft", "@", 
        RowBox[{"Flatten", "@", 
         RowBox[{"Transpose", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{"-", "#"}], ",", "#"}], "}"}], "&"}], "@", 
           RowBox[{"GetIndicesOfVBundle", "[", 
            RowBox[{"TangentM3", ",", "n"}], "]"}]}], "]"}]}]}], ",", "2"}], 
      "]"}], ",", 
     RowBox[{"{", "1", "}"}]}], "]"}]}]}]], "Input",
 CellLabel->
  "In[329]:=",ExpressionUUID->"a97db65d-aa6f-4e2c-8596-38103839f2e5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"productAnti", "[", "1", "]"}]], "Input",
 CellLabel->
  "In[330]:=",ExpressionUUID->"81792521-ba50-437a-892a-2af2d7f16f90"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"A", 
      StyleBox[GridBox[{
         {"a", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Anti[$CellContext`a, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289779590927477`*^9},
 CellLabel->"Out[330]=",
 CellID->1232442392,ExpressionUUID->"90966409-7be8-4612-aff7-55d4a98a8717"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"productAnti", "[", "7", "]"}]], "Input",
 CellLabel->
  "In[331]:=",ExpressionUUID->"aa483db2-bc25-47f5-9259-b7794da341a7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"a", " "},
          {" ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"b", " "},
          {" ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`b, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"c", " "},
          {" ", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`c, -$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"d", " "},
          {" ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`d, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"e", " "},
          {" ", "f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`e, -$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"f", " "},
          {" ", "g"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`f, -$CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"A", 
       StyleBox[GridBox[{
          {"g", " "},
          {" ", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Anti[$CellContext`g, -$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977959202917*^9},
 CellLabel->"Out[331]=",
 CellID->1179507029,ExpressionUUID->"65f541c7-88dd-4b90-8b6f-e62822efa2a3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We shall need up to 59 different indices:", "MathCaption",ExpressionUUID->"58278466-d99d-4b97-83ee-1ec983186a32"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GetIndicesOfVBundle", "[", 
  RowBox[{"TangentM3", ",", "59", ",", 
   RowBox[{"{", "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[332]:=",ExpressionUUID->"03049667-d549-42e2-bdda-8f2adad19d79"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", "h", 
   ",", "i", ",", "j", ",", "j1", ",", "j2", ",", "j3", ",", "j4", ",", "j5", 
   ",", "j6", ",", "j7", ",", "j8", ",", "j9", ",", "j10", ",", "j11", ",", 
   "j12", ",", "j13", ",", "j14", ",", "j15", ",", "j16", ",", "j17", ",", 
   "j18", ",", "j19", ",", "j20", ",", "j21", ",", "j22", ",", "j23", ",", 
   "j24", ",", "j25", ",", "j26", ",", "j27", ",", "j28", ",", "j29", ",", 
   "j30", ",", "j31", ",", "j32", ",", "j33", ",", "j34", ",", "j35", ",", 
   "j36", ",", "j37", ",", "j38", ",", "j39", ",", "j40", ",", "j41", ",", 
   "j42", ",", "j43", ",", "j44", ",", "j45", ",", "j46", ",", "j47", ",", 
   "j48", ",", "j49"}], "}"}]], "Output",
 CellChangeTimes->{3.728977959312731*^9},
 CellLabel->"Out[332]=",
 CellID->553090946,ExpressionUUID->"09830161-d42a-4f93-833f-ba2c0d78ee3c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "And this canonicalizes it, giving the ",
 StyleBox["AbsoluteTiming",
  FontFamily->"Courier"],
 " spent. Note that we can get products of up to 25 tensors in less than half \
a second (in that case the dummy group D has more than 5 10^32 elements)."
}], "MathCaption",ExpressionUUID->"7b06b920-242b-408a-bb0a-1e41c5c38091"],

Cell[BoxData[
 RowBox[{
  RowBox[{"canAnti", "[", "n_", "]"}], ":=", 
  RowBox[{"Flatten", "[", 
   RowBox[{"{", 
    RowBox[{"n", ",", 
     RowBox[{"With", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"product", "=", 
         RowBox[{"productAnti", "[", "n", "]"}]}], "}"}], ",", 
       RowBox[{"AbsoluteTiming", "[", 
        RowBox[{"ToCanonical", "[", "product", "]"}], "]"}]}], "]"}]}], "}"}],
    "]"}]}]], "Input",
 CellLabel->
  "In[333]:=",ExpressionUUID->"a57f1837-0fb7-4e26-b145-ae0d180a3d7a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canAnti", "[", "6", "]"}]], "Input",
 CellLabel->
  "In[334]:=",ExpressionUUID->"04931ec2-cf33-471f-b029-68b11f36f523"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"6", ",", "0.008214`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Anti[-$CellContext`a, $CellContext`c],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`b, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "e"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`c, $CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", "f"},
            {"d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`d, $CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Anti[-$CellContext`e, -$CellContext`f],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977959486445*^9},
 CellLabel->"Out[334]=",
 CellID->1490305206,ExpressionUUID->"31470277-d0a6-4d53-a2b8-45d53c5fa269"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canAnti", "[", "7", "]"}]], "Input",
 CellLabel->
  "In[335]:=",ExpressionUUID->"d39e6661-d757-44c9-be98-9bc7db8f7f96"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"7", ",", "0.00847`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.7289779596188307`*^9},
 CellLabel->"Out[335]=",
 CellID->1457822281,ExpressionUUID->"3261b232-0b24-40f5-bbcd-5d7058f4118f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canAnti", "[", "25", "]"}]], "Input",
 CellLabel->
  "In[336]:=",ExpressionUUID->"a0f35cac-b5fd-46a1-805e-266da2fda802"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"25", ",", "0.107618`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977959852089*^9},
 CellLabel->"Out[336]=",
 CellID->978148144,ExpressionUUID->"29950182-f47c-4bb2-9abf-f6ef920590ba"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Let us see how the timings scale with the number of antisymmetric tensors. \
Evaluating this cell takes about a minute in a 1.8GHz Dual core processor. \
The executable xperm.linux requires here up to 30 Mbytes. Each computation is \
done only once, and that invariably introduces small deviations depending on \
which other tasks the computer is busy with.\
\>", "MathCaption",ExpressionUUID->"9e378747-c9a2-4c1a-a41d-9068cb4d1062"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"data", "=", 
  RowBox[{"canAnti", "/@", 
   RowBox[{"Range", "[", 
    RowBox[{"1", ",", "59", ",", "2"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[337]:=",ExpressionUUID->"2ff056ab-3e3d-45c0-8abc-0bf50f7cce6a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0.001215`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3", ",", "0.004188`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"5", ",", "0.005516`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"7", ",", "0.007171`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"9", ",", "0.009748`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"11", ",", "0.011858`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"13", ",", "0.019129`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"15", ",", "0.024334`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"17", ",", "0.034803`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"19", ",", "0.03962`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"21", ",", "0.060073`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"23", ",", "0.084288`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"25", ",", "0.099601`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"27", ",", "0.134321`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"29", ",", "0.178059`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"31", ",", "0.223807`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"33", ",", "0.280071`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"35", ",", "0.353737`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"37", ",", "0.445857`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"39", ",", "0.579452`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"41", ",", "0.723847`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"43", ",", "0.947962`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"45", ",", "1.161231`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"47", ",", "1.423106`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"49", ",", "1.618648`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"51", ",", "2.05579`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"53", ",", "2.366349`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"55", ",", "2.931515`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"57", ",", "3.480804`", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"59", ",", "4.177864`", ",", "0"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.72897798345789*^9},
 CellLabel->"Out[337]=",
 CellID->1077164391,ExpressionUUID->"36fe7410-2339-4c82-86f8-8c8813a5c7f3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TableForm", "[", 
  RowBox[{"data", ",", 
   RowBox[{"TableHeadings", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"None", ",", 
      RowBox[{"{", 
       RowBox[{"\"\<n\>\"", ",", "\"\<time\>\"", ",", "\"\<result\>\""}], 
       "}"}]}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[338]:=",ExpressionUUID->"89c512e7-7580-4529-a6d3-0ef04abd224a"],

Cell[BoxData[
 TagBox[
  TagBox[GridBox[{
     {
      TagBox["\<\"n\"\>",
       HoldForm], 
      TagBox["\<\"time\"\>",
       HoldForm], 
      TagBox["\<\"result\"\>",
       HoldForm]},
     {"1", "0.001215`", "0"},
     {"3", "0.004188`", "0"},
     {"5", "0.005516`", "0"},
     {"7", "0.007171`", "0"},
     {"9", "0.009748`", "0"},
     {"11", "0.011858`", "0"},
     {"13", "0.019129`", "0"},
     {"15", "0.024334`", "0"},
     {"17", "0.034803`", "0"},
     {"19", "0.03962`", "0"},
     {"21", "0.060073`", "0"},
     {"23", "0.084288`", "0"},
     {"25", "0.099601`", "0"},
     {"27", "0.134321`", "0"},
     {"29", "0.178059`", "0"},
     {"31", "0.223807`", "0"},
     {"33", "0.280071`", "0"},
     {"35", "0.353737`", "0"},
     {"37", "0.445857`", "0"},
     {"39", "0.579452`", "0"},
     {"41", "0.723847`", "0"},
     {"43", "0.947962`", "0"},
     {"45", "1.161231`", "0"},
     {"47", "1.423106`", "0"},
     {"49", "1.618648`", "0"},
     {"51", "2.05579`", "0"},
     {"53", "2.366349`", "0"},
     {"55", "2.931515`", "0"},
     {"57", "3.480804`", "0"},
     {"59", "4.177864`", "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxDividers->{
     "Columns" -> {{False}}, "ColumnsIndexed" -> {}, 
      "Rows" -> {False, True, {False}, False}, "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[2.0999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   {None, OutputFormsDump`HeadedColumns}],
  Function[BoxForm`e$, 
   TableForm[
   BoxForm`e$, TableHeadings -> {None, {"n", "time", "result"}}]]]], "Output",\

 CellChangeTimes->{3.728977983625675*^9},
 CellLabel->"Out[338]//TableForm=",
 CellID->614775327,ExpressionUUID->"bb713164-1237-4569-bd3f-e75dfa8df2ca"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is a log-log plot. A straight line represents a power-law. The fit on \
the last half of points seems to favour a fourth power-law scaling.\
\>", "MathCaption",ExpressionUUID->"e01e3f92-c6d7-4118-80f2-d79e14f1e348"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"fit", "[", "x_", "]"}], "=", 
  RowBox[{"Fit", "[", 
   RowBox[{
    RowBox[{"Log", "[", 
     RowBox[{"data", "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"Range", "[", 
         RowBox[{"15", ",", "30"}], "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"1", ",", "2"}], "}"}]}], "]"}], "]"}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"1", ",", "x"}], "}"}], ",", "x"}], "]"}]}]], "Input",
 CellLabel->
  "In[339]:=",ExpressionUUID->"addcdb82-571c-45b0-bc11-5008a436bd73"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "17.04871797938694`"}], "+", 
  RowBox[{"4.516926434106734`", " ", "x"}]}]], "Output",
 CellChangeTimes->{3.728977983738082*^9},
 CellLabel->"Out[339]=",
 CellID->985095172,ExpressionUUID->"46e18a45-d7cd-42ad-83ff-3ecaa2c3eb40"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"ListLogLogPlot", "[", 
    RowBox[{
     RowBox[{"data", "[", 
      RowBox[{"[", 
       RowBox[{"All", ",", 
        RowBox[{"{", 
         RowBox[{"1", ",", "2"}], "}"}]}], "]"}], "]"}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"PointSize", "[", "0.02", "]"}], ",", 
        RowBox[{"Hue", "[", "0.7", "]"}]}], "}"}]}], ",", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], "]"}], ",", 
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"fit", "[", "x", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", "0", ",", "5"}], "}"}], ",", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", 
   RowBox[{"Axes", "\[Rule]", "False"}]}], "]"}]], "Input",
 CellLabel->
  "In[340]:=",ExpressionUUID->"960cfb89-bdb6-402f-bd23-5f54b3b7c63b"],

Cell[BoxData[
 GraphicsBox[{{{}, {{}, 
     {Hue[0.7, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
      PointBox[CompressedData["
1:eJxTTMoPSmViYGCQA2IQDQPqEWdmy9+SOsDNvLbq1eSP9ouuxd/++Ez0gHHv
2++OB37a520LmP/8jMiB8CijdUVK/+3d8+6Lmx8QhqpndAg8EJ9zpVXogI95
/fbHeswO3cErjujvFTwQ2GOozdLK4nCI7+sE5uX8B+Q9fC3FVrE6mIbwrJXf
xXugqv7XwexVbA5uP01MvO5yHdDZzMJm1sfu4FHR23zuMueBjevyp+hFczjc
nx/9O76e7cCHzRf/1opyOjxdZLL40ElmqPs4HUxj11x+WMx0YEIL8477sVwO
1Sf/yGfwMxyQ/3f2tMUHLgeVw7/YLOf83h+lPdHjXAW3w2c3M5EZn7/vv7jj
6pcbP7gdzh/ZmcYe+2X/0c/8wVGFPA7/zb9cnbjgw/69x63eaTzmcdCcP8VC
9ebL/dd4OzJ++vE68M/ZLKhW9nC/35o+19StvA4N7gZ6z9dd2S/4RkTsgxif
w9Hv2Udmxqzev+Tk4jLuEj6H+CqlSXpKh+23fZigL3WWz6HP0+Xs1ynX7CHh
ye/wPfMqR8qle/b/GQL/J5TwO+Rc5PmzlP+5/WobpQyNw/wO0XaFVloTXtt/
/rcrsoldwKGNx/tEsOlH+8hZaxeK6gg4iL2RiT315bN9pebxn6n+Ag6hby+J
lD34Zg8ADx3gvQ==
       "]]}, {}}, {}, {}, {}, {}}, {{{}, {}, 
     TagBox[
      {RGBColor[0.368417, 0.506779, 0.709798], AbsoluteThickness[1.6], 
       Opacity[1.], LineBox[CompressedData["
1:eJwB4QQe+yFib1JlAgAAAE0AAAACAAAAnX5iBjFkez7XmFTAeAwxwMTTIExX
IFk/9DRrzrIKMcD/xr6DICBpPxHRgdzsCDHAncCNHwUgeT9MCa/4YAUxwGw9
dW33H4k/wHkJMUn+MMDT+2iU8B+ZP6havqEZ8DDAB9viJ+0fqT96HCiDutMw
wKHKn3HrH7k/HKD7RfyaMMCur4nbbC7KP15E4w70HzDAZz/Rmexy0z873mgW
JlovwG7lKeO0rtk/YvrC1eV4LsAKIYQn3zjgP4JQnhuOhC3AjhDLKqFg4z+f
s0zIhKAswGLbVL8DzOY/tVB8+2OpK8BxxYRibyfqP3K0Y/XEtirAsXAMUxhJ
7T8sJR5WdNQpwKF7a+owV/A/389ZPQzfKMCCn/xR9OzxP4+HaIvy+SfAANPg
QLx68z/mBS+gWhknwCZ0ZnhUKvU/Nr52O6slJsDk9ZdWC732P4ODkT1KQiXA
SuVqfZJx+D/Kgi3G0UskwE7kkCseHvo/t0iBFdtZI8Dqw2KAyK37P6IbqMsy
eCLALhHWHUNf/T+FKFAIc4MhwAo/9WHc8/4/ZkLLqwGfIMBH7Vr3IlUAQIAs
j6vxTh/A2MJkAVosAUCCYfeM42gdwLWIxF4g9QFAfrAFPHKjG8BmBXXgzs4C
QGxzFvjRtxnAY3J7tQyaA0BUUM2BzuwXwC9nK85MYQRAirrzmM4qFsDPEiwL
dTkFQLKYHL2fQhTAu66CmywDBkDWkOuuDXsSwHsBKlDM3QZA6vy8rUyNEMAK
3HpIbrQHQJjs+3MeUQ3A5aYhlJ98CEBQE8on3cgJwJQoGQS5VQlA8CGd9T30
BcCQmmbHYSAKQIBkvF7YYALAWpRdzgznCkBYg3XF8779v/hEpfmfvgtAgA18
AXsj9r/j5UJ4wocMQAD/NejqFO6/oj0xG81hDUBABQ9r0GLdv62FdRFnLQ5A
AIh1AvyyhL+HVWNLA/UOQADY7tNnh9s/NdyhqYfND0CgDuVbqwvtP5gpm63N
SxBAsPBQMF6n9T9/wI2YS7kQQFCKpX6iYf0/TRtVpcokEUBQ951L7HsCQEFu
x1sRiRFAiPWcvM0FBkCfHGIkzPURQNgLlxMN3AlAJMOnlk5bEkBA7kTPEnEN
QJAtwirSvhJABNsJuAh6EEBm8wTRySoTQPTKbvu2YRJAYrHyIImPE0DsoC1x
yCgUQGyCU/xKkRNA3HevabgwFEB1U7TXDJMTQMxOMWKoOBRAiPV1jpCWE0Co
/DRTiEgUQK45+fuXnRNAZFg8NUhoFED5wf/WpqsTQNgPS/nHpxRAkNIMjcTH
E0DEfmiBxyYVQJqjbWiGyRNAtFXqebcuFUCjdM5DSMsTQKAsbHKnNhVAthaQ
+svOE0CA2m9jh0YVQNxaE2jT1RNAPDZ3RUdmFUAn4xlD4uMTQKzthQnHpRVA
MLR6HqTlE0CcxAcCt60VQDqF2/ll5xNAjJuJ+qa1FUBMJ52w6eoTQGRJjeuG
xRVAcmsgHvHxE0AgpZTNRuUVQHw8gfmy8xNAFHwWxjbtFUCFDeLUdPUTQABT
mL4m9RVAmK+ji/j4E0DcAJyvBgUWQKKABGe6+hNA0NcdqPYMFkCrUWVCfPwT
QLyun6DmFBZAtCLGHT7+E0CohSGZ1hwWQL7zJvn//xNAmFyjkcYkFkAJnT6E

        "]]},
      Annotation[#, "Charting`Private`Tag$15394#1"]& ]}, {}, {}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->False,
  AxesLabel->{None, None},
  AxesOrigin->{-0.14998754269575076`, -7.304021700738073},
  DisplayFunction->Identity,
  Frame->True,
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->FrontEndValueCache[{{
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}, {
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}}, {{{{-6.907755278982137, 
        FormBox["0.001`", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.010\"", 0.01, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.100\"", 0.1, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox["1", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-9.210340371976182, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.517193191416238, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.111728083308073, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.824046010856292, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.0794415416798357`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.1972245773362196`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.302585092994046, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.4011973816621555`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{-6.907755278982137, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-9.210340371976182, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.517193191416238, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.111728083308073, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.824046010856292, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.0794415416798357`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.1972245773362196`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.302585092994046, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.4011973816621555`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}, {{{0., 
        FormBox["1", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox["2", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox["5", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {2.302585092994046, 
        FormBox["10", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {2.995732273553991, 
        FormBox["20", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {3.912023005428146, 
        FormBox["50", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.4054651081081644, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.0794415416798357`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.1972245773362196`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.70805020110221, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.4011973816621555`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.6888794541139363`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.0943445622221, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.248495242049359, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.382026634673881, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.499809670330265, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.605170185988092, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {2.302585092994046, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.4054651081081644, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.0794415416798357`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.1972245773362196`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.70805020110221, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.4011973816621555`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.6888794541139363`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.0943445622221, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.248495242049359, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.382026634673881, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.499809670330265, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {4.605170185988092, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}}],
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Exp[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Exp[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{-0.08494869674803583, 4.07753744390572}, {-7.165389608485968, 
   1.4298001111445349`}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.05]}},
  Ticks->{
    Charting`ScaledTicks[{Log, Exp}], 
    Charting`ScaledTicks[{Log, Exp}]}]], "Output",
 CellChangeTimes->{3.728977984065475*^9},
 CellLabel->"Out[340]=",
 CellID->1543121081,ExpressionUUID->"a8ff1d5e-8314-4a61-9ca8-e05214aa273e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is a linear-log plot. A straight line represents an exponential:\
\>", "MathCaption",ExpressionUUID->"dbd6ca66-204d-42ed-9eac-bb8b2997fd79"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ListLogPlot", "[", 
  RowBox[{
   RowBox[{"data", "[", 
    RowBox[{"[", 
     RowBox[{"All", ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", "2"}], "}"}]}], "]"}], "]"}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"PointSize", "[", "0.02", "]"}], ",", 
      RowBox[{"Hue", "[", "0.7", "]"}]}], "}"}]}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", 
   RowBox[{"Axes", "\[Rule]", "False"}]}], "]"}]], "Input",
 CellLabel->
  "In[341]:=",ExpressionUUID->"790ea5d1-f8b9-443a-a0de-5dadc8a53ef8"],

Cell[BoxData[
 GraphicsBox[{{}, {{}, 
    {Hue[0.7, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
     PointBox[CompressedData["
1:eJxTTMoPSmViYGCQA2IQDQEf7NUjzsyWvyV1AMLncFh0Lf72x2eiUL6IQ962
gPnPz4hA+TIO7nn3xc0PCEP5Sg6BB+JzrrQKQflqDt3BK47o7xWE8rUcDvF9
ncC8nB/K13MwDeFZK7+LF8o3dHD7aWLidZcLyjd28KjobT53mRPKN3W4Pz/6
d3w9G5Rv7vB0kcniQyeZoXxLB9PYNZcfFjNB+dYO1Sf/yGfwM0D5tg4qh3+x
Wc75vR/Ct3f47GYmMuPzdwi/wcHh/JGdaeyxX6B8R4f/5l+uTlzwAcp3ctCc
P8VC9eZLKN/ZgX/OZkG1sodQvotDg7uB3vN1V6B8V4ej37OPzIxZDeW7OcRX
KU3SUzpsD+G7O/R5upz9OuUalO/h8D3zKkfKpXtQvqdDzkWeP0v5n0P5Xg7R
doVWWhNeQ/neDm083ieCTT9C+T4OYm9kYk99+Qzl+zqEvr0kUvbgmz0A/1CV
YA==
      "]]}, {}}, {}, {}, {}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{0, -7.304021700738073},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->FrontEndValueCache[{{
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}, {
     Automatic, Automatic}}, {{{{-6.907755278982137, 
        FormBox["0.001`", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.010\"", 0.01, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.100\"", 0.1, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox["1", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-9.210340371976182, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.517193191416238, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.111728083308073, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.824046010856292, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.0794415416798357`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.1972245773362196`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.302585092994046, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.4011973816621555`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{-6.907755278982137, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-9.210340371976182, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.517193191416238, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-8.111728083308073, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.824046010856292, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.0794415416798357`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.1972245773362196`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.302585092994046, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {3.4011973816621555`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}, {Automatic, Automatic}}],
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{0, 59.}, {-7.165389608485968, 1.4298001111445349`}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.05]}},
  Ticks->{Automatic, 
    Charting`ScaledTicks[{Log, Exp}]}]], "Output",
 CellChangeTimes->{3.728977984188106*^9},
 CellLabel->"Out[341]=",
 CellID->76003548,ExpressionUUID->"6dcd8b22-ce7f-45ac-af64-10672ccd3513"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "It is difficult to decide from the plots whether the behaviour is polynomic \
or exponential, but seems polynomic. This is misleading. The Dummy algorithm \
is based on the Intersection algorithm, which is known to be exponential in \
the worst case, though highly efficient. The meaning of this will become \
clear in the second example below.\nThis linear-log plot (not produced in \
this notebook) compares the timings of several packages on the same problem. \
On the x-axis, the number of antisymmetric tensors to canonicalize (both when \
the result is zero and when it is not zero). On the y-axis, the timings in \
Seconds. The color code is the following:\n\tRed: MathTensor\n\tYello: dhPark\
\n\tGreen: Tools of Tensor Calculus\n\tBlue (upper): xTensor with pure ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code in xPerm\n\tMagenta: Canon (R. Portugal's own implementation in Maple \
of his canonicalization algorithms)\n\tBlue (lower): xTensor with the \
external C-executable xperm\nThe efficiency of the permutation-based \
algorithms with respect to more classic algorithms is clear."
}], "MathCaption",ExpressionUUID->"b3c38f86-5c21-4a7f-b07a-f93a645ea233"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
%%ImageSize: 500 309.017 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-6.07153e-18 0.019802 0.218177 0.0727013 [
[.19802 -0.0125 -6 -9 ]
[.19802 -0.0125 6 0 ]
[.39604 -0.0125 -6 -9 ]
[.39604 -0.0125 6 0 ]
[.59406 -0.0125 -6 -9 ]
[.59406 -0.0125 6 0 ]
[.79208 -0.0125 -6 -9 ]
[.79208 -0.0125 6 0 ]
[.9901 -0.0125 -6 -9 ]
[.9901 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 7e-05 -34 -4.90625 ]
[-0.0125 7e-05 0 4.90625 ]
[-0.0125 .07277 -28 -4.90625 ]
[-0.0125 .07277 0 4.90625 ]
[-0.0125 .14548 -22 -4.90625 ]
[-0.0125 .14548 0 4.90625 ]
[-0.0125 .21818 -10 -4.90625 ]
[-0.0125 .21818 0 4.90625 ]
[-0.0125 .29088 -16 -4.90625 ]
[-0.0125 .29088 0 4.90625 ]
[-0.0125 .36358 -22 -4.90625 ]
[-0.0125 .36358 0 4.90625 ]
[-0.0125 .43628 -28 -4.90625 ]
[-0.0125 .43628 0 4.90625 ]
[-0.0125 .50898 -34 -4.90625 ]
[-0.0125 .50898 0 4.90625 ]
[-0.0125 .58168 -40 -4.90625 ]
[-0.0125 .58168 0 4.90625 ]
[ 0 0 -0.125 0 ]
[ 0 .61803 .125 0 ]
[1.0125 7e-05 0 -4.90625 ]
[1.0125 7e-05 34 4.90625 ]
[1.0125 .07277 0 -4.90625 ]
[1.0125 .07277 28 4.90625 ]
[1.0125 .14548 0 -4.90625 ]
[1.0125 .14548 22 4.90625 ]
[1.0125 .21818 0 -4.90625 ]
[1.0125 .21818 10 4.90625 ]
[1.0125 .29088 0 -4.90625 ]
[1.0125 .29088 16 4.90625 ]
[1.0125 .36358 0 -4.90625 ]
[1.0125 .36358 22 4.90625 ]
[1.0125 .43628 0 -4.90625 ]
[1.0125 .43628 28 4.90625 ]
[1.0125 .50898 0 -4.90625 ]
[1.0125 .50898 34 4.90625 ]
[1.0125 .58168 0 -4.90625 ]
[1.0125 .58168 40 4.90625 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.19802 0 m
.19802 .00625 L
s
[(10)] .19802 -0.0125 0 1 Mshowa
.39604 0 m
.39604 .00625 L
s
[(20)] .39604 -0.0125 0 1 Mshowa
.59406 0 m
.59406 .00625 L
s
[(30)] .59406 -0.0125 0 1 Mshowa
.79208 0 m
.79208 .00625 L
s
[(40)] .79208 -0.0125 0 1 Mshowa
.9901 0 m
.9901 .00625 L
s
[(50)] .9901 -0.0125 0 1 Mshowa
.125 Mabswid
.0396 0 m
.0396 .00375 L
s
.07921 0 m
.07921 .00375 L
s
.11881 0 m
.11881 .00375 L
s
.15842 0 m
.15842 .00375 L
s
.23762 0 m
.23762 .00375 L
s
.27723 0 m
.27723 .00375 L
s
.31683 0 m
.31683 .00375 L
s
.35644 0 m
.35644 .00375 L
s
.43564 0 m
.43564 .00375 L
s
.47525 0 m
.47525 .00375 L
s
.51485 0 m
.51485 .00375 L
s
.55446 0 m
.55446 .00375 L
s
.63366 0 m
.63366 .00375 L
s
.67327 0 m
.67327 .00375 L
s
.71287 0 m
.71287 .00375 L
s
.75248 0 m
.75248 .00375 L
s
.83168 0 m
.83168 .00375 L
s
.87129 0 m
.87129 .00375 L
s
.91089 0 m
.91089 .00375 L
s
.9505 0 m
.9505 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 7e-05 m
.00625 7e-05 L
s
gsave
-0.0125 7e-05 -95 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.001) show
93.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .07277 m
.00625 .07277 L
s
gsave
-0.0125 .07277 -89 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.01) show
87.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .14548 m
.00625 .14548 L
s
gsave
-0.0125 .14548 -83 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
81.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .21818 m
.00625 .21818 L
s
gsave
-0.0125 .21818 -71 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
69.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .29088 m
.00625 .29088 L
s
gsave
-0.0125 .29088 -77 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .36358 m
.00625 .36358 L
s
gsave
-0.0125 .36358 -83 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
81.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .43628 m
.00625 .43628 L
s
gsave
-0.0125 .43628 -89 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1000) show
87.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .50898 m
.00625 .50898 L
s
gsave
-0.0125 .50898 -95 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10000) show
93.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .58168 m
.00625 .58168 L
s
gsave
-0.0125 .58168 -101 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100000) show
99.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
0 .61803 L
s
0 .61178 m
0 .61803 L
s
.19802 .61178 m
.19802 .61803 L
s
.39604 .61178 m
.39604 .61803 L
s
.59406 .61178 m
.59406 .61803 L
s
.79208 .61178 m
.79208 .61803 L
s
.9901 .61178 m
.9901 .61803 L
s
.125 Mabswid
.0396 .61428 m
.0396 .61803 L
s
.07921 .61428 m
.07921 .61803 L
s
.11881 .61428 m
.11881 .61803 L
s
.15842 .61428 m
.15842 .61803 L
s
.23762 .61428 m
.23762 .61803 L
s
.27723 .61428 m
.27723 .61803 L
s
.31683 .61428 m
.31683 .61803 L
s
.35644 .61428 m
.35644 .61803 L
s
.43564 .61428 m
.43564 .61803 L
s
.47525 .61428 m
.47525 .61803 L
s
.51485 .61428 m
.51485 .61803 L
s
.55446 .61428 m
.55446 .61803 L
s
.63366 .61428 m
.63366 .61803 L
s
.67327 .61428 m
.67327 .61803 L
s
.71287 .61428 m
.71287 .61803 L
s
.75248 .61428 m
.75248 .61803 L
s
.83168 .61428 m
.83168 .61803 L
s
.87129 .61428 m
.87129 .61803 L
s
.91089 .61428 m
.91089 .61803 L
s
.9505 .61428 m
.9505 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 7e-05 m
1 7e-05 L
s
gsave
1.0125 7e-05 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.001) show
93.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .07277 m
1 .07277 L
s
gsave
1.0125 .07277 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.01) show
87.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .14548 m
1 .14548 L
s
gsave
1.0125 .14548 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
81.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .21818 m
1 .21818 L
s
gsave
1.0125 .21818 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
69.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .29088 m
1 .29088 L
s
gsave
1.0125 .29088 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .36358 m
1 .36358 L
s
gsave
1.0125 .36358 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
81.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .43628 m
1 .43628 L
s
gsave
1.0125 .43628 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1000) show
87.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .50898 m
1 .50898 L
s
gsave
1.0125 .50898 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10000) show
93.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 .58168 m
1 .58168 L
s
gsave
1.0125 .58168 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100000) show
99.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.01 w
.0198 .04384 Mdot
.0396 .0509 Mdot
.05941 .06573 Mdot
.07921 .07853 Mdot
.09901 .10042 Mdot
.11881 .13511 Mdot
.13861 .2085 Mdot
.15842 .29808 Mdot
.17822 .39103 Mdot
.19802 .48602 Mdot
.21782 .58405 Mdot
.5 Mabswid
.0198 .04384 m
.0396 .0509 L
.05941 .06573 L
.07921 .07853 L
.09901 .10042 L
.11881 .13511 L
.13861 .2085 L
.15842 .29808 L
.17822 .39103 L
.19802 .48602 L
.21782 .58405 L
s
1 .9 0 r
.0198 .04384 m
.0396 .06573 L
.05941 .08761 L
.07921 .13138 L
.09901 .1095 L
.11881 .19393 L
.13861 .20428 L
.15842 .35695 L
.17822 .38323 L
.19802 .55449 L
.21782 .58037 L
s
.01 w
.0198 .04384 Mdot
.0396 .06573 Mdot
.05941 .08761 Mdot
.07921 .13138 Mdot
.09901 .1095 Mdot
.11881 .19393 Mdot
.13861 .20428 Mdot
.15842 .35695 Mdot
.17822 .38323 Mdot
.19802 .55449 Mdot
.21782 .58037 Mdot
0 1 .4 r
.5 Mabswid
.0198 .06573 m
.0396 .07853 L
.05941 .13997 L
.07921 .1715 L
.09901 .19283 L
.11881 .22704 L
.13861 .24608 L
.15842 .28471 L
.17822 .30954 L
.19802 .34494 L
.21782 .37724 L
.23762 .41674 L
.25743 .45419 L
.27723 .49718 L
.29703 .54016 L
s
.01 w
.0198 .06573 Mdot
.0396 .07853 Mdot
.05941 .13997 Mdot
.07921 .1715 Mdot
.09901 .19283 Mdot
.11881 .22704 Mdot
.13861 .24608 Mdot
.15842 .28471 Mdot
.17822 .30954 Mdot
.19802 .34494 Mdot
.21782 .37724 Mdot
.23762 .41674 Mdot
.25743 .45419 Mdot
.27723 .49718 Mdot
.29703 .54016 Mdot
.2 0 1 r
.0198 .06741 Mdot
.0396 .10758 Mdot
.05941 .13736 Mdot
.07921 .15888 Mdot
.09901 .1801 Mdot
.11881 .19547 Mdot
.13861 .20873 Mdot
.15842 .2241 Mdot
.17822 .23454 Mdot
.19802 .24797 Mdot
.21782 .2571 Mdot
.23762 .26827 Mdot
.25743 .28015 Mdot
.27723 .28636 Mdot
.29703 .29739 Mdot
.31683 .30227 Mdot
.33663 .31292 Mdot
.35644 .3171 Mdot
.37624 .32819 Mdot
.39604 .33045 Mdot
.41584 .34034 Mdot
.43564 .3428 Mdot
.45545 .35159 Mdot
.47525 .35523 Mdot
.49505 .36285 Mdot
.51485 .36892 Mdot
.53465 .37271 Mdot
.55446 .37898 Mdot
.57426 .38054 Mdot
.59406 .38823 Mdot
.61386 .39183 Mdot
.63366 .39516 Mdot
.65347 .40129 Mdot
.67327 .40348 Mdot
.69307 .40909 Mdot
.71287 .40972 Mdot
.73267 .41515 Mdot
.75248 .42064 Mdot
.77228 .42382 Mdot
.79208 .42878 Mdot
.81188 .43131 Mdot
.83168 .43565 Mdot
.85149 .43809 Mdot
.87129 .44246 Mdot
.89109 .44512 Mdot
.91089 .44856 Mdot
.93069 .4515 Mdot
.9505 .45284 Mdot
.9703 .45592 Mdot
.9901 .46014 Mdot
.5 Mabswid
.0198 .06741 m
.0396 .10758 L
.05941 .13736 L
.07921 .15888 L
.09901 .1801 L
.11881 .19547 L
.13861 .20873 L
.15842 .2241 L
.17822 .23454 L
.19802 .24797 L
.21782 .2571 L
.23762 .26827 L
.25743 .28015 L
.27723 .28636 L
.29703 .29739 L
.31683 .30227 L
.33663 .31292 L
.35644 .3171 L
.37624 .32819 L
.39604 .33045 L
.41584 .34034 L
.43564 .3428 L
.45545 .35159 L
.47525 .35523 L
.49505 .36285 L
.51485 .36892 L
.53465 .37271 L
.55446 .37898 L
.57426 .38054 L
.59406 .38823 L
.61386 .39183 L
.63366 .39516 L
.65347 .40129 L
.67327 .40348 L
.69307 .40909 L
.71287 .40972 L
.73267 .41515 L
.75248 .42064 L
.77228 .42382 L
.79208 .42878 L
.81188 .43131 L
.83168 .43565 L
.85149 .43809 L
.87129 .44246 L
.89109 .44512 L
.91089 .44856 L
.93069 .4515 L
.9505 .45284 L
.9703 .45592 L
.9901 .46014 L
Mistroke
Mfstroke
.9901 .46014 m
1 .46025 L
s
1 0 .9 r
.0198 .06573 m
.0396 .09133 L
.05941 .1017 L
.07921 .1095 L
.09901 .12164 L
.11881 .14352 L
.13861 .17245 L
.15842 .17441 L
.17822 .18448 L
.19802 .19526 L
.21782 .20409 L
.23762 .21396 L
.25743 .22319 L
.27723 .23129 L
.29703 .23802 L
.31683 .25065 L
.33663 .25286 L
.35644 .26145 L
.37624 .26789 L
.39604 .2754 L
.41584 .28276 L
.43564 .28592 L
.45545 .29228 L
.47525 .29679 L
.49505 .30154 L
.51485 .30569 L
.53465 .30968 L
.55446 .3135 L
.57426 .31735 L
.59406 .32101 L
.61386 .32498 L
.63366 .32774 L
.65347 .33117 L
.67327 .33466 L
.69307 .3377 L
.71287 .34164 L
.73267 .34499 L
.75248 .34808 L
.77228 .3513 L
.79208 .35358 L
.81188 .35702 L
.83168 .35937 L
.85149 .36204 L
.87129 .36424 L
.89109 .36692 L
.91089 .36953 L
.93069 .37212 L
.9505 .37464 L
.9703 .37704 L
.9901 .37947 L
Mistroke
Mfstroke
.9901 .37947 m
1 .38046 L
s
.01 w
.0198 .06573 Mdot
.0396 .09133 Mdot
.05941 .1017 Mdot
.07921 .1095 Mdot
.09901 .12164 Mdot
.11881 .14352 Mdot
.13861 .17245 Mdot
.15842 .17441 Mdot
.17822 .18448 Mdot
.19802 .19526 Mdot
.21782 .20409 Mdot
.23762 .21396 Mdot
.25743 .22319 Mdot
.27723 .23129 Mdot
.29703 .23802 Mdot
.31683 .25065 Mdot
.33663 .25286 Mdot
.35644 .26145 Mdot
.37624 .26789 Mdot
.39604 .2754 Mdot
.41584 .28276 Mdot
.43564 .28592 Mdot
.45545 .29228 Mdot
.47525 .29679 Mdot
.49505 .30154 Mdot
.51485 .30569 Mdot
.53465 .30968 Mdot
.55446 .3135 Mdot
.57426 .31735 Mdot
.59406 .32101 Mdot
.61386 .32498 Mdot
.63366 .32774 Mdot
.65347 .33117 Mdot
.67327 .33466 Mdot
.69307 .3377 Mdot
.71287 .34164 Mdot
.73267 .34499 Mdot
.75248 .34808 Mdot
.77228 .3513 Mdot
.79208 .35358 Mdot
.81188 .35702 Mdot
.83168 .35937 Mdot
.85149 .36204 Mdot
.87129 .36424 Mdot
.89109 .36692 Mdot
.91089 .36953 Mdot
.93069 .37212 Mdot
.9505 .37464 Mdot
.9703 .37704 Mdot
.9901 .37947 Mdot
.2 0 1 r
.0198 .032 Mdot
.0396 .0702 Mdot
.05941 .07487 Mdot
.07921 .08554 Mdot
.09901 .08668 Mdot
.11881 .09319 Mdot
.13861 .0972 Mdot
.15842 .10421 Mdot
.17822 .10756 Mdot
.19802 .1143 Mdot
.21782 .11693 Mdot
.23762 .12306 Mdot
.25743 .1261 Mdot
.27723 .13198 Mdot
.29703 .13636 Mdot
.31683 .13966 Mdot
.33663 .14325 Mdot
.35644 .1483 Mdot
.37624 .15172 Mdot
.39604 .15798 Mdot
.41584 .16006 Mdot
.43564 .16539 Mdot
.45545 .16699 Mdot
.47525 .17201 Mdot
.49505 .17474 Mdot
.51485 .17963 Mdot
.53465 .1797 Mdot
.55446 .18779 Mdot
.57426 .18643 Mdot
.59406 .19454 Mdot
.61386 .19299 Mdot
.63366 .19755 Mdot
.65347 .20062 Mdot
.67327 .20398 Mdot
.69307 .20738 Mdot
.71287 .21109 Mdot
.73267 .21417 Mdot
.75248 .21792 Mdot
.77228 .21925 Mdot
.79208 .22438 Mdot
.81188 .22592 Mdot
.83168 .23046 Mdot
.85149 .23213 Mdot
.87129 .23537 Mdot
.89109 .24039 Mdot
.91089 .24143 Mdot
.93069 .24422 Mdot
.9505 .24748 Mdot
.9703 .25075 Mdot
.9901 .25288 Mdot
.5 Mabswid
.0198 .032 m
.0396 .0702 L
.05941 .07487 L
.07921 .08554 L
.09901 .08668 L
.11881 .09319 L
.13861 .0972 L
.15842 .10421 L
.17822 .10756 L
.19802 .1143 L
.21782 .11693 L
.23762 .12306 L
.25743 .1261 L
.27723 .13198 L
.29703 .13636 L
.31683 .13966 L
.33663 .14325 L
.35644 .1483 L
.37624 .15172 L
.39604 .15798 L
.41584 .16006 L
.43564 .16539 L
.45545 .16699 L
.47525 .17201 L
.49505 .17474 L
.51485 .17963 L
.53465 .1797 L
.55446 .18779 L
.57426 .18643 L
.59406 .19454 L
.61386 .19299 L
.63366 .19755 L
.65347 .20062 L
.67327 .20398 L
.69307 .20738 L
.71287 .21109 L
.73267 .21417 L
.75248 .21792 L
.77228 .21925 L
.79208 .22438 L
.81188 .22592 L
.83168 .23046 L
.85149 .23213 L
.87129 .23537 L
.89109 .24039 L
.91089 .24143 L
.93069 .24422 L
.9505 .24748 L
.9703 .25075 L
.9901 .25288 L
Mistroke
Mfstroke
.9901 .25288 m
1 .25542 L
s
0 g
0 .21818 m
1 .21818 L
s
0 .34745 m
1 .34745 L
s
0 .47672 m
1 .47672 L
s
0 .57707 m
1 .57707 L
s
gsave
.88119 .20364 -81 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Second) show
99.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.88119 .33291 -81 -9.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 18.125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Minute) show
99.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.88119 .46218 -75 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Hour) show
87.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.88119 .56253 -72 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Day) show
81.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
 ImageSize->{510.625, 315.5},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 
  1}},ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>", "ImageResolution" -> 72.],ImageRangeCache->{{{0., 1146.66}, {707.625, \
0.}} -> {-5.45087, -4.14462, 0.120484, \
0.0328168}},ExpressionUUID->"05777d37-7e27-4146-8ec0-ea9558f56baf"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"UndefTensor", "[", "Anti", "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"Remove", "[", 
  RowBox[{"productAnti", ",", "canAnti"}], "]"}]}], "Input",
 CellLabel->
  "In[342]:=",ExpressionUUID->"7381c63d-68d4-494b-8195-637c9fb9b3fd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Anti"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Anti],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897798433132*^9},
 CellLabel->"During evaluation of In[342]:=",
 CellID->1786344918,ExpressionUUID->"755fe360-d4a6-4a13-98fb-96a58aca7e86"]
}, Open  ]]
}, Open  ]],

Cell[TextData[StyleBox["Example 2:",
 FontWeight->"Bold"]], "Text",ExpressionUUID->"75290d70-a3f6-4a06-b46e-b1c44f8cf380"],

Cell["\<\
As a second example, let us work with the canonicalization of products of \
Riemann tensors.\
\>", "Text",ExpressionUUID->"bc7a5b46-1e2e-4867-8312-afe620e72123"],

Cell[CellGroupData[{

Cell["Define the Riemann tensor:", "MathCaption",ExpressionUUID->"f8fb447d-f240-48b6-9298-8a9fe6f534aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"R", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], ",", "M3", ",", 
   RowBox[{"RiemannSymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], "]"}]}], 
  "]"}]], "Input",
 CellLabel->
  "In[344]:=",ExpressionUUID->"d486fceb-da49-4a59-b705-951519642fee"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"R", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`R[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728977984473402*^9},
 CellLabel->"During evaluation of In[344]:=",
 CellID->1277651766,ExpressionUUID->"fa14ab72-61f6-4345-94c2-bd22f0bade93"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["with the expected permutation symmetries:", "MathCaption",ExpressionUUID->"4a57779e-4954-4cf5-a39a-8cea84420dd0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"R", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "a"}], ",", "b", ",", "c"}], "]"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[345]:=",ExpressionUUID->"0a4263be-ef76-4c96-99c0-51e1ff9a4204"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728977984691874*^9},
 CellLabel->"Out[345]=",
 CellID->804244809,ExpressionUUID->"6c7e54c1-7c64-401e-8f15-db942d4fb686"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"R", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "a"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[346]:=",ExpressionUUID->"019ab91a-91b1-47e2-bdf0-536d93539a67"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "d", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`a, -$CellContext`d, -$CellContext`b, \
-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977984808028*^9},
 CellLabel->"Out[346]=",
 CellID->680430645,ExpressionUUID->"7500004f-97dd-4430-8f2a-f075a3fcff6d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The function ",
 StyleBox["productR",
  FontFamily->"Courier"],
 " constructs random products of Riemann tensors. The second argument gives \
the list of free indices of the expression (always with even length). There \
is no defined conversion to Ricci or RicciScalar tensors."
}], "MathCaption",ExpressionUUID->"616311d9-73e9-4798-8f8e-45fd6504df5f"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"canonindices", "[", 
    RowBox[{"npairs_", ",", "frees_"}], "]"}], ":=", 
   RowBox[{"Join", "[", 
    RowBox[{"frees", ",", 
     RowBox[{"Flatten", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"{", 
         RowBox[{"#", ",", 
          RowBox[{"-", "#"}]}], "}"}], "&"}], "/@", 
       RowBox[{"GetIndicesOfVBundle", "[", 
        RowBox[{"TangentM3", ",", "npairs", ",", 
         RowBox[{"Join", "[", 
          RowBox[{"frees", ",", 
           RowBox[{"-", "frees"}]}], "]"}]}], "]"}]}], "]"}]}], "]"}]}], 
  ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"productR", "[", 
    RowBox[{"n_", ",", 
     RowBox[{"frees_", ":", 
      RowBox[{"{", "}"}]}]}], "]"}], ":=", 
   RowBox[{
    RowBox[{"Apply", "[", 
     RowBox[{"Times", ",", 
      RowBox[{"Apply", "[", 
       RowBox[{"R", ",", 
        RowBox[{"Partition", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"canonindices", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"2", "n"}], "-", 
              RowBox[{
               RowBox[{"Length", "[", "frees", "]"}], "/", "2"}]}], ",", 
             "frees"}], "]"}], "[", 
           RowBox[{"[", 
            RowBox[{"First", "@", 
             RowBox[{"RandomPerm", "[", 
              RowBox[{"4", "n"}], "]"}]}], "]"}], "]"}], ",", "4"}], "]"}], 
        ",", "1"}], "]"}]}], "]"}], "/;", 
    RowBox[{"EvenQ", "[", 
     RowBox[{"Length", "[", "frees", "]"}], "]"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[347]:=",ExpressionUUID->"a50ed49d-b47a-4da1-ad0f-eb0dc97c8f2f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"productR", "[", "1", "]"}]], "Input",
 CellLabel->
  "In[349]:=",ExpressionUUID->"d7dd2de1-38a1-4e42-a772-3c23af96f8a9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"R", 
      StyleBox[GridBox[{
         {"a", " ", " ", "b"},
         {" ", "a", "b", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`R[$CellContext`a, -$CellContext`a, -$CellContext`b, \
$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897798497915*^9},
 CellLabel->"Out[349]=",
 CellID->647430240,ExpressionUUID->"da83b066-3505-4b4e-977f-9167c40930ae"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"productR", "[", "10", "]"}]], "Input",
 CellLabel->
  "In[350]:=",ExpressionUUID->"e6415d7e-0ab9-489a-bab1-b007a22a5c7a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", " ", "j14", " "},
          {"b", "j12", " ", "j14"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`b, -$CellContext`j12, $CellContext`j14, \
-$CellContext`j14],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", "j", "j13", "i"},
          {"g", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`g, $CellContext`j, $CellContext`j13, \
$CellContext`i],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {"g", "j15", " ", "a"},
          {" ", " ", "j15", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[$CellContext`g, $CellContext`j15, -$CellContext`j15, \
$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", " ", " ", "f"},
          {"i", "h", "a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`i, -$CellContext`h, -$CellContext`a, \
$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", " ", " ", "j18"},
          {"j1", "f", "d", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`j1, -$CellContext`f, -$CellContext`d, \
$CellContext`j18],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {"j1", " ", " ", "j17"},
          {" ", "e", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[$CellContext`j1, -$CellContext`e, -$CellContext`c, \
$CellContext`j17],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {"j10", "b", " ", " "},
          {" ", " ", "j16", "j10"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[$CellContext`j10, $CellContext`b, -$CellContext`j16, \
-$CellContext`j10],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", " ", "j11", "h"},
          {"j11", "j13", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`j11, -$CellContext`j13, $CellContext`j11, \
$CellContext`h],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", " ", "c", "e"},
          {"j17", "j", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`j17, -$CellContext`j, $CellContext`c, \
$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", "j16", "j12", "d"},
          {"j18", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`j18, $CellContext`j16, $CellContext`j12, \
$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977985101462*^9},
 CellLabel->"Out[350]=",
 CellID->60404331,ExpressionUUID->"179db550-674f-40bb-87fa-5832a63336e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"productR", "[", 
  RowBox[{"3", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", 
     RowBox[{"-", "d"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[351]:=",ExpressionUUID->"eb6880de-7bcd-417d-bce4-b757c5d980a0"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {"a", "f", " ", " "},
          {" ", " ", "d", "h"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[$CellContext`a, $CellContext`f, -$CellContext`d, \
-$CellContext`h],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {"b", " ", "c", "g"},
          {" ", "e", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[$CellContext`b, -$CellContext`e, $CellContext`c, \
$CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"R", 
       StyleBox[GridBox[{
          {" ", "h", " ", "e"},
          {"f", " ", "g", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`R[-$CellContext`f, $CellContext`h, -$CellContext`g, \
$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728977985224517*^9},
 CellLabel->"Out[351]=",
 CellID->1603309733,ExpressionUUID->"380b9c9b-bdc2-4e2d-8ab7-cb58ca7f9606"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"productR", "[", 
  RowBox[{"3", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[352]:=",ExpressionUUID->"129aa6ef-2d57-42f2-a76f-dd4307d6d6bc"],

Cell[BoxData[
 RowBox[{"productR", "[", 
  RowBox[{"3", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c"}], "}"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728977985342948*^9},
 CellLabel->"Out[352]=",
 CellID->201879975,ExpressionUUID->"ec286e7f-12d8-4e74-bd53-e2e75d889ec4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This second function canonicalizes the products of Riemanns, timing the \
process:\
\>", "MathCaption",ExpressionUUID->"b9d45f12-b7f4-434f-9054-d97bd8049a33"],

Cell[BoxData[
 RowBox[{
  RowBox[{"canR", "[", 
   RowBox[{"n_", ",", 
    RowBox[{"frees_", ":", 
     RowBox[{"{", "}"}]}]}], "]"}], ":=", 
  RowBox[{"Flatten", "[", 
   RowBox[{"{", 
    RowBox[{"n", ",", 
     RowBox[{"With", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"product", "=", 
         RowBox[{"productR", "[", 
          RowBox[{"n", ",", "frees"}], "]"}]}], "}"}], ",", 
       RowBox[{"AbsoluteTiming", "[", 
        RowBox[{"ToCanonical", "[", "product", "]"}], "]"}]}], "]"}]}], "}"}],
    "]"}]}]], "Input",
 CellLabel->
  "In[353]:=",ExpressionUUID->"5c60c886-6fe6-43a2-a72d-874e7e44e7f1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Examples:", "MathCaption",ExpressionUUID->"412b6862-162a-4456-9488-915be9bf5cda"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "1", "]"}]], "Input",
 CellLabel->
  "In[354]:=",ExpressionUUID->"f5af0254-856f-4def-8bdb-92f3b8ca1c72"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1", ",", "0.001532`", ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {"a", "b", " ", " "},
            {" ", " ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[$CellContext`a, $CellContext`b, -$CellContext`a, \
-$CellContext`b],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977985452059*^9},
 CellLabel->"Out[354]=",
 CellID->2120061212,ExpressionUUID->"98ab902e-5284-45b7-94be-7402c799bec8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "2", "]"}]], "Input",
 CellLabel->
  "In[355]:=",ExpressionUUID->"6eb20774-d3c8-40f4-a2f4-e63d0c87b56d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2", ",", "0.004183`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977985559325*^9},
 CellLabel->"Out[355]=",
 CellID->820119060,ExpressionUUID->"4e079120-6de3-4b22-9250-a80a9e2b7719"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "3", "]"}]], "Input",
 CellLabel->
  "In[356]:=",ExpressionUUID->"7d1bb697-efc9-4f7e-ba3f-a301240025d1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"3", ",", "0.006365`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977985707755*^9},
 CellLabel->"Out[356]=",
 CellID->1190994974,ExpressionUUID->"682767dd-8e40-4bf0-b8a7-10abb259991c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "4", "]"}]], "Input",
 CellLabel->
  "In[357]:=",ExpressionUUID->"7b38f26b-7e84-4972-a4e1-96fb146432f1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"4", ",", "0.008595`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {"a", "b", " ", "c"},
            {" ", " ", "a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[$CellContext`a, $CellContext`b, -$CellContext`a, \
$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "d", "e", "f"},
            {"b", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`b, $CellContext`d, $CellContext`e, \
$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", " ", "g"},
            {"c", "e", "f", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`c, -$CellContext`e, -$CellContext`f, \
$CellContext`g],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "h", " ", " "},
            {"d", " ", "g", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`d, $CellContext`h, -$CellContext`g, \
-$CellContext`h],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977985857621*^9},
 CellLabel->"Out[357]=",
 CellID->357454331,ExpressionUUID->"cf5bb4d2-a10c-43f7-bfb6-8a11bf302581"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "10", "]"}]], "Input",
 CellLabel->
  "In[358]:=",ExpressionUUID->"c4c207bf-5248-4cdd-b1a7-6922d04411ae"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"10", ",", "0.024734`", ",", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"R", 
          StyleBox[GridBox[{
             {"a", "b", " ", "c"},
             {" ", " ", "a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`R[$CellContext`a, $CellContext`b, -$CellContext`a, \
$CellContext`c],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "d", " ", "e"},
            {"b", " ", "c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`b, $CellContext`d, -$CellContext`c, \
$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "f", "g", "h"},
            {"d", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`d, $CellContext`f, $CellContext`g, \
$CellContext`h],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "i", " ", "j"},
            {"e", " ", "g", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`e, $CellContext`i, -$CellContext`g, \
$CellContext`j],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j1", "j10", "j11"},
            {"f", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`f, $CellContext`j1, $CellContext`j10, \
$CellContext`j11],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j12", " ", "j13"},
            {"h", " ", "j1", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`h, $CellContext`j12, -$CellContext`j1, \
$CellContext`j13],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j14", "j15", "j16"},
            {"i", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`i, $CellContext`j14, $CellContext`j15, \
$CellContext`j16],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j17", " ", " "},
            {"j", " ", "j10", "j13"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j, $CellContext`j17, -$CellContext`j10, \
-$CellContext`j13],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"j11", "j14", "j15", "j16"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j11, -$CellContext`j14, -$CellContext`j15, \
-$CellContext`j16],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j18", " ", " "},
            {"j12", " ", "j17", "j18"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j12, $CellContext`j18, -$CellContext`j17, \
-$CellContext`j18],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977986015621*^9},
 CellLabel->"Out[358]=",
 CellID->1223180458,ExpressionUUID->"68762d59-ecd2-4b3e-9a8f-ae415ce0dc1e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "20", "]"}]], "Input",
 CellLabel->
  "In[359]:=",ExpressionUUID->"936c3368-c43c-4049-9303-00516971bf61"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"20", ",", "0.209571`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", "e", "f"},
            {"a", "c", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`a, -$CellContext`c, $CellContext`e, \
$CellContext`f],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {"a", "b", "c", "d"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "g", " ", "h"},
            {"b", " ", "d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`b, $CellContext`g, -$CellContext`d, \
$CellContext`h],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "i", " ", "j"},
            {"e", " ", "f", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`e, $CellContext`i, -$CellContext`f, \
$CellContext`j],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j1", "j10", "j11"},
            {"g", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`g, $CellContext`j1, $CellContext`j10, \
$CellContext`j11],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j12", "j13", "j14"},
            {"h", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`h, $CellContext`j12, $CellContext`j13, \
$CellContext`j14],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", "j15", "j16"},
            {"i", "j1", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`i, -$CellContext`j1, $CellContext`j15, \
$CellContext`j16],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j17", "j18", "j19"},
            {"j", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j, $CellContext`j17, $CellContext`j18, \
$CellContext`j19],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j2", " ", "j20"},
            {"j10", " ", "j15", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j10, $CellContext`j2, -$CellContext`j15, \
$CellContext`j20],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j21", "j22", "j23"},
            {"j11", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j11, $CellContext`j21, $CellContext`j22, \
$CellContext`j23],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j24", " ", "j25"},
            {"j12", " ", "j21", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j12, $CellContext`j24, -$CellContext`j21, \
$CellContext`j25],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", "j26", "j27"},
            {"j13", "j18", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j13, -$CellContext`j18, $CellContext`j26, \
$CellContext`j27],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j28", " ", "j29"},
            {"j14", " ", "j24", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j14, $CellContext`j28, -$CellContext`j24, \
$CellContext`j29],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j3", " ", " "},
            {"j16", " ", "j22", "j29"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j16, $CellContext`j3, -$CellContext`j22, \
-$CellContext`j29],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j30", " ", "j31"},
            {"j17", " ", "j23", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j17, $CellContext`j30, -$CellContext`j23, \
$CellContext`j31],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j32", " ", "j33"},
            {"j19", " ", "j26", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j19, $CellContext`j32, -$CellContext`j26, \
$CellContext`j33],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", " ", "j34"},
            {"j2", "j33", "j3", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j2, -$CellContext`j33, -$CellContext`j3, \
$CellContext`j34],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", "j35", " ", " "},
            {"j20", " ", "j30", "j34"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j20, $CellContext`j35, -$CellContext`j30, \
-$CellContext`j34],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", " ", "j36"},
            {"j25", "j32", "j28", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j25, -$CellContext`j32, -$CellContext`j28, \
$CellContext`j36],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"R", 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"j27", "j31", "j35", "j36"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`R[-$CellContext`j27, -$CellContext`j31, -$CellContext`j35, \
-$CellContext`j36],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728977986340464*^9},
 CellLabel->"Out[359]=",
 CellID->133004875,ExpressionUUID->"783061e0-ed2e-4021-b47c-a833b800d32f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"canR", "[", "25", "]"}]], "Input",
 CellLabel->
  "In[360]:=",ExpressionUUID->"9fa72213-ca81-44f1-b720-8ca2381e53f6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"25", ",", "0.094011`", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728977986546266*^9},
 CellLabel->"Out[360]=",
 CellID->2029941245,ExpressionUUID->"06c70c2d-4524-45ba-a29f-85ecec8950ca"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Scaling of timings. Evaluating next cell could take between half a minute and \
a minute:\
\>", "MathCaption",ExpressionUUID->"26c109e5-2e2f-4f4a-9820-55db413f2b93"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data1", "=", 
   RowBox[{"canR", "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data2", "=", 
   RowBox[{"canR", "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data3", "=", 
   RowBox[{"canR", "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[361]:=",ExpressionUUID->"f466d4a3-f2e4-4418-a8dc-cf044cfd15b5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"plot0", "=", 
  RowBox[{"ListLogPlot", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Join", "[", 
      RowBox[{"data1", ",", "data2", ",", "data3"}], "]"}], "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "2"}], "}"}]}], "]"}], "]"}], ",", 
    RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
    RowBox[{"Frame", "\[Rule]", "True"}], ",", 
    RowBox[{"PlotStyle", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"PointSize", "[", "0.02", "]"}], ",", 
       RowBox[{"Hue", "[", "0.7", "]"}]}], "}"}]}]}], "]"}]}]], "Input",
 CellLabel->
  "In[364]:=",ExpressionUUID->"2045601b-73f0-4120-bf3a-d87848bea561"],

Cell[BoxData[
 GraphicsBox[{{}, {{}, 
    {Hue[0.7, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
     PointBox[CompressedData["
1:eJxF030s1HEcwPEfzsNxnh+S0DHPYefinKI+FM08LI8ly2MqW7m1SBZTaC7L
w4ZczcPclvOQSXkKOdGidjtPkWhMHlLStJ3IQ6vm+/39brvd3vf9/n5/fF77
mMXxghPkCYLw+/f9/7v3WTvufklASI8Z9aI/oMfSkpOXqY9aBT7kjRXxruqh
1oLBpB/Z1nO6qPXAULFxhRmvg9oQMn19ulbjtVEbQxx3ZSjPURM1E749da7d
YWuhNoeBfP8iboEGags4uWQrk6Opo7YCepj7pFCZgdoGboYXGo9+oqO2g1L/
iMtlGvjcHoIf8gWzIYqoHcH0Xl2kUILvs2BMm2/AbSJQO0F/68yqohwNNRsa
DKSPY4UqqA+Dp+x+4o38P+K9doblJJaVnAZ+3gV0s8PHc5s30DkHrlR4aq7K
ZKhdoa0uqmShFb+fC0EJw96K0R/RuRvY0fqZNgrrYuzBTg/nlNgfID3CJPtE
1wgD0iNqQqfmQjLlYeIhX/62Cs9fDwTn05uNOvD8DSFZjab6PpXyeJaR9aiz
EjcT+MMbO1s87GMOuSMDTIcAyuOLVcbpPn18bgWu/sal54opD2JmQWgcqkp6
lBCB183n8Lzt4WBtZqJHCp6nI2Qvt3Jcq/F9FqSsEBbqVUqkh013oyTPDPux
oX1i0aG9YkeMPfQN1e4sPsD3nWFSdmpbqRu3C4yaeHjzaL9ID6eAny93X2+S
HrPzpURh8IQYe4yL2vyiRL9JD9aJnOJw2RDp4U/rEqy8ofZD4YxXAaOS2g+L
RbFs24zycJCExqS2UB49OZX0gK+Uh6hctHRLRHl0uBSn6cXifWCCNKszJSiS
2o/2/YyRs+/wPlhAqq9tPv0u5THYN8T2FuJ524C71lSsbyDlcVvi65O0oUZ6
bHOmvOvDlEmPaevvT55P43mzIC0zsaZpc0uMPfiMQ9LUF9R+tAxHW0Zsy/di
j1f1CT78Ycpjfb3Ma2Aez98FLlYnF2fH4/dzoMHvSIEZfZf0MP2sfZTo2SU9
GCHBkZHTm6THYowfseam0PsXco5Fkg==
      "]]}, {}}, {}, {}, {}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0.625, -7.268493894781159},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->FrontEndValueCache[{{
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}, {
     Automatic, Automatic}}, {{{{-6.907755278982137, 
        FormBox["0.001`", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.005\"", 0.005, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.010\"", 0.01, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.050\"", 0.05, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.100\"", 0.1, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.500\"", 0.5, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox["1", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{-6.907755278982137, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}, {Automatic, Automatic}}],
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{1., 25.}, {-7.157083580733111, -0.2496441097541946}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.05]}},
  Ticks->{Automatic, 
    Charting`ScaledTicks[{Log, Exp}]}]], "Output",
 CellChangeTimes->{3.728977993754455*^9},
 CellLabel->"Out[364]=",
 CellID->149804826,ExpressionUUID->"29262ecc-83ab-4277-8a11-b41440edb4b9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We repeat the computation, but now leaving two and four free indices in the \
expressions:\
\>", "MathCaption",ExpressionUUID->"3466479f-10e9-4cb3-b104-b0f38e953f71"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data1", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"canR", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"a", ",", "b"}], "}"}]}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data2", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"canR", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"a", ",", "b"}], "}"}]}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data3", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"canR", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"a", ",", "b"}], "}"}]}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[365]:=",ExpressionUUID->"be790f1f-7919-4d17-89bc-6000a4ef7f6e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"plot1", "=", 
  RowBox[{"ListLogPlot", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Join", "[", 
      RowBox[{"data1", ",", "data2", ",", "data3"}], "]"}], "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "2"}], "}"}]}], "]"}], "]"}], ",", 
    RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
    RowBox[{"Frame", "\[Rule]", "True"}], ",", 
    RowBox[{"PlotStyle", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"PointSize", "[", "0.02", "]"}], ",", 
       RowBox[{"Hue", "[", "0", "]"}]}], "}"}]}]}], "]"}]}]], "Input",
 CellLabel->
  "In[368]:=",ExpressionUUID->"aa25ab63-7880-4635-bd01-0f094ce929f3"],

Cell[BoxData[
 GraphicsBox[{{}, {{}, 
    {Hue[0, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
     PointBox[CompressedData["
1:eJxN030s1HEcwPFD8vwsUziP3RnO43l+uE8q68LZjjyUUXmYUNaDmGuWIbsm
WrVa5UJK52aFxoiGpZYUQ2yessvYhDoSOU/VfL+/b7/tt9/e3+/vr89rH8vT
GfwkRRqNFvz3/ffdeWSc4MXkemMtkw50AOzMY+NrfXtQq0Jfvc5++WtD1LrA
TAycedZqgNoQbqTNVPA4+qiNIUHxToDMBLcp+C4sqw5P6aK2AAFXv/nDVdxW
UFL8OYAeqIPaBuomxFahh7RQM6A2xDNs7LgmalvwoE++bPLHbQfi4Gq+Z4A6
agcIj7RYFziponYEM+ucusb03aidwTJU/kkjD7cLRI/mP36UqozaFWrFB5a6
IxVRu8F1Ucx0QaIKajZUmVVK3VhKqN3hLOdW00obDbUHsL6lDnb2ydt32hPa
ixkefvc3UHtB7ID0waz9KmpvuBvl02Ivx/cyTkpBhJX2KeIhZJ1UWaIZUR7S
3HU1HwnxOCLOkltbEY/JafMtRwnx6Gr5rhXVqUd5+E0La9a6iYeQx+8pjyMe
dt7SFNks8QiU9Ov5fyUeqwplDrY52pRHemXZz+UrxCPI5W2hVFmD8hCOdXG5
pWqUR01+wo+HT7CPM2SOu19WjMPzdQHmoOh9Ujn2cYWYoL3X+u9hHzewYfLL
zJvx/NkQXcR+Z5eHvdyBFdbNDVnC/3tA6XDH+WR1hQ7swSyuTTXvx/deYJ5p
zX2et015RGhOsH7f3KY8BgSiBr4T8RBVb82LD5L9qK8yalNtwPPXhexOmbGr
gHiINhbnhJeIByP4i7pOHfGYbJ2amF/A87aAoZQSJYMs4lEk6S1M/28/UkMa
Ljhs4vkzwKli2f7XONmP4vjZVzNC4rEc5phWeIbsR+zIZvIAHfs4wtHOtqzc
WHzvDLXdt9/sE+L5uACviO5Tb4h9XAGy1Q7zW4lHw8ey+KeVuNkQ3hjEUxAS
jwzehmBlbrMde5zrubjwoo14rK+YlorYuyiPbInT6AnfdcpjhDOkQe/dbP8D
BP9MYw==
      "]]}, {}}, {}, {}, {}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0.625, -7.2364662608581884`},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->FrontEndValueCache[{{
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}, {
     Automatic, Automatic}}, {{{{-6.907755278982137, 
        FormBox["0.001`", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.005\"", 0.005, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.010\"", 0.01, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.050\"", 0.05, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.100\"", 0.1, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.09531017980432493, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.1823215567939546, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.26236426446749106`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.3364722366212129, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.4054651081081644, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.47000362924573563`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{-6.907755278982137, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.09531017980432493, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.1823215567939546, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.26236426446749106`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.3364722366212129, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.4054651081081644, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.47000362924573563`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}, {Automatic, Automatic}}],
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{1., 25.}, {-7.1445893492568135`, -1.448220829971567}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.05]}},
  Ticks->{Automatic, 
    Charting`ScaledTicks[{Log, Exp}]}]], "Output",
 CellChangeTimes->{3.728977999450864*^9},
 CellLabel->"Out[368]=",
 CellID->918876298,ExpressionUUID->"cb849551-d8bf-4e38-b54a-3f9d2942e8be"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"data1", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"canR", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "}"}]}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data2", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"canR", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "}"}]}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"data3", "=", 
   RowBox[{
    RowBox[{
     RowBox[{"canR", "[", 
      RowBox[{"#", ",", 
       RowBox[{"{", 
        RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "}"}]}], "]"}], "&"}], "/@", 
    RowBox[{"Range", "[", 
     RowBox[{"1", ",", "25"}], "]"}]}]}], ";"}]}], "Input",
 CellLabel->
  "In[369]:=",ExpressionUUID->"a387db89-fb5f-4460-b887-71913b0c88e2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"plot2", "=", 
  RowBox[{"ListLogPlot", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"Join", "[", 
      RowBox[{"data1", ",", "data2", ",", "data3"}], "]"}], "[", 
     RowBox[{"[", 
      RowBox[{"All", ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "2"}], "}"}]}], "]"}], "]"}], ",", 
    RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
    RowBox[{"Frame", "\[Rule]", "True"}], ",", 
    RowBox[{"PlotStyle", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"PointSize", "[", "0.02", "]"}], ",", 
       RowBox[{"Hue", "[", "0.3", "]"}]}], "}"}]}]}], "]"}]}]], "Input",
 CellLabel->
  "In[372]:=",ExpressionUUID->"6d844840-04d0-4eef-9f84-91eba2baae07"],

Cell[BoxData[
 GraphicsBox[{{}, {{}, 
    {Hue[0.3, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
     PointBox[CompressedData["
1:eJxF038s1HEcx3HFCfl5ZOZuq+PosmN37o6T292bm6WJZOnHOK2QWunmmPk9
NtP0hyGmH0pFGS5qYatWJz/SSDOuDEfNFElDh9zu6Md8Pt/Pd/vuu+f3j+8f
n8f3xTqnjEneaWZmFvHv/v/cvpZldaul97uU7l3oBUgdWkPW1HtQW8FEzSt3
uRduR+jfmgGVzhm1C+Q++pFnn0VH7QZBhyLW5fVOqJmgY4xmXeHg3gfTN+O1
JzSOqD0gdeaY3p+Omw3OnPzY9zEOqL1h2islfXXQFjUHWBYbvX4FuH2g/VRc
IUO6GzUXXpzP/GkxY43aD+bp3Q2e+VaoeaCzLoO9KkvUfHB9OJSqbcTtD6yR
1RvqYdwCeNq3bFiItkAthJDDJY6swZ2oRSAZaKy9fMYcdQCoF0dyYsd2oA6E
PtnXmk/DJs12iyE7vfrgSusW6iBQ8eRJ13gGDfZ4/ngiSpHDoDxaqtiudpXE
w9agXxkcdaE8ikMzhKvjxEM3Nb402UY8QkY71/Zb4GZC4u3lwmoW8bgw5cb0
ziYe6vjo4wdM+PzZ0CD1r6zPIR6l6SPsuTI7ysMyuf9jRbg95cHKNerZQDwy
+aXGwXc2lIdW3OTWdgf78GCg5I9CcA/78EHn5THRak48jL6SLa2CeCSkK/N9
nGmUB2fTuSqMQTxaOr65cIN2UR6TAjmD30k8fvOyR8QLuMVQG2ylOanUUx6+
pvpe7i38vWVZD3928S6XSXl0N6teKhiulIf0Qc8RfSbxcF9QfS4qJx55kYbX
tAziUSr0/RDaRPYxHxm12WxFPHL7n1UIs4gHj9MuSmskHpK5Oq+zkfj8vaHY
SP/+K4nsg/5EpI4dJ/u47iLTKB1xc8EYFiz9oiAewiJaIs2J7GOtnEkTzxIP
G9VG7mg88Sjg9a1Px+HzF4B7rX9Vyhuyj/lAWZM8De9BBGNXNwUXh/D/HgBa
c/uE9g7iESrRHj1ts0TtQ9QW/nbikoHymPaUZ+Q4mTR/AUs2SNU=
      
      "]]}, {}}, {}, {}, {}, {}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0.625, -7.484398029593986},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->FrontEndValueCache[{{
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}, {
     Automatic, Automatic}}, {{{{-6.907755278982137, 
        FormBox["0.001`", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.005\"", 0.005, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.010\"", 0.01, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.050\"", 0.05, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.100\"", 0.1, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.500\"", 0.5, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.4054651081081644, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{-6.907755278982137, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.4054651081081644, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}, {Automatic, Automatic}}],
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{1., 25.}, {-7.380995097224639, -0.9700132903252625}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.05]}},
  Ticks->{Automatic, 
    Charting`ScaledTicks[{Log, Exp}]}]], "Output",
 CellChangeTimes->{3.728978004304397*^9},
 CellLabel->"Out[372]=",
 CellID->999034909,ExpressionUUID->"5fbc540e-bc39-41bf-98e7-f2298cc8837c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Combining the three plots, we do not see any important difference in the \
timings:\
\>", "MathCaption",ExpressionUUID->"8abc4310-5721-4430-a922-4fc6cf23bb01"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{"plot0", ",", "plot1", ",", "plot2"}], "]"}]], "Input",
 CellLabel->
  "In[373]:=",ExpressionUUID->"d1ab9a61-e620-4c02-b28d-f3c9b5a5e646"],

Cell[BoxData[
 GraphicsBox[{{{}, {{}, 
     {Hue[0.7, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
      PointBox[CompressedData["
1:eJxF030s1HEcwPEfzsNxnh+S0DHPYefinKI+FM08LI8ly2MqW7m1SBZTaC7L
w4ZczcPclvOQSXkKOdGidjtPkWhMHlLStJ3IQ6vm+/39brvd3vf9/n5/fF77
mMXxghPkCYLw+/f9/7v3WTvufklASI8Z9aI/oMfSkpOXqY9aBT7kjRXxruqh
1oLBpB/Z1nO6qPXAULFxhRmvg9oQMn19ulbjtVEbQxx3ZSjPURM1E749da7d
YWuhNoeBfP8iboEGags4uWQrk6Opo7YCepj7pFCZgdoGboYXGo9+oqO2g1L/
iMtlGvjcHoIf8gWzIYqoHcH0Xl2kUILvs2BMm2/AbSJQO0F/68yqohwNNRsa
DKSPY4UqqA+Dp+x+4o38P+K9doblJJaVnAZ+3gV0s8PHc5s30DkHrlR4aq7K
ZKhdoa0uqmShFb+fC0EJw96K0R/RuRvY0fqZNgrrYuzBTg/nlNgfID3CJPtE
1wgD0iNqQqfmQjLlYeIhX/62Cs9fDwTn05uNOvD8DSFZjab6PpXyeJaR9aiz
EjcT+MMbO1s87GMOuSMDTIcAyuOLVcbpPn18bgWu/sal54opD2JmQWgcqkp6
lBCB183n8Lzt4WBtZqJHCp6nI2Qvt3Jcq/F9FqSsEBbqVUqkh013oyTPDPux
oX1i0aG9YkeMPfQN1e4sPsD3nWFSdmpbqRu3C4yaeHjzaL9ID6eAny93X2+S
HrPzpURh8IQYe4yL2vyiRL9JD9aJnOJw2RDp4U/rEqy8ofZD4YxXAaOS2g+L
RbFs24zycJCExqS2UB49OZX0gK+Uh6hctHRLRHl0uBSn6cXifWCCNKszJSiS
2o/2/YyRs+/wPlhAqq9tPv0u5THYN8T2FuJ524C71lSsbyDlcVvi65O0oUZ6
bHOmvOvDlEmPaevvT55P43mzIC0zsaZpc0uMPfiMQ9LUF9R+tAxHW0Zsy/di
j1f1CT78Ycpjfb3Ma2Aez98FLlYnF2fH4/dzoMHvSIEZfZf0MP2sfZTo2SU9
GCHBkZHTm6THYowfseam0PsXco5Fkg==
       
       "]]}, {}}, {}, {}, {}, {}}, {{}, {{}, 
     {Hue[0, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
      PointBox[CompressedData["
1:eJxN030s1HEcwPFD8vwsUziP3RnO43l+uE8q68LZjjyUUXmYUNaDmGuWIbsm
WrVa5UJK52aFxoiGpZYUQ2yessvYhDoSOU/VfL+/b7/tt9/e3+/vr89rH8vT
GfwkRRqNFvz3/ffdeWSc4MXkemMtkw50AOzMY+NrfXtQq0Jfvc5++WtD1LrA
TAycedZqgNoQbqTNVPA4+qiNIUHxToDMBLcp+C4sqw5P6aK2AAFXv/nDVdxW
UFL8OYAeqIPaBuomxFahh7RQM6A2xDNs7LgmalvwoE++bPLHbQfi4Gq+Z4A6
agcIj7RYFziponYEM+ucusb03aidwTJU/kkjD7cLRI/mP36UqozaFWrFB5a6
IxVRu8F1Ucx0QaIKajZUmVVK3VhKqN3hLOdW00obDbUHsL6lDnb2ydt32hPa
ixkefvc3UHtB7ID0waz9KmpvuBvl02Ivx/cyTkpBhJX2KeIhZJ1UWaIZUR7S
3HU1HwnxOCLOkltbEY/JafMtRwnx6Gr5rhXVqUd5+E0La9a6iYeQx+8pjyMe
dt7SFNks8QiU9Ov5fyUeqwplDrY52pRHemXZz+UrxCPI5W2hVFmD8hCOdXG5
pWqUR01+wo+HT7CPM2SOu19WjMPzdQHmoOh9Ujn2cYWYoL3X+u9hHzewYfLL
zJvx/NkQXcR+Z5eHvdyBFdbNDVnC/3tA6XDH+WR1hQ7swSyuTTXvx/deYJ5p
zX2et015RGhOsH7f3KY8BgSiBr4T8RBVb82LD5L9qK8yalNtwPPXhexOmbGr
gHiINhbnhJeIByP4i7pOHfGYbJ2amF/A87aAoZQSJYMs4lEk6S1M/28/UkMa
Ljhs4vkzwKli2f7XONmP4vjZVzNC4rEc5phWeIbsR+zIZvIAHfs4wtHOtqzc
WHzvDLXdt9/sE+L5uACviO5Tb4h9XAGy1Q7zW4lHw8ey+KeVuNkQ3hjEUxAS
jwzehmBlbrMde5zrubjwoo14rK+YlorYuyiPbInT6AnfdcpjhDOkQe/dbP8D
BP9MYw==
       "]]}, {}}, {}, {}, {}, {}}, {{}, {{}, 
     {Hue[0.3, 1, 1], PointSize[0.02], AbsoluteThickness[1.6], 
      PointBox[CompressedData["
1:eJxF038s1HEcx3HFCfl5ZOZuq+PosmN37o6T292bm6WJZOnHOK2QWunmmPk9
NtP0hyGmH0pFGS5qYatWJz/SSDOuDEfNFElDh9zu6Md8Pt/Pd/vuu+f3j+8f
n8f3xTqnjEneaWZmFvHv/v/cvpZldaul97uU7l3oBUgdWkPW1HtQW8FEzSt3
uRduR+jfmgGVzhm1C+Q++pFnn0VH7QZBhyLW5fVOqJmgY4xmXeHg3gfTN+O1
JzSOqD0gdeaY3p+Omw3OnPzY9zEOqL1h2islfXXQFjUHWBYbvX4FuH2g/VRc
IUO6GzUXXpzP/GkxY43aD+bp3Q2e+VaoeaCzLoO9KkvUfHB9OJSqbcTtD6yR
1RvqYdwCeNq3bFiItkAthJDDJY6swZ2oRSAZaKy9fMYcdQCoF0dyYsd2oA6E
PtnXmk/DJs12iyE7vfrgSusW6iBQ8eRJ13gGDfZ4/ngiSpHDoDxaqtiudpXE
w9agXxkcdaE8ikMzhKvjxEM3Nb402UY8QkY71/Zb4GZC4u3lwmoW8bgw5cb0
ziYe6vjo4wdM+PzZ0CD1r6zPIR6l6SPsuTI7ysMyuf9jRbg95cHKNerZQDwy
+aXGwXc2lIdW3OTWdgf78GCg5I9CcA/78EHn5THRak48jL6SLa2CeCSkK/N9
nGmUB2fTuSqMQTxaOr65cIN2UR6TAjmD30k8fvOyR8QLuMVQG2ylOanUUx6+
pvpe7i38vWVZD3928S6XSXl0N6teKhiulIf0Qc8RfSbxcF9QfS4qJx55kYbX
tAziUSr0/RDaRPYxHxm12WxFPHL7n1UIs4gHj9MuSmskHpK5Oq+zkfj8vaHY
SP/+K4nsg/5EpI4dJ/u47iLTKB1xc8EYFiz9oiAewiJaIs2J7GOtnEkTzxIP
G9VG7mg88Sjg9a1Px+HzF4B7rX9Vyhuyj/lAWZM8De9BBGNXNwUXh/D/HgBa
c/uE9g7iESrRHj1ts0TtQ9QW/nbikoHymPaUZ+Q4mTR/AUs2SNU=
       
       "]]}, {}}, {}, {}, {}, {}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, True},
  AxesLabel->{None, None},
  AxesOrigin->{0.625, -7.268493894781159},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{None, None}, {None, None}},
  FrameTicks->FrontEndValueCache[{{
      Charting`ScaledTicks[{Log, Exp}], 
      Charting`ScaledFrameTicks[{Log, Exp}]}, {
     Automatic, Automatic}}, {{{{-6.907755278982137, 
        FormBox["0.001`", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.005\"", 0.005, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.010\"", 0.01, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.050\"", 0.05, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.100\"", 0.1, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TagBox[
          InterpretationBox["\"0.500\"", 0.5, AutoDelete -> True], 
          NumberForm[#, {
            DirectedInfinity[1], 3}]& ], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox["1", TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}, {{-6.907755278982137, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.298317366548036, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.605170185988091, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.995732273553991, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.3025850929940455`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {0., 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.01, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.600902459542082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.418580902748128, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.264430222920869, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.1308988302963465`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-7.013115794639964, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-6.214608098422191, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.809142990314028, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.521460917862246, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-5.115995809754082, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.961845129926823, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.8283137373023015`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-4.710530701645918, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.912023005428146, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.506557897319982, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-3.2188758248682006`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.8134107167600364`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.659260036932778, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.5257286443082556`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-2.4079456086518722`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-1.2039728043259361`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.916290731874155, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.5108256237659907, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.35667494393873245`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.2231435513142097, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {-0.10536051565782628`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {0.6931471805599453, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.0986122886681098`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.3862943611198906`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.6094379124341003`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.791759469228055, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}, {1.9459101490553132`, 
        FormBox[
         TemplateBox[{0., 0.}, "Spacer2"], TraditionalForm], {0.005, 0.}, {
         AbsoluteThickness[0.1]}}}}, {Automatic, Automatic}}],
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImagePadding->All,
  Method->{"CoordinatesToolOptions" -> {"DisplayFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& ), "CopiedValueFunction" -> ({
        (Identity[#]& )[
         Part[#, 1]], 
        (Exp[#]& )[
         Part[#, 2]]}& )}},
  PlotRange->{{1., 25.}, {-7.157083580733111, -0.2496441097541946}},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.02], 
     Scaled[0.05]}},
  Ticks->{Automatic, 
    Charting`ScaledTicks[{Log, Exp}]}]], "Output",
 CellChangeTimes->{3.7289780045882053`*^9},
 CellLabel->"Out[373]=",
 CellID->1752563317,ExpressionUUID->"1451c39e-d12e-49f6-ae1e-e752003f4d7a"]
}, Open  ]]
}, Open  ]],

Cell[TextData[StyleBox["Example 3:",
 FontWeight->"Bold"]], "Text",ExpressionUUID->"1f6cdcce-45f6-4960-8e52-6911fed6e1c9"],

Cell["Now let us concentrate on the products of seven Riemanns.", "Text",ExpressionUUID->"6a657e17-4ff5-4b35-b0ea-f0a001c61803"],

Cell[CellGroupData[{

Cell["\<\
This figure (not constructed in this notebook) shows an histogram of the \
timings of canonicalization of a hundred thousand ramdom products of seven \
Riemanns. Note in the smaller plot that there are some cases which take much \
more time (and much more memory) to canonicalize. This is a consequence of \
the algorithm not being polynomial, but exponential in the worst case. Those \
are cases in which the symmetry is much higher than usual, for example \
because the product of Riemanns can be separated in products of decoupled \
monomials. There are cases in which, however, the reason is not at all clear, \
so that in general it is not possible to prepare the canonicalization \
algorithms to avoid all possible \"worst cases\".\
\>", "MathCaption",ExpressionUUID->"59a01a97-dc36-430c-96ed-2027d076f997"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
%%ImageSize: 400 247.214 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.588604 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0147151 0.97619 0.603319 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.025 0.05 0.0147151 1.80942e-05 [
[.005 -0.0125 -3 -9 ]
[.005 -0.0125 3 0 ]
[.2 -0.0125 -12 -9 ]
[.2 -0.0125 12 0 ]
[.4 -0.0125 -12 -9 ]
[.4 -0.0125 12 0 ]
[.6 -0.0125 -12 -9 ]
[.6 -0.0125 12 0 ]
[.8 -0.0125 -12 -9 ]
[.8 -0.0125 12 0 ]
[1 -0.0125 -9 -9 ]
[1 -0.0125 9 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -26 -19.375 ]
[.5 -0.0125 26 -9 ]
[-0.0125 .01472 -6 -4.5 ]
[-0.0125 .01472 0 4.5 ]
[-0.0125 .10519 -24 -4.5 ]
[-0.0125 .10519 0 4.5 ]
[-0.0125 .19566 -30 -4.5 ]
[-0.0125 .19566 0 4.5 ]
[-0.0125 .28613 -30 -4.5 ]
[-0.0125 .28613 0 4.5 ]
[-0.0125 .3766 -30 -4.5 ]
[-0.0125 .3766 0 4.5 ]
[-0.0125 .46707 -30 -4.5 ]
[-0.0125 .46707 0 4.5 ]
[-0.0125 .55754 -30 -4.5 ]
[-0.0125 .55754 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.005 0 m
.005 .00625 L
s
[(0)] .005 -0.0125 0 1 Mshowa
.2 0 m
.2 .00625 L
s
[(0.02)] .2 -0.0125 0 1 Mshowa
.4 0 m
.4 .00625 L
s
[(0.04)] .4 -0.0125 0 1 Mshowa
.6 0 m
.6 .00625 L
s
[(0.06)] .6 -0.0125 0 1 Mshowa
.8 0 m
.8 .00625 L
s
[(0.08)] .8 -0.0125 0 1 Mshowa
1 0 m
1 .00625 L
s
[(0.1)] 1 -0.0125 0 1 Mshowa
0 0 m
1 0 L
s
gsave
.5 -0.0125 -87 -23.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 18.375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Time) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
93.000 11.562 moveto
(H) show
99.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(s) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
105.000 11.562 moveto
(L) show
111.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
[(0)] -0.0125 .01472 1 0 Mshowa
0 .10519 m
.00625 .10519 L
s
[(5000)] -0.0125 .10519 1 0 Mshowa
0 .19566 m
.00625 .19566 L
s
[(10000)] -0.0125 .19566 1 0 Mshowa
0 .28613 m
.00625 .28613 L
s
[(15000)] -0.0125 .28613 1 0 Mshowa
0 .3766 m
.00625 .3766 L
s
[(20000)] -0.0125 .3766 1 0 Mshowa
0 .46707 m
.00625 .46707 L
s
[(25000)] -0.0125 .46707 1 0 Mshowa
0 .55754 m
.00625 .55754 L
s
[(30000)] -0.0125 .55754 1 0 Mshowa
.125 Mabswid
0 .03281 m
.00375 .03281 L
s
0 .0509 m
.00375 .0509 L
s
0 .069 m
.00375 .069 L
s
0 .08709 m
.00375 .08709 L
s
0 .12328 m
.00375 .12328 L
s
0 .14137 m
.00375 .14137 L
s
0 .15947 m
.00375 .15947 L
s
0 .17756 m
.00375 .17756 L
s
0 .21375 m
.00375 .21375 L
s
0 .23185 m
.00375 .23185 L
s
0 .24994 m
.00375 .24994 L
s
0 .26803 m
.00375 .26803 L
s
0 .30422 m
.00375 .30422 L
s
0 .32232 m
.00375 .32232 L
s
0 .34041 m
.00375 .34041 L
s
0 .3585 m
.00375 .3585 L
s
0 .39469 m
.00375 .39469 L
s
0 .41279 m
.00375 .41279 L
s
0 .43088 m
.00375 .43088 L
s
0 .44898 m
.00375 .44898 L
s
0 .48516 m
.00375 .48516 L
s
0 .50326 m
.00375 .50326 L
s
0 .52135 m
.00375 .52135 L
s
0 .53945 m
.00375 .53945 L
s
0 .57563 m
.00375 .57563 L
s
0 .59373 m
.00375 .59373 L
s
0 .61182 m
.00375 .61182 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
0 .61803 m
1 .61803 L
s
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.005 .01472 m
.005 .01472 L
.045 .01472 L
.045 .01472 L
F
0 g
.5 Mabswid
.005 .01472 m
.005 .01472 L
.045 .01472 L
.045 .01472 L
.005 .01472 L
s
1 0 0 r
.055 .01472 m
.055 .01472 L
.095 .01472 L
.095 .01472 L
F
0 g
.055 .01472 m
.055 .01472 L
.095 .01472 L
.095 .01472 L
.055 .01472 L
s
1 0 0 r
.105 .01472 m
.105 .01472 L
.145 .01472 L
.145 .01472 L
F
0 g
.105 .01472 m
.105 .01472 L
.145 .01472 L
.145 .01472 L
.105 .01472 L
s
1 0 0 r
.155 .01472 m
.155 .01495 L
.195 .01495 L
.195 .01472 L
F
0 g
.155 .01472 m
.155 .01495 L
.195 .01495 L
.195 .01472 L
.155 .01472 L
s
1 0 0 r
.205 .01472 m
.205 .0242 L
.245 .0242 L
.245 .01472 L
F
0 g
.205 .01472 m
.205 .0242 L
.245 .0242 L
.245 .01472 L
.205 .01472 L
s
1 0 0 r
.255 .01472 m
.255 .10367 L
.295 .10367 L
.295 .01472 L
F
0 g
.255 .01472 m
.255 .10367 L
.295 .10367 L
.295 .01472 L
.255 .01472 L
s
1 0 0 r
.305 .01472 m
.305 .35979 L
.345 .35979 L
.345 .01472 L
F
0 g
.305 .01472 m
.305 .35979 L
.345 .35979 L
.345 .01472 L
.305 .01472 L
s
1 0 0 r
.355 .01472 m
.355 .28718 L
.395 .28718 L
.395 .01472 L
F
0 g
.355 .01472 m
.355 .28718 L
.395 .28718 L
.395 .01472 L
.355 .01472 L
s
1 0 0 r
.405 .01472 m
.405 .08402 L
.445 .08402 L
.445 .01472 L
F
0 g
.405 .01472 m
.405 .08402 L
.445 .08402 L
.445 .01472 L
.405 .01472 L
s
1 0 0 r
.455 .01472 m
.455 .02345 L
.495 .02345 L
.495 .01472 L
F
0 g
.455 .01472 m
.455 .02345 L
.495 .02345 L
.495 .01472 L
.455 .01472 L
s
1 0 0 r
.505 .01472 m
.505 .01889 L
.545 .01889 L
.545 .01472 L
F
0 g
.505 .01472 m
.505 .01889 L
.545 .01889 L
.545 .01472 L
.505 .01472 L
s
1 0 0 r
.555 .01472 m
.555 .01544 L
.595 .01544 L
.595 .01472 L
F
0 g
.555 .01472 m
.555 .01544 L
.595 .01544 L
.595 .01472 L
.555 .01472 L
s
1 0 0 r
.605 .01472 m
.605 .01479 L
.645 .01479 L
.645 .01472 L
F
0 g
.605 .01472 m
.605 .01479 L
.645 .01479 L
.645 .01472 L
.605 .01472 L
s
1 0 0 r
.655 .01472 m
.655 .01475 L
.695 .01475 L
.695 .01472 L
F
0 g
.655 .01472 m
.655 .01475 L
.695 .01475 L
.695 .01472 L
.655 .01472 L
s
1 0 0 r
.705 .01472 m
.705 .01475 L
.745 .01475 L
.745 .01472 L
F
0 g
.705 .01472 m
.705 .01475 L
.745 .01475 L
.745 .01472 L
.705 .01472 L
s
1 0 0 r
.755 .01472 m
.755 .01472 L
.795 .01472 L
.795 .01472 L
F
0 g
.755 .01472 m
.755 .01472 L
.795 .01472 L
.795 .01472 L
.755 .01472 L
s
1 0 0 r
.805 .01472 m
.805 .01472 L
.845 .01472 L
.845 .01472 L
F
0 g
.805 .01472 m
.805 .01472 L
.845 .01472 L
.845 .01472 L
.805 .01472 L
s
1 0 0 r
.855 .01472 m
.855 .01472 L
.895 .01472 L
.895 .01472 L
F
0 g
.855 .01472 m
.855 .01472 L
.895 .01472 L
.895 .01472 L
.855 .01472 L
s
1 0 0 r
.905 .01472 m
.905 .01472 L
.945 .01472 L
.945 .01472 L
F
0 g
.905 .01472 m
.905 .01472 L
.945 .01472 L
.945 .01472 L
.905 .01472 L
s
1 0 0 r
.955 .01472 m
.955 .01472 L
.995 .01472 L
.995 .01472 L
F
0 g
.955 .01472 m
.955 .01472 L
.995 .01472 L
.995 .01472 L
.955 .01472 L
s
0 .4 1 r
.005 .01472 m
.005 .01472 L
.045 .01472 L
.045 .01472 L
F
0 g
.005 .01472 m
.005 .01472 L
.045 .01472 L
.045 .01472 L
.005 .01472 L
s
0 .4 1 r
.055 .01472 m
.055 .01472 L
.095 .01472 L
.095 .01472 L
F
0 g
.055 .01472 m
.055 .01472 L
.095 .01472 L
.095 .01472 L
.055 .01472 L
s
0 .4 1 r
.105 .01472 m
.105 .01472 L
.145 .01472 L
.145 .01472 L
F
0 g
.105 .01472 m
.105 .01472 L
.145 .01472 L
.145 .01472 L
.105 .01472 L
s
0 .4 1 r
.155 .01495 m
.155 .01522 L
.195 .01522 L
.195 .01495 L
F
0 g
.155 .01495 m
.155 .01522 L
.195 .01522 L
.195 .01495 L
.155 .01495 L
s
0 .4 1 r
.205 .0242 m
.205 .03136 L
.245 .03136 L
.245 .0242 L
F
0 g
.205 .0242 m
.205 .03136 L
.245 .03136 L
.245 .0242 L
.205 .0242 L
s
0 .4 1 r
.255 .10367 m
.255 .14322 L
.295 .14322 L
.295 .10367 L
F
0 g
.255 .10367 m
.255 .14322 L
.295 .14322 L
.295 .10367 L
.255 .10367 L
s
0 .4 1 r
.305 .35979 m
.305 .60332 L
.345 .60332 L
.345 .35979 L
F
0 g
.305 .35979 m
.305 .60332 L
.345 .60332 L
.345 .35979 L
.305 .35979 L
s
0 .4 1 r
.355 .28718 m
.355 .55691 L
.395 .55691 L
.395 .28718 L
F
0 g
.355 .28718 m
.355 .55691 L
.395 .55691 L
.395 .28718 L
.355 .28718 L
s
0 .4 1 r
.405 .08402 m
.405 .23474 L
.445 .23474 L
.445 .08402 L
F
0 g
.405 .08402 m
.405 .23474 L
.445 .23474 L
.445 .08402 L
.405 .08402 L
s
0 .4 1 r
.455 .02345 m
.455 .17782 L
.495 .17782 L
.495 .02345 L
F
0 g
.455 .02345 m
.455 .17782 L
.495 .17782 L
.495 .02345 L
.455 .02345 L
s
0 .4 1 r
.505 .01889 m
.505 .13446 L
.545 .13446 L
.545 .01889 L
F
0 g
.505 .01889 m
.505 .13446 L
.545 .13446 L
.545 .01889 L
.505 .01889 L
s
0 .4 1 r
.555 .01544 m
.555 .03426 L
.595 .03426 L
.595 .01544 L
F
0 g
.555 .01544 m
.555 .03426 L
.595 .03426 L
.595 .01544 L
.555 .01544 L
s
0 .4 1 r
.605 .01479 m
.605 .02099 L
.645 .02099 L
.645 .01479 L
F
0 g
.605 .01479 m
.605 .02099 L
.645 .02099 L
.645 .01479 L
.605 .01479 L
s
0 .4 1 r
.655 .01475 m
.655 .0169 L
.695 .0169 L
.695 .01475 L
F
0 g
.655 .01475 m
.655 .0169 L
.695 .0169 L
.695 .01475 L
.655 .01475 L
s
0 .4 1 r
.705 .01475 m
.705 .01573 L
.745 .01573 L
.745 .01475 L
F
0 g
.705 .01475 m
.705 .01573 L
.745 .01573 L
.745 .01475 L
.705 .01475 L
s
0 .4 1 r
.755 .01472 m
.755 .01524 L
.795 .01524 L
.795 .01472 L
F
0 g
.755 .01472 m
.755 .01524 L
.795 .01524 L
.795 .01472 L
.755 .01472 L
s
0 .4 1 r
.805 .01472 m
.805 .01484 L
.845 .01484 L
.845 .01472 L
F
0 g
.805 .01472 m
.805 .01484 L
.845 .01484 L
.845 .01472 L
.805 .01472 L
s
0 .4 1 r
.855 .01472 m
.855 .01477 L
.895 .01477 L
.895 .01472 L
F
0 g
.855 .01472 m
.855 .01477 L
.895 .01477 L
.895 .01472 L
.855 .01472 L
s
0 .4 1 r
.905 .01472 m
.905 .01472 L
.945 .01472 L
.945 .01472 L
F
0 g
.905 .01472 m
.905 .01472 L
.945 .01472 L
.945 .01472 L
.905 .01472 L
s
0 .4 1 r
.955 .01472 m
.955 .01473 L
.995 .01473 L
.995 .01472 L
F
0 g
.955 .01472 m
.955 .01473 L
.995 .01473 L
.995 .01472 L
.955 .01472 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.471429 0.309017 0.947619 0.603319 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00678334 0 0.0416667 [
[.02381 -0.0125 -3 -9 ]
[.02381 -0.0125 3 0 ]
[.15948 -0.0125 -8 -9.8125 ]
[.15948 -0.0125 8 0 ]
[.29514 -0.0125 -8 -9.8125 ]
[.29514 -0.0125 8 0 ]
[.43081 -0.0125 -8 -9.8125 ]
[.43081 -0.0125 8 0 ]
[.56648 -0.0125 -8 -9.8125 ]
[.56648 -0.0125 8 0 ]
[.70214 -0.0125 -8 -9.8125 ]
[.70214 -0.0125 8 0 ]
[.83781 -0.0125 -8 -9.8125 ]
[.83781 -0.0125 8 0 ]
[.97348 -0.0125 -8 -9.8125 ]
[.97348 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -26 -20.1875 ]
[.5 -0.0125 26 -9.8125 ]
[-0.0125 .08333 -10 -4.90625 ]
[-0.0125 .08333 0 4.90625 ]
[-0.0125 .16667 -10 -4.90625 ]
[-0.0125 .16667 0 4.90625 ]
[-0.0125 .25 -10 -4.90625 ]
[-0.0125 .25 0 4.90625 ]
[-0.0125 .33333 -10 -4.90625 ]
[-0.0125 .33333 0 4.90625 ]
[ 0 0 -0.125 0 ]
[ 0 .33333 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .33333 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
[(0)] .02381 -0.0125 0 1 Mshowa
.15948 0 m
.15948 .00625 L
s
gsave
.15948 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.1) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.29514 0 m
.29514 .00625 L
s
gsave
.29514 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.2) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.43081 0 m
.43081 .00625 L
s
gsave
.43081 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.3) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.56648 0 m
.56648 .00625 L
s
gsave
.56648 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.4) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.70214 0 m
.70214 .00625 L
s
gsave
.70214 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.5) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.83781 0 m
.83781 .00625 L
s
gsave
.83781 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.6) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.97348 0 m
.97348 .00625 L
s
gsave
.97348 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(.7) show
75.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
s
gsave
.5 -0.0125 -87 -24.1875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 18.375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Time) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
93.000 11.562 moveto
(H) show
99.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(s) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
105.000 11.562 moveto
(L) show
111.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .04167 m
.00625 .04167 L
s
0 .08333 m
.00625 .08333 L
s
gsave
-0.0125 .08333 -71 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
69.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .125 m
.00625 .125 L
s
0 .16667 m
.00625 .16667 L
s
gsave
-0.0125 .16667 -71 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
69.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .20833 m
.00625 .20833 L
s
0 .25 m
.00625 .25 L
s
gsave
-0.0125 .25 -71 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
69.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .29167 m
.00625 .29167 L
s
0 .33333 m
.00625 .33333 L
s
gsave
-0.0125 .33333 -71 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
69.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
0 .33333 L
s
0 .33333 m
1 .33333 L
s
1 0 m
1 .33333 L
s
0 0 m
1 0 L
1 .33333 L
0 .33333 L
closepath
clip
newpath
1 0 0 r
.02788 0 m
.02788 0 L
.03331 0 L
.03331 0 L
F
.03466 0 m
.03466 0 L
.04009 0 L
.04009 0 L
F
.04145 0 m
.04145 0 L
.04687 0 L
.04687 0 L
F
.04823 0 m
.04823 .54167 L
.05366 .54167 L
.05366 0 L
F
.05501 0 m
.05501 21.8333 L
.06044 21.8333 L
.06044 0 L
F
.0618 0 m
.0618 204.833 L
.06722 204.833 L
.06722 0 L
F
.06858 0 m
.06858 794.625 L
.07401 794.625 L
.07401 0 L
F
.07536 0 m
.07536 627.417 L
.08079 627.417 L
.08079 0 L
F
.08215 0 m
.08215 159.583 L
.08757 159.583 L
.08757 0 L
F
.08893 0 m
.08893 20.125 L
.09436 20.125 L
.09436 0 L
F
.09571 0 m
.09571 9.625 L
.10114 9.625 L
.10114 0 L
F
.1025 0 m
.1025 1.66667 L
.10792 1.66667 L
.10792 0 L
F
.10928 0 m
.10928 .16667 L
.11471 .16667 L
.11471 0 L
F
.11606 0 m
.11606 .08333 L
.12149 .08333 L
.12149 0 L
F
.12285 0 m
.12285 .08333 L
.12827 .08333 L
.12827 0 L
F
.12963 0 m
.12963 0 L
.13506 0 L
.13506 0 L
F
.13641 0 m
.13641 0 L
.14184 0 L
.14184 0 L
F
.1432 0 m
.1432 0 L
.14862 0 L
.14862 0 L
F
.14998 0 m
.14998 0 L
.15541 0 L
.15541 0 L
F
.15676 0 m
.15676 0 L
.16219 0 L
.16219 0 L
F
.16355 0 m
.16355 0 L
.16897 0 L
.16897 0 L
F
.17033 0 m
.17033 0 L
.17576 0 L
.17576 0 L
F
.17711 0 m
.17711 0 L
.18254 0 L
.18254 0 L
F
.1839 0 m
.1839 0 L
.18932 0 L
.18932 0 L
F
.19068 0 m
.19068 0 L
.19611 0 L
.19611 0 L
F
.19746 0 m
.19746 0 L
.20289 0 L
.20289 0 L
F
.20425 0 m
.20425 0 L
.20967 0 L
.20967 0 L
F
.21103 0 m
.21103 0 L
.21646 0 L
.21646 0 L
F
.21781 0 m
.21781 0 L
.22324 0 L
.22324 0 L
F
.2246 0 m
.2246 0 L
.23002 0 L
.23002 0 L
F
.23138 0 m
.23138 0 L
.23681 0 L
.23681 0 L
F
.23816 0 m
.23816 0 L
.24359 0 L
.24359 0 L
F
.24495 0 m
.24495 0 L
.25037 0 L
.25037 0 L
F
.25173 0 m
.25173 0 L
.25716 0 L
.25716 0 L
F
.25851 0 m
.25851 0 L
.26394 0 L
.26394 0 L
F
.2653 0 m
.2653 0 L
.27072 0 L
.27072 0 L
F
.27208 0 m
.27208 0 L
.27751 0 L
.27751 0 L
F
.27886 0 m
.27886 0 L
.28429 0 L
.28429 0 L
F
.28565 0 m
.28565 0 L
.29107 0 L
.29107 0 L
F
.29243 0 m
.29243 0 L
.29786 0 L
.29786 0 L
F
.29921 0 m
.29921 0 L
.30464 0 L
.30464 0 L
F
.306 0 m
.306 0 L
.31142 0 L
.31142 0 L
F
.31278 0 m
.31278 0 L
.31821 0 L
.31821 0 L
F
.31956 0 m
.31956 0 L
.32499 0 L
.32499 0 L
F
.32635 0 m
.32635 0 L
.33177 0 L
.33177 0 L
F
.33313 0 m
.33313 0 L
.33856 0 L
.33856 0 L
F
.33991 0 m
.33991 0 L
.34534 0 L
.34534 0 L
F
.3467 0 m
.3467 0 L
.35212 0 L
.35212 0 L
F
.35348 0 m
.35348 0 L
.35891 0 L
.35891 0 L
F
.36026 0 m
.36026 0 L
.36569 0 L
.36569 0 L
F
.36705 0 m
.36705 0 L
.37247 0 L
.37247 0 L
F
.37383 0 m
.37383 0 L
.37926 0 L
.37926 0 L
F
.38061 0 m
.38061 0 L
.38604 0 L
.38604 0 L
F
.3874 0 m
.3874 0 L
.39282 0 L
.39282 0 L
F
.39418 0 m
.39418 0 L
.39961 0 L
.39961 0 L
F
.40096 0 m
.40096 0 L
.40639 0 L
.40639 0 L
F
.40775 0 m
.40775 0 L
.41317 0 L
.41317 0 L
F
.41453 0 m
.41453 0 L
.41996 0 L
.41996 0 L
F
.42131 0 m
.42131 0 L
.42674 0 L
.42674 0 L
F
.4281 0 m
.4281 0 L
.43352 0 L
.43352 0 L
F
.43488 0 m
.43488 0 L
.44031 0 L
.44031 0 L
F
.44166 0 m
.44166 0 L
.44709 0 L
.44709 0 L
F
.44845 0 m
.44845 0 L
.45387 0 L
.45387 0 L
F
.45523 0 m
.45523 0 L
.46066 0 L
.46066 0 L
F
.46201 0 m
.46201 0 L
.46744 0 L
.46744 0 L
F
.4688 0 m
.4688 0 L
.47422 0 L
.47422 0 L
F
.47558 0 m
.47558 0 L
.48101 0 L
.48101 0 L
F
.48236 0 m
.48236 0 L
.48779 0 L
.48779 0 L
F
.48915 0 m
.48915 0 L
.49457 0 L
.49457 0 L
F
.49593 0 m
.49593 0 L
.50136 0 L
.50136 0 L
F
.50271 0 m
.50271 0 L
.50814 0 L
.50814 0 L
F
.5095 0 m
.5095 0 L
.51492 0 L
.51492 0 L
F
.51628 0 m
.51628 0 L
.52171 0 L
.52171 0 L
F
.52306 0 m
.52306 0 L
.52849 0 L
.52849 0 L
F
.52985 0 m
.52985 0 L
.53527 0 L
.53527 0 L
F
.53663 0 m
.53663 0 L
.54206 0 L
.54206 0 L
F
.54341 0 m
.54341 0 L
.54884 0 L
.54884 0 L
F
.5502 0 m
.5502 0 L
.55562 0 L
.55562 0 L
F
.55698 0 m
.55698 0 L
.56241 0 L
.56241 0 L
F
.56376 0 m
.56376 0 L
.56919 0 L
.56919 0 L
F
.57055 0 m
.57055 0 L
.57597 0 L
.57597 0 L
F
.57733 0 m
.57733 0 L
.58276 0 L
.58276 0 L
F
.58411 0 m
.58411 0 L
.58954 0 L
.58954 0 L
F
.5909 0 m
.5909 0 L
.59632 0 L
.59632 0 L
F
.59768 0 m
.59768 0 L
.60311 0 L
.60311 0 L
F
.60446 0 m
.60446 0 L
.60989 0 L
.60989 0 L
F
.61125 0 m
.61125 0 L
.61667 0 L
.61667 0 L
F
.61803 0 m
.61803 0 L
.62346 0 L
.62346 0 L
F
.62481 0 m
.62481 0 L
.63024 0 L
.63024 0 L
F
.6316 0 m
.6316 0 L
.63702 0 L
.63702 0 L
F
.63838 0 m
.63838 0 L
.64381 0 L
.64381 0 L
F
.64516 0 m
.64516 0 L
.65059 0 L
.65059 0 L
F
.65195 0 m
.65195 0 L
.65737 0 L
.65737 0 L
F
.65873 0 m
.65873 0 L
.66416 0 L
.66416 0 L
F
.66551 0 m
.66551 0 L
.67094 0 L
.67094 0 L
F
.6723 0 m
.6723 0 L
.67772 0 L
.67772 0 L
F
.67908 0 m
.67908 0 L
.68451 0 L
.68451 0 L
F
.68586 0 m
.68586 0 L
.69129 0 L
.69129 0 L
F
.69265 0 m
.69265 0 L
.69807 0 L
.69807 0 L
F
.69943 0 m
.69943 0 L
.70486 0 L
.70486 0 L
F
.70621 0 m
.70621 0 L
.71164 0 L
.71164 0 L
F
.713 0 m
.713 0 L
.71842 0 L
.71842 0 L
F
.71978 0 m
.71978 0 L
.72521 0 L
.72521 0 L
F
.72656 0 m
.72656 0 L
.73199 0 L
.73199 0 L
F
.73335 0 m
.73335 0 L
.73877 0 L
.73877 0 L
F
.74013 0 m
.74013 0 L
.74556 0 L
.74556 0 L
F
.74691 0 m
.74691 0 L
.75234 0 L
.75234 0 L
F
.7537 0 m
.7537 0 L
.75912 0 L
.75912 0 L
F
.76048 0 m
.76048 0 L
.76591 0 L
.76591 0 L
F
.76726 0 m
.76726 0 L
.77269 0 L
.77269 0 L
F
.77405 0 m
.77405 0 L
.77947 0 L
.77947 0 L
F
.78083 0 m
.78083 0 L
.78626 0 L
.78626 0 L
F
.78761 0 m
.78761 0 L
.79304 0 L
.79304 0 L
F
.7944 0 m
.7944 0 L
.79982 0 L
.79982 0 L
F
.80118 0 m
.80118 0 L
.80661 0 L
.80661 0 L
F
.80796 0 m
.80796 0 L
.81339 0 L
.81339 0 L
F
.81475 0 m
.81475 0 L
.82017 0 L
.82017 0 L
F
.82153 0 m
.82153 0 L
.82696 0 L
.82696 0 L
F
.82831 0 m
.82831 0 L
.83374 0 L
.83374 0 L
F
.8351 0 m
.8351 0 L
.84052 0 L
.84052 0 L
F
.84188 0 m
.84188 0 L
.84731 0 L
.84731 0 L
F
.84866 0 m
.84866 0 L
.85409 0 L
.85409 0 L
F
.85545 0 m
.85545 0 L
.86087 0 L
.86087 0 L
F
.86223 0 m
.86223 0 L
.86766 0 L
.86766 0 L
F
.86901 0 m
.86901 0 L
.87444 0 L
.87444 0 L
F
.8758 0 m
.8758 0 L
.88122 0 L
.88122 0 L
F
.88258 0 m
.88258 0 L
.88801 0 L
.88801 0 L
F
.88936 0 m
.88936 0 L
.89479 0 L
.89479 0 L
F
.89615 0 m
.89615 0 L
.90157 0 L
.90157 0 L
F
.90293 0 m
.90293 0 L
.90836 0 L
.90836 0 L
F
.90971 0 m
.90971 0 L
.91514 0 L
.91514 0 L
F
.9165 0 m
.9165 0 L
.92192 0 L
.92192 0 L
F
.92328 0 m
.92328 0 L
.92871 0 L
.92871 0 L
F
.93006 0 m
.93006 0 L
.93549 0 L
.93549 0 L
F
.93685 0 m
.93685 0 L
.94227 0 L
.94227 0 L
F
.94363 0 m
.94363 0 L
.94906 0 L
.94906 0 L
F
.95041 0 m
.95041 0 L
.95584 0 L
.95584 0 L
F
.9572 0 m
.9572 0 L
.96262 0 L
.96262 0 L
F
.96398 0 m
.96398 0 L
.96941 0 L
.96941 0 L
F
.97076 0 m
.97076 0 L
.97619 0 L
.97619 0 L
F
0 .4 1 r
.02788 0 m
.02788 0 L
.03331 0 L
.03331 0 L
F
.03466 0 m
.03466 0 L
.04009 0 L
.04009 0 L
F
.04145 0 m
.04145 0 L
.04687 0 L
.04687 0 L
F
.04823 .54167 m
.04823 1.16667 L
.05366 1.16667 L
.05366 .54167 L
F
.05501 21.8333 m
.05501 38.3333 L
.06044 38.3333 L
.06044 21.8333 L
F
.0618 204.833 m
.0618 295.917 L
.06722 295.917 L
.06722 204.833 L
F
.06858 794.625 m
.06858 1355.42 L
.07401 1355.42 L
.07401 794.625 L
F
.07536 627.417 m
.07536 1248.54 L
.08079 1248.54 L
.08079 627.417 L
F
.08215 159.583 m
.08215 506.667 L
.08757 506.667 L
.08757 159.583 L
F
.08893 20.125 m
.08893 375.583 L
.09436 375.583 L
.09436 20.125 L
F
.09571 9.625 m
.09571 275.75 L
.10114 275.75 L
.10114 9.625 L
F
.1025 1.66667 m
.1025 45 L
.10792 45 L
.10792 1.66667 L
F
.10928 .16667 m
.10928 14.4583 L
.11471 14.4583 L
.11471 .16667 L
F
.11606 .08333 m
.11606 5.04167 L
.12149 5.04167 L
.12149 .08333 L
F
.12285 .08333 m
.12285 2.33333 L
.12827 2.33333 L
.12827 .08333 L
F
.12963 0 m
.12963 1.20833 L
.13506 1.20833 L
.13506 0 L
F
.13641 0 m
.13641 .29167 L
.14184 .29167 L
.14184 0 L
F
.1432 0 m
.1432 .125 L
.14862 .125 L
.14862 0 L
F
.14998 0 m
.14998 0 L
.15541 0 L
.15541 0 L
F
.15676 0 m
.15676 .04167 L
.16219 .04167 L
.16219 0 L
F
.16355 0 m
.16355 0 L
.16897 0 L
.16897 0 L
F
.17033 0 m
.17033 .04167 L
.17576 .04167 L
.17576 0 L
F
.17711 0 m
.17711 0 L
.18254 0 L
.18254 0 L
F
.1839 0 m
.1839 0 L
.18932 0 L
.18932 0 L
F
.19068 0 m
.19068 0 L
.19611 0 L
.19611 0 L
F
.19746 0 m
.19746 .08333 L
.20289 .08333 L
.20289 0 L
F
.20425 0 m
.20425 .04167 L
.20967 .04167 L
.20967 0 L
F
.21103 0 m
.21103 .125 L
.21646 .125 L
.21646 0 L
F
.21781 0 m
.21781 .04167 L
.22324 .04167 L
.22324 0 L
F
.2246 0 m
.2246 .125 L
.23002 .125 L
.23002 0 L
F
.23138 0 m
.23138 .08333 L
.23681 .08333 L
.23681 0 L
F
.23816 0 m
.23816 .04167 L
.24359 .04167 L
.24359 0 L
F
.24495 0 m
.24495 .04167 L
.25037 .04167 L
.25037 0 L
F
.25173 0 m
.25173 .04167 L
.25716 .04167 L
.25716 0 L
F
.25851 0 m
.25851 0 L
.26394 0 L
.26394 0 L
F
.2653 0 m
.2653 0 L
.27072 0 L
.27072 0 L
F
.27208 0 m
.27208 0 L
.27751 0 L
.27751 0 L
F
.27886 0 m
.27886 0 L
.28429 0 L
.28429 0 L
F
.28565 0 m
.28565 0 L
.29107 0 L
.29107 0 L
F
.29243 0 m
.29243 0 L
.29786 0 L
.29786 0 L
F
.29921 0 m
.29921 0 L
.30464 0 L
.30464 0 L
F
.306 0 m
.306 .04167 L
.31142 .04167 L
.31142 0 L
F
.31278 0 m
.31278 0 L
.31821 0 L
.31821 0 L
F
.31956 0 m
.31956 0 L
.32499 0 L
.32499 0 L
F
.32635 0 m
.32635 0 L
.33177 0 L
.33177 0 L
F
.33313 0 m
.33313 0 L
.33856 0 L
.33856 0 L
F
.33991 0 m
.33991 0 L
.34534 0 L
.34534 0 L
F
.3467 0 m
.3467 0 L
.35212 0 L
.35212 0 L
F
.35348 0 m
.35348 0 L
.35891 0 L
.35891 0 L
F
.36026 0 m
.36026 0 L
.36569 0 L
.36569 0 L
F
.36705 0 m
.36705 0 L
.37247 0 L
.37247 0 L
F
.37383 0 m
.37383 0 L
.37926 0 L
.37926 0 L
F
.38061 0 m
.38061 0 L
.38604 0 L
.38604 0 L
F
.3874 0 m
.3874 0 L
.39282 0 L
.39282 0 L
F
.39418 0 m
.39418 0 L
.39961 0 L
.39961 0 L
F
.40096 0 m
.40096 0 L
.40639 0 L
.40639 0 L
F
.40775 0 m
.40775 0 L
.41317 0 L
.41317 0 L
F
.41453 0 m
.41453 0 L
.41996 0 L
.41996 0 L
F
.42131 0 m
.42131 0 L
.42674 0 L
.42674 0 L
F
.4281 0 m
.4281 0 L
.43352 0 L
.43352 0 L
F
.43488 0 m
.43488 0 L
.44031 0 L
.44031 0 L
F
.44166 0 m
.44166 0 L
.44709 0 L
.44709 0 L
F
.44845 0 m
.44845 0 L
.45387 0 L
.45387 0 L
F
.45523 0 m
.45523 0 L
.46066 0 L
.46066 0 L
F
.46201 0 m
.46201 0 L
.46744 0 L
.46744 0 L
F
.4688 0 m
.4688 0 L
.47422 0 L
.47422 0 L
F
.47558 0 m
.47558 0 L
.48101 0 L
.48101 0 L
F
.48236 0 m
.48236 0 L
.48779 0 L
.48779 0 L
F
.48915 0 m
.48915 0 L
.49457 0 L
.49457 0 L
F
.49593 0 m
.49593 0 L
.50136 0 L
.50136 0 L
F
.50271 0 m
.50271 0 L
.50814 0 L
.50814 0 L
F
.5095 0 m
.5095 0 L
.51492 0 L
.51492 0 L
F
.51628 0 m
.51628 0 L
.52171 0 L
.52171 0 L
F
.52306 0 m
.52306 0 L
.52849 0 L
.52849 0 L
F
.52985 0 m
.52985 0 L
.53527 0 L
.53527 0 L
F
.53663 0 m
.53663 0 L
.54206 0 L
.54206 0 L
F
.54341 0 m
.54341 0 L
.54884 0 L
.54884 0 L
F
.5502 0 m
.5502 0 L
.55562 0 L
.55562 0 L
F
.55698 0 m
.55698 0 L
.56241 0 L
.56241 0 L
F
.56376 0 m
.56376 0 L
.56919 0 L
.56919 0 L
F
.57055 0 m
.57055 0 L
.57597 0 L
.57597 0 L
F
.57733 0 m
.57733 0 L
.58276 0 L
.58276 0 L
F
.58411 0 m
.58411 0 L
.58954 0 L
.58954 0 L
F
.5909 0 m
.5909 0 L
.59632 0 L
.59632 0 L
F
.59768 0 m
.59768 0 L
.60311 0 L
.60311 0 L
F
.60446 0 m
.60446 0 L
.60989 0 L
.60989 0 L
F
.61125 0 m
.61125 0 L
.61667 0 L
.61667 0 L
F
.61803 0 m
.61803 .04167 L
.62346 .04167 L
.62346 0 L
F
.62481 0 m
.62481 0 L
.63024 0 L
.63024 0 L
F
.6316 0 m
.6316 0 L
.63702 0 L
.63702 0 L
F
.63838 0 m
.63838 0 L
.64381 0 L
.64381 0 L
F
.64516 0 m
.64516 0 L
.65059 0 L
.65059 0 L
F
.65195 0 m
.65195 0 L
.65737 0 L
.65737 0 L
F
.65873 0 m
.65873 0 L
.66416 0 L
.66416 0 L
F
.66551 0 m
.66551 0 L
.67094 0 L
.67094 0 L
F
.6723 0 m
.6723 0 L
.67772 0 L
.67772 0 L
F
.67908 0 m
.67908 0 L
.68451 0 L
.68451 0 L
F
.68586 0 m
.68586 0 L
.69129 0 L
.69129 0 L
F
.69265 0 m
.69265 0 L
.69807 0 L
.69807 0 L
F
.69943 0 m
.69943 0 L
.70486 0 L
.70486 0 L
F
.70621 0 m
.70621 0 L
.71164 0 L
.71164 0 L
F
.713 0 m
.713 0 L
.71842 0 L
.71842 0 L
F
.71978 0 m
.71978 0 L
.72521 0 L
.72521 0 L
F
.72656 0 m
.72656 0 L
.73199 0 L
.73199 0 L
F
.73335 0 m
.73335 0 L
.73877 0 L
.73877 0 L
F
.74013 0 m
.74013 0 L
.74556 0 L
.74556 0 L
F
.74691 0 m
.74691 0 L
.75234 0 L
.75234 0 L
F
.7537 0 m
.7537 0 L
.75912 0 L
.75912 0 L
F
.76048 0 m
.76048 0 L
.76591 0 L
.76591 0 L
F
.76726 0 m
.76726 0 L
.77269 0 L
.77269 0 L
F
.77405 0 m
.77405 0 L
.77947 0 L
.77947 0 L
F
.78083 0 m
.78083 0 L
.78626 0 L
.78626 0 L
F
.78761 0 m
.78761 0 L
.79304 0 L
.79304 0 L
F
.7944 0 m
.7944 0 L
.79982 0 L
.79982 0 L
F
.80118 0 m
.80118 0 L
.80661 0 L
.80661 0 L
F
.80796 0 m
.80796 0 L
.81339 0 L
.81339 0 L
F
.81475 0 m
.81475 0 L
.82017 0 L
.82017 0 L
F
.82153 0 m
.82153 0 L
.82696 0 L
.82696 0 L
F
.82831 0 m
.82831 0 L
.83374 0 L
.83374 0 L
F
.8351 0 m
.8351 0 L
.84052 0 L
.84052 0 L
F
.84188 0 m
.84188 0 L
.84731 0 L
.84731 0 L
F
.84866 0 m
.84866 0 L
.85409 0 L
.85409 0 L
F
.85545 0 m
.85545 0 L
.86087 0 L
.86087 0 L
F
.86223 0 m
.86223 0 L
.86766 0 L
.86766 0 L
F
.86901 0 m
.86901 .04167 L
.87444 .04167 L
.87444 0 L
F
.8758 0 m
.8758 0 L
.88122 0 L
.88122 0 L
F
.88258 0 m
.88258 0 L
.88801 0 L
.88801 0 L
F
.88936 0 m
.88936 0 L
.89479 0 L
.89479 0 L
F
.89615 0 m
.89615 0 L
.90157 0 L
.90157 0 L
F
.90293 0 m
.90293 0 L
.90836 0 L
.90836 0 L
F
.90971 0 m
.90971 0 L
.91514 0 L
.91514 0 L
F
.9165 0 m
.9165 0 L
.92192 0 L
.92192 0 L
F
.92328 0 m
.92328 0 L
.92871 0 L
.92871 0 L
F
.93006 0 m
.93006 0 L
.93549 0 L
.93549 0 L
F
.93685 0 m
.93685 0 L
.94227 0 L
.94227 0 L
F
.94363 0 m
.94363 0 L
.94906 0 L
.94906 0 L
F
.95041 0 m
.95041 0 L
.95584 0 L
.95584 0 L
F
.9572 0 m
.9572 0 L
.96262 0 L
.96262 0 L
F
.96398 0 m
.96398 0 L
.96941 0 L
.96941 0 L
F
.97076 0 m
.97076 0 L
.97619 0 L
.97619 0 L
F
MathSubEnd
P
% End of sub-graphic
0 g
gsave
.24286 .51503 -71.7188 -10.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20.125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.562 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
75.000 9.312 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 7.125 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
80.438 13.562 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.24286 .46794 -86.7188 -10.25 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20.5 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def
63.000 13.938 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 13.938 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Riemann) show
105.000 9.312 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 7.125 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(7) show
110.438 13.938 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.24286 .40908 -84 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3.2) show
87.000 11.250 moveto
(GHz) show
105.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.24286 .27959 -93 -9.0625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 18.125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(bin) show
87.000 11.562 moveto
(.005) show
117.000 11.562 moveto
(s) show
123.000 11.562 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
1 0 0 r
.59524 .27959 m
.59524 .30902 L
.64286 .30902 L
.64286 .27959 L
F
0 g
gsave
.75714 .29136 -96 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Zero) show
87.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(:) show
99.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(44174) show
129.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .4 1 r
.59524 .22073 m
.59524 .25016 L
.64286 .25016 L
.64286 .22073 L
F
0 g
gsave
.78571 .2325 -105 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 17.8125 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Nonzero) show
105.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(:) show
117.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(55826) show
147.000 11.250 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
 ImageSize->{480.625, 296.5},
 ImageMargins->{{0, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 
  1}},ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>", "ImageResolution" -> 72.],ImageRangeCache->{{{0., 1079.16}, {664.875, \
0.}} -> {-0.0266444, -0.0250052, 0.00219607, 0.00355331}, {{27.2812, \
1051.74}, {648.986, 15.75}} -> {-1.79914, -5246.35, 0.0486418, 134.413}, \
{{508.782, 1021.08}, {332.438, 15.75}} -> {-169.034, -21.6282, 0.693196, \
0.112852}},ExpressionUUID->"097fbb40-4e2b-4ac0-8018-f4c5230929ca"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up", "MathCaption",ExpressionUUID->"df2e193a-b645-4761-965f-3cbaf77bca98"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"UndefTensor", "[", "R", "]"}], "\[IndentingNewLine]", 
 RowBox[{"Remove", "[", 
  RowBox[{
  "canonindices", ",", "productR", ",", "canR", ",", "data1", ",", "data2", 
   ",", "data3", ",", "plot0", ",", "plot1", ",", "plot2"}], "]"}]}], "Input",\

 CellLabel->
  "In[374]:=",ExpressionUUID->"a6634d38-a5ef-415d-89da-995b757baeb5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "R"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`R],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978004764112*^9},
 CellLabel->"During evaluation of In[374]:=",
 CellID->445966567,ExpressionUUID->"4d87c0fb-6d77-4f7e-b652-9499cf93206c"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"MetricsOfVBundle", "[", "TangentM3", "]"}], "^=", 
   RowBox[{"{", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Clear", "[", "metricg", "]"}], ";"}]}], "Input",
 CellLabel->
  "In[376]:=",ExpressionUUID->"89ae8c23-6009-426b-94e6-8d078e605141"]
}, Open  ]],

Cell[TextData[{
 "Summarizing, the general canonicalization algorithms of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " are highly efficient in most cases, much faster than the algorithms used \
by other tensor computer algebra systems. However, from time to time you will \
find special cases which take several orders of magnitude longer to \
canonicalize than other similar cases (for example a simple reordering of the \
indices keeping the same tensors), usually also swallowing up a large chunk \
of memory in your computer. This is normal behaviour, and it is due to the \
fact that the Group Intersection algorithm unavoidably behaves like that. \
Fortunately the special cases are always that: special; if you understand the \
meaning of \"special\" in your problem, you can prepare the system in advance \
to switch to \"special\" alternative algorithms.\nIn hindsight, this is the \
price we pay for using a general algorithm, but for the case of permutation \
symmetries it is worth paying it, because only in a very small fraction of \
cases the algorithm behaves exponentially, instead of polynomially. The case \
with multiterm symmetries is precisely the opposite: all known algorithms \
behave exponentially in most cases. "
}], "Text",ExpressionUUID->"2feb7a0f-9e2d-4102-ae8a-967ca9c9a8c6"]
}, Closed]],

Cell[CellGroupData[{

Cell["4.4. Canonicalization in a 1-dimensional space", "Subsection",ExpressionUUID->"c977018a-28ed-40e4-b06f-6c06e25dd990"],

Cell[TextData[{
 StyleBox["xTensor`", "MR"],
 " deals with 1-dimensional spaces in exactly the same way as with any other \
space. In particular, indices cannot be repeated, even though there is just \
one indiependent direction. This is solved by imposing full symmetry at \
canonicalization time."
}], "Text",ExpressionUUID->"34d74bf7-9c7e-40e5-89b5-1f2a7ed22854"],

Cell[CellGroupData[{

Cell["Define a 1-dimensional space:", "MathCaption",ExpressionUUID->"06f691bd-a442-485e-9949-ed7ad0041da2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M1", ",", "1", ",", 
   RowBox[{"{", 
    RowBox[{"a1", ",", "b1", ",", "c1", ",", "d1"}], "}"}]}], "]"}]], "Input",\

 CellLabel->
  "In[378]:=",ExpressionUUID->"2bba5c34-b32d-4dfd-8d41-0e461203103f"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M1", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M1, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780049898653`*^9},
 CellLabel->"During evaluation of In[378]:=",
 CellID->642163539,ExpressionUUID->"b1ca3830-467c-4457-9eaf-a0a2de5ce68b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM1", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM1, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780050056343`*^9},
 CellLabel->"During evaluation of In[378]:=",
 CellID->1737273798,ExpressionUUID->"1d4b9514-4e3c-47ed-a1cb-3372705937c4"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and some tensors in this space :", "MathCaption",ExpressionUUID->"2bbcdbf3-c4fb-41b7-ad75-d3c1f3a9c62c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"a1", ",", "b1"}], "]"}], ",", "M1"}], "]"}]], "Input",
 CellLabel->
  "In[379]:=",ExpressionUUID->"d5502234-febe-4e8b-a8ba-5e060a99311c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"W", "[", 
    RowBox[{"a1", ",", "b1"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`W[$CellContext`a1, $CellContext`b1], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978005173623*^9},
 CellLabel->"During evaluation of In[379]:=",
 CellID->1173058374,ExpressionUUID->"29c166f6-3fd3-4fc9-af74-e716d8a27966"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"R", "[", 
    RowBox[{"a1", ",", "b1"}], "]"}], ",", "M1"}], "]"}]], "Input",
 CellLabel->
  "In[380]:=",ExpressionUUID->"7be6047a-c266-4b2e-a56b-15a517a5b8cf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"R", "[", 
    RowBox[{"a1", ",", "b1"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`R[$CellContext`a1, $CellContext`b1], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780054330797`*^9},
 CellLabel->"During evaluation of In[380]:=",
 CellID->1692212358,ExpressionUUID->"ad0c209b-f0ed-48bc-abf7-cc01b0fe0d8e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now something like this is zero :", "MathCaption",ExpressionUUID->"98ad7da8-2749-4bfb-9c10-1f5cce99ddf5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"W", "[", 
     RowBox[{"a1", ",", "b1"}], "]"}], 
    RowBox[{"R", "[", 
     RowBox[{"c1", ",", "d1"}], "]"}]}], "-", 
   RowBox[{
    RowBox[{"W", "[", 
     RowBox[{"d1", ",", "a1"}], "]"}], 
    RowBox[{"R", "[", 
     RowBox[{"c1", ",", "b1"}], "]"}]}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[381]:=",ExpressionUUID->"db6674db-fd5b-4695-9699-b0df15d64639"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289780059344263`*^9},
 CellLabel->"Out[381]=",
 CellID->1887163685,ExpressionUUID->"13dd1644-fc28-4064-985a-e38e92609559"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up :", "MathCaption",ExpressionUUID->"0bfe26ac-834f-4dac-8bb3-df962330ee2d"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"W", ",", "R"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"UndefManifold", "[", "M1", "]"}]}], "Input",
 CellLabel->
  "In[382]:=",ExpressionUUID->"c5a38fa5-edcc-4fa6-a2e8-e2da5b848dce"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780063240232`*^9},
 CellLabel->"During evaluation of In[382]:=",
 CellID->1374071252,ExpressionUUID->"271256fa-f5c5-4b3d-bcba-2e58bd04eab1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "R"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`R],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780063412733`*^9},
 CellLabel->"During evaluation of In[382]:=",
 CellID->2123876173,ExpressionUUID->"4ef216bb-f455-4714-82b6-10aa3d15e503"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M1",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM1,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM1],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978006417029*^9},
 CellLabel->"During evaluation of In[382]:=",
 CellID->513659177,ExpressionUUID->"4a7d502f-2ad1-4104-9326-711e0429daf3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M1"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M1],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897800643714*^9},
 CellLabel->"During evaluation of In[382]:=",
 CellID->185526899,ExpressionUUID->"8fcb35c6-3329-46b4-aa71-80bbb31edad9"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5. Rules and patterns", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1736504676,ExpressionUUID->"b37ffd4e-e1c9-4992-9806-1ad019ce6449"],

Cell[CellGroupData[{

Cell["5.1. Unique dummies and screening", "Subsection",ExpressionUUID->"fb65e018-3f67-4957-966e-50fbeeb77714"],

Cell[TextData[{
 "One of the most important issues in a tensor packages is how to deal with \
index contractions, i.e. with dummy indices. Until now this was not a real \
issue for us because all input expressions were simply canonicalized, a \
process which does not introduce new indices, but rather, eliminates indices. \
Now we face the opposite problem: expanding expressions into new expressions \
with more dummy indices, for example the conversion of a covariant derivative \
of a tensor in terms of a different covariant derivative. There are three \
problems with conversions that we must address.\nThis subsection explains in \
detail what those three problems are and how they will be solved. Though the \
explanations might seem a bit technical, it is important to understand them, \
because they are based on core concepts of ",
 StyleBox["Mathematica.",
  FontSlant->"Italic"],
 "\nAs an aside, there are two ways to define conversions of expressions into \
new expressions in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ": assignments (",
 StyleBox["Set",
  FontFamily->"Courier"],
 " and its variants ",
 StyleBox["SetDelayed",
  FontFamily->"Courier"],
 ", ",
 StyleBox["TagSet",
  FontFamily->"Courier"],
 ", ",
 StyleBox["UpSet",
  FontFamily->"Courier"],
 ", etc.) and rules (",
 StyleBox["Rule",
  FontFamily->"Courier"],
 " and its variant ",
 StyleBox["RuleDelayed",
  FontFamily->"Courier"],
 "). Asignments can be considered as global or automatic rules. Rules can be \
considered as local assignments. See the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " Help.\nFirst problem. Let us see a simple example using ",
 StyleBox["Set",
  FontFamily->"Courier"],
 ":"
}], "Text",ExpressionUUID->"d3dff436-06bc-43bb-9675-5625baf754cc"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define the vector ",
 StyleBox["w",
  FontFamily->"Courier"],
 " in terms of ",
 StyleBox["S",
  FontFamily->"Courier"],
 " and ",
 StyleBox["v",
  FontFamily->"Courier"],
 ". Introduce a simple dummy:"
}], "MathCaption",ExpressionUUID->"4268c75a-8c25-45ed-99be-a9155307e8cd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[384]:=",ExpressionUUID->"316b5dc0-ab8b-4957-a7ac-152463176573"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"w", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`w[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978014950527*^9},
 CellLabel->"During evaluation of In[384]:=",
 CellID->632950163,ExpressionUUID->"c21c3d7f-7e15-4067-8713-fff2b3164065"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_", "]"}], "=", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[385]:=",ExpressionUUID->"9abd9aeb-f4b2-428d-b9a9-740dbdfa40a4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978015122283*^9},
 CellLabel->"Out[385]=",
 CellID->1506885950,ExpressionUUID->"d68162c6-b4ac-4593-bb8b-3ced9c1b739d"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "[Note that we have used a Pattern on the right hand side. That is normal ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", and a helpful feature which allows us to indicate the generality of the \
rule (which indices or tensors are accepted in the rule)]."
}], "MathCaption",ExpressionUUID->"b661ed81-f866-4d07-af16-f0b3bb9fe1a8"],

Cell[CellGroupData[{

Cell["Then these expressions would be totally wrong:", "MathCaption",ExpressionUUID->"47f18ded-ae51-42db-b892-cfda4b690003"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", "b", "]"}]], "Input",
 CellLabel->
  "In[386]:=",ExpressionUUID->"ce59fe18-18e3-4ee7-8d45-1512d9558661"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"b", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`b, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289780152513313`*^9},
 CellLabel->"Out[386]=",
 CellID->957512579,ExpressionUUID->"f58a3058-a108-422d-9741-a4c04eda09db"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a", "]"}], 
  RowBox[{"w", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellLabel->
  "In[387]:=",ExpressionUUID->"57b774f9-7a95-4a7a-ab67-6ae661dd6224"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "b"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.728978015363159*^9},
 CellLabel->"Out[387]=",
 CellID->1818948282,ExpressionUUID->"137cd222-41fc-4bd6-9c8c-e00a21c1c6de"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " recommends the following solution to this problem:"
}], "Text",ExpressionUUID->"54915ffb-f592-4c93-b31a-49de1c339cf6"],

Cell[CellGroupData[{

Cell[TextData[{
 "Use a ",
 StyleBox["Module",
  FontFamily->"Courier"],
 " construct, such that the dummy index b is replaced by a unique symbol \
denoted with a $ sign, and a unique integer number. Note that we must now use \
",
 StyleBox["SetDelayed",
  FontFamily->"Courier"],
 " (:=), and not ",
 StyleBox["Set",
  FontFamily->"Courier"],
 " (=), to avoid inmediate evaluation of ",
 StyleBox["Module",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"f68b2a59-0964-430b-ad08-991896e2aed4"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "b", "}"}], ",", 
    RowBox[{
     RowBox[{"S", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}]}]}], "]"}]}]], "Input",
 CellLabel->
  "In[388]:=",ExpressionUUID->"ba402cdd-c2b4-4f73-ab23-f4d8b3f802ce"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", "b", "]"}]], "Input",
 CellLabel->
  "In[389]:=",ExpressionUUID->"193a66c5-7a1d-4706-95d0-ed64d757477a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"b", "b$20151"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`b, $CellContext`b$20151],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b$20151"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b$20151],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289780156344852`*^9},
 CellLabel->"Out[389]=",
 CellID->1634060529,ExpressionUUID->"0761ba47-d10e-4a2c-93e7-ff9f7d725a1f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a", "]"}], 
  RowBox[{"w", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellLabel->
  "In[390]:=",ExpressionUUID->"74ae534d-16b5-4a26-bdd8-47b763bc3aea"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "b$20153"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`a, $CellContext`b$20153],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b$20152"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b$20152],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b$20152"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b$20152],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b$20153"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b$20153],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978015867324*^9},
 CellLabel->"Out[390]=",
 CellID->1634433067,ExpressionUUID->"8336e0ed-c44c-4c6f-8dbd-b5cbdaafa257"]
}, Open  ]]
}, Open  ]],

Cell["\<\
This solves the first problem but generates two more, which are solved by \
\"screening\" the dollar-indices:\
\>", "Text",ExpressionUUID->"5f8ba793-a9fd-41dc-880c-15c702c03305"],

Cell[TextData[{
 StyleBox[ButtonBox["ScreenDollarIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ScreenDollarIndices.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Hide internal unique dummy indices", "Text"]
}], "DefinitionBox",ExpressionUUID->"042bc6a1-bf2e-4c9d-8de0-4cd372fdcac4"],

Cell["Screening of indices.", "Text",ExpressionUUID->"12763ed4-61d4-4e40-964b-73a23bc64c99"],

Cell[CellGroupData[{

Cell["\<\
First, each time we use the assignment we get a different object, what \
complicates comparisons. The following is not automatically recognized as \
zero!\
\>", "MathCaption",ExpressionUUID->"ecd54cd4-8210-4993-9546-26d11d4523c4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a", "]"}], "-", 
  RowBox[{"w", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[391]:=",ExpressionUUID->"5f17f755-8b65-4164-8cd0-6eda88d8ab6c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b$20154"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`b$20154],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b$20154"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b$20154],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b$20155"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`b$20155],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b$20155"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b$20155],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978015970345*^9},
 CellLabel->"Out[391]=",
 CellID->1453228292,ExpressionUUID->"e9dc0fa9-00a3-4051-8267-13675bf59338"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[392]:=",ExpressionUUID->"4fdf5386-509d-4588-8539-e9c44c873070"],

Cell["S[a, b$20154]*v[-b$20154] - S[a, b$20155]*v[-b$20155]", "Output",
 CellChangeTimes->{3.728978016117714*^9},
 CellLabel->"Out[392]//InputForm=",
 CellID->1076386313,ExpressionUUID->"e9b28141-bec0-4c75-b94d-49b74d88223a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The second problem is purely aesthetical: the dollar-indices are terribly \
ugly. ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has the function ",
 StyleBox["ScreenDollarIndices",
  FontFamily->"Courier"],
 " to convert the dollar-indices into normal-looking indices:"
}], "MathCaption",ExpressionUUID->"a374b46d-75f8-4456-85b5-2a6596e2770a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", "b", "]"}], ",", 
   RowBox[{"w", "[", "c", "]"}], ",", 
   RowBox[{
    RowBox[{"w", "[", "a", "]"}], 
    RowBox[{"w", "[", 
     RowBox[{"-", "a"}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"w", "[", "a", "]"}], "-", 
    RowBox[{"w", "[", "a", "]"}]}]}], "}"}]], "Input",
 CellLabel->
  "In[393]:=",ExpressionUUID->"88818591-43dd-4353-bd1d-ac8debaa3688"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", "b$20157"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, $CellContext`b$20157],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b$20157"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b$20157],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"b", "b$20158"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`b, $CellContext`b$20158],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b$20158"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b$20158],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "b$20159"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`b$20159],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b$20159"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b$20159],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", "b$20161"},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[-$CellContext`a, $CellContext`b$20161],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", "b$20160"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, $CellContext`b$20160],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b$20160"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b$20160],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b$20161"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b$20161],
     Editable->False]}], ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"a", "b$20162"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[$CellContext`a, $CellContext`b$20162],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b$20162"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`b$20162],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"a", "b$20163"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[$CellContext`a, $CellContext`b$20163],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b$20163"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`b$20163],
      Editable->False]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978016261725*^9},
 CellLabel->"Out[393]=",
 CellID->422520263,ExpressionUUID->"bc225fae-2bf5-4f18-82e4-0e1505e35232"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ScreenDollarIndices", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[394]:=",ExpressionUUID->"9482811c-7de7-4619-aed0-5ff2ccf620bd"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"b", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`b, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", "c"},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[-$CellContext`a, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`c],
     Editable->False]}], ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728978016376299*^9},
 CellLabel->"Out[394]=",
 CellID->1717123259,ExpressionUUID->"941e23b8-88b7-453f-ba7e-d58466c00d64"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[395]:=",ExpressionUUID->"e07eccbc-856e-4b23-891f-b899df7056f1"],

Cell["\<\
{S[a, b]*v[-b], S[b, a]*v[-a], S[c, a]*v[-a], S[-a, c]*S[a, b]*v[-b]*v[-c], 0}\
\>", "Output",
 CellChangeTimes->{3.7289780165130243`*^9},
 CellLabel->"Out[395]//InputForm=",
 CellID->1805365627,ExpressionUUID->"89aa5028-5d27-46f5-aaa4-12292c9d07ad"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The ",
 StyleBox["Module",
  FontFamily->"Courier"],
 " construct is not the solution for all problems in tensor conversions. \
There are two more problems, as seen in the following examples. Second \
problem:"
}], "Text",ExpressionUUID->"c2286cba-dc17-46e2-a31c-8c6c8b4801ef"],

Cell[CellGroupData[{

Cell[TextData[{
 "Objects and indices cannot share the same symbol names. Suppose there were \
a tensor also called ",
 StyleBox["a",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"59f3ef4e-2783-46eb-b5a1-a33421c67f21"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_", "]"}], ":=", 
  RowBox[{"a", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[396]:=",ExpressionUUID->"2cca7070-0258-4ce4-a5cf-116d910740e3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", "b", "]"}]], "Input",
 CellLabel->
  "In[397]:=",ExpressionUUID->"f3d5ccc8-b95b-46a5-b960-19b5379e86b7"],

Cell[BoxData[
 RowBox[{"b", "[", "b", "]"}]], "Output",
 CellChangeTimes->{3.7289780167090397`*^9},
 CellLabel->"Out[397]=",
 CellID->983644526,ExpressionUUID->"6cb68dd5-61d7-41ac-8dd4-3a9bd8487d67"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", this is solved by enforcing type declarations. Every object must be \
defined supplying a symbol, and that symbol is given a type. A symbol cannot \
have two different types. This can be seen as a restriction, but it is \
actually a useful safety feature."
}], "Text",ExpressionUUID->"1781e635-fdf9-4e11-a44a-e6e9091bde9d"],

Cell[CellGroupData[{

Cell[TextData[{
 "The tensor ",
 StyleBox["a[b]",
  FontFamily->"Courier"],
 " cannot be defined:"
}], "MathCaption",ExpressionUUID->"f669974f-c03d-43b5-a7db-a69d54dbd525"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"a", "[", "b", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[398]:=",ExpressionUUID->"5e7fc455-89dd-411e-a9af-8a6c5d3619ea"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "used", 
   "\"Symbol \\!\\(\\*RowBox[{\\\"\\\\\\\"a\\\\\\\"\\\"}]\\) is already used \
\\!\\(\\*RowBox[{\\\"\\\\\\\"as an abstract index\\\\\\\"\\\"}]\\).\"", 2, 
   398, 53, 19253573948027831207, "Local", "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978016876*^9},
 CellLabel->"During evaluation of In[398]:=",
 CellID->784433876,ExpressionUUID->"42189f44-be8a-4a3e-9b41-b9c504d6f746"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The final (third) problem involves dummies again, and happens in the few \
cases in which the same symbol represents a dummy on both sides and there are \
additional dummies on the right. The only possible solution is changing the \
name of the conflicting dummies either on the left or on the right hand side.\
\
\>", "Text",ExpressionUUID->"8166847b-0ae7-4bd8-8920-7609679d48f7"],

Cell[CellGroupData[{

Cell[TextData[{
 "Define the following relation with dummies both on the left and right hand \
sides. Note that the b dummy is present in both sides and is a pattern on the \
left hand side (",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " already points out the error by reddening the dummy):"
}], "MathCaption",ExpressionUUID->"07154e53-e5a9-41ee-9dc7-39cc51a2f82e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", ",", "b_", ",", 
    RowBox[{"-", "b_"}]}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"b", ",", "c"}], "}"}], ",", 
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b", ",", 
      RowBox[{"-", "b"}], ",", "c", ",", 
      RowBox[{"-", "c"}]}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[399]:=",ExpressionUUID->"f8804b8d-6d45-4e63-93e6-808a823c8d9c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", 
  RowBox[{"a", ",", "b", ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->
  "In[400]:=",ExpressionUUID->"9aaa997e-f6fa-44c8-b127-03409e217cee"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "b$20178", " ", "c$20178", " "},
         {" ", " ", "b$20178", " ", "c$20178"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`b$20178, -$CellContext`b$20178, \
$CellContext`c$20178, -$CellContext`c$20178],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978017126254*^9},
 CellLabel->"Out[400]=",
 CellID->578044999,ExpressionUUID->"cb19883c-8afc-4def-a9e6-7d4392fd0e22"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", 
  RowBox[{"a", ",", "c", ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[401]:=",ExpressionUUID->"6aa48fff-8363-47f9-b753-53fb5f4f8278"],

Cell[BoxData[
 TemplateBox[{
  "Module", "dups", 
   "\"Conflicting local variables \\!\\(\\*RowBox[{\\\"c\\\"}]\\) and \
\\!\\(\\*RowBox[{\\\"c$\\\"}]\\) found in local variable specification \\!\\(\
\\*RowBox[{\\\"{\\\", RowBox[{\\\"c\\\", \\\",\\\", \\\"c$\\\"}], \\\"}\\\"}]\
\\).\"", 2, 401, 54, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728978017292831*^9},
 CellLabel->"During evaluation of In[401]:=",
 CellID->2133762812,ExpressionUUID->"881a1b66-d3e4-4ccf-bd7b-412100fec1ac"],

Cell[BoxData[
 RowBox[{"Module", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"c", ",", "c$"}], "}"}], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "c", " ", "c$", " "},
           {" ", " ", "c", " ", "c$"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`c, -$CellContext`c, \
$CellContext`c$, -$CellContext`c$],
    Editable->False]}], "]"}]], "Output",
 CellChangeTimes->{3.7289780173240833`*^9},
 CellLabel->"Out[401]=",
 CellID->1322160031,ExpressionUUID->"2f41f3ac-d0e4-4ee8-b917-17d77627840f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "For example, we can change the name of the dummy index ",
 StyleBox["b",
  FontFamily->"Courier"],
 " on the right hand side:"
}], "MathCaption",ExpressionUUID->"f6cd7ef1-3732-4114-9c6c-af95b3e95eb2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", ",", "b_", ",", 
    RowBox[{"-", "b_"}]}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"d", ",", "c"}], "}"}], ",", 
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "d", ",", 
      RowBox[{"-", "d"}], ",", "c", ",", 
      RowBox[{"-", "c"}]}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[402]:=",ExpressionUUID->"283f45b1-7dd1-4fee-be9c-27b25af6fd4c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", 
  RowBox[{"a", ",", "b", ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->
  "In[403]:=",ExpressionUUID->"2be57b45-87a4-418b-8eea-a535109da820"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "d$20185", " ", "c$20185", " "},
         {" ", " ", "d$20185", " ", "c$20185"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`d$20185, -$CellContext`d$20185, \
$CellContext`c$20185, -$CellContext`c$20185],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289780175257397`*^9},
 CellLabel->"Out[403]=",
 CellID->1530345508,ExpressionUUID->"3060b369-ee4e-4dfd-8c85-bb3470c233d3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", 
  RowBox[{"a", ",", "c", ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[404]:=",ExpressionUUID->"ef53ecb2-734e-45eb-b57e-3014a5105797"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "d$20186", " ", "c$20186", " "},
         {" ", " ", "d$20186", " ", "c$20186"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`d$20186, -$CellContext`d$20186, \
$CellContext`c$20186, -$CellContext`c$20186],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978017676153*^9},
 CellLabel->"Out[404]=",
 CellID->105312837,ExpressionUUID->"8bbe5dc6-f268-4bbf-ba46-37916afb122d"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "As we said, the second problem is solved in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " simply by enforcing type declarations. The first and third problems will \
be solved introducing a new collection of ",
 StyleBox["Index*",
  FontFamily->"Courier"],
 " assignment and rule functions which properly prepare the required ",
 StyleBox["Module",
  FontFamily->"Courier"],
 " constructs. They are ",
 StyleBox["IndexSet",
  FontFamily->"Courier"],
 ", ",
 StyleBox["IndexSetDelayed",
  FontFamily->"Courier"],
 ", ",
 StyleBox["IndexRule",
  FontFamily->"Courier"],
 " and ",
 StyleBox["IndexRuleDelayed",
  FontFamily->"Courier"],
 ", imitating their respective ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " counterparts, and they are described in detail in the next subsections."
}], "Text",ExpressionUUID->"3696f86d-fcd6-4ee8-8f49-abf82beff2f2"],

Cell[TextData[{
 "We finish this section coming back to the function ",
 StyleBox["ScreenDollarIndices",
  FontFamily->"Courier"],
 ". We recommend to automate its use by assigning one of the global variables \
",
 StyleBox["$Post",
  FontFamily->"Courier"],
 " or ",
 StyleBox["$PrePrint",
  FontFamily->"Courier"],
 " to it. There is a small difference between these two global variables \
(mentioned below), and it is not clear that one of them is always better than \
the other for screening, but I generally prefer ",
 StyleBox["$PrePrint",
  FontFamily->"Courier"],
 ":"
}], "Text",ExpressionUUID->"83c9e87e-91d2-4781-8bfc-d100ea675342"],

Cell[CellGroupData[{

Cell[TextData[{
 "Assign ",
 StyleBox["$PrePrint",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"5b81429c-a91e-4cf9-a82c-014270ce2d3e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellLabel->
  "In[405]:=",ExpressionUUID->"d2303095-f81e-42d7-850f-6e57cb8f7307"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now the screening is automatic:", "MathCaption",ExpressionUUID->"109db051-9e08-45c3-88b5-d572cb5db11f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", 
  RowBox[{"a", ",", "c", ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[406]:=",ExpressionUUID->"b9976076-3dd2-43f0-b068-51c3b1f5640d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "c", " ", "b", " "},
         {" ", " ", "c", " ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`c, -$CellContext`c, \
$CellContext`b, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978017942565*^9},
 CellLabel->"Out[406]=",
 CellID->466326331,ExpressionUUID->"b69e2613-b865-4c03-a329-ea36e7105cbb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The screening happened after the result was assigned to ",
 StyleBox["Out[...]",
  FontFamily->"Courier"],
 " and hence if we temporarily remove ",
 StyleBox["$PrePrint",
  FontFamily->"Courier"],
 ", we still see the dollar-indices. (That wouldn't happen using ",
 StyleBox["$Post",
  FontFamily->"Courier"],
 ".)"
}], "MathCaption",ExpressionUUID->"d2676e1e-ce9e-4746-908b-c4ca89c62a16"],

Cell[BoxData[
 RowBox[{"$PrePrint", "=."}]], "Input",
 CellLabel->
  "In[407]:=",ExpressionUUID->"9ffac273-6ff1-45ef-b9e0-4bf1ea6891d0"],

Cell[CellGroupData[{

Cell[BoxData["%%"], "Input",
 CellLabel->
  "In[408]:=",ExpressionUUID->"7ebfd0d9-92e8-4cb4-98f7-17e9dbefe819"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "d$20187", " ", "c$20187", " "},
         {" ", " ", "d$20187", " ", "c$20187"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`d$20187, -$CellContext`d$20187, \
$CellContext`c$20187, -$CellContext`c$20187],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978018133378*^9},
 CellLabel->"Out[408]=",
 CellID->1378817214,ExpressionUUID->"c71784b1-493c-48b8-9902-4e328bfa052f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up", "MathCaption",ExpressionUUID->"0d7dfc4e-6a8e-4a1f-b394-14aa36497f8d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", ",", "b_", ",", 
    RowBox[{"-", "b_"}]}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[409]:=",ExpressionUUID->"89f031cf-812d-4a3f-a905-aa87c9622072"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
From now on, in this notebook, we shall use automatic screening:\
\>", "MathCaption",ExpressionUUID->"3f1c1679-755a-404e-8b12-135eabf41364"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellLabel->
  "In[410]:=",ExpressionUUID->"945e2f72-7998-4ed2-b110-7984d0664f54"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["5.2. IndexSet and IndexSetDelayed", "Subsection",ExpressionUUID->"6d9737a9-7f5f-44c4-a03b-7edbf4a7b3a2"],

Cell[TextData[{
 "The ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " functions ",
 StyleBox["IndexSet",
  FontFamily->"Courier"],
 " and ",
 StyleBox["IndexSetDelayed",
  FontFamily->"Courier"],
 " imitate the behaviour of ",
 StyleBox["Set",
  FontFamily->"Courier"],
 " and ",
 StyleBox["SetDelayed",
  FontFamily->"Courier"],
 " (which delays the evaluation of the rhs to the moment when the lhs is \
converted into the rhs)."
}], "Text",ExpressionUUID->"7b2c5242-bfb1-4dca-ba70-43a75fc98651"],

Cell[TextData[{
 StyleBox[ButtonBox["IndexSet",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexSet.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Set value of an indexed expression, at definition time\n", 
  "Text"],
 StyleBox[ButtonBox["IndexSetDelayed",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexSetDelayed.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Set value of an indexed expression, at evaluation time", "Text"]
}], "DefinitionBox",ExpressionUUID->"4327af69-e03a-4d07-8889-188973731e98"],

Cell["Set functions for indexed expressions.", "Text",ExpressionUUID->"11094ee7-0193-4846-bfb6-a34d1100e80c"],

Cell[CellGroupData[{

Cell[TextData[{
 "Example of use of ",
 StyleBox["IndexSet",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"e7b3aa73-c6fd-4b70-b1a0-112a9a71a331"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"w", "[", "a_", "]"}], ",", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[411]:=",ExpressionUUID->"37855b90-c630-4c23-8c40-b484a6440f23"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978028434578*^9},
 CellLabel->"Out[411]=",
 CellID->499350247,ExpressionUUID->"95813ca7-af9e-44eb-9fbb-de17cd9f36c5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "w"}]], "Input",
 CellLabel->
  "In[412]:=",ExpressionUUID->"bbef108f-c27d-4f5b-b4a5-96c566cb06e6"],

Cell[CellGroupData[{

Cell["Global`w", "Print", "PrintUsage",
 CellChangeTimes->{3.728978028587973*^9},
 CellTags->"Info4123728981628-9181623",
 CellID->213508853,ExpressionUUID->"81374258-63c6-4abc-8ad3-dfd7f1e48801"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "w", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`w],
          Editable->False], "^=", "w"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "w", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "w", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "w", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "w", "]"}], "^=", "\<\"w\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "w", "]"}], "^=", 
         RowBox[{"{", 
          InterpretationBox[
           StyleBox["\[DoubleStruckCapitalT]M3",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`TangentM3,
           Editable->False], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "w", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "w", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "w", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"w", 
              StyleBox[GridBox[{
                 {
                  UnderscriptBox["a", "_"]},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`w[
           Pattern[$CellContext`a, 
            Blank[]]],
          Editable->False], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "b", "}"}], ",", 
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"S", 
                 StyleBox[GridBox[{
                    {"a", "b"},
                    {" ", " "}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`S[$CellContext`a, $CellContext`b],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"v", 
                 StyleBox[GridBox[{
                    {" "},
                    {"b"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`v[-$CellContext`b],
             Editable->False]}]}], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`w],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978028653574*^9},
 CellTags->"Info4123728981628-9181623",
 CellID->2145601881,ExpressionUUID->"93334796-27eb-4b34-88b2-aede27817ead"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"w", "[", "b", "]"}]], "Input",
 CellLabel->
  "In[413]:=",ExpressionUUID->"361dfd15-763c-46fd-9025-a53e3b6cca3e"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"b", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897802904045*^9},
 CellLabel->"Out[413]=",
 CellID->585413543,ExpressionUUID->"bfe14cf8-fbe5-4ea1-ab4c-ba6c1380a3c6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In this case the code finds the conflicting dummies, and then replaces the \
offending dummy on the right hand side:\
\>", "MathCaption",ExpressionUUID->"ebbcebd6-78e1-490f-810d-125b46220498"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[414]:=",ExpressionUUID->"719050be-1a15-4c48-b3bb-3d2e94b1dede"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978029323633*^9},
 CellLabel->"During evaluation of In[414]:=",
 CellID->1422830294,ExpressionUUID->"12f33035-cfb2-4eb9-8472-a07135cd6cf1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a_", ",", "b_", ",", 
     RowBox[{"-", "b_"}]}], "]"}], ",", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "b"}], ",", "c", ",", 
     RowBox[{"-", "c"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[415]:=",ExpressionUUID->"59e3cab6-e2d3-4d19-a4f8-a3479bc2cd5d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "b", " ", "c", " "},
         {" ", " ", "b", " ", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`b, -$CellContext`b, \
$CellContext`c, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897802945444*^9},
 CellLabel->"Out[415]=",
 CellID->1278341913,ExpressionUUID->"a89ee616-250a-4260-90f9-962208d9252a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "T"}]], "Input",
 CellLabel->
  "In[416]:=",ExpressionUUID->"d555789b-a738-4b5c-bd72-7cc5eea70872"],

Cell[CellGroupData[{

Cell["Global`T", "Print", "PrintUsage",
 CellChangeTimes->{3.728978029908429*^9},
 CellTags->"Info4163728981629-9181623",
 CellID->1666127203,ExpressionUUID->"952d080a-d3f2-4c6b-b336-ee40ea619ec5"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "T", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`T],
          Editable->False], "^=", "T"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "T", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "T", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "T", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "T", "]"}], "^=", "\<\"T\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "T", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "T", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "T", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "T", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {
                  UnderscriptBox["a", "_"], 
                  UnderscriptBox["b", "_"], " "},
                 {" ", " ", 
                  UnderscriptBox["b", "_"]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[
           Pattern[$CellContext`a, 
            Blank[]], 
           Pattern[$CellContext`b, 
            Blank[]], -Pattern[$CellContext`b, 
             Blank[]]],
          Editable->False], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"j$20206", ",", "c"}], "}"}], ",", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"S", 
                StyleBox[GridBox[{
                   {"a", "j$20206", " ", "c", " "},
                   {" ", " ", "j$20206", " ", "c"}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`S[$CellContext`a, $CellContext`j$20206, \
-$CellContext`j$20206, $CellContext`c, -$CellContext`c],
            Editable->False]}], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780299352703`*^9},
 CellTags->"Info4163728981629-9181623",
 CellID->1560627588,ExpressionUUID->"dbb65ac4-a485-4871-a20e-6176bd7da37d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"T", "[", 
  RowBox[{"a", ",", "c", ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[417]:=",ExpressionUUID->"ae2b357d-12d1-4214-be5d-41cb3e71cfe2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"a", "c", " ", "b", " "},
         {" ", " ", "c", " ", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`a, $CellContext`c, -$CellContext`c, \
$CellContext`b, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978030291391*^9},
 CellLabel->"Out[417]=",
 CellID->102607058,ExpressionUUID->"d10f3e71-c106-4011-a990-24e0c00812f2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Example of use of ",
 StyleBox["IndexSetDelayed",
  FontFamily->"Courier"],
 ". Note that we only use a pattern in the first index of ",
 StyleBox["U",
  FontFamily->"Courier"],
 " and that ",
 StyleBox["w",
  FontFamily->"Courier"],
 " is still not evaluated:"
}], "MathCaption",ExpressionUUID->"8f72aec1-180c-4368-84eb-2340b268d358"],

Cell[BoxData[
 RowBox[{"IndexSetDelayed", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a_"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{
    RowBox[{"w", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"w", "[", 
     RowBox[{"-", "b"}], "]"}], 
    RowBox[{"w", "[", 
     RowBox[{"-", "c"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[418]:=",ExpressionUUID->"72953c91-ddf4-4c6e-a4da-d35fd4962b44"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "T"}]], "Input",
 CellLabel->
  "In[419]:=",ExpressionUUID->"df876275-f259-4244-b3c9-5711b7f2a3ff"],

Cell[CellGroupData[{

Cell["Global`T", "Print", "PrintUsage",
 CellChangeTimes->{3.728978030561453*^9},
 CellTags->"Info4193728981630-9181623",
 CellID->87455265,ExpressionUUID->"bf548868-f80a-4fce-add5-25287ccbf669"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "T", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`T],
          Editable->False], "^=", "T"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "T", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "T", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "T", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "T", "]"}], "^=", "\<\"T\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "T", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "T", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "T", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "T", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {
                  UnderscriptBox["a", "_"], 
                  UnderscriptBox["b", "_"], " "},
                 {" ", " ", 
                  UnderscriptBox["b", "_"]}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[
           Pattern[$CellContext`a, 
            Blank[]], 
           Pattern[$CellContext`b, 
            Blank[]], -Pattern[$CellContext`b, 
             Blank[]]],
          Editable->False], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"j$20206", ",", "c"}], "}"}], ",", 
           InterpretationBox[
            StyleBox[GridBox[{
               {"S", 
                StyleBox[GridBox[{
                   {"a", "j$20206", " ", "c", " "},
                   {" ", " ", "j$20206", " ", "c"}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`S[$CellContext`a, $CellContext`j$20206, \
-$CellContext`j$20206, $CellContext`c, -$CellContext`c],
            Editable->False]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " ", " "},
                 {
                  UnderscriptBox["a", "_"], "b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[-Pattern[$CellContext`a, 
             Blank[]], -$CellContext`b, -$CellContext`c],
          Editable->False], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            InterpretationBox[
             StyleBox[GridBox[{
                {"w", 
                 StyleBox[GridBox[{
                    {" "},
                    {"a"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w[-$CellContext`a],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"w", 
                 StyleBox[GridBox[{
                    {" "},
                    {"b"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w[-$CellContext`b],
             Editable->False], " ", 
            InterpretationBox[
             StyleBox[GridBox[{
                {"w", 
                 StyleBox[GridBox[{
                    {" "},
                    {"c"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`w[-$CellContext`c],
             Editable->False]}]}], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289780305900517`*^9},
 CellTags->"Info4193728981630-9181623",
 CellID->981824307,ExpressionUUID->"061e13da-7d82-4a5b-9daf-db3bdf7c3cf0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"T", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[420]:=",ExpressionUUID->"5f408341-de08-4e05-adb0-66ca8b8fde3b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "d"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`a, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "e"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`b, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "f"},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`c, $CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`f],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289780307638397`*^9},
 CellLabel->"Out[420]=",
 CellID->256784201,ExpressionUUID->"af48e65b-d8b9-4396-af7d-899bf8a7264b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We can change the definition of ",
 StyleBox["w",
  FontFamily->"Courier"],
 " and the conversion from ",
 StyleBox["S",
  FontFamily->"Courier"],
 " to ",
 StyleBox["w",
  FontFamily->"Courier"],
 " will be still valid:"
}], "MathCaption",ExpressionUUID->"95ed0ed4-ae47-4af6-8a8a-7642709d1d46"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"w", "[", "a_", "]"}], ",", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"v", "[", "c", "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[421]:=",ExpressionUUID->"f5b28f07-1a3d-4835-bda6-7d2ac68adc80"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`b, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289780310003757`*^9},
 CellLabel->"Out[421]=",
 CellID->1477419384,ExpressionUUID->"6f248225-79e0-46aa-ad12-b0405f682313"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"T", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[422]:=",ExpressionUUID->"be29d106-80c3-4119-bec0-652684241156"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "d"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`a, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "e"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`b, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "f"},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`c, $CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", " "},
          {"d", "g"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`d, -$CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", " "},
          {"e", "h"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`e, -$CellContext`h],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", " "},
          {"f", "i"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`f, -$CellContext`i],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"g"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"h"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`h],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"i"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`i],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978031323295*^9},
 CellLabel->"Out[422]=",
 CellID->1781660274,ExpressionUUID->"4fa72d53-38cb-4ce3-84dc-ad58b6884db5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"c9560b09-6b59-4643-914b-27dc64fec181"],

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a_", ",", "b_", ",", 
    RowBox[{"-", "b_"}]}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[423]:=",ExpressionUUID->"3973e3ce-62fb-4998-b3ad-79013d4812d9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[424]:=",ExpressionUUID->"dc4c7c18-c64d-43be-9007-6f7d1fccb5a9"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_", "]"}], "=."}]], "Input",
 CellLabel->
  "In[425]:=",ExpressionUUID->"56ed6a8f-6559-423d-ac00-fd86c2596098"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["5.3. Index patterns", "Subsection",ExpressionUUID->"0c54486b-ed81-4326-8c43-9a6c4133b556"],

Cell[TextData[{
 "In the previous definitions of the vector ",
 StyleBox["w[a]",
  FontFamily->"Courier"],
 " we always used the pattern ",
 StyleBox["a_",
  FontFamily->"Courier"],
 " which stands for any possible input, including up-indices and down-indices \
of any manifold, or even directional or label (or basis or component) \
indices. Sometimes this is what we want, but often it is not. In the \
following definitions there are no dummies or conflicts between indices and \
tensor names, so that we can safely use the usual ",
 StyleBox["SetDelayed",
  FontFamily->"Courier"],
 " (:=) definitions to simplify the examples.\nThe notation for tensors and \
indices in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " is very simple, but that complicates working with patterns. On the other \
hand, there are always several ways to represent the same pattern."
}], "Text",ExpressionUUID->"8ba0453c-a6e9-4c44-a05b-fb8abdd02d7a"],

Cell[CellGroupData[{

Cell["Define, for the sake of simplicity:", "MathCaption",ExpressionUUID->"36ebbdd5-f694-4037-9fbb-2ce37999e5a6"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"dirvup", "=", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", "d", "]"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dirvdown", "=", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", 
     RowBox[{"-", "d"}], "]"}], "]"}]}], ";"}]}], "Input",
 CellLabel->
  "In[426]:=",ExpressionUUID->"ee71972b-d025-4784-a9ed-a4f47641c5e1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "This pattern stands for everything. Every ",
 StyleBox["w",
  FontFamily->"Courier"],
 " is converted into ",
 StyleBox["T",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"9f7bfa3c-2429-427e-9b88-14cfedef94df"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_", "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[428]:=",ExpressionUUID->"51206389-df84-434d-af94-22eea7f5f965"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[429]:=",ExpressionUUID->"2fdb8634-e4cb-4c67-84f7-8d1ef0735729"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978071894261*^9},
 CellLabel->"Out[429]=",
 CellID->1976281324,ExpressionUUID->"cfd1c666-a5a0-4aea-a20b-6d4cd8f55fb2"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_", "]"}], "=."}]], "Input",
 CellLabel->
  "In[430]:=",ExpressionUUID->"a9c10f52-19f4-4eba-9fd8-2794b1c54b3f"]
}, Open  ]],

Cell["\<\
Let us first work with the character of the indices: use the functions\
\>", "Text",ExpressionUUID->"ced5b990-150a-4c6e-a07f-9d8c3aa77145"],

Cell[TextData[{
 StyleBox[ButtonBox["UpIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UpIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect up-indices\n", "Text"],
 StyleBox[ButtonBox["DownIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DownIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect down-indices", "Text"]
}], "DefinitionBox",ExpressionUUID->"437d8ce8-b970-4e81-bf18-418678b3c1d2"],

Cell["Index character.", "Text",
 CellChangeTimes->{{3.72897807925891*^9, 3.7289780853865643`*^9}},
 CellID->476091498,ExpressionUUID->"b0ada00d-de8e-4014-95b5-62bf8e63d420"],

Cell[CellGroupData[{

Cell["This pattern matches only up-indices:", "MathCaption",ExpressionUUID->"6d80eeeb-e0a8-4b40-a422-6660021980ea"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "UpIndexQ"}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[431]:=",ExpressionUUID->"f8191a5e-9341-4167-b266-99f5dd10c4d7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[432]:=",ExpressionUUID->"6eb71841-a7ce-4c2a-9a56-4da149b6735b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289780721203823`*^9},
 CellLabel->"Out[432]=",
 CellID->230644319,ExpressionUUID->"f3caa6ad-2c8f-4224-a954-3576826c3415"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "UpIndexQ"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[433]:=",ExpressionUUID->"3556676a-c865-4ba1-9ff6-c13c1069c582"]
}, Open  ]],

Cell[CellGroupData[{

Cell["This pattern matches only down-indices:", "MathCaption",ExpressionUUID->"2ef8cbf3-599a-44fa-b6bf-ce78be298b0d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "DownIndexQ"}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[434]:=",ExpressionUUID->"6d79c9ad-905e-4e66-9dce-bce66d072e6a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[435]:=",ExpressionUUID->"a44699cc-b758-4c40-a3a6-b88b5e843a7f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289780723497753`*^9},
 CellLabel->"Out[435]=",
 CellID->231086598,ExpressionUUID->"6950d25d-e27a-4fec-b863-cd2d6bbb0b10"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "DownIndexQ"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[436]:=",ExpressionUUID->"c95f5328-d72c-41a9-90cd-42cf3252f5c3"]
}, Open  ]],

Cell["And now we work with the type of index: use the functions", "Text",ExpressionUUID->"ae6efe61-4ecd-4f28-a510-fb36a8833e9d"],

Cell[TextData[{
 StyleBox[ButtonBox["AIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "AIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect abstract indices\n", "Text"],
 StyleBox[ButtonBox["BIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "BIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect basis indices\n", "Text"],
 StyleBox[ButtonBox["CIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "CIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect component indices\n", "Text"],
 StyleBox[ButtonBox["DIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect directional indices\n", "Text"],
 StyleBox[ButtonBox["LIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect label indices\n", "Text"],
 StyleBox[ButtonBox["GIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "GIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect all generalized indices, but not patterns\n", "Text"],
 StyleBox[ButtonBox["ABIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ABIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect contractible indices (abstract or basis indices)\n", "Text"],
 StyleBox[ButtonBox["BCIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "BCIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect indices associated to a basis or chart (basis or component \
indices)\n", "Text"],
 StyleBox[ButtonBox["CDIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "CDIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect indices representing a direction (component or directional \
indices)\n", "Text"],
 StyleBox[ButtonBox["PIndexQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PIndexQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Detect pattern indices", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.7289780460358047`*^9, 
  3.7289780515551577`*^9}},ExpressionUUID->"26e7a798-4ea3-4aa1-8ead-\
05da3633526f"],

Cell["Detect types of indices.", "Text",ExpressionUUID->"fc651664-c941-41b4-990b-e5357d96a4e7"],

Cell[CellGroupData[{

Cell["Introduce another manifold, with a new set of indices:", "MathCaption",ExpressionUUID->"f097ca9d-1f91-4029-8828-7f1301905571"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"S2", ",", "2", ",", 
   RowBox[{"{", 
    RowBox[{"A", ",", "B", ",", "C", ",", "D", ",", "F", ",", "G"}], "}"}]}], 
  "]"}]], "Input",
 CellLabel->
  "In[437]:=",ExpressionUUID->"2e8184f9-4f2c-4c03-a110-6bd62f359a5e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "S2", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`S2, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978072517372*^9},
 CellLabel->"During evaluation of In[437]:=",
 CellID->1222018810,ExpressionUUID->"097a9043-b4e6-4c98-b606-709277669351"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentS2", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentS2, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978072533938*^9},
 CellLabel->"During evaluation of In[437]:=",
 CellID->2084023409,ExpressionUUID->"fdc598e1-8ae8-49f1-8efe-9c10804be408"]
}, Open  ]],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"C\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 437, 55, 19253573948027831207, 
   "Local", "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978072551712*^9},
 CellLabel->"During evaluation of In[437]:=",
 CellID->723152493,ExpressionUUID->"dd97c8dc-ec88-43d8-8027-81821787580f"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"D\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 437, 56, 19253573948027831207, 
   "Local", "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289780725743217`*^9},
 CellLabel->"During evaluation of In[437]:=",
 CellID->1413991478,ExpressionUUID->"395c56a3-76b1-4355-a9b5-25a0a7dc5895"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This pattern matches only abstract indices, from any vector bundle:\
\>", "MathCaption",ExpressionUUID->"55dba7d2-f262-46ac-b27a-ce6e66f3fda4"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "AIndexQ"}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[438]:=",ExpressionUUID->"ceeb0815-22b4-4413-bf72-b4289fa91eb5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[439]:=",ExpressionUUID->"4d6a4570-e9ac-4be1-87db-6ca78a5af46b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289780727833433`*^9},
 CellLabel->"Out[439]=",
 CellID->900733044,ExpressionUUID->"8ab0d10b-cf74-4c17-b7fb-940c9e399e02"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "AIndexQ"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[440]:=",ExpressionUUID->"af2befa4-3824-4379-ad24-dcb58bae11d4"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
There is a simple and efficient pattern for abstract up-indices or abstract \
down-indices:\
\>", "MathCaption",ExpressionUUID->"bb7a1e4d-17cb-4293-82bc-5ac4a8743215"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_Symbol", "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[441]:=",ExpressionUUID->"de2b45e8-8957-4b19-86cb-35f305b3c619"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[442]:=",ExpressionUUID->"e699b822-b73b-4d8d-85c2-cb171a6aeab1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289780730331182`*^9},
 CellLabel->"Out[442]=",
 CellID->1008042805,ExpressionUUID->"f8d04a7b-baea-4679-a505-2e8fade1b924"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_Symbol", "]"}], "=."}]], "Input",
 CellLabel->
  "In[443]:=",ExpressionUUID->"0c73b946-d8f4-4fc7-ba87-6bd7af16f156"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"-", "a_Symbol"}], "]"}], ":=", 
  RowBox[{"S", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellLabel->
  "In[444]:=",ExpressionUUID->"1d1fb3d3-21d9-4985-9650-67f51d31efd0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[445]:=",ExpressionUUID->"22bda2a5-71dd-443a-af89-31a674e8cdcd"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978073283608*^9},
 CellLabel->"Out[445]=",
 CellID->676932452,ExpressionUUID->"763c1130-b0df-4acc-9bde-1dc77d29e6ff"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"-", "a_Symbol"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[446]:=",ExpressionUUID->"a51dd1a3-e92a-481f-ba3a-0da8407ba1dd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "This pattern matches only directional indices. Recall that directions do \
not have a sign in front. A message is sent complaining about the character \
of the vector ",
 StyleBox["v",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"016a86b5-b99f-403a-8e40-59ca1e34ae16"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "DIndexQ"}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[447]:=",ExpressionUUID->"fbc860d7-c346-4310-9a6f-507c2828e1aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[448]:=",ExpressionUUID->"d52a7bd1-911b-4062-ad06-ea612be3053f"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor v\\\\\
\\\"\\\"}]\\)\"", 2, 448, 57, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978073493888*^9},
 CellLabel->"During evaluation of In[448]:=",
 CellID->1982842605,ExpressionUUID->"261dd749-52bc-45b7-ab25-d739655831e6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978073549459*^9},
 CellLabel->"Out[448]=",
 CellID->984428089,ExpressionUUID->"54c578a0-7f03-475c-b658-7bdc7bef19fc"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "DIndexQ"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[449]:=",ExpressionUUID->"675daec1-5fcb-441d-a950-f12b85f76ed5"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
There is also a simpler version (which does not find the problem with the \
position of the ultraindex):\
\>", "MathCaption",ExpressionUUID->"cde8af92-bff1-4610-913d-58239f8c5e1f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_Dir", "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[450]:=",ExpressionUUID->"414e6fb1-2755-4ff6-b191-c75a6d63601c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[451]:=",ExpressionUUID->"41be1a05-e5eb-4253-bbdc-ec4fdfaf264b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289780738189907`*^9},
 CellLabel->"Out[451]=",
 CellID->1068875335,ExpressionUUID->"2a8cc52d-5df6-4548-8a3c-459488dd905d"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", "a_Dir", "]"}], "=."}]], "Input",
 CellLabel->
  "In[452]:=",ExpressionUUID->"a93043b7-51ff-48fc-a241-bd8960979fb9"]
}, Open  ]],

Cell[TextData[{
 "The functions ",
 StyleBox["AIndexQ",
  FontFamily->"Courier"],
 ", ",
 StyleBox["BIndexQ",
  FontFamily->"Courier"],
 ", ",
 StyleBox["CIndexQ",
  FontFamily->"Courier"],
 ", ",
 StyleBox["DIndexQ",
  FontFamily->"Courier"],
 " and ",
 StyleBox["GIndexQ",
  FontFamily->"Courier"],
 " admit a second argument restricting the manifold to which the indices must \
belong."
}], "Text",ExpressionUUID->"f58db9c6-609f-43ea-819c-35a4c2e5d203"],

Cell[CellGroupData[{

Cell[TextData[{
 "This pattern only matches abstract indices on ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"d7c7522f-2601-41d8-afe1-f69345c711da"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"AIndexQ", "[", 
       RowBox[{"#", ",", "TangentM3"}], "]"}], "&"}], ")"}]}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[453]:=",ExpressionUUID->"ff8fec80-16cd-4006-9898-cc92318a64d7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[454]:=",ExpressionUUID->"449d2114-6de5-4c6e-b869-afc49d75a5bb"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289780740401382`*^9},
 CellLabel->"Out[454]=",
 CellID->887048190,ExpressionUUID->"914596a6-bab1-4d39-bf67-cc850fabcafe"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"AIndexQ", "[", 
       RowBox[{"#", ",", "TangentM3"}], "]"}], "&"}], ")"}]}], "]"}], 
  "=."}]], "Input",
 CellLabel->
  "In[455]:=",ExpressionUUID->"71b035a4-9bc2-47fe-9fa5-7d8442c8064e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Again there are simpler and cleaner versions, also discriminating between \
upindices and downindices (",
 StyleBox["Q",
  FontFamily->"Courier"],
 "-functions) or not (",
 StyleBox["pmQ",
  FontFamily->"Courier"],
 "-functions). The use of ",
 StyleBox["Symbol",
  FontFamily->"Courier"],
 " is not needed now."
}], "MathCaption",ExpressionUUID->"5d54706a-9b7b-4705-9639-377f3057507e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "TangentM3`Q"}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[456]:=",ExpressionUUID->"64f457e2-1123-4be9-ae15-93052e29b3d2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[457]:=",ExpressionUUID->"a7c96a73-0f8d-4265-b6f9-fa09903ce333"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978074301056*^9},
 CellLabel->"Out[457]=",
 CellID->787451856,ExpressionUUID->"ec5252f5-6523-49f9-9f4b-37d352d558d8"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "TangentM3`Q"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[458]:=",ExpressionUUID->"c4ca18e4-920a-46dd-973b-8b0a1f691b11"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"-", 
    RowBox[{"a_", "?", "TangentM3`Q"}]}], "]"}], ":=", 
  RowBox[{"S", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellLabel->
  "In[459]:=",ExpressionUUID->"66b9f7f1-d2ee-4f9c-b00f-1774528df694"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[460]:=",ExpressionUUID->"6763dab0-c483-4296-bb3e-f37361f73b5f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978074550058*^9},
 CellLabel->"Out[460]=",
 CellID->1909110978,ExpressionUUID->"6c919914-0917-493f-9fb7-823f010f2626"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"-", 
    RowBox[{"a_", "?", "TangentM3`Q"}]}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[461]:=",ExpressionUUID->"1bc5dbfa-9ab3-4214-96b5-532027cfb0dc"],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "TangentM3`pmQ"}], "]"}], ":=", 
  RowBox[{"S", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[462]:=",ExpressionUUID->"3f196dba-7207-4e59-bd2d-2bc0e64d94bf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"w", "[", "a", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"w", "[", "dirvup", "]"}], ",", 
   RowBox[{"w", "[", "dirvdown", "]"}], ",", 
   RowBox[{"w", "[", "A", "]"}], ",", 
   RowBox[{"w", "[", 
    RowBox[{"-", "A"}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[463]:=",ExpressionUUID->"eb505c38-109a-4746-941e-68888f66c9b6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"v"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"v"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[
     xAct`xTensor`Dir[
      $CellContext`v[-$CellContext`d]]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"A"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`A],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {" "},
           {"A"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[-$CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978074833568*^9},
 CellLabel->"Out[463]=",
 CellID->875037419,ExpressionUUID->"9da9f4fc-ecbe-4013-809e-a5f066521541"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"w", "[", 
   RowBox[{"a_", "?", "TangentM3`pmQ"}], "]"}], "=."}]], "Input",
 CellLabel->
  "In[464]:=",ExpressionUUID->"da793f29-7a78-46a4-999f-a01941b6d611"]
}, Open  ]],

Cell[TextData[{
 "In case of doubt use the function ",
 StyleBox["PatternIndex",
  FontFamily->"Courier"],
 ", which constructs patterns of the required form."
}], "Text",ExpressionUUID->"0f9d9002-5a02-4b69-969a-2cec7583d11f"],

Cell[CellGroupData[{

Cell[TextData[{
 "These are examples of different patterns for abstract indices named ",
 StyleBox["a",
  FontFamily->"Courier"],
 ". Note the minus sign in front of ",
 StyleBox["a",
  FontFamily->"Courier"],
 " for down-indices. The pattern is for the symbol of the index, and not for \
the whole index."
}], "MathCaption",ExpressionUUID->"486f14ab-f014-470b-b191-62914509d30a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"a", ",", "AIndex"}], "]"}]], "Input",
 CellLabel->
  "In[465]:=",ExpressionUUID->"d825c4b2-4ee6-421c-bad2-281fc20dafc2"],

Cell[BoxData[
 RowBox[{"a_", "?", "AIndexQ"}]], "Output",
 CellChangeTimes->{3.728978075042198*^9},
 CellLabel->"Out[465]=",
 CellID->785886477,ExpressionUUID->"d6064a0c-1809-4bdd-9b09-19620bced63e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"a", ",", "AIndex", ",", "Up"}], "]"}]], "Input",
 CellLabel->
  "In[466]:=",ExpressionUUID->"02a3e6fb-14da-47e1-be5a-b3d472ee38ac"],

Cell[BoxData[
 RowBox[{"a_Symbol", "?", "AIndexQ"}]], "Output",
 CellChangeTimes->{3.728978075183324*^9},
 CellLabel->"Out[466]=",
 CellID->1369589889,ExpressionUUID->"8d526ad7-e5fd-4411-aac0-f8d24bfea9ed"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"a", ",", "AIndex", ",", "Down", ",", "TangentM3"}], "]"}]], "Input",\

 CellLabel->
  "In[467]:=",ExpressionUUID->"6f39331c-df82-42b1-9c6d-d20952183bfc"],

Cell[BoxData[
 RowBox[{"-", 
  RowBox[{"a_Symbol", "?", "TangentM3`Q"}]}]], "Output",
 CellChangeTimes->{3.72897807529519*^9},
 CellLabel->"Out[467]=",
 CellID->1967535877,ExpressionUUID->"36a2b13e-4301-49ba-a41b-ce5d82c80da9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"a", ",", "AIndex", ",", "Null", ",", "TangentM3"}], "]"}]], "Input",\

 CellLabel->
  "In[468]:=",ExpressionUUID->"f802c802-505b-454f-833d-07361b5af857"],

Cell[BoxData[
 RowBox[{"a_", "?", "TangentM3`pmQ"}]], "Output",
 CellChangeTimes->{3.728978075412672*^9},
 CellLabel->"Out[468]=",
 CellID->1422288233,ExpressionUUID->"de4ff2d0-292a-4e95-8a7c-44e26ceb235f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "These are patterns for directional indices named ",
 StyleBox["d",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"921b65c1-4d65-4a34-b51b-8623031b4be6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"d", ",", "DIndex"}], "]"}]], "Input",
 CellLabel->
  "In[469]:=",ExpressionUUID->"8887fe1b-9677-4145-9b40-91fbdf150e07"],

Cell[BoxData["d_Dir"], "Output",
 CellChangeTimes->{3.728978075550022*^9},
 CellLabel->"Out[469]=",
 CellID->757857274,ExpressionUUID->"1200ff46-45a3-4566-add6-8b6a40c2e49a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"d", ",", "DIndex", ",", "Up"}], "]"}]], "Input",
 CellLabel->
  "In[470]:=",ExpressionUUID->"0bd9852b-d71a-4712-8149-9e34182c36b7"],

Cell[BoxData[
 RowBox[{"d_Dir", "?", "UpIndexQ"}]], "Output",
 CellChangeTimes->{3.7289780756627607`*^9},
 CellLabel->"Out[470]=",
 CellID->581693761,ExpressionUUID->"6ffa627e-603f-4c34-af4d-296c562ace22"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"d", ",", "DIndex", ",", "Down", ",", "TangentM3"}], "]"}]], "Input",\

 CellLabel->
  "In[471]:=",ExpressionUUID->"bbc3e3fe-75e1-45a4-97bc-46d3a3599e20"],

Cell[BoxData[
 RowBox[{"d_Dir", "?", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"DownIndexQ", "[", "#1", "]"}], "&&", 
     RowBox[{"DIndexQ", "[", 
      RowBox[{"#1", ",", 
       InterpretationBox[
        StyleBox["\[DoubleStruckCapitalT]M3",
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`TangentM3,
        Editable->False]}], "]"}]}], "&"}], ")"}]}]], "Output",
 CellChangeTimes->{3.728978075779697*^9},
 CellLabel->"Out[471]=",
 CellID->1181479983,ExpressionUUID->"840170f6-ef0e-4256-b985-368afb225bc3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PatternIndex", "[", 
  RowBox[{"d", ",", "DIndex", ",", "Null", ",", "TangentM3"}], "]"}]], "Input",\

 CellLabel->
  "In[472]:=",ExpressionUUID->"7e8f09dc-70d2-493b-82bb-4775106db488"],

Cell[BoxData[
 RowBox[{"d_Dir", "?", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"DIndexQ", "[", 
     RowBox[{"#1", ",", 
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]M3",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentM3,
       Editable->False]}], "]"}], "&"}], ")"}]}]], "Output",
 CellChangeTimes->{3.7289780759192142`*^9},
 CellLabel->"Out[472]=",
 CellID->1500274119,ExpressionUUID->"82a847c8-06cb-47f8-a830-c5a1459823d2"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["5.4. IndexRule and IndexRuleDelayed", "Subsection",ExpressionUUID->"d7e8ca92-13b6-4601-8c16-25cbe48767af"],

Cell[TextData[{
 "The function ",
 StyleBox["IndexRule",
  FontFamily->"Courier"],
 " relates to ",
 StyleBox["Rule",
  FontFamily->"Courier"],
 " exactly in the same way that ",
 StyleBox["IndexSet",
  FontFamily->"Courier"],
 " relates to ",
 StyleBox["Set",
  FontFamily->"Courier"],
 ". It just introduces Module constructs. The symbol \[RightTeeArrow] has \
been introduced. It is input as \[Backslash][RightTeeArrow], but this input \
sequence doesn't work with the style used in this notebook (I don't know why)."
}], "Text",ExpressionUUID->"84a1dbae-2441-42ae-892b-e338260ed0d2"],

Cell[TextData[{
 StyleBox[ButtonBox["IndexRule",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexRule.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Construct rule for an indexed expression, at definition time\n", 
  "Text"],
 StyleBox[ButtonBox["IndexRuleDelayed",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexRuleDelayed.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Construct rule for an indexed expression, at evaluation time", 
  "Text"]
}], "DefinitionBox",ExpressionUUID->"e72a0d6e-eec2-4bef-be6e-a5d2b4c1bd5e"],

Cell["Rule  functions for indexed expressions.", "Text",ExpressionUUID->"d67261fd-ffb6-4400-9b20-7740fbf39f9d"],

Cell[CellGroupData[{

Cell[TextData[{
 "Here we have again the problem of repeated indices ",
 StyleBox["b",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"da3aa5e8-9fc6-4f2f-8e48-05638e8ac8dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"w", "[", "a", "]"}], ",", 
    RowBox[{"w", "[", "b", "]"}]}], "}"}], "/.", 
  RowBox[{
   RowBox[{"w", "[", "a_", "]"}], "\[Rule]", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}]}]}]], "Input",
 CellLabel->
  "In[473]:=",ExpressionUUID->"a375752c-54fa-4c2b-a5bd-0965de4fd5a1"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "repeated", 
   "\"Found indices with the same name \\!\\(\\*RowBox[{\\\"\\\\\\\"b\\\\\\\"\
\\\"}]\\).\"", 2, 473, 58, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978099977803*^9},
 CellLabel->"During evaluation of In[473]:=",
 CellID->11055515,ExpressionUUID->"798d69f3-0291-490a-b1ce-bffe04affaa5"],

Cell[BoxData[
 TemplateBox[{
  "Throw", "nocatch", 
   "\"Uncaught \\!\\(\\*RowBox[{\\\"Throw\\\", \\\"[\\\", \\\"Null\\\", \\\"]\
\\\"}]\\) returned to top level.\"", 2, 473, 59, 19253573948027831207, 
   "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728978100019848*^9},
 CellLabel->"During evaluation of In[473]:=",
 CellID->1126446963,ExpressionUUID->"e6c27563-0f1b-48cf-a6ca-a685368e9c9a"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{"Throw", "[", "Null", "]"}], "]"}]], "Output",
 CellChangeTimes->{3.728978100034358*^9},
 CellLabel->"Out[473]=",
 CellID->2125758836,ExpressionUUID->"3bce8da1-fbfa-47f5-8f67-e4ddba866a24"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Using ",
 StyleBox["IndexRule",
  FontFamily->"Courier"],
 " the problem disappears:"
}], "MathCaption",ExpressionUUID->"f83008dc-e96c-4728-a8b6-702d21631da1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"w", "[", "a", "]"}], ",", 
    RowBox[{"w", "[", "b", "]"}]}], "}"}], "/.", 
  RowBox[{"IndexRule", "[", 
   RowBox[{
    RowBox[{"w", "[", "a_", "]"}], ",", 
    RowBox[{
     RowBox[{"S", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}]}]}], "]"}]}]], "Input",
 CellLabel->
  "In[474]:=",ExpressionUUID->"e15ee38b-1700-4870-a790-75c121f26279"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"b", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`b, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978100195264*^9},
 CellLabel->"Out[474]=",
 CellID->2090805160,ExpressionUUID->"46317b68-5bc7-4aca-bbb0-f466327203c2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"w", "[", "a", "]"}], ",", 
    RowBox[{"w", "[", "b", "]"}]}], "}"}], "/.", 
  RowBox[{
   RowBox[{"w", "[", "a_", "]"}], "\[RightTeeArrow]", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}]}]}]], "Input",
 CellLabel->
  "In[475]:=",ExpressionUUID->"69df4d90-cd0e-4ab8-8524-02bdf1adc3cd"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"b", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`b, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978100323357*^9},
 CellLabel->"Out[475]=",
 CellID->887408259,ExpressionUUID->"9c4661ce-4fe6-4c03-849c-aeb92dbc6cf5"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["5.5. MakeRule", "Subsection",ExpressionUUID->"80e795fd-f232-4a73-87ac-aff5ef846f9f"],

Cell[TextData[{
 "Apart from the coding aspects solved by the ",
 StyleBox["Index*",
  FontFamily->"Courier"],
 " functions, there are other, more mathematical issues which must be \
adressed. For instance, given a rule LHS\[Rule]RHS how can we apply the rule \
on expressions which are equivalent to LHS but indices are permuted due to \
symmetries or metric shifting? Can we apply the rule on all vbundles or just \
on some of them? These kind of options are controlled by ",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 ", the second most complicated function in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " after ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 ". This function uses the properties of the metric tensors, and hence we \
delay its description to subsection 9.1 below."
}], "Text",ExpressionUUID->"8909e5f1-e7bc-423f-8695-d8514a0aa07a"]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["6. Derivatives", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}, {3.728976670882564*^9, 
  3.728976674720446*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1740316966,ExpressionUUID->"4e72983f-5bfb-4fb2-bf7e-973e9afda9de"],

Cell[CellGroupData[{

Cell["6.1. Covariant derivatives", "Subsection",ExpressionUUID->"a690d2bf-571c-4ad2-aa9f-9984e4e62f1a"],

Cell[TextData[{
 "We understand a covariant derivative as a covariant derivative operator \
(see Wald). In particular, ordinary (\"partial\") derivative operators are \
obtained as covariant derivatives with the options ",
 StyleBox["Curvature->False",
  FontFamily->"Courier"],
 " and ",
 StyleBox["Torsion->False",
  FontFamily->"Courier"],
 ". Covariant derivatives are real operators acting on (possibly complex) \
vector bundles. The treatment of derivatives on ",
 StyleBox["xTensor`", "DisplayMath"],
 " is fully abstract, and hence we do not explain here how to take components \
of a derivative of a tensor. This is dealt with in the companion package ",
 StyleBox["xCoba`", "DisplayMath"],
 ", and explained in its documentation."
}], "Text",
 CellChangeTimes->{{3.458663370797729*^9, 
  3.458663455965334*^9}},ExpressionUUID->"133463b9-e5a9-4187-9a52-\
093c6ba680ac"],

Cell[TextData[{
 StyleBox[ButtonBox["DefCovD",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefCovD.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Define a covariant derivative operator\n", "Text"],
 StyleBox[ButtonBox["UndefCovD",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefCovD.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Undefine a covariant derivative", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["$CovDs",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$CovDs.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined covariant derivatives\n", "Text"],
 StyleBox[ButtonBox["CovDQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "CovDQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["Validate a covariant derivative symbol", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978113472513*^9},ExpressionUUID->"8afd3f5a-0cca-49d9-b74a-\
c2ccb0401d51"],

Cell["Definition of a covariant derivative.", "Text",ExpressionUUID->"c9e6948a-71d7-4eef-a215-a43f2eb97815"],

Cell[CellGroupData[{

Cell[TextData[{
 "Default options for ",
 StyleBox["DefCovD",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"37df5ee4-d77f-4d53-9d14-9ce3c78f7d3d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Options", "[", "DefCovD", "]"}]], "Input",
 CellLabel->
  "In[476]:=",ExpressionUUID->"0d52f1ba-3a86-45dc-b1a1-3ea99384db07"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"SymbolOfCovD", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\<\";\"\>", ",", "\<\"\[EmptyDownTriangle]\"\>"}], "}"}]}], ",", 
   RowBox[{"Torsion", "\[Rule]", "False"}], ",", 
   RowBox[{"Curvature", "\[Rule]", "True"}], ",", 
   RowBox[{"FromMetric", "\[Rule]", "Null"}], ",", 
   RowBox[{"CurvatureRelations", "\[Rule]", "True"}], ",", 
   RowBox[{"ExtendedFrom", "\[Rule]", "Null"}], ",", 
   RowBox[{"OtherDependencies", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"OrthogonalTo", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"ProjectedWith", "\[Rule]", 
    RowBox[{"{", "}"}]}], ",", 
   RowBox[{"WeightedWithBasis", "\[Rule]", "Null"}], ",", 
   RowBox[{"ProtectNewSymbol", "\[RuleDelayed]", "$ProtectNewSymbols"}], ",", 
   
   RowBox[{"Master", "\[Rule]", "Null"}], ",", 
   RowBox[{"DefInfo", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"\<\"covariant derivative\"\>", ",", "\<\"\"\>"}], "}"}]}]}], 
  "}"}]], "Output",
 CellChangeTimes->{3.7289781391250687`*^9},
 CellLabel->"Out[476]=",
 CellID->492296792,ExpressionUUID->"5fde7438-599a-47eb-94ae-61491e770b22"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a covariant derivative ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 " on ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 " (the default vbundle obtained from the index given in the derivative). \
Automatically the Christoffel symbols and the Riemann and Ricci tensors are \
defined. By default ",
 StyleBox["xTensor`", "DisplayMath"],
 " does not assume that this derivative is associated to a metric and \
therefore the symmetries of the Riemann tensor are not as expected: there is \
just antisymmetry in the first pair of indices. The Ricci tensor has no \
symmetry at all. There is no Ricci scalar."
}], "MathCaption",ExpressionUUID->"afdb5d09-70d7-48ef-83b5-3b3832386b4d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[477]:=",ExpressionUUID->"79183604-22b4-4485-b58e-7b98759e8bd9"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"Cd", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`Cd[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781392914877`*^9},
 CellLabel->"During evaluation of In[477]:=",
 CellID->1925174900,ExpressionUUID->"1bd42072-895c-4893-9579-052efcab4f94"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCd[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781393091516`*^9},
 CellLabel->"During evaluation of In[477]:=",
 CellID->278434951,ExpressionUUID->"dd8ccbb1-f904-4f1d-8930-f7f960053c54"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978139327992*^9},
 CellLabel->"During evaluation of In[477]:=",
 CellID->289354644,ExpressionUUID->"3e7f4cf6-ed3b-4187-8bad-3892757cf16c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978139346546*^9},
 CellLabel->"During evaluation of In[477]:=",
 CellID->1280524900,ExpressionUUID->"4a65eff7-186e-4377-800e-d3f9eb9ab1af"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978139365079*^9},
 CellLabel->"During evaluation of In[477]:=",
 CellID->1848367713,ExpressionUUID->"7887f91e-c263-48b9-adde-e21ed590c047"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.7289781393827553`*^9},
 CellLabel->"During evaluation of In[477]:=",
 CellID->17714889,ExpressionUUID->"b0a29026-525e-405b-b429-133e8febea5b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "RiemannCd", "]"}]], "Input",
 CellLabel->
  "In[478]:=",ExpressionUUID->"d97aae0d-3cac-49a7-84fa-b0f724f7dcb8"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "1", ",", "2", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728978139526023*^9},
 CellLabel->"Out[478]=",
 CellID->1381396137,ExpressionUUID->"028a63d1-6afa-4956-a580-414be978afa7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCd", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "b"}], "]"}]], "Input",
 CellLabel->
  "In[479]:=",ExpressionUUID->"b0981e3f-f5e6-45d6-ad7d-84ea6f2826fe"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RicciCd[-$CellContext`a, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978139674316*^9},
 CellLabel->"Out[479]=",
 CellID->679166842,ExpressionUUID->"805cbad3-a5aa-4714-9547-9c6b8d157ddc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "RicciCd", "]"}]], "Input",
 CellLabel->
  "In[480]:=",ExpressionUUID->"ac559eb5-88b0-4554-b0dd-5cd1b22a19ef"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"GenSet", "[", "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7289781398108397`*^9},
 CellLabel->"Out[480]=",
 CellID->881529689,ExpressionUUID->"207ef9a5-0164-4b6d-a4a4-d90167bc7102"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The names of the associated tensors are determined by the function ",
 StyleBox["GiveSymbol",
  FontFamily->"Courier"],
 ". Their output symbols are determined by the function ",
 StyleBox["GiveOutputString.",
  FontFamily->"Courier"],
 "If you want to change any of those, modify the corresponding function \
before defining the object."
}], "MathCaption",ExpressionUUID->"38520613-bd95-4dd7-a01a-908a5bf57002"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GiveSymbol", "[", 
  RowBox[{"Riemann", ",", "Cd"}], "]"}]], "Input",
 CellLabel->
  "In[481]:=",ExpressionUUID->"2acaf54c-fa83-4202-a855-3a7e0fdb506f"],

Cell[BoxData["RiemannCd"], "Output",
 CellChangeTimes->{3.728978139960977*^9},
 CellLabel->"Out[481]=",
 CellID->379802370,ExpressionUUID->"f644bc4c-abe0-4084-81fc-3ed46b004f2c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GiveOutputString", "[", 
  RowBox[{"Riemann", ",", "Cd"}], "]"}]], "Input",
 CellLabel->
  "In[482]:=",ExpressionUUID->"d53cfcbe-78ca-4320-8290-81b302a3ec04"],

Cell[BoxData["\<\"R[\[EmptyDownTriangle]]\"\>"], "Output",
 CellChangeTimes->{3.728978140107952*^9},
 CellLabel->"Out[482]=",
 CellID->1017033706,ExpressionUUID->"8ca67209-9fcc-4ed1-86e7-53c2213387bc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Derivatives can be represented in two ways, encoded as the strings ",
 StyleBox["\"Prefix\"",
  FontFamily->"Courier"],
 " or ",
 StyleBox["\"Postfix\"",
  FontFamily->"Courier"],
 " in the global variable ",
 StyleBox["$CovDFormat",
  FontFamily->"Courier"],
 ". The default value is ",
 StyleBox["\"Prefix\"",
  FontFamily->"Courier"],
 ". Directional derivatives are always given in ",
 StyleBox["\"Prefix\"",
  FontFamily->"Courier"],
 " notation."
}], "MathCaption",ExpressionUUID->"07a44507-6585-406e-b64b-2ae3ad854432"],

Cell[CellGroupData[{

Cell[BoxData["$CovDFormat"], "Input",
 CellLabel->
  "In[483]:=",ExpressionUUID->"50150507-d5b2-4dc0-9646-99f049be0933"],

Cell[BoxData["\<\"Prefix\"\>"], "Output",
 CellChangeTimes->{3.728978140274942*^9},
 CellLabel->"Out[483]=",
 CellID->1463150269,ExpressionUUID->"cc0dfa08-bb35-4f7c-a270-6d20d5d18428"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b", ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[484]:=",ExpressionUUID->"6f8d56ff-52ad-400b-975f-5aff724e4aac"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", " "},
          {" ", " ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897814042612*^9},
 CellLabel->"Out[484]=",
 CellID->331329286,ExpressionUUID->"090e344f-6d27-4300-ae89-5a71cbdae147"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CovDFormat", "=", "\"\<Postfix\>\""}], ";"}]], "Input",
 CellLabel->
  "In[485]:=",ExpressionUUID->"fbf14784-23b7-4418-914c-73a37faf7c68"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b", ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[486]:=",ExpressionUUID->"3392d8b8-a285-48ad-902e-9b64c82f47aa"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {"a", "b", " ", "  "},
         {" ", " ", "c", 
          RowBox[{";", "a"}]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978140635421*^9},
 CellLabel->"Out[486]=",
 CellID->226213131,ExpressionUUID->"919f223d-83df-43cd-8e09-362719f0a3bb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"Dir", "[", 
    RowBox[{"v", "[", "d", "]"}], "]"}], "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b", ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[487]:=",ExpressionUUID->"b396d441-e5a0-4ddb-8f30-c25b4e37cbde"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "v"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", " "},
          {" ", " ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]]][
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978140771884*^9},
 CellLabel->"Out[487]=",
 CellID->918037527,ExpressionUUID->"cdb08fd9-acb9-4ef6-86f5-dcedc088c7a4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Derivatives are linear operators and implement automatically the Leibnitz \
rule. The final ordering of terms is decided by ",
 StyleBox["Mathematica.",
  FontSlant->"Italic"]
}], "MathCaption",ExpressionUUID->"7ddcc178-4ed1-47c5-b794-7ad914bdcba3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"r", "[", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[488]:=",ExpressionUUID->"e6260682-f6b7-489c-a22c-41e8e5982d10"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"r", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`r[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781408678493`*^9},
 CellLabel->"During evaluation of In[488]:=",
 CellID->1934196241,ExpressionUUID->"afd6daf7-a429-452a-a5b5-896fa560222f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"7", 
    RowBox[{"r", "[", "]"}], 
    RowBox[{"v", "[", "a", "]"}]}], "+", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[489]:=",ExpressionUUID->"ffa63eb5-6b01-43dc-9205-460059e34ac3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b", "  "},
           {" ", " ", 
            RowBox[{";", "a"}]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Cd[-$CellContext`a][
     $CellContext`S[$CellContext`a, $CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{"7", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`Cd[-$CellContext`a][
         $CellContext`r[]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a", "  "},
              {" ", 
               RowBox[{";", "a"}]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Cd[-$CellContext`a][
        $CellContext`v[$CellContext`a]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" ", "  "},
           {"b", 
            RowBox[{";", "a"}]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Cd[-$CellContext`a][
     $CellContext`v[-$CellContext`b]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978141007257*^9},
 CellLabel->"Out[489]=",
 CellID->1576257848,ExpressionUUID->"246a090c-410f-4981-bd83-ef2ef0a5771a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Dummy indices are, by default, pushed to the right by the canonicalization \
process\
\>", "MathCaption",ExpressionUUID->"b6ffecf8-5711-4d7b-9d33-f07334d74bc2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Cd", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"f", ",", "c", ",", 
      RowBox[{"-", "b"}]}], "]"}], 
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "e"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[490]:=",ExpressionUUID->"dba33b63-46ec-4d6f-a1e2-3175425f57df"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"d", "e", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`c],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f", "c", " ", "  "},
           {" ", " ", "b", 
            RowBox[{";", "a"}]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Cd[-$CellContext`a][
     $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f", "c", " "},
           {" ", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " ", "  "},
           {"d", "e", "c", 
            RowBox[{";", "a"}]}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Cd[-$CellContext`a][
     $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`c]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289781411674137`*^9},
 CellLabel->"Out[490]=",
 CellID->807381944,ExpressionUUID->"8775fe9d-80df-4362-a95f-7acbc7bb21ce"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["SortCovDs",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SortCovDs.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Commute covariant derivatives, adding Riemann tensors if needed\n",
   "Text"],
 StyleBox[ButtonBox["SortCovDsStart",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SortCovDsStart.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Automatic commutation of covariant derivatives\n", "Text"],
 StyleBox[ButtonBox["SortCovDsStop",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SortCovDsStop.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Remove automatic commutation of covariant derivatives\n", "Text"],
 StyleBox["$CommuteCovDsOnScalars", "Text",
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Automatic commutation of torsionless derivatives on scalar \
expressions", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.728978128758627*^9, 
  3.7289781307343807`*^9}},ExpressionUUID->"f65163c3-72a3-4923-81ba-\
ff7d40cdd0fe"],

Cell["Commutation of derivatives.", "Text",ExpressionUUID->"e22ac9f9-cb6a-4b37-82c8-f23370c6f1a4"],

Cell[CellGroupData[{

Cell[TextData[{
 "By default, a special ordinary derivative called ",
 StyleBox["PD",
  FontFamily->"Courier"],
 " is defined. It is one of the many covariant derivative operators without \
curvature and without torsion, but no further assumptions are made about it. \
It can be used as a generic \"partial derivative\" in most computations, but \
its meaning is different. Note that they are not automatically sorted:"
}], "MathCaption",ExpressionUUID->"6545f5fa-7433-469b-9daf-e26376c5d2e5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"T", "[", 
    RowBox[{"c", ",", "d", ",", 
     RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[491]:=",ExpressionUUID->"817c56c9-5373-493c-aa9e-c20eebee77c0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {"c", "d", " ", "  ", "  "},
         {" ", " ", "e", 
          RowBox[{",", "b"}], 
          RowBox[{",", "a"}]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978141316661*^9},
 CellLabel->"Out[491]=",
 CellID->1866651062,ExpressionUUID->"0c12541e-152d-4186-8882-b25d2d510593"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$CovDFormat", "=", "\"\<Prefix\>\""}], ";"}]], "Input",
 CellLabel->
  "In[492]:=",ExpressionUUID->"e451f13c-3f44-48b8-a035-2f42166c3781"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"T", "[", 
    RowBox[{"c", ",", "d", ",", 
     RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[493]:=",ExpressionUUID->"2d69238a-5569-4699-b795-867d9f930362"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d", " "},
          {" ", " ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897814150274*^9},
 CellLabel->"Out[493]=",
 CellID->215800011,ExpressionUUID->"5a8f21fc-a918-4de0-9f66-8a04d2768034"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "b"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{"c", ",", "d", ",", 
      RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}], "-", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{"c", ",", "d", ",", 
      RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[494]:=",ExpressionUUID->"a07b2027-2e3f-4f6e-bc3c-9fb68c4c6a95"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], 
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", "d", " "},
           {" ", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], 
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", "d", " "},
           {" ", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978141627575*^9},
 CellLabel->"Out[494]=",
 CellID->1998388741,ExpressionUUID->"be22bbe1-fb5e-4017-bc50-9ec38c5ab479"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "SortCovDs"}]], "Input",
 CellLabel->
  "In[495]:=",ExpressionUUID->"c6d10605-96fd-4ff9-b77d-d740b09db4a0"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289781417706337`*^9},
 CellLabel->"Out[495]=",
 CellID->275330907,ExpressionUUID->"b7fcd682-e328-4698-b68d-15bb164da53e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can automate the commutation of derivatives:", "MathCaption",ExpressionUUID->"f892c410-0d1a-43e7-a4d5-daa994c70d5b"],

Cell[BoxData[
 RowBox[{"SortCovDsStart", "[", "PD", "]"}]], "Input",
 CellLabel->
  "In[496]:=",ExpressionUUID->"8fa5c7c0-c485-4d3f-b3df-72a62097e3d1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "b"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{"c", ",", "d", ",", 
      RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}], "-", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{"c", ",", "d", ",", 
      RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[497]:=",ExpressionUUID->"6f1d57b6-1f7b-403f-ad59-c8d4bf9a7ad2"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978141948246*^9},
 CellLabel->"Out[497]=",
 CellID->1806778311,ExpressionUUID->"2af75073-88e2-4f83-a4e7-00419b4aa9d8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SortCovDsStop", "[", "PD", "]"}]], "Input",
 CellLabel->
  "In[498]:=",ExpressionUUID->"62d34abf-293e-43a8-a5f8-476d69a2b843"],

Cell[BoxData["\<\"Note that $CommuteCovDsOnScalars is still True.\"\>"], \
"Print",
 CellChangeTimes->{3.7289781420650454`*^9},
 CellLabel->"During evaluation of In[498]:=",
 CellID->1104304397,ExpressionUUID->"94d607ee-f870-4df3-902f-f1480707a58a"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "All symmetric (torsionless) connections commute (with themselves) on scalar \
expressions. That can be avoided switching off the variable ",
 StyleBox["$CommuteCovDsOnScalars",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"87dc076f-3fd8-4227-aa28-3d8c46d85766"],

Cell[TextData[{
 "The input of high-order derivatives is certainly slow due to the sharp \
syntactic restrictions of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ". There are several options to avoid this problem. Of course, you cannot \
use ",
 StyleBox["Validate",
  FontFamily->"Courier"],
 " unless you modify it to recognize the new syntactic extensions."
}], "Text",ExpressionUUID->"0c136a5f-4137-4d8a-83ed-c7c82ab6723f"],

Cell[CellGroupData[{

Cell["The standard way to input a third order derivative is this:", \
"MathCaption",ExpressionUUID->"ae199384-d3a8-499c-b3df-56931f3c3385"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "c"}], "]"}], "[", 
    RowBox[{"T", "[", 
     RowBox[{"c", ",", "d", ",", 
      RowBox[{"-", "e"}]}], "]"}], "]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[499]:=",ExpressionUUID->"aa5e924f-2bbe-4966-b485-0d530a484af7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], 
    SubscriptBox["\[PartialD]", "c"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d", " "},
          {" ", " ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781421922483`*^9},
 CellLabel->"Out[499]=",
 CellID->1490323005,ExpressionUUID->"f4f2fb1d-b3a8-4cf4-bb02-384925fdd669"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We can use the prefix notation of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ":"
}], "MathCaption",ExpressionUUID->"86d2b13f-db6c-4d81-876a-7cbace375d4e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "c"}], "]"}], "@", 
    RowBox[{"T", "[", 
     RowBox[{"c", ",", "d", ",", 
      RowBox[{"-", "e"}]}], "]"}]}]}]}]], "Input",
 CellLabel->
  "In[500]:=",ExpressionUUID->"598ec24d-3e7b-458a-9027-e398b0c8f8a7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], 
    SubscriptBox["\[PartialD]", "c"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d", " "},
          {" ", " ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781424676237`*^9},
 CellLabel->"Out[500]=",
 CellID->1756970074,ExpressionUUID->"f2830ff8-bb57-4961-a302-d2d3ed13e7ca"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "or we can define our own rules to input derivatives, breaking the standard \
syntax. The simplest possibility, based on the fact that the first bracket of \
",
 StyleBox["PD",
  FontFamily->"Courier"],
 " always contains a single index, is:"
}], "MathCaption",ExpressionUUID->"2115c7f8-e577-4ce7-bd2e-fbdf324e6ec2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Unprotect", "[", "PD", "]"}], ";"}]], "Input",
 CellLabel->
  "In[501]:=",ExpressionUUID->"ae8b5f7c-73a3-4ea9-ae06-4955efc4d9f1"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"expr_", ",", "a__", ",", "b_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"PD", "[", "b", "]"}], "[", 
   RowBox[{"PD", "[", 
    RowBox[{"expr", ",", "a"}], "]"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"expr_", ",", "a_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"PD", "[", "a", "]"}], "[", "expr", "]"}]}]}], "Input",
 CellLabel->
  "In[502]:=",ExpressionUUID->"09cf970e-1e3a-475a-aae7-94d2903187f5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PD", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"c", ",", "d", ",", 
     RowBox[{"-", "e"}]}], "]"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "a"}]}], "]"}]], "Input",
 CellLabel->
  "In[504]:=",ExpressionUUID->"cb12f830-4d59-4eb9-bfe1-d7f00eb04e5b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], 
    SubscriptBox["\[PartialD]", "c"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d", " "},
          {" ", " ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781430683947`*^9},
 CellLabel->"Out[504]=",
 CellID->1495887468,ExpressionUUID->"22af3180-353d-45ce-a6df-0041138b3307"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[505]:=",ExpressionUUID->"cd871c40-4d76-4629-bd46-2e366e656c92"],

Cell["PD[-a][PD[-b][PD[-c][T[c, d, -e]]]]", "Output",
 CellChangeTimes->{3.7289781432546597`*^9},
 CellLabel->"Out[505]//InputForm=",
 CellID->789619247,ExpressionUUID->"9e7821dd-d848-4067-8f29-5cc29d521e47"]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"expr_", ",", "a__", ",", "b_"}], "]"}], 
  "=."}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"expr_", ",", "a_"}], "]"}], "=."}]}], "Input",
 CellLabel->
  "In[506]:=",ExpressionUUID->"f0c68c66-1bc1-451a-96d4-7d637db6ec55"]
}, Open  ]],

Cell[CellGroupData[{

Cell["or an intermediate situation, based on the same fact, is:", \
"MathCaption",ExpressionUUID->"f96241e0-3cdd-463a-9627-012c631ae4a9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"a_", ",", "b__"}], "]"}], "[", "expr_", "]"}], ":=", 
  RowBox[{
   RowBox[{"PD", "[", "a", "]"}], "[", 
   RowBox[{
    RowBox[{"PD", "[", "b", "]"}], "[", "expr", "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[508]:=",ExpressionUUID->"76e0b143-41c9-4db1-bc86-5d6a022f05ce"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "[", 
  RowBox[{"T", "[", 
   RowBox[{"c", ",", "d", ",", "e"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[509]:=",ExpressionUUID->"7b61996f-0875-47e9-b8fa-abd87cfe5b2c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], 
    SubscriptBox["\[PartialD]", "c"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d", "e"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`T[$CellContext`c, $CellContext`d, $CellContext`e]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978143950625*^9},
 CellLabel->"Out[509]=",
 CellID->1829229516,ExpressionUUID->"a1fd4b0e-8415-48a6-b1e9-4d010ae7d5a1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[510]:=",ExpressionUUID->"85d1f6dc-80e6-4f0d-b79c-37a788d7dac4"],

Cell["PD[-a][PD[-b][PD[-c][T[c, d, e]]]]", "Output",
 CellChangeTimes->{3.728978144255743*^9},
 CellLabel->"Out[510]//InputForm=",
 CellID->484584960,ExpressionUUID->"6008a6ed-d7a1-4b74-ac5f-e37b8abdaaf3"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"a_", ",", "b__"}], "]"}], "[", "expr_", "]"}], "=."}]], "Input",
 CellLabel->
  "In[511]:=",ExpressionUUID->"e4d458cf-7efc-467c-8751-1cdcb9aaa662"]
}, Open  ]],

Cell[CellGroupData[{

Cell["A totally different notation could be", "MathCaption",ExpressionUUID->"4098e5a8-e662-4cbe-be9d-c5af2811a9c6"],

Cell[BoxData[
 RowBox[{"PD", "/:", 
  RowBox[{
   RowBox[{"tensor_", "?", "xTensorQ"}], "[", 
   RowBox[{"indices___", ",", 
    RowBox[{"PD", "[", "index_", "]"}]}], "]"}], ":=", 
  RowBox[{
   RowBox[{"PD", "[", "index", "]"}], "[", 
   RowBox[{"tensor", "[", "indices", "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[512]:=",ExpressionUUID->"2d0b9916-e868-43ec-890c-c972051da3df"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"T", "[", 
  RowBox[{"c", ",", "d", ",", 
   RowBox[{"-", "e"}], ",", 
   RowBox[{"PD", "[", 
    RowBox[{"-", "c"}], "]"}], ",", 
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], ",", 
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[513]:=",ExpressionUUID->"9057620d-db5a-4ddc-a2a0-c8a8cd4ee83b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], 
    SubscriptBox["\[PartialD]", "c"], GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d", " "},
          {" ", " ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`e]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978144719783*^9},
 CellLabel->"Out[513]=",
 CellID->1161433296,ExpressionUUID->"fe840463-c02e-4e43-af91-ea2351e2e015"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[514]:=",ExpressionUUID->"63f5cfad-18f7-4a59-95e8-0cc0a20988bb"],

Cell["PD[-a][PD[-b][PD[-c][T[c, d, -e]]]]", "Output",
 CellChangeTimes->{3.7289781448315496`*^9},
 CellLabel->"Out[514]//InputForm=",
 CellID->1730538289,ExpressionUUID->"49146ef4-877b-4f11-a6d9-240c20e3d7e2"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"PD", "/:", 
  RowBox[{
   RowBox[{
    RowBox[{"tensor_", "?", "xTensorQ"}], "[", 
    RowBox[{"indices___", ",", 
     RowBox[{"PD", "[", "index_", "]"}]}], "]"}], "=."}]}]], "Input",
 CellLabel->
  "In[515]:=",ExpressionUUID->"839585ee-de19-4e04-a256-5bd736244d5b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Protect", "[", "PD", "]"}], ";"}]], "Input",
 CellLabel->
  "In[516]:=",ExpressionUUID->"bf8281a6-bd4e-4859-a39f-4cc71ff38c03"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.2A. Implosion and explosion (I)", "Subsection",ExpressionUUID->"623c0ab1-7911-4324-a525-e01e0bc7d55a"],

Cell[TextData[{
 "It is important to understand the following notational problem with \
derivatives. From a mathematical point of view, the derivative of a function \
or a tensor is another function or tensor, and hence should share its same \
notation. However, we must input the derivative using an operator, and hence \
a notational change should happen at some point. For instance, in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " the operator ",
 StyleBox["D[f[x], x]",
  FontFamily->"Courier"],
 " is inmediately converted into ",
 StyleBox["Derivative[1][f][x]",
  FontFamily->"Courier"],
 ". The former doesn't have the \"function form\" head[x] but the latter \
does, with the head being ",
 StyleBox["Derivative[1][f]",
  FontFamily->"Courier"],
 ". There are many obvious advantages in this change, because now we can use \
a consistent treatment for all functions, but there are also disadvantages."
}], "Text",ExpressionUUID->"eb26b2ac-882a-4352-ba13-ad0d7d6c506b"],

Cell[CellGroupData[{

Cell[TextData[{
 "The main disadvantage of the change is, a mystery for all new users of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", that this does not work:"
}], "MathCaption",ExpressionUUID->"dcf035c9-a4a6-4cd0-ada7-7123834233bf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"D", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", "x"}], "]"}], "/.", 
  RowBox[{
   RowBox[{"f", "[", "x", "]"}], "\[Rule]", 
   RowBox[{"Sin", "[", "x", "]"}]}]}]], "Input",
 CellLabel->
  "In[517]:=",ExpressionUUID->"e12f2e99-61bb-4339-9c6a-c7aea3b99ca1"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["f", "\[Prime]",
   MultilineFunction->None], "[", "x", "]"}]], "Output",
 CellChangeTimes->{3.7289781503158827`*^9},
 CellLabel->"Out[517]=",
 CellID->924892270,ExpressionUUID->"941f0f12-57b8-4718-8ee9-11cf6d6c0ea5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but this does work:", "MathCaption",ExpressionUUID->"2ab14c93-829b-4394-bee6-c86593cd9568"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"D", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", "x"}], "]"}], "/.", 
  RowBox[{"f", "\[Rule]", "Sin"}]}]], "Input",
 CellLabel->
  "In[518]:=",ExpressionUUID->"56ab050b-b341-4396-9fa1-24241c7b76bd"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox["Cos",
    ShowAutoStyles->False,
    AutoSpacing->False],
   Cos,
   Editable->False], "[", "x", "]"}]], "Output",
 CellChangeTimes->{3.72897815043445*^9},
 CellLabel->"Out[518]=",
 CellID->1177744904,ExpressionUUID->"241ad05e-7aee-44e0-be72-8d1d2fe72103"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " we work essentially with the operator form only, and hence rules are \
simpler to use, but we have the problem that tensors and their covariant \
derivatives are treated differently. To alleviate that problem, a new device \
has been introduced in version 0.9.6: we now have the concepts of \"implosion\
\" and \"explosion\":"
}], "Text",ExpressionUUID->"e031fd1e-5827-4acf-9041-55f07c2bd855"],

Cell[CellGroupData[{

Cell["Suppose we start from the covariant derivative", "MathCaption",ExpressionUUID->"76c11376-99af-475d-bdfd-1aa0c00a557c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"v", "[", "b", "]"}], "]"}]], "Input",
 CellLabel->
  "In[519]:=",ExpressionUUID->"a97802de-a084-4d04-bde9-b07c9efd428c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978150555696*^9},
 CellLabel->"Out[519]=",
 CellID->1001827129,ExpressionUUID->"744c6fbc-0743-4ab0-be70-a40b82d01207"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[520]:=",ExpressionUUID->"a887fb36-ce66-4fd5-a8d1-1639b9ee8e2b"],

Cell["Cd[-a][v[b]]", "Output",
 CellChangeTimes->{3.728978150700719*^9},
 CellLabel->"Out[520]//InputForm=",
 CellID->1436764623,ExpressionUUID->"108772cf-0da4-44c8-bc0b-5a263bff65d2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now we \"implode\" this expression, resulting in a new tensor being defined \
(there is no message reporting the creation of this new tensor, but that can \
be activated setting the global variable ",
 StyleBox["$ImplodeInfo = Info",
  FontFamily->"Courier"],
 "). Compare with the previous expression, and please pay attention to the \
positions of indices!"
}], "MathCaption",ExpressionUUID->"c7843ef5-1860-4f66-9502-814c048d9348"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Implode", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[521]:=",ExpressionUUID->"9b15c20e-3fa7-4ab8-a0a8-f1bd3dd29509"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[EmptyDownTriangle]v", 
      StyleBox[GridBox[{
         {"b", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cdv[$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781509017487`*^9},
 CellLabel->"Out[521]=",
 CellID->1396891893,ExpressionUUID->"20dfee45-0ddd-4ec7-b93c-2b980c7ce6a9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[522]:=",ExpressionUUID->"e804fbe0-6a57-4d93-b7d0-14c10680008c"],

Cell["Cdv[b, -a]", "Output",
 CellChangeTimes->{3.728978151272244*^9},
 CellLabel->"Out[522]//InputForm=",
 CellID->1270816647,ExpressionUUID->"71f459c9-3ff4-4636-9890-9d41b534d7d5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The expression can be \"exploded\", returning to the original:", \
"MathCaption",ExpressionUUID->"7644152a-ecd6-46cf-9757-147aeba88a85"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Explode", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[523]:=",ExpressionUUID->"665521ea-e3cf-494c-b365-88273943fdb6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978151562675*^9},
 CellLabel->"Out[523]=",
 CellID->1890516714,ExpressionUUID->"cebb9864-509c-45ba-bff7-716b8a63310f"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Implosion involves the construction of new tensors, and in this process their \
symmetries must also be constructed.\
\>", "Text",ExpressionUUID->"88dfefb5-59e8-445a-8315-8e941aad7a00"],

Cell[CellGroupData[{

Cell["This object is symmetric:", "MathCaption",ExpressionUUID->"0ffd951b-f96a-4851-b6aa-c5da79045b99"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{"v", "[", "c", "]"}]}]}]], "Input",
 CellLabel->
  "In[524]:=",ExpressionUUID->"efb583e9-bc83-4437-9535-d23549aa543b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], 
    SubscriptBox["\[PartialD]", "b"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`v[$CellContext`c]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978152136099*^9},
 CellLabel->"Out[524]=",
 CellID->1893854833,ExpressionUUID->"79d3dd44-ff73-439e-9823-150bdec8e1b9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[525]:=",ExpressionUUID->"9eee5ff2-7aa0-4162-9e98-1afe1e6c19ef"],

Cell[BoxData[
 RowBox[{"Symmetry", "[", 
  RowBox[{"3", ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "\[FilledCircle]3"], 
      SuperscriptBox["\[PartialD]", "\[FilledCircle]2"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"\[FilledCircle]1"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[
     xAct`xTensor`Private`slot[3]][
     xAct`xTensor`PD[
      xAct`xTensor`Private`slot[2]][
      $CellContext`v[
       xAct`xTensor`Private`slot[1]]]],
    Editable->False], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\<\"\[FilledCircle]1\"\>", "\[Rule]", "c"}], ",", 
     RowBox[{"\<\"\[FilledCircle]2\"\>", "\[Rule]", 
      RowBox[{"-", "b"}]}], ",", 
     RowBox[{"\<\"\[FilledCircle]3\"\>", "\[Rule]", 
      RowBox[{"-", "a"}]}]}], "}"}], ",", 
   RowBox[{"StrongGenSet", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"2", ",", "3"}], "}"}], ",", 
     RowBox[{"GenSet", "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{"(", "2", ",", "3", ")"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xPerm`Cycles[{2, 3}],
       Editable->False], "]"}]}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728978152310381*^9},
 CellLabel->"Out[525]=",
 CellID->15562397,ExpressionUUID->"30784372-2e76-40d5-86c6-45f9f31e3a89"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If we implode it, that symmetry must be stored into the new tensor:\
\>", "MathCaption",ExpressionUUID->"df692ab9-a1de-4763-b6dd-39c6431988d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Implode", "[", "%%", "]"}]], "Input",
 CellLabel->
  "In[526]:=",ExpressionUUID->"018784fc-c821-4c24-b466-84c7cb94a582"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[PartialD]", 
       RowBox[{"\[PartialD]", "v"}]}], 
      StyleBox[GridBox[{
         {"c", " ", " "},
         {" ", "b", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`PDPDv[$CellContext`c, -$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978152400223*^9},
 CellLabel->"Out[526]=",
 CellID->1711685092,ExpressionUUID->"a22045d5-b27c-495f-b499-88991547c7f4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "PDPDv", "]"}]], "Input",
 CellLabel->
  "In[527]:=",ExpressionUUID->"b65543ae-a0c0-4284-a692-5d02b46c2acc"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "2", "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "2", ",", "3", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{2, 3}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728978152555881*^9},
 CellLabel->"Out[527]=",
 CellID->600406032,ExpressionUUID->"601a9c67-f142-483c-b333-3e4541bdd2e7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Other torsionless derivatives are commuted on scalars :", "MathCaption",ExpressionUUID->"a68a8b7e-af63-4ec3-bab5-fd3a06c6bedb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"Cd", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{"r", "[", "]"}]}]}]], "Input",
 CellLabel->
  "In[528]:=",ExpressionUUID->"a36d5b22-f1a2-40f0-99ed-292b4a891145"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "b"], 
    SubscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`b][
   $CellContext`Cd[-$CellContext`a][
    $CellContext`r[]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781526858187`*^9},
 CellLabel->"Out[528]=",
 CellID->2146006218,ExpressionUUID->"2383a42b-4e46-4be4-b4c9-2bc444c91edf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Implode", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[529]:=",ExpressionUUID->"1abdb638-1d99-4617-9ba2-2ce841c52d23"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[EmptyDownTriangle]\[EmptyDownTriangle]r", 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CdCdr[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978152855795*^9},
 CellLabel->"Out[529]=",
 CellID->938654561,ExpressionUUID->"0c5695c5-afdf-4c99-ae6b-d7f345441654"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "CdCdr", "]"}]], "Input",
 CellLabel->
  "In[530]:=",ExpressionUUID->"a0398411-ad29-44fc-a611-4aead834b4c2"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "1", "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"(", "1", ",", "2", ")"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728978153005115*^9},
 CellLabel->"Out[530]=",
 CellID->552335865,ExpressionUUID->"f47f15ea-9baf-4f11-a1b6-f677918d572c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but not on other tensors :", "MathCaption",ExpressionUUID->"cbb069d9-e30c-4c68-bac6-902658a76d5c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "@", 
  RowBox[{
   RowBox[{"Cd", "[", 
    RowBox[{"-", "b"}], "]"}], "@", 
   RowBox[{"v", "[", "a", "]"}]}]}]], "Input",
 CellLabel->
  "In[531]:=",ExpressionUUID->"a7c74cf7-e5d2-46e1-8ed1-7af7a0da271b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], 
    SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`Cd[-$CellContext`b][
    $CellContext`v[$CellContext`a]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978153143219*^9},
 CellLabel->"Out[531]=",
 CellID->1684560721,ExpressionUUID->"acb0fc5b-7b3b-407c-94e1-b4856b690484"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Implode", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[532]:=",ExpressionUUID->"eefde0e9-2758-4271-838d-f3a0430fabd6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[EmptyDownTriangle]\[EmptyDownTriangle]v", 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CdCdv[$CellContext`a, -$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978153285036*^9},
 CellLabel->"Out[532]=",
 CellID->2135119446,ExpressionUUID->"4350cd76-32f2-4759-93d4-131c7af75868"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", "CdCdv", "]"}]], "Input",
 CellLabel->
  "In[533]:=",ExpressionUUID->"eb382130-444a-426a-82cd-19c7f2ed328e"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", "}"}], ",", 
   RowBox[{"GenSet", "[", "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7289781534031467`*^9},
 CellLabel->"Out[533]=",
 CellID->500787552,ExpressionUUID->"1aa56282-833f-42bc-9115-a091b6d2fe4b"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The problem of implosion becomes more complicated when metrics (and bases) \
are present, but we defer the corresponding analysis to the next section, \
after metrics are introduced.\
\>", "Text",ExpressionUUID->"558fd9c3-6418-429d-be70-8e1ec50d84cb"]
}, Closed]],

Cell[CellGroupData[{

Cell["6.2B. TensorDerivative", "Subsection",ExpressionUUID->"29c87aca-5f16-4fff-b041-275a77e254cf"],

Cell[TextData[{
 "Now that we can have tensors with non-atomic heads, it is natural to absorb \
the imploded information in the head of a new tensor. Following ",
 StyleBox["Derivative",
  FontFamily->"Courier"],
 " in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " we introduce the head ",
 StyleBox["TensorDerivative",
  FontFamily->"Courier"],
 "."
}], "Text",
 CellChangeTimes->{{3.4946100213772717`*^9, 
  3.494610102092366*^9}},ExpressionUUID->"560f40ba-9226-426f-b4a5-\
a4a38879a4e7"],

Cell[CellGroupData[{

Cell["Take a tensor expression containing derivatives:", "MathCaption",
 CellChangeTimes->{{3.494610124243494*^9, 
  3.49461014104458*^9}},ExpressionUUID->"d966590a-008b-49b2-8ddb-\
cdbb37ded23e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"v", "[", "b", "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.494610144033408*^9, 3.4946101469465446`*^9}},
 CellLabel->
  "In[534]:=",ExpressionUUID->"68c792cf-28eb-4f5b-a6a0-a7c42b4ee4d3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978156767712*^9},
 CellLabel->"Out[534]=",
 CellID->1246017163,ExpressionUUID->"25810f6e-c342-4a75-8018-22c5cad3fa73"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Convert it into the new notation (note that the derivative index is placed \
last):\
\>", "MathCaption",
 CellChangeTimes->{{3.49461018389324*^9, 3.494610196594267*^9}, {
  3.494610236120812*^9, 
  3.4946102473986273`*^9}},ExpressionUUID->"2726d7a2-4784-4842-94fe-\
da05be0357d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToTensorDerivative"}]], "Input",
 CellChangeTimes->{{3.4946101499036083`*^9, 3.49461015326237*^9}},
 CellLabel->
  "In[535]:=",ExpressionUUID->"984c6475-63f4-4dc1-9e32-7018e2b67f65"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[EmptyDownTriangle]", "v"}], 
      StyleBox[GridBox[{
         {"b", " "},
         {" ", "a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`TensorDerivative[$CellContext`v, \
$CellContext`Cd][$CellContext`b, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781569040737`*^9},
 CellLabel->"Out[535]=",
 CellID->1697416644,ExpressionUUID->"b192aae8-e1f0-4e0a-a8c9-3f0467b0bb9a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellChangeTimes->{{3.494610156694859*^9, 3.494610158619482*^9}},
 CellLabel->
  "In[536]:=",ExpressionUUID->"f558af18-ca8d-4dc8-98bb-e45b11df316b"],

Cell["TensorDerivative[v, Cd][b, -a]", "Output",
 CellChangeTimes->{3.728978157018217*^9},
 CellLabel->"Out[536]//InputForm=",
 CellID->1308768960,ExpressionUUID->"de8e1f6e-3944-4c29-be89-ccb32f8b2b58"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and back:", "MathCaption",
 CellChangeTimes->{{3.494610199034997*^9, 
  3.494610209232727*^9}},ExpressionUUID->"b8eb3c01-ff30-4d5e-b7ec-\
aca046fb693d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "FromTensorDerivative"}]], "Input",
 CellChangeTimes->{{3.4946101667558193`*^9, 3.494610172783784*^9}},
 CellLabel->
  "In[537]:=",ExpressionUUID->"c75266c6-326c-48bc-9c4f-37ac00cd609b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978157105207*^9},
 CellLabel->"Out[537]=",
 CellID->1952398146,ExpressionUUID->"5dc65c81-cc8d-4201-976b-c1a398f57b46"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellChangeTimes->{{3.49461017482341*^9, 3.494610176830839*^9}},
 CellLabel->
  "In[538]:=",ExpressionUUID->"9468b8ca-fe1c-4b04-b264-4513f4d6dc08"],

Cell["Cd[-a][v[b]]", "Output",
 CellChangeTimes->{3.728978157219108*^9},
 CellLabel->"Out[538]//InputForm=",
 CellID->405513662,ExpressionUUID->"b39ba3fd-964c-4b2a-9528-b29980bfc0c6"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.3. Change of covariant derivative", "Subsection",ExpressionUUID->"866d0406-c99f-4c83-873e-360202063718"],

Cell[TextData[{
 "In ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " we can work simultaneously with any number of covariant derivatives."
}], "Text",ExpressionUUID->"9bb62229-a8dc-4dfc-8419-8f81c4e38495"],

Cell[CellGroupData[{

Cell["The list of all covariant derivatives currently defined is", \
"MathCaption",ExpressionUUID->"15d5fd4f-e3f3-47c2-8059-f99ff94332a4"],

Cell[CellGroupData[{

Cell[BoxData["$CovDs"], "Input",
 CellLabel->
  "In[539]:=",ExpressionUUID->"800cb60b-a217-4d01-bb9d-ed6f7be38204"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"PD", ",", "Cd"}], "}"}]], "Output",
 CellChangeTimes->{3.7289781598236094`*^9},
 CellLabel->"Out[539]=",
 CellID->493916519,ExpressionUUID->"08d03c21-68a5-4b3a-97b0-ee5f17c35de7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We define another one", "MathCaption",ExpressionUUID->"20778d52-197f-4236-b331-75f1dceb6a88"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<|\>\"", ",", "\"\<D\>\""}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[540]:=",ExpressionUUID->"eef1e896-72f1-48a9-886f-364dbeb487a6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978160006521*^9},
 CellLabel->"During evaluation of In[540]:=",
 CellID->1230639178,ExpressionUUID->"62ef7723-3f3a-45c2-9812-e8f245de7356"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781600260267`*^9},
 CellLabel->"During evaluation of In[540]:=",
 CellID->937660389,ExpressionUUID->"ba85928e-6ba1-4bc3-96f9-436e1de61ef1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978160043438*^9},
 CellLabel->"During evaluation of In[540]:=",
 CellID->1023156237,ExpressionUUID->"25c01830-08b5-4b8c-bfff-963d26ff1d54"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978160061049*^9},
 CellLabel->"During evaluation of In[540]:=",
 CellID->1285313730,ExpressionUUID->"946f7961-4d4f-443d-a63e-85fb8fe18067"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781600775642`*^9},
 CellLabel->"During evaluation of In[540]:=",
 CellID->729854937,ExpressionUUID->"ddd3f2d7-51b8-4ea6-81eb-d87f3de9c05e"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728978160094405*^9},
 CellLabel->"During evaluation of In[540]:=",
 CellID->245334495,ExpressionUUID->"71a25a5e-ed64-4b2f-84ee-713626eafb25"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The difference between two covariant derivatives of the same tensor can be \
expressed in terms of Christoffel tensors (the C tensors of Wald), and in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " we shall always use this point of view of Christoffels: they are always \
tensors, but associated to two covariant derivatives. The Christoffel tensor \
defined together with each derivative is the tensor associated to that \
derivative and the derivative ",
 StyleBox["PD",
  FontFamily->"Courier"],
 " associated to a generic chart."
}], "Text",ExpressionUUID->"b7a2707d-2bdc-47a7-81e8-6fb49f7db199"],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["Christoffel",
  FontFamily->"Courier"],
 " constructs the Christoffel tensor relating two derivatives. It is \
antisymmetric in its derivatives arguments."
}], "MathCaption",ExpressionUUID->"eada7ce8-a4e4-4f27-942d-4d48b901e055"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Christoffel", "[", 
   RowBox[{"CD", ",", "PD"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[541]:=",ExpressionUUID->"7e2009d1-8995-4aea-9c32-b1bdf719a590"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
      StyleBox[GridBox[{
         {"a", " ", " "},
         {" ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781602220078`*^9},
 CellLabel->"Out[541]=",
 CellID->168255225,ExpressionUUID->"93b899ee-ad4d-4416-90bc-822979752576"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[542]:=",ExpressionUUID->"02640ad5-5ccc-4b58-9f6a-da64b21b8842"],

Cell["ChristoffelCD[a, -b, -c]", "Output",
 CellChangeTimes->{3.72897816052466*^9},
 CellLabel->"Out[542]//InputForm=",
 CellID->1777087123,ExpressionUUID->"acd465b1-2b15-49d0-8e19-97a9d723ddea"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Christoffel", "[", 
   RowBox[{"PD", ",", "CD"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[543]:=",ExpressionUUID->"98e0657a-f42c-408b-96d0-60f4b1f9c195"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {"a", " ", " "},
          {" ", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289781609415693`*^9},
 CellLabel->"Out[543]=",
 CellID->300713785,ExpressionUUID->"b4e9b041-fef4-45fb-be1b-20a234668e0a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[544]:=",ExpressionUUID->"2096d0d6-4b66-4c78-8179-41712d33dac7"],

Cell["-ChristoffelCD[a, -b, -c]", "Output",
 CellChangeTimes->{3.72897816145223*^9},
 CellLabel->"Out[544]//InputForm=",
 CellID->1067563104,ExpressionUUID->"e8f939e0-ef11-451c-9994-c353a6e4b13c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The Christoffel tensor relating two non-",
 StyleBox["PD",
  FontFamily->"Courier"],
 " derivatives is constructed automatically whenever needed, with \
(lexicographically) sorted derivatives, to avoid duplicity:"
}], "MathCaption",ExpressionUUID->"9dc2606d-d7bd-4526-8f2a-8c7a511fd7d8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Christoffel", "[", 
   RowBox[{"CD", ",", "Cd"}], "]"}], "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[545]:=",ExpressionUUID->"656b1ab0-0bfa-4479-af16-028a1b73b379"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCdCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897816163828*^9},
 CellLabel->"During evaluation of In[545]:=",
 CellID->307218243,ExpressionUUID->"9e87e89a-8430-445e-a1bc-2312ba82e546"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
       StyleBox[GridBox[{
          {"a", " ", " "},
          {" ", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978161658037*^9},
 CellLabel->"Out[545]=",
 CellID->866322713,ExpressionUUID->"493384bd-0716-4b38-98aa-cb492e52a857"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[546]:=",ExpressionUUID->"70af6ea8-545e-43e4-8cf7-d478db21dead"],

Cell["-ChristoffelCdCD[a, -b, -c]", "Output",
 CellChangeTimes->{3.728978162027939*^9},
 CellLabel->"Out[546]//InputForm=",
 CellID->346710731,ExpressionUUID->"52fb7749-5c2a-469e-a313-fd5734339108"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The origin of this curious conversion from ",
 StyleBox["Christoffel[Cd,CD]",
  FontFamily->"Courier"],
 " to ",
 StyleBox["ChristoffelCdCD",
  FontFamily->"Courier"],
 " is again the fact that we cannot associate information to the former."
}], "MathCaption",ExpressionUUID->"3626e143-dd10-46f4-b73d-1fe9e5135ee5"],

Cell["\<\
Using this structure of Christoffel tensors we can relate any two derivatives \
of any tensor. \
\>", "Text",ExpressionUUID->"4b94c430-a305-4cb6-bff3-be761e76f09b"],

Cell[CellGroupData[{

Cell["Suppose this derivative:", "MathCaption",ExpressionUUID->"effccc54-4ffe-4c98-a97b-2c947951a2b6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "d"}], "]"}], "[", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[547]:=",ExpressionUUID->"84f98ec1-d558-4daf-9d54-3f914e11f7ae"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "d"], GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", "b", " "},
          {"a", " ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`d][
   $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978162137796*^9},
 CellLabel->"Out[547]=",
 CellID->1044121606,ExpressionUUID->"5072d705-595a-4c32-a1b6-26073f5e97d9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ChangeCovD",
  FontFamily->"Courier"],
 " (aka ",
 StyleBox["CovDToChristoffel",
  FontFamily->"Courier"],
 " in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " version 0.7) changes by default to ",
 StyleBox["PD",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"fad22346-8041-4f3d-8db0-de8311109d4e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[548]:=",ExpressionUUID->"1d333c1f-7944-482a-8623-978473b00aea"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "d", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`d, \
-$CellContext`c],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`b, -$CellContext`d, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "e", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`e, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`d, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"e", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`e, $CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978162453116*^9},
 CellLabel->"Out[548]=",
 CellID->1462952405,ExpressionUUID->"ba19239f-437e-4c47-8091-fced0020753f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["That is equivalent to", "MathCaption",ExpressionUUID->"fcc2196f-c91b-4472-b894-653fd7d31cf3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"expr", ",", "CD"}], "]"}]], "Input",
 CellLabel->
  "In[549]:=",ExpressionUUID->"c63e4165-8446-4170-9003-e1d433372aba"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "d", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`d, \
-$CellContext`c],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`b, -$CellContext`d, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "e", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`e, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`d, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"e", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`e, $CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978162610963*^9},
 CellLabel->"Out[549]=",
 CellID->1300796332,ExpressionUUID->"90d10878-a28f-480e-96af-f5dc9c9effb6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but we can also change to any desired derivative:", "MathCaption",ExpressionUUID->"e6aa7a5b-c6ef-442a-8a71-edbb2099770a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"expr", ",", "CD", ",", "Cd"}], "]"}]], "Input",
 CellLabel->
  "In[550]:=",ExpressionUUID->"d176231d-98a6-4b23-9abb-67b28941bed9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`e, -$CellContext`d, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`b, -$CellContext`d, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "e", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`e, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`e, -$CellContext`d, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"e", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`e, $CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`d][
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978162789381*^9},
 CellLabel->"Out[550]=",
 CellID->1100425464,ExpressionUUID->"65661f01-2d3b-45e5-a40b-14f1b176087b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "Cd", ",", "CD"}], "]"}]], "Input",
 CellLabel->
  "In[551]:=",ExpressionUUID->"fee87d64-b59a-4299-93e5-968a8449cc64"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "d"], GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", "b", " "},
          {"a", " ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`d][
   $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978162938037*^9},
 CellLabel->"Out[551]=",
 CellID->721605484,ExpressionUUID->"fb333ce6-9b3a-42fd-8360-c5dc40a2e129"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "If you do not like the double-derivative Christoffels, you can always break \
them to the CovD-",
 StyleBox["PD",
  FontFamily->"Courier"],
 " Christoffels:"
}], "MathCaption",ExpressionUUID->"7efae191-10dd-4865-bc5a-23cdd3f9b0b5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "BreakChristoffel"}]], "Input",
 CellLabel->
  "In[552]:=",ExpressionUUID->"655d09de-6231-49b9-be0e-4418ccbb16a4"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "d", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCd[$CellContext`e, -$CellContext`d, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "d", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`d, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"b", " ", " "},
             {" ", "d", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCd[$CellContext`b, -$CellContext`d, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
          StyleBox[GridBox[{
             {"b", " ", " "},
             {" ", "d", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`b, -$CellContext`d, \
-$CellContext`e],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "e", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, $CellContext`e, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "d", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCd[$CellContext`e, -$CellContext`d, \
-$CellContext`a],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "d", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`d, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"e", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`e, $CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`d][
    $CellContext`U[-$CellContext`a, $CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897816308776*^9},
 CellLabel->"Out[552]=",
 CellID->1183214979,ExpressionUUID->"eee795fe-986c-4378-a66e-73ec9d85d25b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The expansion is recursive:", "MathCaption",ExpressionUUID->"3a61f797-9edc-4ba0-8abf-678382460e5f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Cd", "[", 
     RowBox[{"-", "e"}], "]"}], "[", 
    RowBox[{
     RowBox[{"Cd", "[", 
      RowBox[{"-", "a"}], "]"}], "[", 
     RowBox[{"T", "[", 
      RowBox[{"f", ",", "c", ",", 
       RowBox[{"-", "b"}]}], "]"}], "]"}], "]"}], ",", "Cd"}], "]"}]], "Input",\

 CellLabel->
  "In[553]:=",ExpressionUUID->"c333ba2a-bd73-415d-b94e-5c73ef9365bd"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"i", " ", " "},
            {" ", "e", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCd[$CellContext`i, -$CellContext`e, \
-$CellContext`b],
     Editable->False]}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"d", " ", " "},
               {" ", "a", "i"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCd[$CellContext`d, -$CellContext`a, \
-$CellContext`i],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "c", " "},
              {" ", " ", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"c", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`c, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "g", " "},
              {" ", " ", "i"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`g, -$CellContext`i],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`f, -$CellContext`a, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"h", "c", " "},
              {" ", " ", "i"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`h, $CellContext`c, -$CellContext`i],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "c", " "},
              {" ", " ", "i"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`i]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "e", "i"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`c, -$CellContext`e, \
-$CellContext`i],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"i", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`i, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "g", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`g, -$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`d, -$CellContext`a, \
-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "i", " "},
              {" ", " ", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`i, -$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`f, -$CellContext`a, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"h", "i", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`h, $CellContext`i, -$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "i", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`T[$CellContext`f, $CellContext`i, -$CellContext`b]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "e", "i"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`f, -$CellContext`e, \
-$CellContext`i],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"i", " ", " "},
              {" ", "a", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`i, -$CellContext`a, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"h", "c", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`h, $CellContext`c, -$CellContext`b],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`d, -$CellContext`a, \
-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"i", "c", " "},
              {" ", " ", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`i, $CellContext`c, -$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"c", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`c, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"i", "g", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`i, $CellContext`g, -$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"i", "c", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`T[$CellContext`i, $CellContext`c, -$CellContext`b]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f", "d", " "},
           {" ", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`f, $CellContext`d, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`c, -$CellContext`a, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f", "c", " "},
           {" ", " ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`d, -$CellContext`a, \
-$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"d", "c", " "},
           {" ", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`d, $CellContext`c, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`ChristoffelCd[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"d", "c", " "},
            {" ", " ", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`T[$CellContext`d, $CellContext`c, -$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`d, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"f", "c", " "},
            {" ", " ", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`c, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"f", "d", " "},
            {" ", " ", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`T[$CellContext`f, $CellContext`d, -$CellContext`b]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "e"], 
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"f", "c", " "},
           {" ", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`e][
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"i", " ", " "},
           {" ", "e", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`i, -$CellContext`e, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"d", " ", " "},
               {" ", "i", "b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCd[$CellContext`d, -$CellContext`i, \
-$CellContext`b],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "c", " "},
              {" ", " ", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"c", " ", " "},
              {" ", "i", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`c, -$CellContext`i, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "g", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`g, -$CellContext`b],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "i", "h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCd[$CellContext`f, -$CellContext`i, \
-$CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"h", "c", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`h, $CellContext`c, -$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "i"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "c", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`i][
       $CellContext`T[$CellContext`f, $CellContext`c, -$CellContext`b]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978163268766*^9},
 CellLabel->"Out[553]=",
 CellID->595580833,ExpressionUUID->"99a9f03d-cbdf-4f8c-960c-0e9b5442056c"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["Christoffel",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Christoffel.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Construct Christoffel tensor relating two derivatives", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["BreakChristoffel",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "BreakChristoffel.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Rewrite a Christoffel tensor as the difference of two other \
Christoffel tensors\n", "Text"],
 StyleBox[ButtonBox["ChangeCovD",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ChangeCovD.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Rewrite the covariant derivative of a tensor in terms of a second \
derivative and Christoffels ", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978166124424*^9},ExpressionUUID->"ebb00c11-15d1-4f90-af87-\
59f5d08c0b10"],

Cell["Change of covariant derivative.", "Text",ExpressionUUID->"37768638-83d1-45a0-9860-23c2fad19d3c"],

Cell["\<\
If the relation between two covariant derivatives is fully described by a \
Christoffel tensor, then the curvature and torsion tensors associated to them \
must be also related by those Christoffel tensors.\
\>", "Text",ExpressionUUID->"21872d92-f2ba-4b0d-bf88-6d22ee7350f7"],

Cell[CellGroupData[{

Cell["This is the curvature tensor of the derivative CD:", "MathCaption",ExpressionUUID->"30606ad9-3ba3-4daa-88e5-8bb2ab82b0c5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "d"}], "]"}]], "Input",
 CellLabel->
  "In[554]:=",ExpressionUUID->"791b7fb7-2546-4b24-ab29-d06a7cd0c15b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "D", "]"}], 
      StyleBox[GridBox[{
         {" ", " ", " ", "d"},
         {"a", "b", "c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978163481202*^9},
 CellLabel->"Out[554]=",
 CellID->505781564,ExpressionUUID->"f7ebda4a-9ca5-47a9-8bcc-bf56aab34d78"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["ChangeCurvature",
  FontFamily->"Courier"],
 " (aka ",
 StyleBox["RiemannToChristoffel",
  FontFamily->"Courier"],
 " in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " version 0.7) changes any curvature tensor of a derivative to the curvature \
tensor of other derivative (by default ",
 StyleBox["PD",
  FontFamily->"Courier"],
 ", with zero curvature)."
}], "MathCaption",ExpressionUUID->"fc6c4717-dc5e-457c-a5bb-471911dfb305"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCurvature", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[555]:=",ExpressionUUID->"0ec2618f-a699-4338-9128-05c1576c977c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289781636311617`*^9},
 CellLabel->"Out[555]=",
 CellID->930471322,ExpressionUUID->"52a1826f-59d4-435b-9924-e6276e7fca68"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but we can convert to any derivative:", "MathCaption",ExpressionUUID->"7506df57-9f7b-48b5-aa64-7081731d64ee"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCurvature", "[", 
  RowBox[{"%%", ",", "CD", ",", "Cd"}], "]"}]], "Input",
 CellLabel->
  "In[556]:=",ExpressionUUID->"2c010025-8124-49ed-8a78-9c298a20b8b2"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"a", "b", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`a][
    $CellContext`ChristoffelCdCD[$CellContext`d, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`b][
    $CellContext`ChristoffelCdCD[$CellContext`d, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978163761724*^9},
 CellLabel->"Out[556]=",
 CellID->455591400,ExpressionUUID->"ee4da196-ae8c-4538-9443-6fbb2264c9d6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCurvature", "[", 
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "a"}], "]"}], ",", "CD", ",", "Cd"}], 
  "]"}]], "Input",
 CellLabel->
  "In[557]:=",ExpressionUUID->"ad8291c8-156a-46cd-b86f-26d99ddafe23"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`d, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCD[$CellContext`d, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCd[-$CellContext`b, -$CellContext`c],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`a][
    $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cd[-$CellContext`b][
    $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978163927684*^9},
 CellLabel->"Out[557]=",
 CellID->1345458245,ExpressionUUID->"4ce94a6c-ae84-441e-b0a4-b8a036b78a88"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Define a covariant derivative with torsion, but not metric-compatible. Now \
the Christoffel tensor is non-symmetric:\
\>", "MathCaption",ExpressionUUID->"0d899c47-3032-436e-9eb7-70d07b5d2de4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"CDT", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<#\>\"", ",", "\"\<DT\>\""}], "}"}]}], ",", 
   RowBox[{"Torsion", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[558]:=",ExpressionUUID->"a0cf2d5a-66b0-42ec-aca2-f033c82522aa"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CDT", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CDT[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781640586452`*^9},
 CellLabel->"During evaluation of In[558]:=",
 CellID->113920805,ExpressionUUID->"532cc24a-5b23-4e96-ae36-ac10121b0e8a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionCDT[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164074646*^9},
 CellLabel->"During evaluation of In[558]:=",
 CellID->1259040158,ExpressionUUID->"db04ebce-e3b6-4b2b-95bf-e6a9e6a27044"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164092724*^9},
 CellLabel->"During evaluation of In[558]:=",
 CellID->1531561615,ExpressionUUID->"f0d8a7ca-bef9-4c91-a2f3-0b2b01c38e68"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCDT[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164111292*^9},
 CellLabel->"During evaluation of In[558]:=",
 CellID->1217070363,ExpressionUUID->"83999169-0297-425d-9382-e0926739c2b2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCDT[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781641286077`*^9},
 CellLabel->"During evaluation of In[558]:=",
 CellID->519853266,ExpressionUUID->"08fbd7d4-2e0e-461c-8704-94fe67248675"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.7289781641489887`*^9},
 CellLabel->"During evaluation of In[558]:=",
 CellID->1210365666,ExpressionUUID->"1276d03d-21aa-4152-b671-573dbe8edd5e"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now the formulas involving ",
 StyleBox["CDT",
  FontFamily->"Courier"],
 " will contain torsion terms:"
}], "MathCaption",ExpressionUUID->"e5f7ef7f-6967-4c1a-ad55-49faeafe63cd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCurvature", "[", 
  RowBox[{
   RowBox[{"RiemannCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "a"}], "]"}], ",", "Cd", ",", "CDT"}], 
  "]"}]], "Input",
 CellLabel->
  "In[559]:=",ExpressionUUID->"25566c72-d2f9-40df-b7e5-5d4cd1900fb3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCdCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897816431199*^9},
 CellLabel->"During evaluation of In[559]:=",
 CellID->1650644078,ExpressionUUID->"ec84d557-b6fd-494f-a4e4-f33f466ac9fb"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCDT[$CellContext`d, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCDT[$CellContext`d, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "DT", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCDT[-$CellContext`b, -$CellContext`c],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`d, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "DT", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCDT[$CellContext`d, -$CellContext`b, -$CellContext`a],
    
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["DT", "a"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CDT[-$CellContext`a][
    $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`c]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["DT", "b"], GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CDT[-$CellContext`b][
    $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`a, \
-$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978164361699*^9},
 CellLabel->"Out[559]=",
 CellID->1181267000,ExpressionUUID->"968a3072-327f-4128-a5a0-3f04ac99fb72"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[560]:=",ExpressionUUID->"64b3c03c-6d04-485c-8089-34e6689ba954"],

Cell["\<\
-(ChristoffelCdCDT[j$22259, -j$22259, -j$22269]*ChristoffelCdCDT[j$22269, -b, \
-c]) + 
 ChristoffelCdCDT[j$22259, -b, -j$22269]*ChristoffelCdCDT[j$22269, -j$22259, \
-c] - 
 RicciCDT[-b, -c] + ChristoffelCdCDT[j$22259, -j$22269, \
-c]*TorsionCDT[j$22269, -b, -j$22259] + 
 CDT[-b][ChristoffelCdCDT[j$22259, -j$22259, -c]] - 
 CDT[-j$22259][ChristoffelCdCDT[j$22259, -b, -c]]\
\>", "Output",
 CellChangeTimes->{3.728978164511882*^9},
 CellLabel->"Out[560]//InputForm=",
 CellID->68152485,ExpressionUUID->"acecbec6-73d3-4cf2-bd92-733232a56855"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The difference between the torsion tensors of two derivatives is given by \
the antisymmetric part of the Christoffel relating them. In this case the \
torsion of ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 " is zero. The change is performed by ",
 StyleBox["ChangeTorsion",
  FontFamily->"Courier"],
 " (aka ",
 StyleBox["TorsionToChristoffel",
  FontFamily->"Courier"],
 " in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " version 0.7):"
}], "MathCaption",ExpressionUUID->"4b6413aa-575a-49c7-bae1-512b19313fd5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeTorsion", "[", 
  RowBox[{
   RowBox[{"TorsionCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", "CDT", ",", "Cd"}], "]"}]], "Input",
 CellLabel->
  "In[561]:=",ExpressionUUID->"a02bc13a-ad78-44ad-821c-1b0fd7bfdc5b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[CapitalGamma]", "[", 
        RowBox[{"\[EmptyDownTriangle]", ",", "DT"}], "]"}], 
       StyleBox[GridBox[{
          {"a", " ", " "},
          {" ", "c", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ChristoffelCdCDT[$CellContext`a, -$CellContext`c, \
-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289781646589746`*^9},
 CellLabel->"Out[561]=",
 CellID->830031544,ExpressionUUID->"13dca5a3-9a1f-4679-96c9-19fa0d56f64c"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["ChangeCurvature",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ChangeCurvature.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Change in curvature when changing between two covariant \
derivatives\n", "Text"],
 StyleBox[ButtonBox["ChangeTorsion",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ChangeTorsion.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Change in torsion when changing between two covariant derivatives",
   "Text"]
}], "DefinitionBox",ExpressionUUID->"a5a86fff-914f-4cd1-ab4c-207be4d7047f"],

Cell["Induced changes in curvature and torsion tensors.", "Text",ExpressionUUID->"9fdfda3c-9d70-4fa1-8d66-ec9131f4123a"],

Cell[CellGroupData[{

Cell["Undefine some derivatives:", "MathCaption",ExpressionUUID->"583303dd-f2ec-48b8-8ddc-1ff7ce5f0630"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefCovD", "/@", 
   RowBox[{"{", 
    RowBox[{"CD", ",", "CDT"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[562]:=",ExpressionUUID->"7caf7908-52d0-449f-8dca-7a8028be4dd1"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781648111753`*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->286389096,ExpressionUUID->"2e78a874-31de-42c8-aab0-600b65ccf633"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCdCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCdCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164827755*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->830683545,ExpressionUUID->"5c3c07f5-2332-4aae-b9a1-2d17a3b0bcb5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164846345*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->507151668,ExpressionUUID->"2569eda0-cf27-4667-9d77-781f7aea01bb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781648652143`*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->335489133,ExpressionUUID->"c550e499-7b79-46c4-8b5c-a9a8e17a31f4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781648839006`*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->1421284956,ExpressionUUID->"0d17c501-f28c-4c11-be8f-2c38b18093aa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164902277*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->1075649911,ExpressionUUID->"3eac565b-0e67-46ef-b2a8-e8688912ae1a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCdCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCdCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164920608*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->930713731,ExpressionUUID->"ea74b6e0-3620-424c-954d-4ad75e7fb8b3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164938925*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->234792169,ExpressionUUID->"e644f1ad-ea7d-4d8d-a025-2184541f7628"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164956386*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->1224524844,ExpressionUUID->"684e7e66-a470-4cad-b24d-b29b344efb3a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164975408*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->1253263907,ExpressionUUID->"185e1e0c-91e7-498a-af26-0f6aa3a83ff8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978164994878*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->1086875173,ExpressionUUID->"08687b25-63c2-48b8-b0c9-c64e6398a307"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781650132513`*^9},
 CellLabel->"During evaluation of In[562]:=",
 CellID->615131229,ExpressionUUID->"1f3fc87e-45bb-464a-ada4-f89d7a288bc2"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Note that ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " also allows metric-compatible connections with torsion. The symmetry \
properties of the associated curvature tensors are not complete. We shall \
later illustrate this type of derivatives, after defining metric fields."
}], "Text",ExpressionUUID->"1e49e14d-2ffb-45a8-a1c8-0677d22cd53d"]
}, Closed]],

Cell[CellGroupData[{

Cell["6.4. General Bianchi identities", "Subsection",ExpressionUUID->"f87d0846-1b86-4f36-a2a8-41e6b8db3f25"],

Cell[TextData[{
 "Let us now check the Bianchi identities for general covariant derivatives \
with torsion. Note that these identities are not directly encoded in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", but can be easily computed."
}], "Text",ExpressionUUID->"bc930580-ed21-4357-893a-e6593af4b119"],

Cell[CellGroupData[{

Cell["\<\
Define a covariant derivative with torsion. Note the special symmetry \
properties of the defined tensors:\
\>", "MathCaption",ExpressionUUID->"fbbe1de0-a173-4fc2-9a70-4e3cc69a759e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"Torsion", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[563]:=",ExpressionUUID->"9510dcf1-73a8-4f16-ae18-fb03f7c67260"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978181225565*^9},
 CellLabel->"During evaluation of In[563]:=",
 CellID->1117298015,ExpressionUUID->"1822de1b-cad8-4275-bfce-86efa4cda7a7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978181243128*^9},
 CellLabel->"During evaluation of In[563]:=",
 CellID->830728737,ExpressionUUID->"6bb9f321-bc42-4424-bc08-230ddf769aef"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781812601423`*^9},
 CellLabel->"During evaluation of In[563]:=",
 CellID->1266693612,ExpressionUUID->"294e7038-56d2-4e22-b225-07fec430e77f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897818127743*^9},
 CellLabel->"During evaluation of In[563]:=",
 CellID->1314144173,ExpressionUUID->"ee5d6336-2613-4f50-aecb-5a713db3ecc5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897818129548*^9},
 CellLabel->"During evaluation of In[563]:=",
 CellID->2099289863,ExpressionUUID->"3f752cc7-f4d3-40c9-9f37-f982e4f2facb"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.72897818131336*^9},
 CellLabel->"During evaluation of In[563]:=",
 CellID->1755890878,ExpressionUUID->"fade3f37-f93d-4d72-91fe-0b837eabd4fb"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the general form of the first Bianchi identity:", "MathCaption",ExpressionUUID->"e72e4e15-9e11-4495-a5de-a4d0ae663522"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "d"}], "]"}], "+", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", " ", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"d", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], " ", "]"}], "-", 
  RowBox[{
   RowBox[{"TorsionCD", "[", 
    RowBox[{"e", ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"d", ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "e"}]}], "]"}]}]}]], "Input",
 CellLabel->
  "In[564]:=",ExpressionUUID->"15ee41dd-57e7-475a-a412-a2f2166264c4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"a", "b", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`c]],
   
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897818154132*^9},
 CellLabel->"Out[564]=",
 CellID->537709472,ExpressionUUID->"0b78c8d6-cc38-486f-9036-891bf5d8c5a5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"6", 
  RowBox[{"Antisymmetrize", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[565]:=",ExpressionUUID->"7aad89ba-2a9e-4998-bb89-ea2879a6a040"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"a", "b", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"a", "c", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, -$CellContext`b, \
$CellContext`d],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"b", "a", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`c, \
$CellContext`d],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"b", "c", "a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`a, \
$CellContext`d],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"c", "a", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, -$CellContext`b, \
$CellContext`d],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"c", "b", "a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, -$CellContext`a, \
$CellContext`d],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`c]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`b]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`c]],
   
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`a]],
   
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`b]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`a]],
   
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978182149106*^9},
 CellLabel->"Out[565]=",
 CellID->671909328,ExpressionUUID->"ba6a087b-e66c-4fe7-849d-1f1201c30e94"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The computation can be performed by transforming the derivative CD and its \
associated tensors into PD and Christoffel tensors:\
\>", "MathCaption",ExpressionUUID->"af8a9621-05b7-413a-8a0b-896840faa124"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ChangeCovD"}]], "Input",
 CellLabel->
  "In[566]:=",ExpressionUUID->"5587cfce-5c55-448d-af7a-76278e3c6983"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"a", "b", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"a", "c", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, -$CellContext`b, \
$CellContext`d],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"b", "a", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`c, \
$CellContext`d],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"b", "c", "a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`a, \
$CellContext`d],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"c", "a", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, -$CellContext`b, \
$CellContext`d],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "d"},
          {"c", "b", "a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, -$CellContext`a, \
$CellContext`d],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`c]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`b]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`c]],
   
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`a]],
   
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`b]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`a]],
   
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978182714324*^9},
 CellLabel->"Out[566]=",
 CellID->1311059770,ExpressionUUID->"93642775-b1e9-44be-a6f6-1aa4ebabe52d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "RiemannToChristoffel"}]], "Input",
 CellLabel->
  "In[567]:=",ExpressionUUID->"f43c7626-8972-4b74-a84b-9293a54a61ed"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "e", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`e, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`e, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`c]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`b]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`c]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`b]],
   
   Editable->False], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`a]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`c]],
   
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`c, -$CellContext`a]],
   
   Editable->False], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"d", " ", " "},
            {" ", "b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`a]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`a, -$CellContext`b]],
   
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`TorsionCD[$CellContext`d, -$CellContext`b, -$CellContext`a]],
   
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978182920506*^9},
 CellLabel->"Out[567]=",
 CellID->1799336170,ExpressionUUID->"90bef3ed-a01b-487c-98b3-3c21551fcce5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "TorsionToChristoffel"}]], "Input",
 CellLabel->
  "In[568]:=",ExpressionUUID->"0ead20f3-dd2b-44e7-b0ed-bbe9e02f104f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "c", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "b", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "b", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "c", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "c", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "c", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`c],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "a", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "b", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "b", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "b", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "b", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`b, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`c],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`a],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"d", " ", " "},
              {" ", "a", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "a", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`a, \
-$CellContext`e],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`e, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`c],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"e", " ", " "},
             {" ", "c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`b],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978183638836*^9},
 CellLabel->"Out[568]=",
 CellID->578292136,ExpressionUUID->"b660f6f8-1048-42f2-9ffb-ce6a73886b12"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[569]:=",ExpressionUUID->"cb6c1a5e-c217-4049-bd1b-14f8d2a6547b"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978184174458*^9},
 CellLabel->"Out[569]=",
 CellID->1992427077,ExpressionUUID->"36c4f8af-764f-4b7e-b777-662bce75ef73"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the general form of the second Bianchi identity:", "MathCaption",ExpressionUUID->"c87b02a2-fe67-4786-9dc5-26fe14dba555"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", " ", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", "e"}], "]"}], " ", "]"}], "-", 
  RowBox[{
   RowBox[{"TorsionCD", "[", 
    RowBox[{"f", ",", 
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "f"}], ",", 
     RowBox[{"-", "d"}], ",", "e"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[570]:=",ExpressionUUID->"0b33cdd0-b621-4392-943e-4b580315db5d"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", "e"},
            {"c", "f", "d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, \
-$CellContext`d, $CellContext`e],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`d, \
$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978184465444*^9},
 CellLabel->"Out[570]=",
 CellID->1009375468,ExpressionUUID->"c34683a7-269f-48d3-a1dc-127bda1f3c49"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"6", 
  RowBox[{"Antisymmetrize", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[571]:=",ExpressionUUID->"46d6c294-043b-4908-a077-d010bc96a5e9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", "e"},
            {"c", "f", "d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, \
-$CellContext`d, $CellContext`e],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`d, \
$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289781859741907`*^9},
 CellLabel->"Out[571]=",
 CellID->1600163023,ExpressionUUID->"de03aafb-ae30-45a9-9f39-274abc449096"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The computation proceeds along the same lines:", "MathCaption",ExpressionUUID->"708a4e32-a664-427f-979c-79237bd2442a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ChangeCovD"}]], "Input",
 CellLabel->
  "In[572]:=",ExpressionUUID->"91a7e00d-9fb8-4dc6-b8c1-e319d387d044"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "f"},
           {"a", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "b", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`f, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "f"},
           {"a", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, -$CellContext`d, \
$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "c", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, -$CellContext`f, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "f"},
           {"b", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`d, \
$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "a", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`f, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "f"},
           {"b", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`d, \
$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "c", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`f, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "f"},
           {"c", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, -$CellContext`d, \
$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "a", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, -$CellContext`f, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "f"},
           {"c", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, -$CellContext`d, \
$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "b", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, -$CellContext`f, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"f", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`f, -$CellContext`a, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"f", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`f, -$CellContext`a, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"f", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`f, -$CellContext`b, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"f", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`f, -$CellContext`b, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"f", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`f, -$CellContext`c, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"f", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`f, -$CellContext`c, -$CellContext`d, \
$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`b, -$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`c, -$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "f", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`d, \
$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`c, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`b, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "c", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`c, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"c", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`a, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"a", "b", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
$CellContext`e]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "e"},
           {"b", "a", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`d, \
$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978186151469*^9},
 CellLabel->"Out[572]=",
 CellID->1002216180,ExpressionUUID->"ce42d8e2-35a5-4df0-9c60-b25a32335937"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "RiemannToChristoffel"}]], "Input",
 CellLabel->
  "In[573]:=",ExpressionUUID->"095718d5-0070-4b62-9a48-27b85cbccbfb"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "b", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"f", " ", " "},
               {" ", "b", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"f", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"f", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`c, -$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"T", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TorsionCD[$CellContext`f, -$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978186505559*^9},
 CellLabel->"Out[573]=",
 CellID->1533093839,ExpressionUUID->"aa7cbb20-d3d6-42af-aec9-d42cc67d234a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "TorsionToChristoffel"}]], "Input",
 CellLabel->
  "In[574]:=",ExpressionUUID->"118d0df2-cafb-46b3-828d-22ee5195157e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "b", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"f", " ", " "},
               {" ", "b", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "f"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"f", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"f", " ", " "},
               {" ", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"f", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`c],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`b],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "c", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`c],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "c", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {"e", " ", " "},
               {" ", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"f", " ", " "},
              {" ", "a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"f", " ", " "},
             {" ", "b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`a],
      Editable->False]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"g", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`f, \
-$CellContext`d],
       Editable->False]}], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "f", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`f, \
-$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {
           RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {"e", " ", " "},
              {" ", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289781868425627`*^9},
 CellLabel->"Out[574]=",
 CellID->1199355367,ExpressionUUID->"8682f9ac-d088-4124-8796-8e03b2bad44e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[575]:=",ExpressionUUID->"09a85fd6-bbad-49ee-84f6-e81b2718d64c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`a, \
-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "c", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`b, \
-$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "b", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`a, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"e", " ", " "},
           {" ", "a", "g"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`g],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"f", " ", " "},
           {" ", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`f, -$CellContext`b, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"g", " ", " "},
           {" ", "c", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`g, -$CellContext`c, \
-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`c, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`b, \
-$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"e", " ", " "},
            {" ", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`ChristoffelCD[$CellContext`e, -$CellContext`a, \
-$CellContext`d]]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978187098949*^9},
 CellLabel->"Out[575]=",
 CellID->1284142427,ExpressionUUID->"b13a67bc-1a06-4736-88c4-c151d8533942"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[576]:=",ExpressionUUID->"549b35dd-b674-40f2-83e3-8f9be8d0a357"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978187495422*^9},
 CellLabel->"Out[576]=",
 CellID->371118239,ExpressionUUID->"eec8e411-a994-4b07-8c42-b9d1d84a8b14"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"fb60f043-f295-4e88-922a-25f37c9bb1ee"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefCovD", "[", "CD", "]"}]], "Input",
 CellLabel->
  "In[577]:=",ExpressionUUID->"aafdb914-e2f5-4ccf-a213-b98b35580641"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978187661109*^9},
 CellLabel->"During evaluation of In[577]:=",
 CellID->1091731985,ExpressionUUID->"89900bfd-778b-4fb5-8d5c-84fd96515585"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978187686255*^9},
 CellLabel->"During evaluation of In[577]:=",
 CellID->639191927,ExpressionUUID->"ed7d98b6-aa10-4cc5-83d7-16546d96698c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978187706502*^9},
 CellLabel->"During evaluation of In[577]:=",
 CellID->1185514795,ExpressionUUID->"2d0eb9b9-d3a5-4e8c-a438-6db5e962a7d7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289781877251797`*^9},
 CellLabel->"During evaluation of In[577]:=",
 CellID->593882699,ExpressionUUID->"ef44147d-41c3-4c6a-aacb-b0264e988660"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978187743479*^9},
 CellLabel->"During evaluation of In[577]:=",
 CellID->2030014484,ExpressionUUID->"35fe2e3d-924e-4cee-835a-a0ac03ab2644"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.5. Constant symbols", "Subsection",ExpressionUUID->"e53b0c02-7a31-444e-af12-e664d149db42"],

Cell[TextData[{
 "It is important to discuss the concept of a constant in ",
 StyleBox["xTensor`", "DisplayMath"],
 ". A scalar field ",
 StyleBox["c[]",
  FontFamily->"Courier"],
 " is a constant if it is defined on no manifold. However a nonscalar tensor \
field cannot be constant on the base manifold of its indices (actually ",
 StyleBox["xTensor`", "DisplayMath"],
 " automatically defines the tensor to be a field on the base manifolds on \
its indices). On the other hand we have to use symbols denoting constants (as \
Newton's constant). We shall use the command ",
 StyleBox["DefConstantSymbol",
  FontFamily->"Courier"],
 " to define those constants."
}], "Text",ExpressionUUID->"66acdbf6-cfe0-4c5d-9195-5ac7c3a48a92"],

Cell[TextData[{
 StyleBox[ButtonBox["DefConstantSymbol",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefConstantSymbol.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Define a constant\n", "Text"],
 StyleBox[ButtonBox["UndefConstantSymbol",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefConstantSymbol.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Undefine a constant\n", "Text"],
 StyleBox[ButtonBox["$ConstantSymbols",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$ConstantSymbols.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["List of defined constants\n", "Text"],
 StyleBox[ButtonBox["ConstantSymbolQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ConstantSymbolQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Check a constant symbol", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978193595676*^9},ExpressionUUID->"9c916998-07b6-4d7f-a5dc-\
9c0454a0bed8"],

Cell["Functions that operate with constants.", "Text",ExpressionUUID->"75f4892b-629e-4473-a591-88818030bf40"],

Cell[CellGroupData[{

Cell["This defines a constant scalar field:", "MathCaption",ExpressionUUID->"22773fbb-d315-44c5-84ba-b1ae3966642f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"const", "[", "]"}], ",", 
   RowBox[{"{", "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[578]:=",ExpressionUUID->"769d1396-2962-42e4-90f2-171bbf2a623f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"const", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`const[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978196222909*^9},
 CellLabel->"During evaluation of In[578]:=",
 CellID->1175468544,ExpressionUUID->"f3d7bd1b-69ac-41ce-900b-bf49f203e3b8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "const"}]], "Input",
 CellLabel->
  "In[579]:=",ExpressionUUID->"8d3dc582-8a41-48b6-a2eb-a1697bbdb4c3"],

Cell[CellGroupData[{

Cell["Global`const", "Print", "PrintUsage",
 CellChangeTimes->{3.728978196375587*^9},
 CellTags->"Info5793728981796-9181623",
 CellID->1397750255,ExpressionUUID->"c227df3b-fdfd-490b-a692-412f36cca699"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "const", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`const],
          Editable->False], "^=", "const"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "const", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "const", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "const", "]"}], "^=", "\<\"const\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "const", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "const", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "const", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "const", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`const],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978196398527*^9},
 CellTags->"Info5793728981796-9181623",
 CellID->664921252,ExpressionUUID->"86dcd4d2-d60f-4dc8-8d2c-deaa78c12e5d"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
By default, it is not automatic to check that a derivative on an object \
living on a different manifold is zero because that is a slow process that \
could be internally happening too often.\
\>", "MathCaption",ExpressionUUID->"da88c5ae-3389-43a8-9e37-7fd8deb1d5af"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"const", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[580]:=",ExpressionUUID->"59c548bb-9730-456c-851f-653dedf61ef1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], "const"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`const[]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978196517425*^9},
 CellLabel->"Out[580]=",
 CellID->1974067023,ExpressionUUID->"dd5b3192-5173-4cd7-9e92-1029acdb9978"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckZeroDerivative", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[581]:=",ExpressionUUID->"6f3c6176-a501-4c05-9560-981d00bc557b"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978196691395*^9},
 CellLabel->"Out[581]=",
 CellID->1669593058,ExpressionUUID->"7e16bc55-71b7-4905-9acf-1e9cee5f874e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can automate it for a particular derivative (or for all):", \
"MathCaption",ExpressionUUID->"fa6f71bf-1bd8-4f8f-91d6-7ded973828e4"],

Cell[BoxData[
 RowBox[{"CheckZeroDerivativeStart", "[", "PD", "]"}]], "Input",
 CellLabel->
  "In[582]:=",ExpressionUUID->"552cb858-df77-4d95-9eb0-db0c8a4e63c3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"const", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[583]:=",ExpressionUUID->"b73f45b2-67f8-4625-8286-4b18796c929a"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978197383325*^9},
 CellLabel->"Out[583]=",
 CellID->1829194104,ExpressionUUID->"e259591a-1051-4655-864e-0b8b917e2364"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"CheckZeroDerivativeStop", "[", "PD", "]"}]], "Input",
 CellLabel->
  "In[584]:=",ExpressionUUID->"846fd6be-23f9-4508-8baf-b0a75683bc77"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
An arbitrary symbol is not understood by default as a constant. We must \
define it as a constant:\
\>", "MathCaption",ExpressionUUID->"5941d6dd-d795-466a-b0f7-9c86790d986b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", "GNewton", "]"}]], "Input",
 CellLabel->
  "In[585]:=",ExpressionUUID->"a945d173-b062-4002-8bfa-49c8f9c45d22"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], "GNewton"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][$CellContext`GNewton],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289781975904417`*^9},
 CellLabel->"Out[585]=",
 CellID->2036899123,ExpressionUUID->"3b120647-f42c-40d7-98b9-06f81147168b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefConstantSymbol", "[", "GNewton", "]"}]], "Input",
 CellLabel->
  "In[586]:=",ExpressionUUID->"f87de4cc-ffb9-4c36-bbdb-28914852d7b0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol \"\>", "\[InvisibleSpace]", 
   "GNewton", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefConstantSymbol, ": Defining ", 
   "constant symbol ", $CellContext`GNewton, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978197708646*^9},
 CellLabel->"During evaluation of In[586]:=",
 CellID->1126743434,ExpressionUUID->"2fb4a181-3c60-4280-a7d1-ecca648bbe81"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", "GNewton", "]"}]], "Input",
 CellLabel->
  "In[587]:=",ExpressionUUID->"1b285f34-ef14-4e6a-957d-a51f0753071f"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978197835034*^9},
 CellLabel->"Out[587]=",
 CellID->666095992,ExpressionUUID->"e80e1ac4-ae45-4505-b193-f4538d6fbd4f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefConstantSymbol", "[", "GNewton", "]"}]], "Input",
 CellLabel->
  "In[588]:=",ExpressionUUID->"681aca36-cd36-46c0-a4f5-4a1a8e6a2c12"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefConstantSymbol", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"constant symbol\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["GNewton",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`GNewton,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefConstantSymbol, ": Undefined ", "constant symbol", 
   " ", $CellContext`GNewton],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978197924679*^9},
 CellLabel->"During evaluation of In[588]:=",
 CellID->243085733,ExpressionUUID->"a233a52a-72c1-4805-ad2c-68915034335d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", "GNewton", "]"}]], "Input",
 CellLabel->
  "In[589]:=",ExpressionUUID->"97cc96e3-865e-4352-a1cd-cf6953fffd7e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "a"], "GNewton"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[-$CellContext`a][$CellContext`GNewton],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978198039319*^9},
 CellLabel->"Out[589]=",
 CellID->1028017937,ExpressionUUID->"ece36969-2c21-4896-98f7-3b4d3293936d"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "GNewton", "]"}]], "Input",
 CellLabel->
  "In[590]:=",ExpressionUUID->"4172c80d-4d44-4b23-be79-3f386109b90c"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.6. Parameter derivatives", "Subsection",ExpressionUUID->"0704b121-e23a-4a34-b2da-ab4444ebb3de"],

Cell[TextData[{
 "Given a parameter defined with ",
 StyleBox["DefParameter",
  FontFamily->"Courier"],
 " we can take parametric derivatives of expressions which depend on that \
parameter, using ",
 StyleBox["ParamD",
  FontFamily->"Courier"],
 ". For historical reasons ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has a generic parametric derivative called ",
 StyleBox["OverDot",
  FontFamily->"Courier"],
 " such that every field is considered to depend on the corresponding \
(undefined) parameter."
}], "Text",ExpressionUUID->"2552928f-29d9-4573-9340-1837b8ffaeb1"],

Cell[TextData[{
 StyleBox[ButtonBox["ParamD",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ParamD.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Parametric derivative with respect to a defined parameter\n", 
  "Text"],
 StyleBox[ButtonBox["OverDot",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "OverDot.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Generic parametric derivative with respect to an unspecified \
parameter", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978205546434*^9},ExpressionUUID->"45628cc3-8169-471c-95dd-\
a09e3e1c7100"],

Cell["Parametric derivatives.", "Text",ExpressionUUID->"d9bff57a-c643-47d5-8bf8-5feb8e98f66c"],

Cell[CellGroupData[{

Cell[TextData[{
 "The generic parametric derivative ",
 StyleBox["OverDot",
  FontFamily->"Courier"],
 " (Dot is used by ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " as inner scalar product) is represented by a (very) small over dot:"
}], "MathCaption",ExpressionUUID->"431facde-7d95-4aff-a35d-31e682d084c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"OverDot", "[", 
  RowBox[{
   RowBox[{"r", "[", "]"}], "^", "2"}], "]"}]], "Input",
 CellLabel->
  "In[591]:=",ExpressionUUID->"ecdaff66-c610-4e86-89e1-87a4979ac18c"],

Cell[BoxData[
 RowBox[{"2", " ", 
  OverscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "."], " ", 
  InterpretationBox[
   StyleBox["r",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`r[],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289782100253477`*^9},
 CellLabel->"Out[591]=",
 CellID->656372838,ExpressionUUID->"9cd7bf8f-e1a2-41c6-8a61-0b7e180ed6ce"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["OverDot",
  FontFamily->"Courier"],
 " has been defined to commute with partial derivatives, but not with other \
covariant derivatives or with Lie derivatives:"
}], "MathCaption",ExpressionUUID->"2c2319e2-06be-4e1e-a453-973e1294e76c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"OverDot", "[", 
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"r", "[", "]"}], "]"}], "]"}], "//", "InputForm"}]], "Input",
 CellLabel->
  "In[592]:=",ExpressionUUID->"5fba2b9f-c335-4326-8e3d-985324837dd8"],

Cell["PD[-a][OverDot[r[]]]", "Output",
 CellChangeTimes->{3.7289782101464653`*^9},
 CellLabel->"Out[592]//InputForm=",
 CellID->34793627,ExpressionUUID->"ddadd07b-a25e-486a-ad96-f0084b1aed58"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"OverDot", "[", 
   RowBox[{
    RowBox[{"Cd", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"r", "[", "]"}], "]"}], "]"}], "//", "InputForm"}]], "Input",
 CellLabel->
  "In[593]:=",ExpressionUUID->"b0f59510-bbe6-4863-bb59-398cbcae60c2"],

Cell["OverDot[Cd[-a][r[]]]", "Output",
 CellChangeTimes->{3.728978210267165*^9},
 CellLabel->"Out[593]//InputForm=",
 CellID->660937005,ExpressionUUID->"da209a7c-43c2-4490-95b3-a21c71435510"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
All fields currently defined do not depend on external parameters, so that we \
define new objects:\
\>", "MathCaption",ExpressionUUID->"6434f3d2-02e6-409a-98d5-eba0e36a2db2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefParameter", "[", 
  RowBox[{"time", ",", 
   RowBox[{"PrintAs", "->", "\"\<t\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[594]:=",ExpressionUUID->"a7525c81-cda1-4b78-850d-dfea1b3b6f42"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefParameter", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"parameter \"\>", "\[InvisibleSpace]", "time", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefParameter, ": Defining ", 
   "parameter ", $CellContext`time, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897821039218*^9},
 CellLabel->"During evaluation of In[594]:=",
 CellID->1582820751,ExpressionUUID->"77c535b5-45ae-43fa-8b15-86638c451455"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", "time", "]"}], "[", 
  RowBox[{"r", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[595]:=",ExpressionUUID->"7d44797d-14b6-4359-a76c-44e9c6b70b13"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978210510548*^9},
 CellLabel->"Out[595]=",
 CellID->1588426668,ExpressionUUID->"1bb34982-01c5-42ca-afb7-2e016988e206"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"M3", ",", "time"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[596]:=",ExpressionUUID->"b90a6b90-f93e-4df1-8b92-58b9557e569d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`Q[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978210633469*^9},
 CellLabel->"During evaluation of In[596]:=",
 CellID->1317793219,ExpressionUUID->"1bc0d25b-eb46-49c9-88c2-629407b50730"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Q"}]], "Input",
 CellLabel->
  "In[597]:=",ExpressionUUID->"bb97df03-06da-4e15-84f2-621ea286c4c3"],

Cell[CellGroupData[{

Cell["Global`Q", "Print", "PrintUsage",
 CellChangeTimes->{3.7289782107517033`*^9},
 CellTags->"Info5973728981810-9181623",
 CellID->1151780222,ExpressionUUID->"47a658f0-5129-4f86-9b44-b5ae579d9de6"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "Q", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`Q],
          Editable->False], "^=", "Q"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "Q", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "Q", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["t",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`time,
            Editable->False], ",", "M3"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "Q", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False], ",", 
           InterpretationBox[
            StyleBox["t",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`time,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "Q", "]"}], "^=", "\<\"Q\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "Q", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"-", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "Q", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "Q", "]"}], "^=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "Q", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`Q],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978210774721*^9},
 CellTags->"Info5973728981810-9181623",
 CellID->552597096,ExpressionUUID->"73efd163-d740-46b6-873d-021591318f7f"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["V1.0 UPDATE:",
  FontColor->RGBColor[1, 0, 0]],
 " From version 1.0.2 there are several posible representations of a ",
 StyleBox["ParamD",
  FontFamily->"Courier"],
 " derivative, controlled by a global variable called ",
 StyleBox["$ParamDFormat",
  FontFamily->"Courier"],
 ", with default value ",
 StyleBox["\"Prefix\"",
  FontFamily->"Courier"],
 ". Let us first simplify the printing of the parameter:"
}], "MathCaption",ExpressionUUID->"72b9653c-57ad-4dd8-8d84-dc0294b79014"],

Cell[CellGroupData[{

Cell[BoxData["$ParamDFormat"], "Input",
 CellLabel->
  "In[598]:=",ExpressionUUID->"517b52ea-01d9-4729-9af6-39b8ae7d413b"],

Cell[BoxData["\<\"Prefix\"\>"], "Output",
 CellChangeTimes->{3.728978210937985*^9},
 CellLabel->"Out[598]=",
 CellID->1780102844,ExpressionUUID->"d7bc4cf1-9800-48c9-b430-457b2b372b4e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", "time", "]"}], "[", 
  RowBox[{"3", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[599]:=",ExpressionUUID->"e87154fa-28dc-4dc6-b9c7-c36f3843da31"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "t"], GridBox[{
       {"Q", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ParamD[$CellContext`time][
    $CellContext`Q[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289782111135*^9},
 CellLabel->"Out[599]=",
 CellID->1593812751,ExpressionUUID->"18fe6429-1cdb-4bad-80d0-984afd072f62"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", 
   RowBox[{"time", ",", "time"}], "]"}], "[", 
  RowBox[{"3", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[600]:=",ExpressionUUID->"f1733c28-b855-4ae7-87e3-1db43b96dcd0"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "t"], 
     SubscriptBox["\[PartialD]", "t"], GridBox[{
       {"Q", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ParamD[$CellContext`time, $CellContext`time][
    $CellContext`Q[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978211297696*^9},
 CellLabel->"Out[600]=",
 CellID->1647377026,ExpressionUUID->"51cf5f54-f27c-4937-946c-d23b7a514e34"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$ParamDFormat", "=", "\"\<SinglePrefix\>\""}], ";"}]], "Input",
 CellLabel->
  "In[601]:=",ExpressionUUID->"75990947-29f6-4495-ac12-8008e55da988"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", "time", "]"}], "[", 
  RowBox[{"3", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[602]:=",ExpressionUUID->"47d7ef43-e76d-4e35-94ce-946fe8ef8269"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "t"], GridBox[{
       {"Q", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ParamD[$CellContext`time][
    $CellContext`Q[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289782114578648`*^9},
 CellLabel->"Out[602]=",
 CellID->2106498858,ExpressionUUID->"68a04909-898d-4cea-95e4-56081b80dd04"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", 
   RowBox[{"time", ",", "time"}], "]"}], "[", 
  RowBox[{"3", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[603]:=",ExpressionUUID->"a0f6933e-345b-421b-9996-1ec7fe877ce8"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "tt"], GridBox[{
       {"Q", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ParamD[$CellContext`time, $CellContext`time][
    $CellContext`Q[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289782115960712`*^9},
 CellLabel->"Out[603]=",
 CellID->510901964,ExpressionUUID->"066d6fbd-3de7-4297-b1fb-7b8d1f391149"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"$ParamDFormat", "=", "\"\<Postfix\>\""}], ";"}]], "Input",
 CellLabel->
  "In[604]:=",ExpressionUUID->"361d07b6-0be9-45a8-bce6-70641d4539ac"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", "time", "]"}], "[", 
  RowBox[{"3", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[605]:=",ExpressionUUID->"aa0e8b99-63d4-4074-85f6-a4ea22c66023"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"Q", 
       StyleBox[GridBox[{
          {" ", "", " "},
          {"a", ",", "t"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ParamD[$CellContext`time][
    $CellContext`Q[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978211792398*^9},
 CellLabel->"Out[605]=",
 CellID->1089688242,ExpressionUUID->"794a4899-9adb-4568-9435-d3b99319bd89"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ParamD", "[", 
   RowBox[{"time", ",", "time"}], "]"}], "[", 
  RowBox[{"3", 
   RowBox[{"Q", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[606]:=",ExpressionUUID->"102f8a36-6784-410f-ba11-a6dd1a5b3dc6"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"Q", 
       StyleBox[GridBox[{
          {" ", "", " ", " "},
          {"a", ",", "t", "t"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`ParamD[$CellContext`time, $CellContext`time][
    $CellContext`Q[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978211928198*^9},
 CellLabel->"Out[606]=",
 CellID->649741678,ExpressionUUID->"99ef7db5-78d8-4244-9be3-a72541f49da9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"0893f67b-8882-4761-aa1a-99673c70eaad"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "Q", "]"}]], "Input",
 CellLabel->
  "In[607]:=",ExpressionUUID->"de8f687e-9d73-4da9-adf7-318ec17e80c5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "Q"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`Q],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978212060626*^9},
 CellLabel->"During evaluation of In[607]:=",
 CellID->254037151,ExpressionUUID->"2c553688-f03e-4ed8-b565-6c651e92edff"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefParameter", "[", "time", "]"}]], "Input",
 CellLabel->
  "In[608]:=",ExpressionUUID->"814a3b84-f815-40ea-a914-51a9bd722572"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefParameter", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"parameter\"\>", "\[InvisibleSpace]", "\<\" \"\>",
    "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["t",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`time,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefParameter, ": Undefined ", "parameter", 
   " ", $CellContext`time],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978212189172*^9},
 CellLabel->"During evaluation of In[608]:=",
 CellID->404906621,ExpressionUUID->"91bb81ac-a553-4136-b01e-7554c6db00bf"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.7. Lie derivatives", "Subsection",ExpressionUUID->"1ce5e5c3-cd56-4397-a63c-c19b8f523078"],

Cell[TextData[{
 StyleBox["xTensor`", "DisplayMath"],
 " does not define independent Lie derivatives, but a single command ",
 StyleBox["LieD",
  FontFamily->"Courier"],
 " which depends both on a contravariant vector field and on the expression \
that we want to derive."
}], "Text",ExpressionUUID->"54546be5-3498-4f7d-aec5-95603a2ad162"],

Cell[TextData[{
 StyleBox[ButtonBox["LieD",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LieD.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t", "DisplayMath"],
 StyleBox["Lie derivative\n", "Text"],
 StyleBox[ButtonBox["LieDToCovD",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LieDToCovD.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Expansion of Lie derivative in terms of covariant derivatives", 
  "Text"]
}], "DefinitionBox",ExpressionUUID->"c6742e28-54b3-4bcb-b390-196e36e2e3ed"],

Cell["Lie derivatives.", "Text",ExpressionUUID->"4950f6ad-4d93-4243-8a87-f055ae9dd624"],

Cell[CellGroupData[{

Cell["\<\
Lie derivatives of expressions with respect to a contravariant vector field \
are not automatically expanded in terms of derivatives.\
\>", "MathCaption",ExpressionUUID->"b06186aa-a51a-4720-9f70-3f15386d89ca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LieD", "[", 
   RowBox[{"v", "[", "a", "]"}], "]"}], "[", 
  RowBox[{"S", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", "c"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[609]:=",ExpressionUUID->"9b605bac-b45b-4e31-800d-cd28a5e6efe0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[ScriptCapitalL]", "v"], GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "c"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`LieD[
   $CellContext`v[$CellContext`a]][
   $CellContext`S[-$CellContext`b, $CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782197953653`*^9},
 CellLabel->"Out[609]=",
 CellID->1760093074,ExpressionUUID->"55dde540-ae92-4161-a7ee-0362c2c38632"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "However, we can expand them using any derivative (",
 StyleBox["PD ",
  FontFamily->"Courier"],
 "is the default)."
}], "MathCaption",ExpressionUUID->"2a3edfc0-e6e4-4a78-bd11-803fcb7234ec"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"LieDToCovD", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"LieD", "[", 
     RowBox[{"w", "[", "a", "]"}], "]"}], "[", 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", "c"}], "]"}], "]"}], ",", "PD"}], 
  "]"}]], "Input",
 CellLabel->
  "In[610]:=",ExpressionUUID->"ea973fe8-5e9c-41f6-b297-37fc66152ec1"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", "c"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`S[-$CellContext`b, $CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"w", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`w[$CellContext`c]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"w", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`w[$CellContext`a]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978219961639*^9},
 CellLabel->"Out[610]=",
 CellID->297057293,ExpressionUUID->"200407c5-c80a-436e-8c3c-042cc29330e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"LieDToCovD", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"LieD", "[", 
     RowBox[{"w", "[", "a", "]"}], "]"}], "[", 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", "c"}], "]"}], "]"}], ",", "Cd"}], 
  "]"}]], "Input",
 CellLabel->
  "In[611]:=",ExpressionUUID->"520ffa17-6ab4-4593-8681-86c74a0763a0"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`S[-$CellContext`b, $CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"w", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`w[$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {"w", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`b][
      $CellContext`w[$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978220112823*^9},
 CellLabel->"Out[611]=",
 CellID->1889927823,ExpressionUUID->"77b5a529-7d0b-4135-916e-00ced6d8776d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The derivative can also be expanded directly:", "MathCaption",ExpressionUUID->"34f08f9e-0658-453b-9d73-ab2cf419bcf4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LieD", "[", 
   RowBox[{
    RowBox[{"w", "[", "a", "]"}], ",", "Cd"}], "]"}], "[", 
  RowBox[{"S", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", "c"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[612]:=",ExpressionUUID->"1b00e22f-e1ed-43b3-a756-396321deb724"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"w", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`S[-$CellContext`b, $CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"w", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`w[$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {"w", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`b][
      $CellContext`w[$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289782202819033`*^9},
 CellLabel->"Out[612]=",
 CellID->586298384,ExpressionUUID->"cebc9cc4-df90-4d70-b96e-838a6be7acfd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["LieD",
  FontFamily->"Courier"],
 " knows the Leibniz rule:"
}], "MathCaption",ExpressionUUID->"4a5d6dbc-1894-44f3-9669-40986ab51978"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LieD", "[", 
   RowBox[{"w", "[", "a", "]"}], "]"}], "[", 
  RowBox[{
   RowBox[{"S", "[", "a", "]"}], 
   RowBox[{"T", "[", 
    RowBox[{"-", "a"}], "]"}]}], " ", "]"}]], "Input",
 CellLabel->
  "In[613]:=",ExpressionUUID->"6522b774-ec6c-4a0a-9dba-54a429f6d482"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[ScriptCapitalL]", "w"], GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`LieD[
      $CellContext`w[$CellContext`a]][
      $CellContext`S[$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[ScriptCapitalL]", "w"], GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`LieD[
      $CellContext`w[$CellContext`a]][
      $CellContext`T[-$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978220412878*^9},
 CellLabel->"Out[613]=",
 CellID->1261967452,ExpressionUUID->"866b912d-5ffd-4dc6-943d-16879891c955"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can have any expression in the \"directional argument\". Recall however \
that it is not a tensorial slot (it hence cannot be described with an \
abstract index):\
\>", "MathCaption",ExpressionUUID->"5317df4d-5265-47b2-9abb-6cb64a6d7102"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LieD", "[", 
   RowBox[{
    RowBox[{"7", 
     RowBox[{"r", "[", "]"}], 
     RowBox[{"w", "[", "a", "]"}]}], "+", 
    RowBox[{"v", "[", "a", "]"}]}], "]"}], "[", 
  RowBox[{"T", "[", "a", "]"}], "]"}]], "Input",
 CellLabel->
  "In[614]:=",ExpressionUUID->"1b7c6f2e-91d9-4699-a3ea-cf2edcb1ec67"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[ScriptCapitalL]", "v"], GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`LieD[
    $CellContext`v[$CellContext`a]][
    $CellContext`T[$CellContext`a]],
   Editable->False], "+", 
  RowBox[{"7", " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[ScriptCapitalL]", 
        TagBox[
         RowBox[{
          InterpretationBox[
           StyleBox["r",
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`r[],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"w", 
               StyleBox[GridBox[{
                  {"a"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`w[$CellContext`a],
           Editable->False]}],
         Short]], GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`LieD[$CellContext`r[] $CellContext`w[$CellContext`a]][
      $CellContext`T[$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978220547928*^9},
 CellLabel->"Out[614]=",
 CellID->1338467992,ExpressionUUID->"4967cf15-c6d0-44d3-a8a8-6ffbd8cbd01c"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.8. Directional derivatives", "Subsection",ExpressionUUID->"ba293095-3b69-43c8-a591-3ce503ab871e"],

Cell[TextData[{
 "From the mathematical point of view, the simplest concept of a derivative \
is that of a \"derivation\", a mapping from scalar fields to scalar fields. \
In fact, the concept of tanget space at a point is constructed as the space \
of derivations at that point. We can work with operators acting as \
derivations (or \"directional derivatives\") using the ",
 StyleBox["Dir",
  FontFamily->"Courier"],
 " head."
}], "Text",ExpressionUUID->"2d24f4fb-a1c1-4199-8ba9-d311b57bc149"],

Cell[CellGroupData[{

Cell[TextData[{
 "The object ",
 StyleBox["PD[Dir[v[a]]]",
  FontFamily->"Courier"],
 " can be interpreted as the directional derivative along the vector ",
 StyleBox["v[a]",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"64759844-33cb-499e-a3a1-036337811057"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"Dir", "[", 
    RowBox[{"w", "[", "a", "]"}], "]"}], "]"}], "[", 
  RowBox[{"r", "[", "]"}], "]"}]], "Input",
 CellLabel->
  "In[615]:=",ExpressionUUID->"3d5e679c-8409-47a0-914f-8730a1dbbc75"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "w"], "r"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[
   xAct`xTensor`Dir[
    $CellContext`w[$CellContext`a]]][
   $CellContext`r[]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782240101023`*^9},
 CellLabel->"Out[615]=",
 CellID->736370330,ExpressionUUID->"b24940b6-26e6-4667-92b6-fb5032918e4a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and it has all the expected properties:", "MathCaption",ExpressionUUID->"91f0e302-277a-4a5f-b69b-4cf6db973374"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"Dir", "[", " ", 
    RowBox[{
     RowBox[{"3", 
      RowBox[{"S", "[", 
       RowBox[{"a", ",", 
        RowBox[{"Dir", "[", 
         RowBox[{"v", "[", 
          RowBox[{"-", "b"}], "]"}], "]"}]}], "]"}]}], "+", 
     RowBox[{"w", "[", "a", "]"}]}], " ", "]"}], "]"}], "[", " ", 
  RowBox[{"1", "/", 
   RowBox[{
    RowBox[{"r", "[", "]"}], "^", "2"}]}], " ", "]"}]], "Input",
 CellLabel->
  "In[616]:=",ExpressionUUID->"b1a5fe00-c00c-4ba4-add1-d6f604767120"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"3", " ", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[PartialD]", "#"], "r"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`PD[
         xAct`xTensor`Dir[
          $CellContext`S[$CellContext`a, 
           xAct`xTensor`Dir[
            $CellContext`v[-$CellContext`b]]]]][
         $CellContext`r[]],
        Editable->False]}], "+", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", "w"], "r"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[
        xAct`xTensor`Dir[
         $CellContext`w[$CellContext`a]]][
        $CellContext`r[]],
       Editable->False]}], ")"}]}], 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "3"]]}]], "Output",
 CellChangeTimes->{3.7289782241604767`*^9},
 CellLabel->"Out[616]=",
 CellID->227878649,ExpressionUUID->"908981c2-9554-48ff-a769-591b06171d76"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Actually, it is possible to work directly with the derivations acting on \
scalars, saving 6 brackets (!) in the notation. For example we define the \
derivation ",
 Cell[BoxData[
  StyleBox["wder",
   FontWeight->"Plain",
   FontSlant->"Plain",
   FontTracking->"Plain",
   FontVariations->{"CompatibilityType"->0,
   "Outline"->False,
   "Shadow"->False,
   "StrikeThrough"->False,
   "Underline"->False},
   FontVariations -> {"Masked" -> False},
   FontVariations -> {"RotationAngle" -> 0}]], "Input",ExpressionUUID->
  "51450830-72ba-4046-ad86-27835d971ab8"],
 " associated to the vector ",
 Cell[BoxData[
  StyleBox[
   RowBox[{"w", "[", "a", "]"}],
   FontWeight->"Plain",
   FontSlant->"Plain",
   FontTracking->"Plain",
   FontVariations->{"CompatibilityType"->0,
   "Outline"->False,
   "Shadow"->False,
   "StrikeThrough"->False,
   "Underline"->False},
   FontVariations -> {"Masked" -> False},
   FontVariations -> {"RotationAngle" -> 0}]], "Input",ExpressionUUID->
  "1a1b93b5-f469-4250-bfd8-8b70817e9f9d"],
 ":"
}], "MathCaption",ExpressionUUID->"1ea0640e-0d6e-49e4-918c-0258aec23067"],

Cell[BoxData[
 RowBox[{"wder", ":=", 
  RowBox[{"PD", "[", 
   RowBox[{"Dir", "[", 
    RowBox[{"w", "[", "a", "]"}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[617]:=",ExpressionUUID->"47903dce-bacd-4dc5-adf4-7d5a16ee5c11"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"wder", "[", " ", 
  RowBox[{"r", "[", "]"}], " ", "]"}]], "Input",
 CellLabel->
  "In[618]:=",ExpressionUUID->"dbd9b37b-9efc-495c-b66b-3e26a85ed20f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "w"], "r"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[
   xAct`xTensor`Dir[
    $CellContext`w[$CellContext`a]]][
   $CellContext`r[]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978224287879*^9},
 CellLabel->"Out[618]=",
 CellID->993111142,ExpressionUUID->"b47e1803-4be1-49c7-80d2-fc47bb159f05"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Or even better, in prefix notation:", "MathCaption",ExpressionUUID->"412d1f19-b538-4195-a6bc-c6d17935364f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"wder", "@", 
  RowBox[{"r", "[", "]"}]}]], "Input",
 CellLabel->
  "In[619]:=",ExpressionUUID->"a5994b6e-4ab4-460d-bec7-b778f5f86f85"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[PartialD]", "w"], "r"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[
   xAct`xTensor`Dir[
    $CellContext`w[$CellContext`a]]][
   $CellContext`r[]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978224411665*^9},
 CellLabel->"Out[619]=",
 CellID->976059110,ExpressionUUID->"85522cee-e09c-4afb-8a50-8164de963da5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ScalarQ", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[620]:=",ExpressionUUID->"f8fa0ca4-59be-40be-8dea-206395358983"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978224543147*^9},
 CellLabel->"Out[620]=",
 CellID->2145365984,ExpressionUUID->"891eaf7b-0fd5-427c-a305-2b20199dc6d2"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "wder", "]"}]], "Input",
 CellLabel->
  "In[621]:=",ExpressionUUID->"1bf5b19f-f635-4d69-b6a7-5bb99c719086"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.9. Lie brackets", "Subsection",ExpressionUUID->"b6020147-9053-46ad-9704-cf7314c5ef80"],

Cell["\<\
The Lie bracket of two vector fields is a third vector field.\
\>", "Text",ExpressionUUID->"b5e18dd5-99fb-4aa3-95ba-da933a922f50"],

Cell[CellGroupData[{

Cell["Using abstract indices a bracket can be expressed as", "MathCaption",ExpressionUUID->"a533edd7-2548-4d98-bb20-9c9cee4ec5b1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Bracket", "[", 
   RowBox[{
    RowBox[{"v", "[", "b", "]"}], ",", 
    RowBox[{"w", "[", "b", "]"}]}], "]"}], "[", "a", "]"}]], "Input",
 CellLabel->
  "In[622]:=",ExpressionUUID->"16961d18-74cc-42f7-ab22-d9c93db109b3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"[", GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}], ",", GridBox[{
         {"w", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}], "]"}], 
      StyleBox[GridBox[{
         {"a"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Bracket[
   $CellContext`v[$CellContext`b], 
   $CellContext`w[$CellContext`b]][$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978226995029*^9},
 CellLabel->"Out[622]=",
 CellID->1216730947,ExpressionUUID->"35e06ab9-e692-4eea-a675-b869cd07aa50"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Note that the indices of the vector fields are ultraindices. The index of the \
resulting vector field is represented outside the bracket.\
\>", "Text",ExpressionUUID->"12ed87d3-944f-4628-80e8-8d66fa18aeec"],

Cell[CellGroupData[{

Cell["\<\
It has almost all expected properties: bilinearity, antisymmetry, derivation, \
..., but not yet the Jacobi rule.\
\>", "MathCaption",ExpressionUUID->"e2db8e34-9db5-4483-9ffb-c7c9d5efb56e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Bracket", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"v", "[", "b", "]"}], "+", 
     RowBox[{"w", "[", "b", "]"}]}], ",", 
    RowBox[{
     RowBox[{"3", 
      RowBox[{"w", "[", "b", "]"}]}], "+", 
     RowBox[{
      RowBox[{"r", "[", "]"}], 
      RowBox[{"v", "[", "b", "]"}]}]}]}], "]"}], "[", "a", "]"}]], "Input",
 CellLabel->
  "In[623]:=",ExpressionUUID->"d8667d58-fb8e-4b98-bae2-43d3b3103664"],

Cell[BoxData[
 RowBox[{
  RowBox[{"3", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"[", GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}], ",", GridBox[{
           {"w", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}], "]"}], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Bracket[
     $CellContext`v[$CellContext`b], 
     $CellContext`w[$CellContext`b]][$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"[", GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}], ",", GridBox[{
           {"w", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}], "]"}], 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Bracket[
     $CellContext`v[$CellContext`b], 
     $CellContext`w[$CellContext`b]][$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "v"], "r"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[
     xAct`xTensor`Dir[
      $CellContext`v[$CellContext`b]]][
     $CellContext`r[]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "w"], "r"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[
     xAct`xTensor`Dir[
      $CellContext`w[$CellContext`b]]][
     $CellContext`r[]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978227143029*^9},
 CellLabel->"Out[623]=",
 CellID->1257739495,ExpressionUUID->"a70cfffd-452d-4991-aae4-18d44773dea9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "[", "w", "]"}], ";"}]], "Input",
 CellLabel->
  "In[624]:=",ExpressionUUID->"acc83492-4121-4650-b6f8-01b1c3b9537a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "w"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`w],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978227262026*^9},
 CellLabel->"During evaluation of In[624]:=",
 CellID->1146155044,ExpressionUUID->"13c7f4c4-608d-477e-b1f3-636d64b9bc17"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["6.10. Derivatives on inner vbundles", "Subsection",ExpressionUUID->"5baa5a84-1a5b-47e6-ae3b-f75fea059ce5"],

Cell["\<\
It is possible to define a derivative which acts on an inner vector bundle \
(as used in Yang-Mills gauge theories).\
\>", "Text",ExpressionUUID->"766a2028-dd53-4216-91be-2d9d69c393e4"],

Cell[CellGroupData[{

Cell["Define a complex vector bundle:", "MathCaption",ExpressionUUID->"001d1f57-7f67-468d-a614-132d6edf9a41"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefVBundle", "[", 
  RowBox[{"InnerC", ",", "M3", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{
    "\[GothicCapitalA]", ",", "\[GothicCapitalB]", ",", "\[GothicCapitalC]", 
     ",", "\[GothicCapitalD]", ",", "\[GothicCapitalE]", ",", 
     "\[GothicCapitalF]", ",", "\[GothicCapitalG]", ",", 
     "\[GothicCapitalH]"}], "}"}], ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->
  "In[625]:=",ExpressionUUID->"9939a251-ca48-4cff-bcff-4468637a767f"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "InnerC", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`InnerC, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230115168*^9},
 CellLabel->"During evaluation of In[625]:=",
 CellID->659454615,ExpressionUUID->"f19ce1be-8245-44be-b032-d47330ae1e94"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle \"\>", "\[InvisibleSpace]", 
   "InnerC\[Dagger]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Assuming fixed anti-isomorphism between InnerC \
and InnerC\[Dagger]\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "conjugated vbundle ", $CellContext`InnerC\[Dagger], ". ", 
   "Assuming fixed anti-isomorphism between InnerC and InnerC\[Dagger]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230132415*^9},
 CellLabel->"During evaluation of In[625]:=",
 CellID->290612138,ExpressionUUID->"37c014a3-1496-410f-9025-f33cd1f0a880"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define a connection on a complex inner vbundle:", "MathCaption",ExpressionUUID->"0a251e7b-6ed8-4ceb-9cb6-8391edce4646"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"ICD", "[", 
    RowBox[{"-", "a"}], "]"}], ",", "InnerC", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<*\>\"", ",", "\"\<D\>\""}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[626]:=",ExpressionUUID->"f70d5cdf-3dc4-4103-92c2-8bef4f154530"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"ICD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`ICD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230267747*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->940390088,ExpressionUUID->"552d49bd-f8a7-490b-a10c-fdd59ea45b1d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionICD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionICD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782302845583`*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->1804050743,ExpressionUUID->"2de43790-d042-426c-b197-064c55ec4cdb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelICD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelICD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230300942*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->359225608,ExpressionUUID->"ba5f94d6-95db-464a-8a7a-794755907d54"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannICD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannICD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782303185987`*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->928261939,ExpressionUUID->"57f02b9b-efad-4760-912b-2df325d335ee"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciICD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciICD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230335187*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->1977196965,ExpressionUUID->"35d7fe87-f13d-4421-91ce-a33574161cd7"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.7289782303507347`*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->587437077,ExpressionUUID->"8570cbe2-6713-4d42-aa49-ef277110a0f0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelICD", "[", 
    RowBox[{"\[GothicCapitalA]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelICD[$CellContext`\[GothicCapitalA], \
-$CellContext`b, -$CellContext`\[GothicCapitalC]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230366996*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->1072627880,ExpressionUUID->"7c4bf037-3126-455f-9418-705dcd2da0d5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelICD\[Dagger]", "[", 
    RowBox[{"\[GothicCapitalA]\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]\[Dagger]"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelICD\[Dagger][$CellContext`\[GothicCapitalA]\
\[Dagger], -$CellContext`b, -$CellContext`\[GothicCapitalC]\[Dagger]], ". ", 
   ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230383051*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->1679046242,ExpressionUUID->"d5e179ff-71db-4b5e-97f2-d0810b2068ce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"FRiemannICD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]"}], ",", "\[GothicCapitalD]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "FRiemann tensor ", 
   $CellContext`FRiemannICD[-$CellContext`a, -$CellContext`b, -$CellContext`\
\[GothicCapitalC], $CellContext`\[GothicCapitalD]], ". ", 
   "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230399522*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->1795067714,ExpressionUUID->"1527bd07-f64c-4dc0-acfc-72c8ded6fe87"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"FRiemannICD\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]\[Dagger]"}], ",", 
     "\[GothicCapitalD]\[Dagger]"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "FRiemann tensor ", 
   $CellContext`FRiemannICD\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`\[GothicCapitalC]\[Dagger], $CellContext`\[GothicCapitalD]\
\[Dagger]], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978230416738*^9},
 CellLabel->"During evaluation of In[626]:=",
 CellID->1852796142,ExpressionUUID->"47bb244f-4532-4248-8793-dec77e9df754"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We see that a single torsion tensor has been defined, but two Christoffel \
and two Riemann tensors. The inner Christoffel is called ",
 StyleBox["AChristoffelICD",
  FontFamily->"Courier"],
 " and the inner Riemann tensor is called ",
 StyleBox["FRiemannICD",
  FontFamily->"Courier"],
 ". They are both complex tensors and therefore have their complex \
conjugates. Note the different vbundles of their indices."
}], "MathCaption",ExpressionUUID->"26e4fac4-e4cd-42d0-9fa8-408deb01147b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "AChristoffelICD"}]], "Input",
 CellLabel->
  "In[627]:=",ExpressionUUID->"7c4b7e34-53f5-413f-8602-259c046cda73"],

Cell[CellGroupData[{

Cell["Global`AChristoffelICD", "Print", "PrintUsage",
 CellChangeTimes->{3.728978230570784*^9},
 CellTags->"Info6273728981830-9181623",
 CellID->1431877094,ExpressionUUID->"519da795-919b-40cf-b068-f23f25068295"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "AChristoffelICD", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`AChristoffelICD],
          Editable->False], "^=", "AChristoffelICD\[Dagger]"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "AChristoffelICD", "]"}], "^=", 
         RowBox[{"{", 
          
          RowBox[{"\<\"nonsymmetric AChristoffel tensor\"\>", 
           ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "AChristoffelICD", "]"}], "^=", 
         
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "AChristoffelICD", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC,
            Editable->False], ",", "M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", "AChristoffelICD", "]"}], "^=", "ICD"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "AChristoffelICD", "]"}], 
         "^=", "\<\"A[D]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ServantsOf", "[", "AChristoffelICD", "]"}], "^=", 
         RowBox[{"{", "AChristoffelICD\[Dagger]", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "AChristoffelICD", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC,
            Editable->False], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["InnerC",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`InnerC,
             Editable->False]}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "AChristoffelICD", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "AChristoffelICD", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"AChristoffel", ",", "ICD", ",", "PD"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "AChristoffelICD", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`AChristoffelICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782305930023`*^9},
 CellTags->"Info6273728981830-9181623",
 CellID->856570576,ExpressionUUID->"7cc7d0aa-97b2-46e4-8186-fe70e9c01262"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "FRiemannICD"}]], "Input",
 CellLabel->
  "In[628]:=",ExpressionUUID->"68f8f6eb-5c45-457d-aae9-1fa3ca02237a"],

Cell[CellGroupData[{

Cell["Global`FRiemannICD", "Print", "PrintUsage",
 CellChangeTimes->{3.728978231113431*^9},
 CellTags->"Info6283728981830-9181623",
 CellID->1798784991,ExpressionUUID->"25bd47ae-1722-4fac-9b5f-ad3965650133"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "FRiemannICD", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`FRiemannICD],
          Editable->False], "^=", "FRiemannICD\[Dagger]"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"{", 
          
          RowBox[{"\<\"FRiemann tensor\"\>", 
           ",", "\<\"Antisymmetric only in the first pair.\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC,
            Editable->False], ",", "M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", "FRiemannICD", "]"}], "^=", "ICD"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "FRiemannICD", "]"}], "^=", "\<\"F[D]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"ServantsOf", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"{", "FRiemannICD\[Dagger]", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["InnerC",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`InnerC,
             Editable->False]}], ",", 
           InterpretationBox[
            StyleBox["InnerC",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"1", ",", "2"}], "}"}], ",", 
           RowBox[{"GenSet", "[", 
            InterpretationBox[
             StyleBox[
              RowBox[{"-", 
               RowBox[{"(", "1", ",", "2", ")"}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             -xAct`xPerm`Cycles[{1, 2}],
             Editable->False], "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "FRiemannICD", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"FRiemann", ",", "ICD"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "FRiemannICD", "]"}], "^=", "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`FRiemannICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978231137081*^9},
 CellTags->"Info6283728981830-9181623",
 CellID->2052044918,ExpressionUUID->"8d196b37-dc33-4d02-aecc-a9fcbf76e838"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "AChristoffelICD\[Dagger]"}]], "Input",
 CellLabel->
  "In[629]:=",ExpressionUUID->"4c5ea162-50a2-4492-8e00-d7cb539072d4"],

Cell[CellGroupData[{

Cell["Global`AChristoffelICD\[Dagger]", "Print", "PrintUsage",
 CellChangeTimes->{3.728978231646302*^9},
 CellTags->"Info6293728981831-9181623",
 CellID->131566006,ExpressionUUID->"5fcf77bd-09a9-4707-b110-ac2ccc6bfe59"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "AChristoffelICD\[Dagger]", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`AChristoffelICD\[Dagger]],
          Editable->False], "^=", "AChristoffelICD"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          
          RowBox[{"\<\"nonsymmetric AChristoffel tensor\"\>", 
           ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{
         "DependenciesOfTensor", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC\[Dagger]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC\[Dagger],
            Editable->False], ",", "M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         "AChristoffelICD"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "AChristoffelICD\[Dagger]", "]"}], 
         "^=", "\<\"A[D]\[Dagger]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC\[Dagger]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC\[Dagger],
            Editable->False], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["InnerC\[Dagger]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`InnerC\[Dagger],
             Editable->False]}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{
         "SymmetryGroupOfTensor", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{"GenSet", "[", "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"AChristoffel", ",", "ICD", ",", "PD"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "AChristoffelICD\[Dagger]", "]"}], "^=", 
         "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`AChristoffelICD\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978231668528*^9},
 CellTags->"Info6293728981831-9181623",
 CellID->1469190079,ExpressionUUID->"e36f34d4-fbb6-4f39-a840-90e6ad5ebd86"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "FRiemannICD\[Dagger]"}]], "Input",
 CellLabel->
  "In[630]:=",ExpressionUUID->"862db3f2-a585-4bcf-a28a-3b888e787291"],

Cell[CellGroupData[{

Cell["Global`FRiemannICD\[Dagger]", "Print", "PrintUsage",
 CellChangeTimes->{3.728978232133226*^9},
 CellTags->"Info6303728981831-9181623",
 CellID->910684547,ExpressionUUID->"e2ece981-4b04-4bc6-98a0-fb537d10ef48"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "FRiemannICD\[Dagger]", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`FRiemannICD\[Dagger]],
          Editable->False], "^=", "FRiemannICD"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          
          RowBox[{"\<\"FRiemann tensor\"\>", 
           ",", "\<\"Antisymmetric only in the first pair.\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           InterpretationBox[
            StyleBox["InnerC\[Dagger]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC\[Dagger],
            Editable->False], ",", "M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"MasterOf", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         "FRiemannICD"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "FRiemannICD\[Dagger]", "]"}], 
         "^=", "\<\"F[D]\[Dagger]\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["InnerC\[Dagger]",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`InnerC\[Dagger],
             Editable->False]}], ",", 
           InterpretationBox[
            StyleBox["InnerC\[Dagger]",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`InnerC\[Dagger],
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "FRiemannICD\[Dagger]", "]"}], 
         "^=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"1", ",", "2"}], "}"}], ",", 
           RowBox[{"GenSet", "[", 
            InterpretationBox[
             StyleBox[
              RowBox[{"-", 
               RowBox[{"(", "1", ",", "2", ")"}]}],
              ShowAutoStyles->False,
              AutoSpacing->False],
             -xAct`xPerm`Cycles[{1, 2}],
             Editable->False], "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"FRiemann", ",", "ICD"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "FRiemannICD\[Dagger]", "]"}], "^=", 
         "True"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`FRiemannICD\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978232158225*^9},
 CellTags->"Info6303728981831-9181623",
 CellID->759229260,ExpressionUUID->"3ddd1743-5669-46ff-aed9-4d27b5ffd144"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The derivative ",
 StyleBox["ICD",
  FontFamily->"Courier"],
 " has curvature in both the tangent bundle ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 " and in the inner bundle ",
 StyleBox["InnerC",
  FontFamily->"Courier"],
 ". That becomes most apparent when changing to a different derivative:"
}], "MathCaption",ExpressionUUID->"4bbba8af-7283-424f-b0da-86d37c92c637"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"X", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}], ",", "M3", ",", 
   RowBox[{"Dagger", "\[Rule]", "Complex"}]}], "]"}]], "Input",
 CellLabel->
  "In[631]:=",ExpressionUUID->"dca1197c-db66-4443-9a88-53b997314897"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"X", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalB]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782327331467`*^9},
 CellLabel->"During evaluation of In[631]:=",
 CellID->18342346,ExpressionUUID->"efa2f69d-f92b-4237-875c-4e4e2eb6cc31"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"X\[Dagger]", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]\[Dagger]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalB]\
\[Dagger]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978232751936*^9},
 CellLabel->"During evaluation of In[631]:=",
 CellID->1189647701,ExpressionUUID->"b4b786c4-6283-4dbb-a2d8-9f9123a0b2f5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ICD", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"X", "[", 
   RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[632]:=",ExpressionUUID->"f71e6421-db58-46ed-aaf8-6fd9f6fa9224"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "c"], GridBox[{
      {"X", 
       StyleBox[GridBox[{
          {"a", "\[GothicCapitalB]"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ICD[-$CellContext`c][
   $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalB]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782329653482`*^9},
 CellLabel->"Out[632]=",
 CellID->1336157137,ExpressionUUID->"efce94b3-3fb7-4035-8b0f-dfcb54d319a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "ICD"}], "]"}]], "Input",
 CellLabel->
  "In[633]:=",ExpressionUUID->"4317f374-5a4c-4392-a194-13863464fd4f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"A", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"\[GothicCapitalB]", " ", " "},
           {" ", "c", "\[GothicCapitalA]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`AChristoffelICD[$CellContext`\[GothicCapitalB], \
-$CellContext`c, -$CellContext`\[GothicCapitalA]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalA]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalA]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelICD[$CellContext`a, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X", 
        StyleBox[GridBox[{
           {"b", "\[GothicCapitalB]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X[$CellContext`b, $CellContext`\[GothicCapitalB]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {"X", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalB]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalB]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897823310289*^9},
 CellLabel->"Out[633]=",
 CellID->556504880,ExpressionUUID->"06e998a2-b538-41b2-b3dc-c946e177eb0e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "PD", ",", "ICD"}], "]"}]], "Input",
 CellLabel->
  "In[634]:=",ExpressionUUID->"1c4d9b76-b26b-4081-a248-bec7d522bea4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "c"], GridBox[{
      {"X", 
       StyleBox[GridBox[{
          {"a", "\[GothicCapitalB]"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ICD[-$CellContext`c][
   $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalB]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782332995253`*^9},
 CellLabel->"Out[634]=",
 CellID->886235475,ExpressionUUID->"f87e6009-880f-4c6c-840a-fff673f91e48"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or when commuting two derivatives:", "MathCaption",ExpressionUUID->"127adeab-ebf1-434e-bee9-2453fd82a41a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ICD", "[", 
   RowBox[{"-", "c"}], "]"}], "@", 
  RowBox[{
   RowBox[{"ICD", "[", 
    RowBox[{"-", "d"}], "]"}], "@", 
   RowBox[{"X", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[635]:=",ExpressionUUID->"8a81acea-58e3-4c19-9933-801be4615b80"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "c"], 
    SubscriptBox["D", "d"], GridBox[{
      {"X", 
       StyleBox[GridBox[{
          {"a", "\[GothicCapitalB]"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ICD[-$CellContext`c][
   $CellContext`ICD[-$CellContext`d][
    $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalB]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782337129*^9},
 CellLabel->"Out[635]=",
 CellID->13052733,ExpressionUUID->"df53dfaa-391d-41b7-8d16-ccb1779bfd5f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SortCovDs", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[636]:=",ExpressionUUID->"66a241f2-e7a9-4688-91e5-d0487777e60b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"F", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "\[GothicCapitalB]"},
           {"d", "c", "\[GothicCapitalA]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`FRiemannICD[-$CellContext`d, -$CellContext`c, -$CellContext`\
\[GothicCapitalA], $CellContext`\[GothicCapitalB]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalA]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalA]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"d", "c", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannICD[-$CellContext`d, -$CellContext`c, \
-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X", 
        StyleBox[GridBox[{
           {"b", "\[GothicCapitalB]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X[$CellContext`b, $CellContext`\[GothicCapitalB]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "d"], 
     SubscriptBox["D", "c"], GridBox[{
       {"X", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalB]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ICD[-$CellContext`d][
    $CellContext`ICD[-$CellContext`c][
     $CellContext`X[$CellContext`a, $CellContext`\[GothicCapitalB]]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978233971252*^9},
 CellLabel->"Out[636]=",
 CellID->333951537,ExpressionUUID->"201bf36c-8ec0-438c-95f5-77ff9c867b73"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["For their complex conjugates:", "MathCaption",ExpressionUUID->"36e1b7d9-c936-4c85-8818-388bd6fa554e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ICD", "[", 
    RowBox[{"-", "c"}], "]"}], "[", 
   RowBox[{"X", "[", 
    RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}], "]"}], "//", 
  "Dagger"}]], "Input",
 CellLabel->
  "In[637]:=",ExpressionUUID->"ad97cd95-e30e-4141-b7fe-f69bd478eac7"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "c"], GridBox[{
      {"X\[Dagger]", 
       StyleBox[GridBox[{
          {"a", "\[GothicCapitalB]\[Dagger]"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ICD[-$CellContext`c][
   $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalB]\
\[Dagger]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978234246624*^9},
 CellLabel->"Out[637]=",
 CellID->1348712748,ExpressionUUID->"f2287a8a-c75d-49b8-a61f-c4a365eafd31"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "ICD"}], "]"}]], "Input",
 CellLabel->
  "In[638]:=",ExpressionUUID->"5d3fbb29-ecac-44fa-8695-55c9eeece61c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{
         RowBox[{"A", "[", "D", "]"}], "\[Dagger]"}], 
        StyleBox[GridBox[{
           {"\[GothicCapitalB]\[Dagger]", " ", " "},
           {" ", "c", "\[GothicCapitalA]\[Dagger]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`AChristoffelICD\[Dagger][$CellContext`\[GothicCapitalB]\
\[Dagger], -$CellContext`c, -$CellContext`\[GothicCapitalA]\[Dagger]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X\[Dagger]", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalA]\[Dagger]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalA]\
\[Dagger]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelICD[$CellContext`a, -$CellContext`c, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X\[Dagger]", 
        StyleBox[GridBox[{
           {"b", "\[GothicCapitalB]\[Dagger]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X\[Dagger][$CellContext`b, $CellContext`\[GothicCapitalB]\
\[Dagger]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "c"], GridBox[{
       {"X\[Dagger]", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalB]\[Dagger]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`c][
    $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalB]\
\[Dagger]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978234832972*^9},
 CellLabel->"Out[638]=",
 CellID->1125261551,ExpressionUUID->"a282bfbe-be47-491f-9450-2e70f3724d80"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ICD", "[", 
    RowBox[{"-", "c"}], "]"}], "@", 
   RowBox[{
    RowBox[{"ICD", "[", 
     RowBox[{"-", "d"}], "]"}], "@", 
    RowBox[{"X", "[", 
     RowBox[{"a", ",", "\[GothicCapitalB]"}], "]"}]}]}], "//", 
  "Dagger"}]], "Input",
 CellLabel->
  "In[639]:=",ExpressionUUID->"7f2e09cd-6399-4066-8a6c-7d8d9986c3af"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "c"], 
    SubscriptBox["D", "d"], GridBox[{
      {"X\[Dagger]", 
       StyleBox[GridBox[{
          {"a", "\[GothicCapitalB]\[Dagger]"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ICD[-$CellContext`c][
   $CellContext`ICD[-$CellContext`d][
    $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalB]\
\[Dagger]]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978234967828*^9},
 CellLabel->"Out[639]=",
 CellID->1529892175,ExpressionUUID->"94826702-3337-4442-9d43-65d4610ee5f2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SortCovDs", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[640]:=",ExpressionUUID->"b363bf6e-13b5-43ea-af2a-75d93b1bc31f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{
         RowBox[{"F", "[", "D", "]"}], "\[Dagger]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "\[GothicCapitalB]\[Dagger]"},
           {"d", "c", "\[GothicCapitalA]\[Dagger]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`FRiemannICD\[Dagger][-$CellContext`d, -$CellContext`c, \
-$CellContext`\[GothicCapitalA]\[Dagger], $CellContext`\[GothicCapitalB]\
\[Dagger]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X\[Dagger]", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalA]\[Dagger]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalA]\
\[Dagger]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"d", "c", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannICD[-$CellContext`d, -$CellContext`c, \
-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"X\[Dagger]", 
        StyleBox[GridBox[{
           {"b", "\[GothicCapitalB]\[Dagger]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`X\[Dagger][$CellContext`b, $CellContext`\[GothicCapitalB]\
\[Dagger]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "d"], 
     SubscriptBox["D", "c"], GridBox[{
       {"X\[Dagger]", 
        StyleBox[GridBox[{
           {"a", "\[GothicCapitalB]\[Dagger]"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ICD[-$CellContext`d][
    $CellContext`ICD[-$CellContext`c][
     $CellContext`X\[Dagger][$CellContext`a, $CellContext`\[GothicCapitalB]\
\[Dagger]]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978235105771*^9},
 CellLabel->"Out[640]=",
 CellID->1076079694,ExpressionUUID->"af80de5c-c56b-4cb3-999b-95849585cd9d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "X", "]"}]], "Input",
 CellLabel->
  "In[641]:=",ExpressionUUID->"163854a0-a861-4903-bc9f-17cd15268a64"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "X\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`X\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978235272023*^9},
 CellLabel->"During evaluation of In[641]:=",
 CellID->1711625671,ExpressionUUID->"bc9ad22d-9018-4139-8821-b7b1e4926cf5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "X"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`X],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978235297348*^9},
 CellLabel->"During evaluation of In[641]:=",
 CellID->1749200932,ExpressionUUID->"5b9cb26f-7ad5-42bc-b1d2-a8f39378c62d"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "An important concept, introduced in version 0.9.1 of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", is that of extension of a derivative.  Given a derivative ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 " it is possible to define a second one ",
 StyleBox["CD",
  FontFamily->"Courier"],
 " in such a way that the associated tensors of ",
 StyleBox["CD",
  FontFamily->"Courier"],
 " on its tangent bundle will be those of ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 ". We say then that ",
 StyleBox["CD",
  FontFamily->"Courier"],
 " extends ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 ", or that ",
 StyleBox["CD",
  FontFamily->"Courier"],
 " has been constructed by extension of ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"00e25242-9ad9-468c-b4a1-f4283104d420"],

Cell[CellGroupData[{

Cell[TextData[{
 "Let us take the ",
 StyleBox["Cd",
  FontFamily->"Courier"],
 " derivative as base covd:"
}], "MathCaption",ExpressionUUID->"e9907f0c-d120-49ef-9b82-cc69c7e26bbf"],

Cell[CellGroupData[{

Cell[BoxData["$CovDs"], "Input",
 CellLabel->
  "In[642]:=",ExpressionUUID->"afb606c6-79d4-44af-a698-2072e0c25507"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"PD", ",", "Cd", ",", "ICD"}], "}"}]], "Output",
 CellChangeTimes->{3.728978235453761*^9},
 CellLabel->"Out[642]=",
 CellID->1916420341,ExpressionUUID->"ea848941-fbee-4591-8ebe-7ceeba2e6304"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We now extend that derivative to ",
 StyleBox["CD",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"cf043f7c-ea17-4ad9-8e1c-6d20afec86b4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], ",", "InnerC", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<#\>\"", ",", "\"\<D\>\""}], "}"}]}], ",", 
   RowBox[{"ExtendedFrom", "\[Rule]", "Cd"}]}], "]"}]], "Input",
 CellLabel->
  "In[643]:=",ExpressionUUID->"c3129cdb-c2db-4561-afa9-b9341f909618"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978235621216*^9},
 CellLabel->"During evaluation of In[643]:=",
 CellID->1316933481,ExpressionUUID->"cbca5d5f-bf0f-4c18-a160-965ab455e5e4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCdCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing tensor ", 
   $CellContext`ChristoffelCdCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978235642514*^9},
 CellLabel->"During evaluation of In[643]:=",
 CellID->1734549185,ExpressionUUID->"189038da-e105-45df-aa1b-20cb119e9598"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelCD", "[", 
    RowBox[{"\[GothicCapitalA]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelCD[$CellContext`\[GothicCapitalA], \
-$CellContext`b, -$CellContext`\[GothicCapitalC]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782356679573`*^9},
 CellLabel->"During evaluation of In[643]:=",
 CellID->477644270,ExpressionUUID->"ff80f269-feba-46e4-83ca-23215b3f5c71"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"AChristoffelCD\[Dagger]", "[", 
    RowBox[{"\[GothicCapitalA]\[Dagger]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]\[Dagger]"}]}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "nonsymmetric AChristoffel tensor ", 
   $CellContext`AChristoffelCD\[Dagger][$CellContext`\[GothicCapitalA]\
\[Dagger], -$CellContext`b, -$CellContext`\[GothicCapitalC]\[Dagger]], ". ", 
   ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978235690667*^9},
 CellLabel->"During evaluation of In[643]:=",
 CellID->473373404,ExpressionUUID->"fa0c5709-6927-4335-b207-536afaa3b80a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"FRiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]"}], ",", "\[GothicCapitalD]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "FRiemann tensor ", 
   $CellContext`FRiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`\
\[GothicCapitalC], $CellContext`\[GothicCapitalD]], ". ", 
   "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782357117043`*^9},
 CellLabel->"During evaluation of In[643]:=",
 CellID->1707372540,ExpressionUUID->"f0cce46d-8ab8-41e6-9645-b3bdaf7f26ae"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"FRiemannCD\[Dagger]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]\[Dagger]"}], ",", 
     "\[GothicCapitalD]\[Dagger]"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "FRiemann tensor ", 
   $CellContext`FRiemannCD\[Dagger][-$CellContext`a, -$CellContext`b, \
-$CellContext`\[GothicCapitalC]\[Dagger], $CellContext`\[GothicCapitalD]\
\[Dagger]], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782357328243`*^9},
 CellLabel->"During evaluation of In[643]:=",
 CellID->1133788566,ExpressionUUID->"2f149c8c-ae94-4aac-bf0a-974ec1c28f7b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now the tensors associated to the extended derivative will be those of the \
base derivative:\
\>", "MathCaption",ExpressionUUID->"f9cf64d5-459a-4a61-962f-e9b42825abde"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], ",", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[644]:=",ExpressionUUID->"1f027cf2-8da2-4fed-9840-dad987a32405"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "d"},
           {"a", "b", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
    Editable->False], ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728978235884254*^9},
 CellLabel->"Out[644]=",
 CellID->962028931,ExpressionUUID->"d6a1217c-cdf1-4dfc-843c-529e5704731d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
And then we have the inner curvature of the extended derivative:\
\>", "MathCaption",ExpressionUUID->"701004f9-4705-4297-94eb-712725739681"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"AChristoffelCD", "[", 
    RowBox[{"\[GothicCapitalA]", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]"}]}], "]"}], ",", 
   RowBox[{"FRiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "\[GothicCapitalC]"}], ",", "\[GothicCapitalD]"}], "]"}]}], 
  "}"}]], "Input",
 CellLabel->
  "In[645]:=",ExpressionUUID->"39142340-08af-4887-a36a-4c0a348b7051"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"A", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"\[GothicCapitalA]", " ", " "},
           {" ", "b", "\[GothicCapitalC]"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`AChristoffelCD[$CellContext`\[GothicCapitalA], \
-$CellContext`b, -$CellContext`\[GothicCapitalC]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"F", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "\[GothicCapitalD]"},
           {"a", "b", "\[GothicCapitalC]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`FRiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`\
\[GothicCapitalC], $CellContext`\[GothicCapitalD]],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978236066374*^9},
 CellLabel->"Out[645]=",
 CellID->415865054,ExpressionUUID->"e352f8ed-e7ac-4794-beda-ebd5da151d06"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can undefine the extended derivative, without removing any of the base \
derivative objects:\
\>", "MathCaption",ExpressionUUID->"d8ecfcb8-cdbb-4e0e-b992-419a9912a4c3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefCovD", "[", "CD", "]"}]], "Input",
 CellLabel->
  "In[646]:=",ExpressionUUID->"a5f0721b-1222-492f-8e91-0ba2c63b58f6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "AChristoffelCD\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", 
   " ", $CellContext`AChristoffelCD\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978236151801*^9},
 CellLabel->"During evaluation of In[646]:=",
 CellID->1532175240,ExpressionUUID->"373cd9a0-6592-4413-bd55-db658750312b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "AChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", " ", $CellContext`AChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978236175071*^9},
 CellLabel->"During evaluation of In[646]:=",
 CellID->792847503,ExpressionUUID->"d821b322-5ae1-40da-b4d7-af5c83cee540"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCdCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCdCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978236194686*^9},
 CellLabel->"During evaluation of In[646]:=",
 CellID->251953286,ExpressionUUID->"5152685e-9c0f-4deb-9034-e5ff173b1154"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "FRiemannCD\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannCD\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978236213135*^9},
 CellLabel->"During evaluation of In[646]:=",
 CellID->1888609565,ExpressionUUID->"cd21f6fb-1f33-4318-a335-845b9b2d3aa7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "FRiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978236231531*^9},
 CellLabel->"During evaluation of In[646]:=",
 CellID->2052656295,ExpressionUUID->"b0bf257a-928a-421f-a413-e31b434705b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978236251218*^9},
 CellLabel->"During evaluation of In[646]:=",
 CellID->1817044657,ExpressionUUID->"84135a63-52d7-4fac-97f3-9c3eea2c6f31"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ChristoffelCd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"TorsionCd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"RiemannCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], ",", 
   RowBox[{"RicciCd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[647]:=",ExpressionUUID->"9de6d6bb-2bc6-40b3-a6a1-1db1685966b6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCd[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False], ",", "0", ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "d"},
           {"a", "b", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCd[-$CellContext`a, -$CellContext`b],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289782364138727`*^9},
 CellLabel->"Out[647]=",
 CellID->1852089859,ExpressionUUID->"d5d074ce-3de0-4cb1-8353-8a86a005da8c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], ",", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[648]:=",ExpressionUUID->"0dc6c194-7f0b-4ca6-8568-0d230d68c792"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], ",", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978236547894*^9},
 CellLabel->"Out[648]=",
 CellID->298796948,ExpressionUUID->"d46d2d73-e8bc-4191-9663-b5a90a490dc2"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["7. Metrics", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}, {3.728976670882564*^9, 
  3.728976674720446*^9}, {3.7289768454424953`*^9, 3.728976848920129*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1069634512,ExpressionUUID->"ba260751-a6a0-4d09-85a7-cab77ca50fa1"],

Cell[CellGroupData[{

Cell["7.1. Definition", "Subsection",ExpressionUUID->"c5eda06a-00e0-44e6-b436-fa80f5bb631b"],

Cell["Now we introduce metrics on the vbundles.", "Text",ExpressionUUID->"53619325-7d62-40b1-bc18-bd1c28d73315"],

Cell[TextData[{
 StyleBox[ButtonBox["DefMetric",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefMetric.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Define a metric on a single ", "Text"],
 "vbundle",
 StyleBox["\n", "Text"],
 StyleBox[ButtonBox["DefProductMetric",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefProductMetric.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Define a metric on a product ", "Text"],
 "vbundle",
 StyleBox["\n", "Text"],
 StyleBox[ButtonBox["UndefMetric",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefMetric.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Undefine a metric", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["$Metrics",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$Metrics.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined metrics\n", "Text"],
 StyleBox[ButtonBox["$ProductMetrics",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$ProductMetrics.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["List of defined product metrics", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["MetricQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MetricQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Validate name of metric", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.728978243418363*^9, 
  3.728978244769761*^9}},ExpressionUUID->"57bc3550-d40d-4c40-a3c3-\
6308bb201068"],

Cell["Definition of a metric.", "Text",ExpressionUUID->"06a9f87d-6ec6-48ab-87da-d6541351219d"],

Cell[CellGroupData[{

Cell[TextData[{
 "We define a metric ",
 StyleBox["metricg",
  FontFamily->"Courier"],
 " with negative determinant. The associated covariant derivative will be \
called ",
 StyleBox["CD",
  FontFamily->"Courier"],
 " and will be represented with a semicolon under ",
 StyleBox["IndexForm",
  FontFamily->"Courier"],
 ". All the needed associated tensors are defined (note that epsilon has a \
suffix ",
 StyleBox["metricg",
  FontFamily->"Courier"],
 " and the other tensors have a suffix ",
 StyleBox["CD",
  FontFamily->"Courier"],
 "). Because we use indices of ",
 StyleBox["\[DoubleStruckCapitalT]M",
  FontFamily->"Courier"],
 " it is understood that ",
 StyleBox["metricg",
  FontFamily->"Courier"],
 " will be a metric on ",
 StyleBox["M3",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"d64e1593-1fd5-4647-899d-d71b66d45c0c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
   RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[649]:=",ExpressionUUID->"6dca4d8c-1eb7-439c-8f3e-ac41944b22c2"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metricg[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250010631*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->829774745,ExpressionUUID->"34c64916-6003-4cef-9bb2-a4eee95b600c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetricg[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250026477*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->265335597,ExpressionUUID->"e2315ba0-e805-435a-8de3-86b08a383563"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250044648*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1319327607,ExpressionUUID->"9e82f6aa-57ff-4a03-a67f-c8e944cea348"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782500621567`*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1170435574,ExpressionUUID->"80a12a66-47d1-4273-b28b-45121cacc07e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250080269*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->560485698,ExpressionUUID->"0857f64d-25ac-4e77-8a5a-e87007c60414"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250098236*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1219850544,ExpressionUUID->"8105c94d-e565-4c34-a850-1473d1f1a122"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250116066*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1715099,ExpressionUUID->"5d77d0fd-f4d2-448b-a3d4-49fb791fb8a9"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.7289782501331997`*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->908381482,ExpressionUUID->"c4cea385-a714-4213-8b9c-093adea4acb8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782501499233`*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->706321451,ExpressionUUID->"047426d4-b18e-4ac7-9ad5-f29468d57faa"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728978250165893*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->2002509988,ExpressionUUID->"d7714970-e8e1-45b6-8abe-641dc265916a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250182768*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->821652532,ExpressionUUID->"d6d5bd93-4fb8-4462-82b1-23e40e08e153"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250200748*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1194255114,ExpressionUUID->"da8ee768-0ce6-46d1-9f92-49846f265f32"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782502178717`*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1463536136,ExpressionUUID->"30a8eca9-850d-4b6f-a5fb-a9c0e4cc9924"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897825023512*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->374145014,ExpressionUUID->"32e7435a-a072-475e-9aa3-9c94813b8323"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782502524757`*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->423131882,ExpressionUUID->"2c0655e6-7f8d-4e3d-a602-c966d441c5a7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250269287*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1259185557,ExpressionUUID->"030e58d4-589b-4cd7-b1db-9746b1cdfc58"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289782502859077`*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->1827798961,ExpressionUUID->"6bc3025c-5f84-4579-a495-cd771816efa4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetricg", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetricg[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978250303742*^9},
 CellLabel->"During evaluation of In[649]:=",
 CellID->63567192,ExpressionUUID->"df278643-30b9-4912-9320-042464b57020"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Now we have", "MathCaption",ExpressionUUID->"493f36b8-f608-45dc-b233-d813e77fc8a2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", "b", ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}]}], "]"}]], "Input",
 CellLabel->
  "In[650]:=",ExpressionUUID->"226f16f6-3a2b-4348-8fb6-ab3550da4077"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978250488392*^9},
 CellLabel->"Out[650]=",
 CellID->1074423385,ExpressionUUID->"883f1966-ea23-4664-9c7c-3b00abb5b657"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[651]:=",ExpressionUUID->"7a22a559-4394-43e6-8819-36fa6b2b0753"],

Cell["-RicciCD[-a, -c]", "Output",
 CellChangeTimes->{3.728978250640695*^9},
 CellLabel->"Out[651]//InputForm=",
 CellID->1791327719,ExpressionUUID->"e6c75474-a260-44c0-ab6e-c62c8a795f9c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", "b", ",", "a"}], "]"}]], "Input",
 CellLabel->
  "In[652]:=",ExpressionUUID->"476c5f6d-ee2e-4b0d-b307-6f2d9d4a7aa1"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarCD[],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978250788796*^9},
 CellLabel->"Out[652]=",
 CellID->1271725940,ExpressionUUID->"8ccf4593-9514-4d30-9127-cfccbea1dce7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[653]:=",ExpressionUUID->"6a2ad638-4ada-420d-94c5-4e119fa4189a"],

Cell["-RicciScalarCD[]", "Output",
 CellChangeTimes->{3.728978250911673*^9},
 CellLabel->"Out[653]//InputForm=",
 CellID->2140618636,ExpressionUUID->"a00d8838-dd24-4afe-bded-478e39af129b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"metricg", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[654]:=",ExpressionUUID->"1a1514a7-0a1a-4357-a95d-e38ebb1b8a3e"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978251024322*^9},
 CellLabel->"Out[654]=",
 CellID->583877061,ExpressionUUID->"a3551895-904d-4a0c-9695-e8cbd0940255"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"metricg", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[655]:=",ExpressionUUID->"d0671f74-dec3-45d8-9155-a83dd04ca381"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`c][
   $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978251158292*^9},
 CellLabel->"Out[655]=",
 CellID->1368981084,ExpressionUUID->"74b3fee8-05f2-41cc-b04c-dde855c1c428"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[656]:=",ExpressionUUID->"01719be4-35dd-46bd-bf23-e258d4134d50"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978251290207*^9},
 CellLabel->"Out[656]=",
 CellID->325284830,ExpressionUUID->"e8b1de85-146d-44f1-a409-10579f2436f6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"EinsteinCD", "[", 
   RowBox[{"b", ",", "c"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[657]:=",ExpressionUUID->"b711b162-c513-47f3-ad2f-841f5bbc7598"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {
       RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"b", "c"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`a][
   $CellContext`EinsteinCD[$CellContext`b, $CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782514658003`*^9},
 CellLabel->"Out[657]=",
 CellID->1716256195,ExpressionUUID->"7ef6601f-1e38-43f9-88b9-a3a41fb03a71"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"EinsteinCD", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[658]:=",ExpressionUUID->"182c4435-0a60-4a56-a95f-ce156954a058"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978251607841*^9},
 CellLabel->"Out[658]=",
 CellID->117522861,ExpressionUUID->"518b14f8-8a63-49fa-a7f8-3cee87d163cb"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The determinant of the basis in the default basis ",
 StyleBox["AIndex",
  FontFamily->"Courier"],
 " (the double tilde denotes weight +2 in that basis; this is Astekhar \
notation):"
}], "MathCaption",ExpressionUUID->"875d9f9e-9853-4d79-b057-e6299a4b5b87"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Detmetricg", "[", "]"}]], "Input",
 CellLabel->
  "In[659]:=",ExpressionUUID->"046cd19f-c17d-41b1-a5ff-68f31c9a33f1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   OverscriptBox["g", 
    OverscriptBox[
     StyleBox["~",
      FontColor->RGBColor[0, 0, 0]], 
     StyleBox["~",
      FontColor->RGBColor[0, 0, 0]]]],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Detmetricg[],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289782517306137`*^9},
 CellLabel->"Out[659]=",
 CellID->1669253234,ExpressionUUID->"c55dded4-0361-49dd-b705-8ddf70a1061b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Its Weyl tensor is zero because the manifold is 3d:", "MathCaption",ExpressionUUID->"3c01bc27-f702-423d-9c3d-656cb6db21df"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeylCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}]}], "]"}]], "Input",
 CellLabel->
  "In[660]:=",ExpressionUUID->"5e124beb-30ff-44ec-97c1-093ac0e41830"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978251839717*^9},
 CellLabel->"Out[660]=",
 CellID->412739430,ExpressionUUID->"a2f82dfd-5a72-43a3-8a71-5af2c91689ca"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["There is also the traceless part of the Ricci tensor:", "MathCaption",ExpressionUUID->"7476d2f3-12c6-46bd-984d-0e3cc469c628"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TFRicciCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->
  "In[661]:=",ExpressionUUID->"be194554-b2d4-4e96-8152-9f24bfe9f659"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978251989274*^9},
 CellLabel->"Out[661]=",
 CellID->539420200,ExpressionUUID->"286811ce-c000-466e-9ef9-dd87597608ba"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TFRicciCD", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "a"}]}], "]"}]], "Input",
 CellLabel->
  "In[662]:=",ExpressionUUID->"144239ed-7062-4a22-a984-9aac9035a3fb"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978252348454*^9},
 CellLabel->"Out[662]=",
 CellID->1506827413,ExpressionUUID->"fe659049-4f49-4e21-bf96-611b553a57b5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now ",
 StyleBox["xTensor`", "DisplayMath"],
 " accepts index structures that do not correspond to the original \
definition, as long as they only involve defined metrics:"
}], "MathCaption",ExpressionUUID->"e32cd178-bbd0-4c87-aa1a-f249e1603a48"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "U", "]"}]], "Input",
 CellLabel->
  "In[663]:=",ExpressionUUID->"299fd2ba-cb2e-4a8c-ab35-1a413704a471"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M3",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM3,
     Editable->False]}], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M3",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM3,
     Editable->False]}], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M3",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM3,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978252769844*^9},
 CellLabel->"Out[663]=",
 CellID->2121164867,ExpressionUUID->"536f4bcc-1851-4e9f-9dda-ce6803e3d4b7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[664]:=",ExpressionUUID->"a7924818-9ba2-4d90-8384-aca80272d058"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"U", 
      StyleBox[GridBox[{
         {"a", "b", "c"},
         {" ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978253233532*^9},
 CellLabel->"Out[664]=",
 CellID->41895822,ExpressionUUID->"7f665a72-c883-457a-8097-37197b25ab31"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"X", "[", 
    RowBox[{"a", ",", "b", ",", "A"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"M3", ",", "S2"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[665]:=",ExpressionUUID->"b2022f06-2f21-42cd-9836-dd4863b7f31d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"X", "[", 
    RowBox[{"a", ",", "b", ",", "A"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`X[$CellContext`a, $CellContext`b, $CellContext`A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978253348819*^9},
 CellLabel->"During evaluation of In[665]:=",
 CellID->1915927785,ExpressionUUID->"a710aabc-87a4-415b-aa33-6db98801af57"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "X", "]"}]], "Input",
 CellLabel->
  "In[666]:=",ExpressionUUID->"fae06c02-6681-4554-9a7e-3500e27251fe"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]S2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentS2,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289782536167583`*^9},
 CellLabel->"Out[666]=",
 CellID->1628479377,ExpressionUUID->"bdec889f-163b-4476-a8df-c678443a6519"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", 
  RowBox[{"X", "[", 
   RowBox[{"a", ",", "b", ",", 
    RowBox[{"-", "A"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[667]:=",ExpressionUUID->"ed8a1d02-d2be-47ef-a115-f58d3faa09bf"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "error", 
   "\"\\!\\(\\*RowBox[{\\\"\\\\\\\"Invalid character of index in tensor X\\\\\
\\\"\\\"}]\\)\"", 2, 667, 60, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978253852234*^9},
 CellLabel->"During evaluation of In[667]:=",
 CellID->173808224,ExpressionUUID->"d52fa4e0-b263-4b88-b490-30b27aecb66f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["ERROR",
     FontColor->Hue[0.]], "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"X", 
         StyleBox[GridBox[{
            {"a", "b", " "},
            {" ", " ", "A"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`X[$CellContext`a, $CellContext`b, -$CellContext`A],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`ERROR[
   $CellContext`X[$CellContext`a, $CellContext`b, -$CellContext`A]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978253885462*^9},
 CellLabel->"Out[667]=",
 CellID->617060848,ExpressionUUID->"d3978747-d853-4e60-bf2a-974fc339dd0b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "X", "]"}]], "Input",
 CellLabel->
  "In[668]:=",ExpressionUUID->"6fccf0eb-37c3-463b-a049-2e2eec9e0967"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "X"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`X],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978254128549*^9},
 CellLabel->"During evaluation of In[668]:=",
 CellID->743140973,ExpressionUUID->"613be26c-33e8-409d-a251-ac2c1d47be1e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Expression of the Christoffel symbols (with respect to ",
 StyleBox["PD",
  FontFamily->"Courier"],
 ") in terms of the metric:"
}], "MathCaption",ExpressionUUID->"8e11a471-ba59-4402-9a7b-519e6066dd3f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}], "]"}], "//", "ChangeCovD"}]], "Input",
 CellLabel->
  "In[669]:=",ExpressionUUID->"651ec90d-f1d2-4c04-a41a-4503c39829b3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", " ", " "},
            {" ", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`a, \
-$CellContext`b],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978254280581*^9},
 CellLabel->"Out[669]=",
 CellID->205442259,ExpressionUUID->"fbfdeb40-061d-4e93-a8cc-3f23139711d8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ChristoffelToGradMetric"}]], "Input",
 CellLabel->
  "In[670]:=",ExpressionUUID->"0774afae-ae71-414c-80e3-21262a85d13c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[-$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "a"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`a][
       $CellContext`metricg[-$CellContext`b, -$CellContext`d]],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
      Editable->False], "-", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978254428368*^9},
 CellLabel->"Out[670]=",
 CellID->1861826284,ExpressionUUID->"fd4e4059-faf3-43da-85a5-bde33c93d3dd"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.2. Contraction", "Subsection",ExpressionUUID->"729bc3fa-0125-43cd-af49-29a91841ae85"],

Cell[TextData[{
 "Metrics are not contracted by default because in general that obscures the \
manipulation of indices and the canonicalization process becomes slower. \
However, ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has a number of commands that perform metric contraction. When the \
expression is a pure tensor contraction is a simple process, but when there \
are derivatives which are not associated to the (first) metric of the vbundle \
then problematic situations could develop. There are two situations we need \
to worry about: one is the process of contraction itself, shown in this \
subsection; the other is the influence of the metric in the process of \
canonicalization, studied in the next subsection."
}], "Text",ExpressionUUID->"6d4a01c2-2565-4a3e-93c7-72907fa67f7c"],

Cell[TextData[{
 StyleBox[ButtonBox["ContractMetric",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ContractMetric.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Perform metric contractions without differentiating metrics\n", 
  "Text"],
 StyleBox[ButtonBox["OverDerivatives",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "OverDerivatives.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Option to perform metric contractions with metric differentiation \
if needed\n", "Text"],
 StyleBox[ButtonBox["AllowUpperDerivatives",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "AllowUpperDerivatives.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Option stating if it is possible to contract an inverse metric \
with a derivative operator", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.7289782614002666`*^9, 
  3.728978262400227*^9}},ExpressionUUID->"aefcc30d-9833-4dd5-8ae6-\
4a2ce6907983"],

Cell["Metric contraction.", "Text",ExpressionUUID->"7b516783-a40c-48d0-ad93-a09957515020"],

Cell[CellGroupData[{

Cell["Metric factors are automatically contracted with themselve:", \
"MathCaption",ExpressionUUID->"2fba4edd-de0c-4fc9-9d12-fc60e952c557"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"metricg", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[676]:=",ExpressionUUID->"7bbe88c4-69f7-4bd2-a59d-8b156b8b8d42"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "a"},
         {"c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`c, $CellContext`a],
  Editable->False]], "Output",
 CellLabel->
  "Out[676]=",ExpressionUUID->"cb4f743e-9d21-4cb9-855f-a295925379c7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"metricg", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "a"}]}], "]"}]], "Input",
 CellLabel->
  "In[677]:=",ExpressionUUID->"e733692f-9d3f-4ce0-90f4-4705f976cef1"],

Cell[BoxData["3"], "Output",
 CellLabel->
  "Out[677]=",ExpressionUUID->"1fd002b2-971b-4e68-958b-6458a6adfcf3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However they are only contracted with other objects if required. This is to \
have more control on expressions formed by metric factors, and mostly to \
avoid unexpected index-shiftings.\
\>", "MathCaption",ExpressionUUID->"74a57a0c-4f43-423d-9e60-445aa1727c87"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"metricg", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"S", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "c"}], "]"}]}]], "Input",
 CellLabel->
  "In[678]:=",ExpressionUUID->"caa7cc6b-814d-41b3-bbab-ee7a79e9c6fa"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {" ", "c"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[-$CellContext`a, $CellContext`c],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[678]=",ExpressionUUID->"3ce05824-3e3f-4e5a-87e2-91fe9b0028a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{"%", ",", "metricg"}], "]"}]], "Input",
 CellLabel->
  "In[679]:=",ExpressionUUID->"0b74dc0d-e9af-414e-b92d-db741212fa91"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"S", 
      StyleBox[GridBox[{
         {"b", "c"},
         {" ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`S[$CellContext`b, $CellContext`c],
  Editable->False]], "Output",
 CellLabel->
  "Out[679]=",ExpressionUUID->"6ea8097e-ba54-402f-9866-fe455aeeb760"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Derivatives of a metric are automatically converted so that the indices of \
the metric are covariant, introducing new dummies. Note that derivatives of \
mixed-indices metrics are zero. That means that all non-zero derivatives of \
metrics will have covariant indices:\
\>", "MathCaption",ExpressionUUID->"acdc3df6-8f77-4390-a6ba-a1c9c62367e2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"metricg", "[", 
   RowBox[{"b", ",", "c"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[680]:=",ExpressionUUID->"ea4be80b-a9e1-4c53-9e0f-cebe6138f8e5"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`d],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"c", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`c, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[680]=",ExpressionUUID->"e1e7fd0f-439a-4c75-95d1-2018eb17cb7e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"metricg", "[", 
   RowBox[{"b", ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[681]:=",ExpressionUUID->"6e7cb503-4e77-490f-aeb1-afd9f61fe9f0"],

Cell[BoxData["0"], "Output",
 CellLabel->
  "Out[681]=",ExpressionUUID->"bcc4504c-8f41-4b42-a08b-2a6556b21e6c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Contraction of derivative indices is not allowed by default, but can be \
forced using the option ",
 StyleBox["AllowUpperDerivatives",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"8af64153-3662-46ac-8e01-44226966a578"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"S", "[", 
      RowBox[{"c", ",", "d"}], "]"}], "]"}]}], ",", "metricg"}], 
  "]"}]], "Input",
 CellLabel->
  "In[682]:=",ExpressionUUID->"363a9176-94c8-4052-806a-c49952df56a9"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`S[$CellContext`c, $CellContext`d]],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[682]=",ExpressionUUID->"356aea68-4410-410b-aa44-0e0c7a0a797f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "b"}], "]"}], "[", 
     RowBox[{"S", "[", 
      RowBox[{"c", ",", "d"}], "]"}], "]"}]}], ",", "metricg", ",", 
   RowBox[{"AllowUpperDerivatives", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[683]:=",ExpressionUUID->"f19a9d7c-f615-4399-a92e-1cc57247f2c0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SuperscriptBox["\[PartialD]", "a"], GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"c", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PD[$CellContext`a][
   $CellContext`S[$CellContext`c, $CellContext`d]],
  Editable->False]], "Output",
 CellLabel->
  "Out[683]=",ExpressionUUID->"6c3b90b5-fc87-4fd5-aa69-231e6e0ee154"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "With ",
 StyleBox["ContractMetric",
  FontFamily->"Courier"],
 " a metric can be contracted through its Levi-Civita derivative, but not \
through other derivatives:"
}], "MathCaption",ExpressionUUID->"e768066d-b5a2-42be-a860-6f1494c00caa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{"S", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", "c"}], "]"}], "]"}]}], ",", "metricg"}], 
  "]"}]], "Input",
 CellLabel->
  "In[684]:=",ExpressionUUID->"9345c044-5e5b-49ad-9e08-9bb85a9a76dd"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"b", "c"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`d][
   $CellContext`S[$CellContext`b, $CellContext`c]],
  Editable->False]], "Output",
 CellLabel->
  "Out[684]=",ExpressionUUID->"e0a72705-93cc-42ad-96cf-86abb9342098"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{"S", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", "c"}], "]"}], "]"}]}], ",", "metricg"}], 
  "]"}]], "Input",
 CellLabel->
  "In[685]:=",ExpressionUUID->"dcb80acd-f916-4951-9a92-02cc42d66e12"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`S[-$CellContext`a, $CellContext`c]],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[685]=",ExpressionUUID->"ece2356e-9a14-418d-889b-72843b2752ca"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In those cases we need to use a different function which forces contraction \
over derivatives, but at the price of producing new terms with derivatives of \
the metric:\
\>", "MathCaption",ExpressionUUID->"b57bc7c4-a2b4-43cd-85fd-669ca9f76684"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{"S", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", "c"}], "]"}], "]"}]}], ",", "metricg", ",", 
   RowBox[{"OverDerivatives", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[686]:=",ExpressionUUID->"c1ac7d4b-a1ee-43b1-9fb9-ca89c7a7ecbe"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"a", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`S[$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`e]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"S", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`S[$CellContext`b, $CellContext`c]],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[686]=",ExpressionUUID->"b5588248-ceb0-4d90-8a00-229ff58e9b40"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "7.3A. Canonicalization with a metric",
 StyleBox[". IMPORTANT BUG",
  FontColor->RGBColor[0.5, 0, 0.5]]
}], "Subsection",ExpressionUUID->"d26708be-fc7e-4080-aa5e-f61d015ebfbe"],

Cell[TextData[{
 "In the presence of a metric new symmetries can be added to the process of \
canonicalization. The command ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 " has an option ",
 StyleBox["UseMetricOnVBundle",
  FontFamily->"Courier"],
 " that specifies on which vbundles the metric (if there is one) must be used \
during canonicalization. Possible values are ",
 StyleBox["All",
  FontFamily->"Courier"],
 " (default), ",
 StyleBox["None",
  FontFamily->"Courier"],
 " or a list of (maybe one) vbundles."
}], "Text",ExpressionUUID->"6f1f20a3-4fd7-43f8-8a28-acdaac257764"],

Cell[CellGroupData[{

Cell[TextData[{
 "Tensors ",
 StyleBox["S",
  FontFamily->"Courier"],
 " and ",
 StyleBox["T",
  FontFamily->"Courier"],
 " have no symmetries and therefore their canonicalization without metric is \
trivial:"
}], "MathCaption",ExpressionUUID->"24450e09-7be9-41ed-8d0c-94bfb39b636d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"c", ",", "a"}], "]"}], 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "b"}], "]"}]}], ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}]}], "]"}]], "Input",
 CellLabel->
  "In[671]:=",ExpressionUUID->"e514a8ad-b2c1-41dc-8813-1cbea531bdca"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"c", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`c, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " ", "b"},
          {"a", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b, $CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978274187221*^9},
 CellLabel->"Out[671]=",
 CellID->28637891,ExpressionUUID->"7df159bd-bd76-4dce-b066-af258a3dd335"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However with metric the canonical form is different because pairs of dummies \
are sorted such that the up-index comes first:\
\>", "MathCaption",ExpressionUUID->"1755d33c-f713-4354-8eff-bacca0203c51"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"c", ",", "a"}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", "b"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[672]:=",ExpressionUUID->"77c9b493-fbdf-4794-ad19-0a2016c726e4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"c", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`c, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", "b", " "},
          {"a", " ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, $CellContext`b, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978274323007*^9},
 CellLabel->"Out[672]=",
 CellID->280697109,ExpressionUUID->"7d08f7d8-6507-44c8-8a2c-27cc598bdeff"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "When derivatives are present canonicalization with a metric is harder, \
unless the only derivative is its Levi-Civita associated operator. In other \
cases, ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 " shows a warning message and changes to a different algorithm rather than \
using the default algorithm which just moves indices around and that could \
get easily confused. There are two possible internal algorithms, controlled \
with the option ",
 StyleBox["Method",
  FontFamily->"Courier"],
 " of ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 ". Before version 0.9.8 of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " the only possibility was (and still is the default) ",
 StyleBox["Method->ChangeCovD",
  FontFamily->"Courier"],
 ", which changes the non-compatible derivatives into Levi-Civita \
connections. Now there is a second option, ",
 StyleBox["Method->Implode",
  FontFamily->"Courier"],
 ", which uses the new idea of converting all derivatives into tensors. The \
latter is faster and more general, but hasn't been tested much. Use it with \
care. Both methods considerably slow down the process of canonicalization \
because they require several internal calls to ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"4aa9f8cb-b443-4c21-a144-5346e5286463"],

Cell[CellGroupData[{

Cell[TextData[{
 "We use the same tensors ",
 StyleBox["S",
  FontFamily->"Courier"],
 " and ",
 StyleBox["T",
  FontFamily->"Courier"],
 ", with metric-compatible derivatives:"
}], "MathCaption",ExpressionUUID->"60b93444-1bee-46e2-bb91-6f1def9e7093"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"S", "[", 
      RowBox[{"c", ",", "a"}], "]"}], 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"T", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}], ",", "b"}], "]"}], "]"}]}], ",", 
    RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[673]:=",ExpressionUUID->"a69199cf-b7f7-4249-8936-7870554be859"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.003378`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`a],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", " ", "b"},
              {"a", "b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`T[-$CellContext`a, -$CellContext`b, $CellContext`b]],
      Editable->False], ")"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978274475149*^9},
 CellLabel->"Out[673]=",
 CellID->1802222151,ExpressionUUID->"07b0ee99-d9ee-43b4-9f15-d6665de5e3a8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"c", ",", "a"}], "]"}], 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{"T", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", "b"}], "]"}], "]"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[674]:=",ExpressionUUID->"293caad7-f96f-480a-b0b3-cd72ea5fd170"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.003491`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`a],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" ", "b", " "},
              {"a", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`T[-$CellContext`a, $CellContext`b, -$CellContext`b]],
      Editable->False], ")"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978274642151*^9},
 CellLabel->"Out[674]=",
 CellID->1838334569,ExpressionUUID->"12cd9674-152c-4aa6-90e4-ba00ea8360b1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "But now if we use ",
 StyleBox["PD",
  FontFamily->"Courier"],
 " instead:"
}], "MathCaption",ExpressionUUID->"e387efa2-837e-4948-ae56-7be100ae0000"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"S", "[", 
      RowBox[{"c", ",", "a"}], "]"}], 
     RowBox[{
      RowBox[{"PD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"T", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}], ",", "b"}], "]"}], "]"}]}], ",", 
    RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[675]:=",ExpressionUUID->"3766483b-dca2-4214-bd0c-da67e4016776"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.004159`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "d"], GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " ", "b"},
             {"a", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`T[-$CellContext`a, -$CellContext`b, $CellContext`b]],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978274770886*^9},
 CellLabel->"Out[675]=",
 CellID->1041871794,ExpressionUUID->"9caaaba4-04ba-4887-9a35-80fc0ed9f7d6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"c", ",", "a"}], "]"}], 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{"T", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", "b"}], "]"}], "]"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[676]:=",ExpressionUUID->"9f123f4e-9dc3-4855-9166-753a0d2c9669"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 676, 61, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978274911991*^9},
 CellLabel->"During evaluation of In[676]:=",
 CellID->157584120,ExpressionUUID->"abb712e9-212c-4481-a17b-57e363d2118b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.021684`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "d"], GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", "b", " "},
             {"a", " ", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`T[-$CellContext`a, $CellContext`b, -$CellContext`b]],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978274952014*^9},
 CellLabel->"Out[676]=",
 CellID->668968089,ExpressionUUID->"2ee6d7b4-f1e2-4baa-b82b-5b924bd371e4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "This form of canonicalization leaves lots of Christoffel hanging around. \
You can get rid of them in three steps: first use ",
 StyleBox["ChristoffelToGradMetric",
  FontFamily->"Courier"],
 "; then use ",
 StyleBox["ContractMetric",
  FontFamily->"Courier"],
 ","
}], "MathCaption",ExpressionUUID->"8c60933b-99f1-4732-b5a0-524af687f94b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Last", "[", "%", "]"}], "//", "ChristoffelToGradMetric"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[677]:=",ExpressionUUID->"c0dac4ab-9b07-4381-b212-2ec1afd8594f"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"c", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`c, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`T[-$CellContext`a, $CellContext`b, -$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978275077084*^9},
 CellLabel->"Out[677]=",
 CellID->1269779778,ExpressionUUID->"92b6c5ef-8d18-43bd-b8c1-af148227914f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Finally, canonicalize again, but this time without moving indices, to avoid \
bringing back the Christoffels:\
\>", "MathCaption",ExpressionUUID->"5cbd2977-d36d-4829-9441-90e29e300eb0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{"%", ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}]}], "]"}]], "Input",
 CellLabel->
  "In[678]:=",ExpressionUUID->"0760b1d2-a6d5-4722-98f6-5dc26f5e7d02"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"S", 
       StyleBox[GridBox[{
          {"c", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`S[$CellContext`c, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", "b", " "},
           {"a", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`T[-$CellContext`a, $CellContext`b, -$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289782752041492`*^9},
 CellLabel->"Out[678]=",
 CellID->1643728607,ExpressionUUID->"b70b3931-4ce6-44d3-a36b-b1e8e6f2c683"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The whole process is automatic with the option ",
 StyleBox["ExpandChristoffel",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"da6be665-8ce2-44d3-8a33-9c4061eea484"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"S", "[", 
      RowBox[{"c", ",", "a"}], "]"}], 
     RowBox[{
      RowBox[{"PD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"T", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}], ",", "b"}], "]"}], "]"}]}], ",", 
    RowBox[{"ExpandChristoffel", "\[Rule]", "True"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[679]:=",ExpressionUUID->"d5cd31ad-0cba-42e4-baa7-ed857c8c02db"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 679, 62, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289782753319902`*^9},
 CellLabel->"During evaluation of In[679]:=",
 CellID->917062841,ExpressionUUID->"f443f3e0-26c1-4a0c-a3b6-8b698fb76567"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 679, 63, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728978275368126*^9},
 CellLabel->"During evaluation of In[679]:=",
 CellID->585034168,ExpressionUUID->"d042badb-6e6d-4f66-86ef-ab3415437ef3"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 679, 64, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978275435573*^9},
 CellLabel->"During evaluation of In[679]:=",
 CellID->1489125610,ExpressionUUID->"88a9fdb5-c0e7-4c38-942f-70a81a3d59aa"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 679, 65, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728978275491085*^9},
 CellLabel->"During evaluation of In[679]:=",
 CellID->959705132,ExpressionUUID->"de1596c1-ff15-4e93-acfd-07c932d46e89"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"OptionValue\\\", \\\"::\
\\\", \\\"nodef\\\"}], \\\"MessageName\\\"]\\) will be suppressed during this \
calculation.\"", 2, 679, 66, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.72897827554751*^9},
 CellLabel->"During evaluation of In[679]:=",
 CellID->14480907,ExpressionUUID->"bea8c403-eb60-4b10-83ce-a7d3deb1d781"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.04141`", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[$CellContext`c, $CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "d"], GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", "b", " "},
             {"a", " ", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`T[-$CellContext`a, $CellContext`b, -$CellContext`b]],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978275636218*^9},
 CellLabel->"Out[679]=",
 CellID->714951838,ExpressionUUID->"daf6ffd7-1d4a-421e-b1b9-ecb00430a385"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The alternative algorithm based on implosion is similar (note that it would \
require a further use of ",
 StyleBox["ContractMetric",
  FontFamily->"Courier"],
 " to arrive at the same result):"
}], "MathCaption",ExpressionUUID->"b435c987-dcb0-4749-9b2a-07dc3fbac95b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"S", "[", 
      RowBox[{"c", ",", "a"}], "]"}], 
     RowBox[{
      RowBox[{"PD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"T", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}], ",", "b"}], "]"}], "]"}]}], ",", 
    RowBox[{"Method", "->", "Implode"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[680]:=",ExpressionUUID->"51cb68b8-b11a-43da-92a5-317a822df33b"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 680, 67, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978275773612*^9},
 CellLabel->"During evaluation of In[680]:=",
 CellID->1024536301,ExpressionUUID->"c119ff55-305d-4992-a805-ba0f53adea43"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.040908`", ",", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"c", "a"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[$CellContext`c, $CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"b", "e", " "},
             {" ", " ", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`b, $CellContext`e, -$CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"c", "a"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[$CellContext`c, $CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", "b", "e"},
             {"a", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[-$CellContext`a, $CellContext`b, $CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`metricg[-$CellContext`b, -$CellContext`e]],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[-$CellContext`b, -$CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"c", " "},
             {" ", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[$CellContext`c, -$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "d"], GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b", "e"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`d][
       $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`e]],
      Editable->False]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289782758659563`*^9},
 CellLabel->"Out[680]=",
 CellID->857729261,ExpressionUUID->"1020b617-d71a-4533-acc3-4fa1bad576ea"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is another, somehow harder, example. Partial derivatives do not commute \
when having upper indices:\
\>", "MathCaption",ExpressionUUID->"05818001-6b59-4d49-a488-5eec5102318f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "b"}], "]"}], "@", 
    RowBox[{
     RowBox[{"PD", "[", "a", "]"}], "@", 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}]}]}], "-", 
   RowBox[{
    RowBox[{"PD", "[", "a", "]"}], "@", 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "b"}], "]"}], "@", 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}]}]}]}], "//", "Simplification"}]], "Input",
 CellLabel->
  "In[681]:=",ExpressionUUID->"b575d3c4-d4b8-4713-b4eb-9194a009464f"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 681, 68, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978276015203*^9},
 CellLabel->"During evaluation of In[681]:=",
 CellID->1945450963,ExpressionUUID->"feddb210-91d9-4378-865a-2095c7432b58"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 681, 69, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978276115917*^9},
 CellLabel->"During evaluation of In[681]:=",
 CellID->1755809566,ExpressionUUID->"88dd3740-dba2-45b4-8db1-690e9e689c81"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 681, 70, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289782761426153`*^9},
 CellLabel->"During evaluation of In[681]:=",
 CellID->1318582335,ExpressionUUID->"52d2546b-db12-468d-889f-13cbd05f2b21"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"ToCanonical\\\", \\\"::\
\\\", \\\"cmods\\\"}], \\\"MessageName\\\"]\\) will be suppressed during this \
calculation.\"", 2, 681, 71, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.72897827616951*^9},
 CellLabel->"During evaluation of In[681]:=",
 CellID->1529550952,ExpressionUUID->"26eca71b-9fad-4fdf-8e61-bc50f7ef11f5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", "d", " "},
            {"c", " ", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[-$CellContext`c, $CellContext`d, \
-$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "c"},
           {"a", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`a, -$CellContext`b, \
$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", "d", " "},
             {"c", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[-$CellContext`c, $CellContext`d, \
-$CellContext`d],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`d],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], 
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`v[$CellContext`a]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", "c", " "},
            {"a", " ", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[-$CellContext`a, $CellContext`c, \
-$CellContext`c]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", " ", " "},
            {" ", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`a, \
-$CellContext`c]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], 
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`v[$CellContext`a]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", "c"},
            {"a", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[-$CellContext`a, -$CellContext`b, \
$CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", " ", " "},
            {" ", "b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`b, \
-$CellContext`a]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "c"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`c][
     $CellContext`v[$CellContext`a]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "c"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`c][
     $CellContext`v[$CellContext`a]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978276407987*^9},
 CellLabel->"Out[681]=",
 CellID->1783595674,ExpressionUUID->"0c4257dd-290e-474a-a6e9-8a5ea86d1871"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we can drop the two second derivatives, and remove the Christoffels:\
\>", "MathCaption",ExpressionUUID->"4f3729e4-cb08-44da-b848-38cc2cb5fb37"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "SortCovDs"}]], "Input",
 CellLabel->
  "In[682]:=",ExpressionUUID->"ffe634cc-4f71-4509-854c-c4a141c026f6"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", "d", " "},
            {"c", " ", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[-$CellContext`c, $CellContext`d, \
-$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"d", " ", " "},
           {" ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "c"},
           {"a", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`a, -$CellContext`b, \
$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", "d", " "},
             {"c", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[-$CellContext`c, $CellContext`d, \
-$CellContext`d],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"d", " ", " "},
             {" ", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`ChristoffelCD[$CellContext`d, -$CellContext`c, \
-$CellContext`d],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", "c", " "},
            {"a", " ", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[-$CellContext`a, $CellContext`c, \
-$CellContext`c]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", " ", " "},
            {" ", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`a, \
-$CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", "c"},
            {"a", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[-$CellContext`a, -$CellContext`b, \
$CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", " ", " "},
            {" ", "b", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`ChristoffelCD[$CellContext`c, -$CellContext`b, \
-$CellContext`a]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "c"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`c][
     $CellContext`v[$CellContext`a]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`c, -$CellContext`b, \
-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "c"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`c][
     $CellContext`v[$CellContext`a]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978276719771*^9},
 CellLabel->"Out[682]=",
 CellID->168264445,ExpressionUUID->"f71853eb-5230-4c36-84fa-1ef98e57b1f8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"result1", "=", 
  RowBox[{
   RowBox[{"Simplification", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"%", "//", "ChristoffelToGradMetric"}], "//", 
      "ContractMetric"}], ",", 
     RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}]}], "]"}], "//", 
   "AbsoluteTiming"}]}]], "Input",
 CellLabel->
  "In[683]:=",ExpressionUUID->"82793961-b9a1-4cbd-85e1-0fbf81ab7402"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.449143`", ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[PartialD]", "b"], GridBox[{
           {"g", 
            StyleBox[GridBox[{
               {" ", " "},
               {"a", "c"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`PD[-$CellContext`b][
        $CellContext`metricg[-$CellContext`a, -$CellContext`c]],
       Editable->False]}], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[PartialD]", "c"], GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[$CellContext`c][
       $CellContext`v[$CellContext`a]],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"c", "d"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`c, $CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"e", "f"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`e, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
      Editable->False]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289782773060083`*^9},
 CellLabel->"Out[683]=",
 CellID->1932047663,ExpressionUUID->"b9e2e9e6-4134-41fa-85eb-8d767f4c1613"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In this case the newer algorithm is faster, and again differs from the \
previous result just in the use of ",
 StyleBox["ContractMetric",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"16779222-2915-4df0-bade-dba12c398814"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"PD", "[", 
       RowBox[{"-", "b"}], "]"}], "@", 
      RowBox[{
       RowBox[{"PD", "[", "a", "]"}], "@", 
       RowBox[{"v", "[", 
        RowBox[{"-", "a"}], "]"}]}]}], "-", 
     RowBox[{
      RowBox[{"PD", "[", "a", "]"}], "@", 
      RowBox[{
       RowBox[{"PD", "[", 
        RowBox[{"-", "b"}], "]"}], "@", 
       RowBox[{"v", "[", 
        RowBox[{"-", "a"}], "]"}]}]}]}], ",", 
    RowBox[{"Method", "->", "Implode"}]}], "]"}], "//", 
  "AbsoluteTiming"}]], "Input",
 CellLabel->
  "In[684]:=",ExpressionUUID->"d4124387-34cd-409c-92d0-9ed686affe3d"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 684, 72, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289782774477777`*^9},
 CellLabel->"During evaluation of In[684]:=",
 CellID->1972329401,ExpressionUUID->"b1d399bd-9ffa-4d65-865d-ff5f068491ae"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 684, 73, 19253573948027831207, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978277536667*^9},
 CellLabel->"During evaluation of In[684]:=",
 CellID->382667515,ExpressionUUID->"32bcbaf2-7703-46a9-8891-06a1791498c7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0.096447`", ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`c, $CellContext`d],
       Editable->False]}], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "c"], GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`c][
       $CellContext`v[$CellContext`a]],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"c", "d"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`c, $CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"e", "f"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`e, $CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "b"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`b][
       $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[PartialD]", "f"], GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`PD[-$CellContext`f][
       $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
      Editable->False]}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978277617975*^9},
 CellLabel->"Out[684]=",
 CellID->1926885487,ExpressionUUID->"61349af1-6fbf-4bdd-9295-1e17eb5f6197"]
}, Open  ]]
}, Open  ]],

Cell["\<\
General recommendations:
\t1. The simplest situation is having just one tangent bundle, with one \
metric and its Levi-Civita connection. In that case the metric can be always \
contracted, and we recommend to do it, just to decrease the number of \
indices.
\t2. If there are several tangent bundles, each one having one metric with \
its Levi-Civita connection, then the metrics can be freely contracted, and \
the only additional problem is that you decide whether to use \
CheckZeroDerivative at some intermediate steps of the calculation, or else \
automate that point.
\t3. If there is one vbundle, one metric with its Levi-Civita connection and \
an additional derivative (ordinary or not, that is irrelevant) then the \
canonicalizer must change to the slower algorithm. I recommend not to change \
the character of the slots of the tensors. To do that set \
UseMetricOnVBundle->False in ToCanonical and never use ContractMetric. The \
metric factors will be always explicit, and the number of indices in the \
expressions will be larger, but overall the canonicalizations will be faster.
\t4. If there are several vbundles with several derivatives, combine comments \
2 and 3.
\t5. If furthermore there are several metrics on the vbundles, each with its \
own Levi-Civita connection, then before start crying, see next subsection.\
\>", "Text",ExpressionUUID->"92fe3149-4a3c-46f6-8139-ff7d651d9f12"],

Cell["\<\
The canonicalizer also takes into account whether indices can be moved up and \
down or not. This is done internally using the concept of metric-state for \
indices. Study next example carefully:\
\>", "Text",ExpressionUUID->"8d80d6d6-f980-4c4b-a874-2c8d98634b69"],

Cell[CellGroupData[{

Cell["\<\
This expression is clearly antisymmetric under the exchange of derivatives:\
\>", "MathCaption",ExpressionUUID->"36dbe083-df68-4b13-bd56-c753de526981"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", "d", "]"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"v", "[", "e", "]"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[685]:=",ExpressionUUID->"24773395-7626-4ae0-ab8c-3d6b315f4220"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {" ", " "},
          {"d", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[-$CellContext`d, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[$CellContext`d]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`v[$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289782777731323`*^9},
 CellLabel->"Out[685]=",
 CellID->1110983249,ExpressionUUID->"ff4b644b-9d83-4527-8f06-c4d261928fe8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and it is thus zero:", "MathCaption",ExpressionUUID->"eb250d2f-42f7-4e3e-baf9-569d2374378e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[686]:=",ExpressionUUID->"f67f61ae-5735-4bf4-bae4-3c656c2a727c"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.72897827793489*^9},
 CellLabel->"Out[686]=",
 CellID->2072638925,ExpressionUUID->"197e5707-72dc-4b7c-930e-c825f3af7602"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We contract the metric through a derivative. The expression however is now \
different and not zero:\
\>", "MathCaption",ExpressionUUID->"ae80212d-ad2c-4bb1-9e55-312eedf7ecbe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"-", "d"}], "]"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"v", "[", "e", "]"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{"d", ",", 
    RowBox[{"-", "e"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[687]:=",ExpressionUUID->"a39167ba-3a9e-4c9c-a542-13fe9785ac60"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[-$CellContext`e]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`v[$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978278085394*^9},
 CellLabel->"Out[687]=",
 CellID->510174488,ExpressionUUID->"ce70711b-104a-4fdd-9243-a398f18564d6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
but the system detects it and change to the alternative algorithm (THIS IS \
BROKEN NOW):\
\>", "MathCaption",ExpressionUUID->"57e04643-39b2-4e4f-ae99-191a8ca2cbb4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{"%", ",", 
   RowBox[{"Method", "->", "ChangeCovD"}], ",", 
   RowBox[{"ExpandChristoffel", "->", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[688]:=",ExpressionUUID->"040ef5b7-bb33-4a2f-9102-6e73af827ff5"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 688, 74, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289782782166557`*^9},
 CellLabel->"During evaluation of In[688]:=",
 CellID->1747900992,ExpressionUUID->"95300994-a28f-43a8-8d05-867220ea277b"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 688, 75, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289782782556667`*^9},
 CellLabel->"During evaluation of In[688]:=",
 CellID->1543165231,ExpressionUUID->"44eb245e-968d-4955-b477-6377fa921450"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289782782687883`*^9},
 CellLabel->"Out[688]=",
 CellID->818032598,ExpressionUUID->"af72e56b-03d5-43e1-bb52-0f3ebe3e6326"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", 
   RowBox[{"%%", ",", 
    RowBox[{"Method", "->", "Implode"}]}], "]"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[689]:=",ExpressionUUID->"99e63dde-c725-4c5a-b9c7-a850219af135"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289782784026423`*^9},
 CellLabel->"Out[689]=",
 CellID->483566492,ExpressionUUID->"1b67ca0d-984b-4515-bfde-63da7e25c168"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Another example:", "MathCaption",ExpressionUUID->"d909b7df-3a2c-4e5d-baa8-fb49c9029f28"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}], ",", "c"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "d"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "e"}], "]"}], "[", 
   RowBox[{"v", "[", "b", "]"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{"d", ",", "e"}], "]"}]}]], "Input",
 CellLabel->
  "In[690]:=",ExpressionUUID->"f780ec05-6d0e-4dd8-80ce-2ebaafe4a433"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"d", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`d, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", " ", "c"},
          {" ", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, -$CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`v[-$CellContext`a]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "e"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`e][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978278546865*^9},
 CellLabel->"Out[690]=",
 CellID->1223910796,ExpressionUUID->"36252d6c-cb6a-4e3b-8cd7-8401d19732d7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[707]:=",ExpressionUUID->"a29959b0-08ab-4cae-ad00-2a088b5c5a60"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 707, 79, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[707]:=",
 CellID->990163873,ExpressionUUID->"ec8cc230-bb66-4c02-afe0-990bbd8f3a62"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 707, 80, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[707]:=",
 CellID->558299888,ExpressionUUID->"0208eb44-7e25-43c3-aa46-c9cead9abf45"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 707, 81, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[707]:=",
 CellID->1276487174,ExpressionUUID->"179a05d0-1f13-4b96-a8c6-33567d57ea87"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"ToCanonical\\\", \\\"::\
\\\", \\\"cmods\\\"}], \\\"MessageName\\\"]\\) will be suppressed during this \
calculation.\"", 2, 707, 82, 19250698866719454101, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellLabel->"During evaluation of In[707]:=",
 CellID->2102327895,ExpressionUUID->"97234cd6-d27c-4001-b841-ce391ae67e7f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", "f"},
            {"b", "d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[-$CellContext`b, -$CellContext`d, \
$CellContext`f],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`e][
     $CellContext`v[-$CellContext`a]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "f"},
           {"d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`b, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`e][
     $CellContext`v[-$CellContext`a]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "d"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`d][
     $CellContext`v[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`e][
     $CellContext`v[-$CellContext`b]],
    Editable->False]}]}]], "Output",
 CellLabel->"Out[707]=",
 CellID->1431695669,ExpressionUUID->"516d1b8d-916c-4d3f-9e4a-466ddf54baef"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{"%%", ",", 
   RowBox[{"ExpandChristoffel", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[708]:=",ExpressionUUID->"58f7a425-4258-49b9-beca-18b21bd039d3"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 708, 83, 
   19250698866719454101, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->1398306970,ExpressionUUID->"2ece1536-384e-4e6c-bcce-8dc3bc23c46b"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 708, 84, 
   19250698866719454101, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->1311674775,ExpressionUUID->"044c4d2b-9ff7-41fd-8e00-514a4010afdf"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 708, 85, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->1097777336,ExpressionUUID->"08144853-a199-418c-90b1-e71de5c1e420"],

Cell[BoxData[
 TemplateBox[{
  "OptionValue", "nodef", 
   "\"Unknown option \\!\\(\\*RowBox[{\\\"\\\\\\\"ExpandChristoffel\\\\\\\"\\\
\"}]\\) for \\!\\(\\*RowBox[{\\\"ToCanonical\\\"}]\\).\"", 2, 708, 86, 
   19250698866719454101, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->1910785282,ExpressionUUID->"107551a2-7dd5-487b-9ae0-f4a2d08f93a1"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"OptionValue\\\", \\\"::\
\\\", \\\"nodef\\\"}], \\\"MessageName\\\"]\\) will be suppressed during this \
calculation.\"", 2, 708, 87, 19250698866719454101, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->1684291827,ExpressionUUID->"db78caba-9c06-492d-a581-51f619098d0f"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 708, 88, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->120718085,ExpressionUUID->"1b7c0750-1132-4750-98a8-234b59a6d495"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 708, 89, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->681804811,ExpressionUUID->"7760eaaa-f487-407b-9410-f660f19cf72d"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"ToCanonical\\\", \\\"::\
\\\", \\\"cmods\\\"}], \\\"MessageName\\\"]\\) will be suppressed during this \
calculation.\"", 2, 708, 90, 19250698866719454101, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellLabel->"During evaluation of In[708]:=",
 CellID->274369560,ExpressionUUID->"5519e53e-81cc-49ae-948c-96b90d7afc56"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", "f"},
            {"b", "d", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ChristoffelCD[-$CellContext`b, -$CellContext`d, \
$CellContext`f],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`e][
     $CellContext`v[-$CellContext`a]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "f"},
           {"d", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[-$CellContext`d, -$CellContext`b, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`e][
     $CellContext`v[-$CellContext`a]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "d"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`d][
     $CellContext`v[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[$CellContext`e][
     $CellContext`v[-$CellContext`b]],
    Editable->False]}]}]], "Output",
 CellLabel->"Out[708]=",
 CellID->674914202,ExpressionUUID->"6139e7e2-ce80-4dcf-aa77-9adddc0f5462"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", 
  RowBox[{"%%%", ",", 
   RowBox[{"Method", "->", "Implode"}]}], "]"}]], "Input",
 CellLabel->
  "In[709]:=",ExpressionUUID->"4c85cac3-8262-44eb-9bfb-50073742e631"],

Cell[BoxData[
 TemplateBox[{
  "ToCanonical", "cmods", 
   "\"Detected metric-incompatible derivatives \\!\\(\\*RowBox[{\\\"{\\\", \\\
\"PD\\\", \\\"}\\\"}]\\).\"", 2, 709, 91, 19250698866719454101, "Local", 
   "xAct`xTensor`ToCanonical"},
  "MessageTemplate2"]], "Message", "MSG",
 CellLabel->"During evaluation of In[709]:=",
 CellID->458362513,ExpressionUUID->"65806eca-5490-45b4-8826-39a7fd6c9c24"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " ", "c"},
           {"a", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a, -$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`v[$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`v[$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " ", "c"},
           {" ", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, -$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`v[$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False]}]}]], "Output",
 CellLabel->"Out[709]=",
 CellID->1934640820,ExpressionUUID->"3cdb740e-ad61-4d2f-81a4-4fbadd8d6e43"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.3B. Canonicalization with a metric using TensorDerivative", \
"Subsection",ExpressionUUID->"b15d6210-1bbd-4601-ab7e-a617bbcf0637"],

Cell[TextData[{
 "The head ",
 StyleBox["TensorDerivative",
  FontFamily->"Courier"],
 " allows an easier handling of this problem."
}], "Text",
 CellChangeTimes->{{3.494610528465789*^9, 
  3.494610557581745*^9}},ExpressionUUID->"8f6fdb9b-11a9-42fa-98a2-\
32bb103d2a97"],

Cell[CellGroupData[{

Cell["\<\
Take the same example we had before. We arrive at a result much faster:\
\>", "MathCaption",
 CellChangeTimes->{{3.4946105855553207`*^9, 3.4946105938016157`*^9}, {
  3.494611328134512*^9, 
  3.4946113370125847`*^9}},ExpressionUUID->"f0bd2eb9-8903-44ec-af44-\
c5dae6f84d36"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"PD", "[", 
     RowBox[{"-", "b"}], "]"}], "@", 
    RowBox[{
     RowBox[{"PD", "[", "a", "]"}], "@", 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}]}]}], "-", 
   RowBox[{
    RowBox[{"PD", "[", "a", "]"}], "@", 
    RowBox[{
     RowBox[{"PD", "[", 
      RowBox[{"-", "b"}], "]"}], "@", 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}]}]}]}], "//", "ToTensorDerivative"}]], "Input",\

 CellChangeTimes->{{3.494610598661261*^9, 3.494610629125098*^9}, 
   3.494611301776803*^9},
 CellLabel->
  "In[694]:=",ExpressionUUID->"b9d9a77d-d317-4ea3-87b4-8598690cb347"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[PartialD]", "g"}], 
         StyleBox[GridBox[{
            {"c", "a", " "},
            {" ", " ", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`TensorDerivative[$CellContext`metricg, 
      xAct`xTensor`PD][$CellContext`c, $CellContext`a, -$CellContext`b],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "v"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`v, 
     xAct`xTensor`PD][-$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"\[PartialD]", "g"}], 
            StyleBox[GridBox[{
               {" ", " ", " "},
               {"c", "a", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`TensorDerivative[$CellContext`metricg, 
         xAct`xTensor`PD][-$CellContext`c, -$CellContext`a, -$CellContext`d],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"\[PartialD]", "g"}], 
           StyleBox[GridBox[{
              {"d", "a", " "},
              {" ", " ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`TensorDerivative[$CellContext`metricg, 
        xAct`xTensor`PD][$CellContext`d, $CellContext`a, -$CellContext`b],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[PartialD]", 
           RowBox[{"\[PartialD]", "g"}]}], 
          StyleBox[GridBox[{
             {" ", " ", "a", " "},
             {"c", "a", " ", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`TensorDerivative[$CellContext`metricg, xAct`xTensor`PD, 
       xAct`xTensor`PD][-$CellContext`c, -$CellContext`a, $CellContext`a, \
-$CellContext`b],
      Editable->False]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", 
         RowBox[{"\[PartialD]", "g"}]}], 
        StyleBox[GridBox[{
           {" ", " ", " ", "a"},
           {"c", "a", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`metricg, xAct`xTensor`PD, 
     xAct`xTensor`PD][-$CellContext`c, -$CellContext`a, -$CellContext`b, \
$CellContext`a],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", 
        RowBox[{"\[PartialD]", "v"}]}], 
       StyleBox[GridBox[{
          {" ", "a", " "},
          {"a", " ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, xAct`xTensor`PD, 
    xAct`xTensor`PD][-$CellContext`a, $CellContext`a, -$CellContext`b],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", 
        RowBox[{"\[PartialD]", "v"}]}], 
       StyleBox[GridBox[{
          {" ", " ", "a"},
          {"a", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, xAct`xTensor`PD, 
    xAct`xTensor`PD][-$CellContext`a, -$CellContext`b, $CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978318443945*^9},
 CellLabel->"Out[694]=",
 CellID->609986205,ExpressionUUID->"c5ff7d7a-2be9-45a9-9741-91ab95708c7b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellChangeTimes->{{3.4946113146218863`*^9, 3.49461131657568*^9}},
 CellLabel->
  "In[695]:=",ExpressionUUID->"72cae8d6-8a1f-4412-b642-5162b62f5109"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "g"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`metricg, 
     xAct`xTensor`PD][-$CellContext`a, -$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "g"}], 
        StyleBox[GridBox[{
           {"c", "d", " "},
           {" ", " ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`metricg, 
     xAct`xTensor`PD][$CellContext`c, $CellContext`d, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "g"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`metricg, 
     xAct`xTensor`PD][-$CellContext`a, -$CellContext`c, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "v"}], 
        StyleBox[GridBox[{
           {"a", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`v, 
     xAct`xTensor`PD][$CellContext`a, $CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978318635496*^9},
 CellLabel->"Out[695]=",
 CellID->2098700304,ExpressionUUID->"d3a40bc5-c0e0-43e0-8bee-9bb7cb129b34"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Other examples shown before:", "MathCaption",
 CellChangeTimes->{{3.494611410178259*^9, 
  3.494611417781898*^9}},ExpressionUUID->"3d6ecafc-2760-4ad2-b8e1-\
4f0754831b6d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", "d", "]"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"v", "[", "e", "]"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[696]:=",ExpressionUUID->"1661c5d8-b395-4e34-bf03-03a8d3f1d97f"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {" ", " "},
          {"d", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[-$CellContext`d, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[$CellContext`d]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`v[$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783187903843`*^9},
 CellLabel->"Out[696]=",
 CellID->437107353,ExpressionUUID->"c1716e50-eec1-4ad4-bd40-7253d10d76cd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToTensorDerivative"}]], "Input",
 CellChangeTimes->{{3.494611427826297*^9, 3.494611435620492*^9}},
 CellLabel->
  "In[697]:=",ExpressionUUID->"d6f66c33-0bc2-42d9-bc10-fd05db66e3de"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {" ", " "},
          {"d", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[-$CellContext`d, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", "v"}], 
       StyleBox[GridBox[{
          {"d", " "},
          {" ", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, 
    xAct`xTensor`PD][$CellContext`d, -$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", "v"}], 
       StyleBox[GridBox[{
          {"e", " "},
          {" ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, 
    xAct`xTensor`PD][$CellContext`e, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897831892756*^9},
 CellLabel->"Out[697]=",
 CellID->2069209131,ExpressionUUID->"a2afdda6-7f4c-4443-bd7c-9638d24fb348"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[698]:=",ExpressionUUID->"470fef24-3b0c-459e-b071-2c9dcc9db444"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.72897831926164*^9},
 CellLabel->"Out[698]=",
 CellID->847685199,ExpressionUUID->"87ce841a-7743-4416-942c-f2213ac777d3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Another example:", "MathCaption",ExpressionUUID->"938dbf0c-3d3b-481c-b990-511bd3b9556a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"-", "d"}], "]"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"v", "[", "e", "]"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{"d", ",", 
    RowBox[{"-", "e"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[699]:=",ExpressionUUID->"a14e6196-6f72-41d7-b101-7022754ae20c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`a][
    $CellContext`v[-$CellContext`e]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "b"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`b][
    $CellContext`v[$CellContext`e]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978319595317*^9},
 CellLabel->"Out[699]=",
 CellID->638625395,ExpressionUUID->"d01bec0e-9f0e-4681-8e56-62d3eb2e0b1e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToTensorDerivative"}]], "Input",
 CellChangeTimes->{{3.494611427826297*^9, 3.494611435620492*^9}},
 CellLabel->
  "In[700]:=",ExpressionUUID->"4bb4a716-efba-4b1c-8cab-16e33351c4cd"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[PartialD]", "g"}], 
          StyleBox[GridBox[{
             {" ", " ", " "},
             {"d", "e", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`TensorDerivative[$CellContext`metricg, 
       xAct`xTensor`PD][-$CellContext`d, -$CellContext`e, -$CellContext`a],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[PartialD]", "v"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`TensorDerivative[$CellContext`v, 
      xAct`xTensor`PD][-$CellContext`e, -$CellContext`a],
     Editable->False]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", "v"}], 
       StyleBox[GridBox[{
          {"e", " "},
          {" ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, 
    xAct`xTensor`PD][$CellContext`e, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978319824088*^9},
 CellLabel->"Out[700]=",
 CellID->253746059,ExpressionUUID->"1e2611fe-93c8-431c-a527-c02aead2ca6a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[701]:=",ExpressionUUID->"9851ba43-f488-46bd-9e9a-a181e8ad08d6"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"c", " ", " "},
           {" ", "e", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`c, -$CellContext`e, -$CellContext`d],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", "g"}], 
       StyleBox[GridBox[{
          {" ", " ", "d"},
          {"a", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`metricg, 
    xAct`xTensor`PD][-$CellContext`a, -$CellContext`b, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", "v"}], 
       StyleBox[GridBox[{
          {"b", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, 
    xAct`xTensor`PD][$CellContext`b, $CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978319958747*^9},
 CellLabel->"Out[701]=",
 CellID->1954974318,ExpressionUUID->"685a13fc-f937-4b58-9f0b-063dc59f496e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Another example:", "MathCaption",ExpressionUUID->"fb479851-7a59-4d36-96b1-bbb96866d2e6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}], ",", "c"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "d"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], "]"}], 
  RowBox[{
   RowBox[{"PD", "[", 
    RowBox[{"-", "e"}], "]"}], "[", 
   RowBox[{"v", "[", "b", "]"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{"d", ",", "e"}], "]"}]}]], "Input",
 CellLabel->
  "In[702]:=",ExpressionUUID->"34fb6a35-0a43-4537-97f8-28a2e6531b59"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"d", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`d, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", " ", "c"},
          {" ", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, -$CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`d][
    $CellContext`v[-$CellContext`a]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[PartialD]", "e"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PD[-$CellContext`e][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978320075346*^9},
 CellLabel->"Out[702]=",
 CellID->753563526,ExpressionUUID->"01be0494-3630-49d8-8208-a21709933345"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToTensorDerivative"}]], "Input",
 CellChangeTimes->{{3.4946114857743177`*^9, 3.494611489393757*^9}},
 CellLabel->
  "In[703]:=",ExpressionUUID->"95b8a649-7d95-4e74-952d-b73cab63f546"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"d", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`d, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", " ", "c"},
          {" ", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, -$CellContext`b, $CellContext`c],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"f"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`f],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"\[PartialD]", "g"}], 
          StyleBox[GridBox[{
             {" ", " ", " "},
             {"f", "a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`TensorDerivative[$CellContext`metricg, 
       xAct`xTensor`PD][-$CellContext`f, -$CellContext`a, -$CellContext`d],
      Editable->False]}], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"\[PartialD]", "v"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`TensorDerivative[$CellContext`v, 
      xAct`xTensor`PD][-$CellContext`a, -$CellContext`d],
     Editable->False]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[PartialD]", "v"}], 
       StyleBox[GridBox[{
          {"b", " "},
          {" ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`TensorDerivative[$CellContext`v, 
    xAct`xTensor`PD][$CellContext`b, -$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783202111063`*^9},
 CellLabel->"Out[703]=",
 CellID->1431115123,ExpressionUUID->"6527c72b-5ed5-48e9-a46a-422af268654b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellChangeTimes->{{3.4946114911567593`*^9, 3.494611493554122*^9}},
 CellLabel->
  "In[704]:=",ExpressionUUID->"1e3f16ea-29dc-4527-bf9f-307a11c44b93"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "g"}], 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`metricg, 
     xAct`xTensor`PD][-$CellContext`a, -$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "v"}], 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`v, 
     xAct`xTensor`PD][-$CellContext`b, $CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "v"}], 
        StyleBox[GridBox[{
           {" ", "d"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`v, 
     xAct`xTensor`PD][-$CellContext`a, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[PartialD]", "v"}], 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`TensorDerivative[$CellContext`v, 
     xAct`xTensor`PD][-$CellContext`b, $CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978320342108*^9},
 CellLabel->"Out[704]=",
 CellID->935025861,ExpressionUUID->"0dc7e481-4627-4fb9-9842-b79b489baf9c"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.4. Several metrics", "Subsection",ExpressionUUID->"19a788d8-1dac-4761-82cc-fdfd4a205663"],

Cell[TextData[{
 "It is possible to work with several metrics on the same vbundle. Of course, \
not all of them can raise and lower the indices of other tensors because we \
would inmediately loose track of the metric that was used to move a given \
index. The situation in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " is as follows. The list of metrics on a vbundle is stored as an upvalue of \
the vbundle for the function ",
 StyleBox["MetricsOfVBundle",
  FontFamily->"Courier"],
 ". The metrics are placed in that list in order of definition. The first \
metric is considered special and it is the only one which can move indices. \
All other metrics (called \"frozen\" metrics) are simply symmetric \
two-tensors with a Levi-Civita associated connection and a number of \
associated tensors. For a frozen metric, many of the expected rules do not \
work.\nIt is possible to remove the first metric, suddenly converting the \
second metric into the first of the list. I cannot think of any situation \
where that could be valid or just safe.\nI have never seriously tested this \
part of the code. Please use it with maximum care!"
}], "Text",ExpressionUUID->"d83de66a-b28b-4fa9-8ddf-5ef0c2ffa7de"],

Cell[CellGroupData[{

Cell[TextData[{
 "Currently there is only one metric on ",
 StyleBox["TangentM3",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"565f8f3a-4967-4358-b087-ccd69e823395"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricsOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[705]:=",ExpressionUUID->"f60be583-9499-4ebb-beca-00de190cb958"],

Cell[BoxData[
 RowBox[{"{", "metricg", "}"}]], "Output",
 CellChangeTimes->{3.728978325990116*^9},
 CellLabel->"Out[705]=",
 CellID->2065006001,ExpressionUUID->"91589141-ae61-4a31-b8bf-d59a0deb040a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We define a second (frozen) metric:", "MathCaption",ExpressionUUID->"af6edc8d-9fd0-4fbc-bcb7-817a4dfec718"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{"1", ",", 
   RowBox[{"frozen", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD2", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<|\>\"", ",", "\"\<D\>\""}], "}"}]}], ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<f\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[706]:=",ExpressionUUID->"bdeba044-43c9-4700-b525-a919e3f0a480"],

Cell[BoxData[
 TemplateBox[{
  "DefMetric", "old", 
   "\"There are already metrics \\!\\(\\*RowBox[{\\\"{\\\", \\\"metricg\\\", \
\\\"}\\\"}]\\) in vbundle \\!\\(\\*InterpretationBox[StyleBox[\\\"\
\[DoubleStruckCapitalT]M3\\\", Rule[AutoSpacing, False], Rule[ShowAutoStyles, \
False]], TangentM3, Rule[Editable, False]]\\).\"", 2, 706, 89, 
   19253573948027831207, "Local", "xAct`xTensor`DefMetric"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978326136835*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1072639640,ExpressionUUID->"6dc93ce5-e681-4645-95e3-3ce506dfb353"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"frozen", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`frozen[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783261630993`*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1876696562,ExpressionUUID->"3e969c8b-1e60-4b07-a8bd-2c47fe20e29f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inverse metric tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Invfrozen", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Metric is frozen!\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "inverse metric tensor ", 
   $CellContext`Invfrozen[$CellContext`a, $CellContext`b], ". ", 
   "Metric is frozen!"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326175161*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1546794045,ExpressionUUID->"54f8cebf-d8d7-411d-ae64-dd8fbeda0933"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonfrozen", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonfrozen[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326192724*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1677967380,ExpressionUUID->"54fda6ee-6a1a-4452-bc93-4eb226366100"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD2", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD2[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326210272*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->842503256,ExpressionUUID->"04cca175-82a5-4d28-bff4-b52f4de6086c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD2", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD2[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783262278137`*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1592538921,ExpressionUUID->"3be5411f-f830-4bec-b366-406e5244b634"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD2", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD2[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897832624508*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1712952686,ExpressionUUID->"e31c8cab-bc57-44d2-aee9-bb560fe9a305"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannDownCD2", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannDownCD2[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783262623987`*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->430021920,ExpressionUUID->"1617cb6a-3edc-4400-a1e1-1c46ac9feac7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD2", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", "d"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"Antisymmetric only in the first pair.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD2[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d], ". ", "Antisymmetric only in the first pair."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326280714*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1095937285,ExpressionUUID->"f73f28a8-e2aa-48d5-9f66-349c81af1aff"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD2", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD2[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326298191*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->450708676,ExpressionUUID->"4222a580-dccf-49c4-b53c-fac2c9e16154"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728978326314692*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->885894563,ExpressionUUID->"f6322896-412d-48e0-86d9-7d962dfd0cc5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD2", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD2[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326332567*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->725795690,ExpressionUUID->"cbfd9153-feb0-4aeb-b7ac-d614e007967c"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728978326349752*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->741046870,ExpressionUUID->"8f26d478-b777-487c-9942-6e89b3f8bf74"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD2", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD2[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326367429*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1517075137,ExpressionUUID->"c0334b33-6e92-40e3-b2af-6d11ea388a47"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD2", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCD2[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326384348*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->408486728,ExpressionUUID->"0bf8e675-7a1a-461c-b693-b4d2070c8ba0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD2", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD2[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326405879*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->2078701684,ExpressionUUID->"89f20f82-d433-4e34-a961-f5993a9c9472"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD2", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD2[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326427348*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1475233592,ExpressionUUID->"75223229-1581-45f4-baae-1888dd067d88"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326447776*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1552276129,ExpressionUUID->"56057232-3575-48cb-ad01-b202885b8112"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326466885*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1473556347,ExpressionUUID->"8a9a3b74-f180-405a-9a47-3cf2c300ce9e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detfrozen", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detfrozen[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978326484229*^9},
 CellLabel->"During evaluation of In[706]:=",
 CellID->1280106825,ExpressionUUID->"565e7c15-2f8b-40be-a00d-d21216df5ee2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MetricsOfVBundle", "[", "TangentM3", "]"}]], "Input",
 CellLabel->
  "In[707]:=",ExpressionUUID->"90c7de95-9167-49db-939e-ff03faeb5356"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"metricg", ",", "frozen"}], "}"}]], "Output",
 CellChangeTimes->{3.7289783268276377`*^9},
 CellLabel->"Out[707]=",
 CellID->1264232129,ExpressionUUID->"0de02cbb-dca1-4e6e-80e9-5137da472a86"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compare these two results:", "MathCaption",ExpressionUUID->"e994c9fc-b228-491e-ab58-58a54b886ece"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"frozen", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[708]:=",ExpressionUUID->"65f50658-960b-422e-84d6-c4b79326475d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"f", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`frozen[$CellContext`a, -$CellContext`b],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289783272289553`*^9},
 CellLabel->"Out[708]=",
 CellID->746188685,ExpressionUUID->"18c16b5f-4278-4ea2-b50f-286859bafbbd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That is because these three expressions behave very differently:\
\>", "MathCaption",ExpressionUUID->"eaa046e8-c311-411d-b9b4-d1fd191d43d3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"frozen", "[", 
     RowBox[{"a", ",", "c"}], "]"}], 
    RowBox[{"frozen", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "c"}], "]"}], 
    RowBox[{"frozen", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "c"}], "]"}], 
    RowBox[{"metricg", "[", 
     RowBox[{
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}]}], "}"}]], "Input",
 CellLabel->
  "In[709]:=",ExpressionUUID->"3ab0ac4d-2e04-4bf9-b953-354986eb5bc7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"f", 
         StyleBox[GridBox[{
            {"a", "c"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`frozen[$CellContext`a, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"f", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`frozen[-$CellContext`c, -$CellContext`b],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"f", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`frozen[-$CellContext`c, -$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {"a", "c"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[$CellContext`a, $CellContext`c],
     Editable->False]}], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978327545805*^9},
 CellLabel->"Out[709]=",
 CellID->2036213962,ExpressionUUID->"ddf6f53c-a46d-46bd-95c1-7dc0daa5c9dd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[710]:=",ExpressionUUID->"f8c299b9-9b74-4655-b0c5-6abac751f942"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"f", 
         StyleBox[GridBox[{
            {"a", "c"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`frozen[$CellContext`a, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"f", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`frozen[-$CellContext`c, -$CellContext`b],
     Editable->False]}], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"f", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`frozen[$CellContext`a, -$CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978327746771*^9},
 CellLabel->"Out[710]=",
 CellID->348660742,ExpressionUUID->"409f52a9-13c6-40f2-a7c6-d17f1e8d6d38"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In particular the inverse of ",
 StyleBox["frozen[-a,-b]",
  FontFamily->"Courier"],
 " is not ",
 StyleBox["frozen[a,b]",
  FontFamily->"Courier"],
 ", but ",
 StyleBox["Invfrozen[a,b]",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"22abeb6d-4427-43ca-b0af-747627b252a9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Invfrozen", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"frozen", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[711]:=",ExpressionUUID->"6ea67654-6fd0-44b4-aa4b-f6ce902fe8ba"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "a"},
         {"c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`c, $CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978327875778*^9},
 CellLabel->"Out[711]=",
 CellID->1605701778,ExpressionUUID->"9ef7e4f1-a488-4d3f-b61f-af8654082b81"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is problematic to generalize typical expressions. For example this is not \
zero now:\
\>", "MathCaption",ExpressionUUID->"838a3c28-6591-4068-b72b-93a193e07b20"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD2", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"EinsteinCD2", "[", 
   RowBox[{"a", ",", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[712]:=",ExpressionUUID->"1618bef4-5bda-44fd-95d1-d15edad861cc"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "a"], GridBox[{
      {
       RowBox[{"G", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD2[-$CellContext`a][
   $CellContext`EinsteinCD2[$CellContext`a, $CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289783279897127`*^9},
 CellLabel->"Out[712]=",
 CellID->1750274644,ExpressionUUID->"65896c37-674e-4f0f-a66d-5231b87f74e7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but this is zero:", "MathCaption",
 CellChangeTimes->{
  3.49461159658919*^9},ExpressionUUID->"1c84daf4-0921-41cd-84fb-8efd89cb9e79"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Invfrozen", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{
   RowBox[{"CD2", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"EinsteinCD2", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[713]:=",ExpressionUUID->"7f08fb4c-3a52-4816-be6b-3345d349a421"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978328156356*^9},
 CellLabel->"Out[713]=",
 CellID->634945102,ExpressionUUID->"b584e013-1ff2-4078-9374-1bb345513b7f"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox["V1.0 UPDATE",
  FontColor->RGBColor[1, 0, 0]],
 ": An important problem is that of the symmetries of the Riemann tensor. \
Only if we lower the upper index of the Riemann tensor with the appropriate \
metric we get the correct answer. But when that appropriate metric is not the \
first metric then the all-lower-indices Riemann tensor cannot be identified \
with the original tensor directly. Hence, we introduce the ",
 StyleBox["RiemannDown",
  FontFamily->"Courier"],
 " tensor."
}], "Text",
 CellChangeTimes->{{3.494611607813137*^9, 
  3.494611749668467*^9}},ExpressionUUID->"510f21ea-fe11-4b0d-95bb-\
64b3e8d06f77"],

Cell[CellGroupData[{

Cell["Note the different symmetries of these two tensors:", "MathCaption",
 CellChangeTimes->{
  3.49461159658919*^9, {3.494611862940583*^9, 
   3.494611873388983*^9}},ExpressionUUID->"4b45bc2b-27f2-406a-8985-\
7aedb9056d62"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"RiemannCD2", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "d"}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.494611753389804*^9, 3.4946117659043427`*^9}},
 CellLabel->
  "In[714]:=",ExpressionUUID->"2dc77a75-1962-4edb-833e-c04d53ed4f28"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{"-", 
       RowBox[{"(", "1", ",", "2", ")"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     -xAct`xPerm`Cycles[{1, 2}],
     Editable->False], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.7289783282917137`*^9},
 CellLabel->"Out[714]=",
 CellID->676087371,ExpressionUUID->"7ef8034a-2069-4b3c-bcce-47fa9d35026d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SymmetryGroupOfTensor", "[", 
  RowBox[{"RiemannDownCD2", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}]}], "]"}], "]"}]], "Input",
 CellChangeTimes->{{3.494611771724618*^9, 3.494611782549254*^9}},
 CellLabel->
  "In[715]:=",ExpressionUUID->"d8a829e9-565f-4c96-91e1-5175ea6a06bf"],

Cell[BoxData[
 RowBox[{"StrongGenSet", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2", ",", "3", ",", "4"}], "}"}], ",", 
   RowBox[{"GenSet", "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "1", ",", "2", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{1, 2}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{"-", 
        RowBox[{"(", "3", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      -xAct`xPerm`Cycles[{3, 4}],
      Editable->False], ",", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        RowBox[{"(", "1", ",", "3", ")"}], 
        RowBox[{"(", "2", ",", "4", ")"}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xPerm`Cycles[{1, 3}, {2, 4}],
      Editable->False]}], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.728978328436138*^9},
 CellLabel->"Out[715]=",
 CellID->664468120,ExpressionUUID->"a2839592-2b57-430d-a70c-4ce4e1f88211"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is possible to convert between those two tensors (note the different \
formatted representation):\
\>", "MathCaption",
 CellChangeTimes->{
  3.49461159658919*^9, {3.494611862940583*^9, 
   3.494611911132543*^9}},ExpressionUUID->"1b09f8de-9e60-4637-ba44-\
9480f797f34f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCD2", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "d"}], "]"}]], "Input",
 CellChangeTimes->{{3.494611820054744*^9, 3.494611825200121*^9}},
 CellLabel->
  "In[716]:=",ExpressionUUID->"0df1c0e4-1c8f-4fdd-8570-edb83d751a68"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "D", "]"}], 
      StyleBox[GridBox[{
         {" ", " ", " ", "d"},
         {"a", "b", "c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RiemannCD2[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978328575889*^9},
 CellLabel->"Out[716]=",
 CellID->800728440,ExpressionUUID->"aaba159a-ed2a-42a6-bb39-91b4034b2448"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "RiemannToRiemannDown"}]], "Input",
 CellChangeTimes->{{3.4946118279979258`*^9, 3.494611833092725*^9}},
 CellLabel->
  "In[717]:=",ExpressionUUID->"79e37963-494e-4c29-bae9-72b912162898"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"if", 
       StyleBox[GridBox[{
          {"e", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Invfrozen[$CellContext`e, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[ScriptCapitalR]", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannDownCD2[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783287192698`*^9},
 CellLabel->"Out[717]=",
 CellID->1710358978,ExpressionUUID->"5c7e77a8-fc9f-45fe-86e9-5cd50385ef37"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "RiemannDownToRiemann"}]], "Input",
 CellChangeTimes->{{3.494611838853174*^9, 3.4946118430469837`*^9}},
 CellLabel->
  "In[718]:=",ExpressionUUID->"34b6f441-73da-40aa-9496-189da2075fe2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "D", "]"}], 
      StyleBox[GridBox[{
         {" ", " ", " ", "d"},
         {"a", "b", "c", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RiemannCD2[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978328991477*^9},
 CellLabel->"Out[718]=",
 CellID->1934728151,ExpressionUUID->"118a4dcd-f568-489c-82fa-8f8f0ab40c75"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "frozen", "]"}]], "Input",
 CellLabel->
  "In[719]:=",ExpressionUUID->"f3006008-56de-4b39-af62-61795b12add5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detfrozen"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detfrozen],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329508254*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1926134630,ExpressionUUID->"a3753980-a0cb-4465-a2e7-2c5733632507"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329524263*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1376231532,ExpressionUUID->"f7caa540-1a07-4f24-a23f-209be01f3c08"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783295435343`*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1924480134,ExpressionUUID->"72727013-c27c-4115-9a8a-9a0ff3a3f4a4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897832956177*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1486404671,ExpressionUUID->"aa8d238e-3d75-4625-9e7a-57d0ba06b7e7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783295798388`*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->326039946,ExpressionUUID->"2a1f1dff-1323-4341-a2fd-ab967bfef03a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329598239*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1522588925,ExpressionUUID->"3fadeb9d-b807-47d6-a179-d27b606e048e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783296169977`*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->736924823,ExpressionUUID->"4518d517-09b4-4bb1-a2c6-c6c915a1221a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannDownCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannDownCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329635145*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->957307912,ExpressionUUID->"efdee0ad-68a2-473f-b1e8-e4618ea99b09"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329654335*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->526593660,ExpressionUUID->"f19f8f15-ccbe-4228-a614-2041ea397220"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329673616*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->679534333,ExpressionUUID->"fb40b726-2c52-448b-829b-47234d4cc7d7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329692998*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->615499985,ExpressionUUID->"294f7e2b-7d49-4a3b-a6b1-8806a83c199b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329711255*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->279520296,ExpressionUUID->"152058f7-9ff5-4cf9-bba2-dc368ea1ee45"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonfrozen"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonfrozen],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783297287197`*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1350878883,ExpressionUUID->"95199d39-320b-465c-b8c1-49a05b757d5a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Invfrozen"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "inverse metric tensor", 
   " ", $CellContext`Invfrozen],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329746159*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->1024270497,ExpressionUUID->"b44e9bc3-6913-4c3c-a2a5-49ac0b45a7fb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "frozen"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`frozen],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978329763467*^9},
 CellLabel->"During evaluation of In[719]:=",
 CellID->685328727,ExpressionUUID->"a360f435-5ff2-4efa-9cc7-9602a07f1755"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell["\<\
Other commands and algorithms have been updated in V1 .0 to work properly \
with frozen metrics.\
\>", "Text",
 CellChangeTimes->{{3.494611927247816*^9, 
  3.494611966717822*^9}},ExpressionUUID->"c1da5237-5097-415f-9826-\
b2d5585aaf5c"]
}, Closed]],

Cell[CellGroupData[{

Cell["7.5. Curvature tensors of the metric", "Subsection",ExpressionUUID->"c9e7f10e-d301-41ba-95e3-c8658b4c8e4c"],

Cell[TextData[{
 "When we define a metric, a number of curvature tensors are defined. \
Currently they are: ",
 StyleBox["Riemann",
  FontFamily->"Courier"],
 ", ",
 StyleBox["Ricci",
  FontFamily->"Courier"],
 ", ",
 StyleBox["RicciScalar",
  FontFamily->"Courier"],
 ", ",
 StyleBox["Einstein",
  FontFamily->"Courier"],
 ", ",
 StyleBox["TFRici",
  FontFamily->"Courier"],
 ", ",
 StyleBox["Weyl",
  FontFamily->"Courier"],
 ". Note that they are all associated to the Levi-Civita connection of the \
metric, and not directly to the metric. There are a number of functions to \
change among them."
}], "Text",ExpressionUUID->"2819f13a-e559-4131-aa25-7d0abe82c8bb"],

Cell[CellGroupData[{

Cell[TextData[{
 "First we have some automatic contractions. These can be prevented with the \
option ",
 StyleBox["CurvatureRelations",
  FontFamily->"Courier"],
 " of ",
 StyleBox["DefCovD",
  FontFamily->"Courier"],
 ". By default, contractions of the Riemann tensor are automatically replaced \
by the Ricci tensor:"
}], "MathCaption",ExpressionUUID->"eeee7a0a-d524-4c12-b94a-2393fa69c85a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RiemannCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", "a"}], "]"}]], "Input",
 CellLabel->
  "In[720]:=",ExpressionUUID->"b4fa354f-1932-46f4-a35b-bcf27753d8e5"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`b, -$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978339258285*^9},
 CellLabel->"Out[720]=",
 CellID->1233400720,ExpressionUUID->"46b31d8d-6016-486c-9188-1c6a97987660"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[721]:=",ExpressionUUID->"8d0f1729-1bbe-4d8c-af1b-3d980096eedd"],

Cell["-RicciCD[-b, -c]", "Output",
 CellChangeTimes->{3.7289783394037437`*^9},
 CellLabel->"Out[721]//InputForm=",
 CellID->106256549,ExpressionUUID->"00f97cea-96aa-4673-a2ac-857b51fba085"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
And contractions of the Ricci tensor are replaced by the Ricci scalar:\
\>", "MathCaption",ExpressionUUID->"8cde665a-eee1-4a0f-a7e1-4d9daed2ac8e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"metricg", "[", 
   RowBox[{"c", ",", "d"}], "]"}], 
  RowBox[{"RicciCD", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[722]:=",ExpressionUUID->"62faeb0d-29e3-4951-8dfa-130e3a8eef2e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RicciScalarCD[],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978339525382*^9},
 CellLabel->"Out[722]=",
 CellID->1294908386,ExpressionUUID->"5bc6f494-bb42-46a4-b446-894ace6bff1b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[723]:=",ExpressionUUID->"8be25018-45df-4868-a79a-83257f314630"],

Cell["RicciScalarCD[]", "Output",
 CellChangeTimes->{3.728978339654249*^9},
 CellLabel->"Out[723]//InputForm=",
 CellID->925964804,ExpressionUUID->"b17a6d3c-f374-4f12-9114-49da68f06d37"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RicciCD", "[", 
  RowBox[{"a", ",", 
   RowBox[{"-", "a"}]}], "]"}]], "Input",
 CellLabel->
  "In[724]:=",ExpressionUUID->"8dd2301b-17ea-40f8-8c76-39ae7bdf3f06"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RicciScalarCD[],
  Editable->False]], "Output",
 CellChangeTimes->{3.72897833981872*^9},
 CellLabel->"Out[724]=",
 CellID->2023954666,ExpressionUUID->"ea32b9e1-ee64-4645-b8af-66db1c95381f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Then we can change among the tensors:", "MathCaption",ExpressionUUID->"430ee0e2-99c7-43b4-8316-4f729d80503b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"EinsteinCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "//", "EinsteinToRicci"}]], "Input",
 CellLabel->
  "In[725]:=",ExpressionUUID->"58c5d7c8-83d2-4c35-ad3f-1d2235f35ce6"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciScalarCD[],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978340231461*^9},
 CellLabel->"Out[725]=",
 CellID->465102982,ExpressionUUID->"148e0ccc-7fb7-40f4-9f18-816f5722f937"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "RicciToEinstein"}]], "Input",
 CellLabel->
  "In[726]:=",ExpressionUUID->"e1580a7b-1ab8-435a-912a-be3e24e8c59b"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"G", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978340476632*^9},
 CellLabel->"Out[726]=",
 CellID->145343394,ExpressionUUID->"b46c4d42-7bba-4a6f-b4c7-6b320ee5b25e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["In a 3d manifold the Weyl tensor is zero:", "MathCaption",ExpressionUUID->"f1c1db60-4479-4a50-aaa6-23c870613374"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"WeylCD", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "d"}]}], "]"}]], "Input",
 CellLabel->
  "In[727]:=",ExpressionUUID->"6fcf7d89-ecfe-4dd0-860a-0265eb71b876"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978340625886*^9},
 CellLabel->"Out[727]=",
 CellID->1097516319,ExpressionUUID->"131d0359-8d44-4efe-971f-54ce3c866d53"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}]}], "]"}], "//", "RiemannToWeyl"}]], "Input",
 CellLabel->
  "In[728]:=",ExpressionUUID->"e80eadee-e367-45dc-aea3-02cd4602bfb5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciScalarCD[],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciScalarCD[],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978340774992*^9},
 CellLabel->"Out[728]=",
 CellID->1219729350,ExpressionUUID->"3c36ebdc-1e5d-44df-8675-a374e7a30e3b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can also change between Ricci and TFRicci:", "MathCaption",ExpressionUUID->"58f48673-1713-4801-ba9b-4dae90081aac"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RicciCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "//", "RicciToTFRicci"}]], "Input",
 CellLabel->
  "In[729]:=",ExpressionUUID->"a66c7c3f-e420-4696-aa40-57e9b4e6a2ae"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "3"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciScalarCD[],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"S", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978340993915*^9},
 CellLabel->"Out[729]=",
 CellID->112291188,ExpressionUUID->"96ff3c6a-2f70-4442-980e-5832eac6216f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "TFRicciToRicci"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[730]:=",ExpressionUUID->"0285038f-7773-4063-873c-644297289d55"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978341463336*^9},
 CellLabel->"Out[730]=",
 CellID->1787354650,ExpressionUUID->"1517d70e-c799-420d-be1e-a9c15f1717ec"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["RiemannToWeyl",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "RiemannToWeyl.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand Riemann tensors into Weyl, Ricci and RicciScalar tensors", 
  "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["WeylToRiemann",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "WeylToRiemann.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand Weyl tensors into Riemann, Ricci and RicciScalar tensors", 
  "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["RicciToEinstein",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "RicciToEinstein.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand Ricci tensors into Einstein and RicciScalar tensors", 
  "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["EinsteinToRicci",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "EinsteinToRicci.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand Einstein tensors into Ricci and RicciScalar tensors\n", 
  "Text"],
 StyleBox[ButtonBox["RicciToTFRicci",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "RicciToTFRicci.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand Ricci tensors into TFRicci and RicciScalar tensors", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["TFRicciToRicci",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TFRicciToRicci.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand TFRicci tensors into Ricci and RicciScalar tensors", 
  "Text"]
}], "DefinitionBox",ExpressionUUID->"e90876ee-f938-4ba5-9b09-7ae1a3d38dd4"],

Cell["Relations among curvature objects.", "Text",ExpressionUUID->"8b50e31a-1981-40e6-8c72-ca668477739c"],

Cell["\<\
As we said, we can work with connections which are compatible with a given \
metric field but have torsion too.\
\>", "Text",ExpressionUUID->"7aa3cdb2-26b3-4e8f-9be8-e157f8ea7f65"],

Cell[CellGroupData[{

Cell["\<\
We define a new connection associated to our metric field, but this time with \
torsion. The Riemann tensor is antisymmetric in both pairs, but now those \
pairs cannot be exchanged. Hence the Ricci tensor is not symmetric. Currently \
Weyl is not even defined:\
\>", "MathCaption",ExpressionUUID->"b03ccc55-d410-4992-bcf4-afc1922e11b8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefCovD", "[", 
  RowBox[{
   RowBox[{"CDT", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<#\>\"", ",", "\"\<D\>\""}], "}"}]}], ",", 
   RowBox[{"Torsion", "\[Rule]", "True"}], ",", 
   RowBox[{"FromMetric", "\[Rule]", "metricg"}]}], "]"}]], "Input",
 CellLabel->
  "In[731]:=",ExpressionUUID->"ffc21272-330c-4232-9ae3-0d0f84cf54cc"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CDT", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CDT[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341649054*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1068348314,ExpressionUUID->"f2167a3a-23f8-421b-8a8a-4b13d4bd5d2a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"TorsionCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "torsion tensor ", 
   $CellContext`TorsionCDT[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341665655*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->619340831,ExpressionUUID->"b120262a-c2d1-45ca-ae02-a33764f5eb3c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341684154*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->408109608,ExpressionUUID->"a43ecaf6-219b-4e99-8889-a90d27beeda0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Antisymmetric pairs cannot be exchanged.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCDT[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", "Antisymmetric pairs cannot be exchanged."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783417025948`*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->35413138,ExpressionUUID->"27ba4c4c-c7e9-4d29-aca2-653d5f6a489d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "non-symmetric Ricci tensor ", 
   $CellContext`RicciCDT[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897834172243*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->335640147,ExpressionUUID->"61db10a6-9903-4c99-a45c-3eb38b0ee7f9"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728978341740056*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1811893607,ExpressionUUID->"59c95aae-605d-4ec1-86c2-a67db6f25f16"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCDT", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCDT[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341757044*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1177738389,ExpressionUUID->"0e625634-2a82-4576-b5c3-a797f8525a06"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.7289783417741632`*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->892329524,ExpressionUUID->"ce8a4db4-23a8-4e20-929a-fc2efbd8a04b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric Einstein tensor ", 
   $CellContext`EinsteinCDT[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341791006*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1504002867,ExpressionUUID->"c8121aab-17ed-4441-9340-bc1c2817c31c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Antisymmetric pairs cannot be exchanged.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCDT[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", "Antisymmetric pairs cannot be exchanged."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341808159*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1893660479,ExpressionUUID->"a423dacd-9679-4d39-8e2b-dd6567e66d24"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCDT", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "non-symmetric TFRicci tensor ", 
   $CellContext`TFRicciCDT[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341825509*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1044022013,ExpressionUUID->"e9f2c5c9-72e1-40a1-8f52-722b57dfa5c8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCDT", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCDT[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897834184591*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1936576501,ExpressionUUID->"62cc6729-5e8b-4f1c-ae3d-dd912ffcb6f4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978341867754*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->779098375,ExpressionUUID->"579170c6-6ae0-4ed3-affe-cd06842ac5d1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783418894367`*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->1098592866,ExpressionUUID->"3b0cfec6-6956-4491-985e-3e6c22b9045e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783419096107`*^9},
 CellLabel->"During evaluation of In[731]:=",
 CellID->2133385603,ExpressionUUID->"e456c2f2-80ac-4d1d-8b20-3ca415a1edb7"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The difference of covariant derivatives", "MathCaption",ExpressionUUID->"417fd3fe-942e-46f5-a7c8-447592556373"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CDT", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", "b", "]"}], "]"}], "-", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{"v", "[", "b", "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[732]:=",ExpressionUUID->"fe17a7b2-8534-4388-9f4a-4458200cc735"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "a"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CDT[-$CellContext`a][
    $CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897834211731*^9},
 CellLabel->"Out[732]=",
 CellID->1424859309,ExpressionUUID->"d1e78e72-5a2e-46e6-9394-a333fbc3445f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["is given by a Christoffel tensor", "MathCaption",ExpressionUUID->"16524a6d-a0d6-40b0-b006-52bedf3e5ab7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeCovD", "[", 
  RowBox[{"%", ",", "CDT", ",", "CD"}], "]"}]], "Input",
 CellLabel->
  "In[733]:=",ExpressionUUID->"e3792122-8606-47ae-936f-0b62fcc472c6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCDCDT", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`ChristoffelCDCDT[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342253257*^9},
 CellLabel->"During evaluation of In[733]:=",
 CellID->1020258666,ExpressionUUID->"d4197de6-5044-46f0-98b7-1173cbf96796"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", 
         RowBox[{"\[EmptyDownTriangle]", ",", "D"}], "]"}], 
        StyleBox[GridBox[{
           {"b", " ", " "},
           {" ", "a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCDCDT[$CellContext`b, -$CellContext`a, \
-$CellContext`c],
    Editable->False]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978342284642*^9},
 CellLabel->"Out[733]=",
 CellID->1977961814,ExpressionUUID->"ae21565f-47e6-4b1a-a0fc-c32ec2ca655f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
which can be expressed uniquely in terms of torsion and metric fields:\
\>", "MathCaption",ExpressionUUID->"8d4b16d9-9d1f-4264-bf1d-dacb82c3c478"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ChristoffelToGradMetric"}]], "Input",
 CellLabel->
  "In[734]:=",ExpressionUUID->"4208acf7-41f2-4e07-aebe-6e3c48b84b19"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"b", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`b, $CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"T", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"a", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionCDT[-$CellContext`a, -$CellContext`c, \
-$CellContext`d],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"T", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"c", "a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionCDT[-$CellContext`c, -$CellContext`a, \
-$CellContext`d],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"T", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"d", "a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TorsionCDT[-$CellContext`d, -$CellContext`a, \
-$CellContext`c],
     Editable->False]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978342433175*^9},
 CellLabel->"Out[734]=",
 CellID->557352338,ExpressionUUID->"20e68122-9fb9-4890-918e-04deb4d93bd7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefCovD", "[", "CDT", "]"}]], "Input",
 CellLabel->
  "In[735]:=",ExpressionUUID->"de14bbed-f695-447e-b2b3-c2a3924fa5a5"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "ChristoffelCDCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`ChristoffelCDCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342569006*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->97076552,ExpressionUUID->"92ec3068-c305-499d-a799-ffc0b8a09ca5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Christoffel tensor", " ", $CellContext`ChristoffelCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342584766*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->1625521391,ExpressionUUID->"9b477e20-3352-49dc-8869-4d7909ce6ae7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Einstein tensor", " ", $CellContext`EinsteinCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342605234*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->1998545050,ExpressionUUID->"5f345735-401c-4c17-91ff-4b8fcc6b3c05"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342623867*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->754932623,ExpressionUUID->"0f9049f3-3ac7-4037-99c2-cf88923439e4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897834264314*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->819208285,ExpressionUUID->"e2087b04-c60c-42fa-883a-06dc04c88f72"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783426617393`*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->915985078,ExpressionUUID->"a036a213-8d3d-49d9-9e21-973fea995efa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342679574*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->1798424250,ExpressionUUID->"76c81ada-78cb-4f4c-b955-047db66128f0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric TFRicci tensor", " ", $CellContext`TFRicciCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342697052*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->284038225,ExpressionUUID->"085ea881-ffeb-420f-9019-385ad3dc4929"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783427157393`*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->2121783941,ExpressionUUID->"49ab64af-aa6f-4939-b9f0-ea22c1ce7fac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978342732996*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->1856737952,ExpressionUUID->"e764238d-8e94-489e-ba23-6bf37f7cb480"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CDT"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CDT],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783427511168`*^9},
 CellLabel->"During evaluation of In[735]:=",
 CellID->1109884707,ExpressionUUID->"378c41e5-f9ec-4001-b9ec-aabcf43506e8"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.6. Product metrics", "Subsection",ExpressionUUID->"8feec4d9-2dd2-4696-85e4-204e145be129"],

Cell["\<\
Spacetimes with a high degree of symmetry can be sometimes locally decomposed \
as products of reduced manifolds and the orbits of symmetry. The metric of \
the whole manifold can then be given in terms of metrics on simpler vbundles. \
Currently there is a stupid limitation to two subvbundles. It will be dropped \
soon.\
\>", "Text",ExpressionUUID->"c3e7588b-fba1-4178-a94d-14000e1e9e16"],

Cell[CellGroupData[{

Cell["Define a product manifold:", "MathCaption",ExpressionUUID->"fb254167-c88e-4af7-b941-27e64f4dc52e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M5", ",", 
   RowBox[{"{", 
    RowBox[{"M3", ",", "S2"}], "}"}], ",", " ", 
   RowBox[{"{", 
    RowBox[{
    "\[Mu]", ",", " ", "\[Nu]", ",", " ", "\[Lambda]", ",", " ", "\[Sigma]", 
     ",", " ", "\[Eta]", ",", " ", "\[Rho]"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[736]:=",ExpressionUUID->"96a48b35-84c8-43eb-ad8b-53ea42abe70b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M5", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", 
   "manifold ", $CellContext`M5, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978350876842*^9},
 CellLabel->"During evaluation of In[736]:=",
 CellID->1357370834,ExpressionUUID->"042496d1-c24f-4f5f-bd56-64c013f86b88"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM5", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM5, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783508944693`*^9},
 CellLabel->"During evaluation of In[736]:=",
 CellID->612824957,ExpressionUUID->"6f398dac-3ce6-410e-be39-0c7854b21a13"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a product metric. We already have a scalar ",
 StyleBox["r[]",
  FontFamily->"Courier"],
 " on ",
 StyleBox["M3",
  FontFamily->"Courier"],
 ". Now we define another scalar ",
 StyleBox["w[]",
  FontFamily->"Courier"],
 " on ",
 StyleBox["S2",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"97a33204-3f58-4eaa-b471-485978fd7536"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"w", "[", "]"}], ",", "S2"}], "]"}]], "Input",
 CellLabel->
  "In[737]:=",ExpressionUUID->"01f703f3-fc2a-402b-a278-20d21d80830b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"w", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`w[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978351062262*^9},
 CellLabel->"During evaluation of In[737]:=",
 CellID->417481287,ExpressionUUID->"090e8cf6-7ccf-41ae-bbd5-8fecf861f731"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefProductMetric", "[", 
  RowBox[{
   RowBox[{"g5", "[", 
    RowBox[{
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"TangentM3", ",", 
       RowBox[{"w", "[", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"TangentS2", ",", 
       RowBox[{"r", "[", "]"}]}], "}"}]}], "}"}], ",", "Cd5", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<#\>\"", ",", "\"\<\[GothicCapitalD]\>\""}], "}"}]}]}], 
  "]"}]], "Input",
 CellLabel->
  "In[738]:=",ExpressionUUID->"6fb4c958-ffa0-45f7-b0c4-95bc1c2d5cc1"],

Cell[BoxData[
 TemplateBox[{
  "DefProductMetric", "nometric", 
   "\"VBundle \
\\!\\(\\*InterpretationBox[StyleBox[\\\"\[DoubleStruckCapitalT]S2\\\", \
Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], TangentS2, \
Rule[Editable, False]]\\) does not have a metric.\"", 2, 738, 90, 
   19253573948027831207, "Local", "xAct`xTensor`DefProductMetric"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978351247184*^9},
 CellLabel->"During evaluation of In[738]:=",
 CellID->778915860,ExpressionUUID->"4735604a-e080-480e-98ee-4054db6ee92b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{"1", ",", 
   RowBox[{"gamma", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], ",", "CdS", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<:\>\"", ",", "\"\<D\>\""}], "}"}]}], ",", 
   RowBox[{"PrintAs", "\[Rule]", "\"\<\[Gamma]\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[739]:=",ExpressionUUID->"e080aa83-81c9-4bb2-a392-f61627144701"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"gamma", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`gamma[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783517433243`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->178546908,ExpressionUUID->"d7e07fbe-09e6-47ab-959d-50bcd2b3d69b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilongamma", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilongamma[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783517606564`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->803227897,ExpressionUUID->"461f9d05-6b70-45ba-85c7-f7995c2eb6d8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CdS", "[", 
    RowBox[{"-", "A"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CdS[-$CellContext`A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978351779358*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->769059837,ExpressionUUID->"d94900f6-786b-48f2-b1ed-d403c1c028c5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCdS", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCdS[$CellContext`A, -$CellContext`B, -C], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978351797072*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->2032092813,ExpressionUUID->"fe1f4d19-a26f-40d7-8976-d3339492d2a5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCdS", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCdS[$CellContext`A, -$CellContext`B, -C], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783518139553`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1951469850,ExpressionUUID->"8e6ed137-bbcb-4616-89d9-0b8408c0f5a7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCdS", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCdS[-$CellContext`A, -$CellContext`B, -C, -D], ". ", 
   ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783518314037`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1983511966,ExpressionUUID->"22ee8ddc-a012-4b65-88eb-ab8a40b78c59"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCdS", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCdS[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978351848401*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1496960181,ExpressionUUID->"2b928888-366b-4d12-b824-d262064e36f7"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728978351866496*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1638837462,ExpressionUUID->"97fc5034-c768-4bab-87a4-46081f023035"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCdS", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCdS[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783518839693`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->315967412,ExpressionUUID->"1d51a4f5-eb3f-42d9-a296-c6c2741daf19"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728978351901134*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1884718100,ExpressionUUID->"2e090813-7cfa-4f7c-8811-ba8ac3ec8fce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCdS", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing Einstein tensor ", 
   $CellContext`EinsteinCdS[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978351917987*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1073313450,ExpressionUUID->"0dc380bc-bbb9-425b-9ed4-d1608df96eea"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCdS", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCdS[-$CellContext`A, -$CellContext`B, -C, -D], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835193487*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->316519350,ExpressionUUID->"36d2d141-3902-4140-b129-ccecf47dd3e8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCdS", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing TFRicci tensor ", 
   $CellContext`TFRicciCdS[-$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783519527473`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->423721831,ExpressionUUID->"8df546ae-3a0d-49f1-9a9d-2c34cd61a587"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCdS", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCdS[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978351970253*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->437040165,ExpressionUUID->"eac3c76a-6c5b-47ff-9d7d-e3a9de7bd153"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "2"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783519885197`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->939980415,ExpressionUUID->"0233b1ee-fed7-4a79-b5c1-0d470eebf6de"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "2"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352004897*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1361127573,ExpressionUUID->"8b670bbe-27d8-4f2b-80b5-77884b697bcf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "2"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783520219917`*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->1475311567,ExpressionUUID->"265a2972-def2-4d08-9ee2-03d7c8be6df7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detgamma", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detgamma[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352040003*^9},
 CellLabel->"During evaluation of In[739]:=",
 CellID->715318307,ExpressionUUID->"22eae6a0-6c52-4a35-98c4-dc42835c0bc6"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We define the warped metric     ",
 StyleBox[" w[]^2 MetricOfVBundle[TangentM3]+r[]^2 MetricOfVBundle[TangentS2] \
", "DisplayMath"],
 ":"
}], "MathCaption",ExpressionUUID->"736dfdd3-5a7a-45ec-803f-f733d4808ae1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefProductMetric", "[", 
  RowBox[{
   RowBox[{"g5", "[", 
    RowBox[{
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"TangentM3", ",", 
       RowBox[{"w", "[", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"TangentS2", ",", 
       RowBox[{"r", "[", "]"}]}], "}"}]}], "}"}], ",", "Cd5", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<#\>\"", ",", "\"\<\[GothicCapitalD]\>\""}], "}"}]}]}], 
  "]"}]], "Input",
 CellLabel->
  "In[740]:=",ExpressionUUID->"99955222-ce72-41e5-8b18-70067fb7c6ca"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"g5", "[", 
    RowBox[{
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`g5[-$CellContext`\[Mu], -$CellContext`\[Nu]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783524804564`*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->741329843,ExpressionUUID->"8e9daef9-6f57-4b1a-b7df-505955bf84a6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilong5", "[", 
    RowBox[{
     RowBox[{"-", "\[Eta]"}], ",", 
     RowBox[{"-", "\[Lambda]"}], ",", 
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}], ",", 
     RowBox[{"-", "\[Rho]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilong5[-$CellContext`\[Eta], -$CellContext`\[Lambda], \
-$CellContext`\[Mu], -$CellContext`\[Nu], -$CellContext`\[Rho]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352498281*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1978035054,ExpressionUUID->"36143434-64ff-4a56-9fe6-7cc4eac6e534"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"Cd5", "[", 
    RowBox[{"-", "\[Mu]"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`Cd5[-$CellContext`\[Mu]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783525170383`*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1788297018,ExpressionUUID->"010a347c-af20-4b1a-bcf1-6f73c97017a8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCd5", "[", 
    RowBox[{"\[Eta]", ",", 
     RowBox[{"-", "\[Lambda]"}], ",", 
     RowBox[{"-", "\[Mu]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCd5[$CellContext`\[Eta], -$CellContext`\[Lambda], \
-$CellContext`\[Mu]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783525350847`*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1874341761,ExpressionUUID->"502036cf-dae6-4eaa-8227-df2ea6026c37"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCd5", "[", 
    RowBox[{"\[Eta]", ",", 
     RowBox[{"-", "\[Lambda]"}], ",", 
     RowBox[{"-", "\[Mu]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCd5[$CellContext`\[Eta], -$CellContext`\[Lambda], \
-$CellContext`\[Mu]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352552651*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->624359284,ExpressionUUID->"6d5d6341-6f69-4444-beda-4b156b47caee"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCd5", "[", 
    RowBox[{
     RowBox[{"-", "\[Eta]"}], ",", 
     RowBox[{"-", "\[Lambda]"}], ",", 
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCd5[-$CellContext`\[Eta], -$CellContext`\[Lambda], \
-$CellContext`\[Mu], -$CellContext`\[Nu]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352570586*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1001386946,ExpressionUUID->"0ad97094-7c39-49d3-a722-15dd51aaacec"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCd5", "[", 
    RowBox[{
     RowBox[{"-", "\[Eta]"}], ",", 
     RowBox[{"-", "\[Lambda]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCd5[-$CellContext`\[Eta], -$CellContext`\[Lambda]], ". ",
    ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783525876713`*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->481059883,ExpressionUUID->"cd3971bb-79bb-415b-837f-76d80b499f91"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCd5", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCd5[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835260533*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->2045006273,ExpressionUUID->"d80fad38-44e5-425c-8ec7-de384067bf42"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCd5", "[", 
    RowBox[{
     RowBox[{"-", "\[Eta]"}], ",", 
     RowBox[{"-", "\[Lambda]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCd5[-$CellContext`\[Eta], -$CellContext`\[Lambda]], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835262281*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->2092143723,ExpressionUUID->"95254926-4096-424e-af71-4d9f87adfc30"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCd5", "[", 
    RowBox[{
     RowBox[{"-", "\[Eta]"}], ",", 
     RowBox[{"-", "\[Lambda]"}], ",", 
     RowBox[{"-", "\[Mu]"}], ",", 
     RowBox[{"-", "\[Nu]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "Weyl tensor ", 
   $CellContext`WeylCd5[-$CellContext`\[Eta], -$CellContext`\[Lambda], \
-$CellContext`\[Mu], -$CellContext`\[Nu]], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352640767*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1909800130,ExpressionUUID->"e7b846f1-0a2a-402e-9c2b-3daf90f7edce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCd5", "[", 
    RowBox[{
     RowBox[{"-", "\[Eta]"}], ",", 
     RowBox[{"-", "\[Lambda]"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCd5[-$CellContext`\[Eta], -$CellContext`\[Lambda]], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352658461*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1720196848,ExpressionUUID->"acd4b90e-055a-4274-b95b-9deeef17124f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCd5", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCd5[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352677684*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1923568422,ExpressionUUID->"5e78d282-a9fc-481d-8ecc-baf3e4640b90"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "5"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352695684*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1195646616,ExpressionUUID->"8e868cb3-6596-4e3d-9d9c-251e6e98f508"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "5"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352801338*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1228190133,ExpressionUUID->"f43336cd-2a3c-4369-a6a8-5cf2db35afa2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "5"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352820281*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->578750367,ExpressionUUID->"044bdc6b-8fd5-4191-8b1b-ace978ef741f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detg5", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detg5[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978352837143*^9},
 CellLabel->"During evaluation of In[740]:=",
 CellID->1103980906,ExpressionUUID->"2519bce2-0437-4b6b-8234-15ec0dcd26f5"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["ExpandProducMetric",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ExpandProductMetric.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Expansion of the metric of a product manifold into objects of its \
submanifolds", "Text"]
}], "DefinitionBox",ExpressionUUID->"409cdb9c-da8a-4fd4-8ff0-a5c50734fedd"],

Cell["Computations with product metrics.", "Text",ExpressionUUID->"cf19eaea-e9ee-4840-a162-6f1dcbed5993"],

Cell[CellGroupData[{

Cell["The delta tensors are automatically expanded:", "MathCaption",ExpressionUUID->"af523ea9-5b16-42c7-979e-793677d7d71c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"g5", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"g5", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "B"}]}], "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[741]:=",ExpressionUUID->"edbf4805-bf65-420d-8e11-2cc6f69930a6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "A"},
           {"B", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`B, $CellContext`A],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978353003278*^9},
 CellLabel->"Out[741]=",
 CellID->348377062,ExpressionUUID->"e7e87a04-3027-4d97-8483-650a652aef4e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "/@", "%"}]], "Input",
 CellLabel->
  "In[742]:=",ExpressionUUID->"f9c9c8ec-ebce-4411-8592-2016b45a22ba"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"delta", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", " ", "a"}], "]"}],
     ShowStringCharacters->True,
     NumberMarks->True],
    InputForm[
     xAct`xTensor`delta[-$CellContext`b, $CellContext`a]],
    AutoDelete->True,
    Editable->True], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"delta", "[", 
      RowBox[{
       RowBox[{"-", "B"}], ",", " ", "A"}], "]"}],
     ShowStringCharacters->True,
     NumberMarks->True],
    InputForm[
     xAct`xTensor`delta[-$CellContext`B, $CellContext`A]],
    AutoDelete->True,
    Editable->True]}], "}"}]], "Output",
 CellChangeTimes->{3.728978353133271*^9},
 CellLabel->"Out[742]=",
 CellID->1132859312,ExpressionUUID->"ca6b3b3a-5b58-4401-a678-bb9bbd35a242"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now we can compute any object on ",
 StyleBox["M5",
  FontFamily->"Courier"],
 " in terms of objects of ",
 StyleBox["M3",
  FontFamily->"Courier"],
 " and ",
 StyleBox["S2",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"7f8f4c72-8ef3-485a-a26a-53be4fe5901f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ExpandProductMetric", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"g5", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], ",", 
     RowBox[{"g5", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "B"}]}], "]"}], ",", 
     RowBox[{"g5", "[", 
      RowBox[{
       RowBox[{"-", "A"}], ",", 
       RowBox[{"-", "B"}]}], "]"}]}], "}"}], ",", "g5"}], "]"}]], "Input",
 CellLabel->
  "In[743]:=",ExpressionUUID->"6165cf7e-205e-48e3-9ff3-d8edf6926e08"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[-$CellContext`a, -$CellContext`b],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["w",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`w[],
      Editable->False], "2"]}], ",", "0", ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Gamma]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"A", "B"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`gamma[-$CellContext`A, -$CellContext`B],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978353318025*^9},
 CellLabel->"Out[743]=",
 CellID->376442482,ExpressionUUID->"dea33014-4b1a-45ac-956a-983b7b8ec221"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ExpandProductMetric", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"ChristoffelCd5", "[", 
       RowBox[{"A", ",", 
        RowBox[{"-", "B"}], ",", 
        RowBox[{"-", "C"}]}], "]"}], ",", 
      RowBox[{"ChristoffelCd5", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "B"}], ",", 
        RowBox[{"-", "C"}]}], "]"}], ",", 
      RowBox[{"ChristoffelCd5", "[", 
       RowBox[{"A", ",", "b", ",", 
        RowBox[{"-", "C"}]}], "]"}], ",", 
      RowBox[{"ChristoffelCd5", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "C"}]}], "]"}], ",", 
      RowBox[{"ChristoffelCd5", "[", 
       RowBox[{"A", ",", 
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "c"}]}], "]"}], ",", 
      RowBox[{"ChristoffelCd5", "[", 
       RowBox[{"a", ",", 
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "c"}]}], "]"}]}], "}"}], ",", "g5"}], "]"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[744]:=",ExpressionUUID->"ab8d12b0-6e88-4a0d-bf75-c21380c73697"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "D", "]"}], 
        StyleBox[GridBox[{
           {"A", " ", " "},
           {" ", "B", "C"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCdS[$CellContext`A, -$CellContext`B, -C],
    Editable->False], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Gamma]", 
           StyleBox[GridBox[{
              {" ", " "},
              {"B", "C"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`gamma[-$CellContext`B, -C],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"a", "b"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`a, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "b"], "r"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`b][
         $CellContext`r[]],
        Editable->False], ")"}]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w[],
       Editable->False], "2"]]}], ",", 
   FractionBox[
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "A"},
             {"C", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-C, $CellContext`A],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"b", "a"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`b, $CellContext`a],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`a][
        $CellContext`r[]],
       Editable->False], ")"}]}], 
    RowBox[{
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], " ", 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w[],
       Editable->False], "2"]}]], ",", 
   FractionBox[
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "a"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
      Editable->False], " ", 
     RowBox[{"(", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["D", "C"], "w"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CdS[-C][
        $CellContext`w[]],
       Editable->False], ")"}]}], 
    InterpretationBox[
     StyleBox["w",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`w[],
     Editable->False]], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Gamma]", 
           StyleBox[GridBox[{
              {"A", "B"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`gamma[$CellContext`A, $CellContext`B],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`b, -$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox["w",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`w[],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "B"], "w"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CdS[-$CellContext`B][
         $CellContext`w[]],
        Editable->False], ")"}]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]]}], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"\[CapitalGamma]", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", " ", " "},
           {" ", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289783540618353`*^9},
 CellLabel->"Out[744]=",
 CellID->381970912,ExpressionUUID->"5166d641-dc71-4243-8385-f292f5e45ac2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"RiemannCd5", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], "//", "ExpandProductMetric"}], "//", 
   "ContractMetric"}], "//", "Simplify"}]], "Input",
 CellLabel->
  "In[745]:=",ExpressionUUID->"fa85f148-7855-4871-98f7-7e9a7698599f"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox["w",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`w[],
    Editable->False], "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "b", "c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
     Editable->False], "+", 
    FractionBox[
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Gamma]", 
           StyleBox[GridBox[{
              {"A", "B"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`gamma[$CellContext`A, $CellContext`B],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metricg[-$CellContext`a, -$CellContext`d],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metricg[-$CellContext`b, -$CellContext`c],
          Editable->False]}], "-", 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metricg[-$CellContext`a, -$CellContext`c],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"g", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`metricg[-$CellContext`b, -$CellContext`d],
          Editable->False]}]}], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "A"], "w"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CdS[-$CellContext`A][
         $CellContext`w[]],
        Editable->False], ")"}], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "B"], "w"}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CdS[-$CellContext`B][
         $CellContext`w[]],
        Editable->False], ")"}]}], 
     SuperscriptBox[
      InterpretationBox[
       StyleBox["r",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`r[],
       Editable->False], "2"]]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728978354477344*^9},
 CellLabel->"Out[745]=",
 CellID->956497101,ExpressionUUID->"389b13c3-f182-4b56-abdc-6ab90a6bac58"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"RicciScalarCd5", "[", "]"}], "//", "ExpandProductMetric"}], "//",
    "ContractMetric"}], "//", "Simplification"}]], "Input",
 CellLabel->
  "In[746]:=",ExpressionUUID->"8b38849a-425b-4af7-9b32-2899b7640259"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox[
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciScalarCD[],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"R", "[", "D", "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciScalarCdS[],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["w",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`w[],
      Editable->False], "2"]}], "-", 
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "b"], 
        SubscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`b][
       $CellContext`CD[-$CellContext`a][
        $CellContext`r[]]],
      Editable->False], ")"}]}], "-", 
   RowBox[{"2", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[-$CellContext`a, -$CellContext`b],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[$CellContext`a][
       $CellContext`r[]],
      Editable->False], ")"}], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[EmptyDownTriangle]", "b"], "r"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[$CellContext`b][
       $CellContext`r[]],
      Editable->False], ")"}]}], "-", 
   RowBox[{"6", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Gamma]", 
         StyleBox[GridBox[{
            {"A", "B"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`gamma[$CellContext`A, $CellContext`B],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox["w",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`w[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["D", "B"], 
        SubscriptBox["D", "A"], "w"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CdS[-$CellContext`B][
       $CellContext`CdS[-$CellContext`A][
        $CellContext`w[]]],
      Editable->False], ")"}]}], "-", 
   RowBox[{"6", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Gamma]", 
         StyleBox[GridBox[{
            {" ", " "},
            {"A", "B"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`gamma[-$CellContext`A, -$CellContext`B],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["D", "A"], "w"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CdS[$CellContext`A][
       $CellContext`w[]],
      Editable->False], ")"}], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["D", "B"], "w"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CdS[$CellContext`B][
       $CellContext`w[]],
      Editable->False], ")"}]}]}], 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["w",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`w[],
     Editable->False], "2"]}]]], "Output",
 CellChangeTimes->{3.728978355186089*^9},
 CellLabel->"Out[746]=",
 CellID->1931678712,ExpressionUUID->"004ceac4-e828-433b-9a20-f50557ed45d2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "By default derivative indices cannot be contracted with metric tensors. \
This behaviour can be changed using the option ",
 StyleBox["AllowUpperDerivatives",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"edc3d274-37cf-4dd1-bf15-b27946780313"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ContractMetric", "[", 
   RowBox[{"%", ",", 
    RowBox[{"AllowUpperDerivatives", "\[Rule]", "True"}]}], "]"}], "//", 
  "Simplification"}]], "Input",
 CellLabel->
  "In[747]:=",ExpressionUUID->"ad9564ab-1ee2-4221-9ccf-a383507c9f23"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   RowBox[{
    SuperscriptBox[
     InterpretationBox[
      StyleBox["r",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`r[],
      Editable->False], "2"], " ", 
    InterpretationBox[
     StyleBox[
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciScalarCD[],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"R", "[", "D", "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciScalarCdS[],
     Editable->False], " ", 
    SuperscriptBox[
     InterpretationBox[
      StyleBox["w",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`w[],
      Editable->False], "2"]}], "-", 
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "a"], 
        SuperscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`a][
       $CellContext`CD[$CellContext`a][
        $CellContext`r[]]],
      Editable->False], ")"}]}], "-", 
   RowBox[{"2", " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`a][
       $CellContext`r[]],
      Editable->False], ")"}], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["\[EmptyDownTriangle]", "a"], "r"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[$CellContext`a][
       $CellContext`r[]],
      Editable->False], ")"}]}], "-", 
   RowBox[{"6", " ", 
    InterpretationBox[
     StyleBox["w",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`w[],
     Editable->False], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["D", "A"], 
        SuperscriptBox["D", "A"], "w"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CdS[-$CellContext`A][
       $CellContext`CdS[$CellContext`A][
        $CellContext`w[]]],
      Editable->False], ")"}]}], "-", 
   RowBox[{"6", " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["D", "A"], "w"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CdS[-$CellContext`A][
       $CellContext`w[]],
      Editable->False], ")"}], " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SuperscriptBox["D", "A"], "w"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CdS[$CellContext`A][
       $CellContext`w[]],
      Editable->False], ")"}]}]}], 
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox["r",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`r[],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox["w",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`w[],
     Editable->False], "2"]}]]], "Output",
 CellChangeTimes->{3.728978355335689*^9},
 CellLabel->"Out[747]=",
 CellID->66337238,ExpressionUUID->"647edd95-8c5a-43d8-b9dd-cd3a99d6895c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up", "MathCaption",ExpressionUUID->"c438bef6-fc05-4fdb-bc9c-f0d672a6d8c1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "g5", "]"}]], "Input",
 CellLabel->
  "In[748]:=",ExpressionUUID->"e6953740-2e96-462e-b7a3-2dcfffa52ce2"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detg5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detg5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355469303*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->385946273,ExpressionUUID->"9bcbc04b-4f90-4f57-adda-c02cda688f96"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355485526*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->1198396371,ExpressionUUID->"80e924df-e2f3-4d0c-a810-45cfeb699c1d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355503942*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->198682184,ExpressionUUID->"35306c64-a2c3-4bc4-bd7d-3cfd352de8ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355522566*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->2061879050,ExpressionUUID->"76601480-4efd-42cb-a70e-88acdfa86cec"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355540556*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->57581711,ExpressionUUID->"4a38b645-ddf8-4bb0-aa8a-0882a3b1f16a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783555586843`*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->1408175627,ExpressionUUID->"8d52c105-8fc7-4cc5-a6f6-a73525d88f99"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783555772543`*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->1937973049,ExpressionUUID->"07af8537-20d1-47b4-915f-7627c810b87b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783555950537`*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->628560494,ExpressionUUID->"2e34563c-6e9c-451d-8303-81c1b1bf8cce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783556134357`*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->744163065,ExpressionUUID->"a2f8fb32-8f18-41d4-8267-4807d14c4cd7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355632416*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->203873327,ExpressionUUID->"58f49ab5-4db6-4bde-a3cf-221aa7d81a0f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Cd5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`Cd5],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835565022*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->1262589924,ExpressionUUID->"a8a0c491-7acb-425a-a227-0efd9c653abb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilong5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilong5],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835567153*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->1062936661,ExpressionUUID->"6af02f5f-4fb8-4bee-a05c-2d0b9c92b6fb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "g5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`g5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355690752*^9},
 CellLabel->"During evaluation of In[748]:=",
 CellID->2007485681,ExpressionUUID->"1a212853-70f0-421f-afb9-5cf1fabd02e3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metricg", "]"}]], "Input",
 CellLabel->
  "In[749]:=",ExpressionUUID->"d7119c8e-ad90-40a5-b1f1-42401a833f67"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355902851*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->696262550,ExpressionUUID->"556d32c3-36b5-491e-b6e7-b6bb13aacec7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355919222*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->446285347,ExpressionUUID->"e89113b5-bc85-4751-8af8-e6b3f19e204e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783559382267`*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->1704931705,ExpressionUUID->"be2a5666-c865-4426-a2a6-d1faca75cd90"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355956839*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->912864014,ExpressionUUID->"b5c6b856-9d70-40ca-b7f9-ba9792fb3668"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978355975779*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->882309130,ExpressionUUID->"2b220c96-b130-4796-bcf7-c33f327d72c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783559945602`*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->594965375,ExpressionUUID->"861efe30-f1e7-4956-b303-2e826a856989"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783560131187`*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->899197193,ExpressionUUID->"5a318d00-8ef5-4cb4-bddc-0d1e340e21be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835603445*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->964798812,ExpressionUUID->"d11207fe-86fb-43d9-b6c1-9c603fa1f5fd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783560518837`*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->1872098434,ExpressionUUID->"95f5d7b3-4d5d-4050-92bc-0a599ba86412"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356071289*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->1575468041,ExpressionUUID->"cd4a2522-a723-4f19-bbdf-2db04b202cab"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356089328*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->417997577,ExpressionUUID->"d5f28504-8fa2-43ab-b697-fb3959cfa6e7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356107821*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->866267302,ExpressionUUID->"7c4dc29f-b783-4528-9a26-192d4b86ebc3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356126597*^9},
 CellLabel->"During evaluation of In[749]:=",
 CellID->1562901701,ExpressionUUID->"467fd898-e57b-40be-814e-331c7964dc21"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "gamma", "]"}]], "Input",
 CellLabel->
  "In[750]:=",ExpressionUUID->"987a4cfa-08d7-469f-ba84-5d4a498fdca4"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detgamma"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detgamma],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356268882*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1816402250,ExpressionUUID->"ce149a00-318b-4e8a-90bf-dd9e79185c07"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356284978*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1792573645,ExpressionUUID->"a0cb8cd3-d205-4067-ad49-d69c480d6ffc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Einstein tensor", 
   " ", $CellContext`EinsteinCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356303438*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->737167752,ExpressionUUID->"ead943a8-e838-4cc0-9284-c184c4ffb3c8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783563225403`*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->745248321,ExpressionUUID->"96b9fb0c-dc99-4be2-b40a-fedd4764c8c0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356340766*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1263901743,ExpressionUUID->"57104a82-407f-4bdd-bbe9-9cf985996b28"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783563585863`*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1635842124,ExpressionUUID->"d4342f5d-5e55-45d5-a9c2-0f2dfcd152e9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356378447*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1512929174,ExpressionUUID->"880353ba-c39b-40e0-b0fa-3d1f2775ecb2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "TFRicci tensor", 
   " ", $CellContext`TFRicciCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356397008*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1594246938,ExpressionUUID->"2770eda5-c4cb-4221-b406-f53f465589c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356415156*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->354823347,ExpressionUUID->"bab29aa1-9271-461f-85cc-90607100524f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897835643314*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->2101388957,ExpressionUUID->"40d07db7-b1ab-4ee8-8a60-378c6f99f6ac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CdS"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CdS],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356451147*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->528141737,ExpressionUUID->"97ab3683-927f-45d1-870d-b2fe483add39"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilongamma"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilongamma],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356468527*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->928260708,ExpressionUUID->"accc993f-d65e-4f32-ade3-8ca76606c645"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "gamma"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`gamma],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783564863167`*^9},
 CellLabel->"During evaluation of In[750]:=",
 CellID->1956507548,ExpressionUUID->"23294db7-290a-454e-a4c7-714917a3eb99"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M5", "]"}]], "Input",
 CellLabel->
  "In[751]:=",ExpressionUUID->"88e092ec-768a-43b1-adb7-b8e298cb539f"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M5",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM5,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM5],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978356783688*^9},
 CellLabel->"During evaluation of In[751]:=",
 CellID->752756372,ExpressionUUID->"198a6599-5309-4d2c-9166-99d244f526c9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M5"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M5],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783567985363`*^9},
 CellLabel->"During evaluation of In[751]:=",
 CellID->750179727,ExpressionUUID->"4910fd08-3656-4bc7-97e1-be5944d016c1"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.7. Flat and Cartesian metrics", "Subsection",ExpressionUUID->"5b95623d-0d2b-4cf5-925f-4956c23651ea"],

Cell["\<\
A flat metric is one without curvature. We define its Christoffel and epsilon \
objects, and then zero curvature tensors.\
\>", "Text",ExpressionUUID->"01f65013-28df-469b-bea6-ea42be0ea907"],

Cell[CellGroupData[{

Cell[TextData[{
 "We can define a flat metric using the option ",
 StyleBox["FlatMetric",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"df5a273e-86b5-4764-b99e-9b8c108815ca"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"gflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "Cdflat", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<%\>\"", ",", "\"\<\[PartialD]\>\""}], "}"}]}], ",", 
   RowBox[{"FlatMetric", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[752]:=",ExpressionUUID->"e96cd80b-2b78-4531-89cf-2a62b395d38c"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"gflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`gflat[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897836594951*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->695821353,ExpressionUUID->"dd69f961-1171-49b8-9c31-cc9036510a73"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilongflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilongflat[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783659673*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1621021956,ExpressionUUID->"223b5d2d-b436-43c8-8bbf-684dbf793001"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"Cdflat", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`Cdflat[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978365983954*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1538468650,ExpressionUUID->"317e40e1-2f18-4348-bec8-ce2d677db301"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCdflat", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCdflat[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366000811*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1339490670,ExpressionUUID->"8be054b6-da36-4c0f-b6ff-e8e809caf2fe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCdflat", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCdflat[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366018038*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->748619189,ExpressionUUID->"4877a68d-0dfc-4f93-acfd-b0ddd8c08886"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Riemann tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RiemannCdflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Riemann tensor ", 
   $CellContext`RiemannCdflat[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366036335*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->2083505397,ExpressionUUID->"034c99e4-fc4f-45ad-91b8-1af0f84f8cf8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCdflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci tensor ", 
   $CellContext`RicciCdflat[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366052713*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->606859397,ExpressionUUID->"bf9b8100-c229-4464-bed2-41c868f79477"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Ricci scalar \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCdflat", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Ricci scalar ", 
   $CellContext`RicciScalarCdflat[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366068905*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1086045776,ExpressionUUID->"8b387915-a3fc-4305-9d70-76a36489b0e5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCdflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing Einstein tensor ", 
   $CellContext`EinsteinCdflat[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366085017*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1707841379,ExpressionUUID->"ae308982-aeea-4397-8dd5-94818b3111e5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCdflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCdflat[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366102551*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->423631051,ExpressionUUID->"9606baa0-fe39-49d5-8738-486377542418"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCdflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing TFRicci tensor ", 
   $CellContext`TFRicciCdflat[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366118661*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1058784352,ExpressionUUID->"6529fee7-41d2-4692-9ec6-983db9c89240"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Kretschmann scalar \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"KretschmannCdflat", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>",
    "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing Kretschmann scalar ", 
   $CellContext`KretschmannCdflat[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366201367*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->939105022,ExpressionUUID->"dfbf130b-069b-4f21-90dc-70416a375042"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detgflat", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detgflat[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366219*^9},
 CellLabel->"During evaluation of In[752]:=",
 CellID->1680782951,ExpressionUUID->"56a4b494-2cd2-4d7f-84e0-49963fef97c7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "gflat", "]"}]], "Input",
 CellLabel->
  "In[753]:=",ExpressionUUID->"81cbfef4-405f-4363-ae3c-451e7c4e4eef"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detgflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detgflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366847475*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->919164819,ExpressionUUID->"07526731-701b-4a09-a715-bd4c1f9974d7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ChristoffelCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366870678*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1553240062,ExpressionUUID->"b82b37dc-b6bb-4133-be71-c16b7f695fc1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Einstein tensor", 
   " ", $CellContext`EinsteinCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366893149*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->494509102,ExpressionUUID->"195097d8-5ab1-4de1-9e1a-c5b32c2e1dc2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "KretschmannCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366915758*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->452763424,ExpressionUUID->"34f733f0-fd4b-40cf-8c84-0878e7e4600b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci tensor", 
   " ", $CellContext`RicciCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366938258*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1060305847,ExpressionUUID->"88c2e5ef-3c3f-41e8-bafa-dcea7307b6cc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "RicciScalarCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366961143*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->741067723,ExpressionUUID->"eb8046a4-7f95-4342-b1e1-c1115c9c1457"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978366986248*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1867551508,ExpressionUUID->"58a8f942-c946-4c3e-91f2-33b7a33b9612"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "TFRicci tensor", 
   " ", $CellContext`TFRicciCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783670087013`*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->317210404,ExpressionUUID->"de52af17-14b9-43b6-85bc-14b3147fddff"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897836703133*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1300848174,ExpressionUUID->"ec3bdcc6-9c58-488e-8110-b15b9b27f59c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367054484*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1971733958,ExpressionUUID->"06a3cc89-b9c1-4534-a96c-c85a25b798fd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Cdflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`Cdflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367075266*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1112475249,ExpressionUUID->"b826d113-9109-444e-8480-5356a269f5fc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilongflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilongflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783670965567`*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1378879161,ExpressionUUID->"ef7eab35-1291-4edc-b04e-07441d66b181"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "gflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`gflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367118034*^9},
 CellLabel->"During evaluation of In[753]:=",
 CellID->1326024150,ExpressionUUID->"9e8c17bf-61a3-4108-ac02-16e952cdab0e"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "A delicate issue is that of a Cartesian metric. The concept of a flat \
metric is covariantly defined and therefore perfectly fits in the current \
structure of ",
 StyleBox["xTensor`", "DisplayMath"],
 ". However we say that a metric in a given basis of vectors is Cartesian if \
all metric components in that basis are constant (not necessarily of unit \
modulus). For a flat metric coordinate systems of that kind always exist. We \
introduce the possibility of associating ",
 StyleBox["PD",
  FontFamily->"Courier"],
 " as the covariant derivative of a flat metric. That means that ",
 StyleBox["PD",
  FontFamily->"Courier"],
 " is the partial derivative of one of those coordinate systems with respect \
to that metric and hence ",
 StyleBox["PD",
  FontFamily->"Courier"],
 " is not general anymore. This is an ugly trick for a problem which is only \
properly solved in the twin package ",
 StyleBox["xCoba`",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"232c1865-fe41-40e4-985a-c4bb7f84e3f4"],

Cell[CellGroupData[{

Cell["\<\
We redefine the flat metric. In this case not even the Christoffel symbol is \
defined. By definition it is zero.\
\>", "MathCaption",ExpressionUUID->"b8a6c7e0-6359-409f-8edb-723a9175517b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"gflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "PD", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<,\>\"", ",", "\"\<\[PartialD]\>\""}], "}"}]}], ",", 
   RowBox[{"FlatMetric", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[754]:=",ExpressionUUID->"e15f39cb-b782-481a-a5fc-46da894cee2e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"gflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`gflat[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897836743032*^9},
 CellLabel->"During evaluation of In[754]:=",
 CellID->2050084131,ExpressionUUID->"d53f452d-8dd1-4886-b349-f6b65bdc3c55"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilongflat", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilongflat[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367455524*^9},
 CellLabel->"During evaluation of In[754]:=",
 CellID->1536357249,ExpressionUUID->"ad386c9d-4851-4264-b19b-6860bd36de56"],

Cell[BoxData["\<\"** DefMetric: Associating fiducial flat derivative PD to \
metric.\"\>"], "Print",
 CellChangeTimes->{3.728978367475741*^9},
 CellLabel->"During evaluation of In[754]:=",
 CellID->212912415,ExpressionUUID->"24cd41c1-ccab-49fa-9373-d94329599baa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detgflat", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detgflat[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367496611*^9},
 CellLabel->"During evaluation of In[754]:=",
 CellID->711203003,ExpressionUUID->"b679ba94-39c2-4177-9a17-a96269a5391a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"gflat", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[755]:=",ExpressionUUID->"755cf158-3089-4129-9be9-e5c838f7afce"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289783676835747`*^9},
 CellLabel->"Out[755]=",
 CellID->294892219,ExpressionUUID->"71deb919-6c0c-4d81-a313-69f79cfda858"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "gflat", "]"}]], "Input",
 CellLabel->
  "In[756]:=",ExpressionUUID->"3ac1d263-d2bd-4921-988a-fc0d545f0b7a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detgflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detgflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783678351297`*^9},
 CellLabel->"During evaluation of In[756]:=",
 CellID->2008751104,ExpressionUUID->"2ee8446f-8154-488b-b4b9-2f8c1cd06d1d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilongflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilongflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367859516*^9},
 CellLabel->"During evaluation of In[756]:=",
 CellID->489390441,ExpressionUUID->"5e73aaa5-c36f-4de9-90b9-ed50867ba2c4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "gflat"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`gflat],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978367880127*^9},
 CellLabel->"During evaluation of In[756]:=",
 CellID->322773877,ExpressionUUID->"d045421b-ba5c-40bd-891a-2844a54e9f6f"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.8. Induced metrics", "Subsection",ExpressionUUID->"cfa3ecda-cd9c-4be7-a543-398c4e19470c"],

Cell[TextData[{
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " has a simple way to define the induced metric obtained from projection of \
another metric along hypersurface-orthogonal vector fields, as it is \
typically done in the ADM 3+1 decomposition of the spacetime metric:"
}], "Text",ExpressionUUID->"509ca7d7-5855-454b-bba4-3e1b59338048"],

Cell[CellGroupData[{

Cell["We first define the global ambient metric:", "MathCaption",ExpressionUUID->"3036de7d-3bdf-42c9-968c-1bb08cb30151"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"DefMetric", "[", 
   RowBox[{
    RowBox[{"-", "1"}], ",", 
    RowBox[{"metricg", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
    RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}], ";"}]], "Input",
 CellLabel->
  "In[757]:=",ExpressionUUID->"55a07874-51d5-4766-a57a-b5b013f7a261"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metricg[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783732480907`*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->701636648,ExpressionUUID->"be3a7142-4223-417f-a8c2-6b759f11f425"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetricg[-$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373265662*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->574437259,ExpressionUUID->"3791b567-95d7-4dfa-93ce-00a6a66cbcbd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"CD", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`CD[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783732832537`*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->1628018748,ExpressionUUID->"951e5298-c5d5-4062-8ae6-a91ebdfeb29c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TorsionCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`TorsionCD[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373301264*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->1021951209,ExpressionUUID->"b03a5d9c-95e1-4f0d-8609-87b12e2cfb57"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ChristoffelCD", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`ChristoffelCD[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373319562*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->359760957,ExpressionUUID->"c6ca4e71-40ec-4119-92cf-28ebfda4ae89"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783733372498`*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->1335937994,ExpressionUUID->"b10676cb-0be5-4e8f-9dc5-0f6344344a5b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373355151*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->1168335773,ExpressionUUID->"33b78ac1-6fb0-45e5-b808-ef03edd5ab41"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728978373370739*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->1776352290,ExpressionUUID->"05d688c3-790c-466e-b702-1d8fca1cc697"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783733874683`*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->835277436,ExpressionUUID->"bf67f6ac-de7c-4689-aa53-64a34dfb4f48"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728978373404394*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->407266413,ExpressionUUID->"51ece204-fb72-469d-966a-a4b6edd4661a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"EinsteinCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Einstein tensor ", 
   $CellContext`EinsteinCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373421049*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->896220302,ExpressionUUID->"4eac071e-5a36-41b5-a415-618f92853c05"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"WeylCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`WeylCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783734384403`*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->324315656,ExpressionUUID->"a18d7c21-a7fe-4fce-8ca4-f9931496b219"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric TFRicci tensor ", 
   $CellContext`TFRicciCD[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373454863*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->459734306,ExpressionUUID->"bf9aec6d-d45b-4c80-a2eb-c15a343c0a75"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"KretschmannCD", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`KretschmannCD[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373471023*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->108319036,ExpressionUUID->"2075aa61-ea78-456e-a6de-c0cab61a1e7d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783734884768`*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->1599389043,ExpressionUUID->"bc9655d3-44ec-4fc2-843e-47981d248207"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897837350511*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->878836202,ExpressionUUID->"615194d3-a4ca-4dde-9673-21d70410eac8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "3"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373521574*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->207522948,ExpressionUUID->"e3dc52d4-a6dd-46a6-bc8b-a4b76f4a8dcf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetricg", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetricg[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978373538899*^9},
 CellLabel->"During evaluation of In[757]:=",
 CellID->324744308,ExpressionUUID->"71caa11f-826b-4c95-8a3a-8706efb0b687"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define the timelike normal to the slices. The use of ",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 " and ",
 StyleBox["AutomaticRules",
  FontFamily->"Courier"],
 " will be explained later."
}], "MathCaption",ExpressionUUID->"7bf0c0d2-7414-41bf-97a3-34e74a462786"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"n", "[", "a", "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[758]:=",ExpressionUUID->"43fe3905-ffa5-446f-8867-a5d6adf16845"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"n", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`n[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978374201317*^9},
 CellLabel->"During evaluation of In[758]:=",
 CellID->1221637529,ExpressionUUID->"96ea8474-5ba7-4b39-8186-e8defe63287c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AutomaticRules", "[", 
  RowBox[{"n", ",", 
   RowBox[{"MakeRule", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{"n", "[", 
        RowBox[{"-", "a"}], "]"}]}], ",", 
      RowBox[{"-", "1"}]}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[759]:=",ExpressionUUID->"e662822f-379f-436a-9c3c-25a8280a85f9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"   Rules \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", "1", "}"}], 
   "\[InvisibleSpace]", "\<\" have been declared as UpValues for \"\>", 
   "\[InvisibleSpace]", "n", "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm["   Rules ", 
   Shallow[{1}], " have been declared as UpValues for ", $CellContext`n, "."],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728978374500704*^9},
 CellLabel->"During evaluation of In[759]:=",
 CellID->2118284583,ExpressionUUID->"d525bcd9-44eb-47b6-9aa8-8ccd6657e2b7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AutomaticRules", "[", 
  RowBox[{"n", ",", 
   RowBox[{"MakeRule", "[", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"metricg", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "b"}]}], "]"}], 
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{"n", "[", "b", "]"}]}], ",", 
      RowBox[{"-", "1"}]}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[760]:=",ExpressionUUID->"1366ddf3-0160-45ab-8424-cbd943f94780"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"   Rules \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], 
   "\[InvisibleSpace]", "\<\" have been declared as UpValues for \"\>", 
   "\[InvisibleSpace]", "n", "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm["   Rules ", 
   Shallow[{1, 2}], " have been declared as UpValues for ", $CellContext`n, 
   "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783747725487`*^9},
 CellLabel->"During evaluation of In[760]:=",
 CellID->1025086515,ExpressionUUID->"96dd61e3-7268-4002-85a7-9ce87e7471ad"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We choose the standard conventions in Numerical Relativity:", \
"MathCaption",ExpressionUUID->"39814496-8bf1-40a9-bcc1-c65e26f46537"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"$ExtrinsicKSign", "=", 
   RowBox[{"-", "1"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$AccelerationSign", "=", 
   RowBox[{"-", "1"}]}], ";"}]}], "Input",
 CellLabel->
  "In[761]:=",ExpressionUUID->"cde3dd99-4f7f-4fcc-b3aa-dbc9ddd04abe"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we can define the induced metric associated to the pair {ambient metric, \
orthogonal vector field}. Note that even though there is another metric on \
the tangent bundle, an induced metric is not considered to be a frozen metric.\
\>", "MathCaption",ExpressionUUID->"42fd3cba-f88d-4b02-a089-afc7fab41157"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMetric", "[", 
  RowBox[{"1", ",", 
   RowBox[{"metrich", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "cd", ",", 
   RowBox[{"SymbolOfCovD", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<|\>\"", ",", "\"\<D\>\""}], "}"}]}], ",", 
   RowBox[{"InducedFrom", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"metricg", ",", "n"}], "}"}]}], ",", 
   RowBox[{"PrintAs", "->", "\"\<h\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[763]:=",ExpressionUUID->"7830ab30-549a-48f1-9d07-fb022390e44f"],

Cell[BoxData[
 TemplateBox[{
  "DefMetric", "old", 
   "\"There are already metrics \\!\\(\\*RowBox[{\\\"{\\\", \\\"metricg\\\", \
\\\"}\\\"}]\\) in vbundle \\!\\(\\*InterpretationBox[StyleBox[\\\"\
\[DoubleStruckCapitalT]M3\\\", Rule[AutoSpacing, False], Rule[ShowAutoStyles, \
False]], TangentM3, Rule[Editable, False]]\\).\"", 2, 763, 91, 
   19253573948027831207, "Local", "xAct`xTensor`DefMetric"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289783753583307`*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1515241371,ExpressionUUID->"997f40cd-0972-400d-bd59-4d6c257348e3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"metrich", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric metric tensor ", 
   $CellContext`metrich[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375384367*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1800636271,ExpressionUUID->"d13f05ed-e699-45ec-af90-ea45babb45d5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"epsilonmetrich", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "antisymmetric tensor ", 
   $CellContext`epsilonmetrich[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375396715*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->954253173,ExpressionUUID->"5d3ab98d-cfd4-4606-b885-37d08b50b2d9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefCovD", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative \"\>", "\[InvisibleSpace]", 
   
   RowBox[{"cd", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefCovD, ": Defining ", "covariant derivative ", 
   $CellContext`cd[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375415061*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->725202815,ExpressionUUID->"2f1367b8-b441-4654-9bbd-6d4be431ee4e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing torsion tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Torsioncd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing torsion tensor ", 
   $CellContext`Torsioncd[$CellContext`a, -$CellContext`b, -$CellContext`c], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375432947*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1514014980,ExpressionUUID->"d3edb312-cb59-47bc-aa7c-6abde4cd9ef9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Christoffelcd", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "symmetric Christoffel tensor ", 
   $CellContext`Christoffelcd[$CellContext`a, -$CellContext`b, \
-$CellContext`c], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783754515247`*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->516035557,ExpressionUUID->"1ae48d4e-55f7-45d7-8186-d1c25bf557f4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Riemanncd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Riemann tensor ", 
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375470467*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1470639913,ExpressionUUID->"08bd624f-ea29-4f99-acfc-43443aaffb24"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Riccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "symmetric Ricci tensor ", 
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783754892693`*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1655724468,ExpressionUUID->"a46e5085-bbb9-44d6-b6a3-264b529d14e9"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Riemann automatically replaced \
by Ricci.\"\>"], "Print",
 CellChangeTimes->{3.728978375507246*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->668035850,ExpressionUUID->"0c9c73cc-110c-42b6-a773-a3aa9bb95e60"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"RicciScalarcd", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Ricci scalar ", 
   $CellContext`RicciScalarcd[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375525487*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->634024434,ExpressionUUID->"9f156de3-a51a-4307-957d-698d48f4fb59"],

Cell[BoxData["\<\"** DefCovD:  Contractions of Ricci automatically replaced \
by RicciScalar.\"\>"], "Print",
 CellChangeTimes->{3.728978375544207*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->234805824,ExpressionUUID->"9d1f9650-3199-4855-ae9e-7264ffc3e77e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Einstein tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"Einsteincd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "vanishing Einstein tensor ", 
   $CellContext`Einsteincd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375562099*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1454024429,ExpressionUUID->"a0c611e9-8b4c-4476-b1d5-75ae089af348"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing Weyl tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"Weylcd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing Weyl tensor ", 
   $CellContext`Weylcd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375580208*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1572158990,ExpressionUUID->"cdc032b1-7b9a-4260-b31c-dbcbaa072171"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vanishing TFRicci tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"TFRiccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "vanishing TFRicci tensor ", 
   $CellContext`TFRiccicd[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375597107*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->637352242,ExpressionUUID->"f770e74d-1591-4b4b-b32d-384beef3942e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar \"\>", "\[InvisibleSpace]", 
   RowBox[{"Kretschmanncd", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "Kretschmann scalar ", 
   $CellContext`Kretschmanncd[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375614531*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->330700058,ExpressionUUID->"a18e410d-b1a9-423a-858c-5926d2181156"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RiemannToWeylRules for dim \"\>", 
   "\[InvisibleSpace]", "2"}],
  SequenceForm["** DefCovD:  Computing RiemannToWeylRules for dim ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783756319447`*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->388196370,ExpressionUUID->"d2ea5b6f-1ff1-4c32-b04b-b3d7d2a992ae"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToTFRicci for dim \"\>", 
   "\[InvisibleSpace]", "2"}],
  SequenceForm["** DefCovD:  Computing RicciToTFRicci for dim ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375649477*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->361071004,ExpressionUUID->"4caae49f-a874-4d33-b01f-a207d00dd3b4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** DefCovD:  Computing RicciToEinsteinRules for dim \"\>", 
   "\[InvisibleSpace]", "2"}],
  SequenceForm["** DefCovD:  Computing RicciToEinsteinRules for dim ", 2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375667038*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1879061453,ExpressionUUID->"0583c9e3-3031-4180-80c9-9cb855e66fa1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density \"\>", "\[InvisibleSpace]", 
   RowBox[{"Detmetrich", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Determinant.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "weight +2 density ", 
   $CellContext`Detmetrich[], ". ", "Determinant."],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289783756843863`*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->491327789,ExpressionUUID->"7b6c84a3-180d-44fc-9962-fc16f361ceb7"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"extrinsic curvature tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"ExtrinsicKmetrich", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Associated to vector n\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "extrinsic curvature tensor ", 
   $CellContext`ExtrinsicKmetrich[$CellContext`a, $CellContext`b], ". ", 
   "Associated to vector n"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375701749*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->671527008,ExpressionUUID->"11fcf667-47a8-4e92-ad4a-ec00f31d08c6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"acceleration vector \"\>", "\[InvisibleSpace]", 
   RowBox[{"Accelerationn", "[", "a", "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Associated to vector n\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", "acceleration vector ", 
   $CellContext`Accelerationn[$CellContext`a], ". ", "Associated to vector n"],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375720134*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->1327858471,ExpressionUUID->"122eef35-238a-4a8b-b2e6-50299b244071"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefInertHead", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"projector inert-head \"\>", "\[InvisibleSpace]", 
   "Projectormetrich", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefInertHead, ": Defining ", 
   "projector inert-head ", $CellContext`Projectormetrich, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978375737219*^9},
 CellLabel->"During evaluation of In[763]:=",
 CellID->667262473,ExpressionUUID->"d1b0fa12-84d3-4a97-8ee6-1fb771e5118e"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
For a hypersurface orthogonal vector we can express the acceleration vector \
in terms of a lapse function:\
\>", "MathCaption",ExpressionUUID->"a38aa410-eff7-4e19-889b-8715a2111af5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"lapse", "[", "]"}], ",", "M3", ",", 
   RowBox[{"PrintAs", "->", "\"\<\[Alpha]\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[764]:=",ExpressionUUID->"e54b9bbc-9b58-4894-870f-302255832284"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"lapse", "[", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`lapse[], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978376217369*^9},
 CellLabel->"During evaluation of In[764]:=",
 CellID->520779077,ExpressionUUID->"6980c9eb-3079-4c18-8e3a-495e2b41d304"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"LapseRule", "=", 
  RowBox[{
   RowBox[{"Accelerationn", "[", "a_", "]"}], "\[Rule]", 
   RowBox[{
    RowBox[{
     RowBox[{"cd", "[", "a", "]"}], "[", 
     RowBox[{"lapse", "[", "]"}], "]"}], "/", 
    RowBox[{"lapse", "[", "]"}]}]}]}]], "Input",
 CellLabel->
  "In[765]:=",ExpressionUUID->"b374831a-1330-440a-b39e-6b0d0c630ff6"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"An", 
       StyleBox[GridBox[{
          {
           UnderscriptBox["a", "_"]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Accelerationn[
    Pattern[$CellContext`a, 
     Blank[]]],
   Editable->False], "\[Rule]", 
  FractionBox[
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["D", "a"], "\[Alpha]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`cd[$CellContext`a][
     $CellContext`lapse[]],
    Editable->False], 
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`lapse[],
    Editable->False]]}]], "Output",
 CellChangeTimes->{3.728978376434761*^9},
 CellLabel->"Out[765]=",
 CellID->1738085614,ExpressionUUID->"cc9a1c31-fd8e-4a5e-a622-3f65eb603d71"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We define a generic vector field and a spatial vector field:", \
"MathCaption",ExpressionUUID->"ea052db3-9d0c-404b-9d69-36e724f2290c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"V", "[", 
    RowBox[{"-", "a"}], "]"}], ",", "M3"}], "]"}]], "Input",
 CellLabel->
  "In[766]:=",ExpressionUUID->"385545ed-2d4d-420a-b8e8-63e43efa566d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"V", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`V[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978376596573*^9},
 CellLabel->"During evaluation of In[766]:=",
 CellID->1932348004,ExpressionUUID->"e88f489c-e860-4d5e-94f9-26d6a0801d40"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"-", "a"}], "]"}], ",", "M3", ",", 
   RowBox[{"OrthogonalTo", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"n", "[", "a", "]"}], "}"}]}], ",", 
   RowBox[{"ProjectedWith", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"metrich", "[", 
      RowBox[{"a", ",", 
       RowBox[{"-", "b"}]}], "]"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[767]:=",ExpressionUUID->"024069b5-f93c-4141-a4a1-2aa04f4fcafb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"W", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`W[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978376918994*^9},
 CellLabel->"During evaluation of In[767]:=",
 CellID->577709647,ExpressionUUID->"0f300746-b9d1-4f4b-8ba2-9fde5962dd73"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The first operation we need is the decomposition of any expression along \
the normal vector. This is given by the command ",
 StyleBox["InducedDecomposition",
  FontFamily->"Courier"],
 ", which requires as second argument the pair {projector, normal}. The \
decomposition process does not change the input; it simply rewrites it in a \
different way:"
}], "Text",ExpressionUUID->"e25d19ea-cc67-4e0d-b0f4-833909cb29c6"],

Cell[CellGroupData[{

Cell[TextData[{
 "Any expression can be decomposed in projected and orthogonal parts with \
respect to the vector field. Note the special output of the head ",
 StyleBox["Projectormetrich",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"b9ac5637-0f38-43d5-b6de-6320ec1617d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InducedDecomposition", "[", 
  RowBox[{
   RowBox[{"W", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[768]:=",ExpressionUUID->"34c64ef0-1bfc-4180-aa2c-0abcb3a34c54"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"W", 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`W[-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978377403737*^9},
 CellLabel->"Out[768]=",
 CellID->792469042,ExpressionUUID->"6500ec17-4539-4d71-a29a-e1a36e9d5ae2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InducedDecomposition", "[", 
  RowBox[{
   RowBox[{"V", "[", 
    RowBox[{"-", "a"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[769]:=",ExpressionUUID->"43764773-7ded-47ab-90b2-f636e3bb07d3"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"V", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V[-$CellContext`a],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`V[-$CellContext`a]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["(",
       FontColor->RGBColor[1, 0, 0]], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`a],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"V", 
            StyleBox[GridBox[{
               {" "},
               {"a"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`V[-$CellContext`a],
        Editable->False]}], 
      StyleBox[")",
       FontColor->RGBColor[1, 0, 0]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Scalar[$CellContext`n[$CellContext`a] \
$CellContext`V[-$CellContext`a]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978377667103*^9},
 CellLabel->"Out[769]=",
 CellID->333610200,ExpressionUUID->"849cb0d2-3553-41e9-92b8-168d717401aa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InducedDecomposition", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"7", 
     RowBox[{"V", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"V", "[", 
      RowBox[{"-", "b"}], "]"}]}], "+", 
    RowBox[{
     RowBox[{"W", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"W", "[", 
      RowBox[{"-", "b"}], "]"}]}]}], ",", 
   RowBox[{"{", 
    RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[770]:=",ExpressionUUID->"1976425a-62f5-447e-be47-4f905ddaad80"],

Cell[BoxData[
 RowBox[{
  RowBox[{"7", " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        UnderscriptBox["P", "h"], "[", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"V", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`V[-$CellContext`a],
         Editable->False], "]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Projectormetrich[
       $CellContext`V[-$CellContext`a]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         StyleBox["(",
          FontColor->RGBColor[1, 0, 0]], 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {"a"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`n[$CellContext`a],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"V", 
               StyleBox[GridBox[{
                  {" "},
                  {"a"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`V[-$CellContext`a],
           Editable->False]}], 
         StyleBox[")",
          FontColor->RGBColor[1, 0, 0]]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Scalar[$CellContext`n[$CellContext`a] \
$CellContext`V[-$CellContext`a]],
       Editable->False]}]}], ")"}], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[
       RowBox[{
        UnderscriptBox["P", "h"], "[", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"V", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`V[-$CellContext`b],
         Editable->False], "]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Projectormetrich[
       $CellContext`V[-$CellContext`b]],
      Editable->False], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         StyleBox["(",
          FontColor->RGBColor[1, 0, 0]], 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {"a"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`n[$CellContext`a],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"V", 
               StyleBox[GridBox[{
                  {" "},
                  {"a"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`V[-$CellContext`a],
           Editable->False]}], 
         StyleBox[")",
          FontColor->RGBColor[1, 0, 0]]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Scalar[$CellContext`n[$CellContext`a] \
$CellContext`V[-$CellContext`a]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[-$CellContext`b],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289783782547493`*^9},
 CellLabel->"Out[770]=",
 CellID->2054312730,ExpressionUUID->"a21bafb3-bc0a-4036-b6a0-275d9779bbaa"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InducedDecomposition", "[", 
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[771]:=",ExpressionUUID->"7c27b8c0-d85d-4ab1-8da9-cac4a6b6d7cb"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"a", "b", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "c", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "e", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`e, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`e, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "e", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`e, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`b, -$CellContext`e, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "f", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`f, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`f, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"g"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`g],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`f, -$CellContext`g],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[(($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`n[$CellContext`g]) \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, -$CellContext`f, \
-$CellContext`g]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "b", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "b", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`e, \
-$CellContext`b, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "b", "f", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, \
-$CellContext`f, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`e, \
-$CellContext`b, -$CellContext`f, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"g"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`g],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "b", "f", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, \
-$CellContext`f, -$CellContext`g],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[(($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`n[$CellContext`g]) \
$CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, -$CellContext`f, \
-$CellContext`g]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "f", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`e, \
-$CellContext`f, -$CellContext`c, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"g"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`g],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "f", "c", "g"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, \
-$CellContext`c, -$CellContext`g],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[(($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`n[$CellContext`g]) \
$CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, -$CellContext`c, \
-$CellContext`g]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"g"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`g],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"e", "f", "g", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, \
-$CellContext`g, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[(($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`n[$CellContext`g]) \
$CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, -$CellContext`g, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["(",
       FontColor->RGBColor[1, 0, 0]], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`a],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`b],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"d"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], 
      StyleBox[")",
       FontColor->RGBColor[1, 0, 0]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Scalar[((($CellContext`n[$CellContext`a] \
$CellContext`n[$CellContext`b]) $CellContext`n[$CellContext`c]) \
$CellContext`n[$CellContext`d]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`b, -$CellContext`c, -$CellContext`d]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978378483261*^9},
 CellLabel->"Out[771]=",
 CellID->1739460291,ExpressionUUID->"3669b57e-87bc-4de0-8103-5c078dd7016e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ToCanonical", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[772]:=",ExpressionUUID->"23f5084e-04e8-4f25-a06c-0e6cfb5f18ca"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"a", "b", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "c", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "d", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
-$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978379026804*^9},
 CellLabel->"Out[772]=",
 CellID->1414224226,ExpressionUUID->"b33d1030-2cd2-4422-8ca3-bcf8d71f5958"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Projection along the normal vector is automatic:", "MathCaption",ExpressionUUID->"3fb159f4-3f59-4f20-90a6-e22c158f9810"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"n", "[", "d", "]"}], "%"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[773]:=",ExpressionUUID->"dfaf6e7e-855c-40dc-9ac8-2d122c3d6198"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"a", "b", "c", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`d] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"d"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "d", "c", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, \
-$CellContext`c, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`d] \
$CellContext`n[$CellContext`e]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`d, -$CellContext`c, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"d"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "d", "c", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`d, \
-$CellContext`c, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`d] \
$CellContext`n[$CellContext`e]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`d, -$CellContext`c, -$CellContext`e]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289783791526957`*^9},
 CellLabel->"Out[773]=",
 CellID->517561386,ExpressionUUID->"98fa84a9-a110-4f62-afaa-3729a4b1b8d5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"n", "[", "c", "]"}], "%"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[774]:=",ExpressionUUID->"59834dbd-99d8-4da7-8ed0-176fc556a5fe"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978379283415*^9},
 CellLabel->"Out[774]=",
 CellID->1345444152,ExpressionUUID->"92964102-44f4-456a-892c-d5c61649071c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"n", "[", "b", "]"}], "%%"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[775]:=",ExpressionUUID->"861edced-869e-490b-b3ab-c343d4eaa08f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"a", "b", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
      Editable->False]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[($CellContext`n[$CellContext`b] \
$CellContext`n[$CellContext`d]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`b, -$CellContext`c, -$CellContext`d]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978379417858*^9},
 CellLabel->"Out[775]=",
 CellID->2024543401,ExpressionUUID->"09d5cac1-b4c0-4c03-a3fa-98b23084b657"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The head ",
 StyleBox["Projectormetrich",
  FontFamily->"Courier"],
 " is the formal projector, but we now need a command which actually performs \
the projection. This is done with ",
 StyleBox["ProjectWith[metrich]",
  FontFamily->"Courier"],
 ". Do not confuse them. They have been constructed as a matched pair, and it \
is very useful to go from the former to the latter."
}], "Text",ExpressionUUID->"5181de5e-f4d7-4d66-98db-1d1f5c2c3aab"],

Cell[CellGroupData[{

Cell[TextData[{
 "As long as we need to work only with the projected part of ",
 StyleBox["RicciCD",
  FontFamily->"Courier"],
 ", we can use this object:"
}], "MathCaption",ExpressionUUID->"6d241c2a-5fec-4b1e-a3ae-f8dafb993c37"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Projectormetrich", "[", 
  RowBox[{"RicciCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[776]:=",ExpressionUUID->"56532404-b8e1-4c21-86bb-c930342ab471"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978379568306*^9},
 CellLabel->"Out[776]=",
 CellID->1770037539,ExpressionUUID->"42f06f82-0c63-4653-b016-d6befec251fd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However, if we need to perform the projection, we can do this:\
\>", "MathCaption",ExpressionUUID->"020ed742-1c27-4ba3-8717-c58ba81f2182"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Projectormetrich", "\[Rule]", 
   RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}]], "Input",
 CellLabel->
  "In[777]:=",ExpressionUUID->"b5dded42-5dde-42c6-923c-dad0ed6a7d59"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "c"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`a, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "d"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`b, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978379730627*^9},
 CellLabel->"Out[777]=",
 CellID->2077439729,ExpressionUUID->"6ab27f5f-a299-4bfd-9856-07954b36d268"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Finally, we can expand the projectors as", "MathCaption",ExpressionUUID->"b08d3cba-a484-4f1e-b52d-74afaccbb1a0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ProjectorToMetric"}]], "Input",
 CellLabel->
  "In[778]:=",ExpressionUUID->"d871482d-026c-4637-8f69-70532d17bff7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "c"},
            {"a", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`a, $CellContext`c],
     Editable->False], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`c],
      Editable->False]}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`b, $CellContext`d],
     Editable->False], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`d],
      Editable->False]}]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978380131299*^9},
 CellLabel->"Out[778]=",
 CellID->1740392377,ExpressionUUID->"c808fa89-40e0-43c1-8f0a-facacaa43c46"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[779]:=",ExpressionUUID->"6c7a4ffb-8291-4427-bd60-f1a85bf3e8a2"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289783803224382`*^9},
 CellLabel->"Out[779]=",
 CellID->2047408099,ExpressionUUID->"0515b4e6-d638-4ba1-9bab-58b53baf2afd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The projection process can be easily undone using ",
 StyleBox["InducedDecomposition",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"e21d621c-e1f0-4159-8fb8-334851df42a1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InducedDecomposition", "[", 
   RowBox[{"%", ",", 
    RowBox[{"{", 
     RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[780]:=",ExpressionUUID->"051426ca-e44e-4f37-be5e-d204a8553c96"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978380470448*^9},
 CellLabel->"Out[780]=",
 CellID->1657444012,ExpressionUUID->"57a6315f-c43d-4eb4-849d-10f66b2080bc"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["InducedDecomposition",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "InducedDecomposition.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t", "DisplayMath"],
 StyleBox["Decompose an expression along the normal vector", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["Projector",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Projector.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Formal projector head", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["ProjectWith",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ProjectWith.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Actual projection operator, using the projected metric", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["ProjectorToMetric",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ProjectorToMetric.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Convert projected metric into sum of metric and projector onto the \
normal\n", "Text"],
 StyleBox[ButtonBox["MetricToProjector",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MetricToProjector.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Convert metric into sum of projected metric and projector onto the \
normal", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.728978383102869*^9, 
  3.728978385119171*^9}},ExpressionUUID->"d44273b4-46a0-4111-8cbb-\
f40ff8939202"],

Cell["Projection operators.", "Text",ExpressionUUID->"06e1d5f6-2c74-4cd5-9ad0-e9db4972a015"],

Cell[TextData[{
 "The second issue is the manipulation of derivatives. The projected metric \
has its own Levi-Civita connection, which is a proper covariant derivative \
only when acting on objects on the projected manifold. The relation between \
the original and the new covariant derivatives is given by  a kind of \
\"Christoffel\" encoded in two objects: the extrinsic curvature ",
 StyleBox["ExtrinsicK",
  FontFamily->"Courier"],
 " tensor and the ",
 StyleBox["Acceleration",
  FontFamily->"Courier"],
 " vector:"
}], "Text",ExpressionUUID->"d32d75b5-f77d-4c2f-a865-7fbc5ed87b84"],

Cell[CellGroupData[{

Cell["\<\
This is the derivative of the normal (note the sign convention):\
\>", "MathCaption",ExpressionUUID->"a7c917e1-b407-4bb0-b408-1a634f3c25aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"n", "[", "b", "]"}], "]"}]], "Input",
 CellLabel->
  "In[781]:=",ExpressionUUID->"42cb6533-235f-4fe3-b675-b327c749d475"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"n", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`a][
   $CellContext`n[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978380609627*^9},
 CellLabel->"Out[781]=",
 CellID->1651124447,ExpressionUUID->"d8a57b05-44af-4ce0-b70b-c35d27f4342a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "GradNormalToExtrinsicK"}]], "Input",
 CellLabel->
  "In[782]:=",ExpressionUUID->"0f08fb59-b4e1-4cc1-a035-245774613750"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978380713962*^9},
 CellLabel->"Out[782]=",
 CellID->664894195,ExpressionUUID->"e0ea26de-fb09-4071-a352-dea565c225fc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is useful to remove derivatives in favour of these two tensors whenever \
possible. For example for \"spatial\" vectors:\
\>", "MathCaption",ExpressionUUID->"4bd7057c-ec59-40e3-8e29-eb6b1a81b4b5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "[", 
   RowBox[{"-", "a"}], "]"}], 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "b"}], "]"}], "[", 
   RowBox[{"W", "[", "a", "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[783]:=",ExpressionUUID->"57b88724-41c8-4910-8f83-4f93b76d5b22"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"n", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`n[-$CellContext`a],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`b][
     $CellContext`W[$CellContext`a]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.7289783808572063`*^9},
 CellLabel->"Out[783]=",
 CellID->1519320024,ExpressionUUID->"0ca5d11b-0838-4e8c-b11a-94735196397d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[784]:=",ExpressionUUID->"7dbffcac-1fb1-4378-847f-368ec11d481d"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978380986813*^9},
 CellLabel->"Out[784]=",
 CellID->1631080538,ExpressionUUID->"58370695-6755-4b68-ada6-b29652d9ae6b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The opposite command also exists:", "MathCaption",ExpressionUUID->"12595416-9ed0-42ad-8a5f-388f49eee419"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ExtrinsicKmetrich", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "b"}]}], "]"}]], "Input",
 CellLabel->
  "In[785]:=",ExpressionUUID->"0159f067-e530-4559-82ff-3bfb13647309"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"Kh", 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978381125216*^9},
 CellLabel->"Out[785]=",
 CellID->941848211,ExpressionUUID->"460ebd2a-2960-469c-8f16-06c56394b1a7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ExtrinsicKToGradNormal"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[786]:=",ExpressionUUID->"27e1cf07-c6d2-4a13-9626-95d2a9200886"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`b],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`n[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978381270727*^9},
 CellLabel->"Out[786]=",
 CellID->540920240,ExpressionUUID->"e3462d92-79de-4bae-9678-0ba597fee1ff"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us now project the derivative on a \"spatial\" vector:", \
"MathCaption",ExpressionUUID->"7f56dd22-9685-466f-bfed-dd8346cef492"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"W", "[", 
   RowBox[{"-", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[787]:=",ExpressionUUID->"eaa2b14a-2538-4c57-9c19-7ce0bdd587a2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`c][
   $CellContext`W[-$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289783813738737`*^9},
 CellLabel->"Out[787]=",
 CellID->948195929,ExpressionUUID->"2c969b03-01ba-449c-880b-df3f4d6d3e93"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"ProjectWith", "[", "metrich", "]"}]}]], "Input",
 CellLabel->
  "In[788]:=",ExpressionUUID->"dfce37a2-0f15-477f-b09b-efc7982b7296"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "a"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "d"},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`W[-$CellContext`a]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978381487236*^9},
 CellLabel->"Out[788]=",
 CellID->2000317963,ExpressionUUID->"1f519fc4-ee54-4fec-847a-7ee84ef8a39c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ProjectorToMetric"}]], "Input",
 CellLabel->
  "In[789]:=",ExpressionUUID->"3aa82a23-30d0-4096-b111-ac9e464dbfaa"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "a"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
     Editable->False], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`b],
      Editable->False]}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`c, $CellContext`d],
     Editable->False], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`d],
      Editable->False]}]}], ")"}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`W[-$CellContext`a]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.7289783816049547`*^9},
 CellLabel->"Out[789]=",
 CellID->487510356,ExpressionUUID->"0aeae5c6-5e16-4a16-a8a0-5662b116786e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[790]:=",ExpressionUUID->"decc9b8c-f666-42de-a36e-bc1c58e7c239"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`W[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`W[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`W[-$CellContext`b]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`W[-$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978381756514*^9},
 CellLabel->"Out[790]=",
 CellID->936439987,ExpressionUUID->"5bbcc8bd-cb5a-4290-9bb7-5279a0b406e2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "GradNormalToExtrinsicK"}]], "Input",
 CellLabel->
  "In[791]:=",ExpressionUUID->"3e2eb936-7586-4fff-a4bb-6e0d30b15bf3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`n[-$CellContext`b],
     Editable->False]}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "a"},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`a],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "a"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`d, $CellContext`a],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`W[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`W[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978381904604*^9},
 CellLabel->"Out[791]=",
 CellID->1986696293,ExpressionUUID->"ca9f73eb-f317-49db-b82d-36d4d510b454"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->
  "In[792]:=",ExpressionUUID->"4e9b2857-6b00-40cd-af91-c8d42ea2734a"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`W[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`W[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978382072831*^9},
 CellLabel->"Out[792]=",
 CellID->1328852895,ExpressionUUID->"1740eb1f-ea55-4499-9000-d5cce8d78255"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A different way to do it is this:", "MathCaption",ExpressionUUID->"92408cf4-dc70-4dd7-998b-c59937e11a7d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"W", "[", 
   RowBox[{"-", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[793]:=",ExpressionUUID->"f2acbc6b-f57b-42ab-8ab1-45c8709e2e6e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "c"], GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`cd[-$CellContext`c][
   $CellContext`W[-$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978382210218*^9},
 CellLabel->"Out[793]=",
 CellID->200677465,ExpressionUUID->"c7a5ead0-8f49-43e4-b050-2d01b6f887b4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ProjectDerivative", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[794]:=",ExpressionUUID->"ab708605-2164-43c6-90ac-818153d8ad65"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`W[-$CellContext`b]],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[
   $CellContext`CD[-$CellContext`c][
    $CellContext`W[-$CellContext`b]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978382372007*^9},
 CellLabel->"Out[794]=",
 CellID->1236828465,ExpressionUUID->"c8fb3416-9e89-4c84-ada1-4418d66f7c1e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Projectormetrich", "\[Rule]", 
   RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}]], "Input",
 CellLabel->
  "In[795]:=",ExpressionUUID->"ea9afe11-c509-4995-a680-8c19d7b80d17"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "a"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "d"},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
        {"W", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`W[-$CellContext`a]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.7289783824698467`*^9},
 CellLabel->"Out[795]=",
 CellID->1881791942,ExpressionUUID->"d073bbbd-bcb4-4238-80a5-e40d6c6403b1"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["ExtrinsicK",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ExtrinsicK.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Extrinsic curvature tensor", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["Acceleration",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Acceleration.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Acceleration vector", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["GradNormalToExtrinsicK",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "GradNormalToExtrinsicK.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t", "DisplayMath"],
 StyleBox["Convert from derivatives of the normal to the extrinsic curvature \
tensor", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["ExtrinsicKToGradNorma",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ExtrinsicKToGradNormal.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["l\t", "DisplayMath"],
 StyleBox["Convert form the extrinsic curvature tensor to derivatives of the \
normal\n", "Text"],
 StyleBox[ButtonBox["ProjectDerivative",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ProjectDerivative.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Convert derivatives on the projected vbundle into projected \
derivatives", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.72897839353584*^9},ExpressionUUID->"4322fa9b-6a43-4874-9206-b4005df2c599"],

Cell["Induced derivatives.", "Text",ExpressionUUID->"db1547c8-31dc-4aea-8a41-2df6422a3e64"],

Cell["\<\
A very important detail is that of the Leibnitz rule for induced derivatives. \
It is not valid on expressions not on the projected vbundle!\
\>", "Text",ExpressionUUID->"946d0d03-0149-416f-8966-05871a3a261c"],

Cell[CellGroupData[{

Cell[TextData[{
 "The vector ",
 StyleBox["W",
  FontFamily->"Courier"],
 " lives on the projected ",
 "vbundle",
 " and therefore the Leibnitz rule is valid:"
}], "MathCaption",ExpressionUUID->"d148de44-461f-4717-bd5f-f12c4cba85dd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"W", "[", "b", "]"}], 
   RowBox[{"W", "[", "c", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[796]:=",ExpressionUUID->"c6657f47-fb91-4433-9445-80986c68d1d5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`W[$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`W[$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978382589012*^9},
 CellLabel->"Out[796]=",
 CellID->1285499825,ExpressionUUID->"c84a2a41-4bd8-4041-aa2a-a164de0af840"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If there are free non-spatial indices we send an error message, and the input \
expression is returned unevaluated:\
\>", "MathCaption",ExpressionUUID->"d7175ee7-f7d7-4c16-b327-c25df842b190"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"V", "[", "b", "]"}], 
   RowBox[{"V", "[", "c", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[797]:=",ExpressionUUID->"3114835a-0d06-49fc-a7cf-a9e16e0991e5"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "nonproj", 
   "\"Induced derivative acting on the non-projected expression \
\\!\\(\\*RowBox[{InterpretationBox[StyleBox[GridBox[{{\\\"V\\\", \
StyleBox[GridBox[{{\\\"b\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[b], \
Rule[Editable, False]], \\\" \\\", InterpretationBox[StyleBox[GridBox[{{\\\"V\
\\\", StyleBox[GridBox[{{\\\"c\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[c], \
Rule[Editable, False]]}]\\).\"", 2, 797, 92, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978382738721*^9},
 CellLabel->"During evaluation of In[797]:=",
 CellID->1710069955,ExpressionUUID->"96db6553-ffb6-4294-9229-1b684546f538"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "nonproj", 
   "\"Induced derivative acting on the non-projected expression \
\\!\\(\\*RowBox[{InterpretationBox[StyleBox[GridBox[{{\\\"V\\\", \
StyleBox[GridBox[{{\\\"b\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[b], \
Rule[Editable, False]], \\\" \\\", InterpretationBox[StyleBox[GridBox[{{\\\"V\
\\\", StyleBox[GridBox[{{\\\"c\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[c], \
Rule[Editable, False]]}]\\).\"", 2, 797, 93, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978382765388*^9},
 CellLabel->"During evaluation of In[797]:=",
 CellID->1372085884,ExpressionUUID->"8d8ed186-18fb-4d3e-a262-1979f6232106"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "a"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"V", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V[$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"V", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V[$CellContext`c],
      Editable->False]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`cd[-$CellContext`a][$CellContext`V[$CellContext`b] \
$CellContext`V[$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289783828485937`*^9},
 CellLabel->"Out[797]=",
 CellID->968050902,ExpressionUUID->"c7b4635d-ea1d-4272-821e-58a4e30f8e4a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
With dummy indices we can expand both spatial indices and non-spatial indices:\
\>", "MathCaption",ExpressionUUID->"e928471c-6152-43bc-8bf0-1cf436fc9ce6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"W", "[", "a", "]"}], 
   RowBox[{"W", "[", "b", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[798]:=",ExpressionUUID->"db506d0d-6d24-43af-9579-5bf3f40a4829"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`W[$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`W[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"W", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`W[$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783830054913`*^9},
 CellLabel->"Out[798]=",
 CellID->695400242,ExpressionUUID->"01557f69-a661-4886-b6dc-50fd2ab8cc39"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"cd", "[", 
    RowBox[{"-", "a"}], "]"}], "[", 
   RowBox[{
    RowBox[{"V", "[", "b", "]"}], 
    RowBox[{"V", "[", 
     RowBox[{"-", "b"}], "]"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[799]:=",ExpressionUUID->"4f728459-655d-443d-9318-ad4f7089426d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"V", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`V[$CellContext`b],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[
     $CellContext`V[$CellContext`b]],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], 
       UnderscriptBox["P", "h"], "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"V", 
            StyleBox[GridBox[{
               {" "},
               {"b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`V[-$CellContext`b],
        Editable->False], "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`Projectormetrich[
       $CellContext`V[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["(",
       FontColor->RGBColor[1, 0, 0]], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`a],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"V", 
            StyleBox[GridBox[{
               {" "},
               {"a"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`V[-$CellContext`a],
        Editable->False]}], 
      StyleBox[")",
       FontColor->RGBColor[1, 0, 0]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Scalar[$CellContext`n[$CellContext`a] \
$CellContext`V[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["D", "a"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`b],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"V", 
            StyleBox[GridBox[{
               {" "},
               {"b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`V[-$CellContext`b],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`cd[-$CellContext`a][$CellContext`n[$CellContext`b] \
$CellContext`V[-$CellContext`b]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978383223494*^9},
 CellLabel->"Out[799]=",
 CellID->1956173085,ExpressionUUID->"e9cd8496-84df-49b7-8255-2c42c1b42c84"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In other cases, we prefer to show a warning and leave the expression \
untouched:\
\>", "MathCaption",
 CellChangeTimes->{{3.4716459815597343`*^9, 
  3.4716459822111607`*^9}},ExpressionUUID->"7282b404-affb-4801-ad7f-\
a887503eaa0a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{
   RowBox[{"V", "[", "a", "]"}], 
   RowBox[{"V", "[", "b", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[800]:=",ExpressionUUID->"55ceed0f-2c24-407b-8815-916119b2ccba"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "nonproj", 
   "\"Induced derivative acting on the non-projected expression \
\\!\\(\\*RowBox[{InterpretationBox[StyleBox[GridBox[{{\\\"V\\\", \
StyleBox[GridBox[{{\\\"a\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[a], \
Rule[Editable, False]], \\\" \\\", InterpretationBox[StyleBox[GridBox[{{\\\"V\
\\\", StyleBox[GridBox[{{\\\"b\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[b], \
Rule[Editable, False]]}]\\).\"", 2, 800, 94, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978383388501*^9},
 CellLabel->"During evaluation of In[800]:=",
 CellID->1573887672,ExpressionUUID->"275cf7c8-88fc-4b29-8b8c-813a79675abb"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "nonproj", 
   "\"Induced derivative acting on the non-projected expression \
\\!\\(\\*RowBox[{InterpretationBox[StyleBox[GridBox[{{\\\"V\\\", \
StyleBox[GridBox[{{\\\"a\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[a], \
Rule[Editable, False]], \\\" \\\", InterpretationBox[StyleBox[GridBox[{{\\\"V\
\\\", StyleBox[GridBox[{{\\\"b\\\"}, {\\\" \\\"}}, Rule[RowSpacings, 0], \
Rule[ColumnSpacings, 0.05`], Rule[GridFrameMargins, List[List[0, 0], List[0, \
0]]]], Rule[FontSize, 9]]}}, Rule[ColumnSpacings, 0.05`], Rule[RowAlignments, \
Center]], Rule[AutoSpacing, False], Rule[ShowAutoStyles, False]], V[b], \
Rule[Editable, False]]}]\\).\"", 2, 800, 95, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978383414803*^9},
 CellLabel->"During evaluation of In[800]:=",
 CellID->1774023205,ExpressionUUID->"5b022a4c-1cce-40ab-a69c-166b9ab8a913"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["D", "a"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"V", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"V", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`V[$CellContext`b],
      Editable->False]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`cd[-$CellContext`a][$CellContext`V[$CellContext`a] \
$CellContext`V[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978383498144*^9},
 CellLabel->"Out[800]=",
 CellID->345051487,ExpressionUUID->"e57c3cb7-377e-4c6b-b189-ff14197f0983"]
}, Open  ]]
}, Open  ]],

Cell["\<\
The most important application of this in GR is the ADM-type decomposition of \
a spacetime. In the rest of this section we reproduce some equations from the \
well-organized notes on ADM by M. Choptuik.\
\>", "Text",ExpressionUUID->"a891101e-2d10-44a9-b3d6-bfcc0f46cbc6"],

Cell[CellGroupData[{

Cell["\<\
1) Foliations and normals. We use his same notations for the normal vector \
and the acceleration vector:\
\>", "MathCaption",ExpressionUUID->"81eb0886-f0da-4930-9bdb-df4053db42c1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"n", "[", "a", "]"}], 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "a"}], "]"}], "[", 
     RowBox[{"n", "[", "b", "]"}], "]"}]}], "//", "GradNormalToExtrinsicK"}], 
  "//", "Expand"}]], "Input",
 CellLabel->
  "In[801]:=",ExpressionUUID->"48277dae-31b8-4156-8767-640436af5c7c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"An", 
      StyleBox[GridBox[{
         {"b"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Accelerationn[$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978383638309*^9},
 CellLabel->"Out[801]=",
 CellID->1074495109,ExpressionUUID->"bb409634-bbb4-4dff-b04f-36bb01a16c0e"]
}, Open  ]]
}, Open  ]],

Cell["\<\
2) The projection tensor and the spatial metric. We do not follow the sign \
switch of York. The orthogonal projector is our pair Projectormetrich / \
ProjectWith[metrich] as explained above.\
\>", "MathCaption",ExpressionUUID->"950800d7-4e74-4710-ad96-19eef081d217"],

Cell[CellGroupData[{

Cell["3) The spatial derivative operator and curvature tensor.", "MathCaption",ExpressionUUID->"4d9e887a-4901-4234-8d9d-862fff0af6bb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"metrich", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[802]:=",ExpressionUUID->"8140c199-4bce-446e-a4ae-cc31b48104f4"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978383790574*^9},
 CellLabel->"Out[802]=",
 CellID->849309340,ExpressionUUID->"05ec780b-4916-4d04-a134-98be9d736481"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"cd", "[", 
       RowBox[{"-", "a"}], "]"}], "@", 
      RowBox[{
       RowBox[{"cd", "[", 
        RowBox[{"-", "b"}], "]"}], "[", "#", "]"}]}], "-", 
     RowBox[{
      RowBox[{"cd", "[", 
       RowBox[{"-", "b"}], "]"}], "@", 
      RowBox[{
       RowBox[{"cd", "[", 
        RowBox[{"-", "a"}], "]"}], "[", "#", "]"}]}]}], ")"}], "&"}], "@", 
  RowBox[{"W", "[", 
   RowBox[{"-", "c"}], "]"}]}]], "Input",
 CellLabel->
  "In[803]:=",ExpressionUUID->"ba163dd5-a531-4d93-a30f-2a5ffc97c27c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "a"], 
     SubscriptBox["D", "b"], GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`a][
    $CellContext`cd[-$CellContext`b][
     $CellContext`W[-$CellContext`c]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], 
     SubscriptBox["D", "a"], GridBox[{
       {"W", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`cd[-$CellContext`a][
     $CellContext`W[-$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783839392347`*^9},
 CellLabel->"Out[803]=",
 CellID->18275771,ExpressionUUID->"0dd05c87-6e2d-4b25-8fba-06775d81e40b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "SortCovDs"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[804]:=",ExpressionUUID->"199a1aa8-ad00-4382-af03-d68d92f594e2"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {"d"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`W[$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783840557623`*^9},
 CellLabel->"Out[804]=",
 CellID->70721676,ExpressionUUID->"82a9a4eb-ef28-4fec-b85d-a836d027a0e7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Riemanncd", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "d"}], "]"}], 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"n", "[", "a", "]"}], ",", 
    RowBox[{"n", "[", "b", "]"}], ",", 
    RowBox[{"n", "[", "c", "]"}], ",", 
    RowBox[{"n", "[", 
     RowBox[{"-", "d"}], "]"}]}], "}"}]}]], "Input",
 CellLabel->
  "In[805]:=",ExpressionUUID->"a14ae639-205d-4a2d-83cf-05feab8103ee"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{3.728978384172439*^9},
 CellLabel->"Out[805]=",
 CellID->1060751741,ExpressionUUID->"0288b88f-f9c7-45e2-bf36-11987136fa5a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Riemanncd", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", 
   RowBox[{"-", "c"}], ",", 
   RowBox[{"-", "b"}], ",", "c"}], "]"}]], "Input",
 CellLabel->
  "In[806]:=",ExpressionUUID->"08d7337c-245f-4314-94fd-a904e2f6296d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "D", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978384286434*^9},
 CellLabel->"Out[806]=",
 CellID->1792757240,ExpressionUUID->"697419b9-bdb4-4928-922e-0877fa4d7706"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The contraction of Ricci into RicciScalar does not work:", "MathCaption",ExpressionUUID->"850fa9c0-f74a-417a-9a2c-ddcde079cf18"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Riccicd", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "a"}], "]"}], ",", 
   RowBox[{
    RowBox[{"metrich", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Riccicd", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"Riccicd", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}]}], "}"}]], "Input",
 CellLabel->
  "In[807]:=",ExpressionUUID->"d11bc6bf-d718-418c-a04f-7502f3d9af6e"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox[
     RowBox[{"R", "[", "D", "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciScalarcd[],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"R", "[", "D", "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciScalarcd[],
    Editable->False], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {"a", "b"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[$CellContext`a, $CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "D", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978384405715*^9},
 CellLabel->"Out[807]=",
 CellID->1675365270,ExpressionUUID->"44b953fe-a864-4224-8f17-cbbbe761814b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["4) The extrinsic curvature tensor", "MathCaption",ExpressionUUID->"8aa58e4c-9317-4668-ace2-d2dfaf41ca8f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "a"}], "]"}], "[", 
    RowBox[{"n", "[", 
     RowBox[{"-", "b"}], "]"}], "]"}], "//", "GradNormalToExtrinsicK"}], "//",
   "ContractMetric"}]], "Input",
 CellLabel->
  "In[808]:=",ExpressionUUID->"a4d6aa28-a724-4952-a127-cd1bae6267dd"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978384523757*^9},
 CellLabel->"Out[808]=",
 CellID->58513426,ExpressionUUID->"1e3dc39a-5b5d-4702-926d-26df56036304"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Antisymmetrize", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", "%", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "}"}]}], "]"}], "//", "ContractMetric"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[809]:=",ExpressionUUID->"14b47a10-48cc-4fea-9acb-31a89a18b885"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978384661337*^9},
 CellLabel->"Out[809]=",
 CellID->2035124103,ExpressionUUID->"c003739a-458c-4167-a07c-228489860ac6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"-", "1"}], "/", "2"}], 
      RowBox[{
       RowBox[{"LieD", "[", 
        RowBox[{
         RowBox[{"n", "[", "c", "]"}], ",", "CD"}], "]"}], "[", 
       RowBox[{"metricg", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "b"}]}], "]"}], "]"}]}], "//", 
     "GradNormalToExtrinsicK"}], "]"}], "//", "ContractMetric"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[810]:=",ExpressionUUID->"043a5706-7642-4ae1-bbce-f2616fd93366"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"Kh", 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978384783965*^9},
 CellLabel->"Out[810]=",
 CellID->1358753352,ExpressionUUID->"e2aa869b-8579-4faf-80f7-80ca6182a5cc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"-", "1"}], "/", "2"}], 
      RowBox[{
       RowBox[{"LieD", "[", 
        RowBox[{
         RowBox[{"n", "[", "c", "]"}], ",", "CD"}], "]"}], "[", 
       RowBox[{"metrich", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "b"}]}], "]"}], "]"}]}], "//", "ProjectorToMetric"}], "//",
     "GradNormalToExtrinsicK"}], "//", "ContractMetric"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[811]:=",ExpressionUUID->"ada00239-b41a-4fa3-af0b-0398a9c50ec1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"Kh", 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "b"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289783849234133`*^9},
 CellLabel->"Out[811]=",
 CellID->226484866,ExpressionUUID->"cf1e9294-1c2d-4b49-b1b0-aba9ec088424"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5) The Gauss-Codazzi equations. Single derivative:", "MathCaption",ExpressionUUID->"d22d1995-fc26-403b-9bd3-4f5c7890c674"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "v", "]"}]], "Input",
 CellLabel->
  "In[812]:=",ExpressionUUID->"ab387ecb-66de-455b-accd-7050bad5053e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978385055627*^9},
 CellLabel->"During evaluation of In[812]:=",
 CellID->1190120378,ExpressionUUID->"efbc8fbc-0353-45cb-9a71-f67c886d8b47"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3", ",", 
   RowBox[{"OrthogonalTo", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], "}"}]}], ",", 
   RowBox[{"ProjectedWith", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"metrich", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", "b"}], "]"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[813]:=",ExpressionUUID->"56e5d15a-38f1-48ad-9e8f-21b41269a27c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897838520605*^9},
 CellLabel->"During evaluation of In[813]:=",
 CellID->1437879317,ExpressionUUID->"b4d38a62-844d-4571-993c-00cdacbe2df0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "c"}], "]"}], "[", 
  RowBox[{"v", "[", 
   RowBox[{"-", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[814]:=",ExpressionUUID->"be53ac32-2d78-4b0a-b58e-8a217455fbe3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`CD[-$CellContext`c][
   $CellContext`v[-$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978385323153*^9},
 CellLabel->"Out[814]=",
 CellID->804201128,ExpressionUUID->"2832be64-b87a-4708-887c-98f5a3cc3def"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", 
  RowBox[{"ProjectWith", "[", "metrich", "]"}]}]], "Input",
 CellLabel->
  "In[815]:=",ExpressionUUID->"14bff954-44b8-4863-a887-4dee9f4c02db"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "a"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"h", 
       StyleBox[GridBox[{
          {" ", "d"},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metrich[-$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`a]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978385439849*^9},
 CellLabel->"Out[815]=",
 CellID->2093767725,ExpressionUUID->"b9b6e115-b600-4444-8785-b693694f9264"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ProjectorToMetric"}]], "Input",
 CellLabel->
  "In[816]:=",ExpressionUUID->"20debca3-3216-434f-a4b7-ff313f802198"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "a"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`b, $CellContext`a],
     Editable->False], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`b],
      Editable->False]}]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "d"},
            {"c", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[-$CellContext`c, $CellContext`d],
     Editable->False], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`d],
      Editable->False]}]}], ")"}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`a]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978385536092*^9},
 CellLabel->"Out[816]=",
 CellID->1287114333,ExpressionUUID->"98b3dad1-1191-4cfa-ac25-3d3e11a77879"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Expand"}]], "Input",
 CellLabel->
  "In[817]:=",ExpressionUUID->"e8364c72-27e6-4aa6-9bf1-2d21d072ff40"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`v[-$CellContext`b]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783856562*^9},
 CellLabel->"Out[817]=",
 CellID->927818500,ExpressionUUID->"57fc4f71-7a64-475f-85b0-e746b6dafce4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "GradNormalToExtrinsicK"}]], "Input",
 CellLabel->
  "In[818]:=",ExpressionUUID->"44140647-b549-4927-a788-e5c95e5aeac2"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`n[-$CellContext`b],
     Editable->False]}], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "a"},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`a],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "a"},
              {"d", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`d, $CellContext`a],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`v[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897838578573*^9},
 CellLabel->"Out[818]=",
 CellID->966768633,ExpressionUUID->"fcc022c9-1619-4a9d-92c7-fd3201c0223f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->
  "In[819]:=",ExpressionUUID->"471b8453-739d-4d1d-8ee3-b9047d10a47c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`v[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978385923937*^9},
 CellLabel->"Out[819]=",
 CellID->619141629,ExpressionUUID->"9ed3f6d9-6f46-40dc-91d4-15f4c59c06a2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We compare with expression (47) of Choptuik's notes:", "MathCaption",ExpressionUUID->"d0a981e5-fbec-4c2e-9f51-428276d2802d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "%"}], "+", 
  RowBox[{
   RowBox[{"CD", "[", 
    RowBox[{"-", "c"}], "]"}], "[", 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}], "]"}], "-", 
  RowBox[{
   RowBox[{"n", "[", 
    RowBox[{"-", "b"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "f"}], "]"}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "c"}], "]"}], "[", 
    RowBox[{"n", "[", "f", "]"}], "]"}]}], "+", 
  RowBox[{
   RowBox[{"n", "[", 
    RowBox[{"-", "c"}], "]"}], 
   RowBox[{"n", "[", "e", "]"}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "[", 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}], "]"}]}], "-", 
  RowBox[{
   RowBox[{"n", "[", 
    RowBox[{"-", "c"}], "]"}], 
   RowBox[{"n", "[", 
    RowBox[{"-", "b"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "f"}], "]"}], 
   RowBox[{"Accelerationn", "[", "f", "]"}]}]}]], "Input",
 CellLabel->
  "In[820]:=",ExpressionUUID->"caf3067e-e44b-49b9-bc94-bfc4a1b32c2e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"f"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[$CellContext`f]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978386091324*^9},
 CellLabel->"Out[820]=",
 CellID->1355088888,ExpressionUUID->"54896bc4-5806-4777-9cfb-9f0a0203e089"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "GradNormalToExtrinsicK"}]], "Input",
 CellLabel->
  "In[821]:=",ExpressionUUID->"f93f740a-c394-4435-9d5f-b5dab002160d"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "f"},
              {"c", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`f],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {"f"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[$CellContext`f],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978386272078*^9},
 CellLabel->"Out[821]=",
 CellID->914864181,ExpressionUUID->"4b292053-543a-4e72-b100-d51969b8208f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->
  "In[822]:=",ExpressionUUID->"6458fb8b-c5cf-4ebf-90e1-635d2c49b1dc"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978386439844*^9},
 CellLabel->"Out[822]=",
 CellID->139377547,ExpressionUUID->"9de7627e-4ce3-415b-b76e-1b7d2e76777f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Second derivatives:", "MathCaption",ExpressionUUID->"8b602bf8-def1-45b7-b908-b9f8c75cfa6e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "d"}], "]"}], "[", 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "c"}], "]"}], "[", 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}], "]"}], "]"}], "]"}], "-", 
  RowBox[{
   RowBox[{"cd", "[", 
    RowBox[{"-", "d"}], "]"}], "[", 
   RowBox[{
    RowBox[{"cd", "[", 
     RowBox[{"-", "c"}], "]"}], "[", 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}], "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[823]:=",ExpressionUUID->"248dc607-74e5-40e9-a62d-f5778d92a251"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], 
       SubscriptBox["D", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`cd[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783865905733`*^9},
 CellLabel->"Out[823]=",
 CellID->1761257199,ExpressionUUID->"1c086fd5-2497-4744-9f5e-01e3ae0bdd41"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ProjectDerivative"}], "//", 
  "ProjectDerivative"}]], "Input",
 CellLabel->
  "In[824]:=",ExpressionUUID->"7406f188-b98c-4bd2-a4f3-2d9bddf99067"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "d"], 
         UnderscriptBox["P", "h"], "[", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {"b"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CD[-$CellContext`c][
           $CellContext`v[-$CellContext`b]],
          Editable->False], "]"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`d][
        $CellContext`Projectormetrich[
         $CellContext`CD[-$CellContext`c][
          $CellContext`v[-$CellContext`b]]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[
     $CellContext`CD[-$CellContext`d][
      $CellContext`Projectormetrich[
       $CellContext`CD[-$CellContext`c][
        $CellContext`v[-$CellContext`b]]]]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978386739194*^9},
 CellLabel->"Out[824]=",
 CellID->535153345,ExpressionUUID->"6864c629-81d4-4148-a346-e50c2b42c955"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"Projectormetrich", "\[Rule]", 
   RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}]], "Input",
 CellLabel->
  "In[825]:=",ExpressionUUID->"fdbb5bf5-7f56-4f40-a962-cd9b6e9a74e8"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"h", 
         StyleBox[GridBox[{
            {" ", "f"},
            {"b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metrich[-$CellContext`b, $CellContext`f],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "g"},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`d, $CellContext`g],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "g"], GridBox[{
         {"h", 
          StyleBox[GridBox[{
             {" ", "a"},
             {"f", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`g][
      $CellContext`metrich[-$CellContext`f, $CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`c, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "g"},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`d, $CellContext`g],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "g"], GridBox[{
         {"h", 
          StyleBox[GridBox[{
             {" ", "e"},
             {"f", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`g][
      $CellContext`metrich[-$CellContext`f, $CellContext`e]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783869064417`*^9},
 CellLabel->"Out[825]=",
 CellID->660755779,ExpressionUUID->"8195bff4-c067-41d3-b928-2740e71784e7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[826]:=",ExpressionUUID->"f7c2aa06-7d86-4345-9dcc-c34718f76cc6"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`n[$CellContext`a],
     Editable->False]}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`f]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"g"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`g],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "g"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`g][
      $CellContext`n[-$CellContext`f]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978387112753*^9},
 CellLabel->"Out[826]=",
 CellID->684439016,ExpressionUUID->"600ffc97-240e-4a07-94c8-57af64c48eca"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[827]:=",ExpressionUUID->"23aebe17-2c0b-42da-bca5-761c36d5ad3f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.72897838736957*^9},
 CellLabel->"Out[827]=",
 CellID->1444128580,ExpressionUUID->"35f4935b-ad34-4eb6-ba69-f4a47f515fb2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ContractMetric"}]], "Input",
 CellLabel->
  "In[828]:=",ExpressionUUID->"b245d85c-bbb7-474c-bcac-110695affdbb"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978387540292*^9},
 CellLabel->"Out[828]=",
 CellID->608865666,ExpressionUUID->"c9bf446e-f9a2-4658-8abd-af0686cb598d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->
  "In[829]:=",ExpressionUUID->"f38a1752-e389-497d-b365-d8adb2e5147c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783877223186`*^9},
 CellLabel->"Out[829]=",
 CellID->445670507,ExpressionUUID->"18d99ee8-398d-4e5d-9603-83e37a483984"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["First Gauss-Codazzi equation:", "MathCaption",ExpressionUUID->"9220c3f2-233a-4b85-9803-cd7b301b6632"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1", "=", 
  RowBox[{"Projectormetrich", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "c"}], "]"}], "[", 
      RowBox[{"v", "[", 
       RowBox[{"-", "b"}], "]"}], "]"}], "]"}], "-", 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "c"}], "]"}], "[", 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"v", "[", 
       RowBox[{"-", "b"}], "]"}], "]"}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[830]:=",ExpressionUUID->"dcf0b184-8de3-410c-90d2-7874b6184a6c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "c"], 
         SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {" "},
               {"b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`c][
        $CellContext`CD[-$CellContext`d][
         $CellContext`v[-$CellContext`b]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], 
        SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`CD[-$CellContext`c][
        $CellContext`v[-$CellContext`b]]],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`CD[-$CellContext`d][
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`b]]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978387873282*^9},
 CellLabel->"Out[830]=",
 CellID->2108734060,ExpressionUUID->"2268dc17-35ff-439a-9e64-65bd85edcd50"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1A", "=", 
  RowBox[{"expr1", "//", "SortCovDs"}]}]], "Input",
 CellLabel->
  "In[831]:=",ExpressionUUID->"dbc1f46f-8a73-4c6c-a166-68693b251994"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"d", "c", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`d, -$CellContext`c, \
-$CellContext`b, $CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`a],
      Editable->False]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[$CellContext`RiemannCD[-$CellContext`d, \
-$CellContext`c, -$CellContext`b, $CellContext`a] \
$CellContext`v[-$CellContext`a]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978387986929*^9},
 CellLabel->"Out[831]=",
 CellID->1570719979,ExpressionUUID->"4644d6f7-e3a2-4c4f-a799-ad0d6de4a28d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"expr1", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[832]:=",ExpressionUUID->"a1cac09b-2564-4338-b395-7d316d45c283"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`CD[-$CellContext`c][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`f][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978388191828*^9},
 CellLabel->"Out[832]=",
 CellID->74214482,ExpressionUUID->"73396972-fe19-4219-9e74-083bce127dc5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1B", "=", 
  RowBox[{
   RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", 
   "Simplification"}]}]], "Input",
 CellLabel->
  "In[833]:=",ExpressionUUID->"cc6884e1-534c-4fca-bd51-b2cb26187a31"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`CD[-$CellContext`c][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"-", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], 
          SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`CD[-$CellContext`d][
          $CellContext`v[-$CellContext`a]]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], 
        SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`CD[-$CellContext`c][
        $CellContext`v[-$CellContext`a]]],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"e"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "d"], 
          SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`d][
         $CellContext`CD[-$CellContext`e][
          $CellContext`v[-$CellContext`a]]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"e"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{"(", 
          InterpretationBox[
           StyleBox[
            RowBox[{
             SubscriptBox["\[EmptyDownTriangle]", "c"], 
             SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
               {"v", 
                StyleBox[GridBox[{
                   {" "},
                   {"a"}
                  },
                  GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                 FontSize->9]}
              },
              
              GridBoxAlignment->{
               "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                "Rows" -> {{Center}}, "RowsIndexed" -> {}},
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.27999999999999997`], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                "Rows" -> {
                  Offset[0.2], {
                   Offset[0.4]}, 
                  Offset[0.2]}, "RowsIndexed" -> {}}]}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`CD[-$CellContext`c][
            $CellContext`CD[-$CellContext`e][
             $CellContext`v[-$CellContext`a]]],
           Editable->False], ")"}]}], "+", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], 
           SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"a"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`CD[-$CellContext`c][
           $CellContext`v[-$CellContext`a]]],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> 
           {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"e"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "e"], 
          SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`e][
         $CellContext`CD[-$CellContext`d][
          $CellContext`v[-$CellContext`a]]],
        Editable->False], ")"}]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978388466999*^9},
 CellLabel->"Out[833]=",
 CellID->1782283714,ExpressionUUID->"1e789b15-d357-4bff-b704-3c8bc61b339a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"cd", "[", 
     RowBox[{"-", "d"}], "]"}], "[", 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "c"}], "]"}], "[", 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}], "]"}], "]"}], "-", 
   RowBox[{
    RowBox[{"cd", "[", 
     RowBox[{"-", "c"}], "]"}], "[", 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}], "]"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[834]:=",ExpressionUUID->"b6113762-d2e2-4e83-ada5-5f8ea3eaaf75"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], 
       SubscriptBox["D", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`cd[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "d"], 
     SubscriptBox["D", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`d][
    $CellContext`cd[-$CellContext`c][
     $CellContext`v[-$CellContext`b]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978388604958*^9},
 CellLabel->"Out[834]=",
 CellID->1740472842,ExpressionUUID->"94032fe5-e022-433b-833c-bcbb8e750729"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2A", "=", 
  RowBox[{"expr2", "//", "SortCovDs"}]}]], "Input",
 CellLabel->
  "In[835]:=",ExpressionUUID->"6c891b48-cc0d-4cd4-8be2-aacc4b290ba7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "a"},
          {"d", "c", "b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`d, -$CellContext`c, -$CellContext`b, \
$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978388722826*^9},
 CellLabel->"Out[835]=",
 CellID->1252299707,ExpressionUUID->"b096926c-6b6a-480e-a9ae-7ed10abdc6f1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"expr2", "//", "ProjectDerivative"}], "//", 
  "ProjectDerivative"}]], "Input",
 CellLabel->
  "In[836]:=",ExpressionUUID->"a752487c-6e52-4597-89fb-1799f4c6faa3"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "c"], 
         UnderscriptBox["P", "h"], "[", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {"b"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CD[-$CellContext`d][
           $CellContext`v[-$CellContext`b]],
          Editable->False], "]"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`c][
        $CellContext`Projectormetrich[
         $CellContext`CD[-$CellContext`d][
          $CellContext`v[-$CellContext`b]]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[
     $CellContext`CD[-$CellContext`c][
      $CellContext`Projectormetrich[
       $CellContext`CD[-$CellContext`d][
        $CellContext`v[-$CellContext`b]]]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], 
        UnderscriptBox["P", "h"], "[", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"b"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`c][
          $CellContext`v[-$CellContext`b]],
         Editable->False], "]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`Projectormetrich[
        $CellContext`CD[-$CellContext`c][
         $CellContext`v[-$CellContext`b]]]],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`CD[-$CellContext`d][
     $CellContext`Projectormetrich[
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978388840487*^9},
 CellLabel->"Out[836]=",
 CellID->2061434949,ExpressionUUID->"5871785a-2877-4cbe-bf01-4afff8b66e6b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[837]:=",ExpressionUUID->"6944dd13-2689-4769-95eb-100ac716eb27"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`CD[-$CellContext`c][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`n[-$CellContext`f]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`d]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`f]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`n[-$CellContext`f]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"g"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`g],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"g"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`n[-$CellContext`g]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"g"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`g],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`v[-$CellContext`a]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "g"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`g][
      $CellContext`n[-$CellContext`f]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978389142853*^9},
 CellLabel->"Out[837]=",
 CellID->1183510792,ExpressionUUID->"19b59072-2cf9-4d6c-a944-cd5f406eb110"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[838]:=",ExpressionUUID->"63e4d41f-c2b0-42bc-b5f6-2c63a409eaca"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`v[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`CD[-$CellContext`c][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783893875217`*^9},
 CellLabel->"Out[838]=",
 CellID->1272342630,ExpressionUUID->"839dc0b1-4ffe-4d6c-8729-80fbe5e964a4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2B", "=", 
  RowBox[{"%", "//", "Simplification"}]}]], "Input",
 CellLabel->
  "In[839]:=",ExpressionUUID->"91e81c5f-0bb1-405b-9e00-462723c0f8d9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`CD[-$CellContext`d][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`v[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`CD[-$CellContext`c][
     $CellContext`v[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`e][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`c][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`d][
       $CellContext`v[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978389639183*^9},
 CellLabel->"Out[839]=",
 CellID->1767909231,ExpressionUUID->"c379fca7-cbd8-4337-a185-817e58bda7ad"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", 
    RowBox[{"expr1A", "-", "expr1B"}], ")"}], "-", 
   RowBox[{"(", 
    RowBox[{"expr2A", "-", "expr2B"}], ")"}]}], "//", 
  "Simplification"}]], "Input",
 CellLabel->
  "In[840]:=",ExpressionUUID->"1aa74944-89ad-49cc-84bd-81332f9338e1"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "a", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, \
-$CellContext`c, -$CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[$CellContext`a],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`a, -$CellContext`c, -$CellContext`d] \
$CellContext`v[$CellContext`a]],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", " "},
               {"b", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`a],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "D", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"b", "a", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Riemanncd[-$CellContext`b, -$CellContext`a, \
-$CellContext`c, -$CellContext`d],
      Editable->False]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978389793508*^9},
 CellLabel->"Out[840]=",
 CellID->666705525,ExpressionUUID->"4d53ca9d-5bc7-41c3-86f4-c6de129f9dc0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"GCRiemann1", "[", 
    RowBox[{
     RowBox[{"-", "a_"}], ",", 
     RowBox[{"-", "b_"}], ",", 
     RowBox[{"-", "c_"}], ",", 
     RowBox[{"-", "d_"}]}], "]"}], ",", 
   RowBox[{"Simplification", "[", 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"%", "/.", 
       RowBox[{
        RowBox[{"v", "[", "_", "]"}], "\[Rule]", "1"}]}], "//", 
      "ScreenDollarIndices"}], ")"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[841]:=",ExpressionUUID->"3036fcf4-7f5f-4abf-a39c-d9435c169b21"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"a", "b", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783901685963`*^9},
 CellLabel->"Out[841]=",
 CellID->1897572276,ExpressionUUID->"0b230a1b-cdd5-4e56-b334-ec3376f688cc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["If we decompose the Riemann tensor along the vector n", "MathCaption",ExpressionUUID->"8fbe4d8a-ea32-4de8-af7e-e56fd978f972"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRiemann", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}], ",", 
    RowBox[{"-", "c_"}], ",", 
    RowBox[{"-", "d_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{"InducedDecomposition", "[", 
    RowBox[{
     RowBox[{"RiemannCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "d"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}], "//", 
   "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[842]:=",ExpressionUUID->"d6f4ccee-a446-45d4-9ed9-f92d9629cd8a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"a", "b", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "c", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "d", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
-$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978390479409*^9},
 CellLabel->"Out[842]=",
 CellID->217601135,ExpressionUUID->"18782321-8445-4df4-b53e-ab3896cb8fb0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["then we can express it as", "MathCaption",ExpressionUUID->"645e61a7-dbab-4821-95bb-7b3edbc299f3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRiemann", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}], ",", 
    RowBox[{"-", "c_"}], ",", 
    RowBox[{"-", "d_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"inducedRiemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], "-", 
    RowBox[{"GCRiemann1", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[843]:=",ExpressionUUID->"d75d2ee6-dbae-4efd-b25b-a1532a39fdef"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "c", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "b", "d", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, \
-$CellContext`d, -$CellContext`e],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`d, \
-$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978391397525*^9},
 CellLabel->"Out[843]=",
 CellID->57625104,ExpressionUUID->"da01718d-78b9-4941-8372-6bc3580912c7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Second Gauss-Codazzi equation:", "MathCaption",ExpressionUUID->"1c776f3b-7b53-4eca-b60a-644aac1ef447"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1", "=", 
  RowBox[{"Projectormetrich", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "d"}], "]"}], "[", 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "c"}], "]"}], "[", 
      RowBox[{"n", "[", 
       RowBox[{"-", "b"}], "]"}], "]"}], "]"}], "-", 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "c"}], "]"}], "[", 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"n", "[", 
       RowBox[{"-", "b"}], "]"}], "]"}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[844]:=",ExpressionUUID->"4b104a24-2a7e-4dad-b004-137529e445ee"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "c"], 
         SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {" "},
               {"b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`c][
        $CellContext`CD[-$CellContext`d][
         $CellContext`n[-$CellContext`b]]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`b]]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], 
        SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`CD[-$CellContext`c][
        $CellContext`n[-$CellContext`b]]],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`CD[-$CellContext`d][
     $CellContext`CD[-$CellContext`c][
      $CellContext`n[-$CellContext`b]]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978391652082*^9},
 CellLabel->"Out[844]=",
 CellID->2137481778,ExpressionUUID->"50d7b4af-6a42-46e2-950f-541ae2b9c886"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1A", "=", 
  RowBox[{"expr1", "//", "SortCovDs"}]}]], "Input",
 CellLabel->
  "In[845]:=",ExpressionUUID->"805cb987-ad32-419a-ae7a-15df50792d9e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"d", "c", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`d, -$CellContext`c, \
-$CellContext`b, $CellContext`a],
      Editable->False]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[$CellContext`n[-$CellContext`a] \
$CellContext`RiemannCD[-$CellContext`d, -$CellContext`c, -$CellContext`b, \
$CellContext`a]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289783918371763`*^9},
 CellLabel->"Out[845]=",
 CellID->1549795830,ExpressionUUID->"cb407dc8-0856-41fd-947b-db69c37224da"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"expr1", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[846]:=",ExpressionUUID->"e868a243-edcf-4744-b1e8-c5dd302bff28"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`c][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`e][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`CD[-$CellContext`d][
     $CellContext`n[-$CellContext`b]]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`e][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`CD[-$CellContext`c][
     $CellContext`n[-$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`e][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`a][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`c][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`f][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`CD[-$CellContext`e][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978392042131*^9},
 CellLabel->"Out[846]=",
 CellID->589360347,ExpressionUUID->"885449f8-4baa-45ff-b8f6-b71a7d27d26b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1B", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", "ContractMetric"}], 
    "//", "GradNormalToExtrinsicK"}], "//", "Simplification"}]}]], "Input",
 CellLabel->
  "In[847]:=",ExpressionUUID->"bfae4782-47c8-4d6d-952d-e10fdfe44336"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", " "},
               {"d", "a"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`a],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978392273409*^9},
 CellLabel->"Out[847]=",
 CellID->1184772065,ExpressionUUID->"9827c46f-3188-4207-a734-290db5e7f5c6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2A", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"cd", "[", 
     RowBox[{"-", "c"}], "]"}], "[", 
    RowBox[{"ExtrinsicKmetrich", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], "]"}], "-", 
   RowBox[{
    RowBox[{"cd", "[", 
     RowBox[{"-", "d"}], "]"}], "[", 
    RowBox[{"ExtrinsicKmetrich", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[848]:=",ExpressionUUID->"91aaa073-3887-48a7-8c8c-324638921de9"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "c"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`c][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "d"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`d][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289783924657516`*^9},
 CellLabel->"Out[848]=",
 CellID->995721471,ExpressionUUID->"12ce8edd-aaf0-4708-955a-54644109a555"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2A", "//", "ProjectDerivative"}]], "Input",
 CellLabel->
  "In[849]:=",ExpressionUUID->"0b8acb87-30a2-4df3-88c2-5e2477681d71"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`c][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`CD[-$CellContext`c][
     $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`CD[-$CellContext`d][
     $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978392586802*^9},
 CellLabel->"Out[849]=",
 CellID->1898185673,ExpressionUUID->"be552039-825c-4c23-81c3-36981e7497cb"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[850]:=",ExpressionUUID->"db15b3fd-0e16-4b4a-a8cb-ba593513392c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`n[$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`e]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`e]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "f"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`f]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"f"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`f],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`f][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`e]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978392775831*^9},
 CellLabel->"Out[850]=",
 CellID->1937029926,ExpressionUUID->"b9e540f0-a118-4405-b31a-1690f77f8a0c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2B", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", "ContractMetric"}], "//",
    "Simplification"}]}]], "Input",
 CellLabel->
  "In[851]:=",ExpressionUUID->"7a03111b-313c-46a2-b426-c5d6e336e722"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", " "},
               {"d", "a"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`a],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}]}], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978393024456*^9},
 CellLabel->"Out[851]=",
 CellID->1957361409,ExpressionUUID->"57325239-b282-4b4a-94f2-4220e0008aaf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"GCRiemann2", "[", 
    RowBox[{
     RowBox[{"-", "b_"}], ",", 
     RowBox[{"-", "c_"}], ",", 
     RowBox[{"-", "d_"}]}], "]"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"expr1A", "-", "expr1B"}], ")"}], "-", 
     RowBox[{"(", 
      RowBox[{"expr2A", "-", "expr2B"}], ")"}]}], "//", "Simplification"}]}], 
  "]"}]], "Input",
 CellLabel->
  "In[852]:=",ExpressionUUID->"6402e71d-ae9f-478c-8298-87614c7c2ebf"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`a],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "a", "c", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, \
-$CellContext`c, -$CellContext`d],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`a] \
$CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`c, \
-$CellContext`d]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "c"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`c][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "d"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`d][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978393207261*^9},
 CellLabel->"Out[852]=",
 CellID->71212670,ExpressionUUID->"b580a972-fd42-43ac-8727-68a57a3ead0a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can further decompose the Riemann tensor now:", "MathCaption",ExpressionUUID->"375ee021-f39e-431d-a9bb-979c3293d949"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRiemann", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}], ",", 
    RowBox[{"-", "c_"}], ",", 
    RowBox[{"-", "d_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"inducedRiemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], "-", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "d"}], "]"}], 
     RowBox[{"GCRiemann2", "[", 
      RowBox[{
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}]}], "+", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "c"}], "]"}], 
     RowBox[{"GCRiemann2", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}]}], "-", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "b"}], "]"}], 
     RowBox[{"GCRiemann2", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "d"}]}], "]"}]}], "+", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"GCRiemann2", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}], ",", 
       RowBox[{"-", "d"}]}], "]"}]}]}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[853]:=",ExpressionUUID->"20b2587b-c994-4104-81a2-dba51d7934cd"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"a", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`a, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`a, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "c", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`c, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"f"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`f],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " ", " ", " "},
               {"b", "e", "d", "f"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RiemannCD[-$CellContext`b, -$CellContext`e, \
-$CellContext`d, -$CellContext`f],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[($CellContext`n[$CellContext`e] \
$CellContext`n[$CellContext`f]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`e, -$CellContext`d, -$CellContext`f]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.72897839346703*^9},
 CellLabel->"Out[853]=",
 CellID->722953311,ExpressionUUID->"79dc49ad-e02f-4b69-a639-e3b5181cf9d2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\"Third Gauss-Codazzi equation\":", "MathCaption",ExpressionUUID->"6525bcad-7378-4aeb-b6b3-470375a630c7"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1", "=", 
  RowBox[{"Projectormetrich", "[", 
   RowBox[{
    RowBox[{"n", "[", "d", "]"}], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       RowBox[{"CD", "[", 
        RowBox[{"-", "d"}], "]"}], "[", 
       RowBox[{
        RowBox[{"CD", "[", 
         RowBox[{"-", "c"}], "]"}], "[", 
        RowBox[{"n", "[", 
         RowBox[{"-", "b"}], "]"}], "]"}], "]"}], "-", 
      RowBox[{
       RowBox[{"CD", "[", 
        RowBox[{"-", "c"}], "]"}], "[", 
       RowBox[{
        RowBox[{"CD", "[", 
         RowBox[{"-", "d"}], "]"}], "[", 
        RowBox[{"n", "[", 
         RowBox[{"-", "b"}], "]"}], "]"}], "]"}]}], ")"}]}], "]"}]}]], "Input",\

 CellLabel->
  "In[854]:=",ExpressionUUID->"96303464-b6c1-4189-86ea-c82f7583a860"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`d],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{"(", 
         InterpretationBox[
          StyleBox[
           RowBox[{
            SubscriptBox["\[EmptyDownTriangle]", "c"], 
            SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {" "},
                  {"b"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}]}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`CD[-$CellContext`c][
           $CellContext`CD[-$CellContext`d][
            $CellContext`n[-$CellContext`b]]],
          Editable->False], ")"}]}], "+", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "d"], 
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"n", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`d][
         $CellContext`CD[-$CellContext`c][
          $CellContext`n[-$CellContext`b]]],
        Editable->False]}], ")"}]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[$CellContext`n[$CellContext`d] \
(-$CellContext`CD[-$CellContext`c][
       $CellContext`CD[-$CellContext`d][
        $CellContext`n[-$CellContext`b]]] + $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`n[-$CellContext`b]]])],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978393623658*^9},
 CellLabel->"Out[854]=",
 CellID->1052526960,ExpressionUUID->"82604d8c-2f0c-45f8-8797-6e9c3ae6fb53"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1A", "=", 
  RowBox[{"expr1", "//", "SortCovDs"}]}]], "Input",
 CellLabel->
  "In[855]:=",ExpressionUUID->"7b019246-29ce-4837-9166-bacd9484ccf1"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    UnderscriptBox["P", "h"], "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`n[$CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", "a"},
             {"d", "c", "b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RiemannCD[-$CellContext`d, -$CellContext`c, \
-$CellContext`b, $CellContext`a],
      Editable->False]}], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Projectormetrich[($CellContext`n[-$CellContext`a] \
$CellContext`n[$CellContext`d]) $CellContext`RiemannCD[-$CellContext`d, \
-$CellContext`c, -$CellContext`b, $CellContext`a]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978393773682*^9},
 CellLabel->"Out[855]=",
 CellID->997816381,ExpressionUUID->"b661502b-c350-47cb-a9c0-a027d7966718"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"expr1", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[856]:=",ExpressionUUID->"e04feeb5-d70a-4450-bb74-9eccfa59523d"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`n[$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`a][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`n[-$CellContext`b]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`e][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], 
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"n", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`CD[-$CellContext`d][
       $CellContext`n[-$CellContext`a]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978393944701*^9},
 CellLabel->"Out[856]=",
 CellID->600116044,ExpressionUUID->"aa650f7e-1c5c-4bab-9b42-15e032514421"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr1B", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", "ContractMetric"}], 
    "//", "GradNormalToExtrinsicK"}], "//", "Simplification"}]}]], "Input",
 CellLabel->
  "In[857]:=",ExpressionUUID->"32e52a9e-1d49-45a5-9d92-f879f1612d17"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`b],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`Accelerationn[-$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978394119581*^9},
 CellLabel->"Out[857]=",
 CellID->1567145196,ExpressionUUID->"ded4a535-a314-41fc-b461-1b21c0d1a03e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2A", "=", 
  RowBox[{
   RowBox[{"-", 
    RowBox[{
     RowBox[{"cd", "[", 
      RowBox[{"-", "c"}], "]"}], "[", 
     RowBox[{"Accelerationn", "[", 
      RowBox[{"-", "b"}], "]"}], "]"}]}], "-", 
   RowBox[{"Projectormetrich", "[", 
    RowBox[{
     RowBox[{"n", "[", "d", "]"}], 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "d"}], "]"}], "[", 
      RowBox[{"ExtrinsicKmetrich", "[", 
       RowBox[{
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "c"}]}], "]"}], "]"}]}], "]"}]}]}]], "Input",
 CellLabel->
  "In[858]:=",ExpressionUUID->"753031c4-2bc6-404e-b001-f60fbf387dba"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"d"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`d],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`d][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`d] \
$CellContext`CD[-$CellContext`d][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "c"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`c][
    $CellContext`Accelerationn[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978394275156*^9},
 CellLabel->"Out[858]=",
 CellID->122378717,ExpressionUUID->"0488b665-bbb1-4c04-8f03-fa978bcdac56"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2A", "//", "ProjectDerivative"}]], "Input",
 CellLabel->
  "In[859]:=",ExpressionUUID->"4ebd329d-6c03-4b00-8036-b7bd2a88c2f2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
           {"An", 
            StyleBox[GridBox[{
               {" "},
               {"b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`CD[-$CellContext`c][
        $CellContext`Accelerationn[-$CellContext`b]],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[
     $CellContext`CD[-$CellContext`c][
      $CellContext`Accelerationn[-$CellContext`b]]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`d],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"b", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`d][
         $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`d] \
$CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978394440625*^9},
 CellLabel->"Out[859]=",
 CellID->1677246440,ExpressionUUID->"53485bfa-2a07-433e-bebd-49cf48a8da05"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ProjectorToMetric"}], "//", "Expand"}]], "Input",
 CellLabel->
  "In[860]:=",ExpressionUUID->"074bd6ad-4f49-4582-9078-e6e4f828460c"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"n", 
         StyleBox[GridBox[{
            {"a"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`n[$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`Accelerationn[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`Accelerationn[-$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`e]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978394611408*^9},
 CellLabel->"Out[860]=",
 CellID->1610566558,ExpressionUUID->"aee152d4-dbe4-42e7-8c4a-f8eba00e148a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr2B", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"%", "//", "GradNormalToExtrinsicK"}], "//", "ContractMetric"}], "//",
    "Simplification"}]}]], "Input",
 CellLabel->
  "In[861]:=",ExpressionUUID->"fa8dec79-8466-47bb-b31d-909f0b5371d4"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`Accelerationn[-$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
            {"An", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`d][
         $CellContext`Accelerationn[-$CellContext`b]],
        Editable->False], ")"}]}], "+", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`CD[-$CellContext`d][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
      Editable->False]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783948406963`*^9},
 CellLabel->"Out[861]=",
 CellID->1908253518,ExpressionUUID->"b17c98d9-7aa0-4f70-a827-1ebf6d3e5da8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we have (note that this pair of equations is equivalent to (105) of \
Choptuik):\
\>", "MathCaption",ExpressionUUID->"259a9919-b12f-4750-8423-899d75746bdd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"GCRiemann3", "[", 
    RowBox[{
     RowBox[{"-", "b_"}], ",", 
     RowBox[{"-", "c_"}]}], "]"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{"expr1A", "-", "expr1B"}], ")"}], "-", 
     RowBox[{"(", 
      RowBox[{"expr2A", "-", "expr2B"}], ")"}]}], "//", "Simplification"}]}], 
  "]"}]], "Input",
 CellLabel->
  "In[862]:=",ExpressionUUID->"e9f7b50e-ed27-41ac-87b9-c08cc415d450"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " ", " ", " "},
              {"b", "d", "c", "a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RiemannCD[-$CellContext`b, -$CellContext`d, \
-$CellContext`c, -$CellContext`a],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[($CellContext`n[$CellContext`a] \
$CellContext`n[$CellContext`d]) $CellContext`RiemannCD[-$CellContext`b, \
-$CellContext`d, -$CellContext`c, -$CellContext`a]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`d],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"b", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`d][
         $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`d] \
$CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "c"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`c][
    $CellContext`Accelerationn[-$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978395041*^9},
 CellLabel->"Out[862]=",
 CellID->1496092728,ExpressionUUID->"625e9764-3633-4f15-afd1-072706e35e0c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"LieD", "[", 
        RowBox[{
         RowBox[{"n", "[", "a", "]"}], ",", "CD"}], "]"}], "[", 
       RowBox[{"ExtrinsicKmetrich", "[", 
        RowBox[{
         RowBox[{"-", "b"}], ",", 
         RowBox[{"-", "c"}]}], "]"}], "]"}], "-", 
      RowBox[{
       RowBox[{"n", "[", "a", "]"}], 
       RowBox[{
        RowBox[{"CD", "[", 
         RowBox[{"-", "a"}], "]"}], "[", 
        RowBox[{"ExtrinsicKmetrich", "[", 
         RowBox[{
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}]}], "]"}], "]"}]}]}], "//", 
     "GradNormalToExtrinsicK"}], "]"}], "//", "Simplification"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[863]:=",ExpressionUUID->"35e8c90f-069d-46b0-88a1-72cc74bd0f7f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "2"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"Kh", 
       StyleBox[GridBox[{
          {" ", "a"},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"Kh", 
       StyleBox[GridBox[{
          {" ", " "},
          {"c", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978395225894*^9},
 CellLabel->"Out[863]=",
 CellID->2091940420,ExpressionUUID->"2dbba7fb-936a-4bb7-b13f-71a2ef8097e0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Finally we can combine the three Gauss-Codazzi equations into a single \
expression:\
\>", "MathCaption",ExpressionUUID->"f88ae0c0-061e-4586-a235-d04b0845e738"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRiemann", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}], ",", 
    RowBox[{"-", "c_"}], ",", 
    RowBox[{"-", "d_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"inducedRiemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], "+", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "b"}], "]"}], 
     RowBox[{"n", "[", 
      RowBox[{"-", "d"}], "]"}], 
     RowBox[{"GCRiemann3", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "c"}]}], "]"}]}], "-", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "b"}], "]"}], 
     RowBox[{"n", "[", 
      RowBox[{"-", "c"}], "]"}], 
     RowBox[{"GCRiemann3", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "d"}]}], "]"}]}], "-", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"n", "[", 
      RowBox[{"-", "d"}], "]"}], 
     RowBox[{"GCRiemann3", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}]}], "+", 
    RowBox[{
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"n", "[", 
      RowBox[{"-", "c"}], "]"}], 
     RowBox[{"GCRiemann3", "[", 
      RowBox[{
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "d"}]}], "]"}]}]}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[864]:=",ExpressionUUID->"840ede90-4ef9-4736-be39-fb78275f734e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`Accelerationn[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978395575583*^9},
 CellLabel->"Out[864]=",
 CellID->640370256,ExpressionUUID->"a4bd2824-d186-4e16-9fe6-64069e4d19e0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GaussCodazziRiemannRule", "=", 
  RowBox[{"IndexRule", "[", 
   RowBox[{
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], ",", 
    RowBox[{"inducedRiemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[865]:=",ExpressionUUID->"8a8a986b-f900-4f98-ac31-a7fbad66c2b6"],

Cell[BoxData[
 RowBox[{
  RowBox[{"HoldPattern", "[", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
    Editable->False], "]"}], "\[RuleDelayed]", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "e", "}"}], ",", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"Kh", 
            StyleBox[GridBox[{
               {" ", " "},
               {"a", "d"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"b", "d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "e"},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "e"},
              {"a", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "e"},
              {"b", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"An", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Accelerationn[-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", "e"},
              {"a", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"Kh", 
           StyleBox[GridBox[{
              {" ", " "},
              {"c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         UnderscriptBox["P", "h"], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {"e"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`n[$CellContext`e],
           Editable->False], " ", 
          RowBox[{"(", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
                {"Kh", 
                 StyleBox[GridBox[{
                    {" ", " "},
                    {"a", "c"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}]}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CD[-$CellContext`e][
             $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
            Editable->False], ")"}]}], "]"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         UnderscriptBox["P", "h"], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {"e"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`n[$CellContext`e],
           Editable->False], " ", 
          RowBox[{"(", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
                {"Kh", 
                 StyleBox[GridBox[{
                    {" ", " "},
                    {"a", "d"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}]}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CD[-$CellContext`e][
             $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
            Editable->False], ")"}]}], "]"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         UnderscriptBox["P", "h"], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {"e"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`n[$CellContext`e],
           Editable->False], " ", 
          RowBox[{"(", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
                {"Kh", 
                 StyleBox[GridBox[{
                    {" ", " "},
                    {"b", "c"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}]}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CD[-$CellContext`e][
             $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
            Editable->False], ")"}]}], "]"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[
        RowBox[{
         UnderscriptBox["P", "h"], "[", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"n", 
               StyleBox[GridBox[{
                  {"e"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`n[$CellContext`e],
           Editable->False], " ", 
          RowBox[{"(", 
           InterpretationBox[
            StyleBox[
             RowBox[{
              SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
                {"Kh", 
                 StyleBox[GridBox[{
                    {" ", " "},
                    {"b", "d"}
                   },
                   GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                  FontSize->9]}
               },
               
               GridBoxAlignment->{
                "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {{Center}}, "RowsIndexed" -> {}},
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                 "Rows" -> {
                   Offset[0.2], {
                    Offset[0.4]}, 
                   Offset[0.2]}, "RowsIndexed" -> {}}]}],
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`CD[-$CellContext`e][
             $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
            Editable->False], ")"}]}], "]"}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]]],
       Editable->False]}], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "D", "]"}], 
          StyleBox[GridBox[{
             {" ", " ", " ", " "},
             {"a", "b", "c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, \
-$CellContext`c, -$CellContext`d],
      Editable->False], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "a"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"b", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`a][
         $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "a"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"b", "d"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`a][
         $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "b"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`b][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "b"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "d"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`b][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "c"], GridBox[{
            {"An", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`c][
         $CellContext`Accelerationn[-$CellContext`a]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"d"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`d],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "c"], GridBox[{
            {"An", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`c][
         $CellContext`Accelerationn[-$CellContext`b]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "d"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"b", "d"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "d"], GridBox[{
            {"An", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`d][
         $CellContext`Accelerationn[-$CellContext`a]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "d"], GridBox[{
            {"An", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`d][
         $CellContext`Accelerationn[-$CellContext`b]],
        Editable->False], ")"}]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`b],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "d"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {\

               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`d][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
        Editable->False], ")"}]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[-$CellContext`a],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["D", "d"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"b", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`cd[-$CellContext`d][
         $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
        Editable->False], ")"}]}]}]}], "]"}]}]], "Output",
 CellChangeTimes->{3.728978395845191*^9},
 CellLabel->"Out[865]=",
 CellID->1662337475,ExpressionUUID->"e7a6ae8d-bfeb-4435-85c4-83d3c5d20467"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This expression is stored in xTensor` as", "MathCaption",ExpressionUUID->"26c2e768-af0c-4ffd-a7cd-51442496fcd3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GaussCodazzi", "[", 
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], ",", "metrich"}], "]"}]], "Input",
 CellLabel->
  "In[866]:=",ExpressionUUID->"1a16bcf7-26cd-463e-b415-406b32bb908a"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"Kh", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riemanncd[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`Accelerationn[-$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783960314903`*^9},
 CellLabel->"Out[866]=",
 CellID->143694855,ExpressionUUID->"d0687f23-f08c-4a42-a85f-d1a76210c49c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "-", 
   RowBox[{"Last", "[", "GaussCodazziRiemannRule", "]"}]}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[867]:=",ExpressionUUID->"e8996e60-e964-4e5c-97d0-31bfbbc57a1b"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978396207159*^9},
 CellLabel->"Out[867]=",
 CellID->1307808757,ExpressionUUID->"39f0597e-9463-433e-bcb3-0d8461166511"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Most of the nontrivial computations in this area can now be easily performed \
using the Gauss-Codazzi formula. For example, let us compute the projections \
of the Ricci tensor and the Ricci scalar:\
\>", "Text",ExpressionUUID->"0d677ecc-eef6-4169-8761-01c3d82e6358"],

Cell[CellGroupData[{

Cell["Decompose:", "MathCaption",ExpressionUUID->"840b16f2-0e78-47ab-a8d8-a0e22a1b7b15"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRicci", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{"InducedDecomposition", "[", 
    RowBox[{
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"metrich", ",", "n"}], "}"}]}], "]"}], "//", 
   "NoScalar"}]}]], "Input",
 CellLabel->
  "In[868]:=",ExpressionUUID->"ff79c8a5-9e4c-4026-a7cc-251824b75e33"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " "},
               {"a", "c"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`RicciCD[-$CellContext`a, -$CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " "},
               {"c", "b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RicciCD[-$CellContext`c, -$CellContext`b],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`RicciCD[-$CellContext`c, -$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289783963742943`*^9},
 CellLabel->"Out[868]=",
 CellID->846651854,ExpressionUUID->"62ac2f10-41d1-4b39-bfb4-c8a961642ebf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["First formula:", "MathCaption",ExpressionUUID->"f7140a1c-592a-475f-9030-08caaa2ee72b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zero", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], "-", 
     RowBox[{
      RowBox[{"inducedRiemann", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "c"}], ",", 
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "d"}]}], "]"}], " ", 
      RowBox[{"metricg", "[", 
       RowBox[{"c", ",", "d"}], "]"}]}]}], "//", "ContractMetric"}], "//", 
   "NoScalar"}]}]], "Input",
 CellLabel->
  "In[869]:=",ExpressionUUID->"d8304f9b-982f-4b8d-9313-c4e3e9742040"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`c]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`Accelerationn[-$CellContext`a]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"c", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`c]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[$CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "d"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`d]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`e],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"c", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`e][
      $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289783965461607`*^9},
 CellLabel->"Out[869]=",
 CellID->451065156,ExpressionUUID->"d7da88f2-b68c-49fc-b571-c508739f3255"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zeroA", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", "%", "]"}], "//", 
    "ContractMetric"}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[870]:=",ExpressionUUID->"00cca68e-ee79-49ce-8deb-65efd0806cd8"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`Accelerationn[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978396741169*^9},
 CellLabel->"Out[870]=",
 CellID->434802982,ExpressionUUID->"c7675050-1c19-4a0c-94e8-3507999b7ee1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zeroB", "=", 
  RowBox[{
   RowBox[{"Projectormetrich", "[", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], "-", 
   RowBox[{
    RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[871]:=",ExpressionUUID->"a5cad8a6-1ec9-4527-bb12-ed79fa8b5433"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "d"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.72897839690727*^9},
 CellLabel->"Out[871]=",
 CellID->1994351380,ExpressionUUID->"abad15aa-0c6d-4a26-a3ec-1a92432d2acc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"GCRicci1", "[", 
    RowBox[{
     RowBox[{"-", "a_"}], ",", 
     RowBox[{"-", "b_"}]}], "]"}], ",", 
   RowBox[{
    RowBox[{"zeroA", "+", "zeroB"}], "//", "ToCanonical"}]}], "]"}]], "Input",\

 CellLabel->
  "In[872]:=",ExpressionUUID->"a6e399f1-7f72-4460-a806-18f8f0c9548e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`Accelerationn[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978397089267*^9},
 CellLabel->"Out[872]=",
 CellID->1138881284,ExpressionUUID->"ed928bdf-2d7c-494b-8f75-8cb478ce1afd"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRicci", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"inducedRicci", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "-", 
    RowBox[{"GCRicci1", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[873]:=",ExpressionUUID->"46a818b0-d7a9-4919-a106-a859471f72e0"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " "},
               {"a", "c"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`RicciCD[-$CellContext`a, -$CellContext`c]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {
            RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
            StyleBox[GridBox[{
               {" ", " "},
               {"b", "c"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`RicciCD[-$CellContext`b, -$CellContext`c],
        Editable->False]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`RicciCD[-$CellContext`b, -$CellContext`c]],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`Accelerationn[-$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978397258284*^9},
 CellLabel->"Out[873]=",
 CellID->655994477,ExpressionUUID->"ee5c52bc-04ad-47d8-8b28-f78264bbdd1d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Second formula:", "MathCaption",ExpressionUUID->"81a4acd7-75b9-4195-9791-897e5adf300e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zero", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"RicciCD", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "c"}]}], "]"}], "-", 
       RowBox[{
        RowBox[{"inducedRiemann", "[", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}], " ", 
        RowBox[{"metricg", "[", 
         RowBox[{"b", ",", "d"}], "]"}]}]}], ")"}], 
     RowBox[{"n", "[", "c", "]"}]}], "//", "ContractMetric"}], "//", 
   "NoScalar"}]}]], "Input",
 CellLabel->
  "In[874]:=",ExpressionUUID->"e550f6b2-7140-4743-a5b0-3e8b57d9c880"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "a"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`a][
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`b]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`b]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978398134104*^9},
 CellLabel->"Out[874]=",
 CellID->118131241,ExpressionUUID->"d913852b-3bcb-411b-b372-b09b18b61ff1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zeroA", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", "%", "]"}], "//", 
    "MetricToProjector"}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[875]:=",ExpressionUUID->"9fe55e4a-aeef-44d0-810b-0291e6fa41cf"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "a"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`a][
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`b]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978398616543*^9},
 CellLabel->"Out[875]=",
 CellID->1156877659,ExpressionUUID->"081e2fb6-ea4c-4646-930b-4eb0cba683e8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zeroB", "=", 
  RowBox[{
   RowBox[{"Projectormetrich", "[", 
    RowBox[{
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"n", "[", "b", "]"}]}], "]"}], "-", 
   RowBox[{
    RowBox[{"ProjectWith", "[", "metrich", "]"}], "[", 
    RowBox[{
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], 
     RowBox[{"n", "[", "b", "]"}]}], "]"}]}]}]], "Input",
 CellLabel->
  "In[876]:=",ExpressionUUID->"ecd40e86-b7d2-4136-980c-8c6491a8528a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`b] \
$CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`b],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.72897839882401*^9},
 CellLabel->"Out[876]=",
 CellID->325874875,ExpressionUUID->"257a8b54-46fe-4b66-bcfb-724b6bdc7f2a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexSet", "[", 
  RowBox[{
   RowBox[{"GCRicci2", "[", 
    RowBox[{"-", "a_"}], "]"}], ",", 
   RowBox[{
    RowBox[{"zeroA", "+", "zeroB"}], "//", "ToCanonical"}]}], "]"}]], "Input",\

 CellLabel->
  "In[877]:=",ExpressionUUID->"4cfcf066-584f-4f95-a167-3940d0a0a937"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {
           RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
           StyleBox[GridBox[{
              {" ", " "},
              {"a", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`b] \
$CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "a"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`a][
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`b]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978399317628*^9},
 CellLabel->"Out[877]=",
 CellID->895724275,ExpressionUUID->"bd8490b9-4563-4c80-a864-f295e1024b74"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRicci", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{"inducedRicci", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "+", 
    RowBox[{
     RowBox[{"GCRicci2", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"n", "[", 
      RowBox[{"-", "b"}], "]"}]}], "+", 
    RowBox[{
     RowBox[{"GCRicci2", "[", 
      RowBox[{"-", "b"}], "]"}], 
     RowBox[{"n", "[", 
      RowBox[{"-", "a"}], "]"}]}]}], "//", "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[878]:=",ExpressionUUID->"782af030-0090-4d47-9310-dad2d290545b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {"c", " "},
             {" ", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`Accelerationn[-$CellContext`a]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {"c", " "},
             {" ", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289784005011063`*^9},
 CellLabel->"Out[878]=",
 CellID->577524455,ExpressionUUID->"1fa67787-b2d9-4a13-b06c-b0661c467a86"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Third formula:", "MathCaption",ExpressionUUID->"7cafe600-619f-44a2-876b-4e6ec3f56a8f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"zero", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"RicciCD", "[", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "c"}]}], "]"}], "-", 
       RowBox[{
        RowBox[{"inducedRiemann", "[", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}], " ", 
        RowBox[{"metricg", "[", 
         RowBox[{"b", ",", "d"}], "]"}]}]}], ")"}], 
     RowBox[{"n", "[", "a", "]"}], 
     RowBox[{"n", "[", "c", "]"}]}], "//", "Expand"}], "//", 
   "ReplaceDummies"}]}]], "Input",
 CellLabel->
  "In[879]:=",ExpressionUUID->"fe6df6d1-e6bd-443c-9d3b-f65adc023569"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "c"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`c],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"a", "b"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`c][
          $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
       $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`Accelerationn[-$CellContext`a]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978401296364*^9},
 CellLabel->"Out[879]=",
 CellID->1974914392,ExpressionUUID->"0c66d77e-1827-4ddb-b530-01fcb8faa468"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRicci", "[", 
   RowBox[{
    RowBox[{"-", "a_"}], ",", 
    RowBox[{"-", "b_"}]}], "]"}], "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"inducedRicci", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}], "-", 
     RowBox[{
      RowBox[{"n", "[", 
       RowBox[{"-", "a"}], "]"}], 
      RowBox[{"n", "[", 
       RowBox[{"-", "b"}], "]"}], "zero"}]}], "//", "Expand"}], "//", 
   "SameDummies"}]}]], "Input",
 CellLabel->
  "In[880]:=",ExpressionUUID->"7bc9ed29-8571-4b80-bfbb-45a4758a1b7e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"c", " "},
           {" ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "e"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     UnderscriptBox["P", "h"], "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"n", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`n[$CellContext`c],
       Editable->False], " ", 
      RowBox[{"(", 
       InterpretationBox[
        StyleBox[
         RowBox[{
          SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
            {"Kh", 
             StyleBox[GridBox[{
                {" ", " "},
                {"a", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}]}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`CD[-$CellContext`c][
         $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
        Editable->False], ")"}]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Projectormetrich[$CellContext`n[$CellContext`c] \
$CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]]],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      UnderscriptBox["P", "h"], "[", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"n", 
            StyleBox[GridBox[{
               {"e"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`n[$CellContext`e],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
             {"Kh", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"c", "d"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`e][
          $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d]],
         Editable->False], ")"}]}], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Projectormetrich[$CellContext`n[$CellContext`e] \
$CellContext`CD[-$CellContext`e][
       $CellContext`ExtrinsicKmetrich[-$CellContext`c, -$CellContext`d]]],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "D", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Riccicd[-$CellContext`a, -$CellContext`b],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {"c", " "},
             {" ", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c]],
     Editable->False], ")"}]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "b"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`b][
    $CellContext`Accelerationn[-$CellContext`a]],
   Editable->False], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "b"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {"c", " "},
             {" ", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`b][
      $CellContext`ExtrinsicKmetrich[$CellContext`c, -$CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`c]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[-$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "d"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`d][
      $CellContext`Accelerationn[-$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289784035664587`*^9},
 CellLabel->"Out[880]=",
 CellID->878938312,ExpressionUUID->"1c656b1a-f937-4775-93e0-2267eaa9c6d0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Finally, the Ricci scalar:", "MathCaption",ExpressionUUID->"a2593739-009d-487e-b60e-cf1e284e7510"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"inducedRiemann", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}], 
    RowBox[{"metricg", "[", 
     RowBox[{"b", ",", "d"}], "]"}], 
    RowBox[{"metricg", "[", 
     RowBox[{"a", ",", "c"}], "]"}]}], "//", "ContractMetric"}], "//", 
  "NoScalar"}]], "Input",
 CellLabel->
  "In[881]:=",ExpressionUUID->"2dee5a99-82d4-4a12-985f-4cecd091e514"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"An", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Accelerationn[-$CellContext`a],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, -$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"a", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "b"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", "a"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`a],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "D", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarcd[],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "c"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`c][
    $CellContext`Accelerationn[$CellContext`c]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["D", "d"], GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`cd[-$CellContext`d][
    $CellContext`Accelerationn[$CellContext`d]],
   Editable->False], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978403720858*^9},
 CellLabel->"Out[881]=",
 CellID->1923685718,ExpressionUUID->"ddd35677-78e6-453e-a7dc-7a1eacb0ad5f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"%", "/.", 
    RowBox[{"Projectormetrich", "\[Rule]", 
     RowBox[{"ProjectWith", "[", "metrich", "]"}]}]}], "//", 
   "ContractMetric"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[882]:=",ExpressionUUID->"73cbf227-43fc-44eb-b0e8-5b85dce15de3"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "D", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarcd[],
   Editable->False], "-", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`Accelerationn[$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"h", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metrich[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978403847693*^9},
 CellLabel->"Out[882]=",
 CellID->1019556841,ExpressionUUID->"f76f7d1f-7aab-4ac1-a779-3c72de569680"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ProjectorToMetric"}], "//", "ContractMetric"}]], "Input",\

 CellLabel->
  "In[883]:=",ExpressionUUID->"97b21467-7a2b-4450-8256-6bd2d1fb2a61"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "D", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarcd[],
   Editable->False], "-", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`Accelerationn[$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "b"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, -$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978403991784*^9},
 CellLabel->"Out[883]=",
 CellID->919807274,ExpressionUUID->"bce6c191-e62e-44d0-a816-89b4e76a8fb6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "GradNormalToExtrinsicK"}]], "Input",
 CellLabel->
  "In[884]:=",ExpressionUUID->"4d210fe2-7625-4f00-9a5b-1461713404c8"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", "2"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"An", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Accelerationn[$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "D", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarcd[],
   Editable->False], "-", 
  RowBox[{"2", " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["D", "a"], GridBox[{
         {"An", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`cd[-$CellContext`a][
      $CellContext`Accelerationn[$CellContext`a]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "b"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978404158909*^9},
 CellLabel->"Out[884]=",
 CellID->1613563012,ExpressionUUID->"e17638fa-bbfd-47b8-a41b-4403f005172a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["With a lapse:", "MathCaption",ExpressionUUID->"f2d8ec63-aac1-4d5f-bcc1-18c7f50dd3ee"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"inducedRicciScalar", "[", "]"}], "=", 
  RowBox[{
   RowBox[{"%", "/.", "LapseRule"}], "//", "Expand"}]}]], "Input",
 CellLabel->
  "In[885]:=",ExpressionUUID->"d5c5e2c7-392f-4f60-ba8b-cf34ffed42e1"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, $CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`a, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"Kh", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`ExtrinsicKmetrich[$CellContext`b, -$CellContext`b],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{"R", "[", "D", "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciScalarcd[],
   Editable->False], "-", 
  FractionBox[
   RowBox[{"2", " ", 
    RowBox[{"(", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        SubscriptBox["D", "a"], 
        SuperscriptBox["D", "a"], "\[Alpha]"}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`cd[-$CellContext`a][
       $CellContext`cd[$CellContext`a][
        $CellContext`lapse[]]],
      Editable->False], ")"}]}], 
   InterpretationBox[
    StyleBox["\[Alpha]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`lapse[],
    Editable->False]], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"n", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`n[$CellContext`a],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"Kh", 
          StyleBox[GridBox[{
             {" ", "b"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`ExtrinsicKmetrich[-$CellContext`b, $CellContext`b]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289784043034782`*^9},
 CellLabel->"Out[885]=",
 CellID->730048368,ExpressionUUID->"42b2259e-b441-4ec8-b82a-fa8729aedd4d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metrich", "]"}]], "Input",
 CellLabel->
  "In[886]:=",ExpressionUUID->"cfe5a1b5-e3c7-4492-a842-5e13e20197d3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefInertHead", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"projector inert-head\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Projectormetrich"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefInertHead, ": Undefined ", "projector inert-head", 
   " ", $CellContext`Projectormetrich],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784044360857`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1348845665,ExpressionUUID->"a1daf3cf-3275-4173-ad7f-84adbdc3224f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"acceleration vector\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Accelerationn"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "acceleration vector", 
   " ", $CellContext`Accelerationn],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784044578857`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1211713923,ExpressionUUID->"dba1e8ac-8097-44f4-8195-fe4e04366015"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"extrinsic curvature tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "ExtrinsicKmetrich"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "extrinsic curvature tensor", " ", $CellContext`ExtrinsicKmetrich],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404476974*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->688479360,ExpressionUUID->"906a668e-52d3-494e-ae3e-8ed2024b41c8"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetrich"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetrich],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784044962397`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->630713484,ExpressionUUID->"351d2b2d-d690-484b-829e-32bc60adb435"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Christoffelcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`Christoffelcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404515152*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->422243996,ExpressionUUID->"c29efb97-ff77-417b-adbf-06401133cb5f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Einsteincd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Einstein tensor", 
   " ", $CellContext`Einsteincd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404533783*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1368871084,ExpressionUUID->"27749748-81c6-4a26-8ee4-4084490cb60a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Kretschmanncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`Kretschmanncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784045530863`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->623998477,ExpressionUUID->"f3f0c654-6f8c-4d55-a21f-483df6573720"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Riccicd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`Riccicd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404572027*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1379754638,ExpressionUUID->"045fa821-87e5-4676-99d5-fd57823452ce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404590831*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1007780560,ExpressionUUID->"5270c5ef-7dbc-4772-9d04-ab40434fa0d9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Riemanncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`Riemanncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784046100397`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->564348031,ExpressionUUID->"40215329-6b9d-4264-bb1a-37b59a1bc047"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRiccicd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "TFRicci tensor", 
   " ", $CellContext`TFRiccicd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784046304502`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1709571865,ExpressionUUID->"ab8d08d8-3d03-4154-968f-15b41c661b6d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Torsioncd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`Torsioncd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784046486397`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1590621842,ExpressionUUID->"ceccc465-1ce8-4751-9575-b98d34b54c89"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Weylcd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`Weylcd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784046668987`*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1704860638,ExpressionUUID->"0bba2ed3-70e9-46ab-99bd-e345404720f9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "cd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`cd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404684696*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1865924592,ExpressionUUID->"876235e2-6882-46c9-8519-4d4dfc1b8e36"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetrich"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetrich],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404702476*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->890920603,ExpressionUUID->"61029f9f-7446-4122-a09c-2591ad3da720"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metrich"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metrich],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404721505*^9},
 CellLabel->"During evaluation of In[886]:=",
 CellID->1446587737,ExpressionUUID->"362872ce-b693-4c81-8bd8-658854257998"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"V", ",", "W"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[887]:=",ExpressionUUID->"00fea31e-a751-4b23-aa25-837a74582d36"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "V"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`V],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784049168577`*^9},
 CellLabel->"During evaluation of In[887]:=",
 CellID->1137728072,ExpressionUUID->"99ee8363-bb64-424e-a10e-af7a08b073ae"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978404935811*^9},
 CellLabel->"During evaluation of In[887]:=",
 CellID->601313216,ExpressionUUID->"de4c7b3f-91b2-44f0-8157-deabe2c101ef"]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "zero", ",", "zeroA", ",", "zeroB", ",", "expr1", ",", "expr1A", ",", 
   "expr1B", ",", "expr2", ",", "expr2A", ",", "expr2B", ",", "GCRiemann1", 
   ",", "GCRiemann2", ",", "GCRiemann3", ",", "GCRicci1", ",", "GCRicci2", 
   ",", "GaussCodazziRiemannRule", ",", "LapseRule", ",", "inducedRiemann", 
   ",", "inducedRicci", ",", "inducedRicciScalar"}], "]"}]], "Input",
 CellLabel->
  "In[888]:=",ExpressionUUID->"8c77d418-24a4-4ef1-b3dd-85ecd03e42ce"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.9. Products of epsilon tensors", "Subsection",ExpressionUUID->"d1713bc0-fc25-4242-bac3-eef50b2341d3"],

Cell["\<\
In many applications we need to manipulate products of two epsilon tensors. \
They are automatically converted into products of metrics:\
\>", "Text",ExpressionUUID->"034a42b1-5182-440f-9fcb-313eff8765db"],

Cell[CellGroupData[{

Cell["Products without contracted indices:", "MathCaption",ExpressionUUID->"c994e0d0-b499-45d5-85b4-a2b781554c77"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], 
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}], ",", "f"}], "]"}]}]], "Input",
 CellLabel->
  "In[889]:=",ExpressionUUID->"24d53a9b-4933-4e48-851c-9d43d5b4ab2e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`c, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "f"},
           {"b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[-$CellContext`b, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`c, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`c, -$CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978410394332*^9},
 CellLabel->"Out[889]=",
 CellID->678200266,ExpressionUUID->"d3bdb3ac-cb3e-47a3-97d2-ec0cfc1d4686"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Or with contracted indices:", "MathCaption",ExpressionUUID->"38f28c85-ed91-4076-91e3-933b38272b6e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{
    RowBox[{"-", "d"}], ",", 
    RowBox[{"-", "e"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[890]:=",ExpressionUUID->"c806032d-0300-4624-a43f-b6dd0b482836"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`a, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`b, -$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {"b", " "},
           {" ", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`delta[$CellContext`b, -$CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978410550702*^9},
 CellLabel->"Out[890]=",
 CellID->1816156033,ExpressionUUID->"431bba4c-d5bf-44dc-b6d3-59ff628d554a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "e"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[891]:=",ExpressionUUID->"03ea2de4-1d75-469c-b44c-d7041b914035"],

Cell[BoxData[
 RowBox[{"2", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[Delta]", 
       StyleBox[GridBox[{
          {"a", " "},
          {" ", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`delta[$CellContext`a, -$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978410687214*^9},
 CellLabel->"Out[891]=",
 CellID->1949371820,ExpressionUUID->"e77bbeba-5a2b-4e56-a360-206b475920f5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{"epsilonmetricg", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[892]:=",ExpressionUUID->"0834c7a2-cc39-41d5-af02-f4c87280142e"],

Cell[BoxData[
 RowBox[{"-", "6"}]], "Output",
 CellChangeTimes->{3.728978410868333*^9},
 CellLabel->"Out[892]=",
 CellID->1081660108,ExpressionUUID->"86c27222-5915-492d-9d05-d87703c9fc9d"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.10. Implosion and explosion (II)", "Subsection",ExpressionUUID->"0b31d9ad-ad3c-4de9-94e2-36a87c8f4864"],

Cell["\<\
Implosion of a tensor is only possible directly when its indices have the \
correct characters (those specified at definition time). In all other cases \
we need to pull out a number of metric factors first, to correct them.\
\>", "Text",ExpressionUUID->"02d733ef-a4e6-4656-ab0a-774274031576"],

Cell[CellGroupData[{

Cell["\<\
The tensor v was defined with an upper index. Hence here we get an additional \
derivative of the metric, which is also imploded:\
\>", "MathCaption",ExpressionUUID->"1e53bbbf-b600-4454-9dd2-107d62ad823c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Cd", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"v", "[", 
   RowBox[{"-", "b"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[893]:=",ExpressionUUID->"da42ec84-f369-4972-b054-adbc1c658512"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`v[-$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289784136530533`*^9},
 CellLabel->"Out[893]=",
 CellID->882303186,ExpressionUUID->"90525cd4-c2ea-4827-b6dc-720f1ca6773f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Implode", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[894]:=",ExpressionUUID->"439f5cd8-d7a9-45c9-8957-548c6d1c1986"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[EmptyDownTriangle]v", 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`Cdv[-$CellContext`b, -$CellContext`a],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[EmptyDownTriangle]g", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`Cdmetricg[-$CellContext`c, -$CellContext`b, -$CellContext`a],
    
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289784138178587`*^9},
 CellLabel->"Out[894]=",
 CellID->495064567,ExpressionUUID->"c4b69e7b-7653-4b30-8ddd-7de29d24993e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[895]:=",ExpressionUUID->"34241b67-3055-47c1-b88d-04958f98dd06"],

Cell["Cdv[-b, -a] + Cdmetricg[-j$31537, -b, -a]*v[j$31537]", "Output",
 CellChangeTimes->{3.72897841393425*^9},
 CellLabel->"Out[895]//InputForm=",
 CellID->1198529091,ExpressionUUID->"dcc9dc85-880f-4fc8-9318-626eb6e610ee"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Explosion now does not return to the original expression:", \
"MathCaption",ExpressionUUID->"1ed5ddbd-95a4-47a4-acb1-1267bb68120e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Explode", "[", "%%", "]"}]], "Input",
 CellLabel->
  "In[896]:=",ExpressionUUID->"3c51bc16-4e64-4166-b99a-c255451d6385"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {" ", " "},
             {"c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`metricg[-$CellContext`c, -$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`c, -$CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`Cd[-$CellContext`a][
      $CellContext`v[$CellContext`c]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978414071271*^9},
 CellLabel->"Out[896]=",
 CellID->281823577,ExpressionUUID->"a15648d6-c9a1-4f7d-ba1d-0155b786b943"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["but this expression is equivalent to the original one:", "MathCaption",ExpressionUUID->"cc955863-7460-4f02-938f-8cf4558a70ae"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{"%", ",", 
   RowBox[{"OverDerivatives", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[897]:=",ExpressionUUID->"68337242-61d6-471a-b2d0-b084173455c3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`Cd[-$CellContext`a][
   $CellContext`v[-$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978414202918*^9},
 CellLabel->"Out[897]=",
 CellID->1407417004,ExpressionUUID->"f47ad45e-dda9-42b5-bd46-1f115bee5b92"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.11. Variational derivatives", "Subsection",ExpressionUUID->"5f03f11f-137e-4963-9f65-9721950bf5ba"],

Cell[TextData[{
 "There is not yet a concept of integration in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ". Instead of working with a variational derivative of a functional, we \
shall assume that such a functional is the integral of some integrand (or \
\"Lagrangian\") and that the integral is computed with respect to a volume \
form which vanishes under the action of some derivative (this is the \
derivative which will be \"integrated by parts\")."
}], "Text",ExpressionUUID->"486b539d-d3b4-475a-8d82-9f2ea778f31b"],

Cell[CellGroupData[{

Cell["\<\
For instance, let us compute the RicciScalar field of our metric:\
\>", "MathCaption",ExpressionUUID->"e54c0851-8430-4ba0-ae7a-05a854d4369e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rs", "=", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"RicciScalarCD", "[", "]"}], "//", "RiemannToChristoffel"}], "//",
      "NoScalar"}], "//", "ChristoffelToGradMetric"}], "//", 
   "Expand"}]}]], "Input",
 CellLabel->
  "In[898]:=",ExpressionUUID->"3e609e06-a845-4f2f-9378-f8efdd464ebe"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "2"]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`metricg[-$CellContext`c, -$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`metricg[-$CellContext`b, -$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], 
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`metricg[-$CellContext`c, -$CellContext`b]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`c, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`b, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`a, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`b, -$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`metricg[-$CellContext`b, -$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`metricg[-$CellContext`a, -$CellContext`d]]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`metricg[-$CellContext`a, -$CellContext`b]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`c, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`c, -$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`a, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`c, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`a, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`c, -$CellContext`b]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`c, -$CellContext`b]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`c, -$CellContext`b]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289784175273*^9},
 CellLabel->"Out[898]=",
 CellID->1653966926,ExpressionUUID->"6f5c5225-2fcb-4171-b3be-bb16a3eebaa6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Working intensively with partial derivatives, it is interesting to set\
\>", "MathCaption",ExpressionUUID->"3824b418-a4de-4e98-b575-ff636945cd69"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"ToCanonical", ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}]}], "]"}]], "Input",
 CellLabel->
  "In[899]:=",ExpressionUUID->"f8d9d3aa-eff5-46c9-8477-43be4fe1c059"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Verbose", "\[Rule]", "False"}], ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "None"}], ",", 
   RowBox[{"Method", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"ChangeCovD", ",", 
      RowBox[{"\<\"ExpandChristoffel\"\>", "\[Rule]", "False"}]}], "}"}]}], 
   ",", 
   RowBox[{"MathLink", "\[RuleDelayed]", "$xpermQ"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978417752191*^9},
 CellLabel->"Out[899]=",
 CellID->1193798514,ExpressionUUID->"cf58506c-23fd-4500-998a-07cb5058ef91"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we perform a direct variational derivative of the Einstein-Hilbert \
Lagrangian with respect to the inverse metric. It is a long computation:\
\>", "MathCaption",ExpressionUUID->"44621bfd-a70e-417f-8307-5c8813cfaf44"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"Length", "[", 
   RowBox[{"result", "=", 
    RowBox[{"ToCanonical", "@", 
     RowBox[{
      RowBox[{"VarD", "[", 
       RowBox[{
        RowBox[{"metricg", "[", 
         RowBox[{"a", ",", "b"}], "]"}], ",", "PD"}], "]"}], "[", 
      RowBox[{
       RowBox[{"Sqrt", "[", 
        RowBox[{"-", 
         RowBox[{"Detmetricg", "[", "]"}]}], "]"}], "rs"}], "]"}]}]}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[900]:=",ExpressionUUID->"bbf44674-b511-4f6e-8b27-5c4fb3f07483"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"10.069411`", ",", "20"}], "}"}]], "Output",
 CellChangeTimes->{3.7289784279547*^9},
 CellLabel->"Out[900]=",
 CellID->231910877,ExpressionUUID->"08a14196-94f4-4b0d-8bb1-c77d208c39b6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The expected result is the Einstein tensor:", "MathCaption",ExpressionUUID->"82a7904d-6421-4e39-acca-6c5f18839a60"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"EinsteinCD", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}]}], "]"}], "//", "EinsteinToRicci"}], "//", 
     "RiemannToChristoffel"}], "//", "NoScalar"}], "//", 
   "ChristoffelToGradMetric"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[901]:=",ExpressionUUID->"ac935daa-2601-46b3-8f1b-663835db5f89"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`d, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`metricg[-$CellContext`c, -$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`b, -$CellContext`c]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`a, -$CellContext`c]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"e", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`e, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`metricg[-$CellContext`b, -$CellContext`c]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`metricg[-$CellContext`a, -$CellContext`c]]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], 
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`metricg[-$CellContext`a, -$CellContext`b]]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`b, -$CellContext`f]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`c]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`c]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`b, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`metricg[-$CellContext`b, -$CellContext`c]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`metricg[-$CellContext`a, -$CellContext`c]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`a, -$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "8"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"g", "h"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`g, $CellContext`h],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"g", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`g, -$CellContext`h]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], 
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`metricg[-$CellContext`c, -$CellContext`e]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], 
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     xAct`xTensor`PD[-$CellContext`e][
      $CellContext`metricg[-$CellContext`c, -$CellContext`d]]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"g", "h"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`g, $CellContext`h],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "f"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "h"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`f][
     $CellContext`metricg[-$CellContext`d, -$CellContext`h]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "g"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`g][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["3", "8"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"g", "h"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`g, $CellContext`h],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "g"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`g][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "h"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "f"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`h][
     $CellContext`metricg[-$CellContext`d, -$CellContext`f]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"g", "h"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`g, $CellContext`h],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`c, -$CellContext`e]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "h"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"f", "g"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`h][
     $CellContext`metricg[-$CellContext`f, -$CellContext`g]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"e", "f"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`e, $CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"g", "h"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`g, $CellContext`h],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "h"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"f", "g"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`h][
     $CellContext`metricg[-$CellContext`f, -$CellContext`g]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978428803789*^9},
 CellLabel->"Out[901]=",
 CellID->43290434,ExpressionUUID->"c02f7a4b-966e-4bec-a5c9-89bfe1c58bc8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"result", "-", 
   RowBox[{
    RowBox[{"Sqrt", "[", 
     RowBox[{"-", 
      RowBox[{"Detmetricg", "[", "]"}]}], "]"}], "%"}]}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[902]:=",ExpressionUUID->"2e64f155-5441-4cc6-bdf3-bb3ad6e7c5f1"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289784291577806`*^9},
 CellLabel->"Out[902]=",
 CellID->1732851552,ExpressionUUID->"1af94ba1-4bc8-4d48-8021-57beb8f73603"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us now define the electromagnetic fields:", "MathCaption",ExpressionUUID->"9b8c4ae5-3b0b-4000-89cb-76c82e59f15f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"MaxwellA", "[", 
    RowBox[{"-", "a"}], "]"}], ",", "M3", ",", 
   RowBox[{"PrintAs", "->", "\"\<A\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[903]:=",ExpressionUUID->"090bef6b-e5d5-4d9f-b7ef-362f3aa23e2d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"MaxwellA", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`MaxwellA[-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978429707572*^9},
 CellLabel->"During evaluation of In[903]:=",
 CellID->2010297497,ExpressionUUID->"d6f3c3f6-b6ae-49d6-8395-be798d5e63d0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"MaxwellF", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}], ",", 
   RowBox[{"PrintAs", "->", "\"\<F\>\""}]}], "]"}]], "Input",
 CellLabel->
  "In[904]:=",ExpressionUUID->"f02d4fbf-75b7-42aa-9904-ad575ba8b4b6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"MaxwellF", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`MaxwellF[-$CellContext`a, -$CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978430058198*^9},
 CellLabel->"During evaluation of In[904]:=",
 CellID->709813828,ExpressionUUID->"04283114-1508-4877-9f5a-6c3662c8c233"]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"MaxwellF", "[", 
   RowBox[{"a_", ",", "b_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"PD", "[", "a", "]"}], "[", 
    RowBox[{"MaxwellA", "[", "b", "]"}], "]"}], "-", 
   RowBox[{
    RowBox[{"PD", "[", "b", "]"}], "[", 
    RowBox[{"MaxwellA", "[", "a", "]"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[905]:=",ExpressionUUID->"0f1202db-3c09-4dd1-867c-1d071ff6ca78"]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is the electromagnetic Lagrangian:", "MathCaption",ExpressionUUID->"481cadbe-1cad-4943-883c-cee3cf243bdf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"metricg", "[", 
   RowBox[{"a", ",", "b"}], "]"}], 
  RowBox[{"metricg", "[", 
   RowBox[{"c", ",", "d"}], "]"}], 
  RowBox[{"MaxwellF", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "c"}]}], "]"}], 
  RowBox[{
   RowBox[{"MaxwellF", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], "/", "4"}]}]], "Input",
 CellLabel->
  "In[906]:=",ExpressionUUID->"008507e6-c9a9-4b55-8229-c2903eb287ea"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"c", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "a"], GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`MaxwellA[-$CellContext`c]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "c"], GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`c][
      $CellContext`MaxwellA[-$CellContext`a]],
     Editable->False]}], ")"}], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "b"], GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" "},
             {"d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`MaxwellA[-$CellContext`d]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[PartialD]", "d"], GridBox[{
         {"A", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`PD[-$CellContext`d][
      $CellContext`MaxwellA[-$CellContext`b]],
     Editable->False]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728978430322659*^9},
 CellLabel->"Out[906]=",
 CellID->569348111,ExpressionUUID->"5b0d2c69-f8bc-446c-8ed7-d57b68a637fc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["and these are the Maxwell equations on a curved background:", \
"MathCaption",ExpressionUUID->"efb7a19c-7f8c-4ed5-abe4-4c2e8a8f2729"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"result", "=", 
  RowBox[{
   RowBox[{"VarD", "[", 
    RowBox[{
     RowBox[{"MaxwellA", "[", "a", "]"}], ",", "PD"}], "]"}], "[", 
   RowBox[{
    RowBox[{"Sqrt", "[", 
     RowBox[{"-", 
      RowBox[{"Detmetricg", "[", "]"}]}], "]"}], "%"}], "]"}]}]], "Input",
 CellLabel->
  "In[907]:=",ExpressionUUID->"d639e9b6-ea71-4fbf-b6b2-63f5418350f9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`MaxwellA[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`MaxwellA[-$CellContext`b]]],
    Editable->False]}], "-", 
  RowBox[{
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`MaxwellA[-$CellContext`a]]],
    Editable->False]}], "-", 
  RowBox[{
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   SqrtBox[
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[
       OverscriptBox["g", 
        OverscriptBox[
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]], 
         StyleBox["~",
          FontColor->RGBColor[0, 0, 0]]]],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`Detmetricg[],
      Editable->False]}]], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`MaxwellA[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978430819104*^9},
 CellLabel->"Out[907]=",
 CellID->1994054192,ExpressionUUID->"cc9bac43-6a0c-4888-988d-b941f05f0b77"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"metricg", "[", 
      RowBox[{"b", ",", "c"}], "]"}], 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "b"}], "]"}], "[", 
      RowBox[{"MaxwellF", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "c"}]}], "]"}], "]"}]}], "//", "ChangeCovD"}], "//", 
   "ChristoffelToGradMetric"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[908]:=",ExpressionUUID->"1ba91924-9c4e-41af-b6e2-29df80c4503a"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`MaxwellA[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`metricg[-$CellContext`d, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`a][
      $CellContext`MaxwellA[-$CellContext`b]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], 
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     xAct`xTensor`PD[-$CellContext`b][
      $CellContext`MaxwellA[-$CellContext`a]]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "d"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`d][
     $CellContext`metricg[-$CellContext`a, -$CellContext`e]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"c", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`c, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "c"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`c][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`a, -$CellContext`d]],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "a"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`a][
     $CellContext`MaxwellA[-$CellContext`b]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"g", 
        StyleBox[GridBox[{
           {"d", "e"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`metricg[$CellContext`d, $CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "b"], GridBox[{
        {"A", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`b][
     $CellContext`MaxwellA[-$CellContext`a]],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[PartialD]", "e"], GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"c", "d"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`PD[-$CellContext`e][
     $CellContext`metricg[-$CellContext`c, -$CellContext`d]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978431082127*^9},
 CellLabel->"Out[908]=",
 CellID->439009862,ExpressionUUID->"972bddae-1d34-41ec-9933-779e657c11e5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"result", "-", 
   RowBox[{
    RowBox[{"Sqrt", "[", 
     RowBox[{"-", 
      RowBox[{"Detmetricg", "[", "]"}]}], "]"}], "%"}]}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[909]:=",ExpressionUUID->"94d9ef2c-c451-4910-aed8-8ff39da3ee37"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978431257518*^9},
 CellLabel->"Out[909]=",
 CellID->1832423189,ExpressionUUID->"cea9ddfc-6c30-497a-bd83-1bfb92b3c6d8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"MaxwellA", ",", "MaxwellF"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[910]:=",ExpressionUUID->"f4109034-4633-4018-9a05-077300cd3613"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "MaxwellA"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`MaxwellA],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897843140491*^9},
 CellLabel->"During evaluation of In[910]:=",
 CellID->473555896,ExpressionUUID->"badb77ec-5b1b-4f6f-baa1-4104f5d65f53"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "MaxwellF"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`MaxwellF],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897843142225*^9},
 CellLabel->"During evaluation of In[910]:=",
 CellID->479548290,ExpressionUUID->"668905b2-5b5d-472c-b923-5708f4b88f93"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Restore standard options:", "MathCaption",ExpressionUUID->"2cd10378-4151-462a-a455-63db2f7f25bf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetOptions", "[", 
  RowBox[{"ToCanonical", ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "All"}]}], "]"}]], "Input",
 CellLabel->
  "In[911]:=",ExpressionUUID->"1a9b2cea-9c0f-4ec3-b462-cf27abac59a8"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Verbose", "\[Rule]", "False"}], ",", 
   RowBox[{"UseMetricOnVBundle", "\[Rule]", "All"}], ",", 
   RowBox[{"Method", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"ChangeCovD", ",", 
      RowBox[{"\<\"ExpandChristoffel\"\>", "\[Rule]", "False"}]}], "}"}]}], 
   ",", 
   RowBox[{"MathLink", "\[RuleDelayed]", "$xpermQ"}], ",", 
   RowBox[{"TimeVerbose", "\[Rule]", "False"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978431633123*^9},
 CellLabel->"Out[911]=",
 CellID->261982839,ExpressionUUID->"b935d184-ded2-4fbb-bd3c-2ff7e9d7eccf"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["7.12. Clean up", "Subsection",ExpressionUUID->"d6923f02-2b66-412e-af12-321a0ce33e88"],

Cell[TextData[{
 "Again, we remove all objects in our session. This must be done in the \
correct order, to respect the relations among the symbols. Otherwise, ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " will refuse to remove the corresponding symbols, to avoid breaking \
internal relations."
}], "Text",ExpressionUUID->"6081aed3-f802-41cd-aeed-1dc8971c90ca"],

Cell[CellGroupData[{

Cell["Start by undefining the metric fields:", "MathCaption",ExpressionUUID->"c17f0cb6-dc4e-4630-8a27-ade9604d919b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metricg", "]"}]], "Input",
 CellLabel->
  "In[912]:=",ExpressionUUID->"3e3b79c6-abff-4e45-a53b-9a17cbec3289"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"weight +2 density\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Detmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "weight +2 density", 
   " ", $CellContext`Detmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897843517458*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1013965578,ExpressionUUID->"127cb54b-9257-44dd-8e4a-cfa1f30ada32"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435192692*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1436730501,ExpressionUUID->"c239eeaa-7433-480a-a778-cf497aba09a5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Einstein tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "EinsteinCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Einstein tensor", " ", $CellContext`EinsteinCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784352095222`*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->823403439,ExpressionUUID->"8ff7242d-7202-4146-93bd-91a20535ef7e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Kretschmann scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "KretschmannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Kretschmann scalar", 
   " ", $CellContext`KretschmannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435226075*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->556978005,ExpressionUUID->"0af6b051-23f3-4613-bbd1-ebf124a22288"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric Ricci tensor", 
   " ", $CellContext`RicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435242837*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->248112762,ExpressionUUID->"648031af-aec6-4184-9071-e3f3cfb83b80"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Ricci scalar\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciScalarCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Ricci scalar", 
   " ", $CellContext`RicciScalarCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784352596197`*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1765432107,ExpressionUUID->"ec2e35ee-f2ee-4ca9-afb1-ebb03aefe50c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897843527632*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1983355397,ExpressionUUID->"48c0afb8-26ff-423f-84f8-ce2722cb3e5d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric TFRicci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TFRicciCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric TFRicci tensor",
    " ", $CellContext`TFRicciCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435292708*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1012988645,ExpressionUUID->"c2752829-86ff-487b-9803-e3808eb11fac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435308701*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->49803099,ExpressionUUID->"5297cb02-7bd0-4e39-a409-5c3b7b3ad8a2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Weyl tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "WeylCD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Weyl tensor", 
   " ", $CellContext`WeylCD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435324636*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1669546210,ExpressionUUID->"bc69bac2-647d-45aa-8d18-ac235a88695c"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784353918467`*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1041819768,ExpressionUUID->"ab28b520-f625-4f0b-98cb-335c30478810"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897843540906*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1430666785,ExpressionUUID->"c5134cd4-b9ec-47e4-ba51-f52fe2a27720"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435426268*^9},
 CellLabel->"During evaluation of In[912]:=",
 CellID->1989743683,ExpressionUUID->"1e053e9f-ff79-423f-9ca2-f401e289780c"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Undefine tensors with imploded variants before removing the derivative:\
\>", "MathCaption",ExpressionUUID->"52f28339-ca18-4e06-8d46-fac8ac790249"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", 
   RowBox[{"{", 
    RowBox[{"r", ",", "v", ",", "T"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[913]:=",ExpressionUUID->"cd4edb72-92f9-49d6-b928-44e48d5265f6"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "r"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`r],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784356048613`*^9},
 CellLabel->"During evaluation of In[913]:=",
 CellID->1006729097,ExpressionUUID->"375ce098-22fd-4eea-9b9a-7e7c6ba9119f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784356228037`*^9},
 CellLabel->"During evaluation of In[913]:=",
 CellID->72278566,ExpressionUUID->"70079b5e-7274-48a4-b0fc-3430137fa049"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435640012*^9},
 CellLabel->"During evaluation of In[913]:=",
 CellID->1457598207,ExpressionUUID->"84d17d86-c008-475b-8a64-534c8bdc00e4"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine the derivatives:", "MathCaption",ExpressionUUID->"0d9073bf-7aac-47d8-91f7-a1ca40646413"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefCovD", "[", "ICD", "]"}]], "Input",
 CellLabel->
  "In[914]:=",ExpressionUUID->"9d07abea-f93d-4f01-86b9-82387d207df3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "AChristoffelICD\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", 
   " ", $CellContext`AChristoffelICD\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435909821*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->1507103720,ExpressionUUID->"38ebc133-2088-4b4f-9e3a-e412bbea51cb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"nonsymmetric AChristoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "AChristoffelICD"}],
  
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "nonsymmetric AChristoffel tensor", " ", $CellContext`AChristoffelICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435928075*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->322605675,ExpressionUUID->"032d23bc-fdd8-458d-82c9-3c0cd50fd746"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelICD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784359460983`*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->1784854697,ExpressionUUID->"3d98b748-94ee-4e89-8635-419ccf108916"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   "FRiemannICD\[Dagger]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannICD\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435963238*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->2001311183,ExpressionUUID->"29cd1e11-7f8a-471c-a4b2-808524c63417"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"FRiemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "FRiemannICD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "FRiemann tensor", 
   " ", $CellContext`FRiemannICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435980692*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->689720,ExpressionUUID->"3005ea08-db95-4831-b6ff-2628254cd586"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciICD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978435997922*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->854705805,ExpressionUUID->"754d7d38-521d-459a-b42e-f19e4f8085cf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannICD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436015625*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->522349852,ExpressionUUID->"68897784-080d-4b67-bb00-7f130affcab1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionICD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436033259*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->220813628,ExpressionUUID->"8f01818c-31ab-4d7e-9959-7df760b9493e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ICD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`ICD],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436051305*^9},
 CellLabel->"During evaluation of In[914]:=",
 CellID->362903780,ExpressionUUID->"c18d0f36-1c27-48e8-a522-109a13fc49f2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefCovD", "[", "Cd", "]"}]], "Input",
 CellLabel->
  "In[915]:=",ExpressionUUID->"cf1c0fae-d5c3-4f8f-ada6-224a65f01fc0"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric Christoffel tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "ChristoffelCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "symmetric Christoffel tensor", " ", $CellContext`ChristoffelCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436356511*^9},
 CellLabel->"During evaluation of In[915]:=",
 CellID->470272980,ExpressionUUID->"aee60881-e717-4910-b5af-16f8444fc8b1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"non-symmetric Ricci tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RicciCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "non-symmetric Ricci tensor", " ", $CellContext`RicciCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784363760357`*^9},
 CellLabel->"During evaluation of In[915]:=",
 CellID->1138153900,ExpressionUUID->"48cb4962-e628-4ae9-b2ed-dc4716b4355f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"Riemann tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "RiemannCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "Riemann tensor", 
   " ", $CellContext`RiemannCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436394813*^9},
 CellLabel->"During evaluation of In[915]:=",
 CellID->1315753471,ExpressionUUID->"5d8c16c4-a089-4c7c-bde8-143820efd640"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"torsion tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TorsionCd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "torsion tensor", 
   " ", $CellContext`TorsionCd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784364123297`*^9},
 CellLabel->"During evaluation of In[915]:=",
 CellID->1229473938,ExpressionUUID->"8791e39d-816d-4ebe-807e-1d8591e624ce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "Cd"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`Cd],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784364287157`*^9},
 CellLabel->"During evaluation of In[915]:=",
 CellID->680824532,ExpressionUUID->"93286f54-09f5-4f8b-bd25-34d6e470a10b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine the rest of the tensors:", "MathCaption",ExpressionUUID->"e3aafc61-9d3b-4b59-87ff-0ccc9cff0bdc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", "$Tensors"}], ";"}]], "Input",
 CellLabel->
  "In[916]:=",ExpressionUUID->"e6e2b3ff-8b96-4ff8-aa7c-2c096068287b"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784368746567`*^9},
 CellLabel->"During evaluation of In[916]:=",
 CellID->825038108,ExpressionUUID->"d862fdb2-7656-4ec7-abc1-cda50a28ed06"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436893005*^9},
 CellLabel->"During evaluation of In[916]:=",
 CellID->113572477,ExpressionUUID->"c098217c-b723-41a5-b384-9751a36deefa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "const"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`const],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784369104233`*^9},
 CellLabel->"During evaluation of In[916]:=",
 CellID->1847142403,ExpressionUUID->"3545c4d2-4990-424c-8f5f-f4c84374bddf"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "w"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`w],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436927493*^9},
 CellLabel->"During evaluation of In[916]:=",
 CellID->898766189,ExpressionUUID->"e9bd3c90-ec57-4725-b255-93d36d520588"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "n"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`n],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436944276*^9},
 CellLabel->"During evaluation of In[916]:=",
 CellID->163365525,ExpressionUUID->"bb84fa3b-f475-4641-b68c-3742288380db"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "lapse"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`lapse],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978436961937*^9},
 CellLabel->"During evaluation of In[916]:=",
 CellID->1200277809,ExpressionUUID->"8781ba9f-9f19-4f4c-a013-6f12f829f2f9"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine inner vector bundles:", "MathCaption",ExpressionUUID->"0a16e88e-adbf-4ba1-8a5a-748b90d81ce0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefVBundle", "[", "InnerC", "]"}]], "Input",
 CellLabel->
  "In[917]:=",ExpressionUUID->"5321598a-431e-4e4d-9af6-1775f6ad95d7"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"conjugated vbundle\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC\[Dagger]",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC\[Dagger],
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "conjugated vbundle", 
   " ", $CellContext`InnerC\[Dagger]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978437361784*^9},
 CellLabel->"During evaluation of In[917]:=",
 CellID->1755239592,ExpressionUUID->"28d28b77-1155-4337-8c3e-56d76eaba4ce"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["InnerC",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`InnerC,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`InnerC],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784375250883`*^9},
 CellLabel->"During evaluation of In[917]:=",
 CellID->328963905,ExpressionUUID->"ebc54623-5d8c-45f3-98ac-258068966fe9"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Finally undefine the manifolds:", "MathCaption",ExpressionUUID->"31294351-fbc9-4e40-9c3e-ac31e717c208"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "S2", "]"}]], "Input",
 CellLabel->
  "In[918]:=",ExpressionUUID->"0237f5d4-8cca-42ef-bdd1-e956ef969156"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]S2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentS2,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentS2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978437808621*^9},
 CellLabel->"During evaluation of In[918]:=",
 CellID->1273443957,ExpressionUUID->"282f6553-137d-4040-b273-9a55cb7b14a1"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`S2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978437827074*^9},
 CellLabel->"During evaluation of In[918]:=",
 CellID->950559297,ExpressionUUID->"1fab8531-0e54-4341-a9c7-0a3915a5f4bc"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellLabel->
  "In[919]:=",ExpressionUUID->"21b93252-a4b6-42c1-9a74-c7a84ab41fea"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978439063818*^9},
 CellLabel->"During evaluation of In[919]:=",
 CellID->919254646,ExpressionUUID->"9e4fab67-8496-4d65-be2c-6bd2f442ea89"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978439082636*^9},
 CellLabel->"During evaluation of In[919]:=",
 CellID->915478804,ExpressionUUID->"a93cccef-26ee-4eaa-9743-9dea115574d2"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There are still symbols in the Global` , not under control of ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ". We can remove them all now:"
}], "MathCaption",ExpressionUUID->"03b42bf4-fb63-4e7d-9734-5c58720b7106"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Global`*"}]], "Input",
 CellLabel->
  "In[920]:=",ExpressionUUID->"4aa7c635-10b7-419a-869d-fe7ed3e510a7"],

Cell[BoxData[
 DynamicModuleBox[{Typeset`open$$ = True}, 
  PaneSelectorBox[{False->
   RowBox[{
    OpenerBox[Dynamic[Typeset`open$$],
     ImageSize->Small], 
    StyleBox["Global`", "InfoHeading"]}], True->GridBox[{
     {
      RowBox[{
       OpenerBox[Dynamic[Typeset`open$$],
        ImageSize->Small], 
       StyleBox["Global`", "InfoHeading"]}]},
     {GridBox[{
        {
         ButtonBox["data",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"data", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["expr",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"expr", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["indices",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"indices", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["result",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"result", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["x",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"x", "Global`"}},
          ButtonNote->"Global`"]},
        {
         ButtonBox["dirvdown",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"dirvdown", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["fit",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"fit", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["npairs",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"npairs", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["result1",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"result1", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["$UseTextSearchQ",
          BaseStyle->"InformationLink",
          
          ButtonData:>{
           "Info9203728982039-9181623", {"$UseTextSearchQ", "Global`"}},
          ButtonNote->"Global`"]},
        {
         ButtonBox["dirvup",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"dirvup", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["frees",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"frees", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["product",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"product", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["rs",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"rs", "Global`"}},
          ButtonNote->"Global`"], ""},
        {
         ButtonBox["ExpandChristoffel",
          BaseStyle->"InformationLink",
          
          ButtonData:>{
           "Info9203728982039-9181623", {"ExpandChristoffel", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["index",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"index", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["product$",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"product$", "Global`"}},
          ButtonNote->"Global`"], 
         ButtonBox["tensor",
          BaseStyle->"InformationLink",
          ButtonData:>{"Info9203728982039-9181623", {"tensor", "Global`"}},
          ButtonNote->"Global`"], ""}
       },
       DefaultBaseStyle->"InfoGrid",
       GridBoxItemSize->{"Columns" -> {{
            Scaled[0.19]}}}]}
    },
    GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}]}, 
   Dynamic[Typeset`open$$],
   ImageSize->Automatic]]], "Print", "InfoCell",
 CellChangeTimes->{3.728978439247221*^9},
 CellID->1879411720,ExpressionUUID->"8165ecdd-6b68-4070-a5c1-785212a284bf"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", "\"\<Global`*\>\"", "]"}]], "Input",
 CellLabel->
  "In[921]:=",ExpressionUUID->"3c538320-e0ae-4be6-8c39-e45459b92feb"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Our session is clean again:", "MathCaption",ExpressionUUID->"5a937aa9-735f-443f-bd3a-81ac79191ccc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Global`*"}]], "Input",
 CellLabel->
  "In[922]:=",ExpressionUUID->"6f00becf-5abe-4699-a4ec-9231680d7635"],

Cell[BoxData[
 TemplateBox[{
  "Information", "nomatch", 
   "\"No symbol matching \
\\!\\(\\*RowBox[{\\\"\\\\\\\"Global`*\\\\\\\"\\\"}]\\) found.\"", 2, 922, 96, 
   19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.7289784394142447`*^9},
 CellLabel->"During evaluation of In[922]:=",
 CellID->2083297964,ExpressionUUID->"efc46c41-5a88-47ca-90b0-af740c406989"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "7. Mappings and tensors along mappings. ",
 StyleBox["V1.1.0 NEW",
  FontColor->RGBColor[0, 0, 1]]
}], "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}, {3.728976670882564*^9, 
  3.728976674720446*^9}, {3.7289768454424953`*^9, 3.728976848920129*^9}, {
  3.728976971716473*^9, 3.728976983868394*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1088456401,ExpressionUUID->"4d1d0042-6342-4549-8253-0be8111126d2"],

Cell[CellGroupData[{

Cell["8.1. Diffeomorphic mapping", "Subsection",ExpressionUUID->"bd19e30b-c5c6-48ee-b788-df3215499082"],

Cell[TextData[{
 "Since version 1.1.0 of ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " we can define mappings between manifolds and use them to push/pull tensor \
fields from one vbundle to another. As in the rest of ",
 StyleBox["xAct",
  FontSlant->"Italic"],
 ", objects must be considered only locally. All manifolds in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " are real, and so are the mappings between any two of them."
}], "Text",ExpressionUUID->"ce2aa235-efec-4671-8a23-ed7b3c3506f1"],

Cell[TextData[{
 StyleBox[ButtonBox["DefMapping",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefMapping.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Define a mapping between two manifolds\n", "Text"],
 StyleBox[ButtonBox["UndefMapping",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefMapping.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Undefine a mapping\n", "Text"],
 StyleBox[ButtonBox["MappingQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MappingQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check a mapping\n", "Text"],
 StyleBox[ButtonBox["$Mappings",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$Mappings.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["List of defined mappings", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.7289770174916286`*^9},ExpressionUUID->"4b2373aa-d26c-4660-9aa2-\
d668c0f79239"],

Cell["Definition of mappings.", "Text",ExpressionUUID->"e9100e92-e074-46c3-8e37-8ecf73b21270"],

Cell[CellGroupData[{

Cell["\<\
Define two manifolds of the same dimension, with their respective tangent \
bundles:\
\>", "MathCaption",ExpressionUUID->"f4079698-6ce3-4813-afa0-461d46cbdec6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"M", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[923]:=",ExpressionUUID->"31ba2f4e-92df-41c2-b37e-2d860cbd03e9"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "M", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", "manifold ", $CellContext`M,
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784581574707`*^9, 3.792804080514573*^9},
 CellLabel->"During evaluation of In[923]:=",
 CellID->1342712479,ExpressionUUID->"3f14afae-a7e8-4610-8112-455e21b96ebe"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentM", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentM, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784581574707`*^9, 3.7928040805199423`*^9},
 CellLabel->"During evaluation of In[923]:=",
 CellID->1210471877,ExpressionUUID->"c62ab0c1-889b-43b8-9728-4440208ba040"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"P", ",", "4", ",", 
   RowBox[{"{", 
    RowBox[{"A", ",", "B", ",", "C", ",", "D"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[924]:=",ExpressionUUID->"201218f8-8db8-4abf-910b-405b21d65829"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "P", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", "manifold ", $CellContext`P,
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784583399982`*^9, 3.7928040822899446`*^9},
 CellLabel->"During evaluation of In[924]:=",
 CellID->1877787831,ExpressionUUID->"acc84c39-4d7f-43d3-adfd-179693a62192"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentP", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentP, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784583399982`*^9, 3.792804082296945*^9},
 CellLabel->"During evaluation of In[924]:=",
 CellID->560399305,ExpressionUUID->"d6294237-48be-4b32-bafb-31d820d8b668"]
}, Open  ]],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"C\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 924, 1, 19671867426941700138, "Local",
    "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978458382565*^9, 3.7928040823226557`*^9},
 CellLabel->"During evaluation of In[924]:=",
 CellID->1915327040,ExpressionUUID->"50c779b0-1b9d-4d47-9edb-64ecd7555a21"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"D\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 924, 2, 19671867426941700138, "Local",
    "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978458382565*^9, 3.792804082350729*^9},
 CellLabel->"During evaluation of In[924]:=",
 CellID->121692750,ExpressionUUID->"37624fce-ecb5-46ed-8bd2-772685f5764e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Define a mapping from the manifold ",
 StyleBox["M",
  FontFamily->"Courier"],
 " to the manifold ",
 StyleBox["P",
  FontFamily->"Courier"],
 ". By default mappings between manifolds of the same dimension are assumed \
to be diffeomorphisms (we will see later how to change this)."
}], "MathCaption",ExpressionUUID->"ccf8a0e1-3ad2-4255-94ef-b7f2a7bc2141"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMapping", "[", 
  RowBox[{"\[Phi]", ",", 
   RowBox[{"M", "\[Rule]", "P"}]}], "]"}]], "Input",
 CellLabel->
  "In[925]:=",ExpressionUUID->"97842234-8dd8-494b-bed4-a1189bae3fdc"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping \"\>", "\[InvisibleSpace]", "\[Phi]", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "mapping ", $CellContext`\[Phi], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978458540165*^9, 3.792804125655573*^9},
 CellLabel->"During evaluation of In[925]:=",
 CellID->1229540967,ExpressionUUID->"ff4f7f4e-900b-489a-a5cf-b30e6cdd16d0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping \"\>", "\[InvisibleSpace]", 
   "i\[Phi]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "inverse mapping ", $CellContext`i\[Phi], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978458540165*^9, 3.792804125662232*^9},
 CellLabel->"During evaluation of In[925]:=",
 CellID->1807380935,ExpressionUUID->"9d2eb126-12ad-4fa2-9293-e2ab55d3e318"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"di\[Phi]", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", "i\[Phi]a"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`di\[Phi][-$CellContext`A, $CellContext`i\[Phi]a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978458540165*^9, 3.7928041256699543`*^9},
 CellLabel->"During evaluation of In[925]:=",
 CellID->495187730,ExpressionUUID->"64f5a93f-023b-4646-b0fd-6abc66c7cd9e"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"d\[Phi]", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "\[Phi]A"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`d\[Phi][-$CellContext`a, $CellContext`\[Phi]A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978458540165*^9, 3.7928041256739683`*^9},
 CellLabel->"During evaluation of In[925]:=",
 CellID->1121045985,ExpressionUUID->"78baae71-eca0-4328-b767-cfd89726997b"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell["\<\
Together with the mapping \[Phi] we have defined its inverse mapping, and \
also their differentials, i.e., the linear operators relating the tangent \
bundles. Let us first have a look at the mappings themselves, and their \
properties.\
\>", "Text",ExpressionUUID->"9325c653-c73d-487e-a615-18313bddbab6"],

Cell[TextData[{
 StyleBox[ButtonBox["MappingDomain",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MappingDomain.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Domain manifold of a mapping\n", "Text"],
 StyleBox[ButtonBox["MappingImage",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MappingImage.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Image manifold of a mapping\n", "Text"],
 StyleBox[ButtonBox["InverseMapping",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "InverseMapping.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Inverse of a mapping\n", "Text"],
 StyleBox[ButtonBox["ImmersionQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ImmersionQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check whether a mapping is an immersion\n", "Text"],
 StyleBox[ButtonBox["SubmersionQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SubmersionQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Check whether a mapping is a submersion\n", "Text"],
 StyleBox[ButtonBox["DiffeomorphismQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DiffeomorphismQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Check whether a mapping is a diffeomorphism", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978450560616*^9},ExpressionUUID->"cafe33b9-d6af-491a-a756-\
2e3dee610c23"],

Cell["Properties of mappings.", "Text",ExpressionUUID->"0a81f6e4-c054-4628-9cd9-c6f069d42312"],

Cell[CellGroupData[{

Cell["\<\
A mapping has a domain manifold and an image (or codomain) manifold.\
\>", "MathCaption",ExpressionUUID->"7a5d775f-ce16-43b9-ab5a-acd1938ff8aa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MappingDomain", "[", "\[Phi]", "]"}], "\[Rule]", 
  RowBox[{"MappingImage", "[", "\[Phi]", "]"}]}]], "Input",
 CellLabel->
  "In[926]:=",ExpressionUUID->"22b40760-e387-4d74-840c-c4750f513e93"],

Cell[BoxData[
 RowBox[{"M", "\[Rule]", "P"}]], "Output",
 CellChangeTimes->{3.7289784588617687`*^9, 3.792804160826948*^9},
 CellLabel->"Out[926]=",
 CellID->1484713041,ExpressionUUID->"218d97ca-8c08-402c-9e99-9b8612ea43c4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the inverse mapping of \[Phi]. By default it is named by prepending i:\
\>", "MathCaption",ExpressionUUID->"be04e4c7-01e2-4312-bcbb-4e0098921d9c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InverseMapping", "[", "\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[927]:=",ExpressionUUID->"5efebdac-ebf9-4cd2-82ad-f39b8d6d2b07"],

Cell[BoxData["i\[Phi]"], "Output",
 CellChangeTimes->{3.728978459113825*^9, 3.792804165932309*^9},
 CellLabel->"Out[927]=",
 CellID->966467739,ExpressionUUID->"f8c74aa8-1be5-4699-8039-32f669a28894"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The inverse mapping is defined from the image to the domain of the original \
mapping:\
\>", "MathCaption",ExpressionUUID->"1e94c03d-2022-4082-baed-57ca743ef091"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MappingDomain", "[", "i\[Phi]", "]"}], "\[Rule]", 
  RowBox[{"MappingImage", "[", "i\[Phi]", "]"}]}]], "Input",
 CellLabel->
  "In[928]:=",ExpressionUUID->"467f4c6a-f063-4f16-90eb-492fcabe36cb"],

Cell[BoxData[
 RowBox[{"P", "\[Rule]", "M"}]], "Output",
 CellChangeTimes->{3.728978459288693*^9, 3.792804169844322*^9},
 CellLabel->"Out[928]=",
 CellID->987277630,ExpressionUUID->"ed04c486-e87d-4e54-96a4-9afa6ccc7756"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Both mappings have all good properties in the (local) diffeomorphic case:\
\>", "MathCaption",ExpressionUUID->"942cd52c-6e9a-40cf-b703-5cba375db5e9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ImmersionQ", "[", "\[Phi]", "]"}], ",", 
   RowBox[{"SubmersionQ", "[", "\[Phi]", "]"}], ",", 
   RowBox[{"DiffeomorphismQ", "[", "\[Phi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[929]:=",ExpressionUUID->"2a44ddd1-7fdc-4d41-8a04-29a77139ef05"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{3.728978459425131*^9, 3.792804178964137*^9},
 CellLabel->"Out[929]=",
 CellID->1529220533,ExpressionUUID->"d1ff0ccb-06e5-4e95-95ca-13728ae9d216"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ImmersionQ", "[", "i\[Phi]", "]"}], ",", 
   RowBox[{"SubmersionQ", "[", "i\[Phi]", "]"}], ",", 
   RowBox[{"DiffeomorphismQ", "[", "i\[Phi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[930]:=",ExpressionUUID->"48e8b980-ed4f-415e-ab9b-e35dd82ad06a"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{3.728978459711199*^9, 3.79280418049541*^9},
 CellLabel->"Out[930]=",
 CellID->1632892784,ExpressionUUID->"1f838fd9-b3a4-463a-b87b-4fd19e5875ec"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Because \[Phi] is a submersion, the image manifold coincides with the range:\
\>", "MathCaption",ExpressionUUID->"5f2bcb82-b954-45d1-930d-295f391e0b73"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Phi]", "[", "M", "]"}]], "Input",
 CellLabel->
  "In[931]:=",ExpressionUUID->"845b3dcd-f2d4-44c3-b467-f0338b71a477"],

Cell[BoxData["P"], "Output",
 CellChangeTimes->{3.7289784599782257`*^9, 3.792804189677649*^9},
 CellLabel->"Out[931]=",
 CellID->617915879,ExpressionUUID->"2f44ba53-1cd8-4dcf-9d77-0019ac66b36c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["The same can be said of its inverse:", "MathCaption",ExpressionUUID->"0fdb29a9-f384-4d48-981c-a99487b2e35d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"i\[Phi]", "[", "P", "]"}]], "Input",
 CellLabel->
  "In[932]:=",ExpressionUUID->"79d83be9-55df-44ae-83b4-fb459adf1d3f"],

Cell[BoxData["M"], "Output",
 CellChangeTimes->{3.728978460158435*^9, 3.792804192565185*^9},
 CellLabel->"Out[932]=",
 CellID->1533455672,ExpressionUUID->"53684c69-163e-4388-920f-edc917f244dd"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.2. Non-diffeomorphic mapping", "Subsection",ExpressionUUID->"ba14c322-0a96-4de9-b204-78bbc49afd82"],

Cell["\<\
Mappings do not have to relate manifolds of the same dimension. Or even if \
they do, they might fail to be immersions or submersions.\
\>", "Text",ExpressionUUID->"20890f5e-cad6-46c4-8b10-efd4bd00f6dc"],

Cell[CellGroupData[{

Cell["Define a manifold of higher dimension:", "MathCaption",ExpressionUUID->"d6ef2031-0e6b-4fa7-b877-bb8ca3872550"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefManifold", "[", 
  RowBox[{"L", ",", "5", ",", 
   RowBox[{"{", 
    RowBox[{"\[Alpha]", ",", "\[Beta]", ",", "\[Gamma]", ",", "\[Delta]"}], 
    "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[933]:=",ExpressionUUID->"44c2a0b6-e89c-48cc-915b-e9cc0fc5986a"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefManifold", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"manifold \"\>", "\[InvisibleSpace]", "L", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefManifold, ": Defining ", "manifold ", $CellContext`L,
    ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784646601963`*^9, 3.7928042497466583`*^9},
 CellLabel->"During evaluation of In[933]:=",
 CellID->243441307,ExpressionUUID->"fb5bd5d5-0e64-477f-8e41-b35b7cce19fa"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefVBundle", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle \"\>", "\[InvisibleSpace]", "TangentL", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefVBundle, ": Defining ", 
   "vbundle ", $CellContext`TangentL, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289784646601963`*^9, 3.792804249753831*^9},
 CellLabel->"During evaluation of In[933]:=",
 CellID->561699214,ExpressionUUID->"fb84e27c-861e-4d85-8ca3-c0aa65cea3a1"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now we can define another mapping, this time being only an immersion:\
\>", "MathCaption",ExpressionUUID->"d2a24ddf-69ff-43ff-bab6-9fb2efec0df1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefMapping", "[", 
  RowBox[{"\[Psi]", ",", 
   RowBox[{"P", "\[Rule]", "L"}], ",", 
   RowBox[{"SubmersionQ", "\[Rule]", "False"}]}], "]"}]], "Input",
 CellLabel->
  "In[934]:=",ExpressionUUID->"228bc87a-ac89-49e6-b1dd-891e561ed4bb"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping \"\>", "\[InvisibleSpace]", "\[Psi]", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "mapping ", $CellContext`\[Psi], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978464846589*^9, 3.792804255364479*^9},
 CellLabel->"During evaluation of In[934]:=",
 CellID->497035243,ExpressionUUID->"df7b4ed7-7072-433e-94a7-c6af4fbe4661"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefMapping", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"left-inverse mapping \"\>", "\[InvisibleSpace]", 
   "i\[Psi]", "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"Assuming arbitrary extension.\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefMapping, ": Defining ", 
   "left-inverse mapping ", $CellContext`i\[Psi], ". ", 
   "Assuming arbitrary extension."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978464846589*^9, 3.792804255371584*^9},
 CellLabel->"During evaluation of In[934]:=",
 CellID->2107425918,ExpressionUUID->"dbb9f3db-a67f-4173-8b07-bb0f31ad0754"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"di\[Psi]", "[", 
    RowBox[{
     RowBox[{"-", "\[Alpha]"}], ",", "i\[Psi]A"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`di\[Psi][-$CellContext`\[Alpha], $CellContext`i\[Psi]A], ". ",
    ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978464846589*^9, 3.7928042553778963`*^9},
 CellLabel->"During evaluation of In[934]:=",
 CellID->1059613855,ExpressionUUID->"e06b57b1-0756-4211-a43a-3b14ffc5511f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"d\[Psi]", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", "\[Psi]\[Alpha]"}], "]"}], 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefTensor, ": Defining ", 
   "mapping differential tensor ", 
   $CellContext`d\[Psi][-$CellContext`A, $CellContext`\[Psi]\[Alpha]], ". ", 
   ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978464846589*^9, 3.792804255381695*^9},
 CellLabel->"During evaluation of In[934]:=",
 CellID->1995660061,ExpressionUUID->"f7db3191-eb1f-49c6-80d3-6204a1078029"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These are the properties of the mapping \[Psi]:", "MathCaption",ExpressionUUID->"dce76985-b9ea-48d9-ace5-4dc48bb53098"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ImmersionQ", "[", "\[Psi]", "]"}], ",", 
   RowBox[{"SubmersionQ", "[", "\[Psi]", "]"}], ",", 
   RowBox[{"DiffeomorphismQ", "[", "\[Psi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[935]:=",ExpressionUUID->"5056e74b-c88e-4b49-b3c9-33b85ac171c2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "False", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.728978465179954*^9, 3.792804262393531*^9},
 CellLabel->"Out[935]=",
 CellID->1522510521,ExpressionUUID->"f9ff3b4a-8e45-439e-a5aa-5ff013b0fa97"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MappingDomain", "[", "\[Psi]", "]"}], "\[Rule]", 
  RowBox[{"MappingImage", "[", "\[Psi]", "]"}]}]], "Input",
 CellLabel->
  "In[936]:=",ExpressionUUID->"e5f472bd-60ce-445e-ac35-ccb0d502c691"],

Cell[BoxData[
 RowBox[{"P", "\[Rule]", "L"}]], "Output",
 CellChangeTimes->{3.7289784654263372`*^9, 3.792804265631938*^9},
 CellLabel->"Out[936]=",
 CellID->438371102,ExpressionUUID->"7e7cba8d-6516-4853-8669-935773b996e7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The range does not coincide with all of the image manifold. It is just a \
submanifold of the image:\
\>", "MathCaption",ExpressionUUID->"49cd33e8-a753-4d83-bc5e-9135a8068600"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Psi]", "[", "P", "]"}]], "Input",
 CellLabel->
  "In[937]:=",ExpressionUUID->"7f283785-1f55-4521-a14a-25247ae9ed4f"],

Cell[BoxData[
 RowBox[{"\[Psi]", "[", "P", "]"}]], "Output",
 CellChangeTimes->{3.728978465616881*^9, 3.792804270879578*^9},
 CellLabel->"Out[937]=",
 CellID->24126788,ExpressionUUID->"094c6a46-792a-4b40-ab3b-0a7083bac496"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SubmanifoldQ", "[", 
  RowBox[{"L", ",", 
   RowBox[{"\[Psi]", "[", "P", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[938]:=",ExpressionUUID->"03fd7cde-172f-4c6d-b676-34ae28ab6f86"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978465749744*^9, 3.79280427253878*^9},
 CellLabel->"Out[938]=",
 CellID->408857426,ExpressionUUID->"ec9a505d-66c6-4bf9-9d22-ac669839495a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The inverse mapping is defined on all of the image of \[Psi], because we \
assume it has been extended in some smooth arbitrary way. This is just a \
simplification to avoid working with subvbundles of ",
 StyleBox["TangentL",
  FontFamily->"Courier"],
 ", and might be improved in some future version of ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 "."
}], "MathCaption",ExpressionUUID->"bfa4bc4a-b164-4737-aafa-77971369a371"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MappingDomain", "[", "i\[Psi]", "]"}], "\[Rule]", 
  RowBox[{"MappingImage", "[", "i\[Psi]", "]"}]}]], "Input",
 CellLabel->
  "In[939]:=",ExpressionUUID->"10435631-1d8f-44bf-91d8-59aa3909cad2"],

Cell[BoxData[
 RowBox[{"L", "\[Rule]", "P"}]], "Output",
 CellChangeTimes->{3.728978465887426*^9, 3.792804373735746*^9},
 CellLabel->"Out[939]=",
 CellID->1088131848,ExpressionUUID->"0b033c06-e6ff-4f92-97f9-0c4af366470b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Properties of the inverse mapping i\[Psi]:", "MathCaption",ExpressionUUID->"f6077cf3-fb94-439c-ae03-8bf7e9beb426"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ImmersionQ", "[", "i\[Psi]", "]"}], ",", 
   RowBox[{"SubmersionQ", "[", "i\[Psi]", "]"}], ",", 
   RowBox[{"DiffeomorphismQ", "[", "i\[Psi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[940]:=",ExpressionUUID->"7e8fea0f-c075-43dd-8fbe-da4a517d5bbb"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"False", ",", "True", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.728978465991518*^9, 3.7928043757942657`*^9},
 CellLabel->"Out[940]=",
 CellID->907850413,ExpressionUUID->"e8893599-2f9b-4f62-846c-d5dda25c1f4a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is only a left-inverse mapping, but not a right-inverse mapping of \[Psi]:\
\
\>", "MathCaption",ExpressionUUID->"933fa1d3-312a-46b5-aba4-2d6be451c452"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"i\[Psi]", "[", 
  RowBox[{"\[Psi]", "[", "P", "]"}], "]"}]], "Input",
 CellLabel->
  "In[941]:=",ExpressionUUID->"9224ab65-d05b-474a-9102-069a6cc21885"],

Cell[BoxData["P"], "Output",
 CellChangeTimes->{3.728978466121024*^9, 3.792804378061294*^9},
 CellLabel->"Out[941]=",
 CellID->1263115465,ExpressionUUID->"60c90f84-25a4-4840-ac64-a099e8ea3683"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Psi]", "[", 
  RowBox[{"i\[Psi]", "[", "L", "]"}], "]"}]], "Input",
 CellLabel->
  "In[942]:=",ExpressionUUID->"d989c0b4-fe37-4a00-9301-982c39e21a1a"],

Cell[BoxData[
 RowBox[{"\[Psi]", "[", "P", "]"}]], "Output",
 CellChangeTimes->{3.728978466241234*^9, 3.792804379221299*^9},
 CellLabel->"Out[942]=",
 CellID->1762305658,ExpressionUUID->"945eadf9-bdea-40e8-b2df-d18c6a90adb0"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.3. Mapping composition", "Subsection",ExpressionUUID->"ec16d493-45ad-487f-b888-2b5bf158c405"],

Cell[TextData[{
 "Mappings can be composed, resulting in other mappings. To avoid \
interferences with ",
 StyleBox["Composition",
  FontFamily->"Courier"],
 ", and to have nicer typesetting and an alias, we use ",
 StyleBox["SmallCircle",
  FontFamily->"Courier"],
 " to represent mapping composition. "
}], "Text",ExpressionUUID->"22079b7e-013a-4594-b299-fb60e80f25c4"],

Cell[TextData[{
 StyleBox[ButtonBox["SmallCircle",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SmallCircle.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Composition of several mappings\n", "Text"],
 StyleBox[ButtonBox["InverseMapping",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "InverseMapping.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Inverse of a mapping\n", "Text"],
 StyleBox[ButtonBox["IdentityMapping",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IdentityMapping.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["[M]\t\t\t", "DisplayMath"],
 StyleBox["Identity mapping on a manifold", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978472405828*^9},ExpressionUUID->"171ba3a9-2d73-4e12-830b-\
f564315eb8cb"],

Cell["Composition of mappings.", "Text",ExpressionUUID->"b66c8ada-e039-4279-843e-64f3b7bb47ba"],

Cell[CellGroupData[{

Cell[TextData[{
 "Compose the mappings \[Phi] and \[Psi], following the standard \
right-to-left convention. Note the alias esc-sc-esc for ",
 StyleBox["SmallCircle",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"03ae8c17-e6a1-4869-b68f-d21d3bf4d88d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"map", "=", 
  RowBox[{"\[Psi]", "\[SmallCircle]", "\[Phi]"}]}]], "Input",
 CellLabel->
  "In[943]:=",ExpressionUUID->"e3bddfc3-a5ff-4603-a1a1-864a49de7226"],

Cell[BoxData[
 RowBox[{"\[Psi]", "\[SmallCircle]", "\[Phi]"}]], "Output",
 CellChangeTimes->{3.728978470042137*^9, 3.792804410380517*^9},
 CellLabel->"Out[943]=",
 CellID->1115016536,ExpressionUUID->"081de4c1-8148-4092-ad47-b97668e3bc89"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MappingDomain", "[", "map", "]"}], "\[Rule]", 
  RowBox[{"MappingImage", "[", "map", "]"}]}]], "Input",
 CellLabel->
  "In[944]:=",ExpressionUUID->"4161397e-ea4f-4c43-8533-4b56b4ca39eb"],

Cell[BoxData[
 RowBox[{"M", "\[Rule]", "L"}]], "Output",
 CellChangeTimes->{3.728978470191555*^9, 3.7928044161323013`*^9},
 CellLabel->"Out[944]=",
 CellID->1188860967,ExpressionUUID->"d966591d-a28a-4a40-be1e-c4d862440e18"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Inverse of the composition. This requires all individual mappings to be \
invertible (at least left- or right- invertible). Note the reversal of \
ordering.\
\>", "MathCaption",ExpressionUUID->"142660cf-596c-4ad3-8958-1288e5b40a73"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"imap", "=", 
  RowBox[{"InverseMapping", "[", "map", "]"}]}]], "Input",
 CellLabel->
  "In[945]:=",ExpressionUUID->"e08a9064-1c8d-4368-a1e4-d28650736d46"],

Cell[BoxData[
 RowBox[{"i\[Phi]", "\[SmallCircle]", "i\[Psi]"}]], "Output",
 CellChangeTimes->{3.728978470309476*^9, 3.7928044195900917`*^9},
 CellLabel->"Out[945]=",
 CellID->1484132930,ExpressionUUID->"5acf51b6-3b59-4d0d-b923-ce3f77cf7f83"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Properties of the composition:", "MathCaption",ExpressionUUID->"3bac8eea-ba41-43b0-8ce7-b7155cc8b33b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ImmersionQ", "[", "map", "]"}], ",", 
   RowBox[{"SubmersionQ", "[", "map", "]"}], ",", 
   RowBox[{"DiffeomorphismQ", "[", "map", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[946]:=",ExpressionUUID->"0adfd2a3-4ecd-4392-9b2f-99863a4cd7f2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "False", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.728978470425653*^9, 3.792804430600642*^9},
 CellLabel->"Out[946]=",
 CellID->285100635,ExpressionUUID->"a43b3bf0-a85f-4a39-aa64-31dbdd25dd3a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Because the composed map is an immersion but not a submersion, its inverse is \
a left-inverse but not a right-inverse:\
\>", "MathCaption",ExpressionUUID->"311d8741-bb03-453b-a19a-37bffd0d264c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"imap", "\[SmallCircle]", "map"}]], "Input",
 CellLabel->
  "In[947]:=",ExpressionUUID->"12c52d38-3235-4b69-9e7d-69b60cd5eb3d"],

Cell[BoxData[
 SubscriptBox["Id", "M"]], "Output",
 CellChangeTimes->{3.728978470560626*^9, 3.792804438716001*^9},
 CellLabel->"Out[947]=",
 CellID->1411403213,ExpressionUUID->"bd6a3955-c8c7-4c5e-bba5-bdb5f37ba424"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[948]:=",ExpressionUUID->"f1bcdaf1-0e4e-4bdd-8c52-1be143f647a6"],

Cell["IdentityMapping[M]", "Output",
 CellChangeTimes->{3.728978470641161*^9, 3.792804439819817*^9},
 CellLabel->"Out[948]//InputForm=",
 CellID->1427366760,ExpressionUUID->"6897d6e4-4c73-4f11-8373-a290d34e8ee4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"map", "\[SmallCircle]", "imap"}]], "Input",
 CellLabel->
  "In[949]:=",ExpressionUUID->"160627a7-27bc-482e-9b0d-5d5a7f5da1c2"],

Cell[BoxData[
 RowBox[{"\[Psi]", "\[SmallCircle]", "i\[Psi]"}]], "Output",
 CellChangeTimes->{3.728978470758889*^9, 3.792804440613188*^9},
 CellLabel->"Out[949]=",
 CellID->1071935285,ExpressionUUID->"84c65189-12a6-4e68-9710-c943eba4c51a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["More explicitly:", "MathCaption",ExpressionUUID->"6d94bb4b-29d7-40d0-ad14-4850682e3d87"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"imap", "[", 
  RowBox[{"map", "[", "M", "]"}], "]"}]], "Input",
 CellLabel->
  "In[950]:=",ExpressionUUID->"ab254e1c-a3bc-43e2-8ed1-84706f26d2c4"],

Cell[BoxData["M"], "Output",
 CellChangeTimes->{3.72897847085607*^9, 3.792804443633301*^9},
 CellLabel->"Out[950]=",
 CellID->837399513,ExpressionUUID->"3d09e326-bca2-4abc-827e-b4b5d1d97f88"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"map", "[", 
  RowBox[{"imap", "[", "L", "]"}], "]"}]], "Input",
 CellLabel->
  "In[951]:=",ExpressionUUID->"03e415a2-5f63-413b-9be7-c229004ef9f1"],

Cell[BoxData[
 RowBox[{"\[Psi]", "[", "P", "]"}]], "Output",
 CellChangeTimes->{3.72897847095192*^9, 3.792804445034412*^9},
 CellLabel->"Out[951]=",
 CellID->324926915,ExpressionUUID->"69f29f14-ac41-4498-8eef-0717abccac41"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The identity mapping is denoted by ",
 StyleBox["IdentityMapping[manifold]",
  FontFamily->"Courier"],
 ", explicitly indicating the manifold where it acts. It has all the expected \
properties."
}], "Text",ExpressionUUID->"b1b31f59-477e-4cc5-b17e-3e42511c4ea7"],

Cell[CellGroupData[{

Cell[TextData[{
 "The identity mapping on the manifold ",
 StyleBox["M",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"437043e7-74e4-4443-8f0c-5c4192f50262"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"idM", "=", 
  RowBox[{"IdentityMapping", "[", "M", "]"}]}]], "Input",
 CellLabel->
  "In[952]:=",ExpressionUUID->"1de03d3f-2372-4752-81a0-ce5269abe12f"],

Cell[BoxData[
 SubscriptBox["Id", "M"]], "Output",
 CellChangeTimes->{3.728978471055626*^9, 3.7928044540508423`*^9},
 CellLabel->"Out[952]=",
 CellID->313260877,ExpressionUUID->"e7f35533-8843-44bc-82e3-424a77a5140c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"MappingDomain", "[", "idM", "]"}], "\[Rule]", 
  RowBox[{"MappingImage", "[", "idM", "]"}]}]], "Input",
 CellChangeTimes->{{3.792804461721429*^9, 3.792804461992343*^9}},
 CellLabel->
  "In[953]:=",ExpressionUUID->"ebeafe23-a4dc-4a6e-97fb-cd9a4ea53b05"],

Cell[BoxData[
 RowBox[{"M", "\[Rule]", "M"}]], "Output",
 CellChangeTimes->{
  3.728978471202463*^9, {3.792804454795545*^9, 3.7928044728341827`*^9}},
 CellLabel->"Out[953]=",
 CellID->1451966044,ExpressionUUID->"14bd38a3-5d64-4c08-b006-c08d65182588"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It has all good properties, and it is its own inverse:", "MathCaption",ExpressionUUID->"b90c82eb-9fc8-4d91-9cf3-b30aaec49f20"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"ImmersionQ", "[", "idM", "]"}], ",", 
   RowBox[{"SubmersionQ", "[", "idM", "]"}], ",", 
   RowBox[{"DiffeomorphismQ", "[", "idM", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[954]:=",ExpressionUUID->"5b8064e6-508c-451c-bf44-0b6a8ed3c976"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{3.728978471313446*^9, 3.792804475519474*^9},
 CellLabel->"Out[954]=",
 CellID->184692807,ExpressionUUID->"ce165c7f-e829-48a0-9734-6425aeb65a71"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InverseMapping", "[", "idM", "]"}]], "Input",
 CellLabel->
  "In[955]:=",ExpressionUUID->"a7b1389d-4dc1-4ab0-9d17-4623cd084476"],

Cell[BoxData[
 SubscriptBox["Id", "M"]], "Output",
 CellChangeTimes->{3.728978471440206*^9, 3.792804477576268*^9},
 CellLabel->"Out[955]=",
 CellID->1192473938,ExpressionUUID->"88a871e8-b1a1-4e99-826a-fb3fc92352d7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"idM", "[", "M", "]"}]], "Input",
 CellLabel->
  "In[956]:=",ExpressionUUID->"0024beb5-ae49-46ac-86ce-b8401c95a4c0"],

Cell[BoxData["M"], "Output",
 CellChangeTimes->{3.728978471534943*^9, 3.792804478994257*^9},
 CellLabel->"Out[956]=",
 CellID->1362085225,ExpressionUUID->"a2e54119-62b4-4348-81a9-151931d3d264"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It is the identity of composition:", "MathCaption",ExpressionUUID->"9ddfa442-78a1-4777-b2df-e6d2a17e9081"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Phi]", "\[SmallCircle]", "idM"}]], "Input",
 CellLabel->
  "In[957]:=",ExpressionUUID->"c0b02ce0-6156-4d73-8655-737e6133b05b"],

Cell[BoxData["\[Phi]"], "Output",
 CellChangeTimes->{3.7289784716564007`*^9, 3.792804483241184*^9},
 CellLabel->"Out[957]=",
 CellID->957605907,ExpressionUUID->"56417482-0e9c-488f-9b1e-36654ba105ed"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IdentityMapping", "[", "P", "]"}], "\[SmallCircle]", 
  "\[Phi]"}]], "Input",
 CellLabel->
  "In[958]:=",ExpressionUUID->"cc162166-3860-433c-bbe6-06c31caec72e"],

Cell[BoxData["\[Phi]"], "Output",
 CellChangeTimes->{3.7289784718178377`*^9, 3.792804486219458*^9},
 CellLabel->"Out[958]=",
 CellID->1292709219,ExpressionUUID->"d1ae09e6-e3c2-4349-9668-239e9872af0b"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.4. The pullback bundle and its tensor fields", "Subsection",ExpressionUUID->"d87f07d4-4d04-4f31-be9d-2e446937c5b7"],

Cell[TextData[{
 "Given a vector bundle ",
 StyleBox["VB",
  FontFamily->"Courier"],
 " with base manifold ",
 StyleBox["P",
  FontFamily->"Courier"],
 " and a mapping ",
 StyleBox["\[Phi]: M->P",
  FontFamily->"Courier"],
 " we can construct another bundle with base ",
 StyleBox["M",
  FontFamily->"Courier"],
 " and fibers those of ",
 StyleBox["VB",
  FontFamily->"Courier"],
 ". It is called the pullback bundle of ",
 StyleBox["VB",
  FontFamily->"Courier"],
 " by ",
 StyleBox["\[Phi]",
  FontFamily->"Courier"],
 ". This is effectively an abstraction of the concept of reparametrization."
}], "Text",ExpressionUUID->"adc30f64-3f0b-4c9f-b8b3-f205c9a6910f"],

Cell[TextData[{
 StyleBox[ButtonBox["PullBackVBundle",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PullBackVBundle.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Pull back of a vector bundle", "Text"]
}], "DefinitionBox",ExpressionUUID->"54b8cafc-8ce7-4197-8414-15455a64638b"],

Cell["Pull back vector bundle.", "Text",ExpressionUUID->"b72f3ffb-df0a-4441-be80-2f88c110f3ec"],

Cell[CellGroupData[{

Cell[TextData[{
 "This is the tangent bundle of the manifold ",
 StyleBox["P",
  FontFamily->"Courier"],
 ", with 4-dimensional fibers: "
}], "MathCaption",ExpressionUUID->"b397e042-a772-4269-815c-d7216d17f405"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Tangent", "[", "P", "]"}]], "Input",
 CellLabel->
  "In[959]:=",ExpressionUUID->"08c297a6-3ee1-4233-99b1-f12f1f497b6c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox["\[DoubleStruckCapitalT]P",
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TangentP,
  Editable->False]], "Output",
 CellChangeTimes->{3.728978485955963*^9, 3.792804548903894*^9},
 CellLabel->"Out[959]=",
 CellID->1990684632,ExpressionUUID->"aa068460-dbdd-497f-9084-2d1825cb79c5"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"VBundleQ", "[", "%", "]"}], ",", 
   RowBox[{"BaseOfVBundle", "[", "%", "]"}], ",", 
   RowBox[{"DimOfVBundle", "[", "%", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[960]:=",ExpressionUUID->"911b2e85-5839-4e03-990b-73607dd6a601"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "P", ",", "4"}], "}"}]], "Output",
 CellChangeTimes->{3.728978486101933*^9, 3.792804556296316*^9},
 CellLabel->"Out[960]=",
 CellID->2039037263,ExpressionUUID->"aad8cd6d-f596-4383-8233-1d40b09511fe"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "This is its pullback bundle by the mapping \[Phi], now with base ",
 StyleBox["M",
  FontFamily->"Courier"],
 " and the same 4-dimensional fibers:"
}], "MathCaption",ExpressionUUID->"cb6187b0-05f3-4393-8d07-fd1626ed870e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PullBackVBundle", "[", 
  RowBox[{"TangentP", ",", "\[Phi]"}], "]"}]], "Input",
 CellLabel->
  "In[961]:=",ExpressionUUID->"63f6b7dc-72b2-4416-a5f3-579b36dbdf73"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["\[Phi]", "*"], 
  InterpretationBox[
   StyleBox["\[DoubleStruckCapitalT]P",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TangentP,
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978486208871*^9, 3.7928045940865383`*^9},
 CellLabel->"Out[961]=",
 CellID->263760201,ExpressionUUID->"3d04867c-4619-4007-8ae0-237ab937c146"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"VBundleQ", "[", "%", "]"}], ",", 
   RowBox[{"BaseOfVBundle", "[", "%", "]"}], ",", 
   RowBox[{"DimOfVBundle", "[", "%", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[962]:=",ExpressionUUID->"a5b3b0ef-0dba-43b8-a21a-6af054c0b557"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "M", ",", "4"}], "}"}]], "Output",
 CellChangeTimes->{3.7289784863525963`*^9, 3.792804602175761*^9},
 CellLabel->"Out[962]=",
 CellID->1344428859,ExpressionUUID->"d44e912e-452e-4d56-ae65-4e472b0c4ade"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Note its internal non-symbolic representation", "MathCaption",ExpressionUUID->"48695cc1-d373-4cb2-8f69-afb87ab3675f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[963]:=",ExpressionUUID->"7e0cdcbd-59b9-48f4-8a10-3cf6f28cbee0"],

Cell["PullBackVBundle[TangentP, \[Phi]]", "Output",
 CellChangeTimes->{3.728978486440577*^9, 3.792804613999029*^9},
 CellLabel->"Out[963]//InputForm=",
 CellID->1764624780,ExpressionUUID->"77a1456c-4cb6-4b05-86cc-f8a52e9172b1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Even though the fibers are isomorphic, the associated indices must be \
different because indices must identify vbundles uniquely in ",
 StyleBox["xAct",
  FontSlant->"Italic"],
 ". To avoid requesting new indices from the user, we prepend the name of the \
mapping to the indices:"
}], "MathCaption",ExpressionUUID->"561c965a-d8b4-4e5d-92dc-cb2ad5b532cf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "TangentP", "]"}]], "Input",
 CellLabel->
  "In[964]:=",ExpressionUUID->"555c3601-251f-45c4-a212-e9344ec59159"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"A", ",", "B", ",", "C", ",", "D"}], "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978486583853*^9, 3.792804619766375*^9},
 CellLabel->"Out[964]=",
 CellID->498710995,ExpressionUUID->"bdeec7d8-2d57-4656-a52c-24a9cc65bada"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", 
  RowBox[{"PullBackVBundle", "[", 
   RowBox[{"TangentP", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[965]:=",ExpressionUUID->"8d180f07-ad3b-421f-b5fb-c752a1cd7614"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\[Phi]A", ",", "\[Phi]B", ",", "\[Phi]C", ",", "\[Phi]D"}], 
    "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978486701942*^9, 3.792805460196883*^9},
 CellLabel->"Out[965]=",
 CellID->1394865606,ExpressionUUID->"11bd1512-88d8-434a-b88d-cfe67fb3a67e"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Tensor fields defined on the pullback bundles are called \
\[OpenCurlyDoubleQuote]tensors along a map\[CloseCurlyDoubleQuote] or \
\[OpenCurlyDoubleQuote]pullback sections\[CloseCurlyDoubleQuote]. They will \
carry distinctive \[OpenCurlyDoubleQuote]starred indices\
\[CloseCurlyDoubleQuote], as a reminder that dependencies have been changed \
(this is a bit like the DeWitt index notation). In this way, for example, we \
will not be able to contract tensors belonging to different vbundles. The \
most important example is the differential of the mapping, described in the \
next section. We will then see two other examples: precomposition and linear \
push of tensor fields.\
\>", "Text",ExpressionUUID->"3f3fad20-3df4-4b9b-82e9-a5e75a043bdd"],

Cell[CellGroupData[{

Cell[TextData[{
 "Since version 1.1.0 of ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ", tensors can be natively tensors along a map. This is a section of the \
pullback bundle \[Phi]*\[DoubleStruckCapitalT]P, because we use an index of \
that vbundle:"
}], "MathCaption",ExpressionUUID->"abc38903-6859-4247-9057-8a37a08dfff3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"W", "[", "\[Phi]A", "]"}], ",", "M"}], "]"}]], "Input",
 CellLabel->
  "In[966]:=",ExpressionUUID->"4f68c0a6-049c-4839-b19f-05eb295f232a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"W", "[", "\[Phi]A", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`W[$CellContext`\[Phi]A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978486828211*^9, 3.792805462770262*^9},
 CellLabel->"During evaluation of In[966]:=",
 CellID->1629058196,ExpressionUUID->"b53223de-ff49-4104-bb7b-152ca4d74a2e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "W", "]"}]], "Input",
 CellLabel->
  "In[967]:=",ExpressionUUID->"a923efa3-8437-4e9b-97be-0c14629773f0"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SuperscriptBox["\[Phi]", "*"], 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978486962223*^9, 3.792805463474318*^9},
 CellLabel->"Out[967]=",
 CellID->1431452682,ExpressionUUID->"39d51741-5883-4029-8024-e211e3a2aae9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "W", "]"}]], "Input",
 CellLabel->
  "In[968]:=",ExpressionUUID->"ad5b95ea-66c1-435b-b483-6675072a8fac"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978487095113*^9, 3.792805464138433*^9},
 CellLabel->"Out[968]=",
 CellID->1879865030,ExpressionUUID->"c085b3c9-febb-4402-97c3-ac40262e273a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Indices of a pullback bundle are represented with a star. We need to look at \
their input form to see the actual mapping.\
\>", "MathCaption",ExpressionUUID->"3bac2e83-233b-467f-8c5f-7bbe9fcfcbb2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"W", "[", "\[Phi]A", "]"}]], "Input",
 CellLabel->
  "In[969]:=",ExpressionUUID->"d633dfd4-6fa3-4a8b-bd31-9392a1a84b6c"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"W", 
      StyleBox[GridBox[{
         {
          SuperscriptBox["A", "*"]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`W[$CellContext`\[Phi]A],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978487224628*^9, 3.792805465504715*^9},
 CellLabel->"Out[969]=",
 CellID->870884032,ExpressionUUID->"a4e661b8-c007-4a3f-8683-021868c90906"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Algebraic operations proceed normally:", "MathCaption",ExpressionUUID->"4f495177-a20e-44d3-8860-1e90bcc9e2bb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"W", "[", "\[Phi]A", "]"}], 
   RowBox[{"delta", "[", 
    RowBox[{
     RowBox[{"-", "\[Phi]A"}], ",", "\[Phi]B"}], "]"}]}], "-", 
  RowBox[{"7", 
   RowBox[{"W", "[", "\[Phi]B", "]"}]}]}]], "Input",
 CellLabel->
  "In[970]:=",ExpressionUUID->"d95aa074-b024-4322-ae51-056821f3113d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", "6"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"W", 
       StyleBox[GridBox[{
          {
           SuperscriptBox["B", "*"]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`W[$CellContext`\[Phi]B],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978487344174*^9, 3.7928054761785707`*^9},
 CellLabel->"Out[970]=",
 CellID->1799799912,ExpressionUUID->"a0008b71-fdd7-4141-ad32-c4b2562f8741"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Undefine the field:", "MathCaption",ExpressionUUID->"3ccf19cc-4caf-45b3-a0c9-25d8c7b6bdf2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "W", "]"}]], "Input",
 CellLabel->
  "In[971]:=",ExpressionUUID->"840f1352-b6dc-4248-9151-423ae53da654"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "W"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`W],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978487460059*^9, 3.7928054816892633`*^9},
 CellLabel->"During evaluation of In[971]:=",
 CellID->1735349372,ExpressionUUID->"32af6adb-383f-4ed0-9a6f-aa9dbc00eb5a"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Note: in our scheme of things, \[OpenCurlyDoubleQuote]pull back vbundle\
\[CloseCurlyDoubleQuote] is not the best name. It should have been \
\[OpenCurlyDoubleQuote]precomposed vbundle\[CloseCurlyDoubleQuote], but the \
former is standard, so that is what we use.\
\>", "Text",ExpressionUUID->"a26bdf63-9b31-4fc3-b789-542f5e191297"]
}, Closed]],

Cell[CellGroupData[{

Cell["8.5. Differential of a mapping", "Subsection",ExpressionUUID->"03843fde-55e9-4cb0-a90f-902509ec21a5"],

Cell[TextData[{
 "Associated to each smooth mapping we have its differential, also known as \
\[OpenCurlyDoubleQuote]tangent map\[CloseCurlyDoubleQuote]. This is a linear \
operator mapping vectors on a vbundle to vectors on a pullback bundle of the \
same base. It will hence be represented in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " as a 2-tensor with a normal index and a starred index."
}], "Text",ExpressionUUID->"d6c263a2-c6a2-4478-a3a1-bb4f22d39203"],

Cell[TextData[{
 StyleBox[ButtonBox["TangentTensor",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TangentTensor.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Differential of a mapping\n", "Text"],
 StyleBox[ButtonBox["Inv",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Inv.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t\t\t", "DisplayMath"],
 StyleBox["Inverse of a 2-tensor", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978493276936*^9},ExpressionUUID->"45705199-d5cc-4326-897b-\
2f5b1970316d"],

Cell["Mapping differentials.", "Text",ExpressionUUID->"4db538bd-45ab-4414-a536-90cdc3680279"],

Cell[CellGroupData[{

Cell[TextData[{
 "This is the tangent map of ",
 StyleBox["\[Phi]",
  FontFamily->"Courier"],
 ", by default constructed by prepending ",
 StyleBox["d",
  FontFamily->"Courier"],
 " to the name of the mapping:"
}], "MathCaption",ExpressionUUID->"cadd3ec2-22b1-4f48-9e6f-635af6c69cb5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentTensor", "[", "\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[972]:=",ExpressionUUID->"1128b329-76d7-4dff-8110-f5fed822ca13"],

Cell[BoxData["d\[Phi]"], "Output",
 CellChangeTimes->{3.728978497178871*^9, 3.792805528560298*^9},
 CellLabel->"Out[972]=",
 CellID->753785169,ExpressionUUID->"2474e17a-645a-4363-b0cb-adecb2d586c3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "It is a tensor field on the manifold ",
 StyleBox["M",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"50298c0f-da86-47a0-865c-3694fbe3769e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xTensorQ", "[", "d\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[973]:=",ExpressionUUID->"05460950-b2f1-43f6-bf87-289669da142a"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.7289784973002*^9, 3.7928055315438538`*^9},
 CellLabel->"Out[973]=",
 CellID->160295255,ExpressionUUID->"28b8a97b-193a-4dd1-a6f8-9619532bf359"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "d\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[974]:=",ExpressionUUID->"c73bf1cb-fdd6-42d0-b218-319c6f218c84"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978497413406*^9, 3.792805532958851*^9},
 CellLabel->"Out[974]=",
 CellID->379308330,ExpressionUUID->"00343745-6f4c-40b8-883f-bfed8bcadeee"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Its two slots belong to different vbundles, the second one being a pullback \
bundle:\
\>", "MathCaption",ExpressionUUID->"faf4806f-11c4-49db-a401-accaa6a56242"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "d\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[975]:=",ExpressionUUID->"65fcc3e7-de6e-470b-95c6-a7b9d0749e44"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM,
     Editable->False]}], ",", 
   RowBox[{
    SuperscriptBox["\[Phi]", "*"], 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]P",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentP,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978497699992*^9, 3.7928055368004932`*^9},
 CellLabel->"Out[975]=",
 CellID->586021695,ExpressionUUID->"9febfc27-e1e6-4d05-bf83-23874d6454c1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Both bundles have the same base, which coincides with the dependency of the \
tangent map (that\[CloseCurlyQuote]s why it would be incorrect to have ",
 StyleBox["\[DoubleStruckCapitalT]P",
  FontFamily->"Courier"],
 " as a slot of ",
 StyleBox["d\[Phi]",
  FontFamily->"Courier"],
 ": its dependency would be ",
 StyleBox["P",
  FontFamily->"Courier"],
 "):"
}], "MathCaption",ExpressionUUID->"1b7d5cea-1eaf-4f48-8c77-c941dbee9f82"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BaseOfVBundle", "/@", "%"}]], "Input",
 CellLabel->
  "In[976]:=",ExpressionUUID->"37bdcb9b-b3ff-4e47-ade7-14dc7a0b7fbe"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"M", ",", "M"}], "}"}]], "Output",
 CellChangeTimes->{3.728978497935701*^9, 3.792805546361824*^9},
 CellLabel->"Out[976]=",
 CellID->1406927212,ExpressionUUID->"97a1ee0d-d465-4b5f-a10a-6e8b228525b8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
With indices we have the following expression, where we need a starred index.\
\
\>", "MathCaption",ExpressionUUID->"20d7e44e-768a-4cf8-af29-7a7943bdfca8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"d\[Phi]", "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", "\[Phi]B"}], "]"}]], "Input",
 CellLabel->
  "In[977]:=",ExpressionUUID->"ec1fc1de-b8aa-468a-bb8c-c8ad7f032ae6"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[DifferentialD]", "\[Phi]"}], 
      StyleBox[GridBox[{
         {" ", 
          SuperscriptBox["B", "*"]},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`d\[Phi][-$CellContext`a, $CellContext`\[Phi]B],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289784981458*^9, 3.79280554951196*^9},
 CellLabel->"Out[977]=",
 CellID->827479770,ExpressionUUID->"1ade713b-68d7-47ba-be75-ec2e22efdf8a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The inverse mapping ",
 StyleBox["i\[Phi]",
  FontFamily->"Courier"],
 " has its own tangent map:"
}], "MathCaption",ExpressionUUID->"2ec69836-96a6-499e-9e03-91c87b3734b2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TangentTensor", "[", "i\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[978]:=",ExpressionUUID->"306db61e-2be1-411d-a9cf-de47af112275"],

Cell[BoxData["di\[Phi]"], "Output",
 CellChangeTimes->{3.7289784984008427`*^9, 3.7928055530794287`*^9},
 CellLabel->"Out[978]=",
 CellID->1424869359,ExpressionUUID->"a8b3b216-e209-4222-a24c-29c6d06c9055"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "This is now a field on ",
 StyleBox["P",
  FontFamily->"Courier"],
 ", and has slots in ",
 StyleBox["-\[DoubleStruckCapitalT]P",
  FontFamily->"Courier"],
 " (this is ",
 StyleBox["xAct",
  FontSlant->"Italic"],
 " notation for the dual ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[DoubleStruckCapitalT]", "*"], TraditionalForm]],
  ExpressionUUID->"33da5424-34b9-4e06-8906-7777842da548"],
 StyleBox["P",
  FontFamily->"Courier"],
 " of ",
 StyleBox["\[DoubleStruckCapitalT]P",
  FontFamily->"Courier"],
 ") and the pullback by ",
 StyleBox["i\[Phi]",
  FontFamily->"Courier"],
 " of ",
 StyleBox["\[DoubleStruckCapitalT]M",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"6889b5f5-4ca1-4dc3-a66c-8fb55106be38"],

Cell[CellGroupData[{

Cell["\<\
(Jose: The -1 of the inverse is inconvenient. Use i\[Phi] instead)\
\>", "Text",ExpressionUUID->"29912bc3-4150-4410-9013-d46b7cffafa6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "di\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[979]:=",ExpressionUUID->"3d14eb67-b1e4-4136-aaff-b1a0a40f9f3b"],

Cell[BoxData[
 RowBox[{"{", "P", "}"}]], "Output",
 CellChangeTimes->{3.728978498601367*^9, 3.792805569590612*^9},
 CellLabel->"Out[979]=",
 CellID->142462808,ExpressionUUID->"0e9b37e1-2796-4fa2-a311-92d8a5645649"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "di\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[980]:=",ExpressionUUID->"49c9a0c9-92c9-41de-b7b9-250df8da171d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]P",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentP,
     Editable->False]}], ",", 
   RowBox[{
    SuperscriptBox[
     SuperscriptBox["\[Phi]", 
      RowBox[{"-", "1"}]], "*"], 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.72897849903052*^9, 3.7928055718705072`*^9},
 CellLabel->"Out[980]=",
 CellID->1400460588,ExpressionUUID->"e862672a-e774-4846-bc24-55dd284c7d01"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BaseOfVBundle", "/@", "%"}]], "Input",
 CellLabel->
  "In[981]:=",ExpressionUUID->"726a4856-180a-421b-be7f-cab56a80a581"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"P", ",", "P"}], "}"}]], "Output",
 CellChangeTimes->{3.728978499200142*^9, 3.792805577281076*^9},
 CellLabel->"Out[981]=",
 CellID->1231328694,ExpressionUUID->"44c6da18-a252-456d-93a7-f35ad52bf1c5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Using the corresponding indices we have:", "MathCaption",ExpressionUUID->"c90df122-2231-4ea4-a904-e743d4b70783"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"di\[Phi]", "[", 
  RowBox[{
   RowBox[{"-", "A"}], ",", "i\[Phi]b"}], "]"}]], "Input",
 CellLabel->
  "In[982]:=",ExpressionUUID->"29ad5e6c-8ffd-4a23-ba17-47f769eede72"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"\[DifferentialD]", "i\[Phi]"}], 
      StyleBox[GridBox[{
         {" ", 
          SuperscriptBox["b", "*"]},
         {"A", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`di\[Phi][-$CellContext`A, $CellContext`i\[Phi]b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978499596693*^9, 3.792805579232092*^9},
 CellLabel->"Out[982]=",
 CellID->510434630,ExpressionUUID->"1d6b3067-c12f-4f4c-ba90-8fca6d64ccbf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["d\[Phi]",
  FontFamily->"Courier"],
 " and ",
 StyleBox["di\[Phi]",
  FontFamily->"Courier"],
 " are NOT mutual inverses. That\[CloseCurlyQuote]s clear in two facts: 1) \
they are fields on different manifolds, and 2) they have indices on different \
vbundles. However there is a close relation. We use ",
 StyleBox["Inv",
  FontFamily->"Courier"],
 " to compute the inverse of a 2-tensor:"
}], "MathCaption",ExpressionUUID->"3472033a-0c5e-4a94-8d72-8ea3751acad3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Inv", "[", "d\[Phi]", "]"}], "[", 
  RowBox[{
   RowBox[{"-", "\[Phi]A"}], ",", "b"}], "]"}]], "Input",
 CellLabel->
  "In[983]:=",ExpressionUUID->"1e969b38-8057-49a1-a92a-37877fcf2e0d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[DifferentialD]", "i\[Phi]"}], "\[SmallCircle]", "\[Phi]"}],
        ")"}], 
      StyleBox[GridBox[{
         {" ", "b"},
         {
          SuperscriptBox["A", "*"], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Precompose[$CellContext`di\[Phi], \
$CellContext`\[Phi]][-$CellContext`\[Phi]A, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978499778391*^9, 3.792805604168486*^9},
 CellLabel->"Out[983]=",
 CellID->59630773,ExpressionUUID->"eae09354-afed-4af6-b625-95dbeca41c51"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[984]:=",ExpressionUUID->"8e7d3a41-2d46-439f-be98-cd56d80839af"],

Cell["Precompose[di\[Phi], \[Phi]][-\[Phi]A, b]", "Output",
 CellChangeTimes->{3.728978499903768*^9, 3.7928056051988297`*^9},
 CellLabel->"Out[984]//InputForm=",
 CellID->420825556,ExpressionUUID->"a8c7059e-a3f4-4925-844c-141d6ff1f2f2"]
}, Open  ]]
}, Open  ]],

Cell["\<\
If the mapping is not a diffeomorphism then its differential may be only \
left-invertible or right-invertible or none.\
\>", "Text",ExpressionUUID->"dee68c3d-7afb-4fd7-8bf0-006e05d25cc8"],

Cell["\<\
(Jose: The second argument of InvertibleQ is ugly. Use separate functions \
LeftInvertibleQ and RightInvertibleQ)\
\>", "Text",ExpressionUUID->"f5e6c6df-c8e9-4911-8a90-82ecb1269ee0"],

Cell[CellGroupData[{

Cell[TextData[{
 "Because ",
 StyleBox["\[Phi]",
  FontFamily->"Courier"],
 " is a diffeomorphism its tangent map is invertible from both sides:"
}], "MathCaption",ExpressionUUID->"b930d27e-52a3-4fcf-9f3f-065462436ea6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"d\[Phi]", ",", "Left"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"d\[Phi]", ",", "Right"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", "d\[Phi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[985]:=",ExpressionUUID->"87bd98ea-07f4-411d-9769-6c04b5cf0013"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{3.728978500083744*^9, 3.792805630071053*^9},
 CellLabel->"Out[985]=",
 CellID->1531935448,ExpressionUUID->"c65bce88-bebf-4a61-b130-09a3be1cdaaa"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And so is the differential of its inverse:", "MathCaption",ExpressionUUID->"e6cc5a2f-e8be-4e11-965c-6090528fc675"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"di\[Phi]", ",", "Left"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"di\[Phi]", ",", "Right"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", "di\[Phi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[986]:=",ExpressionUUID->"8a50fbb5-8e0d-4b97-9e62-5c7a554fec12"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "True", ",", "True"}], "}"}]], "Output",
 CellChangeTimes->{3.728978500219616*^9, 3.7928056328390303`*^9},
 CellLabel->"Out[986]=",
 CellID->1718202283,ExpressionUUID->"5dc54a68-6c07-43be-a66d-550c92623445"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "However, the mapping ",
 StyleBox["\[Psi]",
  FontFamily->"Courier"],
 " is only an immersion (note the unfortunate fact that ",
 StyleBox["\[Psi]",
  FontFamily->"Courier"],
 " is left-invertible but its differential is right invertible, due to our \
choice down-up for the delta, instead of the standard up-down):"
}], "MathCaption",ExpressionUUID->"f3c97a73-0044-40ed-8d8d-8161d51ac56b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"d\[Psi]", ",", "Left"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"d\[Psi]", ",", "Right"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", "d\[Psi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[987]:=",ExpressionUUID->"43633553-75a9-47fd-a86d-03a24d5b2390"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"False", ",", "True", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.7289785003644876`*^9, 3.7928056347087593`*^9},
 CellLabel->"Out[987]=",
 CellID->168189705,ExpressionUUID->"1e866e49-535f-43f9-b93a-a469b3cf61c2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"d\[Psi]", "[", 
   RowBox[{
    RowBox[{"-", "A"}], ",", "\[Psi]\[Alpha]"}], "]"}], 
  RowBox[{
   RowBox[{"Inv", "[", "d\[Psi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "\[Psi]\[Alpha]"}], ",", "B"}], "]"}]}]], "Input",
 CellLabel->
  "In[988]:=",ExpressionUUID->"4e9979ed-9b17-43db-b5bb-783e3f55f8bd"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "B"},
         {"A", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`A, $CellContext`B],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978500498498*^9, 3.7928056377323723`*^9},
 CellLabel->"Out[988]=",
 CellID->1060894942,ExpressionUUID->"aac8e908-cb32-4077-9083-c2779798d895"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Inv", "[", "d\[Psi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "\[Psi]\[Alpha]"}], ",", "A"}], "]"}], 
  RowBox[{"d\[Psi]", "[", 
   RowBox[{
    RowBox[{"-", "A"}], ",", "\[Psi]\[Beta]"}], "]"}]}]], "Input",
 CellLabel->
  "In[989]:=",ExpressionUUID->"e7ffb67d-2496-4879-9146-c414099cc46b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Psi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["\[Beta]", "*"]},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Psi][-$CellContext`A, $CellContext`\[Psi]\[Beta]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "i\[Psi]"}], "\[SmallCircle]", 
         "\[Psi]"}], ")"}], 
       StyleBox[GridBox[{
          {" ", "A"},
          {
           SuperscriptBox["\[Alpha]", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`di\[Psi], \
$CellContext`\[Psi]][-$CellContext`\[Psi]\[Alpha], $CellContext`A],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978500645775*^9, 3.792805638597969*^9},
 CellLabel->"Out[989]=",
 CellID->931757711,ExpressionUUID->"c073d59b-2570-4c0a-ac10-2c28ad428ce8"]
}, Open  ]]
}, Open  ]],

Cell["\<\
That\[CloseCurlyQuote]s a rather complicated expression, but it is what it is.\
\>", "MathCaption",ExpressionUUID->"1440df59-5266-4938-a972-8e43fd01f697"],

Cell[CellGroupData[{

Cell[TextData[{
 "The properties of the differential of ",
 StyleBox["i\[Psi]",
  FontFamily->"Courier"],
 " are reversed with respect to those of ",
 StyleBox["\[Psi]",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"e76b60c5-f3e9-42b3-a982-75cfadaa353a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"di\[Psi]", ",", "Left"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", 
    RowBox[{"di\[Psi]", ",", "Right"}], "]"}], ",", 
   RowBox[{"InvertibleQ", "[", "di\[Psi]", "]"}]}], "}"}]], "Input",
 CellLabel->
  "In[990]:=",ExpressionUUID->"98c8a931-a040-44d7-afe6-dbf34c75f1e2"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"True", ",", "False", ",", "False"}], "}"}]], "Output",
 CellChangeTimes->{3.728978500795618*^9, 3.792805657573265*^9},
 CellLabel->"Out[990]=",
 CellID->1748009716,ExpressionUUID->"f7d5e6d8-862e-4783-b234-4c7508460b02"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"di\[Psi]", "[", 
   RowBox[{
    RowBox[{"-", "\[Psi]\[Alpha]"}], ",", "A"}], "]"}], 
  RowBox[{
   RowBox[{"Inv", "[", "di\[Psi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "A"}], ",", "\[Psi]\[Beta]"}], "]"}]}]], "Input",
 CellLabel->
  "In[991]:=",ExpressionUUID->"083a860e-8575-46c4-9c46-2ab50b63923b"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "i\[Psi]"}], 
       StyleBox[GridBox[{
          {" ", "A"},
          {
           SuperscriptBox["\[Alpha]", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`di\[Psi][-$CellContext`\[Psi]\[Alpha], $CellContext`A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "\[Psi]"}], "\[SmallCircle]", 
         SuperscriptBox["\[Psi]", 
          RowBox[{"-", "1"}]]}], ")"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["\[Beta]", "*"]},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`d\[Psi], \
$CellContext`i\[Psi]][-$CellContext`A, $CellContext`\[Psi]\[Beta]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978500932432*^9, 3.7928056585148277`*^9},
 CellLabel->"Out[991]=",
 CellID->779430251,ExpressionUUID->"64e4e4a6-b5ce-413e-aded-219769116212"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Inv", "[", "di\[Psi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "A"}], ",", "\[Psi]\[Alpha]"}], "]"}], 
  RowBox[{"di\[Psi]", "[", 
   RowBox[{
    RowBox[{"-", "\[Psi]\[Alpha]"}], ",", "B"}], "]"}]}]], "Input",
 CellLabel->
  "In[992]:=",ExpressionUUID->"6bb8f5dc-9f93-4e9a-81c4-5e34b8c338fe"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "B"},
         {"A", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`A, $CellContext`B],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978501064232*^9, 3.792805659268145*^9},
 CellLabel->"Out[992]=",
 CellID->1247137483,ExpressionUUID->"e09b9507-a6b1-4007-865c-94c390a8a556"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.6. Precomposition", "Subsection",ExpressionUUID->"e2123d25-7e55-4223-968d-a6dce190ac14"],

Cell[TextData[{
 "Using precomposition we modify the dependencies of a tensor, but do not \
change essentially its slot structure (we only \[OpenCurlyDoubleQuote]add \
stars\[CloseCurlyDoubleQuote]). In particular, this operation is still \
nontrivial on scalar fields.\nPrecomposition is very similar to composition (",
 StyleBox["SmallCircle",
  FontFamily->"Courier"],
 "). We need to distinguish them because mappings and tensor fields are two \
different types of objects in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 ". That is, a mapping is ",
 StyleBox["composed",
  FontSlant->"Italic"],
 " with another mapping, and a tensor field is ",
 StyleBox["precomposed",
  FontSlant->"Italic"],
 " with a mapping. This is a fundamental design decision."
}], "Text",ExpressionUUID->"cde23a58-c689-42fb-87fd-3be3baef3cfa"],

Cell[TextData[{
 StyleBox[ButtonBox["Precompose",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Precompose.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Precomposition of a tensor field by a mapping", "Text"]
}], "DefinitionBox",ExpressionUUID->"b65fb66c-aead-4a20-b827-77aa136dcc98"],

Cell["Precomposition.", "Text",ExpressionUUID->"e5727659-7120-47f5-8e59-bdf44f6a1500"],

Cell[CellGroupData[{

Cell["\<\
By precomposition we change the properties of a tensor. For example:\
\>", "MathCaption",ExpressionUUID->"3f481555-13dc-4c06-9240-c0c1e3ff538b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "B"}]}], "]"}], ",", "P"}], "]"}]], "Input",
 CellLabel->
  "In[993]:=",ExpressionUUID->"0098a684-2717-4a37-b653-b886c8ef9f4f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"T", "[", 
    RowBox[{"A", ",", 
     RowBox[{"-", "B"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`T[$CellContext`A, -$CellContext`B], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785069132442`*^9, 3.792805714217061*^9},
 CellLabel->"During evaluation of In[993]:=",
 CellID->1425323066,ExpressionUUID->"8dc0f115-e04d-4c8b-a377-962eff512748"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "T", "]"}]], "Input",
 CellLabel->
  "In[994]:=",ExpressionUUID->"67d2bd67-c3da-412e-9dfb-2571f0952926"],

Cell[BoxData[
 RowBox[{"{", "P", "}"}]], "Output",
 CellChangeTimes->{3.7289785070246897`*^9, 3.792805716009985*^9},
 CellLabel->"Out[994]=",
 CellID->628488876,ExpressionUUID->"29cf62fc-519f-4877-82d3-37b24eddc6c4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "T", "]"}]], "Input",
 CellLabel->
  "In[995]:=",ExpressionUUID->"59bace9d-ef9a-4cc2-a2ad-159c442e703f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]P",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentP,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289785071455803`*^9, 3.792805717186069*^9},
 CellLabel->"Out[995]=",
 CellID->452688226,ExpressionUUID->"cff1484f-ebb7-413f-b292-337e88c794dd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["If we precompose the tensor we get:", "MathCaption",ExpressionUUID->"2cc75d7c-8646-4349-bb36-8ea96a499ed1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", 
  RowBox[{"Precompose", "[", 
   RowBox[{"T", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[996]:=",ExpressionUUID->"89670031-bbf8-4b20-8577-c13757f43a5e"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978507264328*^9, 3.7928057256577377`*^9},
 CellLabel->"Out[996]=",
 CellID->1903426450,ExpressionUUID->"deee789f-aeaf-48cc-a503-683649c85e6d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", 
  RowBox[{"Precompose", "[", 
   RowBox[{"T", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[997]:=",ExpressionUUID->"fa289ff0-222f-4909-92e7-761f279a68da"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SuperscriptBox["\[Phi]", "*"], 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]P",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentP,
     Editable->False]}], ",", 
   RowBox[{"-", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["\[Phi]", "*"], 
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]P",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentP,
       Editable->False]}], ")"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978507535049*^9, 3.7928057315940523`*^9},
 CellLabel->"Out[997]=",
 CellID->824102103,ExpressionUUID->"94688858-5f23-4824-9354-719fb7132edd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "That explains why the properties of the precomposition of ",
 StyleBox["di\[Phi]",
  FontFamily->"Courier"],
 " are compatible with those of ",
 StyleBox["d\[Phi]",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"34575066-a9ef-4bb1-85c0-a48aa87c2079"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", 
  RowBox[{"Precompose", "[", 
   RowBox[{"di\[Phi]", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[998]:=",ExpressionUUID->"8a38463a-8159-4146-bcb0-7af78548c6ee"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.72897850787978*^9, 3.7928057720223*^9},
 CellLabel->"Out[998]=",
 CellID->1310595457,ExpressionUUID->"935138ad-fcf9-4c0b-a8a9-0b1af59d48ca"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", 
  RowBox[{"Precompose", "[", 
   RowBox[{"di\[Phi]", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[999]:=",ExpressionUUID->"c07e384d-d93d-41fa-a6ca-a295198cd369"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["\[Phi]", "*"], 
      InterpretationBox[
       StyleBox["\[DoubleStruckCapitalT]P",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`TangentP,
       Editable->False]}], ")"}]}], ",", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.7289785080151987`*^9, 3.79280577316812*^9},
 CellLabel->"Out[999]=",
 CellID->2127789970,ExpressionUUID->"bc2b2fed-76e1-47a2-8e26-d0fa9d26af93"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "d\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[1000]:=",ExpressionUUID->"ee320269-9fd4-4438-b3b3-adab5173888f"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978508319058*^9, 3.79280577381566*^9},
 CellLabel->"Out[1000]=",
 CellID->236700852,ExpressionUUID->"c97cfcbb-f0b0-4f11-885d-eda00c8d83ef"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "d\[Phi]", "]"}]], "Input",
 CellLabel->
  "In[1001]:=",ExpressionUUID->"1f7a9b5e-bba0-436a-a56a-9f4a20c74d44"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]M",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentM,
     Editable->False]}], ",", 
   RowBox[{
    SuperscriptBox["\[Phi]", "*"], 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]P",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentP,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978508499723*^9, 3.7928057744481*^9},
 CellLabel->"Out[1001]=",
 CellID->28232387,ExpressionUUID->"a1ad6c8b-a8d5-4a45-8aed-82db0ca523e9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["And hence now we can contract in four different ways:", "MathCaption",ExpressionUUID->"e78f6828-c888-4ffe-9e8d-57deeddda654"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"d\[Phi]", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "\[Phi]B"}], "]"}], 
  RowBox[{
   RowBox[{"Inv", "[", "d\[Phi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "\[Phi]B"}], ",", "c"}], "]"}]}]], "Input",
 CellLabel->
  "In[1002]:=",ExpressionUUID->"47885fa1-ddab-488b-9055-a290e413bd04"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "c"},
         {"a", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`a, $CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289785088852262`*^9, 3.792805778944458*^9},
 CellLabel->"Out[1002]=",
 CellID->504730059,ExpressionUUID->"269511c5-0a84-4fa2-a065-8bb0c463f9bf"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Inv", "[", "d\[Phi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "\[Phi]A"}], ",", "b"}], "]"}], 
  RowBox[{"d\[Phi]", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", "\[Phi]C"}], "]"}]}]], "Input",
 CellLabel->
  "In[1003]:=",ExpressionUUID->"2488b393-9ded-485f-8f2b-ddac9d6402ce"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", 
          SuperscriptBox["C", "*"]},
         {
          SuperscriptBox["A", "*"], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`\[Phi]A, $CellContext`\[Phi]C],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978509030024*^9, 3.792805779648797*^9},
 CellLabel->"Out[1003]=",
 CellID->370035437,ExpressionUUID->"d53c4b6b-82c8-4c54-8e9c-1b10d69c0d11"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"di\[Phi]", "[", 
   RowBox[{
    RowBox[{"-", "A"}], ",", "i\[Phi]b"}], "]"}], 
  RowBox[{
   RowBox[{"Inv", "[", "di\[Phi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "i\[Phi]b"}], ",", "C"}], "]"}]}]], "Input",
 CellLabel->
  "In[1004]:=",ExpressionUUID->"1f23fcbc-285d-4c96-9291-42ca1de2326a"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", "C"},
         {"A", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`A, C],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289785095355587`*^9, 3.7928057802325773`*^9},
 CellLabel->"Out[1004]=",
 CellID->1106157631,ExpressionUUID->"47c93f95-fe6e-4fb8-9d5e-4eb57f81c1ac"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Inv", "[", "di\[Phi]", "]"}], "[", 
   RowBox[{
    RowBox[{"-", "i\[Phi]a"}], ",", "B"}], "]"}], 
  RowBox[{"di\[Phi]", "[", 
   RowBox[{
    RowBox[{"-", "B"}], ",", "i\[Phi]c"}], "]"}]}]], "Input",
 CellLabel->
  "In[1005]:=",ExpressionUUID->"5978cfa1-4ebe-4e4f-b63c-58ff27a03d71"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", 
          SuperscriptBox["c", "*"]},
         {
          SuperscriptBox["a", "*"], " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`delta[-$CellContext`i\[Phi]a, $CellContext`i\[Phi]c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978509701207*^9, 3.792805780843279*^9},
 CellLabel->"Out[1005]=",
 CellID->451529215,ExpressionUUID->"6921614a-5fff-4458-a83b-f6d6ddca7765"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Another possible example of the need to be careful with indices is in the \
computation of the differential of a composition. The differential of ",
 StyleBox["\[Psi]\[SmallCircle]\[Phi]",
  FontFamily->"Courier"],
 " is the product of their differentials, but these are fields in different \
vbundles, so there must be precomposition."
}], "Text",ExpressionUUID->"ff59333d-4889-4982-8fb3-41431a6f3275"],

Cell[CellGroupData[{

Cell["We will work with a double pullback vbundle,", "MathCaption",ExpressionUUID->"276ef982-e27c-4ec7-b333-de66dd89a424"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PullBackVBundle", "[", 
  RowBox[{"TangentL", ",", "\[Psi]"}], "]"}]], "Input",
 CellLabel->
  "In[1006]:=",ExpressionUUID->"2f13f3c9-bea6-4d76-bf3f-7fd1481b1967"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox["\[Psi]", "*"], 
  InterpretationBox[
   StyleBox["\[DoubleStruckCapitalT]L",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TangentL,
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978509867621*^9, 3.7928057947434072`*^9},
 CellLabel->"Out[1006]=",
 CellID->2105705276,ExpressionUUID->"91c4380f-04fb-4215-a083-2f9e4d3c4085"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PullBackVBundle", "[", 
  RowBox[{"%", ",", "\[Phi]"}], "]"}]], "Input",
 CellLabel->
  "In[1007]:=",ExpressionUUID->"0f37d232-d1b9-4500-85ee-ea80a848c9e0"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   RowBox[{"(", 
    RowBox[{"\[Psi]", "\[SmallCircle]", "\[Phi]"}], ")"}], "*"], 
  InterpretationBox[
   StyleBox["\[DoubleStruckCapitalT]L",
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`TangentL,
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785100015697`*^9, 3.792805795668355*^9},
 CellLabel->"Out[1007]=",
 CellID->1500552726,ExpressionUUID->"b1b18b94-0d4e-4ba0-8c9a-ce7379ae6e18"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1008]:=",ExpressionUUID->"54929481-42c0-496e-83fe-59e90be0b196"],

Cell["PullBackVBundle[TangentL, \[Psi] \[SmallCircle] \[Phi]]", "Output",
 CellChangeTimes->{3.728978510125551*^9, 3.792805796287016*^9},
 CellLabel->"Out[1008]//InputForm=",
 CellID->906924670,ExpressionUUID->"60c889ca-1dfd-46e6-ba3e-1831c80e7192"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["whose indices are \[OpenCurlyDoubleQuote]double mapped\
\[CloseCurlyDoubleQuote]:", "MathCaption",ExpressionUUID->"34da694f-acfe-47f9-\
9443-c01015d9eed2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndicesOfVBundle", "[", "%%", "]"}]], "Input",
 CellLabel->
  "In[1009]:=",ExpressionUUID->"1c87b923-d2a2-47cd-bbdc-73cc56912f40"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "\[Phi]\[Psi]\[Alpha]", ",", "\[Phi]\[Psi]\[Beta]", ",", 
     "\[Phi]\[Psi]\[Gamma]", ",", "\[Phi]\[Psi]\[Delta]"}], "}"}], ",", 
   RowBox[{"{", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289785103644753`*^9, 3.792805797431673*^9},
 CellLabel->"Out[1009]=",
 CellID->1493694214,ExpressionUUID->"5e667112-1ad3-4a0b-b527-6954e492b61d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The order of mappings is reversed in the index names (as a result of pull \
",
 StyleBox["back",
  FontSlant->"Italic"],
 "). It would have been more natural to write the mappings on the right or \
the indices, to keep the order, but we already have the dagger, dollar and \
number postfixes for indices, and I prefer not to mess up with those other \
more important marks."
}], "MathCaption",ExpressionUUID->"1d431271-f7fb-47b9-986f-42f7dff414f2"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellLabel->
  "In[1010]:=",ExpressionUUID->"37fb181d-a668-4afc-af69-994f659e0ca5"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This is the differential of the composition \[Psi]\[SmallCircle]\[Phi] (as a \
contraction of linear operators, the order is reversed unfortunately, as we \
mentioned before):\
\>", "MathCaption",ExpressionUUID->"461d2f03-410a-49dd-8913-93920a8bb7cb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TangentTensor", "[", 
   RowBox[{"\[Psi]", "\[SmallCircle]", "\[Phi]"}], "]"}], "[", 
  RowBox[{
   RowBox[{"-", "a"}], ",", "\[Phi]\[Psi]\[Alpha]"}], "]"}]], "Input",
 CellLabel->
  "In[1011]:=",ExpressionUUID->"969dda22-3c72-4f4a-8052-e775e87bd69d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["A", "*"]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`a, $CellContext`\[Phi]A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "\[Psi]"}], "\[SmallCircle]", "\[Phi]"}],
         ")"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["\[Alpha]", "**"]},
          {
           SuperscriptBox["A", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`d\[Psi], \
$CellContext`\[Phi]][-$CellContext`\[Phi]A, $CellContext`\[Phi]\[Psi]\[Alpha]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785105608807`*^9, 3.79280582358348*^9},
 CellLabel->"Out[1011]=",
 CellID->1323687785,ExpressionUUID->"c974f4b5-52e8-402a-b9b5-e5ea4e12b402"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
As we had anticipated, we need precomposition of one of the differentials in \
order to have consistent dependencies:\
\>", "MathCaption",ExpressionUUID->"b0f15d98-e44c-4158-804f-c54e0d5d4d59"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1012]:=",ExpressionUUID->"dc187a7b-6286-442a-8e2e-195d913fd752"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978510682675*^9, 3.792805826294669*^9},
 CellLabel->"Out[1012]=",
 CellID->1718687517,ExpressionUUID->"b0afc6ef-9d3d-4904-a98e-ee61801ee656"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.7. Pullback", "Subsection",ExpressionUUID->"e751d01c-1f5f-4abf-81d1-945766380919"],

Cell["\<\
Here we find the first fundamental operation of working with tensors and \
mappings: the pullback of a tensor field on the image manifold to a tensor \
field on the domain manifold. This operation is always defined for covariant \
tensor fields, and can be extended to contravariant tensor fields if the \
mapping is an immersion (QUESTION: is the latter the right statement?).\
\>", "Text",ExpressionUUID->"bee120b3-5dc5-4aca-9db7-59dcb14a3579"],

Cell["\<\
TODO:
\t- Document PullBackTensor.
\t- Document PullBackCovD and complete its encoding.
\t- Encode how standard operations (like Lie derivatives and brackets) behave \
under pull back and precomposition.\
\>", "Text",ExpressionUUID->"9a32e720-3e23-49a2-9a4f-566fc30afdb3"],

Cell[TextData[{
 StyleBox[ButtonBox["PullBack",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PullBack.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Pull back a tensor field along a mapping\n", "Text"],
 StyleBox[ButtonBox["ExpandPullBack",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ExpandPullBack.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand a ", "Text"],
 StyleBox["PullBack", "Text",
  FontFamily->"Courier"],
 StyleBox[" expression into precompositions and contractions", "Text"]
}], "DefinitionBox",ExpressionUUID->"e5c65272-7bfd-4be8-8488-9aee2d79f3b2"],

Cell["Pull back.", "Text",ExpressionUUID->"0499d1ba-ec50-4b77-8692-67fd7e34603b"],

Cell[CellGroupData[{

Cell["Declare a covector field on -\[DoubleStruckCapitalT]P:", "MathCaption",ExpressionUUID->"079dc867-c043-4216-99a1-c51f6181561e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"\[CapitalOmega]", "[", 
    RowBox[{"-", "A"}], "]"}], ",", "P"}], "]"}]], "Input",
 CellLabel->
  "In[1013]:=",ExpressionUUID->"a6266728-9e3d-4f92-ab21-919f323fc564"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[CapitalOmega]", "[", 
    RowBox[{"-", "A"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[CapitalOmega][-$CellContext`A], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978517045735*^9, 3.792805839164556*^9},
 CellLabel->"During evaluation of In[1013]:=",
 CellID->75949296,ExpressionUUID->"631230f8-a18e-4744-9354-ed3b14077d1b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", "\[CapitalOmega]", "]"}]], "Input",
 CellLabel->
  "In[1014]:=",ExpressionUUID->"8fc433c6-c1ef-4ed2-94be-67e3b9beded4"],

Cell[BoxData[
 RowBox[{"{", "P", "}"}]], "Output",
 CellChangeTimes->{3.728978517193242*^9, 3.7928058401815453`*^9},
 CellLabel->"Out[1014]=",
 CellID->1317092768,ExpressionUUID->"1d0ba8e9-e47b-4d4b-8e57-d1ebad85190e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", "\[CapitalOmega]", "]"}]], "Input",
 CellLabel->
  "In[1015]:=",ExpressionUUID->"7219b24b-0c5c-45fd-98ff-fe382c0f1107"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"-", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978517313595*^9, 3.792805840766396*^9},
 CellLabel->"Out[1015]=",
 CellID->908186115,ExpressionUUID->"ae34b12b-8b03-458e-a72f-2baaf32c7677"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can pull it back to -\[DoubleStruckCapitalT]M. We need to specify how the \
free indices must be renamed:\
\>", "MathCaption",ExpressionUUID->"885c13e0-d84c-4c10-8638-9016e9f6213f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"pbexpr", "=", 
  RowBox[{"PullBack", "[", 
   RowBox[{
    RowBox[{"\[CapitalOmega]", "[", 
     RowBox[{"-", "A"}], "]"}], ",", "\[Phi]", ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "A"}], "\[Rule]", 
      RowBox[{"-", "a"}]}], "}"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[1016]:=",ExpressionUUID->"21a8de32-c5f6-47fd-9747-ed32d9af60b8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Phi]", "*"], "\[CapitalOmega]"}], ")"}], 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PullBackTensor[$CellContext`\[CapitalOmega], $CellContext`\
\[Phi]][-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978517434354*^9, 3.792805845080714*^9},
 CellLabel->"Out[1016]=",
 CellID->1611076884,ExpressionUUID->"b56972a3-57cc-4551-af2e-cfb517d4a3c4"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1017]:=",ExpressionUUID->"934e9c58-17f1-4975-af6f-228aee6427d8"],

Cell["PullBackTensor[\[CapitalOmega], \[Phi]][-a]", "Output",
 CellChangeTimes->{3.728978517580454*^9, 3.7928058501730337`*^9},
 CellLabel->"Out[1017]//InputForm=",
 CellID->1604784060,ExpressionUUID->"8714cc32-6a01-46c1-a714-8a26daf15ac9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOfTensor", "[", 
  RowBox[{"PullBackTensor", "[", 
   RowBox[{"\[CapitalOmega]", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[1018]:=",ExpressionUUID->"2bcf295d-875b-457a-abe7-27e84f3ee4c5"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978517698262*^9, 3.7928058511170588`*^9},
 CellLabel->"Out[1018]=",
 CellID->621234410,ExpressionUUID->"6a23e6d6-fbc7-4db9-a765-c1429e4ddf4e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", 
  RowBox[{"PullBackTensor", "[", 
   RowBox[{"\[CapitalOmega]", ",", "\[Phi]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[1019]:=",ExpressionUUID->"e48abc57-98b5-449b-a5ee-86fcf0cbb998"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"-", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.72897851785423*^9, 3.7928058517576714`*^9},
 CellLabel->"Out[1019]=",
 CellID->861310320,ExpressionUUID->"78262c5e-f487-48c4-b568-acf181f1c122"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can expand the pulled-back expression in terms of precomposition and \
contraction with the tangent maps:\
\>", "MathCaption",ExpressionUUID->"c95e733e-ce02-43c2-8842-138e1d82980b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"pbexpr", "//", "ExpandPullBack"}]], "Input",
 CellLabel->
  "In[1020]:=",ExpressionUUID->"36f0fb6d-01ce-4524-b79b-413a4281e8f2"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["A", "*"]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`a, $CellContext`\[Phi]A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{"\[CapitalOmega]", "\[SmallCircle]", "\[Phi]"}], ")"}], 
       StyleBox[GridBox[{
          {" "},
          {
           SuperscriptBox["A", "*"]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`\[CapitalOmega], \
$CellContext`\[Phi]][-$CellContext`\[Phi]A],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785182872257`*^9, 3.792805854926005*^9},
 CellLabel->"Out[1020]=",
 CellID->549252194,ExpressionUUID->"895568ff-ced3-45fe-b97a-0fa9eecba6fc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1021]:=",ExpressionUUID->"32a7b706-272a-4df2-9f46-4d764d2768fd"],

Cell["d\[Phi][-a, \[Phi]D$3930]*Precompose[\[CapitalOmega], \
\[Phi]][-\[Phi]D$3930]", "Output",
 CellChangeTimes->{3.728978518702568*^9, 3.792805855757025*^9},
 CellLabel->"Out[1021]//InputForm=",
 CellID->85798042,ExpressionUUID->"e99102f0-777c-4671-9902-d6699b896742"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A higher-rank covariant field:", "MathCaption",ExpressionUUID->"5647658e-3e60-4ffd-a344-88ebec0459b1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"\[CapitalUpsilon]", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], ",", "P"}], "]"}]], "Input",
 CellLabel->
  "In[1022]:=",ExpressionUUID->"56c68915-1e97-4ebf-bdb2-aba7216d0d15"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[CapitalUpsilon]", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[CapitalUpsilon][-$CellContext`A, -$CellContext`B, -C], ". ",
    ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978518883521*^9, 3.792805859032031*^9},
 CellLabel->"During evaluation of In[1022]:=",
 CellID->1199195019,ExpressionUUID->"90a46002-116f-4035-b9ae-f226eee2cf32"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PullBack", "[", 
  RowBox[{
   RowBox[{"\[CapitalUpsilon]", "[", 
    RowBox[{
     RowBox[{"-", "A"}], ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], ",", "\[Phi]", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "A"}], "\[Rule]", 
      RowBox[{"-", "a"}]}], ",", 
     RowBox[{
      RowBox[{"-", "B"}], "\[Rule]", 
      RowBox[{"-", "b"}]}], ",", 
     RowBox[{
      RowBox[{"-", "C"}], "\[Rule]", 
      RowBox[{"-", "c"}]}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1023]:=",ExpressionUUID->"1182a3b1-73c4-4119-afe9-9bcf881bb5b9"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Phi]", "*"], "\[CapitalUpsilon]"}], ")"}], 
      StyleBox[GridBox[{
         {" ", " ", " "},
         {"a", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PullBackTensor[$CellContext`\[CapitalUpsilon], $CellContext`\
\[Phi]][-$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289785191870947`*^9, 3.7928058597895517`*^9},
 CellLabel->"Out[1023]=",
 CellID->839803071,ExpressionUUID->"ab557008-1364-4e8f-8790-4f68f2e2300b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ExpandPullBack"}]], "Input",
 CellLabel->
  "In[1024]:=",ExpressionUUID->"9ff46d8c-54fc-4940-bfaa-7795dd00bfa1"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["A", "*"]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`a, $CellContext`\[Phi]A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["B", "*"]},
          {"b", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`b, $CellContext`\[Phi]B],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["C", "*"]},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`c, $CellContext`\[Phi]C],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{"\[CapitalUpsilon]", "\[SmallCircle]", "\[Phi]"}], ")"}], 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {
           SuperscriptBox["A", "*"], 
           SuperscriptBox["B", "*"], 
           SuperscriptBox["C", "*"]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`\[CapitalUpsilon], \
$CellContext`\[Phi]][-$CellContext`\[Phi]A, -$CellContext`\[Phi]B, \
-$CellContext`\[Phi]C],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978519502921*^9, 3.7928058605436077`*^9},
 CellLabel->"Out[1024]=",
 CellID->229621416,ExpressionUUID->"2a005abb-0413-4e5e-8cba-46b2ddd9e8f2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1025]:=",ExpressionUUID->"4694b36f-a1f7-46d9-9b65-6b7dd7056ddb"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978519646226*^9, 3.792805861309249*^9},
 CellLabel->"Out[1025]=",
 CellID->972789631,ExpressionUUID->"fcaea77c-42a3-45b2-a5bc-e1cab274cf4a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now with contravariant slots and a contraction. We only need to specify the \
change of free indices. Note the need of the differential inverse, but there \
is no need to use the inverse mapping:\
\>", "MathCaption",ExpressionUUID->"5bd19bdc-74ff-4ccf-ab15-6b974793c6b9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"\[CapitalLambda]", "[", 
    RowBox[{"A", ",", "B", ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], ",", "P"}], "]"}]], "Input",
 CellLabel->
  "In[1026]:=",ExpressionUUID->"10bcd954-2732-4867-8dda-3b9a24d2e197"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[CapitalLambda]", "[", 
    RowBox[{"A", ",", "B", ",", 
     RowBox[{"-", "C"}], ",", 
     RowBox[{"-", "D"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[CapitalLambda][$CellContext`A, $CellContext`B, -C, -D], 
   ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785197822123`*^9, 3.79280586450906*^9},
 CellLabel->"During evaluation of In[1026]:=",
 CellID->488810969,ExpressionUUID->"d832d0a8-dfc9-4207-a597-880926cc2c7b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PullBack", "[", 
  RowBox[{
   RowBox[{"\[CapitalLambda]", "[", 
    RowBox[{"A", ",", "B", ",", 
     RowBox[{"-", "B"}], ",", 
     RowBox[{"-", "C"}]}], "]"}], ",", "\[Phi]", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"A", "\[Rule]", "a"}], ",", 
     RowBox[{
      RowBox[{"-", "C"}], "\[Rule]", 
      RowBox[{"-", "c"}]}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1027]:=",ExpressionUUID->"b595eb23-9765-4210-8784-b13d3f684c89"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["\[Phi]", "*"], "\[CapitalLambda]"}], ")"}], 
      StyleBox[GridBox[{
         {"a", "b", " ", " "},
         {" ", " ", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PullBackTensor[$CellContext`\[CapitalLambda], $CellContext`\
\[Phi]][$CellContext`a, $CellContext`b, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289785199189873`*^9, 3.792805865157652*^9},
 CellLabel->"Out[1027]=",
 CellID->1225988727,ExpressionUUID->"29e2653d-9db0-4ba5-a529-f91f91efde3e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ExpandPullBack"}]], "Input",
 CellLabel->
  "In[1028]:=",ExpressionUUID->"6f842fa8-6bbd-4b1c-9f5a-9ad521d0e3f7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["C", "*"]},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`c, $CellContext`\[Phi]C],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "i\[Phi]"}], "\[SmallCircle]", 
         "\[Phi]"}], ")"}], 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           SuperscriptBox["A", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`di\[Phi], \
$CellContext`\[Phi]][-$CellContext`\[Phi]A, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{"\[CapitalLambda]", "\[SmallCircle]", "\[Phi]"}], ")"}], 
       StyleBox[GridBox[{
          {
           SuperscriptBox["A", "*"], 
           SuperscriptBox["B", "*"], " ", " "},
          {" ", " ", 
           SuperscriptBox["B", "*"], 
           SuperscriptBox["C", "*"]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`\[CapitalLambda], \
$CellContext`\[Phi]][$CellContext`\[Phi]A, $CellContext`\[Phi]B, \
-$CellContext`\[Phi]B, -$CellContext`\[Phi]C],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978520065147*^9, 3.7928058658046618`*^9},
 CellLabel->"Out[1028]=",
 CellID->391383441,ExpressionUUID->"f10efa48-1636-4bac-9385-46ea2b637120"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A further contraction gives a precomposed scalar:", "MathCaption",ExpressionUUID->"9ed35e7d-7a7e-48c7-a24e-6ece7ddbefdd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", " ", 
  RowBox[{"delta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "c"}], "]"}]}]], "Input",
 CellLabel->
  "In[1029]:=",ExpressionUUID->"40778bba-a9f0-46d9-97ab-42c53497a8c0"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{"\[CapitalLambda]", "\[SmallCircle]", "\[Phi]"}], ")"}], 
      StyleBox[GridBox[{
         {
          SuperscriptBox["A", "*"], 
          SuperscriptBox["B", "*"], " ", " "},
         {" ", " ", 
          SuperscriptBox["B", "*"], 
          SuperscriptBox["A", "*"]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Precompose[$CellContext`\[CapitalLambda], \
$CellContext`\[Phi]][$CellContext`\[Phi]A, $CellContext`\[Phi]B, \
-$CellContext`\[Phi]B, -$CellContext`\[Phi]A],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978520197433*^9, 3.79280586752063*^9},
 CellLabel->"Out[1029]=",
 CellID->1664870535,ExpressionUUID->"499b4a89-84a4-4532-a2ea-068253dd2238"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.8. Linear push with the mapping differential", "Subsection",ExpressionUUID->"bda7e719-a98a-45b8-9a0b-6f778f5b88b8"],

Cell["\<\
This operation is dual with respect to precomposition. Here we do not change \
the dependencies of a tensor field, but we do act linearly on its indices. \
Effectively, this is nothing but a condensed notation to denote contractions \
with the mapping differentials. This operation is trivial on scalar fields.
The name is far from ideal, but I don\[CloseCurlyQuote]t see what else to \
choose.\
\>", "Text",ExpressionUUID->"dd56b798-3eaa-4e25-9de8-e0b7a4c21c0d"],

Cell["\<\
TODO:
\t- Analyze what happens to the covariant derivatives of linear push fields.
\t- Encode how the standard operations (like Lie derivatives and brackets) \
behave under linear push and pushforward.\
\>", "Text",ExpressionUUID->"9d626024-f7d1-4e0a-8928-d31ab6c9e7e9"],

Cell[TextData[{
 StyleBox[ButtonBox["LinearPush",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LinearPush.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Linear push of a tensor field by a mapping differential", "Text"]
}], "DefinitionBox",ExpressionUUID->"b30f4dd4-e479-407d-bdb1-ce2f6a5e0584"],

Cell["Linear push.", "Text",ExpressionUUID->"d9f9d2e4-a3bc-4761-a551-aec3903d1a43"],

Cell[CellGroupData[{

Cell[TextData[{
 "Take a tensor field in ",
 StyleBox["\[DoubleStruckCapitalT]M",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"4e1ea5ae-1733-45d3-82ad-a2429be305e6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M"}], "]"}]], "Input",
 CellLabel->
  "In[1030]:=",ExpressionUUID->"d79ca26c-341a-4c66-9da7-452e86abe853"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"v", "[", "a", "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`v[$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978525948946*^9, 3.792805877318437*^9},
 CellLabel->"During evaluation of In[1030]:=",
 CellID->25109116,ExpressionUUID->"f4dea6df-6f73-472b-abe9-38b265bb31bf"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Push it to the pullback bundle ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["\[Phi]", "*"], TraditionalForm]],ExpressionUUID->
  "e9275b72-435b-4e76-ab6e-713764b3c10a"],
 StyleBox["\[DoubleStruckCapitalT]P",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"b30470ed-2aa6-40fa-a092-f8079169da38"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LinearPush", "[", 
   RowBox[{"v", ",", "d\[Phi]"}], "]"}], "[", "\[Phi]A", "]"}]], "Input",
 CellLabel->
  "In[1031]:=",ExpressionUUID->"ed2917de-926f-422b-8fa2-0089a2fd1a9f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[DifferentialD]", "\[Phi]"}], "\[CenterDot]", "v"}], ")"}], 
      
      StyleBox[GridBox[{
         {
          SuperscriptBox["A", "*"]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`LinearPush[$CellContext`v, $CellContext`d\[Phi]][$CellContext`\
\[Phi]A],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289785260844393`*^9, 3.792805878693383*^9},
 CellLabel->"Out[1031]=",
 CellID->542313864,ExpressionUUID->"1286c9b1-4a47-4a4d-9928-51d45608f791"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["It is a vector field on the pullback bundle:", "MathCaption",ExpressionUUID->"824f90eb-5053-4cba-a2bb-57875ea504c3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", 
  RowBox[{"Head", "[", "%", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1032]:=",ExpressionUUID->"1c897e38-2c96-46a5-b72c-34187b632bc4"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   SuperscriptBox["\[Phi]", "*"], 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978526198522*^9, 3.792805879908882*^9},
 CellLabel->"Out[1032]=",
 CellID->1397078346,ExpressionUUID->"efa39507-59db-49d4-8502-ecf997ed9134"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "It is still a field on ",
 StyleBox["M",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"e6daddfd-350e-4819-8737-9f0a9248c6bb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOf", "[", "%%", "]"}]], "Input",
 CellLabel->
  "In[1033]:=",ExpressionUUID->"3f5edd83-a7ef-4dc9-9614-10ab45a33d33"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978526310813*^9, 3.7928058810710163`*^9},
 CellLabel->"Out[1033]=",
 CellID->236134905,ExpressionUUID->"ff57b13f-5e64-4bef-a9a0-c4642fe6ce3f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Expand a ",
 StyleBox["LinearPush",
  FontFamily->"Courier"],
 " expression:"
}], "MathCaption",ExpressionUUID->"52986d2f-4a3c-489d-b91f-b00279b96000"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"LinearPush", "[", 
    RowBox[{"v", ",", "d\[Phi]"}], "]"}], "[", "\[Phi]A", "]"}], "//", 
  "ExpandPushForward"}]], "Input",
 CellLabel->
  "In[1034]:=",ExpressionUUID->"6ba3acd8-7554-4a4d-ba89-645b1fb4c659"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["A", "*"]},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`d\[Phi][-$CellContext`a, $CellContext`\[Phi]A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785266695337`*^9, 3.792805882173531*^9},
 CellLabel->"Out[1034]=",
 CellID->556348291,ExpressionUUID->"636aadad-80b9-4a45-b617-757d99fd3942"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1035]:=",ExpressionUUID->"80a91cbc-46dd-4659-b9bf-d63e1fc1d27e"],

Cell["d\[Phi][-d$3975, \[Phi]A]*v[d$3975]", "Output",
 CellChangeTimes->{3.728978526970348*^9, 3.792805884180791*^9},
 CellLabel->"Out[1035]//InputForm=",
 CellID->407674122,ExpressionUUID->"91b915e1-ec5e-4ace-aa94-a39e28d46290"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Now take a covector field in ",
 StyleBox["-\[DoubleStruckCapitalT]M",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"ed019e62-1097-46a1-8145-ee4b56e5e4c1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"\[Omega]", "[", 
    RowBox[{"-", "a"}], "]"}], ",", "M"}], "]"}]], "Input",
 CellLabel->
  "In[1036]:=",ExpressionUUID->"e50811c0-861f-4218-9844-1b7ab1a5e54f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"\[Omega]", "[", 
    RowBox[{"-", "a"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`\[Omega][-$CellContext`a], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785272669077`*^9, 3.792805885572711*^9},
 CellLabel->"During evaluation of In[1036]:=",
 CellID->1776391051,ExpressionUUID->"3ec5a9a9-a20d-45a5-8123-0117c6c9af86"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is possible to push covector fields in this case because the mapping \
differential is left-invertible:\
\>", "MathCaption",ExpressionUUID->"fbc82e85-5edf-4a1b-8b99-ab412feadf41"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LinearPush", "[", 
   RowBox[{"\[Omega]", ",", "d\[Phi]"}], "]"}], "[", 
  RowBox[{"-", "\[Phi]A"}], "]"}]], "Input",
 CellLabel->
  "In[1037]:=",ExpressionUUID->"69367213-5861-4bac-91e5-7b2b2809fc7d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        RowBox[{"\[DifferentialD]", "\[Phi]"}], "\[CenterDot]", "\[Omega]"}], 
       ")"}], 
      StyleBox[GridBox[{
         {" "},
         {
          SuperscriptBox["A", "*"]}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`LinearPush[$CellContext`\[Omega], \
$CellContext`d\[Phi]][-$CellContext`\[Phi]A],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978527401929*^9, 3.792805887108563*^9},
 CellLabel->"Out[1037]=",
 CellID->654170841,ExpressionUUID->"c3df70a0-c899-484e-b64c-c63742e69637"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", 
  RowBox[{"Head", "[", "%", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1038]:=",ExpressionUUID->"204b4243-d566-4840-80f7-2566975d2b48"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"-", 
   RowBox[{"(", 
    RowBox[{
     SuperscriptBox["\[Phi]", "*"], 
     InterpretationBox[
      StyleBox["\[DoubleStruckCapitalT]P",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`TangentP,
      Editable->False]}], ")"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978527531467*^9, 3.7928058877158813`*^9},
 CellLabel->"Out[1038]=",
 CellID->1621705353,ExpressionUUID->"46df10fa-26b8-470e-a72b-52f86b5b7283"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
That is clearer if we expand the expression into precomposition and \
contraction:\
\>", "MathCaption",ExpressionUUID->"db439dca-d81f-483c-8255-1cfb7cdc6046"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"LinearPush", "[", 
    RowBox[{"\[Omega]", ",", "d\[Phi]"}], "]"}], "[", 
   RowBox[{"-", "\[Phi]A"}], "]"}], "//", "ExpandPushForward"}]], "Input",
 CellLabel->
  "In[1039]:=",ExpressionUUID->"7297081d-8201-4ecb-a0c3-c5379b0af18c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[Omega]", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`\[Omega][-$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "i\[Phi]"}], "\[SmallCircle]", 
         "\[Phi]"}], ")"}], 
       StyleBox[GridBox[{
          {" ", "a"},
          {
           SuperscriptBox["A", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`di\[Phi], \
$CellContext`\[Phi]][-$CellContext`\[Phi]A, $CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978527666154*^9, 3.7928058891572742`*^9},
 CellLabel->"Out[1039]=",
 CellID->1329065620,ExpressionUUID->"aaf1ffbc-1e73-4e2c-93a1-4b3d6f5282b2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1040]:=",ExpressionUUID->"6ee39914-f67a-4068-a1a1-0db9c55d9e4a"],

Cell["\[Omega][-d$3986]*Precompose[di\[Phi], \[Phi]][-\[Phi]A, d$3986]", \
"Output",
 CellChangeTimes->{3.7289785277791348`*^9, 3.792805890237452*^9},
 CellLabel->"Out[1040]//InputForm=",
 CellID->1518038754,ExpressionUUID->"66792c29-cfa1-4653-ac25-d39bb8303d94"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "8.9. PushForward",
 StyleBox[". MINOR BUG",
  FontColor->RGBColor[0.5, 0, 0.5]]
}], "Subsection",ExpressionUUID->"711acf14-f170-4019-9d21-08f27d71786c"],

Cell["\<\
And this is our second fundamental operation of working with tensors and \
mappings: the pushforward of a tensor field on the domain manifold to a \
tensor field on the image manifold. This operation always requires an \
invertible mapping, because we need to use precomposition along the inverse \
mapping to change the dependencies. QUESTION: Fully invertible or is it \
enough with being a submersion?\
\>", "Text",
 CellChangeTimes->{
  3.792805910508025*^9},ExpressionUUID->"17738ee1-db9d-4b76-a748-\
0304da15a10f"],

Cell["\<\
TODO:
\t- Document PushForwardTensor.\
\>", "Text",ExpressionUUID->"e806c476-060b-415f-af0a-ba300f6f747f"],

Cell[TextData[{
 StyleBox[ButtonBox["PushForward",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PushForward.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Push forward a tensor field along a mapping\n", "Text"],
 StyleBox[ButtonBox["ExpandPullBack",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ExpandPushForward.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Expand a ", "Text"],
 StyleBox["PushForward", "Text",
  FontFamily->"Courier"],
 StyleBox[" expression into precompositions and contractions", "Text"]
}], "DefinitionBox",ExpressionUUID->"eea02c1d-07e3-43c7-b34f-38f37ec1fe13"],

Cell["Push forward.", "Text",ExpressionUUID->"fbf379b0-9033-4d14-ad03-2a9983eda162"],

Cell[CellGroupData[{

Cell[TextData[{
 "Push forward the vector field ",
 StyleBox["v",
  FontFamily->"Courier"],
 " on ",
 StyleBox["\[DoubleStruckCapitalT]M",
  FontFamily->"Courier"],
 " to a vector field on ",
 StyleBox["\[DoubleStruckCapitalT]P",
  FontFamily->"Courier"],
 ". Again, we need to specify the renaming of the free indices:"
}], "MathCaption",ExpressionUUID->"0b94e55b-670c-42db-abc7-7ffe90584c48"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PushForward", "[", 
  RowBox[{
   RowBox[{"3", 
    RowBox[{"v", "[", "a", "]"}]}], ",", "\[Phi]", ",", 
   RowBox[{"{", 
    RowBox[{"a", "\[Rule]", "A"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1041]:=",ExpressionUUID->"086f204b-8155-4f11-8c85-5b70e67735e1"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         SubscriptBox["\[Phi]", "*"], "v"}], ")"}], 
       StyleBox[GridBox[{
          {"A"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PushForwardTensor[$CellContext`v, \
$CellContext`\[Phi]][$CellContext`A],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978534216819*^9, 3.7928059290354013`*^9},
 CellLabel->"Out[1041]=",
 CellID->308568836,ExpressionUUID->"05d5bd4c-5012-4f75-a903-11e38fdda88a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1042]:=",ExpressionUUID->"a0a02fe6-5464-4e8c-a74a-9673326568ca"],

Cell["3*PushForwardTensor[v, \[Phi]][A]", "Output",
 CellChangeTimes->{3.728978534334032*^9, 3.7928059304738083`*^9},
 CellLabel->"Out[1042]//InputForm=",
 CellID->659725678,ExpressionUUID->"0266dbc8-15e8-4dc2-b5bd-1c568bbadf90"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "It is now a vector field on ",
 StyleBox["P",
  FontFamily->"Courier"],
 ": (TODO: THIS IS WRONG!!!)"
}], "MathCaption",ExpressionUUID->"c505edab-16c2-4523-8ca9-916f1e7b153d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DependenciesOf", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1043]:=",ExpressionUUID->"8c994e47-d86a-418f-a6a7-e12772c6af02"],

Cell[BoxData[
 RowBox[{"{", "M", "}"}]], "Output",
 CellChangeTimes->{3.728978534449218*^9, 3.792805936722458*^9},
 CellLabel->"Out[1043]=",
 CellID->1940298987,ExpressionUUID->"1448fe62-2317-475c-9ae5-99798642f7ec"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We see that the actual operation requires the use of the inverse mapping:\
\>", "MathCaption",ExpressionUUID->"e5d03bf1-e7d9-4afd-ad41-b453154227d1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%%", "//", "ExpandPushForward"}]], "Input",
 CellLabel->
  "In[1044]:=",ExpressionUUID->"33d134b7-8b39-401a-96f4-a46dc866336f"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "\[Phi]"}], "\[SmallCircle]", 
         SuperscriptBox["\[Phi]", 
          RowBox[{"-", "1"}]]}], ")"}], 
       StyleBox[GridBox[{
          {" ", "A"},
          {
           SuperscriptBox["a", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`d\[Phi], \
$CellContext`i\[Phi]][-$CellContext`i\[Phi]a, $CellContext`A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{"v", "\[SmallCircle]", 
         SuperscriptBox["\[Phi]", 
          RowBox[{"-", "1"}]]}], ")"}], 
       StyleBox[GridBox[{
          {
           SuperscriptBox["a", "*"]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`v, \
$CellContext`i\[Phi]][$CellContext`i\[Phi]a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785345694323`*^9, 3.792805969292282*^9},
 CellLabel->"Out[1044]=",
 CellID->631203911,ExpressionUUID->"5b227170-8aed-4f45-b558-0490b06a3943"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
In fact, push forward is equivalent to a pull back along the inverse mapping:\
\
\>", "MathCaption",ExpressionUUID->"b9c24005-d3ab-4c5e-834f-74b9fc3ad8ff"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PullBack", "[", 
  RowBox[{
   RowBox[{"3", 
    RowBox[{"v", "[", "a", "]"}]}], ",", "i\[Phi]", ",", 
   RowBox[{"{", 
    RowBox[{"a", "\[Rule]", "A"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1045]:=",ExpressionUUID->"bb0540f0-083b-437d-8486-b499e5bef89d"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         SuperscriptBox[
          SuperscriptBox["\[Phi]", 
           RowBox[{"-", "1"}]], "*"], "v"}], ")"}], 
       StyleBox[GridBox[{
          {"A"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`PullBackTensor[$CellContext`v, \
$CellContext`i\[Phi]][$CellContext`A],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785347141933`*^9, 3.79280597056382*^9},
 CellLabel->"Out[1045]=",
 CellID->198240297,ExpressionUUID->"047d96ac-1c16-4363-81df-20defa909573"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ExpandPullBack"}]], "Input",
 CellLabel->
  "In[1046]:=",ExpressionUUID->"3e32c0ee-c9a0-4ad0-9ab8-04e8681d93cf"],

Cell[BoxData[
 RowBox[{"3", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[DifferentialD]", "\[Phi]"}], "\[SmallCircle]", 
         SuperscriptBox["\[Phi]", 
          RowBox[{"-", "1"}]]}], ")"}], 
       StyleBox[GridBox[{
          {" ", "A"},
          {
           SuperscriptBox["a", "*"], " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`d\[Phi], \
$CellContext`i\[Phi]][-$CellContext`i\[Phi]a, $CellContext`A],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{"v", "\[SmallCircle]", 
         SuperscriptBox["\[Phi]", 
          RowBox[{"-", "1"}]]}], ")"}], 
       StyleBox[GridBox[{
          {
           SuperscriptBox["a", "*"]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`v, \
$CellContext`i\[Phi]][$CellContext`i\[Phi]a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785348459787`*^9, 3.792805971122115*^9},
 CellLabel->"Out[1046]=",
 CellID->808437265,ExpressionUUID->"aedabdd6-d817-4ff4-a825-9a5a4cbda3e2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Push forward a covector field. Now we need both the inverse mapping and its \
differential:\
\>", "MathCaption",ExpressionUUID->"eefc2bee-edec-400b-a784-657b769e61af"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PushForwardTensor", "[", 
   RowBox[{"\[Omega]", ",", "\[Phi]"}], "]"}], "[", 
  RowBox[{"-", "A"}], "]"}]], "Input",
 CellLabel->
  "In[1047]:=",ExpressionUUID->"a7c6e292-bb74-4ed9-8f96-42cc0d9252ab"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["\[Phi]", "*"], "\[Omega]"}], ")"}], 
      StyleBox[GridBox[{
         {" "},
         {"A"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PushForwardTensor[$CellContext`\[Omega], \
$CellContext`\[Phi]][-$CellContext`A],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978535103607*^9, 3.79280597293029*^9},
 CellLabel->"Out[1047]=",
 CellID->1548033283,ExpressionUUID->"36ab1a8f-8e6d-4e70-a1c2-ac6e95b3aeec"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ExpandPushForward"}]], "Input",
 CellLabel->
  "In[1048]:=",ExpressionUUID->"aa365a1e-e22e-4fae-a7b3-d033ee86f96e"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"\[DifferentialD]", "i\[Phi]"}], 
       StyleBox[GridBox[{
          {" ", 
           SuperscriptBox["a", "*"]},
          {"A", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`di\[Phi][-$CellContext`A, $CellContext`i\[Phi]a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"(", 
        RowBox[{"\[Omega]", "\[SmallCircle]", 
         SuperscriptBox["\[Phi]", 
          RowBox[{"-", "1"}]]}], ")"}], 
       StyleBox[GridBox[{
          {" "},
          {
           SuperscriptBox["a", "*"]}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "Rows" -> {{
              Offset[0.]}}}],
        FontSize->9]}
     },
     GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Precompose[$CellContext`\[Omega], \
$CellContext`i\[Phi]][-$CellContext`i\[Phi]a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785354877377`*^9, 3.792805973505096*^9},
 CellLabel->"Out[1048]=",
 CellID->1301376685,ExpressionUUID->"ac8cad06-efb9-4f68-8865-78bef531a32d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Push forward a mixed tensor, and with contractions:", "MathCaption",ExpressionUUID->"d4f93538-9bc2-4146-80dd-ebcdf8057046"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", "M"}], "]"}]], "Input",
 CellLabel->
  "In[1049]:=",ExpressionUUID->"81b74fa6-97d3-4464-84e5-d9b491461ac2"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}]}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`S[$CellContext`a, $CellContext`b, -$CellContext`c], ". ", ""],
  
  Editable->False]], "Print",
 CellChangeTimes->{3.728978535631645*^9, 3.792805974904907*^9},
 CellLabel->"During evaluation of In[1049]:=",
 CellID->1632227167,ExpressionUUID->"fd886551-613c-4f2a-875b-5bc839339756"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PushForward", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "\[Phi]", ",", 
   RowBox[{"{", 
    RowBox[{"a", "\[Rule]", "A"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1050]:=",ExpressionUUID->"cd908dd3-9edf-4cad-bcd4-ae962cee8b0e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"(", 
       RowBox[{
        SubscriptBox["\[Phi]", "*"], "S"}], ")"}], 
      StyleBox[GridBox[{
         {"A", "B", " "},
         {" ", " ", "B"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "Rows" -> {{
             Offset[0.]}}}],
       FontSize->9]}
    },
    GridBoxAlignment->{"Columns" -> {{Center}}, "Rows" -> {{Center}}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`PushForwardTensor[$CellContext`S, \
$CellContext`\[Phi]][$CellContext`A, $CellContext`B, -$CellContext`B],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978535775199*^9, 3.792805975422368*^9},
 CellLabel->"Out[1050]=",
 CellID->1270626084,ExpressionUUID->"90e75fda-5add-4af1-b781-b97fe5f93293"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SlotsOfTensor", "[", 
  RowBox[{"Head", "[", "%", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1051]:=",ExpressionUUID->"9a2b0126-e2c0-4868-a358-15549a3d0a12"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False], ",", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False], ",", 
   RowBox[{"-", 
    InterpretationBox[
     StyleBox["\[DoubleStruckCapitalT]P",
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`TangentP,
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.72897853596773*^9, 3.792805976784355*^9},
 CellLabel->"Out[1051]=",
 CellID->1195123764,ExpressionUUID->"2971e489-240e-43a7-8a1f-2e719c0b7e40"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "8.10. Clean up",
 StyleBox[". TRIVIAL BUG",
  FontColor->RGBColor[0.5, 0, 0.5]]
}], "Subsection",ExpressionUUID->"e857e651-01ac-4e24-b76e-147328f407c5"],

Cell[CellGroupData[{

Cell["Remove the various objects we have declared in this section:", \
"MathCaption",ExpressionUUID->"a15a2f25-001b-46c7-8214-eea650dca245"],

Cell[CellGroupData[{

Cell[BoxData["$Mappings"], "Input",
 CellLabel->
  "In[1052]:=",ExpressionUUID->"4da653f6-c7fd-4fda-9a56-b7529de68adb"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"\[Phi]", ",", "i\[Phi]", ",", "\[Psi]", ",", "i\[Psi]"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.7289785410497513`*^9, 3.792805988088622*^9},
 CellLabel->"Out[1052]=",
 CellID->1632433000,ExpressionUUID->"b7502431-c9fd-40a1-a337-3e27571c22d3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$Tensors"], "Input",
 CellLabel->
  "In[1053]:=",ExpressionUUID->"e6398199-3bf8-43f2-9649-d2d3b2b2f31b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "di\[Phi]", ",", "d\[Phi]", ",", "di\[Psi]", ",", "d\[Psi]", ",", "T", ",", 
   "\[CapitalOmega]", ",", "\[CapitalUpsilon]", ",", "\[CapitalLambda]", ",", 
   "v", ",", "\[Omega]", ",", "S"}], "}"}]], "Output",
 CellChangeTimes->{3.72897854122619*^9, 3.7928059890246887`*^9},
 CellLabel->"Out[1053]=",
 CellID->1439801107,ExpressionUUID->"12dbf9aa-8d35-434d-8221-12e4516db6f1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$Manifolds"], "Input",
 CellLabel->
  "In[1054]:=",ExpressionUUID->"d97a2271-4ee2-41d9-89ec-30eaff22daa1"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"M", ",", "P", ",", "L"}], "}"}]], "Output",
 CellChangeTimes->{3.7289785413561897`*^9, 3.792805989698687*^9},
 CellLabel->"Out[1054]=",
 CellID->428150384,ExpressionUUID->"473a1b1b-8a4f-4eb5-97da-a2085e42b99e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It must be done in this order: first the mappings (to remove their \
differentials too), then the tensors and finally the manifolds:\
\>", "MathCaption",ExpressionUUID->"4cfb0dff-5a8f-42c2-8d14-2cdcb9829a47"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefMapping", "/@", 
   RowBox[{"{", 
    RowBox[{"\[Phi]", ",", "\[Psi]"}], "}"}]}], ";"}]], "Input",
 CellLabel->
  "In[1055]:=",ExpressionUUID->"71764803-344b-49ce-a9f6-b630dd8724c7"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "d\[Phi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`d\[Phi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.792805991582847*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->1593606993,ExpressionUUID->"af4e4d65-8f26-4386-97a6-797444cd069d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "di\[Phi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`di\[Phi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.79280599158892*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->336685967,ExpressionUUID->"a5a1a93d-695d-4b5d-9041-098b367266af"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "i\[Phi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "inverse mapping", 
   " ", $CellContext`i\[Phi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.792805991629937*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->61637524,ExpressionUUID->"a3bb7982-d6b5-44be-b743-129d1ccb4de3"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "\[Phi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "mapping", 
   " ", $CellContext`\[Phi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.792805991676429*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->854748014,ExpressionUUID->"1ec96e0f-b280-4de8-9245-93f5e6c1d692"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "d\[Psi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`d\[Psi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.79280599168121*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->1231756515,ExpressionUUID->"7d92fc12-b0ea-4a13-b417-e29a11a48a33"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping differential tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "di\[Psi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", 
   "mapping differential tensor", " ", $CellContext`di\[Psi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.7928059916856194`*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->389390525,ExpressionUUID->"eed3a22e-a76f-414a-9def-c6913b08a693"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"left-inverse mapping\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "i\[Psi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "left-inverse mapping", 
   " ", $CellContext`i\[Psi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.792805991711796*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->1103881266,ExpressionUUID->"f9ae2128-f6b1-4a64-a3d8-d4885e560315"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefMapping", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"mapping\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "\[Psi]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefMapping, ": Undefined ", "mapping", 
   " ", $CellContext`\[Psi]],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785414652777`*^9, 3.792805991748722*^9},
 CellLabel->"During evaluation of In[1055]:=",
 CellID->821211229,ExpressionUUID->"c5ca0913-e848-485b-8f1f-2266ed1a7c02"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", "$Tensors"}], ";"}]], "Input",
 CellLabel->
  "In[1056]:=",ExpressionUUID->"cf7982e4-0ab3-4a69-bced-f629de53f469"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993153791*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->1731263231,ExpressionUUID->"4f80955b-71cb-4fe1-a9a0-77ecda6160ea"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "\[CapitalOmega]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`\[CapitalOmega]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993160158*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->1085190214,ExpressionUUID->"87bff5d7-b1f1-4394-ade4-9e247f031cbb"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "\[CapitalUpsilon]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`\[CapitalUpsilon]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993170224*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->213312727,ExpressionUUID->"a7261e45-1959-4b67-bf9e-f590c18790ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "\[CapitalLambda]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`\[CapitalLambda]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993174226*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->996795846,ExpressionUUID->"6edba8d1-9ba6-4cd8-be59-8c97bd6f1d46"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993178012*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->627853475,ExpressionUUID->"af170818-9f12-4652-a43d-0c8b49d1eb5f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "\[Omega]"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`\[Omega]],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993182117*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->1746196614,ExpressionUUID->"4fb79b47-0044-4c97-b28f-8323426d162a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978541912341*^9, 3.792805993186338*^9},
 CellLabel->"During evaluation of In[1056]:=",
 CellID->814971596,ExpressionUUID->"dff29a03-c4f8-4cb9-a223-7ccbdeac9460"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefManifold", "/@", "$Manifolds"}], ";"}]], "Input",
 CellLabel->
  "In[1057]:=",ExpressionUUID->"632e0e4b-2f2c-4209-9088-0a061375dcb3"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978542216691*^9, 3.792805995125647*^9},
 CellLabel->"During evaluation of In[1057]:=",
 CellID->1654220525,ExpressionUUID->"0c13e229-f040-4f88-9406-271535100694"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978542216691*^9, 3.792805995133678*^9},
 CellLabel->"During evaluation of In[1057]:=",
 CellID->1161069613,ExpressionUUID->"466f28e9-ff43-4585-b18b-8ef6c8a68146"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]P",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentP,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentP],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978542216691*^9, 3.792805995142399*^9},
 CellLabel->"During evaluation of In[1057]:=",
 CellID->2060569049,ExpressionUUID->"eb6c6ad2-3cc0-4101-a31c-481a2ba068be"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "P"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`P],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978542216691*^9, 3.792805995146888*^9},
 CellLabel->"During evaluation of In[1057]:=",
 CellID->1662715021,ExpressionUUID->"bc18bdc3-5b70-44be-919b-a61de852b453"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]L",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentL,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentL],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978542216691*^9, 3.7928059951748667`*^9},
 CellLabel->"During evaluation of In[1057]:=",
 CellID->46659536,ExpressionUUID->"3986f88e-87d1-4e01-aced-ace1152768af"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "L"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`L],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978542216691*^9, 3.792805995179111*^9},
 CellLabel->"During evaluation of In[1057]:=",
 CellID->375290397,ExpressionUUID->"e4be0448-119e-4178-8e6f-0612ce0ce957"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Remove other symbols used:", "MathCaption",ExpressionUUID->"beef4f91-7ae3-40aa-bd16-e4e181d5b733"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"idM", ",", "map", ",", "imap", ",", "pbexpr"}], "]"}]], "Input",
 CellLabel->
  "In[1058]:=",ExpressionUUID->"0f60d2ab-aa1d-4e54-9b3a-9ff613b95fdf"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
TODO: WE DID A TERRIBLE JOB AT REMOVING INDICES... WHAT ARE THE mapTangentM \
symbols?\
\>", "MathCaption",ExpressionUUID->"2e473975-d54a-4107-9fa3-1de9fa40db71"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Global`*"}]], "Input",
 CellLabel->
  "In[1059]:=",ExpressionUUID->"bd1eace3-24cd-4760-ae3d-c4ced1280bfc"],

Cell[BoxData[
 StyleBox[
  FrameBox[GridBox[{
     {
      DynamicModuleBox[{Typeset`open$$ = True}, 
       PaneSelectorBox[{False->
        ButtonBox[
         RowBox[{
          
          DynamicBox[FEPrivate`FrontEndResource[
           "FEBitmaps", "RightPointerOpener"]], " ", 
          StyleBox["Global`", "InformationGridGroupHeader"]}],
         Appearance->None,
         BaseStyle->"InformationGridLink",
         ButtonFunction:>FEPrivate`Set[Typeset`open$$, True],
         Evaluator->Automatic,
         Method->"Preemptive"], True->
        PaneBox[GridBox[{
           {
            ButtonBox[
             RowBox[{
              
              DynamicBox[FEPrivate`FrontEndResource[
               "FEBitmaps", "DownPointerOpener"],
               ImageSizeCache->{10., {3., 7.}}], " ", 
              StyleBox["Global`", "InformationGridGroupHeader"]}],
             Appearance->None,
             BaseStyle->"InformationGridLink",
             ButtonFunction:>FEPrivate`Set[Typeset`open$$, False],
             Evaluator->Automatic,
             Method->"Preemptive"]},
           {
            PaneBox[GridBox[{
               {
                ButtonBox[
                 StyleBox["fmt", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-63cf7050-8c9f-42e5-890d-57a9750d4c64", {
                   "fmt", "Global`"}, False},
                 ButtonNote->"Global`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["fmt$", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-63cf7050-8c9f-42e5-890d-57a9750d4c64", {
                   "fmt$", "Global`"}, False},
                 ButtonNote->"Global`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["\[Phi]\[Psi]\[Alpha]", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-63cf7050-8c9f-42e5-890d-57a9750d4c64", {
                   "\[Phi]\[Psi]\[Alpha]", "Global`"}, False},
                 ButtonNote->"Global`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["\[Phi]\[Psi]\[Beta]", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-63cf7050-8c9f-42e5-890d-57a9750d4c64", {
                   "\[Phi]\[Psi]\[Beta]", "Global`"}, False},
                 ButtonNote->"Global`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["\[Phi]\[Psi]\[Gamma]", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-63cf7050-8c9f-42e5-890d-57a9750d4c64", {
                   "\[Phi]\[Psi]\[Gamma]", "Global`"}, False},
                 ButtonNote->"Global`",
                 Evaluator->Automatic], 
                ButtonBox[
                 StyleBox["\[Phi]\[Psi]\[Delta]", "InformationGridButton"],
                 Appearance->None,
                 BaseStyle->"InformationGridLink",
                 
                 ButtonData:>{
                  "Info-63cf7050-8c9f-42e5-890d-57a9750d4c64", {
                   "\[Phi]\[Psi]\[Delta]", "Global`"}, False},
                 ButtonNote->"Global`",
                 Evaluator->Automatic]}
              },
              DefaultBaseStyle->"Text",
              
              GridBoxAlignment->{
               "Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
              GridBoxItemSize->{"Columns" -> {{
                   Scaled[0.15833333333333333`]}}}],
             ImageMargins->{{10, 0}, {0, 2}}]}
          },
          GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}}],
         FrameMargins->{{0, 0}, {8, 0}}]}, Dynamic[Typeset`open$$],
        ImageSize->Automatic]]}
    },
    GridBoxAlignment->{"Columns" -> {{Left}}, "Rows" -> {{Baseline}}},
    GridBoxDividers->{"ColumnsIndexed" -> {{False}}, "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.5599999999999999]}, 
        Offset[0.27999999999999997`]}, "Rows" -> {
        Offset[0.2], {
         Offset[0.8]}, 
        Offset[0.2]}}],
   BaseStyle->"InformationTitleFrame"], "InformationGridPlain"]], "Output",
 CellChangeTimes->{3.792806007114993*^9},
 CellLabel->"Out[1059]=",
 CellID->573846044,ExpressionUUID->"56acd225-8c8e-4cba-878b-2a053707dcc2"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["8.11. Graphical summary", "Subsection",ExpressionUUID->"bf812b76-03e7-402c-b038-c9ddefa14da8"],

Cell[CellGroupData[{

Cell[TextData[{
 "In the following diagram we have the connections between the manifolds ",
 StyleBox["M",
  FontFamily->"Courier"],
 " and ",
 StyleBox["P",
  FontFamily->"Courier"],
 " and their respective tangent bundles ",
 StyleBox["\[DoubleStruckCapitalT]M",
  FontFamily->"Courier"],
 " and ",
 StyleBox["\[DoubleStruckCapitalT]P",
  FontFamily->"Courier"],
 ". Fields in this diagram are lines connecting a manifold to a bundle. \
Vertical lines have \[OpenCurlyDoubleQuote]normal\[CloseCurlyDoubleQuote] \
indices, while diagonal lines have \[OpenCurlyDoubleQuote]starred\
\[CloseCurlyDoubleQuote] indices. We see how starting with ",
 StyleBox["v",
  FontFamily->"Courier"],
 " we can move forward its (upper) head using contraction ",
 StyleBox["d\[Phi]",
  FontFamily->"Courier"],
 ", while starting with ",
 StyleBox["W",
  FontFamily->"Courier"],
 " we can move backward its tail using precomposition with ",
 StyleBox["\[Phi]",
  FontFamily->"Courier"],
 ". Adding the lines for ",
 StyleBox["i\[Phi]",
  FontFamily->"Courier"],
 " is trivial. Moving the vertical line of ",
 StyleBox["v",
  FontFamily->"Courier"],
 " to ",
 StyleBox["W",
  FontFamily->"Courier"],
 " would be a full \[OpenCurlyDoubleQuote]push \
forward\[CloseCurlyDoubleQuote] and involves both moving the head and the \
tail. Moving ",
 StyleBox["W",
  FontFamily->"Courier"],
 " to ",
 StyleBox["v",
  FontFamily->"Courier"],
 " would be a full \[OpenCurlyDoubleQuote]pull back\[CloseCurlyDoubleQuote], \
also involving both operations."
}], "Text",ExpressionUUID->"c13b7c22-44ec-422c-9ab2-14f75be11eef"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"LayeredGraphPlot", "[", 
   RowBox[{
    RowBox[{"{", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"\"\<M\>\"", ",", "Large"}], "]"}], "\[Rule]", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<P\>\"", ",", "Large"}], "]"}]}], ",", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<\[Phi]\>\"", ",", "Large"}], "]"}]}], "}"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"\"\<M\>\"", ",", "Large"}], "]"}], "\[Rule]", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<\[DoubleStruckCapitalT]M\>\"", ",", "Large"}], "]"}]}],
         ",", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<v\>\"", ",", "Large"}], "]"}]}], "}"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"\"\<\[DoubleStruckCapitalT]M\>\"", ",", "Large"}], "]"}], 
         "\[Rule]", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<\[DoubleStruckCapitalT]P\>\"", ",", "Large"}], "]"}]}],
         ",", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<d\[Phi]\>\"", ",", "Large"}], "]"}]}], "}"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"\"\<P\>\"", ",", "Large"}], "]"}], "\[Rule]", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<\[DoubleStruckCapitalT]P\>\"", ",", "Large"}], "]"}]}],
         ",", 
        RowBox[{"Style", "[", 
         RowBox[{"\"\<W\>\"", ",", "Large"}], "]"}]}], "}"}], ",", 
      "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"Style", "[", 
          RowBox[{"\"\<M\>\"", ",", "Large"}], "]"}], "\[Rule]", 
         RowBox[{"Style", "[", 
          RowBox[{"\"\<\[DoubleStruckCapitalT]P\>\"", ",", "Large"}], "]"}]}],
         ",", 
        RowBox[{"Style", "[", 
         RowBox[{
         "\"\<d\[Phi]\[CenterDot]v or W\[SmallCircle]\[Phi]\>\"", ",", 
          "Large"}], "]"}]}], "}"}]}], "\[IndentingNewLine]", "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{"VertexLabeling", "\[Rule]", "True"}], ",", 
    RowBox[{"FormatType", "\[Rule]", "StandardForm"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{"VertexCoordinateRules", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0", ",", "1"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1", ",", "1"}], "}"}]}], "}"}]}]}], "\[IndentingNewLine]", 
   "]"}], "/.", 
  RowBox[{"Network`GraphPlot`wrapper", "\[Rule]", "Identity"}]}]], "Input",
 CellLabel->
  "In[1060]:=",ExpressionUUID->"25097137-6aeb-472d-b5d5-8b9d7f00c425"],

Cell[BoxData[
 GraphicsBox[
  TagBox[GraphicsComplexBox[{{0., 0.}, {1., 0.}, {0., 1.}, {1., 1.}}, {
     {RGBColor[0.5, 0., 0.], Arrowheads[{{0.041582132564841494`, 0.7}}], 
      {Arrowheads[{{0.5, 0.5, 
          GraphicsBox[{
            GrayLevel[0], 
            InsetBox[
             BoxData[
              FormBox[
               StyleBox[
                StyleBox["\"\[Phi]\"", Large, StripOnInput -> False], 
                StripOnInput -> False], TraditionalForm]], {0, 0}, 
             ImageScaled[{0.5, 0.5}], Automatic, None, Background -> 
             GrayLevel[1]]}]}, {0.041582132564841494`, 0.8}}], 
       ArrowBox[{1, 2}]}, 
      {Arrowheads[{{0.5, 0.5, 
          GraphicsBox[{
            GrayLevel[0], 
            InsetBox[
             BoxData[
              FormBox[
               StyleBox[
                StyleBox["\"v\"", Large, StripOnInput -> False], StripOnInput -> 
                False], TraditionalForm]], {0, 0}, 
             ImageScaled[{0.5, 0.5}], Automatic, None, Background -> 
             GrayLevel[1]]}]}, {0.041582132564841494`, 0.8}}], 
       ArrowBox[{1, 3}]}, 
      {Arrowheads[{{0.5, 0.5, 
          GraphicsBox[{
            GrayLevel[0], 
            InsetBox[
             BoxData[
              FormBox[
               StyleBox[
                StyleBox[
                "\"d\[Phi]\[CenterDot]v or W\[SmallCircle]\[Phi]\"", Large, 
                 StripOnInput -> False], StripOnInput -> False], 
               TraditionalForm]], {0, 0}, 
             ImageScaled[{0.5, 0.5}], Automatic, None, Background -> 
             GrayLevel[1]]}]}, {0.041582132564841494`, 0.8}}], 
       ArrowBox[{1, 4}]}, 
      {Arrowheads[{{0.5, 0.5, 
          GraphicsBox[{
            GrayLevel[0], 
            InsetBox[
             BoxData[
              FormBox[
               StyleBox[
                StyleBox["\"W\"", Large, StripOnInput -> False], StripOnInput -> 
                False], TraditionalForm]], {0, 0}, 
             ImageScaled[{0.5, 0.5}], Automatic, None, Background -> 
             GrayLevel[1]]}]}, {0.041582132564841494`, 0.8}}], 
       ArrowBox[{2, 4}]}, 
      {Arrowheads[{{0.5, 0.5, 
          GraphicsBox[{
            GrayLevel[0], 
            InsetBox[
             BoxData[
              FormBox[
               StyleBox[
                StyleBox["\"d\[Phi]\"", Large, StripOnInput -> False], 
                StripOnInput -> False], TraditionalForm]], {0, 0}, 
             ImageScaled[{0.5, 0.5}], Automatic, None, Background -> 
             GrayLevel[1]]}]}, {0.041582132564841494`, 0.8}}], 
       ArrowBox[{3, 4}]}}, {InsetBox[
       FrameBox[
        StyleBox["\<\"M\"\>",
         StripOnInput->False,
         FontSize->Large],
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 1], InsetBox[
       FrameBox[
        StyleBox["\<\"P\"\>",
         StripOnInput->False,
         FontSize->Large],
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 2], InsetBox[
       FrameBox[
        StyleBox["\<\"\[DoubleStruckCapitalT]M\"\>",
         StripOnInput->False,
         FontSize->Large],
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 3], InsetBox[
       FrameBox[
        StyleBox["\<\"\[DoubleStruckCapitalT]P\"\>",
         StripOnInput->False,
         FontSize->Large],
        Background->RGBColor[1, 1, 0.8],
        FrameStyle->RGBColor[0.94, 0.85, 0.36],
        StripOnInput->False], 4]}}],
   Annotation[#, 
    VertexCoordinateRules -> {{0., 0.}, {1., 0.}, {0., 1.}, {1., 1.}}]& ],
  AspectRatio->Automatic,
  FormatType->StandardForm,
  FrameTicks->None,
  PlotRange->All,
  PlotRangePadding->Scaled[0.1]]], "Output",
 CellChangeTimes->{3.728978548468009*^9},
 CellLabel->"Out[1060]=",
 CellID->1784840896,ExpressionUUID->"1f217a17-0181-4ed3-9570-29bb4cd5a988"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["9. More on rules", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}, {3.728976670882564*^9, 
  3.728976674720446*^9}, {3.7289768454424953`*^9, 3.728976848920129*^9}, {
  3.728977115725004*^9, 3.7289771217954493`*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->717570716,ExpressionUUID->"4b0a4311-6f29-4ca4-b89d-024d36e40a11"],

Cell[CellGroupData[{

Cell["9.1. MakeRule", "Subsection",ExpressionUUID->"761534c8-fc5e-4f60-98a9-c5582fe712e1"],

Cell[TextData[{
 "Defining rules is one of the most important components of a tensor package. \
The function ",
 StyleBox["IndexRule",
  FontFamily->"Courier"],
 " takes care of dummy generation, but there are other points to consider \
when defining a rule:\n\t1. Checking consistency between left and right hand \
sides.\n\t2. Defining rules that work only for a given ",
 "vbundle",
 ".\n\t3. Extending the rule to pattern expressions with equivalent \
index-structure under symmetries.\n\t4. Extendig the rule to patterns with a \
different up/down character if metrics are present.\nWe define the function \
",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 ", which takes care of all these things. Note that there is not a parallel \
function for ",
 StyleBox["Set",
  FontFamily->"Courier"],
 ", but below we present a way to automatize a collection of rules."
}], "Text",ExpressionUUID->"1ffec475-bdef-4e2c-b908-94aefb68a921"],

Cell[TextData[{
 StyleBox[ButtonBox["MakeRule",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "MakeRule.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Construct rules between ", "Text"],
 StyleBox["xTensor`", "Text",
  FontFamily->"Courier"],
 StyleBox[" expressions\n", "Text"],
 StyleBox[ButtonBox["AutomaticRules",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "AutomaticRules.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Automatize rules", "Text"]
}], "DefinitionBox",ExpressionUUID->"6dcfa298-e04b-4031-beeb-026f2bd330af"],

Cell["Construction of rules.", "Text",ExpressionUUID->"f88eaf3c-9bb8-4581-9b6b-c3227f3bdb1d"],

Cell[CellGroupData[{

Cell["\<\
Define again the environment for examples. This time we quite the info \
messages:\
\>", "MathCaption",ExpressionUUID->"192ea1b4-8633-4601-86cb-0f91fdbcafa1"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$DefInfoQ", "=", "False"}], ";"}]], "Input",
 CellLabel->
  "In[1061]:=",ExpressionUUID->"5463ce9b-a045-42d0-be25-7079f00fc50b"],

Cell[BoxData[{
 RowBox[{"DefManifold", "[", 
  RowBox[{"M3", ",", "3", ",", 
   RowBox[{"{", 
    RowBox[{
    "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", 
     "h"}], "}"}]}], "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"DefManifold", "[", 
   RowBox[{"S2", ",", "2", ",", 
    RowBox[{"{", 
     RowBox[{"A", ",", "B", ",", "C", ",", "D"}], "}"}]}], "]"}], "//", 
  "Quiet"}], "\[IndentingNewLine]", 
 RowBox[{"DefMetric", "[", 
  RowBox[{
   RowBox[{"-", "1"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", "CD", ",", 
   RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3"}], "}"}], "]"}]}], "]"}]}], "Input",
 CellLabel->
  "In[1062]:=",ExpressionUUID->"3fc62a92-c215-41f6-a20b-9e5526bba196"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$PrePrint", "=", "ScreenDollarIndices"}], ";"}]], "Input",
 CellLabel->
  "In[1068]:=",ExpressionUUID->"9432d1ce-43e9-47df-bdf2-1c4998c8cfe3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The simplest way to define a rule would easily give corrupted answers:\
\>", "MathCaption",ExpressionUUID->"b445bc29-d8e3-453e-9dcb-d60cdfaff1bd"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rule", "=", 
  RowBox[{
   RowBox[{"v", "[", "a_", "]"}], "\[Rule]", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "b"}]}], "]"}], 
    RowBox[{"v", "[", "b", "]"}]}]}]}]], "Input",
 CellLabel->
  "In[1069]:=",ExpressionUUID->"f6514494-ae48-4b57-b507-7d706efe6e6d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {
           UnderscriptBox["a", "_"]},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[
    Pattern[$CellContext`a, 
     Blank[]]],
   Editable->False], "\[Rule]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " "},
           {" ", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978571320746*^9},
 CellLabel->"Out[1069]=",
 CellID->1276669088,ExpressionUUID->"35deabfe-96fb-46af-b4b6-b8b2c7f1080b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "b", "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1070]:=",ExpressionUUID->"7c223395-44bc-40e1-837e-f3c00fac5e14"],

Cell[BoxData[
 TemplateBox[{
  "Validate", "repeated", 
   "\"Found indices with the same name \\!\\(\\*RowBox[{\\\"\\\\\\\"b\\\\\\\"\
\\\"}]\\).\"", 2, 1070, 99, 19253573948027831207, "Local", 
   "xAct`xTensor`Validate"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.7289785719274807`*^9},
 CellLabel->"During evaluation of In[1070]:=",
 CellID->221778279,ExpressionUUID->"45cab096-4fd0-424e-abeb-a582a8677e33"],

Cell[BoxData[
 TemplateBox[{
  "Throw", "nocatch", 
   "\"Uncaught \\!\\(\\*RowBox[{\\\"Throw\\\", \\\"[\\\", \\\"Null\\\", \\\"]\
\\\"}]\\) returned to top level.\"", 2, 1070, 100, 19253573948027831207, 
   "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.72897857196731*^9},
 CellLabel->"During evaluation of In[1070]:=",
 CellID->1543407608,ExpressionUUID->"d1508ed1-e0d7-4ffc-a376-061827ad9908"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{"Throw", "[", "Null", "]"}], "]"}]], "Output",
 CellChangeTimes->{3.728978571981594*^9},
 CellLabel->"Out[1070]=",
 CellID->596695543,ExpressionUUID->"5012442f-444a-43d2-8b58-32d4ca2ce7d2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We have seen that changing to ",
 StyleBox["IndexRule",
  FontFamily->"Courier"],
 " we can solve those problems (underlined indices denote patterns):"
}], "MathCaption",ExpressionUUID->"15e0a955-db0a-4080-ba25-cefa37cb0f6d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rule", "=", 
  RowBox[{
   RowBox[{"v", "[", "a_", "]"}], "\[RightTeeArrow]", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}]}]}]], "Input",
 CellLabel->
  "In[1071]:=",ExpressionUUID->"59bada2f-3310-4391-b913-2abe2b44af10"],

Cell[BoxData[
 RowBox[{
  RowBox[{"HoldPattern", "[", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {
            UnderscriptBox["a", "_"]},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[
     Pattern[$CellContext`a, 
      Blank[]]],
    Editable->False], "]"}], "\[RuleDelayed]", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "b", "}"}], ",", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "b"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`b],
      Editable->False]}]}], "]"}]}]], "Output",
 CellChangeTimes->{3.7289785730679293`*^9},
 CellLabel->"Out[1071]=",
 CellID->1586186550,ExpressionUUID->"4a46b256-f786-4a5d-9bb2-705b3e6aaca7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "b", "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1072]:=",ExpressionUUID->"ef80d674-d317-446e-b03f-592cd2d3300a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"b", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978573705282*^9},
 CellLabel->"Out[1072]=",
 CellID->566591278,ExpressionUUID->"c581f5f9-bec6-4c88-ac6e-78dea16b60fc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", "rule"}]], "Input",
 CellLabel->
  "In[1073]:=",ExpressionUUID->"a201fd4f-9806-4fd7-98a3-f15cf4939eaa"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", "c"},
          {"a", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"b", "a"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`b, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978574317881*^9},
 CellLabel->"Out[1073]=",
 CellID->1979744497,ExpressionUUID->"7a7834cc-d7f8-4063-84f1-83bdaa5030d5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "We can also define the same rule using ",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 ". Note that there is no pattern on the lhs; by default all indices on the \
LHS are converted into patterns:"
}], "MathCaption",ExpressionUUID->"bc48e003-56bc-4675-ba58-98d5572a001e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rule", "=", 
  RowBox[{"MakeRule", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"v", "[", "a", "]"}], ",", 
      RowBox[{
       RowBox[{"T", "[", 
        RowBox[{"a", ",", 
         RowBox[{"-", "b"}]}], "]"}], 
       RowBox[{"v", "[", "b", "]"}]}]}], "}"}], ",", 
    RowBox[{"MetricOn", "\[Rule]", "None"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[1074]:=",ExpressionUUID->"a159baac-85cf-4b58-8904-db23a2b49551"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"HoldPattern", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             UnderscriptBox[
              UnderscriptBox["a", "_"], "_"]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[
      PatternTest[
       Pattern[$CellContext`a, 
        Blank[Symbol]], TangentM3`Q]],
     Editable->False], "]"}], "\[RuleDelayed]", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "b", "}"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", " "},
              {" ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, -$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`b],
       Editable->False]}]}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978575501432*^9},
 CellLabel->"Out[1074]=",
 CellID->1070771511,ExpressionUUID->"2b1d1567-c419-466a-8796-29f3d4a0d307"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], 
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{"v", "[", "b", "]"}]}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1075]:=",ExpressionUUID->"7c0ebce2-4e21-4068-bea2-0322b63e1ec4"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", " "},
          {" ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"b", " "},
          {" ", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`b, -$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"c"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"d"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978576088382*^9},
 CellLabel->"Out[1075]=",
 CellID->1750465817,ExpressionUUID->"33a5b4f9-adb0-4707-b041-bc5fa43ad401"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "rule", "]"}]], "Input",
 CellLabel->
  "In[1076]:=",ExpressionUUID->"11ec3f79-f70e-4680-a2c1-e78f1372c4c7"],

Cell["\<\
{HoldPattern[v[(a_Symbol)?TangentM3`Q]] :> Module[{h$44689}, T[a, \
-h$44689]*v[h$44689]]}\
\>", "Output",
 CellChangeTimes->{3.7289785767187757`*^9},
 CellLabel->"Out[1076]//InputForm=",
 CellID->1089903271,ExpressionUUID->"bf7b287d-d6a7-433f-9754-ad6678eecc56"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Due to the option ",
 StyleBox["TestIndices\[Rule]True",
  FontFamily->"Courier"],
 ", membership of the indices of the LHS is checked. The function ",
 StyleBox["M3`Q",
  FontFamily->"Courier"],
 " checks that an up-index belongs to ",
 StyleBox["M3",
  FontFamily->"Courier"],
 ", but it does not accept down-indices (this is due to the option ",
 StyleBox["MetricOn->None",
  FontFamily->"Courier"],
 "):"
}], "MathCaption",ExpressionUUID->"157f86c5-9615-43e8-ba25-e16d4855cf46"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1077]:=",ExpressionUUID->"70747862-6180-45b2-827a-b668a0744bd8"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", " "},
          {" ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289785782218246`*^9},
 CellLabel->"Out[1077]=",
 CellID->1751303676,ExpressionUUID->"28e476d6-f2d0-4f75-95b9-53509ec5c272"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1078]:=",ExpressionUUID->"f48258d3-d6fe-47af-bcab-7798a36c2256"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {" "},
         {"a"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[-$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978578809684*^9},
 CellLabel->"Out[1078]=",
 CellID->713601693,ExpressionUUID->"2513e7c8-6c1c-42d8-a9ea-e87ac5e011b0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "A", "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1079]:=",ExpressionUUID->"ed3e9c49-3a3d-4f45-ae73-d3247ab9e19f"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"v", 
      StyleBox[GridBox[{
         {"A"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`v[$CellContext`A],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978579426157*^9},
 CellLabel->"Out[1079]=",
 CellID->1271861996,ExpressionUUID->"1d2b66c4-38a7-4023-aeba-6a0542363a8f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In order to allow both up/down-indices of a given vbundle we use ",
 StyleBox["MetricOn",
  FontFamily->"Courier"],
 ". We need to have a metric first. (If the ",
 "vbundle",
 " of the index did not have a metric ",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 " would refuse to change the up/down character):"
}], "MathCaption",ExpressionUUID->"c2b3f63c-e4d7-4574-92f0-db770e5be306"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rule", "=", 
  RowBox[{"MakeRule", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{
      RowBox[{"v", "[", "a", "]"}], ",", 
      RowBox[{
       RowBox[{"T", "[", 
        RowBox[{"a", ",", 
         RowBox[{"-", "b"}]}], "]"}], 
       RowBox[{"v", "[", "b", "]"}]}]}], "}"}], ",", 
    RowBox[{"MetricOn", "\[Rule]", 
     RowBox[{"{", "a", "}"}]}], ",", 
    RowBox[{"ContractMetrics", "\[Rule]", "True"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[1080]:=",ExpressionUUID->"344c9768-08e7-4e27-a1fb-3d55fae69edb"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"HoldPattern", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {
             UnderscriptBox[
              UnderscriptBox["a", "_"], "_"]},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[
      PatternTest[
       Pattern[$CellContext`a, 
        Blank[]], TangentM3`pmQ]],
     Editable->False], "]"}], "\[RuleDelayed]", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "b", "}"}], ",", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", " "},
              {" ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, -$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`b],
       Editable->False]}]}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978580673114*^9},
 CellLabel->"Out[1080]=",
 CellID->1602880341,ExpressionUUID->"f35af55c-6b99-43c5-846d-e958b1529361"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "rule", "]"}]], "Input",
 CellLabel->
  "In[1081]:=",ExpressionUUID->"d3d435e2-b281-47d7-9cf1-38945c4469d5"],

Cell["\<\
{HoldPattern[v[(a_)?TangentM3`pmQ]] :> Module[{h$44706}, T[a, \
-h$44706]*v[h$44706]]}\
\>", "Output",
 CellChangeTimes->{3.728978581268605*^9},
 CellLabel->"Out[1081]//InputForm=",
 CellID->379773373,ExpressionUUID->"9535121c-c622-4a9b-91de-99bc00dc5c5c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1082]:=",ExpressionUUID->"b7cda420-df12-451f-aba6-5b1156b6467f"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", " "},
          {" ", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978581939163*^9},
 CellLabel->"Out[1082]=",
 CellID->1666256820,ExpressionUUID->"0d99160b-cdde-4006-92df-8f0be2ee8d18"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", 
   RowBox[{"-", "a"}], "]"}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1083]:=",ExpressionUUID->"170234ba-b43f-4af0-8d06-9a2f11cd835d"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978582558051*^9},
 CellLabel->"Out[1083]=",
 CellID->1740720426,ExpressionUUID->"c064f791-cf9c-4c13-ad56-88ecfccf6596"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If we have symmetries the situation gets even more complicated:\
\>", "MathCaption",ExpressionUUID->"f91379e6-6cef-4381-84cf-37d89a670974"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"(", 
  RowBox[{"rule", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"U", "[", 
         RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
        RowBox[{"v", "[", 
         RowBox[{"-", "c"}], "]"}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"v", "[", "a", "]"}], 
         RowBox[{"T", "[", "b", "]"}]}], "-", 
        RowBox[{
         RowBox[{"v", "[", "b", "]"}], 
         RowBox[{"T", "[", "a", "]"}]}]}]}], "}"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}], ",", 
     RowBox[{"ContractMetrics", "\[Rule]", "True"}]}], "]"}]}], 
  ")"}]], "Input",
 CellLabel->
  "In[1084]:=",ExpressionUUID->"814ecefb-6203-484f-9df2-5e64d5f4bdb6"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"], " "},
              {" ", " ", 
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ], -PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {"b"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[$CellContext`b],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"], " ", 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"]},
              {" ", 
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ], -PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {"b"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[$CellContext`b],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"], " ", 
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"]},
              {" ", 
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ], -PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", 
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"]},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "-", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"]},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {"b"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[$CellContext`b],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"HoldPattern", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", 
               UnderscriptBox[
                UnderscriptBox["b", "_"], "_"], 
               UnderscriptBox[
                UnderscriptBox["a", "_"], "_"]},
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"], " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-PatternTest[
          Pattern[$CellContext`c, 
           Blank[Symbol]], TangentM3`Q], 
        PatternTest[
         Pattern[$CellContext`b, 
          Blank[]], TangentM3`pmQ], 
        PatternTest[
         Pattern[$CellContext`a, 
          Blank[]], TangentM3`pmQ]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {
               UnderscriptBox[
                UnderscriptBox["c", "_"], "_"]},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[
        PatternTest[
         Pattern[$CellContext`c, 
          Blank[Symbol]], TangentM3`Q]],
       Editable->False]}], "]"}], "\[RuleDelayed]", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "}"}], ",", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {"b"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[$CellContext`b],
          Editable->False]}], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}]}]}], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978583741191*^9},
 CellLabel->"Out[1084]=",
 CellID->2036193382,ExpressionUUID->"e9781e6d-3ed1-4abe-aff6-d599a3e21c27"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "b", ",", "c"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}]}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1085]:=",ExpressionUUID->"b1160f47-0d43-424a-bd44-4d5d45659466"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`c],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978584772752*^9},
 CellLabel->"Out[1085]=",
 CellID->2072497326,ExpressionUUID->"6dfe6d05-d804-4a65-969d-0ce54a014ed0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", "b", ",", "c"}], "]"}], 
   RowBox[{"v", "[", "a", "]"}]}], "/.", "rule"}]], "Input",
 CellLabel->
  "In[1086]:=",ExpressionUUID->"5c516a3f-c7a1-4af3-8d41-1fdb2f5c4ea5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978585406519*^9},
 CellLabel->"Out[1086]=",
 CellID->1969517102,ExpressionUUID->"21e17d17-0f69-4944-83b2-d3096d4ae0d2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", "c", ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"v", "[", "a", "]"}]}], "/.", "rule"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[1087]:=",ExpressionUUID->"15dd5045-cfd4-4025-98e5-78deb2f2222f"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289785860187283`*^9},
 CellLabel->"Out[1087]=",
 CellID->1006228165,ExpressionUUID->"e4171b49-ed86-4d79-bf0a-8fb453dd5fc6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "It is possible to automate rules, declaring them as definitions for a \
symbol, in this case the tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"8b840f1b-4144-4c72-bc8c-6d43fcb486f0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AutomaticRules", "[", 
  RowBox[{"U", ",", "rule"}], "]"}]], "Input",
 CellLabel->
  "In[1088]:=",ExpressionUUID->"2dfb92d4-1a57-413f-bd12-4609063e0a76"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"   Rules \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
    "1", ",", "2", ",", "3", ",", "4", ",", "5", ",", "6", ",", "7", ",", "8",
      ",", "9", ",", "10"}], "}"}], 
   "\[InvisibleSpace]", "\<\" have been declared as UpValues for \"\>", 
   "\[InvisibleSpace]", "U", "\[InvisibleSpace]", "\<\".\"\>"}],
  SequenceForm["   Rules ", 
   Shallow[{1, 2, 3, 4, 5, 6, 7, 8, 9, 10}], 
   " have been declared as UpValues for ", $CellContext`U, "."],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978587269072*^9},
 CellLabel->"During evaluation of In[1088]:=",
 CellID->2010194070,ExpressionUUID->"2fb8372d-e71a-496f-a56a-8be6adb57665"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "b", ",", "c"}], "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"-", "b"}], "]"}]}]], "Input",
 CellLabel->
  "In[1089]:=",ExpressionUUID->"23ee9058-9d3d-47b9-ac84-1c1ae6626821"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`c],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978587886525*^9},
 CellLabel->"Out[1089]=",
 CellID->722037073,ExpressionUUID->"f154215c-9702-4bf5-98e0-b98d82822925"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"U", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "b", ",", "c"}], "]"}], 
  RowBox[{"v", "[", "a", "]"}]}]], "Input",
 CellLabel->
  "In[1090]:=",ExpressionUUID->"58663585-91bb-4ee2-9d38-1e3ccee6783b"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289785885377493`*^9},
 CellLabel->"Out[1090]=",
 CellID->1484724578,ExpressionUUID->"d54daa98-1253-437b-b1bf-f7c7f1e5946c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "U"}]], "Input",
 CellLabel->
  "In[1091]:=",ExpressionUUID->"065675b8-2453-4073-9d8f-4da9bc6851a0"],

Cell[CellGroupData[{

Cell["Global`U", "Print", "PrintUsage",
 CellChangeTimes->{3.728978589289118*^9},
 CellTags->"Info10913728982189-9181623",
 CellID->1587009223,ExpressionUUID->"43e2b182-5295-45a0-97ed-b7fecafa9b13"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         InterpretationBox[
          StyleBox[
           RowBox[{"Dagger", "[", "U", "]"}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          xAct`xTensor`Dagger[$CellContext`U],
          Editable->False], "^:=", "U"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "U", "]"}], "^:=", 
         RowBox[{"{", 
          RowBox[{"\<\"tensor\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DependenciesOfTensor", "[", "U", "]"}], "^:=", 
         RowBox[{"{", "M3", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"HostsOf", "[", "U", "]"}], "^:=", 
         RowBox[{"{", 
          RowBox[{"M3", ",", 
           InterpretationBox[
            StyleBox["\[DoubleStruckCapitalT]M3",
             ShowAutoStyles->False,
             AutoSpacing->False],
            $CellContext`TangentM3,
            Editable->False]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "U", "]"}], "^:=", "\<\"U\"\>"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SlotsOfTensor", "[", "U", "]"}], "^:=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}], ",", 
           RowBox[{"-", 
            InterpretationBox[
             StyleBox["\[DoubleStruckCapitalT]M3",
              ShowAutoStyles->False,
              AutoSpacing->False],
             $CellContext`TangentM3,
             Editable->False]}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SymmetryGroupOfTensor", "[", "U", "]"}], "^:=", 
         RowBox[{"StrongGenSet", "[", 
          RowBox[{
           RowBox[{"{", 
            RowBox[{"1", ",", "2", ",", "3"}], "}"}], ",", 
           RowBox[{"GenSet", "[", 
            RowBox[{
             InterpretationBox[
              StyleBox[
               RowBox[{"-", 
                RowBox[{"(", "1", ",", "2", ")"}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              -xAct`xPerm`Cycles[{1, 2}],
              Editable->False], ",", 
             InterpretationBox[
              StyleBox[
               RowBox[{"-", 
                RowBox[{"(", "2", ",", "3", ")"}]}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              -xAct`xPerm`Cycles[{2, 3}],
              Editable->False]}], "]"}]}], "]"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"TensorID", "[", "U", "]"}], "^:=", 
         RowBox[{"{", "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"xTensorQ", "[", "U", "]"}], "^:=", "True"}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`b],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "-", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"], " "},
                  {" ", " ", 
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ], -PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`b],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "-", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"]},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             RowBox[{"-", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"T", 
                   StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T[$CellContext`b],
               Editable->False]}], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"], " ", 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"]},
                  {" ", 
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ], -PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             RowBox[{"-", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"T", 
                   StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T[$CellContext`b],
               Editable->False]}], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"]},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`b],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "-", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"], " ", 
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"]},
                  {" ", 
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ], -PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`b],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "-", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"]},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`b],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "-", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {" ", 
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"]},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`b],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "-", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"]},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             RowBox[{"-", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"T", 
                   StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T[$CellContext`b],
               Editable->False]}], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]},
       {" "},
       {
        RowBox[{"U", "/:", 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {" ", 
                   UnderscriptBox[
                    UnderscriptBox["b", "_"], "_"], 
                   UnderscriptBox[
                    UnderscriptBox["a", "_"], "_"]},
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"], " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[-PatternTest[
              Pattern[$CellContext`c, 
               Blank[Symbol]], TangentM3`Q], 
            PatternTest[
             Pattern[$CellContext`b, 
              Blank[]], TangentM3`pmQ], 
            PatternTest[
             Pattern[$CellContext`a, 
              Blank[]], TangentM3`pmQ]],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {
                   UnderscriptBox[
                    UnderscriptBox["c", "_"], "_"]},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[
            PatternTest[
             Pattern[$CellContext`c, 
              Blank[Symbol]], TangentM3`Q]],
           Editable->False]}], ":=", 
         RowBox[{"Module", "[", 
          RowBox[{
           RowBox[{"{", "}"}], ",", 
           RowBox[{
            RowBox[{
             RowBox[{"-", 
              InterpretationBox[
               StyleBox[GridBox[{
                  {"T", 
                   StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                    FontSize->9]}
                 },
                 
                 GridBoxAlignment->{
                  "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                   "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
                ShowAutoStyles->False,
                AutoSpacing->False],
               $CellContext`T[$CellContext`b],
               Editable->False]}], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`a],
              Editable->False]}], "+", 
            RowBox[{
             InterpretationBox[
              StyleBox[GridBox[{
                 {"T", 
                  StyleBox[GridBox[{
                    {"a"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`T[$CellContext`a],
              Editable->False], " ", 
             InterpretationBox[
              StyleBox[GridBox[{
                 {"v", 
                  StyleBox[GridBox[{
                    {"b"},
                    {" "}
                    },
                    GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                   FontSize->9]}
                },
                
                GridBoxAlignment->{
                 "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {{Center}}, "RowsIndexed" -> {}},
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.27999999999999997`], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
                  "Rows" -> {
                    Offset[0.2], {
                    Offset[0.4]}, 
                    Offset[0.2]}, "RowsIndexed" -> {}}],
               ShowAutoStyles->False,
               AutoSpacing->False],
              $CellContext`v[$CellContext`b],
              Editable->False]}]}]}], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289785893730793`*^9},
 CellTags->"Info10913728982189-9181623",
 CellID->1117115221,ExpressionUUID->"8248b8f0-89c7-4a25-a0f3-d2e87fb130d5"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The simplest way to remove those definitions for U is removing the tensor and \
defining it again:\
\>", "MathCaption",ExpressionUUID->"eaf34494-1f4c-4f40-898f-b3c02ea804bf"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefTensor", "[", "U", "]"}]], "Input",
 CellLabel->
  "In[1092]:=",ExpressionUUID->"046aad13-b58e-4dbf-8fdd-57b81633cb81"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978591042122*^9},
 CellLabel->"During evaluation of In[1092]:=",
 CellID->1139610945,ExpressionUUID->"c632a3f9-4aa1-4652-b42d-aa8b9ae9a188"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], ",", "M3", ",", 
   RowBox[{"Antisymmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2", ",", "3"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1093]:=",ExpressionUUID->"1d3ba07f-bb3a-49d9-88ed-e1ff729ce904"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["9.2. Riemann expressions", "Subsection",ExpressionUUID->"4e047b77-a755-46c9-be86-b00ebd7e5b66"],

Cell[TextData[{
 "As we said, there is no algorithm at present in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " to canonicalize expressions with multiterm symmetries. That is an obstacle \
in GR when one wants to work with the Riemann tensor, which has a cyclic \
symmetry. A simple, brute-force, solution is to prepare a number of rules \
which are only valid for the Riemann tensor. We follow MathTensor's \
discussion of this point (see MathTensor book).\nThere are 48 rules in total, \
numbered from 1 to 40, some of them having subcases a, b, c, ...\n14 of those \
rules are simple consequences of the monoterm symmetries of the Riemann \
tensor, and hence can be considered already implemented in ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 ":"
}], "Text",ExpressionUUID->"68d2fc2a-30c7-4870-b309-54629143979a"],

Cell[CellGroupData[{

Cell["RiemannRule1:", "MathCaption",ExpressionUUID->"eb30f429-b4e3-4116-b398-86ec8a694030"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "a", ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "d"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1094]:=",ExpressionUUID->"3618dc1a-a6b6-4e89-8ba9-69d47f81e75a"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289785995333033`*^9},
 CellLabel->"Out[1094]=",
 CellID->1162546407,ExpressionUUID->"8c8580ce-5d3d-4e01-a05b-f9b8f0308014"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "c"}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1095]:=",ExpressionUUID->"9df949b3-b28a-4890-8124-eb93d5ae66fd"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289786003080387`*^9},
 CellLabel->"Out[1095]=",
 CellID->1333229371,ExpressionUUID->"771db3a0-c230-4ee3-a241-8b809137806f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule2:", "MathCaption",ExpressionUUID->"21328013-d11e-4ab6-a5bc-ecfeb6b27d5c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", "a", ",", 
    RowBox[{"-", "d"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1096]:=",ExpressionUUID->"cffe264e-c7d9-485f-a476-0b857ec26f77"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"b", "d"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RicciCD[-$CellContext`b, -$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978602127598*^9},
 CellLabel->"Out[1096]=",
 CellID->1546588497,ExpressionUUID->"18aa9b5e-05d2-43d3-94ab-39b8d6470a84"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule3:", "MathCaption",ExpressionUUID->"488d9216-9806-4285-8488-71b33bd00e0b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "a"}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1097]:=",ExpressionUUID->"0b3df27e-40c3-42c6-8cd9-b4e32a6b7a29"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`b, -$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978603431882*^9},
 CellLabel->"Out[1097]=",
 CellID->687942202,ExpressionUUID->"2be60db1-5075-4a8c-adad-89c5cf8ffa68"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule4:", "MathCaption",ExpressionUUID->"5b59ed10-9fad-46b9-938b-56c0d32570dc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", "b", ",", 
    RowBox[{"-", "d"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1098]:=",ExpressionUUID->"b5a73ad4-e5c4-493a-9c69-40421cad7874"],

Cell[BoxData[
 RowBox[{"-", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978605800941*^9},
 CellLabel->"Out[1098]=",
 CellID->519529191,ExpressionUUID->"152b13dc-5a3e-4fcd-ad8c-9f1ba189103d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule5:", "MathCaption",ExpressionUUID->"99177654-52ef-4afc-b38b-9f33c549f6f6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"RiemannCD", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", 
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", "b"}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1099]:=",ExpressionUUID->"c3f5e872-0457-43ea-b0a6-872cd7da1f5e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {
      RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
      StyleBox[GridBox[{
         {" ", " "},
         {"a", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`RicciCD[-$CellContext`a, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978607642878*^9},
 CellLabel->"Out[1099]=",
 CellID->372594429,ExpressionUUID->"8bcad7ae-9895-4fd3-979f-c926e8335e44"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule6:", "MathCaption",ExpressionUUID->"2c972818-2c95-4f63-a8ed-29d0d7480560"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RicciCD", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1100]:=",ExpressionUUID->"bb9246aa-7aff-4776-93e0-16c3b8d5485d"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978610392332*^9},
 CellLabel->"Out[1100]=",
 CellID->975912735,ExpressionUUID->"6e1ae10d-0cd0-45aa-b529-83498e15e90b"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RicciCD", "[", 
    RowBox[{"c", ",", "d"}], "]"}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1101]:=",ExpressionUUID->"5e092416-608c-4704-a227-13305a8eeaaa"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978611046129*^9},
 CellLabel->"Out[1101]=",
 CellID->1677838393,ExpressionUUID->"c1066037-eb51-4b59-9394-ecbf39f4ec54"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule31:", "MathCaption",ExpressionUUID->"58604d53-f0cc-4811-9db8-e8b9e5d4dd21"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{"RicciCD", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1102]:=",ExpressionUUID->"10b8feeb-0d2a-4a83-bcad-f3432de20454"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289786121426973`*^9},
 CellLabel->"Out[1102]=",
 CellID->2017726656,ExpressionUUID->"53974fec-36f2-46ec-8b2b-06b41d12e4fc"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{"RicciCD", "[", 
    RowBox[{"c", ",", "d"}], "]"}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1103]:=",ExpressionUUID->"4b40eefd-0ecb-410b-bcb2-f31100811675"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978612752234*^9},
 CellLabel->"Out[1103]=",
 CellID->794599648,ExpressionUUID->"24cd6c3c-8563-4097-aa74-d7d672e867cd"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule32:", "MathCaption",ExpressionUUID->"3d964f36-d20f-456c-96d3-2487e55906a2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "f"}], "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{"a", ",", "b"}], "]"}]}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1104]:=",ExpressionUUID->"c86ba01e-8e21-44d8-951a-9cbdec0b7e4c"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289786137536383`*^9},
 CellLabel->"Out[1104]=",
 CellID->2073260556,ExpressionUUID->"7f0ab981-baf2-4e05-aa59-9ad89cc0e97d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "f"}], "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{"c", ",", "d"}], "]"}]}]}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1105]:=",ExpressionUUID->"75d7f3e2-e004-4f2e-be1e-2befc57dd2dc"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289786147689543`*^9},
 CellLabel->"Out[1105]=",
 CellID->1837683274,ExpressionUUID->"d3d01595-4841-4a60-b4b7-6d2c6d6fbbd3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule33:", "MathCaption",ExpressionUUID->"5e361936-1380-46ec-b718-4f652c419cde"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "c"}], "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"a", ",", "b", ",", "d", ",", "e"}], "]"}]}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[1106]:=",ExpressionUUID->"1d4d819e-e146-4a3f-bb66-824ea453d58e"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978615733417*^9},
 CellLabel->"Out[1106]=",
 CellID->202417881,ExpressionUUID->"5c9bd9fb-75f0-4efa-89ee-2e82f8d906a6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "c"}], "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"d", ",", "e", ",", "a", ",", "b"}], "]"}]}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[1107]:=",ExpressionUUID->"390fe4eb-29d7-4060-9b0d-226e25cd5455"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978616339088*^9},
 CellLabel->"Out[1107]=",
 CellID->565207425,ExpressionUUID->"ec707050-51d8-40ee-9c53-f60d44e293ca"]
}, Open  ]]
}, Open  ]],

Cell["\<\
There are 2 more rules which can be obtained from the fact that the Einstein \
tensor is divergence free:\
\>", "Text",ExpressionUUID->"07f1d41e-aeb7-4a88-9f81-a5313fc05b87"],

Cell[CellGroupData[{

Cell["RiemannRule16:", "MathCaption",
 CellChangeTimes->{{3.728978617574692*^9, 
  3.728978618414003*^9}},ExpressionUUID->"679fa4c8-be0b-4eaf-9c57-\
cae8b1e6038d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", "b", "]"}], "[", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], "//", "RicciToEinstein"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[1108]:=",ExpressionUUID->"b24780f4-7683-407b-bea2-d676dbdb3a96"],

Cell[BoxData[
 FractionBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], 
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`RicciScalarCD[]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728978619565283*^9},
 CellLabel->"Out[1108]=",
 CellID->353549070,ExpressionUUID->"f1bbe9a2-315e-4567-a080-b895614c4762"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule17:", "MathCaption",ExpressionUUID->"10632beb-7aea-4694-bbae-6eb2cff4ef69"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", "a", "]"}], "[", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "]"}], "//", "RicciToEinstein"}], "//", 
  "ContractMetric"}]], "Input",
 CellLabel->
  "In[1109]:=",ExpressionUUID->"b7e7aee0-15e6-4880-862e-13c488c79896"],

Cell[BoxData[
 FractionBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], 
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`RicciScalarCD[]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728978620712853*^9},
 CellLabel->"Out[1109]=",
 CellID->1155081646,ExpressionUUID->"afcd036e-10b3-4657-b955-16de85725fe6"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "The other 32 rules are actual consequences of the cyclic symmetry of the \
Riemann tensor, and therefore must be programmed independently. Using ",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 " we can avoid typing equivalent rules (up to permutations of indices), \
reducing the number to 17 independent rules.\nWe first study 13 rules having \
derivatives of a single curvature tensor on the LHS. They are in fact only 6 \
independent rules:\n\t",
 StyleBox["RuleDivRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleDivGradRicci",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleBoxRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleDivGradRiemann\n",
  FontFamily->"Courier"],
 "\t",
 StyleBox["RuleD4RicciScalar",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleD4Ricci",
  FontFamily->"Courier"]
}], "Text",ExpressionUUID->"5f143f82-3dc4-4fb4-bb96-aa402fe86fe9"],

Cell[CellGroupData[{

Cell["The four rules 20--23 are all the same:", "MathCaption",ExpressionUUID->"5787172f-9759-44e6-bd45-eba1ae92e0d7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleDivRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"CD", "[", "a", "]"}], "@", 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "d"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "b"}], ",", 
            RowBox[{"-", "c"}]}], "]"}]}]}], "+", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "c"}], "]"}], "@", 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1110]:=",ExpressionUUID->"1c0e0d19-446e-4838-8c4c-a26870c18f70"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule20:", "MathCaption",ExpressionUUID->"e30d8502-c3f0-4be7-a745-1b51f8aba136"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "a", "]"}], "@", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "/.", "RuleDivRiemann"}]], "Input",
 CellLabel->
  "In[1111]:=",ExpressionUUID->"ac18f7a3-f1f6-4279-8d9f-ca1a1d279a00"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`c][
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`d]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978626812598*^9},
 CellLabel->"Out[1111]=",
 CellID->838488775,ExpressionUUID->"7353b8ea-af54-4046-85dd-553522be6860"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule21:", "MathCaption",ExpressionUUID->"844e07f2-1c2c-4be8-b14b-8357613c819e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "b", "]"}], "@", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "/.", "RuleDivRiemann"}]], "Input",
 CellLabel->
  "In[1112]:=",ExpressionUUID->"a79c6459-5fa1-46d2-847f-0b71acca9667"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`d]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`d][
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`c]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978627793942*^9},
 CellLabel->"Out[1112]=",
 CellID->411981408,ExpressionUUID->"5b0e5cb6-57bf-494a-8d56-bad43524c793"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule22:", "MathCaption",ExpressionUUID->"235349fe-ff76-48fc-a505-352233681c2a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "c", "]"}], "@", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "/.", "RuleDivRiemann"}]], "Input",
 CellLabel->
  "In[1113]:=",ExpressionUUID->"318bc09f-19a0-4bfa-af51-861c03e37eca"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`RicciCD[-$CellContext`d, -$CellContext`b]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`RicciCD[-$CellContext`d, -$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978628807454*^9},
 CellLabel->"Out[1113]=",
 CellID->1558873201,ExpressionUUID->"0faa4075-adae-4f29-84b5-98e182205bc9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule23:", "MathCaption",ExpressionUUID->"898bc111-8f28-4631-972c-4ed900fadd60"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "d", "]"}], "@", 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "/.", "RuleDivRiemann"}]], "Input",
 CellLabel->
  "In[1114]:=",ExpressionUUID->"738e6b76-f1a0-43b1-815b-6d4b6ad05f1f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"c", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`RicciCD[-$CellContext`c, -$CellContext`b]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`a]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289786297767153`*^9},
 CellLabel->"Out[1114]=",
 CellID->1276281109,ExpressionUUID->"a6c1a68f-fa6c-4d4f-826e-d7fcaec6f463"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Create more indices:", "MathCaption",ExpressionUUID->"40ca9974-4c0c-4e87-9e9d-2a53711a425d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"GetIndicesOfVBundle", "[", 
  RowBox[{"TangentM3", ",", "14"}], "]"}]], "Input",
 CellLabel->
  "In[1115]:=",ExpressionUUID->"923815a3-d557-4bd8-a712-666bacbe1b77"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "a", ",", "b", ",", "c", ",", "d", ",", "e", ",", "f", ",", "g", ",", "h", 
   ",", "h1", ",", "h2", ",", "h3", ",", "h4", ",", "h5", ",", "h6"}], 
  "}"}]], "Output",
 CellChangeTimes->{3.728978630835025*^9},
 CellLabel->"Out[1115]=",
 CellID->388405617,ExpressionUUID->"848afb47-8e20-4bd8-bb4b-9830928596d3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Two rules with second derivatives of Ricci can be combined in a single rule:\
\>", "MathCaption",ExpressionUUID->"3ef9ba87-65d2-497f-b0ec-5ea795ec1baa"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleDivGradRicci", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"CD", "[", "b", "]"}], "@", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "c"}], "]"}], "@", 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}]}], "]"}]}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"1", "/", "2"}], 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "c"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "a"}], "]"}], "@", 
           RowBox[{"RicciScalarCD", "[", "]"}]}]}]}], "+", 
        RowBox[{
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "a"}]}], "]"}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "c"}], ",", "h1"}], "]"}]}], "-", 
        RowBox[{
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", "h2", ",", 
           RowBox[{"-", "c"}], ",", "h1"}], "]"}]}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1116]:=",ExpressionUUID->"829dcf18-00aa-482c-95a0-0815d880b6b4"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule18:", "MathCaption",ExpressionUUID->"f276cfcd-062c-4f07-9d4a-778bf727546f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "b", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "c"}], "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}]}], "/.", "RuleDivGradRicci"}]], "Input",
 CellLabel->
  "In[1117]:=",ExpressionUUID->"ef0e109f-06bb-4efa-9a51-614801d3e996"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "b"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, $CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "d", " ", "b"},
           {"a", " ", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`d, -$CellContext`c, \
$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`a][
       $CellContext`RicciScalarCD[]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978633095573*^9},
 CellLabel->"Out[1117]=",
 CellID->1335071829,ExpressionUUID->"08e62f19-50f1-414c-876d-8788adfdce96"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule19:", "MathCaption",ExpressionUUID->"71ec6c65-d70b-4b7e-9cea-e87e899c0d4b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "a", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "c"}], "]"}], "@", 
    RowBox[{"RicciCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}]}]}], "/.", "RuleDivGradRicci"}]], "Input",
 CellLabel->
  "In[1118]:=",ExpressionUUID->"768a75ea-74c3-47b9-a769-98049c746812"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "d", " ", "a"},
           {"b", " ", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`d, -$CellContext`c, \
$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "b"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`CD[-$CellContext`b][
       $CellContext`RicciScalarCD[]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978634145083*^9},
 CellLabel->"Out[1118]=",
 CellID->2090386549,ExpressionUUID->"1271fa8c-c57d-4c03-b9ef-8957d98f7d94"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Rules 26 to 30 are now two rules in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"53dcc64d-6486-4310-b27b-ddb58946ef02"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleBoxRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"CD", "[", "e", "]"}], "@", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}]}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "b"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "d"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "a"}], ",", 
            RowBox[{"-", "c"}]}], "]"}]}]}], "-", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "b"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "c"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "a"}], ",", 
            RowBox[{"-", "d"}]}], "]"}]}]}], "-", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "a"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "d"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "b"}], ",", 
            RowBox[{"-", "c"}]}], "]"}]}]}], "+", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "a"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "c"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "b"}], ",", 
            RowBox[{"-", "d"}]}], "]"}]}]}], "+", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "b"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", "h1", ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}], "-", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "a"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "b"}], ",", "h1", ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}], "-", "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", "h1", ",", 
           RowBox[{"-", "b"}], ",", "h2"}], "]"}]}], "-", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", "h1", ",", 
           RowBox[{"-", "c"}], ",", "h2"}], "]"}]}], "+", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "c"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", "h1", ",", 
           RowBox[{"-", "d"}], ",", "h2"}], "]"}]}], "+", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "b"}], ",", "h1", ",", 
           RowBox[{"-", "c"}], ",", "h2"}], "]"}]}], "-", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "c"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "b"}], ",", "h1", ",", 
           RowBox[{"-", "d"}], ",", "h2"}], "]"}]}], "-", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "b"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}], ",", "h1", ",", "h2"}], "]"}]}]}]}], "}"}], 
     ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1119]:=",ExpressionUUID->"1745cfbb-48a4-4306-8b87-67422778d505"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleDivGradRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"CD", "[", "a", "]"}], "@", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}]}], ",", 
       RowBox[{
        RowBox[{"-", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "e"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "d"}], "]"}], "@", 
           RowBox[{"RicciCD", "[", 
            RowBox[{
             RowBox[{"-", "b"}], ",", 
             RowBox[{"-", "c"}]}], "]"}]}]}]}], "+", 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "c"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "b"}], ",", 
            RowBox[{"-", "d"}]}], "]"}]}]}], "-", 
        RowBox[{
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "e"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "b"}], ",", "h1", ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}], "+", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "b"}], ",", "h1", ",", 
           RowBox[{"-", "e"}], ",", "h2"}], "]"}]}], "-", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "c"}], ",", "h2", ",", 
           RowBox[{"-", "e"}], ",", "h1"}], "]"}]}], "+", 
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "c"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "d"}], ",", "h2", ",", 
           RowBox[{"-", "e"}], ",", "h1"}], "]"}]}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1120]:=",ExpressionUUID->"c24797a3-ae3b-4def-8131-732773456dd4"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule26:", "MathCaption",ExpressionUUID->"b02f1b7f-0579-45aa-ba4b-9a6c6564cac1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "e", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}]}], "/.", "RuleBoxRiemann"}]], "Input",
 CellLabel->
  "In[1121]:=",ExpressionUUID->"237a8781-1685-46f4-a9e3-bae475d3bb05"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"e", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`e, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", " "},
           {"a", " ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`e, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"e", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`e, -$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", " "},
           {"b", " ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`e, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "e", "f"},
           {"c", "d", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`d, $CellContext`e, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"e", "a", "f", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`a, -$CellContext`f, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", "f"},
           {"b", " ", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`e, -$CellContext`d, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"e", "a", "f", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`a, -$CellContext`f, \
-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", "f"},
           {"b", " ", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`e, -$CellContext`c, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"e", "a", "f", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`a, -$CellContext`f, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", "f"},
           {"a", " ", "d", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`e, -$CellContext`d, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"e", "b", "f", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, -$CellContext`f, \
-$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", "f"},
           {"a", " ", "c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`e, -$CellContext`c, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"e", "b", "f", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`b, -$CellContext`f, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "e", " ", "f"},
           {"a", " ", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`e, -$CellContext`b, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"e", "f", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`CD[-$CellContext`c][
     $CellContext`RicciCD[-$CellContext`b, -$CellContext`d]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "a"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`a][
    $CellContext`CD[-$CellContext`d][
     $CellContext`RicciCD[-$CellContext`b, -$CellContext`c]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`CD[-$CellContext`c][
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`d]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "b"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`b][
    $CellContext`CD[-$CellContext`d][
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289786379321136`*^9},
 CellLabel->"Out[1121]=",
 CellID->275472198,ExpressionUUID->"b5961346-f56b-4c54-84c5-acd600f4ce52"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule27:", "MathCaption",ExpressionUUID->"dea448ee-919a-4df3-8cc6-2b1db90e670b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "a", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}]}], "/.", "RuleDivGradRiemann"}]], "Input",\

 CellLabel->
  "In[1122]:=",ExpressionUUID->"fa447cb9-6443-4f26-8e79-284bcdb9dffe"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`e],
     Editable->False]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "a", " ", " "},
           {"b", " ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`a, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "f", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "a", " ", "f"},
           {"b", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`a, -$CellContext`e, \
$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "b", "f", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`f, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "a"},
           {"c", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, $CellContext`f, -$CellContext`e, \
$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "b", "f", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`f, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "a"},
           {"d", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, $CellContext`f, -$CellContext`e, \
$CellContext`a],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`c][
     $CellContext`RicciCD[-$CellContext`b, -$CellContext`d]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`d][
     $CellContext`RicciCD[-$CellContext`b, -$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978639074306*^9},
 CellLabel->"Out[1122]=",
 CellID->19996476,ExpressionUUID->"91b9a257-f14c-4bd4-8496-752c6ad545a1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule28:", "MathCaption",ExpressionUUID->"c5a12d90-cd26-4507-9184-aad098e8c77f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "b", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}]}], "/.", "RuleDivGradRiemann"}]], "Input",\

 CellLabel->
  "In[1123]:=",ExpressionUUID->"de6b9e74-bfcd-41db-bb6e-93e5f2b1b299"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "b", " ", " "},
           {"a", " ", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`b, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "b", " ", "f"},
           {"a", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`b, -$CellContext`e, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"b", "f", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`f, -$CellContext`c, \
-$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"b", "a", "f", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`f, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "b"},
           {"c", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, $CellContext`f, -$CellContext`e, \
$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"b", "a", "f", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, -$CellContext`a, -$CellContext`f, \
-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "b"},
           {"d", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, $CellContext`f, -$CellContext`e, \
$CellContext`b],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`c][
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`d]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "d"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`d][
     $CellContext`RicciCD[-$CellContext`a, -$CellContext`c]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289786401438704`*^9},
 CellLabel->"Out[1123]=",
 CellID->593130850,ExpressionUUID->"2ce58a6c-670d-441a-88e5-a1f22040174d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule29:", "MathCaption",ExpressionUUID->"7886fee3-2fd5-4360-a3ed-bf07203091d1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "c", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}]}], "/.", "RuleDivGradRiemann"}]], "Input",\

 CellLabel->
  "In[1124]:=",ExpressionUUID->"8baafca8-ab74-438b-9009-b2ba7d59144a"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "c"},
           {"b", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`f, -$CellContext`e, \
$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"c", "d", "f", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`d, -$CellContext`f, \
-$CellContext`a],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "c"},
           {"a", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`f, -$CellContext`e, \
$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"c", "d", "f", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`d, -$CellContext`f, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "c", " ", " "},
           {"d", " ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, $CellContext`c, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"c", "f", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, -$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "c", " ", "f"},
           {"d", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, $CellContext`c, -$CellContext`e, \
$CellContext`f],
    Editable->False]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`a][
     $CellContext`RicciCD[-$CellContext`d, -$CellContext`b]]],
   Editable->False], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`b][
     $CellContext`RicciCD[-$CellContext`d, -$CellContext`a]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978641226193*^9},
 CellLabel->"Out[1124]=",
 CellID->1279502412,ExpressionUUID->"6ac9b906-ae86-4431-9b23-8b4656e8e60f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule30:", "MathCaption",ExpressionUUID->"fdf14473-f8ff-420a-8a9b-39b51951e97c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "d", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}]}], "/.", "RuleDivGradRiemann"}]], "Input",\

 CellLabel->
  "In[1125]:=",ExpressionUUID->"586a5b1b-52ef-4457-9114-e0f262a21fb9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "d", " ", " "},
           {"c", " ", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, $CellContext`d, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "d"},
           {"b", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`b, $CellContext`f, -$CellContext`e, \
$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"d", "c", "f", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, -$CellContext`c, -$CellContext`f, \
-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "f", " ", "d"},
           {"a", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, $CellContext`f, -$CellContext`e, \
$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"d", "c", "f", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, -$CellContext`c, -$CellContext`f, \
-$CellContext`b],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "d", " ", "f"},
           {"c", " ", "e", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`c, $CellContext`d, -$CellContext`e, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"d", "f", "a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`d, -$CellContext`f, -$CellContext`a, \
-$CellContext`b],
    Editable->False]}], "-", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`a][
     $CellContext`RicciCD[-$CellContext`c, -$CellContext`b]]],
   Editable->False], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SubscriptBox["\[EmptyDownTriangle]", "e"], 
     SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[-$CellContext`e][
    $CellContext`CD[-$CellContext`b][
     $CellContext`RicciCD[-$CellContext`c, -$CellContext`a]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978642280842*^9},
 CellLabel->"Out[1125]=",
 CellID->1569544593,ExpressionUUID->"9c4b4b55-c0c9-4943-88e1-b9735fb47012"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Fourth derivative of the Ricci scalar: conversion to Box^2:", \
"MathCaption",ExpressionUUID->"0f8d7d58-88e1-4755-b542-e667865c995c"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD4RicciScalar", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"CD", "[", "b", "]"}], "@", 
        RowBox[{
         RowBox[{"CD", "[", "a", "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "b"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "a"}], "]"}], "@", 
           RowBox[{"RicciScalarCD", "[", "]"}]}]}]}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"1", "/", "2"}], 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h1"}], "]"}], "@", 
          RowBox[{"RicciScalarCD", "[", "]"}]}], 
         RowBox[{
          RowBox[{"CD", "[", "h1", "]"}], "@", 
          RowBox[{"RicciScalarCD", "[", "]"}]}]}], "+", 
        RowBox[{
         RowBox[{"CD", "[", "h2", "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h2"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", "h1", "]"}], "@", 
           RowBox[{
            RowBox[{"CD", "[", 
             RowBox[{"-", "h1"}], "]"}], "@", 
            RowBox[{"RicciScalarCD", "[", "]"}]}]}]}]}], "+", 
        RowBox[{
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h2"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "h1"}], "]"}], "@", 
           RowBox[{"RicciScalarCD", "[", "]"}]}]}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{"h1", ",", "h2"}], "]"}]}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1126]:=",ExpressionUUID->"971a69f8-62b1-4bfd-a956-9d2d5dcb720f"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule24:", "MathCaption",ExpressionUUID->"be1e9a6d-d9f2-412d-aa35-0b2d7959af8e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "b", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", "a", "]"}], "@", 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "b"}], "]"}], "@", 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "a"}], "]"}], "@", 
      RowBox[{"RicciScalarCD", "[", "]"}]}]}]}]}], "/.", 
  "RuleD4RicciScalar"}]], "Input",
 CellLabel->
  "In[1127]:=",ExpressionUUID->"3c9eea6f-14fe-4db9-863d-06a9fc9f1793"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`RicciScalarCD[]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SuperscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[$CellContext`a][
      $CellContext`RicciScalarCD[]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`b][
      $CellContext`CD[-$CellContext`a][
       $CellContext`RicciScalarCD[]]],
     Editable->False], ")"}]}], "+", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     SuperscriptBox["\[EmptyDownTriangle]", "b"], 
     SubscriptBox["\[EmptyDownTriangle]", "b"], 
     SuperscriptBox["\[EmptyDownTriangle]", "a"], 
     SubscriptBox["\[EmptyDownTriangle]", "a"], 
     RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`CD[$CellContext`b][
    $CellContext`CD[-$CellContext`b][
     $CellContext`CD[$CellContext`a][
      $CellContext`CD[-$CellContext`a][
       $CellContext`RicciScalarCD[]]]]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978644464241*^9},
 CellLabel->"Out[1127]=",
 CellID->1876290669,ExpressionUUID->"adf8399e-562b-4d11-bcd7-8fd3ce4f0591"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Fourth derivative of the Ricci tensor:", "MathCaption",ExpressionUUID->"589c8697-e474-4b98-9e9b-c33e98d6959f"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD4Ricci", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"CD", "[", "b", "]"}], "@", 
        RowBox[{
         RowBox[{"CD", "[", "a", "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", "c", "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "c"}], "]"}], "@", 
           RowBox[{"RicciCD", "[", 
            RowBox[{
             RowBox[{"-", "a"}], ",", 
             RowBox[{"-", "b"}]}], "]"}]}]}]}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"1", "/", "2"}], 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h1"}], "]"}], "@", 
          RowBox[{"RicciScalarCD", "[", "]"}]}], 
         RowBox[{
          RowBox[{"CD", "[", "h1", "]"}], "@", 
          RowBox[{"RicciScalarCD", "[", "]"}]}]}], "-", 
        RowBox[{"3", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h3"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "h1"}], ",", 
            RowBox[{"-", "h2"}]}], "]"}]}], 
         RowBox[{
          RowBox[{"CD", "[", "h3", "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{"h1", ",", "h2"}], "]"}]}]}], "+", 
        RowBox[{"4", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h3"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "h1"}], ",", 
            RowBox[{"-", "h2"}]}], "]"}]}], 
         RowBox[{
          RowBox[{"CD", "[", "h2", "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{"h1", ",", "h3"}], "]"}]}]}], "+", 
        RowBox[{
         RowBox[{"1", "/", "2"}], 
         RowBox[{
          RowBox[{"CD", "[", "h2", "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "h2"}], "]"}], "@", 
           RowBox[{
            RowBox[{"CD", "[", "h1", "]"}], "@", 
            RowBox[{
             RowBox[{"CD", "[", 
              RowBox[{"-", "h1"}], "]"}], "@", 
             RowBox[{"RicciScalarCD", "[", "]"}]}]}]}]}]}], "+", 
        RowBox[{"2", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h2"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "h1"}], "]"}], "@", 
           RowBox[{"RicciScalarCD", "[", "]"}]}]}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{"h1", ",", "h2"}], "]"}]}], "-", 
        RowBox[{
         RowBox[{
          RowBox[{"CD", "[", "h3", "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "h3"}], "]"}], "@", 
           RowBox[{"RicciCD", "[", 
            RowBox[{
             RowBox[{"-", "h1"}], ",", 
             RowBox[{"-", "h2"}]}], "]"}]}]}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{"h1", ",", "h2"}], "]"}]}], "+", 
        RowBox[{"2", 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}]}], "]"}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h3"}], ",", "h1"}], "]"}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{"h2", ",", "h3"}], "]"}]}], "-", 
        RowBox[{"2", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h4"}], "]"}], "@", 
          RowBox[{
           RowBox[{"CD", "[", 
            RowBox[{"-", "h3"}], "]"}], "@", 
           RowBox[{"RicciCD", "[", 
            RowBox[{
             RowBox[{"-", "h1"}], ",", 
             RowBox[{"-", "h2"}]}], "]"}]}]}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "h3", ",", "h2", ",", "h4"}], "]"}]}], "-", 
        RowBox[{"2", 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}]}], "]"}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h3"}], ",", 
           RowBox[{"-", "h4"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "h3", ",", "h2", ",", "h4"}], "]"}]}]}]}], "}"}],
      ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1128]:=",ExpressionUUID->"0ee2cb12-1d86-485c-8590-a7a0360f6844"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule25:", "MathCaption",ExpressionUUID->"f7c99a1e-c733-44a8-9712-f4c390255c50"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"CD", "[", "b", "]"}], "@", 
   RowBox[{
    RowBox[{"CD", "[", "a", "]"}], "@", 
    RowBox[{
     RowBox[{"CD", "[", "c", "]"}], "@", 
     RowBox[{
      RowBox[{"CD", "[", 
       RowBox[{"-", "c"}], "]"}], "@", 
      RowBox[{"RicciCD", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}]}], "]"}]}]}]}]}], "/.", "RuleD4Ricci"}]], "Input",
 CellLabel->
  "In[1129]:=",ExpressionUUID->"36b5760a-060c-460a-8bcb-0f4bf491a64e"],

Cell[BoxData[
 RowBox[{
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"b", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[$CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "a"},
           {"c", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, $CellContext`a],
    Editable->False]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " "},
           {"c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[-$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", "c", "b", "d"},
           {" ", " ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[$CellContext`a, $CellContext`c, $CellContext`b, \
$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`a][
      $CellContext`RicciScalarCD[]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SuperscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[$CellContext`a][
      $CellContext`RicciScalarCD[]],
     Editable->False], ")"}]}], "+", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "b"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`b][
      $CellContext`CD[-$CellContext`a][
       $CellContext`RicciScalarCD[]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SuperscriptBox["\[EmptyDownTriangle]", "b"], 
       SubscriptBox["\[EmptyDownTriangle]", "b"], 
       SuperscriptBox["\[EmptyDownTriangle]", "a"], 
       SubscriptBox["\[EmptyDownTriangle]", "a"], 
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[$CellContext`b][
      $CellContext`CD[-$CellContext`b][
       $CellContext`CD[$CellContext`a][
        $CellContext`CD[-$CellContext`a][
         $CellContext`RicciScalarCD[]]]]],
     Editable->False], ")"}]}], "+", 
  RowBox[{"4", " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SuperscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"a", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[$CellContext`b][
      $CellContext`RicciCD[$CellContext`a, $CellContext`c]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"3", " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`c][
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]],
     Editable->False], ")"}], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SuperscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {"a", "b"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[$CellContext`c][
      $CellContext`RicciCD[$CellContext`a, $CellContext`b]],
     Editable->False], ")"}]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RicciCD[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SuperscriptBox["\[EmptyDownTriangle]", "c"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[$CellContext`c][
      $CellContext`CD[-$CellContext`c][
       $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]]],
     Editable->False], ")"}]}], "-", 
  RowBox[{"2", " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"a", "c", "b", "d"},
           {" ", " ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[$CellContext`a, $CellContext`c, $CellContext`b, \
$CellContext`d],
    Editable->False], " ", 
   RowBox[{"(", 
    InterpretationBox[
     StyleBox[
      RowBox[{
       SubscriptBox["\[EmptyDownTriangle]", "d"], 
       SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
         {
          RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
          StyleBox[GridBox[{
             {" ", " "},
             {"a", "b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`CD[-$CellContext`d][
      $CellContext`CD[-$CellContext`c][
       $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]]],
     Editable->False], ")"}]}]}]], "Output",
 CellChangeTimes->{3.7289786467448883`*^9},
 CellLabel->"Out[1129]=",
 CellID->1823700841,ExpressionUUID->"b4930427-e77c-4d6a-812b-78b872a94332"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Now we study six rules with products of two Riemann tensors on the LHS. \
There are only two different rules, depending on whether there are two or \
three contracted indices:\n\t",
 StyleBox["RuleTwoRiemann3",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleTwoRiemann2",
  FontFamily->"Courier"]
}], "Text",ExpressionUUID->"6f2600a8-ab19-4a2c-99de-d1ac3e09fa72"],

Cell[CellGroupData[{

Cell["These are the independent rules:", "MathCaption",ExpressionUUID->"5280a227-e9bb-4280-b021-3baac7b7ba85"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleTwoRiemann3", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "e"}], ",", "a", ",", "b", ",", "c"}], "]"}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "1"}], "/", "2"}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "f"}], ",", 
          RowBox[{"-", "d"}], ",", 
          RowBox[{"-", "g"}], ",", 
          RowBox[{"-", "h"}]}], "]"}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "e"}], ",", "f", ",", "g", ",", "h"}], "]"}]}]}], 
      "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1130]:=",ExpressionUUID->"81efa92c-8b32-4f7e-a4c6-fef5b91df159"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleTwoRiemann2", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "e"}], ",", "a", ",", 
          RowBox[{"-", "g"}], ",", "b"}], "]"}]}], ",", 
       RowBox[{
        RowBox[{"1", "/", "2"}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "f"}], ",", 
          RowBox[{"-", "h"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "e"}], ",", 
          RowBox[{"-", "g"}], ",", "f", ",", "h"}], "]"}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1131]:=",ExpressionUUID->"b86aa0b6-822e-499c-921a-01fc7651f161"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule7:", "MathCaption",ExpressionUUID->"836e9b5e-c49a-4700-a3b6-5b51905ce956"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", "a", ",", "b", ",", "c"}], "]"}]}], "/.", 
  "RuleTwoRiemann3"}]], "Input",
 CellLabel->
  "In[1132]:=",ExpressionUUID->"b06ac3ad-6a4a-4ecf-a231-a39b19dd0644"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "d", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, -$CellContext`b, \
-$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", "a", "b", "c"},
          {"e", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, $CellContext`a, $CellContext`b, \
$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978651775392*^9},
 CellLabel->"Out[1132]=",
 CellID->2028499984,ExpressionUUID->"de152a61-edbd-4862-818e-332087336192"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule8:", "MathCaption",ExpressionUUID->"38ce75ba-8373-4f24-b666-8825427efe65"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"a", ",", "c", ",", "b", ",", "h"}], "]"}]}], "/.", 
  "RuleTwoRiemann3"}]], "Input",
 CellLabel->
  "In[1133]:=",ExpressionUUID->"066f00de-648d-42b2-8959-15505afd3707"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"h", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`h, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "d", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, -$CellContext`b, \
-$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", "a", "b", "c"},
          {"e", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, $CellContext`a, $CellContext`b, \
$CellContext`c],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978653248361*^9},
 CellLabel->"Out[1133]=",
 CellID->796341276,ExpressionUUID->"257987bc-d1fd-40a4-a9a2-8b3524d45724"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule9a:", "MathCaption",ExpressionUUID->"42a0654a-d004-48ab-8354-38b4344989d3"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", "a", ",", 
     RowBox[{"-", "g"}], ",", "b"}], "]"}]}], "/.", 
  "RuleTwoRiemann2"}]], "Input",
 CellLabel->
  "In[1134]:=",ExpressionUUID->"910fcd2d-21b0-4eee-9593-95383d0d5f75"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "a", "b"},
          {"e", "g", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`g, $CellContext`a, \
$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978654365656*^9},
 CellLabel->"Out[1134]=",
 CellID->1301679146,ExpressionUUID->"0c19a7ca-d787-46dc-b4a0-b1f0b2b146c6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule9b:", "MathCaption",ExpressionUUID->"edecdc48-9fa5-49ab-9387-eef1704699ba"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"e", ",", "a", ",", "b", ",", "h"}], "]"}]}], "/.", 
  "RuleTwoRiemann2"}]], "Input",
 CellLabel->
  "In[1135]:=",ExpressionUUID->"3c7f0957-53c9-49d1-9727-2f2b92defe43"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"e", "f"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`e, $CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"h", "g"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`h, $CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "a", "b"},
          {"f", "g", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`f, -$CellContext`g, $CellContext`a, \
$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978655444376*^9},
 CellLabel->"Out[1135]=",
 CellID->923999403,ExpressionUUID->"a4199c44-4235-4178-8525-1f44f26e634a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule9c:", "MathCaption",ExpressionUUID->"cd371940-b188-4dc2-a8c6-83b0ba2cce0b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"a", ",", "f", ",", "b", ",", "h"}], "]"}]}], "/.", 
  "RuleTwoRiemann2"}]], "Input",
 CellLabel->
  "In[1136]:=",ExpressionUUID->"3d209a27-fe0a-4e60-98f2-bdab7e7d28db"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"f", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`f, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"h", "g"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`h, $CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "a", "b"},
          {"e", "g", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`g, $CellContext`a, \
$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.72897865650949*^9},
 CellLabel->"Out[1136]=",
 CellID->972971938,ExpressionUUID->"f8acad0d-203a-480c-a051-d728e085c2ed"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule9d:", "MathCaption",ExpressionUUID->"e625fea5-2a1d-4aeb-958e-5d551138808d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"a", ",", "f", ",", "g", ",", "b"}], "]"}]}], "/.", 
  "RuleTwoRiemann2"}]], "Input",
 CellLabel->
  "In[1137]:=",ExpressionUUID->"b822f7ad-340b-4c75-a417-6a298f07f330"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"f", "e"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`f, $CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"g", "h"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`g, $CellContext`h],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "a", "b"},
          {"e", "h", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`h, $CellContext`a, \
$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978657615779*^9},
 CellLabel->"Out[1137]=",
 CellID->994892794,ExpressionUUID->"e03eba45-32f3-451c-8aa8-ea335506a20a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There are seven more rules involving two curvature tensors and derivatives, \
corresponding to five independent rules:\n\t",
 StyleBox["RuleD01TwoRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleD11TwoRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleD2RicciRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleD2RicciDRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleD11bisTwoRiemann",
  FontFamily->"Courier"]
}], "Text",ExpressionUUID->"f8f9e044-f2ef-4a5f-af23-df639eab60a5"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD01TwoRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"a", ",", "f", ",", "b", ",", "c"}], "]"}]}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "a"}], ",", 
          RowBox[{"-", "b"}], ",", 
          RowBox[{"-", "c"}], ",", 
          RowBox[{"-", "d"}]}], "]"}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "1"}], "/", "2"}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "f", ",", "h2", ",", "h3"}], "]"}]}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{
          RowBox[{"-", "h1"}], ",", 
          RowBox[{"-", "d"}], ",", 
          RowBox[{"-", "h2"}], ",", 
          RowBox[{"-", "h3"}]}], "]"}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1138]:=",ExpressionUUID->"36d58aff-1e85-4d69-98e7-72761b62c67b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD11TwoRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "f"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"a", ",", "g", ",", "b", ",", "c"}], "]"}]}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"-", "1"}], "/", "2"}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "d"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "h3"}]}], "]"}]}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "f"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "g", ",", "h2", ",", "h3"}], "]"}]}]}]}], "}"}], 
     ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1139]:=",ExpressionUUID->"52ddb2cb-ad8b-496f-9e7c-ee63772b6f36"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD2RicciRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "d"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "c"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "a"}], ",", 
            RowBox[{"-", "b"}]}], "]"}]}]}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{"a", ",", "d", ",", "b", ",", "c"}], "]"}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "h4"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h3"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "h1"}], ",", 
            RowBox[{"-", "h2"}]}], "]"}]}]}], 
        RowBox[{"RiemannCD", "[", 
         RowBox[{"h1", ",", "h3", ",", "h2", ",", "h4"}], "]"}]}]}], "}"}], 
     ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1140]:=",ExpressionUUID->"676d0a2b-4311-4a33-97e0-fc03a2fb4ccb"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD2RicciDRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"a", ",", "d", ",", "b", ",", "c"}], "]"}]}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "d"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "c"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "a"}], ",", 
            RowBox[{"-", "b"}]}], "]"}]}]}]}], ",", 
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "e"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "h2", ",", "h3", ",", "h4"}], "]"}]}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "h4"}], "]"}], "@", 
         RowBox[{
          RowBox[{"CD", "[", 
           RowBox[{"-", "h2"}], "]"}], "@", 
          RowBox[{"RicciCD", "[", 
           RowBox[{
            RowBox[{"-", "h1"}], ",", 
            RowBox[{"-", "h3"}]}], "]"}]}]}]}]}], "}"}], ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1141]:=",ExpressionUUID->"655d01bb-0e02-42de-8498-3d9b8032aa64"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This rule has a typo in MathTensor's book. I assume the index c on the RHS is \
actually f:\
\>", "MathCaption",ExpressionUUID->"0c2844a3-1270-4e5c-88a3-395a08ec8aef"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RuleD11bisTwoRiemann", "=", 
   RowBox[{"MakeRule", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "g"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}]}], 
        RowBox[{
         RowBox[{"CD", "[", "g", "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"c", ",", "e", ",", "d", ",", "f"}], "]"}]}]}], ",", 
       RowBox[{
        RowBox[{"1", "/", "2"}], 
        RowBox[{
         RowBox[{"CD", "[", 
          RowBox[{"-", "h3"}], "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}]}], "]"}]}], 
        RowBox[{
         RowBox[{"CD", "[", "h3", "]"}], "@", 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "h2", ",", "f", ",", "e"}], "]"}]}]}]}], "}"}], 
     ",", 
     RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
     RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], ";"}]], "Input",
 CellLabel->
  "In[1142]:=",ExpressionUUID->"ff523cb0-aad5-4fa9-b08b-ba025f2ceb3a"]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule34:", "MathCaption",ExpressionUUID->"947ebd8a-9958-4318-8c54-79b8f87753e4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{"a", ",", "f", ",", "b", ",", "c"}], "]"}]}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "/.", "RuleD01TwoRiemann"}]], "Input",
 CellLabel->
  "In[1143]:=",ExpressionUUID->"b398ab77-99d7-42e9-a9d2-39e3227a46e7"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "d", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, -$CellContext`b, \
-$CellContext`c],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", "f", "b", "c"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`RiemannCD[$CellContext`a, $CellContext`f, $CellContext`b, \
$CellContext`c]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.72897866642899*^9},
 CellLabel->"Out[1143]=",
 CellID->1029087669,ExpressionUUID->"371dcb9a-4409-42ca-87c5-d2508c9472d6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule35:", "MathCaption",ExpressionUUID->"caa74a4b-ffd6-46fe-b73f-43f3b461bf91"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "f"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{"a", ",", "g", ",", "b", ",", "c"}], "]"}]}]}], "/.", 
  "RuleD11TwoRiemann"}]], "Input",
 CellLabel->
  "In[1144]:=",ExpressionUUID->"1f1707e0-ab8a-450e-bb91-f77199e2633d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "d", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, \
-$CellContext`b, -$CellContext`c]],
    Editable->False], ")"}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", "g", "b", "c"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`f][
     $CellContext`RiemannCD[$CellContext`a, $CellContext`g, $CellContext`b, \
$CellContext`c]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.7289786667169123`*^9},
 CellLabel->"Out[1144]=",
 CellID->1506847844,ExpressionUUID->"39c62921-40e2-4509-bee9-977e51e4a919"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule36:", "MathCaption",ExpressionUUID->"45307c60-1c5b-49d0-ae95-d597762c20d8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{"a", ",", "c", ",", "b", ",", "f"}], "]"}]}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "/.", "RuleD01TwoRiemann"}]], "Input",
 CellLabel->
  "In[1145]:=",ExpressionUUID->"c23224a3-96a8-4224-a7f8-6384c4a37421"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "d", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, -$CellContext`b, \
-$CellContext`c],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", "f", "b", "c"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`RiemannCD[$CellContext`a, $CellContext`f, $CellContext`b, \
$CellContext`c]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978667745606*^9},
 CellLabel->"Out[1145]=",
 CellID->1228034384,ExpressionUUID->"6da631b7-ec8d-4c0d-aa5a-0ae2d0408141"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule37:", "MathCaption",ExpressionUUID->"56919c3d-4074-46d3-b6db-e55f26f68f83"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "f"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{"a", ",", "c", ",", "b", ",", "g"}], "]"}]}]}], "/.", 
  "RuleD11TwoRiemann"}]], "Input",
 CellLabel->
  "In[1146]:=",ExpressionUUID->"893f5d18-50db-4a58-a915-698351501c31"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"a", "d", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`RiemannCD[-$CellContext`a, -$CellContext`d, \
-$CellContext`b, -$CellContext`c]],
    Editable->False], ")"}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "f"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", "g", "b", "c"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`f][
     $CellContext`RiemannCD[$CellContext`a, $CellContext`g, $CellContext`b, \
$CellContext`c]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978668746291*^9},
 CellLabel->"Out[1146]=",
 CellID->633460511,ExpressionUUID->"9008fe4a-0652-45e3-a447-8ec5e4a49b9b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule38:", "MathCaption",ExpressionUUID->"cd6ea944-b3dc-49f0-8181-264fe367e377"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "g"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}], ",", 
      RowBox[{"-", "d"}]}], "]"}]}], 
   RowBox[{
    RowBox[{"CD", "[", "g", "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{"c", ",", "e", ",", "d", ",", "f"}], "]"}]}]}], "/.", 
  "RuleD11bisTwoRiemann"}]], "Input",
 CellLabel->
  "In[1147]:=",ExpressionUUID->"cd391048-ae43-46ef-8720-9e6cde9b22bd"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "g"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " ", " ", " "},
            {"c", "d", "a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`g][
     $CellContext`RiemannCD[-$CellContext`c, -$CellContext`d, \
-$CellContext`a, -$CellContext`b]],
    Editable->False], ")"}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SuperscriptBox["\[EmptyDownTriangle]", "g"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"c", "d", "f", "e"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[$CellContext`g][
     $CellContext`RiemannCD[$CellContext`c, $CellContext`d, $CellContext`f, \
$CellContext`e]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978670545567*^9},
 CellLabel->"Out[1147]=",
 CellID->1657064808,ExpressionUUID->"c888b4d7-dfad-4ba8-99a3-5848d8d37a69"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule39:", "MathCaption",ExpressionUUID->"53913b63-4165-453d-bb07-afb563b664cc"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "@", 
    RowBox[{"RiemannCD", "[", 
     RowBox[{"a", ",", "d", ",", "b", ",", "c"}], "]"}]}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "d"}], "]"}], "@", 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "c"}], "]"}], "@", 
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}]}]}]}], "/.", 
  "RuleD2RicciDRiemann"}]], "Input",
 CellLabel->
  "In[1148]:=",ExpressionUUID->"c90c3246-5928-45d4-bbdb-1b2a531b7b5b"],

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], 
      SubscriptBox["\[EmptyDownTriangle]", "b"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`CD[-$CellContext`b][
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`c]]],
    Editable->False], ")"}], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {"a", "b", "c", "d"},
            {" ", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`RiemannCD[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978671800338*^9},
 CellLabel->"Out[1148]=",
 CellID->671595088,ExpressionUUID->"e9e450ac-e953-4a04-a54c-456a0e29b240"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule40:", "MathCaption",ExpressionUUID->"485731cf-0096-47fb-b1ae-9ab32a555558"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "d"}], "]"}], "@", 
    RowBox[{
     RowBox[{"CD", "[", 
      RowBox[{"-", "c"}], "]"}], "@", 
     RowBox[{"RicciCD", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}]}], "]"}]}]}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"a", ",", "d", ",", "b", ",", "c"}], "]"}]}], "/.", 
  "RuleD2RicciRiemann"}]], "Input",
 CellLabel->
  "In[1149]:=",ExpressionUUID->"7269d999-b525-4012-a6ed-474aee69ecab"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"a", "c", "b", "d"},
          {" ", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[$CellContext`a, $CellContext`c, $CellContext`b, \
$CellContext`d],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "d"], 
      SubscriptBox["\[EmptyDownTriangle]", "c"], GridBox[{
        {
         RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`d][
     $CellContext`CD[-$CellContext`c][
      $CellContext`RicciCD[-$CellContext`a, -$CellContext`b]]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978673014798*^9},
 CellLabel->"Out[1149]=",
 CellID->979087761,ExpressionUUID->"2cd3ca1c-eed6-42ac-948a-4e6e831ccc21"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Finally we study six rules with products of three curvature tensors, only \
four of which are independent:\n\t",
 StyleBox["RuleRicciTwoRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleThreeRiemann",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleThreeRiemannB",
  FontFamily->"Courier"],
 "\n\t",
 StyleBox["RuleThreeRiemannC",
  FontFamily->"Courier"]
}], "Text",ExpressionUUID->"7555d3e2-f9e0-4a0f-aa21-723e4d78252d"],

Cell[CellGroupData[{

Cell["\<\
Constructing the independent rules takes rather long because there are lots \
of equivalent cases:\
\>", "MathCaption",ExpressionUUID->"76ae39a2-d747-4d93-8296-51d766975932"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"Length", "[", 
   RowBox[{"RuleRicciTwoRiemann", "=", 
    RowBox[{"MakeRule", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}], ",", 
           RowBox[{"-", "e"}], ",", "a"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"b", ",", "c", ",", "d", ",", "e"}], "]"}]}], ",", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "1"}], "/", "2"}], 
         RowBox[{"RicciCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h3"}], ",", 
           RowBox[{"-", "h4"}], ",", 
           RowBox[{"-", "h5"}], ",", "h1"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h2", ",", "h5", ",", "h3", ",", "h4"}], "]"}]}]}], "}"}], 
      ",", 
      RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
      RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[1166]:=",ExpressionUUID->"3e935a59-ad9b-44dd-971e-04a65e62707f"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"6.063203`", ",", "4096"}], "}"}]], "Output",
 CellLabel->
  "Out[1166]=",ExpressionUUID->"b5f60cc0-2d2d-4318-b6f6-36a74c75f733"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"Length", "[", 
   RowBox[{"RuleThreeRiemann", "=", 
    RowBox[{"MakeRule", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "e"}], ",", 
           RowBox[{"-", "f"}], ",", "a", ",", "b"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"c", ",", "e", ",", "d", ",", "f"}], "]"}]}], ",", 
        RowBox[{
         RowBox[{"1", "/", "2"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "h3"}], ",", 
           RowBox[{"-", "h4"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h5"}], ",", 
           RowBox[{"-", "h6"}], ",", "h1", ",", "h2"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h3", ",", "h4", ",", "h5", ",", "h6"}], "]"}]}]}], "}"}], 
      ",", 
      RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
      RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[1167]:=",ExpressionUUID->"40be1608-5050-443f-82ac-1e2a46a67621"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"63.885789`", ",", "32768"}], "}"}]], "Output",
 CellLabel->
  "Out[1167]=",ExpressionUUID->"5da3050a-a247-416c-b405-7457ca3c757e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"Length", "[", 
   RowBox[{"RuleThreeRiemannB", "=", 
    RowBox[{"MakeRule", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "e"}], ",", "a", ",", 
           RowBox[{"-", "f"}], ",", "b"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"c", ",", "e", ",", "d", ",", "f"}], "]"}]}], ",", 
        RowBox[{
         RowBox[{"1", "/", "4"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h1"}], ",", 
           RowBox[{"-", "h2"}], ",", 
           RowBox[{"-", "h3"}], ",", 
           RowBox[{"-", "h4"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "h5"}], ",", 
           RowBox[{"-", "h6"}], ",", "h3", ",", "h4"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"h1", ",", "h2", ",", "h5", ",", "h6"}], "]"}]}]}], "}"}], 
      ",", 
      RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
      RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[1168]:=",ExpressionUUID->"0af9875b-f23d-47a0-b3d0-e5dfd22227b3"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"63.892907`", ",", "32768"}], "}"}]], "Output",
 CellLabel->
  "Out[1168]=",ExpressionUUID->"ddc25b93-f0ad-488b-9e19-05695857eb03"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"AbsoluteTiming", "[", 
  RowBox[{"Length", "[", 
   RowBox[{"RuleThreeRiemannC", "=", 
    RowBox[{"MakeRule", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "a"}], ",", 
           RowBox[{"-", "b"}], ",", 
           RowBox[{"-", "c"}], ",", 
           RowBox[{"-", "d"}]}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{
           RowBox[{"-", "e"}], ",", "a", ",", 
           RowBox[{"-", "f"}], ",", "c"}], "]"}], 
         RowBox[{"RiemannCD", "[", 
          RowBox[{"b", ",", "f", ",", "d", ",", "e"}], "]"}]}], ",", 
        RowBox[{
         RowBox[{
          RowBox[{"RiemannCD", "[", 
           RowBox[{
            RowBox[{"-", "h1"}], ",", 
            RowBox[{"-", "h2"}], ",", 
            RowBox[{"-", "h3"}], ",", 
            RowBox[{"-", "h4"}]}], "]"}], 
          RowBox[{"RiemannCD", "[", 
           RowBox[{
            RowBox[{"-", "h5"}], ",", "h1", ",", 
            RowBox[{"-", "h6"}], ",", "h3"}], "]"}], 
          RowBox[{"RiemannCD", "[", 
           RowBox[{"h2", ",", "h5", ",", "h4", ",", "h6"}], "]"}]}], "-", 
         RowBox[{
          RowBox[{"1", "/", "4"}], 
          RowBox[{"RiemannCD", "[", 
           RowBox[{
            RowBox[{"-", "h1"}], ",", 
            RowBox[{"-", "h2"}], ",", 
            RowBox[{"-", "h3"}], ",", 
            RowBox[{"-", "h4"}]}], "]"}], 
          RowBox[{"RiemannCD", "[", 
           RowBox[{
            RowBox[{"-", "h5"}], ",", 
            RowBox[{"-", "h6"}], ",", "h1", ",", "h2"}], "]"}], 
          RowBox[{"RiemannCD", "[", 
           RowBox[{"h3", ",", "h4", ",", "h5", ",", "h6"}], "]"}]}]}]}], 
       "}"}], ",", 
      RowBox[{"UseSymmetries", "\[Rule]", "True"}], ",", 
      RowBox[{"MetricOn", "\[Rule]", "All"}]}], "]"}]}], "]"}], 
  "]"}]], "Input",
 CellLabel->
  "In[1169]:=",ExpressionUUID->"a96214c7-3eef-4080-8319-3a13900a07cd"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"103.372059`", ",", "32768"}], "}"}]], "Output",
 CellLabel->
  "Out[1169]=",ExpressionUUID->"729cf3a5-fd36-4e58-9a7f-e045ac5e4e61"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule10:", "MathCaption",ExpressionUUID->"d26156d6-589a-43eb-a820-c8e8b584bb73"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RicciCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}], ",", "a"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"b", ",", "c", ",", "d", ",", "e"}], "]"}]}], "/.", 
  "RuleRicciTwoRiemann"}]], "Input",
 CellLabel->
  "In[1170]:=",ExpressionUUID->"309a3465-1ffd-4c73-9832-446ea85a7c45"],

Cell[BoxData[
 RowBox[{
  RowBox[{"-", 
   FractionBox["1", "2"]}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RicciCD[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"b", "e", "c", "d"},
          {" ", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[$CellContext`b, $CellContext`e, $CellContext`c, \
$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", "a"},
          {"c", "d", "e", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`c, -$CellContext`d, -$CellContext`e, \
$CellContext`a],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[1170]=",ExpressionUUID->"22bbefd4-025a-4433-83d4-13dff87bba91"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule11:", "MathCaption",ExpressionUUID->"debd4b8d-e108-4b0b-9c68-c73f9ffbae3a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", 
     RowBox[{"-", "f"}], ",", "a", ",", "b"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"c", ",", "e", ",", "d", ",", "f"}], "]"}]}], "/.", 
  "RuleThreeRiemann"}]], "Input",
 CellLabel->
  "In[1171]:=",ExpressionUUID->"7c3d8ca6-e7e8-4793-9c43-a10744069eca"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"c", "d", "e", "f"},
          {" ", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[$CellContext`c, $CellContext`d, $CellContext`e, \
$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "a", "b"},
          {"e", "f", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, $CellContext`a, \
$CellContext`b],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[1171]=",ExpressionUUID->"bb1acacb-e689-4b9b-9450-ea49182e60b1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule12:", "MathCaption",ExpressionUUID->"3c9902a2-c109-4b53-806d-e23ca248f7f1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", "a", ",", 
     RowBox[{"-", "f"}], ",", "b"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"c", ",", "e", ",", "d", ",", "f"}], "]"}]}], "/.", 
  "RuleThreeRiemannB"}]], "Input",
 CellLabel->
  "In[1172]:=",ExpressionUUID->"e9d30a21-72e6-4a9e-bbbb-8533e08884c7"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"a", "b", "e", "f"},
          {" ", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[$CellContext`a, $CellContext`b, $CellContext`e, \
$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "c", "d"},
          {"e", "f", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, $CellContext`c, \
$CellContext`d],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[1172]=",ExpressionUUID->"607d087e-4798-4f22-baff-b17dc18481e2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule13:", "MathCaption",ExpressionUUID->"c3a4c268-0329-407b-94ff-bd8e08d13930"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", "a", ",", 
     RowBox[{"-", "f"}], ",", "b"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"c", ",", "d", ",", "e", ",", "f"}], "]"}]}], "/.", 
  "RuleThreeRiemann"}]], "Input",
 CellLabel->
  "In[1173]:=",ExpressionUUID->"bfe655b8-47fc-48dd-9560-8b9e0398eec8"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"c", "d", "e", "f"},
          {" ", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[$CellContext`c, $CellContext`d, $CellContext`e, \
$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "a", "b"},
          {"e", "f", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, $CellContext`a, \
$CellContext`b],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[1173]=",ExpressionUUID->"87d05fd6-2d7f-4b93-8b4f-c09f0b9fc22a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule14:", "MathCaption",ExpressionUUID->"e9d3cd67-115d-486f-8f85-eb471465433d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", "a", ",", 
     RowBox[{"-", "f"}], ",", "c"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"b", ",", "d", ",", "e", ",", "f"}], "]"}]}], "/.", 
  "RuleThreeRiemannB"}]], "Input",
 CellLabel->
  "In[1174]:=",ExpressionUUID->"b2f09501-e040-439e-bcae-c774f8af7fd4"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "4"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", " ", " "},
          {"a", "b", "c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {"a", "b", "e", "f"},
          {" ", " ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[$CellContext`a, $CellContext`b, $CellContext`e, \
$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {
       RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
       StyleBox[GridBox[{
          {" ", " ", "c", "d"},
          {"e", "f", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, $CellContext`c, \
$CellContext`d],
   Editable->False]}]], "Output",
 CellLabel->
  "Out[1174]=",ExpressionUUID->"27c52738-aba9-47be-8a6c-0500f4e60223"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["RiemannRule15:", "MathCaption",ExpressionUUID->"aa14db38-9a1b-491d-99d0-918611075d4c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{
     RowBox[{"-", "e"}], ",", "a", ",", 
     RowBox[{"-", "f"}], ",", "c"}], "]"}], 
   RowBox[{"RiemannCD", "[", 
    RowBox[{"b", ",", "f", ",", "d", ",", "e"}], "]"}]}], "/.", 
  "RuleThreeRiemannC"}]], "Input",
 CellLabel->
  "In[1175]:=",ExpressionUUID->"8fb2918c-8f5a-4646-b7bc-577d4d69ffda"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"b", "e", "d", "f"},
           {" ", " ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[$CellContext`b, $CellContext`e, $CellContext`d, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", "a", " ", "c"},
           {"e", " ", "f", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, $CellContext`a, -$CellContext`f, \
$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "4"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", " ", " "},
           {"a", "b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`a, -$CellContext`b, -$CellContext`c, \
-$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {"c", "d", "e", "f"},
           {" ", " ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[$CellContext`c, $CellContext`d, $CellContext`e, \
$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {
        RowBox[{"R", "[", "\[EmptyDownTriangle]", "]"}], 
        StyleBox[GridBox[{
           {" ", " ", "a", "b"},
           {"e", "f", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`RiemannCD[-$CellContext`e, -$CellContext`f, $CellContext`a, \
$CellContext`b],
    Editable->False]}]}]], "Output",
 CellLabel->
  "Out[1175]=",ExpressionUUID->"d378f07c-5895-47f1-931c-4c7f03cb19d9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Remove rules:", "MathCaption",ExpressionUUID->"ddc26dad-c949-49d0-9c60-f22636982523"],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{
  "RuleBoxRiemann", ",", "RuleD01TwoRiemann", ",", "RuleD11bisTwoRiemann", 
   ",", "RuleD11TwoRiemann", ",", "RuleD2RicciDRiemann", ",", 
   "RuleD2RicciRiemann", ",", "RuleD4Ricci", ",", "RuleD4RicciScalar", ",", 
   "RuleDivGradRicci", ",", "RuleDivGradRiemann", ",", "RuleDivRiemann", ",", 
   "RuleRicciTwoRiemann", ",", "RuleThreeRiemann", ",", "RuleThreeRiemannB", 
   ",", "RuleThreeRiemannC", ",", "RuleTwoRiemann2", ",", "RuleTwoRiemann3"}],
   "]"}]], "Input",
 CellLabel->
  "In[1160]:=",ExpressionUUID->"3589a430-9fb6-425a-9918-18589f632bbe"]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["10. Manipulation of expressions", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}, {3.728976670882564*^9, 
  3.728976674720446*^9}, {3.7289768454424953`*^9, 3.728976848920129*^9}, {
  3.728977115725004*^9, 3.7289771217954493`*^9}, {3.728977171203144*^9, 
  3.728977176113048*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->1149066679,ExpressionUUID->"4ad3e1c7-5134-49fb-8ac5-06e6eda5f7a3"],

Cell[CellGroupData[{

Cell["10.1. Find indices", "Subsection",ExpressionUUID->"4a1d9509-c2b0-464d-b189-034c5dce8f04"],

Cell[TextData[{
 "Given an expression, very often we want to extract all indices of the \
expression, or those of certain type, or character or state. We have already \
seen how to select indices of any type and any character. In this subsection \
we shall define functions to extract all indices according to a number of ` \
`selectors\.b4 \.b4.\nThere are two sets of functions. There is first the \
internal function ",
 StyleBox["FindIndices",
  FontFamily->"Courier"],
 ", which returns all indices of an expression, without first evaluating it, \
and its relatives ",
 StyleBox["FindFreeIndice",
  FontFamily->"Courier"],
 "s, ",
 StyleBox["FindDummyIndices",
  FontFamily->"Courier"],
 " and ",
 StyleBox["FindBlockedIndices",
  FontFamily->"Courier"],
 ". There is then the user driver for finding indices, called ",
 StyleBox["IndicesOf",
  FontFamily->"Courier"],
 ", which admits a number of criteria for index selection, and evaluates its \
input (as the user typically expects)."
}], "Text",ExpressionUUID->"ec90ca91-73e8-4982-b3ad-a87b6b269135"],

Cell[CellGroupData[{

Cell["Define some more objects:", "MathCaption",ExpressionUUID->"08b54f7b-2565-46fb-bd26-d3fbd3f10964"],

Cell[BoxData[{
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"r", "[", "]"}], ",", "M3"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"S", "[", 
    RowBox[{
     RowBox[{"LI", "[", "_", "]"}], ",", "a", ",", "b"}], "]"}], ",", "M3", 
   ",", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"{", 
     RowBox[{"2", ",", "3"}], "}"}], "]"}]}], "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"n", "[", "a", "]"}], ",", "M3"}], "]"}]}], "Input",
 CellLabel->
  "In[1161]:=",ExpressionUUID->"9d30b384-5571-4cee-93b7-75df564d881d"],

Cell[BoxData[
 RowBox[{
  RowBox[{"$DefInfoQ", "=", "True"}], ";"}]], "Input",
 CellLabel->
  "In[1164]:=",ExpressionUUID->"aeec4dee-b6df-4b4c-8726-1939e11fb6f9"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Let us start with this expression", "MathCaption",ExpressionUUID->"08a9687d-9bc4-4073-a0db-3c9ff2cbe796"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{"4", 
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"Dir", "[", 
      RowBox[{"v", "[", "d", "]"}], "]"}]}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{
    RowBox[{"r", "[", "]"}], "^", "2"}], 
   RowBox[{
    RowBox[{"CD", "[", 
     RowBox[{"-", "e"}], "]"}], "[", 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"LI", "[", "1", "]"}], ",", 
      RowBox[{"-", "b"}], ",", "d"}], "]"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[1165]:=",ExpressionUUID->"bf813901-6563-450c-8a46-b575be9b1512"],

Cell[BoxData[
 RowBox[{"4", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", " ", " "},
          {" ", " ", "c", "v"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c, 
    xAct`xTensor`Dir[
     $CellContext`v[$CellContext`d]]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"1", " ", "d"},
            {" ", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`S[
      xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.7289788219190702`*^9},
 CellLabel->"Out[1165]=",
 CellID->1909078976,ExpressionUUID->"0bf6d017-b842-47a7-886f-434008a978a7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The three ",
 StyleBox["Find*Indices",
  FontFamily->"Courier"],
 " functions have attribute ",
 StyleBox["HoldFirst",
  FontFamily->"Courier"],
 ", and hence require the use of ",
 StyleBox["Evaluate",
  FontFamily->"Courier"],
 " when we do not feed the expression itself:"
}], "MathCaption",ExpressionUUID->"5b0fd041-9fbd-424e-b8e3-54272a6cc6d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindIndices", "[", 
  RowBox[{"Evaluate", "[", "expr", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1166]:=",ExpressionUUID->"cc154ed0-3b24-44d2-af25-a2a6b0acaccf"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", 
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, -$CellContext`c, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], -$CellContext`a, 
   xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d, -$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978822054653*^9},
 CellLabel->"Out[1166]=",
 CellID->337179805,ExpressionUUID->"e4af8fbc-b8cd-4637-92e5-136ce6d59176"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindFreeIndices", "[", 
  RowBox[{"Evaluate", "[", "expr", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1167]:=",ExpressionUUID->"1093c74e-dcda-4487-9d3d-b0dec9c53943"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`c, $CellContext`d, -$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978822358879*^9},
 CellLabel->"Out[1167]=",
 CellID->1919121849,ExpressionUUID->"1c716952-bd70-47fe-b137-c0b5d8bb5622"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindDummyIndices", "[", 
  RowBox[{"Evaluate", "[", "expr", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1168]:=",ExpressionUUID->"26bd0cf6-4eea-43ed-bbef-c210aeba3357"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788226463957`*^9},
 CellLabel->"Out[1168]=",
 CellID->1624021850,ExpressionUUID->"33221d10-6e70-475a-be52-41aa07c619e9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindBlockedIndices", "[", 
  RowBox[{"Evaluate", "[", "expr", "]"}], "]"}]], "Input",
 CellLabel->
  "In[1169]:=",ExpressionUUID->"9c22e424-6182-4ecc-8a9b-d56846c34361"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], 
   xAct`xTensor`LI[1]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978822789974*^9},
 CellLabel->"Out[1169]=",
 CellID->418079580,ExpressionUUID->"f7cc9725-14a6-473e-be6c-22efacfa94f1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
This attribute is given to be able to extract indices of expressions before \
they get evaluated:\
\>", "MathCaption",ExpressionUUID->"54cf3993-472b-4ea2-a1ee-73451f43bbac"],

Cell[BoxData[
 RowBox[{"n", "/:", 
  RowBox[{
   RowBox[{"n", "[", "a_", "]"}], 
   RowBox[{"n", "[", 
    RowBox[{"-", "a_"}], "]"}]}], ":=", 
  RowBox[{"-", "1"}]}]], "Input",
 CellLabel->
  "In[1170]:=",ExpressionUUID->"8b65680a-4ff5-46e2-b555-c722321fc144"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"n", "[", "a", "]"}], 
  RowBox[{"n", "[", 
   RowBox[{"-", "a"}], "]"}]}]], "Input",
 CellLabel->
  "In[1171]:=",ExpressionUUID->"1874ad9d-4211-415b-b32f-32f312bf29e5"],

Cell[BoxData[
 RowBox[{"-", "1"}]], "Output",
 CellChangeTimes->{3.728978822985046*^9},
 CellLabel->"Out[1171]=",
 CellID->97008076,ExpressionUUID->"7a3dc2e3-9433-4ab4-bb4c-ee17e9cbe777"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindIndices", "[", 
  RowBox[{
   RowBox[{"n", "[", "a", "]"}], 
   RowBox[{"n", "[", 
    RowBox[{"-", "a"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1172]:=",ExpressionUUID->"ebc891be-03ba-4e5e-9574-29660d9d625a"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", 
     RowBox[{"-", "a"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, -$CellContext`a],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788231215982`*^9},
 CellLabel->"Out[1172]=",
 CellID->474645259,ExpressionUUID->"53ec0e41-54be-4efc-a670-c4d5a8bb9ed5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Note that the returned lists of indices have head ",
 StyleBox["IndexList",
  FontFamily->"Courier"],
 ", in order to avoid confusion with basis and component indices:"
}], "MathCaption",ExpressionUUID->"7a7257e7-26f4-4a22-b15a-3b975d701814"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InputForm", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1173]:=",ExpressionUUID->"d8cd172e-64a5-4e8e-9d68-71eb2f69f9cb"],

Cell["IndexList[a, -a]", "Output",
 CellChangeTimes->{3.7289788232475567`*^9},
 CellLabel->"Out[1173]//InputForm=",
 CellID->1608603055,ExpressionUUID->"6910bb66-d207-477e-b8da-6168a2161218"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["FindIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "FindIndices.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Get all indices of an expression\n", "Text"],
 StyleBox[ButtonBox["FindFreeIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "FindFreeIndices.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Get all free indices of an expression\n", "Text"],
 StyleBox[ButtonBox["FindDummyIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "FindDummyIndices.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Get all dummy indices of an expression\n", "Text"],
 StyleBox["FindBlockedIndices\t\t\t\t", "Text",
  FontFamily->"Courier"],
 StyleBox["Get all blocked indices of an expression", "Text"],
 StyleBox["\n", "DisplayMath"],
 StyleBox[ButtonBox["IndexList",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexList.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Head for a list of indices", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.7289788108087263`*^9, 
  3.728978813502325*^9}},ExpressionUUID->"0e23b0c4-4651-44a1-a3ee-\
482c5df8f2a5"],

Cell["Finding indices. Internal functions", "Text",ExpressionUUID->"a2c4f575-feb8-4edf-ab08-26c30d77788d"],

Cell["Then we have the user driver, with two pairs of brackets:", "Text",ExpressionUUID->"73f4e5cf-2827-4ae2-bbb6-9bc69cc9b5b5"],

Cell[CellGroupData[{

Cell["With no selectors, we get all indices in the expression:", "MathCaption",ExpressionUUID->"b487dc67-fe4e-4949-8750-1eccd2e7be2d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1174]:=",ExpressionUUID->"3987f991-8621-4236-b018-3919f802bf0e"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", 
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, -$CellContext`c, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], -$CellContext`a, 
   xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d, -$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788235415773`*^9},
 CellLabel->"Out[1174]=",
 CellID->1119913302,ExpressionUUID->"615241d1-5c6a-4c2e-a469-b403af5fe989"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can select indices by type:", "MathCaption",ExpressionUUID->"5e9a7637-15ec-40a6-984c-5eff5328942d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "AIndex", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1175]:=",ExpressionUUID->"9c392366-afd0-47ae-b94f-024692108dce"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, -$CellContext`c, \
-$CellContext`a, -$CellContext`b, $CellContext`d, -$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978823820766*^9},
 CellLabel->"Out[1175]=",
 CellID->1836184406,ExpressionUUID->"c6c65fe9-d5ad-4361-8da8-6bb337a60613"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "DIndex", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1176]:=",ExpressionUUID->"fa2f8566-ee93-43ed-a926-aed2ffbe537d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"Dir", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`d],
      Editable->False], "]"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788239699173`*^9},
 CellLabel->"Out[1176]=",
 CellID->1411304252,ExpressionUUID->"35ebb4fe-469d-4040-b455-ba95b54c0810"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "LIndex", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1177]:=",ExpressionUUID->"929ad7af-3096-4593-8f23-c329c00e6205"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"LI", "[", "1", "]"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[
   xAct`xTensor`LI[1]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978824119091*^9},
 CellLabel->"Out[1177]=",
 CellID->833682249,ExpressionUUID->"0fa56880-707d-46f2-8b3e-bf998b0e9472"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or by character:", "MathCaption",ExpressionUUID->"7f5c625f-734f-4588-982b-7f532f5f253c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "Up", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1178]:=",ExpressionUUID->"1237657c-684c-408f-9331-953c427b2a71"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}], ",", "\[ThinSpace]", "d"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, 
   xAct`xTensor`LI[1], $CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978824358161*^9},
 CellLabel->"Out[1178]=",
 CellID->1514245915,ExpressionUUID->"18a33b13-8f57-4934-acd3-1a014710c30a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "Down", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1179]:=",ExpressionUUID->"a664330c-0928-4016-b1a9-be729781757d"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", 
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`c, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], -$CellContext`a, -$CellContext`b, \
-$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978824490089*^9},
 CellLabel->"Out[1179]=",
 CellID->102392808,ExpressionUUID->"38a0d0eb-8851-4dd5-a484-3c564af2b8c9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or by state:", "MathCaption",ExpressionUUID->"b6542f97-4204-415d-b2a5-dcc32ff3e912"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "Free", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1180]:=",ExpressionUUID->"03991c10-3d36-4745-b173-ca20212f9887"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`c, $CellContext`d, -$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788246908073`*^9},
 CellLabel->"Out[1180]=",
 CellID->781522809,ExpressionUUID->"e683677e-0da6-4f25-b3fb-4e6dcc5582de"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "Dummy", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1181]:=",ExpressionUUID->"dbe0d3a9-0487-465e-b803-32fc1b3843a8"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", "a", ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "b"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`a, $CellContext`a, -$CellContext`b, \
$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788251389217`*^9},
 CellLabel->"Out[1181]=",
 CellID->1043865811,ExpressionUUID->"27eb3542-084b-4b35-8885-f7fdeab98bb2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "Blocked", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1182]:=",ExpressionUUID->"ed2e8b93-887a-4d0e-892d-7cfa24604bff"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], 
   xAct`xTensor`LI[1]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978825312847*^9},
 CellLabel->"Out[1182]=",
 CellID->1458773134,ExpressionUUID->"172d4218-2fab-4c12-92fc-52681e63ec15"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can also specify a vbundle (or a basis):", "MathCaption",ExpressionUUID->"1645b5ba-fee1-4383-b782-f728ec514af4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "TangentM3", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1183]:=",ExpressionUUID->"feb66604-48ae-41a3-991c-b246ef12bc66"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{"a", ",", "\[ThinSpace]", "b", ",", "\[ThinSpace]", 
     RowBox[{"-", "c"}], ",", "\[ThinSpace]", 
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"-", "e"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`a, $CellContext`b, -$CellContext`c, 
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], -$CellContext`a, -$CellContext`b, \
$CellContext`d, -$CellContext`e],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978825461932*^9},
 CellLabel->"Out[1183]=",
 CellID->1926748986,ExpressionUUID->"a3ec2f84-a3bd-4c07-95af-120ddabd08ad"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or a tensor:", "MathCaption",ExpressionUUID->"a71749fd-50b6-4a16-9de0-b162a2e23a8c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "S", "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1184]:=",ExpressionUUID->"890c6c2e-4789-4395-9bf4-aee37ece4fb3"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "d", ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`b, $CellContext`d, 
   xAct`xTensor`LI[1]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978825595778*^9},
 CellLabel->"Out[1184]=",
 CellID->964294589,ExpressionUUID->"ab8534f6-b8b5-4529-865a-f6fc69fef193"]
}, Open  ]]
}, Open  ]],

Cell["\<\
We can combine several selectors, in two possible ways: a list of selectors \
represents the union, and a sequence represents the intersection.\
\>", "Text",ExpressionUUID->"29fe073d-cb3a-49da-834d-9ec3421e0c07"],

Cell[CellGroupData[{

Cell["Directions or labels:", "MathCaption",ExpressionUUID->"8900d4b8-df42-4bd2-a20f-cd5d57cc8eea"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"{", 
    RowBox[{"DIndex", ",", "LIndex"}], "}"}], "]"}], "[", "expr", 
  "]"}]], "Input",
 CellLabel->
  "In[1185]:=",ExpressionUUID->"bee2bb1d-d978-4918-b1ef-540fa6545495"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"Dir", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"d"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`d],
       Editable->False], "]"}], ",", "\[ThinSpace]", 
     RowBox[{"LI", "[", "1", "]"}]}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[
   xAct`xTensor`Dir[
    $CellContext`v[$CellContext`d]], 
   xAct`xTensor`LI[1]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788257464733`*^9},
 CellLabel->"Out[1185]=",
 CellID->1987557061,ExpressionUUID->"86ab5e1c-9cc9-420e-a9a6-3f3368f2974a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Free indices on the tensor S:", "MathCaption",ExpressionUUID->"75fa086f-3cd8-409a-a1bd-8d0f8a21bd63"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"Free", ",", "S"}], "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1186]:=",ExpressionUUID->"fca19177-3c73-4e80-9138-18dd0f5b3dda"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], "d", 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978825895564*^9},
 CellLabel->"Out[1186]=",
 CellID->763395142,ExpressionUUID->"c6d13128-76ba-40eb-a39d-a50b232239ef"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Members of a selectors list are conmutative, but not members of a sequence: \
the construction of the final list is performed from left to right. Compare \
with the previous result:\
\>", "MathCaption",ExpressionUUID->"334be659-ef87-4933-93d4-328b122cb3e6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", 
   RowBox[{"S", ",", "Free"}], "]"}], "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1187]:=",ExpressionUUID->"dd83d1c5-dbe1-4ebf-8b40-1c588e98a97a"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "d"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`b, $CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788260115623`*^9},
 CellLabel->"Out[1187]=",
 CellID->1314844416,ExpressionUUID->"8b8c0d3f-5a12-4bfd-bc6d-b95cb7fca734"]
}, Open  ]]
}, Open  ]],

Cell["\<\
This powerful function is very flexible and has more possibilities once we \
allow the presence of bases (see xCobaDoc.nb).\
\>", "Text",ExpressionUUID->"3df83f9b-37cc-496d-8fd1-3aa4f09a2516"],

Cell[TextData[{
 StyleBox[ButtonBox["IndicesOf",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndicesOf.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Get all indices of an expression according to a number of \
selecting criteria", "Text"]
}], "DefinitionBox",ExpressionUUID->"462a50ad-7a2f-47cf-9937-55736894057f"],

Cell["Finding indices. User driver", "Text",ExpressionUUID->"1e627b5b-15af-4eea-a552-cb87e10ee62d"]
}, Closed]],

Cell[CellGroupData[{

Cell["10.2. Replace indices", "Subsection",ExpressionUUID->"68a77a50-2a2b-464e-be3d-aea7c596e152"],

Cell[TextData[{
 "A second, commonly required operation is the change of indices. In \
principle, the user should never need to replace indices directly in an \
expression. But sometimes it is needed, either because two expressions with \
different free indices must be compared, or to avoid problems with the dummy \
indices. This subsection introduces the basic function ",
 StyleBox["ReplaceIndex",
  FontFamily->"Courier"],
 ", and five more functions based on it.\nReplacing indices is a basic \
critical operation: done without care would inmediately lead to wrong \
results. Whenever possible, we recommend the use of delta contractions to \
change indices, which is a safer and mathematically clearer operation."
}], "Text",ExpressionUUID->"8c454804-295a-4df4-9665-194a0ee1939b"],

Cell["\<\
Above all, avoid the temptation of changing indices using normal rules, \
unless you are sure that index-collisions will not appear.\
\>", "Text",ExpressionUUID->"2b35b02a-c8bf-4f14-bed2-8147e6c0d6aa"],

Cell[TextData[{
 StyleBox[ButtonBox["ReplaceIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ReplaceIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Replace an index by another index in an expression\n", "Text"],
 StyleBox[ButtonBox["ReplaceDummies",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ReplaceDummies.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Replace dummy indices by new dummy indices\n", "Text"],
 StyleBox[ButtonBox["SameDummies",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SameDummies.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Minimize number of different dummy indices among several terms of \
a sum\n", "Text"],
 StyleBox[ButtonBox["ChangeFreeIndices",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ChangeFreeIndices.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Change free indices of an expression\n", "Text"],
 StyleBox[ButtonBox["EqualExpressionsQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "EqualExpressionsQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Check whether two expressions are proportional to each other apart \
from a renaming of free indices\n", "Text"],
 StyleBox[ButtonBox["SplitIndex",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "SplitIndex.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Return a list of expressions where an index has been replace by \
different possibilities in a list\n", "Text"],
 StyleBox[ButtonBox["TraceDummy",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "TraceDummy.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Return a sum expressions where a dummy pair has been replace by \
different possibilities in a list", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.7289788460856133`*^9, 
  3.72897885722967*^9}},ExpressionUUID->"3f6ce36e-5c90-432c-a098-\
4423b2bb4792"],

Cell["Index replacements.", "Text",ExpressionUUID->"35178b9b-4b2c-44ba-9287-02bfed95b5f8"],

Cell[CellGroupData[{

Cell["Let us go back to our expression:", "MathCaption",ExpressionUUID->"1ea4b711-f943-49cf-866b-97d2be47f759"],

Cell[CellGroupData[{

Cell[BoxData["expr"], "Input",
 CellLabel->
  "In[1188]:=",ExpressionUUID->"5f02793c-a854-4c2c-a92f-510be884e9fa"],

Cell[BoxData[
 RowBox[{"4", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", " ", " "},
          {" ", " ", "c", "v"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c, 
    xAct`xTensor`Dir[
     $CellContext`v[$CellContext`d]]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"1", " ", "d"},
            {" ", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`S[
      xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978874875111*^9},
 CellLabel->"Out[1188]=",
 CellID->101426694,ExpressionUUID->"a7c8fac9-0283-49fa-a4ab-08024f8eaf01"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "For similar reasons, ",
 StyleBox["ReplaceIndex",
  FontFamily->"Courier"],
 " has attribute ",
 StyleBox["HoldFirst",
  FontFamily->"Courier"],
 ", and hence we need ",
 StyleBox["Evaluate",
  FontFamily->"Courier"],
 ". We can replace a single index:"
}], "MathCaption",ExpressionUUID->"0dd1d9b6-37ba-4247-98a7-326bc91b7ec0"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{"Evaluate", "[", "expr", "]"}], ",", 
   RowBox[{"a", "\[Rule]", "f"}]}], "]"}]], "Input",
 CellLabel->
  "In[1189]:=",ExpressionUUID->"cae724e6-0487-456e-8869-216c7fe76746"],

Cell[BoxData[
 RowBox[{"4", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"f", "b", " ", " "},
          {" ", " ", "c", "v"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`f, $CellContext`b, -$CellContext`c, 
    xAct`xTensor`Dir[
     $CellContext`v[$CellContext`d]]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"1", " ", "d"},
            {" ", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`S[
      xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.72897887500665*^9},
 CellLabel->"Out[1189]=",
 CellID->1723913987,ExpressionUUID->"8ce8d018-eba7-4205-9cc1-f3c4c626c6f4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Note that the paired index ",
 StyleBox["-a",
  FontFamily->"Courier"],
 " has not been replaced. We need separate rules for that (this gives greater \
flexibility):"
}], "MathCaption",ExpressionUUID->"33b3e14f-9ec0-4f3f-aab9-fac51b3861b6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{"Evaluate", "[", "expr", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"a", "\[Rule]", "f"}], ",", 
     RowBox[{
      RowBox[{"-", "a"}], "\[Rule]", 
      RowBox[{"-", "f"}]}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1190]:=",ExpressionUUID->"05e945ee-5191-4533-9a44-ada8367ba4de"],

Cell[BoxData[
 RowBox[{"4", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"f", "b", " ", " "},
          {" ", " ", "c", "v"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`f, $CellContext`b, -$CellContext`c, 
    xAct`xTensor`Dir[
     $CellContext`v[$CellContext`d]]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`f],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"1", " ", "d"},
            {" ", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`S[
      xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978875122147*^9},
 CellLabel->"Out[1190]=",
 CellID->2120497832,ExpressionUUID->"36664888-4387-4549-aa4a-ff491c478820"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can change the character of an index, or even the type:", \
"MathCaption",ExpressionUUID->"16c5cc11-ec75-401f-a4b1-58f13116440d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{"Evaluate", "[", "expr", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"a", "\[Rule]", 
      RowBox[{"-", "f"}]}], ",", 
     RowBox[{
      RowBox[{"-", "a"}], "\[Rule]", "f"}], ",", 
     RowBox[{
      RowBox[{"Dir", "[", 
       RowBox[{"v", "[", "d", "]"}], "]"}], "\[Rule]", 
      RowBox[{"LI", "[", "2", "]"}]}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1191]:=",ExpressionUUID->"aff528fe-5a5b-4e98-b971-b58a1d123480"],

Cell[BoxData[
 RowBox[{"4", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", "b", " ", "2"},
          {"f", " ", "c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`f, $CellContext`b, -$CellContext`c, 
    xAct`xTensor`LI[2]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"f"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`f],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"1", " ", "d"},
            {" ", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`S[
      xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978875236174*^9},
 CellLabel->"Out[1191]=",
 CellID->1715940459,ExpressionUUID->"38f4585e-dc98-4a36-a920-2884996ea1f6"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["If the rules do not apply, nothing happens:", "MathCaption",ExpressionUUID->"1dc78cd9-8486-4725-9123-15d6509810fa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{"Evaluate", "[", "expr", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"f", "\[Rule]", "a"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1192]:=",ExpressionUUID->"a5706088-171f-4209-ad6d-6a7ce9714825"],

Cell[BoxData[
 RowBox[{"4", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", " ", " "},
          {" ", " ", "c", "v"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c, 
    xAct`xTensor`Dir[
     $CellContext`v[$CellContext`d]]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False], " ", 
  RowBox[{"(", 
   InterpretationBox[
    StyleBox[
     RowBox[{
      SubscriptBox["\[EmptyDownTriangle]", "e"], GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {"1", " ", "d"},
            {" ", "b", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`CD[-$CellContext`e][
     $CellContext`S[
      xAct`xTensor`LI[1], -$CellContext`b, $CellContext`d]],
    Editable->False], ")"}]}]], "Output",
 CellChangeTimes->{3.728978875352461*^9},
 CellLabel->"Out[1192]=",
 CellID->1253398850,ExpressionUUID->"468b0d29-b24c-48fd-9086-ba525ac34c28"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Pattern indices can be changed, but this should never be required:\
\>", "MathCaption",ExpressionUUID->"6e97997d-9ad6-41cd-b188-ed3fcce9fbc4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{"T", "[", "a_", "]"}], ",", 
   RowBox[{"a", "\[Rule]", "b"}]}], "]"}]], "Input",
 CellLabel->
  "In[1193]:=",ExpressionUUID->"59274bc5-e245-4819-b937-0359f467bccb"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {
          UnderscriptBox["b", "_"]},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`T[
   Pattern[$CellContext`b, 
    Blank[]]],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788754404593`*^9},
 CellLabel->"Out[1193]=",
 CellID->1207378527,ExpressionUUID->"906018a3-2b3f-4f7e-8a3e-1234a3071628"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The replacement is performed with great care. In the following list T is a \
tensor, but X is not a tensor:\
\>", "MathCaption",ExpressionUUID->"dd271608-1cce-4f7b-982d-7922eddfd796"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ReplaceIndex", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"a", ",", 
     RowBox[{"T", "[", "a", "]"}], ",", 
     RowBox[{"T", "[", 
      RowBox[{"-", "a"}], "]"}], ",", 
     RowBox[{"X", "[", "a", "]"}]}], "}"}], ",", 
   RowBox[{"a", "\[Rule]", "b"}]}], "]"}]], "Input",
 CellLabel->
  "In[1194]:=",ExpressionUUID->"cddb3c08-4e2e-49e4-b7c0-b7ec23245170"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"a", ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`b],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a],
    Editable->False], ",", 
   RowBox[{"X", "[", "a", "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978875538023*^9},
 CellLabel->"Out[1194]=",
 CellID->837922786,ExpressionUUID->"3c3ef9d9-4733-44f8-bfda-07d793082cde"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "As I said, the user should try to refrain from using ",
 StyleBox["ReplaceIndex",
  FontFamily->"Courier"],
 ". If a free index must be changed, use contractions with delta tensors: if \
we want to change from ",
 StyleBox["T[a]",
  FontFamily->"Courier"],
 " to ",
 StyleBox["T[b]",
  FontFamily->"Courier"],
 " we can do this:"
}], "MathCaption",ExpressionUUID->"9d384b72-75ba-48c8-beda-4d191dff22af"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"T", "[", "a", "]"}], 
  RowBox[{"delta", "[", 
   RowBox[{
    RowBox[{"-", "a"}], ",", "b"}], "]"}]}]], "Input",
 CellLabel->
  "In[1195]:=",ExpressionUUID->"f1839c86-0e9a-44b0-949a-4456f09e48d2"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {"b"},
         {" "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`T[$CellContext`b],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289788756518383`*^9},
 CellLabel->"Out[1195]=",
 CellID->1917990923,ExpressionUUID->"6cfe3277-edd2-4f91-8f98-d7b8812b2c0c"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "In some cases we need to change the dummy indices of an expression. For \
example when we want to multiply a previous expression by itself. ",
 StyleBox["ReplaceDummies",
  FontFamily->"Courier"],
 " is the function which calls ",
 StyleBox["ReplaceIndex",
  FontFamily->"Courier"],
 " to manipulate all possible cases with dummies. A particular case, always \
required in the canonicalization process, is implemented in ",
 StyleBox["SameDummies",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"0f28b103-1f3c-451f-92f4-a23c34372f91"],

Cell[CellGroupData[{

Cell["Suppose now this scalar expression:", "MathCaption",ExpressionUUID->"4f9aab16-ab41-4b86-935b-176894b39011"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}]}]}]], "Input",
 CellLabel->
  "In[1196]:=",ExpressionUUID->"bffc8524-eb60-4749-9859-340eadcd48df"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978875752359*^9},
 CellLabel->"Out[1196]=",
 CellID->1943901391,ExpressionUUID->"6e4052a9-9889-4ccd-b8d4-1ecd42a99960"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["This is incorrect:", "MathCaption",ExpressionUUID->"99012966-5b75-46bb-b44b-87c258438ef4"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", " ", "expr"}]], "Input",
 CellLabel->
  "In[1197]:=",ExpressionUUID->"100a5072-2743-49ff-8ba7-56cf190c0ea6"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"a", "b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a, -$CellContext`b],
    Editable->False], "2"], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.728978875886117*^9},
 CellLabel->"Out[1197]=",
 CellID->1481956726,ExpressionUUID->"daadf511-6f13-4a91-b113-30d3f4d6c6c5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The following is correct. ",
 StyleBox["ReplaceDummies",
  FontFamily->"Courier"],
 " changes the dummy by the given indices"
}], "MathCaption",ExpressionUUID->"929903a5-6ee2-4d13-af7f-b41bd4370310"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", " ", 
  RowBox[{"ReplaceDummies", "[", 
   RowBox[{"expr", ",", 
    RowBox[{"IndexList", "[", 
     RowBox[{"e", ",", "f"}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[1198]:=",ExpressionUUID->"da498164-adb3-4e7e-aa7a-1a798dd60837"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"e", "f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`e, -$CellContext`f],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"e", "f"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`e, $CellContext`f],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978876018105*^9},
 CellLabel->"Out[1198]=",
 CellID->731355976,ExpressionUUID->"a324e78f-07d6-4274-b551-8f60d7826ca0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
If not enough indices (or no index at all) are given, the dummies are \
replaced by unique  dollar-indices:\
\>", "MathCaption",ExpressionUUID->"4850485e-cf10-4ed6-b042-a7313d8fcbf1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", " ", 
  RowBox[{"ReplaceDummies", "[", "expr", "]"}]}]], "Input",
 CellLabel->
  "In[1199]:=",ExpressionUUID->"43306e30-242a-48d4-82c3-ea9169785cab"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"a", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`a, -$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {" ", " "},
          {"c", "d"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[-$CellContext`c, -$CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`c, $CellContext`d],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978876152823*^9},
 CellLabel->"Out[1199]=",
 CellID->1850976851,ExpressionUUID->"b7cb7749-0df4-48ec-a5fb-f28ce932a99e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndicesOf", "[", "]"}], "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1200]:=",ExpressionUUID->"0ff4b29f-3819-4961-8a5e-5f399ec359ed"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["{",
     FontColor->RGBColor[0, 0, 1]], 
    RowBox[{
     RowBox[{"-", "a"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "b"}], ",", "\[ThinSpace]", "a", ",", "\[ThinSpace]", "b", 
     ",", "\[ThinSpace]", 
     RowBox[{"-", "h$46992"}], ",", "\[ThinSpace]", 
     RowBox[{"-", "h$46993"}], ",", "\[ThinSpace]", "h$46992", ",", 
     "\[ThinSpace]", "h$46993"}], 
    StyleBox["}",
     FontColor->RGBColor[0, 0, 1]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`IndexList[-$CellContext`a, -$CellContext`b, $CellContext`a, \
$CellContext`b, -$CellContext`h$46992, -$CellContext`h$46993, \
$CellContext`h$46992, $CellContext`h$46993],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978876268751*^9},
 CellLabel->"Out[1200]=",
 CellID->919824315,ExpressionUUID->"076e08f9-0503-46cf-91bd-c5032a87cd58"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "A particular case of dummy manipulation is the minimization of the number \
of dummies in a sum of terms. This is achieved using ",
 StyleBox["SameDummies",
  FontFamily->"Courier"],
 ", which is simply a combined call to ",
 StyleBox["FindDummyIndices",
  FontFamily->"Courier"],
 " and ",
 StyleBox["ReplaceDummies",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"86993165-44a8-4b2e-a22d-dbdd7f1ecd46"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SameDummies", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b", ",", "d", ",", 
      RowBox[{"-", "d"}]}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "c", ",", "e", ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1201]:=",ExpressionUUID->"8c6b02de-2419-47f2-8349-8b6e4472f7de"],

Cell[BoxData[
 RowBox[{"2", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", "c", " "},
          {" ", " ", " ", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c, \
-$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289788764021053`*^9},
 CellLabel->"Out[1201]=",
 CellID->1899971099,ExpressionUUID->"a98ce367-bdde-466e-8dee-2e2c261ab3e2"]
}, Open  ]]
}, Open  ]],

Cell["\<\
In parallel, sometimes we need to change the free indices of an expression. \
This happens for example when we need to compare the results of several \
independent computations. If we do not know the index translation table which \
brings one expression to the other, then we need to search among all \
permutations, what could take quite some time...\
\>", "Text",ExpressionUUID->"331ef9a3-c41e-4169-88d8-7674afe1c480"],

Cell[CellGroupData[{

Cell["\<\
This changes the free indices a,c to b,e. The index b was a dummy in the \
original expression, and hence it must be replaced (by a unique dollar-dummy) \
before actually changing the free indices:\
\>", "MathCaption",ExpressionUUID->"57336c3e-2ae6-458b-b1f2-8aa13319c26c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ChangeFreeIndices", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", "c", ",", "d", ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "d"}]}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"b", ",", "e"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1202]:=",ExpressionUUID->"ecf81d73-b7cc-4e1a-ab8b-8362167da0aa"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"T", 
      StyleBox[GridBox[{
         {"b", "a", "e", "d", " ", " "},
         {" ", " ", " ", " ", "a", "d"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`T[$CellContext`b, $CellContext`a, $CellContext`e, \
$CellContext`d, -$CellContext`a, -$CellContext`d],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978876506719*^9},
 CellLabel->"Out[1202]=",
 CellID->1879913650,ExpressionUUID->"8e0f888f-3e7b-44fe-83b1-ad89fe9fd200"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
It is not obvious to see whether these two expressions are the same or not \
under a renaming of free indices:\
\>", "MathCaption",ExpressionUUID->"63433919-b7ed-4060-b27a-59382e72ea60"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"expr1", "=", 
  RowBox[{"4", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}]}], "]"}], 
   RowBox[{"U", "[", 
    RowBox[{"b", ",", 
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "e"}]}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{"c", ",", "d"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"expr2", "=", 
  RowBox[{"2", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "d"}], ",", 
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "a"}]}], "]"}], 
   RowBox[{"U", "[", 
    RowBox[{"a", ",", 
     RowBox[{"-", "f"}], ",", "g"}], "]"}], 
   RowBox[{"T", "[", 
    RowBox[{"d", ",", 
     RowBox[{"-", "g"}]}], "]"}]}]}]}], "Input",
 CellLabel->
  "In[1203]:=",ExpressionUUID->"1ec771b7-f5f1-4180-b0ab-2efd66266226"],

Cell[BoxData[
 RowBox[{"4", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"c", "d"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`c, $CellContext`d],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"a", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"b", " ", " "},
          {" ", "d", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978876600754*^9},
 CellLabel->"Out[1203]=",
 CellID->1929300886,ExpressionUUID->"f729d0ec-d17e-4cb8-8ab8-37cb66991c32"],

Cell[BoxData[
 RowBox[{"2", " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"d", " "},
          {" ", "g"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`d, -$CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {"a", " ", "g"},
          {" ", "f", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[$CellContext`a, -$CellContext`f, $CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"d", "b", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[-$CellContext`d, -$CellContext`b, -$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978876622154*^9},
 CellLabel->"Out[1204]=",
 CellID->904404949,ExpressionUUID->"0e629277-9a32-4426-bbdc-a4e4052292ad"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "The function ",
 StyleBox["EqualExpressionsQ",
  FontFamily->"Courier"],
 " gives ",
 StyleBox["True",
  FontFamily->"Courier"],
 " if one is simply a constant times the other, and ",
 StyleBox["False",
  FontFamily->"Courier"],
 " otherwise:"
}], "MathCaption",ExpressionUUID->"088a6e9c-0ee1-44df-aad1-c3e6c8059dcb"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"EqualExpressionsQ", "[", 
  RowBox[{"expr1", ",", "expr2"}], "]"}]], "Input",
 CellLabel->
  "In[1205]:=",ExpressionUUID->"48a5a25b-e796-489e-be4a-b8d0199583f4"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"LEFT  == \"\>", "\[InvisibleSpace]", 
   RowBox[{"-", "2"}], "\[InvisibleSpace]", "\<\" RIGHT   with rules: \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "a"}], "\[Rule]", 
      RowBox[{"-", "b"}]}], ",", 
     RowBox[{
      RowBox[{"-", "e"}], "\[Rule]", 
      RowBox[{"-", "f"}]}]}], "}"}]}],
  SequenceForm[
  "LEFT  == ", -2, 
   " RIGHT   with rules: ", {-$CellContext`a -> -$CellContext`b, \
-$CellContext`e -> -$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978876723538*^9},
 CellLabel->"During evaluation of In[1205]:=",
 CellID->444724377,ExpressionUUID->"9a36a885-2676-4d0c-bcf3-0f87ee5555c3"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978876736678*^9},
 CellLabel->"Out[1205]=",
 CellID->1239512679,ExpressionUUID->"69a310c2-ea43-4750-9ecd-13b1ae2fc5e8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"EqualExpressionsQ", "[", 
  RowBox[{"expr1", ",", 
   RowBox[{
    RowBox[{"T", "[", "]"}], "expr2"}]}], "]"}]], "Input",
 CellLabel->
  "In[1206]:=",ExpressionUUID->"4bab9c9c-34bc-4aba-8251-5ad860cdabe5"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.728978876888626*^9},
 CellLabel->"Out[1206]=",
 CellID->1898220053,ExpressionUUID->"ce0fbcb8-d1b6-4b4e-9979-6dd1cfe23b3f"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "Finally, we introduce the operation of index splitting. The key idea is the \
concept of a splitrule of the form   ",
 StyleBox["a -> IndexList[ a1, a2, ... ]",
  FontFamily->"Courier"],
 ".  The function ",
 StyleBox["SplitIndex",
  FontFamily->"Courier"],
 " is rather efficient because it uses ",
 StyleBox["ReplaceIndex",
  FontFamily->"Courier"],
 " only once, replacing the index by a temporary variable, which is later \
used to do all other replacements with ",
 StyleBox["ReplaceAll",
  FontFamily->"Courier"],
 ", which is much faster."
}], "Text",ExpressionUUID->"4f840722-7826-490a-9e87-e58551852f62"],

Cell[CellGroupData[{

Cell["We can construct a list of replacements as follows:", "MathCaption",ExpressionUUID->"14e43814-0a7b-482a-93cb-989a4415f4fa"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SplitIndex", "[", 
  RowBox[{"expr1", ",", 
   RowBox[{
    RowBox[{"-", "a"}], "\[Rule]", 
    RowBox[{"IndexList", "[", 
     RowBox[{
      RowBox[{"-", "A"}], ",", 
      RowBox[{"-", "B"}], ",", 
      RowBox[{"-", "C"}]}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1207]:=",ExpressionUUID->"0b657521-d0d2-4e7e-8dd2-2281e9e6b9fb"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"c", "d"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`c, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"A", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[-$CellContext`A, -$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"b", " ", " "},
            {" ", "d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
     Editable->False]}], ",", 
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"c", "d"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`c, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"b", " ", " "},
            {" ", "d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"B", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[-$CellContext`B, -$CellContext`b, -$CellContext`c],
     Editable->False]}], ",", 
   RowBox[{"4", " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"c", "d"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`c, $CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"b", " ", " "},
            {" ", "d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"C", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[-C, -$CellContext`b, -$CellContext`c],
     Editable->False]}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978877052256*^9},
 CellLabel->"Out[1207]=",
 CellID->367370939,ExpressionUUID->"a7e40621-21da-4b12-8798-83144f028ec9"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or an array:", "MathCaption",ExpressionUUID->"28062b0f-a0a6-4e31-b569-9ed6389d5a77"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SplitIndex", "[", 
  RowBox[{"expr1", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "a"}], "\[Rule]", 
      RowBox[{"IndexList", "[", 
       RowBox[{
        RowBox[{"-", "A"}], ",", 
        RowBox[{"-", "B"}]}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"-", "b"}], "\[Rule]", 
      RowBox[{"IndexList", "[", 
       RowBox[{"C", ",", "D"}], "]"}]}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1208]:=",ExpressionUUID->"8256cd02-1a59-4f7f-9f01-f9b5c1aba6db"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", "C", " "},
              {"A", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-$CellContext`A, C, -$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", "D", " "},
              {"A", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-$CellContext`A, D, -$CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", "C", " "},
              {"B", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-$CellContext`B, C, -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", "D", " "},
              {"B", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-$CellContext`B, D, -$CellContext`c],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289788772393503`*^9},
 CellLabel->"Out[1208]=",
 CellID->387886648,ExpressionUUID->"63821e0f-b1a0-443c-b498-ff5572caa16a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Dimensions", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1209]:=",ExpressionUUID->"be0afcd4-87f9-4a72-8e8b-24e4e4512a9b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2", ",", "2"}], "}"}]], "Output",
 CellChangeTimes->{3.728978877340312*^9},
 CellLabel->"Out[1209]=",
 CellID->1598024291,ExpressionUUID->"99c61762-2fae-4934-9916-26a2da43cff7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "There is also a combined notation for multiple expansions over the same \
range. Again, this is performed through a single call to ",
 StyleBox["ReplaceIndex",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"65a7c07b-05fd-4128-87d1-24b55ec809e5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SplitIndex", "[", 
  RowBox[{"expr1", ",", 
   RowBox[{
    RowBox[{"IndexList", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "]"}], "\[Rule]", 
    RowBox[{"IndexList", "[", 
     RowBox[{
      RowBox[{"LI", "[", "1", "]"}], ",", 
      RowBox[{"LI", "[", "2", "]"}], ",", 
      RowBox[{"LI", "[", "3", "]"}]}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1210]:=",ExpressionUUID->"971ad077-8690-47f7-b065-a37d0763409b"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"1", "1", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[1], 
        xAct`xTensor`LI[1], -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"1", "2", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[1], 
        xAct`xTensor`LI[2], -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"1", "3", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[1], 
        xAct`xTensor`LI[3], -$CellContext`c],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"2", "1", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[2], 
        xAct`xTensor`LI[1], -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"2", "2", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[2], 
        xAct`xTensor`LI[2], -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"2", "3", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[2], 
        xAct`xTensor`LI[3], -$CellContext`c],
       Editable->False]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"3", "1", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[3], 
        xAct`xTensor`LI[1], -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"3", "2", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[3], 
        xAct`xTensor`LI[2], -$CellContext`c],
       Editable->False]}], ",", 
     RowBox[{"4", " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"c", "d"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`c, $CellContext`d],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"b", " ", " "},
              {" ", "d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[$CellContext`b, -$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {"3", "3", " "},
              {" ", " ", "c"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[
        xAct`xTensor`LI[3], 
        xAct`xTensor`LI[3], -$CellContext`c],
       Editable->False]}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978877456197*^9},
 CellLabel->"Out[1210]=",
 CellID->1563769455,ExpressionUUID->"0d6e8811-665f-491a-a874-cf9382994925"]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Remove", "[", 
  RowBox[{"expr1", ",", "expr2"}], "]"}]], "Input",
 CellLabel->
  "In[1211]:=",ExpressionUUID->"f11b00ca-2ad1-45d5-ae7f-ec6964794603"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.3. Scalar and scalar functions", "Subsection",ExpressionUUID->"34d6e14e-978f-4ad5-8c42-a9b088439972"],

Cell[TextData[{
 "Another important problem is the following. In the same way that we are \
using ",
 StyleBox["Times",
  FontFamily->"Courier"],
 " for a product of tensors, we shall use ",
 StyleBox["Power",
  FontFamily->"Courier"],
 ", ",
 StyleBox["Sqrt",
  FontFamily->"Courier"],
 ", ",
 StyleBox["Exp",
  FontFamily->"Courier"],
 " and other functions acting on scalar tensor expressions. There is a \
problem: there are built-in definitions for those functions that could be \
incorrect for tensors. For example we have ",
 StyleBox["(a b)^n = a^n b^n",
  FontFamily->"Courier"],
 ". However ",
 StyleBox["(v[a] v[-a])^n",
  FontFamily->"Courier"],
 " is not equal to ",
 StyleBox["v[a]^n v[-a]^n",
  FontFamily->"Courier"],
 ". To prevent those problems we have two options: we could define new \
functions IndexPower, IndexSqrt, etc, or we can introduce a new head ",
 StyleBox["Scalar",
  FontFamily->"Courier"],
 " such that the arguments of those functions must be always wrapped with \
that head."
}], "Text",ExpressionUUID->"d6b7cae7-6cbc-4b7a-8f76-4397526299c9"],

Cell[TextData[{
 StyleBox[ButtonBox["Scalar",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Scalar.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Wrapper to isolate scalar expressions\n", "Text"],
 StyleBox[ButtonBox["PutScalar",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PutScalar.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Wrap scalars with the head Scalar\n", "Text"],
 StyleBox[ButtonBox["BreakScalars",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "BreakScalars.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Break Scalar expressions into irreducible Scalar expressions\n", 
  "Text"],
 StyleBox[ButtonBox["NoScalar",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "NoScalar.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Remove all Scalar heads, introducing new dummies wherever needed", 
  "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.7289788812028923`*^9},ExpressionUUID->"ccc23ef0-80f2-499f-ac44-\
b818dc825725"],

Cell["Scalar expressions.", "Text",ExpressionUUID->"dfd5dd34-fc9d-4079-ac37-f8622311e238"],

Cell[CellGroupData[{

Cell[TextData[{
 "The direct use of ",
 StyleBox["Power",
  FontFamily->"Courier"],
 " could lead to unconsistent expressions:"
}], "MathCaption",ExpressionUUID->"02809a8f-7935-42fe-b9d3-87c59bf6e556"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(", 
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}]}], ")"}], "^", "2"}]], "Input",
 CellLabel->
  "In[1212]:=",ExpressionUUID->"7e28da1b-4600-4317-b9f1-4ee3c66fa768"],

Cell[BoxData[
 RowBox[{
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], "2"], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"a"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`a],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.728978886928447*^9},
 CellLabel->"Out[1212]=",
 CellID->1052016746,ExpressionUUID->"e893da69-2f50-456c-902b-720c3baeb5c4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Instead, the right way to write it would be", "MathCaption",ExpressionUUID->"89afdd22-2fed-4caa-b129-f4242e28eaf5"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Scalar", "[", 
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}]}], "]"}], "^", "2"}]], "Input",
 CellLabel->
  "In[1213]:=",ExpressionUUID->"2a926f40-ec11-4b98-8d4b-2f34888ea30f"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.728978887060541*^9},
 CellLabel->"Out[1213]=",
 CellID->1012584271,ExpressionUUID->"c60a4b11-acaa-494f-a22d-2cbe60519d86"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " can work with those objects as usual:"
}], "MathCaption",ExpressionUUID->"f3fba0e3-ed58-4501-8902-895a8267fe48"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Scalar", "[", 
    RowBox[{
     RowBox[{"v", "[", "a", "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}]}], "]"}], "^", "2"}], " ", 
  RowBox[{
   RowBox[{
    RowBox[{"Scalar", "[", 
     RowBox[{
      RowBox[{"v", "[", "b", "]"}], 
      RowBox[{"v", "[", 
       RowBox[{"-", "b"}], "]"}]}], "]"}], "^", "3"}], "/", 
   RowBox[{
    RowBox[{"Scalar", "[", 
     RowBox[{
      RowBox[{"v", "[", "c", "]"}], 
      RowBox[{"v", "[", 
       RowBox[{"-", "c"}], "]"}]}], "]"}], "^", "4"}]}]}]], "Input",
 CellLabel->
  "In[1214]:=",ExpressionUUID->"2a1b3afe-1680-41f2-ba36-3fd29147134a"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   SuperscriptBox[
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[-$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
     Editable->False], "2"], " ", 
   SuperscriptBox[
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[-$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"b"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`b],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[-$CellContext`b] \
$CellContext`v[$CellContext`b]],
     Editable->False], "3"]}], 
  SuperscriptBox[
   InterpretationBox[
    StyleBox[
     RowBox[{
      StyleBox["(",
       FontColor->RGBColor[1, 0, 0]], 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {" "},
               {"c"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[-$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[$CellContext`c],
        Editable->False]}], 
      StyleBox[")",
       FontColor->RGBColor[1, 0, 0]]}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Scalar[$CellContext`v[-$CellContext`c] \
$CellContext`v[$CellContext`c]],
    Editable->False], "4"]]], "Output",
 CellChangeTimes->{3.728978887189312*^9},
 CellLabel->"Out[1214]=",
 CellID->1843423869,ExpressionUUID->"7a38a75b-3952-4011-bb56-b4b218f1ad70"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1215]:=",ExpressionUUID->"64be45db-78a2-45df-a8b4-0433bf8ae7ee"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["(",
     FontColor->RGBColor[1, 0, 0]], 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`a],
      Editable->False]}], 
    StyleBox[")",
     FontColor->RGBColor[1, 0, 0]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978887426001*^9},
 CellLabel->"Out[1215]=",
 CellID->678391014,ExpressionUUID->"3802aa67-8e6b-4326-b2d5-ec70241ddeea"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Note that ",
 StyleBox["Scalar",
  FontFamily->"Courier"],
 " effectively shields its argument from the rest of an expression and \
therefore there can be repeated indices:"
}], "MathCaption",ExpressionUUID->"a23430fd-4f21-48fc-bd82-1d783d79446e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"v", "[", "a", "]"}], 
  RowBox[{"Scalar", "[", 
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[1216]:=",ExpressionUUID->"1d2435ee-543a-48ad-819d-18469e47d9a7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978887865926*^9},
 CellLabel->"Out[1216]=",
 CellID->727090375,ExpressionUUID->"c063ba00-21d8-4127-abf7-81f32cb00bed"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Validate", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1217]:=",ExpressionUUID->"f4e9ad89-1248-4411-a143-6a3dc3e2a5e7"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978888332727*^9},
 CellLabel->"Out[1217]=",
 CellID->1025981195,ExpressionUUID->"3d2f6192-691e-40f7-aa2e-f792f62f7520"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "By default the function ",
 StyleBox["Simplification",
  FontFamily->"Courier"],
 " does not add new ",
 StyleBox["Scalar",
  FontFamily->"Courier"],
 " heads, because that takes some time for large expressions:"
}], "MathCaption",ExpressionUUID->"5e7bc697-5e66-4d09-a94d-8aee390a8881"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"v", "[", "b", "]"}], 
   RowBox[{
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}], "/", 
    RowBox[{"Scalar", "[", 
     RowBox[{
      RowBox[{"v", "[", "a", "]"}], 
      RowBox[{"v", "[", 
       RowBox[{"-", "a"}], "]"}]}], "]"}]}]}], "//", 
  "Simplification"}]], "Input",
 CellLabel->
  "In[1218]:=",ExpressionUUID->"f9d1632a-49af-4c4b-a9f4-dc0e9645bbc1"],

Cell[BoxData[
 FractionBox[
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False]}], 
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False]]], "Output",
 CellChangeTimes->{3.728978888545436*^9},
 CellLabel->"Out[1218]=",
 CellID->1519742602,ExpressionUUID->"8a7c8ff0-de02-42ee-99e0-2a3d4b443ef1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "You can force it using ",
 StyleBox["PutScalar",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"ef0cf605-7fcb-4dab-874c-fd1ba838cd67"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"PutScalar", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1219]:=",ExpressionUUID->"831d34d5-c825-49b7-bfa2-3e3bc81f8ee2"],

Cell[BoxData[
 FractionBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`b],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`b] \
$CellContext`v[$CellContext`b]],
   Editable->False], 
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False]]], "Output",
 CellChangeTimes->{3.728978888917203*^9},
 CellLabel->"Out[1219]=",
 CellID->143414396,ExpressionUUID->"9988d172-c4c2-4cd2-aef3-9803db18c0d7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1220]:=",ExpressionUUID->"82ae8eae-4525-40e8-a4cd-87cf0d1beeae"],

Cell[BoxData["1"], "Output",
 CellChangeTimes->{3.728978889216895*^9},
 CellLabel->"Out[1220]=",
 CellID->889101638,ExpressionUUID->"b2bab2e6-40e1-4d20-8a0c-c281b03b7e2b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Two other functions to manipulate Scalar expressions are ",
 StyleBox["BreakScalars",
  FontFamily->"Courier"],
 " and ",
 StyleBox["NoScalar",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"afc80288-f75e-4254-82cb-0a383607b6ec"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Scalar", "[", 
  RowBox[{
   RowBox[{"v", "[", "a", "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "a"}], "]"}], 
   RowBox[{"v", "[", "b", "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1221]:=",ExpressionUUID->"238c28a7-0568-4a0d-b606-b68eac0c4de4"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{
    StyleBox["(",
     FontColor->RGBColor[1, 0, 0]], 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`b],
      Editable->False]}], 
    StyleBox[")",
     FontColor->RGBColor[1, 0, 0]]}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Scalar[(($CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]) $CellContext`v[-$CellContext`b]) \
$CellContext`v[$CellContext`b]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978889350171*^9},
 CellLabel->"Out[1221]=",
 CellID->940014040,ExpressionUUID->"00a549f2-e608-4b61-878c-391bd5ae3104"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "BreakScalars"}]], "Input",
 CellLabel->
  "In[1222]:=",ExpressionUUID->"6c45551a-43eb-4f09-96eb-ae05a5750bf2"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"b"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`b],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`b] \
$CellContext`v[$CellContext`b]],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978889495236*^9},
 CellLabel->"Out[1222]=",
 CellID->1926262948,ExpressionUUID->"69ee8eb8-ec20-49b3-ae0a-3d8998eb3502"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "Simplification"}]], "Input",
 CellLabel->
  "In[1223]:=",ExpressionUUID->"804d690d-8ec8-4c0e-a017-090f827590dd"],

Cell[BoxData[
 SuperscriptBox[
  InterpretationBox[
   StyleBox[
    RowBox[{
     StyleBox["(",
      FontColor->RGBColor[1, 0, 0]], 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], 
     StyleBox[")",
      FontColor->RGBColor[1, 0, 0]]}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
   Editable->False], "2"]], "Output",
 CellChangeTimes->{3.7289788896456947`*^9},
 CellLabel->"Out[1223]=",
 CellID->1385388511,ExpressionUUID->"9026885d-242f-4c97-b679-16ea593fc002"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "NoScalar"}]], "Input",
 CellLabel->
  "In[1224]:=",ExpressionUUID->"8aa7cb1c-a863-435c-94d1-830aa0c3ba4c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"a"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`b],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"b"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978889780858*^9},
 CellLabel->"Out[1224]=",
 CellID->1207003603,ExpressionUUID->"924d21a1-dd0e-4ce9-9bfd-8c4346ba0cbf"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["DefScalarFunction",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefScalarFunction.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Define a scalar function\n", "Text"],
 StyleBox[ButtonBox["UndefScalarFunction",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefScalarFunction.nb", 
     CharacterEncoding -> "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Undefine a scalar function\n", "Text"],
 StyleBox[ButtonBox["$ScalarFunctions",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$ScalarFunctions.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["List of defined scalar functions", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.505362003311414*^9, 
  3.505362018206544*^9}},ExpressionUUID->"c16ed14c-a9db-48b2-8046-\
7985ef86d7ab"],

Cell["Definition of a scalar function.", "Text",ExpressionUUID->"cb026765-8d55-4693-aa38-9cc24b0d450f"],

Cell[CellGroupData[{

Cell["\<\
A scalar function is one that only accepts scalar arguments and returns \
scalars.\
\>", "MathCaption",ExpressionUUID->"0526b846-a9e2-4fbc-9fb6-f2d0474909e5"],

Cell[CellGroupData[{

Cell[BoxData["$ScalarFunctions"], "Input",
 CellLabel->
  "In[1225]:=",ExpressionUUID->"c6ec4450-6498-4cb7-a576-5098e9d26a99"],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{3.728978889907815*^9},
 CellLabel->"Out[1225]=",
 CellID->979026059,ExpressionUUID->"b614e995-1a44-4ce8-9335-e3a5033d2275"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We define a scalar function SF:", "MathCaption",ExpressionUUID->"3d0d2018-13ec-4050-be38-032986c2c27e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefScalarFunction", "[", "SF", "]"}]], "Input",
 CellLabel->
  "In[1226]:=",ExpressionUUID->"099e2601-145b-4497-852a-ec0f65bb57bd"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefScalarFunction", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"scalar function \"\>", "\[InvisibleSpace]", "SF",
    "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefScalarFunction, ": Defining ", 
   "scalar function ", $CellContext`SF, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978890029471*^9},
 CellLabel->"During evaluation of In[1226]:=",
 CellID->471270299,ExpressionUUID->"37e7dc5f-87e0-453e-878c-a1fd963d1594"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ScalarQ", "[", 
  RowBox[{"SF", "[", 
   RowBox[{
    RowBox[{"v", "[", "a", "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}]}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[1227]:=",ExpressionUUID->"dda1ecc1-315b-48ec-acb1-a8b07f754fda"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978890163661*^9},
 CellLabel->"Out[1227]=",
 CellID->720225157,ExpressionUUID->"4d25fdf3-6ab1-4962-a100-77e407725531"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData["$CovDs"], "Input",
 CellLabel->
  "In[1228]:=",ExpressionUUID->"b94db8d1-562d-48c2-940c-2ae087dfc283"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"PD", ",", "CD"}], "}"}]], "Output",
 CellChangeTimes->{3.7289788902924223`*^9},
 CellLabel->"Out[1228]=",
 CellID->1606379207,ExpressionUUID->"e20dedf7-ec3b-481f-9a46-78d73dac1440"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"CD", "[", 
   RowBox[{"-", "a"}], "]"}], "[", 
  RowBox[{"1", "/", 
   RowBox[{"SF", "[", 
    RowBox[{"Scalar", "[", 
     RowBox[{
      RowBox[{"v", "[", "a", "]"}], 
      RowBox[{"v", "[", 
       RowBox[{"-", "a"}], "]"}]}], "]"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1229]:=",ExpressionUUID->"6f56fbdc-a1ea-479f-8cf0-209b86efcb4a"],

Cell[BoxData[
 RowBox[{"-", 
  FractionBox[
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[$CellContext`b],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"b"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`a][
          $CellContext`v[-$CellContext`b]],
         Editable->False], ")"}]}], "+", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {" "},
               {"b"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[-$CellContext`b],
        Editable->False], " ", 
       RowBox[{"(", 
        InterpretationBox[
         StyleBox[
          RowBox[{
           SubscriptBox["\[EmptyDownTriangle]", "a"], GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"b"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}]}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`CD[-$CellContext`a][
          $CellContext`v[$CellContext`b]],
         Editable->False], ")"}]}]}], ")"}], " ", 
    RowBox[{
     SuperscriptBox[
      InterpretationBox[
       StyleBox["SF",
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`SF,
       Editable->False], "\[Prime]",
      MultilineFunction->None], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        StyleBox["(",
         FontColor->RGBColor[1, 0, 0]], 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"a"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[-$CellContext`a],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"a"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[$CellContext`a],
          Editable->False]}], 
        StyleBox[")",
         FontColor->RGBColor[1, 0, 0]]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
      Editable->False], "]"}]}], 
   SuperscriptBox[
    RowBox[{
     InterpretationBox[
      StyleBox["SF",
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`SF,
      Editable->False], "[", 
     InterpretationBox[
      StyleBox[
       RowBox[{
        StyleBox["(",
         FontColor->RGBColor[1, 0, 0]], 
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"a"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[-$CellContext`a],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"a"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[$CellContext`a],
          Editable->False]}], 
        StyleBox[")",
         FontColor->RGBColor[1, 0, 0]]}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
      Editable->False], "]"}], "2"]]}]], "Output",
 CellChangeTimes->{3.728978890459882*^9},
 CellLabel->"Out[1229]=",
 CellID->250921965,ExpressionUUID->"b0b75048-4c72-45ac-bfd8-b7b35a9ac844"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.4. Inert heads", "Subsection",ExpressionUUID->"c8c4f3cf-4587-4a19-9809-f676b4bd4292"],

Cell[TextData[{
 "Sometimes we need to wrap tensors with some head. This is an extremely \
general concept and it is prepared to play the role of an arbitary head with \
the only property of being \"transparent\" to the symmetry and \
canonicalization routines. ",
 StyleBox["Dagger",
  FontFamily->"Courier"],
 " and ",
 StyleBox["ERROR",
  FontFamily->"Courier"],
 " are both inert heads."
}], "Text",ExpressionUUID->"6fa9b11f-a727-49e7-ab64-160e4d699f9a"],

Cell[TextData[{
 StyleBox[ButtonBox["DefInertHead",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "DefInertHead.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Define an inert head\n", "Text"],
 StyleBox[ButtonBox["UndefInertHead",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "UndefInertHead.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Undefine an inert head\n", "Text"],
 StyleBox[ButtonBox["$InertHeads",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "$InertHeads.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["List of defined inert heads\n", "Text"],
 StyleBox[ButtonBox["LinearQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Linear.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Boolean option to specify whether an inert head is linear or not\n\
", "Text"],
 StyleBox[ButtonBox["LinearQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "LinearQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Detect whether an inert head is linear\n", "Text"],
 StyleBox[ButtonBox["ContractThrough",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ContractThrough.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Option to specify which metrics, projectors or delta can be \
contracted through the inert head\n", "Text"],
 StyleBox[ButtonBox["ContractThroughQ",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ContractThroughQ.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Detect whether a metric, projector or delta can be contracted \
through an inert head", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{{3.728978899947547*^9, 
  3.7289789063952723`*^9}},ExpressionUUID->"b80fd34e-e0a6-43b9-aa68-\
531d399ef4e5"],

Cell["Definition of an inert head.", "Text",ExpressionUUID->"80fe4883-cd03-4d3e-865d-0b430de5569c"],

Cell[CellGroupData[{

Cell[TextData[{
 "A prototypical example could be traceless, transverse tensors: ",
 "We define an inert head called ",
 StyleBox["TTPart",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"733b0f1e-ad8e-483f-a535-40f4d4d1f905"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefInertHead", "[", "TTPart", "]"}]], "Input",
 CellLabel->
  "In[1230]:=",ExpressionUUID->"0c98f142-6e48-4840-8e65-030e39381dac"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefInertHead", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inert head \"\>", "\[InvisibleSpace]", "TTPart", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefInertHead, ": Defining ", 
   "inert head ", $CellContext`TTPart, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978915281704*^9},
 CellLabel->"During evaluation of In[1230]:=",
 CellID->1265143671,ExpressionUUID->"1c41b4b3-ce33-4c7b-953b-097bcc707856"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"TTPart", "[", 
   RowBox[{"U", "[", 
    RowBox[{
     RowBox[{"-", "b"}], ",", 
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "a"}]}], "]"}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1231]:=",ExpressionUUID->"09d3db03-2ad7-4b1f-af7d-f8ea2941b7b5"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[
   RowBox[{"TTPart", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {" ", " ", " "},
            {"a", "b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
     Editable->False], "]"}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`TTPart[
   $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978915400893*^9},
 CellLabel->"Out[1231]=",
 CellID->1034089550,ExpressionUUID->"485c6a60-2478-4ceb-a1a7-852406d78846"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UpVectorQ", "[", 
  RowBox[{"TTPart", "[", 
   RowBox[{"v", "[", "a", "]"}], "]"}], "]"}]], "Input",
 CellLabel->
  "In[1232]:=",ExpressionUUID->"51a627cb-2737-4c3e-89fa-136cf80e43e8"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978915532283*^9},
 CellLabel->"Out[1232]=",
 CellID->357911679,ExpressionUUID->"8d453e70-e9b1-4e11-a200-565148176d41"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefInertHead", "[", "TTPart", "]"}]], "Input",
 CellLabel->
  "In[1233]:=",ExpressionUUID->"2253f3f5-0604-4a8a-9fec-4d85c4832100"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefInertHead", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inert head\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "TTPart"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefInertHead, ": Undefined ", "inert head", 
   " ", $CellContext`TTPart],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789156437063`*^9},
 CellLabel->"During evaluation of In[1233]:=",
 CellID->319893706,ExpressionUUID->"710782d0-3167-46f5-8bfb-346a19352fac"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "With help from Guillaume Faye, this concept has been improved in version \
0.9.2 of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " in three different ways:"
}], "Text",ExpressionUUID->"5ca9d54d-ab49-4ed8-b2c0-a9a3faa724ca"],

Cell[CellGroupData[{

Cell[TextData[{
 "1) There is an option ",
 StyleBox["LinearQ",
  FontFamily->"Courier"],
 " to make the inert head linear with respect to constants. The default value \
is ",
 StyleBox["False",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"829ab694-be18-4d4c-93e4-664f484d74c9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefInertHead", "[", 
  RowBox[{"lhead", ",", 
   RowBox[{"LinearQ", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellLabel->
  "In[1234]:=",ExpressionUUID->"e6c54ea5-6f93-448c-9316-36b4d120ff45"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefInertHead", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inert head \"\>", "\[InvisibleSpace]", "lhead", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefInertHead, ": Defining ", 
   "inert head ", $CellContext`lhead, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978915764558*^9},
 CellLabel->"During evaluation of In[1234]:=",
 CellID->1301948301,ExpressionUUID->"fd74ea41-a589-42b0-9ffc-5ade58e61d7a"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"LinearQ", "[", "lhead", "]"}]], "Input",
 CellLabel->
  "In[1235]:=",ExpressionUUID->"43520ac1-3e71-482b-a187-09c76978fd0e"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978915941139*^9},
 CellLabel->"Out[1235]=",
 CellID->1134215624,ExpressionUUID->"075aba0d-81c0-4737-a691-8e3aae9bff16"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"lhead", "[", 
  RowBox[{
   RowBox[{"3", 
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b"}], "]"}]}], "+", 
   RowBox[{"S", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1236]:=",ExpressionUUID->"303b98cd-830c-4a67-9c99-0f3f5fc49e31"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[
    RowBox[{"lhead", "[", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {"a", "b"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[$CellContext`a, $CellContext`b],
      Editable->False], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`lhead[
    $CellContext`S[$CellContext`a, $CellContext`b]],
   Editable->False], "+", 
  RowBox[{"3", " ", 
   InterpretationBox[
    StyleBox[
     RowBox[{"lhead", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, $CellContext`b],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`lhead[
     $CellContext`T[$CellContext`a, $CellContext`b]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289789163518877`*^9},
 CellLabel->"Out[1236]=",
 CellID->1697141796,ExpressionUUID->"ce1cffdd-2177-4ead-8e19-b09f5850d64a"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "2) There is an option ",
 StyleBox["ContractThrough",
  FontFamily->"Courier"],
 " to specify a list of metrics, projectors and/or the delta tensor which can \
be contracted through the inert head. In this example only the delta tensor \
can be contracted:"
}], "MathCaption",ExpressionUUID->"4bdcc9f6-76da-4d61-abf6-3ef98ef65885"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefInertHead", "[", 
  RowBox[{"chead", ",", 
   RowBox[{"ContractThrough", "\[Rule]", 
    RowBox[{"{", "delta", "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1237]:=",ExpressionUUID->"f905aeaf-678d-4a74-b6a5-1019eb025c59"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefInertHead", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inert head \"\>", "\[InvisibleSpace]", "chead", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefInertHead, ": Defining ", 
   "inert head ", $CellContext`chead, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978916682494*^9},
 CellLabel->"During evaluation of In[1237]:=",
 CellID->984365918,ExpressionUUID->"3f38ca6c-52b5-406f-bdac-f2d801bf6e4e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefInertHead", "[", 
  RowBox[{"mhead", ",", 
   RowBox[{"ContractThrough", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"delta", ",", "metricg"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1238]:=",ExpressionUUID->"f0750651-cc9b-4093-9c65-327ac95e491f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefInertHead", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"inert head \"\>", "\[InvisibleSpace]", "mhead", 
   "\[InvisibleSpace]", "\<\". \"\>", "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm[
  "** ", xAct`xTensor`DefInertHead, ": Defining ", 
   "inert head ", $CellContext`mhead, ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897891678382*^9},
 CellLabel->"During evaluation of In[1238]:=",
 CellID->23967338,ExpressionUUID->"0e56d285-1f8f-48c7-850b-9170397720a1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "mhead"}]], "Input",
 CellLabel->
  "In[1239]:=",ExpressionUUID->"3b73ef04-c158-4ad4-ab84-823a579ce6c1"],

Cell[CellGroupData[{

Cell["Global`mhead", "Print", "PrintUsage",
 CellChangeTimes->{3.728978917134647*^9},
 CellTags->"Info12393728982516-9181623",
 CellID->1239607777,ExpressionUUID->"f7c28f10-04f1-4f64-97a8-4dd64ff8b439"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{"mhead", "/:", 
         RowBox[{"ContractThroughQ", "[", 
          RowBox[{"mhead", ",", "delta"}], "]"}], "=", "True"}]},
       {" "},
       {
        RowBox[{"mhead", "/:", 
         RowBox[{"ContractThroughQ", "[", 
          RowBox[{"mhead", ",", "metricg"}], "]"}], "=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"DefInfo", "[", "mhead", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{"\<\"inert head\"\>", ",", "\<\"\"\>"}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"InertHeadQ", "[", "mhead", "]"}], "^=", "True"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"LinearQ", "[", "mhead", "]"}], "^=", "False"}]},
       {" "},
       {
        RowBox[{
         RowBox[{"PrintAs", "[", "mhead", "]"}], "^=", "\<\"mhead\"\>"}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`mhead],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789171564713`*^9},
 CellTags->"Info12393728982516-9181623",
 CellID->1349048492,ExpressionUUID->"e7556ba5-d1ca-4503-aebf-03e321a473f9"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compare now these cases:", "MathCaption",ExpressionUUID->"0de789bd-b2b7-4ae0-b428-8fae5544dcd6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"delta", "[", 
   RowBox[{"a", ",", 
    RowBox[{"-", "b"}]}], "]"}], 
  RowBox[{"{", 
   RowBox[{
    RowBox[{"lhead", "[", 
     RowBox[{"T", "[", "b", "]"}], "]"}], ",", 
    RowBox[{"chead", "[", 
     RowBox[{"T", "[", "b", "]"}], "]"}], ",", 
    RowBox[{"mhead", "[", 
     RowBox[{"T", "[", "b", "]"}], "]"}]}], "}"}]}]], "Input",
 CellLabel->
  "In[1240]:=",ExpressionUUID->"5f5df07e-aaf0-498e-93af-7fc09637ba77"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"a", " "},
            {" ", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`delta[$CellContext`a, -$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[
      RowBox[{"lhead", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[$CellContext`b],
        Editable->False], "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lhead[
      $CellContext`T[$CellContext`b]],
     Editable->False]}], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"chead", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`chead[
     $CellContext`T[$CellContext`a]],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"mhead", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`mhead[
     $CellContext`T[$CellContext`a]],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978917804023*^9},
 CellLabel->"Out[1240]=",
 CellID->325660018,ExpressionUUID->"8fd22edf-e763-4e7a-9982-770731dd7135"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ContractMetric", "[", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"lhead", "[", 
      RowBox[{"T", "[", "b", "]"}], "]"}], ",", 
     RowBox[{"chead", "[", 
      RowBox[{"T", "[", "b", "]"}], "]"}], ",", 
     RowBox[{"mhead", "[", 
      RowBox[{"T", "[", "b", "]"}], "]"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1241]:=",ExpressionUUID->"9da0f548-1b46-4408-89ea-da4e96b50aa7"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"lhead", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[$CellContext`b],
        Editable->False], "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`lhead[
      $CellContext`T[$CellContext`b]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[-$CellContext`a, -$CellContext`b],
     Editable->False]}], ",", 
   RowBox[{
    InterpretationBox[
     StyleBox[
      RowBox[{"chead", "[", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[$CellContext`b],
        Editable->False], "]"}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`chead[
      $CellContext`T[$CellContext`b]],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"g", 
         StyleBox[GridBox[{
            {" ", " "},
            {"a", "b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`metricg[-$CellContext`a, -$CellContext`b],
     Editable->False]}], ",", 
   InterpretationBox[
    StyleBox[
     RowBox[{"mhead", "[", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {" "},
              {"a"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[-$CellContext`a],
       Editable->False], "]"}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`mhead[
     $CellContext`T[-$CellContext`a]],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978918182239*^9},
 CellLabel->"Out[1241]=",
 CellID->1579589970,ExpressionUUID->"bc61fdda-48b1-430e-9c51-339643791517"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "3) It is possible to have additional arguments, including arguments with \
synchronized lists of indices. This does not even have to be specified as an \
argument. The indices must be surrounded with the head ",
 StyleBox["IndexList",
  FontFamily->"Courier"],
 ". The system keeps a one to one relation between the indices in the \
tensorial expression and the indices in the lists of arguments. This \
correspondence is based on the slots: for example in the following expression \
the IndexList marks the {3, 2} slots of the tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 ", and this is kept through the use of the ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " functions:"
}], "MathCaption",ExpressionUUID->"fafb08f7-c6c0-4df5-bf1f-b1c9adc35881"],

Cell[BoxData[
 RowBox[{
  RowBox[{"expr", "=", 
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "d"}], "]"}], 
    RowBox[{"mhead", "[", 
     RowBox[{
      RowBox[{"U", "[", 
       RowBox[{"b", ",", "c", ",", "a"}], "]"}], ",", 
      RowBox[{"{", 
       RowBox[{"a", ",", "c"}], "}"}], ",", 
      RowBox[{"IndexList", "[", 
       RowBox[{"a", ",", "c"}], "]"}]}], "]"}]}]}], ";"}]], "Input",
 CellLabel->
  "In[1242]:=",ExpressionUUID->"af717f45-7c64-4008-a3e4-1dd89b2e3a77"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ReplaceIndex", "[", 
   RowBox[{
    RowBox[{"Evaluate", "[", "expr", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"c", "\[Rule]", "f"}], "}"}]}], "]"}], "//", 
  "InputForm"}]], "Input",
 CellLabel->
  "In[1243]:=",ExpressionUUID->"8d33abb9-e21c-4b36-8f7a-7e5246b821c5"],

Cell["mhead[U[b, f, a], {a, c}, IndexList[a, f]]*U[-a, -b, d]", "Output",
 CellChangeTimes->{3.72897891839081*^9},
 CellLabel->"Out[1243]//InputForm=",
 CellID->975781576,ExpressionUUID->"0e957366-467e-475f-a9dc-fcdda8548d12"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ToCanonical", "[", "expr", "]"}], "//", "InputForm"}]], "Input",
 CellLabel->
  "In[1244]:=",ExpressionUUID->"dec5f89f-2b59-49f1-a9ee-bd96d8e00a63"],

Cell["mhead[U[c, -a, -b], {a, c}, IndexList[-b, -a]]*U[d, a, b]", "Output",
 CellChangeTimes->{3.728978918517589*^9},
 CellLabel->"Out[1244]//InputForm=",
 CellID->1775819340,ExpressionUUID->"5f9a3770-0f25-42b5-8f7a-40e80437800d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"metricg", "[", 
     RowBox[{
      RowBox[{"-", "f"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], "expr"}], "//", "ContractMetric"}], "//", 
  "InputForm"}]], "Input",
 CellLabel->
  "In[1245]:=",ExpressionUUID->"0801157d-4b09-4d11-a643-47b08cdad97b"],

Cell["mhead[U[b, -f, a], {a, c}, IndexList[a, -f]]*U[-a, -b, d]", "Output",
 CellChangeTimes->{3.728978918650518*^9},
 CellLabel->"Out[1245]//InputForm=",
 CellID->467921374,ExpressionUUID->"eb47b289-562a-4c4e-8def-7e3461fddb49"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Perhaps there should be something more general than slot-syncronization, but \
we do not know of a complete set of possibilities, and no other case has been \
asked for until now. \
\>", "MathCaption",ExpressionUUID->"c71baacf-2edb-4e19-944d-8b633139f9a8"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefInertHead", "[", "lhead", "]"}]], "Input",
 CellLabel->
  "In[1246]:=",ExpressionUUID->"0d35e924-327e-4796-9a6e-8f987cd0a6bc"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefInertHead", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inert head\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "lhead"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefInertHead, ": Undefined ", "inert head", 
   " ", $CellContext`lhead],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978918794364*^9},
 CellLabel->"During evaluation of In[1246]:=",
 CellID->269417504,ExpressionUUID->"26e8e7fe-d6e6-4988-b7c6-c067df673f04"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefInertHead", "[", "chead", "]"}]], "Input",
 CellLabel->
  "In[1247]:=",ExpressionUUID->"035b9896-a7a9-4761-90e6-8aa4568c4b5f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefInertHead", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inert head\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "chead"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefInertHead, ": Undefined ", "inert head", 
   " ", $CellContext`chead],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978918915001*^9},
 CellLabel->"During evaluation of In[1247]:=",
 CellID->1204461852,ExpressionUUID->"19679f16-102d-4247-bcfc-3ce506894d58"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefInertHead", "[", "mhead", "]"}]], "Input",
 CellLabel->
  "In[1248]:=",ExpressionUUID->"0b974f2c-c0aa-4392-993d-12c29641e2c0"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefInertHead", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"inert head\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "mhead"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefInertHead, ": Undefined ", "inert head", 
   " ", $CellContext`mhead],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978919049817*^9},
 CellLabel->"During evaluation of In[1248]:=",
 CellID->1209489942,ExpressionUUID->"1f9a0f91-7e95-4756-b51b-1772a43adfd8"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.5. Monomials", "Subsection",ExpressionUUID->"0e34af1e-3617-4ec4-b436-adf5525614c9"],

Cell[TextData[{
 "A monomial is defined as a product of tensors such that it cannot be \
further reduced into products with different dummies. The function ",
 StyleBox["BreakInMonomials",
  FontFamily->"Courier"],
 " decomposes an arbitrary expression in monomials with (inert) head ",
 StyleBox["Monomial",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"402e611c-deb3-472c-be07-5789b9c53c4d"],

Cell[CellGroupData[{

Cell["\<\
We break this expressions in monomials. Note that scalars are left outside:\
\>", "MathCaption",ExpressionUUID->"09956d3d-9850-4f27-aa8b-7ae3db749c6f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"7", 
  RowBox[{
   RowBox[{"r", "[", "]"}], "^", "2"}], 
  RowBox[{"T", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
  RowBox[{"U", "[", 
   RowBox[{
    RowBox[{"-", "b"}], ",", 
    RowBox[{"-", "c"}], ",", 
    RowBox[{"-", "e"}]}], "]"}], 
  RowBox[{"v", "[", "e", "]"}], 
  RowBox[{"T", "[", 
   RowBox[{"f", ",", "g", ",", "h"}], "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"-", "g"}], "]"}], 
  RowBox[{"v", "[", 
   RowBox[{"-", "h"}], "]"}]}]], "Input",
 CellLabel->
  "In[1249]:=",ExpressionUUID->"7c7c9070-81c0-4189-9870-4d83f7f5365f"],

Cell[BoxData[
 RowBox[{"7", " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"a", "b", "c"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"T", 
       StyleBox[GridBox[{
          {"f", "g", "h"},
          {" ", " ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`T[$CellContext`f, $CellContext`g, $CellContext`h],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"b", "c", "e"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {"e"},
          {" "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[$CellContext`e],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"g"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`g],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"h"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`h],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289789289684258`*^9},
 CellLabel->"Out[1249]=",
 CellID->165667155,ExpressionUUID->"8299ace9-4766-42e4-97f6-8c209bd7fae8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BreakInMonomials", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1250]:=",ExpressionUUID->"d3a7ef85-a5c2-4362-9b4a-4f2c9f3da99c"],

Cell[BoxData[
 RowBox[{"7", " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{"Monomial", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b", "c"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {"b", "c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"e"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`e],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Monomial[($CellContext`T[$CellContext`a, $CellContext`b, \
$CellContext`c] $CellContext`U[-$CellContext`b, -$CellContext`c, \
-$CellContext`e]) $CellContext`v[$CellContext`e]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{"Monomial", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "g", "h"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`g, $CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`h],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Monomial[($CellContext`T[$CellContext`f, $CellContext`g, \
$CellContext`h] $CellContext`v[-$CellContext`g]) \
$CellContext`v[-$CellContext`h]],
   Editable->False], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.72897892913451*^9},
 CellLabel->"Out[1250]=",
 CellID->1229067573,ExpressionUUID->"28717485-f25e-496d-bde3-2a9c5e7369b4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Now for example each monomial could be canonicalized independently:\
\>", "MathCaption",ExpressionUUID->"bbc36c11-f5b2-4dd2-ba3d-e4582c404f3e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "/.", 
  RowBox[{"expr_Monomial", "\[RuleDelayed]", 
   RowBox[{"ToCanonical", "[", "expr", "]"}]}]}]], "Input",
 CellLabel->
  "In[1251]:=",ExpressionUUID->"527aeb2b-6fdd-4aee-b18d-2c4a04a06779"],

Cell[BoxData[
 RowBox[{"7", " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{"Monomial", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b", "c"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"U", 
           StyleBox[GridBox[{
              {" ", " ", " "},
              {"b", "c", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`e],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"e"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`e],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Monomial[($CellContext`T[$CellContext`a, $CellContext`b, \
$CellContext`c] $CellContext`U[-$CellContext`b, -$CellContext`c, \
-$CellContext`e]) $CellContext`v[$CellContext`e]],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[
    RowBox[{"Monomial", "[", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"f", "g", "h"},
              {" ", " ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`f, $CellContext`g, $CellContext`h],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"g"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`g],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"h"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`h],
       Editable->False]}], "]"}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`Monomial[($CellContext`T[$CellContext`f, $CellContext`g, \
$CellContext`h] $CellContext`v[-$CellContext`g]) \
$CellContext`v[-$CellContext`h]],
   Editable->False], " ", 
  SuperscriptBox[
   InterpretationBox[
    StyleBox["r",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`r[],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.7289789292687607`*^9},
 CellLabel->"Out[1251]=",
 CellID->293498918,ExpressionUUID->"b6ed1c5f-f169-46d6-9739-85ac80ec2fef"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 StyleBox[ButtonBox["Monomial",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Monomial.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Head for a monomial expression\n", "Text"],
 StyleBox[ButtonBox["BreakInMonomials",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "BreakInMonomials.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Separate a term in monomials", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.7289789266021423`*^9},ExpressionUUID->"d5b56b42-ce19-4ca0-985a-\
bebb332d43f7"],

Cell["Monomials.", "Text",ExpressionUUID->"e9611c4b-44b8-43e7-8bb6-15f12df0cd6c"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
 "10.6. Symmetrization of an expression. ",
 StyleBox["V1.1.0 UPDATE",
  FontColor->RGBColor[0, 0, 1]]
}], "Subsection",ExpressionUUID->"b6edd169-b03e-4e7a-b1b9-5fc72a028f71"],

Cell[TextData[{
 StyleBox[ButtonBox["ImposeSymmetry",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "ImposeSymmetry.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Symmetrize an expression as given by a permutations group\n", 
  "Text"],
 StyleBox[ButtonBox["Symmetrize",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Symmetrize.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t\t", "DisplayMath"],
 StyleBox["Symmetrize a set of indices\n", "Text"],
 StyleBox[ButtonBox["Antisymmetrize",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "Antisymmetrize.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Antisymmetrize a set of indices\n", "Text"],
 StyleBox[ButtonBox["PairSymmetrize",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PairSymmetrize.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Symmetrize a set of pairs of indices\n", "Text"],
 StyleBox[ButtonBox["PairAntisymmetrize",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "PairAntisymmetrize.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t\t", "DisplayMath"],
 StyleBox["Antisymmetrize a set of pairs of indices", "Text"]
}], "DefinitionBox",
 CellChangeTimes->{
  3.728978933664846*^9},ExpressionUUID->"73e3778f-e53c-431f-8c3c-\
8d8bc89db3b9"],

Cell["Symmetrization.", "Text",
 CellChangeTimes->{
  3.728978938751334*^9},ExpressionUUID->"85735a22-355f-40e2-8b2a-\
46343efddcdd"],

Cell[CellGroupData[{

Cell[TextData[{
 "Symmetrization of an expression under a given group of permutations. \
Numbers on the third argument refer to positions in the second argument of ",
 StyleBox["ImposeSymmetry",
  FontFamily->"Courier"],
 ". Note that ",
 StyleBox["ToCanonical",
  FontFamily->"Courier"],
 " is always applied after the symmetrization functions in order to take into \
account the antisymmetry of the tensor ",
 StyleBox["U",
  FontFamily->"Courier"],
 ". It is not automatically implemented into ",
 StyleBox["ImposeSymmetry",
  FontFamily->"Courier"],
 " because for large numbers of indices it would take ages to canonicalize \
all terms!"
}], "MathCaption",ExpressionUUID->"45881c2e-a25a-4c3e-adec-dab3b227f836"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ImposeSymmetry", "[", 
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "}"}], ",", 
    RowBox[{"Symmetric", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[1252]:=",ExpressionUUID->"6db88852-8d91-4836-8e13-0c99f0492e09"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978946460074*^9},
 CellLabel->"Out[1252]=",
 CellID->2054448071,ExpressionUUID->"d396edad-297d-418b-a164-a99b51a24390"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ImposeSymmetry", "[", 
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}]}], "}"}], ",", 
    RowBox[{"Antisymmetric", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[1253]:=",ExpressionUUID->"0ea0dd93-d9f4-4954-b0f8-ebd2fc0f4def"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"U", 
      StyleBox[GridBox[{
         {" ", " ", " "},
         {"a", "b", "c"}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
  Editable->False]], "Output",
 CellChangeTimes->{3.7289789465938*^9},
 CellLabel->"Out[1253]=",
 CellID->1866652831,ExpressionUUID->"1d66ce3e-88c0-4c36-8404-fe8cc604c413"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ImposeSymmetry", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "f"}]}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}]}], "}"}], ",", 
    RowBox[{"Antisymmetric", "[", 
     RowBox[{"{", 
      RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}], "//", 
  "ToCanonical"}]], "Input",
 CellLabel->
  "In[1254]:=",ExpressionUUID->"bef4318e-b1bf-4f76-8142-29874ecfdf54"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "2"]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`f],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978946711246*^9},
 CellLabel->"Out[1254]=",
 CellID->245531594,ExpressionUUID->"9a00163f-9ce7-4453-84ef-852fe2a9d06d"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"xAct`xTensor`Symmetrize", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "f"}]}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}]}], "}"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1255]:=",ExpressionUUID->"faa75994-6051-451d-ac35-84b6a83f3f9f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`f],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978946830111*^9},
 CellLabel->"Out[1255]=",
 CellID->1333414471,ExpressionUUID->"4c6aaaad-b777-4d4d-b638-3702d1162a9e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"xAct`xTensor`Symmetrize", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "f"}]}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "e"}], ",", 
      RowBox[{"-", "d"}]}], "}"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1256]:=",ExpressionUUID->"900abf6c-ba1a-4dd7-8be9-77842b2c3b31"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.72897894694354*^9},
 CellLabel->"Out[1256]=",
 CellID->1705157652,ExpressionUUID->"0db8d8ea-03b3-40fc-861e-d83d0eac35a9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Antisymmetrize", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "f"}]}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "d"}]}], "}"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1257]:=",ExpressionUUID->"8757895e-9503-4769-9c8d-6e4204d8916e"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"-", 
    FractionBox["1", "2"]}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`f],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289789470636787`*^9},
 CellLabel->"Out[1257]=",
 CellID->415387949,ExpressionUUID->"724ee41d-e8e6-4cf7-8120-8286680d2466"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Antisymmetrize", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "a"}], ",", 
       RowBox[{"-", "b"}], ",", 
       RowBox[{"-", "c"}]}], "]"}], 
     RowBox[{"U", "[", 
      RowBox[{
       RowBox[{"-", "d"}], ",", 
       RowBox[{"-", "e"}], ",", 
       RowBox[{"-", "f"}]}], "]"}]}], ",", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "e"}], ",", 
      RowBox[{"-", "d"}]}], "}"}]}], "]"}], "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1258]:=",ExpressionUUID->"cdad5b95-38bb-4284-859a-2d882a7e188c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"a", "b", "c"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"U", 
       StyleBox[GridBox[{
          {" ", " ", " "},
          {"d", "e", "f"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`f],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978947194907*^9},
 CellLabel->"Out[1258]=",
 CellID->2103406093,ExpressionUUID->"11875416-cceb-4de8-8aed-8348f93eefd8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ps", "=", 
  RowBox[{
   RowBox[{"PairSymmetrize", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"U", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "c"}]}], "]"}], 
      RowBox[{"U", "[", 
       RowBox[{
        RowBox[{"-", "d"}], ",", 
        RowBox[{"-", "e"}], ",", 
        RowBox[{"-", "f"}]}], "]"}]}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "b"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "c"}], ",", 
         RowBox[{"-", "d"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "e"}], ",", 
         RowBox[{"-", "f"}]}], "}"}]}], "}"}]}], "]"}], "//", 
   "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[1259]:=",ExpressionUUID->"0caf2cec-03b9-41c6-b0b7-44ca8797b0e9"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`d],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`e, -$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`c, -$CellContext`d, -$CellContext`e],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "d", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`c, -$CellContext`d, -$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`c, -$CellContext`e, -$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`f],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978947332575*^9},
 CellLabel->"Out[1259]=",
 CellID->70408137,ExpressionUUID->"90ab9d32-887c-4491-8319-88251615f0c6"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"pa", "=", 
  RowBox[{
   RowBox[{"PairAntisymmetrize", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"U", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", 
        RowBox[{"-", "b"}], ",", 
        RowBox[{"-", "c"}]}], "]"}], 
      RowBox[{"U", "[", 
       RowBox[{
        RowBox[{"-", "d"}], ",", 
        RowBox[{"-", "e"}], ",", 
        RowBox[{"-", "f"}]}], "]"}]}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "a"}], ",", 
         RowBox[{"-", "b"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "c"}], ",", 
         RowBox[{"-", "d"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "e"}], ",", 
         RowBox[{"-", "f"}]}], "}"}]}], "}"}]}], "]"}], "//", 
   "ToCanonical"}]}]], "Input",
 CellLabel->
  "In[1260]:=",ExpressionUUID->"9b23fd11-fbd7-44af-850e-1f3bb0c1df28"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`e, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`c, -$CellContext`d],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "c", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`c, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"b", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`b, -$CellContext`e, -$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`f],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`c, -$CellContext`d, -$CellContext`e],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "d", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`c, -$CellContext`d, -$CellContext`f],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"c", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`c, -$CellContext`e, -$CellContext`f],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"a", "b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`a, -$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {" ", " ", " "},
           {"d", "e", "f"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[-$CellContext`d, -$CellContext`e, -$CellContext`f],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978947452621*^9},
 CellLabel->"Out[1260]=",
 CellID->86069062,ExpressionUUID->"925adde0-ab79-4973-9571-bad63270a556"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"ps", "-", 
   RowBox[{"ReplaceIndex", "[", 
    RowBox[{
     RowBox[{"Evaluate", "[", "ps", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "a"}], "\[Rule]", 
        RowBox[{"-", "c"}]}], ",", 
       RowBox[{
        RowBox[{"-", "b"}], "\[Rule]", 
        RowBox[{"-", "d"}]}], ",", 
       RowBox[{
        RowBox[{"-", "c"}], "\[Rule]", 
        RowBox[{"-", "a"}]}], ",", 
       RowBox[{
        RowBox[{"-", "d"}], "\[Rule]", 
        RowBox[{"-", "b"}]}]}], "}"}]}], "]"}]}], "//", 
  "Simplification"}]], "Input",
 CellLabel->
  "In[1261]:=",ExpressionUUID->"b2bb5f3b-01ee-4704-823a-3e64c9f85999"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978947567192*^9},
 CellLabel->"Out[1261]=",
 CellID->1306230603,ExpressionUUID->"9aa14bf2-d9bb-464d-8b30-3b7d7a413df8"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"pa", "+", 
   RowBox[{"ReplaceIndex", "[", 
    RowBox[{
     RowBox[{"Evaluate", "[", "pa", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "a"}], "\[Rule]", 
        RowBox[{"-", "c"}]}], ",", 
       RowBox[{
        RowBox[{"-", "b"}], "\[Rule]", 
        RowBox[{"-", "d"}]}], ",", 
       RowBox[{
        RowBox[{"-", "c"}], "\[Rule]", 
        RowBox[{"-", "a"}]}], ",", 
       RowBox[{
        RowBox[{"-", "d"}], "\[Rule]", 
        RowBox[{"-", "b"}]}]}], "}"}]}], "]"}]}], "//", 
  "Simplification"}]], "Input",
 CellLabel->
  "In[1262]:=",ExpressionUUID->"bc453b2e-865f-4a35-8d70-40c82ff3a3dc"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289789476825037`*^9},
 CellLabel->"Out[1262]=",
 CellID->92421340,ExpressionUUID->"9eedb2cd-8357-4c82-a7d3-36c54517c407"]
}, Open  ]]
}, Open  ]],

Cell["\<\
Working with Gdelta implements certain identities involving many terms, which \
are even unknown to ToCanonical. For example, for a Gdelta with antisymmetric \
blocks of n indices, antisymmetrizing over more than n indices gives zero. \
ToCanonical does not know it, but the contraction rules encode that \
information. We develop this example in arbitrary dimension.\
\>", "Text",ExpressionUUID->"a73525ef-2616-4b35-89ec-1c3b9a0fecbb"],

Cell[CellGroupData[{

Cell["Define a manifold with metric in arbitrary dimension:", "MathCaption",ExpressionUUID->"950e0afd-9b17-420a-b2d1-5e545b7d6e77"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "False"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{"DefConstantSymbol", "[", "dim", "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefManifold", "[", 
  RowBox[{"Mdim", ",", "dim", ",", 
   RowBox[{"IndexRange", "[", 
    RowBox[{"\[FormalA]", ",", "\[FormalZ]"}], "]"}]}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{"DefMetric", "[", 
  RowBox[{"1", ",", 
   RowBox[{"metric", "[", 
    RowBox[{
     RowBox[{"-", "\[FormalA]"}], ",", 
     RowBox[{"-", "\[FormalB]"}]}], "]"}], ",", "cd", ",", 
   RowBox[{"PrintAs", "->", "\"\<g\>\""}]}], "]"}]}], "Input",
 CellLabel->
  "In[1263]:=",ExpressionUUID->"7b1d9279-3b3d-4e63-8459-b858cbbbe352"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"\[FormalA]\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 1265, 101, 19253573948027831207, 
   "Local", "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978947884632*^9},
 CellLabel->"During evaluation of In[1263]:=",
 CellID->694784599,ExpressionUUID->"1e5ba656-853d-4cb1-9859-7451c0619516"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"\[FormalB]\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 1265, 102, 19253573948027831207, 
   "Local", "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978947912162*^9},
 CellLabel->"During evaluation of In[1263]:=",
 CellID->1376166654,ExpressionUUID->"27f5f983-3518-454b-9de1-eaebcbfdee37"],

Cell[BoxData[
 TemplateBox[{
  "ValidateSymbol", "capital", 
   "\"System name \\!\\(\\*RowBox[{\\\"\\\\\\\"\[FormalC]\\\\\\\"\\\"}]\\) is \
overloaded as an abstract index.\"", 2, 1265, 103, 19253573948027831207, 
   "Local", "xAct`xCore`ValidateSymbol"},
  "MessageTemplate2"]], "Message", "MSG",
 CellChangeTimes->{3.728978947934404*^9},
 CellLabel->"During evaluation of In[1263]:=",
 CellID->834075588,ExpressionUUID->"7af15651-3b69-4fc1-92e8-26e481dae68b"],

Cell[BoxData[
 TemplateBox[{
  "General", "stop", 
   "\"Further output of \\!\\(\\*StyleBox[RowBox[{\\\"ValidateSymbol\\\", \
\\\"::\\\", \\\"capital\\\"}], \\\"MessageName\\\"]\\) will be suppressed \
during this calculation.\"", 2, 1265, 104, 19253573948027831207, "Local"},
  "MessageTemplate"]], "Message", "MSG",
 CellChangeTimes->{3.728978947956596*^9},
 CellLabel->"During evaluation of In[1263]:=",
 CellID->1682643547,ExpressionUUID->"419066dd-b051-44e9-963e-afb76873c5a5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Take a Gdelta, with antisymmetric blocks of 3 indices:", "MathCaption",ExpressionUUID->"b1e84f57-0c66-4a49-9b76-5f81a479a4e2"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Gdelta", "[", 
  RowBox[{
   RowBox[{"-", "\[FormalA]"}], ",", 
   RowBox[{"-", "\[FormalB]"}], ",", 
   RowBox[{"-", "\[FormalC]"}], ",", "\[FormalD]", ",", "\[FormalE]", ",", 
   "\[FormalF]"}], "]"}]], "Input",
 CellLabel->
  "In[1267]:=",ExpressionUUID->"1dbc5b7f-301c-4085-a160-bb5805a507dc"],

Cell[BoxData[
 InterpretationBox[
  StyleBox[GridBox[{
     {"\[Delta]", 
      StyleBox[GridBox[{
         {" ", " ", " ", "\[FormalD]", "\[FormalE]", "\[FormalF]"},
         {"\[FormalA]", "\[FormalB]", "\[FormalC]", " ", " ", " "}
        },
        GridBoxSpacings->{"Columns" -> {
            Offset[0.], {
             Offset[0.034999999999999996`]}, 
            Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
             Offset[0.]}}, "RowsIndexed" -> {}}],
       FontSize->9]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.034999999999999996`]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}],
   ShowAutoStyles->False,
   AutoSpacing->False],
  xAct`xTensor`Gdelta[-\[FormalA], -\[FormalB], -\[FormalC], \[FormalD], \
\[FormalE], \[FormalF]],
  Editable->False]], "Output",
 CellChangeTimes->{3.728978948670308*^9},
 CellLabel->"Out[1267]=",
 CellID->2125430733,ExpressionUUID->"14153213-4d30-4321-9e80-97385e5d71b7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Antisymmetrize in four indices, two in each block:", "MathCaption",ExpressionUUID->"7f4e43bb-ff53-4605-afd6-c02bb4b344d6"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Antisymmetrize", "[", 
  RowBox[{"%", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "\[FormalA]"}], ",", 
     RowBox[{"-", "\[FormalB]"}], ",", "\[FormalE]", ",", "\[FormalF]"}], 
    "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1268]:=",ExpressionUUID->"3ade80a8-faf8-43f3-ab48-f545d73c14fb"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "24"], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", " ", " ", "\[FormalD]", "\[FormalE]", "\[FormalF]"},
            {"\[FormalA]", "\[FormalB]", "\[FormalC]", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalA], -\[FormalB], -\[FormalC], \[FormalD], \
\[FormalE], \[FormalF]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", " ", " ", "\[FormalD]", "\[FormalF]", "\[FormalE]"},
            {"\[FormalA]", "\[FormalB]", "\[FormalC]", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalA], -\[FormalB], -\[FormalC], \[FormalD], \
\[FormalF], \[FormalE]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalE]", " ", "\[FormalD]", " ", "\[FormalF]"},
            {"\[FormalA]", " ", "\[FormalC]", " ", "\[FormalB]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalA], \[FormalE], -\[FormalC], \[FormalD], -\
\[FormalB], \[FormalF]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalE]", " ", "\[FormalD]", "\[FormalF]", " "},
            {"\[FormalA]", " ", "\[FormalC]", " ", " ", "\[FormalB]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalA], \[FormalE], -\[FormalC], \[FormalD], \
\[FormalF], -\[FormalB]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalF]", " ", "\[FormalD]", " ", "\[FormalE]"},
            {"\[FormalA]", " ", "\[FormalC]", " ", "\[FormalB]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalA], \[FormalF], -\[FormalC], \[FormalD], -\
\[FormalB], \[FormalE]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalF]", " ", "\[FormalD]", "\[FormalE]", " "},
            {"\[FormalA]", " ", "\[FormalC]", " ", " ", "\[FormalB]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalA], \[FormalF], -\[FormalC], \[FormalD], \
\[FormalE], -\[FormalB]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", " ", " ", "\[FormalD]", "\[FormalE]", "\[FormalF]"},
            {"\[FormalB]", "\[FormalA]", "\[FormalC]", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalB], -\[FormalA], -\[FormalC], \[FormalD], \
\[FormalE], \[FormalF]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", " ", " ", "\[FormalD]", "\[FormalF]", "\[FormalE]"},
            {"\[FormalB]", "\[FormalA]", "\[FormalC]", " ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalB], -\[FormalA], -\[FormalC], \[FormalD], \
\[FormalF], \[FormalE]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalE]", " ", "\[FormalD]", " ", "\[FormalF]"},
            {"\[FormalB]", " ", "\[FormalC]", " ", "\[FormalA]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalB], \[FormalE], -\[FormalC], \[FormalD], -\
\[FormalA], \[FormalF]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalE]", " ", "\[FormalD]", "\[FormalF]", " "},
            {"\[FormalB]", " ", "\[FormalC]", " ", " ", "\[FormalA]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalB], \[FormalE], -\[FormalC], \[FormalD], \
\[FormalF], -\[FormalA]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalF]", " ", "\[FormalD]", " ", "\[FormalE]"},
            {"\[FormalB]", " ", "\[FormalC]", " ", "\[FormalA]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalB], \[FormalF], -\[FormalC], \[FormalD], -\
\[FormalA], \[FormalE]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {" ", "\[FormalF]", " ", "\[FormalD]", "\[FormalE]", " "},
            {"\[FormalB]", " ", "\[FormalC]", " ", " ", "\[FormalA]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[-\[FormalB], \[FormalF], -\[FormalC], \[FormalD], \
\[FormalE], -\[FormalA]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalE]", " ", " ", "\[FormalD]", " ", "\[FormalF]"},
            {" ", "\[FormalA]", "\[FormalC]", " ", "\[FormalB]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalE], -\[FormalA], -\[FormalC], \[FormalD], -\
\[FormalB], \[FormalF]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalE]", " ", " ", "\[FormalD]", "\[FormalF]", " "},
            {" ", "\[FormalA]", "\[FormalC]", " ", " ", "\[FormalB]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalE], -\[FormalA], -\[FormalC], \[FormalD], \
\[FormalF], -\[FormalB]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalE]", " ", " ", "\[FormalD]", " ", "\[FormalF]"},
            {" ", "\[FormalB]", "\[FormalC]", " ", "\[FormalA]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalE], -\[FormalB], -\[FormalC], \[FormalD], -\
\[FormalA], \[FormalF]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalE]", " ", " ", "\[FormalD]", "\[FormalF]", " "},
            {" ", "\[FormalB]", "\[FormalC]", " ", " ", "\[FormalA]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalE], -\[FormalB], -\[FormalC], \[FormalD], \
\[FormalF], -\[FormalA]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalE]", "\[FormalF]", " ", "\[FormalD]", " ", " "},
            {" ", " ", "\[FormalC]", " ", "\[FormalA]", "\[FormalB]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalE], \[FormalF], -\[FormalC], \[FormalD], -\
\[FormalA], -\[FormalB]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalE]", "\[FormalF]", " ", "\[FormalD]", " ", " "},
            {" ", " ", "\[FormalC]", " ", "\[FormalB]", "\[FormalA]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalE], \[FormalF], -\[FormalC], \[FormalD], -\
\[FormalB], -\[FormalA]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalF]", " ", " ", "\[FormalD]", " ", "\[FormalE]"},
            {" ", "\[FormalA]", "\[FormalC]", " ", "\[FormalB]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalF], -\[FormalA], -\[FormalC], \[FormalD], -\
\[FormalB], \[FormalE]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalF]", " ", " ", "\[FormalD]", "\[FormalE]", " "},
            {" ", "\[FormalA]", "\[FormalC]", " ", " ", "\[FormalB]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalF], -\[FormalA], -\[FormalC], \[FormalD], \
\[FormalE], -\[FormalB]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalF]", " ", " ", "\[FormalD]", " ", "\[FormalE]"},
            {" ", "\[FormalB]", "\[FormalC]", " ", "\[FormalA]", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalF], -\[FormalB], -\[FormalC], \[FormalD], -\
\[FormalA], \[FormalE]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalF]", " ", " ", "\[FormalD]", "\[FormalE]", " "},
            {" ", "\[FormalB]", "\[FormalC]", " ", " ", "\[FormalA]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalF], -\[FormalB], -\[FormalC], \[FormalD], \
\[FormalE], -\[FormalA]],
     Editable->False], "-", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalF]", "\[FormalE]", " ", "\[FormalD]", " ", " "},
            {" ", " ", "\[FormalC]", " ", "\[FormalA]", "\[FormalB]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalF], \[FormalE], -\[FormalC], \[FormalD], -\
\[FormalA], -\[FormalB]],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"\[Delta]", 
         StyleBox[GridBox[{
            {"\[FormalF]", "\[FormalE]", " ", "\[FormalD]", " ", " "},
            {" ", " ", "\[FormalC]", " ", "\[FormalB]", "\[FormalA]"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Gdelta[\[FormalF], \[FormalE], -\[FormalC], \[FormalD], -\
\[FormalB], -\[FormalA]],
     Editable->False]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728978948865247*^9},
 CellLabel->"Out[1268]=",
 CellID->871345333,ExpressionUUID->"e35a0581-084e-4e22-8ad8-894c60e786e4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["ToCanonical cannot see that this is zero:", "MathCaption",ExpressionUUID->"daedf9d1-108b-4f60-ab20-e7951fb9d818"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1269]:=",ExpressionUUID->"07a7b17d-7701-4eb7-a8c2-59d87678b067"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", " ", " ", "\[FormalD]", "\[FormalE]", "\[FormalF]"},
           {"\[FormalA]", "\[FormalB]", "\[FormalC]", " ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-\[FormalA], -\[FormalB], -\[FormalC], \[FormalD], \
\[FormalE], \[FormalF]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", " ", "\[FormalD]", " ", "\[FormalE]", "\[FormalF]"},
           {"\[FormalA]", "\[FormalB]", " ", "\[FormalC]", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-\[FormalA], -\[FormalB], \[FormalD], -\[FormalC], \
\[FormalE], \[FormalF]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", " ", "\[FormalE]", " ", "\[FormalD]", "\[FormalF]"},
           {"\[FormalA]", "\[FormalC]", " ", "\[FormalB]", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-\[FormalA], -\[FormalC], \[FormalE], -\[FormalB], \
\[FormalD], \[FormalF]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", " ", "\[FormalF]", " ", "\[FormalD]", "\[FormalE]"},
           {"\[FormalA]", "\[FormalC]", " ", "\[FormalB]", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-\[FormalA], -\[FormalC], \[FormalF], -\[FormalB], \
\[FormalD], \[FormalE]],
    Editable->False]}], "-", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "\[FormalD]", "\[FormalE]", " ", " ", "\[FormalF]"},
           {"\[FormalA]", " ", " ", "\[FormalB]", "\[FormalC]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-\[FormalA], \[FormalD], \[FormalE], -\[FormalB], -\
\[FormalC], \[FormalF]],
    Editable->False]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"\[Delta]", 
        StyleBox[GridBox[{
           {" ", "\[FormalD]", "\[FormalF]", " ", " ", "\[FormalE]"},
           {"\[FormalA]", " ", " ", "\[FormalB]", "\[FormalC]", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    xAct`xTensor`Gdelta[-\[FormalA], \[FormalD], \[FormalF], -\[FormalB], -\
\[FormalC], \[FormalE]],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289789491571083`*^9},
 CellLabel->"Out[1269]=",
 CellID->1010022838,ExpressionUUID->"50f581d7-41e6-472d-acfe-4ab3872acc9d"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We prove it by expansion into elementary deltas:", "MathCaption",ExpressionUUID->"d59c67b9-9b6c-4843-90d8-966964c43c89"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ExpandGdelta"}]], "Input",
 CellLabel->
  "In[1270]:=",ExpressionUUID->"5697f1a1-b6d2-4bb4-ae97-1a3bd8ab73e7"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "\[FormalF]"},
               {"\[FormalA]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-\[FormalA], \[FormalF]],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalD]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalD]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalE]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalE]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalF]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalF]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "\[FormalF]"},
               {"\[FormalA]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-\[FormalA], \[FormalF]],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalC]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalC]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalE]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalB]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalB], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalE]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalB]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalB], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalF]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalB]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalB], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalF]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalE]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalE], -\[FormalB]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalE]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalE], -\[FormalC]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalE]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalE], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalF]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalE]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalE], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalF]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalE]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalE], \[FormalF]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalE]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalE], \[FormalF]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalE]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalE], -\[FormalB]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalE]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalE], -\[FormalB]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalE]", "\[FormalD]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalE], \[FormalD]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalF]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalF]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalC]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalC], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalE]", "\[FormalD]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalE], \[FormalD]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalE]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalE], \[FormalF]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalC]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalC], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalE]", "\[FormalF]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalE], \[FormalF]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "\[FormalE]"},
               {"\[FormalA]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-\[FormalA], \[FormalE]],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalF]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalF], -\[FormalB]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalF]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalF], -\[FormalC]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalF]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalF], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalE]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalF]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalF], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalD]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalD], \[FormalE]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalF]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalF], \[FormalE]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalD]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalD], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalC]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalC]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalF]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalF], \[FormalE]],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"\[Delta]", 
            StyleBox[GridBox[{
               {" ", "\[FormalE]"},
               {"\[FormalA]", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        xAct`xTensor`delta[-\[FormalA], \[FormalE]],
        Editable->False]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalF]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalF], -\[FormalB]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {"\[FormalF]", " "},
              {" ", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[\[FormalF], -\[FormalB]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalF]", "\[FormalD]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalF], \[FormalD]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalE]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalE]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalC]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalC], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalF]", "\[FormalD]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalF], \[FormalD]],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalC]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalC], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalA]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalA], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalF]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalF], \[FormalE]],
       Editable->False]}], "-", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "\[FormalD]"},
              {"\[FormalA]", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-\[FormalA], \[FormalD]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"\[FormalC]", "\[FormalB]"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[-\[FormalC], -\[FormalB]],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"\[FormalF]", "\[FormalE]"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metric[\[FormalF], \[FormalE]],
       Editable->False]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978949409525*^9},
 CellLabel->"Out[1270]=",
 CellID->1743092378,ExpressionUUID->"d26a8336-b30f-445e-be20-86f8943bd291"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"%", "//", "ToCanonical"}]], "Input",
 CellLabel->
  "In[1271]:=",ExpressionUUID->"61021ebc-174e-4242-bd0f-b3d4e56e66eb"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978949784836*^9},
 CellLabel->"Out[1271]=",
 CellID->181684672,ExpressionUUID->"f9c37a47-ec32-43ac-97e9-c617423e027c"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
However, a different way of antisymmetrizing is contracting with a Gdelta, \
and this gives zero \[LineSeparator]immediately, independently of the \
dimension:\
\>", "MathCaption",ExpressionUUID->"aa9fbe97-08f1-496f-9121-6456f887623e"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Gdelta", "[", 
   RowBox[{
    RowBox[{"-", "\[FormalA]"}], ",", 
    RowBox[{"-", "\[FormalB]"}], ",", 
    RowBox[{"-", "\[FormalC]"}], ",", "\[FormalD]", ",", "\[FormalE]", ",", 
    "\[FormalF]"}], "]"}], 
  RowBox[{"Gdelta", "[", 
   RowBox[{"\[FormalA]", ",", "\[FormalB]", ",", 
    RowBox[{"-", "\[FormalE]"}], ",", 
    RowBox[{"-", "\[FormalF]"}], ",", "\[FormalG]", ",", "\[FormalH]", ",", 
    "\[FormalI]", ",", "\[FormalJ]"}], "]"}]}]], "Input",
 CellLabel->
  "In[1272]:=",ExpressionUUID->"999bf8dd-3cef-4733-a1eb-b43f8945b7bf"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978949970647*^9},
 CellLabel->"Out[1272]=",
 CellID->2000019917,ExpressionUUID->"c5365fcb-4391-43b2-94c8-514554b933a8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Clean up:", "MathCaption",ExpressionUUID->"72b2dc43-e983-4ab5-b3f3-b3d31190637b"],

Cell[BoxData[{
 RowBox[{"UndefMetric", "[", "metric", "]"}], "\[IndentingNewLine]", 
 RowBox[{"UndefManifold", "[", "Mdim", "]"}], "\[IndentingNewLine]", 
 RowBox[{"UndefConstantSymbol", "[", "dim", "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"$DefInfoQ", "=", 
   RowBox[{"$UndefInfoQ", "=", "True"}]}], ";"}]}], "Input",
 CellLabel->
  "In[1273]:=",ExpressionUUID->"760165dd-bf45-4578-8753-12b2d0bfe204"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.7. Symmetric Trace-Free part", "Subsection",ExpressionUUID->"cc00b7aa-4354-4ad2-8a11-e97cec62027d"],

Cell["\<\
The command STFPart returns the symmetric, trace-free part of an expression \
with only free indices. In order for a symmetric object to have a trace we \
need to use a metric.\
\>", "Text",ExpressionUUID->"2d326e97-e85d-40bb-8696-71a5138dfb8a"],

Cell[CellGroupData[{

Cell["\<\
This is the STF part of a 3-index tensor without symmetry, and with respect \
to the metricg:\
\>", "MathCaption",ExpressionUUID->"d11147e3-40ad-4e33-8b3c-a5c906f8610b"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"STFPart", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", "metricg"}], 
  "]"}]], "Input",
 CellLabel->
  "In[1277]:=",ExpressionUUID->"02cad8a7-b143-4bf4-83a3-26b7b8038509"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "6"], " ", 
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "b", "c"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "c", "b"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`c, $CellContext`b],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"b", "a", "c"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`b, $CellContext`a, $CellContext`c],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"b", "c", "a"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`b, $CellContext`c, $CellContext`a],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"c", "a", "b"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`c, $CellContext`a, $CellContext`b],
      Editable->False], "+", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"c", "b", "a"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`c, $CellContext`b, $CellContext`a],
      Editable->False]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "10"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       FractionBox["1", "6"]}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"b", "c"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`b, $CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a", "d", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a, $CellContext`d, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a", "e", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a, $CellContext`e, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "a", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`a, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "e", "a"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`e, $CellContext`a],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "a", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`a, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "d", "a"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {\

               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`d, $CellContext`a],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      FractionBox["1", "6"], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"c", "b"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`c, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a", "d", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a, $CellContext`d, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a", "e", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a, $CellContext`e, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "a", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`a, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "e", "a"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`e, $CellContext`a],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "a", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`a, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "d", "a"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`d, $CellContext`a],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      FractionBox["1", "6"], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"a", "c"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`a, $CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b", "d", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b, $CellContext`d, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b", "e", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b, $CellContext`e, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "b", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`b, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "e", "b"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`e, $CellContext`b],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "b", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`b, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "d", "b"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`d, $CellContext`b],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      FractionBox["1", "6"], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"c", "a"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`c, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b", "d", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b, $CellContext`d, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b", "e", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b, $CellContext`e, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "b", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`b, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "e", "b"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`e, $CellContext`b],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "b", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`b, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "d", "b"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`d, $CellContext`b],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      FractionBox["1", "6"], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"a", "b"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`a, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"c", "d", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`c, $CellContext`d, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"c", "e", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`c, $CellContext`e, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "c", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`c, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "e", "c"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`e, $CellContext`c],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "c", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`c, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "d", "c"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`d, $CellContext`c],
         Editable->False]}], ")"}]}], "-", 
     RowBox[{
      FractionBox["1", "6"], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {"b", "a"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[$CellContext`b, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"d", "e"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`d, -$CellContext`e],
       Editable->False], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"c", "d", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`c, $CellContext`d, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"c", "e", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`c, $CellContext`e, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "c", "e"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`c, $CellContext`e],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"d", "e", "c"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`d, $CellContext`e, $CellContext`c],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "c", "d"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`c, $CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"e", "d", "c"},
                {" ", " ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`e, $CellContext`d, $CellContext`c],
         Editable->False]}], ")"}]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978957192182*^9},
 CellLabel->"Out[1277]=",
 CellID->1677761570,ExpressionUUID->"8b00d4f3-5ae0-4a7f-b0bf-94d1fa31ec42"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The result is not simplified at all. We must ask for explicit metric \
contraction and simplification:\
\>", "MathCaption",ExpressionUUID->"39ba8d23-c3bf-4eb8-8f91-785d6ed73679"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ContractMetric"}], "//", "Simplification"}]], "Input",
 CellLabel->
  "In[1278]:=",ExpressionUUID->"0c130b34-b821-4376-9ea4-1dde287ecfd6"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "30"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "b", "c"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`b, $CellContext`c],
      Editable->False]}], "+", 
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "c", "b"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`c, $CellContext`b],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"b", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`b, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "d", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`d, -$CellContext`d],
      Editable->False]}], "+", 
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"b", "a", "c"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`b, $CellContext`a, $CellContext`c],
      Editable->False]}], "+", 
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"b", "c", "a"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`b, $CellContext`c, $CellContext`a],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"a", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`a, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"b", "d", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`b, $CellContext`d, -$CellContext`d],
      Editable->False]}], "+", 
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"c", "a", "b"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`c, $CellContext`a, $CellContext`b],
      Editable->False]}], "+", 
    RowBox[{"5", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"c", "b", "a"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`c, $CellContext`b, $CellContext`a],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"a", "b"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`a, $CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"c", "d", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`c, $CellContext`d, -$CellContext`d],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"b", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`b, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"d", "a", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`d, $CellContext`a, -$CellContext`d],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"a", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`a, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"d", "b", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`d, $CellContext`b, -$CellContext`d],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"a", "b"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`a, $CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"d", "c", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`d, $CellContext`c, -$CellContext`d],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"b", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`b, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"d", " ", "a"},
             {" ", "d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`d, -$CellContext`d, $CellContext`a],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"a", "c"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`a, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"d", " ", "b"},
             {" ", "d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`d, -$CellContext`d, $CellContext`b],
      Editable->False]}], "-", 
    RowBox[{"2", " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"g", 
          StyleBox[GridBox[{
             {"a", "b"},
             {" ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`metricg[$CellContext`a, $CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"d", " ", "c"},
             {" ", "d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`d, -$CellContext`d, $CellContext`c],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.728978957450172*^9},
 CellLabel->"Out[1278]=",
 CellID->1678421880,ExpressionUUID->"c43db077-e4d5-4728-ae82-565ee7f067f8"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.8. Mapping at specified positions", "Subsection",ExpressionUUID->"657a9bdf-72f6-42dd-9fa5-4695039b9fb2"],

Cell[TextData[{
 "We use the package ",
 StyleBox["ExpressionManipulation` ",
  FontFamily->"Courier"],
 "(written by David Park & Ted Ersek) to have full control on the positions \
of an expression"
}], "Text",ExpressionUUID->"13668bed-de68-4c30-95a4-78d142b13253"],

Cell[CellGroupData[{

Cell["\<\
For versions of xTensor` before 0.9.8 this package was automatically loaded, \
but now we need to do it explicitly, to avoid confusing it with the xAct` \
packages. This package has no copyright messages:\
\>", "MathCaption",ExpressionUUID->"c57b1329-32db-492d-abd7-83f09d5a2635"],

Cell[BoxData[
 RowBox[{"<<", "xAct`ExpressionManipulation`"}]], "Input",
 CellLabel->
  "In[1279]:=",ExpressionUUID->"2275d90e-a848-42eb-a85f-4ce761b931e3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Suppose we have an expression like this, which is clearly ",
 StyleBox["0",
  FontFamily->"Courier"],
 " because ",
 StyleBox["U",
  FontFamily->"Courier"],
 " is antisymmetric and ",
 StyleBox["S",
  FontFamily->"Courier"],
 " is symmetric:"
}], "MathCaption",ExpressionUUID->"ab4332c5-bc82-4ab1-965e-10cd3c9d8a86"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"DefTensor", "[", 
  RowBox[{
   RowBox[{"G", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", "M3", ",", 
   RowBox[{"Symmetric", "[", 
    RowBox[{"{", 
     RowBox[{"1", ",", "2"}], "}"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1280]:=",ExpressionUUID->"da7839a9-4637-4a64-8db5-5222795b5809"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "DefTensor", 
   "\[InvisibleSpace]", "\<\": Defining \"\>", 
   "\[InvisibleSpace]", "\<\"tensor \"\>", "\[InvisibleSpace]", 
   RowBox[{"G", "[", 
    RowBox[{"a", ",", "b"}], "]"}], "\[InvisibleSpace]", "\<\". \"\>", 
   "\[InvisibleSpace]", "\<\"\"\>"}],
  SequenceForm["** ", xAct`xTensor`DefTensor, ": Defining ", "tensor ", 
   $CellContext`G[$CellContext`a, $CellContext`b], ". ", ""],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978960453923*^9},
 CellLabel->"During evaluation of In[1280]:=",
 CellID->894627505,ExpressionUUID->"b332f045-e5fd-450d-887a-29cac9615bec"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"G", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"v", "[", "b", "]"}], 
    RowBox[{"v", "[", "c", "]"}]}], "-", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"v", "[", "d", "]"}], 
    RowBox[{"v", "[", "e", "]"}]}]}]}]], "Input",
 CellLabel->
  "In[1281]:=",ExpressionUUID->"5c9b5b2c-97fb-444c-896e-308e24313366"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"G", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`G[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978960607463*^9},
 CellLabel->"Out[1281]=",
 CellID->1471463888,ExpressionUUID->"b5d4af3e-e969-41ed-82fb-6edcbd0abf9e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "expr", "]"}]], "Input",
 CellLabel->
  "In[1282]:=",ExpressionUUID->"4a8b139b-ad55-4da0-b7bb-7305fc5eadc0"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.7289789607356997`*^9},
 CellLabel->"Out[1282]=",
 CellID->471886764,ExpressionUUID->"a75e5628-3f78-4a38-912b-ff381aa9838b"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Each object in the expression can be identified in a tree-like form by its \
position. In this case there are four positions at level 1:\
\>", "MathCaption",ExpressionUUID->"7aff734c-a431-4430-8bb4-7186a61e1766"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "//", "ColorTerms"}]], "Input",
 CellLabel->
  "In[1283]:=",ExpressionUUID->"07bb6e50-cd48-4d31-bc62-acb1bd4ef9e6"],

Cell[BoxData[
 TagBox[
  RowBox[{
   TagBox[
    StyleBox[
     InterpretationBox[
      RowBox[{
       TagBox[
        FrameBox[
         TagBox[
          StyleBox[
           RowBox[{"{", "1", "}"}],
           FontSize->10],
          StyleForm[#, FontSize -> 10]& ]],
        DisplayForm], "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
       TagBox[
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"G", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`G[-$CellContext`b, -$CellContext`c],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"U", 
              StyleBox[GridBox[{
                 {"a", "b", "c"},
                 {" ", " ", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"a"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[-$CellContext`a],
          Editable->False]}],
        HoldForm]}],
      SequenceForm[
       DisplayForm[
        FrameBox[
         StyleForm[{1}, FontSize -> 10]]], " ", 
       HoldForm[($CellContext`G[-$CellContext`b, -$CellContext`c] \
$CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c]) \
$CellContext`v[-$CellContext`a]]],
      Editable->False],
     Background->Hue[0., 0.4, 0.95]],
    StyleForm[#, Background -> Hue[0., 0.4, 0.95]]& ], "+", 
   TagBox[
    StyleBox[
     InterpretationBox[
      RowBox[{
       TagBox[
        FrameBox[
         TagBox[
          StyleBox[
           RowBox[{"{", "2", "}"}],
           FontSize->10],
          StyleForm[#, FontSize -> 10]& ]],
        DisplayForm], "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
       TagBox[
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"U", 
              StyleBox[GridBox[{
                 {"a", "b", "c"},
                 {" ", " ", " "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"a"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[-$CellContext`a],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"b"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[-$CellContext`b],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {" "},
                 {"c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[-$CellContext`c],
          Editable->False]}],
        HoldForm]}],
      SequenceForm[
       DisplayForm[
        FrameBox[
         StyleForm[{2}, FontSize -> 10]]], " ", 
       HoldForm[(($CellContext`U[$CellContext`a, $CellContext`b, \
$CellContext`c] $CellContext`v[-$CellContext`a]) \
$CellContext`v[-$CellContext`b]) $CellContext`v[-$CellContext`c]]],
      Editable->False],
     Background->Hue[0.2333333333333333, 0.4, 0.95]],
    StyleForm[#, Background -> Hue[0.2333333333333333, 0.4, 0.95]]& ], "+", 
   TagBox[
    StyleBox[
     InterpretationBox[
      RowBox[{
       TagBox[
        FrameBox[
         TagBox[
          StyleBox[
           RowBox[{"{", "3", "}"}],
           FontSize->10],
          StyleForm[#, FontSize -> 10]& ]],
        DisplayForm], "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
       TagBox[
        RowBox[{
         InterpretationBox[
          StyleBox[GridBox[{
             {"T", 
              StyleBox[GridBox[{
                 {" ", " "},
                 {"b", "c"}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`T[-$CellContext`b, -$CellContext`c],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"b"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[$CellContext`b],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"c"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[$CellContext`c],
          Editable->False]}],
        HoldForm]}],
      SequenceForm[
       DisplayForm[
        FrameBox[
         StyleForm[{3}, FontSize -> 10]]], " ", 
       HoldForm[($CellContext`T[-$CellContext`b, -$CellContext`c] \
$CellContext`v[$CellContext`b]) $CellContext`v[$CellContext`c]]],
      Editable->False],
     Background->Hue[0.4666666666666666, 0.4, 0.95]],
    StyleForm[#, Background -> Hue[0.4666666666666666, 0.4, 0.95]]& ], "+", 
   TagBox[
    StyleBox[
     InterpretationBox[
      RowBox[{
       TagBox[
        FrameBox[
         TagBox[
          StyleBox[
           RowBox[{"{", "4", "}"}],
           FontSize->10],
          StyleForm[#, FontSize -> 10]& ]],
        DisplayForm], "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
       TagBox[
        RowBox[{
         RowBox[{"-", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"T", 
               StyleBox[GridBox[{
                  {" ", " "},
                  {"d", "e"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`T[-$CellContext`d, -$CellContext`e],
           Editable->False]}], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"d"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[$CellContext`d],
          Editable->False], " ", 
         InterpretationBox[
          StyleBox[GridBox[{
             {"v", 
              StyleBox[GridBox[{
                 {"e"},
                 {" "}
                },
                GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
               FontSize->9]}
            },
            
            GridBoxAlignment->{
             "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
              "Rows" -> {{Center}}, "RowsIndexed" -> {}},
            GridBoxSpacings->{"Columns" -> {
                Offset[0.27999999999999997`], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
                Offset[0.2], {
                 Offset[0.4]}, 
                Offset[0.2]}, "RowsIndexed" -> {}}],
           ShowAutoStyles->False,
           AutoSpacing->False],
          $CellContext`v[$CellContext`e],
          Editable->False]}],
        HoldForm]}],
      SequenceForm[
       DisplayForm[
        FrameBox[
         StyleForm[{4}, FontSize -> 10]]], " ", 
       HoldForm[((-$CellContext`T[-$CellContext`d, -$CellContext`e]) \
$CellContext`v[$CellContext`d]) $CellContext`v[$CellContext`e]]],
      Editable->False],
     Background->Hue[0.6999999999999998, 0.4, 0.95]],
    StyleForm[#, Background -> Hue[0.6999999999999998, 0.4, 0.95]]& ]}],
  HoldForm]], "Output",
 CellChangeTimes->{3.728978960880972*^9},
 CellLabel->"Out[1283]=",
 CellID->343734177,ExpressionUUID->"856ce393-70ee-4cd6-b252-f6c2ee55a21f"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
We can evaluate a given function at one or several positions:\
\>", "MathCaption",ExpressionUUID->"ae2a9a69-446c-46f3-83b5-f671de712962"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MapAt", "[", 
  RowBox[{"Simplification", ",", "expr", ",", 
   RowBox[{"{", "2", "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1284]:=",ExpressionUUID->"2d1f479c-dcf1-496b-8042-93e5e4fc3a64"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"G", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`G[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978961403676*^9},
 CellLabel->"Out[1284]=",
 CellID->409302409,ExpressionUUID->"4288c058-3668-45ea-be14-10a36270b6ab"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MapAt", "[", 
  RowBox[{"Simplification", ",", "expr", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", "1", "}"}], ",", 
     RowBox[{"{", "2", "}"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1285]:=",ExpressionUUID->"90d8e5f1-2661-4528-9450-b8b4ab5302db"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978962237442*^9},
 CellLabel->"Out[1285]=",
 CellID->755466722,ExpressionUUID->"a255b957-99e7-4111-b91a-1e8f26f13b89"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In order to map a function over several terms simultaneously we need to \
make use of the concept of ",
 StyleBox["ExtendedPosition",
  FontFamily->"Courier"],
 " (also from the ",
 StyleBox["ExpressionManipulation`",
  FontFamily->"Courier"],
 " package) and the new function ",
 StyleBox["NewMapAt",
  FontFamily->"Courier"],
 ". See notes for",
 StyleBox[" ExtendedPosition",
  FontFamily->"Courier"],
 " and ",
 StyleBox["eP",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"8b05dfb5-3cba-432e-bbae-830d55adc084"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MapAt", "[", 
  RowBox[{"f", ",", "expr", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", "3", "}"}], ",", 
     RowBox[{"{", "4", "}"}]}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1286]:=",ExpressionUUID->"ecba2eab-fcbc-49f0-b215-142297c26976"],

Cell[BoxData[
 RowBox[{
  RowBox[{"f", "[", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`c],
     Editable->False]}], "]"}], "+", 
  RowBox[{"f", "[", 
   RowBox[{
    RowBox[{"-", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {"d", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[-$CellContext`d, -$CellContext`e],
      Editable->False]}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"d"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"e"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`e],
     Editable->False]}], "]"}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"G", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`G[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978962469295*^9},
 CellLabel->"Out[1286]=",
 CellID->1205875590,ExpressionUUID->"c2dea06c-478c-4d3a-9b17-71669303306c"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"NewMapAt", "[", 
  RowBox[{"f", ",", "expr", ",", 
   RowBox[{"eP", "[", 
    RowBox[{
     RowBox[{"{", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "4"}], "}"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1287]:=",ExpressionUUID->"7d1015f5-9712-4d4c-9ec2-2cb1d6353bb1"],

Cell[BoxData[
 RowBox[{
  RowBox[{"f", "[", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {"b", "c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[-$CellContext`b, -$CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`c],
      Editable->False]}], "-", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {" ", " "},
             {"d", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[-$CellContext`d, -$CellContext`e],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"e"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`e],
      Editable->False]}]}], "]"}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"G", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`G[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c"},
           {" ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.7289789626339407`*^9},
 CellLabel->"Out[1287]=",
 CellID->1347864391,ExpressionUUID->"b667fd05-76ba-49a5-89a0-b10ebb5b3005"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["We can also localize a given pattern:", "MathCaption",ExpressionUUID->"633e312a-6217-4d3f-845d-8076502815d1"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "//", 
  RowBox[{"ColorPositionsOfPattern", "[", 
   RowBox[{"U", "[", "__", "]"}], "]"}]}]], "Input",
 CellLabel->
  "In[1288]:=",ExpressionUUID->"705d82ff-4f88-45fd-8f3b-174a519371dd"],

Cell[BoxData[
 TagBox[
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"G", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`G[-$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    RowBox[{"(", 
     TagBox[
      StyleBox[
       InterpretationBox[
        RowBox[{
         TagBox[
          FrameBox[
           TagBox[
            StyleBox[
             RowBox[{"{", 
              RowBox[{"1", ",", "2"}], "}"}],
             FontSize->10],
            StyleForm[#, FontSize -> 10]& ]],
          DisplayForm], "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
         TagBox[
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {"a", "b", "c"},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
           Editable->False],
          HoldForm]}],
        SequenceForm[
         DisplayForm[
          FrameBox[
           StyleForm[{1, 2}, FontSize -> 10]]], " ", 
         HoldForm[
          $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c]]],
        Editable->False],
       Background->Hue[0.5, 0.4, 0.95]],
      StyleForm[#, Background -> Hue[0.5, 0.4, 0.95]]& ], ")"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False]}], "+", 
   RowBox[{
    RowBox[{"(", 
     TagBox[
      StyleBox[
       InterpretationBox[
        RowBox[{
         TagBox[
          FrameBox[
           TagBox[
            StyleBox[
             RowBox[{"{", 
              RowBox[{"2", ",", "1"}], "}"}],
             FontSize->10],
            StyleForm[#, FontSize -> 10]& ]],
          DisplayForm], "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
         TagBox[
          InterpretationBox[
           StyleBox[GridBox[{
              {"U", 
               StyleBox[GridBox[{
                  {"a", "b", "c"},
                  {" ", " ", " "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
           Editable->False],
          HoldForm]}],
        SequenceForm[
         DisplayForm[
          FrameBox[
           StyleForm[{2, 1}, FontSize -> 10]]], " ", 
         HoldForm[
          $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c]]],
        Editable->False],
       Background->Hue[0.5, 0.4, 0.95]],
      StyleForm[#, Background -> Hue[0.5, 0.4, 0.95]]& ], ")"}], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`c],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`c],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`d, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"d"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"e"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`e],
     Editable->False]}]}],
  HoldForm]], "Output",
 CellChangeTimes->{3.7289789627603083`*^9},
 CellLabel->"Out[1288]=",
 CellID->1899310981,ExpressionUUID->"dd10be8c-c452-4b1e-bca7-3d2ef812fbe7"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["or map a function onto the occurrences of that pattern:", "MathCaption",ExpressionUUID->"8a8cdb46-20f2-417e-9fc0-d3bf3c4f644c"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Position", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"U", "[", "__", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1289]:=",ExpressionUUID->"99ee2185-faae-463a-ade9-54935d2cdd84"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "2"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.7289789628769484`*^9},
 CellLabel->"Out[1289]=",
 CellID->1522728659,ExpressionUUID->"6765649b-d86a-41d9-a9b2-51fdebf0a6d7"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MapAt", "[", 
  RowBox[{"f", ",", "expr", ",", "%"}], "]"}]], "Input",
 CellLabel->
  "In[1290]:=",ExpressionUUID->"d2ee5476-8d66-4d16-b15e-b4c1f309fe0f"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"G", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`G[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False]}], "+", 
  RowBox[{
   RowBox[{"f", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
     Editable->False], "]"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"a"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`a],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"b", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"b"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}], "-", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " "},
           {"d", "e"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`d, -$CellContext`e],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"e"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`e],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978963014223*^9},
 CellLabel->"Out[1290]=",
 CellID->937950514,ExpressionUUID->"48f0e4c6-27bb-4c8d-8cbf-c3ee024c28ec"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "In case the expression is held, we can use ",
 StyleBox["EvaluateAt",
  FontFamily->"Courier"],
 " (also from the package ",
 StyleBox["ExpressionManipulation`",
  FontFamily->"Courier"],
 ") instead of ",
 StyleBox["MapAt",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"2c9ca0ad-cb53-4e94-8f5d-b97cfb91f380"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"S", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}]}], "+", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "b"}], ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"v", "[", "b", "]"}], 
    RowBox[{"v", "[", "c", "]"}]}], "-", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "d"}], ",", 
      RowBox[{"-", "e"}]}], "]"}], 
    RowBox[{"v", "[", "d", "]"}], 
    RowBox[{"v", "[", "e", "]"}]}], "+", "1", "+", "2"}], "]"}]], "Input",
 CellLabel->
  "In[1291]:=",ExpressionUUID->"b61110fc-4edb-482d-8122-d331a3ebffc4"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"b"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`c],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[-$CellContext`b, -$CellContext`c],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`c],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`d, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"d"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"e"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`e],
     Editable->False]}], "+", "1", "+", "2"}], "]"}]], "Output",
 CellChangeTimes->{3.7289789631757517`*^9},
 CellLabel->"Out[1291]=",
 CellID->1604319818,ExpressionUUID->"547903c1-45cd-4d4c-ace5-5c4ef471331f"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"MapAt", "[", 
  RowBox[{"Simplification", ",", "%", ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "1"}], "}"}]}], "]"}]], "Input",
 CellLabel->
  "In[1292]:=",ExpressionUUID->"b93f51d5-b0e1-427b-8ef0-0dc2f7eb81b6"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{
   RowBox[{"Simplification", "[", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"U", 
          StyleBox[GridBox[{
             {"a", "b", "c"},
             {" ", " ", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"b"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"c"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`c],
      Editable->False]}], "]"}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[-$CellContext`b, -$CellContext`c],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`c],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`d, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"d"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"e"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`e],
     Editable->False]}], "+", "1", "+", "2"}], "]"}]], "Output",
 CellChangeTimes->{3.7289789633156137`*^9},
 CellLabel->"Out[1292]=",
 CellID->2145993041,ExpressionUUID->"ffbd0585-c4da-4d75-b5a9-b1e67db229be"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"EvaluateAt", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"1", ",", "1"}], "}"}], ",", "Simplification"}], "]"}], "[", 
  "%%", "]"}]], "Input",
 CellLabel->
  "In[1293]:=",ExpressionUUID->"be0d5668-72cd-4b74-a936-efee4d51c7be"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{"0", "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"U", 
         StyleBox[GridBox[{
            {"a", "b", "c"},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {" "},
            {"a"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[-$CellContext`a],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"S", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`S[-$CellContext`b, -$CellContext`c],
     Editable->False]}], "+", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"b", "c"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`b, -$CellContext`c],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"b"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`b],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"c"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`c],
     Editable->False]}], "-", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {" ", " "},
            {"d", "e"}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[-$CellContext`d, -$CellContext`e],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"d"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`d],
     Editable->False], " ", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"v", 
         StyleBox[GridBox[{
            {"e"},
            {" "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`v[$CellContext`e],
     Editable->False]}], "+", "1", "+", "2"}], "]"}]], "Output",
 CellChangeTimes->{3.728978963454331*^9},
 CellLabel->"Out[1293]=",
 CellID->717188722,ExpressionUUID->"cdbb3b69-d6bd-4131-9ed3-b0cb72d1a025"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.9. Coefficients", "Subsection",ExpressionUUID->"90f7019f-3ef3-4543-86f2-339abc68ed1e"],

Cell["\<\
Given an expression, sometimes we need to get the (possibly indexed) \
coefficient multiplying a given tensor. When the objects have symmetries the \
coefficient might be more complicated than expected.\
\>", "Text",ExpressionUUID->"a5d086bc-55b6-422a-9038-07fe77be6716"],

Cell[TextData[{
 StyleBox[ButtonBox["IndexCoefficient",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    FrontEnd`FileName[{"xTensor"}, "IndexCoefficient.nb", CharacterEncoding -> 
     "iso8859-1"], None}],
  FontFamily->"Courier"],
 StyleBox["\t\t", "DisplayMath"],
 StyleBox["Find the coefficient of an indexed expression", "Text"]
}], "DefinitionBox",ExpressionUUID->"a6df1543-8041-49ef-93ff-ada9bc9c212a"],

Cell["Coefficients of indexed expressions.", "Text",ExpressionUUID->"107cf7da-cac8-41fe-ad13-773c153bb983"],

Cell[CellGroupData[{

Cell["The basic relation is this:", "MathCaption",ExpressionUUID->"5458d8e6-cb10-4ffd-b719-8673f931c9fe"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCoefficient", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{"T", "[", 
    RowBox[{"c", ",", "d"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1294]:=",ExpressionUUID->"32c52d95-07d7-4e3d-adb4-e3de49ed39c1"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[Delta]", 
       StyleBox[GridBox[{
          {" ", "a"},
          {"c", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`delta[-$CellContext`c, $CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"\[Delta]", 
       StyleBox[GridBox[{
          {" ", "b"},
          {"d", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   xAct`xTensor`delta[-$CellContext`d, $CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978971754344*^9},
 CellLabel->"Out[1294]=",
 CellID->776216801,ExpressionUUID->"4b43a68b-3b23-44de-9123-aca2464bd0ad"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Metrics and symmetries must be taken into account:", "MathCaption",ExpressionUUID->"b8299032-23fc-4d23-b77e-f44a5bbd9487"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCoefficient", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"T", "[", 
    RowBox[{"c", ",", "d"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1295]:=",ExpressionUUID->"3fff252c-b5c0-4dca-b9fa-64d384b159be"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {" ", " "},
          {"c", "a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[-$CellContext`c, -$CellContext`a],
   Editable->False], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {" ", " "},
          {"d", "b"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[-$CellContext`d, -$CellContext`b],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.7289789719221354`*^9},
 CellLabel->"Out[1295]=",
 CellID->2060102529,ExpressionUUID->"e7344982-2e18-486f-bfdb-b77343f82fb3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCoefficient", "[", 
  RowBox[{
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "a"}], ",", 
     RowBox[{"-", "b"}]}], "]"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{
     RowBox[{"-", "c"}], ",", 
     RowBox[{"-", "d"}]}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1296]:=",ExpressionUUID->"82c746d3-5949-4d74-9ef7-089393e3a32b"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "d"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`d],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`c],
      Editable->False]}], "+", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"a", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`a, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {" ", "d"},
             {"b", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[-$CellContext`b, $CellContext`d],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.72897897205991*^9},
 CellLabel->"Out[1296]=",
 CellID->264364322,ExpressionUUID->"a0c106dc-48db-4927-9a00-cb20e45695ee"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IndexCoefficient", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"T", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"-", "a"}], "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}]}], ",", 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}]}], "]"}], "//", "Simplification"}]], "Input",
 CellLabel->
  "In[1297]:=",ExpressionUUID->"90e3f535-a8af-45e5-aedc-2340ae2a20cd"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"a", "c"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`a, $CellContext`c],
     Editable->False], "+", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {"c", "a"},
            {" ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[$CellContext`c, $CellContext`a],
     Editable->False]}], ")"}], " ", 
  InterpretationBox[
   StyleBox[GridBox[{
      {"v", 
       StyleBox[GridBox[{
          {" "},
          {"a"}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`v[-$CellContext`a],
   Editable->False]}]], "Output",
 CellChangeTimes->{3.728978972326285*^9},
 CellLabel->"Out[1297]=",
 CellID->27494211,ExpressionUUID->"07828efe-3a2a-4d00-8e88-a874a8b4c035"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCoefficient", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "a"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "d"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1298]:=",ExpressionUUID->"680b0274-1234-4d20-bcb9-0f437a64dcdf"],

Cell[BoxData[
 RowBox[{
  FractionBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"c", "d"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`c, $CellContext`d],
    Editable->False], "2"], "+", 
  FractionBox[
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"d", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`d, $CellContext`c],
    Editable->False], "2"]}]], "Output",
 CellChangeTimes->{3.728978972770171*^9},
 CellLabel->"Out[1298]=",
 CellID->386220172,ExpressionUUID->"898c8772-f955-48de-a0d4-742667a36504"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Imitating ",
 StyleBox["Coefficient",
  FontFamily->"Courier"],
 ", dependencies must be explicit:"
}], "MathCaption",ExpressionUUID->"8c4b7018-bdc8-4aab-b3b6-f405579c2b9d"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCoefficient", "[", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], ",", 
   RowBox[{"metricg", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1299]:=",ExpressionUUID->"339b4920-2595-46c1-990f-60c624e47847"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.728978972910747*^9},
 CellLabel->"Out[1299]=",
 CellID->1506849570,ExpressionUUID->"57623251-4d6a-4dd9-876a-54450a855ab5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Frequently we need further manipulation of the result:", "MathCaption",ExpressionUUID->"d0f2179a-1f93-417d-8e72-5aa6a5c468d9"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCoefficient", "[", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"T", "[", 
      RowBox[{"a", ",", "b"}], "]"}], 
     RowBox[{"v", "[", 
      RowBox[{"-", "b"}], "]"}], 
     RowBox[{"v", "[", "c", "]"}]}], "+", 
    RowBox[{
     RowBox[{"T", "[", 
      RowBox[{"d", ",", 
       RowBox[{"-", "d"}]}], "]"}], 
     RowBox[{"v", "[", "a", "]"}], 
     RowBox[{"v", "[", "c", "]"}]}]}], ",", 
   RowBox[{"v", "[", "e", "]"}]}], "]"}]], "Input",
 CellLabel->
  "In[1300]:=",ExpressionUUID->"0db62fc3-bc4b-44b2-b50b-7c49787640c5"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "c"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {" "},
              {"b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[-$CellContext`b],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"g", 
           StyleBox[GridBox[{
              {" ", " "},
              {"e", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`metricg[-$CellContext`e, -$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"a", "b"},
              {" ", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`a, $CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`c],
       Editable->False]}]}], ")"}]}], "+", 
  RowBox[{
   FractionBox["1", "2"], " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "c"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`c],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"b", " "},
              {" ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`b, -$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"a"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`a],
       Editable->False]}], "+", 
     RowBox[{
      InterpretationBox[
       StyleBox[GridBox[{
          {"\[Delta]", 
           StyleBox[GridBox[{
              {" ", "a"},
              {"e", " "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`delta[-$CellContext`e, $CellContext`a],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"T", 
           StyleBox[GridBox[{
              {"b", " "},
              {" ", "b"}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`T[$CellContext`b, -$CellContext`b],
       Editable->False], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`c],
       Editable->False]}]}], ")"}]}]}]], "Output",
 CellChangeTimes->{3.728978973077858*^9},
 CellLabel->"Out[1300]=",
 CellID->346073919,ExpressionUUID->"cdfa049a-b657-4df4-9ae0-b11f7835779e"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"%", "//", "ContractMetric"}], "//", "Simplification"}]], "Input",
 CellLabel->
  "In[1301]:=",ExpressionUUID->"9cded25c-f506-4d0a-8959-32ee2e548b55"],

Cell[BoxData[
 RowBox[{
  FractionBox["1", "2"], " ", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"\[Delta]", 
          StyleBox[GridBox[{
             {"c", " "},
             {" ", "e"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      xAct`xTensor`delta[$CellContext`c, -$CellContext`e],
      Editable->False], " ", 
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b", " "},
                {" ", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b, -$CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"a", "b"},
                {" ", " "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`a, $CellContext`b],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[-$CellContext`b],
         Editable->False]}]}], ")"}]}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"T", 
            StyleBox[GridBox[{
               {"a", " "},
               {" ", "e"}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`T[$CellContext`a, -$CellContext`e],
        Editable->False], "+", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"\[Delta]", 
             StyleBox[GridBox[{
                {"a", " "},
                {" ", "e"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         xAct`xTensor`delta[$CellContext`a, -$CellContext`e],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"T", 
             StyleBox[GridBox[{
                {"b", " "},
                {" ", "b"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`T[$CellContext`b, -$CellContext`b],
         Editable->False]}]}], ")"}], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"c"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`c],
      Editable->False]}]}], ")"}]}]], "Output",
 CellChangeTimes->{3.7289789732459*^9},
 CellLabel->"Out[1301]=",
 CellID->1078640557,ExpressionUUID->"4ce7d19b-c48d-4df1-baf9-4ab4802a8af8"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
 "A closely related function is ",
 StyleBox["IndexCollect",
  FontFamily->"Courier"],
 ", which acts as a simple recursive driver for ",
 StyleBox["IndexCoefficient",
  FontFamily->"Courier"],
 ". This pair has been designed to follow closely the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " pair ",
 StyleBox["Collect / Coefficient",
  FontFamily->"Courier"],
 "."
}], "Text",ExpressionUUID->"fb35b0f7-3f30-490e-9e00-17dde2728888"],

Cell[TextData[{
 StyleBox["IndexCollect\t\t", "DisplayMath"],
 StyleBox["Collect terms in an indexed expression", "Text"]
}], "DefinitionBox",ExpressionUUID->"40ef30fe-74eb-449c-9eee-c00c4a08041f"],

Cell[CellGroupData[{

Cell["Suppose an expression like this:", "MathCaption",ExpressionUUID->"53e839fa-c408-421f-9baf-9e7df6bc6263"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"expr", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b", ",", 
      RowBox[{"-", "c"}]}], "]"}], 
    RowBox[{"v", "[", "c", "]"}]}], "+", 
   RowBox[{
    RowBox[{"U", "[", 
     RowBox[{"a", ",", "b", ",", "c", ",", "d"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "d"}], "]"}]}], "+", 
   RowBox[{"metricg", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[1302]:=",ExpressionUUID->"6cd77f16-eb35-4a2a-99a0-46c03a51b92a"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`a, $CellContext`b],
   Editable->False], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b", " "},
           {" ", " ", "c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"c"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`c],
    Editable->False]}], "+", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"U", 
        StyleBox[GridBox[{
           {"a", "b", "c", "d"},
           {" ", " ", " ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`U[$CellContext`a, $CellContext`b, $CellContext`c, \
$CellContext`d],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"d"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`d],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.72897897340326*^9},
 CellLabel->"Out[1302]=",
 CellID->1802377172,ExpressionUUID->"8cc53ff2-11d3-4243-a243-f878fd819c42"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IndexCollect", "[", 
  RowBox[{"expr", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"v", "[", "d", "]"}], ",", 
     RowBox[{"v", "[", "e", "]"}]}], "}"}], ",", "Simplification"}], 
  "]"}]], "Input",
 CellLabel->
  "In[1303]:=",ExpressionUUID->"33287f06-68bf-4008-bdd9-9ee1d083155c"],

Cell[BoxData[
 RowBox[{
  InterpretationBox[
   StyleBox[GridBox[{
      {"g", 
       StyleBox[GridBox[{
          {"a", "b"},
          {" ", " "}
         },
         GridBoxSpacings->{"Columns" -> {
             Offset[0.], {
              Offset[0.034999999999999996`]}, 
             Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
              Offset[0.]}}, "RowsIndexed" -> {}}],
        FontSize->9]}
     },
     GridBoxAlignment->{
      "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.034999999999999996`]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    ShowAutoStyles->False,
    AutoSpacing->False],
   $CellContext`metricg[$CellContext`a, $CellContext`b],
   Editable->False], "+", 
  RowBox[{
   RowBox[{"(", 
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"T", 
          StyleBox[GridBox[{
             {"a", "b", " "},
             {" ", " ", "d"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`T[$CellContext`a, $CellContext`b, -$CellContext`d],
      Editable->False], "+", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"U", 
             StyleBox[GridBox[{
                {"a", "b", " ", " "},
                {" ", " ", "c", "d"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`U[$CellContext`a, $CellContext`b, -$CellContext`c, \
-$CellContext`d],
         Editable->False], "+", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"U", 
             StyleBox[GridBox[{
                {"a", "b", " ", " "},
                {" ", " ", "d", "c"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`U[$CellContext`a, $CellContext`b, -$CellContext`d, \
-$CellContext`c],
         Editable->False]}], ")"}], " ", 
      InterpretationBox[
       StyleBox[GridBox[{
          {"v", 
           StyleBox[GridBox[{
              {"c"},
              {" "}
             },
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                  Offset[0.]}}, "RowsIndexed" -> {}}],
            FontSize->9]}
         },
         GridBoxAlignment->{
          "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
           "Rows" -> {{Center}}, "RowsIndexed" -> {}},
         GridBoxSpacings->{"Columns" -> {
             Offset[0.27999999999999997`], {
              Offset[0.034999999999999996`]}, 
             Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
             Offset[0.2], {
              Offset[0.4]}, 
             Offset[0.2]}, "RowsIndexed" -> {}}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       $CellContext`v[$CellContext`c],
       Editable->False]}]}], ")"}], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {"d"},
           {" "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[$CellContext`d],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978973568578*^9},
 CellLabel->"Out[1303]=",
 CellID->148805222,ExpressionUUID->"24aece55-80c7-4ad2-af0b-a1c556975660"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.10. Tensor equations", "Subsection",ExpressionUUID->"bc25c3e2-b171-420e-abc8-60aea9c198d9"],

Cell[TextData[{
 "It is not simple to deal with tensor equations because there are many \
things to worry about simultaneously. Currently ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " only works with linear equations where the x tensor is not contracted. The \
concept of equality (==) has been generalized to include tensor equalities."
}], "Text",ExpressionUUID->"4f9c13b6-2fbe-45fd-b59f-e05b80cc794d"],

Cell[CellGroupData[{

Cell[TextData[{
 "From the point of view of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " these two expressions are not related:"
}], "MathCaption",ExpressionUUID->"6de3f851-3bef-4bb5-8b4e-27ba06c24f5f"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "b"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "b"}], "]"}]}], "\[Equal]", 
  RowBox[{
   RowBox[{"T", "[", 
    RowBox[{"a", ",", "c"}], "]"}], 
   RowBox[{"v", "[", 
    RowBox[{"-", "c"}], "]"}]}]}]], "Input",
 CellLabel->
  "In[1304]:=",ExpressionUUID->"65e7af33-4841-4448-aef8-4368eb920f22"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "b"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`b],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"b"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`b],
    Editable->False]}], "\[Equal]", 
  RowBox[{
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", "c"},
           {" ", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, $CellContext`c],
    Editable->False], " ", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"v", 
        StyleBox[GridBox[{
           {" "},
           {"c"}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`v[-$CellContext`c],
    Editable->False]}]}]], "Output",
 CellChangeTimes->{3.728978978805379*^9},
 CellLabel->"Out[1304]=",
 CellID->1532473207,ExpressionUUID->"55468b7c-f04a-4b1c-bb57-09d8b71051f0"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "However ",
 StyleBox["xTensor`", "DisplayMath"],
 " recognizes them as equal:"
}], "MathCaption",ExpressionUUID->"61cf9248-7fb6-4821-bf60-390e7a7d6a27"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1305]:=",ExpressionUUID->"51369881-802e-4b76-84e1-518cb1996f5a"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{3.728978978956438*^9},
 CellLabel->"Out[1305]=",
 CellID->1505811578,ExpressionUUID->"efb27cbb-4278-4578-aecd-443c260e7ea8"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["These two are clearly different:", "MathCaption",ExpressionUUID->"ceea1a38-49ec-45fd-a9f0-9ddab3502197"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Simplification", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "b"}], "]"}]}], "\[Equal]", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", "c"}], "]"}], 
    RowBox[{"v", "[", 
     RowBox[{"-", "c"}], "]"}]}]}], "]"}]], "Input",
 CellLabel->
  "In[1306]:=",ExpressionUUID->"b96d886a-552f-4290-9825-7d6362347414"],

Cell[BoxData["False"], "Output",
 CellChangeTimes->{3.728978979092833*^9},
 CellLabel->"Out[1306]=",
 CellID->80411551,ExpressionUUID->"448c2c8b-01ae-4225-ae9b-911f6fd31bf4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Solving equations. The answer is a tensor rule. Options to ",
 StyleBox["IndexSolve",
  FontFamily->"Courier"],
 " are actually options to ",
 StyleBox["MakeRule",
  FontFamily->"Courier"],
 ":"
}], "MathCaption",ExpressionUUID->"3f13f80b-1730-4627-9e9e-07d12d56f1ce"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"rule", "=", 
  RowBox[{"IndexSolve", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"v", "[", "a", "]"}], 
      RowBox[{"v", "[", 
       RowBox[{"-", "a"}], "]"}], 
      RowBox[{"T", "[", 
       RowBox[{"b", ",", "c", ",", "d"}], "]"}]}], "\[Equal]", 
     RowBox[{
      RowBox[{"v", "[", "b", "]"}], 
      RowBox[{"v", "[", "c", "]"}], 
      RowBox[{"v", "[", "a", "]"}], 
      RowBox[{"S", "[", 
       RowBox[{
        RowBox[{"-", "a"}], ",", "d"}], "]"}]}]}], ",", 
    RowBox[{"T", "[", 
     RowBox[{"b", ",", "c", ",", "d"}], "]"}], ",", " ", 
    RowBox[{"MetricOn", "\[Rule]", 
     RowBox[{"{", "b", "}"}]}], ",", 
    RowBox[{"ContractMetrics", "\[Rule]", "True"}]}], "]"}]}]], "Input",
 CellLabel->
  "In[1307]:=",ExpressionUUID->"4b846980-45ef-487f-9af7-70e10a1faa88"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"HoldPattern", "[", 
    InterpretationBox[
     StyleBox[GridBox[{
        {"T", 
         StyleBox[GridBox[{
            {
             UnderscriptBox[
              UnderscriptBox["b", "_"], "_"], 
             UnderscriptBox[
              UnderscriptBox["c", "_"], "_"], 
             UnderscriptBox[
              UnderscriptBox["d", "_"], "_"]},
            {" ", " ", " "}
           },
           GridBoxSpacings->{"Columns" -> {
               Offset[0.], {
                Offset[0.034999999999999996`]}, 
               Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                Offset[0.]}}, "RowsIndexed" -> {}}],
          FontSize->9]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}},
          "RowsIndexed" -> {}},
       GridBoxSpacings->{"Columns" -> {
           Offset[0.27999999999999997`], {
            Offset[0.034999999999999996`]}, 
           Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
           Offset[0.2], {
            Offset[0.4]}, 
           Offset[0.2]}, "RowsIndexed" -> {}}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     $CellContext`T[
      PatternTest[
       Pattern[$CellContext`b, 
        Blank[]], TangentM3`pmQ], 
      PatternTest[
       Pattern[$CellContext`c, 
        Blank[Symbol]], TangentM3`Q], 
      PatternTest[
       Pattern[$CellContext`d, 
        Blank[Symbol]], TangentM3`Q]],
     Editable->False], "]"}], "\[RuleDelayed]", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", "a", "}"}], ",", 
     FractionBox[
      RowBox[{
       InterpretationBox[
        StyleBox[GridBox[{
           {"S", 
            StyleBox[GridBox[{
               {" ", "d"},
               {"a", " "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`S[-$CellContext`a, $CellContext`d],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"b"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[$CellContext`b],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"c"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[$CellContext`c],
        Editable->False], " ", 
       InterpretationBox[
        StyleBox[GridBox[{
           {"v", 
            StyleBox[GridBox[{
               {"a"},
               {" "}
              },
              GridBoxSpacings->{"Columns" -> {
                  Offset[0.], {
                   Offset[0.034999999999999996`]}, 
                  Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                   Offset[0.]}}, "RowsIndexed" -> {}}],
             FontSize->9]}
          },
          
          GridBoxAlignment->{
           "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
            "Rows" -> {{Center}}, "RowsIndexed" -> {}},
          GridBoxSpacings->{"Columns" -> {
              Offset[0.27999999999999997`], {
               Offset[0.034999999999999996`]}, 
              Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
              Offset[0.2], {
               Offset[0.4]}, 
              Offset[0.2]}, "RowsIndexed" -> {}}],
         ShowAutoStyles->False,
         AutoSpacing->False],
        $CellContext`v[$CellContext`a],
        Editable->False]}], 
      InterpretationBox[
       StyleBox[
        RowBox[{
         StyleBox["(",
          FontColor->RGBColor[1, 0, 0]], 
         RowBox[{
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {" "},
                  {"a"}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[-$CellContext`a],
           Editable->False], " ", 
          InterpretationBox[
           StyleBox[GridBox[{
              {"v", 
               StyleBox[GridBox[{
                  {"a"},
                  {" "}
                 },
                 GridBoxSpacings->{"Columns" -> {
                    Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                    Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
                FontSize->9]}
             },
             
             GridBoxAlignment->{
              "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
               "Rows" -> {{Center}}, "RowsIndexed" -> {}},
             GridBoxSpacings->{"Columns" -> {
                 Offset[0.27999999999999997`], {
                  Offset[0.034999999999999996`]}, 
                 Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
               "Rows" -> {
                 Offset[0.2], {
                  Offset[0.4]}, 
                 Offset[0.2]}, "RowsIndexed" -> {}}],
            ShowAutoStyles->False,
            AutoSpacing->False],
           $CellContext`v[$CellContext`a],
           Editable->False]}], 
         StyleBox[")",
          FontColor->RGBColor[1, 0, 0]]}],
        ShowAutoStyles->False,
        AutoSpacing->False],
       xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
       Editable->False]]}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{3.728978979236088*^9},
 CellLabel->"Out[1307]=",
 CellID->709723694,ExpressionUUID->"bbf5732b-bbda-46f4-b8ef-f25e24a613b9"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{
    RowBox[{"T", "[", 
     RowBox[{"a", ",", "b", ",", "c"}], "]"}], ",", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", "b", ",", "c"}], "]"}], ",", 
    RowBox[{"T", "[", 
     RowBox[{"a", ",", 
      RowBox[{"-", "b"}], ",", "c"}], "]"}], ",", 
    RowBox[{"T", "[", 
     RowBox[{
      RowBox[{"-", "a"}], ",", 
      RowBox[{"-", "b"}], ",", "c"}], "]"}]}], "}"}], "/.", "rule"}]], "Input",\

 CellLabel->
  "In[1308]:=",ExpressionUUID->"c941c480-0e60-482e-bee3-4e85c69dff1d"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   FractionBox[
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[-$CellContext`d, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"a"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`d],
      Editable->False]}], 
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[-$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
     Editable->False]], ",", 
   FractionBox[
    RowBox[{
     InterpretationBox[
      StyleBox[GridBox[{
         {"S", 
          StyleBox[GridBox[{
             {" ", "c"},
             {"d", " "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`S[-$CellContext`d, $CellContext`c],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {" "},
             {"a"}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[-$CellContext`a],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"b"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`b],
      Editable->False], " ", 
     InterpretationBox[
      StyleBox[GridBox[{
         {"v", 
          StyleBox[GridBox[{
             {"d"},
             {" "}
            },
            GridBoxSpacings->{"Columns" -> {
                Offset[0.], {
                 Offset[0.034999999999999996`]}, 
                Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                 Offset[0.]}}, "RowsIndexed" -> {}}],
           FontSize->9]}
        },
        GridBoxAlignment->{
         "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Center}}, "RowsIndexed" -> {}},
        GridBoxSpacings->{"Columns" -> {
            Offset[0.27999999999999997`], {
             Offset[0.034999999999999996`]}, 
            Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, 
          "Rows" -> {
            Offset[0.2], {
             Offset[0.4]}, 
            Offset[0.2]}, "RowsIndexed" -> {}}],
       ShowAutoStyles->False,
       AutoSpacing->False],
      $CellContext`v[$CellContext`d],
      Editable->False]}], 
    InterpretationBox[
     StyleBox[
      RowBox[{
       StyleBox["(",
        FontColor->RGBColor[1, 0, 0]], 
       RowBox[{
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {" "},
                {"a"}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[-$CellContext`a],
         Editable->False], " ", 
        InterpretationBox[
         StyleBox[GridBox[{
            {"v", 
             StyleBox[GridBox[{
                {"a"},
                {" "}
               },
               GridBoxSpacings->{"Columns" -> {
                   Offset[0.], {
                    Offset[0.034999999999999996`]}, 
                   Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
                    Offset[0.]}}, "RowsIndexed" -> {}}],
              FontSize->9]}
           },
           
           GridBoxAlignment->{
            "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, 
             "Rows" -> {{Center}}, "RowsIndexed" -> {}},
           GridBoxSpacings->{"Columns" -> {
               Offset[0.27999999999999997`], {
                Offset[0.034999999999999996`]}, 
               Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
               Offset[0.2], {
                Offset[0.4]}, 
               Offset[0.2]}, "RowsIndexed" -> {}}],
          ShowAutoStyles->False,
          AutoSpacing->False],
         $CellContext`v[$CellContext`a],
         Editable->False]}], 
       StyleBox[")",
        FontColor->RGBColor[1, 0, 0]]}],
      ShowAutoStyles->False,
      AutoSpacing->False],
     xAct`xTensor`Scalar[$CellContext`v[-$CellContext`a] \
$CellContext`v[$CellContext`a]],
     Editable->False]], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {"a", " ", "c"},
           {" ", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[$CellContext`a, -$CellContext`b, $CellContext`c],
    Editable->False], ",", 
   InterpretationBox[
    StyleBox[GridBox[{
       {"T", 
        StyleBox[GridBox[{
           {" ", " ", "c"},
           {"a", "b", " "}
          },
          GridBoxSpacings->{"Columns" -> {
              Offset[0.], {
               Offset[0.034999999999999996`]}, 
              Offset[0.]}, "ColumnsIndexed" -> {}, "Rows" -> {{
               Offset[0.]}}, "RowsIndexed" -> {}}],
         FontSize->9]}
      },
      GridBoxAlignment->{
       "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Center}}, 
        "RowsIndexed" -> {}},
      GridBoxSpacings->{"Columns" -> {
          Offset[0.27999999999999997`], {
           Offset[0.034999999999999996`]}, 
          Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
          Offset[0.2], {
           Offset[0.4]}, 
          Offset[0.2]}, "RowsIndexed" -> {}}],
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`T[-$CellContext`a, -$CellContext`b, $CellContext`c],
    Editable->False]}], "}"}]], "Output",
 CellChangeTimes->{3.728978979367464*^9},
 CellLabel->"Out[1308]=",
 CellID->1201949620,ExpressionUUID->"0806bfe2-b54c-470f-9c4e-0b67670d0f34"]
}, Open  ]]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["10.11. Clean up", "Subsection",ExpressionUUID->"9e102fec-dd1b-4b26-8b3d-135b944df6da"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefMetric", "[", "metricg", "]"}]], "Input",
 CellLabel->
  "In[1309]:=",ExpressionUUID->"63987a5e-f110-4dd1-940e-140cf3b4baec"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefCovD", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"covariant derivative\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "CD"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefCovD, ": Undefined ", "covariant derivative", 
   " ", $CellContext`CD],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789821774387`*^9},
 CellLabel->"During evaluation of In[1309]:=",
 CellID->1894719997,ExpressionUUID->"0d699409-e5c6-4ec2-adda-bc0401471fb9"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"antisymmetric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "epsilonmetricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "antisymmetric tensor", 
   " ", $CellContext`epsilonmetricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789821944523`*^9},
 CellLabel->"During evaluation of In[1309]:=",
 CellID->1671471863,ExpressionUUID->"7d8df572-1d72-47bc-8d64-e834fb317518"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"symmetric metric tensor\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", "metricg"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "symmetric metric tensor", 
   " ", $CellContext`metricg],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789822109327`*^9},
 CellLabel->"During evaluation of In[1309]:=",
 CellID->1194015034,ExpressionUUID->"d1b39128-a058-437d-a7b0-cc30cae4bc63"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"UndefTensor", "/@", "$Tensors"}], ";"}]], "Input",
 CellLabel->
  "In[1310]:=",ExpressionUUID->"3188f92f-4028-4cb1-901b-213d0840f6ea"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "v"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`v],
  Editable->False]], "Print",
 CellChangeTimes->{3.72897898238481*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->1723478870,ExpressionUUID->"849a03c8-36f9-4c1c-9edc-f39881820b8a"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "T"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`T],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978982401678*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->1249937354,ExpressionUUID->"4ced1b35-08f4-4cc5-8c5a-c8406aedca9d"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "U"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`U],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789824181633`*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->1012073724,ExpressionUUID->"85cf0a92-4fdf-4fd8-acf1-bc5118d8d30f"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "r"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`r],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978982434326*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->1852595028,ExpressionUUID->"51e08db8-5cd2-4794-bc7a-ac8f6d814243"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`S],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789824505672`*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->199237743,ExpressionUUID->"da4ea3f5-6e60-4e82-8e3b-19bb00f71fc5"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "n"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`n],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978982466817*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->661740928,ExpressionUUID->"e19e43bf-9b20-4d93-b245-6cd7b7a1c1ba"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefTensor", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"tensor\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "G"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefTensor, ": Undefined ", "tensor", 
   " ", $CellContext`G],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789824832287`*^9},
 CellLabel->"During evaluation of In[1310]:=",
 CellID->61969083,ExpressionUUID->"3a6d9e5c-3d8c-4526-8af6-a61b35226cf5"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "S2", "]"}]], "Input",
 CellLabel->
  "In[1311]:=",ExpressionUUID->"03bad5e8-bcb2-425a-bbba-f56daaa3271e"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]S2",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentS2,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentS2],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978982635165*^9},
 CellLabel->"During evaluation of In[1311]:=",
 CellID->2132292833,ExpressionUUID->"74f75195-ea60-443b-9c67-687fb4a21521"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "S2"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`S2],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789826518087`*^9},
 CellLabel->"During evaluation of In[1311]:=",
 CellID->238543095,ExpressionUUID->"fd3743e5-4bfb-49c8-84c9-a0c1b7c3987e"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefManifold", "[", "M3", "]"}]], "Input",
 CellLabel->
  "In[1312]:=",ExpressionUUID->"8635e9e6-ab5c-42be-a250-230a769d71bf"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefVBundle", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"vbundle\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["\[DoubleStruckCapitalT]M3",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`TangentM3,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefVBundle, ": Undefined ", "vbundle", 
   " ", $CellContext`TangentM3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789830482903`*^9},
 CellLabel->"During evaluation of In[1312]:=",
 CellID->1989312363,ExpressionUUID->"0569c43e-59aa-4eb0-b9e2-bff683d40a1b"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefManifold", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"manifold\"\>", "\[InvisibleSpace]", "\<\" \"\>", 
   "\[InvisibleSpace]", "M3"}],
  SequenceForm[
  "** ", xAct`xTensor`UndefManifold, ": Undefined ", "manifold", 
   " ", $CellContext`M3],
  Editable->False]], "Print",
 CellChangeTimes->{3.7289789830661783`*^9},
 CellLabel->"During evaluation of In[1312]:=",
 CellID->199984798,ExpressionUUID->"b1101d58-8cd8-43f1-b3f0-bf542c8286b1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"UndefScalarFunction", "[", "SF", "]"}]], "Input",
 CellLabel->
  "In[1313]:=",ExpressionUUID->"2a2e5a4f-6c77-47c6-9ed1-560d61bac3c6"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"** \"\>", "\[InvisibleSpace]", "UndefScalarFunction", 
   "\[InvisibleSpace]", "\<\": Undefined \"\>", 
   "\[InvisibleSpace]", "\<\"scalar function\"\>", 
   "\[InvisibleSpace]", "\<\" \"\>", "\[InvisibleSpace]", 
   InterpretationBox[
    StyleBox["SF",
     ShowAutoStyles->False,
     AutoSpacing->False],
    $CellContext`SF,
    Editable->False]}],
  SequenceForm[
  "** ", xAct`xTensor`UndefScalarFunction, ": Undefined ", "scalar function", 
   " ", $CellContext`SF],
  Editable->False]], "Print",
 CellChangeTimes->{3.728978983185851*^9},
 CellLabel->"During evaluation of In[1313]:=",
 CellID->332592431,ExpressionUUID->"f3f984b6-1520-4855-b8c6-8e6cce0a0a88"]
}, Open  ]]
}, Closed]]
}, Open  ]],

Cell[CellGroupData[{

Cell["11. Final comments", "Section",
 CellFrame->{{0, 0}, {0, 0.5}},
 CellChangeTimes->{{3.7286627354220657`*^9, 3.728662748172317*^9}, {
  3.728664037492599*^9, 3.72866403885896*^9}, {3.7286704747855062`*^9, 
  3.728670480415387*^9}, {3.728670889883876*^9, 3.728670893540101*^9}, {
  3.728673410238123*^9, 3.728673425886174*^9}, {3.7289761721698513`*^9, 
  3.7289761765432653`*^9}, {3.728976277988899*^9, 3.728976287866613*^9}, {
  3.728976341130004*^9, 3.728976342960009*^9}, {3.72897639317553*^9, 
  3.728976399469426*^9}, {3.728976517721738*^9, 3.7289765237113934`*^9}, {
  3.7289765976539173`*^9, 3.72897660356386*^9}, {3.728976670882564*^9, 
  3.728976674720446*^9}, {3.7289768454424953`*^9, 3.728976848920129*^9}, {
  3.728977115725004*^9, 3.7289771217954493`*^9}, {3.728977171203144*^9, 
  3.728977176113048*^9}, {3.72897727200626*^9, 3.7289772761003437`*^9}},
 FontSize->18,
 FontColor->GrayLevel[0],
 CellTags->{"S5.18.1", "1.4"},
 CellID->292931816,ExpressionUUID->"60102f35-97c9-49fb-ac37-e2c30aabc8e1"],

Cell[CellGroupData[{

Cell["11.1. ToDo list. Version", "Subsection",ExpressionUUID->"16b126b0-e45c-413a-b713-bd224f7c8574"],

Cell[TextData[{
 "There are many things that ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 " cannot do yet. Here there is a list with some of them:\n\t- Differential \
forms\n\t- Grasmann variables"
}], "Text",
 CellChangeTimes->{{3.4612409082777452`*^9, 
  3.461240909343998*^9}},ExpressionUUID->"9bf0f838-d02c-4590-a0f9-\
a84476c860e5"],

Cell[TextData[{
 StyleBox["Note:", "TB"],
 " For further information about ",
 StyleBox["xTensor`", "MR"],
 ", and to be kept informed about new releases, you may contact the author \
electronically at jose@xact.es. This is xTensorDoc.nb, the docfile of ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ", currently in version 1.1.3."
}], "Text",ExpressionUUID->"8130a58f-c2e0-469c-ae40-a12f297a8dd7"],

Cell[TextData[{
 "The latest version of the package can always be downloaded from   ",
 ButtonBox["http://www.xAct.es/",
  BaseStyle->"Hyperlink",
  ButtonData->{
    URL["http://www.xact.es/"], None},
  ButtonNote->"http://www.xact.es/"]
}], "Text",
 CellID->983232810,ExpressionUUID->"8f4bcdcd-88dc-4765-a3c5-264f84a6227e"]
}, Closed]],

Cell[CellGroupData[{

Cell["11.2. Frequently Asked Questions", "Subsection",ExpressionUUID->"445c4b78-68c5-40bf-921f-1b70fa4b00bf"],

Cell["\<\
1. Why is delta contraction automatic? I want to have full control on every \
single detail of the process.
    Why is metric contraction not automatic? It is boring to ask continuously \
for trivial tasks.\
\>", "Text",ExpressionUUID->"773cc8d8-26c1-48ea-8659-4aba568c0e15"],

Cell[TextData[{
 "There are two opposite approaches to the issue of automatic simplification: \
Too little automatic simplification makes the computations hard and slow, \
while too much automatic simplification makes the process too restrictive and \
the result seems just magic. Following ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", we have tried to automatize all those things which are commonly required, \
like contraction of any tensor with delta, or the expansion of derivatives of \
products using the Leibnitz rule. However, most of the processes are not \
automatic: simplification, metric contraction, commutation of derivatives, \
etc.\nThere are some processes for which the user can choose whether they are \
automatic or not: look for the pairs commandStart / commandStop."
}], "MathCaption",ExpressionUUID->"7b2ce55e-9501-48aa-9698-8ac0f91131f2"],

Cell[TextData[{
 "2. I want to define several tensors at the same time. Why not allowing this \
syntax? \n\t\t\t",
 StyleBox["DefTensor[{T[a], S[b,c], R[-a,-b,-c,-d}, M]",
  FontFamily->"Courier"]
}], "Text",ExpressionUUID->"69fe4b03-c970-4d33-b472-c77368ccf6cb"],

Cell[TextData[{
 "We do not include new definitions which do not save time or space with \
respect to their ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " counterparts. This helps learning ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", and save space in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ". In that particular example we recommend to use \n\t\t",
 StyleBox["Map[ DefTensor[#, M]&, {T[a], S[b,c], R[-a,-b,-c,-d]} ]\n",
  FontFamily->"Courier"],
 "or simply threading\n\t\t",
 StyleBox["Thread[ DefTensor[{T[a], S[b,c], R[-a,-b,-c,-d}, M] ]",
  FontFamily->"Courier"]
}], "MathCaption",ExpressionUUID->"6eb445ea-32ea-4bf4-a609-2180aba05b1a"],

Cell[TextData[{
 "3. In ",
 StyleBox["MathTensor",
  FontSlant->"Italic"],
 " many commands have an additional argument to map them only to a certain \
part of an expression. Why don't you have the same thing?"
}], "Text",ExpressionUUID->"43a15891-8e32-4156-b3af-b77a0f4ac331"],

Cell[TextData[{
 "The answer is identical to that of question 2. Use the syntax ",
 StyleBox["MapAt[f, expr, n]",
  FontFamily->"Courier"],
 " to map the function ",
 StyleBox["f",
  FontFamily->"Courier"],
 " on the ",
 StyleBox["n",
  FontFamily->"Courier"],
 "-th term of ",
 StyleBox["expr",
  FontFamily->"Courier"],
 "."
}], "MathCaption",ExpressionUUID->"2848bfa1-d880-4e6e-bf62-a20ed756fbdf"],

Cell[TextData[{
 "4. ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " is great for abstract tensor calculus, but I want to do component \
computations. Can I do that with ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 "?"
}], "Text",ExpressionUUID->"5b07a210-037f-4997-b0c9-7d388ebf5bb7"],

Cell[TextData[{
 "No, there is no support in ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " for component tensor computer algebra. The twin package ",
 StyleBox["xCoba",
  FontSlant->"Italic"],
 " is being developed for that purpose, but it does not have the degree of \
maturity of ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 "."
}], "MathCaption",ExpressionUUID->"d5368906-9eaf-423e-b4b7-b3d834476af1"],

Cell[TextData[{
 "5. I want to do 3+1 decompositions (ADM-type). ",
 StyleBox["xTensor",
  FontSlant->"Italic"],
 " has a four-dimensional approach to that. Can I do the same with the \
traditional {i, 0} index splitting formalism?"
}], "Text",ExpressionUUID->"aa2315df-ece4-4ee0-bf92-e93744aca77e"],

Cell[TextData[{
 "Yes, this is possible, but there are no predefined tools to help in the \
process. You need to code them up using the building blocks: define three \
manifolds: 4d, 3d and 1d. Then use ",
 StyleBox["SplitIndex",
  FontFamily->"Courier"],
 " and ",
 StyleBox["TraceProductDummy",
  FontFamily->"Courier"],
 " for the computations."
}], "MathCaption",ExpressionUUID->"19095ea6-048c-40ff-ba3e-8a9caea50eea"]
}, Closed]],

Cell[CellGroupData[{

Cell["11.3. Statistics", "Subsection",ExpressionUUID->"4a0ec84b-57f6-45d5-ae86-2fd979f0579e"],

Cell[CellGroupData[{

Cell[TextData[{
 "Collection of public symbols in ",
 StyleBox["xTensor`",
  FontFamily->"Courier"],
 ":"
}], "Text",ExpressionUUID->"466fe0e3-1233-4ced-ae6f-3ebb5dd1d87a"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Names", "[", "\"\<xAct`xTensor`*\>\"", "]"}]], "Input",
 CellLabel->
  "In[1314]:=",ExpressionUUID->"14963688-618e-4087-91d9-575435758978"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"\<\"ABIndexQ\"\>", ",", "\<\"AbstractIndex\"\>", 
   ",", "\<\"AbstractIndexQ\"\>", ",", "\<\"Acceleration\"\>", 
   ",", "\<\"AChristoffel\"\>", ",", "\<\"AddIndices\"\>", 
   ",", "\<\"AIndex\"\>", ",", "\<\"AIndexQ\"\>", 
   ",", "\<\"AllowUpperDerivatives\"\>", ",", "\<\"Anticommutator\"\>", 
   ",", "\<\"Antihermitian\"\>", ",", "\<\"AntihermitianQ\"\>", 
   ",", "\<\"Antisymmetrize\"\>", ",", "\<\"AnyDependencies\"\>", 
   ",", "\<\"AnyIndices\"\>", ",", "\<\"AnyVBundles\"\>", 
   ",", "\<\"AssociativeProductQ\"\>", ",", "\<\"At\"\>", 
   ",", "\<\"AutomaticRules\"\>", ",", "\<\"BaseOfVBundle\"\>", 
   ",", "\<\"Basis\"\>", ",", "\<\"BasisQ\"\>", ",", "\<\"BCIndexQ\"\>", 
   ",", "\<\"BIndex\"\>", ",", "\<\"BIndexQ\"\>", ",", "\<\"Blocked\"\>", 
   ",", "\<\"BlockedQ\"\>", ",", "\<\"Bracket\"\>", 
   ",", "\<\"BracketToCovD\"\>", ",", "\<\"BreakChristoffel\"\>", 
   ",", "\<\"BreakInMonomials\"\>", ",", "\<\"BreakScalars\"\>", 
   ",", "\<\"CartesianProduct\"\>", ",", "\<\"CDIndexQ\"\>", 
   ",", "\<\"ChangeCovD\"\>", ",", "\<\"ChangeCurvature\"\>", 
   ",", "\<\"ChangeFreeIndices\"\>", ",", "\<\"ChangeIndex\"\>", 
   ",", "\<\"ChangeTorsion\"\>", ",", "\<\"Chart\"\>", ",", "\<\"ChartQ\"\>", 
   ",", "\<\"CheckZeroDerivative\"\>", 
   ",", "\<\"CheckZeroDerivativeStart\"\>", 
   ",", "\<\"CheckZeroDerivativeStop\"\>", ",", "\<\"Christoffel\"\>", 
   ",", "\<\"ChristoffelToGradConformal\"\>", 
   ",", "\<\"ChristoffelToGradMetric\"\>", ",", "\<\"ChristoffelToMetric\"\>",
    ",", "\<\"CIndex\"\>", ",", "\<\"CIndexForm\"\>", ",", "\<\"CIndexQ\"\>", 
   ",", "\<\"ColorPositionsOfPattern\"\>", ",", "\<\"ColorTerms\"\>", 
   ",", "\<\"CommutativityOfProduct\"\>", ",", "\<\"Commutator\"\>", 
   ",", "\<\"CommuteCovDs\"\>", ",", "\<\"CommutePDs\"\>", 
   ",", "\<\"ConformalFactor\"\>", ",", "\<\"ConformalRules\"\>", 
   ",", "\<\"ConformalTo\"\>", ",", "\<\"ConstantMetric\"\>", 
   ",", "\<\"ConstantQ\"\>", ",", "\<\"ConstantSymbol\"\>", 
   ",", "\<\"ConstantSymbolQ\"\>", ",", "\<\"ContractCurvature\"\>", 
   ",", "\<\"ContractDir\"\>", ",", "\<\"ContractMetric\"\>", 
   ",", "\<\"ContractMetrics\"\>", ",", "\<\"ContractProjector\"\>", 
   ",", "\<\"ContractThrough\"\>", ",", "\<\"ContractThroughQ\"\>", 
   ",", "\<\"CovD\"\>", ",", "\<\"CovDOfMetric\"\>", ",", "\<\"CovDQ\"\>", 
   ",", "\<\"CovDToChristoffel\"\>", ",", "\<\"Curvature\"\>", 
   ",", "\<\"CurvatureQ\"\>", ",", "\<\"CurvatureRelations\"\>", 
   ",", "\<\"Cyclize\"\>", ",", "\<\"Dagger\"\>", ",", "\<\"DaggerIndex\"\>", 
   ",", "\<\"DaggerQ\"\>", ",", "\<\"DefAbstractIndex\"\>", 
   ",", "\<\"DefConstantSymbol\"\>", ",", "\<\"DefCovD\"\>", 
   ",", "\<\"DefineCommutator\"\>", ",", "\<\"DefInertHead\"\>", 
   ",", "\<\"DefInfo\"\>", ",", "\<\"DefManifold\"\>", 
   ",", "\<\"DefMapping\"\>", ",", "\<\"DefMetric\"\>", 
   ",", "\<\"DefParameter\"\>", ",", "\<\"DefProduct\"\>", 
   ",", "\<\"DefProductMetric\"\>", ",", "\<\"DefScalarFunction\"\>", 
   ",", "\<\"DefTensor\"\>", ",", "\<\"DefVBundle\"\>", ",", "\<\"delta\"\>", 
   ",", "\<\"DependenciesOf\"\>", ",", "\<\"DependenciesOfCovD\"\>", 
   ",", "\<\"DependenciesOfInertHead\"\>", 
   ",", "\<\"DependenciesOfTensor\"\>", ",", "\<\"Determinant\"\>", 
   ",", "\<\"DiffeomorphismQ\"\>", ",", "\<\"DimOfManifold\"\>", 
   ",", "\<\"DimOfVBundle\"\>", ",", "\<\"DIndex\"\>", ",", "\<\"DIndexQ\"\>",
    ",", "\<\"Dir\"\>", ",", "\<\"DirCovDToLieD\"\>", 
   ",", "\<\"Disclaimer\"\>", ",", "\<\"DisjointManifoldsQ\"\>", 
   ",", "\<\"DisjointVBundlesQ\"\>", ",", "\<\"DisorderedPairQ\"\>", 
   ",", "\<\"Down\"\>", ",", "\<\"DownIndex\"\>", ",", "\<\"DownIndexQ\"\>", 
   ",", "\<\"DownVectorQ\"\>", ",", "\<\"Dummy\"\>", ",", "\<\"DummyAs\"\>", 
   ",", "\<\"DummyIn\"\>", ",", "\<\"EIndexQ\"\>", ",", "\<\"Einstein\"\>", 
   ",", "\<\"EinsteinToRicci\"\>", ",", "\<\"EmptyManifold\"\>", 
   ",", "\<\"epsilon\"\>", ",", "\<\"epsilonOrientation\"\>", 
   ",", "\<\"epsilonOrientationInBasis\"\>", ",", "\<\"EqualExpressionsQ\"\>",
    ",", "\<\"ERROR\"\>", ",", "\<\"ExpandCommutator\"\>", 
   ",", "\<\"ExpandGdelta\"\>", ",", "\<\"ExpandProductMetric\"\>", 
   ",", "\<\"ExpandPullBack\"\>", ",", "\<\"ExpandPushForward\"\>", 
   ",", "\<\"ExpandSdelta\"\>", ",", "\<\"Explode\"\>", 
   ",", "\<\"ExtendedFrom\"\>", ",", "\<\"ExtrinsicK\"\>", 
   ",", "\<\"ExtrinsicKToGradNormal\"\>", ",", "\<\"FindAllOfType\"\>", 
   ",", "\<\"FindBlockedIndices\"\>", ",", "\<\"FindDummyIndices\"\>", 
   ",", "\<\"FindFreeIndices\"\>", ",", "\<\"FindIndices\"\>", 
   ",", "\<\"FirstDerQ\"\>", ",", "\<\"FlatMetric\"\>", 
   ",", "\<\"FlatMetricQ\"\>", ",", "\<\"ForceSymmetries\"\>", 
   ",", "\<\"Free\"\>", ",", "\<\"FRiemann\"\>", ",", "\<\"FrobeniusQ\"\>", 
   ",", "\<\"FromMetric\"\>", ",", "\<\"FromProjected\"\>", 
   ",", "\<\"FromStripped\"\>", ",", "\<\"FromTensorDerivative\"\>", 
   ",", "\<\"GaussCodazzi\"\>", ",", "\<\"Gdelta\"\>", 
   ",", "\<\"GetIndicesOfVBundle\"\>", ",", "\<\"GIndexQ\"\>", 
   ",", "\<\"GiveOutputString\"\>", ",", "\<\"GiveSymbol\"\>", 
   ",", "\<\"Grade\"\>", ",", "\<\"GradedProductQ\"\>", 
   ",", "\<\"GradeOfCovD\"\>", ",", "\<\"GradeOfProduct\"\>", 
   ",", "\<\"GradeOfTensor\"\>", ",", "\<\"GradMetricToChristoffel\"\>", 
   ",", "\<\"GradNormalToExtrinsicK\"\>", ",", "\<\"HeadOfTensor\"\>", 
   ",", "\<\"Hermitian\"\>", ",", "\<\"HermitianQ\"\>", 
   ",", "\<\"HostsOf\"\>", ",", "\<\"IdentityElementOfProduct\"\>", 
   ",", "\<\"IdentityMapping\"\>", ",", "\<\"Imaginary\"\>", 
   ",", "\<\"ImmersionQ\"\>", ",", "\<\"IModQ\"\>", 
   ",", "\<\"ImplicitTensorDepQ\"\>", ",", "\<\"Implode\"\>", 
   ",", "\<\"ImposeSymmetry\"\>", ",", "\<\"IndexCoefficient\"\>", 
   ",", "\<\"IndexCollect\"\>", ",", "\<\"IndexForm\"\>", 
   ",", "\<\"IndexList\"\>", ",", "\<\"IndexOrderedQ\"\>", 
   ",", "\<\"IndexRange\"\>", ",", "\<\"IndexRule\"\>", 
   ",", "\<\"IndexRuleDelayed\"\>", ",", "\<\"IndexSet\"\>", 
   ",", "\<\"IndexSetDelayed\"\>", ",", "\<\"IndexSolve\"\>", 
   ",", "\<\"IndexSort\"\>", ",", "\<\"IndicesOf\"\>", 
   ",", "\<\"IndicesOfVBundle\"\>", ",", "\<\"InducedDecomposition\"\>", 
   ",", "\<\"InducedFrom\"\>", ",", "\<\"InertHead\"\>", 
   ",", "\<\"InertHeadQ\"\>", ",", "\<\"Inv\"\>", 
   ",", "\<\"InverseMapping\"\>", ",", "\<\"InvertibleQ\"\>", 
   ",", "\<\"IsIndexOf\"\>", ",", "\<\"Keep\"\>", ",", "\<\"Kretschmann\"\>", 
   ",", "\<\"Labels\"\>", ",", "\<\"LastIndex\"\>", ",", "\<\"LC\"\>", 
   ",", "\<\"Leibnitz\"\>", ",", "\<\"LI\"\>", ",", "\<\"LieD\"\>", 
   ",", "\<\"LieDToCovD\"\>", ",", "\<\"LIndex\"\>", ",", "\<\"LIndexQ\"\>", 
   ",", "\<\"LinearPush\"\>", ",", "\<\"LinearQ\"\>", 
   ",", "\<\"MakeProjectors\"\>", ",", "\<\"MakeRule\"\>", 
   ",", "\<\"Manifold\"\>", ",", "\<\"ManifoldBoundary\"\>", 
   ",", "\<\"ManifoldOfCovD\"\>", ",", "\<\"ManifoldQ\"\>", 
   ",", "\<\"ManifoldsOf\"\>", ",", "\<\"Mapping\"\>", 
   ",", "\<\"MappingDomain\"\>", ",", "\<\"MappingImage\"\>", 
   ",", "\<\"MappingQ\"\>", ",", "\<\"Master\"\>", ",", "\<\"MasterOf\"\>", 
   ",", "\<\"Metric\"\>", ",", "\<\"MetricEndowedQ\"\>", 
   ",", "\<\"MetricOfCovD\"\>", ",", "\<\"MetricOn\"\>", 
   ",", "\<\"MetricQ\"\>", ",", "\<\"MetricScalar\"\>", 
   ",", "\<\"MetricsOfVBundle\"\>", ",", "\<\"MetricToProjector\"\>", 
   ",", "\<\"Monomial\"\>", ",", "\<\"MultiplyHead\"\>", 
   ",", "\<\"NewIndexIn\"\>", ",", "\<\"NoScalar\"\>", 
   ",", "\<\"OrthogonalTo\"\>", ",", "\<\"OrthogonalToVectorQ\"\>", 
   ",", "\<\"OrthogonalWithMetric\"\>", ",", "\<\"OtherDependencies\"\>", 
   ",", "\<\"OverDerivatives\"\>", ",", "\<\"Pair\"\>", 
   ",", "\<\"PairAntisymmetrize\"\>", ",", "\<\"PairQ\"\>", 
   ",", "\<\"PairSymmetrize\"\>", ",", "\<\"ParamD\"\>", 
   ",", "\<\"ParameterQ\"\>", ",", "\<\"ParametersOf\"\>", 
   ",", "\<\"PatternIndex\"\>", ",", "\<\"PatternIndices\"\>", 
   ",", "\<\"PD\"\>", ",", "\<\"PermuteIndices\"\>", ",", "\<\"PIndex\"\>", 
   ",", "\<\"PIndexQ\"\>", ",", "\<\"Precompose\"\>", 
   ",", "\<\"PrimeDagger\"\>", ",", "\<\"PrintAs\"\>", 
   ",", "\<\"PrintAsCharacter\"\>", ",", "\<\"ProductByScalar\"\>", 
   ",", "\<\"ProductOfScalars\"\>", ",", "\<\"ProductQ\"\>", 
   ",", "\<\"Project\"\>", ",", "\<\"ProjectDerivative\"\>", 
   ",", "\<\"Projected\"\>", ",", "\<\"ProjectedWith\"\>", 
   ",", "\<\"Projector\"\>", ",", "\<\"ProjectorQ\"\>", 
   ",", "\<\"ProjectorToMetric\"\>", ",", "\<\"ProjectWith\"\>", 
   ",", "\<\"ProtectNewSymbol\"\>", ",", "\<\"PullBack\"\>", 
   ",", "\<\"PullBackCovD\"\>", ",", "\<\"PullBackTensor\"\>", 
   ",", "\<\"PullBackVBundle\"\>", ",", "\<\"PushForward\"\>", 
   ",", "\<\"PushForwardTensor\"\>", ",", "\<\"PutScalar\"\>", 
   ",", "\<\"RegisterIndices\"\>", ",", "\<\"RemoveIndices\"\>", 
   ",", "\<\"ReplaceDummies\"\>", ",", "\<\"ReplaceIndex\"\>", 
   ",", "\<\"Ricci\"\>", ",", "\<\"RicciScalar\"\>", 
   ",", "\<\"RicciToEinstein\"\>", ",", "\<\"RicciToTFRicci\"\>", 
   ",", "\<\"Riemann\"\>", ",", "\<\"RiemannDown\"\>", 
   ",", "\<\"RiemannDownToRiemann\"\>", ",", "\<\"RiemannToChristoffel\"\>", 
   ",", "\<\"RiemannToRiemannDown\"\>", ",", "\<\"RiemannToWeyl\"\>", 
   ",", "\<\"SameDummies\"\>", ",", "\<\"Scalar\"\>", 
   ",", "\<\"ScalarFunction\"\>", ",", "\<\"ScalarFunctionQ\"\>", 
   ",", "\<\"ScalarQ\"\>", ",", "\<\"ScalarsOfProduct\"\>", 
   ",", "\<\"ScreenDollarIndices\"\>", ",", "\<\"Sdelta\"\>", 
   ",", "\<\"SeparateDir\"\>", ",", "\<\"SeparateMetric\"\>", 
   ",", "\<\"ServantsOf\"\>", ",", "\<\"SetCharacters\"\>", 
   ",", "\<\"SetConformalTo\"\>", ",", "\<\"SetIndexSortPriorities\"\>", 
   ",", "\<\"SetOrthogonal\"\>", ",", "\<\"SetProjected\"\>", 
   ",", "\<\"SetSplitTensors\"\>", ",", "\<\"SignatureOfMetric\"\>", 
   ",", "\<\"SignDetOfMetric\"\>", ",", "\<\"Simplification\"\>", 
   ",", "\<\"SlotsOfTensor\"\>", ",", "\<\"SortCovDs\"\>", 
   ",", "\<\"SortCovDsStart\"\>", ",", "\<\"SortCovDsStop\"\>", 
   ",", "\<\"SplitIndex\"\>", ",", "\<\"SplitManifold\"\>", 
   ",", "\<\"SplitTensor\"\>", ",", "\<\"SplittingsOfManifold\"\>", 
   ",", "\<\"SplittingsOfVBundle\"\>", ",", "\<\"SplitVBundle\"\>", 
   ",", "\<\"STFPart\"\>", ",", "\<\"Stripped\"\>", 
   ",", "\<\"SubdummiesIn\"\>", ",", "\<\"SubmanifoldQ\"\>", 
   ",", "\<\"SubmanifoldsOfManifold\"\>", ",", "\<\"SubmersionQ\"\>", 
   ",", "\<\"SubvbundleQ\"\>", ",", "\<\"SubvbundlesOfVBundle\"\>", 
   ",", "\<\"Supercommutator\"\>", ",", "\<\"Sym\"\>", 
   ",", "\<\"SymbolOfCovD\"\>", ",", "\<\"Symmetrize\"\>", 
   ",", "\<\"Symmetry\"\>", ",", "\<\"SymmetryGroupOfCovD\"\>", 
   ",", "\<\"SymmetryGroupOfTensor\"\>", ",", "\<\"SymmetryOf\"\>", 
   ",", "\<\"SymmetryTableauxOfTensor\"\>", ",", "\<\"Tangent\"\>", 
   ",", "\<\"TangentBundleOfManifold\"\>", ",", "\<\"TangentTensor\"\>", 
   ",", "\<\"Tensor\"\>", ",", "\<\"TensorD\"\>", 
   ",", "\<\"TensorDerivative\"\>", ",", "\<\"TensorID\"\>", 
   ",", "\<\"TestIndices\"\>", ",", "\<\"Tetra\"\>", ",", "\<\"TetraRule\"\>",
    ",", "\<\"TFRicci\"\>", ",", "\<\"TFRicciToRicci\"\>", 
   ",", "\<\"ToCanonical\"\>", ",", "\<\"ToInducedDerivative\"\>", 
   ",", "\<\"Torsion\"\>", ",", "\<\"TorsionQ\"\>", 
   ",", "\<\"TorsionToChristoffel\"\>", ",", "\<\"ToRule\"\>", 
   ",", "\<\"ToStripped\"\>", ",", "\<\"ToTensorDerivative\"\>", 
   ",", "\<\"TraceDummy\"\>", ",", "\<\"TraceProductDummy\"\>", 
   ",", "\<\"TransposeDagger\"\>", ",", "\<\"Undef\"\>", 
   ",", "\<\"UndefAbstractIndex\"\>", ",", "\<\"UndefConstantSymbol\"\>", 
   ",", "\<\"UndefCovD\"\>", ",", "\<\"UndefInertHead\"\>", 
   ",", "\<\"UndefManifold\"\>", ",", "\<\"UndefMapping\"\>", 
   ",", "\<\"UndefMetric\"\>", ",", "\<\"UndefParameter\"\>", 
   ",", "\<\"UndefProduct\"\>", ",", "\<\"UndefScalarFunction\"\>", 
   ",", "\<\"UndefTensor\"\>", ",", "\<\"UndefVBundle\"\>", ",", "\<\"Up\"\>",
    ",", "\<\"UpIndex\"\>", ",", "\<\"UpIndexQ\"\>", ",", "\<\"UpVectorQ\"\>",
    ",", "\<\"UseMetricOnVBundle\"\>", ",", "\<\"UseSymmetries\"\>", 
   ",", "\<\"UxSort\"\>", ",", "\<\"Validate\"\>", 
   ",", "\<\"ValidateSymbolInSession\"\>", ",", "\<\"VanishingQ\"\>", 
   ",", "\<\"VarD\"\>", ",", "\<\"VBundle\"\>", ",", "\<\"VBundleOfIndex\"\>",
    ",", "\<\"VBundleOfMetric\"\>", ",", "\<\"VBundleQ\"\>", 
   ",", "\<\"VBundlesOfCovD\"\>", ",", "\<\"VectorOfInducedMetric\"\>", 
   ",", "\<\"VisitorsOf\"\>", ",", "\<\"WeightedWithBasis\"\>", 
   ",", "\<\"WeightOf\"\>", ",", "\<\"WeightOfTensor\"\>", 
   ",", "\<\"Weyl\"\>", ",", "\<\"WeylToRiemann\"\>", ",", "\<\"xSort\"\>", 
   ",", "\<\"xSortOrder\"\>", ",", "\<\"xSortPrecedence\"\>", 
   ",", "\<\"xTensorFormStart\"\>", ",", "\<\"xTensorFormStop\"\>", 
   ",", "\<\"xTensorQ\"\>", ",", "\<\"Zero\"\>", 
   ",", "\<\"$AbstractIndices\"\>", ",", "\<\"$AccelerationSign\"\>", 
   ",", "\<\"$Bases\"\>", ",", "\<\"$Charts\"\>", 
   ",", "\<\"$CheckZeroDerivativeVerbose\"\>", ",", "\<\"$CIndexForm\"\>", 
   ",", "\<\"$CommuteCovDsOnScalars\"\>", ",", "\<\"$ComputeNewDummies\"\>", 
   ",", "\<\"$ConstantSymbols\"\>", ",", "\<\"$CovDFormat\"\>", 
   ",", "\<\"$CovDs\"\>", ",", "\<\"$DefInfoQ\"\>", 
   ",", "\<\"$epsilonSign\"\>", ",", "\<\"$ExtrinsicKSign\"\>", 
   ",", "\<\"$FindIndicesAcceptedHeads\"\>", ",", "\<\"$ImplodeInfoQ\"\>", 
   ",", "\<\"$IndexListColor\"\>", ",", "\<\"$InertHeads\"\>", 
   ",", "\<\"$Manifolds\"\>", ",", "\<\"$Mappings\"\>", 
   ",", "\<\"$Metrics\"\>", ",", "\<\"$MixedDers\"\>", 
   ",", "\<\"$ParamDFormat\"\>", ",", "\<\"$Parameters\"\>", 
   ",", "\<\"$ProductManifolds\"\>", ",", "\<\"$ProductMetrics\"\>", 
   ",", "\<\"$Products\"\>", ",", "\<\"$ProjectedFormat\"\>", 
   ",", "\<\"$ProtectNewSymbols\"\>", ",", "\<\"$ReadingVerbose\"\>", 
   ",", "\<\"$RicciSign\"\>", ",", "\<\"$RiemannSign\"\>", 
   ",", "\<\"$Rules\"\>", ",", "\<\"$ScalarColor\"\>", 
   ",", "\<\"$ScalarFunctions\"\>", ",", "\<\"$SeparateScalarsFromScalar\"\>",
    ",", "\<\"$SumVBundles\"\>", ",", "\<\"$TensorBoxes\"\>", 
   ",", "\<\"$Tensors\"\>", ",", "\<\"$TorsionSign\"\>", 
   ",", "\<\"$TraceDummyVerbose\"\>", ",", "\<\"$UndefInfoQ\"\>", 
   ",", "\<\"$VBundles\"\>", ",", "\<\"$Version\"\>", 
   ",", "\<\"$xPermVersionExpected\"\>"}], "}"}]], "Output",
 CellChangeTimes->{3.728978995157233*^9},
 CellLabel->"Out[1314]=",
 CellID->584626563,ExpressionUUID->"aa20a420-50da-47de-9374-999d5404fb65"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Length", "[", "%", "]"}]], "Input",
 CellLabel->
  "In[1315]:=",ExpressionUUID->"a7b1e781-126d-4adf-afea-fb8a5a94d7ca"],

Cell[BoxData["454"], "Output",
 CellChangeTimes->{3.728978995351515*^9},
 CellLabel->"Out[1315]=",
 CellID->1080959316,ExpressionUUID->"d3452b50-cf7a-473a-acf5-1cac357ae8c2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"TimeUsed", "[", "]"}]], "Input",
 CellLabel->
  "In[1316]:=",ExpressionUUID->"8b648d0b-23b6-48a0-af25-eae60d794b98"],

Cell[BoxData["58.868932`"], "Output",
 CellChangeTimes->{3.728978995488511*^9},
 CellLabel->"Out[1316]=",
 CellID->2119054039,ExpressionUUID->"9f74160c-9d79-4eea-a136-1ec3326406ab"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MemoryInUse", "[", "]"}], "/", "1000"}], "/", "1024"}], "//", 
  "N"}]], "Input",
 CellLabel->
  "In[1317]:=",ExpressionUUID->"5ad2e8f7-061b-4adc-915c-664072b9760b"],

Cell[BoxData["92.8694921875`"], "Output",
 CellChangeTimes->{3.7289789956010036`*^9},
 CellLabel->"Out[1317]=",
 CellID->1072431625,ExpressionUUID->"8f267b3e-f333-440c-b35a-b2bfa10b58e3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"MaxMemoryUsed", "[", "]"}], "/", "1000"}], "/", "1024"}], "//", 
  "N"}]], "Input",
 CellLabel->
  "In[1318]:=",ExpressionUUID->"02498cf6-b487-4e39-bd44-d17253c1e3f0"],

Cell[BoxData["95.077546875`"], "Output",
 CellChangeTimes->{3.7289789957219048`*^9},
 CellLabel->"Out[1318]=",
 CellID->528072104,ExpressionUUID->"1a62664f-4b21-41e8-a21d-400601625b3a"]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]]
},
ScreenStyleEnvironment->"Brackets",
WindowSize->{854, 839},
WindowMargins->{{4, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
PageHeaders->{{Inherited, Inherited, 
   Cell[
    TextData[{
      ValueBox["FileName"]}], "Header"]}, {
   Cell[
    TextData[{
      ValueBox["FileName"]}], "Header"], Inherited, Inherited}},
PageFooters->{{Inherited, 
   Cell[
    TextData[{
     "\[Copyright]2003-2004 Jos\[EAcute] M. Mart\[IAcute]n-Garc\[IAcute]a"}], 
    "PageFooter"], Inherited}, {Inherited, 
   Cell[
    TextData[{
     "\[Copyright]2003-2004 Jos\[EAcute] M. Mart\[IAcute]n-Garc\[IAcute]a"}], 
    "PageFooter"], Inherited}},
PrintingOptions->{"Magnification"->1,
"PaperOrientation"->"Portrait",
"PaperSize"->{597.562, 842.375},
"PostScriptOutputFile":>FrontEnd`FileName[{$RootDirectory, "home", "jmm"}, 
  "example.ps", CharacterEncoding -> "iso8859-1"],
"PrintCellBrackets"->False,
"PrintMultipleHorizontalPages"->False,
"PrintRegistrationMarks"->True,
"PrintingMargins"->{{54, 54}, {108, 108}}},
ShowSelection->True,
CellLabelAutoDelete->False,
FrontEndVersion->"12.1 for Mac OS X x86 (64-bit) (March 13, 2020)",
StyleDefinitions->FrontEnd`FileName[{"Wolfram"}, "TutorialPageStyles.nb", 
  CharacterEncoding -> "UTF-8"],
ExpressionUUID->"1b49e380-8909-4558-9230-0dffaed82783"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "S5.18.1"->{
  Cell[580, 22, 218, 4, 53, "Section",ExpressionUUID->"8f751253-cb93-40fd-ba41-fbaa77bd1a49",
   CellTags->{"S5.18.1", "1.1"},
   CellID->733679349],
  Cell[9100, 245, 522, 10, 54, "Section",ExpressionUUID->"ca3135f0-5abf-4775-a4fc-678f1dffad3e",
   CellTags->{"S5.18.1", "1.4"},
   CellID->853073756],
  Cell[345374, 9594, 602, 10, 54, "Section",ExpressionUUID->"f7d60e03-9b2e-438d-919f-8d580da78398",
   CellTags->{"S5.18.1", "1.4"},
   CellID->390546124],
  Cell[423713, 11978, 619, 11, 54, "Section",ExpressionUUID->"e50a4bbd-8e74-4fce-82b6-11d15ef76e1d",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1215012216],
  Cell[454259, 12909, 684, 12, 54, "Section",ExpressionUUID->"f4f44c18-1ef4-4eff-98cd-53767a44bc77",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1339531280],
  Cell[954994, 27187, 724, 12, 54, "Section",ExpressionUUID->"e07ebc50-8899-4702-bbbe-f4d58d3ca5cc",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1330943070],
  Cell[1497147, 44019, 776, 13, 54, "Section",ExpressionUUID->"b37ffd4e-e1c9-4992-9806-1ad019ce6449",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1736504676],
  Cell[1736747, 51291, 818, 14, 54, "Section",ExpressionUUID->"4e72983f-5bfb-4fb2-bf7e-973e9afda9de",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1740316966],
  Cell[3195180, 91736, 862, 14, 54, "Section",ExpressionUUID->"ba260751-a6a0-4d09-85a7-cab77ca50fa1",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1069634512],
  Cell[6129458, 176089, 1013, 19, 54, "Section",ExpressionUUID->"4d1d0042-6342-4549-8253-0be8111126d2",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1088456401],
  Cell[6320077, 182006, 918, 15, 54, "Section",ExpressionUUID->"4b0a4311-6f29-4ca4-b89d-024d36e40a11",
   CellTags->{"S5.18.1", "1.4"},
   CellID->717570716],
  Cell[6796388, 195701, 983, 16, 54, "Section",ExpressionUUID->"4ad3e1c7-5134-49fb-8ac5-06e6eda5f7a3",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1149066679],
  Cell[7631306, 219433, 1016, 16, 54, "Section",ExpressionUUID->"60102f35-97c9-49fb-ac37-e2c30aabc8e1",
   CellTags->{"S5.18.1", "1.4"},
   CellID->292931816]},
 "1.1"->{
  Cell[580, 22, 218, 4, 53, "Section",ExpressionUUID->"8f751253-cb93-40fd-ba41-fbaa77bd1a49",
   CellTags->{"S5.18.1", "1.1"},
   CellID->733679349]},
 "1.4"->{
  Cell[9100, 245, 522, 10, 54, "Section",ExpressionUUID->"ca3135f0-5abf-4775-a4fc-678f1dffad3e",
   CellTags->{"S5.18.1", "1.4"},
   CellID->853073756],
  Cell[345374, 9594, 602, 10, 54, "Section",ExpressionUUID->"f7d60e03-9b2e-438d-919f-8d580da78398",
   CellTags->{"S5.18.1", "1.4"},
   CellID->390546124],
  Cell[423713, 11978, 619, 11, 54, "Section",ExpressionUUID->"e50a4bbd-8e74-4fce-82b6-11d15ef76e1d",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1215012216],
  Cell[454259, 12909, 684, 12, 54, "Section",ExpressionUUID->"f4f44c18-1ef4-4eff-98cd-53767a44bc77",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1339531280],
  Cell[954994, 27187, 724, 12, 54, "Section",ExpressionUUID->"e07ebc50-8899-4702-bbbe-f4d58d3ca5cc",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1330943070],
  Cell[1497147, 44019, 776, 13, 54, "Section",ExpressionUUID->"b37ffd4e-e1c9-4992-9806-1ad019ce6449",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1736504676],
  Cell[1736747, 51291, 818, 14, 54, "Section",ExpressionUUID->"4e72983f-5bfb-4fb2-bf7e-973e9afda9de",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1740316966],
  Cell[3195180, 91736, 862, 14, 54, "Section",ExpressionUUID->"ba260751-a6a0-4d09-85a7-cab77ca50fa1",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1069634512],
  Cell[6129458, 176089, 1013, 19, 54, "Section",ExpressionUUID->"4d1d0042-6342-4549-8253-0be8111126d2",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1088456401],
  Cell[6320077, 182006, 918, 15, 54, "Section",ExpressionUUID->"4b0a4311-6f29-4ca4-b89d-024d36e40a11",
   CellTags->{"S5.18.1", "1.4"},
   CellID->717570716],
  Cell[6796388, 195701, 983, 16, 54, "Section",ExpressionUUID->"4ad3e1c7-5134-49fb-8ac5-06e6eda5f7a3",
   CellTags->{"S5.18.1", "1.4"},
   CellID->1149066679],
  Cell[7631306, 219433, 1016, 16, 54, "Section",ExpressionUUID->"60102f35-97c9-49fb-ac37-e2c30aabc8e1",
   CellTags->{"S5.18.1", "1.4"},
   CellID->292931816]},
 "Info443728981228-9181623"->{
  Cell[333634, 9263, 212, 3, 43, "Print",ExpressionUUID->"94eae8c2-9b3a-4e88-b43e-adc509ad506c",
   CellTags->"Info443728981228-9181623",
   CellID->1995674241]},
 "Info453728981253-9181623"->{
  Cell[336961, 9350, 677, 10, 133, "Print",ExpressionUUID->"e4d5af28-bacd-48e1-a58c-0f67bfc59604",
   CellTags->"Info453728981253-9181623",
   CellID->1862154774]},
 "Info463728981253-9181623"->{
  Cell[337828, 9371, 318, 5, 44, "Print",ExpressionUUID->"61ca6cff-8a7c-4bca-b966-a93a55cad7c9",
   CellTags->"Info463728981253-9181623",
   CellID->1956098887],
  Cell[338149, 9378, 1591, 48, 159, "Print",ExpressionUUID->"560a3844-0f62-4961-bffb-79612314e255",
   CellTags->"Info463728981253-9181623",
   CellID->326841886]},
 "Info513728981275-9181623"->{
  Cell[350987, 9754, 196, 3, 43, "Print",ExpressionUUID->"0c64b26f-1b6a-426c-8169-13f835a49e44",
   CellTags->"Info513728981275-9181623",
   CellID->580545217],
  Cell[351186, 9759, 1746, 53, 193, "Print",ExpressionUUID->"794dc6e3-9199-42ce-b23c-afc9b0990f95",
   CellTags->"Info513728981275-9181623",
   CellID->1205253798]},
 "Info543728981286-9181623"->{
  Cell[355554, 9892, 348, 9, 44, "Print",ExpressionUUID->"a14a8109-2565-4fe7-84ec-88b5522828b6",
   CellTags->"Info543728981286-9181623",
   CellID->76239291],
  Cell[355905, 9903, 5061, 155, 363, "Print",ExpressionUUID->"2ec7dae1-8a02-4a21-a5a7-9fd5c71f3517",
   CellTags->"Info543728981286-9181623",
   CellID->1453510225]},
 "Info643728981304-9181623"->{
  Cell[374702, 10468, 197, 3, 43, "Print",ExpressionUUID->"495fb146-4919-4a2e-a6a9-01c1ea7ad2db",
   CellTags->"Info643728981304-9181623",
   CellID->1678239341],
  Cell[374902, 10473, 2093, 64, 261, "Print",ExpressionUUID->"696209d9-520d-4605-8a2b-cc1e264a6a62",
   CellTags->"Info643728981304-9181623",
   CellID->1118384489]},
 "Info653728981305-9181623"->{
  Cell[377393, 10556, 352, 9, 44, "Print",ExpressionUUID->"c67e0a3a-aab2-4e76-b372-ae1b984fb9b1",
   CellTags->"Info653728981305-9181623",
   CellID->1548585086],
  Cell[377748, 10567, 7095, 214, 431, "Print",ExpressionUUID->"4f8c5a66-9828-4a8c-a4c2-2c05d8fba923",
   CellTags->"Info653728981305-9181623",
   CellID->1715222409]},
 "Info673728981308-9181623"->{
  Cell[385859, 10821, 482, 7, 89, "Print",ExpressionUUID->"f89e4fe2-2986-45ad-b58e-63513099d114",
   CellTags->"Info673728981308-9181623",
   CellID->1069478814]},
 "Info763728981334-9181623"->{
  Cell[396033, 11122, 330, 9, 44, "Print",ExpressionUUID->"78f21f23-a24d-4d19-9550-1ea66aeba6ec",
   CellTags->"Info763728981334-9181623",
   CellID->1134239889],
  Cell[396366, 11133, 5501, 174, 397, "Print",ExpressionUUID->"86d299a7-dd5d-4f3d-ba67-a9c99a8f3c23",
   CellTags->"Info763728981334-9181623",
   CellID->2104963651]},
 "Info773728981334-9181623"->{
  Cell[402801, 11341, 345, 9, 44, "Print",ExpressionUUID->"e45eb883-9712-4a14-bb63-71266d89737a",
   CellTags->"Info773728981334-9181623",
   CellID->478481679],
  Cell[403149, 11352, 5879, 178, 420, "Print",ExpressionUUID->"1b43b27e-5ef8-4e34-ad3f-b46f987e1e90",
   CellTags->"Info773728981334-9181623",
   CellID->1662804585]},
 "Info793728981334-9181623"->{
  Cell[410680, 11595, 456, 6, 89, "Print",ExpressionUUID->"b624df61-586c-41cd-84c6-be2034fb0f05",
   CellTags->"Info793728981334-9181623",
   CellID->483646526]},
 "Info813728981342-9181623"->{
  Cell[414177, 11695, 324, 9, 44, "Print",ExpressionUUID->"c8622f06-4d6e-4201-bb88-bb26ce55c8ab",
   CellTags->"Info813728981342-9181623",
   CellID->2049167434],
  Cell[414504, 11706, 1704, 51, 91, "Print",ExpressionUUID->"3f1bc420-c985-4a39-be0e-84a2da604eb8",
   CellTags->"Info813728981342-9181623",
   CellID->1161816299]},
 "Info1163728981383-9181623"->{
  Cell[452141, 12840, 212, 3, 43, "Print",ExpressionUUID->"58195580-63ce-46a6-9406-193b5711864f",
   CellTags->"Info1163728981383-9181623",
   CellID->564633603],
  Cell[452356, 12845, 1818, 55, 193, "Print",ExpressionUUID->"0c3eff53-b245-4530-b9f2-53af8ab16efe",
   CellTags->"Info1163728981383-9181623",
   CellID->66729528]},
 "Info1223728981397-9181623"->{
  Cell[463128, 13171, 198, 3, 43, "Print",ExpressionUUID->"51354a4a-940d-40e8-8e3e-9b1c29871105",
   CellTags->"Info1223728981397-9181623",
   CellID->284491250],
  Cell[463329, 13176, 3840, 117, 296, "Print",ExpressionUUID->"14242c38-c27d-43a7-955c-6a18caddad3f",
   CellTags->"Info1223728981397-9181623",
   CellID->1145840528]},
 "Info1263728981397-9181623"->{
  Cell[470720, 13414, 198, 3, 43, "Print",ExpressionUUID->"125cac5c-4025-4a7f-a7b8-a2538f16f51f",
   CellTags->"Info1263728981397-9181623",
   CellID->709603301],
  Cell[470921, 13419, 3150, 97, 296, "Print",ExpressionUUID->"493ff12b-20ae-40f5-99b4-5f0b62a6615a",
   CellTags->"Info1263728981397-9181623",
   CellID->1325988116]},
 "Info1733728981432-9181623"->{
  Cell[543536, 15663, 198, 3, 43, "Print",ExpressionUUID->"ab3ee4a8-7bf1-41bf-a3b2-652eaf0137c9",
   CellTags->"Info1733728981432-9181623",
   CellID->944631114],
  Cell[543737, 15668, 3211, 99, 296, "Print",ExpressionUUID->"579ee2dc-80f0-49f8-9e65-24f5efae300c",
   CellTags->"Info1733728981432-9181623",
   CellID->68530727]},
 "Info1753728981432-9181623"->{
  Cell[548697, 15829, 196, 3, 43, "Print",ExpressionUUID->"a67b04c3-c3c7-4ac1-8a2a-674cc955be23",
   CellTags->"Info1753728981432-9181623",
   CellID->535632163],
  Cell[548896, 15834, 3900, 118, 296, "Print",ExpressionUUID->"78dd0d02-24ec-4d49-9b35-f8d81c30bc4b",
   CellTags->"Info1753728981432-9181623",
   CellID->118315317]},
 "Info2273728981469-9181623"->{
  Cell[873560, 24755, 199, 3, 43, "Print",ExpressionUUID->"d09061f2-d245-4708-9251-86b611186377",
   CellTags->"Info2273728981469-9181623",
   CellID->1217710107],
  Cell[873762, 24760, 2666, 84, 296, "Print",ExpressionUUID->"17855b62-38ea-437b-aaa7-7df201779033",
   CellTags->"Info2273728981469-9181623",
   CellID->534612439]},
 "Info2393728981479-9181623"->{
  Cell[891722, 25320, 196, 3, 43, "Print",ExpressionUUID->"729ec76a-285d-4e29-a392-3a0b59f569e5",
   CellTags->"Info2393728981479-9181623",
   CellID->704900921],
  Cell[891921, 25325, 3255, 102, 330, "Print",ExpressionUUID->"55bb2ae3-f92f-49b6-a6b4-e800ab4bf2c8",
   CellTags->"Info2393728981479-9181623",
   CellID->1749982395]},
 "Info2403728981479-9181623"->{
  Cell[895386, 25440, 204, 3, 43, "Print",ExpressionUUID->"f4ae4598-eca7-4561-b34c-69b4a8d91214",
   CellTags->"Info2403728981479-9181623",
   CellID->18083453],
  Cell[895593, 25445, 3371, 102, 330, "Print",ExpressionUUID->"50884fe2-237c-4232-b241-7b33a9fb2528",
   CellTags->"Info2403728981479-9181623",
   CellID->1133400344]},
 "Info2663728981492-9181623"->{
  Cell[936139, 26600, 197, 3, 43, "Print",ExpressionUUID->"8b37f71b-9bc8-4c6b-a966-48e42cfdf8b6",
   CellTags->"Info2663728981492-9181623",
   CellID->1735986504],
  Cell[936339, 26605, 2476, 77, 262, "Print",ExpressionUUID->"bf829cfa-6dbb-4861-8773-7a3aefbd4dba",
   CellTags->"Info2663728981492-9181623",
   CellID->1012344586]},
 "Info2823728981498-9181623"->{
  Cell[954553, 27172, 212, 3, 43, "Print",ExpressionUUID->"944a01b4-5f58-4e66-a631-2beb8f017d35",
   CellTags->"Info2823728981498-9181623",
   CellID->366539895]},
 "Info3143728981546-9181623"->{
  Cell[1145955, 32371, 401, 6, 67, "Print",ExpressionUUID->"b286d54a-110c-4101-a947-c1930e20065e",
   CellTags->"Info3143728981546-9181623",
   CellID->799431715]},
 "Info3183728981546-9181623"->{
  Cell[1149266, 32475, 408, 6, 67, "Print",ExpressionUUID->"bfc1f777-98e2-405b-8b4a-375be9c9d56c",
   CellTags->"Info3183728981546-9181623",
   CellID->1257358069]},
 "Info4123728981628-9181623"->{
  Cell[1570353, 46231, 196, 3, 43, "Print",ExpressionUUID->"81374258-63c6-4abc-8ad3-dfd7f1e48801",
   CellTags->"Info4123728981628-9181623",
   CellID->213508853],
  Cell[1570552, 46236, 6879, 192, 339, "Print",ExpressionUUID->"93334796-27eb-4b34-88b2-aede27817ead",
   CellTags->"Info4123728981628-9181623",
   CellID->2145601881]},
 "Info4163728981629-9181623"->{
  Cell[1582651, 46595, 197, 3, 43, "Print",ExpressionUUID->"952d080a-d3f2-4c6b-b336-ee40ea619ec5",
   CellTags->"Info4163728981629-9181623",
   CellID->1666127203],
  Cell[1582851, 46600, 6430, 181, 340, "Print",ExpressionUUID->"dbb65ac4-a485-4871-a20e-6176bd7da37d",
   CellTags->"Info4163728981629-9181623",
   CellID->1560627588]},
 "Info4193728981630-9181623"->{
  Cell[1591749, 46872, 195, 3, 43, "Print",ExpressionUUID->"bf548868-f80a-4fce-add5-25287ccbf669",
   CellTags->"Info4193728981630-9181623",
   CellID->87455265],
  Cell[1591947, 46877, 11528, 309, 383, "Print",ExpressionUUID->"061e13da-7d82-4a5b-9daf-db3bdf7c3cf0",
   CellTags->"Info4193728981630-9181623",
   CellID->981824307]},
 "Info5793728981796-9181623"->{
  Cell[3015433, 86319, 201, 3, 43, "Print",ExpressionUUID->"c227df3b-fdfd-490b-a692-412f36cca699",
   CellTags->"Info5793728981796-9181623",
   CellID->1397750255],
  Cell[3015637, 86324, 2136, 66, 262, "Print",ExpressionUUID->"86dcd4d2-d60f-4dc8-8d2c-deaa78c12e5d",
   CellTags->"Info5793728981796-9181623",
   CellID->664921252]},
 "Info5973728981810-9181623"->{
  Cell[3030472, 86819, 199, 3, 43, "Print",ExpressionUUID->"47a658f0-5129-4f86-9b44-b5ae579d9de6",
   CellTags->"Info5973728981810-9181623",
   CellID->1151780222],
  Cell[3030674, 86824, 3117, 98, 296, "Print",ExpressionUUID->"73efd163-d740-46b6-873d-021591318f7f",
   CellTags->"Info5973728981810-9181623",
   CellID->552597096]},
 "Info6273728981830-9181623"->{
  Cell[3117991, 89462, 211, 3, 43, "Print",ExpressionUUID->"519da795-919b-40cf-b068-f23f25068295",
   CellTags->"Info6273728981830-9181623",
   CellID->1431877094],
  Cell[3118205, 89467, 3967, 119, 363, "Print",ExpressionUUID->"7cc7d0aa-97b2-46e4-8186-fe70e9c01262",
   CellTags->"Info6273728981830-9181623",
   CellID->856570576]},
 "Info6283728981830-9181623"->{
  Cell[3122383, 89599, 207, 3, 43, "Print",ExpressionUUID->"25bd47ae-1722-4fac-9b5f-ad3965650133",
   CellTags->"Info6283728981830-9181623",
   CellID->1798784991],
  Cell[3122593, 89604, 4463, 133, 363, "Print",ExpressionUUID->"8d196b37-dc33-4d02-aecc-a9fcbf76e838",
   CellTags->"Info6283728981830-9181623",
   CellID->2052044918]},
 "Info6293728981831-9181623"->{
  Cell[3127280, 89750, 219, 3, 43, "Print",ExpressionUUID->"5fcf77bd-09a9-4707-b110-ac2ccc6bfe59",
   CellTags->"Info6293728981831-9181623",
   CellID->131566006],
  Cell[3127502, 89755, 4001, 117, 329, "Print",ExpressionUUID->"e36f34d4-fbb6-4f39-a840-90e6ad5ebd86",
   CellTags->"Info6293728981831-9181623",
   CellID->1469190079]},
 "Info6303728981831-9181623"->{
  Cell[3131723, 89885, 215, 3, 43, "Print",ExpressionUUID->"e2ece981-4b04-4bc6-98a0-fb537d10ef48",
   CellTags->"Info6303728981831-9181623",
   CellID->910684547],
  Cell[3131941, 89890, 4511, 132, 329, "Print",ExpressionUUID->"3ddd1743-5669-46ff-aed9-4d27b5ffd144",
   CellTags->"Info6303728981831-9181623",
   CellID->759229260]},
 "Info10913728982189-9181623"->{
  Cell[6458772, 186007, 198, 3, 43, "Print",ExpressionUUID->"43e2b182-5295-45a0-97ed-b7fecafa9b13",
   CellTags->"Info10913728982189-9181623",
   CellID->1587009223],
  Cell[6458973, 186012, 87167, 2211, 776, "Print",ExpressionUUID->"8248b8f0-89c7-4a25-a0f3-d2e87fb130d5",
   CellTags->"Info10913728982189-9181623",
   CellID->1117115221]},
 "Info12393728982516-9181623"->{
  Cell[7037161, 202854, 202, 3, 43, "Print",ExpressionUUID->"f7c28f10-04f1-4f64-97a8-4dd64ff8b439",
   CellTags->"Info12393728982516-9181623",
   CellID->1239607777],
  Cell[7037366, 202859, 1620, 47, 193, "Print",ExpressionUUID->"e7556ba5-d1ca-4503-aebf-03e321a473f9",
   CellTags->"Info12393728982516-9181623",
   CellID->1349048492]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"S5.18.1", 7655806, 219960},
 {"1.1", 7657867, 220000},
 {"1.4", 7658028, 220004},
 {"Info443728981228-9181623", 7659960, 220041},
 {"Info453728981253-9181623", 7660154, 220045},
 {"Info463728981253-9181623", 7660350, 220049},
 {"Info513728981275-9181623", 7660708, 220056},
 {"Info543728981286-9181623", 7661066, 220063},
 {"Info643728981304-9181623", 7661424, 220070},
 {"Info653728981305-9181623", 7661785, 220077},
 {"Info673728981308-9181623", 7662147, 220084},
 {"Info763728981334-9181623", 7662342, 220088},
 {"Info773728981334-9181623", 7662704, 220095},
 {"Info793728981334-9181623", 7663065, 220102},
 {"Info813728981342-9181623", 7663259, 220106},
 {"Info1163728981383-9181623", 7663620, 220113},
 {"Info1223728981397-9181623", 7663981, 220120},
 {"Info1263728981397-9181623", 7664345, 220127},
 {"Info1733728981432-9181623", 7664708, 220134},
 {"Info1753728981432-9181623", 7665069, 220141},
 {"Info2273728981469-9181623", 7665432, 220148},
 {"Info2393728981479-9181623", 7665795, 220155},
 {"Info2403728981479-9181623", 7666159, 220162},
 {"Info2663728981492-9181623", 7666522, 220169},
 {"Info2823728981498-9181623", 7666886, 220176},
 {"Info3143728981546-9181623", 7667082, 220180},
 {"Info3183728981546-9181623", 7667279, 220184},
 {"Info4123728981628-9181623", 7667477, 220188},
 {"Info4163728981629-9181623", 7667843, 220195},
 {"Info4193728981630-9181623", 7668210, 220202},
 {"Info5793728981796-9181623", 7668575, 220209},
 {"Info5973728981810-9181623", 7668940, 220216},
 {"Info6273728981830-9181623", 7669305, 220223},
 {"Info6283728981830-9181623", 7669671, 220230},
 {"Info6293728981831-9181623", 7670038, 220237},
 {"Info6303728981831-9181623", 7670404, 220244},
 {"Info10913728982189-9181623", 7670770, 220251},
 {"Info12393728982516-9181623", 7671144, 220258}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 218, 4, 53, "Section",ExpressionUUID->"8f751253-cb93-40fd-ba41-fbaa77bd1a49",
 CellTags->{"S5.18.1", "1.1"},
 CellID->733679349],
Cell[CellGroupData[{
Cell[823, 30, 83, 0, 39, "Subsection",ExpressionUUID->"9bdcbc5a-2e86-4a77-9fb5-564dc8632094"],
Cell[909, 32, 810, 24, 128, "Text",ExpressionUUID->"e5ed7587-e24d-4f73-b8f3-63f0c64152fc"],
Cell[1722, 58, 336, 9, 81, "Text",ExpressionUUID->"55526d07-7028-4064-b2ac-5696ca58a58a"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2095, 72, 82, 0, 39, "Subsection",ExpressionUUID->"58295eb0-9c58-4bc2-922e-93dc4f966c53"],
Cell[2180, 74, 2033, 46, 421, "Text",ExpressionUUID->"dacae0c5-11a2-4fcd-a41c-5ae57fcad21a"]
}, Open  ]],
Cell[CellGroupData[{
Cell[4250, 125, 93, 0, 39, "Subsection",ExpressionUUID->"8914d7d8-31e0-4a20-a60f-843051cb7ed6"],
Cell[CellGroupData[{
Cell[4368, 129, 358, 7, 56, "MathCaption",ExpressionUUID->"63a736e0-39af-4e8d-9b57-37c116af6eb8"],
Cell[CellGroupData[{
Cell[4751, 140, 110, 1, 27, "Input",ExpressionUUID->"1fb58714-653b-4365-aa43-7b82564c3a6d"],
Cell[CellGroupData[{
Cell[4886, 145, 311, 5, 22, "Print",ExpressionUUID->"37e67236-d784-488e-b78b-539616ac34ba",
 CellID->551875384],
Cell[5200, 152, 581, 13, 22, "Print",ExpressionUUID->"96aae411-9519-4432-ad32-11c55c5de4dc",
 CellID->387707495],
Cell[5784, 167, 330, 5, 22, "Print",ExpressionUUID->"a7e6bf32-dba2-4298-bbcd-4fd1c90f3784",
 CellID->734483467],
Cell[6117, 174, 289, 4, 22, "Print",ExpressionUUID->"b5e01409-19db-4b30-985d-669d58f2ae42",
 CellID->727626913],
Cell[6409, 180, 271, 4, 22, "Print",ExpressionUUID->"2cd76ea9-cff8-44c3-87a5-3eb8839e6499",
 CellID->1453640773],
Cell[6683, 186, 312, 5, 22, "Print",ExpressionUUID->"71b7ba64-d48e-4e69-90ab-9547595f46c7",
 CellID->1566503539],
Cell[6998, 193, 584, 13, 22, "Print",ExpressionUUID->"6cf8339f-c224-46ca-9cbc-91d7126fc1ca",
 CellID->75467753],
Cell[7585, 208, 333, 5, 22, "Print",ExpressionUUID->"1fddaec9-f1f6-420e-912d-b2569353640f",
 CellID->1371172941],
Cell[7921, 215, 309, 5, 22, "Print",ExpressionUUID->"5f1102af-0d9a-4abf-ab95-163e9b78483a",
 CellID->656603430],
Cell[8233, 222, 467, 7, 61, "Print",ExpressionUUID->"cb79e0ae-c180-4725-bb3a-218a63bce74d",
 CellID->1762869724],
Cell[8703, 231, 312, 5, 22, "Print",ExpressionUUID->"32c53a62-a534-4654-8d67-c2495637767f",
 CellID->1867057256]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[9100, 245, 522, 10, 54, "Section",ExpressionUUID->"ca3135f0-5abf-4775-a4fc-678f1dffad3e",
 CellTags->{"S5.18.1", "1.4"},
 CellID->853073756],
Cell[CellGroupData[{
Cell[9647, 259, 97, 0, 39, "Subsection",ExpressionUUID->"49ccbd85-b1ee-4ab9-957e-62b0ff70d533"],
Cell[9747, 261, 264, 6, 23, "Text",ExpressionUUID->"23d37c49-4c87-4fd8-8d70-984ab004bf6e"],
Cell[CellGroupData[{
Cell[10036, 271, 162, 5, 40, "MathCaption",ExpressionUUID->"733d4119-e5b1-4b6b-8ed7-5c97b457d9d7"],
Cell[CellGroupData[{
Cell[10223, 280, 258, 6, 27, "Input",ExpressionUUID->"67a3701a-8ac8-4629-aa60-772ddf0caed8"],
Cell[CellGroupData[{
Cell[10506, 290, 591, 12, 22, "Print",ExpressionUUID->"a9b233b4-d6a5-4736-b9bd-6d4662bf9526",
 CellID->280435421],
Cell[11100, 304, 602, 12, 22, "Print",ExpressionUUID->"c96430fd-4926-488a-b3f2-2d098ec42435",
 CellID->1698999229]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[11763, 323, 254, 8, 40, "MathCaption",ExpressionUUID->"008634dc-0c9e-4eb9-9a34-e678dec0fab0"],
Cell[CellGroupData[{
Cell[12042, 335, 524, 15, 47, "Input",ExpressionUUID->"eaa10a04-da2d-4175-bb11-da65b21403e2"],
Cell[CellGroupData[{
Cell[12591, 354, 624, 12, 22, "Print",ExpressionUUID->"9cc31f94-3ae3-4381-ae36-b22473dbeef1",
 CellID->1672621724],
Cell[13218, 368, 711, 15, 22, "Print",ExpressionUUID->"77953aca-f5eb-4a0d-bb4b-0ad8ddd8843f",
 CellID->1192376038]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[13990, 390, 155, 2, 40, "MathCaption",ExpressionUUID->"979b6ed1-be88-4046-84bd-753c6f388b2c"],
Cell[CellGroupData[{
Cell[14170, 396, 242, 7, 27, "Input",ExpressionUUID->"6b294b34-4699-4a9f-b31f-f5523ab89b3b"],
Cell[14415, 405, 2048, 60, 32, "Output",ExpressionUUID->"4bc2242d-cd73-45a7-a86b-249b53db34f0",
 CellID->41016295]
}, Open  ]],
Cell[CellGroupData[{
Cell[16500, 470, 155, 3, 27, "Input",ExpressionUUID->"74ab546e-7541-401f-b7a8-89446501a6b4"],
Cell[16658, 475, 2976, 88, 32, "Output",ExpressionUUID->"46e5f676-c74b-4032-8930-c05720f5e47b",
 CellID->378027064]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[19683, 569, 156, 2, 40, "MathCaption",ExpressionUUID->"ea0f31b7-36a9-4045-b483-43613ab687c9"],
Cell[CellGroupData[{
Cell[19864, 575, 145, 2, 27, "Input",ExpressionUUID->"18eb2e93-cb8e-4c33-b436-517c3a4c17d7"],
Cell[20012, 579, 169, 3, 26, "Output",ExpressionUUID->"53fe00b1-86b1-4a8d-9b45-f3b363f21829",
 CellID->1898734097]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[20230, 588, 104, 0, 40, "MathCaption",ExpressionUUID->"3fd18b89-5491-4bc0-8950-24f3dccb5d5a"],
Cell[CellGroupData[{
Cell[20359, 592, 674, 23, 27, "Input",ExpressionUUID->"76307278-faac-44e1-907f-0f2948d60a9d"],
Cell[21036, 617, 8880, 258, 32, "Output",ExpressionUUID->"ccc3b734-4654-40e2-b0b6-4574e306b524",
 CellID->1570938722]
}, Open  ]],
Cell[CellGroupData[{
Cell[29953, 880, 141, 2, 27, "Input",ExpressionUUID->"2c856d83-a0bf-4f3c-ac46-75e5062541f5"],
Cell[30097, 884, 6118, 177, 32, "Output",ExpressionUUID->"9e732302-e714-475a-88c4-36f34c51e7d1",
 CellID->136888559]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[36264, 1067, 368, 10, 40, "MathCaption",ExpressionUUID->"86bdeab5-7032-45af-ae1e-7e5083ca90a9"],
Cell[CellGroupData[{
Cell[36657, 1081, 140, 2, 27, "Input",ExpressionUUID->"e1714e59-b838-40c8-95a5-b1120d3c6fd6"],
Cell[36800, 1085, 424, 9, 22, "Message",ExpressionUUID->"a22bf1ed-0245-4f15-9d91-91dd92545138",
 CellID->276709775],
Cell[37227, 1096, 425, 9, 22, "Message",ExpressionUUID->"f982f1fb-0bbf-486f-8511-146b6ba801cc",
 CellID->1359092670],
Cell[37655, 1107, 6834, 198, 32, "Output",ExpressionUUID->"c8dc07e1-6548-42dc-bda7-bc1f06ba7eb9",
 CellID->1634435198]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[44538, 1311, 97, 0, 40, "MathCaption",ExpressionUUID->"cfe5bdde-9bf4-4c42-962e-fa585d6d69fa"],
Cell[CellGroupData[{
Cell[44660, 1315, 142, 2, 27, "Input",ExpressionUUID->"0a91f51a-3d40-43b9-acec-8beb86650730"],
Cell[44805, 1319, 218, 3, 26, "Output",ExpressionUUID->"e272e9ab-b216-43ab-9d1d-924ae292e416",
 CellID->9239946]
}, Open  ]]
}, Open  ]],
Cell[45050, 1326, 496, 10, 57, "Text",ExpressionUUID->"e18353cd-049b-480f-baa8-56723d506293"],
Cell[CellGroupData[{
Cell[45571, 1340, 272, 4, 56, "MathCaption",ExpressionUUID->"af4360a0-05fb-459c-933d-7b0d5575970d"],
Cell[CellGroupData[{
Cell[45868, 1348, 145, 2, 27, "Input",ExpressionUUID->"7746ad4a-dbf0-481b-b66c-676aab67bc23"],
Cell[46016, 1352, 1142, 27, 84, "Output",ExpressionUUID->"8a759938-1bbc-4c24-9eb4-169bff48d812",
 CellID->1951053507]
}, Open  ]],
Cell[CellGroupData[{
Cell[47195, 1384, 187, 4, 27, "Input",ExpressionUUID->"bec91a2d-7597-45cf-ba53-e02d48a72bf7"],
Cell[CellGroupData[{
Cell[47407, 1392, 679, 15, 22, "Print",ExpressionUUID->"80e097f9-1af2-4673-9cad-fe343f30725b",
 CellID->2078150768],
Cell[48089, 1409, 800, 18, 22, "Print",ExpressionUUID->"3c6ed23c-f5e7-4c4d-86b3-5c6ac9216fac",
 CellID->805922968],
Cell[48892, 1429, 816, 19, 22, "Print",ExpressionUUID->"86c37c43-9cd6-4ff3-8c98-034d338fe7d8",
 CellID->967961547],
Cell[49711, 1450, 895, 18, 42, "Print",ExpressionUUID->"825ddd1b-2bbc-4ad2-8d31-52e1d933b591",
 CellID->1341094404],
Cell[50609, 1470, 767, 17, 22, "Print",ExpressionUUID->"18cd6e80-9b20-4e8b-997f-d8de4fab9462",
 CellID->1932772763],
Cell[51379, 1489, 268, 4, 22, "Print",ExpressionUUID->"4b2e3e33-9cf9-4be1-acea-008a936c0deb",
 CellID->1274183219]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[51708, 1500, 109, 0, 40, "MathCaption",ExpressionUUID->"3a2a50f3-9824-47ba-b62e-832dea04f25a"],
Cell[CellGroupData[{
Cell[51842, 1504, 334, 10, 27, "Input",ExpressionUUID->"cebc62b2-b1ef-48b4-8815-3d59f557291c"],
Cell[52179, 1516, 4456, 128, 32, "Output",ExpressionUUID->"1c81698b-9f28-4553-aefb-c433f14b9e3e",
 CellID->94542730]
}, Open  ]],
Cell[CellGroupData[{
Cell[56672, 1649, 194, 4, 27, "Input",ExpressionUUID->"80c61803-d97e-40f7-84bb-9afdfe0934e9"],
Cell[56869, 1655, 3158, 92, 32, "Output",ExpressionUUID->"288343d5-ce25-4c41-bcd1-211a6280a42e",
 CellID->170734857]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[60076, 1753, 147, 2, 40, "MathCaption",ExpressionUUID->"93dd870e-0f51-4948-a667-ca0fc7cbfaf7"],
Cell[CellGroupData[{
Cell[60248, 1759, 458, 14, 27, "Input",ExpressionUUID->"1a9c7bde-e83e-4455-a414-9c8430fec8ec"],
Cell[60709, 1775, 168, 3, 26, "Output",ExpressionUUID->"b02ea8d3-2324-463b-ab74-25874d16cf63",
 CellID->1527135633]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[60926, 1784, 147, 2, 40, "MathCaption",ExpressionUUID->"8dd8161c-5d40-40c2-aeac-7216f5523a2e"],
Cell[CellGroupData[{
Cell[61098, 1790, 512, 15, 27, "Input",ExpressionUUID->"11a3a435-4ae6-42db-bdea-c0e37b4d6181"],
Cell[61613, 1807, 170, 3, 26, "Output",ExpressionUUID->"40d712c8-5e76-454e-abf2-344a3f6e3f03",
 CellID->1550310597]
}, Open  ]],
Cell[CellGroupData[{
Cell[61820, 1815, 672, 20, 47, "Input",ExpressionUUID->"04ae657f-ef8f-481b-90f2-04e18cd0faa3"],
Cell[62495, 1837, 169, 3, 26, "Output",ExpressionUUID->"c65979cb-e594-4aa7-bff5-d9b4ebba6cfa",
 CellID->1920925839]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[62713, 1846, 115, 0, 40, "MathCaption",ExpressionUUID->"f2445a32-362f-49d2-9e09-ba5fd0af47d6"],
Cell[CellGroupData[{
Cell[62853, 1850, 522, 16, 27, "Input",ExpressionUUID->"6dddded9-a24e-4f60-bf55-ca1d47508409"],
Cell[63378, 1868, 7859, 211, 40, "Output",ExpressionUUID->"726abc39-2aa2-4162-a69a-986a72514773",
 CellID->668680751]
}, Open  ]],
Cell[CellGroupData[{
Cell[71274, 2084, 160, 3, 27, "Input",ExpressionUUID->"6863af84-7e1b-4a1b-9378-08de11211671"],
Cell[71437, 2089, 3758, 103, 32, "Output",ExpressionUUID->"bc3317e6-ea2a-45cf-98e4-2816c9180f38",
 CellID->904014618]
}, Open  ]],
Cell[CellGroupData[{
Cell[75232, 2197, 145, 2, 27, "Input",ExpressionUUID->"ae6f8bce-4062-45bc-9de5-24742e3cb5d3"],
Cell[75380, 2201, 30370, 835, 104, "Output",ExpressionUUID->"2cd089c2-762c-4ee3-a704-c928c26b9289",
 CellID->1314426350]
}, Open  ]],
Cell[CellGroupData[{
Cell[105787, 3041, 193, 4, 27, "Input",ExpressionUUID->"2d413902-e71d-4cdd-adbf-cb5808b0d9b1"],
Cell[105983, 3047, 137019, 3641, 368, "Output",ExpressionUUID->"25be150a-b1e1-4eb5-9854-99f870cd86e7",
 CellID->582528635]
}, Open  ]],
Cell[CellGroupData[{
Cell[243039, 6693, 186, 4, 27, "Input",ExpressionUUID->"d823ac90-1c4b-4d64-bf23-46879bb96cad"],
Cell[243228, 6699, 222, 5, 26, "Output",ExpressionUUID->"2ca7c995-2ce7-4fe5-89cf-c8aa071b9346",
 CellID->112544729]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[243499, 6710, 430, 12, 72, "MathCaption",ExpressionUUID->"4d7774b3-33c3-457c-be81-ab513a1cd1a1"],
Cell[CellGroupData[{
Cell[243954, 6726, 278, 8, 27, "Input",ExpressionUUID->"c94edf35-faa8-4fd7-8444-acc6ad3f86f9"],
Cell[244235, 6736, 1269, 36, 32, "Output",ExpressionUUID->"a1983510-b84c-48f2-801d-6560ef830e6d",
 CellID->1474478125]
}, Open  ]],
Cell[CellGroupData[{
Cell[245541, 6777, 319, 9, 27, "Input",ExpressionUUID->"cb171f09-f5a5-440e-9e92-c9952c3f6c7b"],
Cell[245863, 6788, 3376, 96, 32, "Output",ExpressionUUID->"5c8fbeca-a455-4972-9450-e02b2b6de003",
 CellID->603282111]
}, Open  ]],
Cell[CellGroupData[{
Cell[249276, 6889, 187, 4, 27, "Input",ExpressionUUID->"1b96d04a-0ec9-47a2-bdae-5d5758cf4214"],
Cell[249466, 6895, 3377, 96, 32, "Output",ExpressionUUID->"37cab0ed-b2d1-4d83-bc57-a414052bcace",
 CellID->1412296109]
}, Open  ]]
}, Open  ]],
Cell[252870, 6995, 347, 12, 23, "Text",ExpressionUUID->"51e18220-a1c8-4ed2-b295-357653e5fbaf"],
Cell[CellGroupData[{
Cell[253242, 7011, 276, 8, 40, "MathCaption",ExpressionUUID->"3271f7c4-a29c-4892-8b8b-ec36185f6654"],
Cell[CellGroupData[{
Cell[253543, 7023, 403, 11, 27, "Input",ExpressionUUID->"ac761454-d7f9-4d33-9eed-0dad0b039e15"],
Cell[CellGroupData[{
Cell[253971, 7038, 749, 17, 22, "Print",ExpressionUUID->"cd4a8451-9375-48df-a0eb-8a364f4841e6",
 CellID->60842753],
Cell[254723, 7057, 810, 19, 22, "Print",ExpressionUUID->"6bff225e-8ec5-43b9-ab2f-57985493db0f",
 CellID->381933699],
Cell[255536, 7078, 678, 15, 22, "Print",ExpressionUUID->"5fbc85a6-50d0-40e5-968d-e37598c06121",
 CellID->331058210],
Cell[256217, 7095, 799, 18, 22, "Print",ExpressionUUID->"514c17b8-75ef-4170-b500-49b022339039",
 CellID->2112689740],
Cell[257019, 7115, 817, 19, 22, "Print",ExpressionUUID->"876ec043-77b4-46ab-a983-87051b44317f",
 CellID->1481840682],
Cell[257839, 7136, 842, 19, 22, "Print",ExpressionUUID->"c25eacb1-13f8-491e-9f34-482431b3227e",
 CellID->928488115],
Cell[258684, 7157, 761, 17, 22, "Print",ExpressionUUID->"e24d0c23-7a7d-40b9-9ab7-15e0b71d2165",
 CellID->1483569703],
Cell[259448, 7176, 270, 4, 22, "Print",ExpressionUUID->"9b1348f6-aaf0-49f8-8d89-8b310913d49c",
 CellID->2070856651],
Cell[259721, 7182, 644, 13, 22, "Print",ExpressionUUID->"8aff0875-f24a-4b75-ad64-a11d063fb7d8",
 CellID->620266428],
Cell[260368, 7197, 273, 4, 22, "Print",ExpressionUUID->"5f799f0e-b882-4d37-92d1-0033bbfbed60",
 CellID->932232858],
Cell[260644, 7203, 777, 18, 22, "Print",ExpressionUUID->"4883ee58-57ef-4a85-b54b-03e1edf78e1f",
 CellID->2144439541],
Cell[261424, 7223, 852, 19, 22, "Print",ExpressionUUID->"10d712eb-17e0-41d0-b476-672ea12170b4",
 CellID->377317986],
Cell[262279, 7244, 767, 17, 22, "Print",ExpressionUUID->"bf248ee2-2699-41b2-9513-57190ffb4409",
 CellID->64181111],
Cell[263049, 7263, 656, 13, 22, "Print",ExpressionUUID->"e7295a78-6d23-491e-9f21-6f252da0b3e7",
 CellID->655141783],
Cell[263708, 7278, 402, 8, 22, "Print",ExpressionUUID->"cc479cf0-18e6-4b31-a4e2-30bf036da699",
 CellID->819288712],
Cell[264113, 7288, 394, 8, 22, "Print",ExpressionUUID->"7fd59c0a-7e0f-4627-ba0d-efd4564d1800",
 CellID->120238020],
Cell[264510, 7298, 405, 8, 22, "Print",ExpressionUUID->"8d634f01-72dc-4572-866f-3c47ad9863f5",
 CellID->58290313],
Cell[264918, 7308, 659, 13, 22, "Print",ExpressionUUID->"d3012688-4195-4d92-a883-598285d5af02",
 CellID->432747559]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[265638, 7328, 147, 2, 40, "MathCaption",ExpressionUUID->"a0034d48-2255-40ac-abff-8e8cefcfa0cf"],
Cell[CellGroupData[{
Cell[265810, 7334, 283, 8, 27, "Input",ExpressionUUID->"e931d299-b6fe-4d3c-93bd-cf0902fb6299"],
Cell[266096, 7344, 169, 3, 26, "Output",ExpressionUUID->"f541b0d6-0f48-4277-8807-48b4859e9d62",
 CellID->994858890]
}, Open  ]],
Cell[CellGroupData[{
Cell[266302, 7352, 252, 6, 27, "Input",ExpressionUUID->"c17da91a-18b8-44db-a347-2cdcb8ab1800"],
Cell[266557, 7360, 169, 3, 26, "Output",ExpressionUUID->"3f28828a-aa7e-45e8-9197-8f9f556c3ea9",
 CellID->945402066]
}, Open  ]],
Cell[CellGroupData[{
Cell[266763, 7368, 175, 4, 27, "Input",ExpressionUUID->"12cb5df4-aab9-4a7f-a5b8-4e3f3fd5e041"],
Cell[266941, 7374, 170, 3, 26, "Output",ExpressionUUID->"430ca87d-03b6-4250-adbd-dbc43d3bb04c",
 CellID->2032759218]
}, Open  ]],
Cell[CellGroupData[{
Cell[267148, 7382, 267, 8, 27, "Input",ExpressionUUID->"e6d23bb3-801f-4340-925c-9fa6adb26f72"],
Cell[267418, 7392, 1095, 31, 32, "Output",ExpressionUUID->"6af945d8-0592-445c-b8f3-000e55be5d3d",
 CellID->1928400476]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[268562, 7429, 170, 3, 40, "MathCaption",ExpressionUUID->"17a21e64-ee23-46dc-b27f-a0e8c5598a4c"],
Cell[CellGroupData[{
Cell[268757, 7436, 448, 14, 27, "Input",ExpressionUUID->"3f44d07e-b351-4f9f-876b-b747f7d0ebd0"],
Cell[269208, 7452, 2273, 64, 32, "Output",ExpressionUUID->"9adf5748-ee7d-4a27-b08e-44fb57a1214f",
 CellID->773979084]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[271530, 7522, 403, 6, 72, "MathCaption",ExpressionUUID->"a9007a20-2ac3-4bae-ab9f-f32f221dc61a"],
Cell[CellGroupData[{
Cell[271958, 7532, 231, 5, 27, "Input",ExpressionUUID->"4d13ad75-10b3-453f-bf3b-dcfbbc50d6f2"],
Cell[272192, 7539, 442, 9, 22, "Message",ExpressionUUID->"3ebbfba0-68a6-4c4c-ae0e-5e8770c0a313",
 CellID->901495669],
Cell[272637, 7550, 442, 9, 22, "Message",ExpressionUUID->"3109e3f3-3662-49e9-9ca7-d57187803cea",
 CellID->940579298],
Cell[273082, 7561, 444, 9, 22, "Message",ExpressionUUID->"e1c40249-2ec5-496f-9f12-dc516588d066",
 CellID->689135919],
Cell[273529, 7572, 471, 9, 22, "Message",ExpressionUUID->"b59aac10-b288-4ec5-a390-2d93b0189ed0",
 CellID->928082362],
Cell[274003, 7583, 25982, 736, 80, "Output",ExpressionUUID->"17eb4c04-ed79-44dd-a6f5-b598c1fb6b10",
 CellID->1106529973]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[300034, 8325, 179, 3, 40, "MathCaption",ExpressionUUID->"9acd67ce-9b8f-4773-bd43-7d014da9f836"],
Cell[CellGroupData[{
Cell[300238, 8332, 179, 3, 27, "Input",ExpressionUUID->"d29c4f3d-bc25-43f4-9a45-f6305c49c24c"],
Cell[300420, 8337, 4860, 137, 32, "Output",ExpressionUUID->"c13f4b23-11c7-43c5-8278-4ec635c0f151",
 CellID->1308641128]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[305329, 8480, 150, 2, 40, "MathCaption",ExpressionUUID->"71998cc3-cd1d-4173-88eb-8ca18b5a5173"],
Cell[CellGroupData[{
Cell[305504, 8486, 180, 3, 27, "Input",ExpressionUUID->"ea659810-066e-48e3-8459-279068cdef0b"],
Cell[305687, 8491, 5106, 145, 40, "Output",ExpressionUUID->"9851259a-4014-4d4b-b3a6-fb13c32eaf04",
 CellID->1675831686]
}, Open  ]],
Cell[CellGroupData[{
Cell[310830, 8641, 142, 2, 27, "Input",ExpressionUUID->"765c0c44-a175-4bd6-aade-764ecee082b1"],
Cell[310975, 8645, 172, 3, 26, "Output",ExpressionUUID->"6f9dafc6-7ca6-4916-9e22-43a8e646edf0",
 CellID->1541888581]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[311196, 8654, 195, 3, 40, "MathCaption",ExpressionUUID->"4d479e15-4ba2-4b16-8537-f5dd10af7403"],
Cell[CellGroupData[{
Cell[311416, 8661, 138, 2, 27, "Input",ExpressionUUID->"084f69a4-8874-48c6-9086-571b72382de7"],
Cell[311557, 8665, 169, 3, 26, "Output",ExpressionUUID->"0fc2aabc-9475-44b8-b32d-78834a554db7",
 CellID->588011226]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[311775, 8674, 117, 0, 40, "MathCaption",ExpressionUUID->"db3e68e8-3010-4329-979f-789cb337196a"],
Cell[CellGroupData[{
Cell[311917, 8678, 356, 11, 27, "Input",ExpressionUUID->"abdd0cc2-c5c4-4a11-bf96-26120436579b"],
Cell[312276, 8691, 2260, 65, 32, "Output",ExpressionUUID->"7683e2d6-d063-4c43-bd0b-37b4b87caf19",
 CellID->887726534]
}, Open  ]],
Cell[CellGroupData[{
Cell[314573, 8761, 353, 10, 27, "Input",ExpressionUUID->"1bbfeb63-fe02-4dd5-90f1-24f64e4f141c"],
Cell[314929, 8773, 4318, 116, 32, "Output",ExpressionUUID->"2ac1cfa7-707c-4ca0-ab41-9e276c415056",
 CellID->207576009]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[319296, 8895, 147, 2, 40, "MathCaption",ExpressionUUID->"3f77be99-090a-49e6-b0c5-bb118d0b1692"],
Cell[CellGroupData[{
Cell[319468, 8901, 143, 2, 27, "Input",ExpressionUUID->"abeaa576-2e54-4724-9711-781417301268"],
Cell[CellGroupData[{
Cell[319636, 8907, 574, 12, 22, "Print",ExpressionUUID->"3a29ab3d-9582-41e9-b8ea-a295e535af89",
 CellID->1828814815],
Cell[320213, 8921, 616, 12, 22, "Print",ExpressionUUID->"48cff370-5f01-49b5-a6a2-f99863def3c0",
 CellID->2104639841],
Cell[320832, 8935, 604, 12, 22, "Print",ExpressionUUID->"09cfae9b-700a-4eef-9153-05e38e820a3f",
 CellID->1478657950],
Cell[321439, 8949, 598, 12, 22, "Print",ExpressionUUID->"dba1c412-6958-4141-af3d-bf8af67bf99c",
 CellID->1123402566],
Cell[322040, 8963, 593, 12, 22, "Print",ExpressionUUID->"d59ed287-1150-420e-9b0e-e534ffe442ef",
 CellID->350702338],
Cell[322636, 8977, 584, 12, 22, "Print",ExpressionUUID->"7ac86513-5211-4652-9b5c-ac92ee705cf9",
 CellID->1559067398],
Cell[323223, 8991, 580, 12, 22, "Print",ExpressionUUID->"0161eb23-2136-4ae0-9553-4238645f5e8e",
 CellID->1797781139],
Cell[323806, 9005, 599, 12, 22, "Print",ExpressionUUID->"2c3c0f16-270b-4dc9-a3ed-9586f457b607",
 CellID->212689883],
Cell[324408, 9019, 580, 12, 22, "Print",ExpressionUUID->"e7809ed8-dbb8-4cea-9717-950f98cbf247",
 CellID->1266074973],
Cell[324991, 9033, 568, 12, 22, "Print",ExpressionUUID->"84270404-7932-4fcc-b76d-c401e8c5fa07",
 CellID->1654096735],
Cell[325562, 9047, 576, 12, 22, "Print",ExpressionUUID->"d0e91fcb-64c8-4210-b4fc-1f3d772ff66e",
 CellID->1228014730],
Cell[326141, 9061, 589, 12, 22, "Print",ExpressionUUID->"dbfea2e9-e9bb-45e6-8527-48bad9dee745",
 CellID->1925719440],
Cell[326733, 9075, 581, 12, 22, "Print",ExpressionUUID->"4245844c-4c89-43fd-a30d-817c279304d4",
 CellID->840584143]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[327363, 9093, 272, 6, 47, "Input",ExpressionUUID->"91e11ed1-820b-4c86-89df-2be8b24baa86"],
Cell[CellGroupData[{
Cell[327660, 9103, 547, 12, 22, "Print",ExpressionUUID->"e9715e34-33d3-4607-8ff6-7438d65a1aa6",
 CellID->1457251913],
Cell[328210, 9117, 548, 12, 22, "Print",ExpressionUUID->"1418575a-76a2-44c3-b12b-e04cb38f80a9",
 CellID->2129391403],
Cell[328761, 9131, 615, 12, 22, "Print",ExpressionUUID->"ca333a7e-fc91-43a0-a37c-fedf286fdeb1",
 CellID->866332966],
Cell[329379, 9145, 599, 12, 22, "Print",ExpressionUUID->"9bdb7152-61f3-4137-b7a5-6f8a1be95005",
 CellID->519032902],
Cell[329981, 9159, 581, 12, 22, "Print",ExpressionUUID->"50a7d844-5da4-41ee-8c0f-1b760a96adf2",
 CellID->307409800],
Cell[330565, 9173, 580, 12, 22, "Print",ExpressionUUID->"5e36962d-7b18-4d18-a363-9076df425baf",
 CellID->1938337565],
Cell[331148, 9187, 573, 12, 22, "Print",ExpressionUUID->"ac560e78-b987-458f-a447-57c1f6c03c6b",
 CellID->312529965]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[331770, 9205, 146, 2, 27, "Input",ExpressionUUID->"5a9af476-fba0-4d6d-a72e-c1eabd4c9ce8"],
Cell[CellGroupData[{
Cell[331941, 9211, 725, 18, 22, "Print",ExpressionUUID->"543a0863-baf2-4e44-87b4-8a91dc82adb6",
 CellID->132224728],
Cell[332669, 9231, 560, 12, 22, "Print",ExpressionUUID->"a8f07585-618c-4e5a-befa-d81792a06a29",
 CellID->1802912498]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[333290, 9250, 186, 5, 40, "MathCaption",ExpressionUUID->"701669b9-41a3-46d5-abd1-294561a8e615"],
Cell[CellGroupData[{
Cell[333501, 9259, 130, 2, 27, "Input",ExpressionUUID->"e8679f61-e704-4ff9-a662-fa5da7ef1d14"],
Cell[333634, 9263, 212, 3, 43, "Print",ExpressionUUID->"94eae8c2-9b3a-4e88-b43e-adc509ad506c",
 CellTags->"Info443728981228-9181623",
 CellID->1995674241]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[333907, 9273, 89, 0, 39, "Subsection",ExpressionUUID->"5889dd12-3523-4e67-99cd-4929c03562c1"],
Cell[333999, 9275, 1923, 36, 927, "Text",ExpressionUUID->"6d0e05f7-f5f5-43a1-9fe1-f0b6022a9853"]
}, Closed]],
Cell[CellGroupData[{
Cell[335959, 9316, 94, 0, 39, "Subsection",ExpressionUUID->"49d482a0-f3b5-4c3e-8b1f-dc27e05614d1"],
Cell[CellGroupData[{
Cell[336078, 9320, 722, 22, 57, "Text",ExpressionUUID->"cd41f060-d80b-4635-8e01-65bab609a452"],
Cell[CellGroupData[{
Cell[336825, 9346, 133, 2, 27, "Input",ExpressionUUID->"89d6fdb0-314a-4340-85fe-3cecc475cf03"],
Cell[336961, 9350, 677, 10, 133, "Print",ExpressionUUID->"e4d5af28-bacd-48e1-a58c-0f67bfc59604",
 CellTags->"Info453728981253-9181623",
 CellID->1862154774]
}, Open  ]],
Cell[CellGroupData[{
Cell[337675, 9365, 128, 2, 27, "Input",ExpressionUUID->"fc20be92-4173-4af0-95d3-9b1a35edd183"],
Cell[CellGroupData[{
Cell[337828, 9371, 318, 5, 44, "Print",ExpressionUUID->"61ca6cff-8a7c-4bca-b966-a93a55cad7c9",
 CellTags->"Info463728981253-9181623",
 CellID->1956098887],
Cell[338149, 9378, 1591, 48, 159, "Print",ExpressionUUID->"560a3844-0f62-4961-bffb-79612314e255",
 CellTags->"Info463728981253-9181623",
 CellID->326841886]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[339813, 9434, 100, 0, 39, "Subsection",ExpressionUUID->"421f06b6-b5e8-46e6-9b8c-07e14a614d08"],
Cell[339916, 9436, 252, 6, 23, "Text",ExpressionUUID->"5cffb5aa-7bf8-46f7-9de6-c8a1c83f4f25"],
Cell[CellGroupData[{
Cell[340193, 9446, 323, 7, 56, "MathCaption",ExpressionUUID->"8f8513b0-163c-471e-a39e-3e01852ab6c4"],
Cell[CellGroupData[{
Cell[340541, 9457, 227, 6, 27, "Input",ExpressionUUID->"3483e09f-16f4-4b30-b162-c683f4ed5e3a"],
Cell[340771, 9465, 611, 19, 26, "Output",ExpressionUUID->"6919d85c-3f2b-4b91-a9d2-0b8ac99f42b1",
 CellID->1240949882]
}, Open  ]]
}, Open  ]],
Cell[341409, 9488, 410, 8, 72, "MathCaption",ExpressionUUID->"4d9bf58f-f25b-4a42-b26d-97acdd4601dd"],
Cell[341822, 9498, 468, 8, 72, "MathCaption",ExpressionUUID->"403cc0c0-39e7-440d-ac09-bfcd7e3d1699"],
Cell[342293, 9508, 275, 8, 40, "MathCaption",ExpressionUUID->"36e8529a-7053-4ef3-a291-9628ad03698f"],
Cell[CellGroupData[{
Cell[342593, 9520, 346, 5, 72, "MathCaption",ExpressionUUID->"47104d1e-2ff2-4c2c-9286-76126b9bf7ee"],
Cell[342942, 9527, 171, 3, 27, "Input",ExpressionUUID->"e615b8ea-04d7-41a0-8745-10fae41ccf8a"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[343162, 9536, 90, 0, 39, "Subsection",ExpressionUUID->"326cef8c-f330-45b8-9cd1-30477fc02a2d"],
Cell[343255, 9538, 874, 22, 74, "Text",ExpressionUUID->"795b24da-db86-46d7-acda-cf427ab5e34a"],
Cell[344132, 9562, 486, 9, 57, "Text",ExpressionUUID->"23fc648e-4045-4311-b83d-7549addc71c7"],
Cell[344621, 9573, 704, 15, 74, "Text",ExpressionUUID->"9e78cb1c-454f-4695-b5d1-507e97f3f100"]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[345374, 9594, 602, 10, 54, "Section",ExpressionUUID->"f7d60e03-9b2e-438d-919f-8d580da78398",
 CellTags->{"S5.18.1", "1.4"},
 CellID->390546124],
Cell[CellGroupData[{
Cell[346001, 9608, 122, 0, 39, "Subsection",ExpressionUUID->"e8030db3-1fe6-4a08-824b-ae44525c0ac0"],
Cell[346126, 9610, 293, 6, 40, "Text",ExpressionUUID->"0e6cf3af-6b03-4690-9088-015deab2bb53"],
Cell[346422, 9618, 1596, 42, 109, "DefinitionBox",ExpressionUUID->"1ef92702-7204-4829-825f-4d0221cb4c06"],
Cell[348021, 9662, 96, 0, 23, "Text",ExpressionUUID->"98f97fb0-2352-438c-9eee-7798834bb3fd"],
Cell[CellGroupData[{
Cell[348142, 9666, 255, 6, 40, "MathCaption",ExpressionUUID->"a102bb63-2a7b-4eae-8d1b-25fabc2ddb5a"],
Cell[CellGroupData[{
Cell[348422, 9676, 297, 7, 27, "Input",ExpressionUUID->"d39feba1-d953-4465-be4f-f38923b0a0ad"],
Cell[CellGroupData[{
Cell[348744, 9687, 594, 12, 22, "Print",ExpressionUUID->"c0dd4935-216e-43a3-81ba-4651922e51f3",
 CellID->894423641],
Cell[349341, 9701, 602, 12, 22, "Print",ExpressionUUID->"af7cc6fa-cafc-4978-8540-2cced64763b0",
 CellID->192605287]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[349992, 9719, 146, 2, 27, "Input",ExpressionUUID->"39af99de-0f20-4702-96c6-86ce614679e2"],
Cell[350141, 9723, 169, 3, 26, "Output",ExpressionUUID->"7079644f-84ec-4707-bbb3-c2328f7b02a9",
 CellID->859202580]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[350359, 9732, 454, 12, 72, "MathCaption",ExpressionUUID->"8ad557fd-7d55-44df-b518-57f5ee4392ca"],
Cell[CellGroupData[{
Cell[350838, 9748, 124, 2, 27, "Input",ExpressionUUID->"97c5c513-db25-4ea5-9b9f-083bc3d3e4cd"],
Cell[CellGroupData[{
Cell[350987, 9754, 196, 3, 43, "Print",ExpressionUUID->"0c64b26f-1b6a-426c-8169-13f835a49e44",
 CellTags->"Info513728981275-9181623",
 CellID->580545217],
Cell[351186, 9759, 1746, 53, 193, "Print",ExpressionUUID->"794dc6e3-9199-42ce-b23c-afc9b0990f95",
 CellTags->"Info513728981275-9181623",
 CellID->1205253798]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[352971, 9817, 739, 14, 74, "Text",ExpressionUUID->"9a5ff845-8fc9-4b04-966b-e13a0861acde"],
Cell[353713, 9833, 384, 9, 41, "DefinitionBox",ExpressionUUID->"93c7beb0-5818-4811-b987-edb23c890510"],
Cell[354100, 9844, 100, 0, 23, "Text",ExpressionUUID->"ec987b65-22b9-4127-82c4-1ff98ca47192"],
Cell[CellGroupData[{
Cell[354225, 9848, 291, 4, 56, "MathCaption",ExpressionUUID->"4387b72c-5344-483c-9fc7-70a077da29a1"],
Cell[CellGroupData[{
Cell[354541, 9856, 140, 2, 27, "Input",ExpressionUUID->"6a3d6e27-68d1-4ed0-b12f-b7b5f0d81b17"],
Cell[354684, 9860, 321, 9, 26, "Output",ExpressionUUID->"49510770-32ed-48b3-a843-9df916670925",
 CellID->702233062]
}, Open  ]],
Cell[CellGroupData[{
Cell[355042, 9874, 137, 2, 27, "Input",ExpressionUUID->"77e79572-07a7-407f-8d75-25de623e6c28"],
Cell[355182, 9878, 179, 3, 26, "Output",ExpressionUUID->"a922cc88-5ad6-4a59-9153-80931a54f97c",
 CellID->163721661]
}, Open  ]],
Cell[CellGroupData[{
Cell[355398, 9886, 131, 2, 27, "Input",ExpressionUUID->"d8756b4b-befd-4be5-98f3-fca67301574e"],
Cell[CellGroupData[{
Cell[355554, 9892, 348, 9, 44, "Print",ExpressionUUID->"a14a8109-2565-4fe7-84ec-88b5522828b6",
 CellTags->"Info543728981286-9181623",
 CellID->76239291],
Cell[355905, 9903, 5061, 155, 363, "Print",ExpressionUUID->"2ec7dae1-8a02-4a21-a5a7-9fd5c71f3517",
 CellTags->"Info543728981286-9181623",
 CellID->1453510225]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[361027, 10065, 887, 32, 88, "MathCaption",ExpressionUUID->"045d3424-06d0-4dc2-b568-5739921d2361"],
Cell[CellGroupData[{
Cell[361939, 10101, 291, 6, 27, "Input",ExpressionUUID->"c829f48e-d04d-4c93-bf2a-d1bb043cdf03"],
Cell[CellGroupData[{
Cell[362255, 10111, 594, 12, 22, "Print",ExpressionUUID->"96aa42ed-b72f-4b51-8ebc-18fbe500705d",
 CellID->391777724],
Cell[362852, 10125, 602, 12, 22, "Print",ExpressionUUID->"70605633-1a52-4f41-95cf-52a2d3d355c9",
 CellID->423345566]
}, Open  ]],
Cell[363469, 10140, 447, 9, 22, "Message",ExpressionUUID->"0e4c297c-4466-49fc-98f1-255a4fe9dcda",
 CellID->555605251],
Cell[363919, 10151, 447, 9, 22, "Message",ExpressionUUID->"3ac2529b-dc46-4b1a-bdab-d9a8bcb005a6",
 CellID->798556401]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[364415, 10166, 325, 9, 56, "MathCaption",ExpressionUUID->"b035769e-73b9-4ef0-8395-d8fa1b839b26"],
Cell[CellGroupData[{
Cell[364765, 10179, 115, 1, 27, "Input",ExpressionUUID->"5c36480d-b101-471e-90cb-85cdc6a583e1"],
Cell[364883, 10182, 217, 5, 26, "Output",ExpressionUUID->"aa376e0c-a3ea-4522-b6ab-259052088c62",
 CellID->1667145854]
}, Open  ]],
Cell[CellGroupData[{
Cell[365137, 10192, 114, 1, 27, "Input",ExpressionUUID->"374d1cc9-99d6-48b1-87e5-0c7b36bc0c82"],
Cell[365254, 10195, 543, 17, 26, "Output",ExpressionUUID->"c310b3d2-229c-46b7-9e58-826d511bbc1a",
 CellID->2144759790]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[365846, 10218, 233, 3, 56, "MathCaption",ExpressionUUID->"7ec5d8f7-916b-4409-a627-cbed7a34d8c3"],
Cell[CellGroupData[{
Cell[366104, 10225, 146, 2, 27, "Input",ExpressionUUID->"d628234a-4a57-4cf7-8f87-bd66ea255b24"],
Cell[CellGroupData[{
Cell[366275, 10231, 725, 18, 22, "Print",ExpressionUUID->"accd7dfe-6599-404b-ac43-3a3fe8ac4b64",
 CellID->390229355],
Cell[367003, 10251, 559, 12, 22, "Print",ExpressionUUID->"a7a278e1-17e9-438d-8a92-6e3862171a1a",
 CellID->169711547]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[367611, 10269, 124, 2, 27, "Input",ExpressionUUID->"d14ddd2c-fbda-4970-bdf8-38eafa357efd"],
Cell[367738, 10273, 384, 8, 22, "Message",ExpressionUUID->"16d090a3-6e27-402c-9f0c-7d3148fa067e",
 CellID->111960166]
}, Open  ]],
Cell[CellGroupData[{
Cell[368159, 10286, 115, 1, 27, "Input",ExpressionUUID->"1f4fe908-f907-469c-bebb-69cb37f5e101"],
Cell[368277, 10289, 192, 4, 26, "Output",ExpressionUUID->"3c5798aa-4f7b-46c8-b82b-58bd42b5b6fa",
 CellID->730204304]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[368518, 10299, 354, 9, 40, "MathCaption",ExpressionUUID->"2865d1d0-4b87-4877-a573-37bb04904229"],
Cell[CellGroupData[{
Cell[368897, 10312, 250, 5, 27, "Input",ExpressionUUID->"66c7fb81-9d85-4f47-b94b-23f14098af8e"],
Cell[CellGroupData[{
Cell[369172, 10321, 594, 12, 22, "Print",ExpressionUUID->"bd1f5700-9263-42b6-909a-c1f719d85a2b",
 CellID->442594184],
Cell[369769, 10335, 603, 12, 22, "Print",ExpressionUUID->"31490e21-67ca-4617-adac-be7e1b06701b",
 CellID->1944580927]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[370421, 10353, 115, 1, 27, "Input",ExpressionUUID->"d208da62-5962-4155-93ee-f6e9ea12979b"],
Cell[370539, 10356, 215, 5, 26, "Output",ExpressionUUID->"cbd49fc0-aad4-45da-bc9e-481109a90198",
 CellID->863685934]
}, Open  ]]
}, Open  ]],
Cell[370781, 10365, 601, 10, 76, "Text",ExpressionUUID->"87931585-5c99-4960-96b8-0cbc8e05ae00"],
Cell[371385, 10377, 260, 4, 40, "Text",ExpressionUUID->"899896a8-1b63-4664-9c40-f6a921808b78"]
}, Closed]],
Cell[CellGroupData[{
Cell[371682, 10386, 176, 4, 39, "Subsection",ExpressionUUID->"cf275d8f-dc19-4950-a603-a03bd3072d50"],
Cell[371861, 10392, 872, 17, 91, "Text",ExpressionUUID->"33450aa3-d5e7-4cc4-ac55-fcb1bd7e05bd"],
Cell[CellGroupData[{
Cell[372758, 10413, 109, 0, 40, "MathCaption",ExpressionUUID->"b118a9ae-0d52-4943-91b2-0cc2f95370a6"],
Cell[CellGroupData[{
Cell[372892, 10417, 391, 9, 27, "Input",ExpressionUUID->"e62bf63c-d578-475a-88d2-b90e2f37f7ef"],
Cell[CellGroupData[{
Cell[373308, 10430, 591, 12, 22, "Print",ExpressionUUID->"50c1bd67-8f9b-4465-96ed-1712efdb71cd",
 CellID->422190744],
Cell[373902, 10444, 602, 12, 22, "Print",ExpressionUUID->"bafbbe97-0fae-440d-9a24-d7ec43e0b226",
 CellID->473864720]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[374553, 10462, 124, 2, 27, "Input",ExpressionUUID->"2b6c1e02-5bab-41b5-ac57-419456dcbcec"],
Cell[CellGroupData[{
Cell[374702, 10468, 197, 3, 43, "Print",ExpressionUUID->"495fb146-4919-4a2e-a6a9-01c1ea7ad2db",
 CellTags->"Info643728981304-9181623",
 CellID->1678239341],
Cell[374902, 10473, 2093, 64, 261, "Print",ExpressionUUID->"696209d9-520d-4605-8a2b-cc1e264a6a62",
 CellTags->"Info643728981304-9181623",
 CellID->1118384489]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[377056, 10544, 156, 2, 40, "MathCaption",ExpressionUUID->"7b872029-8c5d-42c0-b8c4-bbe8ae53acfa"],
Cell[CellGroupData[{
Cell[377237, 10550, 131, 2, 27, "Input",ExpressionUUID->"819acafa-417f-4430-b4a2-8cde5aee1b6e"],
Cell[CellGroupData[{
Cell[377393, 10556, 352, 9, 44, "Print",ExpressionUUID->"c67e0a3a-aab2-4e76-b372-ae1b984fb9b1",
 CellTags->"Info653728981305-9181623",
 CellID->1548585086],
Cell[377748, 10567, 7095, 214, 431, "Print",ExpressionUUID->"4f8c5a66-9828-4a8c-a4c2-2c05d8fba923",
 CellTags->"Info653728981305-9181623",
 CellID->1715222409]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[384904, 10788, 152, 2, 40, "MathCaption",ExpressionUUID->"19b5b905-c580-4aaa-8d99-1fa6e4b1cd0f"],
Cell[CellGroupData[{
Cell[385081, 10794, 122, 1, 27, "Input",ExpressionUUID->"8a20050f-5b08-40ba-ac20-3a2851925ba5"],
Cell[385206, 10797, 193, 4, 26, "Output",ExpressionUUID->"572978e1-b92a-47e1-b5e1-87f76555d6c8",
 CellID->1383227708]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[385448, 10807, 248, 6, 23, "Text",ExpressionUUID->"3f032dea-7b04-4075-bf98-58359ab38213"],
Cell[CellGroupData[{
Cell[385721, 10817, 135, 2, 27, "Input",ExpressionUUID->"a82a7eff-216c-4076-ada8-b0c9515a073e"],
Cell[385859, 10821, 482, 7, 89, "Print",ExpressionUUID->"f89e4fe2-2986-45ad-b58e-63513099d114",
 CellTags->"Info673728981308-9181623",
 CellID->1069478814]
}, Open  ]]
}, Open  ]],
Cell[386368, 10832, 160, 3, 23, "Text",ExpressionUUID->"c5abdda0-498c-44d1-9cee-e7f2ca7d7d83"],
Cell[386531, 10837, 417, 9, 41, "DefinitionBox",ExpressionUUID->"2b1d3866-bcef-4cf6-9cfc-536e2910f6bb"],
Cell[386951, 10848, 102, 0, 23, "Text",ExpressionUUID->"6bbcb532-2ca1-4ca1-9dc2-dec86556317b"],
Cell[CellGroupData[{
Cell[387078, 10852, 118, 0, 40, "MathCaption",ExpressionUUID->"80005056-4171-41b4-9f96-b546be1999e0"],
Cell[CellGroupData[{
Cell[387221, 10856, 200, 4, 27, "Input",ExpressionUUID->"933e0218-cae1-4abb-b4b7-17711468fe81"],
Cell[387424, 10862, 199, 4, 26, "Output",ExpressionUUID->"837146aa-d13f-4041-b227-6069f282aa7f",
 CellID->839651]
}, Open  ]],
Cell[CellGroupData[{
Cell[387660, 10871, 144, 2, 27, "Input",ExpressionUUID->"ceee3743-da5b-4f41-a2b8-f82e37c94a6b"],
Cell[387807, 10875, 173, 3, 26, "Output",ExpressionUUID->"c2853399-a2a9-47a4-be99-edf76576abdb",
 CellID->1946601836]
}, Open  ]],
Cell[CellGroupData[{
Cell[388017, 10883, 148, 2, 27, "Input",ExpressionUUID->"38297d9f-4f24-4c5a-b946-0dd0e2432b18"],
Cell[388168, 10887, 170, 3, 26, "Output",ExpressionUUID->"b1828f0b-4666-4361-ac7b-9dc736f33796",
 CellID->1089153226]
}, Open  ]],
Cell[CellGroupData[{
Cell[388375, 10895, 142, 2, 27, "Input",ExpressionUUID->"bdede592-7995-498d-aa4d-f58c7be3bf5d"],
Cell[388520, 10899, 521, 16, 26, "Output",ExpressionUUID->"ec06b306-d0f0-4217-ad59-0120d74adccb",
 CellID->718487895]
}, Open  ]],
Cell[CellGroupData[{
Cell[389078, 10920, 136, 2, 27, "Input",ExpressionUUID->"ad47a5b6-d7e8-40c3-8c03-ca9a42bee6d3"],
Cell[389217, 10924, 382, 11, 26, "Output",ExpressionUUID->"a5049b68-0212-412c-8656-d998e2a2dd32",
 CellID->2004106280]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[389648, 10941, 87, 0, 40, "MathCaption",ExpressionUUID->"4405d533-bd0c-4e8e-9732-2b0026dda536"],
Cell[CellGroupData[{
Cell[389760, 10945, 146, 2, 27, "Input",ExpressionUUID->"b46e64fa-0fdb-4f84-85e1-2343055f28f4"],
Cell[CellGroupData[{
Cell[389931, 10951, 726, 18, 22, "Print",ExpressionUUID->"9461cfbc-d200-4bfd-ba4f-c6d0962844f5",
 CellID->1972642216],
Cell[390660, 10971, 558, 12, 22, "Print",ExpressionUUID->"fcd65b7e-ed9b-4d27-b95d-8e2edaebfae4",
 CellID->1345181926]
}, Open  ]]
}, Open  ]],
Cell[391245, 10987, 141, 2, 27, "Input",ExpressionUUID->"9372b455-0c25-43f5-bda9-2d171282f8c9"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[391435, 10995, 102, 0, 39, "Subsection",ExpressionUUID->"41df9b57-d4a7-402c-a52f-449095ba56f2"],
Cell[391540, 10997, 429, 10, 40, "Text",ExpressionUUID->"4dcae6d3-f9c5-4c84-a1c5-aa7d6ce12852"],
Cell[391972, 11009, 1568, 42, 109, "DefinitionBox",ExpressionUUID->"e41db040-0eeb-4478-8fa3-25c3c0a85e46"],
Cell[393543, 11053, 95, 0, 23, "Text",ExpressionUUID->"41d0679c-1a13-485d-b03c-a867b3250886"],
Cell[CellGroupData[{
Cell[393663, 11057, 239, 6, 40, "MathCaption",ExpressionUUID->"a610520d-1f6c-4913-96fc-32f9ceab89c9"],
Cell[CellGroupData[{
Cell[393927, 11067, 502, 10, 27, "Input",ExpressionUUID->"e1f8cde2-3386-4628-9c85-b2d541380ea4"],
Cell[CellGroupData[{
Cell[394454, 11081, 596, 12, 22, "Print",ExpressionUUID->"24e59c23-5cf5-4355-9fa4-771438c3c4e0",
 CellID->960640895],
Cell[395053, 11095, 778, 15, 42, "Print",ExpressionUUID->"a8cd3616-0b7a-4b96-a654-72359909efb5",
 CellID->709464119]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[395880, 11116, 128, 2, 27, "Input",ExpressionUUID->"49a3e91a-e823-44b0-bcc3-c0adf147d061"],
Cell[CellGroupData[{
Cell[396033, 11122, 330, 9, 44, "Print",ExpressionUUID->"78f21f23-a24d-4d19-9550-1ea66aeba6ec",
 CellTags->"Info763728981334-9181623",
 CellID->1134239889],
Cell[396366, 11133, 5501, 174, 397, "Print",ExpressionUUID->"86d299a7-dd5d-4f3d-ba67-a9c99a8f3c23",
 CellTags->"Info763728981334-9181623",
 CellID->2104963651]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[401928, 11314, 686, 17, 88, "MathCaption",ExpressionUUID->"51b25888-11e0-4f34-9197-3d72c4553576"],
Cell[CellGroupData[{
Cell[402639, 11335, 137, 2, 27, "Input",ExpressionUUID->"27f253d2-1c2f-4182-a2a5-a4d6b96ff92c"],
Cell[CellGroupData[{
Cell[402801, 11341, 345, 9, 44, "Print",ExpressionUUID->"e45eb883-9712-4a14-bb63-71266d89737a",
 CellTags->"Info773728981334-9181623",
 CellID->478481679],
Cell[403149, 11352, 5879, 178, 420, "Print",ExpressionUUID->"1b43b27e-5ef8-4e34-ad3f-b46f987e1e90",
 CellTags->"Info773728981334-9181623",
 CellID->1662804585]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[409089, 11537, 132, 0, 40, "MathCaption",ExpressionUUID->"281293b9-d69c-4c8b-8a41-b10a3c4440fb"],
Cell[CellGroupData[{
Cell[409246, 11541, 114, 1, 27, "Input",ExpressionUUID->"a4e76137-ed7d-4a74-b496-6490cc2cc946"],
Cell[409363, 11544, 865, 29, 26, "Output",ExpressionUUID->"c3a2e3f5-8e19-4730-a516-9a28e96094d1",
 CellID->1551961684]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[410277, 11579, 241, 8, 23, "Text",ExpressionUUID->"fa959902-aa79-407c-940f-2285556b445f"],
Cell[CellGroupData[{
Cell[410543, 11591, 134, 2, 27, "Input",ExpressionUUID->"3c96b45f-e102-4392-bf32-84e64578de9c"],
Cell[410680, 11595, 456, 6, 89, "Print",ExpressionUUID->"b624df61-586c-41cd-84c6-be2034fb0f05",
 CellTags->"Info793728981334-9181623",
 CellID->483646526]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[411197, 11608, 92, 0, 39, "Subsection",ExpressionUUID->"e88af057-5080-4dba-a53c-e91d18080d3c"],
Cell[411292, 11610, 254, 6, 23, "Text",ExpressionUUID->"ecadba5a-a0cd-4981-ae56-1e080a567e56"],
Cell[411549, 11618, 1283, 34, 92, "DefinitionBox",ExpressionUUID->"7b133ce1-4008-486c-bf5b-9db872b0a2db"],
Cell[412835, 11654, 96, 0, 23, "Text",ExpressionUUID->"af2f997a-c293-4dc2-abb5-f953b1d04d9b"],
Cell[CellGroupData[{
Cell[412956, 11658, 256, 6, 56, "MathCaption",ExpressionUUID->"703ba00d-9850-441c-a4d1-fa1478287f89"],
Cell[CellGroupData[{
Cell[413237, 11668, 147, 2, 27, "Input",ExpressionUUID->"17213dac-b2f6-46c4-bafe-dc0ad267bac7"],
Cell[413387, 11672, 602, 12, 22, "Print",ExpressionUUID->"af4bb84c-e5cf-4949-8f6d-962879c7ad1d",
 CellID->229219215]
}, Open  ]],
Cell[CellGroupData[{
Cell[414026, 11689, 126, 2, 27, "Input",ExpressionUUID->"ad324b11-367e-4fda-bdea-1eaf4530fe4d"],
Cell[CellGroupData[{
Cell[414177, 11695, 324, 9, 44, "Print",ExpressionUUID->"c8622f06-4d6e-4201-bb88-bb26ce55c8ab",
 CellTags->"Info813728981342-9181623",
 CellID->2049167434],
Cell[414504, 11706, 1704, 51, 91, "Print",ExpressionUUID->"3f1bc420-c985-4a39-be0e-84a2da604eb8",
 CellTags->"Info813728981342-9181623",
 CellID->1161816299]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[416247, 11762, 268, 6, 23, "Text",ExpressionUUID->"b54369c3-fed7-4946-bce8-8e08a33b3a19"],
Cell[416518, 11770, 628, 11, 74, "Text",ExpressionUUID->"869cb9c3-a5d7-4433-a5c8-0af58d5e7af5"]
}, Closed]],
Cell[CellGroupData[{
Cell[417183, 11786, 117, 0, 39, "Subsection",ExpressionUUID->"f4a065ec-8746-4f74-990d-0ab375474513"],
Cell[417303, 11788, 260, 6, 23, "Text",ExpressionUUID->"5bf2dfdc-8a1a-4613-9ba4-36ee49938748"],
Cell[CellGroupData[{
Cell[417588, 11798, 201, 3, 56, "MathCaption",ExpressionUUID->"c7cf903b-a7b6-4bfd-a118-2acc357d11a5"],
Cell[CellGroupData[{
Cell[417814, 11805, 195, 4, 27, "Input",ExpressionUUID->"24f3f38a-1ee0-4ea8-8948-eb7857b82456"],
Cell[CellGroupData[{
Cell[418034, 11813, 593, 12, 22, "Print",ExpressionUUID->"877caff4-d563-41eb-8204-14522d3e65f3",
 CellID->2012913861],
Cell[418630, 11827, 268, 4, 22, "Print",ExpressionUUID->"45226275-4177-442a-97af-57c5f2c2008a",
 CellID->1733370636]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[418959, 11838, 212, 3, 56, "MathCaption",ExpressionUUID->"8071bcb4-9e39-4e02-9c88-eaff93fed879"],
Cell[CellGroupData[{
Cell[419196, 11845, 274, 6, 27, "Input",ExpressionUUID->"36d36bdc-c191-40c9-a208-1e01bfb4120e"],
Cell[419473, 11853, 587, 12, 22, "Print",ExpressionUUID->"d3397308-8b78-4841-bb77-be581f7cfcae",
 CellID->2144607697]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[420109, 11871, 164, 2, 40, "MathCaption",ExpressionUUID->"b85a38d6-fa67-46e7-b45d-96783d8b665a"],
Cell[CellGroupData[{
Cell[420298, 11877, 221, 5, 27, "Input",ExpressionUUID->"b069a155-d510-4b25-bc89-3e5500b1c931"],
Cell[420522, 11884, 643, 12, 22, "Print",ExpressionUUID->"20f1dcde-4f43-4ee5-bcdf-6e76b39e110f",
 CellID->1056032231]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[421214, 11902, 99, 0, 40, "MathCaption",ExpressionUUID->"e293c462-6e57-4668-b017-4a2de63c24dd"],
Cell[CellGroupData[{
Cell[421338, 11906, 143, 2, 27, "Input",ExpressionUUID->"4747e910-5168-427c-9784-59dd97f21fc5"],
Cell[421484, 11910, 548, 12, 22, "Print",ExpressionUUID->"e9753a4c-e1bb-463f-bcdf-0d23b383f27a",
 CellID->1928847609]
}, Open  ]],
Cell[CellGroupData[{
Cell[422069, 11927, 144, 2, 27, "Input",ExpressionUUID->"b2001d89-3fa4-49a6-bc39-dfb1098e92c8"],
Cell[422216, 11931, 684, 18, 22, "Print",ExpressionUUID->"da90fa35-b075-4eab-ae0b-ec9c68303200",
 CellID->1876112998]
}, Open  ]],
Cell[CellGroupData[{
Cell[422937, 11954, 145, 2, 27, "Input",ExpressionUUID->"b02d2530-6567-4f11-8dd4-445ef9080e93"],
Cell[423085, 11958, 555, 12, 22, "Print",ExpressionUUID->"a03abb13-cb34-493a-9369-4f47f122a6a7",
 CellID->328768485]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[423713, 11978, 619, 11, 54, "Section",ExpressionUUID->"e50a4bbd-8e74-4fce-82b6-11d15ef76e1d",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1215012216],
Cell[CellGroupData[{
Cell[424357, 11993, 100, 0, 39, "Subsection",ExpressionUUID->"f3e6e9da-0f4e-40b2-b1ed-c50ec9fe6a74"],
Cell[424460, 11995, 767, 12, 91, "Text",ExpressionUUID->"e6ece487-8eb2-43a6-ace3-ff0e4f43fa3c"],
Cell[425230, 12009, 345, 7, 40, "Text",ExpressionUUID->"d4d91dd1-913d-4b76-a80e-e65db9625f77"],
Cell[425578, 12018, 378, 5, 57, "Text",ExpressionUUID->"0499bb65-c6b2-43a8-bc89-8dece0141f98"],
Cell[425959, 12025, 458, 7, 98, "Text",ExpressionUUID->"ba883f62-4a3d-4909-9d34-4850e872ed9d"]
}, Closed]],
Cell[CellGroupData[{
Cell[426454, 12037, 127, 0, 39, "Subsection",ExpressionUUID->"1d6d10f7-5284-4d78-8010-4890039bf855"],
Cell[426584, 12039, 2224, 61, 231, "Text",ExpressionUUID->"78c412bc-f1ee-408f-a244-2126222cca9b"],
Cell[428811, 12102, 575, 16, 23, "Text",ExpressionUUID->"7bcc6a39-f7e2-45a5-9117-6844b149e1cb"],
Cell[429389, 12120, 1464, 36, 108, "Text",ExpressionUUID->"c9dc8329-8018-4ea8-a951-b43f659228d4"],
Cell[430856, 12158, 805, 15, 103, "Text",ExpressionUUID->"5979c260-3a9b-46c4-9832-5a7d8bb9dd1e"]
}, Closed]],
Cell[CellGroupData[{
Cell[431698, 12178, 126, 0, 39, "Subsection",ExpressionUUID->"05082b76-880a-4928-83c2-67d4901eca40"],
Cell[431827, 12180, 1293, 44, 91, "Text",ExpressionUUID->"0384c176-8639-4d9d-bcff-4069a4f44e53"],
Cell[433123, 12226, 746, 20, 75, "DefinitionBox",ExpressionUUID->"4360fc4d-6fca-4955-90c0-3a73c56ba2de"],
Cell[433872, 12248, 93, 0, 23, "Text",ExpressionUUID->"965569c5-aac7-4c15-9732-9ccba76719fc"],
Cell[CellGroupData[{
Cell[433990, 12252, 232, 5, 40, "MathCaption",ExpressionUUID->"e7ce011f-252a-484a-b040-bb1673e24c97"],
Cell[CellGroupData[{
Cell[434247, 12261, 139, 2, 27, "Input",ExpressionUUID->"80c5e0ea-d526-4a38-8257-b213c41e432d"],
Cell[434389, 12265, 173, 3, 26, "Output",ExpressionUUID->"397657ef-c0bb-4fb7-98d5-84c48bf79008",
 CellID->1077251231]
}, Open  ]],
Cell[CellGroupData[{
Cell[434599, 12273, 157, 3, 27, "Input",ExpressionUUID->"71a605ae-4f3e-4c73-8f9e-3d25178f3cc9"],
Cell[434759, 12278, 172, 3, 26, "Output",ExpressionUUID->"0e7c5339-1ef9-43bd-a44e-ef433d673c19",
 CellID->2074295207]
}, Open  ]],
Cell[CellGroupData[{
Cell[434968, 12286, 139, 2, 27, "Input",ExpressionUUID->"721b5a43-7732-4161-817e-96026db6d291"],
Cell[435110, 12290, 173, 3, 26, "Output",ExpressionUUID->"0d5eb9fd-25d8-47b8-bda6-1928e31800df",
 CellID->430018651]
}, Open  ]],
Cell[435298, 12296, 138, 2, 27, "Input",ExpressionUUID->"9546b9f7-b9d9-4966-9d1f-79a7adf1f4f4"]
}, Open  ]],
Cell[CellGroupData[{
Cell[435473, 12303, 140, 1, 40, "MathCaption",ExpressionUUID->"81bd29c2-5f25-41da-8ecd-d8a5ab274c11"],
Cell[CellGroupData[{
Cell[435638, 12308, 170, 3, 27, "Input",ExpressionUUID->"3fdf3655-d045-4c16-9085-809d392e337b"],
Cell[435811, 12313, 173, 3, 26, "Output",ExpressionUUID->"44b8a245-1440-459f-9abe-3ae9e96fa183",
 CellID->1040995202]
}, Open  ]],
Cell[CellGroupData[{
Cell[436021, 12321, 186, 3, 27, "Input",ExpressionUUID->"b52ffada-af64-4caf-bcee-7d5d4b44afa4"],
Cell[436210, 12326, 175, 3, 26, "Output",ExpressionUUID->"3f4c39d4-08f4-4a80-b461-b85d542f94df",
 CellID->457087905]
}, Open  ]],
Cell[CellGroupData[{
Cell[436422, 12334, 183, 3, 27, "Input",ExpressionUUID->"4523e501-bcf2-44ca-84c5-8d130fd60953"],
Cell[436608, 12339, 172, 3, 26, "Output",ExpressionUUID->"dfdb62b5-4db2-4525-8d5e-bf8d2599b854",
 CellID->727414016]
}, Open  ]],
Cell[CellGroupData[{
Cell[436817, 12347, 192, 3, 27, "Input",ExpressionUUID->"e6831a30-14c5-4617-98bf-db129f2a99bb"],
Cell[437012, 12352, 171, 3, 26, "Output",ExpressionUUID->"7803c63a-3420-4271-b6e9-971bb2fcce3b",
 CellID->22964541]
}, Open  ]],
Cell[CellGroupData[{
Cell[437220, 12360, 204, 4, 27, "Input",ExpressionUUID->"5a1dba44-ace2-434d-9f99-d5f162116f4a"],
Cell[437427, 12366, 173, 3, 26, "Output",ExpressionUUID->"396e0ea3-8712-4e7b-bce8-654dec092a3e",
 CellID->1565467774]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[437649, 12375, 637, 15, 88, "MathCaption",ExpressionUUID->"b8055217-df79-4c70-bd9b-bb7fe710db97"],
Cell[CellGroupData[{
Cell[438311, 12394, 143, 2, 27, "Input",ExpressionUUID->"dad30ed6-0200-4030-8551-9bb366ebb246"],
Cell[438457, 12398, 175, 3, 26, "Output",ExpressionUUID->"fd6f2a5d-fd49-48ef-aba8-e713f1cdee68",
 CellID->1994718221]
}, Open  ]],
Cell[CellGroupData[{
Cell[438669, 12406, 161, 3, 27, "Input",ExpressionUUID->"454c5e63-6461-496f-b8d3-f13466121428"],
Cell[438833, 12411, 175, 3, 26, "Output",ExpressionUUID->"0529e84c-04d4-4088-9a8e-03f6e590dbe5",
 CellID->861927030]
}, Open  ]],
Cell[CellGroupData[{
Cell[439045, 12419, 143, 2, 27, "Input",ExpressionUUID->"a2fc3080-0eb3-4b50-a88e-afe72570ae57"],
Cell[439191, 12423, 176, 3, 26, "Output",ExpressionUUID->"2623b412-737b-4874-9c53-ccb18932f5d6",
 CellID->1630394195]
}, Open  ]],
Cell[CellGroupData[{
Cell[439404, 12431, 149, 3, 27, "Input",ExpressionUUID->"a0dcff7c-244e-4028-9ca0-29524d13b40c"],
Cell[439556, 12436, 173, 3, 26, "Output",ExpressionUUID->"e0fb5b01-4b2e-4d13-8f42-45af8107588b",
 CellID->143899645]
}, Open  ]],
Cell[CellGroupData[{
Cell[439766, 12444, 167, 4, 27, "Input",ExpressionUUID->"812c0777-64e1-4c21-ac43-2d4636fce7e3"],
Cell[439936, 12450, 175, 3, 26, "Output",ExpressionUUID->"55567e55-84f0-4f2a-aaf2-a057c5fdedca",
 CellID->454746993]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[440172, 12460, 99, 0, 39, "Subsection",ExpressionUUID->"8c80a1fc-ae59-4563-9ed8-74a9291dbd67"],
Cell[440274, 12462, 576, 14, 57, "Text",ExpressionUUID->"abf05f7d-06d5-41e5-84d5-fb9be5879686"],
Cell[440853, 12478, 1473, 40, 92, "DefinitionBox",ExpressionUUID->"c3266cf3-6f91-4b2f-bc1c-b8156c655034"],
Cell[442329, 12520, 125, 0, 23, "Text",ExpressionUUID->"afe27f77-65c3-4bd6-981b-82149b8b6a85"],
Cell[CellGroupData[{
Cell[442479, 12524, 150, 2, 40, "MathCaption",ExpressionUUID->"d5be2c62-63d1-4d6d-a3c0-9998bb7664e7"],
Cell[CellGroupData[{
Cell[442654, 12530, 160, 3, 27, "Input",ExpressionUUID->"55710039-c479-47a7-b899-8d71ae9e59e9"],
Cell[442817, 12535, 354, 10, 26, "Output",ExpressionUUID->"077638e7-1096-4d3c-b211-46856bc1d771",
 CellID->391059037]
}, Open  ]],
Cell[443186, 12548, 246, 6, 27, "Input",ExpressionUUID->"604a62e1-604b-408e-b03c-cf0fbe723691"],
Cell[CellGroupData[{
Cell[443457, 12558, 160, 3, 27, "Input",ExpressionUUID->"c452fb46-83d3-45b7-a523-573fb51ae540"],
Cell[443620, 12563, 395, 10, 26, "Output",ExpressionUUID->"865fcbc3-138c-43e9-b678-60093bcab966",
 CellID->1229275039]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[444064, 12579, 317, 5, 72, "MathCaption",ExpressionUUID->"fe691718-bca7-490d-810f-0e942c00e537"],
Cell[444384, 12586, 249, 6, 27, "Input",ExpressionUUID->"82c84a3a-3c9b-4bc5-ac1d-227eb375a76d"],
Cell[CellGroupData[{
Cell[444658, 12596, 160, 3, 27, "Input",ExpressionUUID->"f875e084-0440-429b-870c-7f0247c6e44f"],
Cell[444821, 12601, 355, 10, 26, "Output",ExpressionUUID->"8b2ccf57-90e0-4b2c-983a-e04cbc5807bc",
 CellID->1632896333]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[445225, 12617, 220, 6, 40, "MathCaption",ExpressionUUID->"4baf3557-ec02-446c-bbdf-962e9c3128bf"],
Cell[CellGroupData[{
Cell[445470, 12627, 154, 3, 27, "Input",ExpressionUUID->"39cc2e40-646d-4538-8615-073d65df5cf3"],
Cell[445627, 12632, 172, 3, 26, "Output",ExpressionUUID->"4dc095b7-56c6-4548-8b0e-57086340a0c6",
 CellID->1395143718]
}, Open  ]],
Cell[CellGroupData[{
Cell[445836, 12640, 160, 3, 27, "Input",ExpressionUUID->"82886089-dd34-47f4-815f-ea7b75deb5f5"],
Cell[445999, 12645, 361, 10, 26, "Output",ExpressionUUID->"9aba1e9d-6933-42b7-aa8a-ff59927dcc6c",
 CellID->1954010480]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[446409, 12661, 287, 6, 56, "MathCaption",ExpressionUUID->"a0ea6373-2bad-4d25-bee0-9053c79469c7"],
Cell[CellGroupData[{
Cell[446721, 12671, 257, 6, 27, "Input",ExpressionUUID->"438e6e57-0cde-47b7-a514-ca941c825f25"],
Cell[446981, 12679, 353, 8, 26, "Output",ExpressionUUID->"8211a7e3-a7a6-4cc6-9ef0-1124f422f01a",
 CellID->1734207436]
}, Open  ]],
Cell[CellGroupData[{
Cell[447371, 12692, 160, 3, 27, "Input",ExpressionUUID->"0a2288d1-079a-4af0-847c-7c30de238ef1"],
Cell[447534, 12697, 477, 13, 26, "Output",ExpressionUUID->"1bdc9522-c8a1-464b-8788-d3f155c0f8b7",
 CellID->1176319768]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[448060, 12716, 272, 4, 56, "MathCaption",ExpressionUUID->"45b89367-4d86-4a1d-a75f-ea3654f77f5c"],
Cell[CellGroupData[{
Cell[448357, 12724, 151, 3, 27, "Input",ExpressionUUID->"ec29ef21-2bac-478f-b4a9-e1751f168b64"],
Cell[448511, 12729, 187, 3, 26, "Output",ExpressionUUID->"ac8f83d9-0c91-49ba-b023-d82be5b0f99e",
 CellID->671886494]
}, Open  ]],
Cell[CellGroupData[{
Cell[448735, 12737, 157, 3, 27, "Input",ExpressionUUID->"30d30216-18dc-4d89-90ef-f0f90582887c"],
Cell[448895, 12742, 516, 12, 26, "Output",ExpressionUUID->"1d7446f7-5815-4f36-8b31-4bc84990cb50",
 CellID->927288732]
}, Open  ]],
Cell[CellGroupData[{
Cell[449448, 12759, 166, 3, 27, "Input",ExpressionUUID->"325aebc8-1c3b-4e25-a3ae-62faf917301a"],
Cell[449617, 12764, 601, 13, 26, "Output",ExpressionUUID->"c9fa1af9-cab6-4c31-b82a-fbc0f0914464",
 CellID->652892445]
}, Open  ]],
Cell[CellGroupData[{
Cell[450255, 12782, 193, 4, 27, "Input",ExpressionUUID->"8de04d3a-5ca4-47ae-8ef3-87ed5685e24e"],
Cell[450451, 12788, 560, 11, 26, "Output",ExpressionUUID->"916ea33a-4ee0-48db-b622-b0b015a67b2d",
 CellID->1669370592]
}, Open  ]],
Cell[CellGroupData[{
Cell[451048, 12804, 157, 3, 27, "Input",ExpressionUUID->"909ff47a-feb3-49a9-9fcb-f832d17797de"],
Cell[451208, 12809, 560, 14, 26, "Output",ExpressionUUID->"ca4b4e19-05c6-407d-8e66-253b4e4167b5",
 CellID->237691689]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[451817, 12829, 131, 0, 40, "MathCaption",ExpressionUUID->"684bdef6-822b-477a-8bf3-54d8fbcc0ec1"],
Cell[CellGroupData[{
Cell[451973, 12833, 143, 3, 27, "Input",ExpressionUUID->"ca31c5f4-8ea5-457c-96ef-72abed459a54"],
Cell[CellGroupData[{
Cell[452141, 12840, 212, 3, 43, "Print",ExpressionUUID->"58195580-63ce-46a6-9406-193b5711864f",
 CellTags->"Info1163728981383-9181623",
 CellID->564633603],
Cell[452356, 12845, 1818, 55, 193, "Print",ExpressionUUID->"0c3eff53-b245-4530-b9f2-53af8ab16efe",
 CellTags->"Info1163728981383-9181623",
 CellID->66729528]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[454259, 12909, 684, 12, 54, "Section",ExpressionUUID->"f4f44c18-1ef4-4eff-98cd-53767a44bc77",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1339531280],
Cell[CellGroupData[{
Cell[454968, 12925, 97, 0, 39, "Subsection",ExpressionUUID->"6e665f91-497f-409e-85c8-983c34cca83a"],
Cell[455068, 12927, 298, 6, 40, "Text",ExpressionUUID->"b6150745-e0ba-469a-bd7c-7fa6677052d8"],
Cell[455369, 12935, 1213, 33, 92, "DefinitionBox",ExpressionUUID->"77b34b2e-6302-4f68-8cb8-b9dec729d00a"],
Cell[456585, 12970, 93, 0, 23, "Text",ExpressionUUID->"52e4347c-1f61-48cc-96cf-1d5af49db5ed"],
Cell[CellGroupData[{
Cell[456703, 12974, 419, 10, 72, "MathCaption",ExpressionUUID->"cb266cfd-184e-44cb-b1ab-3d2cd4c92074"],
Cell[CellGroupData[{
Cell[457147, 12988, 254, 7, 27, "Input",ExpressionUUID->"e909f5d5-4eb4-4e66-aba4-6c821b279003"],
Cell[457404, 12997, 718, 15, 22, "Print",ExpressionUUID->"1e41444b-094e-4b97-964c-4d443f6a9be6",
 CellID->1083110202]
}, Open  ]],
Cell[CellGroupData[{
Cell[458159, 13017, 223, 6, 27, "Input",ExpressionUUID->"23ddea17-1aa7-4ccd-8c2c-534254357ebd"],
Cell[458385, 13025, 667, 13, 22, "Print",ExpressionUUID->"7622fc00-a7ba-483e-9758-0f9fdc824475",
 CellID->1761773042]
}, Open  ]],
Cell[CellGroupData[{
Cell[459089, 13043, 198, 5, 27, "Input",ExpressionUUID->"bc8594f3-abaa-44d1-b2d7-49f49663aa87"],
Cell[459290, 13050, 627, 12, 22, "Print",ExpressionUUID->"e8c8137e-392b-4e08-b7cf-be74ce60110d",
 CellID->614992058]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[459966, 13068, 205, 5, 40, "MathCaption",ExpressionUUID->"6986e7f3-fc49-4f57-b7fd-35cc6fd8a5c9"],
Cell[CellGroupData[{
Cell[460196, 13077, 193, 5, 27, "Input",ExpressionUUID->"3b7640fa-e6ef-4593-8ddb-ab5073518a94"],
Cell[460392, 13084, 608, 12, 22, "Print",ExpressionUUID->"1d47e5f9-33a7-4449-9f5e-7ff8aff31037",
 CellID->332725795]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[461049, 13102, 369, 12, 56, "MathCaption",ExpressionUUID->"f22202b1-e4a5-4b3d-a847-f93f27bfc2ab"],
Cell[CellGroupData[{
Cell[461443, 13118, 317, 9, 27, "Input",ExpressionUUID->"20af02d0-e91b-4822-904e-334ddfcd005d"],
Cell[461763, 13129, 700, 14, 22, "Print",ExpressionUUID->"4b4d64c5-3907-4145-9d39-5a9fecca207e",
 CellID->75019777]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[462512, 13149, 439, 11, 40, "Text",ExpressionUUID->"a932613c-c10f-409d-bce5-8b63a0006442"],
Cell[CellGroupData[{
Cell[462976, 13164, 127, 3, 27, "Input",ExpressionUUID->"3325b290-4960-4c9d-bcc0-7bee99df921f"],
Cell[CellGroupData[{
Cell[463128, 13171, 198, 3, 43, "Print",ExpressionUUID->"51354a4a-940d-40e8-8e3e-9b1c29871105",
 CellTags->"Info1223728981397-9181623",
 CellID->284491250],
Cell[463329, 13176, 3840, 117, 296, "Print",ExpressionUUID->"14242c38-c27d-43a7-955c-6a18caddad3f",
 CellTags->"Info1223728981397-9181623",
 CellID->1145840528]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[467230, 13300, 164, 2, 40, "MathCaption",ExpressionUUID->"5bec7d37-3441-42aa-ac77-fe2c5ff31cb9"],
Cell[CellGroupData[{
Cell[467419, 13306, 147, 3, 27, "Input",ExpressionUUID->"44a1d28c-a478-43de-b726-175b2cbe502b"],
Cell[467569, 13311, 551, 12, 22, "Print",ExpressionUUID->"04722cc1-71bb-48de-bba7-38ed39adda76",
 CellID->1655891447]
}, Open  ]],
Cell[CellGroupData[{
Cell[468157, 13328, 127, 3, 27, "Input",ExpressionUUID->"372ddd7d-2650-4c09-9477-f0fcfeae250d"],
Cell[468287, 13333, 387, 8, 22, "Message",ExpressionUUID->"45eed1e1-1aaa-479b-8f10-034e7d87737f",
 CellID->1326968256]
}, Open  ]]
}, Open  ]],
Cell[468701, 13345, 326, 6, 40, "Text",ExpressionUUID->"eac2a4fa-aaf4-46e1-8eb8-fac4551dd604"],
Cell[CellGroupData[{
Cell[469052, 13355, 562, 22, 56, "MathCaption",ExpressionUUID->"43e5cec0-c28c-457a-b300-2ba1c20f74e9"],
Cell[CellGroupData[{
Cell[469639, 13381, 223, 6, 27, "Input",ExpressionUUID->"d9bacb74-5924-467a-9802-2a4ff6b729a6"],
Cell[469865, 13389, 666, 13, 22, "Print",ExpressionUUID->"1203a021-fd2d-4d39-872b-2692e1646fc9",
 CellID->737804497]
}, Open  ]],
Cell[CellGroupData[{
Cell[470568, 13407, 127, 3, 27, "Input",ExpressionUUID->"2a6b19c0-33cf-4265-968a-23e42c6a4d2c"],
Cell[CellGroupData[{
Cell[470720, 13414, 198, 3, 43, "Print",ExpressionUUID->"125cac5c-4025-4a7f-a7b8-a2538f16f51f",
 CellTags->"Info1263728981397-9181623",
 CellID->709603301],
Cell[470921, 13419, 3150, 97, 296, "Print",ExpressionUUID->"493ff12b-20ae-40f5-99b4-5f0b62a6615a",
 CellTags->"Info1263728981397-9181623",
 CellID->1325988116]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[474120, 13522, 114, 2, 27, "Input",ExpressionUUID->"b7df524a-d99c-4041-8abb-9cbaca17e715"],
Cell[474237, 13526, 547, 12, 22, "Print",ExpressionUUID->"25395c65-cc6f-46bd-8ab8-6a451c4d2ffc",
 CellID->1444460693]
}, Open  ]]
}, Open  ]],
Cell[474811, 13542, 307, 6, 40, "Text",ExpressionUUID->"c8c17be7-a18e-43f2-8cd3-4b7a909f5c69"]
}, Closed]],
Cell[CellGroupData[{
Cell[475155, 13553, 97, 0, 39, "Subsection",ExpressionUUID->"46b8adf0-3b18-4471-8061-37060ea25fd1"],
Cell[475255, 13555, 911, 26, 57, "Text",ExpressionUUID->"a8f8d757-5dbf-417d-8a23-8240bd97d0c1"],
Cell[CellGroupData[{
Cell[476191, 13585, 347, 9, 56, "MathCaption",ExpressionUUID->"9e6f37b7-0fbd-4a21-8e30-f2652596eaaa"],
Cell[CellGroupData[{
Cell[476563, 13598, 387, 12, 27, "Input",ExpressionUUID->"cebd0320-5f16-4328-9816-e722b7b75061"],
Cell[476953, 13612, 3283, 95, 32, "Output",ExpressionUUID->"71ec79e4-14b4-4f43-a8b2-63acb13d242d",
 CellID->713664600]
}, Open  ]],
Cell[CellGroupData[{
Cell[480273, 13712, 145, 3, 27, "Input",ExpressionUUID->"697f1067-67f7-4b49-82b5-bad6dacbd53e"],
Cell[480421, 13717, 207, 3, 26, "Output",ExpressionUUID->"e2de2219-b328-4620-b812-354bc0a1597c",
 CellID->864763705]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[480677, 13726, 216, 5, 40, "MathCaption",ExpressionUUID->"14262219-47b3-426f-a774-55ed1891688f"],
Cell[CellGroupData[{
Cell[480918, 13735, 147, 3, 27, "Input",ExpressionUUID->"b9f3d553-d9ac-4b3b-b24e-8396dd824f22"],
Cell[481068, 13740, 277, 7, 31, "Output",ExpressionUUID->"bc22edc7-46e3-45e7-b207-5cc70217bf74",
 CellID->2061750686]
}, Open  ]]
}, Open  ]],
Cell[481372, 13751, 392, 7, 40, "Text",ExpressionUUID->"2ddfb04f-79b2-4628-8a54-be7dd6c5d127"],
Cell[481767, 13760, 479, 12, 41, "DefinitionBox",ExpressionUUID->"1fa6c783-fc39-42cb-a9e6-0fadd5dabb13"],
Cell[482249, 13774, 95, 0, 23, "Text",ExpressionUUID->"dfc0d1ae-5874-420d-81a8-7f54fde2895f"],
Cell[CellGroupData[{
Cell[482369, 13778, 360, 12, 56, "MathCaption",ExpressionUUID->"a35238a9-4a66-4702-9e53-5dc85bd52dc6"],
Cell[CellGroupData[{
Cell[482754, 13794, 314, 8, 27, "Input",ExpressionUUID->"34e02ad1-a042-44c9-9e4c-81a5b7cb5bcc"],
Cell[483071, 13804, 691, 14, 22, "Print",ExpressionUUID->"3240a1b8-1453-41c1-9b58-db31ae5c1d0e",
 CellID->47651610]
}, Open  ]],
Cell[CellGroupData[{
Cell[483799, 13823, 180, 5, 27, "Input",ExpressionUUID->"d9ac9c61-3d48-4166-8427-2d96affd6e96"],
Cell[483982, 13830, 1092, 31, 32, "Output",ExpressionUUID->"3d862ef4-be35-4b67-80cc-e593b9070726",
 CellID->2016692586]
}, Open  ]],
Cell[CellGroupData[{
Cell[485111, 13866, 145, 3, 27, "Input",ExpressionUUID->"069da075-c092-4052-a1f2-7d90e01fe546"],
Cell[485259, 13871, 180, 3, 26, "Output",ExpressionUUID->"fb4f208e-488c-48ac-8ab1-7a871566ffea",
 CellID->817892301]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[485488, 13880, 125, 0, 40, "MathCaption",ExpressionUUID->"dc0a0313-a72b-46d0-9678-ef1987cf6b04"],
Cell[485616, 13882, 292, 7, 27, "Input",ExpressionUUID->"f4f3cf21-8ae1-400e-8d8c-071acd584d55"],
Cell[CellGroupData[{
Cell[485933, 13893, 271, 6, 27, "Input",ExpressionUUID->"9db0fada-cdf3-40eb-b024-f57f8c6e6539"],
Cell[486207, 13901, 633, 12, 22, "Print",ExpressionUUID->"335e854b-29c4-4062-b84b-0e61fbaab3fc",
 CellID->273078460]
}, Open  ]],
Cell[CellGroupData[{
Cell[486877, 13918, 141, 3, 27, "Input",ExpressionUUID->"f957743c-546a-43ae-a38c-a06061a88412"],
Cell[487021, 13923, 1064, 31, 32, "Output",ExpressionUUID->"08818c66-c99a-48ca-b9ac-3e5c24b6a929",
 CellID->457123581]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[488134, 13960, 149, 2, 40, "MathCaption",ExpressionUUID->"a7509d5c-c4bf-4d58-88f5-9155a82b7505"],
Cell[CellGroupData[{
Cell[488308, 13966, 198, 5, 27, "Input",ExpressionUUID->"e4a72725-8685-4ace-a4f0-ff911667b4e2"],
Cell[488509, 13973, 197, 3, 26, "Output",ExpressionUUID->"f2898a4f-b588-42d6-95e9-4d5a9d8fb805",
 CellID->1327020548]
}, Open  ]],
Cell[CellGroupData[{
Cell[488743, 13981, 141, 3, 27, "Input",ExpressionUUID->"addd8d6b-1be4-43e2-be3c-d488149b96e7"],
Cell[488887, 13986, 1079, 31, 32, "Output",ExpressionUUID->"031a0caf-12e7-4ec4-b794-3660389bac1c",
 CellID->1616676141]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[490015, 14023, 1131, 28, 168, "MathCaption",ExpressionUUID->"87db2a4b-67fc-4759-a55e-3009a584e9df"],
Cell[CellGroupData[{
Cell[491171, 14055, 187, 3, 27, "Input",ExpressionUUID->"9410297d-59b2-4490-a6df-4f3c20857999"],
Cell[491361, 14060, 187, 3, 26, "Output",ExpressionUUID->"2439ebe4-7ad2-441a-93f8-178a2e3caa29",
 CellID->1512799943]
}, Open  ]],
Cell[CellGroupData[{
Cell[491585, 14068, 209, 4, 27, "Input",ExpressionUUID->"393998c1-3a41-44f3-a7ad-22def2da8c61"],
Cell[491797, 14074, 1079, 31, 32, "Output",ExpressionUUID->"b2fed207-b25e-4660-9772-2e601c5da022",
 CellID->1569745168]
}, Open  ]],
Cell[CellGroupData[{
Cell[492913, 14110, 143, 3, 27, "Input",ExpressionUUID->"e67ac65b-ec58-4095-80dd-11bf04d39008"],
Cell[493059, 14115, 2486, 64, 91, "Output",ExpressionUUID->"0f371ab1-ed8f-49ce-b193-3141a4b7f69d",
 CellID->870435181]
}, Open  ]],
Cell[CellGroupData[{
Cell[495582, 14184, 227, 4, 27, "Input",ExpressionUUID->"506b43dd-d44b-42ab-a324-bb96123030a9"],
Cell[495812, 14190, 186, 3, 26, "Output",ExpressionUUID->"43b9f0f4-6917-4366-887c-97c2d0fa29fb",
 CellID->745483703]
}, Open  ]],
Cell[CellGroupData[{
Cell[496035, 14198, 209, 4, 27, "Input",ExpressionUUID->"ba03388a-e27c-4842-8b54-5886190efa55"],
Cell[496247, 14204, 359, 10, 28, "Output",ExpressionUUID->"2b055c64-9192-4499-a02b-d53c578b909a",
 CellID->924999723]
}, Open  ]],
Cell[CellGroupData[{
Cell[496643, 14219, 211, 4, 27, "Input",ExpressionUUID->"f5b63544-f31a-4d9e-9dd2-9df447de8ecf"],
Cell[496857, 14225, 790, 20, 49, "Output",ExpressionUUID->"6bcf6bcd-c4db-4dcb-8fee-bb29a0a7dd4a",
 CellID->836223828]
}, Open  ]],
Cell[497662, 14248, 243, 5, 27, "Input",ExpressionUUID->"0d58476b-aa43-4d92-a113-7f9bffbce3ae"]
}, Open  ]],
Cell[CellGroupData[{
Cell[497942, 14258, 297, 4, 56, "MathCaption",ExpressionUUID->"8fa345cd-27c2-48f9-82e7-c7044c3487d4"],
Cell[CellGroupData[{
Cell[498264, 14266, 299, 9, 29, "Input",ExpressionUUID->"49841c43-6c3c-4ed6-a911-1f859d13a4cb"],
Cell[498566, 14277, 1081, 31, 32, "Output",ExpressionUUID->"ff67b1e0-c66c-4bb9-bfe8-0e6ad18750a1",
 CellID->1280891228]
}, Open  ]],
Cell[CellGroupData[{
Cell[499684, 14313, 145, 3, 27, "Input",ExpressionUUID->"cfccd1c6-a210-40b1-a1f2-e191d5504a3c"],
Cell[499832, 14318, 180, 3, 26, "Output",ExpressionUUID->"1d42a209-0066-4317-8506-ac987e7738c2",
 CellID->1562791468]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[500061, 14327, 87, 0, 40, "MathCaption",ExpressionUUID->"4820e426-b458-4fec-a2cb-fefb100c0a0d"],
Cell[CellGroupData[{
Cell[500173, 14331, 151, 3, 27, "Input",ExpressionUUID->"268521b9-9c50-43e9-a5a4-953628137e2d"],
Cell[500327, 14336, 557, 12, 22, "Print",ExpressionUUID->"e0b88a5b-6ddf-4927-a354-4a04b4b52a3a",
 CellID->2111199866]
}, Open  ]],
Cell[CellGroupData[{
Cell[500921, 14353, 148, 3, 27, "Input",ExpressionUUID->"6ab9ba5c-51fb-45df-8ac5-ad6366316025"],
Cell[501072, 14358, 549, 12, 22, "Print",ExpressionUUID->"1f52acbc-edfe-4289-aa1c-14500633e66d",
 CellID->63848481]
}, Open  ]],
Cell[501636, 14373, 183, 4, 27, "Input",ExpressionUUID->"f371e37b-52dc-43a6-9e13-b1aed7548114"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[501868, 14383, 108, 0, 39, "Subsection",ExpressionUUID->"017626f1-d144-499d-99e2-9e9a8cf6e2f6"],
Cell[501979, 14385, 538, 14, 57, "Text",ExpressionUUID->"e1a6ff6a-6972-41d5-b2c0-8f6244714a7e"],
Cell[CellGroupData[{
Cell[502542, 14403, 194, 5, 40, "MathCaption",ExpressionUUID->"f070f89e-4c73-4430-a12f-a750214df2b3"],
Cell[CellGroupData[{
Cell[502761, 14412, 226, 7, 27, "Input",ExpressionUUID->"fa8c85c3-cc64-4834-b23f-9cd7f4af216e"],
Cell[502990, 14421, 2051, 60, 32, "Output",ExpressionUUID->"f4c9549e-60ca-4461-95c0-1e2d16be9f11",
 CellID->123842069]
}, Open  ]],
Cell[CellGroupData[{
Cell[505078, 14486, 374, 13, 27, "Input",ExpressionUUID->"a02e5133-37f8-4ce1-b18a-7f9d4b09d9c9"],
Cell[505455, 14501, 2063, 60, 32, "Output",ExpressionUUID->"fd0ce06f-9dd5-4d6d-bba0-773ca51b9900",
 CellID->501999740]
}, Open  ]]
}, Open  ]],
Cell[507545, 14565, 514, 12, 57, "Text",ExpressionUUID->"cc580805-3a9b-48f6-9f1c-49130c8d2f70"],
Cell[508062, 14579, 378, 9, 41, "DefinitionBox",ExpressionUUID->"40129e61-34b7-4eed-b175-dd7b2996c8ae"],
Cell[508443, 14590, 107, 0, 23, "Text",ExpressionUUID->"b47508f7-78da-4e01-84bd-9fee7e825728"],
Cell[CellGroupData[{
Cell[508575, 14594, 134, 0, 40, "MathCaption",ExpressionUUID->"befda43b-6f4b-4822-9008-0849e2738b6d"],
Cell[CellGroupData[{
Cell[508734, 14598, 512, 17, 27, "Input",ExpressionUUID->"4956aadf-f9da-4fcb-b9c7-5b0b99bf9fa2"],
Cell[509249, 14617, 428, 9, 22, "Message",ExpressionUUID->"2be0e124-2f12-49a2-8156-7d6047845521",
 CellID->1793864764],
Cell[509680, 14628, 4635, 135, 32, "Output",ExpressionUUID->"d5bcfe7d-3aab-4be6-b30c-a1a6fbba3e61",
 CellID->962168315]
}, Open  ]],
Cell[CellGroupData[{
Cell[514352, 14768, 512, 17, 27, "Input",ExpressionUUID->"01322a7f-0a83-41ff-ac7f-5d6c21a6ddb8"],
Cell[514867, 14787, 427, 9, 22, "Message",ExpressionUUID->"d0656a01-a745-4c40-8c04-cbfad3c5af75",
 CellID->611008295],
Cell[515297, 14798, 427, 9, 22, "Message",ExpressionUUID->"ba717ef6-1131-4b02-acfa-9386fe6852b7",
 CellID->2097097758],
Cell[515727, 14809, 428, 9, 22, "Message",ExpressionUUID->"bbf1aa9a-b438-4ec3-9410-c70092f3ba68",
 CellID->1420404142],
Cell[516158, 14820, 474, 9, 22, "Message",ExpressionUUID->"c411dd95-1b9e-4b8b-8991-0897856f2c60",
 CellID->1304194542],
Cell[516635, 14831, 5356, 155, 32, "Output",ExpressionUUID->"9118eedc-3111-4f41-b860-c22a1798fa23",
 CellID->232082465]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[522040, 14992, 179, 3, 40, "MathCaption",ExpressionUUID->"536827e2-c1d8-4869-8491-d688c27cabf4"],
Cell[CellGroupData[{
Cell[522244, 14999, 296, 9, 27, "Input",ExpressionUUID->"3af3fe6f-bd80-450b-b313-a5e88fe9b355"],
Cell[522543, 15010, 430, 9, 22, "Message",ExpressionUUID->"077c2b58-223e-4f52-8d28-ec7954fee910",
 CellID->779686303]
}, Open  ]],
Cell[CellGroupData[{
Cell[523010, 15024, 342, 11, 27, "Input",ExpressionUUID->"3449fe64-203f-44fa-a558-cb77be406c9c"],
Cell[523355, 15037, 498, 11, 22, "Message",ExpressionUUID->"1f63d481-27fd-4ed1-a644-1ba8f8164e21",
 CellID->242480527]
}, Open  ]]
}, Open  ]],
Cell[523880, 15052, 479, 12, 40, "Text",ExpressionUUID->"5e6710a7-dc09-44b1-8fa2-06705999d232"],
Cell[CellGroupData[{
Cell[524384, 15068, 118, 0, 40, "MathCaption",ExpressionUUID->"305da4d5-e080-41b0-8e15-3d2100d9a205"],
Cell[CellGroupData[{
Cell[524527, 15072, 207, 6, 27, "Input",ExpressionUUID->"42e5d121-21ae-4256-b11c-fe044e3262b7"],
Cell[524737, 15080, 2050, 60, 32, "Output",ExpressionUUID->"1f832a38-7fd8-4c23-992f-44282a76e042",
 CellID->1593738930]
}, Open  ]],
Cell[CellGroupData[{
Cell[526824, 15145, 144, 3, 27, "Input",ExpressionUUID->"acb84b21-d6d6-45ec-96ef-f30ebc1c8529"],
Cell[526971, 15150, 426, 9, 22, "Message",ExpressionUUID->"22466f4f-fd2c-425d-aaec-702be3ba7b60",
 CellID->406667479]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[527446, 15165, 116, 0, 40, "MathCaption",ExpressionUUID->"ca259f97-b295-45a4-996b-160232aef5aa"],
Cell[CellGroupData[{
Cell[527587, 15169, 113, 2, 27, "Input",ExpressionUUID->"34a653ac-1b3f-449a-8729-badfd84b360b"],
Cell[527703, 15173, 174, 3, 26, "Output",ExpressionUUID->"158d8a28-b8ee-4cad-aa8d-86c0729509a0",
 CellID->1567354799]
}, Open  ]],
Cell[527892, 15179, 160, 4, 27, "Input",ExpressionUUID->"844c637f-5953-432c-b8bf-3f04eeb8154b"],
Cell[CellGroupData[{
Cell[528077, 15187, 207, 6, 27, "Input",ExpressionUUID->"6636582f-d60b-4b13-a9e6-0fc4c5341dae"],
Cell[528287, 15195, 428, 9, 22, "Message",ExpressionUUID->"2388bc2d-ab73-442a-a79d-862836bed025",
 CellID->1497692691]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[528764, 15210, 117, 0, 40, "MathCaption",ExpressionUUID->"3aebedc9-4ca5-4337-b640-2bab07ef3774"],
Cell[528884, 15212, 131, 3, 27, "Input",ExpressionUUID->"26ec62f1-c428-4513-8ce6-5b8470b1ff54"],
Cell[CellGroupData[{
Cell[529040, 15219, 207, 6, 27, "Input",ExpressionUUID->"1156007c-1e8b-4b2b-ac7d-b53b586c1164"],
Cell[529250, 15227, 2051, 60, 32, "Output",ExpressionUUID->"c14e54cf-6bd0-4ec8-b99b-b29eb18daa5a",
 CellID->1065231150]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[531350, 15293, 557, 16, 56, "MathCaption",ExpressionUUID->"b6c5c2b8-a08e-4a7f-bae0-f2098ca48b36"],
Cell[CellGroupData[{
Cell[531932, 15313, 219, 4, 27, "Input",ExpressionUUID->"3b7f4ae2-cd4a-4d7b-a804-b6f8929f371e"],
Cell[532154, 15319, 234, 5, 26, "Output",ExpressionUUID->"f191540b-7955-46b8-a972-70276d7e7d9e",
 CellID->1916256658]
}, Open  ]],
Cell[532403, 15327, 326, 9, 27, "Input",ExpressionUUID->"fe93f975-ccb6-4219-88ca-fdfe1dcd00e8"],
Cell[CellGroupData[{
Cell[532754, 15340, 217, 4, 27, "Input",ExpressionUUID->"4174f44f-7bfd-426a-b259-e9f1853fee8f"],
Cell[532974, 15346, 2052, 60, 32, "Output",ExpressionUUID->"3afbc5ff-cbb4-4716-a67d-58a02e723225",
 CellID->713396947]
}, Open  ]],
Cell[CellGroupData[{
Cell[535063, 15411, 250, 5, 27, "Input",ExpressionUUID->"c83ef039-57cb-4f2f-a087-d860f334c746"],
Cell[535316, 15418, 427, 9, 22, "Message",ExpressionUUID->"fc08460c-4b2d-4862-97ec-117ad058d10b",
 CellID->661287028]
}, Open  ]],
Cell[535758, 15430, 145, 3, 27, "Input",ExpressionUUID->"10db3ef5-88a4-4dc2-b2ac-3b432b42e8a5"],
Cell[CellGroupData[{
Cell[535928, 15437, 147, 3, 27, "Input",ExpressionUUID->"0c53c218-9311-4fad-83b2-f6041a1654da"],
Cell[536078, 15442, 549, 12, 22, "Print",ExpressionUUID->"1b056f06-41f5-46d8-b32b-4a3434ab21be",
 CellID->1023206371]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[536688, 15461, 92, 0, 39, "Subsection",ExpressionUUID->"39425ff7-2e00-4f6f-9dc3-2e26809f04cc"],
Cell[536783, 15463, 1092, 20, 137, "Text",ExpressionUUID->"16786579-b120-4641-8f02-b8857c0e200d"],
Cell[537878, 15485, 1400, 31, 143, "DefinitionBox",ExpressionUUID->"453cb721-f0a1-4ea4-8d34-26a216b7be59"],
Cell[539281, 15518, 102, 0, 23, "Text",ExpressionUUID->"819c8d03-ab78-40e2-b22b-81a1753ffb63"],
Cell[CellGroupData[{
Cell[539408, 15522, 1064, 32, 104, "MathCaption",ExpressionUUID->"d2ff414e-6268-45ff-b8e5-d7a6a227c370"],
Cell[CellGroupData[{
Cell[540497, 15558, 222, 6, 27, "Input",ExpressionUUID->"3b28144c-f079-463e-bef9-d1d71dd7d723"],
Cell[540722, 15566, 527, 15, 26, "Output",ExpressionUUID->"86b51675-63c7-4700-af43-c76ddbe9662d",
 CellID->1356062270]
}, Open  ]],
Cell[CellGroupData[{
Cell[541286, 15586, 226, 6, 27, "Input",ExpressionUUID->"b110fb19-8a02-41c4-a9df-484a1974f4ac"],
Cell[541515, 15594, 551, 16, 26, "Output",ExpressionUUID->"0b170118-49a9-4dfe-b528-c58eb4102ee2",
 CellID->94626279]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[542115, 15616, 211, 7, 40, "MathCaption",ExpressionUUID->"283dc12e-c20d-4c3a-9071-0e8b7bd7488e"],
Cell[CellGroupData[{
Cell[542351, 15627, 326, 9, 27, "Input",ExpressionUUID->"2f768cee-963c-41c4-a51e-395c139dda24"],
Cell[542680, 15638, 667, 13, 22, "Print",ExpressionUUID->"169e4780-3f91-4f30-ba4d-3ad8d73c477e",
 CellID->1245972373]
}, Open  ]],
Cell[CellGroupData[{
Cell[543384, 15656, 127, 3, 27, "Input",ExpressionUUID->"8dd2f949-19f4-451c-96f5-37969fe1aa56"],
Cell[CellGroupData[{
Cell[543536, 15663, 198, 3, 43, "Print",ExpressionUUID->"ab3ee4a8-7bf1-41bf-a3b2-652eaf0137c9",
 CellTags->"Info1733728981432-9181623",
 CellID->944631114],
Cell[543737, 15668, 3211, 99, 296, "Print",ExpressionUUID->"579ee2dc-80f0-49f8-9e65-24f5efae300c",
 CellTags->"Info1733728981432-9181623",
 CellID->68530727]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[547009, 15774, 298, 8, 56, "MathCaption",ExpressionUUID->"1617e0b4-1388-459a-be01-55a7eed330b4"],
Cell[CellGroupData[{
Cell[547332, 15786, 413, 12, 27, "Input",ExpressionUUID->"1ef15a81-375b-4a70-be6c-b3b4374cc5da"],
Cell[547748, 15800, 760, 17, 22, "Print",ExpressionUUID->"27d19452-2983-48a0-9792-a6c3c013858f",
 CellID->744693897]
}, Open  ]],
Cell[CellGroupData[{
Cell[548545, 15822, 127, 3, 27, "Input",ExpressionUUID->"1a077c8d-5660-4c49-ad01-4030aa0bf0b8"],
Cell[CellGroupData[{
Cell[548697, 15829, 196, 3, 43, "Print",ExpressionUUID->"a67b04c3-c3c7-4ac1-8a2a-674cc955be23",
 CellTags->"Info1753728981432-9181623",
 CellID->535632163],
Cell[548896, 15834, 3900, 118, 296, "Print",ExpressionUUID->"78dd0d02-24ec-4d49-9b35-f8d81c30bc4b",
 CellTags->"Info1753728981432-9181623",
 CellID->118315317]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[552835, 15957, 658, 19, 57, "Text",ExpressionUUID->"a6abed5b-67e9-4968-b506-10e9f89ff031"],
Cell[553496, 15978, 123, 0, 23, "Text",ExpressionUUID->"e415ec05-ead8-47aa-9cf4-ce605f0a1841"],
Cell[CellGroupData[{
Cell[553644, 15982, 267, 4, 56, "MathCaption",ExpressionUUID->"b2a6ac3f-ae13-4448-adea-85f6ca073e36"],
Cell[CellGroupData[{
Cell[553936, 15990, 508, 15, 27, "Input",ExpressionUUID->"0c245624-a1f5-404a-889e-0815cb126fd1"],
Cell[554447, 16007, 394, 8, 22, "Message",ExpressionUUID->"93b808d8-1d22-4f0b-a682-11b237aa5ea7",
 CellID->2103159344]
}, Open  ]]
}, Open  ]],
Cell[554868, 16019, 219, 3, 23, "Text",ExpressionUUID->"f6d502ca-da3b-422b-9dd4-34a538cb82a2"],
Cell[CellGroupData[{
Cell[555112, 16026, 162, 2, 40, "MathCaption",ExpressionUUID->"69bcd9ca-2ef7-44df-b030-faa030a529ff"],
Cell[CellGroupData[{
Cell[555299, 16032, 342, 10, 27, "Input",ExpressionUUID->"193fa296-2a2f-4655-87ec-91cfd5a8c8a8"],
Cell[555644, 16044, 408, 8, 22, "Message",ExpressionUUID->"2a3d9b51-3e5e-4126-9965-ef065f2714d1",
 CellID->218264988]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[556101, 16058, 134, 0, 40, "MathCaption",ExpressionUUID->"e5709dcc-1cf8-4aa9-8d43-01c6fc204c2c"],
Cell[CellGroupData[{
Cell[556260, 16062, 399, 11, 27, "Input",ExpressionUUID->"f805382a-8507-41e8-9736-3487e5adf1f8"],
Cell[556662, 16075, 688, 14, 22, "Print",ExpressionUUID->"8bfeaef2-2de7-4fd6-a814-87005107fcc7",
 CellID->477584240]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[557399, 16095, 186, 3, 40, "MathCaption",ExpressionUUID->"ef20c19f-bdf0-4443-84dc-bd477feedc20"],
Cell[CellGroupData[{
Cell[557610, 16102, 470, 12, 27, "Input",ExpressionUUID->"578b54ef-1fa1-436f-af19-ac42c58fb1b8"],
Cell[CellGroupData[{
Cell[558105, 16118, 508, 11, 22, "Print",ExpressionUUID->"779ed934-9078-4257-abea-a144b075fd31",
 CellID->1611932841],
Cell[558616, 16131, 957, 19, 22, "Print",ExpressionUUID->"acae08af-1839-4f77-8dc7-978450fb1098",
 CellID->1249051782]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[559622, 16156, 227, 6, 27, "Input",ExpressionUUID->"545b076d-d3be-4b68-8acb-4c0166b5fd8a"],
Cell[CellGroupData[{
Cell[559874, 16166, 548, 12, 22, "Print",ExpressionUUID->"5e14588e-2f8d-49ef-bd33-b9b801e998c8",
 CellID->1160650649],
Cell[560425, 16180, 565, 12, 22, "Print",ExpressionUUID->"54908e96-ff70-48c2-b1d9-ee06bf3ac8bd",
 CellID->1461492042],
Cell[560993, 16194, 548, 12, 22, "Print",ExpressionUUID->"8cd86757-6047-4c48-bbb8-fa507b9c5aa0",
 CellID->378795508]
}, Open  ]]
}, Open  ]],
Cell[561568, 16210, 142, 3, 27, "Input",ExpressionUUID->"426b0a8c-5cea-49d7-b87d-6103d81733c3"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[561759, 16219, 183, 4, 39, "Subsection",ExpressionUUID->"8149cdd7-01de-4e5a-a962-368625084a4c"],
Cell[561945, 16225, 1884, 42, 205, "Text",ExpressionUUID->"e31ffbb3-2a58-491f-8c79-34c63a181061"],
Cell[563832, 16269, 1281, 38, 75, "DefinitionBox",ExpressionUUID->"1f261d1d-2917-4ae1-b6ee-54bc6f595eb0"],
Cell[565116, 16309, 88, 0, 23, "Text",ExpressionUUID->"72ef4872-06ea-48da-b3b6-afb1ab2c255c"],
Cell[CellGroupData[{
Cell[565229, 16313, 235, 8, 40, "MathCaption",ExpressionUUID->"c45cc6ad-6baa-44ca-9d48-154461fc9297"],
Cell[CellGroupData[{
Cell[565489, 16325, 183, 5, 27, "Input",ExpressionUUID->"366abbff-f545-4141-a509-5e0b85de9b47"],
Cell[565675, 16332, 1096, 31, 32, "Output",ExpressionUUID->"cd4a9c43-f473-4fef-8b4a-4a4da31f78a6",
 CellID->732934548]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[566820, 16369, 137, 1, 40, "MathCaption",ExpressionUUID->"0d09ccd5-64ff-4b3f-9b4b-f3315184fe41"],
Cell[CellGroupData[{
Cell[566982, 16374, 216, 6, 27, "Input",ExpressionUUID->"7dc92267-4c32-4279-a993-0e54963e03a1"],
Cell[567201, 16382, 433, 9, 22, "Message",ExpressionUUID->"6a3b8160-081e-41d8-a69e-dfd42390d676",
 CellID->464749044],
Cell[567637, 16393, 1433, 41, 32, "Output",ExpressionUUID->"73175e5d-beaf-4346-8221-18586e9f4390",
 CellID->639158369]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[569119, 16440, 228, 6, 40, "MathCaption",ExpressionUUID->"157e4412-5a39-43b4-a985-82ca3495540e"],
Cell[CellGroupData[{
Cell[569372, 16450, 183, 5, 27, "Input",ExpressionUUID->"755d651c-c673-41b5-8f5c-80a7583a391b"],
Cell[569558, 16457, 170, 3, 26, "Output",ExpressionUUID->"bf619c5f-bb3f-4f3e-8adf-3aadc847ecf3",
 CellID->621505489]
}, Open  ]],
Cell[CellGroupData[{
Cell[569765, 16465, 279, 9, 27, "Input",ExpressionUUID->"6784a59f-6539-4a7b-bf2b-2d5c4c65213d"],
Cell[570047, 16476, 1098, 31, 32, "Output",ExpressionUUID->"3d056001-eccf-4097-b2d5-2bfd63fce57f",
 CellID->282294615]
}, Open  ]],
Cell[CellGroupData[{
Cell[571182, 16512, 250, 8, 27, "Input",ExpressionUUID->"343a025f-5253-43dc-9409-35f5abf9ddb0"],
Cell[571435, 16522, 1059, 31, 32, "Output",ExpressionUUID->"67fedc8b-b7b8-438b-b291-02765d56a23f",
 CellID->734218082]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[572543, 16559, 221, 5, 40, "MathCaption",ExpressionUUID->"726a0338-d3d6-4ba2-810a-f4507b70c42a"],
Cell[CellGroupData[{
Cell[572789, 16568, 275, 9, 27, "Input",ExpressionUUID->"a847ef3d-b428-44b6-8e14-9ffb14882973"],
Cell[573067, 16579, 1085, 31, 32, "Output",ExpressionUUID->"b307c2f2-4cd9-4554-aed8-73e61afd5207",
 CellID->562268512]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[574201, 16616, 248, 6, 40, "MathCaption",ExpressionUUID->"5dc6d362-d4d8-4cc4-818a-4fe61da0798d"],
Cell[CellGroupData[{
Cell[574474, 16626, 193, 5, 27, "Input",ExpressionUUID->"71ce46af-1159-4ac5-85eb-571ad624f945"],
Cell[574670, 16633, 598, 11, 22, "Message",ExpressionUUID->"8fe1e60f-c66f-4ea8-8c49-61ae816d53fa",
 CellID->1118834384]
}, Open  ]]
}, Open  ]],
Cell[575295, 16648, 399, 8, 40, "Text",ExpressionUUID->"a4ed2797-8ad1-4985-b7ea-46784be6a640"],
Cell[CellGroupData[{
Cell[575719, 16660, 158, 2, 40, "MathCaption",ExpressionUUID->"98862426-6ea5-4f99-81ff-4a54255b1f23"],
Cell[CellGroupData[{
Cell[575902, 16666, 223, 6, 27, "Input",ExpressionUUID->"1c3da9f7-cd50-46bd-b11f-9d4d8584573f"],
Cell[576128, 16674, 1152, 32, 32, "Output",ExpressionUUID->"d868bf2b-aff7-40e2-bbf1-9e3db75aadc8",
 CellID->882678756]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[577329, 16712, 294, 4, 56, "MathCaption",ExpressionUUID->"e15c1a03-1540-4b6b-b08d-72c240c353f8"],
Cell[CellGroupData[{
Cell[577648, 16720, 578, 18, 27, "Input",ExpressionUUID->"e12c9aee-b83d-4156-a0ff-03a9cbff5b0d"],
Cell[578229, 16740, 3433, 94, 32, "Output",ExpressionUUID->"f53f8d93-7578-4196-bcc5-0cf42b58e1f7",
 CellID->364818616]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[581711, 16840, 175, 3, 40, "MathCaption",ExpressionUUID->"72cbd1b8-6a92-4701-b5bc-d0c4cc2f4075"],
Cell[CellGroupData[{
Cell[581911, 16847, 262, 7, 27, "Input",ExpressionUUID->"98ef51f8-6244-49af-a06a-656ae2311db1"],
Cell[582176, 16856, 172, 3, 26, "Output",ExpressionUUID->"210705cc-2b56-43d1-a859-0ac1785174df",
 CellID->440666927]
}, Open  ]],
Cell[CellGroupData[{
Cell[582385, 16864, 262, 7, 27, "Input",ExpressionUUID->"9dd30a31-f8ea-4d1f-850b-861ccf17b282"],
Cell[582650, 16873, 1153, 32, 32, "Output",ExpressionUUID->"c554b6b6-5dfa-44e8-b4e7-10c6016bae22",
 CellID->1765427233]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[583852, 16911, 204, 3, 56, "MathCaption",ExpressionUUID->"3b3bc965-2ee0-495d-be14-a74b91891f73"],
Cell[CellGroupData[{
Cell[584081, 16918, 262, 7, 27, "Input",ExpressionUUID->"0021ce72-5cab-4323-85fb-230a79a5d574"],
Cell[584346, 16927, 171, 3, 26, "Output",ExpressionUUID->"839bb3e3-92a8-4c0b-9bd2-5d144a5e3219",
 CellID->1901438079]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[584566, 16936, 167, 3, 40, "MathCaption",ExpressionUUID->"0b0ba427-fb05-4c18-9116-1fdc8f0ef43e"],
Cell[CellGroupData[{
Cell[584758, 16943, 265, 7, 27, "Input",ExpressionUUID->"7e41b45b-70d7-4105-94bd-5f73ee803904"],
Cell[585026, 16952, 4213, 119, 32, "Output",ExpressionUUID->"ee0cc355-b587-47ed-be74-c069785507dc",
 CellID->765642746]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[589288, 17077, 113, 0, 40, "MathCaption",ExpressionUUID->"25d26999-d31f-4671-be2c-11c244552727"],
Cell[CellGroupData[{
Cell[589426, 17081, 218, 6, 27, "Input",ExpressionUUID->"3d9ab215-f7d9-4e34-9a2a-3c2af8cdc9bf"],
Cell[589647, 17089, 183, 3, 26, "Output",ExpressionUUID->"e116bce3-18ed-4cb6-9378-2281b5458da8",
 CellID->435619198]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[589879, 17098, 156, 2, 40, "MathCaption",ExpressionUUID->"28251a08-e568-4e50-881a-657ad1f6af19"],
Cell[CellGroupData[{
Cell[590060, 17104, 361, 11, 27, "Input",ExpressionUUID->"09a691f4-571b-48ec-9bad-578ffb55b9b5"],
Cell[590424, 17117, 1204, 33, 32, "Output",ExpressionUUID->"14fc77f9-3179-4615-b67c-5c8c28d7c5af",
 CellID->680108163]
}, Open  ]],
Cell[CellGroupData[{
Cell[591665, 17155, 442, 13, 27, "Input",ExpressionUUID->"16151d9c-97cf-49e7-bad8-bdc4a3b1176d"],
Cell[592110, 17170, 1258, 33, 32, "Output",ExpressionUUID->"91962f4e-9b68-4ac4-831d-f1487261b83c",
 CellID->1669926207]
}, Open  ]],
Cell[CellGroupData[{
Cell[593405, 17208, 443, 13, 27, "Input",ExpressionUUID->"10c5c03d-e424-4675-823e-44b6aaca6fe9"],
Cell[593851, 17223, 6959, 186, 32, "Output",ExpressionUUID->"a7cda249-47e2-48fb-a5aa-7a0add85a3e9",
 CellID->985978406]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[600859, 17415, 108, 0, 40, "MathCaption",ExpressionUUID->"5a9f661f-51fa-4b51-af0a-47294192773f"],
Cell[CellGroupData[{
Cell[600992, 17419, 148, 3, 27, "Input",ExpressionUUID->"e8e57c9a-eddb-49a3-90e6-c5b2161937b4"],
Cell[601143, 17424, 65884, 1688, 80, "Output",ExpressionUUID->"a362d989-54ec-4920-aa0b-6ebcfadb8aef",
 CellID->1355249390]
}, Open  ]],
Cell[CellGroupData[{
Cell[667064, 19117, 571, 17, 27, "Input",ExpressionUUID->"d463af76-bfd5-436a-8d5a-7ebf8b92edc6"],
Cell[667638, 19136, 107377, 2737, 128, "Output",ExpressionUUID->"cf1b929f-8cc6-4925-8133-f37dbd7acf27",
 CellID->1464937648]
}, Open  ]],
Cell[CellGroupData[{
Cell[775052, 21878, 138, 3, 27, "Input",ExpressionUUID->"fa5ae39d-4fb6-4938-a829-4150b5c4ade0"],
Cell[775193, 21883, 170, 3, 26, "Output",ExpressionUUID->"29b782fa-86be-469d-bbef-006de51552bd",
 CellID->307037081]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[775412, 21892, 195, 3, 40, "MathCaption",ExpressionUUID->"e2e35eb6-332c-4f4e-8d88-9eb3057813e4"],
Cell[CellGroupData[{
Cell[775632, 21899, 345, 9, 27, "Input",ExpressionUUID->"76ecc4d4-f804-4230-9c93-e82e2a1c80c9"],
Cell[775980, 21910, 692, 13, 22, "Print",ExpressionUUID->"43303e10-cfc4-41c9-bb86-f92a0c2797fa",
 CellID->2116500273]
}, Open  ]],
Cell[CellGroupData[{
Cell[776709, 21928, 346, 10, 27, "Input",ExpressionUUID->"d1e0b314-4b3b-4963-b343-05780c425cd9"],
Cell[777058, 21940, 1163, 32, 32, "Output",ExpressionUUID->"6ff29cb9-2753-410d-9f64-3fdf16552c45",
 CellID->1100959403]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[778270, 21978, 109, 0, 40, "MathCaption",ExpressionUUID->"86d20593-4145-416e-91d0-37390eb6c98c"],
Cell[CellGroupData[{
Cell[778404, 21982, 394, 12, 27, "Input",ExpressionUUID->"a3a4479d-daa2-4923-8612-a8a2a7906341"],
Cell[778801, 21996, 20378, 562, 32, "Output",ExpressionUUID->"2c26fab6-a96e-462f-8340-145e0ee3c7f2",
 CellID->1105583669]
}, Open  ]],
Cell[CellGroupData[{
Cell[799216, 22563, 138, 3, 27, "Input",ExpressionUUID->"66652138-f1d8-4903-b1f6-e3d5da7825d4"],
Cell[799357, 22568, 6041, 172, 32, "Output",ExpressionUUID->"be4c3da9-440e-4b89-a4a9-598f0b07f3d6",
 CellID->1516211039]
}, Open  ]],
Cell[CellGroupData[{
Cell[805435, 22745, 143, 3, 27, "Input",ExpressionUUID->"8791f0ad-41ea-4796-9cd5-530d3a6e8512"],
Cell[805581, 22750, 1160, 32, 32, "Output",ExpressionUUID->"41c15348-4ae4-4a05-b22a-62dc81984341",
 CellID->917977171]
}, Open  ]],
Cell[CellGroupData[{
Cell[806778, 22787, 147, 3, 27, "Input",ExpressionUUID->"47b8b85c-2421-4c20-b56d-adc94d7c0e33"],
Cell[806928, 22792, 548, 12, 22, "Print",ExpressionUUID->"c36347f0-a8d6-43d2-bc01-bc2234b6976a",
 CellID->919792949]
}, Open  ]]
}, Open  ]],
Cell[807503, 22808, 289, 6, 40, "Text",ExpressionUUID->"bc3c20b0-8a9d-47cf-863d-7c09c054085c"],
Cell[CellGroupData[{
Cell[807817, 22818, 261, 6, 40, "MathCaption",ExpressionUUID->"699aa6cc-c06d-4108-99d4-60028dfac6c1"],
Cell[CellGroupData[{
Cell[808103, 22828, 513, 14, 27, "Input",ExpressionUUID->"f76963c5-a5ff-4f92-bfd8-1ddbb8ba850c"],
Cell[808619, 22844, 1439, 38, 32, "Output",ExpressionUUID->"9a3eba78-774d-450c-84e9-4c10988d96c9",
 CellID->1372802737]
}, Open  ]],
Cell[CellGroupData[{
Cell[810095, 22887, 212, 4, 27, "Input",ExpressionUUID->"06139917-d37f-46b5-8d9b-afe2b7303981"],
Cell[810310, 22893, 34248, 943, 66, "Output",ExpressionUUID->"47795137-627c-42fb-8617-858465a90876",
 CellID->131297391]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[844619, 23843, 101, 0, 39, "Subsection",ExpressionUUID->"cee63d08-2f11-43e6-a4bc-9b8db80f5a44"],
Cell[844723, 23845, 985, 24, 91, "Text",ExpressionUUID->"dae88e9f-94c7-466a-b8a0-987970ef7265"],
Cell[845711, 23871, 386, 9, 41, "DefinitionBox",ExpressionUUID->"1ff6fab4-5050-41f8-b2f5-d1b26ae8b70f"],
Cell[846100, 23882, 91, 0, 23, "Text",ExpressionUUID->"4882a813-adcd-480b-9aaf-e30e26795a3b"],
Cell[CellGroupData[{
Cell[846216, 23886, 129, 0, 40, "MathCaption",ExpressionUUID->"092e5e45-7da0-4d07-a504-9698540dd06e"],
Cell[CellGroupData[{
Cell[846370, 23890, 226, 7, 27, "Input",ExpressionUUID->"7120c12d-c856-431e-a321-b43f633d8a94"],
Cell[846599, 23899, 2051, 60, 32, "Output",ExpressionUUID->"423213de-2dc4-40e8-ab9a-46d3a6588a72",
 CellID->923251014]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[848699, 23965, 176, 3, 40, "MathCaption",ExpressionUUID->"05a19e42-7531-406c-a3df-d2c31fa5bbe0"],
Cell[CellGroupData[{
Cell[848900, 23972, 247, 7, 27, "Input",ExpressionUUID->"6ccf66fd-e2b0-48b9-ba72-1da7e66739ab"],
Cell[849150, 23981, 1129, 33, 32, "Output",ExpressionUUID->"0bd88b6e-44c7-4e12-800c-d09d143242fd",
 CellID->1510442723]
}, Open  ]],
Cell[CellGroupData[{
Cell[850316, 24019, 144, 3, 27, "Input",ExpressionUUID->"84a69629-63e3-467d-94d2-cffbfeedea99"],
Cell[850463, 24024, 427, 9, 22, "Message",ExpressionUUID->"29eda692-6e50-417b-8e3d-5f0fdce60f07",
 CellID->606654274],
Cell[850893, 24035, 1129, 33, 32, "Output",ExpressionUUID->"8824e6ab-bda5-48d9-83ef-c50c9dcea625",
 CellID->1933510809]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[852071, 24074, 209, 3, 56, "MathCaption",ExpressionUUID->"a4ebcff7-4407-4381-9825-209611ee90b4"],
Cell[CellGroupData[{
Cell[852305, 24081, 226, 6, 27, "Input",ExpressionUUID->"d3df5355-2307-473c-a818-af8f0aea3651"],
Cell[852534, 24089, 1124, 33, 32, "Output",ExpressionUUID->"3a8634b7-4bb6-4d11-a99f-57a34efeded1",
 CellID->793620459]
}, Open  ]],
Cell[CellGroupData[{
Cell[853695, 24127, 144, 3, 27, "Input",ExpressionUUID->"2ad9d6db-cc99-4e6b-8ffb-62caf9d74a9a"],
Cell[853842, 24132, 425, 9, 22, "Message",ExpressionUUID->"8de8e366-bc8a-45ba-afd4-5208958efdb1",
 CellID->358122896],
Cell[854270, 24143, 1512, 45, 32, "Output",ExpressionUUID->"e0a7ee8c-6c48-4550-b190-f04da1e89f06",
 CellID->1722334178]
}, Open  ]],
Cell[CellGroupData[{
Cell[855819, 24193, 247, 7, 27, "Input",ExpressionUUID->"b8f584b9-a518-4ae7-9e9f-96139f65c0a9"],
Cell[856069, 24202, 1129, 33, 32, "Output",ExpressionUUID->"4189be2a-852b-41da-8ad0-f4180255af00",
 CellID->1338451733]
}, Open  ]],
Cell[CellGroupData[{
Cell[857235, 24240, 144, 3, 27, "Input",ExpressionUUID->"366ee27f-f619-4c08-9667-7cc58346bbd9"],
Cell[857382, 24245, 427, 9, 22, "Message",ExpressionUUID->"65dcb444-608b-4a8b-9a2d-070df1c03e24",
 CellID->467074206],
Cell[857812, 24256, 1129, 33, 32, "Output",ExpressionUUID->"1386a5b6-b773-40cc-bbd2-74c5435b44e9",
 CellID->1063450457]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[858990, 24295, 130, 0, 40, "MathCaption",ExpressionUUID->"c37cbf54-17b1-4660-ac3c-48e8fce6feca"],
Cell[CellGroupData[{
Cell[859145, 24299, 198, 5, 27, "Input",ExpressionUUID->"39c61074-f698-423c-b63e-fbdf19c13967"],
Cell[859346, 24306, 626, 12, 22, "Print",ExpressionUUID->"81c123f3-d8b8-4bf8-b2bb-b85a5ee6e89a",
 CellID->2052920465]
}, Open  ]],
Cell[CellGroupData[{
Cell[860009, 24323, 350, 11, 27, "Input",ExpressionUUID->"b8e3ceec-38d5-4fb2-aa3b-889cd257c8ee"],
Cell[860362, 24336, 2222, 65, 32, "Output",ExpressionUUID->"21651e12-eb84-4a6a-bb97-4d85a7b5065e",
 CellID->1961982553]
}, Open  ]],
Cell[CellGroupData[{
Cell[862621, 24406, 147, 3, 27, "Input",ExpressionUUID->"3716cbf8-85e4-4d70-8284-d73d1f695200"],
Cell[862771, 24411, 548, 12, 22, "Print",ExpressionUUID->"db977d41-2503-4ec2-86d6-5c720a6f0297",
 CellID->429368586]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[863368, 24429, 230, 5, 40, "MathCaption",ExpressionUUID->"52eb0882-e869-4c24-8ba9-47a038d4b5b3"],
Cell[CellGroupData[{
Cell[863623, 24438, 282, 8, 27, "Input",ExpressionUUID->"67cd41f0-8aeb-4eff-a6f6-699f8c96edb4"],
Cell[863908, 24448, 173, 3, 26, "Output",ExpressionUUID->"908c2215-0161-4142-a94a-02eba1119bcc",
 CellID->852770982]
}, Open  ]]
}, Open  ]],
Cell[864108, 24455, 553, 16, 57, "Text",ExpressionUUID->"6608ce0a-e5ee-491a-8663-9165f031c619"],
Cell[CellGroupData[{
Cell[864686, 24475, 219, 5, 40, "MathCaption",ExpressionUUID->"742c9853-0aee-4193-aeeb-203e626202cf"],
Cell[CellGroupData[{
Cell[864930, 24484, 297, 8, 47, "Input",ExpressionUUID->"d278930a-0155-492b-9fcb-723efc59c857"],
Cell[865230, 24494, 1127, 33, 32, "Output",ExpressionUUID->"dc01e41e-9864-4c36-bcc1-39a215b4ed7c",
 CellID->216034396]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[866406, 24533, 267, 6, 40, "MathCaption",ExpressionUUID->"41c015fd-f582-4c55-90f0-e6f4af8adef3"],
Cell[CellGroupData[{
Cell[866698, 24543, 281, 6, 27, "Input",ExpressionUUID->"abb6c5fc-d64a-458f-acc4-d056b6fe3fce"],
Cell[866982, 24551, 2052, 60, 32, "Output",ExpressionUUID->"f2107841-ef55-43ad-9162-bc3066ef2b42",
 CellID->1642730142]
}, Open  ]],
Cell[CellGroupData[{
Cell[869071, 24616, 236, 5, 27, "Input",ExpressionUUID->"6047e091-bcef-4164-85c5-fa5638b79224"],
Cell[869310, 24623, 1128, 33, 32, "Output",ExpressionUUID->"ec13ccc4-8d12-4a50-b4e1-b805742c0685",
 CellID->1315730762]
}, Open  ]],
Cell[870453, 24659, 145, 3, 27, "Input",ExpressionUUID->"1dea45a0-20ba-4a4e-8326-86027e224e7d"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[870647, 24668, 95, 0, 39, "Subsection",ExpressionUUID->"5bcbb47f-5dd2-4b4c-a475-129578912a65"],
Cell[870745, 24670, 604, 16, 57, "Text",ExpressionUUID->"cee59409-3b4a-4d59-822f-63cc0832e3e6"],
Cell[871352, 24688, 709, 18, 58, "DefinitionBox",ExpressionUUID->"f77ec4b8-27b1-4d56-9f5b-4d842fa65f7d"],
Cell[872064, 24708, 85, 0, 23, "Text",ExpressionUUID->"2cc6deed-6db7-4c12-b135-12b8060806f1"],
Cell[CellGroupData[{
Cell[872174, 24712, 118, 0, 40, "MathCaption",ExpressionUUID->"d0c96438-ab0b-4de1-847b-cf9aaea23a4a"],
Cell[CellGroupData[{
Cell[872317, 24716, 277, 8, 27, "Input",ExpressionUUID->"3917762c-0856-45e2-8a62-473dccb37ee9"],
Cell[872597, 24726, 774, 17, 22, "Print",ExpressionUUID->"9bfd987d-3644-454e-bd9f-7362b6c6ed41",
 CellID->736429224]
}, Open  ]],
Cell[CellGroupData[{
Cell[873408, 24748, 127, 3, 27, "Input",ExpressionUUID->"9201c517-efba-4712-b236-84adf1bbfe92"],
Cell[CellGroupData[{
Cell[873560, 24755, 199, 3, 43, "Print",ExpressionUUID->"d09061f2-d245-4708-9251-86b611186377",
 CellTags->"Info2273728981469-9181623",
 CellID->1217710107],
Cell[873762, 24760, 2666, 84, 296, "Print",ExpressionUUID->"17855b62-38ea-437b-aaa7-7df201779033",
 CellTags->"Info2273728981469-9181623",
 CellID->534612439]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[876489, 24851, 314, 6, 56, "MathCaption",ExpressionUUID->"2b50a1c0-7f05-4f3c-a83b-e3913ff7ab34"],
Cell[CellGroupData[{
Cell[876828, 24861, 210, 6, 27, "Input",ExpressionUUID->"323759bd-c4a0-4395-8b17-1d69a22c20e1"],
Cell[877041, 24869, 1126, 33, 32, "Output",ExpressionUUID->"342c4704-c948-4176-ae45-511858e7b0e4",
 CellID->159683225]
}, Open  ]],
Cell[CellGroupData[{
Cell[878204, 24907, 479, 16, 27, "Input",ExpressionUUID->"4380e29d-41c7-48a5-af56-629b92f3aab4"],
Cell[878686, 24925, 2269, 65, 32, "Output",ExpressionUUID->"2c7a8f35-cb62-4308-9956-320fe89a669b",
 CellID->1588887472]
}, Open  ]],
Cell[CellGroupData[{
Cell[880992, 24995, 147, 3, 27, "Input",ExpressionUUID->"399acf47-e35f-47a7-b7de-9b5cc6a2ab37"],
Cell[881142, 25000, 548, 12, 22, "Print",ExpressionUUID->"396107e9-6a05-45c0-b1ed-f6579ab9a925",
 CellID->472591459]
}, Open  ]]
}, Open  ]],
Cell[881717, 25016, 388, 9, 40, "Text",ExpressionUUID->"8b831627-c064-4d91-b9df-20c4447657ba"],
Cell[CellGroupData[{
Cell[882130, 25029, 495, 9, 56, "MathCaption",ExpressionUUID->"05837fd2-6990-400e-85af-320fda22566d"],
Cell[882628, 25040, 1491, 46, 104, "Input",ExpressionUUID->"fd427259-6fb3-4f69-a6e1-5b856b707520"],
Cell[CellGroupData[{
Cell[884144, 25090, 371, 9, 27, "Input",ExpressionUUID->"ee6e225e-5363-4f55-ab8e-7e15915a4868"],
Cell[884518, 25101, 1239, 35, 32, "Output",ExpressionUUID->"c2f754ea-eadb-440f-acd9-a9d8b5653437",
 CellID->1450124807]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[885806, 25142, 1269, 41, 108, "Text",ExpressionUUID->"8ac046d8-9cb4-4b92-ba43-d7ee9f63bc36"],
Cell[887078, 25185, 224, 6, 27, "Input",ExpressionUUID->"d840339c-6241-45e3-962e-97933d64631d"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[887351, 25197, 101, 0, 39, "Subsection",ExpressionUUID->"ea33d657-7f0c-4e11-b67d-e460d1a2abaf"],
Cell[887455, 25199, 378, 11, 40, "Text",ExpressionUUID->"eb2c7916-e5c5-4dde-a668-0bc55813358f"],
Cell[887836, 25212, 1631, 43, 109, "DefinitionBox",ExpressionUUID->"159dc468-ea03-412f-b8d2-10faa70ec4a4"],
Cell[889470, 25257, 90, 0, 23, "Text",ExpressionUUID->"687e2a48-aee8-46e3-88a2-b01d2d89ee12"],
Cell[CellGroupData[{
Cell[889585, 25261, 154, 2, 40, "MathCaption",ExpressionUUID->"e31ec0b5-a939-48cb-843b-1ad3f66cb1df"],
Cell[CellGroupData[{
Cell[889764, 25267, 297, 7, 27, "Input",ExpressionUUID->"234e76b2-f59f-4f60-915a-a90ed8a4347c"],
Cell[CellGroupData[{
Cell[890086, 25278, 697, 13, 22, "Print",ExpressionUUID->"5f3cd4d9-0c91-4c79-a027-27aba6f3165d",
 CellID->34886677],
Cell[890786, 25293, 735, 14, 22, "Print",ExpressionUUID->"bf2013e1-2abd-4a10-8895-95a7e3f8d863",
 CellID->77344708]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[891570, 25313, 127, 3, 27, "Input",ExpressionUUID->"46acec2c-8acb-4ee1-864f-be8dc3d9316e"],
Cell[CellGroupData[{
Cell[891722, 25320, 196, 3, 43, "Print",ExpressionUUID->"729ec76a-285d-4e29-a392-3a0b59f569e5",
 CellTags->"Info2393728981479-9181623",
 CellID->704900921],
Cell[891921, 25325, 3255, 102, 330, "Print",ExpressionUUID->"55bb2ae3-f92f-49b6-a6b4-e800ab4bf2c8",
 CellTags->"Info2393728981479-9181623",
 CellID->1749982395]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[895225, 25433, 136, 3, 27, "Input",ExpressionUUID->"3ff8171b-7faa-4e1a-b515-08f53ee376dd"],
Cell[CellGroupData[{
Cell[895386, 25440, 204, 3, 43, "Print",ExpressionUUID->"f4ae4598-eca7-4561-b34c-69b4a8d91214",
 CellTags->"Info2403728981479-9181623",
 CellID->18083453],
Cell[895593, 25445, 3371, 102, 330, "Print",ExpressionUUID->"50884fe2-237c-4232-b241-7b33a9fb2528",
 CellTags->"Info2403728981479-9181623",
 CellID->1133400344]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[899025, 25554, 164, 2, 40, "MathCaption",ExpressionUUID->"4868cfeb-8e11-41dc-82b0-fa75de150ccd"],
Cell[CellGroupData[{
Cell[899214, 25560, 220, 5, 27, "Input",ExpressionUUID->"99cccb93-2181-412a-b0f6-257cc8584006"],
Cell[899437, 25567, 1154, 32, 32, "Output",ExpressionUUID->"eae13a93-1ae9-482b-a738-13895d0fad76",
 CellID->879632718]
}, Open  ]],
Cell[CellGroupData[{
Cell[900628, 25604, 147, 3, 27, "Input",ExpressionUUID->"69d3165b-e826-4dbe-89cb-cbe7731681f8"],
Cell[CellGroupData[{
Cell[900800, 25611, 565, 12, 22, "Print",ExpressionUUID->"7c23daf0-2746-4780-b2b1-8d7343ea062b",
 CellID->709665478],
Cell[901368, 25625, 548, 12, 22, "Print",ExpressionUUID->"12b06724-e87f-4d37-b0a7-6e1be6257e48",
 CellID->223513308]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[901977, 25644, 152, 2, 40, "MathCaption",ExpressionUUID->"57c9d282-d824-46e2-92d7-5f5219f6f197"],
Cell[CellGroupData[{
Cell[902154, 25650, 291, 7, 27, "Input",ExpressionUUID->"f4ae1704-1355-4c19-b0ac-ebc7d50ba99e"],
Cell[902448, 25659, 316, 7, 26, "Output",ExpressionUUID->"bd61ac98-b6ce-4fa7-86c5-37011cb7da88",
 CellID->639268953]
}, Open  ]]
}, Open  ]],
Cell[902791, 25670, 158, 3, 23, "Text",ExpressionUUID->"5f6c124e-74f3-4d15-bcc5-1bc82c7c10f9"],
Cell[CellGroupData[{
Cell[902974, 25677, 123, 0, 40, "MathCaption",ExpressionUUID->"8ee12e9d-f5dc-4751-a553-eba159640d0a"],
Cell[CellGroupData[{
Cell[903122, 25681, 151, 3, 27, "Input",ExpressionUUID->"ba71e427-4a2a-4770-9a44-9e4da06317b9"],
Cell[903276, 25686, 1023, 24, 65, "Output",ExpressionUUID->"d467c0fc-2330-4794-a200-34f5434b5c0e",
 CellID->332607130]
}, Open  ]],
Cell[CellGroupData[{
Cell[904336, 25715, 198, 5, 27, "Input",ExpressionUUID->"0b10d262-d244-44b9-932b-72be1dba2bc6"],
Cell[904537, 25722, 627, 12, 22, "Print",ExpressionUUID->"676b09d7-1fdc-4fb7-a628-0f7f99a861b2",
 CellID->228100769]
}, Open  ]],
Cell[CellGroupData[{
Cell[905201, 25739, 199, 5, 27, "Input",ExpressionUUID->"521a9185-17fa-4e06-8e11-9081d4d46fa2"],
Cell[905403, 25746, 1150, 33, 32, "Output",ExpressionUUID->"3bb42abd-eab7-4305-a2cf-1305b38ec1ea",
 CellID->431314688]
}, Open  ]],
Cell[CellGroupData[{
Cell[906590, 25784, 147, 3, 27, "Input",ExpressionUUID->"ced94c9d-6356-4fdb-8eee-75b5e7c18ab2"],
Cell[906740, 25789, 549, 12, 22, "Print",ExpressionUUID->"a291bb32-be87-4f18-b9aa-77c5980dd45b",
 CellID->1348652591]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[907338, 25807, 252, 4, 56, "MathCaption",ExpressionUUID->"68df0777-34c1-4a2e-b230-8c62e528c97f"],
Cell[CellGroupData[{
Cell[907615, 25815, 249, 6, 27, "Input",ExpressionUUID->"e90deda9-b906-4960-a78f-8434db06178e"],
Cell[CellGroupData[{
Cell[907889, 25825, 627, 12, 22, "Print",ExpressionUUID->"ac33456f-e11c-4ae1-9c50-6a4516561df6",
 CellID->114202852],
Cell[908519, 25839, 646, 12, 22, "Print",ExpressionUUID->"9baa71f6-8bde-42bf-9826-efd354a4580b",
 CellID->1704436793]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[909214, 25857, 199, 5, 27, "Input",ExpressionUUID->"544a2b37-a264-429b-8c03-2e884a1e6c95"],
Cell[909416, 25864, 1167, 33, 32, "Output",ExpressionUUID->"024d95dd-ce4f-4e67-ab15-0e7d50851c2a",
 CellID->1184812618]
}, Open  ]],
Cell[CellGroupData[{
Cell[910620, 25902, 147, 3, 27, "Input",ExpressionUUID->"d5c3ee27-e635-4347-ab22-78bf4f8bf4d3"],
Cell[CellGroupData[{
Cell[910792, 25909, 569, 12, 22, "Print",ExpressionUUID->"320d4b5b-3a58-40a1-b36e-9043996b2719",
 CellID->1677295742],
Cell[911364, 25923, 550, 12, 22, "Print",ExpressionUUID->"f2899bff-c11b-4843-bb66-53e086f9b551",
 CellID->291701625]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[911975, 25942, 151, 2, 40, "MathCaption",ExpressionUUID->"8fb825ea-90cb-4f52-b12f-bd6f233b4ee5"],
Cell[CellGroupData[{
Cell[912151, 25948, 216, 6, 27, "Input",ExpressionUUID->"51615578-25c7-4472-819c-fc996d122322"],
Cell[912370, 25956, 478, 9, 22, "Message",ExpressionUUID->"8398fd3c-ec66-4827-8e80-54be58173303",
 CellID->2071528921]
}, Open  ]],
Cell[CellGroupData[{
Cell[912885, 25970, 265, 6, 27, "Input",ExpressionUUID->"572a8af7-b017-43af-9a35-bcd91804c95d"],
Cell[CellGroupData[{
Cell[913175, 25980, 658, 12, 22, "Print",ExpressionUUID->"dc81dfc3-bbbc-4d3f-8390-40848bfde7be",
 CellID->1181690083],
Cell[913836, 25994, 695, 12, 22, "Print",ExpressionUUID->"a4c429ee-fcf1-494e-bb12-0b9c89159343",
 CellID->736457525]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[914580, 26012, 215, 5, 27, "Input",ExpressionUUID->"e38a45d1-4b2c-4184-8703-2cea9adc08c8"],
Cell[914798, 26019, 1217, 33, 32, "Output",ExpressionUUID->"dec04b48-cfde-4a66-8a02-849b265f337a",
 CellID->1693365014]
}, Open  ]],
Cell[CellGroupData[{
Cell[916052, 26057, 147, 3, 27, "Input",ExpressionUUID->"eece5558-b330-469d-ae57-2da5e8bf9946"],
Cell[CellGroupData[{
Cell[916224, 26064, 566, 12, 22, "Print",ExpressionUUID->"2ba7993c-bfb3-45d9-bbda-466b265efa57",
 CellID->1932500254],
Cell[916793, 26078, 548, 12, 22, "Print",ExpressionUUID->"f0363b49-b156-4437-b21b-920cf0851a5f",
 CellID->989091750]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[917402, 26097, 283, 4, 56, "MathCaption",ExpressionUUID->"94d5af63-1576-430f-96ee-ba46a6d17bb2"],
Cell[CellGroupData[{
Cell[917710, 26105, 393, 10, 27, "Input",ExpressionUUID->"49de26f0-e4b5-465d-bb04-9fcda065a04c"],
Cell[CellGroupData[{
Cell[918128, 26119, 898, 17, 22, "Print",ExpressionUUID->"fa48d62d-7b46-4675-acc0-65b2103e9c80",
 CellID->2141614470],
Cell[919029, 26138, 917, 17, 22, "Print",ExpressionUUID->"1b982c69-2c28-4e35-8298-13f45a5f6191",
 CellID->140989923]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[919995, 26161, 308, 8, 27, "Input",ExpressionUUID->"f2be409f-f389-4683-945f-166b3198ef4f"],
Cell[920306, 26171, 1315, 34, 32, "Output",ExpressionUUID->"91ecfb7a-e9a1-4728-b042-288a0f2f39fd",
 CellID->943173220]
}, Open  ]],
Cell[CellGroupData[{
Cell[921658, 26210, 314, 7, 27, "Input",ExpressionUUID->"0af59942-68cd-4700-b253-5b265ecddbac"],
Cell[921975, 26219, 1318, 34, 32, "Output",ExpressionUUID->"3353dae0-284c-4fdf-b6ae-98619fd6c94d",
 CellID->1335510033]
}, Open  ]],
Cell[CellGroupData[{
Cell[923330, 26258, 354, 9, 27, "Input",ExpressionUUID->"a174b86d-f4e2-4ed3-abd6-365721f1981d"],
Cell[923687, 26269, 815, 14, 22, "Message",ExpressionUUID->"54c70a22-939f-404f-8183-91e1b9716b23",
 CellID->418046187]
}, Open  ]],
Cell[CellGroupData[{
Cell[924539, 26288, 165, 4, 27, "Input",ExpressionUUID->"a8cac00d-fec6-48a4-9511-791a28520dcf"],
Cell[CellGroupData[{
Cell[924729, 26296, 569, 12, 22, "Print",ExpressionUUID->"80d5bd83-3433-4cc4-8ae3-4eed20bfc38d",
 CellID->1683255572],
Cell[925301, 26310, 549, 12, 22, "Print",ExpressionUUID->"469395c4-fc63-4ef3-b2ef-b5022b75d228",
 CellID->1674316673]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[925911, 26329, 354, 5, 72, "MathCaption",ExpressionUUID->"1468d4f7-eedf-43ee-854c-0616d4346519"],
Cell[CellGroupData[{
Cell[926290, 26338, 460, 11, 27, "Input",ExpressionUUID->"91584caa-7b9b-4295-90a2-dbc7dd057173"],
Cell[CellGroupData[{
Cell[926775, 26353, 1042, 19, 22, "Print",ExpressionUUID->"fb5b66ba-32e6-4d91-8d1f-ba55870a25e0",
 CellID->1762096470],
Cell[927820, 26374, 1057, 19, 22, "Print",ExpressionUUID->"be4c0973-0c7e-4c18-a0b5-d63de7433988",
 CellID->1755879160]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[928926, 26399, 374, 9, 27, "Input",ExpressionUUID->"43aa06d0-6176-4176-ae70-09442e01734e"],
Cell[929303, 26410, 1462, 36, 32, "Output",ExpressionUUID->"aa333319-0399-4a6c-aa64-32743ea920ff",
 CellID->340765046]
}, Open  ]],
Cell[CellGroupData[{
Cell[930802, 26451, 147, 3, 27, "Input",ExpressionUUID->"561e6c71-f23b-404f-8fed-141a9cb9aaf5"],
Cell[CellGroupData[{
Cell[930974, 26458, 567, 12, 22, "Print",ExpressionUUID->"577390e9-8641-4691-8755-fec59989f09e",
 CellID->2046725220],
Cell[931544, 26472, 548, 12, 22, "Print",ExpressionUUID->"79c1785d-d182-47de-8b28-6101ba8ba080",
 CellID->917714894]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[932153, 26491, 347, 8, 56, "MathCaption",ExpressionUUID->"3aa3f2c4-eaf9-424f-b420-7f34e6bae3fd"],
Cell[CellGroupData[{
Cell[932525, 26503, 125, 2, 27, "Input",ExpressionUUID->"24820903-b510-4139-a564-1360d91496f7"],
Cell[932653, 26507, 186, 3, 26, "Output",ExpressionUUID->"0612ca62-1934-467e-bbd6-038136d4c3c0",
 CellID->806670704]
}, Open  ]],
Cell[932854, 26513, 142, 3, 27, "Input",ExpressionUUID->"08a20120-5f20-4b41-b7e4-95cc4ebb2320"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[933045, 26522, 121, 0, 39, "Subsection",ExpressionUUID->"85b1e558-0468-4c77-849b-9e1fc7db8f9e"],
Cell[933169, 26524, 1452, 25, 200, "Text",ExpressionUUID->"0d0845de-9f05-4c78-8761-3b37fb35c005"],
Cell[CellGroupData[{
Cell[934646, 26553, 155, 2, 40, "MathCaption",ExpressionUUID->"a673768f-7822-4a7c-bfcc-e01965389728"],
Cell[CellGroupData[{
Cell[934826, 26559, 243, 7, 27, "Input",ExpressionUUID->"66ddc65a-3c82-4c76-9e58-c165170672ba"],
Cell[935072, 26568, 878, 20, 22, "Print",ExpressionUUID->"8a06c4a2-ebca-4f41-82b0-bfe02d4a4f82",
 CellID->432104373]
}, Open  ]],
Cell[CellGroupData[{
Cell[935987, 26593, 127, 3, 27, "Input",ExpressionUUID->"5c0076a7-85aa-4dcd-94d3-ef9a7df51381"],
Cell[CellGroupData[{
Cell[936139, 26600, 197, 3, 43, "Print",ExpressionUUID->"8b37f71b-9bc8-4c6b-a966-48e42cfdf8b6",
 CellTags->"Info2663728981492-9181623",
 CellID->1735986504],
Cell[936339, 26605, 2476, 77, 262, "Print",ExpressionUUID->"bf829cfa-6dbb-4861-8773-7a3aefbd4dba",
 CellTags->"Info2663728981492-9181623",
 CellID->1012344586]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[938876, 26689, 159, 2, 40, "MathCaption",ExpressionUUID->"64da12c3-3e4e-4553-8e4d-699d0b9dd4ba"],
Cell[CellGroupData[{
Cell[939060, 26695, 157, 3, 27, "Input",ExpressionUUID->"11ed7705-3f7a-47e5-a176-de2207f51c74"],
Cell[939220, 26700, 280, 7, 26, "Output",ExpressionUUID->"d6204b88-6155-4d78-a651-9589f277ada0",
 CellID->2110005368]
}, Open  ]],
Cell[CellGroupData[{
Cell[939537, 26712, 180, 4, 27, "Input",ExpressionUUID->"6ed271e6-2d44-4b7f-9aab-7280d936bcc7"],
Cell[939720, 26718, 280, 7, 26, "Output",ExpressionUUID->"dd655c45-e0bf-4bec-84be-2427894cf6ac",
 CellID->1462757065]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[940049, 26731, 126, 0, 40, "MathCaption",ExpressionUUID->"23d0c9d9-168f-4b4a-b6f5-e6472fa42c9f"],
Cell[940178, 26733, 350, 9, 27, "Input",ExpressionUUID->"611f91bc-969d-4a77-bf4b-c0da37358482"],
Cell[CellGroupData[{
Cell[940553, 26746, 180, 4, 27, "Input",ExpressionUUID->"7a70475a-99e6-40bc-bc9b-6ad0244aaed7"],
Cell[940736, 26752, 225, 5, 26, "Output",ExpressionUUID->"56378924-4e4f-4098-a161-52e53ad92f97",
 CellID->165200599]
}, Open  ]],
Cell[CellGroupData[{
Cell[940998, 26762, 209, 5, 27, "Input",ExpressionUUID->"c4949cd5-aee2-45f2-afb9-e3010500ed40"],
Cell[941210, 26769, 253, 6, 26, "Output",ExpressionUUID->"8491ec76-bc40-40c1-b887-5bccc693370b",
 CellID->1977595469]
}, Open  ]],
Cell[CellGroupData[{
Cell[941500, 26780, 229, 5, 27, "Input",ExpressionUUID->"b8ef38d1-a6ac-418c-bf18-3d7aaddbffb6"],
Cell[941732, 26787, 272, 6, 26, "Output",ExpressionUUID->"cd5ebad8-825d-4576-85ed-dd14ce98fbcf",
 CellID->842642864]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[942053, 26799, 87, 0, 40, "MathCaption",ExpressionUUID->"2e9d9408-7cc3-455e-b0c2-f382085b96da"],
Cell[CellGroupData[{
Cell[942165, 26803, 147, 3, 27, "Input",ExpressionUUID->"d701fdee-130c-48f5-bfa9-7a7f34e3dcd2"],
Cell[942315, 26808, 548, 12, 22, "Print",ExpressionUUID->"f03f6b90-fbe8-43bd-a0d9-33862eede570",
 CellID->1791446930]
}, Open  ]],
Cell[942878, 26823, 145, 3, 27, "Input",ExpressionUUID->"6ca2900a-4c7f-4594-98ef-910509f5c712"]
}, Open  ]],
Cell[943038, 26829, 314, 7, 40, "Text",ExpressionUUID->"e71a8723-3263-4bf9-ab2d-372e7953824e"]
}, Closed]],
Cell[CellGroupData[{
Cell[943389, 26841, 91, 0, 39, "Subsection",ExpressionUUID->"f35a0194-1aa1-46e0-a2f9-7b2af9e581fa"],
Cell[943483, 26843, 631, 13, 86, "Text",ExpressionUUID->"f3ad4d17-e94d-4713-a7e2-d0f8e9f17935"],
Cell[CellGroupData[{
Cell[944139, 26860, 95, 0, 40, "MathCaption",ExpressionUUID->"399f8699-c3cc-44b3-a2e0-64736fb7b37f"],
Cell[CellGroupData[{
Cell[944259, 26864, 117, 2, 27, "Input",ExpressionUUID->"f6fe5140-dea6-415a-a0c0-a1156ed9cb59"],
Cell[944379, 26868, 236, 5, 26, "Output",ExpressionUUID->"37d12554-7e46-4621-9f53-f7d8b1e0793e",
 CellID->1584742292]
}, Open  ]],
Cell[CellGroupData[{
Cell[944652, 26878, 161, 4, 27, "Input",ExpressionUUID->"470f14bf-c925-403c-9055-0f1752574e97"],
Cell[CellGroupData[{
Cell[944838, 26886, 551, 12, 22, "Print",ExpressionUUID->"bbcf6719-8d9a-4f1d-b4bc-b363589df69f",
 CellID->1671511550],
Cell[945392, 26900, 549, 12, 22, "Print",ExpressionUUID->"104d60d0-4c44-495e-ae8d-424a96a8547e",
 CellID->1861034443],
Cell[945944, 26914, 548, 12, 22, "Print",ExpressionUUID->"0603fbf9-f633-4904-b585-769e13837363",
 CellID->306564946],
Cell[946495, 26928, 548, 12, 22, "Print",ExpressionUUID->"5523e2f4-1511-42f4-ae27-ebcc4277cf2d",
 CellID->623778269]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[947104, 26947, 393, 6, 56, "MathCaption",ExpressionUUID->"d25fb823-4acb-427a-92a2-b9b7348d71c7"],
Cell[CellGroupData[{
Cell[947522, 26957, 182, 4, 27, "Input",ExpressionUUID->"12570b0e-504c-4e69-8a89-5d8f2024ca5e"],
Cell[947707, 26963, 497, 10, 22, "Message",ExpressionUUID->"0cea6aed-94ef-44f7-85d0-83ba02f74565",
 CellID->1973686076]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[948253, 26979, 116, 0, 40, "MathCaption",ExpressionUUID->"ce746ea1-b3f1-444b-9406-bed61163aebe"],
Cell[CellGroupData[{
Cell[948394, 26983, 153, 3, 27, "Input",ExpressionUUID->"ad50ddba-0e40-4de4-a129-4c668b7df482"],
Cell[CellGroupData[{
Cell[948572, 26990, 751, 18, 22, "Print",ExpressionUUID->"cdf7af0e-9284-423d-ae18-328219308bd4",
 CellID->1709467770],
Cell[949326, 27010, 702, 18, 22, "Print",ExpressionUUID->"2c9db178-cfb9-447e-aef8-8e22dab60246",
 CellID->1988580824]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[950089, 27035, 112, 0, 40, "MathCaption",ExpressionUUID->"e03a86d6-80c1-4fb6-a2d0-5dd23d7c5510"],
Cell[CellGroupData[{
Cell[950226, 27039, 150, 3, 27, "Input",ExpressionUUID->"7bf774fe-644d-4207-93c7-e7de4221511b"],
Cell[CellGroupData[{
Cell[950401, 27046, 724, 18, 22, "Print",ExpressionUUID->"ab857645-62ec-40cc-9e08-a5bb4f418850",
 CellID->789910707],
Cell[951128, 27066, 558, 12, 22, "Print",ExpressionUUID->"da4651e4-a813-4773-b8dd-6eefcd6914ad",
 CellID->285346795]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[951735, 27084, 150, 3, 27, "Input",ExpressionUUID->"d06634ee-92e5-474e-a3f1-a4c25d21fa70"],
Cell[CellGroupData[{
Cell[951910, 27091, 726, 18, 22, "Print",ExpressionUUID->"a48e0cb1-6e38-467e-8c27-fabfae5bc59c",
 CellID->492599814],
Cell[952639, 27111, 558, 12, 22, "Print",ExpressionUUID->"83f604d3-0a00-436d-9637-383d92a0f904",
 CellID->575614713]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[953258, 27130, 96, 0, 40, "MathCaption",ExpressionUUID->"ba377a37-bed2-41da-b013-76cb48ccc3a2"],
Cell[CellGroupData[{
Cell[953379, 27134, 153, 3, 27, "Input",ExpressionUUID->"3fea85b3-4646-4287-afc8-ebe1f75c26bc"],
Cell[953535, 27139, 702, 18, 22, "Print",ExpressionUUID->"311ad552-186b-43a2-88e4-c83072ef668b",
 CellID->2090135303]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[954286, 27163, 105, 0, 40, "MathCaption",ExpressionUUID->"d925ef7f-fc5c-49c6-86bf-1742ab62a8fc"],
Cell[CellGroupData[{
Cell[954416, 27167, 134, 3, 27, "Input",ExpressionUUID->"560455fe-a648-4382-b925-9c27aa88e4f2"],
Cell[954553, 27172, 212, 3, 43, "Print",ExpressionUUID->"944a01b4-5f58-4e66-a631-2beb8f017d35",
 CellTags->"Info2823728981498-9181623",
 CellID->366539895]
}, Open  ]]
}, Open  ]],
Cell[954792, 27179, 153, 2, 23, "Text",ExpressionUUID->"37ab47cf-a37b-466f-872f-1400001b4863"]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[954994, 27187, 724, 12, 54, "Section",ExpressionUUID->"e07ebc50-8899-4702-bbbe-f4d58d3ca5cc",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1330943070],
Cell[CellGroupData[{
Cell[955743, 27203, 88, 0, 39, "Subsection",ExpressionUUID->"27be6501-297f-4c01-a07b-643e49a7ce19"],
Cell[955834, 27205, 2127, 49, 234, "Text",ExpressionUUID->"0fc444e8-cd2c-4c05-a2ac-fa11fa4967e1"],
Cell[957964, 27256, 826, 22, 58, "DefinitionBox",ExpressionUUID->"a4853172-1344-4f00-a08e-1d0f978a6c6b"],
Cell[958793, 27280, 95, 0, 23, "Text",ExpressionUUID->"ae327cb7-0805-4d8f-a39a-36c11b296e46"],
Cell[CellGroupData[{
Cell[958913, 27284, 120, 0, 40, "MathCaption",ExpressionUUID->"86ecd186-81dc-4d3f-b9be-378a3ba87e08"],
Cell[CellGroupData[{
Cell[959058, 27288, 321, 8, 27, "Input",ExpressionUUID->"9528b0dc-60db-4da0-a1f8-7bad3c6faab4"],
Cell[CellGroupData[{
Cell[959404, 27300, 595, 12, 22, "Print",ExpressionUUID->"ba18926e-63ab-46ea-bd00-bd15d6d59c87",
 CellID->985284070],
Cell[960002, 27314, 605, 12, 22, "Print",ExpressionUUID->"d0c3d622-7ff1-4eef-8750-2f106d0d466b",
 CellID->422962473]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[960656, 27332, 408, 12, 27, "Input",ExpressionUUID->"41ba7f86-db7d-4177-bd90-7743bf5e9519"],
Cell[961067, 27346, 762, 17, 22, "Print",ExpressionUUID->"f96744ac-c004-4ac5-96b7-00ca5d603efd",
 CellID->559412141]
}, Open  ]],
Cell[CellGroupData[{
Cell[961866, 27368, 223, 6, 27, "Input",ExpressionUUID->"362271f3-1082-489b-a812-99f1a5ed85d0"],
Cell[962092, 27376, 668, 13, 22, "Print",ExpressionUUID->"aa088c12-f943-4acb-8d22-b9661ba00c6e",
 CellID->356447221]
}, Open  ]],
Cell[CellGroupData[{
Cell[962797, 27394, 198, 5, 27, "Input",ExpressionUUID->"f8413efe-4e9d-4480-9a74-52c66952f1c0"],
Cell[962998, 27401, 626, 12, 22, "Print",ExpressionUUID->"4573a8e6-df04-4851-82a3-cf83084f370b",
 CellID->1476551864]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[963673, 27419, 227, 5, 40, "MathCaption",ExpressionUUID->"4123c80d-f87b-4825-ba5d-1a693af48dbf"],
Cell[CellGroupData[{
Cell[963925, 27428, 279, 8, 27, "Input",ExpressionUUID->"f5c1bc7a-1e28-48db-a0ad-afefe3521473"],
Cell[964207, 27438, 1157, 32, 32, "Output",ExpressionUUID->"ced005e8-4ea8-4813-ba0d-e3039a0f927d",
 CellID->86153516]
}, Open  ]],
Cell[CellGroupData[{
Cell[965401, 27475, 420, 14, 27, "Input",ExpressionUUID->"97fc91d0-dbe0-4991-a802-ba00273cb21f"],
Cell[965824, 27491, 169, 3, 26, "Output",ExpressionUUID->"3178cd84-80d9-4780-840e-6bf9e9c0b296",
 CellID->578165334]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[966042, 27500, 172, 3, 40, "MathCaption",ExpressionUUID->"8b0e0cc6-8067-48ad-9c1d-7adeb12e7a26"],
Cell[CellGroupData[{
Cell[966239, 27507, 418, 14, 27, "Input",ExpressionUUID->"1cf5e7ce-0395-4517-83f9-621717c4b8ed"],
Cell[966660, 27523, 2062, 60, 32, "Output",ExpressionUUID->"e1be20c2-0433-4c46-98a2-64c9f1bfe34f",
 CellID->2008749510]
}, Open  ]]
}, Open  ]],
Cell[968749, 27587, 245, 6, 23, "Text",ExpressionUUID->"51d3dc2d-1f73-404e-991c-653ffde168ef"],
Cell[968997, 27595, 755, 18, 75, "DefinitionBox",ExpressionUUID->"4ff7dfea-d01e-40d7-85e9-92d66ceb3358"],
Cell[969755, 27615, 164, 5, 23, "Text",ExpressionUUID->"03367d2e-3fd7-494d-8711-ac60fa5e1ecd"],
Cell[CellGroupData[{
Cell[969944, 27624, 486, 12, 72, "MathCaption",ExpressionUUID->"e92c0676-9ea4-404d-bd5a-65b63e06582e"],
Cell[CellGroupData[{
Cell[970455, 27640, 222, 5, 27, "Input",ExpressionUUID->"a467c9e8-2413-4cb1-b28b-884dd2a2a75a"],
Cell[970680, 27647, 408, 9, 26, "Output",ExpressionUUID->"25b1023d-bb6c-438f-9caa-2de7678b0675",
 CellID->462600083]
}, Open  ]],
Cell[CellGroupData[{
Cell[971125, 27661, 349, 10, 27, "Input",ExpressionUUID->"df9663b7-68f8-44d2-a111-46dabfd5ed75"],
Cell[971477, 27673, 455, 10, 22, "Message",ExpressionUUID->"22d685c6-29dd-496a-a846-88629da5daf3",
 CellID->1028792841],
Cell[971935, 27685, 444, 8, 22, "Print",ExpressionUUID->"21d29b50-cad9-47ce-9775-2d17f3c4e66c",
 CellID->1530436690],
Cell[972382, 27695, 456, 10, 22, "Message",ExpressionUUID->"5a44fe39-3612-4fd7-8ad9-058c1dd8927c",
 CellID->1637808711],
Cell[972841, 27707, 450, 9, 22, "Print",ExpressionUUID->"bde5710b-d9f1-4b13-abb0-0cd668c66e91",
 CellID->205618531],
Cell[973294, 27718, 2123, 61, 32, "Output",ExpressionUUID->"abf6fbcc-a539-4231-a624-bcdffc4488af",
 CellID->529462494]
}, Open  ]],
Cell[CellGroupData[{
Cell[975454, 27784, 345, 10, 27, "Input",ExpressionUUID->"14565ad5-3787-4fce-9650-543266d8ea7c"],
Cell[CellGroupData[{
Cell[975824, 27798, 261, 4, 22, "Print",ExpressionUUID->"b4c221b4-4f7b-49b5-825f-3eb0399bdc57",
 CellID->1651387137],
Cell[976088, 27804, 25308, 544, 35, "Print",ExpressionUUID->"a5d885ca-9c7e-497a-8ce6-362cb8054b45",
 CellID->784957731],
Cell[1001399, 28350, 4637, 123, 50, "Print",ExpressionUUID->"72cfd32d-c6da-44a4-bdde-c77589d2745f",
 CellID->792541396],
Cell[1006039, 28475, 943, 25, 22, "Print",ExpressionUUID->"b5d01558-7dbf-4b52-9c2f-69d5a39abea7",
 CellID->1549663878],
Cell[1006985, 28502, 951, 26, 22, "Print",ExpressionUUID->"9cdd7a1f-ca81-400b-8ec5-35dd8938e42d",
 CellID->519126730],
Cell[1007939, 28530, 601, 17, 22, "Print",ExpressionUUID->"73872195-1695-4c2d-944b-9dd319f008cc",
 CellID->1884376996],
Cell[1008543, 28549, 396, 9, 22, "Print",ExpressionUUID->"8abfab5c-f343-4fc8-bc05-99323e6feeaa",
 CellID->1782829463],
Cell[1008942, 28560, 542, 12, 22, "Print",ExpressionUUID->"b355a96f-982f-41b3-bf61-2dd347fb5592",
 CellID->175458050],
Cell[1009487, 28574, 648, 19, 22, "Print",ExpressionUUID->"df60d226-111b-45b1-b068-9ea8b89941c3",
 CellID->428399019],
Cell[1010138, 28595, 742, 18, 22, "Print",ExpressionUUID->"0420c512-3394-48dc-8702-8e61367d8569",
 CellID->1737127589],
Cell[1010883, 28615, 782, 20, 22, "Print",ExpressionUUID->"e07cf8cb-c6be-4ad1-8c40-7d7603eb69b9",
 CellID->863207358],
Cell[1011668, 28637, 416, 9, 22, "Print",ExpressionUUID->"00b32bf5-5862-481e-a1e8-737e83b8d2bd",
 CellID->1674030421],
Cell[1012087, 28648, 2280, 60, 61, "Print",ExpressionUUID->"811cbc84-89b8-47c2-834c-513d5ed30d01",
 CellID->1246026400],
Cell[1014370, 28710, 551, 14, 22, "Print",ExpressionUUID->"111d435f-fc9d-45a5-a1dc-ac15dc27932d",
 CellID->1852662903],
Cell[1014924, 28726, 1016, 27, 22, "Print",ExpressionUUID->"4dab8de6-07e4-4a95-97c0-93a22154ba95",
 CellID->1340012868],
Cell[1015943, 28755, 369, 8, 22, "Print",ExpressionUUID->"bd738d51-aad3-4b89-9423-ccd2c495057b",
 CellID->935266387],
Cell[1016315, 28765, 2849, 75, 28, "Print",ExpressionUUID->"253fc659-50a2-4a7a-9b8c-553024b42fd1",
 CellID->1589353916]
}, Open  ]],
Cell[1019179, 28843, 2126, 61, 32, "Output",ExpressionUUID->"4050f88c-e41b-4557-b8fc-8d8745644cff",
 CellID->1869664826]
}, Open  ]],
Cell[CellGroupData[{
Cell[1021342, 28909, 350, 10, 27, "Input",ExpressionUUID->"c2889709-49ef-47a9-b5ec-0758cfa9dcc7"],
Cell[1021695, 28921, 445, 10, 22, "Message",ExpressionUUID->"5231426f-eb63-4329-9d37-b841d7c96db4",
 CellID->1476007678],
Cell[1022143, 28933, 445, 10, 22, "Message",ExpressionUUID->"398913a0-557a-4c49-a6a9-5555bf9edf36",
 CellID->1677437649],
Cell[CellGroupData[{
Cell[1022613, 28947, 401, 9, 22, "Print",ExpressionUUID->"243e4bdd-4ba1-4f47-aba7-57169bcbe72d",
 CellID->529248927],
Cell[1023017, 28958, 526, 12, 22, "Print",ExpressionUUID->"5ba65501-e9ee-4ed1-9859-6830a916f654",
 CellID->209450215],
Cell[1023546, 28972, 512, 12, 22, "Print",ExpressionUUID->"7e63b828-dfea-4162-bf2f-d96e5462c44c",
 CellID->504066072],
Cell[1024061, 28986, 390, 9, 22, "Print",ExpressionUUID->"79bb82e2-cd27-4d2f-b5f9-acab31769755",
 CellID->15484689],
Cell[1024454, 28997, 511, 10, 22, "Print",ExpressionUUID->"b0317c6e-0c90-4e7a-bdea-e38cc1b05486",
 CellID->405646736],
Cell[1024968, 29009, 388, 8, 22, "Print",ExpressionUUID->"48bca0d5-9802-4c64-ba31-e76d2c4200c0",
 CellID->1591596374],
Cell[1025359, 29019, 376, 9, 22, "Print",ExpressionUUID->"ef60a04c-4c31-4dbb-8f15-1a1cb250ff99",
 CellID->903950786],
Cell[1025738, 29030, 398, 9, 22, "Print",ExpressionUUID->"2543264f-f02a-4ffe-a651-7f74d0666bc8",
 CellID->1401541424],
Cell[1026139, 29041, 514, 10, 22, "Print",ExpressionUUID->"099001d9-f82f-40fa-85f0-d9195f40a54c",
 CellID->2106072872],
Cell[1026656, 29053, 431, 9, 22, "Print",ExpressionUUID->"1e4ea39c-d562-4ef2-80f7-4f2716e92ce7",
 CellID->1941289850],
Cell[1027090, 29064, 377, 9, 22, "Print",ExpressionUUID->"34b134f6-bf7e-4afa-81c1-e492ec2b84cb",
 CellID->539460079],
Cell[1027470, 29075, 430, 10, 22, "Print",ExpressionUUID->"211bb0f0-f78d-435e-9159-aac2286bb8f9",
 CellID->29590119],
Cell[1027903, 29087, 460, 10, 22, "Print",ExpressionUUID->"53b396f8-0c09-49e9-a603-bf9b74b25fec",
 CellID->1252407576],
Cell[1028366, 29099, 522, 10, 22, "Print",ExpressionUUID->"0762422c-c3b0-46ac-a106-92ba42cd9c3d",
 CellID->1726905585],
Cell[1028891, 29111, 387, 8, 22, "Print",ExpressionUUID->"7580a49c-1430-491e-b3eb-9b90235b551d",
 CellID->860677890],
Cell[1029281, 29121, 735, 16, 22, "Print",ExpressionUUID->"f112c2a3-6ac7-4570-a84d-2999ecb9dfd9",
 CellID->2110934685],
Cell[1030019, 29139, 473, 11, 22, "Print",ExpressionUUID->"ac3becae-d0e3-4690-8314-4a6f8354addd",
 CellID->2049402355],
Cell[1030495, 29152, 373, 8, 22, "Print",ExpressionUUID->"c486b8f9-27ab-4811-8235-6ad32de39d0d",
 CellID->815166252],
Cell[1030871, 29162, 972, 25, 22, "Print",ExpressionUUID->"7a46fb19-db54-4da5-8808-4d927d052c16",
 CellID->1701013151],
Cell[1031846, 29189, 968, 25, 22, "Print",ExpressionUUID->"76466d00-b884-4bc5-87a7-362c43a15fba",
 CellID->1625719993],
Cell[1032817, 29216, 753, 19, 22, "Print",ExpressionUUID->"8a8d9258-eec0-4e1c-b3a7-410245c8bed1",
 CellID->1041481570],
Cell[1033573, 29237, 411, 9, 22, "Print",ExpressionUUID->"fffe2cc6-9187-41c1-b4af-f983868935b9",
 CellID->41540293],
Cell[1033987, 29248, 511, 10, 22, "Print",ExpressionUUID->"d3a48a63-f401-4915-aaf0-9cc34acc2ae2",
 CellID->235519176],
Cell[1034501, 29260, 415, 9, 22, "Print",ExpressionUUID->"b59bf584-0952-41c4-995e-c2f427b4b464",
 CellID->558429611],
Cell[1034919, 29271, 380, 9, 22, "Print",ExpressionUUID->"69a067ff-37e7-4652-8197-80a4f86ea1c9",
 CellID->1022934687],
Cell[1035302, 29282, 404, 9, 22, "Print",ExpressionUUID->"ee1da9f2-76a8-4137-af25-b7368925e3a2",
 CellID->1856690174],
Cell[1035709, 29293, 433, 9, 22, "Print",ExpressionUUID->"6ee556bb-f0e1-40c9-b2b8-8cfe61ec716e",
 CellID->306879861],
Cell[1036145, 29304, 520, 10, 22, "Print",ExpressionUUID->"9f36bd8d-d7c6-49d0-882f-9841a5e5e556",
 CellID->1621587380],
Cell[1036668, 29316, 387, 8, 22, "Print",ExpressionUUID->"cbc9c307-277c-4be2-8a60-802c4435fd56",
 CellID->304131583],
Cell[1037058, 29326, 757, 17, 22, "Print",ExpressionUUID->"7502fc87-ab52-4a88-8b43-bc17a28358bb",
 CellID->528435621],
Cell[1037818, 29345, 490, 11, 22, "Print",ExpressionUUID->"f80b17f0-b380-49ef-92f0-dbb7b47f75aa",
 CellID->1572601802],
Cell[1038311, 29358, 372, 8, 22, "Print",ExpressionUUID->"952832eb-83c8-4f92-be1b-23ecaaa6f3ad",
 CellID->1637242585],
Cell[1038686, 29368, 741, 19, 22, "Print",ExpressionUUID->"c67cd88e-7046-437e-89f0-00726e4bfd9b",
 CellID->1428157084],
Cell[1039430, 29389, 738, 19, 22, "Print",ExpressionUUID->"94abc7bb-3c81-4087-9620-3a17c282f029",
 CellID->611388269],
Cell[1040171, 29410, 561, 14, 22, "Print",ExpressionUUID->"4df27b2a-8e93-4a04-807e-ff01b4a9c115",
 CellID->2040869835],
Cell[1040735, 29426, 400, 9, 22, "Print",ExpressionUUID->"3e9aebfd-fe05-4fb6-91a7-5582772a8b13",
 CellID->1290506961],
Cell[1041138, 29437, 577, 14, 22, "Print",ExpressionUUID->"ead7db8e-237b-4454-9847-3964fb40bec2",
 CellID->2113142474],
Cell[1041718, 29453, 554, 14, 22, "Print",ExpressionUUID->"69b8e4de-12a4-4cae-b11b-bdfb1d9ec0c2",
 CellID->501006432],
Cell[1042275, 29469, 758, 20, 22, "Print",ExpressionUUID->"1825dc49-cf5f-4434-bcdf-11174e901eae",
 CellID->131722737],
Cell[1043036, 29491, 382, 9, 22, "Print",ExpressionUUID->"3a294659-444b-4ad1-93a7-4f6b566f055b",
 CellID->1955884349],
Cell[1043421, 29502, 380, 9, 22, "Print",ExpressionUUID->"0d882871-b7bf-4ea4-9754-c435d5f566f8",
 CellID->1016069014],
Cell[1043804, 29513, 503, 12, 22, "Print",ExpressionUUID->"b2ce695d-e9fc-494f-aca8-4fa7821a388d",
 CellID->281957354],
Cell[1044310, 29527, 539, 14, 22, "Print",ExpressionUUID->"df2bec7b-a199-47d8-9aca-c78c89744293",
 CellID->1508323396],
Cell[1044852, 29543, 392, 9, 22, "Print",ExpressionUUID->"03d3b3e1-cafe-4f90-9b11-5dcff5108a03",
 CellID->1494347384],
Cell[1045247, 29554, 506, 13, 22, "Print",ExpressionUUID->"ffc00945-9edb-4198-a6a3-71753593521c",
 CellID->675231223],
Cell[1045756, 29569, 441, 8, 22, "Print",ExpressionUUID->"71e759a6-8198-4788-bb41-bf0b9ac0dda7",
 CellID->1307574213],
Cell[1046200, 29579, 393, 8, 22, "Print",ExpressionUUID->"aaf95585-25ed-46a3-a17e-b42b524fe27d",
 CellID->1199638140],
Cell[1046596, 29589, 849, 21, 22, "Print",ExpressionUUID->"9cbcaa9c-a651-456f-b2fc-634451b0bd1c",
 CellID->1741740944],
Cell[1047448, 29612, 478, 10, 22, "Print",ExpressionUUID->"bc3643dc-d875-46b4-906d-d1a97ad90ec8",
 CellID->1090453551],
Cell[1047929, 29624, 565, 14, 22, "Print",ExpressionUUID->"7c706081-31f7-4b43-bb31-cca6de4c6750",
 CellID->627388159],
Cell[1048497, 29640, 590, 14, 22, "Print",ExpressionUUID->"7bfec5e3-9c3f-401e-8d4d-c7368053d47a",
 CellID->368081543],
Cell[1049090, 29656, 420, 9, 22, "Print",ExpressionUUID->"45bcdb9b-7b7a-474c-9cec-a4223302a81a",
 CellID->1593589841],
Cell[1049513, 29667, 383, 8, 22, "Print",ExpressionUUID->"4a900eb7-3558-44ee-a994-71855baf7b95",
 CellID->48412303],
Cell[1049899, 29677, 497, 11, 22, "Print",ExpressionUUID->"81e1563a-a93e-4c94-9e62-a12da6e6b395",
 CellID->1916757955],
Cell[1050399, 29690, 425, 8, 22, "Print",ExpressionUUID->"5c94b873-4f4d-4818-9bbf-f07c33934a74",
 CellID->548761038],
Cell[1050827, 29700, 748, 20, 22, "Print",ExpressionUUID->"5f539234-b2da-4e92-b955-27f1e0bcbd3c",
 CellID->1731985448],
Cell[1051578, 29722, 506, 13, 22, "Print",ExpressionUUID->"62a1b77f-24e9-41d8-b866-ba601938fc58",
 CellID->338589451],
Cell[1052087, 29737, 773, 18, 22, "Print",ExpressionUUID->"d547ba90-005c-4cd6-99df-f8833b8c100f",
 CellID->1988722054],
Cell[1052863, 29757, 490, 9, 22, "Print",ExpressionUUID->"6720eb18-d867-4266-a7d7-0c9587507852",
 CellID->1016079670],
Cell[1053356, 29768, 503, 11, 22, "Print",ExpressionUUID->"3ba0ddcc-e929-46d8-913b-a6a800586033",
 CellID->461052746],
Cell[1053862, 29781, 323, 7, 22, "Print",ExpressionUUID->"18d4fdfa-9312-4ca3-8437-b16bb79100b9",
 CellID->792503646],
Cell[1054188, 29790, 323, 8, 22, "Print",ExpressionUUID->"ed9eae32-bae6-4184-bbeb-3f35c116ed1f",
 CellID->915199628],
Cell[1054514, 29800, 684, 17, 22, "Print",ExpressionUUID->"3b2a13c1-7676-45da-9bac-cfcae48e9f3f",
 CellID->1482267982],
Cell[1055201, 29819, 561, 13, 22, "Print",ExpressionUUID->"edbb2532-4ccc-4238-ae0e-4361b8af4d33",
 CellID->1667668274],
Cell[1055765, 29834, 378, 8, 22, "Print",ExpressionUUID->"e07c81b0-6d0c-4680-9a5b-a056f4052fee",
 CellID->1691604121],
Cell[1056146, 29844, 657, 13, 22, "Print",ExpressionUUID->"e2f78199-e2d9-4eb7-a658-09ce598fd232",
 CellID->254459014],
Cell[1056806, 29859, 449, 11, 22, "Print",ExpressionUUID->"f7256c8d-784d-4408-9d9f-40e91682fb4e",
 CellID->1052707121],
Cell[1057258, 29872, 334, 8, 22, "Print",ExpressionUUID->"7f841b7c-bc6c-4822-96fa-15cec466db02",
 CellID->1878298661],
Cell[1057595, 29882, 552, 14, 22, "Print",ExpressionUUID->"ac9d1d54-075f-41a2-8bb1-98956e51a64d",
 CellID->555984527],
Cell[1058150, 29898, 374, 9, 22, "Print",ExpressionUUID->"4561b92d-cdf2-434e-a065-ada9cfd93c08",
 CellID->730597192],
Cell[1058527, 29909, 547, 13, 22, "Print",ExpressionUUID->"f843d385-e3d7-406b-97b5-efb29926892b",
 CellID->1982996045],
Cell[1059077, 29924, 406, 8, 22, "Print",ExpressionUUID->"82e5b943-24d9-4334-b5c9-42976217a92e",
 CellID->1415810522],
Cell[1059486, 29934, 507, 13, 22, "Print",ExpressionUUID->"00579408-3e7d-45d8-8edf-20294f115387",
 CellID->1408634494],
Cell[1059996, 29949, 410, 8, 22, "Print",ExpressionUUID->"5db8ec80-7cd5-4e4b-8d4e-e18d396fff55",
 CellID->1060216130],
Cell[1060409, 29959, 500, 13, 22, "Print",ExpressionUUID->"72df186b-5195-4727-bce0-87d0171e7711",
 CellID->1380677751],
Cell[1060912, 29974, 441, 8, 22, "Print",ExpressionUUID->"e52729ff-0f25-4255-aa29-4a0a041c58e4",
 CellID->1490155222],
Cell[1061356, 29984, 391, 8, 22, "Print",ExpressionUUID->"f644a698-9277-43cb-8995-43c287ce0990",
 CellID->1084524840],
Cell[1061750, 29994, 849, 21, 22, "Print",ExpressionUUID->"70760835-627a-4317-a158-59f4a52d0748",
 CellID->1915273791],
Cell[1062602, 30017, 480, 10, 22, "Print",ExpressionUUID->"65a16a5f-bf60-444b-99d3-f710fda3ce95",
 CellID->1374220454],
Cell[1063085, 30029, 567, 14, 22, "Print",ExpressionUUID->"abf82003-886d-4547-b309-d60354f08913",
 CellID->336546893],
Cell[1063655, 30045, 597, 14, 22, "Print",ExpressionUUID->"2796f95e-2205-426f-b6e4-82eb89cfc206",
 CellID->2019188100],
Cell[1064255, 30061, 418, 9, 22, "Print",ExpressionUUID->"bdd910f6-dc85-4526-a663-de5612d3f05d",
 CellID->1957806806],
Cell[1064676, 30072, 388, 8, 22, "Print",ExpressionUUID->"1450ba03-7bfa-4e2b-b89d-843d8bc46c4f",
 CellID->1114549108],
Cell[1065067, 30082, 497, 11, 22, "Print",ExpressionUUID->"6ae1841c-1764-4b88-8775-d299767047b6",
 CellID->1874489022],
Cell[1065567, 30095, 425, 8, 22, "Print",ExpressionUUID->"b15c3bcb-ee4b-4658-8120-c59dd5d025b4",
 CellID->951891264],
Cell[1065995, 30105, 681, 18, 22, "Print",ExpressionUUID->"3becc5dd-c42a-4a1d-a03f-f8c2983c291d",
 CellID->1830787545],
Cell[1066679, 30125, 500, 13, 22, "Print",ExpressionUUID->"bee12b09-57e9-40fb-be21-6e311df1f64a",
 CellID->900574599],
Cell[1067182, 30140, 772, 18, 22, "Print",ExpressionUUID->"202131ae-cca0-4387-ae1b-f466f66e0d25",
 CellID->492541337],
Cell[1067957, 30160, 488, 9, 22, "Print",ExpressionUUID->"21ab58db-d32b-4faf-824b-91a3f5c7639a",
 CellID->1740275421],
Cell[1068448, 30171, 502, 11, 22, "Print",ExpressionUUID->"c1a1fa7a-dfea-45ec-b1ed-3e126aa016eb",
 CellID->81728607],
Cell[1068953, 30184, 326, 7, 22, "Print",ExpressionUUID->"12ddd58c-28f0-4450-bbd0-1acc80c79a02",
 CellID->1370647416],
Cell[1069282, 30193, 331, 8, 22, "Print",ExpressionUUID->"35071e22-8725-4cab-8153-5261168752b7",
 CellID->414039343],
Cell[1069616, 30203, 683, 17, 22, "Print",ExpressionUUID->"0a06c632-3210-4bd9-8a0f-3cbba4e1166f",
 CellID->912221521],
Cell[1070302, 30222, 560, 13, 22, "Print",ExpressionUUID->"a8b6a7c2-7702-4275-8611-28dc5da42b0b",
 CellID->821347514],
Cell[1070865, 30237, 375, 8, 22, "Print",ExpressionUUID->"ac544d9a-fd18-4860-b039-e523b9f5e397",
 CellID->362784882],
Cell[1071243, 30247, 658, 13, 22, "Print",ExpressionUUID->"7ace7537-c0f8-42b8-93e9-2e2c9b77113e",
 CellID->619437941],
Cell[1071904, 30262, 447, 11, 22, "Print",ExpressionUUID->"3debb102-04d1-413d-bdab-e6f02f705fe9",
 CellID->2040237378],
Cell[1072354, 30275, 362, 9, 22, "Print",ExpressionUUID->"90f3c059-dff5-4bab-aead-0cee2b35c538",
 CellID->1398220397],
Cell[1072719, 30286, 551, 14, 22, "Print",ExpressionUUID->"ffe9ca4b-b9b8-458b-a24d-cc2beef48c07",
 CellID->2124146905],
Cell[1073273, 30302, 403, 10, 22, "Print",ExpressionUUID->"f880cd5e-0a6c-4dc6-b929-f418fd3dd237",
 CellID->768884607],
Cell[1073679, 30314, 549, 13, 22, "Print",ExpressionUUID->"2bc2ea7e-4d80-456c-b160-e55926f5f70f",
 CellID->1234485850],
Cell[1074231, 30329, 406, 8, 22, "Print",ExpressionUUID->"03bfd093-09bf-48a1-9a3c-9d8016b8327f",
 CellID->1164167283],
Cell[1074640, 30339, 500, 13, 22, "Print",ExpressionUUID->"0f588641-4eec-4a13-bc66-6379995b3b41",
 CellID->436017564],
Cell[1075143, 30354, 407, 8, 22, "Print",ExpressionUUID->"dd044807-a668-441e-bda1-8733dd9f315a",
 CellID->932994584],
Cell[1075553, 30364, 501, 13, 22, "Print",ExpressionUUID->"6996463d-8d33-4d8c-a32e-2a2ec3d9a26b",
 CellID->2061866541]
}, Open  ]],
Cell[1076069, 30380, 2126, 61, 32, "Output",ExpressionUUID->"0866f0c0-1356-45ee-8ff6-26aa7f3b2859",
 CellID->1994826595]
}, Open  ]]
}, Open  ]],
Cell[1078222, 30445, 853, 24, 74, "Text",ExpressionUUID->"a4507fde-768f-477b-9982-eb13d4221dcf"],
Cell[CellGroupData[{
Cell[1079100, 30473, 379, 9, 56, "MathCaption",ExpressionUUID->"0372fef3-e7c8-4c7e-90fa-c9885d121816"],
Cell[CellGroupData[{
Cell[1079504, 30486, 116, 2, 27, "Input",ExpressionUUID->"148749d1-a620-47ee-9076-03e9bf70f5ab"],
Cell[1079623, 30490, 172, 3, 26, "Output",ExpressionUUID->"aeda6b3e-e1c3-4c8b-84c3-6b13612b6536",
 CellID->535885201]
}, Open  ]]
}, Open  ]],
Cell[1079822, 30497, 511, 13, 41, "DefinitionBox",ExpressionUUID->"27cb64b0-2b65-4816-aab4-cd60a38eaa6b"],
Cell[1080336, 30512, 91, 0, 23, "Text",ExpressionUUID->"8e34f80c-c37d-4984-bf4d-7e8f11abeef1"],
Cell[1080430, 30514, 273, 6, 23, "Text",ExpressionUUID->"71fe7dc1-1a37-43e7-b943-64f39767fca6"],
Cell[CellGroupData[{
Cell[1080728, 30524, 177, 5, 40, "MathCaption",ExpressionUUID->"ed36a11c-e4a3-4cec-8330-d25081aa7142"],
Cell[CellGroupData[{
Cell[1080930, 30533, 377, 10, 27, "Input",ExpressionUUID->"fd3869cf-952b-4a2b-a3df-9716ebaffe6b"],
Cell[1081310, 30545, 670, 13, 22, "Print",ExpressionUUID->"71b9e845-0598-4eae-a23e-61516b68ba6b",
 CellID->76157689]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1082029, 30564, 242, 5, 40, "MathCaption",ExpressionUUID->"ea8a247c-ee40-4523-868a-e0c9ffa61f66"],
Cell[1082274, 30571, 447, 12, 47, "Input",ExpressionUUID->"3e783a9e-1f33-4b90-a49f-3fd68b7e0e05"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1082758, 30588, 228, 5, 40, "MathCaption",ExpressionUUID->"55f21979-4e3f-4aa9-bff7-757a6061ffc0"],
Cell[CellGroupData[{
Cell[1083011, 30597, 222, 4, 27, "Input",ExpressionUUID->"688a4ec1-f134-4f88-9d92-d06a0f0e01b4"],
Cell[1083236, 30603, 407, 9, 26, "Output",ExpressionUUID->"83d38c86-5951-43e8-b7ed-7a93da3171e0",
 CellID->82025411]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1083692, 30618, 358, 9, 40, "MathCaption",ExpressionUUID->"e4869e3c-055e-492e-a642-4b1246131c7e"],
Cell[CellGroupData[{
Cell[1084075, 30631, 285, 6, 27, "Input",ExpressionUUID->"55d129b9-cdb5-4785-abc2-6a627d133ccd"],
Cell[1084363, 30639, 409, 9, 26, "Output",ExpressionUUID->"87cc3c79-d14a-4073-ba68-17ecdea9c79f",
 CellID->2064064950]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1084821, 30654, 206, 3, 40, "MathCaption",ExpressionUUID->"78e69ffc-f2d4-40df-82b9-36f3f9d5e196"],
Cell[CellGroupData[{
Cell[1085052, 30661, 153, 3, 27, "Input",ExpressionUUID->"20567bd8-92e5-4972-9584-5a2e5080ee48"],
Cell[1085208, 30666, 588, 14, 46, "Output",ExpressionUUID->"11e34d29-76e2-4e24-823f-7d1a9987ee9e",
 CellID->288781012]
}, Open  ]],
Cell[CellGroupData[{
Cell[1085833, 30685, 337, 9, 27, "Input",ExpressionUUID->"3d4c6d62-82a5-47c7-a497-e7643dede8bd"],
Cell[1086173, 30696, 454, 10, 22, "Message",ExpressionUUID->"7da13220-0e9d-4ad5-a563-87a9a3163466",
 CellID->229426464],
Cell[1086630, 30708, 446, 8, 22, "Print",ExpressionUUID->"a9970acb-21da-4a3e-8e79-e671d975a424",
 CellID->1237434083],
Cell[1087079, 30718, 456, 10, 22, "Message",ExpressionUUID->"f3bece03-70d4-4485-8d74-7771c2ad94c6",
 CellID->1322995433],
Cell[1087538, 30730, 453, 9, 22, "Print",ExpressionUUID->"0321ceef-c804-4e3d-9675-49ab579b0f62",
 CellID->538561393],
Cell[1087994, 30741, 224, 5, 26, "Output",ExpressionUUID->"d197824e-22d1-4a9d-a7be-250e1d82ee59",
 CellID->2105243693]
}, Open  ]],
Cell[CellGroupData[{
Cell[1088255, 30751, 444, 13, 27, "Input",ExpressionUUID->"72fe68e8-2b7e-41fa-a107-cfc5ffa0cce8"],
Cell[1088702, 30766, 454, 10, 22, "Message",ExpressionUUID->"7085419d-961f-4152-94be-d976e7c96c83",
 CellID->890180279],
Cell[1089159, 30778, 444, 8, 22, "Print",ExpressionUUID->"a20e79d2-b94f-4b22-9af4-a9892eb86305",
 CellID->1888864151],
Cell[1089606, 30788, 456, 10, 22, "Message",ExpressionUUID->"3098ebff-1ec6-4a5a-97e7-cd95107e26eb",
 CellID->2036515903],
Cell[1090065, 30800, 444, 8, 22, "Print",ExpressionUUID->"36f3dee7-9205-45ea-8b5b-6eda1f150274",
 CellID->1128415835],
Cell[1090512, 30810, 2312, 64, 32, "Output",ExpressionUUID->"d97fd5b1-c399-493d-8bbc-b5312f46c9f4",
 CellID->856852188]
}, Open  ]],
Cell[CellGroupData[{
Cell[1092861, 30879, 533, 16, 27, "Input",ExpressionUUID->"27a0dff3-517b-47cb-bdde-53cae0f29f6f"],
Cell[1093397, 30897, 453, 10, 22, "Message",ExpressionUUID->"40717f26-7dc5-4f93-8321-fad7efb1c7d6",
 CellID->89386934],
Cell[1093853, 30909, 449, 8, 22, "Print",ExpressionUUID->"f911e7de-0418-4ad4-8d1d-5ca6da6b356d",
 CellID->1235130485],
Cell[1094305, 30919, 456, 10, 22, "Message",ExpressionUUID->"bf492051-a3a3-4035-b17d-60047159d926",
 CellID->1261129493],
Cell[1094764, 30931, 446, 8, 22, "Print",ExpressionUUID->"5f2434c5-bc90-49f7-a950-357958d744f7",
 CellID->139792961],
Cell[1095213, 30941, 223, 5, 26, "Output",ExpressionUUID->"bf7b4454-96da-453a-875e-29bafda01b0b",
 CellID->153185709]
}, Open  ]],
Cell[CellGroupData[{
Cell[1095473, 30951, 622, 19, 27, "Input",ExpressionUUID->"bc2e1e6d-c305-4307-84cf-e27cebd21a76"],
Cell[1096098, 30972, 457, 10, 22, "Message",ExpressionUUID->"8a910d4e-0ed3-4bfb-9d9b-f6efc8aa1e91",
 CellID->1906802057],
Cell[1096558, 30984, 445, 8, 22, "Print",ExpressionUUID->"38ebff8b-96c0-42cf-9b63-c86e24051a54",
 CellID->723707818],
Cell[1097006, 30994, 452, 10, 22, "Message",ExpressionUUID->"39248624-227e-437b-b75c-e51be1d7390e",
 CellID->9920518],
Cell[1097461, 31006, 444, 8, 22, "Print",ExpressionUUID->"56bbc755-8eac-4e1c-bff4-b4625081566a",
 CellID->1377905207],
Cell[1097908, 31016, 4276, 118, 32, "Output",ExpressionUUID->"6fb881b7-7708-4341-9a01-d45606b5cdd0",
 CellID->2136648448]
}, Open  ]],
Cell[CellGroupData[{
Cell[1102221, 31139, 711, 22, 47, "Input",ExpressionUUID->"62944ed6-4977-42c2-927e-5ecaf2ca13aa"],
Cell[1102935, 31163, 454, 10, 22, "Message",ExpressionUUID->"88ee9675-3407-41ac-bf10-6082be9544a2",
 CellID->428920402],
Cell[1103392, 31175, 444, 8, 22, "Print",ExpressionUUID->"38b9bd6f-b9ef-435c-a17c-d9e7a70d3203",
 CellID->1910077082],
Cell[1103839, 31185, 456, 10, 22, "Message",ExpressionUUID->"f7239f83-2376-427e-b83c-58ee9d650581",
 CellID->2079961818],
Cell[1104298, 31197, 450, 9, 22, "Print",ExpressionUUID->"3ba135a4-5c02-4273-bf50-51386a64c3a9",
 CellID->1179469260],
Cell[1104751, 31208, 224, 5, 26, "Output",ExpressionUUID->"f1e324b4-7e44-42c3-a4b9-260c9e7d155c",
 CellID->2065791010]
}, Open  ]],
Cell[CellGroupData[{
Cell[1105012, 31218, 800, 25, 47, "Input",ExpressionUUID->"c90256d8-4702-47d3-9679-451d7e5f3452"],
Cell[1105815, 31245, 455, 10, 22, "Message",ExpressionUUID->"23bb41c2-651f-4849-9388-f494abfe84d6",
 CellID->1381384021],
Cell[1106273, 31257, 447, 9, 22, "Print",ExpressionUUID->"6dcb4702-2af2-4b63-b666-845486a36db1",
 CellID->475533230],
Cell[1106723, 31268, 456, 10, 22, "Message",ExpressionUUID->"d86318cb-1101-4c9a-bc49-c538d48ac6e0",
 CellID->1500346123],
Cell[1107182, 31280, 441, 8, 22, "Print",ExpressionUUID->"c9b57d7d-6167-4214-b208-4a526244f897",
 CellID->564306187],
Cell[1107626, 31290, 6359, 176, 32, "Output",ExpressionUUID->"c38f0545-96d8-483a-ba7d-719d01d89626",
 CellID->1006299757]
}, Open  ]],
Cell[CellGroupData[{
Cell[1114022, 31471, 889, 28, 66, "Input",ExpressionUUID->"e3293a73-d6ff-4aef-86fa-7244bd2ba255"],
Cell[1114914, 31501, 455, 10, 22, "Message",ExpressionUUID->"8a712f1f-8243-47bd-b3ed-b8bc58cce75d",
 CellID->1446295774],
Cell[1115372, 31513, 447, 9, 22, "Print",ExpressionUUID->"9a37b2f3-af3d-4143-ba06-487471f6a300",
 CellID->525673225],
Cell[1115822, 31524, 455, 10, 22, "Message",ExpressionUUID->"57480c3a-bf20-48e5-8f56-93943f10ab9f",
 CellID->242168817],
Cell[1116280, 31536, 439, 8, 22, "Print",ExpressionUUID->"3bea37a0-ae26-4279-a172-5af0097721ef",
 CellID->649796254],
Cell[1116722, 31546, 224, 5, 26, "Output",ExpressionUUID->"adebffc6-9e2b-449a-b1fc-9222426d370e",
 CellID->1171295983]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1116995, 31557, 138, 1, 40, "MathCaption",ExpressionUUID->"bb635e68-c167-4451-ae8f-ce5009812c1d"],
Cell[CellGroupData[{
Cell[1117158, 31562, 221, 5, 27, "Input",ExpressionUUID->"649c518f-45c4-459d-86ba-899131779675"],
Cell[1117382, 31569, 408, 9, 26, "Output",ExpressionUUID->"2177cfae-2db9-48c0-8eca-7e6ebfa1f6ee",
 CellID->2136477879]
}, Open  ]],
Cell[CellGroupData[{
Cell[1117827, 31583, 627, 20, 27, "Input",ExpressionUUID->"7ec7d309-e127-4be5-a47a-8236224af3f2"],
Cell[1118457, 31605, 224, 5, 26, "Output",ExpressionUUID->"34e0ea7a-90ef-4ef0-8728-1c0674f189f1",
 CellID->1857134513]
}, Open  ]],
Cell[CellGroupData[{
Cell[1118718, 31615, 713, 23, 27, "Input",ExpressionUUID->"d909efeb-86ba-48c1-881d-7033a9cd54d8"],
Cell[1119434, 31640, 6359, 176, 32, "Output",ExpressionUUID->"f30b2d64-a861-4898-8c3d-45032e7fe687",
 CellID->1332434493]
}, Open  ]],
Cell[CellGroupData[{
Cell[1125830, 31821, 799, 26, 47, "Input",ExpressionUUID->"77099857-70ce-4350-8820-30c9a9b83aca"],
Cell[1126632, 31849, 225, 5, 26, "Output",ExpressionUUID->"aa210262-173d-4ed7-9fb7-feb8cdbc586c",
 CellID->267052935]
}, Open  ]],
Cell[CellGroupData[{
Cell[1126894, 31859, 885, 29, 66, "Input",ExpressionUUID->"f0230162-5e6a-46fa-9503-bcdc742d135e"],
Cell[1127782, 31890, 8325, 230, 32, "Output",ExpressionUUID->"5ccc7ea2-b619-4e47-a540-5d51f1e784b9",
 CellID->117856315]
}, Open  ]],
Cell[CellGroupData[{
Cell[1136144, 32125, 971, 32, 66, "Input",ExpressionUUID->"3395f34e-bedc-453c-a4b7-a6a08b9794e4"],
Cell[1137118, 32159, 223, 5, 26, "Output",ExpressionUUID->"fadb01c1-e854-47be-bf58-cdcc1f028e8e",
 CellID->831163671]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1137402, 32171, 99, 0, 39, "Subsection",ExpressionUUID->"3a8486f2-1233-4b73-8f71-ca4c9d665f6e"],
Cell[1137504, 32173, 3724, 81, 573, "Text",ExpressionUUID->"b45e0955-72d2-4720-8b66-69f514156352"],
Cell[1141231, 32256, 2200, 40, 594, "Text",ExpressionUUID->"656ed42e-59f5-4270-9e76-248059fcbae8"],
Cell[1143434, 32298, 1538, 41, 376, "Text",ExpressionUUID->"e355b849-e774-490e-bf88-f6b718337e2c"],
Cell[1144975, 32341, 693, 17, 58, "DefinitionBox",ExpressionUUID->"e59d4de5-3442-44e3-bbbe-abcf86b5212a"],
Cell[CellGroupData[{
Cell[1145693, 32362, 99, 0, 23, "Text",ExpressionUUID->"0a1890cf-def1-478c-bf42-a12395efe46e"],
Cell[CellGroupData[{
Cell[1145817, 32366, 135, 3, 27, "Input",ExpressionUUID->"8e3f8703-33bc-44fe-8e15-7e30b7adefd0"],
Cell[1145955, 32371, 401, 6, 67, "Print",ExpressionUUID->"b286d54a-110c-4101-a947-c1930e20065e",
 CellTags->"Info3143728981546-9181623",
 CellID->799431715]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1146405, 32383, 168, 5, 40, "MathCaption",ExpressionUUID->"5188daf7-454e-438e-beb3-a072c3832375"],
Cell[CellGroupData[{
Cell[1146598, 32392, 160, 3, 27, "Input",ExpressionUUID->"985532c9-3af5-4014-9c1e-6e3ac0238427"],
Cell[1146761, 32397, 375, 10, 26, "Output",ExpressionUUID->"afb82529-fe52-4833-acc1-225f7b962461",
 CellID->1391138373]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1147185, 32413, 158, 2, 40, "MathCaption",ExpressionUUID->"0e9c8122-11ee-4368-aca5-314b0f9764d4"],
Cell[CellGroupData[{
Cell[1147368, 32419, 342, 8, 27, "Input",ExpressionUUID->"28af3cd7-98ec-418d-9c0b-5e1d2d2dbb16"],
Cell[1147713, 32429, 961, 23, 26, "Output",ExpressionUUID->"4fe2e442-3337-4744-b763-5d7f31fff981",
 CellID->1208170122]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1148723, 32458, 134, 0, 40, "MathCaption",ExpressionUUID->"80277130-8651-4d89-9c61-0169b67139c9"],
Cell[1148860, 32460, 243, 6, 27, "Input",ExpressionUUID->"f64c2818-f3b5-4001-8795-1e453885c655"],
Cell[CellGroupData[{
Cell[1149128, 32470, 135, 3, 27, "Input",ExpressionUUID->"9983e5d6-1ee7-490a-b9d9-ff18cb7b2c56"],
Cell[1149266, 32475, 408, 6, 67, "Print",ExpressionUUID->"bfc1f777-98e2-405b-8b4a-375be9c9d56c",
 CellTags->"Info3183728981546-9181623",
 CellID->1257358069]
}, Open  ]],
Cell[CellGroupData[{
Cell[1149711, 32486, 342, 8, 27, "Input",ExpressionUUID->"987e8edb-93c1-4239-86af-d40316e247f6"],
Cell[1150056, 32496, 964, 23, 26, "Output",ExpressionUUID->"28fd081e-9b2c-49b7-8005-d85ee731cd73",
 CellID->1177979203]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1151069, 32525, 411, 14, 56, "MathCaption",ExpressionUUID->"cee1fcb7-94c3-43ef-8bd9-85d4691ea749"],
Cell[1151483, 32541, 234, 5, 27, "Input",ExpressionUUID->"76a2a6da-9d40-4b88-8662-e74e1cd952fd"],
Cell[CellGroupData[{
Cell[1151742, 32550, 342, 8, 27, "Input",ExpressionUUID->"70008b35-5ee4-4f74-a3f7-21bb5273f958"],
Cell[1152087, 32560, 955, 22, 26, "Output",ExpressionUUID->"15af1cda-c5a7-4964-a931-270805949b8a",
 CellID->644723128]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1153091, 32588, 418, 6, 57, "Text",ExpressionUUID->"84e02928-6c04-4c4d-8e97-2fed344ab395"],
Cell[CellGroupData[{
Cell[1153534, 32598, 641, 22, 27, "Input",ExpressionUUID->"418e6062-1975-413d-8064-6ec9003d3f33"],
Cell[1154178, 32622, 5909, 172, 32, "Output",ExpressionUUID->"0614c7cb-56dd-4451-8bc2-8ca2bef911a0",
 CellID->1806613181]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1160136, 32800, 364, 14, 40, "MathCaption",ExpressionUUID->"8a4bbb3d-20c8-48c2-b074-8e7dc059a85c"],
Cell[1160503, 32816, 234, 5, 27, "Input",ExpressionUUID->"dc15dac0-4473-4a73-8bdc-93d65ffe3e3c"],
Cell[CellGroupData[{
Cell[1160762, 32825, 311, 7, 27, "Input",ExpressionUUID->"3bc09fe8-ac08-4acd-9363-cfdf672f9ec0"],
Cell[1161076, 32834, 454, 10, 22, "Message",ExpressionUUID->"bbfac71d-e079-4011-9533-6ced83690e66",
 CellID->476448134],
Cell[1161533, 32846, 444, 8, 22, "Print",ExpressionUUID->"130fc8fa-eb87-43aa-80f8-b5a8f241e38b",
 CellID->1844752122],
Cell[1161980, 32856, 456, 10, 22, "Message",ExpressionUUID->"32507cba-cedf-4833-a991-fb6ae135efb8",
 CellID->1527502715],
Cell[1162439, 32868, 444, 8, 22, "Print",ExpressionUUID->"defa3453-2797-4719-88dd-4622dc278a90",
 CellID->1718614767],
Cell[1162886, 32878, 6359, 176, 32, "Output",ExpressionUUID->"883d24fd-0908-480a-9271-af30ba786f47",
 CellID->1103336819]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1169294, 33060, 331, 5, 72, "MathCaption",ExpressionUUID->"f89b7f57-0db5-415a-a00a-cd93065f6d01"],
Cell[1169628, 33067, 234, 5, 27, "Input",ExpressionUUID->"e6ec04c6-ffb5-4a41-b9fb-4b84a56bd6b7"],
Cell[CellGroupData[{
Cell[1169887, 33076, 311, 7, 27, "Input",ExpressionUUID->"88025c4c-5c88-42e9-b93f-d560278aa43f"],
Cell[1170201, 33085, 454, 10, 22, "Message",ExpressionUUID->"c159de9a-91d6-45e6-9098-01593e54d90b",
 CellID->260625088],
Cell[1170658, 33097, 448, 9, 22, "Print",ExpressionUUID->"6e784f90-9cd1-4739-9480-b251230ff2d5",
 CellID->1606697783],
Cell[1171109, 33108, 456, 10, 22, "Message",ExpressionUUID->"0eac8c16-d240-4142-8fb1-3be4f40fb96b",
 CellID->1310261503],
Cell[1171568, 33120, 448, 9, 22, "Print",ExpressionUUID->"91adbdd2-ba82-4fb3-a883-e6fc8c0324af",
 CellID->1267764583],
Cell[1172019, 33131, 6359, 176, 32, "Output",ExpressionUUID->"83460eac-6841-4d0a-a08c-80b9083e0187",
 CellID->2134084594]
}, Open  ]],
Cell[1178393, 33310, 145, 3, 27, "Input",ExpressionUUID->"41afebc0-4860-4472-b64a-9f47dbab7a79"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1178587, 33319, 94, 0, 39, "Subsection",ExpressionUUID->"218a8fe1-d10f-406f-b7b2-bf7d5c336cf5"],
Cell[1178684, 33321, 1221, 32, 137, "Text",ExpressionUUID->"ce5ef912-807c-49fd-8faf-c335f3310335"],
Cell[CellGroupData[{
Cell[1179930, 33357, 267, 4, 56, "MathCaption",ExpressionUUID->"4d850949-7562-413c-af18-bfd3c0f00b6d"],
Cell[CellGroupData[{
Cell[1180222, 33365, 116, 2, 27, "Input",ExpressionUUID->"d5c06c29-8c1b-4f4b-9b81-9489407ee455"],
Cell[1180341, 33369, 173, 3, 26, "Output",ExpressionUUID->"9d67c1c7-8b4f-48eb-9a5a-86d46bfaaa3a",
 CellID->517171343]
}, Open  ]]
}, Open  ]],
Cell[1180541, 33376, 122, 1, 23, "Text",ExpressionUUID->"50af9ae9-f931-41f8-b8a2-4315729dacf8"],
Cell[1180666, 33379, 148, 2, 23, "Text",ExpressionUUID->"3c3e1fd3-dada-4a5f-939c-63c71cfe19e0"],
Cell[CellGroupData[{
Cell[1180839, 33385, 151, 2, 40, "MathCaption",ExpressionUUID->"440cb950-21db-4f09-a20d-881ff9003f3c"],
Cell[1180993, 33389, 736, 21, 85, "Input",ExpressionUUID->"a97db65d-aa6f-4e2c-8596-38103839f2e5"],
Cell[CellGroupData[{
Cell[1181754, 33414, 147, 3, 27, "Input",ExpressionUUID->"81792521-ba50-437a-892a-2af2d7f16f90"],
Cell[1181904, 33419, 1091, 31, 32, "Output",ExpressionUUID->"90966409-7be8-4612-aff7-55d4a98a8717",
 CellID->1232442392]
}, Open  ]],
Cell[CellGroupData[{
Cell[1183032, 33455, 147, 3, 27, "Input",ExpressionUUID->"aa483db2-bc25-47f5-9259-b7794da341a7"],
Cell[1183182, 33460, 6865, 200, 32, "Output",ExpressionUUID->"65f541c7-88dd-4b90-8b6f-e62822efa2a3",
 CellID->1179507029]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1190096, 33666, 119, 0, 40, "MathCaption",ExpressionUUID->"58278466-d99d-4b97-83ee-1ec983186a32"],
Cell[CellGroupData[{
Cell[1190240, 33670, 216, 5, 27, "Input",ExpressionUUID->"03049667-d549-42e2-bdda-8f2adad19d79"],
Cell[1190459, 33677, 913, 15, 65, "Output",ExpressionUUID->"09830161-d42a-4f93-833f-ba2c0d78ee3c",
 CellID->553090946]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1191421, 33698, 341, 6, 56, "MathCaption",ExpressionUUID->"7b06b920-242b-408a-bb0a-1e41c5c38091"],
Cell[1191765, 33706, 520, 15, 27, "Input",ExpressionUUID->"a57f1837-0fb7-4e26-b145-ae0d180a3d7a"],
Cell[CellGroupData[{
Cell[1192310, 33725, 143, 3, 27, "Input",ExpressionUUID->"04931ec2-cf33-471f-b029-68b11f36f523"],
Cell[1192456, 33730, 6369, 176, 32, "Output",ExpressionUUID->"31470277-d0a6-4d53-a2b8-45d53c5fa269",
 CellID->1490305206]
}, Open  ]],
Cell[CellGroupData[{
Cell[1198862, 33911, 143, 3, 27, "Input",ExpressionUUID->"d39e6661-d757-44c9-be98-9bc7db8f7f96"],
Cell[1199008, 33916, 235, 5, 26, "Output",ExpressionUUID->"3261b232-0b24-40f5-bbcd-5d7058f4118f",
 CellID->1457822281]
}, Open  ]],
Cell[CellGroupData[{
Cell[1199280, 33926, 144, 3, 27, "Input",ExpressionUUID->"a0f35cac-b5fd-46a1-805e-266da2fda802"],
Cell[1199427, 33931, 234, 5, 26, "Output",ExpressionUUID->"29950182-f47c-4bb2-9abf-f6ef920590ba",
 CellID->978148144]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1199710, 33942, 443, 6, 88, "MathCaption",ExpressionUUID->"9e378747-c9a2-4c1a-a41d-9068cb4d1062"],
Cell[CellGroupData[{
Cell[1200178, 33952, 234, 6, 27, "Input",ExpressionUUID->"2ff056ab-3e3d-45c0-8abc-0bf50f7cce6a"],
Cell[1200415, 33960, 2528, 65, 122, "Output",ExpressionUUID->"36fe7410-2339-4c82-86f8-8c8813a5c7f3",
 CellID->1077164391]
}, Open  ]],
Cell[CellGroupData[{
Cell[1202980, 34030, 372, 10, 27, "Input",ExpressionUUID->"89c512e7-7580-4529-a6d3-0ef04abd224a"],
Cell[1203355, 34042, 1994, 61, 536, "Output",ExpressionUUID->"bb713164-1237-4569-bd3f-e75dfa8df2ca",
 CellID->614775327]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1205398, 34109, 230, 3, 56, "MathCaption",ExpressionUUID->"e01e3f92-c6d7-4118-80f2-d79e14f1e348"],
Cell[CellGroupData[{
Cell[1205653, 34116, 528, 16, 27, "Input",ExpressionUUID->"addcdb82-571c-45b0-bc11-5008a436bd73"],
Cell[1206184, 34134, 267, 6, 26, "Output",ExpressionUUID->"46e18a45-d7cd-42ad-83ff-3ecaa2c3eb40",
 CellID->985095172]
}, Open  ]],
Cell[CellGroupData[{
Cell[1206488, 34145, 1000, 26, 66, "Input",ExpressionUUID->"960cfb89-bdb6-402f-bd23-5f54b3b7c63b"],
Cell[1207491, 34173, 26227, 531, 159, "Output",ExpressionUUID->"a8ff1d5e-8314-4a61-9ca8-e05214aa273e",
 CellID->1543121081]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1233767, 34710, 155, 2, 40, "MathCaption",ExpressionUUID->"dbd6ca66-204d-42ed-9eac-bb8b2997fd79"],
Cell[CellGroupData[{
Cell[1233947, 34716, 572, 16, 47, "Input",ExpressionUUID->"790ea5d1-f8b9-443a-a0de-5dadc8a53ef8"],
Cell[1234522, 34734, 16478, 341, 155, "Output",ExpressionUUID->"6dcd8b22-ce7f-45ac-af64-10672ccd3513",
 CellID->76003548]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1251049, 35081, 1200, 17, 312, "MathCaption",ExpressionUUID->"b3c38f86-5c21-4a7f-b07a-f93a645ea233"],
Cell[1252252, 35100, 46996, 1911, 337, 39935, 1793, "GraphicsData", "PostScript", "Graphics",ExpressionUUID->"05777d37-7e27-4146-8ec0-ea9558f56baf"],
Cell[CellGroupData[{
Cell[1299273, 37015, 269, 6, 47, "Input",ExpressionUUID->"7381c63d-68d4-494b-8195-637c9fb9b3fd"],
Cell[1299545, 37023, 554, 12, 22, "Print",ExpressionUUID->"755fe360-d4a6-4a13-98fb-96a58aca7e86",
 CellID->1786344918]
}, Open  ]]
}, Open  ]],
Cell[1300126, 37039, 122, 1, 23, "Text",ExpressionUUID->"75290d70-a3f6-4a06-b46e-b1c44f8cf380"],
Cell[1300251, 37042, 171, 3, 23, "Text",ExpressionUUID->"bc7a5b46-1e2e-4867-8312-afe620e72123"],
Cell[CellGroupData[{
Cell[1300447, 37049, 104, 0, 40, "MathCaption",ExpressionUUID->"f8fb447d-f240-48b6-9298-8a9fe6f534aa"],
Cell[CellGroupData[{
Cell[1300576, 37053, 371, 10, 27, "Input",ExpressionUUID->"d486fceb-da49-4a59-b705-951519642fee"],
Cell[1300950, 37065, 721, 14, 22, "Print",ExpressionUUID->"fa14ab72-61f6-4345-94c2-bd22f0bade93",
 CellID->1277651766]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1301720, 37085, 119, 0, 40, "MathCaption",ExpressionUUID->"4a57779e-4954-4cf5-a39a-8cea84420dd0"],
Cell[CellGroupData[{
Cell[1301864, 37089, 238, 7, 27, "Input",ExpressionUUID->"0a4263be-ef76-4c96-99c0-51e1ff9a4204"],
Cell[1302105, 37098, 170, 3, 26, "Output",ExpressionUUID->"6c7e54c1-7c64-401e-8f15-db942d4fb686",
 CellID->804244809]
}, Open  ]],
Cell[CellGroupData[{
Cell[1302312, 37106, 295, 9, 27, "Input",ExpressionUUID->"019ab91a-91b1-47e2-bdf0-536d93539a67"],
Cell[1302610, 37117, 1187, 33, 32, "Output",ExpressionUUID->"7500004f-97dd-4430-8f2a-f075a3fcff6d",
 CellID->680430645]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1303846, 37156, 370, 7, 72, "MathCaption",ExpressionUUID->"616311d9-73e9-4798-8f8e-45fd6504df5f"],
Cell[1304219, 37165, 1589, 48, 104, "Input",ExpressionUUID->"a50ed49d-b47a-4da1-ad0f-eb0dc97c8f2f"],
Cell[CellGroupData[{
Cell[1305833, 37217, 144, 3, 27, "Input",ExpressionUUID->"d7dd2de1-38a1-4e42-a772-3c23af96f8a9"],
Cell[1305980, 37222, 1139, 32, 32, "Output",ExpressionUUID->"da83b066-3505-4b4e-977f-9167c40930ae",
 CellID->647430240]
}, Open  ]],
Cell[CellGroupData[{
Cell[1307156, 37259, 145, 3, 27, "Input",ExpressionUUID->"e6415d7e-0ab9-489a-bab1-b007a22a5c7a"],
Cell[1307304, 37264, 10327, 294, 32, "Output",ExpressionUUID->"179db550-674f-40bb-87fa-5832a63336e5",
 CellID->60404331]
}, Open  ]],
Cell[CellGroupData[{
Cell[1317668, 37563, 257, 7, 27, "Input",ExpressionUUID->"eb6880de-7bcd-417d-bce4-b757c5d980a0"],
Cell[1317928, 37572, 3196, 91, 32, "Output",ExpressionUUID->"380b9c9b-bdc2-4e2d-8ab7-cb58ca7f9606",
 CellID->1603309733]
}, Open  ]],
Cell[CellGroupData[{
Cell[1321161, 37668, 226, 6, 27, "Input",ExpressionUUID->"129aa6ef-2d57-42f2-a76f-dd4307d6d6bc"],
Cell[1321390, 37676, 286, 7, 26, "Output",ExpressionUUID->"ec286e7f-12d8-4e74-bd53-e2e75d889ec4",
 CellID->201879975]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1321725, 37689, 168, 3, 40, "MathCaption",ExpressionUUID->"b9d45f12-b7f4-434f-9054-d97bd8049a33"],
Cell[1321896, 37694, 624, 19, 47, "Input",ExpressionUUID->"5c60c886-6fe6-43a2-a72d-874e7e44e7f1"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1322557, 37718, 87, 0, 40, "MathCaption",ExpressionUUID->"412b6862-162a-4456-9488-915be9bf5cda"],
Cell[CellGroupData[{
Cell[1322669, 37722, 140, 3, 27, "Input",ExpressionUUID->"f5af0254-856f-4def-8bdb-92f3b8ca1c72"],
Cell[1322812, 37727, 1307, 35, 32, "Output",ExpressionUUID->"98ab902e-5284-45b7-94be-7402c799bec8",
 CellID->2120061212]
}, Open  ]],
Cell[CellGroupData[{
Cell[1324156, 37767, 140, 3, 27, "Input",ExpressionUUID->"6eb20774-d3c8-40f4-a2f4-e63d0c87b56d"],
Cell[1324299, 37772, 233, 5, 26, "Output",ExpressionUUID->"4e079120-6de3-4b22-9250-a80a9e2b7719",
 CellID->820119060]
}, Open  ]],
Cell[CellGroupData[{
Cell[1324569, 37782, 140, 3, 27, "Input",ExpressionUUID->"7d1bb697-efc9-4f7e-ba3f-a301240025d1"],
Cell[1324712, 37787, 234, 5, 26, "Output",ExpressionUUID->"682767dd-8e40-4bf0-b8a7-10abb259991c",
 CellID->1190994974]
}, Open  ]],
Cell[CellGroupData[{
Cell[1324983, 37797, 140, 3, 27, "Input",ExpressionUUID->"7b38f26b-7e84-4972-a4e1-96fb146432f1"],
Cell[1325126, 37802, 4493, 122, 32, "Output",ExpressionUUID->"cf5bb4d2-a10c-43f7-bfb6-8a11bf302581",
 CellID->357454331]
}, Open  ]],
Cell[CellGroupData[{
Cell[1329656, 37929, 141, 3, 27, "Input",ExpressionUUID->"c4c207bf-5248-4cdd-b1a7-6922d04411ae"],
Cell[1329800, 37934, 11014, 298, 32, "Output",ExpressionUUID->"68762d59-ecd2-4b3e-9a8f-ae415ce0dc1e",
 CellID->1223180458]
}, Open  ]],
Cell[CellGroupData[{
Cell[1340851, 38237, 141, 3, 27, "Input",ExpressionUUID->"936c3368-c43c-4049-9303-00516971bf61"],
Cell[1340995, 38242, 21746, 586, 80, "Output",ExpressionUUID->"783061e0-ed2e-4021-b47c-a833b800d32f",
 CellID->133004875]
}, Open  ]],
Cell[CellGroupData[{
Cell[1362778, 38833, 141, 3, 27, "Input",ExpressionUUID->"9fa72213-ca81-44f1-b720-8ca2381e53f6"],
Cell[1362922, 38838, 235, 5, 26, "Output",ExpressionUUID->"06c70c2d-4524-45ba-a29f-85ecec8950ca",
 CellID->2029941245]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1363206, 38849, 175, 3, 40, "MathCaption",ExpressionUUID->"26c109e5-2e2f-4f4a-9820-55db413f2b93"],
Cell[1363384, 38854, 567, 17, 66, "Input",ExpressionUUID->"f466d4a3-f2e4-4418-a8dc-cf044cfd15b5"],
Cell[CellGroupData[{
Cell[1363976, 38875, 695, 19, 47, "Input",ExpressionUUID->"2045601b-73f0-4120-bf3a-d87848bea561"],
Cell[1364674, 38896, 14681, 305, 159, "Output",ExpressionUUID->"29262ecc-83ab-4277-8a11-b41440edb4b9",
 CellID->149804826]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1379404, 39207, 176, 3, 40, "MathCaption",ExpressionUUID->"3466479f-10e9-4cb3-b104-b0f38e953f71"],
Cell[1379583, 39212, 972, 32, 66, "Input",ExpressionUUID->"be790f1f-7919-4d17-89bc-6000a4ef7f6e"],
Cell[CellGroupData[{
Cell[1380580, 39248, 693, 19, 47, "Input",ExpressionUUID->"aa25ab63-7880-4635-bd01-0f094ce929f3"],
Cell[1381276, 39269, 14596, 303, 155, "Output",ExpressionUUID->"cb849551-d8bf-4e38-b54a-3f9d2942e8be",
 CellID->918876298]
}, Open  ]],
Cell[1395887, 39575, 1032, 32, 66, "Input",ExpressionUUID->"a387db89-fb5f-4460-b887-71913b0c88e2"],
Cell[CellGroupData[{
Cell[1396944, 39611, 695, 19, 47, "Input",ExpressionUUID->"6d844840-04d0-4eef-9f84-91eba2baae07"],
Cell[1397642, 39632, 13482, 282, 160, "Output",ExpressionUUID->"5fbc540e-bc39-41bf-98e7-f2298cc8837c",
 CellID->999034909]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1411173, 39920, 169, 3, 40, "MathCaption",ExpressionUUID->"8abc4310-5721-4430-a922-4fc6cf23bb01"],
Cell[CellGroupData[{
Cell[1411367, 39927, 185, 4, 27, "Input",ExpressionUUID->"d1ab9a61-e620-4c02-b28d-f3c9b5a5e646"],
Cell[1411555, 39933, 17066, 348, 159, "Output",ExpressionUUID->"1451c39e-d12e-49f6-ae1e-e752003f4d7a",
 CellID->1752563317]
}, Open  ]]
}, Open  ]],
Cell[1428648, 40285, 122, 1, 23, "Text",ExpressionUUID->"1f6cdcce-45f6-4960-8e52-6911fed6e1c9"],
Cell[1428773, 40288, 128, 0, 23, "Text",ExpressionUUID->"6a657e17-4ff5-4b35-b0ea-f0a001c61803"],
Cell[CellGroupData[{
Cell[1428926, 40292, 823, 11, 136, "MathCaption",ExpressionUUID->"59a01a97-dc36-430c-96ed-2027d076f997"],
Cell[1429752, 40305, 56742, 3415, 318, 52634, 3347, "GraphicsData", "PostScript", "Graphics",ExpressionUUID->"097fbb40-4e2b-4ac0-8018-f4c5230929ca"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1486531, 43725, 86, 0, 40, "MathCaption",ExpressionUUID->"df2e193a-b645-4761-965f-3cbaf77bca98"],
Cell[CellGroupData[{
Cell[1486642, 43729, 356, 8, 47, "Input",ExpressionUUID->"a6634d38-a5ef-415d-89da-995b757baeb5"],
Cell[1487001, 43739, 548, 12, 22, "Print",ExpressionUUID->"4d87c0fb-6d77-4f7e-b652-9499cf93206c",
 CellID->445966567]
}, Open  ]],
Cell[1487564, 43754, 307, 8, 47, "Input",ExpressionUUID->"89ae8c23-6009-426b-94e6-8d078e605141"]
}, Open  ]],
Cell[1487886, 43765, 1316, 19, 171, "Text",ExpressionUUID->"2feb7a0f-9e2d-4102-ae8a-967ca9c9a8c6"]
}, Closed]],
Cell[CellGroupData[{
Cell[1489239, 43789, 123, 0, 39, "Subsection",ExpressionUUID->"c977018a-28ed-40e4-b06f-6c06e25dd990"],
Cell[1489365, 43791, 366, 6, 41, "Text",ExpressionUUID->"34d74bf7-9c7e-40e5-89b5-1f2a7ed22854"],
Cell[CellGroupData[{
Cell[1489756, 43801, 107, 0, 40, "MathCaption",ExpressionUUID->"06f691bd-a442-485e-9949-ed7ad0041da2"],
Cell[CellGroupData[{
Cell[1489888, 43805, 256, 7, 27, "Input",ExpressionUUID->"2bba5c34-b32d-4dfd-8d41-0e461203103f"],
Cell[CellGroupData[{
Cell[1490169, 43816, 595, 12, 22, "Print",ExpressionUUID->"b1ca3830-467c-4457-9eaf-a0a2de5ce68b",
 CellID->642163539],
Cell[1490767, 43830, 606, 12, 22, "Print",ExpressionUUID->"1d4b9514-4e3c-47ed-a1cb-3372705937c4",
 CellID->1737273798]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1491434, 43849, 110, 0, 40, "MathCaption",ExpressionUUID->"2bbcdbf3-c4fb-41b7-ad75-d3c1f3a9c62c"],
Cell[CellGroupData[{
Cell[1491569, 43853, 225, 6, 27, "Input",ExpressionUUID->"d5502234-febe-4e8b-a8ba-5e060a99311c"],
Cell[1491797, 43861, 671, 13, 22, "Print",ExpressionUUID->"29c166f6-3fd3-4fc9-af74-e716d8a27966",
 CellID->1173058374]
}, Open  ]],
Cell[CellGroupData[{
Cell[1492505, 43879, 225, 6, 27, "Input",ExpressionUUID->"7be6047a-c266-4b2e-a56b-15a517a5b8cf"],
Cell[1492733, 43887, 673, 13, 22, "Print",ExpressionUUID->"ad0c209b-f0ed-48bc-abf7-cc01b0fe0d8e",
 CellID->1692212358]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1493455, 43906, 111, 0, 40, "MathCaption",ExpressionUUID->"98ad7da8-2749-4bfb-9c10-1f5cce99ddf5"],
Cell[CellGroupData[{
Cell[1493591, 43910, 436, 14, 27, "Input",ExpressionUUID->"db6674db-fd5b-4695-9699-b0df15d64639"],
Cell[1494030, 43926, 173, 3, 26, "Output",ExpressionUUID->"13dd1644-fc28-4064-985a-e38e92609559",
 CellID->1887163685]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1494252, 43935, 88, 0, 40, "MathCaption",ExpressionUUID->"0bfe26ac-834f-4dac-8bb3-df962330ee2d"],
Cell[CellGroupData[{
Cell[1494365, 43939, 280, 7, 47, "Input",ExpressionUUID->"c5a38fa5-edcc-4fa6-a2e8-e2da5b848dce"],
Cell[CellGroupData[{
Cell[1494670, 43950, 551, 12, 22, "Print",ExpressionUUID->"271256fa-f5c5-4b3d-bcba-2e58bd04eab1",
 CellID->1374071252],
Cell[1495224, 43964, 551, 12, 22, "Print",ExpressionUUID->"4ef216bb-f455-4714-82b6-10aa3d15e503",
 CellID->2123876173],
Cell[1495778, 43978, 724, 18, 22, "Print",ExpressionUUID->"4a7d502f-2ad1-4104-9326-711e0429daf3",
 CellID->513659177],
Cell[1496505, 43998, 557, 12, 22, "Print",ExpressionUUID->"8fcb35c6-3329-46b4-aa71-80bbb31edad9",
 CellID->185526899]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1497147, 44019, 776, 13, 54, "Section",ExpressionUUID->"b37ffd4e-e1c9-4992-9806-1ad019ce6449",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1736504676],
Cell[CellGroupData[{
Cell[1497948, 44036, 110, 0, 39, "Subsection",ExpressionUUID->"fb65e018-3f67-4957-966e-50fbeeb77714"],
Cell[1498061, 44038, 1774, 44, 229, "Text",ExpressionUUID->"d3dff436-06bc-43bb-9675-5625baf754cc"],
Cell[CellGroupData[{
Cell[1499860, 44086, 294, 11, 40, "MathCaption",ExpressionUUID->"4268c75a-8c25-45ed-99be-a9155307e8cd"],
Cell[CellGroupData[{
Cell[1500179, 44101, 198, 5, 27, "Input",ExpressionUUID->"316b5dc0-ab8b-4957-a7ac-152463176573"],
Cell[1500380, 44108, 625, 12, 22, "Print",ExpressionUUID->"c21c3d7f-7e15-4067-8713-fff2b3164065",
 CellID->632950163]
}, Open  ]],
Cell[CellGroupData[{
Cell[1501042, 44125, 282, 9, 27, "Input",ExpressionUUID->"9abd9aeb-f4b2-428d-b9a9-740dbdfa40a4"],
Cell[1501327, 44136, 2052, 60, 32, "Output",ExpressionUUID->"d68162c6-b4ac-4593-bb8b-3ced9c1b739d",
 CellID->1506885950]
}, Open  ]]
}, Open  ]],
Cell[1503406, 44200, 354, 6, 56, "MathCaption",ExpressionUUID->"b661ed81-f866-4d07-af16-f0b3bb9fe1a8"],
Cell[CellGroupData[{
Cell[1503785, 44210, 124, 0, 40, "MathCaption",ExpressionUUID->"47f18ded-ae51-42db-b892-cfda4b690003"],
Cell[CellGroupData[{
Cell[1503934, 44214, 137, 3, 27, "Input",ExpressionUUID->"ce59fe18-18e3-4ee7-8d45-1512d9558661"],
Cell[1504074, 44219, 2053, 60, 32, "Output",ExpressionUUID->"f58a3058-a108-422d-9741-a4c04eda09db",
 CellID->957512579]
}, Open  ]],
Cell[CellGroupData[{
Cell[1506164, 44284, 202, 6, 27, "Input",ExpressionUUID->"57b774f9-7a95-4a7a-ab67-6ae661dd6224"],
Cell[1506369, 44292, 3057, 89, 34, "Output",ExpressionUUID->"137cd222-41fc-4bd6-9c8c-e00a21c1c6de",
 CellID->1818948282]
}, Open  ]]
}, Open  ]],
Cell[1509453, 44385, 186, 4, 23, "Text",ExpressionUUID->"54915ffb-f592-4c93-b31a-49de1c339cf6"],
Cell[CellGroupData[{
Cell[1509664, 44393, 512, 16, 72, "MathCaption",ExpressionUUID->"f68b2a59-0964-430b-ad08-991896e2aed4"],
Cell[1510179, 44411, 375, 12, 27, "Input",ExpressionUUID->"ba402cdd-c2b4-4f73-ab23-f4d8b3f802ce"],
Cell[CellGroupData[{
Cell[1510579, 44427, 137, 3, 27, "Input",ExpressionUUID->"193a66c5-7a1d-4706-95d0-ed64d757477a"],
Cell[1510719, 44432, 2078, 60, 32, "Output",ExpressionUUID->"0761ba47-d10e-4a2c-93e7-ff9f7d725a1f",
 CellID->1634060529]
}, Open  ]],
Cell[CellGroupData[{
Cell[1512834, 44497, 202, 6, 27, "Input",ExpressionUUID->"74ae534d-16b5-4a26-bdd8-47b763bc3aea"],
Cell[1513039, 44505, 3980, 116, 32, "Output",ExpressionUUID->"8336e0ed-c44c-4c6f-8dbd-b5cbdaafa257",
 CellID->1634433067]
}, Open  ]]
}, Open  ]],
Cell[1517046, 44625, 188, 3, 23, "Text",ExpressionUUID->"5f8ba793-a9fd-41dc-880c-15c702c03305"],
Cell[1517237, 44630, 402, 9, 41, "DefinitionBox",ExpressionUUID->"042bc6a1-bf2e-4c9d-8de0-4cd372fdcac4"],
Cell[1517642, 44641, 92, 0, 23, "Text",ExpressionUUID->"12763ed4-61d4-4e40-964b-73a23bc64c99"],
Cell[CellGroupData[{
Cell[1517759, 44645, 240, 4, 56, "MathCaption",ExpressionUUID->"ecd54cd4-8210-4993-9546-26d11d4523c4"],
Cell[CellGroupData[{
Cell[1518024, 44653, 188, 5, 27, "Input",ExpressionUUID->"5f17f755-8b65-4164-8cd0-6eda88d8ab6c"],
Cell[1518215, 44660, 4117, 118, 32, "Output",ExpressionUUID->"e9dc0fa9-00a3-4051-8267-13675bf59338",
 CellID->1453228292]
}, Open  ]],
Cell[CellGroupData[{
Cell[1522369, 44783, 141, 3, 27, "Input",ExpressionUUID->"4fdf5386-509d-4588-8539-e9c44c873070"],
Cell[1522513, 44788, 225, 3, 26, "Output",ExpressionUUID->"e9b28141-bec0-4c75-b94d-49b74d88223a",
 CellID->1076386313]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1522787, 44797, 369, 9, 56, "MathCaption",ExpressionUUID->"a374b46d-75f8-4456-85b5-2a6596e2770a"],
Cell[CellGroupData[{
Cell[1523181, 44810, 456, 14, 27, "Input",ExpressionUUID->"88818591-43dd-4353-bd1d-ac8debaa3688"],
Cell[1523640, 44826, 14557, 408, 56, "Output",ExpressionUUID->"bc225fae-2bf5-4f18-82e4-0e1505e35232",
 CellID->422520263]
}, Open  ]],
Cell[CellGroupData[{
Cell[1538234, 45239, 155, 3, 27, "Input",ExpressionUUID->"9482811c-7de7-4619-aed0-5ff2ccf620bd"],
Cell[1538392, 45244, 10218, 289, 32, "Output",ExpressionUUID->"941e23b8-88b7-453f-ba7e-d58466c00d64",
 CellID->1717123259]
}, Open  ]],
Cell[CellGroupData[{
Cell[1548647, 45538, 145, 3, 27, "Input",ExpressionUUID->"e07eccbc-856e-4b23-891f-b899df7056f1"],
Cell[1548795, 45543, 260, 5, 26, "Output",ExpressionUUID->"89aa5028-5d27-46f5-aaa4-12292c9d07ad",
 CellID->1805365627]
}, Open  ]]
}, Open  ]],
Cell[1549082, 45552, 295, 7, 40, "Text",ExpressionUUID->"c2286cba-dc17-46e2-a31c-8c6c8b4801ef"],
Cell[CellGroupData[{
Cell[1549402, 45563, 238, 6, 40, "MathCaption",ExpressionUUID->"59f3ef4e-2783-46eb-b5a1-a33421c67f21"],
Cell[1549643, 45571, 190, 5, 27, "Input",ExpressionUUID->"2cca7070-0258-4ce4-a5cf-116d910740e3"],
Cell[CellGroupData[{
Cell[1549858, 45580, 137, 3, 27, "Input",ExpressionUUID->"f3d5ccc8-b95b-46a5-b960-19b5379e86b7"],
Cell[1549998, 45585, 199, 4, 26, "Output",ExpressionUUID->"6cb68dd5-61d7-41ac-8dd4-3a9bd8487d67",
 CellID->983644526]
}, Open  ]]
}, Open  ]],
Cell[1550224, 45593, 399, 8, 40, "Text",ExpressionUUID->"1781e635-fdf9-4e11-a44a-e6e9091bde9d"],
Cell[CellGroupData[{
Cell[1550648, 45605, 172, 5, 40, "MathCaption",ExpressionUUID->"f669974f-c03d-43b5-a7db-a69d54dbd525"],
Cell[CellGroupData[{
Cell[1550845, 45614, 198, 5, 27, "Input",ExpressionUUID->"5e7fc455-89dd-411e-a9af-8a6c5d3619ea"],
Cell[1551046, 45621, 487, 9, 22, "Message",ExpressionUUID->"42189f44-be8a-4a3e-9b41-b9c504d6f746",
 CellID->784433876]
}, Open  ]]
}, Open  ]],
Cell[1551560, 45634, 391, 6, 57, "Text",ExpressionUUID->"8166847b-0ae7-4bd8-8920-7609679d48f7"],
Cell[CellGroupData[{
Cell[1551976, 45644, 374, 7, 72, "MathCaption",ExpressionUUID->"07154e53-e5a9-41ee-9dc7-39cc51a2f82e"],
Cell[1552353, 45653, 456, 14, 27, "Input",ExpressionUUID->"f8804b8d-6d45-4e63-93e6-808a823c8d9c"],
Cell[CellGroupData[{
Cell[1552834, 45671, 189, 5, 27, "Input",ExpressionUUID->"9aaa997e-f6fa-44c8-b127-03409e217cee"],
Cell[1553026, 45678, 1214, 32, 32, "Output",ExpressionUUID->"cb19883c-8afc-4def-a9e6-7d4392fd0e22",
 CellID->578044999]
}, Open  ]],
Cell[CellGroupData[{
Cell[1554277, 45715, 189, 5, 27, "Input",ExpressionUUID->"6aa48fff-8363-47f9-b753-53fb5f4f8278"],
Cell[1554469, 45722, 537, 10, 22, "Message",ExpressionUUID->"881a1b66-d3e4-4ccf-bd7b-412100fec1ac",
 CellID->2133762812],
Cell[1555009, 45734, 1334, 36, 32, "Output",ExpressionUUID->"2f41f3ac-d0e4-4ee8-b917-17d77627840f",
 CellID->1322160031]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1556392, 45776, 218, 5, 40, "MathCaption",ExpressionUUID->"f6cd7ef1-3732-4114-9c6c-af95b3e95eb2"],
Cell[1556613, 45783, 456, 14, 27, "Input",ExpressionUUID->"283f45b1-7dd1-4fee-be9c-27b25af6fd4c"],
Cell[CellGroupData[{
Cell[1557094, 45801, 189, 5, 27, "Input",ExpressionUUID->"2be57b45-87a4-418b-8eea-a535109da820"],
Cell[1557286, 45808, 1217, 32, 32, "Output",ExpressionUUID->"3060b369-ee4e-4dfd-8c85-bb3470c233d3",
 CellID->1530345508]
}, Open  ]],
Cell[CellGroupData[{
Cell[1558540, 45845, 189, 5, 27, "Input",ExpressionUUID->"ef53ecb2-734e-45eb-b57e-3014a5105797"],
Cell[1558732, 45852, 1214, 32, 32, "Output",ExpressionUUID->"8bbe5dc6-f268-4bbf-ba46-37916afb122d",
 CellID->105312837]
}, Open  ]]
}, Open  ]],
Cell[1559973, 45888, 884, 27, 74, "Text",ExpressionUUID->"3696f86d-fcd6-4ee8-8f49-abf82beff2f2"],
Cell[1560860, 45917, 643, 17, 57, "Text",ExpressionUUID->"83c9e87e-91d2-4781-8bfc-d100ea675342"],
Cell[CellGroupData[{
Cell[1561528, 45938, 155, 5, 40, "MathCaption",ExpressionUUID->"5b81429c-a91e-4cf9-a82c-014270ce2d3e"],
Cell[1561686, 45945, 176, 4, 27, "Input",ExpressionUUID->"d2303095-f81e-42d7-850f-6e57cb8f7307"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1561899, 45954, 109, 0, 40, "MathCaption",ExpressionUUID->"109db051-9e08-45c3-88b5-d572cb5db11f"],
Cell[CellGroupData[{
Cell[1562033, 45958, 189, 5, 27, "Input",ExpressionUUID->"b9976076-3dd2-43f0-b068-51c3b1f5640d"],
Cell[1562225, 45965, 1166, 32, 32, "Output",ExpressionUUID->"b69e2613-b865-4c03-a329-ea36e7105cbb",
 CellID->466326331]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1563440, 46003, 407, 11, 56, "MathCaption",ExpressionUUID->"d2676e1e-ce9e-4746-908b-c4ca89c62a16"],
Cell[1563850, 46016, 136, 3, 27, "Input",ExpressionUUID->"9ffac273-6ff1-45ef-b9e0-4bf1ea6891d0"],
Cell[CellGroupData[{
Cell[1564011, 46023, 111, 2, 27, "Input",ExpressionUUID->"7ebfd0d9-92e8-4cb4-98f7-17e9dbefe819"],
Cell[1564125, 46027, 1215, 32, 32, "Output",ExpressionUUID->"c71784b1-493c-48b8-9902-4e328bfa052f",
 CellID->1378817214]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1565389, 46065, 86, 0, 40, "MathCaption",ExpressionUUID->"0d7dfc4e-6a8e-4a1f-b394-14aa36497f8d"],
Cell[1565478, 46067, 213, 6, 27, "Input",ExpressionUUID->"89f031cf-812d-4a3f-a905-aa87c9622072"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1565728, 46078, 150, 2, 40, "MathCaption",ExpressionUUID->"3f1c1679-755a-404e-8b12-135eabf41364"],
Cell[1565881, 46082, 176, 4, 27, "Input",ExpressionUUID->"945e2f72-7998-4ed2-b110-7984d0664f54"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1566106, 46092, 110, 0, 39, "Subsection",ExpressionUUID->"6d9737a9-7f5f-44c4-a03b-7edbf4a7b3a2"],
Cell[1566219, 46094, 503, 18, 40, "Text",ExpressionUUID->"7b2c5242-bfb1-4dca-ba70-43a75fc98651"],
Cell[1566725, 46114, 735, 18, 58, "DefinitionBox",ExpressionUUID->"4327af69-e03a-4d07-8889-188973731e98"],
Cell[1567463, 46134, 109, 0, 23, "Text",ExpressionUUID->"11094ee7-0193-4846-bfb6-a34d1100e80c"],
Cell[CellGroupData[{
Cell[1567597, 46138, 165, 5, 40, "MathCaption",ExpressionUUID->"e7b3aa73-c6fd-4b70-b1a0-112a9a71a331"],
Cell[CellGroupData[{
Cell[1567787, 46147, 323, 10, 27, "Input",ExpressionUUID->"37855b90-c630-4c23-8c40-b484a6440f23"],
Cell[1568113, 46159, 2051, 60, 32, "Output",ExpressionUUID->"95813ca7-af9e-44eb-9fbb-de17cd9f36c5",
 CellID->499350247]
}, Open  ]],
Cell[CellGroupData[{
Cell[1570201, 46224, 127, 3, 27, "Input",ExpressionUUID->"bbef108f-c27d-4f5b-b4a5-96c566cb06e6"],
Cell[CellGroupData[{
Cell[1570353, 46231, 196, 3, 43, "Print",ExpressionUUID->"81374258-63c6-4abc-8ad3-dfd7f1e48801",
 CellTags->"Info4123728981628-9181623",
 CellID->213508853],
Cell[1570552, 46236, 6879, 192, 339, "Print",ExpressionUUID->"93334796-27eb-4b34-88b2-aede27817ead",
 CellTags->"Info4123728981628-9181623",
 CellID->2145601881]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1577480, 46434, 137, 3, 27, "Input",ExpressionUUID->"361dfd15-763c-46fd-9025-a53e3b6cca3e"],
Cell[1577620, 46439, 2050, 60, 32, "Output",ExpressionUUID->"bfe14cf8-fbe5-4ea1-ab4c-ba6c1380a3c6",
 CellID->585413543]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1579719, 46505, 202, 3, 56, "MathCaption",ExpressionUUID->"ebbcebd6-78e1-490f-810d-125b46220498"],
Cell[CellGroupData[{
Cell[1579946, 46512, 233, 6, 27, "Input",ExpressionUUID->"719050be-1a15-4c48-b3bb-3d2e94b1dede"],
Cell[1580182, 46520, 693, 13, 22, "Print",ExpressionUUID->"12f33035-cfb2-4eb9-8472-a07135cd6cf1",
 CellID->1422830294]
}, Open  ]],
Cell[CellGroupData[{
Cell[1580912, 46538, 381, 11, 27, "Input",ExpressionUUID->"59e3cab6-e2d3-4d19-a4f8-a3479bc2cd5d"],
Cell[1581296, 46551, 1166, 32, 32, "Output",ExpressionUUID->"a89ee616-250a-4260-90f9-962208d9252a",
 CellID->1278341913]
}, Open  ]],
Cell[CellGroupData[{
Cell[1582499, 46588, 127, 3, 27, "Input",ExpressionUUID->"d555789b-a738-4b5c-bd72-7cc5eea70872"],
Cell[CellGroupData[{
Cell[1582651, 46595, 197, 3, 43, "Print",ExpressionUUID->"952d080a-d3f2-4c6b-b336-ee40ea619ec5",
 CellTags->"Info4163728981629-9181623",
 CellID->1666127203],
Cell[1582851, 46600, 6430, 181, 340, "Print",ExpressionUUID->"dbb65ac4-a485-4871-a20e-6176bd7da37d",
 CellTags->"Info4163728981629-9181623",
 CellID->1560627588]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1589330, 46787, 189, 5, 27, "Input",ExpressionUUID->"ae2b357d-12d1-4214-be5d-41cb3e71cfe2"],
Cell[1589522, 46794, 1166, 32, 32, "Output",ExpressionUUID->"d10f3e71-c106-4011-a990-24e0c00812f2",
 CellID->102607058]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1590737, 46832, 353, 11, 56, "MathCaption",ExpressionUUID->"8f72aec1-180c-4368-84eb-2340b268d358"],
Cell[1591093, 46845, 479, 16, 27, "Input",ExpressionUUID->"72953c91-ddf4-4c6e-a4da-d35fd4962b44"],
Cell[CellGroupData[{
Cell[1591597, 46865, 127, 3, 27, "Input",ExpressionUUID->"df876275-f259-4244-b3c9-5711b7f2a3ff"],
Cell[CellGroupData[{
Cell[1591749, 46872, 195, 3, 43, "Print",ExpressionUUID->"bf548868-f80a-4fce-add5-25287ccbf669",
 CellTags->"Info4193728981630-9181623",
 CellID->87455265],
Cell[1591947, 46877, 11528, 309, 383, "Print",ExpressionUUID->"061e13da-7d82-4a5b-9daf-db3bdf7c3cf0",
 CellTags->"Info4193728981630-9181623",
 CellID->981824307]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1603524, 47192, 227, 7, 27, "Input",ExpressionUUID->"5f408341-de08-4e05-adb0-66ca8b8fde3b"],
Cell[1603754, 47201, 5814, 172, 32, "Output",ExpressionUUID->"af48e65b-d8b9-4396-af7d-899bf8a7264b",
 CellID->256784201]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1609617, 47379, 313, 11, 40, "MathCaption",ExpressionUUID->"95ed0ed4-ae47-4af6-8a8a-7642709d1d46"],
Cell[CellGroupData[{
Cell[1609955, 47394, 407, 13, 27, "Input",ExpressionUUID->"f5b28f07-1a3d-4835-bda6-7d2ac68adc80"],
Cell[1610365, 47409, 3007, 88, 32, "Output",ExpressionUUID->"6f248225-79e0-46aa-ad12-b0405f682313",
 CellID->1477419384]
}, Open  ]],
Cell[CellGroupData[{
Cell[1613409, 47502, 227, 7, 27, "Input",ExpressionUUID->"be29d106-80c3-4119-bec0-652684241156"],
Cell[1613639, 47511, 8672, 256, 32, "Output",ExpressionUUID->"4fa72d53-38cb-4ce3-84dc-ad58b6884db5",
 CellID->1781660274]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1622360, 47773, 87, 0, 40, "MathCaption",ExpressionUUID->"c9560b09-6b59-4643-914b-27dc64fec181"],
Cell[1622450, 47775, 213, 6, 27, "Input",ExpressionUUID->"3973e3ce-62fb-4998-b3ad-79013d4812d9"],
Cell[1622666, 47783, 251, 8, 27, "Input",ExpressionUUID->"dc4c7c18-c64d-43be-9007-6f7d1fccb5a9"],
Cell[1622920, 47793, 157, 4, 27, "Input",ExpressionUUID->"56ed6a8f-6559-423d-ac00-fd86c2596098"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1623126, 47803, 96, 0, 39, "Subsection",ExpressionUUID->"0c54486b-ed81-4326-8c43-9a6c4133b556"],
Cell[1623225, 47805, 938, 20, 120, "Text",ExpressionUUID->"8ba0453c-a6e9-4c44-a05b-fb8abdd02d7a"],
Cell[CellGroupData[{
Cell[1624188, 47829, 113, 0, 40, "MathCaption",ExpressionUUID->"36ebbdd5-f694-4037-9fbb-2ce37999e5a6"],
Cell[1624304, 47831, 375, 11, 47, "Input",ExpressionUUID->"ee71972b-d025-4784-a9ed-a4f47641c5e1"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1624716, 47847, 247, 8, 40, "MathCaption",ExpressionUUID->"9f7bfa3c-2429-427e-9b88-14cfedef94df"],
Cell[1624966, 47857, 190, 5, 27, "Input",ExpressionUUID->"51206389-df84-434d-af94-22eea7f5f965"],
Cell[CellGroupData[{
Cell[1625181, 47866, 421, 12, 27, "Input",ExpressionUUID->"2fdb8634-e4cb-4c67-84f7-8d1ef0735729"],
Cell[1625605, 47880, 6017, 177, 32, "Output",ExpressionUUID->"cfd1c666-a5a0-4aea-a20b-6d4cd8f55fb2",
 CellID->1976281324]
}, Open  ]],
Cell[1631637, 48060, 157, 4, 27, "Input",ExpressionUUID->"a9c10f52-19f4-4eba-9fd8-2794b1c54b3f"]
}, Open  ]],
Cell[1631809, 48067, 149, 2, 23, "Text",ExpressionUUID->"ced5b990-150a-4c6e-a07f-9d8c3aa77145"],
Cell[1631961, 48071, 652, 17, 58, "DefinitionBox",ExpressionUUID->"437d8ce8-b970-4e81-bf18-418678b3c1d2"],
Cell[1632616, 48090, 174, 2, 23, "Text",ExpressionUUID->"b0ada00d-de8e-4014-95b5-62bf8e63d420",
 CellID->476091498],
Cell[CellGroupData[{
Cell[1632815, 48096, 115, 0, 40, "MathCaption",ExpressionUUID->"6d80eeeb-e0a8-4b40-a422-6660021980ea"],
Cell[1632933, 48098, 221, 6, 27, "Input",ExpressionUUID->"f8191a5e-9341-4167-b266-99f5dd10c4d7"],
Cell[CellGroupData[{
Cell[1633179, 48108, 421, 12, 27, "Input",ExpressionUUID->"6eb71841-a7ce-4c2a-9a56-4da149b6735b"],
Cell[1633603, 48122, 6018, 177, 32, "Output",ExpressionUUID->"f3caa6ad-2c8f-4224-a954-3576826c3415",
 CellID->230644319]
}, Open  ]],
Cell[1639636, 48302, 188, 5, 27, "Input",ExpressionUUID->"3556676a-c865-4ba1-9ff6-c13c1069c582"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1639861, 48312, 117, 0, 40, "MathCaption",ExpressionUUID->"2ef8cbf3-599a-44fa-b6bf-ce78be298b0d"],
Cell[1639981, 48314, 223, 6, 27, "Input",ExpressionUUID->"6d79c9ad-905e-4e66-9dce-bce66d072e6a"],
Cell[CellGroupData[{
Cell[1640229, 48324, 421, 12, 27, "Input",ExpressionUUID->"a44699cc-b758-4c40-a3a6-b88b5e843a7f"],
Cell[1640653, 48338, 6018, 177, 32, "Output",ExpressionUUID->"6950d25d-e27a-4fec-b863-cd2d6bbb0b10",
 CellID->231086598]
}, Open  ]],
Cell[1646686, 48518, 190, 5, 27, "Input",ExpressionUUID->"c95f5328-d72c-41a9-90cd-42cf3252f5c3"]
}, Open  ]],
Cell[1646891, 48526, 128, 0, 23, "Text",ExpressionUUID->"ae6efe61-4ecd-4f28-a510-fb36a8833e9d"],
Cell[1647022, 48528, 3159, 86, 194, "DefinitionBox",ExpressionUUID->"26e7a798-4ea3-4aa1-8ead-05da3633526f"],
Cell[1650184, 48616, 95, 0, 23, "Text",ExpressionUUID->"fc651664-c941-41b4-990b-e5357d96a4e7"],
Cell[CellGroupData[{
Cell[1650304, 48620, 132, 0, 40, "MathCaption",ExpressionUUID->"f097ca9d-1f91-4029-8828-7f1301905571"],
Cell[CellGroupData[{
Cell[1650461, 48624, 273, 7, 27, "Input",ExpressionUUID->"2e8184f9-4f2c-4c03-a110-6bd62f359a5e"],
Cell[CellGroupData[{
Cell[1650759, 48635, 594, 12, 22, "Print",ExpressionUUID->"097a9043-b4e6-4c98-b606-709277669351",
 CellID->1222018810],
Cell[1651356, 48649, 604, 12, 22, "Print",ExpressionUUID->"fdc598e1-8ae8-49f1-8efe-9c10804be408",
 CellID->2084023409]
}, Open  ]],
Cell[1651975, 48664, 450, 9, 22, "Message",ExpressionUUID->"dd97c8dc-ec88-43d8-8027-81821787580f",
 CellID->723152493],
Cell[1652428, 48675, 453, 9, 22, "Message",ExpressionUUID->"395c56a3-76b1-4355-a9b5-25a0a7dc5895",
 CellID->1413991478]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1652930, 48690, 153, 2, 40, "MathCaption",ExpressionUUID->"55dba7d2-f262-46ac-b27a-ce6e66f3fda4"],
Cell[1653086, 48694, 220, 6, 27, "Input",ExpressionUUID->"ceeb0815-22b4-4413-bf72-b4289fa91eb5"],
Cell[CellGroupData[{
Cell[1653331, 48704, 421, 12, 27, "Input",ExpressionUUID->"4d6a4570-e9ac-4be1-87db-6ca78a5af46b"],
Cell[1653755, 48718, 6018, 177, 32, "Output",ExpressionUUID->"8ab0d10b-cf74-4c17-b7fb-940c9e399e02",
 CellID->900733044]
}, Open  ]],
Cell[1659788, 48898, 187, 5, 27, "Input",ExpressionUUID->"af2befa4-3824-4379-ad24-dcb58bae11d4"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1660012, 48908, 177, 3, 40, "MathCaption",ExpressionUUID->"bb7a1e4d-17cb-4293-82bc-5ac4a8743215"],
Cell[1660192, 48913, 196, 5, 27, "Input",ExpressionUUID->"de2b45e8-8957-4b19-86cb-35f305b3c619"],
Cell[CellGroupData[{
Cell[1660413, 48922, 421, 12, 27, "Input",ExpressionUUID->"e699b822-b73b-4d8d-85c2-cb171a6aeab1"],
Cell[1660837, 48936, 6019, 177, 32, "Output",ExpressionUUID->"f8d04a7b-baea-4679-a505-2e8fade1b924",
 CellID->1008042805]
}, Open  ]],
Cell[1666871, 49116, 163, 4, 27, "Input",ExpressionUUID->"0c73b946-d8f4-4fc7-ba87-6bd7af16f156"],
Cell[1667037, 49122, 234, 7, 27, "Input",ExpressionUUID->"1d1fb3d3-21d9-4985-9650-67f51d31efd0"],
Cell[CellGroupData[{
Cell[1667296, 49133, 421, 12, 27, "Input",ExpressionUUID->"22bda2a5-71dd-443a-af89-31a674e8cdcd"],
Cell[1667720, 49147, 6016, 177, 32, "Output",ExpressionUUID->"763c1130-b0df-4acc-9bde-1dc77d29e6ff",
 CellID->676932452]
}, Open  ]],
Cell[1673751, 49327, 182, 5, 27, "Input",ExpressionUUID->"a51dd1a3-e92a-481f-ba3a-0da8407ba1dd"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1673970, 49337, 307, 7, 56, "MathCaption",ExpressionUUID->"016a86b5-b99f-403a-8e40-59ca1e34ae16"],
Cell[1674280, 49346, 220, 6, 27, "Input",ExpressionUUID->"fbc860d7-c346-4310-9a6f-507c2828e1aa"],
Cell[CellGroupData[{
Cell[1674525, 49356, 421, 12, 27, "Input",ExpressionUUID->"d52a7bd1-911b-4062-ad06-ea612be3053f"],
Cell[1674949, 49370, 428, 9, 22, "Message",ExpressionUUID->"261dd749-52bc-45b7-ab25-d739655831e6",
 CellID->1982842605],
Cell[1675380, 49381, 6016, 177, 32, "Output",ExpressionUUID->"54c578a0-7f03-475c-b658-7bdc7bef19fc",
 CellID->984428089]
}, Open  ]],
Cell[1681411, 49561, 187, 5, 27, "Input",ExpressionUUID->"675daec1-5fcb-441d-a950-f12b85f76ed5"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1681635, 49571, 190, 3, 40, "MathCaption",ExpressionUUID->"cde8af92-bff1-4610-913d-58239f8c5e1f"],
Cell[1681828, 49576, 193, 5, 27, "Input",ExpressionUUID->"414e6fb1-2755-4ff6-b191-c75a6d63601c"],
Cell[CellGroupData[{
Cell[1682046, 49585, 421, 12, 27, "Input",ExpressionUUID->"41be1a05-e5eb-4253-bbdc-ec4fdfaf264b"],
Cell[1682470, 49599, 6019, 177, 32, "Output",ExpressionUUID->"2a8cc52d-5df6-4548-8a3c-459488dd905d",
 CellID->1068875335]
}, Open  ]],
Cell[1688504, 49779, 160, 4, 27, "Input",ExpressionUUID->"a93043b7-51ff-48fc-a241-bd8960979fb9"]
}, Open  ]],
Cell[1688679, 49786, 456, 18, 40, "Text",ExpressionUUID->"f58db9c6-609f-43ea-819c-35a4c2e5d203"],
Cell[CellGroupData[{
Cell[1689160, 49808, 194, 5, 40, "MathCaption",ExpressionUUID->"d7c7522f-2601-41d8-afe1-f69345c711da"],
Cell[1689357, 49815, 334, 10, 27, "Input",ExpressionUUID->"ff8fec80-16cd-4006-9898-cc92318a64d7"],
Cell[CellGroupData[{
Cell[1689716, 49829, 421, 12, 27, "Input",ExpressionUUID->"449d2114-6de5-4c6e-b869-afc49d75a5bb"],
Cell[1690140, 49843, 6018, 177, 32, "Output",ExpressionUUID->"914596a6-bab1-4d39-bf67-cc850fabcafe",
 CellID->887048190]
}, Open  ]],
Cell[1696173, 50023, 304, 10, 27, "Input",ExpressionUUID->"71b035a4-9bc2-47fe-9fa5-7d8442c8064e"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1696514, 50038, 404, 12, 56, "MathCaption",ExpressionUUID->"5d54706a-9b7b-4705-9639-377f3057507e"],
Cell[1696921, 50052, 224, 6, 27, "Input",ExpressionUUID->"64f457e2-1123-4be9-ae15-93052e29b3d2"],
Cell[CellGroupData[{
Cell[1697170, 50062, 421, 12, 27, "Input",ExpressionUUID->"a7c96a73-0f8d-4265-b6f9-fa09903ce333"],
Cell[1697594, 50076, 6016, 177, 32, "Output",ExpressionUUID->"ec5252f5-6523-49f9-9f4b-37d352d558d8",
 CellID->787451856]
}, Open  ]],
Cell[1703625, 50256, 191, 5, 27, "Input",ExpressionUUID->"c4ca18e4-920a-46dd-973b-8b0a1f691b11"],
Cell[1703819, 50263, 263, 8, 27, "Input",ExpressionUUID->"66b9f7f1-d2ee-4f9c-b00f-1774528df694"],
Cell[CellGroupData[{
Cell[1704107, 50275, 421, 12, 27, "Input",ExpressionUUID->"6763dab0-c483-4296-bb3e-f37361f73b5f"],
Cell[1704531, 50289, 6017, 177, 32, "Output",ExpressionUUID->"6c919914-0917-493f-9fb7-823f010f2626",
 CellID->1909110978]
}, Open  ]],
Cell[1710563, 50469, 211, 6, 27, "Input",ExpressionUUID->"1bc5dbfa-9ab3-4214-96b5-532027cfb0dc"],
Cell[1710777, 50477, 226, 6, 27, "Input",ExpressionUUID->"3f196dba-7207-4e59-bd2d-2bc0e64d94bf"],
Cell[CellGroupData[{
Cell[1711028, 50487, 421, 12, 27, "Input",ExpressionUUID->"eb505c38-109a-4746-941e-68888f66c9b6"],
Cell[1711452, 50501, 6016, 177, 32, "Output",ExpressionUUID->"9da9f4fc-ecbe-4013-809e-a5f066521541",
 CellID->875037419]
}, Open  ]],
Cell[1717483, 50681, 193, 5, 27, "Input",ExpressionUUID->"da793f29-7a78-46a4-999f-a01941b6d611"]
}, Open  ]],
Cell[1717691, 50689, 226, 5, 23, "Text",ExpressionUUID->"0f9d9002-5a02-4b69-969a-2cec7583d11f"],
Cell[CellGroupData[{
Cell[1717942, 50698, 380, 9, 56, "MathCaption",ExpressionUUID->"486f14ab-f014-470b-b191-62914509d30a"],
Cell[CellGroupData[{
Cell[1718347, 50711, 176, 4, 27, "Input",ExpressionUUID->"d825c4b2-4ee6-421c-bad2-281fc20dafc2"],
Cell[1718526, 50717, 199, 4, 26, "Output",ExpressionUUID->"d6064a0c-1809-4bdd-9b09-19620bced63e",
 CellID->785886477]
}, Open  ]],
Cell[CellGroupData[{
Cell[1718762, 50726, 187, 4, 27, "Input",ExpressionUUID->"02a3e6fb-14da-47e1-be5a-b3d472ee38ac"],
Cell[1718952, 50732, 206, 4, 26, "Output",ExpressionUUID->"8d526ad7-e5fd-4411-aac0-f8d24bfea9ed",
 CellID->1369589889]
}, Open  ]],
Cell[CellGroupData[{
Cell[1719195, 50741, 209, 5, 27, "Input",ExpressionUUID->"6f39331c-df82-42b1-9c6d-d20952183bfc"],
Cell[1719407, 50748, 227, 5, 26, "Output",ExpressionUUID->"36a2b13e-4301-49ba-a41b-ce5d82c80da9",
 CellID->1967535877]
}, Open  ]],
Cell[CellGroupData[{
Cell[1719671, 50758, 209, 5, 27, "Input",ExpressionUUID->"f802c802-505b-454f-833d-07361b5af857"],
Cell[1719883, 50765, 206, 4, 26, "Output",ExpressionUUID->"de4ff2d0-292a-4e95-8a7c-44e26ceb235f",
 CellID->1422288233]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1720138, 50775, 189, 5, 40, "MathCaption",ExpressionUUID->"921b65c1-4d65-4a34-b51b-8623031b4be6"],
Cell[CellGroupData[{
Cell[1720352, 50784, 176, 4, 27, "Input",ExpressionUUID->"8887fe1b-9677-4145-9b40-91fbdf150e07"],
Cell[1720531, 50790, 174, 3, 26, "Output",ExpressionUUID->"1200ff46-45a3-4566-add6-8b6a40c2e49a",
 CellID->757857274]
}, Open  ]],
Cell[CellGroupData[{
Cell[1720742, 50798, 187, 4, 27, "Input",ExpressionUUID->"0bd9852b-d71a-4712-8149-9e34182c36b7"],
Cell[1720932, 50804, 205, 4, 26, "Output",ExpressionUUID->"6ffa627e-603f-4c34-af4d-296c562ace22",
 CellID->581693761]
}, Open  ]],
Cell[CellGroupData[{
Cell[1721174, 50813, 209, 5, 27, "Input",ExpressionUUID->"bbc3e3fe-75e1-45a4-97bc-46d3a3599e20"],
Cell[1721386, 50820, 559, 16, 26, "Output",ExpressionUUID->"840170f6-ef0e-4256-b985-368afb225bc3",
 CellID->1181479983]
}, Open  ]],
Cell[CellGroupData[{
Cell[1721982, 50841, 209, 5, 27, "Input",ExpressionUUID->"7e8f09dc-70d2-493b-82bb-4775106db488"],
Cell[1722194, 50848, 486, 14, 26, "Output",ExpressionUUID->"82a847c8-06cb-47f8-a830-c5a1459823d2",
 CellID->1500274119]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1722741, 50869, 112, 0, 39, "Subsection",ExpressionUUID->"d7e8ca92-13b6-4601-8c16-25cbe48767af"],
Cell[1722856, 50871, 587, 16, 57, "Text",ExpressionUUID->"84a1dbae-2441-42ae-892b-e338260ed0d2"],
Cell[1723446, 50889, 752, 19, 58, "DefinitionBox",ExpressionUUID->"e72a0d6e-eec2-4bef-be6e-a5d2b4c1bd5e"],
Cell[1724201, 50910, 111, 0, 23, "Text",ExpressionUUID->"d67261fd-ffb6-4400-9b20-7740fbf39f9d"],
Cell[CellGroupData[{
Cell[1724337, 50914, 191, 5, 40, "MathCaption",ExpressionUUID->"da3aa5e8-9fc6-4f2f-8e48-05638e8ac8dd"],
Cell[CellGroupData[{
Cell[1724553, 50923, 425, 14, 27, "Input",ExpressionUUID->"a375752c-54fa-4c2b-a5bd-0965de4fd5a1"],
Cell[1724981, 50939, 426, 9, 22, "Message",ExpressionUUID->"798d69f3-0291-490a-b1ce-bffe04affaa5",
 CellID->11055515],
Cell[1725410, 50950, 420, 9, 22, "Message",ExpressionUUID->"e6c27563-0f1b-48cf-a6ca-a685368e9c9a",
 CellID->1126446963],
Cell[1725833, 50961, 236, 5, 26, "Output",ExpressionUUID->"3bce8da1-fbfa-47f5-8f67-e4ddba866a24",
 CellID->2125758836]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1726118, 50972, 177, 5, 40, "MathCaption",ExpressionUUID->"f83008dc-e96c-4728-a8b6-702d21631da1"],
Cell[CellGroupData[{
Cell[1726320, 50981, 462, 15, 27, "Input",ExpressionUUID->"e15ee38b-1700-4870-a790-75c121f26279"],
Cell[1726785, 50998, 4206, 119, 32, "Output",ExpressionUUID->"46317b68-5bc7-4aca-bbb0-f466327203c2",
 CellID->2090805160]
}, Open  ]],
Cell[CellGroupData[{
Cell[1731028, 51122, 434, 14, 27, "Input",ExpressionUUID->"69df4d90-cd0e-4ab8-8524-02bdf1adc3cd"],
Cell[1731465, 51138, 4205, 119, 32, "Output",ExpressionUUID->"9c4661ce-4fe6-4c03-849c-aeb92dbc6cf5",
 CellID->887408259]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1735731, 51264, 90, 0, 39, "Subsection",ExpressionUUID->"80e795fd-f232-4a73-87ac-aff5ef846f9f"],
Cell[1735824, 51266, 874, 19, 91, "Text",ExpressionUUID->"8909e5f1-e7bc-423f-8695-d8514a0aa07a"]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1736747, 51291, 818, 14, 54, "Section",ExpressionUUID->"4e72983f-5bfb-4fb2-bf7e-973e9afda9de",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1740316966],
Cell[CellGroupData[{
Cell[1737590, 51309, 103, 0, 39, "Subsection",ExpressionUUID->"a690d2bf-571c-4ad2-aa9f-9984e4e62f1a"],
Cell[1737696, 51311, 878, 19, 91, "Text",ExpressionUUID->"133463b9-e5a9-4187-9a52-093c6ba680ac"],
Cell[1738577, 51332, 1351, 37, 92, "DefinitionBox",ExpressionUUID->"8afd3f5a-0cca-49d9-b74a-c2ccb0401d51"],
Cell[1739931, 51371, 108, 0, 23, "Text",ExpressionUUID->"c9e6948a-71d7-4eef-a215-a43f2eb97815"],
Cell[CellGroupData[{
Cell[1740064, 51375, 166, 5, 40, "MathCaption",ExpressionUUID->"37df5ee4-d77f-4d53-9d14-9ce3c78f7d3d"],
Cell[CellGroupData[{
Cell[1740255, 51384, 149, 3, 27, "Input",ExpressionUUID->"0d52f1ba-3a86-45dc-b1a1-3ea99384db07"],
Cell[1740407, 51389, 1144, 27, 84, "Output",ExpressionUUID->"5fde7438-599a-47eb-94ae-61491e770b22",
 CellID->492296792]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1741600, 51422, 708, 15, 104, "MathCaption",ExpressionUUID->"afdb5d09-70d7-48ef-83b5-3b3832386b4d"],
Cell[CellGroupData[{
Cell[1742333, 51441, 191, 5, 27, "Input",ExpressionUUID->"79183604-22b4-4485-b58e-7b98759e8bd9"],
Cell[CellGroupData[{
Cell[1742549, 51450, 682, 15, 22, "Print",ExpressionUUID->"1bd42072-895c-4893-9579-052efcab4f94",
 CellID->1925174900],
Cell[1743234, 51467, 801, 18, 22, "Print",ExpressionUUID->"dd8ccbb1-f904-4f1d-8930-f7f960053c54",
 CellID->278434951],
Cell[1744038, 51487, 817, 19, 22, "Print",ExpressionUUID->"3e7f4cf6-ed3b-4187-8bad-3892757cf16c",
 CellID->289354644],
Cell[1744858, 51508, 896, 18, 42, "Print",ExpressionUUID->"4a65eff7-186e-4377-800e-d3f9eb9ab1af",
 CellID->1280524900],
Cell[1745757, 51528, 770, 17, 22, "Print",ExpressionUUID->"7887f91e-c263-48b9-adde-e21ed590c047",
 CellID->1848367713],
Cell[1746530, 51547, 269, 4, 22, "Print",ExpressionUUID->"b0a29026-525e-405b-b429-133e8febea5b",
 CellID->17714889]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1746848, 51557, 165, 3, 27, "Input",ExpressionUUID->"d97aae0d-3cac-49a7-84fa-b0f724f7dcb8"],
Cell[1747016, 51562, 525, 16, 26, "Output",ExpressionUUID->"028a63d1-6afa-4956-a580-414be978afa7",
 CellID->1381396137]
}, Open  ]],
Cell[CellGroupData[{
Cell[1747578, 51583, 245, 7, 27, "Input",ExpressionUUID->"b0981e3f-f5e6-45d6-ad7d-84ea6f2826fe"],
Cell[1747826, 51592, 1143, 32, 32, "Output",ExpressionUUID->"805cbad3-a5aa-4714-9547-9c6b8d157ddc",
 CellID->679166842]
}, Open  ]],
Cell[CellGroupData[{
Cell[1749006, 51629, 163, 3, 27, "Input",ExpressionUUID->"ac559eb5-88b0-4554-b0dd-5cd1b22a19ef"],
Cell[1749172, 51634, 281, 7, 26, "Output",ExpressionUUID->"207ef9a5-0164-4b6d-a4a4-d90167bc7102",
 CellID->881529689]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1749502, 51647, 430, 9, 72, "MathCaption",ExpressionUUID->"38520613-bd95-4dd7-a01a-908a5bf57002"],
Cell[CellGroupData[{
Cell[1749957, 51660, 176, 4, 27, "Input",ExpressionUUID->"2acaf54c-fa83-4202-a855-3a7e0fdb506f"],
Cell[1750136, 51666, 178, 3, 26, "Output",ExpressionUUID->"f644bc4c-abe0-4084-81fc-3ed46b004f2c",
 CellID->379802370]
}, Open  ]],
Cell[CellGroupData[{
Cell[1750351, 51674, 182, 4, 27, "Input",ExpressionUUID->"d53cfcbe-78ca-4320-8290-81b302a3ec04"],
Cell[1750536, 51680, 201, 3, 26, "Output",ExpressionUUID->"8ca67209-9fcc-4ed1-86e7-53c2213387bc",
 CellID->1017033706]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1750786, 51689, 544, 17, 72, "MathCaption",ExpressionUUID->"07a44507-6585-406e-b64b-2ae3ad854432"],
Cell[CellGroupData[{
Cell[1751355, 51710, 120, 2, 27, "Input",ExpressionUUID->"50150507-d5b2-4dc0-9646-99f049be0933"],
Cell[1751478, 51714, 184, 3, 26, "Output",ExpressionUUID->"cc0dfa08-bb35-4f7c-a270-6d20d5d18428",
 CellID->1463150269]
}, Open  ]],
Cell[CellGroupData[{
Cell[1751699, 51722, 267, 8, 27, "Input",ExpressionUUID->"6f8d56ff-52ad-400b-975f-5aff724e4aac"],
Cell[1751969, 51732, 1232, 34, 32, "Output",ExpressionUUID->"090e344f-6d27-4300-ae89-5a71cbdae147",
 CellID->331329286]
}, Open  ]],
Cell[1753216, 51769, 174, 4, 27, "Input",ExpressionUUID->"fbf14784-23b7-4418-914c-73a37faf7c68"],
Cell[CellGroupData[{
Cell[1753415, 51777, 267, 8, 27, "Input",ExpressionUUID->"3392d8b8-a285-48ad-902e-9b64c82f47aa"],
Cell[1753685, 51787, 1186, 33, 34, "Output",ExpressionUUID->"919f223d-83df-43cd-8e09-362719f0a3bb",
 CellID->226213131]
}, Open  ]],
Cell[CellGroupData[{
Cell[1754908, 51825, 309, 9, 27, "Input",ExpressionUUID->"b396d441-e5a0-4ddb-8f30-c25b4e37cbde"],
Cell[1755220, 51836, 1275, 36, 32, "Output",ExpressionUUID->"cdb08fd9-acb9-4ef6-86f5-dcedc088c7a4",
 CellID->918037527]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1756544, 51878, 267, 5, 56, "MathCaption",ExpressionUUID->"7ddcc178-4ed1-47c5-b794-7ad914bdcba3"],
Cell[CellGroupData[{
Cell[1756836, 51887, 193, 5, 27, "Input",ExpressionUUID->"e6260682-f6b7-489c-a22c-41e8e5982d10"],
Cell[1757032, 51894, 609, 12, 22, "Print",ExpressionUUID->"afd6daf7-a429-452a-a5b5-896fa560222f",
 CellID->1934196241]
}, Open  ]],
Cell[CellGroupData[{
Cell[1757678, 51911, 413, 14, 27, "Input",ExpressionUUID->"ffa63eb5-6b01-43dc-9205-460059e34ac3"],
Cell[1758094, 51927, 6957, 201, 34, "Output",ExpressionUUID->"246a090c-410f-4981-bd83-ef2ef0a5771a",
 CellID->1576257848]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1765100, 52134, 170, 3, 40, "MathCaption",ExpressionUUID->"b6ffecf8-5711-4d7b-9d33-f07334d74bc2"],
Cell[CellGroupData[{
Cell[1765295, 52141, 460, 15, 27, "Input",ExpressionUUID->"dba33b63-46ec-4d6f-a1e2-3175425f57df"],
Cell[1765758, 52158, 4437, 123, 34, "Output",ExpressionUUID->"8775fe9d-80df-4362-a95f-7acbc7bb21ce",
 CellID->807381944]
}, Open  ]]
}, Open  ]],
Cell[1770222, 52285, 1333, 34, 101, "DefinitionBox",ExpressionUUID->"f65163c3-72a3-4923-81ba-ff7d40cdd0fe"],
Cell[1771558, 52321, 98, 0, 23, "Text",ExpressionUUID->"e22ac9f9-cb6a-4b37-82c8-f23370c6f1a4"],
Cell[CellGroupData[{
Cell[1771681, 52325, 493, 8, 88, "MathCaption",ExpressionUUID->"6545f5fa-7433-469b-9daf-e26376c5d2e5"],
Cell[CellGroupData[{
Cell[1772199, 52337, 348, 11, 27, "Input",ExpressionUUID->"817c56c9-5373-493c-aa9e-c20eebee77c0"],
Cell[1772550, 52350, 1263, 35, 32, "Output",ExpressionUUID->"0c12541e-152d-4186-8882-b25d2d510593",
 CellID->1866651062]
}, Open  ]],
Cell[1773828, 52388, 173, 4, 27, "Input",ExpressionUUID->"e451f13c-3f44-48b8-a035-2f42166c3781"],
Cell[CellGroupData[{
Cell[1774026, 52396, 348, 11, 27, "Input",ExpressionUUID->"2d69238a-5569-4699-b795-867d9f930362"],
Cell[1774377, 52409, 1301, 36, 32, "Output",ExpressionUUID->"5a8f21fc-a918-4de0-9f66-8a04d2768034",
 CellID->215800011]
}, Open  ]],
Cell[CellGroupData[{
Cell[1775715, 52450, 626, 21, 27, "Input",ExpressionUUID->"a07b2027-2e3f-4f6e-bc3c-9fb68c4c6a95"],
Cell[1776344, 52473, 2523, 70, 32, "Output",ExpressionUUID->"be22bbe1-fb5e-4017-bc50-9ec38c5ab479",
 CellID->1998388741]
}, Open  ]],
Cell[CellGroupData[{
Cell[1778904, 52548, 141, 3, 27, "Input",ExpressionUUID->"c6d10605-96fd-4ff9-b77d-d740b09db4a0"],
Cell[1779048, 52553, 172, 3, 26, "Output",ExpressionUUID->"b7fcd682-e328-4698-b68d-15bb164da53e",
 CellID->275330907]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1779269, 52562, 125, 0, 40, "MathCaption",ExpressionUUID->"f892c410-0d1a-43e7-a4d5-daa994c70d5b"],
Cell[1779397, 52564, 151, 3, 27, "Input",ExpressionUUID->"8fa5c7c0-c485-4d3f-b3df-72a62097e3d1"],
Cell[CellGroupData[{
Cell[1779573, 52571, 626, 21, 27, "Input",ExpressionUUID->"6f1d57b6-1f7b-403f-ad59-c8d4bf9a7ad2"],
Cell[1780202, 52594, 171, 3, 26, "Output",ExpressionUUID->"2af75073-88e2-4f83-a4e7-00419b4aa9d8",
 CellID->1806778311]
}, Open  ]],
Cell[CellGroupData[{
Cell[1780410, 52602, 150, 3, 27, "Input",ExpressionUUID->"62d34abf-293e-43a8-a5f8-476d69a2b843"],
Cell[1780563, 52607, 249, 4, 22, "Print",ExpressionUUID->"94d607ee-f870-4df3-902f-f1480707a58a",
 CellID->1104304397]
}, Open  ]]
}, Open  ]],
Cell[1780839, 52615, 299, 6, 56, "MathCaption",ExpressionUUID->"87dc076f-3fd8-4227-aa28-3d8c46d85766"],
Cell[1781141, 52623, 434, 10, 40, "Text",ExpressionUUID->"0c136a5f-4137-4d8a-83ed-c7c82ab6723f"],
Cell[CellGroupData[{
Cell[1781600, 52637, 139, 1, 40, "MathCaption",ExpressionUUID->"ae199384-d3a8-499c-b3df-56931f3c3385"],
Cell[CellGroupData[{
Cell[1781764, 52642, 432, 14, 27, "Input",ExpressionUUID->"aa5e924f-2bbe-4966-b485-0d530a484af7"],
Cell[1782199, 52658, 1384, 38, 32, "Output",ExpressionUUID->"f4f2fb1d-b3a8-4cf4-bb02-384925fdd669",
 CellID->1490323005]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1783632, 52702, 182, 5, 40, "MathCaption",ExpressionUUID->"86d2b13f-db6c-4d81-876a-7cbace375d4e"],
Cell[CellGroupData[{
Cell[1783839, 52711, 417, 14, 27, "Input",ExpressionUUID->"598ec24d-3e7b-458a-9027-e398b0c8f8a7"],
Cell[1784259, 52727, 1384, 38, 32, "Output",ExpressionUUID->"f2830ff8-bb57-4961-a302-d2d3ed13e7ca",
 CellID->1756970074]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1785692, 52771, 333, 7, 56, "MathCaption",ExpressionUUID->"2115c7f8-e577-4ce7-bd2e-fbdf324e6ec2"],
Cell[1786028, 52780, 164, 4, 27, "Input",ExpressionUUID->"ae8b5f7c-73a3-4ea9-ae06-4955efc4d9f1"],
Cell[1786195, 52786, 491, 14, 47, "Input",ExpressionUUID->"09cf970e-1e3a-475a-aae7-94d2903187f5"],
Cell[CellGroupData[{
Cell[1786711, 52804, 323, 10, 27, "Input",ExpressionUUID->"cb12f830-4d59-4eb9-bfe1-d7f00eb04e5b"],
Cell[1787037, 52816, 1384, 38, 32, "Output",ExpressionUUID->"22af3180-353d-45ce-a6df-0041138b3307",
 CellID->1495887468]
}, Open  ]],
Cell[CellGroupData[{
Cell[1788458, 52859, 145, 3, 27, "Input",ExpressionUUID->"cd871c40-4d76-4629-bd46-2e366e656c92"],
Cell[1788606, 52864, 208, 3, 26, "Output",ExpressionUUID->"9e7821dd-d848-4067-8f29-5cc29d521e47",
 CellID->789619247]
}, Open  ]],
Cell[1788829, 52870, 307, 9, 47, "Input",ExpressionUUID->"f0c68c66-1bc1-451a-96d4-7d637db6ec55"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1789173, 52884, 137, 1, 40, "MathCaption",ExpressionUUID->"f96241e0-3cdd-463a-9627-012c631ae4a9"],
Cell[1789313, 52887, 344, 10, 27, "Input",ExpressionUUID->"76e0b143-41c9-4db1-bc86-5d6a022f05ce"],
Cell[CellGroupData[{
Cell[1789682, 52901, 322, 10, 27, "Input",ExpressionUUID->"7b61996f-0875-47e9-b8fa-abd87cfe5b2c"],
Cell[1790007, 52913, 1381, 38, 32, "Output",ExpressionUUID->"a1fd4b0e-8415-48a6-b1e9-4d010ae7d5a1",
 CellID->1829229516]
}, Open  ]],
Cell[CellGroupData[{
Cell[1791425, 52956, 145, 3, 27, "Input",ExpressionUUID->"85d1f6dc-80e6-4f0d-b79c-37a788d7dac4"],
Cell[1791573, 52961, 205, 3, 26, "Output",ExpressionUUID->"6008a6ed-d7a1-4b74-ac5f-e37b8abdaaf3",
 CellID->484584960]
}, Open  ]],
Cell[1791793, 52967, 218, 6, 27, "Input",ExpressionUUID->"e4d458cf-7efc-467c-8751-1cdcb9aaa662"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1792048, 52978, 115, 0, 40, "MathCaption",ExpressionUUID->"4098e5a8-e662-4cbe-be9d-c5af2811a9c6"],
Cell[1792166, 52980, 381, 10, 27, "Input",ExpressionUUID->"2d0b9916-e868-43ec-890c-c972051da3df"],
Cell[CellGroupData[{
Cell[1792572, 52994, 369, 11, 27, "Input",ExpressionUUID->"9057620d-db5a-4ddc-a2a0-c8a8cd4ee83b"],
Cell[1792944, 53007, 1382, 38, 32, "Output",ExpressionUUID->"fe840463-c02e-4e43-af91-ea2351e2e015",
 CellID->1161433296]
}, Open  ]],
Cell[CellGroupData[{
Cell[1794363, 53050, 145, 3, 27, "Input",ExpressionUUID->"63f5cfad-18f7-4a59-95e8-0cc0a20988bb"],
Cell[1794511, 53055, 209, 3, 26, "Output",ExpressionUUID->"49146ef4-877b-4f11-a6d9-240c20e3d7e2",
 CellID->1730538289]
}, Open  ]],
Cell[1794735, 53061, 291, 8, 27, "Input",ExpressionUUID->"839585ee-de19-4e04-a256-5bd736244d5b"],
Cell[1795029, 53071, 162, 4, 27, "Input",ExpressionUUID->"bf8281a6-bd4e-4859-a39f-4cc71ff38c03"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1795240, 53081, 110, 0, 39, "Subsection",ExpressionUUID->"623c0ab1-7911-4324-a525-e01e0bc7d55a"],
Cell[1795353, 53083, 987, 20, 108, "Text",ExpressionUUID->"eb26b2ac-882a-4352-ba13-ad0d7d6c506b"],
Cell[CellGroupData[{
Cell[1796365, 53107, 244, 5, 40, "MathCaption",ExpressionUUID->"dcf035c9-a4a6-4cd0-ada7-7123834233bf"],
Cell[CellGroupData[{
Cell[1796634, 53116, 304, 9, 27, "Input",ExpressionUUID->"e12f2e99-61bb-4339-9c6a-c7aea3b99ca1"],
Cell[1796941, 53127, 258, 6, 26, "Output",ExpressionUUID->"941f0f12-57b8-4718-8ee9-11cf6d6c0ea5",
 CellID->924892270]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1797248, 53139, 97, 0, 40, "MathCaption",ExpressionUUID->"2ab14c93-829b-4394-bee6-c86593cd9568"],
Cell[CellGroupData[{
Cell[1797370, 53143, 246, 7, 27, "Input",ExpressionUUID->"56ab050b-b341-4396-9fa1-24241c7b76bd"],
Cell[1797619, 53152, 314, 10, 26, "Output",ExpressionUUID->"241ad05e-7aee-44e0-be72-8d1d2fe72103",
 CellID->1177744904]
}, Open  ]]
}, Open  ]],
Cell[1797960, 53166, 468, 9, 57, "Text",ExpressionUUID->"e031fd1e-5827-4acf-9041-55f07c2bd855"],
Cell[CellGroupData[{
Cell[1798453, 53179, 124, 0, 40, "MathCaption",ExpressionUUID->"76c11376-99af-475d-bdfd-1aa0c00a557c"],
Cell[CellGroupData[{
Cell[1798602, 53183, 213, 6, 27, "Input",ExpressionUUID->"a97802de-a084-4d04-bde9-b07c9efd428c"],
Cell[1798818, 53191, 1181, 34, 32, "Output",ExpressionUUID->"744c6fbc-0743-4ab0-be70-a40b82d01207",
 CellID->1001827129]
}, Open  ]],
Cell[CellGroupData[{
Cell[1800036, 53230, 145, 3, 27, "Input",ExpressionUUID->"a887fb36-ce66-4fd5-a8d1-1639b9ee8e2b"],
Cell[1800184, 53235, 184, 3, 26, "Output",ExpressionUUID->"108772cf-0da4-44c8-bc0b-5a263bff65d2",
 CellID->1436764623]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1800417, 53244, 451, 8, 72, "MathCaption",ExpressionUUID->"c7843ef5-1860-4f66-9502-814c048d9348"],
Cell[CellGroupData[{
Cell[1800893, 53256, 143, 3, 27, "Input",ExpressionUUID->"9b15c20e-3fa7-4ab8-a0a8-f1bd3dd29509"],
Cell[1801039, 53261, 1110, 31, 32, "Output",ExpressionUUID->"20dfee45-0ddd-4ec7-b93c-2b980c7ce6a9",
 CellID->1396891893]
}, Open  ]],
Cell[CellGroupData[{
Cell[1802186, 53297, 145, 3, 27, "Input",ExpressionUUID->"e804fbe0-6a57-4d93-b7d0-14c10680008c"],
Cell[1802334, 53302, 182, 3, 26, "Output",ExpressionUUID->"71f459c9-3ff4-4636-9890-9d41b534d7d5",
 CellID->1270816647]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1802565, 53311, 142, 1, 40, "MathCaption",ExpressionUUID->"7644152a-ecd6-46cf-9757-147aeba88a85"],
Cell[CellGroupData[{
Cell[1802732, 53316, 143, 3, 27, "Input",ExpressionUUID->"665521ea-e3cf-494c-b365-88273943fdb6"],
Cell[1802878, 53321, 1181, 34, 32, "Output",ExpressionUUID->"cebb9864-509c-45ba-bff7-716b8a63310f",
 CellID->1890516714]
}, Open  ]]
}, Open  ]],
Cell[1804086, 53359, 195, 3, 23, "Text",ExpressionUUID->"88dfefb5-59e8-445a-8315-8e941aad7a00"],
Cell[CellGroupData[{
Cell[1804306, 53366, 103, 0, 40, "MathCaption",ExpressionUUID->"0ffd951b-f96a-4851-b6aa-c5da79045b99"],
Cell[CellGroupData[{
Cell[1804434, 53370, 282, 9, 27, "Input",ExpressionUUID->"efb583e9-bc83-4437-9535-d23549aa543b"],
Cell[1804719, 53381, 1250, 36, 32, "Output",ExpressionUUID->"79d3dd44-ff73-439e-9823-150bdec8e1b9",
 CellID->1893854833]
}, Open  ]],
Cell[CellGroupData[{
Cell[1806006, 53422, 146, 3, 27, "Input",ExpressionUUID->"9eee5ff2-7aa0-4162-9e98-1afe1e6c19ef"],
Cell[1806155, 53427, 2123, 60, 32, "Output",ExpressionUUID->"30784372-2e76-40d5-86c6-45f9f31e3a89",
 CellID->15562397]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1808327, 53493, 153, 2, 40, "MathCaption",ExpressionUUID->"df692ab9-a1de-4763-b6dd-39c6431988d6"],
Cell[CellGroupData[{
Cell[1808505, 53499, 144, 3, 27, "Input",ExpressionUUID->"018784fc-c821-4c24-b466-84c7cb94a582"],
Cell[1808652, 53504, 1182, 33, 32, "Output",ExpressionUUID->"a22045d5-b27c-495f-b499-88991547c7f4",
 CellID->1711685092]
}, Open  ]],
Cell[CellGroupData[{
Cell[1809871, 53542, 161, 3, 27, "Input",ExpressionUUID->"b65543ae-a0c0-4284-a692-5d02b46c2acc"],
Cell[1810035, 53547, 475, 14, 26, "Output",ExpressionUUID->"601a9c67-f142-483c-b333-3e4541bdd2e7",
 CellID->600406032]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1810559, 53567, 133, 0, 40, "MathCaption",ExpressionUUID->"a68a8b7e-af63-4ec3-bab5-fd3a06c6bedb"],
Cell[CellGroupData[{
Cell[1810717, 53571, 277, 9, 27, "Input",ExpressionUUID->"a36d5b22-f1a2-40f0-99ed-292b4a891145"],
Cell[1810997, 53582, 482, 14, 26, "Output",ExpressionUUID->"2383a42b-4e46-4be4-b4c9-2bc444c91edf",
 CellID->2146006218]
}, Open  ]],
Cell[CellGroupData[{
Cell[1811516, 53601, 143, 3, 27, "Input",ExpressionUUID->"1abdb638-1d99-4617-9ba2-2ce841c52d23"],
Cell[1811662, 53606, 1130, 31, 32, "Output",ExpressionUUID->"0c5695c5-afdf-4c99-ae6b-d7f345441654",
 CellID->938654561]
}, Open  ]],
Cell[CellGroupData[{
Cell[1812829, 53642, 161, 3, 27, "Input",ExpressionUUID->"a0398411-ad29-44fc-a611-4aead834b4c2"],
Cell[1812993, 53647, 475, 14, 26, "Output",ExpressionUUID->"f47f15ea-9baf-4f11-a1b6-f677918d572c",
 CellID->552335865]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1813517, 53667, 104, 0, 40, "MathCaption",ExpressionUUID->"cbb069d9-e30c-4c68-bac6-902658a76d5c"],
Cell[CellGroupData[{
Cell[1813646, 53671, 282, 9, 27, "Input",ExpressionUUID->"a7c74cf7-e5d2-46e1-8ed1-7af7a0da271b"],
Cell[1813931, 53682, 1268, 36, 32, "Output",ExpressionUUID->"acb0fc5b-7b3b-407c-94e1-b4856b690484",
 CellID->1684560721]
}, Open  ]],
Cell[CellGroupData[{
Cell[1815236, 53723, 143, 3, 27, "Input",ExpressionUUID->"eefde0e9-2758-4271-838d-f3a0430fabd6"],
Cell[1815382, 53728, 1157, 31, 32, "Output",ExpressionUUID->"4350cd76-32f2-4759-93d4-131c7af75868",
 CellID->2135119446]
}, Open  ]],
Cell[CellGroupData[{
Cell[1816576, 53764, 161, 3, 27, "Input",ExpressionUUID->"eb382130-444a-426a-82cd-19c7f2ed328e"],
Cell[1816740, 53769, 281, 7, 26, "Output",ExpressionUUID->"1aa56282-833f-42bc-9115-a091b6d2fe4b",
 CellID->500787552]
}, Open  ]]
}, Open  ]],
Cell[1817048, 53780, 261, 4, 40, "Text",ExpressionUUID->"558fd9c3-6418-429d-be70-8e1ec50d84cb"]
}, Closed]],
Cell[CellGroupData[{
Cell[1817346, 53789, 99, 0, 39, "Subsection",ExpressionUUID->"29c87aca-5f16-4fff-b041-275a77e254cf"],
Cell[1817448, 53791, 499, 15, 40, "Text",ExpressionUUID->"560f40ba-9226-426f-b4a5-a4a38879a4e7"],
Cell[CellGroupData[{
Cell[1817972, 53810, 196, 3, 40, "MathCaption",ExpressionUUID->"d966590a-008b-49b2-8ddb-cdbb37ded23e"],
Cell[CellGroupData[{
Cell[1818193, 53817, 281, 7, 27, "Input",ExpressionUUID->"68c792cf-28eb-4f5b-a6a0-a7c42b4ee4d3"],
Cell[1818477, 53826, 1181, 34, 32, "Output",ExpressionUUID->"25810f6e-c342-4a75-8018-22c5cad3fa73",
 CellID->1246017163]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1819707, 53866, 290, 7, 40, "MathCaption",ExpressionUUID->"2726d7a2-4784-4842-94fe-da05be0357d6"],
Cell[CellGroupData[{
Cell[1820022, 53877, 217, 4, 27, "Input",ExpressionUUID->"984c6475-63f4-4dc1-9e32-7018e2b67f65"],
Cell[1820242, 53883, 1179, 33, 32, "Output",ExpressionUUID->"b192aae8-e1f0-4e0a-a8c9-3f0467b0bb9a",
 CellID->1697416644]
}, Open  ]],
Cell[CellGroupData[{
Cell[1821458, 53921, 207, 4, 27, "Input",ExpressionUUID->"f558af18-ca8d-4dc8-98bb-e45b11df316b"],
Cell[1821668, 53927, 202, 3, 26, "Output",ExpressionUUID->"de8e1f6e-3944-4c29-be89-ccb32f8b2b58",
 CellID->1308768960]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1821919, 53936, 158, 3, 40, "MathCaption",ExpressionUUID->"b8eb3c01-ff30-4d5e-b7ec-aca046fb693d"],
Cell[CellGroupData[{
Cell[1822102, 53943, 221, 4, 27, "Input",ExpressionUUID->"c75266c6-326c-48bc-9c4f-37ac00cd609b"],
Cell[1822326, 53949, 1181, 34, 32, "Output",ExpressionUUID->"5dc65c81-cc8d-4201-976b-c1a398f57b46",
 CellID->1952398146]
}, Open  ]],
Cell[CellGroupData[{
Cell[1823544, 53988, 206, 4, 27, "Input",ExpressionUUID->"9468b8ca-fe1c-4b04-b264-4513f4d6dc08"],
Cell[1823753, 53994, 183, 3, 26, "Output",ExpressionUUID->"b39ba3fd-964c-4b2a-9528-b29980bfc0c6",
 CellID->405513662]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[1823997, 54004, 112, 0, 39, "Subsection",ExpressionUUID->"866d0406-c99f-4c83-873e-360202063718"],
Cell[1824112, 54006, 211, 5, 23, "Text",ExpressionUUID->"9bb62229-a8dc-4dfc-8419-8f81c4e38495"],
Cell[CellGroupData[{
Cell[1824348, 54015, 138, 1, 40, "MathCaption",ExpressionUUID->"15d5fd4f-e3f3-47c2-8059-f99ff94332a4"],
Cell[CellGroupData[{
Cell[1824511, 54020, 115, 2, 27, "Input",ExpressionUUID->"800cb60b-a217-4d01-bb9d-ed6f7be38204"],
Cell[1824629, 54024, 219, 5, 26, "Output",ExpressionUUID->"08d03c21-68a5-4b3a-97b0-ee5f17c35de7",
 CellID->493916519]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1824897, 54035, 99, 0, 40, "MathCaption",ExpressionUUID->"20778d52-197f-4236-b331-75f1dceb6a88"],
Cell[CellGroupData[{
Cell[1825021, 54039, 319, 9, 27, "Input",ExpressionUUID->"eef1e896-72f1-48a9-886f-364dbeb487a6"],
Cell[CellGroupData[{
Cell[1825365, 54052, 680, 15, 22, "Print",ExpressionUUID->"62ef7723-3f3a-45c2-9812-e8f245de7356",
 CellID->1230639178],
Cell[1826048, 54069, 801, 18, 22, "Print",ExpressionUUID->"ba85928e-6ba1-4bc3-96f9-436e1de61ef1",
 CellID->937660389],
Cell[1826852, 54089, 818, 19, 22, "Print",ExpressionUUID->"25c01830-08b5-4b8c-bfff-963d26ff1d54",
 CellID->1023156237],
Cell[1827673, 54110, 896, 18, 42, "Print",ExpressionUUID->"946f7961-4d4f-443d-a63e-85fb8fe18067",
 CellID->1285313730],
Cell[1828572, 54130, 771, 17, 22, "Print",ExpressionUUID->"ddd3f2d7-51b8-4ea6-81eb-d87f3de9c05e",
 CellID->729854937],
Cell[1829346, 54149, 268, 4, 22, "Print",ExpressionUUID->"71a25a5e-ed64-4b2f-84ee-713626eafb25",
 CellID->245334495]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[1829653, 54158, 625, 12, 74, "Text",ExpressionUUID->"b7a2707d-2bdc-47a7-81e8-6fb49f7db199"],
Cell[CellGroupData[{
Cell[1830303, 54174, 256, 5, 56, "MathCaption",ExpressionUUID->"eada7ce8-a4e4-4f27-942d-4d48b901e055"],
Cell[CellGroupData[{
Cell[1830584, 54183, 272, 8, 27, "Input",ExpressionUUID->"7e2009d1-8995-4aea-9c32-b1bdf719a590"],
Cell[1830859, 54193, 1172, 32, 32, "Output",ExpressionUUID->"93b899ee-ad4d-4416-90bc-822979752576",
 CellID->168255225]
}, Open  ]],
Cell[CellGroupData[{
Cell[1832068, 54230, 141, 3, 27, "Input",ExpressionUUID->"02640ad5-5ccc-4b58-9f6a-da64b21b8842"],
Cell[1832212, 54235, 195, 3, 26, "Output",ExpressionUUID->"acd465b1-2b15-49d0-8e19-97a9d723ddea",
 CellID->1777087123]
}, Open  ]],
Cell[CellGroupData[{
Cell[1832444, 54243, 272, 8, 27, "Input",ExpressionUUID->"98e0657a-f42c-408b-96d0-60f4b1f9c195"],
Cell[1832719, 54253, 1220, 34, 32, "Output",ExpressionUUID->"b4e9b041-fef4-45fb-be1b-20a234668e0a",
 CellID->300713785]
}, Open  ]],
Cell[CellGroupData[{
Cell[1833976, 54292, 141, 3, 27, "Input",ExpressionUUID->"2096d0d6-4b66-4c78-8179-41712d33dac7"],
Cell[1834120, 54297, 196, 3, 26, "Output",ExpressionUUID->"e8f939e0-ef11-451c-9994-c353a6e4b13c",
 CellID->1067563104]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1834365, 54306, 305, 6, 56, "MathCaption",ExpressionUUID->"9dc2606d-d7bd-4526-8f2a-8c7a511fd7d8"],
Cell[CellGroupData[{
Cell[1834695, 54316, 272, 8, 27, "Input",ExpressionUUID->"656b1ab0-0bfa-4479-af16-028a1b73b379"],
Cell[1834970, 54326, 765, 16, 22, "Print",ExpressionUUID->"9e87e89a-8430-445e-a1bc-2312ba82e546",
 CellID->307218243],
Cell[1835738, 54344, 1268, 35, 32, "Output",ExpressionUUID->"493384bd-0716-4b38-98aa-cb492e52a857",
 CellID->866322713]
}, Open  ]],
Cell[CellGroupData[{
Cell[1837043, 54384, 141, 3, 27, "Input",ExpressionUUID->"70af6ea8-545e-43e4-8cf7-d478db21dead"],
Cell[1837187, 54389, 198, 3, 26, "Output",ExpressionUUID->"52fb7749-5c2a-469e-a313-fd5734339108",
 CellID->346710731]
}, Open  ]]
}, Open  ]],
Cell[1837412, 54396, 333, 8, 56, "MathCaption",ExpressionUUID->"3626e143-dd10-46f4-b73d-1fe9e5135ee5"],
Cell[1837748, 54406, 174, 3, 23, "Text",ExpressionUUID->"4b94c430-a305-4cb6-bff3-be761e76f09b"],
Cell[CellGroupData[{
Cell[1837947, 54413, 102, 0, 40, "MathCaption",ExpressionUUID->"effccc54-4ffe-4c98-a97b-2c947951a2b6"],
Cell[CellGroupData[{
Cell[1838074, 54417, 319, 10, 27, "Input",ExpressionUUID->"84f98ec1-d558-4daf-9d54-3f914e11f7ae"],
Cell[1838396, 54429, 1216, 34, 32, "Output",ExpressionUUID->"5072d705-595a-4c32-a1b6-26073f5e97d9",
 CellID->1044121606]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1839661, 54469, 352, 13, 40, "MathCaption",ExpressionUUID->"fad22346-8041-4f3d-8db0-de8311109d4e"],
Cell[CellGroupData[{
Cell[1840038, 54486, 149, 3, 27, "Input",ExpressionUUID->"1d333c1f-7944-482a-8623-978473b00aea"],
Cell[1840190, 54491, 7590, 213, 32, "Output",ExpressionUUID->"ba19239f-437e-4c47-8091-fced0020753f",
 CellID->1462952405]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1847829, 54710, 99, 0, 40, "MathCaption",ExpressionUUID->"fcc2196f-c91b-4472-b894-653fd7d31cf3"],
Cell[CellGroupData[{
Cell[1847953, 54714, 173, 4, 27, "Input",ExpressionUUID->"c63e4165-8446-4170-9003-e1d433372aba"],
Cell[1848129, 54720, 7590, 213, 32, "Output",ExpressionUUID->"90d10878-a28f-480e-96af-f5dc9c9effb6",
 CellID->1300796332]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1855768, 54939, 127, 0, 40, "MathCaption",ExpressionUUID->"e6aa7a5b-c6ef-442a-8a71-edbb2099770a"],
Cell[CellGroupData[{
Cell[1855920, 54943, 184, 4, 27, "Input",ExpressionUUID->"d176231d-98a6-4b23-9abb-67b28941bed9"],
Cell[1856107, 54949, 7704, 215, 32, "Output",ExpressionUUID->"65661f01-2d3b-45e5-a40b-14f1b176087b",
 CellID->1100425464]
}, Open  ]],
Cell[CellGroupData[{
Cell[1863848, 55169, 181, 4, 27, "Input",ExpressionUUID->"fee87d64-b59a-4299-93e5-968a8449cc64"],
Cell[1864032, 55175, 1215, 34, 32, "Output",ExpressionUUID->"fb333ce6-9b3a-42fd-8360-c5dc40a2e129",
 CellID->721605484]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1865296, 55215, 249, 6, 40, "MathCaption",ExpressionUUID->"7efae191-10dd-4865-bc5a-23cdd3f9b0b5"],
Cell[CellGroupData[{
Cell[1865570, 55225, 149, 3, 27, "Input",ExpressionUUID->"655d09de-6231-49b9-be0e-4418ccbb16a4"],
Cell[1865722, 55230, 11348, 314, 32, "Output",ExpressionUUID->"eee795fe-986c-4378-a66e-73ec9d85d25b",
 CellID->1183214979]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1877119, 55550, 105, 0, 40, "MathCaption",ExpressionUUID->"3a61f797-9edc-4ba0-8abf-678382460e5f"],
Cell[CellGroupData[{
Cell[1877249, 55554, 428, 14, 27, "Input",ExpressionUUID->"c333ba2a-bd73-415d-b94e-5c73ef9365bd"],
Cell[1877680, 55570, 51602, 1378, 128, "Output",ExpressionUUID->"99a9f03d-cbdf-4f8c-960c-0e9b5442056c",
 CellID->595580833]
}, Open  ]]
}, Open  ]],
Cell[1929309, 56952, 1199, 31, 75, "DefinitionBox",ExpressionUUID->"ebb00c11-15d1-4f90-af87-59f5d08c0b10"],
Cell[1930511, 56985, 102, 0, 23, "Text",ExpressionUUID->"37768638-83d1-45a0-9860-23c2fad19d3c"],
Cell[1930616, 56987, 285, 4, 40, "Text",ExpressionUUID->"21872d92-f2ba-4b0d-bf88-6d22ee7350f7"],
Cell[CellGroupData[{
Cell[1930926, 56995, 128, 0, 40, "MathCaption",ExpressionUUID->"30606ad9-3ba3-4daa-88e5-8bb2ab82b0c5"],
Cell[CellGroupData[{
Cell[1931079, 56999, 245, 7, 27, "Input",ExpressionUUID->"791b7fb7-2546-4b24-ab29-d06a7cd0c15b"],
Cell[1931327, 57008, 1181, 33, 32, "Output",ExpressionUUID->"f7ebda4a-9ca5-47a9-8bcc-bf56aab34d78",
 CellID->505781564]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1932557, 57047, 463, 14, 56, "MathCaption",ExpressionUUID->"fc6c4717-dc5e-457c-a5bb-471911dfb305"],
Cell[CellGroupData[{
Cell[1933045, 57065, 151, 3, 27, "Input",ExpressionUUID->"0ec2618f-a699-4338-9128-05c1576c977c"],
Cell[1933199, 57070, 6796, 192, 32, "Output",ExpressionUUID->"52a1826f-59d4-435b-9924-e6276e7fca68",
 CellID->930471322]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1940044, 57268, 115, 0, 40, "MathCaption",ExpressionUUID->"7506df57-9f7b-48b5-aa64-7081731d64ee"],
Cell[CellGroupData[{
Cell[1940184, 57272, 187, 4, 27, "Input",ExpressionUUID->"2c010025-8124-49ed-8a78-9c298a20b8b2"],
Cell[1940374, 57278, 8187, 228, 32, "Output",ExpressionUUID->"ee4da196-ae8c-4538-9443-6fbb2264c9d6",
 CellID->455591400]
}, Open  ]],
Cell[CellGroupData[{
Cell[1948598, 57511, 334, 10, 27, "Input",ExpressionUUID->"ad8291c8-156a-46cd-b86f-26d99ddafe23"],
Cell[1948935, 57523, 8131, 227, 32, "Output",ExpressionUUID->"4ce94a6c-ae84-441e-b0a4-b8a036b78a88",
 CellID->1345458245]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1957115, 57756, 203, 3, 56, "MathCaption",ExpressionUUID->"0d899c47-3032-436e-9eb7-70d07b5d2de4"],
Cell[CellGroupData[{
Cell[1957343, 57763, 370, 10, 27, "Input",ExpressionUUID->"a0cf2d5a-66b0-42ec-aca2-f033c82522aa"],
Cell[CellGroupData[{
Cell[1957738, 57777, 683, 15, 22, "Print",ExpressionUUID->"532cc24a-5b23-4e96-ae36-ac10121b0e8a",
 CellID->113920805],
Cell[1958424, 57794, 778, 17, 22, "Print",ExpressionUUID->"db04ebce-e3b6-4b2b-95bf-e6a9e6a27044",
 CellID->1259040158],
Cell[1959205, 57813, 828, 19, 22, "Print",ExpressionUUID->"f0d8a7ca-bef9-4c91-a2f3-0b2b01c38e68",
 CellID->1531561615],
Cell[1960036, 57834, 898, 18, 42, "Print",ExpressionUUID->"83999169-0297-425d-9382-e0926739c2b2",
 CellID->1217070363],
Cell[1960937, 57854, 773, 17, 22, "Print",ExpressionUUID->"08fbd7d4-2e0e-461c-8704-94fe67248675",
 CellID->519853266],
Cell[1961713, 57873, 271, 4, 22, "Print",ExpressionUUID->"1276d03d-21aa-4152-b671-573dbe8edd5e",
 CellID->1210365666]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1962045, 57884, 196, 5, 40, "MathCaption",ExpressionUUID->"e5f7ef7f-6967-4c1a-ad55-49faeafe63cd"],
Cell[CellGroupData[{
Cell[1962266, 57893, 335, 10, 27, "Input",ExpressionUUID->"25566c72-d2f9-40df-b7e5-5d4cd1900fb3"],
Cell[1962604, 57905, 768, 16, 22, "Print",ExpressionUUID->"ec84d557-b6fd-494f-a4e4-f33f466ac9fb",
 CellID->1650644078],
Cell[1963375, 57923, 10285, 289, 32, "Output",ExpressionUUID->"968a3072-327f-4128-a5a0-3f04ac99fb72",
 CellID->1181267000]
}, Open  ]],
Cell[CellGroupData[{
Cell[1973697, 58217, 141, 3, 27, "Input",ExpressionUUID->"64b3c03c-6d04-485c-8089-34e6689ba954"],
Cell[1973841, 58222, 553, 12, 94, "Output",ExpressionUUID->"acecbec6-73d3-4cf2-bd92-733232a56855",
 CellID->68152485]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1974443, 58240, 540, 16, 72, "MathCaption",ExpressionUUID->"4b6413aa-575a-49c7-bae1-512b19313fd5"],
Cell[CellGroupData[{
Cell[1975008, 58260, 300, 8, 27, "Input",ExpressionUUID->"a02bc13a-ad78-44ad-821c-1b0fd7bfdc5b"],
Cell[1975311, 58270, 2408, 67, 32, "Output",ExpressionUUID->"13dca5a3-9a1f-4679-96c9-19fa0d56f64c",
 CellID->830031544]
}, Open  ]]
}, Open  ]],
Cell[1977746, 58341, 769, 19, 58, "DefinitionBox",ExpressionUUID->"a5a86fff-914f-4cd1-ab4c-207be4d7047f"],
Cell[1978518, 58362, 120, 0, 23, "Text",ExpressionUUID->"9fdfda3c-9d70-4fa1-8d66-ec9131f4123a"],
Cell[CellGroupData[{
Cell[1978663, 58366, 104, 0, 40, "MathCaption",ExpressionUUID->"583303dd-f2ec-48b8-8ddc-1ff7ce5f0630"],
Cell[CellGroupData[{
Cell[1978792, 58370, 211, 6, 27, "Input",ExpressionUUID->"7caf7908-52d0-449f-8dca-7a8028be4dd1"],
Cell[CellGroupData[{
Cell[1979028, 58380, 618, 12, 22, "Print",ExpressionUUID->"2e78a874-31de-42c8-aab0-600b65ccf633",
 CellID->286389096],
Cell[1979649, 58394, 576, 12, 22, "Print",ExpressionUUID->"5c3c07f5-2332-4aae-b9a1-2d17a3b0bcb5",
 CellID->830683545],
Cell[1980228, 58408, 600, 12, 22, "Print",ExpressionUUID->"2569eda0-cf27-4667-9d77-781f7aea01bb",
 CellID->507151668],
Cell[1980831, 58422, 582, 12, 22, "Print",ExpressionUUID->"c550e499-7b79-46c4-8b5c-a9a8e17a31f4",
 CellID->335489133],
Cell[1981416, 58436, 583, 12, 22, "Print",ExpressionUUID->"0d17c501-f28c-4c11-be8f-2c38b18093aa",
 CellID->1421284956],
Cell[1982002, 58450, 575, 12, 22, "Print",ExpressionUUID->"3eac565b-0e67-46ef-b2a8-e8688912ae1a",
 CellID->1075649911],
Cell[1982580, 58464, 578, 12, 22, "Print",ExpressionUUID->"ea74b6e0-3620-424c-954d-4ad75e7fb8b3",
 CellID->930713731],
Cell[1983161, 58478, 626, 12, 22, "Print",ExpressionUUID->"e644f1ad-ea7d-4d8d-a025-2184541f7628",
 CellID->234792169],
Cell[1983790, 58492, 603, 12, 22, "Print",ExpressionUUID->"684e7e66-a470-4cad-b24d-b29b344efb3a",
 CellID->1224524844],
Cell[1984396, 58506, 583, 12, 22, "Print",ExpressionUUID->"185e1e0c-91e7-498a-af26-0f6aa3a83ff8",
 CellID->1253263907],
Cell[1984982, 58520, 583, 12, 22, "Print",ExpressionUUID->"08687b25-63c2-48b8-b0c9-c64e6398a307",
 CellID->1086875173],
Cell[1985568, 58534, 578, 12, 22, "Print",ExpressionUUID->"1f3fc87e-45bb-464a-ada4-f89d7a288bc2",
 CellID->615131229]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[1986185, 58551, 369, 7, 40, "Text",ExpressionUUID->"1e49e14d-2ffb-45a8-a1c8-0677d22cd53d"]
}, Closed]],
Cell[CellGroupData[{
Cell[1986591, 58563, 108, 0, 39, "Subsection",ExpressionUUID->"f87d0846-1b86-4f36-a2a8-41e6b8db3f25"],
Cell[1986702, 58565, 313, 6, 40, "Text",ExpressionUUID->"bc930580-ed21-4357-893a-e6593af4b119"],
Cell[CellGroupData[{
Cell[1987040, 58575, 192, 3, 40, "MathCaption",ExpressionUUID->"fbbe1de0-a173-4fc2-9a70-4e3cc69a759e"],
Cell[CellGroupData[{
Cell[1987257, 58582, 255, 7, 27, "Input",ExpressionUUID->"9510dcf1-73a8-4f16-ae18-fb03f7c67260"],
Cell[CellGroupData[{
Cell[1987537, 58593, 680, 15, 22, "Print",ExpressionUUID->"1822de1b-cad8-4275-bfce-86efa4cda7a7",
 CellID->1117298015],
Cell[1988220, 58610, 775, 17, 22, "Print",ExpressionUUID->"6bb9f321-bc42-4424-bc08-230ddf769aef",
 CellID->830728737],
Cell[1988998, 58629, 828, 19, 22, "Print",ExpressionUUID->"294e7038-56d2-4e22-b225-07fec430e77f",
 CellID->1266693612],
Cell[1989829, 58650, 895, 18, 42, "Print",ExpressionUUID->"ee5d6336-2613-4f50-aecb-5a713db3ecc5",
 CellID->1314144173],
Cell[1990727, 58670, 769, 17, 22, "Print",ExpressionUUID->"3f752cc7-f4d3-40c9-9f37-f982e4f2facb",
 CellID->2099289863],
Cell[1991499, 58689, 268, 4, 22, "Print",ExpressionUUID->"fade3f37-f93d-4d72-91fe-0b837eabd4fb",
 CellID->1755890878]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1991828, 58700, 133, 0, 40, "MathCaption",ExpressionUUID->"e72e4e15-9e11-4495-a5de-a4d0ae663522"],
Cell[CellGroupData[{
Cell[1991986, 58704, 730, 24, 27, "Input",ExpressionUUID->"15ee41dd-57e7-475a-a412-a2f2166264c4"],
Cell[1992719, 58730, 4561, 126, 32, "Output",ExpressionUUID->"0b78c8d6-cc38-486f-9036-891bf5d8c5a5",
 CellID->537709472]
}, Open  ]],
Cell[CellGroupData[{
Cell[1997317, 58861, 319, 10, 27, "Input",ExpressionUUID->"7aad89ba-2a9e-4998-bb89-ea2879a6a040"],
Cell[1997639, 58873, 26512, 736, 80, "Output",ExpressionUUID->"ba6a087b-e66c-4fe7-849d-1f1201c30e94",
 CellID->671909328]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2024200, 59615, 214, 3, 56, "MathCaption",ExpressionUUID->"af8a9621-05b7-413a-8a0b-896840faa124"],
Cell[CellGroupData[{
Cell[2024439, 59622, 142, 3, 27, "Input",ExpressionUUID->"5587cfce-5c55-448d-af7a-76278e3c6983"],
Cell[2024584, 59627, 65537, 1816, 152, "Output",ExpressionUUID->"93642775-b1e9-44be-a6f6-1aa4ebabe52d",
 CellID->1311059770]
}, Open  ]],
Cell[CellGroupData[{
Cell[2090158, 61448, 152, 3, 27, "Input",ExpressionUUID->"f43c7626-8972-4b74-a84b-9293a54a61ed"],
Cell[2090313, 61453, 79740, 2207, 176, "Output",ExpressionUUID->"90bef3ed-a01b-487c-98b3-3c21551fcce5",
 CellID->1799336170]
}, Open  ]],
Cell[CellGroupData[{
Cell[2170090, 63665, 152, 3, 27, "Input",ExpressionUUID->"0ead20f3-dd2b-44e7-b0ed-bbe9e02f104f"],
Cell[2170245, 63670, 104429, 2855, 248, "Output",ExpressionUUID->"b660f6f8-1048-42f2-9ffb-ce6a73886b12",
 CellID->578292136]
}, Open  ]],
Cell[CellGroupData[{
Cell[2274711, 66530, 138, 3, 27, "Input",ExpressionUUID->"cb6c1a5e-c217-4049-bd1b-14f8d2a6547b"],
Cell[2274852, 66535, 171, 3, 26, "Output",ExpressionUUID->"36c4f8af-764f-4b7e-b777-662bce75ef73",
 CellID->1992427077]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2275072, 66544, 134, 0, 40, "MathCaption",ExpressionUUID->"c87b02a2-fe67-4786-9dc5-26fe14dba555"],
Cell[CellGroupData[{
Cell[2275231, 66548, 642, 21, 27, "Input",ExpressionUUID->"0b33cdd0-b621-4392-943e-4b580315db5d"],
Cell[2275876, 66571, 3596, 98, 32, "Output",ExpressionUUID->"c34683a7-269f-48d3-a1dc-127bda1f3c49",
 CellID->1009375468]
}, Open  ]],
Cell[CellGroupData[{
Cell[2279509, 66674, 319, 10, 27, "Input",ExpressionUUID->"46d6c294-043b-4908-a077-d010bc96a5e9"],
Cell[2279831, 66686, 20473, 563, 56, "Output",ExpressionUUID->"de03aafb-ae30-45a9-9f39-274abc449096",
 CellID->1600163023]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[2300353, 67255, 124, 0, 40, "MathCaption",ExpressionUUID->"708a4e32-a664-427f-979c-79237bd2442a"],
Cell[CellGroupData[{
Cell[2300502, 67259, 142, 3, 27, "Input",ExpressionUUID->"91a7e00d-9fb8-4dc6-b8c1-e319d387d044"],
Cell[2300647, 67264, 73169, 2026, 152, "Output",ExpressionUUID->"ce42d8e2-35a5-4df0-9c60-b25a32335937",
 CellID->1002216180]
}, Open  ]],
Cell[CellGroupData[{
Cell[2373853, 69295, 152, 3, 27, "Input",ExpressionUUID->"095718d5-0070-4b62-9a48-27b85cbccbfb"],
Cell[2374008, 69300, 289242, 7637, 752, "Output",ExpressionUUID->"aa7cbb20-d3d6-42af-aec9-d42cc67d234a",
 CellID->1533093839]
}, Open  ]],
Cell[CellGroupData[{
Cell[2663287, 76942, 152, 3, 27, "Input",ExpressionUUID->"118d0df2-cafb-46b3-828d-22ee5195157e"],
Cell[2663442, 76947, 297077, 7847, 752, "Output",ExpressionUUID->"8682f9ac-d088-4124-8796-8e03b2bad44e",
 CellID->1199355367]
}, Open  ]],
Cell[CellGroupData[{
Cell[2960556, 84799, 138, 3, 27, "Input",ExpressionUUID->"09a85fd6-bbad-49ee-84f6-e81b2718d64c"],
Cell[2960697, 84804, 47537, 1312, 104, "Output",ExpressionUUID->"b13a67bc-1a06-4736-88c4-c151d8533942",
 CellID->1284142427]
}, Open  ]],
Cell[CellGroupData[{
Cell[3008271, 86121, 143, 3, 27, "Input",ExpressionUUID->"549b35dd-b674-40f2-83e3-8f9be8d0a357"],
Cell[3008417, 86126, 170, 3, 26, "Output",ExpressionUUID->"eec8e411-a994-4b07-8c42-b9d1d84a8b14",
 CellID->371118239]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3008636, 86135, 87, 0, 40, "MathCaption",ExpressionUUID->"fb60f043-f295-4e88-922a-25f37c9bb1ee"],
Cell[CellGroupData[{
Cell[3008748, 86139, 146, 3, 27, "Input",ExpressionUUID->"aafdb914-e2f5-4ccf-a213-b98b35580641"],
Cell[CellGroupData[{
Cell[3008919, 86146, 625, 12, 22, "Print",ExpressionUUID->"89900bfd-778b-4fb5-8d5c-84fd96515585",
 CellID->1091731985],
Cell[3009547, 86160, 600, 12, 22, "Print",ExpressionUUID->"ed7d98b6-aa10-4cc5-83d7-16546d96698c",
 CellID->639191927],
Cell[3010150, 86174, 581, 12, 22, "Print",ExpressionUUID->"2d0eb9b9-d3a5-4e8c-a438-6db5e962a7d7",
 CellID->1185514795],
Cell[3010734, 86188, 582, 12, 22, "Print",ExpressionUUID->"ef44147d-41c3-4c6a-aacb-b0264e988660",
 CellID->593882699],
Cell[3011319, 86202, 575, 12, 22, "Print",ExpressionUUID->"35fe2e3d-924e-4cee-835a-a0ac03ab2644",
 CellID->2030014484]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3011967, 86222, 98, 0, 39, "Subsection",ExpressionUUID->"e53b0c02-7a31-444e-af12-e664d149db42"],
Cell[3012068, 86224, 730, 15, 74, "Text",ExpressionUUID->"66acdbf6-cfe0-4c5d-9195-5ac7c3a48a92"],
Cell[3012801, 86241, 1329, 36, 92, "DefinitionBox",ExpressionUUID->"9c916998-07b6-4d7f-a5dc-9c0454a0bed8"],
Cell[3014133, 86279, 109, 0, 23, "Text",ExpressionUUID->"75f4892b-629e-4473-a591-88818030bf40"],
Cell[CellGroupData[{
Cell[3014267, 86283, 115, 0, 40, "MathCaption",ExpressionUUID->"22773fbb-d315-44c5-84ba-b1ae3966642f"],
Cell[CellGroupData[{
Cell[3014407, 86287, 215, 6, 27, "Input",ExpressionUUID->"769d1396-2962-42e4-90f2-171bbf2a623f"],
Cell[3014625, 86295, 615, 12, 22, "Print",ExpressionUUID->"f3d7bd1b-69ac-41ce-900b-bf49f203e3b8",
 CellID->1175468544]
}, Open  ]],
Cell[CellGroupData[{
Cell[3015277, 86312, 131, 3, 27, "Input",ExpressionUUID->"8d3dc582-8a41-48b6-a2eb-a1697bbdb4c3"],
Cell[CellGroupData[{
Cell[3015433, 86319, 201, 3, 43, "Print",ExpressionUUID->"c227df3b-fdfd-490b-a692-412f36cca699",
 CellTags->"Info5793728981796-9181623",
 CellID->1397750255],
Cell[3015637, 86324, 2136, 66, 262, "Print",ExpressionUUID->"86dcd4d2-d60f-4dc8-8d2c-deaa78c12e5d",
 CellTags->"Info5793728981796-9181623",
 CellID->664921252]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3017834, 86397, 277, 4, 56, "MathCaption",ExpressionUUID->"da88c5ae-3389-43a8-9e37-7fd8deb1d5af"],
Cell[CellGroupData[{
Cell[3018136, 86405, 212, 6, 27, "Input",ExpressionUUID->"59c548bb-9730-456c-851f-653dedf61ef1"],
Cell[3018351, 86413, 401, 12, 26, "Output",ExpressionUUID->"dd5b3192-5173-4cd7-9e92-1029acdb9978",
 CellID->1974067023]
}, Open  ]],
Cell[CellGroupData[{
Cell[3018789, 86430, 155, 3, 27, "Input",ExpressionUUID->"6f3c6176-a501-4c05-9560-981d00bc557b"],
Cell[3018947, 86435, 171, 3, 26, "Output",ExpressionUUID->"7e16bc55-71b7-4905-9acf-1e9cee5f874e",
 CellID->1669593058]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3019167, 86444, 140, 1, 40, "MathCaption",ExpressionUUID->"fa6f71bf-1bd8-4f8f-91d6-7ded973828e4"],
Cell[3019310, 86447, 161, 3, 27, "Input",ExpressionUUID->"552cb858-df77-4d95-9eb0-db0c8a4e63c3"],
Cell[CellGroupData[{
Cell[3019496, 86454, 212, 6, 27, "Input",ExpressionUUID->"b73f45b2-67f8-4625-8286-4b18796c929a"],
Cell[3019711, 86462, 171, 3, 26, "Output",ExpressionUUID->"e259591a-1051-4655-864e-0b8b917e2364",
 CellID->1829194104]
}, Open  ]],
Cell[3019897, 86468, 160, 3, 27, "Input",ExpressionUUID->"846fd6be-23f9-4508-8baf-b0a75683bc77"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3020094, 86476, 184, 3, 40, "MathCaption",ExpressionUUID->"5941d6dd-d795-466a-b0f7-9c86790d986b"],
Cell[CellGroupData[{
Cell[3020303, 86483, 191, 5, 27, "Input",ExpressionUUID->"a945d173-b062-4002-8bfa-49c8f9c45d22"],
Cell[3020497, 86490, 401, 11, 26, "Output",ExpressionUUID->"3b120647-f42c-40d7-98b9-06f81147168b",
 CellID->2036899123]
}, Open  ]],
Cell[CellGroupData[{
Cell[3020935, 86506, 159, 3, 27, "Input",ExpressionUUID->"f87de4cc-ffb9-4c36-bbdb-28914852d7b0"],
Cell[3021097, 86511, 634, 13, 22, "Print",ExpressionUUID->"2fb4a181-3c60-4280-a7d1-ecca648bbe81",
 CellID->1126743434]
}, Open  ]],
Cell[CellGroupData[{
Cell[3021768, 86529, 191, 5, 27, "Input",ExpressionUUID->"1b285f34-ef14-4e6a-957d-a51f0753071f"],
Cell[3021962, 86536, 170, 3, 26, "Output",ExpressionUUID->"e80e1ac4-ae45-4505-b193-f4538d6fbd4f",
 CellID->666095992]
}, Open  ]],
Cell[CellGroupData[{
Cell[3022169, 86544, 161, 3, 27, "Input",ExpressionUUID->"681aca36-cd36-46c0-a4f5-4a1a8e6a2c12"],
Cell[3022333, 86549, 732, 18, 22, "Print",ExpressionUUID->"a233a52a-72c1-4805-ad2c-68915034335d",
 CellID->243085733]
}, Open  ]],
Cell[CellGroupData[{
Cell[3023102, 86572, 191, 5, 27, "Input",ExpressionUUID->"97cc96e3-865e-4352-a1cd-cf6953fffd7e"],
Cell[3023296, 86579, 390, 11, 26, "Output",ExpressionUUID->"ece36969-2c21-4896-98f7-3b4d3293936d",
 CellID->1028017937]
}, Open  ]],
Cell[3023701, 86593, 148, 3, 27, "Input",ExpressionUUID->"4172c80d-4d44-4b23-be79-3f386109b90c"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3023898, 86602, 103, 0, 39, "Subsection",ExpressionUUID->"0704b121-e23a-4a34-b2da-ab4444ebb3de"],
Cell[3024004, 86604, 582, 16, 57, "Text",ExpressionUUID->"2552928f-29d9-4573-9340-1837b8ffaeb1"],
Cell[3024589, 86622, 783, 22, 58, "DefinitionBox",ExpressionUUID->"45628cc3-8169-471c-95dd-a09e3e1c7100"],
Cell[3025375, 86646, 94, 0, 23, "Text",ExpressionUUID->"d9bff57a-c643-47d5-8bf8-5feb8e98f66c"],
Cell[CellGroupData[{
Cell[3025494, 86650, 318, 8, 56, "MathCaption",ExpressionUUID->"431facde-7d95-4aff-a35d-31e682d084c7"],
Cell[CellGroupData[{
Cell[3025837, 86662, 190, 5, 27, "Input",ExpressionUUID->"ecdaff66-c610-4e86-89e1-87a4979ac18c"],
Cell[3026030, 86669, 489, 17, 26, "Output",ExpressionUUID->"9cd7bf8f-e1a2-41c6-8a61-0b7e180ed6ce",
 CellID->656372838]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3026568, 86692, 262, 5, 56, "MathCaption",ExpressionUUID->"2c2319e2-06be-4e1e-a453-973e1294e76c"],
Cell[CellGroupData[{
Cell[3026855, 86701, 281, 8, 27, "Input",ExpressionUUID->"5fba2b9f-c335-4326-8e3d-985324837dd8"],
Cell[3027139, 86711, 192, 3, 26, "Output",ExpressionUUID->"ddadd07b-a25e-486a-ad96-f0084b1aed58",
 CellID->34793627]
}, Open  ]],
Cell[CellGroupData[{
Cell[3027368, 86719, 281, 8, 27, "Input",ExpressionUUID->"b0f59510-bbe6-4863-bb59-398cbcae60c2"],
Cell[3027652, 86729, 191, 3, 26, "Output",ExpressionUUID->"da209a7c-43c2-4490-95b3-a21c71435510",
 CellID->660937005]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3027892, 86738, 185, 3, 40, "MathCaption",ExpressionUUID->"6434f3d2-02e6-409a-98d5-eba0e36a2db2"],
Cell[CellGroupData[{
Cell[3028102, 86745, 213, 5, 27, "Input",ExpressionUUID->"a7525c81-cda1-4b78-850d-dfea1b3b6f42"],
Cell[3028318, 86752, 601, 12, 22, "Print",ExpressionUUID->"77c535b5-45ae-43fa-8b15-86638c451455",
 CellID->1582820751]
}, Open  ]],
Cell[CellGroupData[{
Cell[3028956, 86769, 196, 5, 27, "Input",ExpressionUUID->"7d44797d-14b6-4359-a76c-44e9c6b70b13"],
Cell[3029155, 86776, 171, 3, 26, "Output",ExpressionUUID->"1bb34982-01c5-42ca-afb7-2e016988e206",
 CellID->1588426668]
}, Open  ]],
Cell[CellGroupData[{
Cell[3029363, 86784, 270, 8, 27, "Input",ExpressionUUID->"b90a6b90-f93e-4df1-8b92-58b9557e569d"],
Cell[3029636, 86794, 647, 13, 22, "Print",ExpressionUUID->"1bc0d25b-eb46-49c9-88c2-629407b50730",
 CellID->1317793219]
}, Open  ]],
Cell[CellGroupData[{
Cell[3030320, 86812, 127, 3, 27, "Input",ExpressionUUID->"bb97df03-06da-4e15-84f2-621ea286c4c3"],
Cell[CellGroupData[{
Cell[3030472, 86819, 199, 3, 43, "Print",ExpressionUUID->"47a658f0-5129-4f86-9b44-b5ae579d9de6",
 CellTags->"Info5973728981810-9181623",
 CellID->1151780222],
Cell[3030674, 86824, 3117, 98, 296, "Print",ExpressionUUID->"73efd163-d740-46b6-873d-021591318f7f",
 CellTags->"Info5973728981810-9181623",
 CellID->552597096]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3033852, 86929, 510, 13, 72, "MathCaption",ExpressionUUID->"72b9653c-57ad-4dd8-8d84-dc0294b79014"],
Cell[CellGroupData[{
Cell[3034387, 86946, 122, 2, 27, "Input",ExpressionUUID->"517b52ea-01d9-4729-9af6-39b8ae7d413b"],
Cell[3034512, 86950, 184, 3, 26, "Output",ExpressionUUID->"d7bc4cf1-9800-48c9-b430-457b2b372b4e",
 CellID->1780102844]
}, Open  ]],
Cell[CellGroupData[{
Cell[3034733, 86958, 274, 8, 27, "Input",ExpressionUUID->"e87154fa-28dc-4dc6-b9c7-c36f3843da31"],
Cell[3035010, 86968, 2156, 63, 32, "Output",ExpressionUUID->"18fe6429-1cdb-4bad-80d0-984afd072f62",
 CellID->1593812751]
}, Open  ]],
Cell[CellGroupData[{
Cell[3037203, 87036, 301, 9, 27, "Input",ExpressionUUID->"f1733c28-b855-4ae7-87e3-1db43b96dcd0"],
Cell[3037507, 87047, 2217, 64, 32, "Output",ExpressionUUID->"51cf5f54-f27c-4937-946c-d23b7a514e34",
 CellID->1647377026]
}, Open  ]],
Cell[3039739, 87114, 181, 4, 27, "Input",ExpressionUUID->"75990947-29f6-4495-ac12-8008e55da988"],
Cell[CellGroupData[{
Cell[3039945, 87122, 274, 8, 27, "Input",ExpressionUUID->"47d7ef43-e76d-4e35-94ce-946fe8ef8269"],
Cell[3040222, 87132, 2160, 63, 32, "Output",ExpressionUUID->"68a04909-898d-4cea-95e4-56081b80dd04",
 CellID->2106498858]
}, Open  ]],
Cell[CellGroupData[{
Cell[3042419, 87200, 301, 9, 27, "Input",ExpressionUUID->"a0f6933e-345b-421b-9996-1ec7fe877ce8"],
Cell[3042723, 87211, 2179, 63, 32, "Output",ExpressionUUID->"066d6fbd-3de7-4297-b1fb-7b8d1f391149",
 CellID->510901964]
}, Open  ]],
Cell[3044917, 87277, 176, 4, 27, "Input",ExpressionUUID->"361d07b6-0be9-45a8-bce6-70641d4539ac"],
Cell[CellGroupData[{
Cell[3045118, 87285, 274, 8, 27, "Input",ExpressionUUID->"aa0e8b99-63d4-4074-85f6-a4ea22c66023"],
Cell[3045395, 87295, 2100, 61, 32, "Output",ExpressionUUID->"794a4899-9adb-4568-9435-d3b99319bd89",
 CellID->1089688242]
}, Open  ]],
Cell[CellGroupData[{
Cell[3047532, 87361, 301, 9, 27, "Input",ExpressionUUID->"102f8a36-6784-410f-ba11-a6dd1a5b3dc6"],
Cell[3047836, 87372, 2128, 61, 32, "Output",ExpressionUUID->"99ef7db5-78d8-4244-9be3-a72541f49da9",
 CellID->649741678]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3050013, 87439, 87, 0, 40, "MathCaption",ExpressionUUID->"0893f67b-8882-4761-aa1a-99673c70eaad"],
Cell[CellGroupData[{
Cell[3050125, 87443, 147, 3, 27, "Input",ExpressionUUID->"de8f687e-9d73-4da9-adf7-318ec17e80c5"],
Cell[3050275, 87448, 548, 12, 22, "Print",ExpressionUUID->"2c553688-f03e-4ed8-b565-6c651e92edff",
 CellID->254037151]
}, Open  ]],
Cell[CellGroupData[{
Cell[3050860, 87465, 153, 3, 27, "Input",ExpressionUUID->"814a3b84-f815-40ea-a914-51a9bd722572"],
Cell[3051016, 87470, 698, 18, 22, "Print",ExpressionUUID->"91bb81ac-a553-4136-b01e-7554c6db00bf",
 CellID->404906621]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3051775, 87495, 97, 0, 39, "Subsection",ExpressionUUID->"1ce5e5c3-cd56-4397-a63c-c19b8f523078"],
Cell[3051875, 87497, 339, 7, 40, "Text",ExpressionUUID->"54546be5-3498-4f7d-aec5-95603a2ad162"],
Cell[3052217, 87506, 688, 18, 58, "DefinitionBox",ExpressionUUID->"c6742e28-54b3-4bcb-b390-196e36e2e3ed"],
Cell[3052908, 87526, 87, 0, 23, "Text",ExpressionUUID->"4950f6ad-4d93-4243-8a87-f055ae9dd624"],
Cell[CellGroupData[{
Cell[3053020, 87530, 219, 3, 56, "MathCaption",ExpressionUUID->"b06186aa-a51a-4720-9f70-3f15386d89ca"],
Cell[CellGroupData[{
Cell[3053264, 87537, 269, 8, 27, "Input",ExpressionUUID->"9b605bac-b45b-4e31-800d-cd28a5e6efe0"],
Cell[3053536, 87547, 1228, 35, 32, "Output",ExpressionUUID->"55dde540-ae92-4161-a7ee-0362c2c38632",
 CellID->1760093074]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3054813, 87588, 207, 5, 40, "MathCaption",ExpressionUUID->"2a3edfc0-e6e4-4a78-bd11-803fcb7234ec"],
Cell[CellGroupData[{
Cell[3055045, 87597, 344, 11, 27, "Input",ExpressionUUID->"ea973fe8-5e9c-41f6-b297-37fc66152ec1"],
Cell[3055392, 87610, 6373, 184, 32, "Output",ExpressionUUID->"200407c5-c80a-436e-8c3c-042cc29330e5",
 CellID->297057293]
}, Open  ]],
Cell[CellGroupData[{
Cell[3061802, 87799, 344, 11, 27, "Input",ExpressionUUID->"520ffa17-6ab4-4593-8681-86c74a0763a0"],
Cell[3062149, 87812, 6599, 190, 32, "Output",ExpressionUUID->"77b5a529-7d0b-4135-916e-00ced6d8776d",
 CellID->1889927823]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3068797, 88008, 123, 0, 40, "MathCaption",ExpressionUUID->"34f08f9e-0658-453b-9d73-ab2cf419bcf4"],
Cell[CellGroupData[{
Cell[3068945, 88012, 295, 9, 27, "Input",ExpressionUUID->"1b00e22f-e1ed-43b3-a756-396321deb724"],
Cell[3069243, 88023, 6600, 190, 32, "Output",ExpressionUUID->"cebc9cc4-df90-4d70-b96e-838a6be7acfd",
 CellID->586298384]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3075892, 88219, 161, 4, 40, "MathCaption",ExpressionUUID->"4a5d6dbc-1894-44f3-9669-40986ab51978"],
Cell[CellGroupData[{
Cell[3076078, 88227, 298, 9, 27, "Input",ExpressionUUID->"6522b774-ec6c-4a0a-9dba-54a429f6d482"],
Cell[3076379, 88238, 4447, 130, 32, "Output",ExpressionUUID->"866b912d-5ffd-4dc6-943d-16879891c955",
 CellID->1261967452]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3080875, 88374, 251, 4, 56, "MathCaption",ExpressionUUID->"5317df4d-5265-47b2-9abb-6cb64a6d7102"],
Cell[CellGroupData[{
Cell[3081151, 88382, 332, 10, 27, "Input",ExpressionUUID->"1b7c6f2e-91d9-4699-a3ea-cf2edcb1ec67"],
Cell[3081486, 88394, 3848, 109, 47, "Output",ExpressionUUID->"4967cf15-c6d0-44d3-a8a8-6ffbd8cbd01c",
 CellID->1338467992]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3085395, 88510, 105, 0, 39, "Subsection",ExpressionUUID->"ba293095-3b69-43c8-a591-3ce503ab871e"],
Cell[3085503, 88512, 496, 9, 57, "Text",ExpressionUUID->"2d24f4fb-a1c1-4199-8ba9-d311b57bc149"],
Cell[CellGroupData[{
Cell[3086024, 88525, 279, 8, 40, "MathCaption",ExpressionUUID->"64759844-33cb-499e-a3a1-036337811057"],
Cell[CellGroupData[{
Cell[3086328, 88537, 250, 7, 27, "Input",ExpressionUUID->"3d5e679c-8409-47a0-914f-8730a1dbbc75"],
Cell[3086581, 88546, 427, 14, 26, "Output",ExpressionUUID->"b24940b6-26e6-4667-92b6-fb5032918e4a",
 CellID->736370330]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3087057, 88566, 117, 0, 40, "MathCaption",ExpressionUUID->"91f0e302-277a-4a5f-b69b-4cf6db973374"],
Cell[CellGroupData[{
Cell[3087199, 88570, 523, 16, 27, "Input",ExpressionUUID->"b1a5fe00-c00c-4ba4-add1-d6f604767120"],
Cell[3087725, 88588, 1184, 40, 42, "Output",ExpressionUUID->"908981c2-9554-48ff-a769-591b06171d76",
 CellID->227878649]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3088958, 88634, 1118, 33, 56, "MathCaption",ExpressionUUID->"1ea0640e-0d6e-49e4-918c-0258aec23067"],
Cell[3090079, 88669, 226, 6, 27, "Input",ExpressionUUID->"47903dce-bacd-4dc5-adf4-7d5a16ee5c11"],
Cell[CellGroupData[{
Cell[3090330, 88679, 173, 4, 27, "Input",ExpressionUUID->"dbd9b37b-9efc-495c-b66b-3e26a85ed20f"],
Cell[3090506, 88685, 425, 14, 26, "Output",ExpressionUUID->"b47e1803-4be1-49c7-80d2-fc47bb159f05",
 CellID->993111142]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3090980, 88705, 113, 0, 40, "MathCaption",ExpressionUUID->"412d1f19-b538-4195-a6bc-c6d17935364f"],
Cell[CellGroupData[{
Cell[3091118, 88709, 158, 4, 27, "Input",ExpressionUUID->"a5994b6e-4ab4-460d-bec7-b778f5f86f85"],
Cell[3091279, 88715, 425, 14, 26, "Output",ExpressionUUID->"85522cee-e09c-4afb-8a50-8164de963da5",
 CellID->976059110]
}, Open  ]],
Cell[CellGroupData[{
Cell[3091741, 88734, 143, 3, 27, "Input",ExpressionUUID->"f8fa0ca4-59be-40be-8dea-206395358983"],
Cell[3091887, 88739, 174, 3, 26, "Output",ExpressionUUID->"891eaf7b-0fd5-427c-a305-2b20199dc6d2",
 CellID->2145365984]
}, Open  ]],
Cell[3092076, 88745, 145, 3, 27, "Input",ExpressionUUID->"1bf5b19f-f635-4d69-b6a7-5bb99c719086"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3092270, 88754, 94, 0, 39, "Subsection",ExpressionUUID->"b6020147-9053-46ad-9704-cf7314c5ef80"],
Cell[3092367, 88756, 140, 2, 23, "Text",ExpressionUUID->"b5e18dd5-99fb-4aa3-95ba-da933a922f50"],
Cell[CellGroupData[{
Cell[3092532, 88762, 130, 0, 40, "MathCaption",ExpressionUUID->"a533edd7-2548-4d98-bb20-9c9cee4ec5b1"],
Cell[CellGroupData[{
Cell[3092687, 88766, 255, 7, 27, "Input",ExpressionUUID->"16961d18-74cc-42f7-ab22-d9c93db109b3"],
Cell[3092945, 88775, 2876, 80, 32, "Output",ExpressionUUID->"35e06ab9-e692-4eea-a675-b869cd07aa50",
 CellID->1216730947]
}, Open  ]]
}, Open  ]],
Cell[3095848, 88859, 217, 3, 23, "Text",ExpressionUUID->"12ed87d3-944f-4628-80e8-8d66fa18aeec"],
Cell[CellGroupData[{
Cell[3096090, 88866, 199, 3, 40, "MathCaption",ExpressionUUID->"e2db8e34-9db5-4483-9ffb-c7c9d5efb56e"],
Cell[CellGroupData[{
Cell[3096314, 88873, 440, 14, 27, "Input",ExpressionUUID->"d8667d58-fb8e-4b98-bae2-43d3b3103664"],
Cell[3096757, 88889, 8591, 246, 32, "Output",ExpressionUUID->"a70cfffd-452d-4991-aae4-18d44773dea9",
 CellID->1257739495]
}, Open  ]],
Cell[CellGroupData[{
Cell[3105385, 89140, 165, 4, 27, "Input",ExpressionUUID->"acc83492-4121-4650-b6f8-01b1c3b9537a"],
Cell[3105553, 89146, 549, 12, 22, "Print",ExpressionUUID->"13c7f4c4-608d-477e-b1f3-636d64b9bc17",
 CellID->1146155044]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3106163, 89165, 112, 0, 39, "Subsection",ExpressionUUID->"5baa5a84-1a5b-47e6-ae3b-f75fea059ce5"],
Cell[3106278, 89167, 195, 3, 23, "Text",ExpressionUUID->"766a2028-dd53-4216-91be-2d9d69c393e4"],
Cell[CellGroupData[{
Cell[3106498, 89174, 109, 0, 40, "MathCaption",ExpressionUUID->"001d1f57-7f67-468d-a614-132d6edf9a41"],
Cell[CellGroupData[{
Cell[3106632, 89178, 506, 11, 27, "Input",ExpressionUUID->"9939a251-ca48-4cff-bcff-4468637a767f"],
Cell[CellGroupData[{
Cell[3107163, 89193, 597, 12, 22, "Print",ExpressionUUID->"f19ce1be-8245-44be-b032-d47330ae1e94",
 CellID->659454615],
Cell[3107763, 89207, 779, 15, 42, "Print",ExpressionUUID->"37c014a3-1496-410f-9025-f33cd1f0a880",
 CellID->290612138]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3108603, 89229, 125, 0, 40, "MathCaption",ExpressionUUID->"0a251e7b-6ed8-4ceb-9cb6-8391edce4646"],
Cell[CellGroupData[{
Cell[3108753, 89233, 335, 9, 27, "Input",ExpressionUUID->"f70d5cdf-3dc4-4103-92c2-8bef4f154530"],
Cell[CellGroupData[{
Cell[3109113, 89246, 681, 15, 22, "Print",ExpressionUUID->"552d49bd-f8a7-490b-a10c-fdd59ea45b1d",
 CellID->940390088],
Cell[3109797, 89263, 804, 18, 22, "Print",ExpressionUUID->"2de43790-d042-426c-b197-064c55ec4cdb",
 CellID->1804050743],
Cell[3110604, 89283, 819, 19, 22, "Print",ExpressionUUID->"ba5f94d6-95db-464a-8a7a-794755907d54",
 CellID->359225608],
Cell[3111426, 89304, 899, 18, 42, "Print",ExpressionUUID->"57f02b9b-efad-4760-912b-2df325d335ee",
 CellID->928261939],
Cell[3112328, 89324, 772, 17, 22, "Print",ExpressionUUID->"35d7fe87-f13d-4421-91ce-a33574161cd7",
 CellID->1977196965],
Cell[3113103, 89343, 270, 4, 22, "Print",ExpressionUUID->"8570cbe2-6713-4d42-aa49-ef277110a0f0",
 CellID->587437077],
Cell[3113376, 89349, 894, 19, 22, "Print",ExpressionUUID->"7c4bf037-3126-455f-9418-705dcd2da0d5",
 CellID->1072627880],
Cell[3114273, 89370, 952, 20, 22, "Print",ExpressionUUID->"d5e179ff-71db-4b5e-97f2-d0810b2068ce",
 CellID->1679046242],
Cell[3115228, 89392, 974, 20, 42, "Print",ExpressionUUID->"1527bd07-f64c-4dc0-acfc-72c8ded6fe87",
 CellID->1795067714],
Cell[3116205, 89414, 1028, 20, 42, "Print",ExpressionUUID->"47bb244f-4532-4248-8793-dec77e9df754",
 CellID->1852796142]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3117294, 89441, 506, 10, 72, "MathCaption",ExpressionUUID->"26e4fac4-e4cd-42d0-9fa8-408deb01147b"],
Cell[CellGroupData[{
Cell[3117825, 89455, 141, 3, 27, "Input",ExpressionUUID->"7c4b7e34-53f5-413f-8602-259c046cda73"],
Cell[CellGroupData[{
Cell[3117991, 89462, 211, 3, 43, "Print",ExpressionUUID->"519da795-919b-40cf-b068-f23f25068295",
 CellTags->"Info6273728981830-9181623",
 CellID->1431877094],
Cell[3118205, 89467, 3967, 119, 363, "Print",ExpressionUUID->"7cc7d0aa-97b2-46e4-8186-fe70e9c01262",
 CellTags->"Info6273728981830-9181623",
 CellID->856570576]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3122221, 89592, 137, 3, 27, "Input",ExpressionUUID->"68f8f6eb-5c45-457d-aae9-1fa3ca02237a"],
Cell[CellGroupData[{
Cell[3122383, 89599, 207, 3, 43, "Print",ExpressionUUID->"25bd47ae-1722-4fac-9b5f-ad3965650133",
 CellTags->"Info6283728981830-9181623",
 CellID->1798784991],
Cell[3122593, 89604, 4463, 133, 363, "Print",ExpressionUUID->"8d196b37-dc33-4d02-aecc-a9fcbf76e838",
 CellTags->"Info6283728981830-9181623",
 CellID->2052044918]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3127105, 89743, 150, 3, 27, "Input",ExpressionUUID->"4c5ea162-50a2-4492-8e00-d7cb539072d4"],
Cell[CellGroupData[{
Cell[3127280, 89750, 219, 3, 43, "Print",ExpressionUUID->"5fcf77bd-09a9-4707-b110-ac2ccc6bfe59",
 CellTags->"Info6293728981831-9181623",
 CellID->131566006],
Cell[3127502, 89755, 4001, 117, 329, "Print",ExpressionUUID->"e36f34d4-fbb6-4f39-a840-90e6ad5ebd86",
 CellTags->"Info6293728981831-9181623",
 CellID->1469190079]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3131552, 89878, 146, 3, 27, "Input",ExpressionUUID->"862db3f2-a585-4bcf-a28a-3b888e787291"],
Cell[CellGroupData[{
Cell[3131723, 89885, 215, 3, 43, "Print",ExpressionUUID->"e2ece981-4b04-4bc6-98a0-fb537d10ef48",
 CellTags->"Info6303728981831-9181623",
 CellID->910684547],
Cell[3131941, 89890, 4511, 132, 329, "Print",ExpressionUUID->"3ddd1743-5669-46ff-aed9-4d27b5ffd144",
 CellTags->"Info6303728981831-9181623",
 CellID->759229260]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3136513, 90029, 397, 11, 56, "MathCaption",ExpressionUUID->"4bbba8af-7283-424f-b0da-86d37c92c637"],
Cell[CellGroupData[{
Cell[3136935, 90044, 290, 7, 27, "Input",ExpressionUUID->"dca1197c-db66-4443-9a88-53b997314897"],
Cell[CellGroupData[{
Cell[3137250, 90055, 699, 13, 22, "Print",ExpressionUUID->"efa2f69d-f92b-4237-875c-4e4e2eb6cc31",
 CellID->18342346],
Cell[3137952, 90070, 737, 14, 22, "Print",ExpressionUUID->"b4b786c4-6283-4dbb-a2d8-9f9123a0b2f5",
 CellID->1189647701]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3138738, 90090, 254, 7, 27, "Input",ExpressionUUID->"f71e6421-db58-46ed-aaf8-6fd9f6fa9224"],
Cell[3138995, 90099, 1223, 34, 32, "Output",ExpressionUUID->"efce94b3-3fb7-4035-8b0f-dfcb54d319a6",
 CellID->1336157137]
}, Open  ]],
Cell[CellGroupData[{
Cell[3140255, 90138, 171, 4, 27, "Input",ExpressionUUID->"4317f374-5a4c-4392-a194-13863464fd4f"],
Cell[3140429, 90144, 5514, 153, 32, "Output",ExpressionUUID->"06e998a2-b538-41b2-b3dc-c946e177eb0e",
 CellID->556504880]
}, Open  ]],
Cell[CellGroupData[{
Cell[3145980, 90302, 182, 4, 27, "Input",ExpressionUUID->"1c4d9b76-b26b-4081-a248-bec7d522bea4"],
Cell[3146165, 90308, 1222, 34, 32, "Output",ExpressionUUID->"f87e6009-880f-4c6c-840a-fff673f91e48",
 CellID->886235475]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3147436, 90348, 112, 0, 40, "MathCaption",ExpressionUUID->"127adeab-ebf1-434e-bee9-2453fd82a41a"],
Cell[CellGroupData[{
Cell[3147573, 90352, 325, 10, 27, "Input",ExpressionUUID->"8a81acea-58e3-4c19-9933-801be4615b80"],
Cell[3147901, 90364, 1286, 36, 32, "Output",ExpressionUUID->"df53dfaa-391d-41b7-8d16-ccb1779bfd5f",
 CellID->13052733]
}, Open  ]],
Cell[CellGroupData[{
Cell[3149224, 90405, 145, 3, 27, "Input",ExpressionUUID->"66a241f2-e7a9-4688-91e5-d0487777e60b"],
Cell[3149372, 90410, 5609, 155, 32, "Output",ExpressionUUID->"201bf36c-8ec0-438c-95f5-77ff9c867b73",
 CellID->333951537]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3155030, 90571, 107, 0, 40, "MathCaption",ExpressionUUID->"36e1b7d9-c936-4c85-8818-388bd6fa554e"],
Cell[CellGroupData[{
Cell[3155162, 90575, 290, 9, 27, "Input",ExpressionUUID->"ad97cd95-e30e-4141-b7fe-f69bd478eac7"],
Cell[3155455, 90586, 1259, 35, 32, "Output",ExpressionUUID->"f2287a8a-c75d-49b8-a61f-c4a365eafd31",
 CellID->1348712748]
}, Open  ]],
Cell[CellGroupData[{
Cell[3156751, 90626, 171, 4, 27, "Input",ExpressionUUID->"5d3fbb29-ecac-44fa-8695-55c9eeece61c"],
Cell[3156925, 90632, 5708, 157, 32, "Output",ExpressionUUID->"a282bfbe-be47-491f-9450-2e70f3724d80",
 CellID->1125261551]
}, Open  ]],
Cell[CellGroupData[{
Cell[3162670, 90794, 364, 12, 27, "Input",ExpressionUUID->"7f2e09cd-6399-4066-8a6c-7d8d9986c3af"],
Cell[3163037, 90808, 1328, 37, 32, "Output",ExpressionUUID->"94826702-3337-4442-9d43-65d4610ee5f2",
 CellID->1529892175]
}, Open  ]],
Cell[CellGroupData[{
Cell[3164402, 90850, 145, 3, 27, "Input",ExpressionUUID->"b363bf6e-13b5-43ea-af2a-75d93b1bc31f"],
Cell[3164550, 90855, 5804, 160, 32, "Output",ExpressionUUID->"af80de5c-c56b-4cb3-999b-95849585cd9d",
 CellID->1076079694]
}, Open  ]],
Cell[CellGroupData[{
Cell[3170391, 91020, 147, 3, 27, "Input",ExpressionUUID->"163854a0-a861-4903-bc9f-17cd15268a64"],
Cell[CellGroupData[{
Cell[3170563, 91027, 567, 12, 22, "Print",ExpressionUUID->"bc9ad22d-9018-4139-8821-b7b1e4926cf5",
 CellID->1711625671],
Cell[3171133, 91041, 549, 12, 22, "Print",ExpressionUUID->"5b9cb26f-7ad5-42bc-b1d2-a8f39378c62d",
 CellID->1749200932]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[3171721, 91058, 830, 29, 57, "Text",ExpressionUUID->"00e25242-9ad9-468c-b4a1-f4283104d420"],
Cell[CellGroupData[{
Cell[3172576, 91091, 181, 5, 40, "MathCaption",ExpressionUUID->"e9907f0c-d120-49ef-9b82-cc69c7e26bbf"],
Cell[CellGroupData[{
Cell[3172782, 91100, 115, 2, 27, "Input",ExpressionUUID->"afb606c6-79d4-44af-a698-2072e0c25507"],
Cell[3172900, 91104, 230, 5, 26, "Output",ExpressionUUID->"ea848941-fbee-4591-8ebe-7ceeba2e6304",
 CellID->1916420341]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3173179, 91115, 174, 5, 40, "MathCaption",ExpressionUUID->"cf043f7c-ea17-4ad9-8e1c-6d20afec86b4"],
Cell[CellGroupData[{
Cell[3173378, 91124, 386, 10, 27, "Input",ExpressionUUID->"c3129cdb-c2db-4561-afa9-b9341f909618"],
Cell[CellGroupData[{
Cell[3173789, 91138, 680, 15, 22, "Print",ExpressionUUID->"cbca5d5f-bf0f-4c18-a160-965ab455e5e4",
 CellID->1316933481],
Cell[3174472, 91155, 790, 17, 22, "Print",ExpressionUUID->"189038da-e105-45df-aa1b-20cb119e9598",
 CellID->1734549185],
Cell[3175265, 91174, 893, 19, 22, "Print",ExpressionUUID->"ff80f269-feba-46e4-83ca-23215b3f5c71",
 CellID->477644270],
Cell[3176161, 91195, 949, 20, 22, "Print",ExpressionUUID->"fa0c5709-6927-4335-b207-536afaa3b80a",
 CellID->473373404],
Cell[3177113, 91217, 974, 20, 42, "Print",ExpressionUUID->"f0cce46d-8ab8-41e6-9645-b3bdaf7f26ae",
 CellID->1707372540],
Cell[3178090, 91239, 1028, 20, 42, "Print",ExpressionUUID->"2f149c8c-ae94-4aac-bf0a-974ec1c28f7b",
 CellID->1133788566]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3179179, 91266, 179, 3, 40, "MathCaption",ExpressionUUID->"f9cf64d5-459a-4a61-962f-e9b42825abde"],
Cell[CellGroupData[{
Cell[3179383, 91273, 542, 17, 27, "Input",ExpressionUUID->"1f027cf2-8da2-4fed-9840-dad987a32405"],
Cell[3179928, 91292, 2392, 65, 32, "Output",ExpressionUUID->"d6a1217c-cdf1-4dfc-843c-529e5704731d",
 CellID->962028931]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3182369, 91363, 150, 2, 40, "MathCaption",ExpressionUUID->"701004f9-4705-4297-94eb-712725739681"],
Cell[CellGroupData[{
Cell[3182544, 91369, 487, 14, 27, "Input",ExpressionUUID->"39142340-08af-4887-a36a-4c0a348b7051"],
Cell[3183034, 91385, 2460, 65, 32, "Output",ExpressionUUID->"e352f8ed-e7ac-4794-beda-ebd5da151d06",
 CellID->415865054]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3185543, 91456, 181, 3, 40, "MathCaption",ExpressionUUID->"d8ecfcb8-cdbb-4e0e-b992-419a9912a4c3"],
Cell[CellGroupData[{
Cell[3185749, 91463, 146, 3, 27, "Input",ExpressionUUID->"a5f0721b-1222-492f-8e91-0ba2c63b58f6"],
Cell[CellGroupData[{
Cell[3185920, 91470, 653, 14, 22, "Print",ExpressionUUID->"373cd9a0-6592-4413-bd55-db658750312b",
 CellID->1532175240],
Cell[3186576, 91486, 626, 12, 22, "Print",ExpressionUUID->"d821b322-5ae1-40da-b4d7-af5c83cee540",
 CellID->792847503],
Cell[3187205, 91500, 576, 12, 22, "Print",ExpressionUUID->"5152685e-9c0f-4deb-9034-e5ff173b1154",
 CellID->251953286],
Cell[3187784, 91514, 607, 13, 22, "Print",ExpressionUUID->"cd21f6fb-1f33-4318-a335-845b9b2d3aa7",
 CellID->1888609565],
Cell[3188394, 91529, 585, 12, 22, "Print",ExpressionUUID->"b0bf257a-928a-421f-a413-e31b434705b4",
 CellID->2052656295],
Cell[3188982, 91543, 575, 12, 22, "Print",ExpressionUUID->"84135a63-52d7-4fac-97f3-9c3eea2c6f31",
 CellID->1817044657]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3189606, 91561, 654, 21, 27, "Input",ExpressionUUID->"9de6d6bb-2bc6-40b3-a6a1-1db1685966b6"],
Cell[3190263, 91584, 3436, 94, 32, "Output",ExpressionUUID->"d5d074ce-3de0-4cb1-8353-8a86a005da8c",
 CellID->1852089859]
}, Open  ]],
Cell[CellGroupData[{
Cell[3193736, 91683, 654, 21, 27, "Input",ExpressionUUID->"0dc6c194-7f0b-4ca6-8568-0d230d68c792"],
Cell[3194393, 91706, 714, 22, 26, "Output",ExpressionUUID->"d46d2d73-e8bc-4191-9663-b5a90a490dc2",
 CellID->298796948]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3195180, 91736, 862, 14, 54, "Section",ExpressionUUID->"ba260751-a6a0-4d09-85a7-cab77ca50fa1",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1069634512],
Cell[CellGroupData[{
Cell[3196067, 91754, 92, 0, 39, "Subsection",ExpressionUUID->"c5eda06a-00e0-44e6-b436-fa80f5bb631b"],
Cell[3196162, 91756, 112, 0, 23, "Text",ExpressionUUID->"53619325-7d62-40b1-bc18-bd1c28d73315"],
Cell[3196277, 91758, 2044, 58, 126, "DefinitionBox",ExpressionUUID->"57bc3550-d40d-4c40-a3c3-6308bb201068"],
Cell[3198324, 91818, 94, 0, 23, "Text",ExpressionUUID->"06a9f87d-6ec6-48ab-87da-d6541351219d"],
Cell[CellGroupData[{
Cell[3198443, 91822, 854, 28, 88, "MathCaption",ExpressionUUID->"d64e1593-1fd5-4647-899d-d71b66d45c0c"],
Cell[CellGroupData[{
Cell[3199322, 91854, 349, 10, 27, "Input",ExpressionUUID->"6dca4d8c-1eb7-439c-8f3e-ac41944b22c2"],
Cell[CellGroupData[{
Cell[3199696, 91868, 763, 17, 22, "Print",ExpressionUUID->"34c64916-6003-4cef-9bb2-a4eee95b600c",
 CellID->829774745],
Cell[3200462, 91887, 821, 19, 22, "Print",ExpressionUUID->"e2315ba0-e805-435a-8de3-86b08a383563",
 CellID->265335597],
Cell[3201286, 91908, 680, 15, 22, "Print",ExpressionUUID->"9e82f6aa-57ff-4a03-a67f-c8e944cea348",
 CellID->1319327607],
Cell[3201969, 91925, 802, 18, 22, "Print",ExpressionUUID->"80a12a66-47d1-4273-b28b-45121cacc07e",
 CellID->1170435574],
Cell[3202774, 91945, 817, 19, 22, "Print",ExpressionUUID->"0857f64d-25ac-4e77-8a5a-e87007c60414",
 CellID->560485698],
Cell[3203594, 91966, 844, 19, 22, "Print",ExpressionUUID->"8105c94d-e565-4c34-a850-1473d1f1a122",
 CellID->1219850544],
Cell[3204441, 91987, 759, 17, 22, "Print",ExpressionUUID->"5d77d0fd-f4d2-448b-a3d4-49fb791fb8a9",
 CellID->1715099],
Cell[3205203, 92006, 270, 4, 22, "Print",ExpressionUUID->"c4cea385-a714-4213-8b9c-093adea4acb8",
 CellID->908381482],
Cell[3205476, 92012, 647, 13, 22, "Print",ExpressionUUID->"047426d4-b18e-4ac7-9ad5-f29468d57faa",
 CellID->706321451],
Cell[3206126, 92027, 273, 4, 22, "Print",ExpressionUUID->"d7714970-e8e1-45b6-8abe-641dc265916a",
 CellID->2002509988],
Cell[3206402, 92033, 777, 18, 22, "Print",ExpressionUUID->"d6d5bd93-4fb8-4462-82b1-23e40e08e153",
 CellID->821652532],
Cell[3207182, 92053, 852, 19, 22, "Print",ExpressionUUID->"da8ee768-0ce6-46d1-9f92-49846f265f32",
 CellID->1194255114],
Cell[3208037, 92074, 772, 17, 22, "Print",ExpressionUUID->"30a8eca9-850d-4b6f-a5fb-a9c0e4cc9924",
 CellID->1463536136],
Cell[3208812, 92093, 656, 13, 22, "Print",ExpressionUUID->"32e7435a-a072-475e-9aa3-9c94813b8323",
 CellID->374145014],
Cell[3209471, 92108, 405, 8, 22, "Print",ExpressionUUID->"2c0655e6-7f8d-4e3d-a602-c966d441c5a7",
 CellID->423131882],
Cell[3209879, 92118, 396, 8, 22, "Print",ExpressionUUID->"030e58d4-589b-4cd7-b1db-9746b1cdfc58",
 CellID->1259185557],
Cell[3210278, 92128, 410, 8, 22, "Print",ExpressionUUID->"6bc3025c-5f84-4579-a495-cd771816efa4",
 CellID->1827798961],
Cell[3210691, 92138, 672, 13, 22, "Print",ExpressionUUID->"df278643-30b9-4912-9320-042464b57020",
 CellID->63567192]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3211424, 92158, 89, 0, 40, "MathCaption",ExpressionUUID->"493f36b8-f608-45dc-b233-d813e77fc8a2"],
Cell[CellGroupData[{
Cell[3211538, 92162, 245, 7, 27, "Input",ExpressionUUID->"226f16f6-3a2b-4348-8fb6-ab3550da4077"],
Cell[3211786, 92171, 1190, 33, 32, "Output",ExpressionUUID->"883f1966-ea23-4664-9c7c-3b00abb5b657",
 CellID->1074423385]
}, Open  ]],
Cell[CellGroupData[{
Cell[3213013, 92209, 141, 3, 27, "Input",ExpressionUUID->"7a22a559-4394-43e6-8819-36fa6b2b0753"],
Cell[3213157, 92214, 188, 3, 26, "Output",ExpressionUUID->"e6c75474-a260-44c0-ab6e-c62c8a795f9c",
 CellID->1791327719]
}, Open  ]],
Cell[CellGroupData[{
Cell[3213382, 92222, 226, 6, 27, "Input",ExpressionUUID->"476c5f6d-ee2e-4b0d-b307-6f2d9d4a7aa1"],
Cell[3213611, 92230, 377, 11, 26, "Output",ExpressionUUID->"8ccf4593-9514-4d30-9127-cfccbea1dce7",
 CellID->1271725940]
}, Open  ]],
Cell[CellGroupData[{
Cell[3214025, 92246, 141, 3, 27, "Input",ExpressionUUID->"6a2ad638-4ada-420d-94c5-4e119fa4189a"],
Cell[3214169, 92251, 188, 3, 26, "Output",ExpressionUUID->"a00d8838-dd24-4afe-bded-478e39af129b",
 CellID->2140618636]
}, Open  ]],
Cell[CellGroupData[{
Cell[3214394, 92259, 283, 9, 27, "Input",ExpressionUUID->"1a1514a7-0a1a-4357-a95d-e38ebb1b8a3e"],
Cell[3214680, 92270, 170, 3, 26, "Output",ExpressionUUID->"a3551895-904d-4a0c-9695-e8cbd0940255",
 CellID->583877061]
}, Open  ]],
Cell[CellGroupData[{
Cell[3214887, 92278, 283, 9, 27, "Input",ExpressionUUID->"d0671f74-dec3-45d8-9155-a83dd04ca381"],
Cell[3215173, 92289, 1215, 34, 32, "Output",ExpressionUUID->"74b3fee8-05f2-41cc-b04c-dde855c1c428",
 CellID->1368981084]
}, Open  ]],
Cell[CellGroupData[{
Cell[3216425, 92328, 260, 7, 27, "Input",ExpressionUUID->"01719be4-35dd-46bd-bf23-e258d4134d50"],
Cell[3216688, 92337, 170, 3, 26, "Output",ExpressionUUID->"e8b1de85-146d-44f1-a409-10579f2436f6",
 CellID->325284830]
}, Open  ]],
Cell[CellGroupData[{
Cell[3216895, 92345, 246, 7, 27, "Input",ExpressionUUID->"b711b162-c513-47f3-ad2f-841f5bbc7598"],
Cell[3217144, 92354, 1270, 35, 32, "Output",ExpressionUUID->"7ef6601f-1e38-43f9-88b9-a3a41fb03a71",
 CellID->1716256195]
}, Open  ]],
Cell[CellGroupData[{
Cell[3218451, 92394, 246, 7, 27, "Input",ExpressionUUID->"182c4435-0a60-4a56-a95f-ce156954a058"],
Cell[3218700, 92403, 170, 3, 26, "Output",ExpressionUUID->"518b14f8-8a63-49fa-a7f8-3cee87d163cb",
 CellID->117522861]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3218919, 92412, 275, 6, 56, "MathCaption",ExpressionUUID->"875d9f9e-9853-4d79-b057-e6299a4b5b87"],
Cell[CellGroupData[{
Cell[3219219, 92422, 141, 3, 27, "Input",ExpressionUUID->"046cd19f-c17d-41b1-a5ff-68f31c9a33f1"],
Cell[3219363, 92427, 457, 15, 29, "Output",ExpressionUUID->"c55dded4-0361-49dd-b705-8ddf70a1061b",
 CellID->1669253234]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3219869, 92448, 129, 0, 40, "MathCaption",ExpressionUUID->"3c01bc27-f702-423d-9c3d-656cb6db21df"],
Cell[CellGroupData[{
Cell[3220023, 92452, 261, 8, 27, "Input",ExpressionUUID->"5e124beb-30ff-44ec-97c1-093ac0e41830"],
Cell[3220287, 92462, 170, 3, 26, "Output",ExpressionUUID->"a2f82dfd-5a72-43a3-8a71-5af2c91689ca",
 CellID->412739430]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3220506, 92471, 131, 0, 40, "MathCaption",ExpressionUUID->"7476d2f3-12c6-46bd-984d-0e3cc469c628"],
Cell[CellGroupData[{
Cell[3220662, 92475, 206, 6, 27, "Input",ExpressionUUID->"be194554-b2d4-4e96-8152-9f24bfe9f659"],
Cell[3220871, 92483, 1145, 32, 32, "Output",ExpressionUUID->"286811ce-c000-466e-9ef9-dd87597608ba",
 CellID->539420200]
}, Open  ]],
Cell[CellGroupData[{
Cell[3222053, 92520, 187, 5, 27, "Input",ExpressionUUID->"144239ed-7062-4a22-a984-9aac9035a3fb"],
Cell[3222243, 92527, 171, 3, 26, "Output",ExpressionUUID->"fe659049-4f49-4e21-bf96-611b553a57b5",
 CellID->1506827413]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3222463, 92536, 264, 5, 56, "MathCaption",ExpressionUUID->"e32cd178-bbd0-4c87-aa1a-f249e1603a48"],
Cell[CellGroupData[{
Cell[3222752, 92545, 149, 3, 27, "Input",ExpressionUUID->"299fd2ba-cb2e-4a8c-ab35-1a413704a471"],
Cell[3222904, 92550, 793, 26, 26, "Output",ExpressionUUID->"536f4bcc-1851-4e9f-9dda-ce6803e3d4b7",
 CellID->2121164867]
}, Open  ]],
Cell[CellGroupData[{
Cell[3223734, 92581, 206, 5, 27, "Input",ExpressionUUID->"a7924818-9ba2-4d90-8384-aca80272d058"],
Cell[3223943, 92588, 1109, 31, 32, "Output",ExpressionUUID->"7f665a72-c883-457a-8097-37197b25ab31",
 CellID->41895822]
}, Open  ]],
Cell[CellGroupData[{
Cell[3225089, 92624, 283, 8, 27, "Input",ExpressionUUID->"b2022f06-2f21-42cd-9836-dd4863b7f31d"],
Cell[3225375, 92634, 693, 13, 22, "Print",ExpressionUUID->"a710aabc-87a4-415b-aa33-6db98801af57",
 CellID->1915927785]
}, Open  ]],
Cell[CellGroupData[{
Cell[3226105, 92652, 149, 3, 27, "Input",ExpressionUUID->"fae06c02-6681-4554-9a7e-3500e27251fe"],
Cell[3226257, 92657, 720, 23, 26, "Output",ExpressionUUID->"bdec889f-163b-4476-a8df-c678443a6519",
 CellID->1628479377]
}, Open  ]],
Cell[CellGroupData[{
Cell[3227014, 92685, 226, 6, 27, "Input",ExpressionUUID->"ed8a1d02-d2be-47ef-a115-f58d3faa09bf"],
Cell[3227243, 92693, 427, 9, 22, "Message",ExpressionUUID->"d52fa4e0-b263-4b88-b490-30b27aecb66f",
 CellID->173808224],
Cell[3227673, 92704, 1461, 41, 32, "Output",ExpressionUUID->"d3978747-d853-4e60-bf2a-974fc339dd0b",
 CellID->617060848]
}, Open  ]],
Cell[CellGroupData[{
Cell[3229171, 92750, 147, 3, 27, "Input",ExpressionUUID->"6fccf0eb-37c3-463b-a049-2e2eec9e0967"],
Cell[3229321, 92755, 548, 12, 22, "Print",ExpressionUUID->"613be26c-33e8-409d-a251-ac2c1d47be1e",
 CellID->743140973]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3229918, 92773, 220, 5, 40, "MathCaption",ExpressionUUID->"8e11a471-ba59-4402-9a7b-519e6066dd3f"],
Cell[CellGroupData[{
Cell[3230163, 92782, 269, 8, 27, "Input",ExpressionUUID->"651ec90d-f1d2-4c04-a41a-4503c39829b3"],
Cell[3230435, 92792, 3320, 95, 32, "Output",ExpressionUUID->"fbfdeb40-061d-4e93-a8cc-3f23139711d8",
 CellID->205442259]
}, Open  ]],
Cell[CellGroupData[{
Cell[3233792, 92892, 155, 3, 27, "Input",ExpressionUUID->"0774afae-ae71-414c-80e3-21262a85d13c"],
Cell[3233950, 92897, 6748, 188, 40, "Output",ExpressionUUID->"fd4e4059-faf3-43da-85a5-bde33c93d3dd",
 CellID->1861826284]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3240759, 93092, 93, 0, 39, "Subsection",ExpressionUUID->"729bc3fa-0125-43cd-af49-29a91841ae85"],
Cell[3240855, 93094, 802, 13, 91, "Text",ExpressionUUID->"6d4a01c2-2565-4a3e-93c7-72907fa67f7c"],
Cell[3241660, 93109, 1227, 31, 75, "DefinitionBox",ExpressionUUID->"aefcc30d-9833-4dd5-8ae6-4a2ce6907983"],
Cell[3242890, 93142, 90, 0, 23, "Text",ExpressionUUID->"7b516783-a40c-48d0-ad93-a09957515020"],
Cell[CellGroupData[{
Cell[3243005, 93146, 139, 1, 40, "MathCaption",ExpressionUUID->"2fba4edd-de0c-4fc9-9d12-fc60e952c557"],
Cell[CellGroupData[{
Cell[3243169, 93151, 283, 9, 27, "Input",ExpressionUUID->"7bbe88c4-69f7-4bd2-a59d-8b156b8b8d42"],
Cell[3243455, 93162, 1037, 30, 32, "Output",ExpressionUUID->"cb4f743e-9d21-4cb9-855f-a295925379c7"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3244529, 93197, 185, 5, 27, "Input",ExpressionUUID->"e733692f-9d3f-4ce0-90f4-4705f976cef1"],
Cell[3244717, 93204, 111, 2, 26, "Output",ExpressionUUID->"1fd002b2-971b-4e68-958b-6458a6adfcf3"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3244877, 93212, 272, 4, 56, "MathCaption",ExpressionUUID->"74a57a0c-4f43-423d-9e60-445aa1727c87"],
Cell[CellGroupData[{
Cell[3245174, 93220, 257, 8, 27, "Input",ExpressionUUID->"caa7cc6b-814d-41b3-bbab-ee7a79e9c6fa"],
Cell[3245434, 93230, 2024, 59, 32, "Output",ExpressionUUID->"3ce05824-3e3f-4e5a-87e2-91fe9b0028a6"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3247495, 93294, 179, 4, 27, "Input",ExpressionUUID->"0b74dc0d-e9af-414e-b92d-db741212fa91"],
Cell[3247677, 93300, 1025, 30, 32, "Output",ExpressionUUID->"6ea8097e-ba54-402f-9866-fe455aeeb760"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3248751, 93336, 355, 5, 72, "MathCaption",ExpressionUUID->"acdc3df6-8f77-4390-a6ba-a1c9c62367e2"],
Cell[CellGroupData[{
Cell[3249131, 93345, 243, 7, 27, "Input",ExpressionUUID->"ea4be80b-a9e1-4c53-9e0f-cebe6138f8e5"],
Cell[3249377, 93354, 3151, 91, 32, "Output",ExpressionUUID->"e1e7fd0f-439a-4c75-95d1-2018eb17cb7e"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3252565, 93450, 263, 8, 27, "Input",ExpressionUUID->"6e7cb503-4e77-490f-aeb1-afd9f61fe9f0"],
Cell[3252831, 93460, 111, 2, 26, "Output",ExpressionUUID->"bcc4504c-8f41-4b42-a08b-2a6556b21e6c"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3252991, 93468, 258, 6, 56, "MathCaption",ExpressionUUID->"8af64153-3662-46ac-8e01-44226966a578"],
Cell[CellGroupData[{
Cell[3253274, 93478, 405, 13, 27, "Input",ExpressionUUID->"363a9176-94c8-4052-806a-c49952df56a9"],
Cell[3253682, 93493, 2139, 62, 32, "Output",ExpressionUUID->"356aea68-4410-410b-aa44-0e0c7a0a797f"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3255858, 93560, 465, 13, 27, "Input",ExpressionUUID->"f19a9d7c-f615-4399-a92e-1cc57247f2c0"],
Cell[3256326, 93575, 1139, 33, 32, "Output",ExpressionUUID->"6c3b90b5-fc87-4fd5-aa69-231e6e0ee154"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3257514, 93614, 257, 6, 56, "MathCaption",ExpressionUUID->"e768066d-b5a2-42be-a860-6f1494c00caa"],
Cell[CellGroupData[{
Cell[3257796, 93624, 428, 14, 27, "Input",ExpressionUUID->"9345c044-5e5b-49ad-9e08-9bb85a9a76dd"],
Cell[3258227, 93640, 1147, 33, 32, "Output",ExpressionUUID->"e0a72705-93cc-42ad-96cf-86abb9342098"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3259411, 93678, 428, 14, 27, "Input",ExpressionUUID->"dcb80acd-f916-4951-9a92-02cc42d66e12"],
Cell[3259842, 93694, 2140, 62, 32, "Output",ExpressionUUID->"ece2356e-9a14-418d-889b-72843b2752ca"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3262031, 93762, 255, 4, 56, "MathCaption",ExpressionUUID->"b57bc7c4-a2b4-43cd-85fd-669ca9f76684"],
Cell[CellGroupData[{
Cell[3262311, 93770, 482, 14, 27, "Input",ExpressionUUID->"c1ac7d4b-a1ee-43b1-9fb9-ca89c7a7ecbe"],
Cell[3262796, 93786, 4267, 122, 32, "Output",ExpressionUUID->"b5588248-ceb0-4d90-8a00-229ff58e9b40"]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3267124, 93915, 194, 4, 39, "Subsection",ExpressionUUID->"d26708be-fc7e-4080-aa5e-f61d015ebfbe"],
Cell[3267321, 93921, 593, 16, 57, "Text",ExpressionUUID->"6f1f20a3-4fd7-43f8-8a28-acdaac257764"],
Cell[CellGroupData[{
Cell[3267939, 93941, 283, 9, 40, "MathCaption",ExpressionUUID->"24450e09-7be9-41ed-8d0c-94bfb39b636d"],
Cell[CellGroupData[{
Cell[3268247, 93954, 405, 12, 27, "Input",ExpressionUUID->"e514a8ad-b2c1-41dc-8813-1cbea531bdca"],
Cell[3268655, 93968, 2103, 60, 32, "Output",ExpressionUUID->"7df159bd-bd76-4dce-b066-af258a3dd335",
 CellID->28637891]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3270807, 94034, 211, 3, 56, "MathCaption",ExpressionUUID->"1755d33c-f713-4354-8eff-bacca0203c51"],
Cell[CellGroupData[{
Cell[3271043, 94041, 325, 10, 27, "Input",ExpressionUUID->"77c9b493-fbdf-4794-ad19-0a2016c726e4"],
Cell[3271371, 94053, 2104, 60, 32, "Output",ExpressionUUID->"7d08f7d8-6507-44c8-8a2c-27cc598bdeff",
 CellID->280697109]
}, Open  ]]
}, Open  ]],
Cell[3273502, 94117, 1356, 32, 142, "Text",ExpressionUUID->"4aa9f8cb-b443-4c21-a144-5346e5286463"],
Cell[CellGroupData[{
Cell[3274883, 94153, 251, 8, 40, "MathCaption",ExpressionUUID->"60b93444-1bee-46e2-bb91-6f1def9e7093"],
Cell[CellGroupData[{
Cell[3275159, 94165, 545, 17, 27, "Input",ExpressionUUID->"a69199cf-b7f7-4249-8936-7870554be859"],
Cell[3275707, 94184, 2459, 66, 32, "Output",ExpressionUUID->"07b0ee99-d9ee-43b4-9f15-d6665de5e3a8",
 CellID->1802222151]
}, Open  ]],
Cell[CellGroupData[{
Cell[3278203, 94255, 460, 15, 27, "Input",ExpressionUUID->"293caad7-f96f-480a-b0b3-cd72ea5fd170"],
Cell[3278666, 94272, 2459, 66, 32, "Output",ExpressionUUID->"12cd9674-152c-4aa6-90e4-ba00ea8360b1",
 CellID->1838334569]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3281174, 94344, 167, 5, 40, "MathCaption",ExpressionUUID->"e387efa2-837e-4948-ae56-7be100ae0000"],
Cell[CellGroupData[{
Cell[3281366, 94353, 545, 17, 27, "Input",ExpressionUUID->"3766483b-dca2-4214-bd0c-da67e4016776"],
Cell[3281914, 94372, 2405, 66, 32, "Output",ExpressionUUID->"9caaaba4-04ba-4887-9a35-80fc0ed9f7d6",
 CellID->1041871794]
}, Open  ]],
Cell[CellGroupData[{
Cell[3284356, 94443, 460, 15, 27, "Input",ExpressionUUID->"9f123f4e-9dc3-4855-9166-753a0d2c9669"],
Cell[3284819, 94460, 445, 9, 22, "Message",ExpressionUUID->"abb712e9-212c-4481-a17b-57e363d2118b",
 CellID->157584120],
Cell[3285267, 94471, 2404, 66, 32, "Output",ExpressionUUID->"2ee6d7b4-f1e2-4baa-b82b-5b924bd371e4",
 CellID->668968089]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3287720, 94543, 357, 9, 56, "MathCaption",ExpressionUUID->"8c60933b-99f1-4732-b5a0-524af687f94b"],
Cell[CellGroupData[{
Cell[3288102, 94556, 227, 6, 27, "Input",ExpressionUUID->"c0dac4ab-9b07-4381-b212-2ec1afd8594f"],
Cell[3288332, 94564, 2221, 63, 32, "Output",ExpressionUUID->"92b6c5ef-8d18-43bd-b8c1-af148227914f",
 CellID->1269779778]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3290602, 94633, 195, 3, 40, "MathCaption",ExpressionUUID->"5cbd2977-d36d-4829-9441-90e29e300eb0"],
Cell[CellGroupData[{
Cell[3290822, 94640, 220, 5, 27, "Input",ExpressionUUID->"0760b1d2-a6d5-4722-98f6-5dc26f5e7d02"],
Cell[3291045, 94647, 2223, 63, 32, "Output",ExpressionUUID->"b70b3931-4ce6-44d3-a36b-b1e8e6f2c683",
 CellID->1643728607]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3293317, 94716, 203, 5, 40, "MathCaption",ExpressionUUID->"da6be665-8ce2-44d3-8a33-9c4061eea484"],
Cell[CellGroupData[{
Cell[3293545, 94725, 544, 17, 27, "Input",ExpressionUUID->"d5cd31ad-0cba-42e4-baa7-ed857c8c02db"],
Cell[3294092, 94744, 447, 9, 22, "Message",ExpressionUUID->"f443f3e0-26c1-4a0c-a3b6-8b698fb76567",
 CellID->917062841],
Cell[3294542, 94755, 445, 9, 22, "Message",ExpressionUUID->"d042badb-6e6d-4f66-86ef-ab3415437ef3",
 CellID->585034168],
Cell[3294990, 94766, 446, 9, 22, "Message",ExpressionUUID->"88a9fdb5-c0e7-4c38-942f-70a81a3d59aa",
 CellID->1489125610],
Cell[3295439, 94777, 445, 9, 22, "Message",ExpressionUUID->"de1596c1-ff15-4e93-acfd-07c932d46e89",
 CellID->959705132],
Cell[3295887, 94788, 472, 9, 22, "Message",ExpressionUUID->"bea8c403-eb60-4b10-83ce-a7d3deb1d781",
 CellID->14480907],
Cell[3296362, 94799, 2403, 66, 32, "Output",ExpressionUUID->"daf6ffd7-1d4a-421e-b1b9-ecb00430a385",
 CellID->714951838]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3298814, 94871, 286, 6, 56, "MathCaption",ExpressionUUID->"b435c987-dcb0-4749-9b2a-07dc3fbac95b"],
Cell[CellGroupData[{
Cell[3299125, 94881, 531, 17, 27, "Input",ExpressionUUID->"51cb68b8-b11a-43da-92a5-317a822df33b"],
Cell[3299659, 94900, 446, 9, 22, "Message",ExpressionUUID->"c119ff55-305d-4992-a805-ba0f53adea43",
 CellID->1024536301],
Cell[3300108, 94911, 10144, 276, 32, "Output",ExpressionUUID->"1020b617-d71a-4533-acc3-4fa1bad576ea",
 CellID->857729261]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3310301, 95193, 191, 3, 40, "MathCaption",ExpressionUUID->"05818001-6b59-4d49-a488-5eec5102318f"],
Cell[CellGroupData[{
Cell[3310517, 95200, 542, 18, 27, "Input",ExpressionUUID->"b575d3c4-d4b8-4713-b4eb-9194a009464f"],
Cell[3311062, 95220, 446, 9, 22, "Message",ExpressionUUID->"feddb210-91d9-4378-865a-2095c7432b58",
 CellID->1945450963],
Cell[3311511, 95231, 446, 9, 22, "Message",ExpressionUUID->"88dd3740-dba2-45b4-8db1-690e9e689c81",
 CellID->1755809566],
Cell[3311960, 95242, 448, 9, 22, "Message",ExpressionUUID->"52d2546b-db12-468d-889f-13cbd05f2b21",
 CellID->1318582335],
Cell[3312411, 95253, 474, 9, 22, "Message",ExpressionUUID->"26eca71b-9fad-4fdf-8e61-bc50f7ef11f5",
 CellID->1529550952],
Cell[3312888, 95264, 26224, 744, 56, "Output",ExpressionUUID->"0c4257dd-290e-474a-a6e9-8a5ea86d1871",
 CellID->1783595674]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3339161, 96014, 158, 2, 40, "MathCaption",ExpressionUUID->"4f3729e4-cb08-44da-b848-38cc2cb5fb37"],
Cell[CellGroupData[{
Cell[3339344, 96020, 141, 3, 27, "Input",ExpressionUUID->"ffe634cc-4f71-4509-854c-c4a141c026f6"],
Cell[3339488, 96025, 23979, 678, 56, "Output",ExpressionUUID->"f71853eb-5230-4c36-84fa-1ef98e57b1f8",
 CellID->168264445]
}, Open  ]],
Cell[CellGroupData[{
Cell[3363504, 96708, 395, 11, 47, "Input",ExpressionUUID->"82793961-b9a1-4cbd-85e1-0fbf81ab7402"],
Cell[3363902, 96721, 8093, 221, 32, "Output",ExpressionUUID->"b9e2e9e6-4134-41fa-85eb-8d767f4c1613",
 CellID->1932047663]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3372044, 96948, 261, 6, 56, "MathCaption",ExpressionUUID->"16779222-2915-4df0-bade-dba12c398814"],
Cell[CellGroupData[{
Cell[3372330, 96958, 674, 22, 27, "Input",ExpressionUUID->"d4124387-34cd-409c-92d0-9ed686affe3d"],
Cell[3373007, 96982, 448, 9, 22, "Message",ExpressionUUID->"b1d399bd-9ffa-4d65-865d-ff5f068491ae",
 CellID->1972329401],
Cell[3373458, 96993, 445, 9, 22, "Message",ExpressionUUID->"32bcbaf2-7703-46a9-8891-06a1791498c7",
 CellID->382667515],
Cell[3373906, 97004, 9118, 248, 32, "Output",ExpressionUUID->"61349af1-6fbf-4bdd-9295-1e17eb5f6197",
 CellID->1926885487]
}, Open  ]]
}, Open  ]],
Cell[3383051, 97256, 1418, 22, 287, "Text",ExpressionUUID->"92fe3149-4a3c-46f6-8139-ff7d651d9f12"],
Cell[3384472, 97280, 274, 4, 40, "Text",ExpressionUUID->"8d80d6d6-f980-4c4b-a874-2c8d98634b69"],
Cell[CellGroupData[{
Cell[3384771, 97288, 161, 2, 40, "MathCaption",ExpressionUUID->"36dbe083-df68-4b13-bd56-c753de526981"],
Cell[CellGroupData[{
Cell[3384957, 97294, 511, 17, 27, "Input",ExpressionUUID->"24773395-7626-4ae0-ab8c-3d6b315f4220"],
Cell[3385471, 97313, 4197, 122, 32, "Output",ExpressionUUID->"ff4b644b-9d83-4527-8f06-c4d261928fe8",
 CellID->1110983249]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3389717, 97441, 98, 0, 40, "MathCaption",ExpressionUUID->"eb250d2f-42f7-4e3e-baf9-569d2374378e"],
Cell[CellGroupData[{
Cell[3389840, 97445, 147, 3, 27, "Input",ExpressionUUID->"f67f61ae-5735-4bf4-bae4-3c656c2a727c"],
Cell[3389990, 97450, 170, 3, 26, "Output",ExpressionUUID->"197e5707-72dc-4b7c-930e-c825f3af7602",
 CellID->2072638925]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3390209, 97459, 186, 3, 40, "MathCaption",ExpressionUUID->"ae80212d-ad2c-4bb1-9e55-312eedf7ecbe"],
Cell[CellGroupData[{
Cell[3390420, 97466, 511, 17, 27, "Input",ExpressionUUID->"a39167ba-3a9e-4c9c-a542-13fe9785ac60"],
Cell[3390934, 97485, 3235, 94, 32, "Output",ExpressionUUID->"ce70711b-104a-4fdd-9243-a398f18564d6",
 CellID->510174488]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3394218, 97585, 175, 3, 40, "MathCaption",ExpressionUUID->"57e04643-39b2-4e4f-ae99-191a8ca2cbb4"],
Cell[CellGroupData[{
Cell[3394418, 97592, 263, 6, 27, "Input",ExpressionUUID->"040ef5b7-bb33-4a2f-9102-6e73af827ff5"],
Cell[3394684, 97600, 448, 9, 22, "Message",ExpressionUUID->"95300994-a28f-43a8-8d05-867220ea277b",
 CellID->1747900992],
Cell[3395135, 97611, 448, 9, 22, "Message",ExpressionUUID->"44eb245e-968d-4955-b477-6377fa921450",
 CellID->1543165231],
Cell[3395586, 97622, 172, 3, 26, "Output",ExpressionUUID->"af72e56b-03d5-43e1-bb52-0f3ebe3e6326",
 CellID->818032598]
}, Open  ]],
Cell[CellGroupData[{
Cell[3395795, 97630, 249, 7, 27, "Input",ExpressionUUID->"99e63dde-c725-4c5a-b9c7-a850219af135"],
Cell[3396047, 97639, 172, 3, 26, "Output",ExpressionUUID->"1b67ca0d-984b-4515-bfde-63da7e25c168",
 CellID->483566492]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3396268, 97648, 94, 0, 40, "MathCaption",ExpressionUUID->"d909b7df-3a2c-4e5d-baa8-fb49c9029f28"],
Cell[CellGroupData[{
Cell[3396387, 97652, 511, 17, 27, "Input",ExpressionUUID->"f780ec05-6d0e-4dd8-80ce-2ebaafe4a433"],
Cell[3396901, 97671, 4195, 122, 32, "Output",ExpressionUUID->"36252d6c-cb6a-4e3b-8cd7-8401d19732d7",
 CellID->1223910796]
}, Open  ]],
Cell[CellGroupData[{
Cell[3401133, 97798, 147, 3, 27, "Input",ExpressionUUID->"a29959b0-08ab-4cae-ad00-2a088b5c5a60"],
Cell[3401283, 97803, 403, 8, 22, "Message",ExpressionUUID->"ec8cc230-bb66-4c02-afe0-990bbd8f3a62",
 CellID->990163873],
Cell[3401689, 97813, 403, 8, 22, "Message",ExpressionUUID->"0208eb44-7e25-43c3-aa46-c9cead9abf45",
 CellID->558299888],
Cell[3402095, 97823, 404, 8, 22, "Message",ExpressionUUID->"179a05d0-1f13-4b96-a8c6-33567d57ea87",
 CellID->1276487174],
Cell[3402502, 97833, 433, 8, 22, "Message",ExpressionUUID->"97234cd6-d27c-4001-b841-ce391ae67e7f",
 CellID->2102327895],
Cell[3402938, 97843, 14586, 415, 32, "Output",ExpressionUUID->"516d1b8d-916c-4d3f-9e4a-466ddf54baef",
 CellID->1431695669]
}, Open  ]],
Cell[CellGroupData[{
Cell[3417561, 98263, 220, 5, 27, "Input",ExpressionUUID->"58f7a425-4258-49b9-beca-18b21bd039d3"],
Cell[3417784, 98270, 404, 8, 22, "Message",ExpressionUUID->"2ece1536-384e-4e6c-bcce-8dc3bc23c46b",
 CellID->1398306970],
Cell[3418191, 98280, 404, 8, 22, "Message",ExpressionUUID->"044c4d2b-9ff7-41fd-8e00-514a4010afdf",
 CellID->1311674775],
Cell[3418598, 98290, 404, 8, 22, "Message",ExpressionUUID->"08144853-a199-418c-90b1-e71de5c1e420",
 CellID->1097777336],
Cell[3419005, 98300, 404, 8, 22, "Message",ExpressionUUID->"107551a2-7dd5-487b-9ae0-f4a2d08f93a1",
 CellID->1910785282],
Cell[3419412, 98310, 433, 8, 22, "Message",ExpressionUUID->"db78caba-9c06-492d-a581-51f619098d0f",
 CellID->1684291827],
Cell[3419848, 98320, 403, 8, 22, "Message",ExpressionUUID->"1b7c0750-1132-4750-98a8-234b59a6d495",
 CellID->120718085],
Cell[3420254, 98330, 403, 8, 22, "Message",ExpressionUUID->"7760eaaa-f487-407b-9410-f660f19cf72d",
 CellID->681804811],
Cell[3420660, 98340, 432, 8, 22, "Message",ExpressionUUID->"5519e53e-81cc-49ae-948c-96b90d7afc56",
 CellID->274369560],
Cell[3421095, 98350, 14585, 415, 32, "Output",ExpressionUUID->"6139e7e2-ce80-4dcf-aa77-9adddc0f5462",
 CellID->674914202]
}, Open  ]],
Cell[CellGroupData[{
Cell[3435717, 98770, 208, 5, 27, "Input",ExpressionUUID->"4c85cac3-8262-44eb-9bfb-50073742e631"],
Cell[3435928, 98777, 403, 8, 22, "Message",ExpressionUUID->"65806eca-5490-45b4-8826-39a7fd6c9c24",
 CellID->458362513],
Cell[3436334, 98787, 9424, 269, 32, "Output",ExpressionUUID->"3cdb740e-ad61-4d2f-81a4-4fbadd8d6e43",
 CellID->1934640820]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3445819, 99063, 138, 1, 39, "Subsection",ExpressionUUID->"b15d6210-1bbd-4601-ab7e-a617bbcf0637"],
Cell[3445960, 99066, 270, 8, 23, "Text",ExpressionUUID->"8f6fdb9b-11a9-42fa-98a2-32bb103d2a97"],
Cell[CellGroupData[{
Cell[3446255, 99078, 283, 6, 40, "MathCaption",ExpressionUUID->"f0bd2eb9-8903-44ec-af44-c5dae6f84d36"],
Cell[CellGroupData[{
Cell[3446563, 99088, 640, 21, 27, "Input",ExpressionUUID->"b9d9a77d-d317-4ea3-87b4-8598690cb347"],
Cell[3447206, 99111, 11514, 316, 32, "Output",ExpressionUUID->"c5ff7d7a-2be9-45a9-9741-91ab95708c7b",
 CellID->609986205]
}, Open  ]],
Cell[CellGroupData[{
Cell[3458757, 99432, 210, 4, 27, "Input",ExpressionUUID->"72cae8d6-8a1f-4412-b642-5162b62f5109"],
Cell[3458970, 99438, 5575, 155, 32, "Output",ExpressionUUID->"d3a40bc5-c0e0-43e0-8bee-9bb7cb129b34",
 CellID->2098700304]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3464594, 99599, 177, 3, 40, "MathCaption",ExpressionUUID->"3d6ecafc-2760-4ad2-b8e1-4f0754831b6d"],
Cell[CellGroupData[{
Cell[3464796, 99606, 511, 17, 27, "Input",ExpressionUUID->"1661c5d8-b395-4e34-bf03-03a8d3f1d97f"],
Cell[3465310, 99625, 4196, 122, 32, "Output",ExpressionUUID->"c1716e50-eec1-4ad4-bd40-7253d10d76cd",
 CellID->437107353]
}, Open  ]],
Cell[CellGroupData[{
Cell[3469543, 99752, 216, 4, 27, "Input",ExpressionUUID->"d6f66c33-0bc2-42d9-bc10-fd05db66e3de"],
Cell[3469762, 99758, 4188, 120, 32, "Output",ExpressionUUID->"a2afdda6-7f4c-4443-bd7c-9638d24fb348",
 CellID->2069209131]
}, Open  ]],
Cell[CellGroupData[{
Cell[3473987, 99883, 147, 3, 27, "Input",ExpressionUUID->"470fef24-3b0c-459e-b071-2c9dcc9db444"],
Cell[3474137, 99888, 169, 3, 26, "Output",ExpressionUUID->"87ce841a-7743-4416-942c-f2213ac777d3",
 CellID->847685199]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3474355, 99897, 94, 0, 40, "MathCaption",ExpressionUUID->"938dbf0c-3d3b-481c-b990-511bd3b9556a"],
Cell[CellGroupData[{
Cell[3474474, 99901, 511, 17, 27, "Input",ExpressionUUID->"a14e6196-6f72-41d7-b101-7022754ae20c"],
Cell[3474988, 99920, 3235, 94, 32, "Output",ExpressionUUID->"d01bec0e-9f0e-4681-8e56-62d3eb2e0b1e",
 CellID->638625395]
}, Open  ]],
Cell[CellGroupData[{
Cell[3478260, 100019, 216, 4, 27, "Input",ExpressionUUID->"4bb4a716-efba-4b1c-8cab-16e33351c4cd"],
Cell[3478479, 100025, 5536, 155, 32, "Output",ExpressionUUID->"1e2611fe-93c8-431c-a527-c02aead2ca6a",
 CellID->253746059]
}, Open  ]],
Cell[CellGroupData[{
Cell[3484052, 100185, 147, 3, 27, "Input",ExpressionUUID->"9851ba43-f488-46bd-9e9a-a181e8ad08d6"],
Cell[3484202, 100190, 4235, 121, 32, "Output",ExpressionUUID->"685a13fc-f937-4b58-9f0b-063dc59f496e",
 CellID->1954974318]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3488486, 100317, 94, 0, 40, "MathCaption",ExpressionUUID->"fb479851-7a59-4d36-96b1-bbb96866d2e6"],
Cell[CellGroupData[{
Cell[3488605, 100321, 511, 17, 27, "Input",ExpressionUUID->"34fb6a35-0a43-4537-97f8-28a2e6531b59"],
Cell[3489119, 100340, 4194, 122, 32, "Output",ExpressionUUID->"01be0494-3630-49d8-8208-a21709933345",
 CellID->753563526]
}, Open  ]],
Cell[CellGroupData[{
Cell[3493350, 100467, 218, 4, 27, "Input",ExpressionUUID->"95b8a649-7d95-4e74-952d-b73cab63f546"],
Cell[3493571, 100473, 6498, 183, 32, "Output",ExpressionUUID->"6527c72b-5ed5-48e9-a46a-422af268654b",
 CellID->1431115123]
}, Open  ]],
Cell[CellGroupData[{
Cell[3500106, 100661, 215, 4, 27, "Input",ExpressionUUID->"1e3f16ea-29dc-4527-bf9f-307a11c44b93"],
Cell[3500324, 100667, 8462, 238, 32, "Output",ExpressionUUID->"0dc7e481-4627-4fb9-9842-b79b489baf9c",
 CellID->935025861]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3508847, 100912, 97, 0, 39, "Subsection",ExpressionUUID->"19a788d8-1dac-4761-82cc-fdfd4a205663"],
Cell[3508947, 100914, 1212, 20, 183, "Text",ExpressionUUID->"d83de66a-b28b-4fa9-8ddf-5ef0c2ffa7de"],
Cell[CellGroupData[{
Cell[3510184, 100938, 186, 5, 40, "MathCaption",ExpressionUUID->"565f8f3a-4967-4358-b087-ccd69e823395"],
Cell[CellGroupData[{
Cell[3510395, 100947, 160, 3, 27, "Input",ExpressionUUID->"f60be583-9499-4ebb-beca-00de190cb958"],
Cell[3510558, 100952, 199, 4, 26, "Output",ExpressionUUID->"91589141-ae61-4a31-b8bf-d59a0deb040a",
 CellID->2065006001]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3510806, 100962, 113, 0, 40, "MathCaption",ExpressionUUID->"af6edc8d-9fd0-4fbc-bcb7-817a4dfec718"],
Cell[CellGroupData[{
Cell[3510944, 100966, 448, 12, 27, "Input",ExpressionUUID->"bdeba044-43c9-4700-b525-a919e3f0a480"],
Cell[3511395, 100980, 609, 11, 22, "Message",ExpressionUUID->"6dc93ce5-e681-4645-95e3-3ce506dfb353",
 CellID->1072639640],
Cell[CellGroupData[{
Cell[3512029, 100995, 764, 17, 22, "Print",ExpressionUUID->"3e969c8b-1e60-4b07-a8bd-2c47fe20e29f",
 CellID->1876696562],
Cell[3512796, 101014, 754, 15, 22, "Print",ExpressionUUID->"54f8cebf-d8d7-411d-ae64-dd8fbeda0933",
 CellID->1546794045],
Cell[3513553, 101031, 820, 19, 22, "Print",ExpressionUUID->"54fda6ee-6a1a-4452-bc93-4eb226366100",
 CellID->1677967380],
Cell[3514376, 101052, 681, 15, 22, "Print",ExpressionUUID->"04cca175-82a5-4d28-bff4-b52f4de6086c",
 CellID->842503256],
Cell[3515060, 101069, 804, 18, 22, "Print",ExpressionUUID->"3be5411f-f830-4bec-b366-406e5244b634",
 CellID->1592538921],
Cell[3515867, 101089, 819, 19, 22, "Print",ExpressionUUID->"e31c8cab-bc57-44d2-aee9-bb560fe9a305",
 CellID->1712952686],
Cell[3516689, 101110, 855, 19, 22, "Print",ExpressionUUID->"1617cb6a-3edc-4400-a1e1-1c46ac9feac7",
 CellID->430021920],
Cell[3517547, 101131, 898, 18, 42, "Print",ExpressionUUID->"f73f28a8-e2aa-48d5-9f66-349c81af1aff",
 CellID->1095937285],
Cell[3518448, 101151, 763, 17, 22, "Print",ExpressionUUID->"4222a580-dccf-49c4-b53c-fac2c9e16154",
 CellID->450708676],
Cell[3519214, 101170, 268, 4, 22, "Print",ExpressionUUID->"f6322896-412d-48e0-86d9-7d962dfd0cc5",
 CellID->885894563],
Cell[3519485, 101176, 647, 13, 22, "Print",ExpressionUUID->"cbfd9153-feb0-4aeb-b7ac-d614e007967c",
 CellID->725795690],
Cell[3520135, 101191, 272, 4, 22, "Print",ExpressionUUID->"8f26d478-b777-487c-9942-6e89b3f8bf74",
 CellID->741046870],
Cell[3520410, 101197, 780, 18, 22, "Print",ExpressionUUID->"c0334b33-6e92-40e3-b2af-6d11ea388a47",
 CellID->1517075137],
Cell[3521193, 101217, 853, 19, 22, "Print",ExpressionUUID->"0bf8e675-7a1a-461c-b693-b4d2070c8ba0",
 CellID->408486728],
Cell[3522049, 101238, 772, 17, 22, "Print",ExpressionUUID->"89f20f82-d433-4e34-a961-f5993a9c9472",
 CellID->2078701684],
Cell[3522824, 101257, 660, 13, 22, "Print",ExpressionUUID->"75223229-1581-45f4-baae-1888dd067d88",
 CellID->1475233592],
Cell[3523487, 101272, 404, 8, 22, "Print",ExpressionUUID->"56057232-3575-48cb-ad01-b202885b8112",
 CellID->1552276129],
Cell[3523894, 101282, 408, 8, 22, "Print",ExpressionUUID->"8a9a3b74-f180-405a-9a47-3cf2c300ce9e",
 CellID->1473556347],
Cell[3524305, 101292, 672, 13, 22, "Print",ExpressionUUID->"565e7c15-2f8b-40be-a00d-d21216df5ee2",
 CellID->1280106825]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3525026, 101311, 160, 3, 27, "Input",ExpressionUUID->"90c7de95-9167-49db-939e-ff03faeb5356"],
Cell[3525189, 101316, 229, 5, 26, "Output",ExpressionUUID->"0de02cbb-dca1-4e6e-80e9-5137da472a86",
 CellID->1264232129]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3525467, 101327, 104, 0, 40, "MathCaption",ExpressionUUID->"e994c9fc-b228-491e-ab58-58a54b886ece"],
Cell[CellGroupData[{
Cell[3525596, 101331, 316, 10, 27, "Input",ExpressionUUID->"65f50658-960b-422e-84d6-c4b79326475d"],
Cell[3525915, 101343, 2175, 61, 32, "Output",ExpressionUUID->"18c16b5f-4278-4ea2-b50f-286859bafbbd",
 CellID->746188685]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3528139, 101410, 150, 2, 40, "MathCaption",ExpressionUUID->"eaa046e8-c311-411d-b9b4-d1fd191d43d3"],
Cell[CellGroupData[{
Cell[3528314, 101416, 723, 25, 27, "Input",ExpressionUUID->"3ab0ac4d-2e04-4bf9-b953-354986eb5bc7"],
Cell[3529040, 101443, 5273, 147, 32, "Output",ExpressionUUID->"ddf6f53c-a46d-46bd-95c1-7dc0daa5c9dd",
 CellID->2036213962]
}, Open  ]],
Cell[CellGroupData[{
Cell[3534350, 101595, 150, 3, 27, "Input",ExpressionUUID->"f8c299b9-9b74-4655-b0c5-6abac751f942"],
Cell[3534503, 101600, 4215, 118, 32, "Output",ExpressionUUID->"409f52a9-13c6-40f2-a7c6-d17f1e8d6d38",
 CellID->348660742]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3538767, 101724, 310, 11, 40, "MathCaption",ExpressionUUID->"22abeb6d-4427-43ca-b0af-747627b252a9"],
Cell[CellGroupData[{
Cell[3539102, 101739, 284, 9, 27, "Input",ExpressionUUID->"6ea67654-6fd0-44b4-aa4b-f6ce902fe8ba"],
Cell[3539389, 101750, 1097, 31, 32, "Output",ExpressionUUID->"9ef7e4f1-a488-4d3f-b61f-af8654082b81",
 CellID->1605701778]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3540535, 101787, 174, 3, 40, "MathCaption",ExpressionUUID->"838a3c28-6591-4068-b72b-93a193e07b20"],
Cell[CellGroupData[{
Cell[3540734, 101794, 248, 7, 27, "Input",ExpressionUUID->"1618bef4-5bda-44fd-95d1-d15edad861cc"],
Cell[3540985, 101803, 1234, 35, 32, "Output",ExpressionUUID->"65896c37-674e-4f0f-a66d-5231b87f74e7",
 CellID->1750274644]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3542268, 101844, 139, 2, 40, "MathCaption",ExpressionUUID->"1c84daf4-0921-41cd-84fb-8efd89cb9e79"],
Cell[CellGroupData[{
Cell[3542432, 101850, 372, 12, 27, "Input",ExpressionUUID->"7f08fb4c-3a52-4816-be6b-3345d349a421"],
Cell[3542807, 101864, 170, 3, 26, "Output",ExpressionUUID->"b584e013-1ff2-4078-9374-1bb345513b7f",
 CellID->634945102]
}, Open  ]]
}, Open  ]],
Cell[3543004, 101871, 648, 14, 57, "Text",ExpressionUUID->"510f21ea-fe11-4b0d-95bb-64b3e8d06f77"],
Cell[CellGroupData[{
Cell[3543677, 101889, 225, 4, 40, "MathCaption",ExpressionUUID->"4b45bc2b-27f2-406a-8985-7aedb9056d62"],
Cell[CellGroupData[{
Cell[3543927, 101897, 366, 9, 27, "Input",ExpressionUUID->"2dc77a75-1962-4edb-833e-c04d53ed4f28"],
Cell[3544296, 101908, 526, 16, 26, "Output",ExpressionUUID->"7ef8034a-2069-4b3c-bcce-47fa9d35026d",
 CellID->676087371]
}, Open  ]],
Cell[CellGroupData[{
Cell[3544859, 101929, 388, 10, 27, "Input",ExpressionUUID->"d8a829e9-565f-4c96-91e1-5175ea6a06bf"],
Cell[3545250, 101941, 1069, 34, 26, "Output",ExpressionUUID->"a2839592-2b57-430d-a70c-4ce4e1f88211",
 CellID->664468120]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3546368, 101981, 282, 7, 40, "MathCaption",ExpressionUUID->"1b09f8de-9e60-4637-ba44-9480f797f34f"],
Cell[CellGroupData[{
Cell[3546675, 101992, 312, 8, 27, "Input",ExpressionUUID->"0df1c0e4-1c8f-4fdd-8570-edb83d751a68"],
Cell[3546990, 102002, 1182, 33, 32, "Output",ExpressionUUID->"aaba159a-ed2a-42a6-bb39-91b4034b2448",
 CellID->800728440]
}, Open  ]],
Cell[CellGroupData[{
Cell[3548209, 102040, 220, 4, 27, "Input",ExpressionUUID->"79e37963-494e-4c29-bae9-72b912162898"],
Cell[3548432, 102046, 2208, 62, 32, "Output",ExpressionUUID->"5c7e77a8-fc9f-45fe-86e9-5cd50385ef37",
 CellID->1710358978]
}, Open  ]],
Cell[CellGroupData[{
Cell[3550677, 102113, 220, 4, 27, "Input",ExpressionUUID->"34b6f441-73da-40aa-9496-189da2075fe2"],
Cell[3550900, 102119, 1183, 33, 32, "Output",ExpressionUUID->"118a4dcd-f568-489c-82fa-8f8f0ab40c75",
 CellID->1934728151]
}, Open  ]],
Cell[CellGroupData[{
Cell[3552120, 102157, 152, 3, 27, "Input",ExpressionUUID->"f3006008-56de-4b39-af62-61795b12add5"],
Cell[CellGroupData[{
Cell[3552297, 102164, 587, 12, 22, "Print",ExpressionUUID->"a3753980-a0cb-4465-a2e7-2c5733632507",
 CellID->1926134630],
Cell[3552887, 102178, 619, 12, 22, "Print",ExpressionUUID->"f7caa540-1a07-4f24-a23f-209be01f3c08",
 CellID->1376231532],
Cell[3553509, 102192, 609, 12, 22, "Print",ExpressionUUID->"72727013-c27c-4115-9a8a-9a0ff3a3f4a4",
 CellID->1924480134],
Cell[3554121, 102206, 598, 12, 22, "Print",ExpressionUUID->"aa8d238e-3d75-4625-9e7a-57d0ba06b7e7",
 CellID->1486404671],
Cell[3554722, 102220, 596, 12, 22, "Print",ExpressionUUID->"2a1f1dff-1323-4341-a2fd-ab967bfef03a",
 CellID->326039946],
Cell[3555321, 102234, 587, 12, 22, "Print",ExpressionUUID->"3fadeb9d-b807-47d6-a179-d27b606e048e",
 CellID->1522588925],
Cell[3555911, 102248, 584, 12, 22, "Print",ExpressionUUID->"4518d517-09b4-4bb1-a2c6-c6c915a1221a",
 CellID->736924823],
Cell[3556498, 102262, 590, 12, 22, "Print",ExpressionUUID->"efdee0ad-68a2-473f-b1e8-e4618ea99b09",
 CellID->957307912],
Cell[3557091, 102276, 602, 12, 22, "Print",ExpressionUUID->"f19f8f15-ccbe-4228-a614-2041ea397220",
 CellID->526593660],
Cell[3557696, 102290, 582, 12, 22, "Print",ExpressionUUID->"fb40b726-2c52-448b-829b-47234d4cc7d7",
 CellID->679534333],
Cell[3558281, 102304, 570, 12, 22, "Print",ExpressionUUID->"294f7e2b-7d49-4a3b-a6b1-8806a83c199b",
 CellID->615499985],
Cell[3558854, 102318, 576, 12, 22, "Print",ExpressionUUID->"152058f7-9ff5-4cf9-bba2-dc368ea1ee45",
 CellID->279520296],
Cell[3559433, 102332, 603, 12, 22, "Print",ExpressionUUID->"95199d39-320b-465c-b8c1-49a05b757d5a",
 CellID->1350878883],
Cell[3560039, 102346, 595, 12, 22, "Print",ExpressionUUID->"b44e9bc3-6913-4c3c-a2a5-49ac0b45a7fb",
 CellID->1024270497],
Cell[3560637, 102360, 592, 12, 22, "Print",ExpressionUUID->"a360f435-5ff2-4efa-9cc7-9602a07f1755",
 CellID->685328727]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[3561268, 102377, 246, 6, 23, "Text",ExpressionUUID->"c1da5237-5097-415f-9826-b2d5585aaf5c"]
}, Closed]],
Cell[CellGroupData[{
Cell[3561551, 102388, 113, 0, 39, "Subsection",ExpressionUUID->"c9e7f10e-d301-41ba-95e3-c8658b4c8e4c"],
Cell[3561667, 102390, 667, 23, 57, "Text",ExpressionUUID->"2819f13a-e559-4131-aa25-7d0abe82c8bb"],
Cell[CellGroupData[{
Cell[3562359, 102417, 394, 10, 56, "MathCaption",ExpressionUUID->"eeee7a0a-d524-4c12-b94a-2393fa69c85a"],
Cell[CellGroupData[{
Cell[3562778, 102431, 245, 7, 27, "Input",ExpressionUUID->"b4fa354f-1932-46f4-a35b-bcf27753d8e5"],
Cell[3563026, 102440, 1190, 33, 32, "Output",ExpressionUUID->"46b31d8d-6016-486c-9188-1c6a97987660",
 CellID->1233400720]
}, Open  ]],
Cell[CellGroupData[{
Cell[3564253, 102478, 141, 3, 27, "Input",ExpressionUUID->"8d0f1729-1bbe-4d8c-af1b-3d980096eedd"],
Cell[3564397, 102483, 189, 3, 26, "Output",ExpressionUUID->"00f97cea-96aa-4673-a2ac-857b51fba085",
 CellID->106256549]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3564635, 102492, 156, 2, 40, "MathCaption",ExpressionUUID->"8cde665a-eee1-4a0f-a7e1-4d9daed2ac8e"],
Cell[CellGroupData[{
Cell[3564816, 102498, 283, 9, 27, "Input",ExpressionUUID->"62faeb0d-29e3-4951-8dfa-130e3a8eef2e"],
Cell[3565102, 102509, 353, 10, 26, "Output",ExpressionUUID->"5bc6f494-bb42-46a4-b446-894ace6bff1b",
 CellID->1294908386]
}, Open  ]],
Cell[CellGroupData[{
Cell[3565492, 102524, 141, 3, 27, "Input",ExpressionUUID->"8be25018-45df-4868-a79a-83257f314630"],
Cell[3565636, 102529, 186, 3, 26, "Output",ExpressionUUID->"b17a6d3c-f374-4f12-9114-49da68f06d37",
 CellID->925964804]
}, Open  ]],
Cell[CellGroupData[{
Cell[3565859, 102537, 185, 5, 27, "Input",ExpressionUUID->"8dd2301b-17ea-40f8-8c76-39ae7bdf3f06"],
Cell[3566047, 102544, 352, 10, 26, "Output",ExpressionUUID->"ea32b9e1-ee64-4645-b8af-66db1c95381f",
 CellID->2023954666]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3566448, 102560, 115, 0, 40, "MathCaption",ExpressionUUID->"430ee0e2-99c7-43b4-8316-4f729d80503b"],
Cell[CellGroupData[{
Cell[3566588, 102564, 248, 7, 27, "Input",ExpressionUUID->"58c5d7c8-83d2-4c35-ad3f-1d2235f35ce6"],
Cell[3566839, 102573, 2423, 70, 40, "Output",ExpressionUUID->"148e0ccc-7fb7-40f4-9f18-816f5722f937",
 CellID->465102982]
}, Open  ]],
Cell[CellGroupData[{
Cell[3569299, 102648, 147, 3, 27, "Input",ExpressionUUID->"e1580a7b-1ab8-435a-912a-be3e24e8c59b"],
Cell[3569449, 102653, 1146, 32, 32, "Output",ExpressionUUID->"b46c4d42-7bba-4a6f-b4c7-6b320ee5b25e",
 CellID->145343394]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3570644, 102691, 119, 0, 40, "MathCaption",ExpressionUUID->"f1c1db60-4479-4a50-aaa6-23c870613374"],
Cell[CellGroupData[{
Cell[3570788, 102695, 261, 8, 27, "Input",ExpressionUUID->"6fcf7d89-ecfe-4dd0-860a-0265eb71b876"],
Cell[3571052, 102705, 171, 3, 26, "Output",ExpressionUUID->"131d0359-8d44-4efe-971f-54ce3c866d53",
 CellID->1097516319]
}, Open  ]],
Cell[CellGroupData[{
Cell[3571260, 102713, 305, 9, 27, "Input",ExpressionUUID->"e80eadee-e367-45dc-aea3-02cd4602bfb5"],
Cell[3571568, 102724, 12795, 366, 40, "Output",ExpressionUUID->"3c36ebdc-1e5d-44df-8675-a374e7a30e3b",
 CellID->1219729350]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3584412, 103096, 123, 0, 40, "MathCaption",ExpressionUUID->"58f48673-1713-4801-ba9b-4dae90081aac"],
Cell[CellGroupData[{
Cell[3584560, 103100, 244, 7, 27, "Input",ExpressionUUID->"a66c7c3f-e420-4696-aa40-57e9b4e6a2ae"],
Cell[3584807, 103109, 2425, 70, 40, "Output",ExpressionUUID->"96ff3c6a-2f70-4442-980e-5832eac6216f",
 CellID->112291188]
}, Open  ]],
Cell[CellGroupData[{
Cell[3587269, 103184, 180, 4, 27, "Input",ExpressionUUID->"0285038f-7773-4063-873c-644297289d55"],
Cell[3587452, 103190, 1144, 32, 32, "Output",ExpressionUUID->"1517d70e-c799-420d-be1e-a9c15f1717ec",
 CellID->1787354650]
}, Open  ]]
}, Open  ]],
Cell[3588623, 103226, 2211, 58, 126, "DefinitionBox",ExpressionUUID->"e90876ee-f938-4ba5-9b09-7ae1a3d38dd4"],
Cell[3590837, 103286, 105, 0, 23, "Text",ExpressionUUID->"8b50e31a-1981-40e6-8c72-ca668477739c"],
Cell[3590945, 103288, 190, 3, 23, "Text",ExpressionUUID->"7aa3cdb2-26b3-4e8f-9be8-e157f8ea7f65"],
Cell[CellGroupData[{
Cell[3591160, 103295, 347, 5, 72, "MathCaption",ExpressionUUID->"b03ccc55-d410-4992-bcf4-afc1922e11b8"],
Cell[CellGroupData[{
Cell[3591532, 103304, 424, 11, 27, "Input",ExpressionUUID->"ffc21272-330c-4232-9ae3-0d0f84cf54cc"],
Cell[CellGroupData[{
Cell[3591981, 103319, 682, 15, 22, "Print",ExpressionUUID->"f2167a3a-23f8-421b-8a8a-4b13d4bd5d2a",
 CellID->1068348314],
Cell[3592666, 103336, 777, 17, 22, "Print",ExpressionUUID->"b120262a-c2d1-45ca-ae02-a33764f5eb3c",
 CellID->619340831],
Cell[3593446, 103355, 827, 19, 22, "Print",ExpressionUUID->"a43ecaf6-219b-4e99-8889-a90d27beeda0",
 CellID->408109608],
Cell[3594276, 103376, 926, 19, 42, "Print",ExpressionUUID->"27ba4c4c-c7e9-4d29-aca2-653d5f6a489d",
 CellID->35413138],
Cell[3595205, 103397, 770, 17, 22, "Print",ExpressionUUID->"61db10a6-9903-4c99-a45c-3eb38b0ee7f9",
 CellID->335640147],
Cell[3595978, 103416, 269, 4, 22, "Print",ExpressionUUID->"59c95aae-605d-4ec1-86c2-a67db6f25f16",
 CellID->1811893607],
Cell[3596250, 103422, 648, 13, 22, "Print",ExpressionUUID->"0e625634-2a82-4576-b5c3-a797f8525a06",
 CellID->1177738389],
Cell[3596901, 103437, 274, 4, 22, "Print",ExpressionUUID->"ce8a4db4-23a8-4e20-929a-fc2efbd8a04b",
 CellID->892329524],
Cell[3597178, 103443, 788, 18, 22, "Print",ExpressionUUID->"c8121aab-17ed-4441-9340-bc1c2817c31c",
 CellID->1504002867],
Cell[3597969, 103463, 934, 19, 42, "Print",ExpressionUUID->"a423dacd-9679-4d39-8e2b-dd6567e66d24",
 CellID->1893660479],
Cell[3598906, 103484, 784, 18, 22, "Print",ExpressionUUID->"e9f2c5c9-72e1-40a1-8f52-722b57dfa5c8",
 CellID->1044022013],
Cell[3599693, 103504, 659, 13, 22, "Print",ExpressionUUID->"62cc6729-5e8b-4f1c-ae3d-dd912ffcb6f4",
 CellID->1936576501],
Cell[3600355, 103519, 403, 8, 22, "Print",ExpressionUUID->"579170c6-6ae0-4ed3-affe-cd06842ac5d1",
 CellID->779098375],
Cell[3600761, 103529, 398, 8, 22, "Print",ExpressionUUID->"3b0cfec6-6956-4491-985e-3e6c22b9045e",
 CellID->1098592866],
Cell[3601162, 103539, 410, 8, 22, "Print",ExpressionUUID->"e456c2f2-80ac-4d1d-8b20-3ca415a1edb7",
 CellID->2133385603]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3601633, 103554, 117, 0, 40, "MathCaption",ExpressionUUID->"417fd3fe-942e-46f5-a7c8-447592556373"],
Cell[CellGroupData[{
Cell[3601775, 103558, 347, 11, 27, "Input",ExpressionUUID->"fe17a7b2-8534-4388-9f4a-4458200cc735"],
Cell[3602125, 103571, 2371, 69, 32, "Output",ExpressionUUID->"d1e78e72-5a2e-46e6-9394-a333fbc3445f",
 CellID->1424859309]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3604545, 103646, 110, 0, 40, "MathCaption",ExpressionUUID->"16524a6d-a0d6-40b0-b006-52bedf3e5ab7"],
Cell[CellGroupData[{
Cell[3604680, 103650, 182, 4, 27, "Input",ExpressionUUID->"e3792122-8606-47ae-936f-0b62fcc472c6"],
Cell[3604865, 103656, 769, 16, 22, "Print",ExpressionUUID->"d4197de6-5044-46f0-98b7-1173cbf96796",
 CellID->1020258666],
Cell[3605637, 103674, 2239, 64, 32, "Output",ExpressionUUID->"ae21565f-47e6-4b1a-a0fc-c32ec2ca655f",
 CellID->1977961814]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3607925, 103744, 156, 2, 40, "MathCaption",ExpressionUUID->"8d4b16d9-9d1f-4264-bf1d-dacb82c3c478"],
Cell[CellGroupData[{
Cell[3608106, 103750, 155, 3, 27, "Input",ExpressionUUID->"4208acf7-41f2-4e07-aebe-6e3c48b84b19"],
Cell[3608264, 103755, 5392, 154, 40, "Output",ExpressionUUID->"20e68122-9fb9-4890-918e-04deb4d93bd7",
 CellID->557352338]
}, Open  ]],
Cell[CellGroupData[{
Cell[3613693, 103914, 147, 3, 27, "Input",ExpressionUUID->"de14bbed-f695-447e-b2b3-c2a3924fa5a5"],
Cell[CellGroupData[{
Cell[3613865, 103921, 577, 12, 22, "Print",ExpressionUUID->"92ec3068-c305-499d-a799-ffc0b8a09ca5",
 CellID->97076552],
Cell[3614445, 103935, 627, 12, 22, "Print",ExpressionUUID->"9b477e20-3352-49dc-8869-4d7909ce6ae7",
 CellID->1625521391],
Cell[3615075, 103949, 615, 12, 22, "Print",ExpressionUUID->"5f345735-401c-4c17-91ff-4b8fcc6b3c05",
 CellID->1998545050],
Cell[3615693, 103963, 598, 12, 22, "Print",ExpressionUUID->"0f9049f3-3ac7-4037-99c2-cf88923439e4",
 CellID->754932623],
Cell[3616294, 103977, 601, 12, 22, "Print",ExpressionUUID->"e2087b04-c60c-42fa-883a-06dc04c88f72",
 CellID->819208285],
Cell[3616898, 103991, 588, 12, 22, "Print",ExpressionUUID->"a036a213-8d3d-49d9-9e21-973fea995efa",
 CellID->915985078],
Cell[3617489, 104005, 583, 12, 22, "Print",ExpressionUUID->"76c81ada-78cb-4f4c-b955-047db66128f0",
 CellID->1798424250],
Cell[3618075, 104019, 610, 12, 22, "Print",ExpressionUUID->"085ea881-ffeb-420f-9019-385ad3dc4929",
 CellID->284038225],
Cell[3618688, 104033, 585, 12, 22, "Print",ExpressionUUID->"49ab64af-aa6f-4939-b9f0-ea22c1ce7fac",
 CellID->2121783941],
Cell[3619276, 104047, 571, 12, 22, "Print",ExpressionUUID->"e764238d-8e94-489e-ba23-6bf37f7cb480",
 CellID->1856737952],
Cell[3619850, 104061, 579, 12, 22, "Print",ExpressionUUID->"378c41e5-f9ec-4001-b9ec-aabcf43506e8",
 CellID->1109884707]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3620502, 104081, 97, 0, 39, "Subsection",ExpressionUUID->"8feec4d9-2dd2-4696-85e4-204e145be129"],
Cell[3620602, 104083, 401, 6, 57, "Text",ExpressionUUID->"c3e7588b-fba1-4178-a94d-14000e1e9e16"],
Cell[CellGroupData[{
Cell[3621028, 104093, 104, 0, 40, "MathCaption",ExpressionUUID->"fb254167-c88e-4af7-b941-27e64f4dc52e"],
Cell[CellGroupData[{
Cell[3621157, 104097, 395, 10, 27, "Input",ExpressionUUID->"96a48b35-84c8-43eb-ad8b-53ea42abe70b"],
Cell[CellGroupData[{
Cell[3621577, 104111, 594, 12, 22, "Print",ExpressionUUID->"042496d1-c24f-4f5f-bd56-64c013f86b88",
 CellID->1357370834],
Cell[3622174, 104125, 605, 12, 22, "Print",ExpressionUUID->"6f398dac-3ce6-410e-be39-0c7854b21a13",
 CellID->612824957]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3622840, 104144, 373, 14, 40, "MathCaption",ExpressionUUID->"97a33204-3f58-4eaa-b471-485978fd7536"],
Cell[CellGroupData[{
Cell[3623238, 104162, 193, 5, 27, "Input",ExpressionUUID->"01f703f3-fc2a-402b-a278-20d21d80830b"],
Cell[3623434, 104169, 606, 12, 22, "Print",ExpressionUUID->"090e8cf6-7ccf-41ae-bbd5-8fecf861f731",
 CellID->417481287]
}, Open  ]],
Cell[CellGroupData[{
Cell[3624077, 104186, 651, 20, 27, "Input",ExpressionUUID->"6fb4c958-ffa0-45f7-b0c4-95bc1c2d5cc1"],
Cell[3624731, 104208, 561, 11, 22, "Message",ExpressionUUID->"4735604a-e080-480e-98ee-4054db6ee92b",
 CellID->778915860]
}, Open  ]],
Cell[CellGroupData[{
Cell[3625329, 104224, 454, 12, 27, "Input",ExpressionUUID->"e080aa83-81c9-4bb2-a392-f61627144701"],
Cell[CellGroupData[{
Cell[3625808, 104240, 761, 17, 22, "Print",ExpressionUUID->"d7e07fbe-09e6-47ab-959d-50bcd2b3d69b",
 CellID->178546908],
Cell[3626572, 104259, 769, 17, 22, "Print",ExpressionUUID->"461f9d05-6b70-45ba-85c7-f7995c2eb6d8",
 CellID->803227897],
Cell[3627344, 104278, 681, 15, 22, "Print",ExpressionUUID->"d94900f6-786b-48f2-b1ed-d403c1c028c5",
 CellID->769059837],
Cell[3628028, 104295, 785, 17, 22, "Print",ExpressionUUID->"fe1f4d19-a26f-40d7-8976-d3339492d2a5",
 CellID->2032092813],
Cell[3628816, 104314, 807, 18, 22, "Print",ExpressionUUID->"8e6ed137-bbcb-4616-89d9-0b8408c0f5a7",
 CellID->1951469850],
Cell[3629626, 104334, 824, 19, 22, "Print",ExpressionUUID->"22ee8ddc-a012-4b65-88eb-ab8a40b78c59",
 CellID->1983511966],
Cell[3630453, 104355, 764, 17, 22, "Print",ExpressionUUID->"2b928888-366b-4d12-b824-d262064e36f7",
 CellID->1496960181],
Cell[3631220, 104374, 269, 4, 22, "Print",ExpressionUUID->"97fc5034-c768-4bab-87a4-46081f023035",
 CellID->1638837462],
Cell[3631492, 104380, 649, 13, 22, "Print",ExpressionUUID->"1d51a4f5-eb3f-42d9-a296-c6c2741daf19",
 CellID->315967412],
Cell[3632144, 104395, 273, 4, 22, "Print",ExpressionUUID->"2e090813-7cfa-4f7c-8811-ba8ac3ec8fce",
 CellID->1884718100],
Cell[3632420, 104401, 780, 18, 22, "Print",ExpressionUUID->"0dc380bc-bbb9-425b-9ed4-d1608df96eea",
 CellID->1073313450],
Cell[3633203, 104421, 824, 18, 22, "Print",ExpressionUUID->"36d2d141-3902-4140-b129-ccecf47dd3e8",
 CellID->316519350],
Cell[3634030, 104441, 773, 17, 22, "Print",ExpressionUUID->"8df546ae-3a0d-49f1-9a9d-2c34cd61a587",
 CellID->423721831],
Cell[3634806, 104460, 659, 13, 22, "Print",ExpressionUUID->"eac3c76a-6c5b-47ff-9d7d-e3a9de7bd153",
 CellID->437040165],
Cell[3635468, 104475, 405, 8, 22, "Print",ExpressionUUID->"0233b1ee-fed7-4a79-b5c1-0d470eebf6de",
 CellID->939980415],
Cell[3635876, 104485, 396, 8, 22, "Print",ExpressionUUID->"8b670bbe-27d8-4f2b-80b5-77884b697bcf",
 CellID->1361127573],
Cell[3636275, 104495, 410, 8, 22, "Print",ExpressionUUID->"265a2972-def2-4d08-9ee2-03d7c8be6df7",
 CellID->1475311567],
Cell[3636688, 104505, 669, 13, 22, "Print",ExpressionUUID->"22eae6a0-6c52-4a35-98c4-dc42835c0bc6",
 CellID->715318307]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3637418, 104525, 230, 5, 40, "MathCaption",ExpressionUUID->"736dfdd3-5a7a-45ec-803f-f733d4808ae1"],
Cell[CellGroupData[{
Cell[3637673, 104534, 651, 20, 27, "Input",ExpressionUUID->"99955222-ce72-41e5-8b18-70067fb7c6ca"],
Cell[CellGroupData[{
Cell[3638349, 104558, 771, 17, 22, "Print",ExpressionUUID->"8e9daef9-6f57-4b1a-b7df-505955bf84a6",
 CellID->741329843],
Cell[3639123, 104577, 960, 21, 22, "Print",ExpressionUUID->"36143434-64ff-4a56-9fe6-7cc4eac6e534",
 CellID->1978035054],
Cell[3640086, 104600, 692, 15, 22, "Print",ExpressionUUID->"010a347c-af20-4b1a-bcf1-6f73c97017a8",
 CellID->1788297018],
Cell[3640781, 104617, 836, 18, 22, "Print",ExpressionUUID->"502036cf-dae6-4eaa-8227-df2ea6026c37",
 CellID->1874341761],
Cell[3641620, 104637, 853, 19, 22, "Print",ExpressionUUID->"6d5d6341-6f69-4444-beda-4b156b47caee",
 CellID->624359284],
Cell[3642476, 104658, 888, 19, 22, "Print",ExpressionUUID->"0ad97094-7c39-49d3-a722-15dd51aaacec",
 CellID->1001386946],
Cell[3643367, 104679, 795, 18, 22, "Print",ExpressionUUID->"cd3971bb-79bb-415b-837f-76d80b499f91",
 CellID->481059883],
Cell[3644165, 104699, 647, 13, 22, "Print",ExpressionUUID->"d80fad38-44e5-425c-8ec7-de384067bf42",
 CellID->2045006273],
Cell[3644815, 104714, 809, 19, 22, "Print",ExpressionUUID->"95254926-4096-424e-af71-4d9f87adfc30",
 CellID->2092143723],
Cell[3645627, 104735, 873, 18, 22, "Print",ExpressionUUID->"e7b846f1-0a2a-402e-9c2b-3daf90f7edce",
 CellID->1909800130],
Cell[3646503, 104755, 802, 18, 22, "Print",ExpressionUUID->"acd4b90e-055a-4274-b95b-9deeef17124f",
 CellID->1720196848],
Cell[3647308, 104775, 660, 13, 22, "Print",ExpressionUUID->"5e78d282-a9fc-481d-8ecc-baf3e4640b90",
 CellID->1923568422],
Cell[3647971, 104790, 404, 8, 22, "Print",ExpressionUUID->"8e868cb3-6596-4e3d-9d9c-251e6e98f508",
 CellID->1195646616],
Cell[3648378, 104800, 396, 8, 22, "Print",ExpressionUUID->"f43336cd-2a3c-4369-a6a8-5cf2db35afa2",
 CellID->1228190133],
Cell[3648777, 104810, 407, 8, 22, "Print",ExpressionUUID->"044bdc6b-8fd5-4191-8b1b-ace978ef741f",
 CellID->578750367],
Cell[3649187, 104820, 664, 13, 22, "Print",ExpressionUUID->"2519bce2-0437-4b6b-8234-15ec0dcd26f5",
 CellID->1103980906]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[3649890, 104838, 447, 10, 41, "DefinitionBox",ExpressionUUID->"409cdb9c-da8a-4fd4-8ff0-a5c50734fedd"],
Cell[3650340, 104850, 105, 0, 23, "Text",ExpressionUUID->"cf19eaea-e9ee-4840-a162-6f1dcbed5993"],
Cell[CellGroupData[{
Cell[3650470, 104854, 123, 0, 40, "MathCaption",ExpressionUUID->"af523ea9-5b16-42c7-979e-793677d7d71c"],
Cell[CellGroupData[{
Cell[3650618, 104858, 307, 10, 27, "Input",ExpressionUUID->"edbf4805-bf65-420d-8e11-2cc6f69930a6"],
Cell[3650928, 104870, 2179, 61, 32, "Output",ExpressionUUID->"e7e87a04-3027-4d97-8483-650a652aef4e",
 CellID->348377062]
}, Open  ]],
Cell[CellGroupData[{
Cell[3653144, 104936, 141, 3, 27, "Input",ExpressionUUID->"f9c9c8ec-ebce-4411-8592-2016b45a22ba"],
Cell[3653288, 104941, 818, 27, 26, "Output",ExpressionUUID->"ca6b3b3a-5b58-4401-a678-bb9bbd35a242",
 CellID->1132859312]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3654155, 104974, 297, 11, 40, "MathCaption",ExpressionUUID->"7f8f4c72-8ef3-485a-a26a-53be4fe5901f"],
Cell[CellGroupData[{
Cell[3654477, 104989, 553, 18, 27, "Input",ExpressionUUID->"6165cf7e-205e-48e3-9ff3-d8edf6926e08"],
Cell[3655033, 105009, 2634, 77, 32, "Output",ExpressionUUID->"dea33014-4b1a-45ac-956a-983b7b8ec221",
 CellID->376442482]
}, Open  ]],
Cell[CellGroupData[{
Cell[3657704, 105091, 1073, 31, 66, "Input",ExpressionUUID->"ab8d12b0-6e88-4a0d-bf75-c21380c73697"],
Cell[3658780, 105124, 12326, 359, 50, "Output",ExpressionUUID->"5166d641-dc71-4243-8385-f292f5e45ac2",
 CellID->381970912]
}, Open  ]],
Cell[CellGroupData[{
Cell[3671143, 105488, 397, 12, 27, "Input",ExpressionUUID->"fa85f148-7855-4871-98f7-7e9a7698599f"],
Cell[3671543, 105502, 8139, 220, 56, "Output",ExpressionUUID->"389b13c3-f182-4b56-abdc-6ab90a6bac58",
 CellID->956497101]
}, Open  ]],
Cell[CellGroupData[{
Cell[3679719, 105727, 268, 7, 27, "Input",ExpressionUUID->"8b38849a-425b-4af7-9b32-2899b7640259"],
Cell[3679990, 105736, 7693, 242, 50, "Output",ExpressionUUID->"004ceac4-e828-433b-9a20-f50557ed45d2",
 CellID->1931678712]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3687732, 105984, 281, 6, 56, "MathCaption",ExpressionUUID->"edc3d274-37cf-4dd1-bf15-b27946780313"],
Cell[CellGroupData[{
Cell[3688038, 105994, 268, 7, 27, "Input",ExpressionUUID->"ad9564ab-1ee2-4221-9ccf-a383507c9f23"],
Cell[3688309, 106003, 3621, 130, 47, "Output",ExpressionUUID->"647edd95-8c5a-43d8-b9dd-cd3a99d6895c",
 CellID->66337238]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3691979, 106139, 86, 0, 40, "MathCaption",ExpressionUUID->"c438bef6-fc05-4fdb-bc9c-f0d672a6d8c1"],
Cell[CellGroupData[{
Cell[3692090, 106143, 148, 3, 27, "Input",ExpressionUUID->"e6953740-2e96-462e-b7a3-2dcfffa52ce2"],
Cell[CellGroupData[{
Cell[3692263, 106150, 578, 12, 22, "Print",ExpressionUUID->"9bcbc04b-4f90-4f57-adda-c02cda688f96",
 CellID->385946273],
Cell[3692844, 106164, 619, 12, 22, "Print",ExpressionUUID->"80e924df-e2f3-4d0c-a810-45cfeb699c1d",
 CellID->1198396371],
Cell[3693466, 106178, 606, 12, 22, "Print",ExpressionUUID->"35306c64-a2c3-4bc4-bd7d-3cfd352de8ba",
 CellID->198682184],
Cell[3694075, 106192, 599, 12, 22, "Print",ExpressionUUID->"76601480-4efd-42cb-a70e-88acdfa86cec",
 CellID->2061879050],
Cell[3694677, 106206, 593, 12, 22, "Print",ExpressionUUID->"4a38b645-ddf8-4bb0-aa8a-0882a3b1f16a",
 CellID->57581711],
Cell[3695273, 106220, 589, 12, 22, "Print",ExpressionUUID->"8d52c105-8fc7-4cc5-a6f6-a73525d88f99",
 CellID->1408175627],
Cell[3695865, 106234, 585, 12, 22, "Print",ExpressionUUID->"07af8537-20d1-47b4-915f-7627c810b87b",
 CellID->1937973049],
Cell[3696453, 106248, 604, 12, 22, "Print",ExpressionUUID->"2e34563c-6e9c-451d-8303-81c1b1bf8cce",
 CellID->628560494],
Cell[3697060, 106262, 584, 12, 22, "Print",ExpressionUUID->"a2f8fb32-8f18-41d4-8267-4807d14c4cd7",
 CellID->744163065],
Cell[3697647, 106276, 570, 12, 22, "Print",ExpressionUUID->"58f49ab5-4db6-4bde-a3cf-221aa7d81a0f",
 CellID->203873327],
Cell[3698220, 106290, 576, 12, 22, "Print",ExpressionUUID->"a8a0c491-7acb-425a-a227-0efd9c653abb",
 CellID->1262589924],
Cell[3698799, 106304, 592, 12, 22, "Print",ExpressionUUID->"6af02f5f-4fb8-4bee-a05c-2d0b9c92b6fb",
 CellID->1062936661],
Cell[3699394, 106318, 585, 12, 22, "Print",ExpressionUUID->"1a212853-70f0-421f-afb9-5cf1fabd02e3",
 CellID->2007485681]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3700028, 106336, 153, 3, 27, "Input",ExpressionUUID->"d7119c8e-ad90-40a5-b1f1-42401a833f67"],
Cell[CellGroupData[{
Cell[3700206, 106343, 588, 12, 22, "Print",ExpressionUUID->"556d32c3-36b5-491e-b6e7-b6bb13aacec7",
 CellID->696262550],
Cell[3700797, 106357, 616, 12, 22, "Print",ExpressionUUID->"e89113b5-bc85-4751-8af8-e6b3f19e204e",
 CellID->446285347],
Cell[3701416, 106371, 607, 12, 22, "Print",ExpressionUUID->"be2a5666-c865-4426-a2a6-d1faca75cd90",
 CellID->1704931705],
Cell[3702026, 106385, 596, 12, 22, "Print",ExpressionUUID->"b5c6b856-9d70-40ca-b7f9-ba9792fb3668",
 CellID->912864014],
Cell[3702625, 106399, 592, 12, 22, "Print",ExpressionUUID->"2b220c96-b130-4796-bcf7-c33f327d72c9",
 CellID->882309130],
Cell[3703220, 106413, 586, 12, 22, "Print",ExpressionUUID->"861efe30-f1e7-4956-b303-2e826a856989",
 CellID->594965375],
Cell[3703809, 106427, 582, 12, 22, "Print",ExpressionUUID->"5a318d00-8ef5-4cb4-bddc-0d1e340e21be",
 CellID->899197193],
Cell[3704394, 106441, 599, 12, 22, "Print",ExpressionUUID->"d11207fe-86fb-43d9-b6c1-9c603fa1f5fd",
 CellID->964798812],
Cell[3704996, 106455, 583, 12, 22, "Print",ExpressionUUID->"95f5d7b3-4d5d-4050-92bc-0a599ba86412",
 CellID->1872098434],
Cell[3705582, 106469, 569, 12, 22, "Print",ExpressionUUID->"cd4a2522-a723-4f19-bbdf-2db04b202cab",
 CellID->1575468041],
Cell[3706154, 106483, 574, 12, 22, "Print",ExpressionUUID->"d5f28504-8fa2-43ab-b697-fb3959cfa6e7",
 CellID->417997577],
Cell[3706731, 106497, 602, 12, 22, "Print",ExpressionUUID->"7c4dc29f-b783-4528-9a26-192d4b86ebc3",
 CellID->866267302],
Cell[3707336, 106511, 595, 12, 22, "Print",ExpressionUUID->"467fd898-e57b-40be-814e-331c7964dc21",
 CellID->1562901701]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3707980, 106529, 151, 3, 27, "Input",ExpressionUUID->"987a4cfa-08d7-469f-ba84-5d4a498fdca4"],
Cell[CellGroupData[{
Cell[3708156, 106536, 585, 12, 22, "Print",ExpressionUUID->"ce149a00-318b-4e8a-90bf-dd9e79185c07",
 CellID->1816402250],
Cell[3708744, 106550, 619, 12, 22, "Print",ExpressionUUID->"a0cb8cd3-d205-4067-ad49-d69c480d6ffc",
 CellID->1792573645],
Cell[3709366, 106564, 586, 12, 22, "Print",ExpressionUUID->"ead943a8-e838-4cc0-9284-c184c4ffb3c8",
 CellID->737167752],
Cell[3709955, 106578, 600, 12, 22, "Print",ExpressionUUID->"96b9fb0c-dc99-4be2-b40a-fedd4764c8c0",
 CellID->745248321],
Cell[3710558, 106592, 595, 12, 22, "Print",ExpressionUUID->"57104a82-407f-4bdd-bbe9-9cf985996b28",
 CellID->1263901743],
Cell[3711156, 106606, 589, 12, 22, "Print",ExpressionUUID->"d4342f5d-5e55-45d5-a9c2-0f2dfcd152e9",
 CellID->1635842124],
Cell[3711748, 106620, 583, 12, 22, "Print",ExpressionUUID->"880353ba-c39b-40e0-b0fa-3d1f2775ecb2",
 CellID->1512929174],
Cell[3712334, 106634, 583, 12, 22, "Print",ExpressionUUID->"2770eda5-c4cb-4221-b406-f53f465589c9",
 CellID->1594246938],
Cell[3712920, 106648, 582, 12, 22, "Print",ExpressionUUID->"bab29aa1-9271-461f-85cc-90607100524f",
 CellID->354823347],
Cell[3713505, 106662, 570, 12, 22, "Print",ExpressionUUID->"40d07db7-b1ab-4ee8-8a60-378c6f99f6ac",
 CellID->2101388957],
Cell[3714078, 106676, 576, 12, 22, "Print",ExpressionUUID->"97ab3683-927f-45d1-870d-b2fe483add39",
 CellID->528141737],
Cell[3714657, 106690, 598, 12, 22, "Print",ExpressionUUID->"accc993f-d65e-4f32-ade3-8ca76606c645",
 CellID->928260708],
Cell[3715258, 106704, 593, 12, 22, "Print",ExpressionUUID->"23294db7-290a-454e-a4c7-714917a3eb99",
 CellID->1956507548]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3715900, 106722, 150, 3, 27, "Input",ExpressionUUID->"88e092ec-768a-43b1-adb7-b8e298cb539f"],
Cell[CellGroupData[{
Cell[3716075, 106729, 724, 18, 22, "Print",ExpressionUUID->"198a6599-5309-4d2c-9166-99d244f526c9",
 CellID->752756372],
Cell[3716802, 106749, 560, 12, 22, "Print",ExpressionUUID->"4910fd08-3656-4bc7-97e1-be5944d016c1",
 CellID->750179727]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3717435, 106769, 108, 0, 39, "Subsection",ExpressionUUID->"5b95623d-0d2b-4cf5-925f-4956c23651ea"],
Cell[3717546, 106771, 200, 3, 23, "Text",ExpressionUUID->"01f65013-28df-469b-bea6-ea42be0ea907"],
Cell[CellGroupData[{
Cell[3717771, 106778, 194, 5, 40, "MathCaption",ExpressionUUID->"df5a273e-86b5-4764-b99e-9b8c108815ca"],
Cell[CellGroupData[{
Cell[3717990, 106787, 477, 13, 27, "Input",ExpressionUUID->"e96cd80b-2b78-4531-89cf-2a62b395d38c"],
Cell[CellGroupData[{
Cell[3718492, 106804, 758, 17, 22, "Print",ExpressionUUID->"dd69f961-1171-49b8-9c31-cc9036510a73",
 CellID->695821353],
Cell[3719253, 106823, 816, 19, 22, "Print",ExpressionUUID->"223b5d2d-b436-43c8-8bbf-684dbf793001",
 CellID->1621021956],
Cell[3720072, 106844, 688, 15, 22, "Print",ExpressionUUID->"317e40e1-2f18-4348-bec8-ce2d677db301",
 CellID->1538468650],
Cell[3720763, 106861, 806, 18, 22, "Print",ExpressionUUID->"8be054b6-da36-4c0f-b6ff-e8e809caf2fe",
 CellID->1339490670],
Cell[3721572, 106881, 825, 19, 22, "Print",ExpressionUUID->"4877a68d-0dfc-4f93-acfd-b0ddd8c08886",
 CellID->748619189],
Cell[3722400, 106902, 876, 20, 22, "Print",ExpressionUUID->"034c99e4-fc4f-45ad-91b8-1af0f84f8cf8",
 CellID->2083505397],
Cell[3723279, 106924, 769, 17, 22, "Print",ExpressionUUID->"bf9b8100-c229-4464-bed2-41c868f79477",
 CellID->606859397],
Cell[3724051, 106943, 678, 14, 22, "Print",ExpressionUUID->"8b387915-a3fc-4305-9d70-76a36489b0e5",
 CellID->1086045776],
Cell[3724732, 106959, 786, 18, 22, "Print",ExpressionUUID->"ae308982-aeea-4397-8dd5-94818b3111e5",
 CellID->1707841379],
Cell[3725521, 106979, 859, 19, 22, "Print",ExpressionUUID->"9606baa0-fe39-49d5-8738-486377542418",
 CellID->423631051],
Cell[3726383, 107000, 778, 17, 22, "Print",ExpressionUUID->"6529fee7-41d2-4692-9ec6-983db9c89240",
 CellID->1058784352],
Cell[3727164, 107019, 693, 15, 22, "Print",ExpressionUUID->"dfbf130b-069b-4f21-90dc-70416a375042",
 CellID->939105022],
Cell[3727860, 107036, 667, 13, 22, "Print",ExpressionUUID->"56a4b494-2cd2-4d7f-84e0-49963fef97c7",
 CellID->1680782951]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3728576, 107055, 151, 3, 27, "Input",ExpressionUUID->"81cbfef4-405f-4363-ae3c-451e7c4e4eef"],
Cell[CellGroupData[{
Cell[3728752, 107062, 584, 12, 22, "Print",ExpressionUUID->"07526731-701b-4a09-a715-bd4c1f9974d7",
 CellID->919164819],
Cell[3729339, 107076, 629, 13, 22, "Print",ExpressionUUID->"b82b37dc-b6bb-4133-be71-c16b7f695fc1",
 CellID->1553240062],
Cell[3729971, 107091, 592, 12, 22, "Print",ExpressionUUID->"195097d8-5ab1-4de1-9e1a-c5b32c2e1dc2",
 CellID->494509102],
Cell[3730566, 107105, 608, 13, 22, "Print",ExpressionUUID->"34f733f0-fd4b-40cf-8c84-0878e7e4600b",
 CellID->452763424],
Cell[3731177, 107120, 581, 12, 22, "Print",ExpressionUUID->"88c2e5ef-3c3f-41e8-bafa-dcea7307b6cc",
 CellID->1060305847],
Cell[3731761, 107134, 596, 13, 22, "Print",ExpressionUUID->"eb8046a4-7f95-4342-b1e1-c1115c9c1457",
 CellID->741067723],
Cell[3732360, 107149, 589, 12, 22, "Print",ExpressionUUID->"58a8f942-c946-4c3e-91f2-33b7a33b9612",
 CellID->1867551508],
Cell[3732952, 107163, 590, 12, 22, "Print",ExpressionUUID->"de52af17-14b9-43b6-85bc-14b3147fddff",
 CellID->317210404],
Cell[3733545, 107177, 588, 12, 22, "Print",ExpressionUUID->"ec3bdcc6-9c58-488e-8110-b15b9b27f59c",
 CellID->1300848174],
Cell[3734136, 107191, 577, 12, 22, "Print",ExpressionUUID->"06a3cc89-b9c1-4534-a96c-c85a25b798fd",
 CellID->1971733958],
Cell[3734716, 107205, 583, 12, 22, "Print",ExpressionUUID->"b826d113-9109-444e-8480-5356a269f5fc",
 CellID->1112475249],
Cell[3735302, 107219, 601, 12, 22, "Print",ExpressionUUID->"ef7eab35-1291-4edc-b04e-07441d66b181",
 CellID->1378879161],
Cell[3735906, 107233, 591, 12, 22, "Print",ExpressionUUID->"9e8c17bf-61a3-4108-ac02-16e952cdab0e",
 CellID->1326024150]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[3736536, 107250, 1034, 23, 108, "Text",ExpressionUUID->"232c1865-fe41-40e4-985a-c4bb7f84e3f4"],
Cell[CellGroupData[{
Cell[3737595, 107277, 199, 3, 40, "MathCaption",ExpressionUUID->"b8a6c7e0-6359-409f-8edb-723a9175517b"],
Cell[CellGroupData[{
Cell[3737819, 107284, 473, 13, 27, "Input",ExpressionUUID->"e15f39cb-b782-481a-a5fc-46da894cee2e"],
Cell[CellGroupData[{
Cell[3738317, 107301, 759, 17, 22, "Print",ExpressionUUID->"d53f452d-8dd1-4886-b349-f6b65bdc3c55",
 CellID->2050084131],
Cell[3739079, 107320, 818, 19, 22, "Print",ExpressionUUID->"ad386c9d-4851-4264-b19b-6860bd36de56",
 CellID->1536357249],
Cell[3739900, 107341, 263, 4, 22, "Print",ExpressionUUID->"24cd41c1-ccab-49fa-9373-d94329599baa",
 CellID->212912415],
Cell[3740166, 107347, 669, 13, 22, "Print",ExpressionUUID->"b679ba94-39c2-4177-9a17-a96269a5391a",
 CellID->711203003]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3740884, 107366, 281, 9, 27, "Input",ExpressionUUID->"755cf158-3089-4129-9be9-e5c838f7afce"],
Cell[3741168, 107377, 172, 3, 26, "Output",ExpressionUUID->"71deb919-6c0c-4d81-a313-69f79cfda858",
 CellID->294892219]
}, Open  ]],
Cell[CellGroupData[{
Cell[3741377, 107385, 151, 3, 27, "Input",ExpressionUUID->"3ac1d263-d2bd-4921-988a-fc0d545f0b7a"],
Cell[CellGroupData[{
Cell[3741553, 107392, 587, 12, 22, "Print",ExpressionUUID->"2ee8446f-8154-488b-b4b9-2f8c1cd06d1d",
 CellID->2008751104],
Cell[3742143, 107406, 598, 12, 22, "Print",ExpressionUUID->"5e73aaa5-c36f-4de9-90b9-ed50867ba2c4",
 CellID->489390441],
Cell[3742744, 107420, 590, 12, 22, "Print",ExpressionUUID->"d045421b-ba5c-40bd-891a-2844a54e9f6f",
 CellID->322773877]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[3743407, 107440, 97, 0, 39, "Subsection",ExpressionUUID->"cfa3ecda-cd9c-4be7-a543-398c4e19470c"],
Cell[3743507, 107442, 350, 6, 40, "Text",ExpressionUUID->"509ca7d7-5855-454b-bba4-3e1b59338048"],
Cell[CellGroupData[{
Cell[3743882, 107452, 120, 0, 40, "MathCaption",ExpressionUUID->"3036de7d-3bdf-42c9-968c-1bb08cb30151"],
Cell[CellGroupData[{
Cell[3744027, 107456, 374, 11, 27, "Input",ExpressionUUID->"55a07874-51d5-4766-a57a-b5b013f7a261"],
Cell[CellGroupData[{
Cell[3744426, 107471, 765, 17, 22, "Print",ExpressionUUID->"be3a7142-4223-417f-a8c2-6b759f11f425",
 CellID->701636648],
Cell[3745194, 107490, 821, 19, 22, "Print",ExpressionUUID->"3791b567-95d7-4dfa-93ce-00a6a66cbcbd",
 CellID->574437259],
Cell[3746018, 107511, 682, 15, 22, "Print",ExpressionUUID->"951e5298-c5d5-4062-8ae6-a91ebdfeb29c",
 CellID->1628018748],
Cell[3746703, 107528, 800, 18, 22, "Print",ExpressionUUID->"b03a5d9c-95e1-4f0d-8609-87b12e2cfb57",
 CellID->1021951209],
Cell[3747506, 107548, 817, 19, 22, "Print",ExpressionUUID->"c6ca4e71-40ec-4119-92cf-28ebfda4ae89",
 CellID->359760957],
Cell[3748326, 107569, 846, 19, 22, "Print",ExpressionUUID->"b10676cb-0be5-4e8f-9dc5-0f6344344a5b",
 CellID->1335937994],
Cell[3749175, 107590, 762, 17, 22, "Print",ExpressionUUID->"33b78ac1-6fb0-45e5-b808-ef03edd5ab41",
 CellID->1168335773],
Cell[3749940, 107609, 269, 4, 22, "Print",ExpressionUUID->"05d688c3-790c-466e-b702-1d8fca1cc697",
 CellID->1776352290],
Cell[3750212, 107615, 647, 13, 22, "Print",ExpressionUUID->"bf67f6ac-de7c-4689-aa53-64a34dfb4f48",
 CellID->835277436],
Cell[3750862, 107630, 272, 4, 22, "Print",ExpressionUUID->"51ece204-fb72-469d-966a-a4b6edd4661a",
 CellID->407266413],
Cell[3751137, 107636, 777, 18, 22, "Print",ExpressionUUID->"4eac071e-5a36-41b5-a415-618f92853c05",
 CellID->896220302],
Cell[3751917, 107656, 853, 19, 22, "Print",ExpressionUUID->"a18d7c21-a7fe-4fce-8ca4-f9931496b219",
 CellID->324315656],
Cell[3752773, 107677, 769, 17, 22, "Print",ExpressionUUID->"bf9aec6d-d45b-4c80-a2eb-c15a343c0a75",
 CellID->459734306],
Cell[3753545, 107696, 657, 13, 22, "Print",ExpressionUUID->"2075aa61-ea78-456e-a6de-c0cab61a1e7d",
 CellID->108319036],
Cell[3754205, 107711, 406, 8, 22, "Print",ExpressionUUID->"bc9655d3-44ec-4fc2-843e-47981d248207",
 CellID->1599389043],
Cell[3754614, 107721, 394, 8, 22, "Print",ExpressionUUID->"615194d3-a4ca-4dde-9673-21d70410eac8",
 CellID->878836202],
Cell[3755011, 107731, 407, 8, 22, "Print",ExpressionUUID->"e3dc52d4-a6dd-46a6-bc8b-a4b76f4a8dcf",
 CellID->207522948],
Cell[3755421, 107741, 673, 13, 22, "Print",ExpressionUUID->"71caa11f-826b-4c95-8a3a-8706efb0b687",
 CellID->324744308]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3756155, 107761, 288, 8, 40, "MathCaption",ExpressionUUID->"7bf0c0d2-7414-41bf-97a3-34e74a462786"],
Cell[CellGroupData[{
Cell[3756468, 107773, 198, 5, 27, "Input",ExpressionUUID->"43fe3905-ffa5-446f-8867-a5d6adf16845"],
Cell[3756669, 107780, 626, 12, 22, "Print",ExpressionUUID->"96ea8474-5ba7-4b39-8186-e8defe63287c",
 CellID->1221637529]
}, Open  ]],
Cell[CellGroupData[{
Cell[3757332, 107797, 394, 12, 27, "Input",ExpressionUUID->"e662822f-379f-436a-9c3c-25a8280a85f9"],
Cell[3757729, 107811, 558, 12, 22, "Print",ExpressionUUID->"d525bcd9-44eb-47b6-9aa8-8ccd6657e2b7",
 CellID->2118284583]
}, Open  ]],
Cell[CellGroupData[{
Cell[3758324, 107828, 493, 15, 27, "Input",ExpressionUUID->"1366ddf3-0160-45ab-8424-cbd943f94780"],
Cell[3758820, 107845, 589, 13, 22, "Print",ExpressionUUID->"96dd61e3-7268-4002-85a7-9ce87e7471ad",
 CellID->1025086515]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3759458, 107864, 139, 1, 40, "MathCaption",ExpressionUUID->"39814496-8bf1-40a9-bcc1-c65e26f46537"],
Cell[3759600, 107867, 288, 8, 47, "Input",ExpressionUUID->"cde3dd99-4f7f-4fcc-b3aa-dbc9ddd04abe"]
}, Open  ]],
Cell[CellGroupData[{
Cell[3759925, 107880, 320, 4, 72, "MathCaption",ExpressionUUID->"42fd3cba-f88d-4b02-a089-afc7fab41157"],
Cell[CellGroupData[{
Cell[3760270, 107888, 550, 15, 47, "Input",ExpressionUUID->"7830ab30-549a-48f1-9d07-fb022390e44f"],
Cell[3760823, 107905, 611, 11, 22, "Message",ExpressionUUID->"997f40cd-0972-400d-bd59-4d6c257348e3",
 CellID->1515241371],
Cell[CellGroupData[{
Cell[3761459, 107920, 764, 17, 22, "Print",ExpressionUUID->"d13f05ed-e699-45ec-af90-ea45babb45d5",
 CellID->1800636271],
Cell[3762226, 107939, 771, 17, 22, "Print",ExpressionUUID->"5d3ab98d-cfd4-4606-b885-37d08b50b2d9",
 CellID->954253173],
Cell[3763000, 107958, 679, 15, 22, "Print",ExpressionUUID->"2f1367b8-b441-4654-9bbd-6d4be431ee4e",
 CellID->725202815],
Cell[3763682, 107975, 800, 18, 22, "Print",ExpressionUUID->"d3edb312-cb59-47bc-aa7c-6abde4cd9ef9",
 CellID->1514014980],
Cell[3764485, 107995, 819, 19, 22, "Print",ExpressionUUID->"1ae48d4e-55f7-45d7-8186-d1c25bf557f4",
 CellID->516035557],
Cell[3765307, 108016, 844, 19, 22, "Print",ExpressionUUID->"08bd624f-ea29-4f99-acfc-43443aaffb24",
 CellID->1470639913],
Cell[3766154, 108037, 764, 17, 22, "Print",ExpressionUUID->"a46e5085-bbb9-44d6-b6a3-264b529d14e9",
 CellID->1655724468],
Cell[3766921, 108056, 268, 4, 22, "Print",ExpressionUUID->"0c9c73cc-110c-42b6-a773-a3aa9bb95e60",
 CellID->668035850],
Cell[3767192, 108062, 645, 13, 22, "Print",ExpressionUUID->"9f156de3-a51a-4307-957d-698d48f4fb59",
 CellID->634024434],
Cell[3767840, 108077, 272, 4, 22, "Print",ExpressionUUID->"9d1f9650-3199-4855-ae9e-7264ffc3e77e",
 CellID->234805824],
Cell[3768115, 108083, 778, 18, 22, "Print",ExpressionUUID->"a0c611e9-8b4c-4476-b1d5-75ae089af348",
 CellID->1454024429],
Cell[3768896, 108103, 852, 19, 22, "Print",ExpressionUUID->"cdc032b1-7b9a-4260-b31c-dbcbaa072171",
 CellID->1572158990],
Cell[3769751, 108124, 769, 17, 22, "Print",ExpressionUUID->"f770e74d-1591-4b4b-b32d-384beef3942e",
 CellID->637352242],
Cell[3770523, 108143, 657, 13, 22, "Print",ExpressionUUID->"a18e410d-b1a9-423a-858c-5926d2181156",
 CellID->330700058],
Cell[3771183, 108158, 405, 8, 22, "Print",ExpressionUUID->"d2ea5b6f-1ff1-4c32-b04b-b3d7d2a992ae",
 CellID->388196370],
Cell[3771591, 108168, 395, 8, 22, "Print",ExpressionUUID->"4caae49f-a874-4d33-b01f-a207d00dd3b4",
 CellID->361071004],
Cell[3771989, 108178, 408, 8, 22, "Print",ExpressionUUID->"0583c9e3-3031-4180-80c9-9cb855e66fa1",
 CellID->1879061453],
Cell[3772400, 108188, 675, 13, 22, "Print",ExpressionUUID->"7b6c84a3-180d-44fc-9962-fc16f361ceb7",
 CellID->491327789],
Cell[3773078, 108203, 793, 16, 42, "Print",ExpressionUUID->"11fcf667-47a8-4e92-ad4a-ec00f31d08c6",
 CellID->671527008],
Cell[3773874, 108221, 727, 14, 22, "Print",ExpressionUUID->"122eef35-238a-4a8b-b2e6-50299b244071",
 CellID->1327858471],
Cell[3774604, 108237, 651, 13, 22, "Print",ExpressionUUID->"d1b0fa12-84d3-4a97-8ee6-1fb771e5118e",
 CellID->667262473]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3775316, 108257, 193, 3, 40, "MathCaption",ExpressionUUID->"a38aa410-eff7-4e19-889b-8715a2111af5"],
Cell[CellGroupData[{
Cell[3775534, 108264, 253, 6, 27, "Input",ExpressionUUID->"e54b9bbc-9b58-4894-870f-302255832284"],
Cell[3775790, 108272, 614, 12, 22, "Print",ExpressionUUID->"6980c9eb-3079-4c18-8e3a-495e2b41d304",
 CellID->520779077]
}, Open  ]],
Cell[CellGroupData[{
Cell[3776441, 108289, 356, 10, 27, "Input",ExpressionUUID->"b374831a-1330-440a-b39e-6b0d0c630ff6"],
Cell[3776800, 108301, 1589, 51, 43, "Output",ExpressionUUID->"cc9a1c31-fd8e-4a5e-a622-3f65eb603d71",
 CellID->1738085614]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3778438, 108358, 140, 1, 40, "MathCaption",ExpressionUUID->"ea052db3-9d0c-404b-9d69-36e724f2290c"],
Cell[CellGroupData[{
Cell[3778603, 108363, 218, 6, 27, "Input",ExpressionUUID->"385545ed-2d4d-420a-b8e8-63e43efa566d"],
Cell[3778824, 108371, 647, 13, 22, "Print",ExpressionUUID->"e88f489c-e860-4d5e-94f9-26d6a0801d40",
 CellID->1932348004]
}, Open  ]],
Cell[CellGroupData[{
Cell[3779508, 108389, 489, 14, 27, "Input",ExpressionUUID->"024069b5-f93c-4141-a4a1-2aa04f4fcafb"],
Cell[3780000, 108405, 646, 13, 22, "Print",ExpressionUUID->"0f300746-b9d1-4f4b-8ba2-9fde5962dd73",
 CellID->577709647]
}, Open  ]]
}, Open  ]],
Cell[3780673, 108422, 438, 8, 57, "Text",ExpressionUUID->"e25d19ea-cc67-4e0d-b0f4-833909cb29c6"],
Cell[CellGroupData[{
Cell[3781136, 108434, 294, 6, 56, "MathCaption",ExpressionUUID->"b9ac5637-0f38-43d5-b6de-6320ec1617d6"],
Cell[CellGroupData[{
Cell[3781455, 108444, 283, 8, 27, "Input",ExpressionUUID->"34c64ef0-1bfc-4180-aa2c-0abcb3a34c54"],
Cell[3781741, 108454, 1059, 31, 32, "Output",ExpressionUUID->"6500ec17-4539-4d71-a29a-e1a36e9d5ae2",
 CellID->792469042]
}, Open  ]],
Cell[CellGroupData[{
Cell[3782837, 108490, 283, 8, 27, "Input",ExpressionUUID->"43764773-7ded-47ab-90b2-f636e3bb07d3"],
Cell[3783123, 108500, 4907, 142, 35, "Output",ExpressionUUID->"849cb0d2-3553-41e9-92b8-168d717401aa",
 CellID->333610200]
}, Open  ]],
Cell[CellGroupData[{
Cell[3788067, 108647, 515, 17, 27, "Input",ExpressionUUID->"1976425a-62f5-447e-be47-4f905ddaad80"],
Cell[3788585, 108666, 12527, 346, 38, "Output",ExpressionUUID->"a21bafb3-bc0a-4036-b6a0-275d9779bbaa",
 CellID->2054312730]
}, Open  ]],
Cell[CellGroupData[{
Cell[3801149, 109017, 400, 12, 27, "Input",ExpressionUUID->"7c27b8c0-d85d-4ab1-8da9-cac4a6b6d7cb"],
Cell[3801552, 109031, 91133, 2522, 119, "Output",ExpressionUUID->"3669b57e-87bc-4de0-8103-5c078dd7016e",
 CellID->1739460291]
}, Open  ]],
Cell[CellGroupData[{
Cell[3892722, 111558, 147, 3, 27, "Input",ExpressionUUID->"23f5084e-04e8-4f25-a06c-0e6cfb5f18ca"],
Cell[3892872, 111563, 38866, 1073, 63, "Output",ExpressionUUID->"b33d1030-2cd2-4422-8ca3-bcf8d71f5958",
 CellID->1414224226]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3931787, 112642, 126, 0, 40, "MathCaption",ExpressionUUID->"3fb159f4-3f59-4f20-90a6-e22c158f9810"],
Cell[CellGroupData[{
Cell[3931938, 112646, 185, 5, 27, "Input",ExpressionUUID->"dfaf6e7e-855c-40dc-9ac8-2d122c3d6198"],
Cell[3932126, 112653, 12201, 331, 35, "Output",ExpressionUUID->"98fa84a9-a110-4f62-afaa-3729a4b1b8d5",
 CellID->517561386]
}, Open  ]],
Cell[CellGroupData[{
Cell[3944364, 112989, 185, 5, 27, "Input",ExpressionUUID->"59834dbd-99d8-4da7-8ed0-176fc556a5fe"],
Cell[3944552, 112996, 171, 3, 26, "Output",ExpressionUUID->"92964102-44f4-456a-892c-d5c61649071c",
 CellID->1345444152]
}, Open  ]],
Cell[CellGroupData[{
Cell[3944760, 113004, 186, 5, 27, "Input",ExpressionUUID->"861edced-869e-490b-b3ab-c343d4eaa08f"],
Cell[3944949, 113011, 3736, 103, 35, "Output",ExpressionUUID->"09d5cac1-b4c0-4c03-a3fa-98b23084b657",
 CellID->2024543401]
}, Open  ]]
}, Open  ]],
Cell[3948712, 113118, 462, 10, 57, "Text",ExpressionUUID->"5181de5e-f4d7-4d66-98db-1d1f5c2c3aab"],
Cell[CellGroupData[{
Cell[3949199, 113132, 229, 5, 40, "MathCaption",ExpressionUUID->"6d241c2a-5fec-4b1e-a3ae-f8dafb993c37"],
Cell[CellGroupData[{
Cell[3949453, 113141, 250, 7, 27, "Input",ExpressionUUID->"56532404-b8e1-4c21-86bb-c930342ab471"],
Cell[3949706, 113150, 1481, 41, 35, "Output",ExpressionUUID->"42f06f82-0c63-4653-b016-d6befec251fd",
 CellID->1770037539]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3951236, 113197, 148, 2, 40, "MathCaption",ExpressionUUID->"020ed742-1c27-4ba3-8717-c58ba81f2182"],
Cell[CellGroupData[{
Cell[3951409, 113203, 222, 5, 27, "Input",ExpressionUUID->"b5dded42-5dde-42c6-923c-dad0ed6a7d59"],
Cell[3951634, 113210, 3103, 89, 32, "Output",ExpressionUUID->"6ab27f5f-a299-4bfd-9856-07954b36d268",
 CellID->2077439729]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3954786, 113305, 118, 0, 40, "MathCaption",ExpressionUUID->"b08d3cba-a484-4f1e-b52d-74afaccbb1a0"],
Cell[CellGroupData[{
Cell[3954929, 113309, 149, 3, 27, "Input",ExpressionUUID->"d871482d-026c-4637-8f69-70532d17bff7"],
Cell[3955081, 113314, 7415, 211, 32, "Output",ExpressionUUID->"c808fa89-40e0-43c1-8f0a-facacaa43c46",
 CellID->1740392377]
}, Open  ]],
Cell[CellGroupData[{
Cell[3962533, 113530, 138, 3, 27, "Input",ExpressionUUID->"6c7a4ffb-8291-4427-bd60-f1a85bf3e8a2"],
Cell[3962674, 113535, 11985, 347, 32, "Output",ExpressionUUID->"0515b4e6-d638-4ba1-9bab-58b53baf2afd",
 CellID->2047408099]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3974708, 113888, 209, 5, 40, "MathCaption",ExpressionUUID->"e21d621c-e1f0-4159-8fb8-334851df42a1"],
Cell[CellGroupData[{
Cell[3974942, 113897, 274, 8, 27, "Input",ExpressionUUID->"051426ca-e44e-4f37-be5e-d204a8553c96"],
Cell[3975219, 113907, 1481, 41, 35, "Output",ExpressionUUID->"57a6315f-c43d-4eb4-849d-10f66b2080bc",
 CellID->1657444012]
}, Open  ]]
}, Open  ]],
Cell[3976727, 113952, 1881, 49, 109, "DefinitionBox",ExpressionUUID->"d44273b4-46a0-4111-8cbb-f40ff8939202"],
Cell[3978611, 114003, 92, 0, 23, "Text",ExpressionUUID->"06e1d5f6-2c74-4cd5-9ad0-e9db4972a015"],
Cell[3978706, 114005, 589, 12, 74, "Text",ExpressionUUID->"d32d75b5-f77d-4c2f-a865-7fbc5ed87b84"],
Cell[CellGroupData[{
Cell[3979320, 114021, 150, 2, 40, "MathCaption",ExpressionUUID->"a7c917e1-b407-4bb0-b408-1a634f3c25aa"],
Cell[CellGroupData[{
Cell[3979495, 114027, 213, 6, 27, "Input",ExpressionUUID->"42cb6533-235f-4fe3-b675-b327c749d475"],
Cell[3979711, 114035, 1181, 34, 32, "Output",ExpressionUUID->"d8a57b05-44af-4ce0-b70b-c35d27f4342a",
 CellID->1651124447]
}, Open  ]],
Cell[CellGroupData[{
Cell[3980929, 114074, 154, 3, 27, "Input",ExpressionUUID->"0f08fb59-b4e1-4cc1-a035-245774613750"],
Cell[3981086, 114079, 3123, 90, 32, "Output",ExpressionUUID->"e0ea26de-fb09-4071-a352-dea565c225fc",
 CellID->664894195]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3984258, 114175, 209, 3, 56, "MathCaption",ExpressionUUID->"4bd7057c-ec59-40e3-8e29-eb6b1a81b4b5"],
Cell[CellGroupData[{
Cell[3984492, 114182, 281, 9, 27, "Input",ExpressionUUID->"57b88724-41c8-4910-8f83-4f93b76d5b22"],
Cell[3984776, 114193, 2207, 64, 32, "Output",ExpressionUUID->"0ca5d11b-0838-4e8c-b11a-94735196397d",
 CellID->1519320024]
}, Open  ]],
Cell[CellGroupData[{
Cell[3987020, 114262, 194, 5, 27, "Input",ExpressionUUID->"7dbffcac-1fb1-4378-847f-368ec11d481d"],
Cell[3987217, 114269, 5029, 146, 32, "Output",ExpressionUUID->"58370695-6755-4b68-ada6-b29652d9ae6b",
 CellID->1631080538]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3992295, 114421, 111, 0, 40, "MathCaption",ExpressionUUID->"12595416-9ed0-42ad-8a5f-388f49eee419"],
Cell[CellGroupData[{
Cell[3992431, 114425, 214, 6, 27, "Input",ExpressionUUID->"0159f067-e530-4559-82ff-3bfb13647309"],
Cell[3992648, 114433, 1103, 31, 32, "Output",ExpressionUUID->"460ebd2a-2960-469c-8f16-06c56394b1a7",
 CellID->941848211]
}, Open  ]],
Cell[CellGroupData[{
Cell[3993788, 114469, 194, 5, 27, "Input",ExpressionUUID->"27e1cf07-c6d2-4a13-9626-95d2a9200886"],
Cell[3993985, 114476, 3207, 93, 32, "Output",ExpressionUUID->"e3462d92-79de-4bae-9678-0ba597fee1ff",
 CellID->540920240]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[3997241, 114575, 138, 1, 40, "MathCaption",ExpressionUUID->"7f56dd22-9685-466f-bfed-dd8346cef492"],
Cell[CellGroupData[{
Cell[3997404, 114580, 232, 7, 27, "Input",ExpressionUUID->"eaa2b14a-2538-4c57-9c19-7ce0bdd587a2"],
Cell[3997639, 114589, 1183, 34, 32, "Output",ExpressionUUID->"2c969b03-01ba-449c-880b-df3f4d6d3e93",
 CellID->948195929]
}, Open  ]],
Cell[CellGroupData[{
Cell[3998859, 114628, 177, 4, 27, "Input",ExpressionUUID->"dfce37a2-0f15-477f-b09b-efc7982b7296"],
Cell[3999039, 114634, 3197, 92, 32, "Output",ExpressionUUID->"1f519fc4-ee54-4fec-847a-7ee84ef8a39c",
 CellID->2000317963]
}, Open  ]],
Cell[CellGroupData[{
Cell[4002273, 114731, 149, 3, 27, "Input",ExpressionUUID->"3aa82a23-30d0-4096-b111-ac9e464dbfaa"],
Cell[4002425, 114736, 7510, 214, 32, "Output",ExpressionUUID->"0aeae5c6-5e16-4a16-a8a0-5662b116786e",
 CellID->487510356]
}, Open  ]],
Cell[CellGroupData[{
Cell[4009972, 114955, 138, 3, 27, "Input",ExpressionUUID->"decc9b8c-f666-42de-a36e-bc1c58e7c239"],
Cell[4010113, 114960, 12346, 361, 32, "Output",ExpressionUUID->"5bbcc8bd-cb5a-4290-9bb7-5279a0b406e2",
 CellID->936439987]
}, Open  ]],
Cell[CellGroupData[{
Cell[4022496, 115326, 154, 3, 27, "Input",ExpressionUUID->"3e2eb936-7586-4fff-a4bb-6e0d30b15bf3"],
Cell[4022653, 115331, 16594, 472, 32, "Output",ExpressionUUID->"ca9f73eb-f317-49db-b82d-36d4d510b454",
 CellID->1986696293]
}, Open  ]],
Cell[CellGroupData[{
Cell[4039284, 115808, 146, 3, 27, "Input",ExpressionUUID->"4e9b2857-6b00-40cd-af91-c8d42ea2734a"],
Cell[4039433, 115813, 7217, 210, 32, "Output",ExpressionUUID->"1740eb1f-ea55-4499-9000-d5cce8d78255",
 CellID->1328852895]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4046699, 116029, 111, 0, 40, "MathCaption",ExpressionUUID->"92408cf4-dc70-4dd7-998b-c59937e11a7d"],
Cell[CellGroupData[{
Cell[4046835, 116033, 232, 7, 27, "Input",ExpressionUUID->"f2acbc6b-f57b-42ab-8ab1-45c8709e2e6e"],
Cell[4047070, 116042, 1162, 34, 32, "Output",ExpressionUUID->"c7a5ead0-8f49-43e4-b050-2d01b6f887b4",
 CellID->200677465]
}, Open  ]],
Cell[CellGroupData[{
Cell[4048269, 116081, 153, 3, 27, "Input",ExpressionUUID->"ab708605-2164-43c6-90ac-818153d8ad65"],
Cell[4048425, 116086, 1552, 45, 35, "Output",ExpressionUUID->"c8fb3416-9e89-4c84-ada1-4418d66f7c1e",
 CellID->1236828465]
}, Open  ]],
Cell[CellGroupData[{
Cell[4050014, 116136, 222, 5, 27, "Input",ExpressionUUID->"ea9afe11-c509-4995-a680-8c19d7b80d17"],
Cell[4050239, 116143, 3199, 92, 32, "Output",ExpressionUUID->"d073bbbd-bcb4-4238-80a5-e40d6c6403b1",
 CellID->1881791942]
}, Open  ]]
}, Open  ]],
Cell[4053465, 116239, 1864, 49, 109, "DefinitionBox",ExpressionUUID->"4322fa9b-6a43-4874-9206-b4005df2c599"],
Cell[4055332, 116290, 91, 0, 23, "Text",ExpressionUUID->"db1547c8-31dc-4aea-8a41-2df6422a3e64"],
Cell[4055426, 116292, 219, 3, 23, "Text",ExpressionUUID->"946d0d03-0149-416f-8966-05871a3a261c"],
Cell[CellGroupData[{
Cell[4055670, 116299, 233, 7, 40, "MathCaption",ExpressionUUID->"d148de44-461f-4717-bd5f-f12c4cba85dd"],
Cell[CellGroupData[{
Cell[4055928, 116310, 261, 8, 27, "Input",ExpressionUUID->"c6657f47-fb91-4433-9445-80986c68d1d5"],
Cell[4056192, 116320, 4365, 128, 32, "Output",ExpressionUUID->"c84a2a41-4bd8-4041-aa2a-a164de0af840",
 CellID->1285499825]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4060606, 116454, 201, 3, 40, "MathCaption",ExpressionUUID->"d7175ee7-f7d7-4c16-b327-c25df842b190"],
Cell[CellGroupData[{
Cell[4060832, 116461, 261, 8, 27, "Input",ExpressionUUID->"3114835a-0d06-49fc-a7cf-a9e16e0991e5"],
Cell[4061096, 116471, 1188, 19, 28, "Message",ExpressionUUID->"96db6553-ffb6-4294-9229-1b684546f538",
 CellID->1710069955],
Cell[4062287, 116492, 1188, 19, 28, "Message",ExpressionUUID->"8d8ed186-18fb-4d3e-a262-1979f6232106",
 CellID->1372085884],
Cell[4063478, 116513, 2475, 71, 32, "Output",ExpressionUUID->"c7b4635d-ea1d-4272-821e-58a4e30f8e4a",
 CellID->968050902]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4066002, 116590, 164, 2, 40, "MathCaption",ExpressionUUID->"e928471c-6152-43bc-8bf0-1cf436fc9ce6"],
Cell[CellGroupData[{
Cell[4066191, 116596, 261, 8, 27, "Input",ExpressionUUID->"db506d0d-6d24-43af-9579-5bf3f40a4829"],
Cell[4066455, 116606, 4366, 128, 32, "Output",ExpressionUUID->"01557f69-a661-4886-b6dc-50fd2ab8cc39",
 CellID->695400242]
}, Open  ]],
Cell[CellGroupData[{
Cell[4070858, 116739, 321, 10, 27, "Input",ExpressionUUID->"4f728459-655d-443d-9318-ad4f7089426d"],
Cell[4071182, 116751, 7885, 223, 38, "Output",ExpressionUUID->"e9cd8496-84df-49b7-8255-2c42c1b42c84",
 CellID->1956173085]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4079116, 116980, 242, 6, 40, "MathCaption",ExpressionUUID->"7282b404-affb-4801-ad7f-a887503eaa0a"],
Cell[CellGroupData[{
Cell[4079383, 116990, 261, 8, 27, "Input",ExpressionUUID->"55ceed0f-2c24-407b-8815-916119b2ccba"],
Cell[4079647, 117000, 1188, 19, 28, "Message",ExpressionUUID->"275cf7c8-88fc-4b29-8b8c-813a79675abb",
 CellID->1573887672],
Cell[4080838, 117021, 1188, 19, 28, "Message",ExpressionUUID->"5b022a4c-1cce-40ab-a69c-166b9ab8a913",
 CellID->1774023205],
Cell[4082029, 117042, 2473, 71, 32, "Output",ExpressionUUID->"e57c3cb7-377e-4c6b-b189-ff14197f0983",
 CellID->345051487]
}, Open  ]]
}, Open  ]],
Cell[4084529, 117117, 282, 4, 40, "Text",ExpressionUUID->"a891101e-2d10-44a9-b3d6-bfcc0f46cbc6"],
Cell[CellGroupData[{
Cell[4084836, 117125, 191, 3, 40, "MathCaption",ExpressionUUID->"81eb0886-f0da-4930-9bdb-df4053db42c1"],
Cell[CellGroupData[{
Cell[4085052, 117132, 350, 11, 27, "Input",ExpressionUUID->"48277dae-31b8-4156-8767-640436af5c7c"],
Cell[4085405, 117145, 1072, 31, 32, "Output",ExpressionUUID->"bb409634-bbb4-4dff-b04f-36bb01a16c0e",
 CellID->1074495109]
}, Open  ]]
}, Open  ]],
Cell[4086504, 117180, 277, 4, 56, "MathCaption",ExpressionUUID->"950800d7-4e74-4710-ad96-19eef081d217"],
Cell[CellGroupData[{
Cell[4086806, 117188, 134, 0, 40, "MathCaption",ExpressionUUID->"4d9e887a-4901-4234-8d9d-862fff0af6bb"],
Cell[CellGroupData[{
Cell[4086965, 117192, 283, 9, 27, "Input",ExpressionUUID->"8140c199-4bce-446e-a4ae-cc31b48104f4"],
Cell[4087251, 117203, 170, 3, 26, "Output",ExpressionUUID->"05ec780b-4916-4d04-a134-98be9d736481",
 CellID->849309340]
}, Open  ]],
Cell[CellGroupData[{
Cell[4087458, 117211, 590, 20, 27, "Input",ExpressionUUID->"ba163dd5-a531-4d93-a30f-2a5ffc97c27c"],
Cell[4088051, 117233, 2379, 70, 32, "Output",ExpressionUUID->"0dd05c87-6e2d-4b25-8fba-06775d81e40b",
 CellID->18275771]
}, Open  ]],
Cell[CellGroupData[{
Cell[4090467, 117308, 175, 4, 27, "Input",ExpressionUUID->"199a1aa8-ad00-4382-af03-d68d92f594e2"],
Cell[4090645, 117314, 2150, 62, 32, "Output",ExpressionUUID->"82a9a4eb-ef28-4fec-b85d-a836d027a0e7",
 CellID->70721676]
}, Open  ]],
Cell[CellGroupData[{
Cell[4092832, 117381, 475, 15, 27, "Input",ExpressionUUID->"a14ae639-205d-4a2d-83cf-05feab8103ee"],
Cell[4093310, 117398, 236, 5, 26, "Output",ExpressionUUID->"0288b88f-f9c7-45e2-bf36-11987136fa5a",
 CellID->1060751741]
}, Open  ]],
Cell[CellGroupData[{
Cell[4093583, 117408, 245, 7, 27, "Input",ExpressionUUID->"08d7337c-245f-4314-94fd-a904e2f6296d"],
Cell[4093831, 117417, 1125, 32, 32, "Output",ExpressionUUID->"697419b9-bdb4-4928-922e-0877fa4d7706",
 CellID->1792757240]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4095005, 117455, 134, 0, 40, "MathCaption",ExpressionUUID->"850fa9c0-f74a-417a-9a2c-ddcde079cf18"],
Cell[CellGroupData[{
Cell[4095164, 117459, 620, 21, 27, "Input",ExpressionUUID->"d11bc6bf-d718-418c-a04f-7502f3d9af6e"],
Cell[4095787, 117482, 2649, 77, 32, "Output",ExpressionUUID->"44b953fe-a864-4224-8f17-cbbbe761814b",
 CellID->1675365270]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4098485, 117565, 111, 0, 40, "MathCaption",ExpressionUUID->"8aa58e4c-9317-4668-ace2-d2dfaf41ca8f"],
Cell[CellGroupData[{
Cell[4098621, 117569, 326, 10, 27, "Input",ExpressionUUID->"a4d6aa28-a724-4952-a127-cd1bae6267dd"],
Cell[4098950, 117581, 3124, 90, 32, "Output",ExpressionUUID->"1e3dc39a-5b5d-4702-926d-26df56036304",
 CellID->58513426]
}, Open  ]],
Cell[CellGroupData[{
Cell[4102111, 117676, 428, 13, 27, "Input",ExpressionUUID->"14b47a10-48cc-4fea-9acb-31a89a18b885"],
Cell[4102542, 117691, 171, 3, 26, "Output",ExpressionUUID->"c003739a-458c-4167-a07c-228489860ac6",
 CellID->2035124103]
}, Open  ]],
Cell[CellGroupData[{
Cell[4102750, 117699, 633, 20, 47, "Input",ExpressionUUID->"043a5706-7642-4ae1-bbce-f2616fd93366"],
Cell[4103386, 117721, 1104, 31, 32, "Output",ExpressionUUID->"e2aa869b-8579-4faf-80f7-80ca6182a5cc",
 CellID->1358753352]
}, Open  ]],
Cell[CellGroupData[{
Cell[4104527, 117757, 598, 19, 47, "Input",ExpressionUUID->"ada00239-b41a-4fa3-af0b-0398a9c50ec1"],
Cell[4105128, 117778, 1105, 31, 32, "Output",ExpressionUUID->"cf1e9294-1c2d-4b49-b1b0-aba9ec088424",
 CellID->226484866]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4106282, 117815, 128, 0, 40, "MathCaption",ExpressionUUID->"d22d1995-fc26-403b-9bd3-4f5c7890c674"],
Cell[CellGroupData[{
Cell[4106435, 117819, 147, 3, 27, "Input",ExpressionUUID->"ab387ecb-66de-455b-accd-7050bad5053e"],
Cell[4106585, 117824, 549, 12, 22, "Print",ExpressionUUID->"efbc8fbc-0353-45cb-9a71-f67c886d8b47",
 CellID->1190120378]
}, Open  ]],
Cell[CellGroupData[{
Cell[4107171, 117841, 491, 14, 27, "Input",ExpressionUUID->"56e5d15a-38f1-48ad-9e8f-21b41269a27c"],
Cell[4107665, 117857, 625, 12, 22, "Print",ExpressionUUID->"b4d38a62-844d-4571-993c-00cdacbe2df0",
 CellID->1437879317]
}, Open  ]],
Cell[CellGroupData[{
Cell[4108327, 117874, 232, 7, 27, "Input",ExpressionUUID->"be53ac32-2d78-4b0a-b58e-8a217455fbe3"],
Cell[4108562, 117883, 1181, 34, 32, "Output",ExpressionUUID->"2832be64-b87a-4708-887c-98f5a3cc3def",
 CellID->804201128]
}, Open  ]],
Cell[CellGroupData[{
Cell[4109780, 117922, 177, 4, 27, "Input",ExpressionUUID->"14bff954-44b8-4863-a887-4dee9f4c02db"],
Cell[4109960, 117928, 3197, 92, 32, "Output",ExpressionUUID->"b9b6e115-b600-4444-8785-b693694f9264",
 CellID->2093767725]
}, Open  ]],
Cell[CellGroupData[{
Cell[4113194, 118025, 149, 3, 27, "Input",ExpressionUUID->"20debca3-3216-434f-a4b7-ff313f802198"],
Cell[4113346, 118030, 7509, 214, 32, "Output",ExpressionUUID->"98b3dad1-1191-4cfa-ac25-3d3e11a77879",
 CellID->1287114333]
}, Open  ]],
Cell[CellGroupData[{
Cell[4120892, 118249, 138, 3, 27, "Input",ExpressionUUID->"e8364c72-27e6-4aa6-9bf1-2d21d072ff40"],
Cell[4121033, 118254, 12344, 361, 32, "Output",ExpressionUUID->"57fc4f71-7a64-475f-85b0-e746b6dafce4",
 CellID->927818500]
}, Open  ]],
Cell[CellGroupData[{
Cell[4133414, 118620, 154, 3, 27, "Input",ExpressionUUID->"44140647-b549-4927-a788-e5c95e5aeac2"],
Cell[4133571, 118625, 16592, 472, 32, "Output",ExpressionUUID->"fcc022c9-1619-4a9d-92c7-fd3201c0223f",
 CellID->966768633]
}, Open  ]],
Cell[CellGroupData[{
Cell[4150200, 119102, 146, 3, 27, "Input",ExpressionUUID->"471b8453-739d-4d1d-8ee3-b9047d10a47c"],
Cell[4150349, 119107, 7216, 210, 32, "Output",ExpressionUUID->"9ed3f6d9-6f46-40dc-91d4-15f4c59c06a2",
 CellID->619141629]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4157614, 119323, 130, 0, 40, "MathCaption",ExpressionUUID->"d0a981e5-fbec-4c2e-9f51-428276d2802d"],
Cell[CellGroupData[{
Cell[4157769, 119327, 988, 35, 47, "Input",ExpressionUUID->"caf3067e-e44b-49b9-bc94-bfc4a1b32c2e"],
Cell[4158760, 119364, 16199, 473, 32, "Output",ExpressionUUID->"54896bc4-5806-4777-9cfb-9f0a0203e089",
 CellID->1355088888]
}, Open  ]],
Cell[CellGroupData[{
Cell[4174996, 119842, 154, 3, 27, "Input",ExpressionUUID->"f93f740a-c394-4435-9d5f-b5dab002160d"],
Cell[4175153, 119847, 18298, 528, 32, "Output",ExpressionUUID->"4b292053-543a-4e72-b100-d51969b8208f",
 CellID->914864181]
}, Open  ]],
Cell[CellGroupData[{
Cell[4193488, 120380, 146, 3, 27, "Input",ExpressionUUID->"6458fb8b-c5cf-4ebf-90e1-635d2c49b1dc"],
Cell[4193637, 120385, 170, 3, 26, "Output",ExpressionUUID->"9de7627e-4ce3-415b-b76e-1b7d2e76777f",
 CellID->139377547]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4193856, 120394, 97, 0, 40, "MathCaption",ExpressionUUID->"8b602bf8-def1-45b7-b908-b9f8c75cfa6e"],
Cell[CellGroupData[{
Cell[4193978, 120398, 633, 21, 27, "Input",ExpressionUUID->"248dc607-74e5-40e9-a62d-f5778d92a251"],
Cell[4194614, 120421, 5613, 160, 32, "Output",ExpressionUUID->"1c086fd5-2497-4744-9f5e-01e3ae0bdd41",
 CellID->1761257199]
}, Open  ]],
Cell[CellGroupData[{
Cell[4200264, 120586, 192, 5, 27, "Input",ExpressionUUID->"7406f188-b98c-4bd2-a4f3-2d9bddf99067"],
Cell[4200459, 120593, 6481, 182, 36, "Output",ExpressionUUID->"6864c629-81d4-4148-a346-e50c2b42c955",
 CellID->535153345]
}, Open  ]],
Cell[CellGroupData[{
Cell[4206977, 120780, 222, 5, 27, "Input",ExpressionUUID->"fdbb5bf5-7f56-4f40-a962-cd9b6e9a74e8"],
Cell[4207202, 120787, 10829, 307, 32, "Output",ExpressionUUID->"8195bff4-c067-41d3-b928-2740e71784e7",
 CellID->660755779]
}, Open  ]],
Cell[CellGroupData[{
Cell[4218068, 121099, 178, 4, 27, "Input",ExpressionUUID->"f7c2aa06-7d86-4345-9dcc-c34718f76cc6"],
Cell[4218249, 121105, 82220, 2399, 104, "Output",ExpressionUUID->"600ffc97-240e-4a07-94c8-57af64c48eca",
 CellID->684439016]
}, Open  ]],
Cell[CellGroupData[{
Cell[4300506, 123509, 183, 4, 27, "Input",ExpressionUUID->"23aebe17-2c0b-42da-bca5-761c36d5ad3f"],
Cell[4300692, 123515, 10141, 292, 32, "Output",ExpressionUUID->"35f4935b-ad34-4eb6-ba69-f4a47f515fb2",
 CellID->1444128580]
}, Open  ]],
Cell[CellGroupData[{
Cell[4310870, 123812, 146, 3, 27, "Input",ExpressionUUID->"b245d85c-bbb7-474c-bcac-110695affdbb"],
Cell[4311019, 123817, 10141, 292, 32, "Output",ExpressionUUID->"c9bf446e-f9a2-4658-8abd-af0686cb598d",
 CellID->608865666]
}, Open  ]],
Cell[CellGroupData[{
Cell[4321197, 124114, 146, 3, 27, "Input",ExpressionUUID->"f38a1752-e389-497d-b365-d8adb2e5147c"],
Cell[4321346, 124119, 10190, 293, 32, "Output",ExpressionUUID->"18d99ee8-398d-4e5d-9603-83e37a483984",
 CellID->445670507]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4331585, 124418, 107, 0, 40, "MathCaption",ExpressionUUID->"9220c3f2-233a-4b85-9803-cd7b301b6632"],
Cell[CellGroupData[{
Cell[4331717, 124422, 655, 21, 27, "Input",ExpressionUUID->"dcf0b184-8de3-410c-90d2-7874b6184a6c"],
Cell[4332375, 124445, 3360, 94, 35, "Output",ExpressionUUID->"2268dc17-35ff-439a-9e64-65bd85edcd50",
 CellID->2108734060]
}, Open  ]],
Cell[CellGroupData[{
Cell[4335772, 124544, 173, 4, 27, "Input",ExpressionUUID->"dbc1f46f-8a73-4c6c-a166-68693b251994"],
Cell[4335948, 124550, 2682, 74, 35, "Output",ExpressionUUID->"4644d6f7-e3a2-4c4f-a799-ad0d6de4a28d",
 CellID->1570719979]
}, Open  ]],
Cell[CellGroupData[{
Cell[4338667, 124629, 304, 8, 27, "Input",ExpressionUUID->"a1cac09b-2564-4338-b395-7d316d45c283"],
Cell[4338974, 124639, 65839, 1918, 104, "Output",ExpressionUUID->"73396972-fe19-4219-9e74-083bce127dc5",
 CellID->74214482]
}, Open  ]],
Cell[CellGroupData[{
Cell[4404850, 126562, 224, 6, 27, "Input",ExpressionUUID->"cc6884e1-534c-4fca-bd51-b2cb26187a31"],
Cell[4405077, 126570, 31535, 881, 56, "Output",ExpressionUUID->"1e789b15-d357-4bff-b704-3c8bc61b339a",
 CellID->1782283714]
}, Open  ]],
Cell[CellGroupData[{
Cell[4436649, 127456, 595, 20, 27, "Input",ExpressionUUID->"b6113762-d2e2-4e83-ada5-5f8ea3eaaf75"],
Cell[4437247, 127478, 2498, 73, 32, "Output",ExpressionUUID->"94032fe5-e022-433b-833c-bcbb8e750729",
 CellID->1740472842]
}, Open  ]],
Cell[CellGroupData[{
Cell[4439782, 127556, 173, 4, 27, "Input",ExpressionUUID->"6c891b48-cc0d-4cd4-8be2-aacc4b290ba7"],
Cell[4439958, 127562, 2150, 62, 32, "Output",ExpressionUUID->"b096926c-6b6a-480e-a9ae-7ed10abdc6f1",
 CellID->1252299707]
}, Open  ]],
Cell[CellGroupData[{
Cell[4442145, 127629, 196, 5, 27, "Input",ExpressionUUID->"a752487c-6e52-4597-89fb-1799f4c6faa3"],
Cell[4442344, 127636, 4273, 118, 36, "Output",ExpressionUUID->"5871785a-2877-4cbe-bf01-4afff8b66e6b",
 CellID->2061434949]
}, Open  ]],
Cell[CellGroupData[{
Cell[4446654, 127759, 300, 8, 27, "Input",ExpressionUUID->"6944dd13-2689-4769-95eb-100ac716eb27"],
Cell[4446957, 127769, 191547, 5588, 272, "Output",ExpressionUUID->"19b59072-2cf9-4d6c-a944-cd5f406eb110",
 CellID->1183510792]
}, Open  ]],
Cell[CellGroupData[{
Cell[4638541, 133362, 194, 5, 27, "Input",ExpressionUUID->"63e4d41f-c2b0-42bc-b5f6-2c63a409eaca"],
Cell[4638738, 133369, 61671, 1790, 104, "Output",ExpressionUUID->"839dc0b1-4ffe-4d6c-8729-80fbe5e964a4",
 CellID->1272342630]
}, Open  ]],
Cell[CellGroupData[{
Cell[4700446, 135164, 174, 4, 27, "Input",ExpressionUUID->"91e81c5f-0bb1-405b-9e00-462723c0f8d9"],
Cell[4700623, 135170, 47708, 1385, 80, "Output",ExpressionUUID->"c379fca7-cbd8-4337-a185-817e58bda7ad",
 CellID->1767909231]
}, Open  ]],
Cell[CellGroupData[{
Cell[4748368, 136560, 290, 9, 27, "Input",ExpressionUUID->"1aa74944-89ad-49cc-84bd-81332f9338e1"],
Cell[4748661, 136571, 9423, 254, 35, "Output",ExpressionUUID->"4d53ca9d-5bc7-41c3-86f4-c6de129f9dc0",
 CellID->666705525]
}, Open  ]],
Cell[CellGroupData[{
Cell[4758121, 136830, 553, 17, 27, "Input",ExpressionUUID->"3036fcf4-7f5f-4abf-a39c-d9435c169b21"],
Cell[4758677, 136849, 6713, 189, 35, "Output",ExpressionUUID->"0b230a1b-cdd5-4e56-b334-ec3376f688cc",
 CellID->1897572276]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4765439, 137044, 131, 0, 40, "MathCaption",ExpressionUUID->"8fbe4d8a-ea32-4de8-af7e-e56fd978f972"],
Cell[CellGroupData[{
Cell[4765595, 137048, 649, 21, 47, "Input",ExpressionUUID->"d6f4ccee-a446-45d4-9ed9-f92d9629cd8a"],
Cell[4766247, 137071, 38865, 1073, 63, "Output",ExpressionUUID->"18782321-8445-4df4-b53e-ab3896cb8fb0",
 CellID->217601135]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4805161, 138150, 103, 0, 40, "MathCaption",ExpressionUUID->"645e61a7-dbab-4821-95bb-7b3edbc299f3"],
Cell[CellGroupData[{
Cell[4805289, 138154, 709, 23, 47, "Input",ExpressionUUID->"d75d2ee6-dbae-4efd-b25b-a1532a39fdef"],
Cell[4806001, 138179, 42519, 1177, 63, "Output",ExpressionUUID->"da01718d-78b9-4941-8372-6bc3580912c7",
 CellID->57625104]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4848569, 139362, 108, 0, 40, "MathCaption",ExpressionUUID->"1c776f3b-7b53-4eca-b60a-644aac1ef447"],
Cell[CellGroupData[{
Cell[4848702, 139366, 655, 21, 27, "Input",ExpressionUUID->"4b104a24-2a7e-4dad-b004-137529e445ee"],
Cell[4849360, 139389, 3360, 94, 35, "Output",ExpressionUUID->"50d7b4af-6a42-46e2-950f-541ae2b9c886",
 CellID->2137481778]
}, Open  ]],
Cell[CellGroupData[{
Cell[4852757, 139488, 173, 4, 27, "Input",ExpressionUUID->"805cb987-ad32-419a-ae7a-15df50792d9e"],
Cell[4852933, 139494, 2684, 74, 35, "Output",ExpressionUUID->"cb407dc8-0856-41fd-947b-db69c37224da",
 CellID->1549795830]
}, Open  ]],
Cell[CellGroupData[{
Cell[4855654, 139573, 304, 8, 27, "Input",ExpressionUUID->"e868a243-edcf-4744-b1e8-c5dd302bff28"],
Cell[4855961, 139583, 65840, 1918, 104, "Output",ExpressionUUID->"885449f8-4baa-45ff-b8f6-b71a7d27d26b",
 CellID->589360347]
}, Open  ]],
Cell[CellGroupData[{
Cell[4921838, 141506, 312, 8, 27, "Input",ExpressionUUID->"bfae4782-47c8-4d6d-952d-e10fdfe44336"],
Cell[4922153, 141516, 13982, 393, 32, "Output",ExpressionUUID->"9827c46f-3188-4207-a734-290db5e7f5c6",
 CellID->1184772065]
}, Open  ]],
Cell[CellGroupData[{
Cell[4936172, 141914, 554, 18, 27, "Input",ExpressionUUID->"91aaa073-3887-48a7-8c8c-324638921de9"],
Cell[4936729, 141934, 2328, 66, 32, "Output",ExpressionUUID->"12ce8edd-aaf0-4708-955a-54644109a555",
 CellID->995721471]
}, Open  ]],
Cell[CellGroupData[{
Cell[4939094, 142005, 154, 3, 27, "Input",ExpressionUUID->"0b8acb87-30a2-4df3-88c2-5e2477681d71"],
Cell[4939251, 142010, 3169, 86, 35, "Output",ExpressionUUID->"be552039-825c-4c23-81c3-36981e7497cb",
 CellID->1898185673]
}, Open  ]],
Cell[CellGroupData[{
Cell[4942457, 142101, 300, 8, 27, "Input",ExpressionUUID->"db15b3fd-0e16-4b4a-a8cb-ba593513392c"],
Cell[4942760, 142111, 65112, 1887, 104, "Output",ExpressionUUID->"b9e540f0-a118-4405-b31a-1690f77f8a0c",
 CellID->1937029926]
}, Open  ]],
Cell[CellGroupData[{
Cell[5007909, 144003, 263, 7, 27, "Input",ExpressionUUID->"7a03111b-313c-46a2-b426-c5d6e336e722"],
Cell[5008175, 144012, 13982, 393, 32, "Output",ExpressionUUID->"57325239-b282-4b4a-94f2-4220e0008aaf",
 CellID->1957361409]
}, Open  ]],
Cell[CellGroupData[{
Cell[5022194, 144410, 498, 16, 27, "Input",ExpressionUUID->"6402e71d-ae9f-478c-8298-87614c7c2ebf"],
Cell[5022695, 144428, 4996, 138, 35, "Output",ExpressionUUID->"b580a972-fd42-43ac-8727-68a57a3ead0a",
 CellID->71212670]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5027740, 144572, 126, 0, 40, "MathCaption",ExpressionUUID->"375ee021-f39e-431d-a9bb-979c3293d949"],
Cell[CellGroupData[{
Cell[5027891, 144576, 1442, 49, 66, "Input",ExpressionUUID->"20b2587b-c994-4104-81a2-dba51d7934cd"],
Cell[5029336, 144627, 45171, 1273, 88, "Output",ExpressionUUID->"79dc49ad-e02f-4b69-a639-e3b5181cf9d2",
 CellID->722953311]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5074556, 145906, 111, 0, 40, "MathCaption",ExpressionUUID->"6525bcad-7378-4aeb-b6b3-470375a630c7"],
Cell[CellGroupData[{
Cell[5074692, 145910, 765, 25, 27, "Input",ExpressionUUID->"96303464-b6c1-4189-86ea-c82f7583a860"],
Cell[5075460, 145937, 4467, 119, 35, "Output",ExpressionUUID->"82604d8c-2f0c-45f8-8797-6e9c3ae6fb53",
 CellID->1052526960]
}, Open  ]],
Cell[CellGroupData[{
Cell[5079964, 146061, 173, 4, 27, "Input",ExpressionUUID->"7b019246-29ce-4837-9166-bacd9484ccf1"],
Cell[5080140, 146067, 3735, 103, 35, "Output",ExpressionUUID->"b661502b-c350-47cb-a9c0-a027d7966718",
 CellID->997816381]
}, Open  ]],
Cell[CellGroupData[{
Cell[5083912, 146175, 304, 8, 27, "Input",ExpressionUUID->"e04feeb5-d70a-4450-bb74-9eccfa59523d"],
Cell[5084219, 146185, 33163, 965, 56, "Output",ExpressionUUID->"aa650f7e-1c5c-4bab-9b42-15e032514421",
 CellID->600116044]
}, Open  ]],
Cell[CellGroupData[{
Cell[5117419, 147155, 312, 8, 27, "Input",ExpressionUUID->"32e52a9e-1d49-45a5-9d92-f879f1612d17"],
Cell[5117734, 147165, 13422, 387, 32, "Output",ExpressionUUID->"ded4a535-a314-41fc-b461-1b21c0d1a03e",
 CellID->1567145196]
}, Open  ]],
Cell[CellGroupData[{
Cell[5131193, 147557, 634, 20, 27, "Input",ExpressionUUID->"753031c4-2bc6-404e-b001-f60fbf387dba"],
Cell[5131830, 147579, 3998, 109, 35, "Output",ExpressionUUID->"0488b665-bbb1-4c04-8f03-fa978bcdac56",
 CellID->122378717]
}, Open  ]],
Cell[CellGroupData[{
Cell[5135865, 147693, 154, 3, 27, "Input",ExpressionUUID->"4ebd329d-6c03-4b00-8036-b7bd2a88c2f2"],
Cell[5136022, 147698, 4371, 119, 35, "Output",ExpressionUUID->"53485bfa-2a07-433e-bebd-49cf48a8da05",
 CellID->1677246440]
}, Open  ]],
Cell[CellGroupData[{
Cell[5140430, 147822, 300, 8, 27, "Input",ExpressionUUID->"074bd6ad-4f49-4582-9078-e6e4f828460c"],
Cell[5140733, 147832, 28722, 834, 56, "Output",ExpressionUUID->"aee152d4-dbe4-42e7-8c4a-f8eba00e148a",
 CellID->1610566558]
}, Open  ]],
Cell[CellGroupData[{
Cell[5169492, 148671, 263, 7, 27, "Input",ExpressionUUID->"fa8dec79-8466-47bb-b31d-909f0b5371d4"],
Cell[5169758, 148680, 8683, 244, 32, "Output",ExpressionUUID->"b17c98d9-7aa0-4f70-a827-1ebf6d3e5da8",
 CellID->1908253518]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5178490, 148930, 170, 3, 40, "MathCaption",ExpressionUUID->"259a9919-b12f-4750-8423-899d75746bdd"],
Cell[CellGroupData[{
Cell[5178685, 148937, 466, 15, 27, "Input",ExpressionUUID->"e9f7b50e-ed27-41ac-87b9-c08cc415d450"],
Cell[5179154, 148954, 11493, 318, 35, "Output",ExpressionUUID->"625e9764-3633-4f15-afd1-072706e35e0c",
 CellID->1496092728]
}, Open  ]],
Cell[CellGroupData[{
Cell[5190684, 149277, 881, 27, 66, "Input",ExpressionUUID->"35e8c90f-069d-46b0-88a1-72cc74bd0f7f"],
Cell[5191568, 149306, 2142, 61, 32, "Output",ExpressionUUID->"2dbba7fb-936a-4bb7-b13f-71a2ef8097e0",
 CellID->2091940420]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5193759, 149373, 170, 3, 40, "MathCaption",ExpressionUUID->"f88ae0c0-061e-4586-a235-d04b0845e738"],
Cell[CellGroupData[{
Cell[5193954, 149380, 1542, 53, 66, "Input",ExpressionUUID->"840ede90-4ef9-4736-be39-fb78275f734e"],
Cell[5195499, 149435, 84548, 2429, 133, "Output",ExpressionUUID->"a4bd2824-d186-4e16-9fe6-64069e4d19e0",
 CellID->640370256]
}, Open  ]],
Cell[CellGroupData[{
Cell[5280084, 151869, 563, 17, 27, "Input",ExpressionUUID->"8a8a986b-f900-4f98-ac31-a7fbad66c2b6"],
Cell[5280650, 151888, 93120, 2475, 162, "Output",ExpressionUUID->"e7a6ae8d-bfeb-4435-85c4-83d3c5d20467",
 CellID->1662337475]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5373819, 154369, 118, 0, 40, "MathCaption",ExpressionUUID->"26c2e768-af0c-4ffd-a7cd-51442496fcd3"],
Cell[CellGroupData[{
Cell[5373962, 154373, 343, 10, 27, "Input",ExpressionUUID->"1a16bcf7-26cd-463e-b415-406b32bb908a"],
Cell[5374308, 154385, 84550, 2429, 133, "Output",ExpressionUUID->"d0687f23-f08c-4a42-a85f-d1a76210c49c",
 CellID->143694855]
}, Open  ]],
Cell[CellGroupData[{
Cell[5458895, 156819, 223, 6, 27, "Input",ExpressionUUID->"e8996e60-e964-4e5c-97d0-31bfbbc57a1b"],
Cell[5459121, 156827, 171, 3, 26, "Output",ExpressionUUID->"39f0597e-9463-433e-bcb3-0d8461166511",
 CellID->1307808757]
}, Open  ]]
}, Open  ]],
Cell[5459319, 156834, 278, 4, 40, "Text",ExpressionUUID->"0d677ecc-eef6-4169-8761-01c3d82e6358"],
Cell[CellGroupData[{
Cell[5459622, 156842, 88, 0, 40, "MathCaption",ExpressionUUID->"840b16f2-0e78-47ab-a8d8-a0e22a1b7b15"],
Cell[CellGroupData[{
Cell[5459735, 156846, 514, 17, 27, "Input",ExpressionUUID->"ff79c8a5-9e4c-4026-a7cc-251824b75e33"],
Cell[5460252, 156865, 13465, 381, 35, "Output",ExpressionUUID->"62ac2f10-41d1-4b39-bfb4-c8a961642ebf",
 CellID->846651854]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5473766, 157252, 92, 0, 40, "MathCaption",ExpressionUUID->"f7140a1c-592a-475f-9030-08caaa2ee72b"],
Cell[CellGroupData[{
Cell[5473883, 157256, 624, 20, 27, "Input",ExpressionUUID->"d8304f9b-982f-4b8d-9313-c4e3e9742040"],
Cell[5474510, 157278, 38348, 1103, 85, "Output",ExpressionUUID->"d7da88f2-b68c-49fc-b571-c508739f3255",
 CellID->451065156]
}, Open  ]],
Cell[CellGroupData[{
Cell[5512895, 158386, 290, 8, 27, "Input",ExpressionUUID->"00cca68e-ee79-49ce-8deb-65efd0806cd8"],
Cell[5513188, 158396, 11880, 336, 35, "Output",ExpressionUUID->"c7675050-1c19-4a0c-94e8-3507999b7ee1",
 CellID->434802982]
}, Open  ]],
Cell[CellGroupData[{
Cell[5525105, 158737, 490, 15, 27, "Input",ExpressionUUID->"a5cad8a6-1ec9-4527-bb12-ed79fa8b5433"],
Cell[5525598, 158754, 4569, 129, 35, "Output",ExpressionUUID->"abad15aa-0c6d-4a26-a3ec-1a92432d2acc",
 CellID->1994351380]
}, Open  ]],
Cell[CellGroupData[{
Cell[5530204, 158888, 342, 11, 27, "Input",ExpressionUUID->"a6e399f1-7f72-4460-a806-18f8f0c9548e"],
Cell[5530549, 158901, 10222, 289, 35, "Output",ExpressionUUID->"ed928bdf-2d7c-494b-8f75-8cb478ce1afd",
 CellID->1138881284]
}, Open  ]],
Cell[CellGroupData[{
Cell[5540808, 159195, 513, 17, 27, "Input",ExpressionUUID->"46a818b0-d7a9-4919-a106-a859471f72e0"],
Cell[5541324, 159214, 20821, 589, 35, "Output",ExpressionUUID->"ee5c52bc-04ad-47d8-8b28-f78264bbdd1d",
 CellID->655994477]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5562194, 159809, 93, 0, 40, "MathCaption",ExpressionUUID->"81a4acd7-75b9-4195-9791-897e5adf300e"],
Cell[CellGroupData[{
Cell[5562312, 159813, 729, 23, 27, "Input",ExpressionUUID->"e550f6b2-7140-4743-a5b0-3e8b57d9c880"],
Cell[5563044, 159838, 16410, 474, 32, "Output",ExpressionUUID->"d913852b-3bcb-411b-b372-b09b18b61ff1",
 CellID->118131241]
}, Open  ]],
Cell[CellGroupData[{
Cell[5579491, 160317, 293, 8, 27, "Input",ExpressionUUID->"9fe55e4a-aeef-44d0-810b-0291e6fa41cf"],
Cell[5579787, 160327, 5320, 152, 32, "Output",ExpressionUUID->"081e2fb6-ea4c-4646-930b-4eb0cba683e8",
 CellID->1156877659]
}, Open  ]],
Cell[CellGroupData[{
Cell[5585144, 160484, 600, 19, 27, "Input",ExpressionUUID->"ecd40e86-b7d2-4136-980c-8c6491a8528a"],
Cell[5585747, 160505, 5635, 157, 35, "Output",ExpressionUUID->"257a8b54-46fe-4b66-bcfb-724b6bdc7f2a",
 CellID->325874875]
}, Open  ]],
Cell[CellGroupData[{
Cell[5591419, 160667, 294, 9, 27, "Input",ExpressionUUID->"4cfcf066-584f-4f95-a167-3940d0a0a937"],
Cell[5591716, 160678, 4793, 133, 35, "Output",ExpressionUUID->"bd8490b9-4563-4c80-a864-f295e1024b74",
 CellID->895724275]
}, Open  ]],
Cell[CellGroupData[{
Cell[5596546, 160816, 682, 23, 27, "Input",ExpressionUUID->"782af030-0090-4d47-9310-dad2d290545b"],
Cell[5597231, 160841, 22337, 641, 60, "Output",ExpressionUUID->"1fa67787-b2d9-4a13-b06c-b0661c467a86",
 CellID->577524455]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5619617, 161488, 92, 0, 40, "MathCaption",ExpressionUUID->"7cafe600-619f-44a2-876b-4e6ec3f56a8f"],
Cell[CellGroupData[{
Cell[5619734, 161492, 763, 24, 47, "Input",ExpressionUUID->"fe6df6d1-e6bd-443c-9d3b-f65adc023569"],
Cell[5620500, 161518, 15005, 424, 35, "Output",ExpressionUUID->"0c66d77e-1827-4ddb-b530-01fcb8faa468",
 CellID->1974914392]
}, Open  ]],
Cell[CellGroupData[{
Cell[5635542, 161947, 584, 20, 27, "Input",ExpressionUUID->"7bc9ed29-8571-4b80-bfbb-45a4758a1b7e"],
Cell[5636129, 161969, 36939, 1057, 60, "Output",ExpressionUUID->"1c656b1a-f937-4775-93e0-2267eaa9c6d0",
 CellID->878938312]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5673117, 163032, 104, 0, 40, "MathCaption",ExpressionUUID->"a2593739-009d-487e-b60e-cf1e284e7510"],
Cell[CellGroupData[{
Cell[5673246, 163036, 505, 16, 27, "Input",ExpressionUUID->"2dee5a99-82d4-4a12-985f-4cecd091e514"],
Cell[5673754, 163054, 17578, 506, 32, "Output",ExpressionUUID->"ddd35677-78e6-453e-a7dc-7a1eacb0ad5f",
 CellID->1923685718]
}, Open  ]],
Cell[CellGroupData[{
Cell[5691369, 163565, 302, 8, 27, "Input",ExpressionUUID->"73cbf227-43fc-44eb-b0e8-5b85dce15de3"],
Cell[5691674, 163575, 10669, 307, 32, "Output",ExpressionUUID->"f76f7d1f-7aab-4ac1-a779-3c72de569680",
 CellID->1019556841]
}, Open  ]],
Cell[CellGroupData[{
Cell[5702380, 163887, 188, 5, 27, "Input",ExpressionUUID->"97b21467-7a2b-4450-8256-6bd2d1fb2a61"],
Cell[5702571, 163894, 13759, 397, 32, "Output",ExpressionUUID->"bce6c191-e62e-44d0-a816-89b4e76a8fb6",
 CellID->919807274]
}, Open  ]],
Cell[CellGroupData[{
Cell[5716367, 164296, 154, 3, 27, "Input",ExpressionUUID->"4d210fe2-7625-4f00-9a5b-1461713404c8"],
Cell[5716524, 164301, 9682, 279, 32, "Output",ExpressionUUID->"e17638fa-bbfd-47b8-a41b-4403f005172a",
 CellID->1613563012]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5726255, 164586, 91, 0, 40, "MathCaption",ExpressionUUID->"f2d8ec63-aac1-4d5f-bcc1-18c7f50dd3ee"],
Cell[CellGroupData[{
Cell[5726371, 164590, 234, 6, 27, "Input",ExpressionUUID->"d5c5e2c7-392f-4f60-ba8b-cf34ffed42e1"],
Cell[5726608, 164598, 7093, 207, 45, "Output",ExpressionUUID->"42b2259e-b441-4ec8-b82a-fa8729aedd4d",
 CellID->730048368]
}, Open  ]],
Cell[CellGroupData[{
Cell[5733738, 164810, 153, 3, 27, "Input",ExpressionUUID->"cfe5a1b5-e3c7-4492-a842-5e13e20197d3"],
Cell[CellGroupData[{
Cell[5733916, 164817, 615, 12, 22, "Print",ExpressionUUID->"a1daf3cf-3275-4173-ad7f-84adbdc3224f",
 CellID->1348845665],
Cell[5734534, 164831, 601, 12, 22, "Print",ExpressionUUID->"dba1e8ac-8097-44f4-8195-fe4e04366015",
 CellID->1211713923],
Cell[5735138, 164845, 624, 13, 22, "Print",ExpressionUUID->"906a668e-52d3-494e-ae3e-8ed2024b41c8",
 CellID->688479360],
Cell[5735765, 164860, 590, 12, 22, "Print",ExpressionUUID->"351d2b2d-d690-484b-829e-32bc60adb435",
 CellID->630713484],
Cell[5736358, 164874, 616, 12, 22, "Print",ExpressionUUID->"c29efb97-ff77-417b-adbf-06401133cb5f",
 CellID->422243996],
Cell[5736977, 164888, 585, 12, 22, "Print",ExpressionUUID->"27749748-81c6-4a26-8ee4-4084490cb60a",
 CellID->1368871084],
Cell[5737565, 164902, 598, 12, 22, "Print",ExpressionUUID->"f3f0c654-6f8c-4d55-a21f-483df6573720",
 CellID->623998477],
Cell[5738166, 164916, 593, 12, 22, "Print",ExpressionUUID->"045fa821-87e5-4676-99d5-fd57823452ce",
 CellID->1379754638],
Cell[5738762, 164930, 585, 12, 22, "Print",ExpressionUUID->"5270c5ef-7dbc-4772-9d04-ab40434fa0d9",
 CellID->1007780560],
Cell[5739350, 164944, 582, 12, 22, "Print",ExpressionUUID->"40215329-6b9d-4264-bb1a-37b59a1bc047",
 CellID->564348031],
Cell[5739935, 164958, 583, 12, 22, "Print",ExpressionUUID->"ab8d08d8-3d03-4154-968f-15b41c661b6d",
 CellID->1709571865],
Cell[5740521, 164972, 583, 12, 22, "Print",ExpressionUUID->"ceccc465-1ce8-4751-9575-b98d34b54c89",
 CellID->1590621842],
Cell[5741107, 164986, 571, 12, 22, "Print",ExpressionUUID->"0bba2ed3-70e9-46ab-99bd-e345404720f9",
 CellID->1704860638],
Cell[5741681, 165000, 575, 12, 22, "Print",ExpressionUUID->"876235e2-6882-46c9-8519-4d4dfc1b8e36",
 CellID->1865924592],
Cell[5742259, 165014, 602, 12, 22, "Print",ExpressionUUID->"61029f9f-7446-4122-a09c-2591ad3da720",
 CellID->890920603],
Cell[5742864, 165028, 595, 12, 22, "Print",ExpressionUUID->"362872ce-b693-4c81-8bd8-658854257998",
 CellID->1446587737]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5743508, 165046, 210, 6, 27, "Input",ExpressionUUID->"00fea31e-a751-4b23-aa25-837a74582d36"],
Cell[CellGroupData[{
Cell[5743743, 165056, 551, 12, 22, "Print",ExpressionUUID->"99ee8363-bb64-424e-a10e-af7a08b073ae",
 CellID->1137728072],
Cell[5744297, 165070, 548, 12, 22, "Print",ExpressionUUID->"de4c7b3f-91b2-44f0-8157-deabe2c101ef",
 CellID->601313216]
}, Open  ]]
}, Open  ]],
Cell[5744872, 165086, 505, 9, 66, "Input",ExpressionUUID->"8c77d418-24a4-4ef1-b3dd-85ecd03e42ce"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[5745426, 165101, 109, 0, 39, "Subsection",ExpressionUUID->"d1713bc0-fc25-4242-bac3-eef50b2341d3"],
Cell[5745538, 165103, 215, 3, 23, "Text",ExpressionUUID->"034a42b1-5182-440f-9fcb-313eff8765db"],
Cell[CellGroupData[{
Cell[5745778, 165110, 114, 0, 40, "MathCaption",ExpressionUUID->"c994e0d0-b499-45d5-85b4-a2b781554c77"],
Cell[CellGroupData[{
Cell[5745917, 165114, 357, 11, 27, "Input",ExpressionUUID->"24d53a9b-4933-4e48-851c-9d43d5b4ab2e"],
Cell[5746277, 165127, 18062, 514, 32, "Output",ExpressionUUID->"d3bdb3ac-cb3e-47a3-97d2-ec0cfc1d4686",
 CellID->678200266]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5764388, 165647, 105, 0, 40, "MathCaption",ExpressionUUID->"38f28c85-ed91-4076-91e3-933b38272b6e"],
Cell[CellGroupData[{
Cell[5764518, 165651, 337, 10, 27, "Input",ExpressionUUID->"c806032d-0300-4624-a43f-b6dd0b482836"],
Cell[5764858, 165663, 4167, 118, 32, "Output",ExpressionUUID->"431bba4c-d5bf-44dc-b6d3-59ff628d554a",
 CellID->1816156033]
}, Open  ]],
Cell[CellGroupData[{
Cell[5769062, 165786, 337, 10, 27, "Input",ExpressionUUID->"03ea2de4-1d75-469c-b44c-d7041b914035"],
Cell[5769402, 165798, 1147, 32, 32, "Output",ExpressionUUID->"e77bbeba-5a2b-4e56-a360-206b475920f5",
 CellID->1949371820]
}, Open  ]],
Cell[CellGroupData[{
Cell[5770586, 165835, 337, 10, 27, "Input",ExpressionUUID->"0834c7a2-cc39-41d5-af02-f4c87280142e"],
Cell[5770926, 165847, 188, 4, 26, "Output",ExpressionUUID->"86c27222-5915-492d-9d05-d87703c9fc9d",
 CellID->1081660108]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[5771175, 165858, 111, 0, 39, "Subsection",ExpressionUUID->"0b31d9ad-ad3c-4de9-94e2-36a87c8f4864"],
Cell[5771289, 165860, 303, 4, 40, "Text",ExpressionUUID->"02d733ef-a4e6-4656-ab0a-774274031576"],
Cell[CellGroupData[{
Cell[5771617, 165868, 215, 3, 56, "MathCaption",ExpressionUUID->"1e53bbbf-b600-4454-9dd2-107d62ad823c"],
Cell[CellGroupData[{
Cell[5771857, 165875, 232, 7, 27, "Input",ExpressionUUID->"da42ec84-f369-4972-b054-adbc1c658512"],
Cell[5772092, 165884, 1183, 34, 32, "Output",ExpressionUUID->"90525cd4-c2ea-4827-b6dc-720f1ca6773f",
 CellID->882303186]
}, Open  ]],
Cell[CellGroupData[{
Cell[5773312, 165923, 143, 3, 27, "Input",ExpressionUUID->"439f5cd8-d7a9-45c9-8957-548c6d1c1986"],
Cell[5773458, 165928, 3159, 90, 32, "Output",ExpressionUUID->"c4b69e7b-7653-4b30-8ddd-7de29d24993e",
 CellID->495064567]
}, Open  ]],
Cell[CellGroupData[{
Cell[5776654, 166023, 145, 3, 27, "Input",ExpressionUUID->"34241b67-3055-47c1-b88d-04958f98dd06"],
Cell[5776802, 166028, 223, 3, 26, "Output",ExpressionUUID->"dcc9dc85-880f-4fc8-9318-626eb6e610ee",
 CellID->1198529091]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5777074, 166037, 137, 1, 40, "MathCaption",ExpressionUUID->"1ed5ddbd-95a4-47a4-acb1-1267bb68120e"],
Cell[CellGroupData[{
Cell[5777236, 166042, 144, 3, 27, "Input",ExpressionUUID->"3c51bc16-4e64-4166-b99a-c255451d6385"],
Cell[5777383, 166047, 4470, 128, 32, "Output",ExpressionUUID->"a15648d6-c9a1-4f7d-ba1d-0155b786b943",
 CellID->281823577]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5781902, 166181, 132, 0, 40, "MathCaption",ExpressionUUID->"cc955863-7460-4f02-938f-8cf4558a70ae"],
Cell[CellGroupData[{
Cell[5782059, 166185, 220, 5, 27, "Input",ExpressionUUID->"68337242-61d6-471a-b2d0-b084173455c3"],
Cell[5782282, 166192, 1182, 34, 32, "Output",ExpressionUUID->"f47ad45e-dda9-42b5-bd46-1f115bee5b92",
 CellID->1407417004]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[5783525, 166233, 106, 0, 39, "Subsection",ExpressionUUID->"5f03f11f-137e-4963-9f65-9721950bf5ba"],
Cell[5783634, 166235, 531, 9, 57, "Text",ExpressionUUID->"486b539d-d3b4-475a-8d82-9f2ea778f31b"],
Cell[CellGroupData[{
Cell[5784190, 166248, 151, 2, 40, "MathCaption",ExpressionUUID->"e54c0851-8430-4ba0-ae7a-05a854d4369e"],
Cell[CellGroupData[{
Cell[5784366, 166254, 337, 10, 27, "Input",ExpressionUUID->"3e609e06-a845-4f2f-9378-f8efdd464ebe"],
Cell[5784706, 166266, 144635, 4103, 285, "Output",ExpressionUUID->"6f5c5225-2fcb-4171-b3be-bb16a3eebaa6",
 CellID->1653966926]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5929390, 170375, 156, 2, 40, "MathCaption",ExpressionUUID->"3824b418-a4de-4e98-b575-ff636945cd69"],
Cell[CellGroupData[{
Cell[5929571, 170381, 229, 5, 27, "Input",ExpressionUUID->"f8d9d3aa-eff5-46c9-8477-43be4fe1c059"],
Cell[5929803, 170388, 590, 14, 46, "Output",ExpressionUUID->"cf58506c-23fd-4500-998a-07cb5058ef91",
 CellID->1193798514]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5930442, 170408, 231, 3, 56, "MathCaption",ExpressionUUID->"44621bfd-a70e-417f-8307-5c8813cfaf44"],
Cell[CellGroupData[{
Cell[5930698, 170415, 537, 16, 27, "Input",ExpressionUUID->"bbf44674-b511-4f6e-8b27-5c4fb3f07483"],
Cell[5931238, 170433, 223, 5, 26, "Output",ExpressionUUID->"08a14196-94f4-4b0d-8bb1-c77d208c39b6",
 CellID->231910877]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5931510, 170444, 121, 0, 40, "MathCaption",ExpressionUUID->"82a7904d-6421-4e39-acca-6c5f18839a60"],
Cell[CellGroupData[{
Cell[5931656, 170448, 434, 13, 47, "Input",ExpressionUUID->"ac935daa-2601-46b3-8f1b-663835db5f89"],
Cell[5932093, 170463, 86543, 2454, 180, "Output",ExpressionUUID->"c02f7a4b-966e-4bec-a5c9-89bfe1c58bc8",
 CellID->43290434]
}, Open  ]],
Cell[CellGroupData[{
Cell[6018673, 172922, 283, 9, 27, "Input",ExpressionUUID->"2e64f155-5441-4cc6-bdf3-bb3ad6e7c5f1"],
Cell[6018959, 172933, 173, 3, 26, "Output",ExpressionUUID->"1af94ba1-4bc8-4d48-8021-57beb8f73603",
 CellID->1732851552]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6019181, 172942, 123, 0, 40, "MathCaption",ExpressionUUID->"9b8c4ae5-3b0b-4000-89cb-76c82e59f15f"],
Cell[CellGroupData[{
Cell[6019329, 172946, 274, 7, 27, "Input",ExpressionUUID->"090bef6b-e5d5-4d9f-b7ef-362f3aa23e2d"],
Cell[6019606, 172955, 661, 13, 22, "Print",ExpressionUUID->"d6f3c3f6-b6ae-49d6-8395-be798d5e63d0",
 CellID->2010297497]
}, Open  ]],
Cell[CellGroupData[{
Cell[6020304, 172973, 423, 12, 27, "Input",ExpressionUUID->"f02d4fbf-75b7-42aa-9904-ad575ba8b4b6"],
Cell[6020730, 172987, 724, 15, 22, "Print",ExpressionUUID->"04283114-1508-4877-9f5a-6c3662c8c233",
 CellID->709813828]
}, Open  ]],
Cell[6021469, 173005, 411, 12, 27, "Input",ExpressionUUID->"0f1202db-3c09-4dd1-867c-1d071ff6ca78"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6021917, 173022, 117, 0, 40, "MathCaption",ExpressionUUID->"481cadbe-1cad-4943-883c-cee3cf243bdf"],
Cell[CellGroupData[{
Cell[6022059, 173026, 478, 16, 27, "Input",ExpressionUUID->"008507e6-c9a9-4b55-8229-c2903eb287ea"],
Cell[6022540, 173044, 6685, 193, 40, "Output",ExpressionUUID->"5b0d2c69-f8bc-446c-8ed7-d57b68a637fc",
 CellID->569348111]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6029274, 173243, 139, 1, 40, "MathCaption",ExpressionUUID->"efb7a19c-7f8c-4ed5-abe4-4c2e8a8f2729"],
Cell[CellGroupData[{
Cell[6029438, 173248, 370, 11, 27, "Input",ExpressionUUID->"d639e9b6-ea71-4fbf-b6b2-63f5418350f9"],
Cell[6029811, 173261, 32599, 956, 94, "Output",ExpressionUUID->"cc9bac43-6a0c-4888-988d-b941f05f0b77",
 CellID->1994054192]
}, Open  ]],
Cell[CellGroupData[{
Cell[6062447, 174222, 513, 16, 27, "Input",ExpressionUUID->"1ba91924-9c4e-41af-b6e2-29df80c4503a"],
Cell[6062963, 174240, 29598, 844, 66, "Output",ExpressionUUID->"972bddae-1d34-41ec-9933-779e657c11e5",
 CellID->439009862]
}, Open  ]],
Cell[CellGroupData[{
Cell[6092598, 175089, 283, 9, 27, "Input",ExpressionUUID->"94d9ef2c-c451-4910-aed8-8ff39da3ee37"],
Cell[6092884, 175100, 171, 3, 26, "Output",ExpressionUUID->"cea9ddfc-6c30-497a-bd83-1bfb92b3c6d8",
 CellID->1832423189]
}, Open  ]],
Cell[CellGroupData[{
Cell[6093092, 175108, 224, 6, 27, "Input",ExpressionUUID->"f4109034-4633-4018-9a05-077300cd3613"],
Cell[CellGroupData[{
Cell[6093341, 175118, 561, 12, 22, "Print",ExpressionUUID->"badb77ec-5b1b-4f6f-baa1-4104f5d65f53",
 CellID->473555896],
Cell[6093905, 175132, 561, 12, 22, "Print",ExpressionUUID->"668905b2-5b5d-472c-b923-5708f4b88f93",
 CellID->479548290]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6094527, 175151, 103, 0, 40, "MathCaption",ExpressionUUID->"2cd10378-4151-462a-a455-63db2f7f25bf"],
Cell[CellGroupData[{
Cell[6094655, 175155, 228, 5, 27, "Input",ExpressionUUID->"1a9b2cea-9c0f-4ec3-b462-cf27abac59a8"],
Cell[6094886, 175162, 588, 14, 46, "Output",ExpressionUUID->"b935d184-ded2-4fbb-bd3c-2ff7e9d7eccf",
 CellID->261982839]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6095535, 175183, 91, 0, 39, "Subsection",ExpressionUUID->"d6923f02-2b66-412e-af12-321a0ce33e88"],
Cell[6095629, 175185, 367, 7, 40, "Text",ExpressionUUID->"6081aed3-f802-41cd-aeed-1dc8971c90ca"],
Cell[CellGroupData[{
Cell[6096021, 175196, 116, 0, 40, "MathCaption",ExpressionUUID->"c17f0cb6-dc4e-4630-8a27-ade9604d919b"],
Cell[CellGroupData[{
Cell[6096162, 175200, 153, 3, 27, "Input",ExpressionUUID->"3e3b79c6-abff-4e45-a53b-9a17cbec3289"],
Cell[CellGroupData[{
Cell[6096340, 175207, 588, 12, 22, "Print",ExpressionUUID->"127cb54b-9257-44dd-8e4a-cfa1f30ada32",
 CellID->1013965578],
Cell[6096931, 175221, 617, 12, 22, "Print",ExpressionUUID->"c239eeaa-7433-480a-a778-cf497aba09a5",
 CellID->1436730501],
Cell[6097551, 175235, 606, 12, 22, "Print",ExpressionUUID->"8ff7242d-7202-4146-93bd-91a20535ef7e",
 CellID->823403439],
Cell[6098160, 175249, 596, 12, 22, "Print",ExpressionUUID->"0af6b051-23f3-4613-bbd1-ebf124a22288",
 CellID->556978005],
Cell[6098759, 175263, 592, 12, 22, "Print",ExpressionUUID->"648031af-aec6-4184-9071-e3f3cfb83b80",
 CellID->248112762],
Cell[6099354, 175277, 587, 12, 22, "Print",ExpressionUUID->"ec2e35ee-f2ee-4ca9-afb1-ebb03aefe50c",
 CellID->1765432107],
Cell[6099944, 175291, 580, 12, 22, "Print",ExpressionUUID->"48c0afb8-26ff-423f-84f8-ce2722cb3e5d",
 CellID->1983355397],
Cell[6100527, 175305, 601, 12, 22, "Print",ExpressionUUID->"c2752829-86ff-487b-9803-e3808eb11fac",
 CellID->1012988645],
Cell[6101131, 175319, 579, 12, 22, "Print",ExpressionUUID->"5297cb02-7bd0-4e39-a409-5c3b7b3ad8a2",
 CellID->49803099],
Cell[6101713, 175333, 569, 12, 22, "Print",ExpressionUUID->"bc69bac2-647d-45aa-8d18-ac235a88695c",
 CellID->1669546210],
Cell[6102285, 175347, 577, 12, 22, "Print",ExpressionUUID->"ab28b520-f625-4f0b-98cb-335c30478810",
 CellID->1041819768],
Cell[6102865, 175361, 602, 12, 22, "Print",ExpressionUUID->"c5134cd4-b9ec-47e4-ba51-f52fe2a27720",
 CellID->1430666785],
Cell[6103470, 175375, 595, 12, 22, "Print",ExpressionUUID->"1e053e9f-ff79-423f-9ca2-f401e289780c",
 CellID->1989743683]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6104126, 175394, 157, 2, 40, "MathCaption",ExpressionUUID->"52f28339-ca18-4e06-8d46-fac8ac790249"],
Cell[CellGroupData[{
Cell[6104308, 175400, 220, 6, 27, "Input",ExpressionUUID->"cd4edb72-92f9-49d6-b928-44e48d5265f6"],
Cell[CellGroupData[{
Cell[6104553, 175410, 551, 12, 22, "Print",ExpressionUUID->"375ce098-22fd-4eea-9b9a-7e7c6ba9119f",
 CellID->1006729097],
Cell[6105107, 175424, 549, 12, 22, "Print",ExpressionUUID->"70079b5e-7274-48a4-b0fc-3430137fa049",
 CellID->72278566],
Cell[6105659, 175438, 549, 12, 22, "Print",ExpressionUUID->"84d17d86-c008-475b-8a64-534c8bdc00e4",
 CellID->1457598207]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6106269, 175457, 103, 0, 40, "MathCaption",ExpressionUUID->"0d9073bf-7aac-47d8-91f7-a1ca40646413"],
Cell[CellGroupData[{
Cell[6106397, 175461, 147, 3, 27, "Input",ExpressionUUID->"9d07abea-f93d-4f01-86b9-82387d207df3"],
Cell[CellGroupData[{
Cell[6106569, 175468, 655, 14, 22, "Print",ExpressionUUID->"38ebc133-2088-4b4f-9e3a-e412bbea51cb",
 CellID->1507103720],
Cell[6107227, 175484, 631, 13, 22, "Print",ExpressionUUID->"032d23bc-fdd8-458d-82c9-3c0cd50fd746",
 CellID->322605675],
Cell[6107861, 175499, 621, 12, 22, "Print",ExpressionUUID->"3d98b748-94ee-4e89-8635-419ccf108916",
 CellID->1784854697],
Cell[6108485, 175513, 609, 13, 22, "Print",ExpressionUUID->"29cd1e11-7f8a-471c-a4b2-808524c63417",
 CellID->2001311183],
Cell[6109097, 175528, 583, 12, 22, "Print",ExpressionUUID->"3005ea08-db95-4831-b6ff-2628254cd586",
 CellID->689720],
Cell[6109683, 175542, 602, 12, 22, "Print",ExpressionUUID->"754d7d38-521d-459a-b42e-f19e4f8085cf",
 CellID->854705805],
Cell[6110288, 175556, 582, 12, 22, "Print",ExpressionUUID->"68897784-080d-4b67-bb00-7f130affcab1",
 CellID->522349852],
Cell[6110873, 175570, 582, 12, 22, "Print",ExpressionUUID->"8f01818c-31ab-4d7e-9959-7df760b9493e",
 CellID->220813628],
Cell[6111458, 175584, 576, 12, 22, "Print",ExpressionUUID->"c18d0f36-1c27-48e8-a522-109a13fc49f2",
 CellID->362903780]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6112083, 175602, 146, 3, 27, "Input",ExpressionUUID->"cf1c0fae-d5c3-4f8f-ada6-224a65f01fc0"],
Cell[CellGroupData[{
Cell[6112254, 175609, 616, 12, 22, "Print",ExpressionUUID->"aee60881-e717-4910-b5af-16f8444fc8b1",
 CellID->470272980],
Cell[6112873, 175623, 603, 12, 22, "Print",ExpressionUUID->"48cb4962-e628-4ae9-b2ed-dc4716b4355f",
 CellID->1138153900],
Cell[6113479, 175637, 581, 12, 22, "Print",ExpressionUUID->"5d8c16c4-a089-4c7c-bde8-143820efd640",
 CellID->1315753471],
Cell[6114063, 175651, 583, 12, 22, "Print",ExpressionUUID->"8791e39d-816d-4ebe-807e-1d8591e624ce",
 CellID->1229473938],
Cell[6114649, 175665, 576, 12, 22, "Print",ExpressionUUID->"93286f54-09f5-4f8b-bd25-34d6e470a10b",
 CellID->680824532]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6115286, 175684, 111, 0, 40, "MathCaption",ExpressionUUID->"e3aafc61-9d3b-4b59-87ff-0ccc9cff0bdc"],
Cell[CellGroupData[{
Cell[6115422, 175688, 168, 4, 27, "Input",ExpressionUUID->"e6e2b3ff-8b96-4ff8-aa7c-2c096068287b"],
Cell[CellGroupData[{
Cell[6115615, 175696, 550, 12, 22, "Print",ExpressionUUID->"d862fdb2-7656-4ec7-abc1-cda50a28ed06",
 CellID->825038108],
Cell[6116168, 175710, 548, 12, 22, "Print",ExpressionUUID->"c098217c-b723-41a5-b384-9751a36deefa",
 CellID->113572477],
Cell[6116719, 175724, 559, 12, 22, "Print",ExpressionUUID->"3545c4d2-4990-424c-8f5f-f4c84374bddf",
 CellID->1847142403],
Cell[6117281, 175738, 548, 12, 22, "Print",ExpressionUUID->"e9bd3c90-ec57-4725-b255-93d36d520588",
 CellID->898766189],
Cell[6117832, 175752, 548, 12, 22, "Print",ExpressionUUID->"bb84fa3b-f475-4641-b68c-3742288380db",
 CellID->163365525],
Cell[6118383, 175766, 557, 12, 22, "Print",ExpressionUUID->"8781ba9f-9f19-4f4c-a013-6f12f829f2f9",
 CellID->1200277809]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6119001, 175785, 108, 0, 40, "MathCaption",ExpressionUUID->"0a16e88e-adbf-4ba1-8a5a-748b90d81ce0"],
Cell[CellGroupData[{
Cell[6119134, 175789, 153, 3, 27, "Input",ExpressionUUID->"5321598a-431e-4e4d-9af6-1775f6ad95d7"],
Cell[CellGroupData[{
Cell[6119312, 175796, 749, 18, 22, "Print",ExpressionUUID->"28d28b77-1155-4337-8c3e-56d76eaba4ce",
 CellID->1755239592],
Cell[6120064, 175816, 701, 18, 22, "Print",ExpressionUUID->"ebc54623-5d8c-45f3-98ac-258068966fe9",
 CellID->328963905]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6120826, 175841, 109, 0, 40, "MathCaption",ExpressionUUID->"31294351-fbc9-4e40-9c3e-ac31e717c208"],
Cell[CellGroupData[{
Cell[6120960, 175845, 150, 3, 27, "Input",ExpressionUUID->"0237f5d4-8cca-42ef-bdd1-e956ef969156"],
Cell[CellGroupData[{
Cell[6121135, 175852, 725, 18, 22, "Print",ExpressionUUID->"282f6553-137d-4040-b273-9a55cb7b14a1",
 CellID->1273443957],
Cell[6121863, 175872, 558, 12, 22, "Print",ExpressionUUID->"1fab8531-0e54-4341-a9c7-0a3915a5f4bc",
 CellID->950559297]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6122470, 175890, 150, 3, 27, "Input",ExpressionUUID->"21b93252-a4b6-42c1-9a74-c7a84ab41fea"],
Cell[CellGroupData[{
Cell[6122645, 175897, 724, 18, 22, "Print",ExpressionUUID->"9e4fab67-8496-4d65-be2c-6bd2f442ea89",
 CellID->919254646],
Cell[6123372, 175917, 558, 12, 22, "Print",ExpressionUUID->"a93cccef-26ee-4eaa-9743-9dea115574d2",
 CellID->915478804]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6123991, 175936, 234, 5, 40, "MathCaption",ExpressionUUID->"03b42bf4-fb63-4e7d-9734-5c58720b7106"],
Cell[CellGroupData[{
Cell[6124250, 175945, 134, 3, 27, "Input",ExpressionUUID->"4aa7c635-10b7-419a-869d-fe7ed3e510a7"],
Cell[6124387, 175950, 4117, 102, 123, "Print",ExpressionUUID->"8165ecdd-6b68-4070-a5c1-785212a284bf",
 CellID->1879411720]
}, Open  ]],
Cell[6128519, 176055, 157, 3, 27, "Input",ExpressionUUID->"3c538320-e0ae-4be6-8c39-e45459b92feb"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6128713, 176063, 105, 0, 40, "MathCaption",ExpressionUUID->"5a937aa9-735f-443f-bd3a-81ac79191ccc"],
Cell[CellGroupData[{
Cell[6128843, 176067, 134, 3, 27, "Input",ExpressionUUID->"6f00becf-5abe-4699-a4ec-9231680d7635"],
Cell[6128980, 176072, 405, 9, 22, "Message",ExpressionUUID->"efc46c41-5a88-47ca-90b0-af740c406989",
 CellID->2083297964]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6129458, 176089, 1013, 19, 54, "Section",ExpressionUUID->"4d1d0042-6342-4549-8253-0be8111126d2",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1088456401],
Cell[CellGroupData[{
Cell[6130496, 176112, 103, 0, 39, "Subsection",ExpressionUUID->"bd19e30b-c5c6-48ee-b788-df3215499082"],
Cell[6130602, 176114, 506, 12, 57, "Text",ExpressionUUID->"ce2aa235-efec-4671-8a23-ed7b3c3506f1"],
Cell[6131111, 176128, 1294, 36, 92, "DefinitionBox",ExpressionUUID->"4b2373aa-d26c-4660-9aa2-d668c0f79239"],
Cell[6132408, 176166, 94, 0, 23, "Text",ExpressionUUID->"e9100e92-e074-46c3-8e37-8ecf73b21270"],
Cell[CellGroupData[{
Cell[6132527, 176170, 170, 3, 40, "MathCaption",ExpressionUUID->"f4079698-6ce3-4813-afa0-461d46cbdec6"],
Cell[CellGroupData[{
Cell[6132722, 176177, 249, 6, 27, "Input",ExpressionUUID->"31ba2f4e-92df-41c2-b37e-2d860cbd03e9"],
Cell[CellGroupData[{
Cell[6132996, 176187, 616, 12, 22, "Print",ExpressionUUID->"3f14afae-a7e8-4610-8112-455e21b96ebe",
 CellID->1342712479],
Cell[6133615, 176201, 628, 12, 22, "Print",ExpressionUUID->"c62ab0c1-889b-43b8-9728-4440208ba040",
 CellID->1210471877]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6134292, 176219, 249, 6, 27, "Input",ExpressionUUID->"201218f8-8db8-4abf-910b-405b21d65829"],
Cell[CellGroupData[{
Cell[6134566, 176229, 618, 12, 22, "Print",ExpressionUUID->"acc84c39-4d7f-43d3-adfd-179693a62192",
 CellID->1877787831],
Cell[6135187, 176243, 625, 12, 22, "Print",ExpressionUUID->"d6294237-48be-4b32-bafb-31d820d8b668",
 CellID->560399305]
}, Open  ]],
Cell[6135827, 176258, 474, 9, 22, "Message",ExpressionUUID->"50c779b0-1b9d-4d47-9edb-64ecd7555a21",
 CellID->1915327040],
Cell[6136304, 176269, 471, 9, 22, "Message",ExpressionUUID->"37624fce-ecb5-46ed-8bd2-772685f5764e",
 CellID->121692750]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6136824, 176284, 374, 9, 56, "MathCaption",ExpressionUUID->"ccf8a0e1-3ad2-4255-94ef-b7f2a7bc2141"],
Cell[CellGroupData[{
Cell[6137223, 176297, 204, 5, 27, "Input",ExpressionUUID->"97842234-8dd8-494b-bed4-a1189bae3fdc"],
Cell[CellGroupData[{
Cell[6137452, 176306, 620, 12, 22, "Print",ExpressionUUID->"ff4f7f4e-900b-489a-a5cf-b30e6cdd16d0",
 CellID->1229540967],
Cell[6138075, 176320, 642, 13, 22, "Print",ExpressionUUID->"9d2eb126-12ad-4fa2-9293-e2ab55d3e318",
 CellID->1807380935],
Cell[6138720, 176335, 793, 17, 22, "Print",ExpressionUUID->"64f5a93f-023b-4646-b0fd-6abc66c7cd9e",
 CellID->495187730],
Cell[6139516, 176354, 790, 17, 22, "Print",ExpressionUUID->"78baae71-eca0-4328-b767-cfd89726997b",
 CellID->1121045985]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[6140345, 176376, 316, 5, 40, "Text",ExpressionUUID->"9325c653-c73d-487e-a615-18313bddbab6"],
Cell[6140664, 176383, 1951, 52, 126, "DefinitionBox",ExpressionUUID->"cafe33b9-d6af-491a-a756-2e3dee610c23"],
Cell[6142618, 176437, 94, 0, 23, "Text",ExpressionUUID->"0a81f6e4-c054-4628-9cd9-c6f069d42312"],
Cell[CellGroupData[{
Cell[6142737, 176441, 154, 2, 40, "MathCaption",ExpressionUUID->"7a5d775f-ce16-43b9-ab5a-acd1938ff8aa"],
Cell[CellGroupData[{
Cell[6142916, 176447, 227, 5, 27, "Input",ExpressionUUID->"22b40760-e387-4d74-840c-c4750f513e93"],
Cell[6143146, 176454, 223, 4, 26, "Output",ExpressionUUID->"218d97ca-8c08-402c-9e99-9b8612ea43c4",
 CellID->1484713041]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6143418, 176464, 164, 2, 40, "MathCaption",ExpressionUUID->"be04e4c7-01e2-4312-bcbb-4e0098921d9c"],
Cell[CellGroupData[{
Cell[6143607, 176470, 155, 3, 27, "Input",ExpressionUUID->"5efebdac-ebf9-4cd2-82ad-f39b8d6d2b07"],
Cell[6143765, 176475, 198, 3, 26, "Output",ExpressionUUID->"f8c74aa8-1be5-4699-8039-32f669a28894",
 CellID->966467739]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6144012, 176484, 172, 3, 40, "MathCaption",ExpressionUUID->"1e94c03d-2022-4082-baed-57ca743ef091"],
Cell[CellGroupData[{
Cell[6144209, 176491, 229, 5, 27, "Input",ExpressionUUID->"467f4c6a-f063-4f16-90eb-492fcabe36cb"],
Cell[6144441, 176498, 220, 4, 26, "Output",ExpressionUUID->"ed04c486-e87d-4e54-96a4-9afa6ccc7756",
 CellID->987277630]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6144710, 176508, 159, 2, 40, "MathCaption",ExpressionUUID->"942cd52c-6e9a-40cf-b703-5cba375db5e9"],
Cell[CellGroupData[{
Cell[6144894, 176514, 300, 7, 27, "Input",ExpressionUUID->"2a44ddd1-7fdc-4d41-8a04-29a77139ef05"],
Cell[6145197, 176523, 257, 5, 26, "Output",ExpressionUUID->"d1ff0ccb-06e5-4e95-95ca-13728ae9d216",
 CellID->1529220533]
}, Open  ]],
Cell[CellGroupData[{
Cell[6145491, 176533, 303, 7, 27, "Input",ExpressionUUID->"48e8b980-ed4f-415e-ab9b-e35dd82ad06a"],
Cell[6145797, 176542, 256, 5, 26, "Output",ExpressionUUID->"1f838fd9-b3a4-463a-b87b-4fd19e5875ec",
 CellID->1632892784]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6146102, 176553, 162, 2, 40, "MathCaption",ExpressionUUID->"5f2bcb82-b954-45d1-930d-295f391e0b73"],
Cell[CellGroupData[{
Cell[6146289, 176559, 142, 3, 27, "Input",ExpressionUUID->"845b3dcd-f2d4-44c3-b467-f0338b71a477"],
Cell[6146434, 176564, 194, 3, 26, "Output",ExpressionUUID->"2f44ba53-1cd8-4dcf-9d77-0019ac66b36c",
 CellID->617915879]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6146677, 176573, 114, 0, 40, "MathCaption",ExpressionUUID->"0fdb29a9-f384-4d48-981c-a99487b2e35d"],
Cell[CellGroupData[{
Cell[6146816, 176577, 143, 3, 27, "Input",ExpressionUUID->"79d83be9-55df-44ae-83b4-fb459adf1d3f"],
Cell[6146962, 176582, 193, 3, 26, "Output",ExpressionUUID->"53684c69-163e-4388-920f-edc917f244dd",
 CellID->1533455672]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6147216, 176592, 107, 0, 39, "Subsection",ExpressionUUID->"ba14c322-0a96-4de9-b204-78bbc49afd82"],
Cell[6147326, 176594, 213, 3, 23, "Text",ExpressionUUID->"20890f5e-cad6-46c4-8b10-efd4bd00f6dc"],
Cell[CellGroupData[{
Cell[6147564, 176601, 116, 0, 40, "MathCaption",ExpressionUUID->"d6ef2031-0e6b-4fa7-b877-bb8ca3872550"],
Cell[CellGroupData[{
Cell[6147705, 176605, 281, 7, 27, "Input",ExpressionUUID->"44c2a0b6-e89c-48cc-915b-e9cc0fc5986a"],
Cell[CellGroupData[{
Cell[6148011, 176616, 617, 12, 22, "Print",ExpressionUUID->"fb5bd5d5-0e64-477f-8e41-b35b7cce19fa",
 CellID->243441307],
Cell[6148631, 176630, 625, 12, 22, "Print",ExpressionUUID->"fb84e27c-861e-4d85-8ca3-c0aa65cea3a1",
 CellID->561699214]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6149317, 176649, 155, 2, 40, "MathCaption",ExpressionUUID->"d2a24ddf-69ff-43ff-bab6-9fb2efec0df1"],
Cell[CellGroupData[{
Cell[6149497, 176655, 258, 6, 27, "Input",ExpressionUUID->"228bc87a-ac89-49e6-b1dd-891e561ed4bb"],
Cell[CellGroupData[{
Cell[6149780, 176665, 619, 12, 22, "Print",ExpressionUUID->"df7b4ed7-7072-433e-94a7-c6af4fbe4661",
 CellID->497035243],
Cell[6150402, 176679, 714, 14, 22, "Print",ExpressionUUID->"dbb9f3db-a67f-4173-8b07-bb0f31ad0754",
 CellID->2107425918],
Cell[6151119, 176695, 812, 18, 22, "Print",ExpressionUUID->"e06b57b1-0756-4211-a43a-3b14ffc5511f",
 CellID->1059613855],
Cell[6151934, 176715, 806, 18, 22, "Print",ExpressionUUID->"f7db3191-eb1f-49c6-80d3-6204a1078029",
 CellID->1995660061]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6152801, 176740, 125, 0, 40, "MathCaption",ExpressionUUID->"dce76985-b9ea-48d9-ace5-4dc48bb53098"],
Cell[CellGroupData[{
Cell[6152951, 176744, 300, 7, 27, "Input",ExpressionUUID->"5056e74b-c88e-4b49-b3c9-33b85ac171c2"],
Cell[6153254, 176753, 259, 5, 26, "Output",ExpressionUUID->"f9ff3b4a-8e45-439e-a5aa-5ff013b0fa97",
 CellID->1522510521]
}, Open  ]],
Cell[CellGroupData[{
Cell[6153550, 176763, 227, 5, 27, "Input",ExpressionUUID->"e5f472bd-60ce-445e-ac35-ccb0d502c691"],
Cell[6153780, 176770, 222, 4, 26, "Output",ExpressionUUID->"7e7cba8d-6516-4853-8669-935773b996e7",
 CellID->438371102]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6154051, 176780, 186, 3, 40, "MathCaption",ExpressionUUID->"49cd33e8-a753-4d83-bc5e-9135a8068600"],
Cell[CellGroupData[{
Cell[6154262, 176787, 142, 3, 27, "Input",ExpressionUUID->"7f283785-1f55-4521-a14a-25247ae9ed4f"],
Cell[6154407, 176792, 223, 4, 26, "Output",ExpressionUUID->"094c6a46-792a-4b40-ab3b-0a7083bac496",
 CellID->24126788]
}, Open  ]],
Cell[CellGroupData[{
Cell[6154667, 176801, 205, 5, 27, "Input",ExpressionUUID->"03fd7cde-172f-4c6d-b676-34ae28ab6f86"],
Cell[6154875, 176808, 194, 3, 26, "Output",ExpressionUUID->"ec9a505d-66c6-4bf9-9d22-ac669839495a",
 CellID->408857426]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6155118, 176817, 449, 10, 72, "MathCaption",ExpressionUUID->"bfa4bc4a-b164-4737-aafa-77971369a371"],
Cell[CellGroupData[{
Cell[6155592, 176831, 229, 5, 27, "Input",ExpressionUUID->"10435631-1d8f-44bf-91d8-59aa3909cad2"],
Cell[6155824, 176838, 221, 4, 26, "Output",ExpressionUUID->"0b033c06-e6ff-4f92-97f9-0c4af366470b",
 CellID->1088131848]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6156094, 176848, 120, 0, 40, "MathCaption",ExpressionUUID->"f6077cf3-fb94-439c-ae03-8bf7e9beb426"],
Cell[CellGroupData[{
Cell[6156239, 176852, 303, 7, 27, "Input",ExpressionUUID->"7e8fea0f-c075-43dd-8fbe-da4a517d5bbb"],
Cell[6156545, 176861, 260, 5, 26, "Output",ExpressionUUID->"e8893599-2f9b-4f62-846c-d5dda25c1f4a",
 CellID->907850413]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6156854, 176872, 165, 3, 40, "MathCaption",ExpressionUUID->"933fa1d3-312a-46b5-aba4-2d6be451c452"],
Cell[CellGroupData[{
Cell[6157044, 176879, 176, 4, 27, "Input",ExpressionUUID->"9224ab65-d05b-474a-9102-069a6cc21885"],
Cell[6157223, 176885, 193, 3, 26, "Output",ExpressionUUID->"60c90f84-25a4-4840-ac64-a099e8ea3683",
 CellID->1263115465]
}, Open  ]],
Cell[CellGroupData[{
Cell[6157453, 176893, 176, 4, 27, "Input",ExpressionUUID->"d989c0b4-fe37-4a00-9301-982c39e21a1a"],
Cell[6157632, 176899, 225, 4, 26, "Output",ExpressionUUID->"945eadf9-bdea-40e8-b2df-d18c6a90adb0",
 CellID->1762305658]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6157918, 176910, 101, 0, 39, "Subsection",ExpressionUUID->"ec16d493-45ad-487f-b888-2b5bf158c405"],
Cell[6158022, 176912, 371, 9, 40, "Text",ExpressionUUID->"22079b7e-013a-4594-b299-fb60e80f25c4"],
Cell[6158396, 176923, 1033, 28, 75, "DefinitionBox",ExpressionUUID->"171ba3a9-2d73-4e12-830b-f564315eb8cb"],
Cell[6159432, 176953, 95, 0, 23, "Text",ExpressionUUID->"b66c8ada-e039-4279-843e-64f3b7bb47ba"],
Cell[CellGroupData[{
Cell[6159552, 176957, 271, 6, 56, "MathCaption",ExpressionUUID->"03ae8c17-e6a1-4869-b68f-d21d3bf4d88d"],
Cell[CellGroupData[{
Cell[6159848, 176967, 180, 4, 27, "Input",ExpressionUUID->"e3bddfc3-a5ff-4603-a1a1-864a49de7226"],
Cell[6160031, 176973, 238, 4, 26, "Output",ExpressionUUID->"081de4c1-8148-4092-ad47-b97668e3bc89",
 CellID->1115016536]
}, Open  ]],
Cell[CellGroupData[{
Cell[6160306, 176982, 221, 5, 27, "Input",ExpressionUUID->"4161397e-ea4f-4c43-8533-4b56b4ca39eb"],
Cell[6160530, 176989, 223, 4, 26, "Output",ExpressionUUID->"d966591d-a28a-4a40-be1e-c4d862440e18",
 CellID->1188860967]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6160802, 176999, 242, 4, 56, "MathCaption",ExpressionUUID->"142660cf-596c-4ad3-8958-1288e5b40a73"],
Cell[CellGroupData[{
Cell[6161069, 177007, 178, 4, 27, "Input",ExpressionUUID->"e08a9064-1c8d-4368-a1e4-d28650736d46"],
Cell[6161250, 177013, 242, 4, 26, "Output",ExpressionUUID->"5acf51b6-3b59-4d0d-b923-ce3f77cf7f83",
 CellID->1484132930]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6161541, 177023, 108, 0, 40, "MathCaption",ExpressionUUID->"3bac8eea-ba41-43b0-8ce7-b7155cc8b33b"],
Cell[CellGroupData[{
Cell[6161674, 177027, 291, 7, 27, "Input",ExpressionUUID->"0adfd2a3-4ecd-4392-9b2f-99863a4cd7f2"],
Cell[6161968, 177036, 258, 5, 26, "Output",ExpressionUUID->"a43b3bf0-a85f-4a39-aa64-31dbdd25dd3a",
 CellID->285100635]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6162275, 177047, 205, 3, 40, "MathCaption",ExpressionUUID->"311d8741-bb03-453b-a19a-37bffd0d264c"],
Cell[CellGroupData[{
Cell[6162505, 177054, 150, 3, 27, "Input",ExpressionUUID->"12c52d38-3235-4b69-9e7d-69b60cd5eb3d"],
Cell[6162658, 177059, 215, 4, 26, "Output",ExpressionUUID->"bd6a3955-c8c7-4c5e-bba5-bdb5f37ba424",
 CellID->1411403213]
}, Open  ]],
Cell[CellGroupData[{
Cell[6162910, 177068, 141, 3, 27, "Input",ExpressionUUID->"f1bcdaf1-0e4e-4bdd-8c52-1be143f647a6"],
Cell[6163054, 177073, 212, 3, 26, "Output",ExpressionUUID->"6897d6e4-4c73-4f11-8373-a290d34e8ee4",
 CellID->1427366760]
}, Open  ]],
Cell[CellGroupData[{
Cell[6163303, 177081, 150, 3, 27, "Input",ExpressionUUID->"160627a7-27bc-482e-9b0d-5d5a7f5da1c2"],
Cell[6163456, 177086, 239, 4, 26, "Output",ExpressionUUID->"84c65189-12a6-4e68-9710-c943eba4c51a",
 CellID->1071935285]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6163744, 177096, 94, 0, 40, "MathCaption",ExpressionUUID->"6d94bb4b-29d7-40d0-ad14-4850682e3d87"],
Cell[CellGroupData[{
Cell[6163863, 177100, 170, 4, 27, "Input",ExpressionUUID->"ab254e1c-a3bc-43e2-8ed1-84706f26d2c4"],
Cell[6164036, 177106, 191, 3, 26, "Output",ExpressionUUID->"3d09e326-bca2-4abc-827e-b4b5d1d97f88",
 CellID->837399513]
}, Open  ]],
Cell[CellGroupData[{
Cell[6164264, 177114, 170, 4, 27, "Input",ExpressionUUID->"03e415a2-5f63-413b-9be7-c229004ef9f1"],
Cell[6164437, 177120, 223, 4, 26, "Output",ExpressionUUID->"69f29f14-ac41-4498-8eef-0717abccac41",
 CellID->324926915]
}, Open  ]]
}, Open  ]],
Cell[6164687, 177128, 280, 6, 40, "Text",ExpressionUUID->"b1b31f59-477e-4cc5-b17e-3e42511c4ea7"],
Cell[CellGroupData[{
Cell[6164992, 177138, 177, 5, 40, "MathCaption",ExpressionUUID->"437043e7-74e4-4443-8f0c-5c4192f50262"],
Cell[CellGroupData[{
Cell[6165194, 177147, 176, 4, 27, "Input",ExpressionUUID->"1de03d3f-2372-4752-81a0-ce5269abe12f"],
Cell[6165373, 177153, 216, 4, 26, "Output",ExpressionUUID->"e7f35533-8843-44bc-82e3-424a77a5140c",
 CellID->313260877]
}, Open  ]],
Cell[CellGroupData[{
Cell[6165626, 177162, 287, 6, 27, "Input",ExpressionUUID->"ebeafe23-a4dc-4a6e-97fb-cd9a4ea53b05"],
Cell[6165916, 177170, 250, 5, 26, "Output",ExpressionUUID->"14bd38a3-5d64-4c08-b006-c08d65182588",
 CellID->1451966044]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6166215, 177181, 132, 0, 40, "MathCaption",ExpressionUUID->"b90c82eb-9fc8-4d91-9cf3-b30aaec49f20"],
Cell[CellGroupData[{
Cell[6166372, 177185, 291, 7, 27, "Input",ExpressionUUID->"5b8064e6-508c-451c-bf44-0b6a8ed3c976"],
Cell[6166666, 177194, 256, 5, 26, "Output",ExpressionUUID->"ce165c7f-e829-48a0-9734-6425aeb65a71",
 CellID->184692807]
}, Open  ]],
Cell[CellGroupData[{
Cell[6166959, 177204, 152, 3, 27, "Input",ExpressionUUID->"a7b1389d-4dc1-4ab0-9d17-4623cd084476"],
Cell[6167114, 177209, 215, 4, 26, "Output",ExpressionUUID->"88a871e8-b1a1-4e99-826a-fb3fc92352d7",
 CellID->1192473938]
}, Open  ]],
Cell[CellGroupData[{
Cell[6167366, 177218, 139, 3, 27, "Input",ExpressionUUID->"0024beb5-ae49-46ac-86ce-b8401c95a4c0"],
Cell[6167508, 177223, 193, 3, 26, "Output",ExpressionUUID->"a2e54119-62b4-4348-81a9-151931d3d264",
 CellID->1362085225]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6167750, 177232, 112, 0, 40, "MathCaption",ExpressionUUID->"9ddfa442-78a1-4777-b2df-e6d2a17e9081"],
Cell[CellGroupData[{
Cell[6167887, 177236, 152, 3, 27, "Input",ExpressionUUID->"c0b02ce0-6156-4d73-8655-737e6133b05b"],
Cell[6168042, 177241, 199, 3, 26, "Output",ExpressionUUID->"56417482-0e9c-488f-9b1e-36654ba105ed",
 CellID->957605907]
}, Open  ]],
Cell[CellGroupData[{
Cell[6168278, 177249, 195, 5, 27, "Input",ExpressionUUID->"cc162166-3860-433c-bbe6-06c31caec72e"],
Cell[6168476, 177256, 200, 3, 26, "Output",ExpressionUUID->"d1ae09e6-e3c2-4349-9668-239e9872af0b",
 CellID->1292709219]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6168737, 177266, 123, 0, 39, "Subsection",ExpressionUUID->"d87f07d4-4d04-4f31-be9d-2e446937c5b7"],
Cell[6168863, 177268, 663, 23, 40, "Text",ExpressionUUID->"adc30f64-3f0b-4c9f-b8b3-f205c9a6910f"],
Cell[6169529, 177293, 392, 9, 41, "DefinitionBox",ExpressionUUID->"54b8cafc-8ce7-4197-8414-15455a64638b"],
Cell[6169924, 177304, 95, 0, 23, "Text",ExpressionUUID->"b72f3ffb-df0a-4441-be80-2f88c110f3ec"],
Cell[CellGroupData[{
Cell[6170044, 177308, 211, 5, 40, "MathCaption",ExpressionUUID->"b397e042-a772-4269-815c-d7216d17f405"],
Cell[CellGroupData[{
Cell[6170280, 177317, 143, 3, 27, "Input",ExpressionUUID->"08c297a6-3ee1-4233-99b1-f12f1f497b6c"],
Cell[6170426, 177322, 343, 9, 26, "Output",ExpressionUUID->"aa068460-dbdd-497f-9084-2d1825cb79c5",
 CellID->1990684632]
}, Open  ]],
Cell[CellGroupData[{
Cell[6170806, 177336, 282, 7, 27, "Input",ExpressionUUID->"911b2e85-5839-4e03-990b-73607dd6a601"],
Cell[6171091, 177345, 251, 5, 26, "Output",ExpressionUUID->"aad8cd6d-f596-4383-8233-1d40b09511fe",
 CellID->2039037263]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6171391, 177356, 239, 5, 40, "MathCaption",ExpressionUUID->"cb6187b0-05f3-4393-8d07-fd1626ed870e"],
Cell[CellGroupData[{
Cell[6171655, 177365, 186, 4, 27, "Input",ExpressionUUID->"63f6b7dc-72b2-4416-a5f3-579b36dbdf73"],
Cell[6171844, 177371, 396, 11, 26, "Output",ExpressionUUID->"3d04867c-4619-4007-8ae0-237ab937c146",
 CellID->263760201]
}, Open  ]],
Cell[CellGroupData[{
Cell[6172277, 177387, 282, 7, 27, "Input",ExpressionUUID->"a5b3b0ef-0dba-43b8-a21a-6af054c0b557"],
Cell[6172562, 177396, 253, 5, 26, "Output",ExpressionUUID->"d44e912e-452e-4d56-ae65-4e472b0c4ade",
 CellID->1344428859]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6172864, 177407, 123, 0, 40, "MathCaption",ExpressionUUID->"48695cc1-d373-4cb2-8f69-afb87ab3675f"],
Cell[CellGroupData[{
Cell[6173012, 177411, 142, 3, 27, "Input",ExpressionUUID->"7e0cdcbd-59b9-48f4-8a10-3cf6f28cbee0"],
Cell[6173157, 177416, 227, 3, 26, "Output",ExpressionUUID->"77a1456c-4cb6-4b05-86cc-f8a52e9172b1",
 CellID->1764624780]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6173433, 177425, 374, 7, 56, "MathCaption",ExpressionUUID->"561c965a-d8b4-4e5d-92dc-cb2ad5b532cf"],
Cell[CellGroupData[{
Cell[6173832, 177436, 159, 3, 27, "Input",ExpressionUUID->"555c3601-251f-45c4-a212-e9344ec59159"],
Cell[6173994, 177441, 325, 8, 26, "Output",ExpressionUUID->"bdeec7d8-2d57-4656-a52c-24a9cc65bada",
 CellID->498710995]
}, Open  ]],
Cell[CellGroupData[{
Cell[6174356, 177454, 230, 5, 27, "Input",ExpressionUUID->"8d180f07-ad3b-421f-b5fb-c752a1cd7614"],
Cell[6174589, 177461, 355, 9, 26, "Output",ExpressionUUID->"11bd1512-88d8-434a-b88d-cfe67fb3a67e",
 CellID->1394865606]
}, Open  ]]
}, Open  ]],
Cell[6174971, 177474, 758, 11, 74, "Text",ExpressionUUID->"3f3fad20-3df4-4b9b-82e9-a5e75a043bdd"],
Cell[CellGroupData[{
Cell[6175754, 177489, 331, 7, 56, "MathCaption",ExpressionUUID->"abc38903-6859-4247-9057-8a37a08dfff3"],
Cell[CellGroupData[{
Cell[6176110, 177500, 203, 5, 27, "Input",ExpressionUUID->"4f68c0a6-049c-4839-b19f-05eb295f232a"],
Cell[6176316, 177507, 660, 12, 22, "Print",ExpressionUUID->"b53223de-ff49-4104-bb7b-152ca4d74a2e",
 CellID->1629058196]
}, Open  ]],
Cell[CellGroupData[{
Cell[6177013, 177524, 149, 3, 27, "Input",ExpressionUUID->"a923efa3-8437-4e9b-97be-0c14629773f0"],
Cell[6177165, 177529, 425, 12, 29, "Output",ExpressionUUID->"39d51741-5883-4029-8024-e211e3a2aae9",
 CellID->1431452682]
}, Open  ]],
Cell[CellGroupData[{
Cell[6177627, 177546, 156, 3, 27, "Input",ExpressionUUID->"ad5b95ea-66c1-435b-b483-6675072a8fac"],
Cell[6177786, 177551, 215, 4, 26, "Output",ExpressionUUID->"c085b3c9-febb-4402-97c3-ac40262e273a",
 CellID->1879865030]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6178050, 177561, 208, 3, 56, "MathCaption",ExpressionUUID->"3bac2e83-233b-467f-8c5f-7bbe9fcfcbb2"],
Cell[CellGroupData[{
Cell[6178283, 177568, 143, 3, 27, "Input",ExpressionUUID->"d633dfd4-6fa3-4a8b-bd31-9392a1a84b6c"],
Cell[6178429, 177573, 970, 30, 32, "Output",ExpressionUUID->"a4e661b8-c007-4a3f-8683-021868c90906",
 CellID->870884032]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6179448, 177609, 116, 0, 40, "MathCaption",ExpressionUUID->"4f495177-a20e-44d3-8860-1e90bcc9e2bb"],
Cell[CellGroupData[{
Cell[6179589, 177613, 329, 10, 27, "Input",ExpressionUUID->"d95aa074-b024-4322-ae51-056821f3113d"],
Cell[6179921, 177625, 1040, 32, 32, "Output",ExpressionUUID->"a0008b71-fdd7-4141-ad32-c4b2562f8741",
 CellID->1799799912]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6181010, 177663, 97, 0, 40, "MathCaption",ExpressionUUID->"3ccf19cc-4caf-45b3-a0c9-25d8c7b6bdf2"],
Cell[CellGroupData[{
Cell[6181132, 177667, 147, 3, 27, "Input",ExpressionUUID->"840f1352-b6dc-4248-9151-423ae53da654"],
Cell[6181282, 177672, 573, 12, 22, "Print",ExpressionUUID->"32af6adb-383f-4ed0-9a6f-aa9dbc00eb5a",
 CellID->1735349372]
}, Open  ]]
}, Open  ]],
Cell[6181882, 177688, 342, 5, 40, "Text",ExpressionUUID->"a26bdf63-9b31-4fc3-b789-542f5e191297"]
}, Closed]],
Cell[CellGroupData[{
Cell[6182261, 177698, 107, 0, 39, "Subsection",ExpressionUUID->"03843fde-55e9-4cb0-a90f-902509ec21a5"],
Cell[6182371, 177700, 468, 8, 57, "Text",ExpressionUUID->"d6c263a2-c6a2-4478-a3a1-bb4f22d39203"],
Cell[6182842, 177710, 709, 20, 58, "DefinitionBox",ExpressionUUID->"45705199-d5cc-4326-897b-2f5b1970316d"],
Cell[6183554, 177732, 93, 0, 23, "Text",ExpressionUUID->"4db538bd-45ab-4414-a536-90cdc3680279"],
Cell[CellGroupData[{
Cell[6183672, 177736, 284, 8, 40, "MathCaption",ExpressionUUID->"cadd3ec2-22b1-4f48-9e6f-635af6c69cb5"],
Cell[CellGroupData[{
Cell[6183981, 177748, 154, 3, 27, "Input",ExpressionUUID->"1128b329-76d7-4dff-8110-f5fed822ca13"],
Cell[6184138, 177753, 198, 3, 26, "Output",ExpressionUUID->"2474e17a-645a-4363-b0cb-adecb2d586c3",
 CellID->753785169]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6184385, 177762, 177, 5, 40, "MathCaption",ExpressionUUID->"50298c0f-da86-47a0-865c-3694fbe3769e"],
Cell[CellGroupData[{
Cell[6184587, 177771, 150, 3, 27, "Input",ExpressionUUID->"05460950-b2f1-43f6-bf87-289669da142a"],
Cell[6184740, 177776, 195, 3, 26, "Output",ExpressionUUID->"28b8a97b-193a-4dd1-a6f8-9619532bf359",
 CellID->160295255]
}, Open  ]],
Cell[CellGroupData[{
Cell[6184972, 177784, 162, 3, 27, "Input",ExpressionUUID->"c73bf1cb-fdd6-42d0-b218-319c6f218c84"],
Cell[6185137, 177789, 214, 4, 26, "Output",ExpressionUUID->"00343745-6f4c-40b8-883f-bfed8bcadeee",
 CellID->379308330]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6185400, 177799, 171, 3, 40, "MathCaption",ExpressionUUID->"faf4806f-11c4-49db-a401-accaa6a56242"],
Cell[CellGroupData[{
Cell[6185596, 177806, 155, 3, 27, "Input",ExpressionUUID->"65fcc3e7-de6e-470b-95c6-a7b9d0749e44"],
Cell[6185754, 177811, 644, 20, 29, "Output",ExpressionUUID->"9febfc27-e1e6-4d05-bf83-23874d6454c1",
 CellID->586021695]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6186447, 177837, 451, 12, 56, "MathCaption",ExpressionUUID->"1b7d5cea-1eaf-4f48-8c77-c941dbee9f82"],
Cell[CellGroupData[{
Cell[6186923, 177853, 145, 3, 27, "Input",ExpressionUUID->"37bdcb9b-b3ff-4e47-ade7-14dc7a0b7fbe"],
Cell[6187071, 177858, 238, 5, 26, "Output",ExpressionUUID->"97a1ee0d-d465-4b5f-a10a-6e8b228525b8",
 CellID->1406927212]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6187358, 177869, 165, 3, 40, "MathCaption",ExpressionUUID->"20d7e44e-768a-4cf8-af29-7a7943bdfca8"],
Cell[CellGroupData[{
Cell[6187548, 177876, 191, 5, 27, "Input",ExpressionUUID->"ec1fc1de-b8aa-468a-bb8c-c8ad7f032ae6"],
Cell[6187742, 177883, 1042, 31, 32, "Output",ExpressionUUID->"1ade713b-68d7-47ba-be75-ec2e22efdf8a",
 CellID->827479770]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6188833, 177920, 190, 5, 40, "MathCaption",ExpressionUUID->"2ec69836-96a6-499e-9e03-91c87b3734b2"],
Cell[CellGroupData[{
Cell[6189048, 177929, 155, 3, 27, "Input",ExpressionUUID->"306db61e-2be1-411d-a9cf-de47af112275"],
Cell[6189206, 177934, 204, 3, 26, "Output",ExpressionUUID->"a8b3b216-e209-4222-a24c-29c6d06c9055",
 CellID->1424869359]
}, Open  ]]
}, Open  ]],
Cell[6189437, 177941, 753, 27, 41, "MathCaption",ExpressionUUID->"6889b5f5-4ca1-4dc3-a66c-8fb55106be38"],
Cell[CellGroupData[{
Cell[6190215, 177972, 145, 2, 23, "Text",ExpressionUUID->"29912bc3-4150-4410-9013-d46b7cffafa6"],
Cell[CellGroupData[{
Cell[6190385, 177978, 163, 3, 27, "Input",ExpressionUUID->"3d14eb67-b1e4-4136-aaff-b1a0a40f9f3b"],
Cell[6190551, 177983, 214, 4, 26, "Output",ExpressionUUID->"0e9b37e1-2796-4fa2-a311-92d8a5645649",
 CellID->142462808]
}, Open  ]],
Cell[CellGroupData[{
Cell[6190802, 177992, 156, 3, 27, "Input",ExpressionUUID->"49c9a0c9-92c9-41de-b7b9-250df8da171d"],
Cell[6190961, 177997, 693, 22, 33, "Output",ExpressionUUID->"e862672a-e774-4846-bc24-55dd284c7d01",
 CellID->1400460588]
}, Open  ]],
Cell[CellGroupData[{
Cell[6191691, 178024, 145, 3, 27, "Input",ExpressionUUID->"726a4856-180a-421b-be7f-cab56a80a581"],
Cell[6191839, 178029, 238, 5, 26, "Output",ExpressionUUID->"44c6da18-a252-456d-93a7-f35ad52bf1c5",
 CellID->1231328694]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6192126, 178040, 118, 0, 40, "MathCaption",ExpressionUUID->"c90df122-2231-4ea4-a904-e743d4b70783"],
Cell[CellGroupData[{
Cell[6192269, 178044, 193, 5, 27, "Input",ExpressionUUID->"29ad5e6c-8ffd-4a23-ba17-47f769eede72"],
Cell[6192465, 178051, 1048, 31, 32, "Output",ExpressionUUID->"1d6b3067-c12f-4f4c-ba90-8fca6d64ccbf",
 CellID->510434630]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6193562, 178088, 491, 12, 56, "MathCaption",ExpressionUUID->"3472033a-0c5e-4a94-8d72-8ea3751acad3"],
Cell[CellGroupData[{
Cell[6194078, 178104, 221, 6, 27, "Input",ExpressionUUID->"1e969b38-8057-49a1-a92a-37877fcf2e0d"],
Cell[6194302, 178112, 1177, 35, 32, "Output",ExpressionUUID->"eae09354-afed-4af6-b625-95dbeca41c51",
 CellID->59630773]
}, Open  ]],
Cell[CellGroupData[{
Cell[6195516, 178152, 141, 3, 27, "Input",ExpressionUUID->"8e7d3a41-2d46-439f-be98-cd56d80839af"],
Cell[6195660, 178157, 236, 3, 26, "Output",ExpressionUUID->"a8c7059e-a3f4-4925-844c-141d6ff1f2f2",
 CellID->420825556]
}, Open  ]]
}, Open  ]],
Cell[6195923, 178164, 198, 3, 23, "Text",ExpressionUUID->"dee68c3d-7afb-4fd7-8bf0-006e05d25cc8"],
Cell[6196124, 178169, 192, 3, 23, "Text",ExpressionUUID->"f5e6c6df-c8e9-4911-8a90-82ecb1269ee0"],
Cell[CellGroupData[{
Cell[6196341, 178176, 219, 5, 40, "MathCaption",ExpressionUUID->"b930d27e-52a3-4fcf-9f3f-065462436ea6"],
Cell[CellGroupData[{
Cell[6196585, 178185, 357, 9, 27, "Input",ExpressionUUID->"87bd98ea-07f4-411d-9769-6c04b5cf0013"],
Cell[6196945, 178196, 257, 5, 26, "Output",ExpressionUUID->"c65bce88-bebf-4a61-b130-09a3be1cdaaa",
 CellID->1531935448]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6197251, 178207, 120, 0, 40, "MathCaption",ExpressionUUID->"e6cc5a2f-e8be-4e11-965c-6090528fc675"],
Cell[CellGroupData[{
Cell[6197396, 178211, 360, 9, 27, "Input",ExpressionUUID->"8a50fbb5-8e0d-4b97-9e62-5c7a554fec12"],
Cell[6197759, 178222, 259, 5, 26, "Output",ExpressionUUID->"5dc54a68-6c07-43be-a66d-550c92623445",
 CellID->1718202283]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6198067, 178233, 408, 9, 56, "MathCaption",ExpressionUUID->"f3c97a73-0044-40ed-8d8d-8161d51ac56b"],
Cell[CellGroupData[{
Cell[6198500, 178246, 357, 9, 27, "Input",ExpressionUUID->"43633553-75a9-47fd-a86d-03a24d5b2390"],
Cell[6198860, 178257, 262, 5, 26, "Output",ExpressionUUID->"1e866e49-535f-43f9-b93a-a469b3cf61c2",
 CellID->168189705]
}, Open  ]],
Cell[CellGroupData[{
Cell[6199159, 178267, 340, 10, 27, "Input",ExpressionUUID->"4e9979ed-9b17-43db-b5bb-783e3f55f8bd"],
Cell[6199502, 178279, 973, 29, 32, "Output",ExpressionUUID->"aac8e908-cb32-4077-9083-c2779798d895",
 CellID->1060894942]
}, Open  ]],
Cell[CellGroupData[{
Cell[6200512, 178313, 339, 10, 27, "Input",ExpressionUUID->"e7ffb67d-2496-4879-9146-c414099cc46b"],
Cell[6200854, 178325, 2139, 64, 32, "Output",ExpressionUUID->"c073d59b-2570-4c0a-ac10-2c28ad428ce8",
 CellID->931757711]
}, Open  ]]
}, Open  ]],
Cell[6203020, 178393, 164, 2, 40, "MathCaption",ExpressionUUID->"1440df59-5266-4938-a972-8e43fd01f697"],
Cell[CellGroupData[{
Cell[6203209, 178399, 274, 8, 40, "MathCaption",ExpressionUUID->"e76b60c5-f3e9-42b3-a982-75cfadaa353a"],
Cell[CellGroupData[{
Cell[6203508, 178411, 360, 9, 27, "Input",ExpressionUUID->"98c8a931-a040-44d7-afe6-dbf34c75f1e2"],
Cell[6203871, 178422, 259, 5, 26, "Output",ExpressionUUID->"f7d5e6d8-862e-4783-b234-4c7508460b02",
 CellID->1748009716]
}, Open  ]],
Cell[CellGroupData[{
Cell[6204167, 178432, 341, 10, 27, "Input",ExpressionUUID->"083a860e-8575-46c4-9c46-2ab50b63923b"],
Cell[6204511, 178444, 2189, 65, 32, "Output",ExpressionUUID->"64e4e4a6-b5ce-413e-aded-219769116212",
 CellID->779430251]
}, Open  ]],
Cell[CellGroupData[{
Cell[6206737, 178514, 342, 10, 27, "Input",ExpressionUUID->"6bb8f5dc-9f93-4e9a-81c4-5e34b8c338fe"],
Cell[6207082, 178526, 971, 29, 32, "Output",ExpressionUUID->"e09b9507-a6b1-4007-865c-94c390a8a556",
 CellID->1247137483]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6208114, 178562, 96, 0, 39, "Subsection",ExpressionUUID->"e2123d25-7e55-4223-968d-a6dce190ac14"],
Cell[6208213, 178564, 827, 18, 103, "Text",ExpressionUUID->"cde23a58-c689-42fb-87fd-3be3baef3cfa"],
Cell[6209043, 178584, 399, 9, 41, "DefinitionBox",ExpressionUUID->"b65fb66c-aead-4a20-b827-77aa136dcc98"],
Cell[6209445, 178595, 86, 0, 23, "Text",ExpressionUUID->"e5727659-7120-47f5-8e59-bdf44f6a1500"],
Cell[CellGroupData[{
Cell[6209556, 178599, 154, 2, 40, "MathCaption",ExpressionUUID->"3f481555-13dc-4c06-9240-c0c1e3ff538b"],
Cell[CellGroupData[{
Cell[6209735, 178605, 243, 7, 27, "Input",ExpressionUUID->"0098a684-2717-4a37-b653-b886c8ef9f4f"],
Cell[6209981, 178614, 713, 14, 22, "Print",ExpressionUUID->"8dc0f115-e04d-4c8b-a377-962eff512748",
 CellID->1425323066]
}, Open  ]],
Cell[CellGroupData[{
Cell[6210731, 178633, 156, 3, 27, "Input",ExpressionUUID->"67d2bd67-c3da-412e-9dfb-2571f0952926"],
Cell[6210890, 178638, 216, 4, 26, "Output",ExpressionUUID->"29cf62fc-519f-4877-82d3-37b24eddc6c4",
 CellID->628488876]
}, Open  ]],
Cell[CellGroupData[{
Cell[6211143, 178647, 149, 3, 27, "Input",ExpressionUUID->"59bace9d-ef9a-4cc2-a2ad-159c442e703f"],
Cell[6211295, 178652, 588, 18, 26, "Output",ExpressionUUID->"cff1484f-ebb7-413f-b292-337e88c794dd",
 CellID->452688226]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6211932, 178676, 113, 0, 40, "MathCaption",ExpressionUUID->"2cc75d7c-8646-4349-bb36-8ea96a499ed1"],
Cell[CellGroupData[{
Cell[6212070, 178680, 222, 5, 27, "Input",ExpressionUUID->"89670031-bbf8-4b20-8577-c13757f43a5e"],
Cell[6212295, 178687, 217, 4, 26, "Output",ExpressionUUID->"deee789f-aeaf-48cc-a503-683649c85e6d",
 CellID->1903426450]
}, Open  ]],
Cell[CellGroupData[{
Cell[6212549, 178696, 215, 5, 27, "Input",ExpressionUUID->"fa289ff0-222f-4909-92e7-761f279a68da"],
Cell[6212767, 178703, 735, 23, 29, "Output",ExpressionUUID->"94688858-5f23-4824-9354-719fb7132edd",
 CellID->824102103]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6213551, 178732, 287, 8, 40, "MathCaption",ExpressionUUID->"34575066-a9ef-4bb1-85c0-a48aa87c2079"],
Cell[CellGroupData[{
Cell[6213863, 178744, 229, 5, 27, "Input",ExpressionUUID->"8a38463a-8159-4146-bcb0-7af78548c6ee"],
Cell[6214095, 178751, 212, 4, 26, "Output",ExpressionUUID->"935138ad-fcf9-4c0b-a8a9-0b1af59d48ca",
 CellID->1310595457]
}, Open  ]],
Cell[CellGroupData[{
Cell[6214344, 178760, 222, 5, 27, "Input",ExpressionUUID->"c07e384d-d93d-41fa-a6ca-a295198cd369"],
Cell[6214569, 178767, 679, 21, 29, "Output",ExpressionUUID->"bc2b2fed-76e1-47a2-8e26-d0fa9d26af93",
 CellID->2127789970]
}, Open  ]],
Cell[CellGroupData[{
Cell[6215285, 178793, 163, 3, 27, "Input",ExpressionUUID->"ee320269-9fd4-4438-b3b3-adab5173888f"],
Cell[6215451, 178798, 214, 4, 26, "Output",ExpressionUUID->"c97cfcbb-f0b0-4f11-885d-eda00c8d83ef",
 CellID->236700852]
}, Open  ]],
Cell[CellGroupData[{
Cell[6215702, 178807, 156, 3, 27, "Input",ExpressionUUID->"1f7a9b5e-bba0-436a-a56a-9f4a20c74d44"],
Cell[6215861, 178812, 640, 20, 29, "Output",ExpressionUUID->"a1ad6c8b-a8d5-4a45-8aed-82db0ca523e9",
 CellID->28232387]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6216550, 178838, 131, 0, 40, "MathCaption",ExpressionUUID->"e78f6828-c888-4ffe-9e8d-57deeddda654"],
Cell[CellGroupData[{
Cell[6216706, 178842, 327, 10, 27, "Input",ExpressionUUID->"47885fa1-ddab-488b-9055-a290e413bd04"],
Cell[6217036, 178854, 973, 29, 32, "Output",ExpressionUUID->"269511c5-0a84-4fa2-a065-8bb0c463f9bf",
 CellID->504730059]
}, Open  ]],
Cell[CellGroupData[{
Cell[6218046, 178888, 327, 10, 27, "Input",ExpressionUUID->"2488b393-9ded-485f-8f2b-ddac9d6402ce"],
Cell[6218376, 178900, 1047, 31, 32, "Output",ExpressionUUID->"d53c4b6b-82c8-4c54-8e9c-1b10d69c0d11",
 CellID->370035437]
}, Open  ]],
Cell[CellGroupData[{
Cell[6219460, 178936, 331, 10, 27, "Input",ExpressionUUID->"1f23fcbc-285d-4c96-9291-42ca1de2326a"],
Cell[6219794, 178948, 963, 29, 32, "Output",ExpressionUUID->"47c93f95-fe6e-4fb8-9d5e-4eb57f81c1ac",
 CellID->1106157631]
}, Open  ]],
Cell[CellGroupData[{
Cell[6220794, 178982, 331, 10, 27, "Input",ExpressionUUID->"5978cfa1-4ebe-4e4f-b63c-58ff27a03d71"],
Cell[6221128, 178994, 1049, 31, 32, "Output",ExpressionUUID->"6921614a-5fff-4458-a83b-f6d6ddca7765",
 CellID->451529215]
}, Open  ]]
}, Open  ]],
Cell[6222204, 179029, 421, 7, 40, "Text",ExpressionUUID->"ff59333d-4889-4982-8fb3-41431a6f3275"],
Cell[CellGroupData[{
Cell[6222650, 179040, 122, 0, 40, "MathCaption",ExpressionUUID->"276ef982-e27c-4ec7-b333-de66dd89a424"],
Cell[CellGroupData[{
Cell[6222797, 179044, 187, 4, 27, "Input",ExpressionUUID->"2f13f3c9-bea6-4d76-bf3f-7fd1481b1967"],
Cell[6222987, 179050, 398, 11, 26, "Output",ExpressionUUID->"91c4380f-04fb-4215-a083-2f9e4d3c4085",
 CellID->2105705276]
}, Open  ]],
Cell[CellGroupData[{
Cell[6223422, 179066, 180, 4, 27, "Input",ExpressionUUID->"0f37d232-d1b9-4500-85ee-ea80a848c9e0"],
Cell[6223605, 179072, 465, 13, 26, "Output",ExpressionUUID->"b1b18b94-0d4e-4ba0-8c9a-ce7379ae6e18",
 CellID->1500552726]
}, Open  ]],
Cell[CellGroupData[{
Cell[6224107, 179090, 142, 3, 27, "Input",ExpressionUUID->"54929481-42c0-496e-83fe-59e90be0b196"],
Cell[6224252, 179095, 249, 3, 26, "Output",ExpressionUUID->"60c889ca-1dfd-46e6-ba3e-1831c80e7192",
 CellID->906924670]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6224550, 179104, 161, 2, 40, "MathCaption",ExpressionUUID->"34da694f-acfe-47f9-9443-c01015d9eed2"],
Cell[CellGroupData[{
Cell[6224736, 179110, 154, 3, 27, "Input",ExpressionUUID->"1c87b923-d2a2-47cd-bbdc-73cc56912f40"],
Cell[6224893, 179115, 415, 10, 26, "Output",ExpressionUUID->"5e667112-1ad3-4a0b-b527-6954e492b61d",
 CellID->1493694214]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6225357, 179131, 466, 9, 72, "MathCaption",ExpressionUUID->"1d431271-f7fb-47b9-986f-42f7dff414f2"],
Cell[6225826, 179142, 177, 4, 27, "Input",ExpressionUUID->"37fb181d-a668-4afc-af69-994f659e0ca5"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6226040, 179151, 261, 4, 56, "MathCaption",ExpressionUUID->"461d2f03-410a-49dd-8913-93920a8bb7cb"],
Cell[CellGroupData[{
Cell[6226326, 179159, 286, 7, 27, "Input",ExpressionUUID->"969dda22-3c72-4f4a-8052-e775e87bd69d"],
Cell[6226615, 179168, 2173, 65, 32, "Output",ExpressionUUID->"c974f4b5-52e8-402a-b9b5-e5ea4e12b402",
 CellID->1323687785]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6228837, 179239, 203, 3, 40, "MathCaption",ExpressionUUID->"b0f15d98-e44c-4158-804f-c54e0d5d4d59"],
Cell[CellGroupData[{
Cell[6229065, 179246, 151, 3, 27, "Input",ExpressionUUID->"dc187a7b-6286-442a-8e2e-195d913fd752"],
Cell[6229219, 179251, 216, 4, 26, "Output",ExpressionUUID->"b0afc6ef-9d3d-4904-a98e-ee61801ee656",
 CellID->1718687517]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6229496, 179262, 90, 0, 39, "Subsection",ExpressionUUID->"e751d01c-1f5f-4abf-81d1-945766380919"],
Cell[6229589, 179264, 456, 6, 57, "Text",ExpressionUUID->"bee120b3-5dc5-4aca-9db7-59dcb14a3579"],
Cell[6230048, 179272, 282, 6, 110, "Text",ExpressionUUID->"9a32e720-3e23-49a2-9a4f-566fc30afdb3"],
Cell[6230333, 179280, 799, 20, 58, "DefinitionBox",ExpressionUUID->"e5c65272-7bfd-4be8-8488-9aee2d79f3b2"],
Cell[6231135, 179302, 81, 0, 23, "Text",ExpressionUUID->"0499d1ba-ec50-4b77-8692-67fd7e34603b"],
Cell[CellGroupData[{
Cell[6231241, 179306, 132, 0, 40, "MathCaption",ExpressionUUID->"079dc867-c043-4216-99a1-c51f6181561e"],
Cell[CellGroupData[{
Cell[6231398, 179310, 232, 6, 27, "Input",ExpressionUUID->"a6266728-9e3d-4f92-ab21-919f323fc564"],
Cell[6231633, 179318, 696, 13, 22, "Print",ExpressionUUID->"631230f8-a18e-4744-9354-ed3b14077d1b",
 CellID->75949296]
}, Open  ]],
Cell[CellGroupData[{
Cell[6232366, 179336, 171, 3, 27, "Input",ExpressionUUID->"8fc433c6-c1ef-4ed2-94be-67e3b9beded4"],
Cell[6232540, 179341, 218, 4, 26, "Output",ExpressionUUID->"1d0ba8e9-e47b-4d4b-8e57-d1ebad85190e",
 CellID->1317092768]
}, Open  ]],
Cell[CellGroupData[{
Cell[6232795, 179350, 164, 3, 27, "Input",ExpressionUUID->"7219b24b-0c5c-45fd-98ff-fe382c0f1107"],
Cell[6232962, 179355, 395, 11, 26, "Output",ExpressionUUID->"ae34b12b-8b03-458e-a72f-2baaf32c7677",
 CellID->908186115]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6233406, 179372, 194, 3, 40, "MathCaption",ExpressionUUID->"885c13e0-d84c-4c10-8638-9016e9f6213f"],
Cell[CellGroupData[{
Cell[6233625, 179379, 378, 11, 27, "Input",ExpressionUUID->"21a8de32-c5f6-47fd-9747-ed32d9af60b8"],
Cell[6234006, 179392, 1100, 33, 32, "Output",ExpressionUUID->"b56972a3-57cc-4551-af2e-cfb517d4a3c4",
 CellID->1611076884]
}, Open  ]],
Cell[CellGroupData[{
Cell[6235143, 179430, 142, 3, 27, "Input",ExpressionUUID->"934e9c58-17f1-4975-af6f-228aee6427d8"],
Cell[6235288, 179435, 240, 3, 26, "Output",ExpressionUUID->"8714cc32-6a01-46c1-a714-8a26daf15ac9",
 CellID->1604784060]
}, Open  ]],
Cell[CellGroupData[{
Cell[6235565, 179443, 241, 5, 27, "Input",ExpressionUUID->"2bcf295d-875b-457a-abe7-27e84f3ee4c5"],
Cell[6235809, 179450, 217, 4, 26, "Output",ExpressionUUID->"6a23e6d6-fbc7-4db9-a765-c1429e4ddf4e",
 CellID->621234410]
}, Open  ]],
Cell[CellGroupData[{
Cell[6236063, 179459, 234, 5, 27, "Input",ExpressionUUID->"e48abc57-98b5-449b-a5ee-86fcf0cbb998"],
Cell[6236300, 179466, 396, 11, 26, "Output",ExpressionUUID->"78262c5e-f487-48c4-b568-acf181f1c122",
 CellID->861310320]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6236745, 179483, 194, 3, 40, "MathCaption",ExpressionUUID->"c95e733e-ce02-43c2-8842-138e1d82980b"],
Cell[CellGroupData[{
Cell[6236964, 179490, 152, 3, 27, "Input",ExpressionUUID->"36f0fb6d-01ce-4524-b79b-413a4281e8f2"],
Cell[6237119, 179495, 2055, 62, 32, "Output",ExpressionUUID->"895568ff-ced3-45fe-b97a-0fa9eecba6fc",
 CellID->549252194]
}, Open  ]],
Cell[CellGroupData[{
Cell[6239211, 179562, 142, 3, 27, "Input",ExpressionUUID->"32a7b706-272a-4df2-9f46-4d764d2768fd"],
Cell[6239356, 179567, 271, 4, 26, "Output",ExpressionUUID->"e99102f0-777c-4671-9902-d6699b896742",
 CellID->85798042]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6239676, 179577, 108, 0, 40, "MathCaption",ExpressionUUID->"5647658e-3e60-4ffd-a344-88ebec0459b1"],
Cell[CellGroupData[{
Cell[6239809, 179581, 312, 9, 27, "Input",ExpressionUUID->"56c68915-1e97-4ebf-bdb2-aba7216d0d15"],
Cell[6240124, 179592, 805, 17, 22, "Print",ExpressionUUID->"90a46002-116f-4035-b9ae-f226eee2cf32",
 CellID->1199195019]
}, Open  ]],
Cell[CellGroupData[{
Cell[6240966, 179614, 613, 20, 27, "Input",ExpressionUUID->"1182a3b1-73c4-4119-afe9-9bcf881bb5b9"],
Cell[6241582, 179636, 1161, 33, 32, "Output",ExpressionUUID->"ab557008-1364-4e8f-8790-4f68f2e2300b",
 CellID->839803071]
}, Open  ]],
Cell[CellGroupData[{
Cell[6242780, 179674, 147, 3, 27, "Input",ExpressionUUID->"9ff46d8c-54fc-4940-bfaa-7795dd00bfa1"],
Cell[6242930, 179679, 3975, 121, 32, "Output",ExpressionUUID->"2a005abb-0413-4e5e-8cba-46b2ddd9e8f2",
 CellID->229621416]
}, Open  ]],
Cell[CellGroupData[{
Cell[6246942, 179805, 151, 3, 27, "Input",ExpressionUUID->"4694b36f-a1f7-46d9-9b65-6b7dd7056ddb"],
Cell[6247096, 179810, 215, 4, 26, "Output",ExpressionUUID->"fcaea77c-42a3-45b2-a5bc-e1cab274cf4a",
 CellID->972789631]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6247360, 179820, 281, 4, 56, "MathCaption",ExpressionUUID->"5bd19bdc-74ff-4ccf-ab15-6b974793c6b9"],
Cell[CellGroupData[{
Cell[6247666, 179828, 300, 8, 27, "Input",ExpressionUUID->"10bcd954-2732-4867-8dda-3b9a24d2e197"],
Cell[6247969, 179838, 794, 16, 22, "Print",ExpressionUUID->"d832d0a8-dfc9-4207-a597-880926cc2c7b",
 CellID->488810969]
}, Open  ]],
Cell[CellGroupData[{
Cell[6248800, 179859, 471, 14, 27, "Input",ExpressionUUID->"b595eb23-9765-4210-8784-b13d3f684c89"],
Cell[6249274, 179875, 1183, 33, 32, "Output",ExpressionUUID->"29e2653d-9db0-4ba5-a529-f91f91efde3e",
 CellID->1225988727]
}, Open  ]],
Cell[CellGroupData[{
Cell[6250494, 179913, 147, 3, 27, "Input",ExpressionUUID->"6f842fa8-6bbd-4b1c-9f5a-9ad521d0e3f7"],
Cell[6250644, 179918, 3283, 98, 32, "Output",ExpressionUUID->"f10efa48-1636-4bac-9385-46ea2b637120",
 CellID->391383441]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6253976, 180022, 127, 0, 40, "MathCaption",ExpressionUUID->"9ed35e7d-7a7e-48c7-a24e-6ece7ddbefdd"],
Cell[CellGroupData[{
Cell[6254128, 180026, 209, 6, 27, "Input",ExpressionUUID->"40778bba-a9f0-46d9-97ab-42c53497a8c0"],
Cell[6254340, 180034, 1318, 37, 32, "Output",ExpressionUUID->"499b4a89-84a4-4532-a2ea-068253dd2238",
 CellID->1664870535]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6255719, 180078, 123, 0, 39, "Subsection",ExpressionUUID->"bda7e719-a98a-45b8-9a0b-6f778f5b88b8"],
Cell[6255845, 180080, 473, 7, 86, "Text",ExpressionUUID->"dd56b798-3eaa-4e25-9de8-e0b7a4c21c0d"],
Cell[6256321, 180089, 280, 5, 81, "Text",ExpressionUUID->"9d626024-f7d1-4e0a-8928-d31ab6c9e7e9"],
Cell[6256604, 180096, 409, 9, 41, "DefinitionBox",ExpressionUUID->"b30f4dd4-e479-407d-bdb1-ce2f6a5e0584"],
Cell[6257016, 180107, 83, 0, 23, "Text",ExpressionUUID->"d9f9d2e4-a3bc-4761-a551-aec3903d1a43"],
Cell[CellGroupData[{
Cell[6257124, 180111, 186, 5, 40, "MathCaption",ExpressionUUID->"4e1ea5ae-1733-45d3-82ad-a2429be305e6"],
Cell[CellGroupData[{
Cell[6257335, 180120, 198, 5, 27, "Input",ExpressionUUID->"d79ca26c-341a-4c66-9da7-452e86abe853"],
Cell[6257536, 180127, 647, 12, 22, "Print",ExpressionUUID->"f4dea6df-6f73-472b-abe9-38b265bb31bf",
 CellID->25109116]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6258232, 180145, 332, 9, 42, "MathCaption",ExpressionUUID->"b30470ed-2aa6-40fa-a092-f8079169da38"],
Cell[CellGroupData[{
Cell[6258589, 180158, 211, 5, 27, "Input",ExpressionUUID->"ed2917de-926f-422b-8fa2-0089a2fd1a9f"],
Cell[6258803, 180165, 1139, 35, 32, "Output",ExpressionUUID->"1286c9b1-4a47-4a4d-9928-51d45608f791",
 CellID->542313864]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6259991, 180206, 122, 0, 40, "MathCaption",ExpressionUUID->"824f90eb-5053-4cba-a2bb-57875ea504c3"],
Cell[CellGroupData[{
Cell[6260138, 180210, 181, 4, 27, "Input",ExpressionUUID->"1c897e38-2c96-46a5-b72c-34187b632bc4"],
Cell[6260322, 180216, 426, 12, 29, "Output",ExpressionUUID->"efa39507-59db-49d4-8502-ecf997ed9134",
 CellID->1397078346]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6260797, 180234, 163, 5, 40, "MathCaption",ExpressionUUID->"e6daddfd-350e-4819-8737-9f0a9248c6bb"],
Cell[CellGroupData[{
Cell[6260985, 180243, 152, 3, 27, "Input",ExpressionUUID->"3f5edd83-a7ef-4dc9-9614-10ab45a33d33"],
Cell[6261140, 180248, 217, 4, 26, "Output",ExpressionUUID->"ff57b13f-5e64-4bef-a9a0-c4642fe6ce3f",
 CellID->236134905]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6261406, 180258, 169, 5, 40, "MathCaption",ExpressionUUID->"52986d2f-4a3c-489d-b91f-b00279b96000"],
Cell[CellGroupData[{
Cell[6261600, 180267, 256, 7, 27, "Input",ExpressionUUID->"6ba3acd8-7554-4a4d-ba89-645b1fb4c659"],
Cell[6261859, 180276, 1864, 58, 32, "Output",ExpressionUUID->"636aadad-80b9-4a45-b617-757d99fd3942",
 CellID->556348291]
}, Open  ]],
Cell[CellGroupData[{
Cell[6263760, 180339, 142, 3, 27, "Input",ExpressionUUID->"80a91cbc-46dd-4659-b9bf-d63e1fc1d27e"],
Cell[6263905, 180344, 229, 3, 26, "Output",ExpressionUUID->"91b915e1-ec5e-4ace-aa94-a39e28d46290",
 CellID->407674122]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6264183, 180353, 193, 5, 40, "MathCaption",ExpressionUUID->"ed019e62-1097-46a1-8145-ee4b56e5e4c1"],
Cell[CellGroupData[{
Cell[6264401, 180362, 225, 6, 27, "Input",ExpressionUUID->"e50811c0-861f-4218-9844-1b7ab1a5e54f"],
Cell[6264629, 180370, 686, 13, 22, "Print",ExpressionUUID->"3ec5a9a9-a20d-45a5-8123-0117c6c9af86",
 CellID->1776391051]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6265364, 180389, 192, 3, 40, "MathCaption",ExpressionUUID->"fbc82e85-5edf-4a1b-8b99-ab412feadf41"],
Cell[CellGroupData[{
Cell[6265581, 180396, 236, 6, 27, "Input",ExpressionUUID->"69367213-5861-4bac-91e5-7b2b2809fc7d"],
Cell[6265820, 180404, 1153, 35, 32, "Output",ExpressionUUID->"c3df70a0-c899-484e-b64c-c63742e69637",
 CellID->654170841]
}, Open  ]],
Cell[CellGroupData[{
Cell[6267010, 180444, 181, 4, 27, "Input",ExpressionUUID->"204b4243-d566-4840-80f7-2566975d2b48"],
Cell[6267194, 180450, 486, 14, 29, "Output",ExpressionUUID->"46df10fa-26b8-470e-a72b-52f86b5b7283",
 CellID->1621705353]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6267729, 180470, 168, 3, 40, "MathCaption",ExpressionUUID->"db439dca-d81f-483c-8255-1cfb7cdc6046"],
Cell[CellGroupData[{
Cell[6267922, 180477, 279, 7, 27, "Input",ExpressionUUID->"7297081d-8201-4ecb-a0c3-c5379b0af18c"],
Cell[6268204, 180486, 2017, 62, 32, "Output",ExpressionUUID->"aaf1ffbc-1e73-4e2c-93a1-4b3d6f5282b2",
 CellID->1329065620]
}, Open  ]],
Cell[CellGroupData[{
Cell[6270258, 180553, 142, 3, 27, "Input",ExpressionUUID->"6ee39914-f67a-4068-a1a1-0db9c55d9e4a"],
Cell[6270403, 180558, 263, 4, 26, "Output",ExpressionUUID->"66792c29-cfa1-4653-ac25-d39bb8303d94",
 CellID->1518038754]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6270727, 180569, 170, 4, 39, "Subsection",ExpressionUUID->"711acf14-f170-4019-9d21-08f27d71786c"],
Cell[6270900, 180575, 530, 10, 57, "Text",ExpressionUUID->"17738ee1-db9d-4b76-a748-0304da15a10f"],
Cell[6271433, 180587, 116, 3, 52, "Text",ExpressionUUID->"e806c476-060b-415f-af0a-ba300f6f747f"],
Cell[6271552, 180592, 814, 20, 58, "DefinitionBox",ExpressionUUID->"eea02c1d-07e3-43c7-b34f-38f37ec1fe13"],
Cell[6272369, 180614, 84, 0, 23, "Text",ExpressionUUID->"fbf379b0-9033-4d14-ad03-2a9983eda162"],
Cell[CellGroupData[{
Cell[6272478, 180618, 394, 11, 56, "MathCaption",ExpressionUUID->"0b94e55b-670c-42db-abc7-7ffe90584c48"],
Cell[CellGroupData[{
Cell[6272897, 180633, 290, 8, 27, "Input",ExpressionUUID->"086f204b-8155-4f11-8c85-5b70e67735e1"],
Cell[6273190, 180643, 1124, 34, 32, "Output",ExpressionUUID->"05d5bd4c-5012-4f75-a903-11e38fdda88a",
 CellID->308568836]
}, Open  ]],
Cell[CellGroupData[{
Cell[6274351, 180682, 142, 3, 27, "Input",ExpressionUUID->"a0a02fe6-5464-4e8c-a74a-9673326568ca"],
Cell[6274496, 180687, 229, 3, 26, "Output",ExpressionUUID->"0266dbc8-15e8-4dc2-b5bd-1c568bbadf90",
 CellID->659725678]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6274774, 180696, 193, 5, 40, "MathCaption",ExpressionUUID->"c505edab-16c2-4523-8ca9-916f1e7b153d"],
Cell[CellGroupData[{
Cell[6274992, 180705, 151, 3, 27, "Input",ExpressionUUID->"8c994e47-d86a-418f-a6a7-e12772c6af02"],
Cell[6275146, 180710, 216, 4, 26, "Output",ExpressionUUID->"1448fe62-2317-475c-9ae5-99798642f7ec",
 CellID->1940298987]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6275411, 180720, 159, 2, 40, "MathCaption",ExpressionUUID->"e5d03bf1-e7d9-4afd-ad41-b453154227d1"],
Cell[CellGroupData[{
Cell[6275595, 180726, 151, 3, 27, "Input",ExpressionUUID->"33d134b7-8b39-401a-96f4-a46dc866336f"],
Cell[6275749, 180731, 2279, 69, 32, "Output",ExpressionUUID->"5b227170-8aed-4f45-b558-0490b06a3943",
 CellID->631203911]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6278077, 180806, 165, 3, 40, "MathCaption",ExpressionUUID->"b9c24005-d3ab-4c5e-834f-74b9fc3ad8ff"],
Cell[CellGroupData[{
Cell[6278267, 180813, 288, 8, 27, "Input",ExpressionUUID->"bb0540f0-083b-437d-8486-b499e5bef89d"],
Cell[6278558, 180823, 1182, 36, 32, "Output",ExpressionUUID->"047d96ac-1c16-4363-81df-20defa909573",
 CellID->198240297]
}, Open  ]],
Cell[CellGroupData[{
Cell[6279777, 180864, 147, 3, 27, "Input",ExpressionUUID->"3e32c0ee-c9a0-4ad0-9ab8-04e8681d93cf"],
Cell[6279927, 180869, 2279, 69, 32, "Output",ExpressionUUID->"aedabdd6-d817-4ff4-a825-9a5a4cbda3e2",
 CellID->808437265]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6282255, 180944, 177, 3, 40, "MathCaption",ExpressionUUID->"eefc2bee-edec-400b-a784-657b769e61af"],
Cell[CellGroupData[{
Cell[6282457, 180951, 236, 6, 27, "Input",ExpressionUUID->"a7c6e292-bb74-4ed9-8f96-42cc0d9252ab"],
Cell[6282696, 180959, 1086, 33, 32, "Output",ExpressionUUID->"36ab1a8f-8e6d-4e70-a1c2-ac6e95b3aeec",
 CellID->1548033283]
}, Open  ]],
Cell[CellGroupData[{
Cell[6283819, 180997, 150, 3, 27, "Input",ExpressionUUID->"aa365a1e-e22e-4fae-a7b3-d033ee86f96e"],
Cell[6283972, 181002, 2104, 64, 32, "Output",ExpressionUUID->"ac8cad06-efb9-4f68-8865-78bef531a32d",
 CellID->1301376685]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6286125, 181072, 129, 0, 40, "MathCaption",ExpressionUUID->"d4f93538-9bc2-4146-80dd-ebcdf8057046"],
Cell[CellGroupData[{
Cell[6286279, 181076, 254, 7, 27, "Input",ExpressionUUID->"81b74fa6-97d3-4464-84e5-d9b491461ac2"],
Cell[6286536, 181085, 741, 15, 22, "Print",ExpressionUUID->"fd886551-613c-4f2a-875b-5bc839339756",
 CellID->1632227167]
}, Open  ]],
Cell[CellGroupData[{
Cell[6287314, 181105, 326, 9, 27, "Input",ExpressionUUID->"cd908dd3-9edf-4cad-bcd4-ae962cee8b0e"],
Cell[6287643, 181116, 1125, 33, 32, "Output",ExpressionUUID->"90e75fda-5add-4af1-b781-b97fe5f93293",
 CellID->1270626084]
}, Open  ]],
Cell[CellGroupData[{
Cell[6288805, 181154, 181, 4, 27, "Input",ExpressionUUID->"9a2b0126-e2c0-4868-a358-15549a3d0a12"],
Cell[6288989, 181160, 759, 24, 26, "Output",ExpressionUUID->"2971e489-240e-43a7-8a1f-2e719c0b7e40",
 CellID->1195123764]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6289809, 181191, 170, 4, 39, "Subsection",ExpressionUUID->"e857e651-01ac-4e24-b76e-147328f407c5"],
Cell[CellGroupData[{
Cell[6290004, 181199, 140, 1, 40, "MathCaption",ExpressionUUID->"a15a2f25-001b-46c7-8214-eea650dca245"],
Cell[CellGroupData[{
Cell[6290169, 181204, 119, 2, 27, "Input",ExpressionUUID->"4da653f6-c7fd-4fda-9a56-b7529de68adb"],
Cell[6290291, 181208, 286, 6, 26, "Output",ExpressionUUID->"b7502431-c9fd-40a1-a337-3e27571c22d3",
 CellID->1632433000]
}, Open  ]],
Cell[CellGroupData[{
Cell[6290614, 181219, 118, 2, 27, "Input",ExpressionUUID->"e6398199-3bf8-43f2-9649-d2d3b2b2f31b"],
Cell[6290735, 181223, 419, 8, 26, "Output",ExpressionUUID->"12dbf9aa-8d35-434d-8221-12e4516db6f1",
 CellID->1439801107]
}, Open  ]],
Cell[CellGroupData[{
Cell[6291191, 181236, 120, 2, 27, "Input",ExpressionUUID->"d97a2271-4ee2-41d9-89ec-30eaff22daa1"],
Cell[6291314, 181240, 250, 5, 26, "Output",ExpressionUUID->"473a1b1b-8a4f-4eb5-97da-a2085e42b99e",
 CellID->428150384]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6291613, 181251, 218, 3, 56, "MathCaption",ExpressionUUID->"4cfb0dff-5a8f-42c2-8d14-2cdcb9829a47"],
Cell[CellGroupData[{
Cell[6291856, 181258, 222, 6, 27, "Input",ExpressionUUID->"71764803-344b-49ce-a9f6-b630dd8724c7"],
Cell[CellGroupData[{
Cell[6292103, 181268, 628, 12, 22, "Print",ExpressionUUID->"af4e4d65-8f26-4386-97a6-797444cd069d",
 CellID->1593606993],
Cell[6292734, 181282, 628, 12, 22, "Print",ExpressionUUID->"a5a1a93d-695d-4b5d-9041-098b367266af",
 CellID->336685967],
Cell[6293365, 181296, 604, 12, 22, "Print",ExpressionUUID->"a3bb7982-d6b5-44be-b743-129d1ccb4de3",
 CellID->61637524],
Cell[6293972, 181310, 587, 12, 22, "Print",ExpressionUUID->"1ec96e0f-b280-4de8-9245-93f5e6c1d692",
 CellID->854748014],
Cell[6294562, 181324, 627, 12, 22, "Print",ExpressionUUID->"7d92fc12-b0ea-4a13-b417-e29a11a48a33",
 CellID->1231756515],
Cell[6295192, 181338, 631, 12, 22, "Print",ExpressionUUID->"eed3a22e-a76f-414a-9def-c6913b08a693",
 CellID->389390525],
Cell[6295826, 181352, 616, 12, 22, "Print",ExpressionUUID->"f9ae2128-f6b1-4a64-a3d8-d4885e560315",
 CellID->1103881266],
Cell[6296445, 181366, 587, 12, 22, "Print",ExpressionUUID->"c5ca0913-e848-485b-8f1f-2266ed1a7c02",
 CellID->821211229]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6297081, 181384, 169, 4, 27, "Input",ExpressionUUID->"cf7982e4-0ab3-4a69-bced-f629de53f469"],
Cell[CellGroupData[{
Cell[6297275, 181392, 572, 12, 22, "Print",ExpressionUUID->"4f80955b-71cb-4fe1-a9a0-77ecda6160ea",
 CellID->1731263231],
Cell[6297850, 181406, 600, 12, 22, "Print",ExpressionUUID->"87bff5d7-b1f1-4394-ade4-9e247f031cbb",
 CellID->1085190214],
Cell[6298453, 181420, 603, 12, 22, "Print",ExpressionUUID->"a7261e45-1959-4b67-bf9e-f590c18790ba",
 CellID->213312727],
Cell[6299059, 181434, 601, 12, 22, "Print",ExpressionUUID->"6edba8d1-9ba6-4cd8-be59-8c97bd6f1d46",
 CellID->996795846],
Cell[6299663, 181448, 571, 12, 22, "Print",ExpressionUUID->"af170818-9f12-4652-a43d-0c8b49d1eb5f",
 CellID->627853475],
Cell[6300237, 181462, 586, 12, 22, "Print",ExpressionUUID->"4fb79b47-0044-4c97-b28f-8323426d162a",
 CellID->1746196614],
Cell[6300826, 181476, 571, 12, 22, "Print",ExpressionUUID->"dff29a03-c4f8-4cb9-a223-7ccbdeac9460",
 CellID->814971596]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6301446, 181494, 173, 4, 27, "Input",ExpressionUUID->"632e0e4b-2f2c-4209-9088-0a061375dcb3"],
Cell[CellGroupData[{
Cell[6301644, 181502, 745, 18, 22, "Print",ExpressionUUID->"0c13e229-f040-4f88-9406-271535100694",
 CellID->1654220525],
Cell[6302392, 181522, 580, 12, 22, "Print",ExpressionUUID->"466f28e9-ff43-4585-b18b-8ef6c8a68146",
 CellID->1161069613],
Cell[6302975, 181536, 745, 18, 22, "Print",ExpressionUUID->"eb6c6ad2-3cc0-4101-a31c-481a2ba068be",
 CellID->2060569049],
Cell[6303723, 181556, 580, 12, 22, "Print",ExpressionUUID->"bc18bdc3-5b70-44be-919b-a61de852b453",
 CellID->1662715021],
Cell[6304306, 181570, 745, 18, 22, "Print",ExpressionUUID->"3986f88e-87d1-4e01-aced-ace1152768af",
 CellID->46659536],
Cell[6305054, 181590, 579, 12, 22, "Print",ExpressionUUID->"e4be0448-119e-4178-8e6f-0612ce0ce957",
 CellID->375290397]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6305694, 181609, 104, 0, 40, "MathCaption",ExpressionUUID->"beef4f91-7ae3-40aa-bd16-e4e181d5b733"],
Cell[6305801, 181611, 198, 4, 27, "Input",ExpressionUUID->"0f60d2ab-aa1d-4e54-9b3a-9ff613b95fdf"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6306036, 181620, 172, 3, 40, "MathCaption",ExpressionUUID->"2e473975-d54a-4107-9fa3-1de9fa40db71"],
Cell[CellGroupData[{
Cell[6306233, 181627, 135, 3, 27, "Input",ExpressionUUID->"bd1eace3-24cd-4760-ae3d-c4ced1280bfc"],
Cell[6306371, 181632, 4881, 119, 85, "Output",ExpressionUUID->"56acd225-8c8e-4cba-878b-2a053707dcc2",
 CellID->573846044]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6311313, 181758, 100, 0, 39, "Subsection",ExpressionUUID->"bf812b76-03e7-402c-b038-c9ddefa14da8"],
Cell[CellGroupData[{
Cell[6311438, 181762, 1591, 47, 91, "Text",ExpressionUUID->"c13b7c22-44ec-422c-9ab2-14f75be11eef"],
Cell[CellGroupData[{
Cell[6313054, 181813, 2968, 80, 199, "Input",ExpressionUUID->"25097137-6aeb-472d-b5d5-8b9d7f00c425"],
Cell[6316025, 181895, 3979, 103, 250, "Output",ExpressionUUID->"1f217a17-0181-4ed3-9570-29bb4cd5a988",
 CellID->1784840896]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6320077, 182006, 918, 15, 54, "Section",ExpressionUUID->"4b0a4311-6f29-4ca4-b89d-024d36e40a11",
 CellTags->{"S5.18.1", "1.4"},
 CellID->717570716],
Cell[CellGroupData[{
Cell[6321020, 182025, 90, 0, 39, "Subsection",ExpressionUUID->"761534c8-fc5e-4f60-98a9-c5582fe712e1"],
Cell[6321113, 182027, 936, 20, 202, "Text",ExpressionUUID->"1ffec475-bdef-4e2c-b908-94aefb68a921"],
Cell[6322052, 182049, 753, 20, 58, "DefinitionBox",ExpressionUUID->"6dcfa298-e04b-4031-beeb-026f2bd330af"],
Cell[6322808, 182071, 93, 0, 23, "Text",ExpressionUUID->"f88eaf3c-9bb8-4581-9b6b-c3227f3bdb1d"],
Cell[CellGroupData[{
Cell[6322926, 182075, 168, 3, 40, "MathCaption",ExpressionUUID->"192ea1b4-8633-4601-86cb-0f91fdbcafa1"],
Cell[6323097, 182080, 163, 4, 27, "Input",ExpressionUUID->"5463ce9b-a045-42d0-be25-7079f00fc50b"],
Cell[6323263, 182086, 1356, 39, 123, "Input",ExpressionUUID->"3fc62a92-c215-41f6-a20b-9e5526bba196"],
Cell[6324622, 182127, 177, 4, 27, "Input",ExpressionUUID->"9432d1ce-43e9-47df-bdf2-1c4998c8cfe3"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6324836, 182136, 156, 2, 40, "MathCaption",ExpressionUUID->"b445bc29-d8e3-453e-9dcb-d60cdfaff1bd"],
Cell[CellGroupData[{
Cell[6325017, 182142, 322, 10, 27, "Input",ExpressionUUID->"f6514494-ae48-4b57-b507-7d706efe6e6d"],
Cell[6325342, 182154, 3116, 92, 32, "Output",ExpressionUUID->"35deabfe-96fb-46af-b4b6-b8b2c7f1080b",
 CellID->1276669088]
}, Open  ]],
Cell[CellGroupData[{
Cell[6328495, 182251, 165, 4, 27, "Input",ExpressionUUID->"7c223395-44bc-40e1-837e-f3c00fac5e14"],
Cell[6328663, 182257, 431, 9, 22, "Message",ExpressionUUID->"45cab096-4fd0-424e-abeb-a582a8677e33",
 CellID->221778279],
Cell[6329097, 182268, 422, 9, 22, "Message",ExpressionUUID->"d1508ed1-e0d7-4ffc-a376-061827ad9908",
 CellID->1543407608],
Cell[6329522, 182279, 236, 5, 26, "Output",ExpressionUUID->"5012442f-444a-43d2-8b58-32d4ca2ce7d2",
 CellID->596695543]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6329807, 182290, 243, 5, 40, "MathCaption",ExpressionUUID->"15e0a955-db0a-4080-ba25-cefa37cb0f6d"],
Cell[CellGroupData[{
Cell[6330075, 182299, 330, 10, 27, "Input",ExpressionUUID->"59bada2f-3310-4391-b913-2abe2b44af10"],
Cell[6330408, 182311, 3412, 98, 34, "Output",ExpressionUUID->"4a46b256-f786-4a5d-9bb2-705b3e6aaca7",
 CellID->1586186550]
}, Open  ]],
Cell[CellGroupData[{
Cell[6333857, 182414, 165, 4, 27, "Input",ExpressionUUID->"ef80d674-d317-446e-b03f-592cd2d3300a"],
Cell[6334025, 182420, 2052, 60, 32, "Output",ExpressionUUID->"c581f5f9-bec6-4c88-ac6e-78dea16b60fc",
 CellID->566591278]
}, Open  ]],
Cell[CellGroupData[{
Cell[6336114, 182485, 137, 3, 27, "Input",ExpressionUUID->"a201fd4f-9806-4fd7-98a3-f15cf4939eaa"],
Cell[6336254, 182490, 3006, 88, 32, "Output",ExpressionUUID->"7a7834cc-d7f8-4063-84f1-83bdaa5030d5",
 CellID->1979744497]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6339309, 182584, 293, 6, 56, "MathCaption",ExpressionUUID->"bc48e003-56bc-4675-ba58-98d5572a001e"],
Cell[CellGroupData[{
Cell[6339627, 182594, 461, 14, 27, "Input",ExpressionUUID->"a159baac-85cf-4b58-8904-db23a2b49551"],
Cell[6340091, 182610, 3580, 99, 36, "Output",ExpressionUUID->"2b1d1567-c419-466a-8796-29f3d4a0d307",
 CellID->1070771511]
}, Open  ]],
Cell[CellGroupData[{
Cell[6343708, 182714, 314, 10, 27, "Input",ExpressionUUID->"7c0ebce2-4e21-4068-bea2-0322b63e1ec4"],
Cell[6344025, 182726, 4886, 144, 32, "Output",ExpressionUUID->"33a5b4f9-adb0-4707-b041-bc5fa43ad401",
 CellID->1750465817]
}, Open  ]],
Cell[CellGroupData[{
Cell[6348948, 182875, 149, 3, 27, "Input",ExpressionUUID->"11ec3f79-f70e-4680-a2c1-e78f1372c4c7"],
Cell[6349100, 182880, 273, 6, 26, "Output",ExpressionUUID->"bf7b287d-d6a7-433f-9754-ad6678eecc56",
 CellID->1089903271]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6349422, 182892, 500, 14, 72, "MathCaption",ExpressionUUID->"157f86c5-9615-43e8-ba25-e16d4855cf46"],
Cell[CellGroupData[{
Cell[6349947, 182910, 165, 4, 27, "Input",ExpressionUUID->"70747862-6180-45b2-827a-b668a0744bd8"],
Cell[6350115, 182916, 2055, 60, 32, "Output",ExpressionUUID->"28e476d6-f2d0-4f75-95b9-53509ec5c272",
 CellID->1751303676]
}, Open  ]],
Cell[CellGroupData[{
Cell[6352207, 182981, 184, 5, 27, "Input",ExpressionUUID->"f48258d3-d6fe-47af-bcab-7798a36c2256"],
Cell[6352394, 182988, 1060, 31, 32, "Output",ExpressionUUID->"2513e7c8-6c1c-42d8-a9ea-e87ac5e011b0",
 CellID->713601693]
}, Open  ]],
Cell[CellGroupData[{
Cell[6353491, 183024, 165, 4, 27, "Input",ExpressionUUID->"ed3e9c49-3a3d-4f45-ae73-d3247ab9e19f"],
Cell[6353659, 183030, 1060, 31, 32, "Output",ExpressionUUID->"1d2b66c4-38a7-4023-aeba-6a0542363a8f",
 CellID->1271861996]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6354768, 183067, 406, 10, 56, "MathCaption",ExpressionUUID->"c2b3f63c-e4d7-4574-92f0-db770e5be306"],
Cell[CellGroupData[{
Cell[6355199, 183081, 542, 16, 27, "Input",ExpressionUUID->"344c9768-08e7-4e27-a1fb-3d55fae69edb"],
Cell[6355744, 183099, 3576, 99, 36, "Output",ExpressionUUID->"f35af55c-6b99-43c5-846d-e958b1529361",
 CellID->1602880341]
}, Open  ]],
Cell[CellGroupData[{
Cell[6359357, 183203, 149, 3, 27, "Input",ExpressionUUID->"d3d435e2-b281-47d7-9cf1-38945c4469d5"],
Cell[6359509, 183208, 266, 6, 26, "Output",ExpressionUUID->"9535121c-c622-4a9b-91de-99bc00dc5c5c",
 CellID->379773373]
}, Open  ]],
Cell[CellGroupData[{
Cell[6359812, 183219, 165, 4, 27, "Input",ExpressionUUID->"b7cda420-df12-451f-aba6-5b1156b6467f"],
Cell[6359980, 183225, 2053, 60, 32, "Output",ExpressionUUID->"0d99160b-cdde-4006-92df-8f0be2ee8d18",
 CellID->1666256820]
}, Open  ]],
Cell[CellGroupData[{
Cell[6362070, 183290, 184, 5, 27, "Input",ExpressionUUID->"170234ba-b43f-4af0-8d06-9a2f11cd835d"],
Cell[6362257, 183297, 2054, 60, 32, "Output",ExpressionUUID->"c064f791-cf9c-4c13-ad56-88ecfccf6596",
 CellID->1740720426]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6364360, 183363, 149, 2, 40, "MathCaption",ExpressionUUID->"f91379e6-6cef-4381-84cf-37d89a670974"],
Cell[CellGroupData[{
Cell[6364534, 183369, 748, 23, 27, "Input",ExpressionUUID->"814ecefb-6203-484f-9df2-5e64d5f4bdb6"],
Cell[6365285, 183394, 74171, 2009, 176, "Output",ExpressionUUID->"e9781e6d-3ed1-4abe-aff6-d599a3e21c27",
 CellID->2036193382]
}, Open  ]],
Cell[CellGroupData[{
Cell[6439493, 185408, 289, 9, 27, "Input",ExpressionUUID->"b1160f47-0d43-424a-bd44-4d5d45659466"],
Cell[6439785, 185419, 4065, 119, 32, "Output",ExpressionUUID->"6dfe6d05-d804-4a65-969d-0ce54a014ed0",
 CellID->2072497326]
}, Open  ]],
Cell[CellGroupData[{
Cell[6443887, 185543, 269, 8, 27, "Input",ExpressionUUID->"5c516a3f-c7a1-4af3-8d41-1fdb2f5c4ea5"],
Cell[6444159, 185553, 4016, 118, 32, "Output",ExpressionUUID->"21e17d17-0f69-4944-83b2-d3096d4ae0d2",
 CellID->1969517102]
}, Open  ]],
Cell[CellGroupData[{
Cell[6448212, 185676, 333, 11, 27, "Input",ExpressionUUID->"15dd5045-cfd4-4025-98e5-78deb2f2222f"],
Cell[6448548, 185689, 174, 3, 26, "Output",ExpressionUUID->"e4171b49-ed86-4d79-bf0a-8fb453dd5fc6",
 CellID->1006228165]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6448771, 185698, 244, 6, 40, "MathCaption",ExpressionUUID->"8b840f1b-4144-4c72-bc8c-6d43fcb486f0"],
Cell[CellGroupData[{
Cell[6449040, 185708, 177, 4, 27, "Input",ExpressionUUID->"2dfb92d4-1a57-413f-bd12-4609063e0a76"],
Cell[6449220, 185714, 705, 15, 22, "Print",ExpressionUUID->"2fb8372d-e71a-496f-a56a-8be6adb57665",
 CellID->2010194070]
}, Open  ]],
Cell[CellGroupData[{
Cell[6449962, 185734, 257, 8, 27, "Input",ExpressionUUID->"23ee9058-9d3d-47b9-ac84-1c1ae6626821"],
Cell[6450222, 185744, 4064, 119, 32, "Output",ExpressionUUID->"f154215c-9702-4bf5-98e0-b98d82822925",
 CellID->722037073]
}, Open  ]],
Cell[CellGroupData[{
Cell[6454323, 185868, 238, 7, 27, "Input",ExpressionUUID->"58663585-91bb-4ee2-9d38-1e3ccee6783b"],
Cell[6454564, 185877, 4018, 118, 32, "Output",ExpressionUUID->"d54daa98-1253-437b-b1bf-f7c7f1e5946c",
 CellID->1484724578]
}, Open  ]],
Cell[CellGroupData[{
Cell[6458619, 186000, 128, 3, 27, "Input",ExpressionUUID->"065675b8-2453-4073-9d8f-4da9bc6851a0"],
Cell[CellGroupData[{
Cell[6458772, 186007, 198, 3, 43, "Print",ExpressionUUID->"43e2b182-5295-45a0-97ed-b7fecafa9b13",
 CellTags->"Info10913728982189-9181623",
 CellID->1587009223],
Cell[6458973, 186012, 87167, 2211, 776, "Print",ExpressionUUID->"8248b8f0-89c7-4a25-a0f3-d2e87fb130d5",
 CellTags->"Info10913728982189-9181623",
 CellID->1117115221]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6546201, 188230, 184, 3, 40, "MathCaption",ExpressionUUID->"eaf34494-1f4c-4f40-898f-b3c02ea804bf"],
Cell[CellGroupData[{
Cell[6546410, 188237, 148, 3, 27, "Input",ExpressionUUID->"046aad13-b58e-4dbf-8fdd-57b81633cb81"],
Cell[6546561, 188242, 550, 12, 22, "Print",ExpressionUUID->"c632a3f9-4aa1-4652-b42d-aa8b9ae9a188",
 CellID->1139610945]
}, Open  ]],
Cell[6547126, 188257, 409, 12, 27, "Input",ExpressionUUID->"1d3ba07f-bb3a-49d9-88ed-e1ff729ce904"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6547584, 188275, 101, 0, 39, "Subsection",ExpressionUUID->"4e047b77-a755-46c9-be86-b00ebd7e5b66"],
Cell[6547688, 188277, 836, 15, 132, "Text",ExpressionUUID->"68d2fc2a-30c7-4870-b309-54629143979a"],
Cell[CellGroupData[{
Cell[6548549, 188296, 91, 0, 40, "MathCaption",ExpressionUUID->"eb30f429-b4e3-4116-b398-86ec8a694030"],
Cell[CellGroupData[{
Cell[6548665, 188300, 284, 8, 27, "Input",ExpressionUUID->"3618dc1a-a6b6-4e89-8ba9-69d47f81e75a"],
Cell[6548952, 188310, 174, 3, 26, "Output",ExpressionUUID->"8c8580ce-5d3d-4e01-a05b-f9b8f0308014",
 CellID->1162546407]
}, Open  ]],
Cell[CellGroupData[{
Cell[6549163, 188318, 284, 8, 27, "Input",ExpressionUUID->"9df949b3-b28a-4890-8124-eb93d5ae66fd"],
Cell[6549450, 188328, 174, 3, 26, "Output",ExpressionUUID->"771db3a0-c230-4ee3-a241-8b809137806f",
 CellID->1333229371]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6549673, 188337, 91, 0, 40, "MathCaption",ExpressionUUID->"21328013-d11e-4ab6-a5bc-ecfeb6b27d5c"],
Cell[CellGroupData[{
Cell[6549789, 188341, 284, 8, 27, "Input",ExpressionUUID->"cffe264e-c7d9-485f-a476-0b857ec26f77"],
Cell[6550076, 188351, 1145, 32, 32, "Output",ExpressionUUID->"18aa9b5e-05d2-43d3-94ab-39b8d6470a84",
 CellID->1546588497]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6551270, 188389, 91, 0, 40, "MathCaption",ExpressionUUID->"488d9216-9806-4285-8488-71b33bd00e0b"],
Cell[CellGroupData[{
Cell[6551386, 188393, 284, 8, 27, "Input",ExpressionUUID->"0b3df27e-40c3-42c6-8cd9-b4e32a6b7a29"],
Cell[6551673, 188403, 1190, 33, 32, "Output",ExpressionUUID->"2be60db1-5075-4a8c-adad-89c5cf8ffa68",
 CellID->687942202]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6552912, 188442, 91, 0, 40, "MathCaption",ExpressionUUID->"5b59ed10-9fad-46b9-938b-56c0d32570dc"],
Cell[CellGroupData[{
Cell[6553028, 188446, 284, 8, 27, "Input",ExpressionUUID->"b5a73ad4-e5c4-493a-9c69-40421cad7874"],
Cell[6553315, 188456, 1190, 33, 32, "Output",ExpressionUUID->"152b13dc-5a3e-4fcd-ad8c-9f1ba189103d",
 CellID->519529191]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6554554, 188495, 91, 0, 40, "MathCaption",ExpressionUUID->"99177654-52ef-4afc-b38b-9f33c549f6f6"],
Cell[CellGroupData[{
Cell[6554670, 188499, 284, 8, 27, "Input",ExpressionUUID->"c3f5e872-0457-43ea-b0a6-872cd7da1f5e"],
Cell[6554957, 188509, 1144, 32, 32, "Output",ExpressionUUID->"8bcad7ae-9895-4fd3-979f-c926e8335e44",
 CellID->372594429]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6556150, 188547, 91, 0, 40, "MathCaption",ExpressionUUID->"2c972818-2c95-4f63-a8ed-29d0d7480560"],
Cell[CellGroupData[{
Cell[6556266, 188551, 388, 12, 27, "Input",ExpressionUUID->"bb9246aa-7aff-4776-93e0-16c3b8d5485d"],
Cell[6556657, 188565, 171, 3, 26, "Output",ExpressionUUID->"6e1ae10d-0cd0-45aa-b529-83498e15e90b",
 CellID->975912735]
}, Open  ]],
Cell[CellGroupData[{
Cell[6556865, 188573, 388, 12, 27, "Input",ExpressionUUID->"5e092416-608c-4704-a227-13305a8eeaaa"],
Cell[6557256, 188587, 172, 3, 26, "Output",ExpressionUUID->"c1066037-eb51-4b59-9394-ecbf39f4ec54",
 CellID->1677838393]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6557477, 188596, 92, 0, 40, "MathCaption",ExpressionUUID->"58604d53-f0cc-4811-9db8-e8b9e5d4dd21"],
Cell[CellGroupData[{
Cell[6557594, 188600, 470, 15, 27, "Input",ExpressionUUID->"10b8feeb-0d2a-4a83-bcad-f3432de20454"],
Cell[6558067, 188617, 174, 3, 26, "Output",ExpressionUUID->"53974fec-36f2-46ec-8b2b-06b41d12e4fc",
 CellID->2017726656]
}, Open  ]],
Cell[CellGroupData[{
Cell[6558278, 188625, 470, 15, 27, "Input",ExpressionUUID->"4b40eefd-0ecb-410b-bcb2-f31100811675"],
Cell[6558751, 188642, 171, 3, 26, "Output",ExpressionUUID->"24cd6c3c-8563-4097-aa74-d7d672e867cd",
 CellID->794599648]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6558971, 188651, 92, 0, 40, "MathCaption",ExpressionUUID->"3d964f36-d20f-456c-96d3-2487e55906a2"],
Cell[CellGroupData[{
Cell[6559088, 188655, 548, 18, 27, "Input",ExpressionUUID->"c86ba01e-8e21-44d8-951a-9cbdec0b7e4c"],
Cell[6559639, 188675, 174, 3, 26, "Output",ExpressionUUID->"7f0ab981-baf2-4e05-aa59-9ad89cc0e97d",
 CellID->2073260556]
}, Open  ]],
Cell[CellGroupData[{
Cell[6559850, 188683, 548, 18, 27, "Input",ExpressionUUID->"75d7f3e2-e004-4f2e-be1e-2befc57dd2dc"],
Cell[6560401, 188703, 174, 3, 26, "Output",ExpressionUUID->"d3d01595-4841-4a60-b4b7-6d2c6d6fbbd3",
 CellID->1837683274]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6560624, 188712, 92, 0, 40, "MathCaption",ExpressionUUID->"5e361936-1380-46ec-b718-4f652c419cde"],
Cell[CellGroupData[{
Cell[6560741, 188716, 429, 14, 27, "Input",ExpressionUUID->"1d4d819e-e146-4a3f-bb66-824ea453d58e"],
Cell[6561173, 188732, 171, 3, 26, "Output",ExpressionUUID->"5c9bd9fb-75f0-4efa-89ee-2e82f8d906a6",
 CellID->202417881]
}, Open  ]],
Cell[CellGroupData[{
Cell[6561381, 188740, 429, 14, 27, "Input",ExpressionUUID->"390fe4eb-29d7-4060-9b0d-226e25cd5455"],
Cell[6561813, 188756, 171, 3, 26, "Output",ExpressionUUID->"ec707050-51d8-40ee-9c53-f60d44e293ca",
 CellID->565207425]
}, Open  ]]
}, Open  ]],
Cell[6562011, 188763, 184, 3, 23, "Text",ExpressionUUID->"07f1d41e-aeb7-4a88-9f81-a5313fc05b87"],
Cell[CellGroupData[{
Cell[6562220, 188770, 163, 3, 40, "MathCaption",ExpressionUUID->"679fa4c8-be0b-4eaf-9c57-cae8b1e6038d"],
Cell[CellGroupData[{
Cell[6562408, 188777, 354, 11, 27, "Input",ExpressionUUID->"b24780f4-7683-407b-bea2-d676dbdb3a96"],
Cell[6562765, 188790, 484, 14, 40, "Output",ExpressionUUID->"f1bbe9a2-315e-4567-a080-b895614c4762",
 CellID->353549070]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6563298, 188810, 92, 0, 40, "MathCaption",ExpressionUUID->"10632beb-7aea-4694-bbae-6eb2cff4ef69"],
Cell[CellGroupData[{
Cell[6563415, 188814, 354, 11, 27, "Input",ExpressionUUID->"b7e7aee0-15e6-4880-862e-13c488c79896"],
Cell[6563772, 188827, 485, 14, 40, "Output",ExpressionUUID->"afcd036e-10b3-4657-b955-16de85725fe6",
 CellID->1155081646]
}, Open  ]]
}, Open  ]],
Cell[6564284, 188845, 912, 26, 260, "Text",ExpressionUUID->"5f143f82-3dc4-4fb4-bb96-aa402fe86fe9"],
Cell[CellGroupData[{
Cell[6565221, 188875, 117, 0, 40, "MathCaption",ExpressionUUID->"5787172f-9759-44e6-bd45-eba1ae92e0d7"],
Cell[6565341, 188877, 1128, 34, 66, "Input",ExpressionUUID->"1c0e0d19-446e-4838-8c4c-a26870c18f70"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6566506, 188916, 92, 0, 40, "MathCaption",ExpressionUUID->"e30d8502-c3f0-4be7-a745-1b51f8aba136"],
Cell[CellGroupData[{
Cell[6566623, 188920, 366, 11, 27, "Input",ExpressionUUID->"ac18f7a3-f1f6-4279-8d9f-ca1a1d279a00"],
Cell[6566992, 188933, 2449, 68, 32, "Output",ExpressionUUID->"7353b8ea-af54-4046-85dd-553522be6860",
 CellID->838488775]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6569490, 189007, 92, 0, 40, "MathCaption",ExpressionUUID->"844e07f2-1c2c-4be8-b14b-8357613c819e"],
Cell[CellGroupData[{
Cell[6569607, 189011, 366, 11, 27, "Input",ExpressionUUID->"a79c6459-5fa1-46d2-847f-0b71acca9667"],
Cell[6569976, 189024, 2566, 71, 32, "Output",ExpressionUUID->"5b0e5cb6-57bf-494a-8d56-bad43524c793",
 CellID->411981408]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6572591, 189101, 92, 0, 40, "MathCaption",ExpressionUUID->"235349fe-ff76-48fc-a505-352233681c2a"],
Cell[CellGroupData[{
Cell[6572708, 189105, 366, 11, 27, "Input",ExpressionUUID->"318bc09f-19a0-4bfa-af51-861c03e37eca"],
Cell[6573077, 189118, 2450, 68, 32, "Output",ExpressionUUID->"0faa4075-adae-4f29-84b5-98e182205bc9",
 CellID->1558873201]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6575576, 189192, 92, 0, 40, "MathCaption",ExpressionUUID->"898bc111-8f28-4631-972c-4ed900fadd60"],
Cell[CellGroupData[{
Cell[6575693, 189196, 366, 11, 27, "Input",ExpressionUUID->"738e6b76-f1a0-43b1-815b-6d4b6ad05f1f"],
Cell[6576062, 189209, 2569, 71, 32, "Output",ExpressionUUID->"a6c1a68f-fa6c-4d4f-826e-d7fcaec6f463",
 CellID->1276281109]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6578680, 189286, 98, 0, 40, "MathCaption",ExpressionUUID->"40ca9974-4c0c-4e87-9e9d-2a53711a425d"],
Cell[CellGroupData[{
Cell[6578803, 189290, 188, 4, 27, "Input",ExpressionUUID->"923815a3-d557-4bd8-a712-666bacbe1b77"],
Cell[6578994, 189296, 352, 8, 26, "Output",ExpressionUUID->"848afb47-8e20-4bd8-bb4b-9830928596d3",
 CellID->388405617]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6579395, 189310, 162, 2, 40, "MathCaption",ExpressionUUID->"3ef9ba87-65d2-497f-b0ec-5ea795ec1baa"],
Cell[6579560, 189314, 1538, 46, 85, "Input",ExpressionUUID->"829dcf18-00aa-482c-95a0-0815d880b6b4"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6581135, 189365, 92, 0, 40, "MathCaption",ExpressionUUID->"f276cfcd-062c-4f07-9d4a-778bf727546f"],
Cell[CellGroupData[{
Cell[6581252, 189369, 384, 12, 27, "Input",ExpressionUUID->"ef0e109f-06bb-4efa-9a51-614801d3e996"],
Cell[6581639, 189383, 4904, 138, 40, "Output",ExpressionUUID->"08e62f19-50f1-414c-876d-8788adfdce96",
 CellID->1335071829]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6586592, 189527, 92, 0, 40, "MathCaption",ExpressionUUID->"71ec6c65-d70b-4b7e-9cea-e87e899c0d4b"],
Cell[CellGroupData[{
Cell[6586709, 189531, 384, 12, 27, "Input",ExpressionUUID->"768a75ea-74c3-47b9-a769-98049c746812"],
Cell[6587096, 189545, 4904, 138, 40, "Output",ExpressionUUID->"1271fa8c-c57d-4c03-b9ef-8957d98f7d94",
 CellID->2090386549]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6592049, 189689, 183, 5, 40, "MathCaption",ExpressionUUID->"53dcc64d-6486-4310-b27b-ddb58946ef02"],
Cell[6592235, 189696, 5259, 152, 218, "Input",ExpressionUUID->"1745cfbb-48a4-4306-8b87-67422778d505"],
Cell[6597497, 189850, 2929, 86, 123, "Input",ExpressionUUID->"c24797a3-ae3b-4def-8131-732773456dd4"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6600463, 189941, 92, 0, 40, "MathCaption",ExpressionUUID->"b02f1b7f-0579-45aa-ba4b-9a6c6564cac1"],
Cell[CellGroupData[{
Cell[6600580, 189945, 448, 14, 27, "Input",ExpressionUUID->"237a8781-1685-46f4-a9e3-bae475d3bb05"],
Cell[6601031, 189961, 22641, 626, 56, "Output",ExpressionUUID->"b5961346-f56b-4c54-84c5-acd600f4ce52",
 CellID->275472198]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6623721, 190593, 92, 0, 40, "MathCaption",ExpressionUUID->"dea448ee-919a-4df3-8cc6-2b1db90e670b"],
Cell[CellGroupData[{
Cell[6623838, 190597, 454, 15, 27, "Input",ExpressionUUID->"fa447cb9-6443-4f26-8e79-284bcdb9dffe"],
Cell[6624295, 190614, 11453, 316, 32, "Output",ExpressionUUID->"91b9a257-f14c-4bd4-8496-752c6ad545a1",
 CellID->19996476]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6635797, 190936, 92, 0, 40, "MathCaption",ExpressionUUID->"c5a12d90-cd26-4507-9184-aad098e8c77f"],
Cell[CellGroupData[{
Cell[6635914, 190940, 454, 15, 27, "Input",ExpressionUUID->"de6b9e74-bfcd-41db-bb6e-93e5f2b1b299"],
Cell[6636371, 190957, 11408, 315, 32, "Output",ExpressionUUID->"2ce58a6c-670d-441a-88e5-a1f22040174d",
 CellID->593130850]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6647828, 191278, 92, 0, 40, "MathCaption",ExpressionUUID->"7886fee3-2fd5-4360-a3ed-bf07203091d1"],
Cell[CellGroupData[{
Cell[6647945, 191282, 454, 15, 27, "Input",ExpressionUUID->"8baafca8-ab74-438b-9009-b2ba7d59144a"],
Cell[6648402, 191299, 11407, 315, 32, "Output",ExpressionUUID->"6ac9b906-ae86-4431-9b23-8b4656e8e60f",
 CellID->1279502412]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6659858, 191620, 92, 0, 40, "MathCaption",ExpressionUUID->"fdf14473-f8ff-420a-8a9b-39b51951e97c"],
Cell[CellGroupData[{
Cell[6659975, 191624, 454, 15, 27, "Input",ExpressionUUID->"586a5b1b-52ef-4457-9114-e0f262a21fb9"],
Cell[6660432, 191641, 11407, 315, 32, "Output",ExpressionUUID->"9c4b4b55-c0c9-4943-88e1-b9735fb47012",
 CellID->1569544593]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6671888, 191962, 139, 1, 40, "MathCaption",ExpressionUUID->"0f8d7d58-88e1-4755-b542-e667865c995c"],
Cell[6672030, 191965, 1800, 52, 104, "Input",ExpressionUUID->"971a69f8-62b1-4bfd-a956-9d2d5dcb720f"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6673867, 192022, 92, 0, 40, "MathCaption",ExpressionUUID->"be1e9a6d-d9f2-412d-aa35-0b2d7959af8e"],
Cell[CellGroupData[{
Cell[6673984, 192026, 461, 15, 27, "Input",ExpressionUUID->"3c9eea6f-14fe-4db9-863d-06a9fc9f1793"],
Cell[6674448, 192043, 2981, 87, 40, "Output",ExpressionUUID->"adf8399e-562b-4d11-bcd7-8fd3ce4f0591",
 CellID->1876290669]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6677478, 192136, 116, 0, 40, "MathCaption",ExpressionUUID->"589c8697-e474-4b98-9e9b-c33e98d6959f"],
Cell[6677597, 192138, 4374, 127, 237, "Input",ExpressionUUID->"0ee2cb12-1d86-485c-8590-a7a0360f6844"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6682008, 192270, 92, 0, 40, "MathCaption",ExpressionUUID->"f7c99a1e-c733-44a8-9712-f4c390255c50"],
Cell[CellGroupData[{
Cell[6682125, 192274, 505, 16, 27, "Input",ExpressionUUID->"36b5760a-060c-460a-8bcb-0f4bf491a64e"],
Cell[6682633, 192292, 19260, 538, 101, "Output",ExpressionUUID->"b4930427-e77c-4d6a-812b-78b872a94332",
 CellID->1823700841]
}, Open  ]]
}, Open  ]],
Cell[6701920, 192834, 384, 9, 98, "Text",ExpressionUUID->"6f2600a8-ab19-4a2c-99de-d1ac3e09fa72"],
Cell[CellGroupData[{
Cell[6702329, 192847, 110, 0, 40, "MathCaption",ExpressionUUID->"5280a227-e9bb-4280-b021-3baac7b7ba85"],
Cell[6702442, 192849, 1117, 33, 66, "Input",ExpressionUUID->"81efa92c-8b32-4f7e-a4c6-fef5b91df159"],
Cell[6703562, 192884, 1137, 33, 66, "Input",ExpressionUUID->"b86aa0b6-822e-499c-921a-01fc7651f161"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6704736, 192922, 91, 0, 40, "MathCaption",ExpressionUUID->"836e9b5e-c49a-4700-a3b6-5b51905ce956"],
Cell[CellGroupData[{
Cell[6704852, 192926, 438, 14, 27, "Input",ExpressionUUID->"b06ac3ad-6a4a-4ecf-a231-a39b19dd0644"],
Cell[6705293, 192942, 2361, 66, 40, "Output",ExpressionUUID->"de152a61-edbd-4862-818e-332087336192",
 CellID->2028499984]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6707703, 193014, 91, 0, 40, "MathCaption",ExpressionUUID->"38ce75ba-8373-4f24-b666-8825427efe65"],
Cell[CellGroupData[{
Cell[6707819, 193018, 417, 13, 27, "Input",ExpressionUUID->"066f00de-648d-42b2-8959-15505afd3707"],
Cell[6708239, 193033, 3318, 94, 40, "Output",ExpressionUUID->"257987bc-d1fd-40a4-a9a2-8b3524d45724",
 CellID->796341276]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6711606, 193133, 92, 0, 40, "MathCaption",ExpressionUUID->"42a0654a-d004-48ab-8354-38b4344989d3"],
Cell[CellGroupData[{
Cell[6711723, 193137, 459, 15, 27, "Input",ExpressionUUID->"910fcd2d-21b0-4eee-9593-95383d0d5f75"],
Cell[6712185, 193154, 2343, 65, 40, "Output",ExpressionUUID->"0c19a7ca-d787-46dc-b4a0-b1f0b2b146c6",
 CellID->1301679146]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6714577, 193225, 92, 0, 40, "MathCaption",ExpressionUUID->"edecdc48-9fa5-49ab-9387-eef1704699ba"],
Cell[CellGroupData[{
Cell[6714694, 193229, 417, 13, 27, "Input",ExpressionUUID->"3c7f0957-53c9-49d1-9727-2f2b92defe43"],
Cell[6715114, 193244, 4277, 122, 40, "Output",ExpressionUUID->"a4199c44-4235-4178-8525-1f44f26e634a",
 CellID->923999403]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6719440, 193372, 92, 0, 40, "MathCaption",ExpressionUUID->"cd371940-b188-4dc2-a8c6-83b0ba2cce0b"],
Cell[CellGroupData[{
Cell[6719557, 193376, 417, 13, 27, "Input",ExpressionUUID->"3d209a27-fe0a-4e60-98f2-bdab7e7d28db"],
Cell[6719977, 193391, 4257, 121, 40, "Output",ExpressionUUID->"f8acad0d-203a-480c-a051-d728e085c2ed",
 CellID->972971938]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6724283, 193518, 92, 0, 40, "MathCaption",ExpressionUUID->"e625fea5-2a1d-4aeb-958e-5d551138808d"],
Cell[CellGroupData[{
Cell[6724400, 193522, 417, 13, 27, "Input",ExpressionUUID->"b822f7ad-340b-4c75-a417-6a298f07f330"],
Cell[6724820, 193537, 4277, 122, 40, "Output",ExpressionUUID->"e03eba45-32f3-451c-8aa8-ea335506a20a",
 CellID->994892794]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6729146, 193665, 535, 17, 168, "Text",ExpressionUUID->"f8f9e044-f2ef-4a5f-af23-df639eab60a5"],
Cell[6729684, 193684, 1252, 36, 85, "Input",ExpressionUUID->"36d58aff-1e85-4d69-98e7-72761b62c67b"],
Cell[6730939, 193722, 1452, 43, 85, "Input",ExpressionUUID->"52ddb2cb-ad8b-496f-9e7c-ee63772b6f36"],
Cell[6732394, 193767, 1254, 37, 66, "Input",ExpressionUUID->"676d0a2b-4311-4a33-97e0-fc03a2fb4ccb"],
Cell[6733651, 193806, 1435, 42, 85, "Input",ExpressionUUID->"655d01bb-0e02-42de-8498-3d9b8032aa64"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6735123, 193853, 177, 3, 40, "MathCaption",ExpressionUUID->"0c2844a3-1270-4e5c-88a3-395a08ec8aef"],
Cell[6735303, 193858, 1378, 40, 85, "Input",ExpressionUUID->"ff523cb0-aad5-4fa9-b08b-ba025f2ceb3a"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6736718, 193903, 92, 0, 40, "MathCaption",ExpressionUUID->"947ebd8a-9958-4318-8c54-79b8f87753e4"],
Cell[CellGroupData[{
Cell[6736835, 193907, 494, 15, 27, "Input",ExpressionUUID->"b398ab77-99d7-42e9-a9d2-39e3227a46e7"],
Cell[6737332, 193924, 2540, 70, 40, "Output",ExpressionUUID->"371dcb9a-4409-42ca-87c5-d2508c9472d6",
 CellID->1029087669]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6739921, 194000, 92, 0, 40, "MathCaption",ExpressionUUID->"caa74a4b-ffd6-46fe-b73f-43f3b461bf91"],
Cell[CellGroupData[{
Cell[6740038, 194004, 579, 19, 27, "Input",ExpressionUUID->"1f1707e0-ab8a-450e-bb91-f77199e2633d"],
Cell[6740620, 194025, 2724, 74, 40, "Output",ExpressionUUID->"39c62921-40e2-4509-bee9-977e51e4a919",
 CellID->1506847844]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6743393, 194105, 92, 0, 40, "MathCaption",ExpressionUUID->"45307c60-1c5b-49d0-ae95-d597762c20d8"],
Cell[CellGroupData[{
Cell[6743510, 194109, 494, 15, 27, "Input",ExpressionUUID->"c23224a3-96a8-4224-a7f8-6384c4a37421"],
Cell[6744007, 194126, 2522, 69, 40, "Output",ExpressionUUID->"6da631b7-ec8d-4c0d-aa5a-0ae2d0408141",
 CellID->1228034384]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6746578, 194201, 92, 0, 40, "MathCaption",ExpressionUUID->"56919c3d-4074-46d3-b6db-e55f26f68f83"],
Cell[CellGroupData[{
Cell[6746695, 194205, 579, 19, 27, "Input",ExpressionUUID->"893f5d18-50db-4a58-a915-698351501c31"],
Cell[6747277, 194226, 2702, 73, 40, "Output",ExpressionUUID->"9008fe4a-0652-45e3-a447-8ec5e4a49b9b",
 CellID->633460511]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6750028, 194305, 92, 0, 40, "MathCaption",ExpressionUUID->"cd6ea944-b3dc-49f0-8181-264fe367e377"],
Cell[CellGroupData[{
Cell[6750145, 194309, 561, 18, 27, "Input",ExpressionUUID->"cd391048-ae43-46ef-8720-9e6cde9b22bd"],
Cell[6750709, 194329, 2704, 73, 40, "Output",ExpressionUUID->"c888b4d7-dfad-4ba8-99a3-5848d8d37a69",
 CellID->1657064808]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6753462, 194408, 92, 0, 40, "MathCaption",ExpressionUUID->"53913b63-4165-453d-bb07-afb563b664cc"],
Cell[CellGroupData[{
Cell[6753579, 194412, 598, 20, 27, "Input",ExpressionUUID->"c90c3246-5928-45d4-bbdb-1b2a531b7b5b"],
Cell[6754180, 194434, 2704, 73, 32, "Output",ExpressionUUID->"e9e450ac-e953-4a04-a54c-456a0e29b240",
 CellID->671595088]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6756933, 194513, 92, 0, 40, "MathCaption",ExpressionUUID->"485731cf-0096-47fb-b1ae-9ab32a555558"],
Cell[CellGroupData[{
Cell[6757050, 194517, 519, 17, 27, "Input",ExpressionUUID->"7269d999-b525-4012-a6ed-474aee69ecab"],
Cell[6757572, 194536, 2523, 69, 32, "Output",ExpressionUUID->"2cd3ca1c-eed6-42ac-948a-4e6e831ccc21",
 CellID->979087761]
}, Open  ]]
}, Open  ]],
Cell[6760122, 194609, 452, 14, 139, "Text",ExpressionUUID->"7555d3e2-f9e0-4a0f-aa21-723e4d78252d"],
Cell[CellGroupData[{
Cell[6760599, 194627, 184, 3, 40, "MathCaption",ExpressionUUID->"76ae39a2-d747-4d93-8296-51d766975932"],
Cell[CellGroupData[{
Cell[6760808, 194634, 1373, 39, 104, "Input",ExpressionUUID->"3e935a59-ad9b-44dd-971e-04a65e62707f"],
Cell[6762184, 194675, 168, 4, 26, "Output",ExpressionUUID->"b5f60cc0-2d2d-4318-b6f6-36a74c75f733"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6762389, 194684, 1444, 40, 123, "Input",ExpressionUUID->"40be1608-5050-443f-82ac-1e2a46a67621"],
Cell[6763836, 194726, 170, 4, 26, "Output",ExpressionUUID->"5da3050a-a247-416c-b405-7457ca3c757e"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6764043, 194735, 1445, 40, 123, "Input",ExpressionUUID->"0af9875b-f23d-47a0-b3d0-e5dfd22227b3"],
Cell[6765491, 194777, 170, 4, 26, "Output",ExpressionUUID->"ddc25b93-f0ad-488b-9e19-05695857eb03"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6765698, 194786, 1987, 54, 142, "Input",ExpressionUUID->"a96214c7-3eef-4080-8319-3a13900a07cd"],
Cell[6767688, 194842, 171, 4, 26, "Output",ExpressionUUID->"729cf3a5-fd36-4e58-9a7f-e045ac5e4e61"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6767908, 194852, 92, 0, 40, "MathCaption",ExpressionUUID->"d26156d6-589a-43eb-a820-c8e8b584bb73"],
Cell[CellGroupData[{
Cell[6768025, 194856, 507, 16, 27, "Input",ExpressionUUID->"309a3465-1ffd-4c73-9832-446ea85a7c45"],
Cell[6768535, 194874, 3311, 94, 40, "Output",ExpressionUUID->"22bbefd4-025a-4433-83d4-13dff87bba91"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6771895, 194974, 92, 0, 40, "MathCaption",ExpressionUUID->"debd4b8d-e108-4b0b-9c68-c73f9ffbae3a"],
Cell[CellGroupData[{
Cell[6772012, 194978, 547, 17, 27, "Input",ExpressionUUID->"7c3d8ca6-e7e8-4793-9c43-a10744069eca"],
Cell[6772562, 194997, 3349, 94, 40, "Output",ExpressionUUID->"bb1acacb-e689-4b9b-9450-ea49182e60b1"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6775960, 195097, 92, 0, 40, "MathCaption",ExpressionUUID->"3c9902a2-c109-4b53-806d-e23ca248f7f1"],
Cell[CellGroupData[{
Cell[6776077, 195101, 548, 17, 27, "Input",ExpressionUUID->"e9d30a21-72e6-4a9e-bbbb-8533e08884c7"],
Cell[6776628, 195120, 3349, 94, 40, "Output",ExpressionUUID->"607d087e-4798-4f22-baff-b17dc18481e2"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6780026, 195220, 92, 0, 40, "MathCaption",ExpressionUUID->"c3a4c268-0329-407b-94ff-bd8e08d13930"],
Cell[CellGroupData[{
Cell[6780143, 195224, 547, 17, 27, "Input",ExpressionUUID->"bfe655b8-47fc-48dd-9560-8b9e0398eec8"],
Cell[6780693, 195243, 3349, 94, 40, "Output",ExpressionUUID->"87d05fd6-2d7f-4b93-8b4f-c09f0b9fc22a"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6784091, 195343, 92, 0, 40, "MathCaption",ExpressionUUID->"e9d3cd67-115d-486f-8f85-eb471465433d"],
Cell[CellGroupData[{
Cell[6784208, 195347, 548, 17, 27, "Input",ExpressionUUID->"b2f09501-e040-439e-bcae-c774f8af7fd4"],
Cell[6784759, 195366, 3349, 94, 40, "Output",ExpressionUUID->"27c52738-aba9-47be-8a6c-0500f4e60223"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6788157, 195466, 92, 0, 40, "MathCaption",ExpressionUUID->"aa14db38-9a1b-491d-99d0-918611075d4c"],
Cell[CellGroupData[{
Cell[6788274, 195470, 548, 17, 27, "Input",ExpressionUUID->"8fb2918c-8f5a-4646-b7bc-577d4d69ffda"],
Cell[6788825, 195489, 6754, 186, 40, "Output",ExpressionUUID->"d378f07c-5895-47f1-931c-4c7f03cb19d9"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6795628, 195681, 91, 0, 40, "MathCaption",ExpressionUUID->"ddc26dad-c949-49d0-9c60-f22636982523"],
Cell[6795722, 195683, 605, 11, 85, "Input",ExpressionUUID->"3589a430-9fb6-425a-9918-18589f632bbe"]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6796388, 195701, 983, 16, 54, "Section",ExpressionUUID->"4ad3e1c7-5134-49fb-8ac5-06e6eda5f7a3",
 CellTags->{"S5.18.1", "1.4"},
 CellID->1149066679],
Cell[CellGroupData[{
Cell[6797396, 195721, 95, 0, 39, "Subsection",ExpressionUUID->"4a1d9509-c2b0-464d-b189-034c5dce8f04"],
Cell[6797494, 195723, 1059, 24, 120, "Text",ExpressionUUID->"ec90ca91-73e8-4982-b3ad-a87b6b269135"],
Cell[CellGroupData[{
Cell[6798578, 195751, 103, 0, 40, "MathCaption",ExpressionUUID->"08b54f7b-2565-46fb-bd26-d3fbd3f10964"],
Cell[6798684, 195753, 597, 17, 66, "Input",ExpressionUUID->"9d30b384-5571-4cee-93b7-75df564d881d"],
Cell[6799284, 195772, 162, 4, 27, "Input",ExpressionUUID->"aeec4dee-b6df-4b4c-8726-1939e11fb6f9"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6799483, 195781, 111, 0, 40, "MathCaption",ExpressionUUID->"08a9687d-9bc4-4073-a0db-3c9ff2cbe796"],
Cell[CellGroupData[{
Cell[6799619, 195785, 629, 20, 27, "Input",ExpressionUUID->"bf813901-6563-450c-8a46-b575be9b1512"],
Cell[6800251, 195807, 3500, 102, 32, "Output",ExpressionUUID->"0bf6d017-b842-47a7-886f-434008a978a7",
 CellID->1909078976]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6803800, 195915, 367, 11, 56, "MathCaption",ExpressionUUID->"5b0fd041-9fbd-424e-b8e3-54272a6cc6d6"],
Cell[CellGroupData[{
Cell[6804192, 195930, 186, 4, 27, "Input",ExpressionUUID->"cc154ed0-3b24-44d2-af25-a2a6b0acaccf"],
Cell[6804381, 195936, 2022, 52, 32, "Output",ExpressionUUID->"e4af8fbc-b8cd-4637-92e5-136ce6d59176",
 CellID->337179805]
}, Open  ]],
Cell[CellGroupData[{
Cell[6806440, 195993, 190, 4, 27, "Input",ExpressionUUID->"1093c74e-dcda-4487-9d3d-b0dec9c53943"],
Cell[6806633, 195999, 583, 17, 26, "Output",ExpressionUUID->"1c716952-bd70-47fe-b137-c0b5d8bb5622",
 CellID->1919121849]
}, Open  ]],
Cell[CellGroupData[{
Cell[6807253, 196021, 191, 4, 27, "Input",ExpressionUUID->"26bd0cf6-4eea-43ed-bbef-c210aeba3357"],
Cell[6807447, 196027, 499, 15, 26, "Output",ExpressionUUID->"33221d10-6e70-475a-be52-41aa07c619e9",
 CellID->1624021850]
}, Open  ]],
Cell[CellGroupData[{
Cell[6807983, 196047, 193, 4, 27, "Input",ExpressionUUID->"9c22e424-6182-4ecc-8a9b-d56846c34361"],
Cell[6808179, 196053, 1640, 48, 32, "Output",ExpressionUUID->"f7cc9725-14a6-473e-be6c-22efacfa94f1",
 CellID->418079580]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6809868, 196107, 183, 3, 40, "MathCaption",ExpressionUUID->"54cf3993-472b-4ea2-a1ee-73451f43bbac"],
Cell[6810054, 196112, 261, 8, 27, "Input",ExpressionUUID->"8b65680a-4ff5-46e2-b555-c722321fc144"],
Cell[CellGroupData[{
Cell[6810340, 196124, 203, 6, 27, "Input",ExpressionUUID->"1874ad9d-4211-415b-b32f-32f312bf29e5"],
Cell[6810546, 196132, 187, 4, 26, "Output",ExpressionUUID->"7a3dc2e3-9433-4ab4-bb4c-ee17e9cbe777",
 CellID->97008076]
}, Open  ]],
Cell[CellGroupData[{
Cell[6810770, 196141, 244, 7, 27, "Input",ExpressionUUID->"ebc891be-03ba-4e5e-9574-29660d9d625a"],
Cell[6811017, 196150, 520, 16, 26, "Output",ExpressionUUID->"53ec0e41-54be-4efc-a670-c4d5a8bb9ed5",
 CellID->474645259]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6811586, 196172, 260, 5, 56, "MathCaption",ExpressionUUID->"7a7257e7-26f4-4a22-b15a-3b975d701814"],
Cell[CellGroupData[{
Cell[6811871, 196181, 146, 3, 27, "Input",ExpressionUUID->"d8cd172e-64a5-4e8e-9d68-71eb2f69f9cb"],
Cell[6812020, 196186, 191, 3, 26, "Output",ExpressionUUID->"6910bb66-d207-477e-b8da-6168a2161218",
 CellID->1608603055]
}, Open  ]]
}, Open  ]],
Cell[6812238, 196193, 1545, 40, 118, "DefinitionBox",ExpressionUUID->"0e23b0c4-4651-44a1-a3ee-482c5df8f2a5"],
Cell[6813786, 196235, 106, 0, 23, "Text",ExpressionUUID->"a2c4f575-feb8-4edf-ab08-26c30d77788d"],
Cell[6813895, 196237, 128, 0, 23, "Text",ExpressionUUID->"73f4e5cf-2827-4ae2-bbb6-9bc69cc9b5b5"],
Cell[CellGroupData[{
Cell[6814048, 196241, 134, 0, 40, "MathCaption",ExpressionUUID->"b487dc67-fe4e-4949-8750-1eccd2e7be2d"],
Cell[CellGroupData[{
Cell[6814207, 196245, 172, 4, 27, "Input",ExpressionUUID->"3987f991-8621-4236-b018-3919f802bf0e"],
Cell[6814382, 196251, 2025, 52, 32, "Output",ExpressionUUID->"615241d1-5c6a-4c2e-a469-b403af5fe989",
 CellID->1119913302]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6816456, 196309, 108, 0, 40, "MathCaption",ExpressionUUID->"5e9a7637-15ec-40a6-984c-5eff5328942d"],
Cell[CellGroupData[{
Cell[6816589, 196313, 182, 4, 27, "Input",ExpressionUUID->"9c392366-afd0-47ae-b94f-024692108dce"],
Cell[6816774, 196319, 797, 20, 26, "Output",ExpressionUUID->"c6c65fe9-d5ad-4361-8da8-6bb337a60613",
 CellID->1836184406]
}, Open  ]],
Cell[CellGroupData[{
Cell[6817608, 196344, 182, 4, 27, "Input",ExpressionUUID->"fa2f8566-ee93-43ed-a926-aed2ffbe537d"],
Cell[6817793, 196350, 1528, 46, 32, "Output",ExpressionUUID->"35ebb4fe-469d-4040-b455-ba95b54c0810",
 CellID->1411304252]
}, Open  ]],
Cell[CellGroupData[{
Cell[6819358, 196401, 182, 4, 27, "Input",ExpressionUUID->"929ad7af-3096-4593-8f23-c329c00e6205"],
Cell[6819543, 196407, 478, 16, 26, "Output",ExpressionUUID->"0fa56880-707d-46f2-8b3e-bf998b0e9472",
 CellID->833682249]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6820070, 196429, 94, 0, 40, "MathCaption",ExpressionUUID->"7f5c625f-734f-4588-982b-7f532f5f253c"],
Cell[CellGroupData[{
Cell[6820189, 196433, 178, 4, 27, "Input",ExpressionUUID->"1237657c-684c-408f-9331-953c427b2a71"],
Cell[6820370, 196439, 621, 17, 26, "Output",ExpressionUUID->"18a33b13-8f57-4934-acd3-1a014710c30a",
 CellID->1514245915]
}, Open  ]],
Cell[CellGroupData[{
Cell[6821028, 196461, 180, 4, 27, "Input",ExpressionUUID->"a664330c-0928-4016-b1a9-be729781757d"],
Cell[6821211, 196467, 1816, 51, 32, "Output",ExpressionUUID->"38a0d0eb-8851-4dd5-a484-3c564af2b8c9",
 CellID->102392808]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6823076, 196524, 90, 0, 40, "MathCaption",ExpressionUUID->"b6542f97-4204-415d-b2a5-dcc32ff3e912"],
Cell[CellGroupData[{
Cell[6823191, 196528, 180, 4, 27, "Input",ExpressionUUID->"03991c10-3d36-4745-b173-ca20212f9887"],
Cell[6823374, 196534, 584, 17, 26, "Output",ExpressionUUID->"e683677e-0da6-4f25-b3fb-4e6dcc5582de",
 CellID->781522809]
}, Open  ]],
Cell[CellGroupData[{
Cell[6823995, 196556, 181, 4, 27, "Input",ExpressionUUID->"dbe0d3a9-0487-465e-b803-32fc1b3843a8"],
Cell[6824179, 196562, 629, 18, 26, "Output",ExpressionUUID->"27eb3542-084b-4b35-8885-f7fdeab98bb2",
 CellID->1043865811]
}, Open  ]],
Cell[CellGroupData[{
Cell[6824845, 196585, 183, 4, 27, "Input",ExpressionUUID->"ed2e8b93-887a-4d0e-892d-7cfa24604bff"],
Cell[6825031, 196591, 1641, 48, 32, "Output",ExpressionUUID->"172d4218-2fab-4c12-92fc-52681e63ec15",
 CellID->1458773134]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6826721, 196645, 121, 0, 40, "MathCaption",ExpressionUUID->"1645b5ba-fee1-4383-b782-f728ec514af4"],
Cell[CellGroupData[{
Cell[6826867, 196649, 185, 4, 27, "Input",ExpressionUUID->"feb66604-48ae-41a3-991c-b246ef12bc66"],
Cell[6827055, 196655, 1943, 51, 32, "Output",ExpressionUUID->"a3ec2f84-a3bd-4c07-95af-120ddabd08ad",
 CellID->1926748986]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6829047, 196712, 90, 0, 40, "MathCaption",ExpressionUUID->"a71749fd-50b6-4a16-9de0-b162a2e23a8c"],
Cell[CellGroupData[{
Cell[6829162, 196716, 177, 4, 27, "Input",ExpressionUUID->"890c6c2e-4789-4395-9bf4-aee37ece4fb3"],
Cell[6829342, 196722, 600, 18, 26, "Output",ExpressionUUID->"ab8534f6-b8b5-4529-865a-f6fc69fef193",
 CellID->964294589]
}, Open  ]]
}, Open  ]],
Cell[6829969, 196744, 222, 3, 40, "Text",ExpressionUUID->"29fe073d-cb3a-49da-834d-9ec3421e0c07"],
Cell[CellGroupData[{
Cell[6830216, 196751, 99, 0, 40, "MathCaption",ExpressionUUID->"8900d4b8-df42-4bd2-a20f-cd5d57cc8eea"],
Cell[CellGroupData[{
Cell[6830340, 196755, 239, 7, 27, "Input",ExpressionUUID->"bee2bb1d-d978-4918-b1ef-540fa6545495"],
Cell[6830582, 196764, 1643, 48, 32, "Output",ExpressionUUID->"86ab5e1c-9cc9-420e-a9a6-3f3368f2974a",
 CellID->1987557061]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6832274, 196818, 107, 0, 40, "MathCaption",ExpressionUUID->"75fa086f-3cd8-409a-a1bd-8d0f8a21bd63"],
Cell[CellGroupData[{
Cell[6832406, 196822, 204, 5, 27, "Input",ExpressionUUID->"fca19177-3c73-4e80-9138-18dd0f5b3dda"],
Cell[6832613, 196829, 439, 14, 26, "Output",ExpressionUUID->"c6d13128-76ba-40eb-a39d-a50b232239ef",
 CellID->763395142]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6833101, 196849, 267, 4, 56, "MathCaption",ExpressionUUID->"334be659-ef87-4933-93d4-328b122cb3e6"],
Cell[CellGroupData[{
Cell[6833393, 196857, 204, 5, 27, "Input",ExpressionUUID->"dd83d1c5-dbe1-4ebf-8b40-1c588e98a97a"],
Cell[6833600, 196864, 521, 16, 26, "Output",ExpressionUUID->"8b8c0d3f-5a12-4bfd-bc6d-b95cb7fca734",
 CellID->1314844416]
}, Open  ]]
}, Open  ]],
Cell[6834148, 196884, 202, 3, 23, "Text",ExpressionUUID->"3df83f9b-37cc-496d-8fd1-3aa4f09a2516"],
Cell[6834353, 196889, 430, 10, 41, "DefinitionBox",ExpressionUUID->"462a50ad-7a2f-47cf-9937-55736894057f"],
Cell[6834786, 196901, 99, 0, 23, "Text",ExpressionUUID->"1e627b5b-15af-4eea-a552-cb87e10ee62d"]
}, Closed]],
Cell[CellGroupData[{
Cell[6834922, 196906, 98, 0, 39, "Subsection",ExpressionUUID->"68a77a50-2a2b-464e-be3d-aea7c596e152"],
Cell[6835023, 196908, 787, 12, 103, "Text",ExpressionUUID->"8c454804-295a-4df4-9665-194a0ee1939b"],
Cell[6835813, 196922, 211, 3, 23, "Text",ExpressionUUID->"2b35b02a-c8bf-4f14-bed2-8147e6c0d6aa"],
Cell[6836027, 196927, 2543, 64, 143, "DefinitionBox",ExpressionUUID->"3f6ce36e-5c90-432c-a098-4423b2bb4792"],
Cell[6838573, 196993, 90, 0, 23, "Text",ExpressionUUID->"35178b9b-4b2c-44ba-9287-02bfed95b5f8"],
Cell[CellGroupData[{
Cell[6838688, 196997, 111, 0, 40, "MathCaption",ExpressionUUID->"1ea4b711-f943-49cf-866b-97d2be47f759"],
Cell[CellGroupData[{
Cell[6838824, 197001, 114, 2, 27, "Input",ExpressionUUID->"5f02793c-a854-4c2c-a92f-510be884e9fa"],
Cell[6838941, 197005, 3497, 102, 32, "Output",ExpressionUUID->"a7c8fac9-0283-49fa-a4ab-08024f8eaf01",
 CellID->101426694]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6842487, 197113, 345, 11, 56, "MathCaption",ExpressionUUID->"0dd1d9b6-37ba-4247-98a7-326bc91b7ec0"],
Cell[CellGroupData[{
Cell[6842857, 197128, 241, 6, 27, "Input",ExpressionUUID->"cae724e6-0487-456e-8869-216c7fe76746"],
Cell[6843101, 197136, 3497, 102, 32, "Output",ExpressionUUID->"8ce8d018-eba7-4205-9cc1-f3c4c626c6f4",
 CellID->1723913987]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6846647, 197244, 258, 6, 56, "MathCaption",ExpressionUUID->"33b3e14f-9ec0-4f3f-aab9-fac51b3861b6"],
Cell[CellGroupData[{
Cell[6846930, 197254, 368, 11, 27, "Input",ExpressionUUID->"05e945ee-5191-4533-9a44-ada8367ba4de"],
Cell[6847301, 197267, 3498, 102, 32, "Output",ExpressionUUID->"36664888-4387-4549-aa4a-ff491c478820",
 CellID->2120497832]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6850848, 197375, 138, 1, 40, "MathCaption",ExpressionUUID->"16c5cc11-ec75-401f-a4b1-58f13116440d"],
Cell[CellGroupData[{
Cell[6851011, 197380, 510, 15, 27, "Input",ExpressionUUID->"aff528fe-5a5b-4e98-b971-b58a1d123480"],
Cell[6851524, 197397, 3462, 101, 32, "Output",ExpressionUUID->"38f4585e-dc98-4a36-a920-2884996ea1f6",
 CellID->1715940459]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6855035, 197504, 121, 0, 40, "MathCaption",ExpressionUUID->"1dc78cd9-8486-4725-9123-15d6509810fa"],
Cell[CellGroupData[{
Cell[6855181, 197508, 266, 7, 27, "Input",ExpressionUUID->"a5706088-171f-4209-ad6d-6a7ce9714825"],
Cell[6855450, 197517, 3498, 102, 32, "Output",ExpressionUUID->"468b0d29-b24c-48fd-9086-ba525ac34c28",
 CellID->1253398850]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6858997, 197625, 152, 2, 40, "MathCaption",ExpressionUUID->"6e97997d-9ad6-41cd-b188-ed3fcce9fbc4"],
Cell[CellGroupData[{
Cell[6859174, 197631, 232, 6, 27, "Input",ExpressionUUID->"59274bc5-e245-4819-b937-0359f467bccb"],
Cell[6859409, 197639, 1121, 34, 32, "Output",ExpressionUUID->"906018a3-2b3f-4f7e-8a3e-1234a3071628",
 CellID->1207378527]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6860579, 197679, 193, 3, 40, "MathCaption",ExpressionUUID->"dd271608-1cce-4f7b-982d-7922eddfd796"],
Cell[CellGroupData[{
Cell[6860797, 197686, 386, 11, 27, "Input",ExpressionUUID->"cddb3c08-4e2e-49e4-b7c0-b7ec23245170"],
Cell[6861186, 197699, 2154, 62, 32, "Output",ExpressionUUID->"3c3ef9d9-4733-44f8-bfda-07d793082cde",
 CellID->837922786]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6863389, 197767, 421, 12, 56, "MathCaption",ExpressionUUID->"9d384b72-75ba-48c8-beda-4d191dff22af"],
Cell[CellGroupData[{
Cell[6863835, 197783, 232, 7, 27, "Input",ExpressionUUID->"f1839c86-0e9a-44b0-949a-4456f09e48d2"],
Cell[6864070, 197792, 1062, 31, 32, "Output",ExpressionUUID->"6cfe3277-edd2-4f91-8f98-d7b8812b2c0c",
 CellID->1917990923]
}, Open  ]]
}, Open  ]],
Cell[6865159, 197827, 564, 13, 57, "Text",ExpressionUUID->"0f28b103-1f3c-451f-92f4-a23c34372f91"],
Cell[CellGroupData[{
Cell[6865748, 197844, 113, 0, 40, "MathCaption",ExpressionUUID->"4f9aab16-ab41-4b86-935b-176894b39011"],
Cell[CellGroupData[{
Cell[6865886, 197848, 304, 10, 27, "Input",ExpressionUUID->"bffc8524-eb60-4749-9859-340eadcd48df"],
Cell[6866193, 197860, 2080, 60, 32, "Output",ExpressionUUID->"6e4052a9-9889-4ccd-b8d4-1ecd42a99960",
 CellID->1943901391]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6868322, 197926, 96, 0, 40, "MathCaption",ExpressionUUID->"99012966-5b75-46bb-b44b-87c258438ef4"],
Cell[CellGroupData[{
Cell[6868443, 197930, 139, 3, 27, "Input",ExpressionUUID->"100a5072-2743-49ff-8ba7-56cf190c0ea6"],
Cell[6868585, 197935, 2184, 62, 34, "Output",ExpressionUUID->"daadf511-6f13-4a91-b113-30d3f4d6c6c5",
 CellID->1481956726]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6870818, 198003, 217, 5, 40, "MathCaption",ExpressionUUID->"929903a5-6ee2-4d13-af7f-b41bd4370310"],
Cell[CellGroupData[{
Cell[6871060, 198012, 268, 7, 27, "Input",ExpressionUUID->"da498164-adb3-4e7e-aa7a-1a798dd60837"],
Cell[6871331, 198021, 3985, 116, 32, "Output",ExpressionUUID->"a324e78f-07d6-4274-b551-8f60d7826ca0",
 CellID->731355976]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6875365, 198143, 193, 3, 40, "MathCaption",ExpressionUUID->"4850485e-cf10-4ed6-b042-a7313d8fcbf1"],
Cell[CellGroupData[{
Cell[6875583, 198150, 180, 4, 27, "Input",ExpressionUUID->"43306e30-242a-48d4-82c3-ea9169785cab"],
Cell[6875766, 198156, 3986, 116, 32, "Output",ExpressionUUID->"b7cb7749-0df4-48ec-a5fb-f28ce932a99e",
 CellID->1850976851]
}, Open  ]],
Cell[CellGroupData[{
Cell[6879789, 198277, 169, 4, 27, "Input",ExpressionUUID->"0ff4b29f-3819-4961-8a5e-5f399ec359ed"],
Cell[6879961, 198283, 900, 23, 26, "Output",ExpressionUUID->"076e08f9-0503-46cf-91bd-c5032a87cd58",
 CellID->919824315]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6880910, 198312, 441, 12, 56, "MathCaption",ExpressionUUID->"86993165-44a8-4b2e-a22d-dbdd7f1ecd46"],
Cell[CellGroupData[{
Cell[6881376, 198328, 507, 16, 27, "Input",ExpressionUUID->"8c6b02de-2419-47f2-8349-8b6e4472f7de"],
Cell[6881886, 198346, 2120, 61, 32, "Output",ExpressionUUID->"a98ce367-bdde-466e-8dee-2e2c261ab3e2",
 CellID->1899971099]
}, Open  ]]
}, Open  ]],
Cell[6884033, 198411, 430, 6, 57, "Text",ExpressionUUID->"331ef9a3-c41e-4169-88d8-7674afe1c480"],
Cell[CellGroupData[{
Cell[6884488, 198421, 283, 4, 56, "MathCaption",ExpressionUUID->"57336c3e-2ae6-458b-b1f2-8aa13319c26c"],
Cell[CellGroupData[{
Cell[6884796, 198429, 362, 10, 27, "Input",ExpressionUUID->"ecf81d73-b7cc-4e1a-ab8b-8362167da0aa"],
Cell[6885161, 198441, 1194, 32, 32, "Output",ExpressionUUID->"8e0f888f-3e7b-44fe-83b1-ad89fe9fd200",
 CellID->1879913650]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6886404, 198479, 196, 3, 40, "MathCaption",ExpressionUUID->"63433919-b7ed-4060-b27a-59382e72ea60"],
Cell[CellGroupData[{
Cell[6886625, 198486, 823, 28, 47, "Input",ExpressionUUID->"1ec771b7-f5f1-4180-b0ab-2efd66266226"],
Cell[6887451, 198516, 3097, 88, 32, "Output",ExpressionUUID->"f729d0ec-d17e-4cb8-8ab8-37cb66991c32",
 CellID->1929300886],
Cell[6890551, 198606, 3096, 88, 32, "Output",ExpressionUUID->"0e629277-9a32-4426-bbdc-a4e4052292ad",
 CellID->904404949]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6893696, 198700, 335, 11, 40, "MathCaption",ExpressionUUID->"088a6e9c-0ee1-44df-aad1-c3e6c8059dcb"],
Cell[CellGroupData[{
Cell[6894056, 198715, 185, 4, 27, "Input",ExpressionUUID->"48a5a25b-e796-489e-be4a-b8d0199583f4"],
Cell[6894244, 198721, 731, 20, 22, "Print",ExpressionUUID->"9a36a885-2676-4d0c-bcf3-0f87ee5555c3",
 CellID->444724377],
Cell[6894978, 198743, 175, 3, 26, "Output",ExpressionUUID->"69a310c2-ea43-4750-9ecd-13b1ae2fc5e8",
 CellID->1239512679]
}, Open  ]],
Cell[CellGroupData[{
Cell[6895190, 198751, 229, 6, 27, "Input",ExpressionUUID->"4bab9c9c-34bc-4aba-8251-5ad860cdabe5"],
Cell[6895422, 198759, 176, 3, 26, "Output",ExpressionUUID->"ce0fbcb8-d1b6-4b4e-9979-6dd1cfe23b3f",
 CellID->1898220053]
}, Open  ]]
}, Open  ]],
Cell[6895625, 198766, 631, 16, 57, "Text",ExpressionUUID->"4f840722-7826-490a-9e87-e58551852f62"],
Cell[CellGroupData[{
Cell[6896281, 198786, 129, 0, 40, "MathCaption",ExpressionUUID->"14e43814-0a7b-482a-93cb-989a4415f4fa"],
Cell[CellGroupData[{
Cell[6896435, 198790, 364, 11, 27, "Input",ExpressionUUID->"0b657521-d0d2-4e7e-8dd2-2281e9e6b9fb"],
Cell[6896802, 198803, 9498, 260, 32, "Output",ExpressionUUID->"a7e40621-21da-4b12-8798-83144f028ec9",
 CellID->367370939]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6906349, 199069, 90, 0, 40, "MathCaption",ExpressionUUID->"28062b0f-a0a6-4e31-b569-9ed6389d5a77"],
Cell[CellGroupData[{
Cell[6906464, 199073, 515, 16, 27, "Input",ExpressionUUID->"8256cd02-1a59-4f7f-9f01-f9b5c1aba6db"],
Cell[6906982, 199091, 13320, 349, 70, "Output",ExpressionUUID->"63821e0f-b1a0-443c-b498-ff5572caa16a",
 CellID->387886648]
}, Open  ]],
Cell[CellGroupData[{
Cell[6920339, 199445, 147, 3, 27, "Input",ExpressionUUID->"be0afcd4-87f9-4a72-8e8b-24e4e4512a9b"],
Cell[6920489, 199450, 217, 5, 70, "Output",ExpressionUUID->"99c61762-2fae-4934-9916-26a2da43cff7",
 CellID->1598024291]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6920755, 199461, 282, 6, 56, "MathCaption",ExpressionUUID->"65a7c07b-05fd-4128-87d1-24b55ec809e5"],
Cell[CellGroupData[{
Cell[6921062, 199471, 485, 14, 27, "Input",ExpressionUUID->"971ad077-8690-47f7-b065-a37d0763409b"],
Cell[6921550, 199487, 30005, 794, 70, "Output",ExpressionUUID->"0d6e8811-665f-491a-a874-cf9382994925",
 CellID->1563769455]
}, Open  ]],
Cell[6951570, 200284, 174, 4, 27, "Input",ExpressionUUID->"f11b00ca-2ad1-45d5-ae7f-ec6964794603"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[6951793, 200294, 110, 0, 39, "Subsection",ExpressionUUID->"34d6e14e-978f-4ad5-8c42-a9b088439972"],
Cell[6951906, 200296, 1079, 31, 91, "Text",ExpressionUUID->"d6b7cae7-6cbc-4b7a-8f76-4397526299c9"],
Cell[6952988, 200329, 1385, 38, 92, "DefinitionBox",ExpressionUUID->"ccc23ef0-80f2-499f-ac44-b818dc825725"],
Cell[6954376, 200369, 90, 0, 23, "Text",ExpressionUUID->"dfd5dd34-fc9d-4079-ac37-f8622311e238"],
Cell[CellGroupData[{
Cell[6954491, 200373, 201, 5, 40, "MathCaption",ExpressionUUID->"02809a8f-7935-42fe-b9d3-87c59bf6e556"],
Cell[CellGroupData[{
Cell[6954717, 200382, 256, 8, 27, "Input",ExpressionUUID->"7e28da1b-4600-4317-b9f1-4ee3c66fa768"],
Cell[6954976, 200392, 2131, 62, 34, "Output",ExpressionUUID->"e893da69-2f50-456c-902b-720c3baeb5c4",
 CellID->1052016746]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6957156, 200460, 121, 0, 40, "MathCaption",ExpressionUUID->"89afdd22-2fed-4caa-b129-f4242e28eaf5"],
Cell[CellGroupData[{
Cell[6957302, 200464, 266, 8, 27, "Input",ExpressionUUID->"2a926f40-ec11-4b98-8d4b-2f34888ea30f"],
Cell[6957571, 200474, 2604, 73, 33, "Output",ExpressionUUID->"c60a4b11-acaa-494f-a22d-2cbe60519d86",
 CellID->1012584271]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6960224, 200553, 179, 4, 40, "MathCaption",ExpressionUUID->"f3fba0e3-ed58-4501-8902-895a8267fe48"],
Cell[CellGroupData[{
Cell[6960428, 200561, 651, 22, 27, "Input",ExpressionUUID->"2a1b3afe-1680-41f2-ba36-3fd29147134a"],
Cell[6961082, 200585, 7926, 221, 62, "Output",ExpressionUUID->"7a38a75b-3952-4011-bb56-b4b218f1ad70",
 CellID->1843423869]
}, Open  ]],
Cell[CellGroupData[{
Cell[6969045, 200811, 151, 3, 27, "Input",ExpressionUUID->"64be45db-78a2-45df-a8b4-0433bf8ae7ee"],
Cell[6969199, 200816, 2534, 74, 32, "Output",ExpressionUUID->"3802aa67-8e6b-4326-b2d5-ec70241ddeea",
 CellID->678391014]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6971782, 200896, 265, 6, 56, "MathCaption",ExpressionUUID->"a23430fd-4f21-48fc-bd82-1d783d79446e"],
Cell[CellGroupData[{
Cell[6972072, 200906, 289, 9, 27, "Input",ExpressionUUID->"1d2435ee-543a-48ad-819d-18469e47d9a7"],
Cell[6972364, 200917, 3518, 101, 32, "Output",ExpressionUUID->"c063ba00-21d8-4127-abf7-81f32cb00bed",
 CellID->727090375]
}, Open  ]],
Cell[CellGroupData[{
Cell[6975919, 201023, 145, 3, 27, "Input",ExpressionUUID->"f4e9ad89-1248-4411-a143-6a3dc3e2a5e7"],
Cell[6976067, 201028, 3519, 101, 32, "Output",ExpressionUUID->"3d2f6192-691e-40f7-aa2e-f792f62f7520",
 CellID->1025981195]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6979635, 201135, 304, 8, 56, "MathCaption",ExpressionUUID->"5e7bc697-5e66-4d09-a94d-8aee390a8881"],
Cell[CellGroupData[{
Cell[6979964, 201147, 415, 14, 27, "Input",ExpressionUUID->"f9d1632a-49af-4c4b-a9f4-dc0e9645bbc1"],
Cell[6980382, 201163, 4513, 130, 58, "Output",ExpressionUUID->"8a7c8ff0-de02-42ee-99e0-2a3d4b443ef1",
 CellID->1519742602]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6984944, 201299, 171, 5, 40, "MathCaption",ExpressionUUID->"ef0cf605-7fcb-4dab-874c-fd1ba838cd67"],
Cell[CellGroupData[{
Cell[6985140, 201308, 146, 3, 27, "Input",ExpressionUUID->"831d34d5-c825-49b7-bfa2-3e3bc81f8ee2"],
Cell[6985289, 201313, 5009, 142, 58, "Output",ExpressionUUID->"9988d172-c4c2-4cd2-aef3-9803db18c0d7",
 CellID->143414396]
}, Open  ]],
Cell[CellGroupData[{
Cell[6990335, 201460, 151, 3, 27, "Input",ExpressionUUID->"82ae8eae-4525-40e8-a4cd-87cf0d1beeae"],
Cell[6990489, 201465, 171, 3, 26, "Output",ExpressionUUID->"b2bab2e6-40e1-4d20-8a0c-c281b03b7e2b",
 CellID->889101638]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[6990709, 201474, 266, 8, 40, "MathCaption",ExpressionUUID->"afc80288-f75e-4254-82cb-0a383607b6ec"],
Cell[CellGroupData[{
Cell[6991000, 201486, 327, 10, 27, "Input",ExpressionUUID->"238c28a7-0568-4a0d-b606-b68eac0c4de4"],
Cell[6991330, 201498, 4646, 133, 32, "Output",ExpressionUUID->"00a549f2-e608-4b61-878c-391bd5ae3104",
 CellID->940014040]
}, Open  ]],
Cell[CellGroupData[{
Cell[6996013, 201636, 145, 3, 27, "Input",ExpressionUUID->"6c45551a-43eb-4f09-96eb-ae05a5750bf2"],
Cell[6996161, 201641, 5012, 142, 32, "Output",ExpressionUUID->"69ee8eb8-ec20-49b3-ae0a-3d8998eb3502",
 CellID->1926262948]
}, Open  ]],
Cell[CellGroupData[{
Cell[7001210, 201788, 147, 3, 27, "Input",ExpressionUUID->"804d690d-8ec8-4c0e-a017-090f827590dd"],
Cell[7001360, 201793, 2606, 73, 33, "Output",ExpressionUUID->"9026885d-242f-4c97-b679-16ea593fc002",
 CellID->1385388511]
}, Open  ]],
Cell[CellGroupData[{
Cell[7004003, 201871, 141, 3, 27, "Input",ExpressionUUID->"8aa7cb1c-a863-435c-94d1-830aa0c3ba4c"],
Cell[7004147, 201876, 3880, 116, 32, "Output",ExpressionUUID->"924d21a1-dd0e-4ce9-9bfd-8c4346ba0cbf",
 CellID->1207003603]
}, Open  ]]
}, Open  ]],
Cell[7008054, 201996, 1071, 28, 75, "DefinitionBox",ExpressionUUID->"c16ed14c-a9db-48b2-8046-7985ef86d7ab"],
Cell[7009128, 202026, 103, 0, 23, "Text",ExpressionUUID->"cb026765-8d55-4693-aa38-9cc24b0d450f"],
Cell[CellGroupData[{
Cell[7009256, 202030, 168, 3, 40, "MathCaption",ExpressionUUID->"0526b846-a9e2-4fbc-9fb6-f2d0474909e5"],
Cell[CellGroupData[{
Cell[7009449, 202037, 126, 2, 27, "Input",ExpressionUUID->"c6ec4450-6498-4cb7-a576-5098e9d26a99"],
Cell[7009578, 202041, 188, 4, 26, "Output",ExpressionUUID->"b614e995-1a44-4ce8-9335-e3a5033d2275",
 CellID->979026059]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7009815, 202051, 109, 0, 40, "MathCaption",ExpressionUUID->"3d0d2018-13ec-4050-be38-032986c2c27e"],
Cell[CellGroupData[{
Cell[7009949, 202055, 155, 3, 27, "Input",ExpressionUUID->"099e2601-145b-4497-852a-ec0f65bb57bd"],
Cell[7010107, 202060, 620, 12, 22, "Print",ExpressionUUID->"37e7dc5f-87e0-453e-878c-a1fd963d1594",
 CellID->471270299]
}, Open  ]],
Cell[CellGroupData[{
Cell[7010764, 202077, 273, 8, 27, "Input",ExpressionUUID->"dda1ecc1-315b-48ec-acb1-a8b07f754fda"],
Cell[7011040, 202087, 174, 3, 26, "Output",ExpressionUUID->"4d25fdf3-6ab1-4962-a100-77e407725531",
 CellID->720225157]
}, Open  ]],
Cell[CellGroupData[{
Cell[7011251, 202095, 116, 2, 27, "Input",ExpressionUUID->"b94db8d1-562d-48c2-940c-2ae087dfc283"],
Cell[7011370, 202099, 221, 5, 26, "Output",ExpressionUUID->"e20dedf7-ec3b-481f-9a46-78d73dac1440",
 CellID->1606379207]
}, Open  ]],
Cell[CellGroupData[{
Cell[7011628, 202109, 382, 12, 27, "Input",ExpressionUUID->"6f56fbdc-a1ea-479f-8cf0-209b86efcb4a"],
Cell[7012013, 202123, 10724, 293, 60, "Output",ExpressionUUID->"b0b75048-4c72-45ac-bfd8-b7b35a9ac844",
 CellID->250921965]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7022798, 202423, 94, 0, 39, "Subsection",ExpressionUUID->"c8c4f3cf-4587-4a19-9809-f676b4bd4292"],
Cell[7022895, 202425, 458, 11, 57, "Text",ExpressionUUID->"6fa9b11f-a727-49e7-ab64-160e4d699f9a"],
Cell[7023356, 202438, 2379, 63, 143, "DefinitionBox",ExpressionUUID->"b80fd34e-e0a6-43b9-aa68-531d399ef4e5"],
Cell[7025738, 202503, 99, 0, 23, "Text",ExpressionUUID->"80fe4883-cd03-4d3e-865d-0b430de5569c"],
Cell[CellGroupData[{
Cell[7025862, 202507, 244, 6, 40, "MathCaption",ExpressionUUID->"733b0f1e-ad8e-483f-a535-40f4d4d1f905"],
Cell[CellGroupData[{
Cell[7026131, 202517, 154, 3, 27, "Input",ExpressionUUID->"0c98f142-6e48-4840-8e65-030e39381dac"],
Cell[7026288, 202522, 609, 12, 22, "Print",ExpressionUUID->"1c41b4b3-ce33-4c7b-953b-097bcc707856",
 CellID->1265143671]
}, Open  ]],
Cell[CellGroupData[{
Cell[7026934, 202539, 304, 9, 27, "Input",ExpressionUUID->"09d3db03-2ad7-4b1f-af7d-f8ea2941b7b5"],
Cell[7027241, 202550, 1429, 39, 32, "Output",ExpressionUUID->"485c6a60-2478-4ceb-a1a7-852406d78846",
 CellID->1034089550]
}, Open  ]],
Cell[CellGroupData[{
Cell[7028707, 202594, 208, 5, 27, "Input",ExpressionUUID->"51a627cb-2737-4c3e-89fa-136cf80e43e8"],
Cell[7028918, 202601, 174, 3, 26, "Output",ExpressionUUID->"8d453e70-e9b1-4e11-a200-565148176d41",
 CellID->357911679]
}, Open  ]],
Cell[CellGroupData[{
Cell[7029129, 202609, 156, 3, 27, "Input",ExpressionUUID->"2253f3f5-0604-4a8a-9fec-4d85c4832100"],
Cell[7029288, 202614, 575, 12, 22, "Print",ExpressionUUID->"710782d0-3167-46f5-8bfb-346a19352fac",
 CellID->319893706]
}, Open  ]]
}, Open  ]],
Cell[7029890, 202630, 248, 6, 23, "Text",ExpressionUUID->"5ca9d54d-ab49-4ed8-b2c0-a9a3faa724ca"],
Cell[CellGroupData[{
Cell[7030163, 202640, 299, 9, 40, "MathCaption",ExpressionUUID->"829ab694-be18-4d4c-93e4-664f484d74c9"],
Cell[CellGroupData[{
Cell[7030487, 202653, 215, 5, 27, "Input",ExpressionUUID->"e6c54ea5-6f93-448c-9316-36b4d120ff45"],
Cell[7030705, 202660, 607, 12, 22, "Print",ExpressionUUID->"fd74ea41-a589-42b0-9ffc-5ade58e61d7a",
 CellID->1301948301]
}, Open  ]],
Cell[CellGroupData[{
Cell[7031349, 202677, 148, 3, 27, "Input",ExpressionUUID->"43520ac1-3e71-482b-a187-09c76978fd0e"],
Cell[7031500, 202682, 175, 3, 26, "Output",ExpressionUUID->"075aba0d-81c0-4737-a691-8e3aae9bff16",
 CellID->1134215624]
}, Open  ]],
Cell[CellGroupData[{
Cell[7031712, 202690, 294, 9, 27, "Input",ExpressionUUID->"303b98cd-830c-4a67-9c99-0f3f5fc49e31"],
Cell[7032009, 202701, 2752, 78, 32, "Output",ExpressionUUID->"ce1cffdd-2177-4ead-8e19-b09f5850d64a",
 CellID->1697141796]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7034810, 202785, 349, 7, 56, "MathCaption",ExpressionUUID->"4bdcc9f6-76da-4d61-abf6-3ef98ef65885"],
Cell[CellGroupData[{
Cell[7035184, 202796, 249, 6, 27, "Input",ExpressionUUID->"f905aeaf-678d-4a74-b6a5-1019eb025c59"],
Cell[7035436, 202804, 606, 12, 22, "Print",ExpressionUUID->"3f38ca6c-52b5-406f-bdac-f2d801bf6e4e",
 CellID->984365918]
}, Open  ]],
Cell[CellGroupData[{
Cell[7036079, 202821, 281, 7, 27, "Input",ExpressionUUID->"f0750651-cc9b-4093-9c65-327ac95e491f"],
Cell[7036363, 202830, 604, 12, 22, "Print",ExpressionUUID->"0e56d285-1f8f-48c7-850b-9170397720a1",
 CellID->23967338]
}, Open  ]],
Cell[CellGroupData[{
Cell[7037004, 202847, 132, 3, 27, "Input",ExpressionUUID->"3b73ef04-c158-4ad4-ab84-823a579ce6c1"],
Cell[CellGroupData[{
Cell[7037161, 202854, 202, 3, 43, "Print",ExpressionUUID->"f7c28f10-04f1-4f64-97a8-4dd64ff8b439",
 CellTags->"Info12393728982516-9181623",
 CellID->1239607777],
Cell[7037366, 202859, 1620, 47, 193, "Print",ExpressionUUID->"e7556ba5-d1ca-4503-aebf-03e321a473f9",
 CellTags->"Info12393728982516-9181623",
 CellID->1349048492]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7039047, 202913, 102, 0, 40, "MathCaption",ExpressionUUID->"0de789bd-b2b7-4ae0-b428-8fae5544dcd6"],
Cell[CellGroupData[{
Cell[7039174, 202917, 456, 14, 27, "Input",ExpressionUUID->"5f5df07e-aaf0-498e-93af-7fc09637ba77"],
Cell[7039633, 202933, 5018, 143, 32, "Output",ExpressionUUID->"8fd22edf-e763-4e7a-9982-770731dd7135",
 CellID->325660018]
}, Open  ]],
Cell[CellGroupData[{
Cell[7044688, 203081, 531, 16, 27, "Input",ExpressionUUID->"9da0f548-1b46-4408-89ea-da4e96b50aa7"],
Cell[7045222, 203099, 6094, 173, 32, "Output",ExpressionUUID->"bc61fdda-48b1-430e-9c51-339643791517",
 CellID->1579589970]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7051365, 203278, 780, 16, 104, "MathCaption",ExpressionUUID->"fafb08f7-c6c0-4df5-bf1f-b1c9adc35881"],
Cell[7052148, 203296, 546, 17, 27, "Input",ExpressionUUID->"af717f45-7c64-4008-a3e4-1dd89b2e3a77"],
Cell[CellGroupData[{
Cell[7052719, 203317, 305, 9, 27, "Input",ExpressionUUID->"8d33abb9-e21c-4b36-8f7a-7e5246b821c5"],
Cell[7053027, 203328, 226, 3, 26, "Output",ExpressionUUID->"0e957366-467e-475f-a9dc-fcdda8548d12",
 CellID->975781576]
}, Open  ]],
Cell[CellGroupData[{
Cell[7053290, 203336, 183, 4, 27, "Input",ExpressionUUID->"dec5f89f-2b59-49f1-a9ee-bd96d8e00a63"],
Cell[7053476, 203342, 230, 3, 26, "Output",ExpressionUUID->"5f9a3770-0f25-42b5-8f7a-40e80437800d",
 CellID->1775819340]
}, Open  ]],
Cell[CellGroupData[{
Cell[7053743, 203350, 310, 10, 27, "Input",ExpressionUUID->"0801157d-4b09-4d11-a643-47b08cdad97b"],
Cell[7054056, 203362, 229, 3, 26, "Output",ExpressionUUID->"eb47b289-562a-4c4e-8def-7e3461fddb49",
 CellID->467921374]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7054334, 203371, 266, 4, 56, "MathCaption",ExpressionUUID->"c71baacf-2edb-4e19-944d-8b633139f9a8"],
Cell[CellGroupData[{
Cell[7054625, 203379, 155, 3, 27, "Input",ExpressionUUID->"0d35e924-327e-4796-9a6e-8f987cd0a6bc"],
Cell[7054783, 203384, 571, 12, 22, "Print",ExpressionUUID->"26e8e7fe-d6e6-4988-b7c6-c067df673f04",
 CellID->269417504]
}, Open  ]],
Cell[CellGroupData[{
Cell[7055391, 203401, 155, 3, 27, "Input",ExpressionUUID->"035b9896-a7a9-4761-90e6-8aa4568c4b5f"],
Cell[7055549, 203406, 572, 12, 22, "Print",ExpressionUUID->"19679f16-102d-4247-bcfc-3ce506894d58",
 CellID->1204461852]
}, Open  ]],
Cell[CellGroupData[{
Cell[7056158, 203423, 155, 3, 27, "Input",ExpressionUUID->"0b974f2c-c0aa-4392-993d-12c29641e2c0"],
Cell[7056316, 203428, 572, 12, 22, "Print",ExpressionUUID->"1f9a0f91-7e95-4756-b51b-1772a43adfd8",
 CellID->1209489942]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7056949, 203447, 92, 0, 39, "Subsection",ExpressionUUID->"0e34af1e-3617-4ec4-b436-adf5525614c9"],
Cell[7057044, 203449, 406, 9, 40, "Text",ExpressionUUID->"402e611c-deb3-472c-be07-5789b9c53c4d"],
Cell[CellGroupData[{
Cell[7057475, 203462, 161, 2, 40, "MathCaption",ExpressionUUID->"09956d3d-9850-4f27-aa8b-7ae3db749c6f"],
Cell[CellGroupData[{
Cell[7057661, 203468, 573, 19, 27, "Input",ExpressionUUID->"7c7c9070-81c0-4189-9870-4d83f7f5365f"],
Cell[7058237, 203489, 6070, 179, 32, "Output",ExpressionUUID->"8299ace9-4766-42e4-97f6-8c209bd7fae8",
 CellID->165667155]
}, Open  ]],
Cell[CellGroupData[{
Cell[7064344, 203673, 153, 3, 27, "Input",ExpressionUUID->"d3a7ef85-a5c2-4362-9b4a-4f2c9f3da99c"],
Cell[7064500, 203678, 7413, 199, 32, "Output",ExpressionUUID->"28717485-f25e-496d-bde3-2a9c5e7369b4",
 CellID->1229067573]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7071962, 203883, 153, 2, 40, "MathCaption",ExpressionUUID->"bbc36c11-f5b2-4dd2-ba3d-e4582c404f3e"],
Cell[CellGroupData[{
Cell[7072140, 203889, 224, 5, 27, "Input",ExpressionUUID->"527aeb2b-6fdd-4aee-b18d-2c4a04a06779"],
Cell[7072367, 203896, 7415, 199, 32, "Output",ExpressionUUID->"b6ed1c5f-f169-46d6-9739-85ac80ec2fef",
 CellID->293498918]
}, Open  ]]
}, Open  ]],
Cell[7079809, 204099, 729, 20, 58, "DefinitionBox",ExpressionUUID->"d5b56b42-ce19-4ca0-985a-bebb332d43f7"],
Cell[7080541, 204121, 81, 0, 23, "Text",ExpressionUUID->"e9611c4b-44b8-43e7-8bb6-15f12df0cd6c"]
}, Closed]],
Cell[CellGroupData[{
Cell[7080659, 204126, 191, 4, 39, "Subsection",ExpressionUUID->"b6edd169-b03e-4e7a-b1b9-5fc72a028f71"],
Cell[7080853, 204132, 1691, 45, 109, "DefinitionBox",ExpressionUUID->"73e3778f-e53c-431f-8c3c-8d8bc89db3b9"],
Cell[7082547, 204179, 133, 3, 23, "Text",ExpressionUUID->"85735a22-355f-40e2-8b2a-46343efddcdd"],
Cell[CellGroupData[{
Cell[7082705, 204186, 716, 17, 104, "MathCaption",ExpressionUUID->"45881c2e-a25a-4c3e-adec-dab3b227f836"],
Cell[CellGroupData[{
Cell[7083446, 204207, 540, 18, 27, "Input",ExpressionUUID->"6db88852-8d91-4836-8e13-0c99f0492e09"],
Cell[7083989, 204227, 172, 3, 26, "Output",ExpressionUUID->"d396edad-297d-418b-a164-a99b51a24390",
 CellID->2054448071]
}, Open  ]],
Cell[CellGroupData[{
Cell[7084198, 204235, 544, 18, 27, "Input",ExpressionUUID->"0ea0dd93-d9f4-4954-b0f8-ebd2fc0f4def"],
Cell[7084745, 204255, 1113, 31, 32, "Output",ExpressionUUID->"1d66ce3e-88c0-4c36-8404-fe8cc604c413",
 CellID->1866652831]
}, Open  ]],
Cell[CellGroupData[{
Cell[7085895, 204291, 706, 24, 27, "Input",ExpressionUUID->"bef4318e-b1bf-4f76-8142-29874ecfdf54"],
Cell[7086604, 204317, 4319, 121, 40, "Output",ExpressionUUID->"9a00163f-9ce7-4453-84ef-852fe2a9d06d",
 CellID->245531594]
}, Open  ]],
Cell[CellGroupData[{
Cell[7090960, 204443, 607, 20, 27, "Input",ExpressionUUID->"faa75994-6051-451d-ac35-84b6a83f3f9f"],
Cell[7091570, 204465, 4300, 120, 40, "Output",ExpressionUUID->"4c6aaaad-b777-4d4d-b638-3702d1162a9e",
 CellID->1333414471]
}, Open  ]],
Cell[CellGroupData[{
Cell[7095907, 204590, 607, 20, 27, "Input",ExpressionUUID->"900abf6c-ba1a-4dd7-8be9-77842b2c3b31"],
Cell[7096517, 204612, 171, 3, 26, "Output",ExpressionUUID->"0db8d8ea-03b3-40fc-861e-d83d0eac35a9",
 CellID->1705157652]
}, Open  ]],
Cell[CellGroupData[{
Cell[7096725, 204620, 598, 20, 27, "Input",ExpressionUUID->"8757895e-9503-4769-9c8d-6e4204d8916e"],
Cell[7097326, 204642, 4321, 121, 40, "Output",ExpressionUUID->"724ee41d-e8e6-4cf7-8120-8286680d2466",
 CellID->415387949]
}, Open  ]],
Cell[CellGroupData[{
Cell[7101684, 204768, 598, 20, 27, "Input",ExpressionUUID->"cdad5b95-38bb-4284-859a-2d882a7e188c"],
Cell[7102285, 204790, 2136, 60, 32, "Output",ExpressionUUID->"11875416-cceb-4de8-8aed-8348f93eefd8",
 CellID->2103406093]
}, Open  ]],
Cell[CellGroupData[{
Cell[7104458, 204855, 928, 32, 27, "Input",ExpressionUUID->"0caf2cec-03b9-41c6-b0b7-44ca8797b0e9"],
Cell[7105389, 204889, 12550, 352, 40, "Output",ExpressionUUID->"90ab9d32-887c-4491-8319-88251615f0c6",
 CellID->70408137]
}, Open  ]],
Cell[CellGroupData[{
Cell[7117976, 205246, 932, 32, 27, "Input",ExpressionUUID->"9b23fd11-fbd7-44af-850e-1f3bb0c1df28"],
Cell[7118911, 205280, 12550, 352, 40, "Output",ExpressionUUID->"925adde0-ab79-4973-9571-bad63270a556",
 CellID->86069062]
}, Open  ]],
Cell[CellGroupData[{
Cell[7131498, 205637, 679, 22, 27, "Input",ExpressionUUID->"b2bb5f3b-01ee-4704-823a-3e64c9f85999"],
Cell[7132180, 205661, 172, 3, 26, "Output",ExpressionUUID->"9aa14bf2-d9bb-464d-8b30-3b7d7a413df8",
 CellID->1306230603]
}, Open  ]],
Cell[CellGroupData[{
Cell[7132389, 205669, 679, 22, 27, "Input",ExpressionUUID->"bc453b2e-865f-4a35-8d70-40c82ff3a3dc"],
Cell[7133071, 205693, 172, 3, 26, "Output",ExpressionUUID->"9eedb2cd-8357-4c82-a7d3-36c54517c407",
 CellID->92421340]
}, Open  ]]
}, Open  ]],
Cell[7133270, 205700, 446, 6, 57, "Text",ExpressionUUID->"a73525ef-2616-4b35-89ec-1c3b9a0fecbb"],
Cell[CellGroupData[{
Cell[7133741, 205710, 131, 0, 40, "MathCaption",ExpressionUUID->"950e0afd-9b17-420a-b2d1-5e545b7d6e77"],
Cell[CellGroupData[{
Cell[7133897, 205714, 724, 18, 94, "Input",ExpressionUUID->"7b1d9279-3b3d-4e63-8459-b858cbbbe352"],
Cell[7134624, 205734, 462, 9, 22, "Message",ExpressionUUID->"1e5ba656-853d-4cb1-9859-7451c0619516",
 CellID->694784599],
Cell[7135089, 205745, 463, 9, 22, "Message",ExpressionUUID->"27f5f983-3518-454b-9de1-eaebcbfdee37",
 CellID->1376166654],
Cell[7135555, 205756, 462, 9, 22, "Message",ExpressionUUID->"7af15651-3b69-4fc1-92e8-26e481dae68b",
 CellID->834075588],
Cell[7136020, 205767, 483, 9, 22, "Message",ExpressionUUID->"419066dd-b051-44e9-963e-afb76873c5a5",
 CellID->1682643547]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7136552, 205782, 132, 0, 40, "MathCaption",ExpressionUUID->"b1e84f57-0c66-4a49-9b76-5f81a479a4e2"],
Cell[CellGroupData[{
Cell[7136709, 205786, 321, 8, 35, "Input",ExpressionUUID->"1dbc5b7f-301c-4085-a160-bb5805a507dc"],
Cell[7137033, 205796, 1237, 32, 40, "Output",ExpressionUUID->"14153213-4d30-4321-9e80-97385e5d71b7",
 CellID->2125430733]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7138319, 205834, 128, 0, 40, "MathCaption",ExpressionUUID->"7f4e43bb-ff53-4605-afd6-c02bb4b344d6"],
Cell[CellGroupData[{
Cell[7138472, 205838, 326, 9, 35, "Input",ExpressionUUID->"3ade80a8-faf8-43f3-ab48-f545d73c14fb"],
Cell[7138801, 205849, 28058, 703, 75, "Output",ExpressionUUID->"e35a0581-084e-4e22-8ad8-894c60e786e4",
 CellID->871345333]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7166908, 206558, 119, 0, 40, "MathCaption",ExpressionUUID->"daedf9d1-108b-4f60-ab20-e7951fb9d818"],
Cell[CellGroupData[{
Cell[7167052, 206562, 144, 3, 27, "Input",ExpressionUUID->"07a7b17d-7701-4eb7-a8c2-59d87678b067"],
Cell[7167199, 206567, 7232, 190, 41, "Output",ExpressionUUID->"50f581d7-41e6-472d-acfe-4ab3872acc9d",
 CellID->1010022838]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7174480, 206763, 126, 0, 40, "MathCaption",ExpressionUUID->"d59c67b9-9b6c-4843-90d8-966964c43c89"],
Cell[CellGroupData[{
Cell[7174631, 206767, 145, 3, 27, "Input",ExpressionUUID->"5697f1a1-b6d2-4bb4-ae97-1a3bd8ab73e7"],
Cell[7174779, 206772, 118546, 3096, 220, "Output",ExpressionUUID->"d26a8336-b30f-445e-be20-86f8943bd291",
 CellID->1743092378]
}, Open  ]],
Cell[CellGroupData[{
Cell[7293362, 209873, 144, 3, 27, "Input",ExpressionUUID->"61021ebc-174e-4242-bd0f-b3d4e56e66eb"],
Cell[7293509, 209878, 171, 3, 26, "Output",ExpressionUUID->"f9c37a47-ec32-43ac-97e9-c617423e027c",
 CellID->181684672]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7293729, 209887, 245, 4, 56, "MathCaption",ExpressionUUID->"aa9fbe97-08f1-496f-9121-6456f887623e"],
Cell[CellGroupData[{
Cell[7293999, 209895, 578, 14, 35, "Input",ExpressionUUID->"999bf8dd-3cef-4733-a1eb-b43f8945b7bf"],
Cell[7294580, 209911, 172, 3, 26, "Output",ExpressionUUID->"c5365fcb-4391-43b2-94c8-514554b933a8",
 CellID->2000019917]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7294801, 209920, 87, 0, 40, "MathCaption",ExpressionUUID->"72b2dc43-e983-4ab5-b3f3-b3d31190637b"],
Cell[7294891, 209922, 413, 8, 85, "Input",ExpressionUUID->"760165dd-bf45-4578-8753-12b2d0bfe204"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7295353, 209936, 108, 0, 39, "Subsection",ExpressionUUID->"cc00b7aa-4354-4ad2-8a11-e97cec62027d"],
Cell[7295464, 209938, 255, 4, 40, "Text",ExpressionUUID->"2d326e97-e85d-40bb-8696-71a5138dfb8a"],
Cell[CellGroupData[{
Cell[7295744, 209946, 179, 3, 40, "MathCaption",ExpressionUUID->"d11147e3-40ad-4e33-8b3c-a5c906f8610b"],
Cell[CellGroupData[{
Cell[7295948, 209953, 240, 7, 27, "Input",ExpressionUUID->"02cad8a7-b143-4bf4-83a3-26b7b8038509"],
Cell[7296191, 209962, 61922, 1592, 145, "Output",ExpressionUUID->"8b00d4f3-5ae0-4a7f-b0bf-94d1fa31ec42",
 CellID->1677761570]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7358162, 211560, 188, 3, 40, "MathCaption",ExpressionUUID->"39ba8d23-c3bf-4eb8-8f91-785d6ed73679"],
Cell[CellGroupData[{
Cell[7358375, 211567, 184, 4, 27, "Input",ExpressionUUID->"0c130b34-b821-4376-9ea4-1dde287ecfd6"],
Cell[7358562, 211573, 26199, 718, 66, "Output",ExpressionUUID->"c43db077-e4d5-4728-ae82-565ee7f067f8",
 CellID->1678421880]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7384822, 212298, 113, 0, 39, "Subsection",ExpressionUUID->"657a9bdf-72f6-42dd-9fa5-4695039b9fb2"],
Cell[7384938, 212300, 266, 6, 40, "Text",ExpressionUUID->"13668bed-de68-4c30-95a4-78d142b13253"],
Cell[CellGroupData[{
Cell[7385229, 212310, 290, 4, 56, "MathCaption",ExpressionUUID->"c57b1329-32db-492d-abd7-83f09d5a2635"],
Cell[7385522, 212316, 156, 3, 27, "Input",ExpressionUUID->"2275d90e-a848-42eb-a85f-4ce761b931e3"]
}, Open  ]],
Cell[CellGroupData[{
Cell[7385715, 212324, 334, 11, 40, "MathCaption",ExpressionUUID->"ab4332c5-bc82-4ab1-965e-10cd3c9d8a86"],
Cell[CellGroupData[{
Cell[7386074, 212339, 322, 9, 27, "Input",ExpressionUUID->"da7839a9-4637-4a64-8db5-5222795b5809"],
Cell[7386399, 212350, 667, 13, 22, "Print",ExpressionUUID->"b332f045-e5fd-450d-887a-29cac9615bec",
 CellID->894627505]
}, Open  ]],
Cell[CellGroupData[{
Cell[7387103, 212368, 1036, 36, 27, "Input",ExpressionUUID->"5c9b5b2c-97fb-444c-896e-308e24313366"],
Cell[7388142, 212406, 12820, 372, 32, "Output",ExpressionUUID->"b5d4af3e-e969-41ed-82fb-6edcbd0abf9e",
 CellID->1471463888]
}, Open  ]],
Cell[CellGroupData[{
Cell[7400999, 212783, 154, 3, 27, "Input",ExpressionUUID->"4a8b139b-ad55-4da0-b7bb-7305fc5eadc0"],
Cell[7401156, 212788, 173, 3, 26, "Output",ExpressionUUID->"a75e5628-3f78-4a38-912b-ff381aa9838b",
 CellID->471886764]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7401378, 212797, 222, 3, 56, "MathCaption",ExpressionUUID->"7aff734c-a431-4430-8bb4-7186a61e1766"],
Cell[CellGroupData[{
Cell[7401625, 212804, 146, 3, 27, "Input",ExpressionUUID->"07bb6e50-cd48-4d31-bc62-acb1bd4ef9e6"],
Cell[7401774, 212809, 18293, 483, 35, "Output",ExpressionUUID->"856ce393-70ee-4cd6-b252-f6c2ee55a21f",
 CellID->343734177]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7420116, 213298, 147, 2, 40, "MathCaption",ExpressionUUID->"ae2a9a69-446c-46f3-83b5-f671de712962"],
Cell[CellGroupData[{
Cell[7420288, 213304, 215, 5, 27, "Input",ExpressionUUID->"2d1f479c-dcf1-496b-8042-93e5e4fc3a64"],
Cell[7420506, 213311, 8935, 259, 32, "Output",ExpressionUUID->"4288c058-3668-45ea-be14-10a36270b6ab",
 CellID->409302409]
}, Open  ]],
Cell[CellGroupData[{
Cell[7429478, 213575, 292, 8, 27, "Input",ExpressionUUID->"90d8e5f1-2661-4528-9450-b8b4ab5302db"],
Cell[7429773, 213585, 5979, 174, 32, "Output",ExpressionUUID->"a255b957-99e7-4111-b91a-1e8f26f13b89",
 CellID->755466722]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7435801, 213765, 553, 18, 72, "MathCaption",ExpressionUUID->"8b05dfb5-3cba-432e-bbae-830d55adc084"],
Cell[CellGroupData[{
Cell[7436379, 213787, 279, 8, 27, "Input",ExpressionUUID->"ecba2eab-fcbc-49f0-b215-142297c26976"],
Cell[7436661, 213797, 13105, 376, 32, "Output",ExpressionUUID->"c2dea06c-478c-4d3a-9b17-71669303306c",
 CellID->1205875590]
}, Open  ]],
Cell[CellGroupData[{
Cell[7449803, 214178, 310, 9, 27, "Input",ExpressionUUID->"7d1015f5-9712-4d4c-9ec2-2cb1d6353bb1"],
Cell[7450116, 214189, 13270, 380, 32, "Output",ExpressionUUID->"b667fd05-76ba-49a5-89a0-b10ebb5b3005",
 CellID->1347864391]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7463435, 214575, 115, 0, 40, "MathCaption",ExpressionUUID->"633e312a-6217-4d3f-845d-8076502815d1"],
Cell[CellGroupData[{
Cell[7463575, 214579, 217, 5, 27, "Input",ExpressionUUID->"705d82ff-4f88-45fd-8f3b-174a519371dd"],
Cell[7463795, 214586, 15157, 428, 35, "Output",ExpressionUUID->"dd10be8c-c452-4b1e-bca7-3d2ef812fbe7",
 CellID->1899310981]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7479001, 215020, 133, 0, 40, "MathCaption",ExpressionUUID->"8a8cdb46-20f2-417e-9fc0-d3bf3c4f644c"],
Cell[CellGroupData[{
Cell[7479159, 215024, 201, 5, 27, "Input",ExpressionUUID->"99ee2185-faae-463a-ade9-54935d2cdd84"],
Cell[7479363, 215031, 317, 9, 26, "Output",ExpressionUUID->"6765649b-d86a-41d9-a9b2-51fdebf0a6d7",
 CellID->1522728659]
}, Open  ]],
Cell[CellGroupData[{
Cell[7479717, 215045, 178, 4, 27, "Input",ExpressionUUID->"d2ee5476-8d66-4d16-b15e-b4c1f309fe0f"],
Cell[7479898, 215051, 12933, 374, 32, "Output",ExpressionUUID->"48f0e4c6-27bb-4c8d-8cbf-c3ee024c28ec",
 CellID->937950514]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7492880, 215431, 347, 11, 56, "MathCaption",ExpressionUUID->"2c9ca0ad-cb53-4e94-8f5d-b97cfb91f380"],
Cell[CellGroupData[{
Cell[7493252, 215446, 1061, 36, 47, "Input",ExpressionUUID->"b61110fc-4edb-482d-8122-d331a3ebffc4"],
Cell[7494316, 215484, 13241, 373, 32, "Output",ExpressionUUID->"547903c1-45cd-4d4c-ace5-5c4ef471331f",
 CellID->1604319818]
}, Open  ]],
Cell[CellGroupData[{
Cell[7507594, 215862, 237, 6, 27, "Input",ExpressionUUID->"b93f51d5-b0e1-427b-8ef0-0dc2f7eb81b6"],
Cell[7507834, 215870, 13440, 378, 32, "Output",ExpressionUUID->"ffbd0585-c4da-4d75-b5a9-b1e67db229be",
 CellID->2145993041]
}, Open  ]],
Cell[CellGroupData[{
Cell[7521311, 216253, 267, 8, 27, "Input",ExpressionUUID->"be0d5668-72cd-4b74-a936-efee4d51c7be"],
Cell[7521581, 216263, 9251, 260, 32, "Output",ExpressionUUID->"cdbb3b69-d6bd-4131-9ed3-b0cb72d1a025",
 CellID->717188722]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7530893, 216530, 95, 0, 39, "Subsection",ExpressionUUID->"90f7019f-3ef3-4543-86f2-339abc68ed1e"],
Cell[7530991, 216532, 281, 4, 40, "Text",ExpressionUUID->"a5d086bc-55b6-422a-9038-07fe77be6716"],
Cell[7531275, 216538, 407, 9, 41, "DefinitionBox",ExpressionUUID->"a6df1543-8041-49ef-93ff-ada9bc9c212a"],
Cell[7531685, 216549, 107, 0, 23, "Text",ExpressionUUID->"107cf7da-cac8-41fe-ad13-773c153bb983"],
Cell[CellGroupData[{
Cell[7531817, 216553, 105, 0, 40, "MathCaption",ExpressionUUID->"5458d8e6-cb10-4ffd-b719-8673f931c9fe"],
Cell[CellGroupData[{
Cell[7531947, 216557, 284, 8, 27, "Input",ExpressionUUID->"32c52d95-07d7-4e3d-adb4-e3de49ed39c1"],
Cell[7532234, 216567, 2101, 60, 32, "Output",ExpressionUUID->"4b43a68b-3b23-44de-9123-aca2464bd0ad",
 CellID->776216801]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7534384, 216633, 128, 0, 40, "MathCaption",ExpressionUUID->"b8299032-23fc-4d23-b77e-f44a5bbd9487"],
Cell[CellGroupData[{
Cell[7534537, 216637, 326, 10, 27, "Input",ExpressionUUID->"3fff252c-b5c0-4dca-b9fa-64d384b159be"],
Cell[7534866, 216649, 2096, 60, 32, "Output",ExpressionUUID->"e7344982-2e18-486f-bfdb-b77343f82fb3",
 CellID->2060102529]
}, Open  ]],
Cell[CellGroupData[{
Cell[7536999, 216714, 380, 12, 27, "Input",ExpressionUUID->"82c746d3-5949-4d74-9ef7-089393e3a32b"],
Cell[7537382, 216728, 4506, 125, 40, "Output",ExpressionUUID->"a0c106dc-48db-4927-9a00-cb20e45695ee",
 CellID->264364322]
}, Open  ]],
Cell[CellGroupData[{
Cell[7541925, 216858, 454, 14, 27, "Input",ExpressionUUID->"90e3f535-a8af-45e5-aedc-2340ae2a20cd"],
Cell[7542382, 216874, 3183, 91, 40, "Output",ExpressionUUID->"07828efe-3a2a-4d00-8e88-a874a8b4c035",
 CellID->27494211]
}, Open  ]],
Cell[CellGroupData[{
Cell[7545602, 216970, 479, 16, 27, "Input",ExpressionUUID->"680b0274-1234-4d20-bcb9-0f437a64dcdf"],
Cell[7546084, 216988, 2175, 62, 49, "Output",ExpressionUUID->"898c8772-f955-48de-a0d4-742667a36504",
 CellID->386220172]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7548308, 217056, 191, 5, 40, "MathCaption",ExpressionUUID->"8c4b7018-bdc8-4aab-b3b6-f405579c2b9d"],
Cell[CellGroupData[{
Cell[7548524, 217065, 290, 8, 27, "Input",ExpressionUUID->"339b4920-2595-46c1-990f-60c624e47847"],
Cell[7548817, 217075, 172, 3, 26, "Output",ExpressionUUID->"57623251-4d6a-4dd9-876a-54450a855ab5",
 CellID->1506849570]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7549038, 217084, 132, 0, 40, "MathCaption",ExpressionUUID->"d0f2179a-1f93-417d-8e72-5aa6a5c468d9"],
Cell[CellGroupData[{
Cell[7549195, 217088, 564, 18, 27, "Input",ExpressionUUID->"0db62fc3-bc4b-44b2-b50b-7c49787640c5"],
Cell[7549762, 217108, 13116, 352, 40, "Output",ExpressionUUID->"cdfa049a-b657-4df4-9ae0-b11f7835779e",
 CellID->346073919]
}, Open  ]],
Cell[CellGroupData[{
Cell[7562915, 217465, 184, 4, 27, "Input",ExpressionUUID->"9cded25c-f506-4d0a-8959-32ee2e548b55"],
Cell[7563102, 217471, 10349, 277, 40, "Output",ExpressionUUID->"4ce7d19b-c48d-4df1-baf9-4ab4802a8af8",
 CellID->1078640557]
}, Open  ]]
}, Open  ]],
Cell[7573478, 217752, 459, 14, 40, "Text",ExpressionUUID->"fb35b0f7-3f30-490e-9e00-17dde2728888"],
Cell[7573940, 217768, 197, 3, 41, "DefinitionBox",ExpressionUUID->"40ef30fe-74eb-449c-9eee-c00c4a08041f"],
Cell[CellGroupData[{
Cell[7574162, 217775, 110, 0, 40, "MathCaption",ExpressionUUID->"53e839fa-c408-421f-9baf-9e7df6bc6263"],
Cell[CellGroupData[{
Cell[7574297, 217779, 568, 18, 27, "Input",ExpressionUUID->"6cd77f16-eb35-4a2a-99a0-46c03a51b92a"],
Cell[7574868, 217799, 6062, 175, 32, "Output",ExpressionUUID->"8cc53ff2-11d3-4243-a243-f878fd819c42",
 CellID->1802377172]
}, Open  ]],
Cell[CellGroupData[{
Cell[7580967, 217979, 312, 9, 27, "Input",ExpressionUUID->"33287f06-68bf-4008-bdd9-9ee1d083155c"],
Cell[7581282, 217990, 6721, 184, 40, "Output",ExpressionUUID->"24aece55-80c7-4ad2-af0b-a1c556975660",
 CellID->148805222]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7588064, 218181, 100, 0, 39, "Subsection",ExpressionUUID->"bc25c3e2-b171-420e-abc8-60aea9c198d9"],
Cell[7588167, 218183, 412, 7, 57, "Text",ExpressionUUID->"4f9c13b6-2fbe-45fd-b59f-e05b80cc794d"],
Cell[CellGroupData[{
Cell[7588604, 218194, 212, 5, 40, "MathCaption",ExpressionUUID->"6de3f851-3bef-4bb5-8b4e-27ba06c24f5f"],
Cell[CellGroupData[{
Cell[7588841, 218203, 382, 13, 27, "Input",ExpressionUUID->"65e7af33-4841-4448-aef8-4368eb920f22"],
Cell[7589226, 218218, 4077, 118, 32, "Output",ExpressionUUID->"55468b7c-f04a-4b1c-bb57-09d8b71051f0",
 CellID->1532473207]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7593352, 218342, 170, 4, 40, "MathCaption",ExpressionUUID->"61cf9248-7fb6-4821-bf60-390e7a7d6a27"],
Cell[CellGroupData[{
Cell[7593547, 218350, 151, 3, 27, "Input",ExpressionUUID->"51369881-802e-4b76-84e1-518cb1996f5a"],
Cell[7593701, 218355, 175, 3, 26, "Output",ExpressionUUID->"efb27cbb-4278-4578-aecd-443c260e7ea8",
 CellID->1505811578]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7593925, 218364, 110, 0, 40, "MathCaption",ExpressionUUID->"ceea1a38-49ec-45fd-a9f0-9ddab3502197"],
Cell[CellGroupData[{
Cell[7594060, 218368, 455, 15, 27, "Input",ExpressionUUID->"b96d886a-552f-4290-9825-7d6362347414"],
Cell[7594518, 218385, 174, 3, 26, "Output",ExpressionUUID->"448c2c8b-01ae-4225-ae9b-911f6fd31bf4",
 CellID->80411551]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7594741, 218394, 286, 8, 40, "MathCaption",ExpressionUUID->"3f13f80b-1730-4627-9e9e-07d12d56f1ce"],
Cell[CellGroupData[{
Cell[7595052, 218406, 826, 24, 47, "Input",ExpressionUUID->"4b846980-45ef-487f-9af7-70e10a1faa88"],
Cell[7595881, 218432, 8903, 243, 58, "Output",ExpressionUUID->"bbf5732b-bbda-46f4-b8ef-f25e24a613b9",
 CellID->709723694]
}, Open  ]],
Cell[CellGroupData[{
Cell[7604821, 218680, 558, 18, 27, "Input",ExpressionUUID->"c941c480-0e60-482e-bee3-4e85c69dff1d"],
Cell[7605382, 218700, 15647, 439, 58, "Output",ExpressionUUID->"0806bfe2-b54c-470f-9c4e-0b67670d0f34",
 CellID->1201949620]
}, Open  ]]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[7621090, 219146, 92, 0, 39, "Subsection",ExpressionUUID->"9e102fec-dd1b-4b26-8b3d-135b944df6da"],
Cell[CellGroupData[{
Cell[7621207, 219150, 154, 3, 27, "Input",ExpressionUUID->"63987a5e-f110-4dd1-940e-140cf3b4baec"],
Cell[CellGroupData[{
Cell[7621386, 219157, 578, 12, 22, "Print",ExpressionUUID->"0d699409-e5c6-4ec2-adda-bc0401471fb9",
 CellID->1894719997],
Cell[7621967, 219171, 606, 12, 22, "Print",ExpressionUUID->"7d8df572-1d72-47bc-8d64-e834fb317518",
 CellID->1671471863],
Cell[7622576, 219185, 598, 12, 22, "Print",ExpressionUUID->"d1b39128-a058-437d-a7b0-cc30cae4bc63",
 CellID->1194015034]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7623223, 219203, 169, 4, 27, "Input",ExpressionUUID->"3188f92f-4028-4cb1-901b-213d0840f6ea"],
Cell[CellGroupData[{
Cell[7623417, 219211, 549, 12, 22, "Print",ExpressionUUID->"849a03c8-36f9-4c1c-9edc-f39881820b8a",
 CellID->1723478870],
Cell[7623969, 219225, 550, 12, 22, "Print",ExpressionUUID->"4ced1b35-08f4-4cc5-8c5a-c8406aedca9d",
 CellID->1249937354],
Cell[7624522, 219239, 552, 12, 22, "Print",ExpressionUUID->"85cf0a92-4fdf-4fd8-acf1-bc5118d8d30f",
 CellID->1012073724],
Cell[7625077, 219253, 550, 12, 22, "Print",ExpressionUUID->"51e08db8-5cd2-4794-bc7a-ac8f6d814243",
 CellID->1852595028],
Cell[7625630, 219267, 551, 12, 22, "Print",ExpressionUUID->"da4ea3f5-6e60-4e82-8e3b-19bb00f71fc5",
 CellID->199237743],
Cell[7626184, 219281, 549, 12, 22, "Print",ExpressionUUID->"e19e43bf-9b20-4d93-b245-6cd7b7a1c1ba",
 CellID->661740928],
Cell[7626736, 219295, 550, 12, 22, "Print",ExpressionUUID->"3a6d9e5c-3d8c-4526-8af6-a61b35226cf5",
 CellID->61969083]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7627335, 219313, 151, 3, 27, "Input",ExpressionUUID->"03bad5e8-bcb2-425a-bbba-f56daaa3271e"],
Cell[CellGroupData[{
Cell[7627511, 219320, 726, 18, 22, "Print",ExpressionUUID->"74f75195-ea60-443b-9c67-687fb4a21521",
 CellID->2132292833],
Cell[7628240, 219340, 561, 12, 22, "Print",ExpressionUUID->"fd3743e5-4bfb-49c8-84c9-a0c1b7c3987e",
 CellID->238543095]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7628850, 219358, 151, 3, 27, "Input",ExpressionUUID->"8635e9e6-ab5c-42be-a250-230a769d71bf"],
Cell[CellGroupData[{
Cell[7629026, 219365, 728, 18, 22, "Print",ExpressionUUID->"0569c43e-59aa-4eb0-b9e2-bff683d40a1b",
 CellID->1989312363],
Cell[7629757, 219385, 561, 12, 22, "Print",ExpressionUUID->"b1101d58-8cd8-43f1-b3f0-bf542c8286b1",
 CellID->199984798]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7630367, 219403, 157, 3, 27, "Input",ExpressionUUID->"2a2e5a4f-6c77-47c6-9ed1-560d61bac3c6"],
Cell[7630527, 219408, 718, 18, 22, "Print",ExpressionUUID->"f3f984b6-1520-4855-b8c6-8e6cce0a0a88",
 CellID->332592431]
}, Open  ]]
}, Closed]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7631306, 219433, 1016, 16, 54, "Section",ExpressionUUID->"60102f35-97c9-49fb-ac37-e2c30aabc8e1",
 CellTags->{"S5.18.1", "1.4"},
 CellID->292931816],
Cell[CellGroupData[{
Cell[7632347, 219453, 101, 0, 39, "Subsection",ExpressionUUID->"16b126b0-e45c-413a-b713-bd224f7c8574"],
Cell[7632451, 219455, 344, 9, 81, "Text",ExpressionUUID->"9bf0f838-d02c-4590-a0f9-a84476c860e5"],
Cell[7632798, 219466, 404, 9, 41, "Text",ExpressionUUID->"8130a58f-c2e0-469c-ae40-a12f297a8dd7"],
Cell[7633205, 219477, 325, 8, 23, "Text",ExpressionUUID->"8f4bcdcd-88dc-4765-a3c5-264f84a6227e",
 CellID->983232810]
}, Closed]],
Cell[CellGroupData[{
Cell[7633567, 219490, 109, 0, 39, "Subsection",ExpressionUUID->"445c4b78-68c5-40bf-921f-1b70fa4b00bf"],
Cell[7633679, 219492, 285, 5, 52, "Text",ExpressionUUID->"773cc8d8-26c1-48ea-8659-4aba568c0e15"],
Cell[7633967, 219499, 874, 13, 160, "MathCaption",ExpressionUUID->"7b2ce55e-9501-48aa-9698-8ac0f91131f2"],
Cell[7634844, 219514, 263, 5, 52, "Text",ExpressionUUID->"69fe4b03-c970-4d33-b472-c77368ccf6cb"],
Cell[7635110, 219521, 674, 17, 144, "MathCaption",ExpressionUUID->"6eb445ea-32ea-4bf4-a609-2180aba05b1a"],
Cell[7635787, 219540, 277, 6, 40, "Text",ExpressionUUID->"43a15891-8e32-4156-b3af-b77a0f4ac331"],
Cell[7636067, 219548, 400, 14, 56, "MathCaption",ExpressionUUID->"2848bfa1-d880-4e6e-bf62-a20ed756fbdf"],
Cell[7636470, 219564, 292, 9, 23, "Text",ExpressionUUID->"5b07a210-037f-4997-b0c9-7d388ebf5bb7"],
Cell[7636765, 219575, 413, 12, 56, "MathCaption",ExpressionUUID->"d5368906-9eaf-423e-b4b7-b3d834476af1"],
Cell[7637181, 219589, 299, 6, 40, "Text",ExpressionUUID->"aa2315df-ece4-4ee0-bf92-e93744aca77e"],
Cell[7637483, 219597, 422, 10, 72, "MathCaption",ExpressionUUID->"19095ea6-048c-40ff-ba3e-8a9caea50eea"]
}, Closed]],
Cell[CellGroupData[{
Cell[7637942, 219612, 93, 0, 39, "Subsection",ExpressionUUID->"4a0ec84b-57f6-45d5-ae86-2fd979f0579e"],
Cell[CellGroupData[{
Cell[7638060, 219616, 172, 5, 23, "Text",ExpressionUUID->"466fe0e3-1233-4ced-ae6f-3ebb5dd1d87a"],
Cell[CellGroupData[{
Cell[7638257, 219625, 163, 3, 27, "Input",ExpressionUUID->"14963688-618e-4087-91d9-575435758978"],
Cell[7638423, 219630, 14268, 222, 1319, "Output",ExpressionUUID->"aa20a420-50da-47de-9374-999d5404fb65",
 CellID->584626563]
}, Open  ]],
Cell[CellGroupData[{
Cell[7652728, 219857, 143, 3, 27, "Input",ExpressionUUID->"a7b1e781-126d-4adf-afea-fb8a5a94d7ca"],
Cell[7652874, 219862, 174, 3, 26, "Output",ExpressionUUID->"d3452b50-cf7a-473a-acf5-1cac357ae8c2",
 CellID->1080959316]
}, Open  ]],
Cell[CellGroupData[{
Cell[7653085, 219870, 140, 3, 27, "Input",ExpressionUUID->"8b648d0b-23b6-48a0-af25-eae60d794b98"],
Cell[7653228, 219875, 181, 3, 26, "Output",ExpressionUUID->"9f74160c-9d79-4eea-a136-1ec3326406ab",
 CellID->2119054039]
}, Open  ]],
Cell[CellGroupData[{
Cell[7653446, 219883, 225, 7, 27, "Input",ExpressionUUID->"5ad2e8f7-061b-4adc-915c-664072b9760b"],
Cell[7653674, 219892, 187, 3, 26, "Output",ExpressionUUID->"8f267b3e-f333-440c-b35a-b2bfa10b58e3",
 CellID->1072431625]
}, Open  ]],
Cell[CellGroupData[{
Cell[7653898, 219900, 227, 7, 27, "Input",ExpressionUUID->"02498cf6-b487-4e39-bd44-d17253c1e3f0"],
Cell[7654128, 219909, 185, 3, 26, "Output",ExpressionUUID->"1a62664f-4b21-41e8-a21d-400601625b3a",
 CellID->528072104]
}, Open  ]]
}, Open  ]]
}, Closed]]
}, Open  ]]
}
]
*)

